// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/* This file defines the module loader for the dart runtime.
*/

var dart_library =
  typeof module != "undefined" && module.exports || {};

(function (dart_library) {
  'use strict';

  /** Note that we cannot use dart_utils.throwInternalError from here. */
  function throwLibraryError(message) {
    throw Error(message);
  }

  // Module support.  This is a simplified module system for Dart.
  // Longer term, we can easily migrate to an existing JS module system:
  // ES6, AMD, RequireJS, ....

  class LibraryLoader {
    constructor(name, defaultValue, imports, lazyImports, loader) {
      this._name = name;
      this._library = defaultValue ? defaultValue : {};
      this._imports = imports;
      this._lazyImports = lazyImports;
      this._loader = loader;

      // Cyclic import detection
      this._state = LibraryLoader.NOT_LOADED;
    }

    loadImports(pendingSet) {
      return this.handleImports(this._imports, (lib) => lib.load(pendingSet));
    }

    deferLazyImports(pendingSet) {
      return this.handleImports(this._lazyImports,
        (lib) => {
          pendingSet.add(lib._name);
          return lib.stub();
      });
    }

    loadLazyImports(pendingSet) {
      return this.handleImports(pendingSet, (lib) => lib.load());
    }

    handleImports(list, handler) {
      let results = [];
      for (let name of list) {
        let lib = libraries[name];
        if (!lib) {
          throwLibraryError('Library not available: ' + name);
        }
        results.push(handler(lib));
      }
      return results;
    }

    load(inheritedPendingSet) {
      // Check for cycles
      if (this._state == LibraryLoader.LOADING) {
        throwLibraryError('Circular dependence on library: '
                              + this._name);
      } else if (this._state >= LibraryLoader.LOADED) {
        return this._library;
      }
      this._state = LibraryLoader.LOADING;

      // Handle imports and record lazy imports
      let pendingSet = inheritedPendingSet ? inheritedPendingSet : new Set();
      let args = this.loadImports(pendingSet);
      args = args.concat(this.deferLazyImports(pendingSet));

      // Load the library
      args.unshift(this._library);
      this._loader.apply(null, args);
      this._state = LibraryLoader.LOADED;

      // Handle lazy imports
      if (inheritedPendingSet === void 0) {
        // Drain the queue
        this.loadLazyImports(pendingSet);
      }
      this._state = LibraryLoader.READY;
      return this._library;
    }

    stub() {
      return this._library;
    }
  }
  LibraryLoader.NOT_LOADED = 0;
  LibraryLoader.LOADING = 1;
  LibraryLoader.LOADED = 2;
  LibraryLoader.READY = 3;

  // Map from name to LibraryLoader
  let libraries = new Map();

  function library(name, defaultValue, imports, lazyImports, loader) {
    return libraries[name] =
      new LibraryLoader(name, defaultValue, imports, lazyImports, loader);
  }
  dart_library.library = library;

  function import_(libraryName) {
    bootstrap();
    let loader = libraries[libraryName];
    // TODO(vsm): A user might call this directly from JS (as we do in tests).
    // We may want a different error type.
    if (!loader) throwLibraryError('Library not found: ' + libraryName);
    return loader.load();
  }
  dart_library.import = import_;

  function start(libraryName) {
    let library = import_(libraryName);
    let _isolate_helper = import_('dart/_isolate_helper');
    _isolate_helper.startRootIsolate(library.main, []);
  }
  dart_library.start = start;

  let _bootstrapped = false;
  function bootstrap() {
    if (_bootstrapped) return;
    _bootstrapped = true;

    // Force import of core.
    var core = import_('dart/core');
    core.Object.toString = function() {
      // Interface types are represented by the corresponding constructor
      // function.  This ensures that Dart interface types print properly.
      return this.name;
    }

    // TODO(vsm): DOM facades?
    // See: https://github.com/dart-lang/dev_compiler/issues/173
    if (typeof NodeList !== "undefined") {
      NodeList.prototype.get = function(i) { return this[i]; };
      NamedNodeMap.prototype.get = function(i) { return this[i]; };
      DOMTokenList.prototype.get = function(i) { return this[i]; };
      HTMLCollection.prototype.get = function(i) { return this[i]; };
    }
  }

})(dart_library);
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/* This library defines the operations that define and manipulate Dart
 * classes.  Included in this are:
 *   - Generics
 *   - Class metadata
 *   - Extension methods
 */

// TODO(leafp): Consider splitting some of this out.
dart_library.library('dart/_classes', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_utils',
  'dart/core',
  'dart/_interceptors',
  'dart/_types',
  'dart/_rtti',
], function(exports, dart_utils, core, _interceptors, types, rtti) {
  'use strict';

  const assert = dart_utils.assert_;
  const copyProperties = dart_utils.copyProperties;
  const copyTheseProperties = dart_utils.copyTheseProperties;
  const defineMemoizedGetter = dart_utils.defineMemoizedGetter;
  const safeGetOwnProperty = dart_utils.safeGetOwnProperty;
  const throwInternalError = dart_utils.throwInternalError;

  const defineProperty = Object.defineProperty;
  const getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor;
  const getOwnPropertySymbols = Object.getOwnPropertySymbols;

  /** The Symbol for storing type arguments on a specialized generic type. */
  const _mixins = Symbol('mixins');
  const _implements = Symbol('implements');
  exports.implements = _implements;
  const _metadata = Symbol('metadata');
  exports.metadata = _metadata;

  /**
   * Returns a new type that mixes members from base and all mixins.
   *
   * Each mixin applies in sequence, with further to the right ones overriding
   * previous entries.
   *
   * For each mixin, we only take its own properties, not anything from its
   * superclass (prototype).
   */
  function mixin(base, ...mixins) {
    // Create an initializer for the mixin, so when derived constructor calls
    // super, we can correctly initialize base and mixins.

    // Create a class that will hold all of the mixin methods.
    class Mixin extends base {
      // Initializer method: run mixin initializers, then the base.
      [base.name](...args) {
        // Run mixin initializers. They cannot have arguments.
        // Run them backwards so most-derived mixin is initialized first.
        for (let i = mixins.length - 1; i >= 0; i--) {
          let mixin = mixins[i];
          let init = mixin.prototype[mixin.name];
          if (init) init.call(this);
        }
        // Run base initializer.
        let init = base.prototype[base.name];
        if (init) init.apply(this, args);
      }
    }
    // Copy each mixin's methods, with later ones overwriting earlier entries.
    for (let m of mixins) {
      copyProperties(Mixin.prototype, m.prototype);
    }

    // Set the signature of the Mixin class to be the composition
    // of the signatures of the mixins.
    setSignature(Mixin, {
      methods: () => {
        let s = {};
        for (let m of mixins) {
          copyProperties(s, m[_methodSig]);
        }
        return s;
      }
    });

    // Save mixins for reflection
    Mixin[_mixins] = mixins;
    return Mixin;
  }
  exports.mixin = mixin;

  function getMixins (clazz) {
    return clazz[_mixins];
  }
  exports.getMixins = getMixins;

  function getImplements (clazz) {
    return clazz[_implements];
  }
  exports.getImplements = getImplements;

  /** The Symbol for storing type arguments on a specialized generic type. */
  let _typeArguments = Symbol('typeArguments');
  let _originalDeclaration = Symbol('originalDeclaration');

  /** Memoize a generic type constructor function. */
  function generic(typeConstructor) {
    let length = typeConstructor.length;
    if (length < 1) {
      throwInternalError('must have at least one generic type argument');
    }
    let resultMap = new Map();
    function makeGenericType(...args) {
      if (args.length != length && args.length != 0) {
        throwInternalError('requires ' + length + ' or 0 type arguments');
      }
      while (args.length < length) args.push(types.dynamic);

      let value = resultMap;
      for (let i = 0; i < length; i++) {
        let arg = args[i];
        if (arg == null) {
          throwInternalError('type arguments should not be null: '
                            + typeConstructor);
        }
        let map = value;
        value = map.get(arg);
        if (value === void 0) {
          if (i + 1 == length) {
            value = typeConstructor.apply(null, args);
            // Save the type constructor and arguments for reflection.
            if (value) {
              value[_typeArguments] = args;
              value[_originalDeclaration] = makeGenericType;
            }
          } else {
            value = new Map();
          }
          map.set(arg, value);
        }
      }
      return value;
    }
    return makeGenericType;
  }
  exports.generic = generic;

  function getGenericClass(type) {
    return safeGetOwnProperty(type, _originalDeclaration);
  };
  exports.getGenericClass = getGenericClass;

  function getGenericArgs(type) {
    return safeGetOwnProperty(type, _typeArguments);
  };
  exports.getGenericArgs = getGenericArgs;

  let _constructorSig = Symbol('sigCtor');
  let _methodSig = Symbol("sig");
  let _staticSig = Symbol("sigStatic");

  /// Get the type of a method using the stored signature
  function _getMethodType(obj, name) {
    if (obj === void 0) return void 0;
    if (obj == null) return void 0;
    let sigObj = obj.__proto__.constructor[_methodSig];
    if (sigObj === void 0) return void 0;
    let parts = sigObj[name];
    if (parts === void 0) return void 0;
    return types.definiteFunctionType.apply(null, parts);
  }

  /// Get the type of a constructor from a class using the stored signature
  /// If name is undefined, returns the type of the default constructor
  /// Returns undefined if the constructor is not found.
  function _getConstructorType(cls, name) {
    if(!name) name = cls.name;
    if (cls === void 0) return void 0;
    if (cls == null) return void 0;
    let sigCtor = cls[_constructorSig];
    if (sigCtor === void 0) return void 0;
    let parts = sigCtor[name];
    if (parts === void 0) return void 0;
    return types.definiteFunctionType.apply(null, parts);
  }
  exports.classGetConstructorType = _getConstructorType;

  /// Given an object and a method name, tear off the method.
  /// Sets the runtime type of the torn off method appropriately,
  /// and also binds the object.
  ///
  /// If the optional `f` argument is passed in, it will be used as the method.
  /// This supports cases like `super.foo` where we need to tear off the method
  /// from the superclass, not from the `obj` directly.
  /// TODO(leafp): Consider caching the tearoff on the object?
  function bind(obj, name, f) {
    if (f === void 0) f = obj[name];
    f = f.bind(obj);
    // TODO(jmesserly): track the function's signature on the function, instead
    // of having to go back to the class?
    let sig = _getMethodType(obj, name);
    assert(sig);
    rtti.tag(f, sig);
    return f;
  }
  exports.bind = bind;

  // Set up the method signature field on the constructor
  function _setMethodSignature(f, sigF) {
    defineMemoizedGetter(f, _methodSig, () => {
      let sigObj = sigF();
      sigObj.__proto__ = f.__proto__[_methodSig];
      return sigObj;
    });
  }

  // Set up the constructor signature field on the constructor
  function _setConstructorSignature(f, sigF) {
    defineMemoizedGetter(f, _constructorSig, sigF);
  }

  // Set up the static signature field on the constructor
  function _setStaticSignature(f, sigF) {
    defineMemoizedGetter(f, _staticSig, sigF);
  }

  // Set the lazily computed runtime type field on static methods
  function _setStaticTypes(f, names) {
    for (let name of names) {
      rtti.tagMemoized(f[name], function() {
        let parts = f[_staticSig][name];
        return types.definiteFunctionType.apply(null, parts);
      })
    }
  }

  /// Set up the type signature of a class (constructor object)
  /// f is a constructor object
  /// signature is an object containing optional properties as follows:
  ///  methods: A function returning an object mapping method names
  ///   to method types.  The function is evaluated lazily and cached.
  ///  statics: A function returning an object mapping static method
  ///   names to types.  The function is evalutated lazily and cached.
  ///  names: An array of the names of the static methods.  Used to
  ///   permit eagerly setting the runtimeType field on the methods
  ///   while still lazily computing the type descriptor object.
  function setSignature(f, signature) {
    let constructors =
      ('constructors' in signature) ? signature.constructors : () => ({});
    let methods =
      ('methods' in signature) ? signature.methods : () => ({});
    let statics =
      ('statics' in signature) ? signature.statics : () => ({});
    let names =
      ('names' in signature) ? signature.names : [];
    _setConstructorSignature(f, constructors);
    _setMethodSignature(f, methods);
    _setStaticSignature(f, statics);
    _setStaticTypes(f, names);
    rtti.tagMemoized(f, () => core.Type);
  }
  exports.setSignature = setSignature;

  function hasMethod(obj, name) {
    return _getMethodType(obj, name) !== void 0;
  }
  exports.hasMethod = hasMethod;

  exports.getMethodType = _getMethodType;

  /**
   * This is called whenever a derived class needs to introduce a new field,
   * shadowing a field or getter/setter pair on its parent.
   *
   * This is important because otherwise, trying to read or write the field
   * would end up calling the getter or setter, and one of those might not even
   * exist, resulting in a runtime error. Even if they did exist, that's the
   * wrong behavior if a new field was declared.
   */
  function virtualField(subclass, fieldName) {
    // If the field is already overridden, do nothing.
    let prop = getOwnPropertyDescriptor(subclass.prototype, fieldName);
    if (prop) return;

    let symbol = Symbol(subclass.name + '.' + fieldName);
    defineProperty(subclass.prototype, fieldName, {
      get: function() { return this[symbol]; },
      set: function(x) { this[symbol] = x; }
    });
  }
  exports.virtualField = virtualField;

  /**
   * Given a class and an initializer method name, creates a constructor
   * function with the same name. For example `new SomeClass.name(args)`.
   */
  function defineNamedConstructor(clazz, name) {
    let proto = clazz.prototype;
    let initMethod = proto[name];
    let ctor = function() { return initMethod.apply(this, arguments); };
    ctor.prototype = proto;
    // Use defineProperty so we don't hit a property defined on Function,
    // like `caller` and `arguments`.
    defineProperty(clazz, name, { value: ctor, configurable: true });
  }
  exports.defineNamedConstructor = defineNamedConstructor;

  let _extensionType = Symbol('extensionType');

  let dartx = {};
  exports.dartx = dartx;

  function getExtensionSymbol(name) {
    let sym = dartx[name];
    if (!sym) dartx[name] = sym = Symbol('dartx.' + name);
    return sym;
  }

  function defineExtensionNames(names) {
    names.forEach(getExtensionSymbol);
  }
  exports.defineExtensionNames = defineExtensionNames;

  /**
   * Copy symbols from the prototype of the source to destination.
   * These are the only properties safe to copy onto an existing public
   * JavaScript class.
   */
  function registerExtension(jsType, dartExtType) {
    let extProto = dartExtType.prototype;
    let jsProto = jsType.prototype;

    // Mark the JS type's instances so we can easily check for extensions.
    assert(jsProto[_extensionType] === void 0);
    jsProto[_extensionType] = extProto;

    let dartObjProto = core.Object.prototype;
    while (extProto !== dartObjProto && extProto !== jsProto) {
      copyTheseProperties(jsProto, extProto, getOwnPropertySymbols(extProto));
      extProto = extProto.__proto__;
    }
    let originalSigFn = getOwnPropertyDescriptor(dartExtType, _methodSig).get;
    assert(originalSigFn);
    defineMemoizedGetter(jsType, _methodSig, originalSigFn);
  }
  exports.registerExtension = registerExtension;

  /**
   * Mark a concrete type as implementing extension methods.
   * For example: `class MyIter implements Iterable`.
   *
   * This takes a list of names, which are the extension methods implemented.
   * It will add a forwarder, so the extension method name redirects to the
   * normal Dart method name. For example:
   *
   *     defineExtensionMembers(MyType, ['add', 'remove']);
   *
   * Results in:
   *
   *     MyType.prototype[dartx.add] = MyType.prototype.add;
   *     MyType.prototype[dartx.remove] = MyType.prototype.remove;
   */
  // TODO(jmesserly): essentially this gives two names to the same method.
  // This benefit is roughly equivalent call performance either way, but the
  // cost is we need to call defineExtensionMembers any time a subclass
  // overrides one of these methods.
  function defineExtensionMembers(type, methodNames) {
    let proto = type.prototype;
    for (let name of methodNames) {
      let method = getOwnPropertyDescriptor(proto, name);
      defineProperty(proto, getExtensionSymbol(name), method);
    }
    // Ensure the signature is available too.
    // TODO(jmesserly): not sure if we can do this in a cleaner way. Essentially
    // we need to copy the signature (and in the future, other data like
    // annotations) any time we copy a method as part of our metaprogramming.
    // It might be more friendly to JS metaprogramming if we include this info
    // on the function.
    let originalSigFn = getOwnPropertyDescriptor(type, _methodSig).get;
    defineMemoizedGetter(type, _methodSig, function() {
      let sig = originalSigFn();
      for (let name of methodNames) {
        sig[getExtensionSymbol(name)] = sig[name];
      }
      return sig;
    });
  }
  exports.defineExtensionMembers = defineExtensionMembers;

  function canonicalMember(obj, name) {
    if (obj != null && obj[_extensionType]) return dartx[name];
    // Check for certain names that we can't use in JS
    if (name == 'constructor' || name == 'prototype') {
      name = '+' + name;
    }
    return name;
  }
  exports.canonicalMember = canonicalMember;

  /** Sets the type of `obj` to be `type` */
  function setType(obj, type) {
    obj.__proto__ = type.prototype;
    return obj;
  }

  /** Sets the element type of a list literal. */
  function list(obj, elementType) {
    return setType(obj, _interceptors.JSArray$(elementType));
  }
  exports.list = list;

  function setBaseClass(derived, base) {
    // Link the extension to the type it's extending as a base class.
    derived.prototype.__proto__ = base.prototype;
  }
  exports.setBaseClass = setBaseClass;

});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/*
 * This library encapsulates the core sdk errors that the runtime knows about.
 *
 */

dart_library.library('dart/_errors', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_operations',
  'dart/core',
  'dart/_js_helper'
], function(exports, operations, core, _js_helper) {
  'use strict';

  function throwNoSuchMethod(obj, name, pArgs, nArgs, extras) {
    operations.throw(new core.NoSuchMethodError(obj, name, pArgs, nArgs, extras));
  }
  exports.throwNoSuchMethod = throwNoSuchMethod;

  function throwCastError(actual, type) {
    operations.throw(new _js_helper.CastErrorImplementation(actual, type));
  }
  exports.throwCastError = throwCastError;

  function throwAssertionError() {
    operations.throw(new core.AssertionError());
  }
  exports.throwAssertionError = throwAssertionError;

  function throwNullValueError() {
    // TODO(vsm): Per spec, we should throw an NSM here.  Technically, we ought
    // to thread through method info, but that uglifies the code and can't
    // actually be queried ... it only affects how the error is printed.
    operations.throw(new core.NoSuchMethodError(null,
      new core.Symbol('<Unexpected Null Value>'), null, null, null));
  }
  exports.throwNullValueError = throwNullValueError;
});
dart_library.library('dart/_foreign_helper', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
], function(exports, dart, core) {
  'use strict';
  let dartx = dart.dartx;
  function JS(typeDescription, codeTemplate, arg0, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11) {
    if (arg0 === void 0)
      arg0 = null;
    if (arg1 === void 0)
      arg1 = null;
    if (arg2 === void 0)
      arg2 = null;
    if (arg3 === void 0)
      arg3 = null;
    if (arg4 === void 0)
      arg4 = null;
    if (arg5 === void 0)
      arg5 = null;
    if (arg6 === void 0)
      arg6 = null;
    if (arg7 === void 0)
      arg7 = null;
    if (arg8 === void 0)
      arg8 = null;
    if (arg9 === void 0)
      arg9 = null;
    if (arg10 === void 0)
      arg10 = null;
    if (arg11 === void 0)
      arg11 = null;
  }
  dart.fn(JS, dart.dynamic, [core.String, core.String], [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]);
  function JS_CURRENT_ISOLATE_CONTEXT() {
  }
  dart.fn(JS_CURRENT_ISOLATE_CONTEXT, () => dart.definiteFunctionType(IsolateContext, []));
  class IsolateContext extends core.Object {}
  function JS_CALL_IN_ISOLATE(isolate, func) {
  }
  dart.fn(JS_CALL_IN_ISOLATE, dart.dynamic, [dart.dynamic, core.Function]);
  function JS_SET_CURRENT_ISOLATE(isolate) {
  }
  dart.fn(JS_SET_CURRENT_ISOLATE, dart.void, [dart.dynamic]);
  function JS_CREATE_ISOLATE() {
  }
  dart.fn(JS_CREATE_ISOLATE);
  function JS_DART_OBJECT_CONSTRUCTOR() {
  }
  dart.fn(JS_DART_OBJECT_CONSTRUCTOR);
  function JS_INTERCEPTOR_CONSTANT(type) {
  }
  dart.fn(JS_INTERCEPTOR_CONSTANT, dart.dynamic, [core.Type]);
  function JS_OPERATOR_IS_PREFIX() {
  }
  dart.fn(JS_OPERATOR_IS_PREFIX, core.String, []);
  function JS_OPERATOR_AS_PREFIX() {
  }
  dart.fn(JS_OPERATOR_AS_PREFIX, core.String, []);
  function JS_OBJECT_CLASS_NAME() {
  }
  dart.fn(JS_OBJECT_CLASS_NAME, core.String, []);
  function JS_NULL_CLASS_NAME() {
  }
  dart.fn(JS_NULL_CLASS_NAME, core.String, []);
  function JS_FUNCTION_CLASS_NAME() {
  }
  dart.fn(JS_FUNCTION_CLASS_NAME, core.String, []);
  function JS_IS_INDEXABLE_FIELD_NAME() {
  }
  dart.fn(JS_IS_INDEXABLE_FIELD_NAME, core.String, []);
  function JS_CURRENT_ISOLATE() {
  }
  dart.fn(JS_CURRENT_ISOLATE);
  function JS_SIGNATURE_NAME() {
  }
  dart.fn(JS_SIGNATURE_NAME, core.String, []);
  function JS_TYPEDEF_TAG() {
  }
  dart.fn(JS_TYPEDEF_TAG, core.String, []);
  function JS_FUNCTION_TYPE_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_TAG, core.String, []);
  function JS_FUNCTION_TYPE_VOID_RETURN_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_VOID_RETURN_TAG, core.String, []);
  function JS_FUNCTION_TYPE_RETURN_TYPE_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_RETURN_TYPE_TAG, core.String, []);
  function JS_FUNCTION_TYPE_REQUIRED_PARAMETERS_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_REQUIRED_PARAMETERS_TAG, core.String, []);
  function JS_FUNCTION_TYPE_OPTIONAL_PARAMETERS_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_OPTIONAL_PARAMETERS_TAG, core.String, []);
  function JS_FUNCTION_TYPE_NAMED_PARAMETERS_TAG() {
  }
  dart.fn(JS_FUNCTION_TYPE_NAMED_PARAMETERS_TAG, core.String, []);
  function JS_GET_NAME(name) {
  }
  dart.fn(JS_GET_NAME, core.String, [core.String]);
  function JS_EMBEDDED_GLOBAL(typeDescription, name) {
  }
  dart.fn(JS_EMBEDDED_GLOBAL, dart.dynamic, [core.String, core.String]);
  function JS_GET_FLAG(name) {
  }
  dart.fn(JS_GET_FLAG, core.bool, [core.String]);
  function JS_EFFECT(code) {
    dart.dcall(code, null);
  }
  dart.fn(JS_EFFECT, dart.void, [core.Function]);
  class JS_CONST extends core.Object {
    JS_CONST(code) {
      this.code = code;
    }
  }
  dart.setSignature(JS_CONST, {
    constructors: () => ({JS_CONST: [JS_CONST, [core.String]]})
  });
  function JS_STRING_CONCAT(a, b) {
    return a + b;
  }
  dart.fn(JS_STRING_CONCAT, core.String, [core.String, core.String]);
  // Exports:
  exports.JS = JS;
  exports.JS_CURRENT_ISOLATE_CONTEXT = JS_CURRENT_ISOLATE_CONTEXT;
  exports.IsolateContext = IsolateContext;
  exports.JS_CALL_IN_ISOLATE = JS_CALL_IN_ISOLATE;
  exports.JS_SET_CURRENT_ISOLATE = JS_SET_CURRENT_ISOLATE;
  exports.JS_CREATE_ISOLATE = JS_CREATE_ISOLATE;
  exports.JS_DART_OBJECT_CONSTRUCTOR = JS_DART_OBJECT_CONSTRUCTOR;
  exports.JS_INTERCEPTOR_CONSTANT = JS_INTERCEPTOR_CONSTANT;
  exports.JS_OPERATOR_IS_PREFIX = JS_OPERATOR_IS_PREFIX;
  exports.JS_OPERATOR_AS_PREFIX = JS_OPERATOR_AS_PREFIX;
  exports.JS_OBJECT_CLASS_NAME = JS_OBJECT_CLASS_NAME;
  exports.JS_NULL_CLASS_NAME = JS_NULL_CLASS_NAME;
  exports.JS_FUNCTION_CLASS_NAME = JS_FUNCTION_CLASS_NAME;
  exports.JS_IS_INDEXABLE_FIELD_NAME = JS_IS_INDEXABLE_FIELD_NAME;
  exports.JS_CURRENT_ISOLATE = JS_CURRENT_ISOLATE;
  exports.JS_SIGNATURE_NAME = JS_SIGNATURE_NAME;
  exports.JS_TYPEDEF_TAG = JS_TYPEDEF_TAG;
  exports.JS_FUNCTION_TYPE_TAG = JS_FUNCTION_TYPE_TAG;
  exports.JS_FUNCTION_TYPE_VOID_RETURN_TAG = JS_FUNCTION_TYPE_VOID_RETURN_TAG;
  exports.JS_FUNCTION_TYPE_RETURN_TYPE_TAG = JS_FUNCTION_TYPE_RETURN_TYPE_TAG;
  exports.JS_FUNCTION_TYPE_REQUIRED_PARAMETERS_TAG = JS_FUNCTION_TYPE_REQUIRED_PARAMETERS_TAG;
  exports.JS_FUNCTION_TYPE_OPTIONAL_PARAMETERS_TAG = JS_FUNCTION_TYPE_OPTIONAL_PARAMETERS_TAG;
  exports.JS_FUNCTION_TYPE_NAMED_PARAMETERS_TAG = JS_FUNCTION_TYPE_NAMED_PARAMETERS_TAG;
  exports.JS_GET_NAME = JS_GET_NAME;
  exports.JS_EMBEDDED_GLOBAL = JS_EMBEDDED_GLOBAL;
  exports.JS_GET_FLAG = JS_GET_FLAG;
  exports.JS_EFFECT = JS_EFFECT;
  exports.JS_CONST = JS_CONST;
  exports.JS_STRING_CONCAT = JS_STRING_CONCAT;
});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/**
 * This library adapts ES6 generators to implement Dart's async/await.
 *
 * It's designed to interact with Dart's Future/Stream and follow Dart
 * async/await semantics.
 *
 * See https://github.com/dart-lang/dev_compiler/issues/245 for ideas on
 * reconciling Dart's Future and ES6 Promise.
 *
 * Inspired by `co`: https://github.com/tj/co/blob/master/index.js, which is a
 * stepping stone for proposed ES7 async/await, and uses ES6 Promises.
 */
dart_library.library('dart/_generators', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_operations',
  'dart/_js_helper',
  'dart/core',
  'dart/collection',
  'dart/async'
], function(exports, _operations, _js_helper, core, collection, async) {
  'use strict';

  const _jsIterator = Symbol('_jsIterator');
  const _current = Symbol('_current');

  function syncStar(gen, E, ...args) {
    const SyncIterable_E = _js_helper.SyncIterable$(E);
    return new SyncIterable_E(gen, args);
  }
  exports.syncStar = syncStar;

  function async_(gen, T, ...args) {
    let iter;
    function onValue(res) {
      if (res === void 0) res = null;
      return next(iter.next(res));
    }
    function onError(err) {
      // If the awaited Future throws, we want to convert this to an exception
      // thrown from the `yield` point, as if it was thrown there.
      //
      // If the exception is not caught inside `gen`, it will emerge here, which
      // will send it to anyone listening on this async function's Future<T>.
      //
      // In essence, we are giving the code inside the generator a chance to
      // use try-catch-finally.
      return next(iter.throw(err));
    }
    function next(ret) {
      if (ret.done) return ret.value;
      // Checks if the awaited value is a Future.
      let future = ret.value;
      if (!_operations.instanceOf(future, async.Future$)) {
        future = async.Future.value(future);
      }
      // Chain the Future so `await` receives the Future's value.
      return future.then(onValue, {onError: onError});
    }
    return async.Future$(T).new(function() {
      iter = gen(...args)[Symbol.iterator]();
      return onValue();
    });
  }
  exports.async = async_;

  // Implementation inspired by _AsyncStarStreamController in
  // dart-lang/sdk's runtime/lib/core_patch.dart
  //
  // Given input like:
  //
  //     foo() async* {
  //       yield 1;
  //       yield* bar();
  //       print(await baz());
  //     }
  //
  // This generates as:
  //
  //    function foo() {
  //      return dart.asyncStar(function*(stream) {
  //        if (stream.add(1)) return;
  //        yield;
  //        if (stream.addStream(bar()) return;
  //        yield;
  //        print(yield baz());
  //      });
  //    }
  class _AsyncStarStreamController {
    constructor(generator, T, args) {
      this.isAdding = false;
      this.isWaiting = false;
      this.isScheduled = false;
      this.isSuspendedAtYield = false;
      this.canceler = null;
      this.iterator = generator(this, ...args)[Symbol.iterator]();
      this.controller = async.StreamController$(T).new({
        onListen: () => this.scheduleGenerator(),
        onResume: () => this.onResume(),
        onCancel: () => this.onCancel()
      });
    }

    onResume() {
      if (this.isSuspendedAtYield) {
        this.scheduleGenerator();
      }
    }

    onCancel() {
      if (this.controller.isClosed) {
        return null;
      }
      if (this.canceler == null) {
        this.canceler = async.Completer.new();
        this.scheduleGenerator();
      }
      return this.canceler.future;
    }

    close() {
      if (this.canceler != null && !this.canceler.isCompleted) {
        // If the stream has been cancelled, complete the cancellation future
        // with the error.
        this.canceler.complete();
      }
      this.controller.close();
    }

    scheduleGenerator() {
      // TODO(jmesserly): is this paused check in the right place? Assuming the
      // async* Stream yields, then is paused (by other code), the body will
      // already be scheduled. This will cause at least one more iteration to
      // run (adding another data item to the Stream) before actually pausing.
      // It could be fixed by moving the `isPaused` check inside `runBody`.
      if (this.isScheduled || this.controller.isPaused ||
          this.isAdding || this.isWaiting) {
        return;
      }
      this.isScheduled = true;
      async.scheduleMicrotask(() => this.runBody());
    }

    runBody(opt_awaitValue) {
      this.isScheduled = false;
      this.isSuspendedAtYield = false;
      this.isWaiting = false;
      let iter;
      try {
        iter = this.iterator.next(opt_awaitValue);
      } catch (e) {
        this.addError(e, _operations.stackTrace(e));
        this.close();
        return;
      }
      if (iter.done) {
        this.close();
        return;
      }

      // If we're suspended at a yield/yield*, we're done for now.
      if (this.isSuspendedAtYield || this.isAdding) return;

      // Handle `await`: if we get a value passed to `yield` it means we are
      // waiting on this Future. Make sure to prevent scheduling, and pass the
      // value back as the result of the `yield`.
      //
      // TODO(jmesserly): is the timing here correct? The assumption here is
      // that we should schedule `await` in `async*` the same as in `async`.
      this.isWaiting = true;
      let future = iter.value;
      if (!_operations.instanceOf(future, async.Future$)) {
        future = async.Future.value(future);
      }
      return future.then((x) => this.runBody(x),
          { onError: (e, s) => this.throwError(e, s) });
    }

    // Adds element to stream, returns true if the caller should terminate
    // execution of the generator.
    add(event) {
      // If stream is cancelled, tell caller to exit the async generator.
      if (!this.controller.hasListener) return true;
      this.controller.add(event);
      this.scheduleGenerator();
      this.isSuspendedAtYield = true;
      return false;
    }

    // Adds the elements of stream into this controller's stream.
    // The generator will be scheduled again when all of the
    // elements of the added stream have been consumed.
    // Returns true if the caller should terminate
    // execution of the generator.
    addStream(stream) {
      // If stream is cancelled, tell caller to exit the async generator.
      if (!this.controller.hasListener) return true;

      this.isAdding = true;
      this.controller.addStream(stream, {cancelOnError: false}).then(() => {
        this.isAdding = false;
        this.scheduleGenerator();
      }, { onError: (e, s) => this.throwError(e, s) });
    }

    throwError(error, stackTrace) {
      try {
        this.iterator.throw(error);
      } catch (e) {
        this.addError(e, stackTrace);
      }
    }

    addError(error, stackTrace) {
      if ((this.canceler != null) && !this.canceler.isCompleted) {
        // If the stream has been cancelled, complete the cancellation future
        // with the error.
        this.canceler.completeError(error, stackTrace);
        return;
      }
      if (!this.controller.hasListener) return;
      this.controller.addError(error, stackTrace);
    }
  }

  /** Returns a Stream of T implemented by an async* function. */
  function asyncStar(gen, T, ...args) {
    return new _AsyncStarStreamController(gen, T, args).controller.stream;
  }
  exports.asyncStar = asyncStar;
});
dart_library.library('dart/_interceptors', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/_internal',
  'dart/collection',
  'dart/math'
], /* Lazy imports */[
  'dart/_js_helper'
], function(exports, dart, core, _internal, collection, math, _js_helper) {
  'use strict';
  let dartx = dart.dartx;
  const JSArray$ = dart.generic(function(E) {
    dart.defineExtensionNames([
      'checkGrowable',
      'add',
      'removeAt',
      'insert',
      'insertAll',
      'setAll',
      'removeLast',
      'remove',
      'removeWhere',
      'retainWhere',
      'where',
      'expand',
      'addAll',
      'clear',
      'forEach',
      'map',
      'join',
      'take',
      'takeWhile',
      'skip',
      'skipWhile',
      'reduce',
      'fold',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'elementAt',
      'sublist',
      'getRange',
      'first',
      'last',
      'single',
      'removeRange',
      'setRange',
      'fillRange',
      'replaceRange',
      'any',
      'every',
      'reversed',
      'sort',
      'shuffle',
      'indexOf',
      'lastIndexOf',
      'contains',
      'isEmpty',
      'isNotEmpty',
      'toString',
      'toList',
      'toSet',
      'iterator',
      'hashCode',
      'length',
      'length',
      'get',
      'set',
      'asMap'
    ]);
    class JSArray extends core.Object {
      JSArray() {
      }
      static typed(allocation) {
        return dart.list(allocation, E);
      }
      static markFixed(allocation) {
        return JSArray$(E).typed(JSArray$().markFixedList(dart.as(allocation, core.List)));
      }
      static markGrowable(allocation) {
        return JSArray$().typed(allocation);
      }
      static markFixedList(list) {
        list.fixed$length = Array;
        return list;
      }
      [dartx.checkGrowable](reason) {
        if (this.fixed$length) {
          dart.throw(new core.UnsupportedError(dart.as(reason, core.String)));
        }
      }
      [dartx.add](value) {
        dart.as(value, E);
        this[dartx.checkGrowable]('add');
        this.push(value);
      }
      [dartx.removeAt](index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError(index));
        if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(this[dartx.length])) {
          dart.throw(new core.RangeError.value(index));
        }
        this[dartx.checkGrowable]('removeAt');
        return this.splice(index, 1)[0];
      }
      [dartx.insert](index, value) {
        dart.as(value, E);
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError(index));
        if (dart.notNull(index) < 0 || dart.notNull(index) > dart.notNull(this[dartx.length])) {
          dart.throw(new core.RangeError.value(index));
        }
        this[dartx.checkGrowable]('insert');
        this.splice(index, 0, value);
      }
      [dartx.insertAll](index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        this[dartx.checkGrowable]('insertAll');
        _internal.IterableMixinWorkaround.insertAllList(this, index, iterable);
      }
      [dartx.setAll](index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        _internal.IterableMixinWorkaround.setAllList(this, index, iterable);
      }
      [dartx.removeLast]() {
        this[dartx.checkGrowable]('removeLast');
        if (this[dartx.length] == 0)
          dart.throw(new core.RangeError.value(-1));
        return dart.as(this.pop(), E);
      }
      [dartx.remove](element) {
        this[dartx.checkGrowable]('remove');
        for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.equals(this[dartx.get](i), element)) {
            this.splice(i, 1);
            return true;
          }
        }
        return false;
      }
      [dartx.removeWhere](test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        _internal.IterableMixinWorkaround.removeWhereList(this, test);
      }
      [dartx.retainWhere](test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        _internal.IterableMixinWorkaround.removeWhereList(this, dart.fn(element => !dart.notNull(test(element)), core.bool, [E]));
      }
      [dartx.where](f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return new (_internal.IterableMixinWorkaround$(E))().where(this, f);
      }
      [dartx.expand](f) {
        dart.as(f, dart.functionType(core.Iterable, [E]));
        return _internal.IterableMixinWorkaround.expand(this, f);
      }
      [dartx.addAll](collection) {
        dart.as(collection, core.Iterable$(E));
        for (let e of collection) {
          this[dartx.add](e);
        }
      }
      [dartx.clear]() {
        this[dartx.length] = 0;
      }
      [dartx.forEach](f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        let length = this[dartx.length];
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          f(dart.as(this[i], E));
          if (length != this[dartx.length]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
      }
      [dartx.map](f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return _internal.IterableMixinWorkaround.mapList(this, f);
      }
      [dartx.join](separator) {
        if (separator === void 0)
          separator = "";
        let list = core.List.new(this[dartx.length]);
        for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
          list[dartx.set](i, `${this[dartx.get](i)}`);
        }
        return list.join(separator);
      }
      [dartx.take](n) {
        return new (_internal.IterableMixinWorkaround$(E))().takeList(this, n);
      }
      [dartx.takeWhile](test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.IterableMixinWorkaround$(E))().takeWhile(this, test);
      }
      [dartx.skip](n) {
        return new (_internal.IterableMixinWorkaround$(E))().skipList(this, n);
      }
      [dartx.skipWhile](test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.IterableMixinWorkaround$(E))().skipWhile(this, test);
      }
      [dartx.reduce](combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        return dart.as(_internal.IterableMixinWorkaround.reduce(this, combine), E);
      }
      [dartx.fold](initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        return _internal.IterableMixinWorkaround.fold(this, initialValue, combine);
      }
      [dartx.firstWhere](test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        return dart.as(_internal.IterableMixinWorkaround.firstWhere(this, test, orElse), E);
      }
      [dartx.lastWhere](test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        return dart.as(_internal.IterableMixinWorkaround.lastWhereList(this, test, orElse), E);
      }
      [dartx.singleWhere](test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return dart.as(_internal.IterableMixinWorkaround.singleWhere(this, test), E);
      }
      [dartx.elementAt](index) {
        return this[dartx.get](index);
      }
      [dartx.sublist](start, end) {
        if (end === void 0)
          end = null;
        _js_helper.checkNull(start);
        if (!(typeof start == 'number'))
          dart.throw(new core.ArgumentError(start));
        if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(this[dartx.length])) {
          dart.throw(new core.RangeError.range(start, 0, this[dartx.length]));
        }
        if (end == null) {
          end = this[dartx.length];
        } else {
          if (!(typeof end == 'number'))
            dart.throw(new core.ArgumentError(end));
          if (dart.notNull(end) < dart.notNull(start) || dart.notNull(end) > dart.notNull(this[dartx.length])) {
            dart.throw(new core.RangeError.range(end, start, this[dartx.length]));
          }
        }
        if (start == end)
          return dart.list([], E);
        return JSArray$(E).typed(this.slice(start, end));
      }
      [dartx.getRange](start, end) {
        return new (_internal.IterableMixinWorkaround$(E))().getRangeList(this, start, end);
      }
      get [dartx.first]() {
        if (dart.notNull(this[dartx.length]) > 0)
          return this[dartx.get](0);
        dart.throw(new core.StateError("No elements"));
      }
      get [dartx.last]() {
        if (dart.notNull(this[dartx.length]) > 0)
          return this[dartx.get](dart.notNull(this[dartx.length]) - 1);
        dart.throw(new core.StateError("No elements"));
      }
      get [dartx.single]() {
        if (this[dartx.length] == 1)
          return this[dartx.get](0);
        if (this[dartx.length] == 0)
          dart.throw(new core.StateError("No elements"));
        dart.throw(new core.StateError("More than one element"));
      }
      [dartx.removeRange](start, end) {
        this[dartx.checkGrowable]('removeRange');
        let receiverLength = this[dartx.length];
        if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(receiverLength)) {
          dart.throw(new core.RangeError.range(start, 0, receiverLength));
        }
        if (dart.notNull(end) < dart.notNull(start) || dart.notNull(end) > dart.notNull(receiverLength)) {
          dart.throw(new core.RangeError.range(end, start, receiverLength));
        }
        _internal.Lists.copy(this, end, this, start, dart.notNull(receiverLength) - dart.notNull(end));
        this[dartx.length] = dart.notNull(receiverLength) - (dart.notNull(end) - dart.notNull(start));
      }
      [dartx.setRange](start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        _internal.IterableMixinWorkaround.setRangeList(this, start, end, iterable, skipCount);
      }
      [dartx.fillRange](start, end, fillValue) {
        if (fillValue === void 0)
          fillValue = null;
        dart.as(fillValue, E);
        _internal.IterableMixinWorkaround.fillRangeList(this, start, end, fillValue);
      }
      [dartx.replaceRange](start, end, iterable) {
        dart.as(iterable, core.Iterable$(E));
        _internal.IterableMixinWorkaround.replaceRangeList(this, start, end, iterable);
      }
      [dartx.any](f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return _internal.IterableMixinWorkaround.any(this, f);
      }
      [dartx.every](f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return _internal.IterableMixinWorkaround.every(this, f);
      }
      get [dartx.reversed]() {
        return new (_internal.IterableMixinWorkaround$(E))().reversedList(this);
      }
      [dartx.sort](compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, dart.functionType(core.int, [E, E]));
        _internal.IterableMixinWorkaround.sortList(this, compare);
      }
      [dartx.shuffle](random) {
        if (random === void 0)
          random = null;
        _internal.IterableMixinWorkaround.shuffleList(this, random);
      }
      [dartx.indexOf](element, start) {
        if (start === void 0)
          start = 0;
        return _internal.IterableMixinWorkaround.indexOfList(this, element, start);
      }
      [dartx.lastIndexOf](element, start) {
        if (start === void 0)
          start = null;
        return _internal.IterableMixinWorkaround.lastIndexOfList(this, element, start);
      }
      [dartx.contains](other) {
        for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.equals(this[dartx.get](i), other))
            return true;
        }
        return false;
      }
      get [dartx.isEmpty]() {
        return this[dartx.length] == 0;
      }
      get [dartx.isNotEmpty]() {
        return !dart.notNull(this[dartx.isEmpty]);
      }
      toString() {
        return collection.ListBase.listToString(this);
      }
      [dartx.toList](opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let list = this.slice();
        if (!dart.notNull(growable))
          JSArray$().markFixedList(dart.as(list, core.List));
        return JSArray$(E).typed(list);
      }
      [dartx.toSet]() {
        return core.Set$(E).from(this);
      }
      get [dartx.iterator]() {
        return new (_internal.ListIterator$(E))(this);
      }
      get hashCode() {
        return _js_helper.Primitives.objectHashCode(this);
      }
      get [dartx.length]() {
        return this.length;
      }
      set [dartx.length](newLength) {
        if (!(typeof newLength == 'number'))
          dart.throw(new core.ArgumentError(newLength));
        if (dart.notNull(newLength) < 0)
          dart.throw(new core.RangeError.value(newLength));
        this[dartx.checkGrowable]('set length');
        this.length = newLength;
      }
      [dartx.get](index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError(index));
        if (dart.notNull(index) >= dart.notNull(this[dartx.length]) || dart.notNull(index) < 0)
          dart.throw(new core.RangeError.value(index));
        return dart.as(this[index], E);
      }
      [dartx.set](index, value) {
        dart.as(value, E);
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError(index));
        if (dart.notNull(index) >= dart.notNull(this[dartx.length]) || dart.notNull(index) < 0)
          dart.throw(new core.RangeError.value(index));
        this[index] = value;
        return value;
      }
      [dartx.asMap]() {
        return new (_internal.IterableMixinWorkaround$(E))().asMapList(this);
      }
    }
    dart.setBaseClass(JSArray, dart.global.Array);
    JSArray[dart.implements] = () => [core.List$(E), JSIndexable];
    dart.setSignature(JSArray, {
      constructors: () => ({
        JSArray: [JSArray$(E), []],
        typed: [JSArray$(E), [dart.dynamic]],
        markFixed: [JSArray$(E), [dart.dynamic]],
        markGrowable: [JSArray$(E), [dart.dynamic]]
      }),
      methods: () => ({
        [dartx.checkGrowable]: [dart.dynamic, [dart.dynamic]],
        [dartx.add]: [dart.void, [E]],
        [dartx.removeAt]: [E, [core.int]],
        [dartx.insert]: [dart.void, [core.int, E]],
        [dartx.insertAll]: [dart.void, [core.int, core.Iterable$(E)]],
        [dartx.setAll]: [dart.void, [core.int, core.Iterable$(E)]],
        [dartx.removeLast]: [E, []],
        [dartx.remove]: [core.bool, [core.Object]],
        [dartx.removeWhere]: [dart.void, [dart.functionType(core.bool, [E])]],
        [dartx.retainWhere]: [dart.void, [dart.functionType(core.bool, [E])]],
        [dartx.where]: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        [dartx.expand]: [core.Iterable, [dart.functionType(core.Iterable, [E])]],
        [dartx.addAll]: [dart.void, [core.Iterable$(E)]],
        [dartx.clear]: [dart.void, []],
        [dartx.forEach]: [dart.void, [dart.functionType(dart.void, [E])]],
        [dartx.map]: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        [dartx.join]: [core.String, [], [core.String]],
        [dartx.take]: [core.Iterable$(E), [core.int]],
        [dartx.takeWhile]: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        [dartx.skip]: [core.Iterable$(E), [core.int]],
        [dartx.skipWhile]: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        [dartx.reduce]: [E, [dart.functionType(E, [E, E])]],
        [dartx.fold]: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        [dartx.firstWhere]: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        [dartx.lastWhere]: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        [dartx.singleWhere]: [E, [dart.functionType(core.bool, [E])]],
        [dartx.elementAt]: [E, [core.int]],
        [dartx.sublist]: [core.List$(E), [core.int], [core.int]],
        [dartx.getRange]: [core.Iterable$(E), [core.int, core.int]],
        [dartx.removeRange]: [dart.void, [core.int, core.int]],
        [dartx.setRange]: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        [dartx.fillRange]: [dart.void, [core.int, core.int], [E]],
        [dartx.replaceRange]: [dart.void, [core.int, core.int, core.Iterable$(E)]],
        [dartx.any]: [core.bool, [dart.functionType(core.bool, [E])]],
        [dartx.every]: [core.bool, [dart.functionType(core.bool, [E])]],
        [dartx.sort]: [dart.void, [], [dart.functionType(core.int, [E, E])]],
        [dartx.shuffle]: [dart.void, [], [math.Random]],
        [dartx.indexOf]: [core.int, [core.Object], [core.int]],
        [dartx.lastIndexOf]: [core.int, [core.Object], [core.int]],
        [dartx.contains]: [core.bool, [core.Object]],
        [dartx.toList]: [core.List$(E), [], {growable: core.bool}],
        [dartx.toSet]: [core.Set$(E), []],
        [dartx.get]: [E, [core.int]],
        [dartx.set]: [dart.void, [core.int, E]],
        [dartx.asMap]: [core.Map$(core.int, E), []]
      }),
      statics: () => ({markFixedList: [core.List, [core.List]]}),
      names: ['markFixedList']
    });
    JSArray[dart.metadata] = () => [dart.const(new _js_helper.JsPeerInterface({name: 'Array'}))];
    return JSArray;
  });
  let JSArray = JSArray$();
  dart.registerExtension(dart.global.Array, JSArray);
  const JSMutableArray$ = dart.generic(function(E) {
    class JSMutableArray extends JSArray$(E) {
      JSMutableArray() {
        super.JSArray();
      }
    }
    JSMutableArray[dart.implements] = () => [JSMutableIndexable];
    return JSMutableArray;
  });
  let JSMutableArray = JSMutableArray$();
  const JSFixedArray$ = dart.generic(function(E) {
    class JSFixedArray extends JSMutableArray$(E) {}
    return JSFixedArray;
  });
  let JSFixedArray = JSFixedArray$();
  const JSExtendableArray$ = dart.generic(function(E) {
    class JSExtendableArray extends JSMutableArray$(E) {}
    return JSExtendableArray;
  });
  let JSExtendableArray = JSExtendableArray$();
  class Interceptor extends core.Object {
    Interceptor() {
    }
  }
  dart.setSignature(Interceptor, {
    constructors: () => ({Interceptor: [Interceptor, []]})
  });
  const _isInt32 = Symbol('_isInt32');
  const _tdivSlow = Symbol('_tdivSlow');
  const _shlPositive = Symbol('_shlPositive');
  const _shrOtherPositive = Symbol('_shrOtherPositive');
  const _shrBothPositive = Symbol('_shrBothPositive');
  dart.defineExtensionNames([
    'compareTo',
    'isNegative',
    'isNaN',
    'isInfinite',
    'isFinite',
    'remainder',
    'abs',
    'sign',
    'toInt',
    'truncate',
    'ceil',
    'floor',
    'round',
    'ceilToDouble',
    'floorToDouble',
    'roundToDouble',
    'truncateToDouble',
    'clamp',
    'toDouble',
    'toStringAsFixed',
    'toStringAsExponential',
    'toStringAsPrecision',
    'toRadixString',
    'toString',
    'hashCode',
    'unary-',
    '+',
    '-',
    '/',
    '*',
    '%',
    '~/',
    '<<',
    '>>',
    '&',
    '|',
    '^',
    '<',
    '>',
    '<=',
    '>=',
    'isEven',
    'isOdd',
    'toUnsigned',
    'toSigned',
    'bitLength',
    '~'
  ]);
  class JSNumber extends Interceptor {
    JSNumber() {
      super.Interceptor();
    }
    [dartx.compareTo](b) {
      if (this < dart.notNull(b)) {
        return -1;
      } else if (this > dart.notNull(b)) {
        return 1;
      } else if (dart.equals(this, b)) {
        if (dart.equals(this, 0)) {
          let bIsNegative = b[dartx.isNegative];
          if (this[dartx.isNegative] == bIsNegative)
            return 0;
          if (dart.notNull(this[dartx.isNegative]))
            return -1;
          return 1;
        }
        return 0;
      } else if (dart.notNull(this[dartx.isNaN])) {
        if (dart.notNull(b[dartx.isNaN])) {
          return 0;
        }
        return 1;
      } else {
        return -1;
      }
    }
    get [dartx.isNegative]() {
      return dart.equals(this, 0) ? 1 / this < 0 : this < 0;
    }
    get [dartx.isNaN]() {
      return isNaN(this);
    }
    get [dartx.isInfinite]() {
      return this == Infinity || this == -Infinity;
    }
    get [dartx.isFinite]() {
      return isFinite(this);
    }
    [dartx.remainder](b) {
      _js_helper.checkNull(b);
      return this % b;
    }
    [dartx.abs]() {
      return Math.abs(this);
    }
    get [dartx.sign]() {
      return this > 0 ? 1 : this < 0 ? -1 : this;
    }
    [dartx.toInt]() {
      if (this >= dart.notNull(JSNumber._MIN_INT32) && this <= dart.notNull(JSNumber._MAX_INT32)) {
        return this | 0;
      }
      if (isFinite(this)) {
        return this[dartx.truncateToDouble]() + 0;
      }
      dart.throw(new core.UnsupportedError('' + this));
    }
    [dartx.truncate]() {
      return this[dartx.toInt]();
    }
    [dartx.ceil]() {
      return this[dartx.ceilToDouble]()[dartx.toInt]();
    }
    [dartx.floor]() {
      return this[dartx.floorToDouble]()[dartx.toInt]();
    }
    [dartx.round]() {
      return this[dartx.roundToDouble]()[dartx.toInt]();
    }
    [dartx.ceilToDouble]() {
      return Math.ceil(this);
    }
    [dartx.floorToDouble]() {
      return Math.floor(this);
    }
    [dartx.roundToDouble]() {
      if (this < 0) {
        return -Math.round(-this);
      } else {
        return Math.round(this);
      }
    }
    [dartx.truncateToDouble]() {
      return this < 0 ? this[dartx.ceilToDouble]() : this[dartx.floorToDouble]();
    }
    [dartx.clamp](lowerLimit, upperLimit) {
      if (dart.notNull(lowerLimit[dartx.compareTo](upperLimit)) > 0) {
        dart.throw(new core.ArgumentError(lowerLimit));
      }
      if (dart.notNull(this[dartx.compareTo](lowerLimit)) < 0)
        return lowerLimit;
      if (dart.notNull(this[dartx.compareTo](upperLimit)) > 0)
        return upperLimit;
      return this;
    }
    [dartx.toDouble]() {
      return this;
    }
    [dartx.toStringAsFixed](fractionDigits) {
      _js_helper.checkInt(fractionDigits);
      if (dart.notNull(fractionDigits) < 0 || dart.notNull(fractionDigits) > 20) {
        dart.throw(new core.RangeError(fractionDigits));
      }
      let result = this.toFixed(fractionDigits);
      if (dart.equals(this, 0) && dart.notNull(this[dartx.isNegative]))
        return `-${result}`;
      return result;
    }
    [dartx.toStringAsExponential](fractionDigits) {
      if (fractionDigits === void 0)
        fractionDigits = null;
      let result = null;
      if (fractionDigits != null) {
        _js_helper.checkInt(fractionDigits);
        if (dart.notNull(fractionDigits) < 0 || dart.notNull(fractionDigits) > 20) {
          dart.throw(new core.RangeError(fractionDigits));
        }
        result = this.toExponential(fractionDigits);
      } else {
        result = this.toExponential();
      }
      if (dart.equals(this, 0) && dart.notNull(this[dartx.isNegative]))
        return `-${result}`;
      return result;
    }
    [dartx.toStringAsPrecision](precision) {
      _js_helper.checkInt(precision);
      if (dart.notNull(precision) < 1 || dart.notNull(precision) > 21) {
        dart.throw(new core.RangeError(precision));
      }
      let result = this.toPrecision(precision);
      if (dart.equals(this, 0) && dart.notNull(this[dartx.isNegative]))
        return `-${result}`;
      return result;
    }
    [dartx.toRadixString](radix) {
      _js_helper.checkInt(radix);
      if (dart.notNull(radix) < 2 || dart.notNull(radix) > 36)
        dart.throw(new core.RangeError(radix));
      let result = this.toString(radix);
      let rightParenCode = 41;
      if (result[dartx.codeUnitAt](dart.notNull(result[dartx.length]) - 1) != rightParenCode) {
        return result;
      }
      return JSNumber._handleIEtoString(result);
    }
    static _handleIEtoString(result) {
      let match = /^([\da-z]+)(?:\.([\da-z]+))?\(e\+(\d+)\)$/.exec(result);
      if (match == null) {
        dart.throw(new core.UnsupportedError(`Unexpected toString result: ${result}`));
      }
      result = dart.dindex(match, 1);
      let exponent = +dart.dindex(match, 3);
      if (dart.dindex(match, 2) != null) {
        result = result + dart.dindex(match, 2);
        exponent = dart.notNull(exponent) - dart.dindex(match, 2).length;
      }
      return dart.notNull(result) + "0"[dartx['*']](exponent);
    }
    toString() {
      if (dart.equals(this, 0) && 1 / this < 0) {
        return '-0.0';
      } else {
        return "" + this;
      }
    }
    get hashCode() {
      return this & 0x1FFFFFFF;
    }
    [dartx['unary-']]() {
      return -this;
    }
    [dartx['+']](other) {
      _js_helper.checkNull(other);
      return this + other;
    }
    [dartx['-']](other) {
      _js_helper.checkNull(other);
      return this - other;
    }
    [dartx['/']](other) {
      _js_helper.checkNull(other);
      return this / other;
    }
    [dartx['*']](other) {
      _js_helper.checkNull(other);
      return this * other;
    }
    [dartx['%']](other) {
      _js_helper.checkNull(other);
      let result = this % other;
      if (result == 0)
        return 0;
      if (dart.notNull(result) > 0)
        return result;
      if (other < 0) {
        return dart.notNull(result) - other;
      } else {
        return dart.notNull(result) + other;
      }
    }
    [_isInt32](value) {
      return (value | 0) === value;
    }
    [dartx['~/']](other) {
      if (dart.notNull(this[_isInt32](this)) && dart.notNull(this[_isInt32](other)) && 0 != other && -1 != other) {
        return this / other | 0;
      } else {
        return this[_tdivSlow](other);
      }
    }
    [_tdivSlow](other) {
      _js_helper.checkNull(other);
      return (this / other)[dartx.toInt]();
    }
    [dartx['<<']](other) {
      if (dart.notNull(other) < 0)
        dart.throw(new core.ArgumentError(other));
      return this[_shlPositive](other);
    }
    [_shlPositive](other) {
      return other > 31 ? 0 : this << other >>> 0;
    }
    [dartx['>>']](other) {
      if (dart.notNull(other) < 0)
        dart.throw(new core.ArgumentError(other));
      return this[_shrOtherPositive](other);
    }
    [_shrOtherPositive](other) {
      return this > 0 ? this[_shrBothPositive](other) : this >> (dart.notNull(other) > 31 ? 31 : other) >>> 0;
    }
    [_shrBothPositive](other) {
      return other > 31 ? 0 : this >>> other;
    }
    [dartx['&']](other) {
      _js_helper.checkNull(other);
      return (this & other) >>> 0;
    }
    [dartx['|']](other) {
      _js_helper.checkNull(other);
      return (this | other) >>> 0;
    }
    [dartx['^']](other) {
      _js_helper.checkNull(other);
      return (this ^ other) >>> 0;
    }
    [dartx['<']](other) {
      _js_helper.checkNull(other);
      return this < other;
    }
    [dartx['>']](other) {
      _js_helper.checkNull(other);
      return this > other;
    }
    [dartx['<=']](other) {
      _js_helper.checkNull(other);
      return this <= other;
    }
    [dartx['>=']](other) {
      _js_helper.checkNull(other);
      return this >= other;
    }
    get [dartx.isEven]() {
      return (this & 1) == 0;
    }
    get [dartx.isOdd]() {
      return (this & 1) == 1;
    }
    [dartx.toUnsigned](width) {
      return this & (1 << dart.notNull(width)) - 1;
    }
    [dartx.toSigned](width) {
      let signMask = 1 << dart.notNull(width) - 1;
      return (this & dart.notNull(signMask) - 1) - (this & dart.notNull(signMask));
    }
    get [dartx.bitLength]() {
      let nonneg = this < 0 ? -this - 1 : this;
      if (dart.notNull(nonneg) >= 4294967296) {
        nonneg = (dart.notNull(nonneg) / 4294967296)[dartx.truncate]();
        return dart.notNull(JSNumber._bitCount(JSNumber._spread(nonneg))) + 32;
      }
      return JSNumber._bitCount(JSNumber._spread(nonneg));
    }
    static _bitCount(i) {
      i = dart.notNull(JSNumber._shru(i, 0)) - (dart.notNull(JSNumber._shru(i, 1)) & 1431655765);
      i = (dart.notNull(i) & 858993459) + (dart.notNull(JSNumber._shru(i, 2)) & 858993459);
      i = 252645135 & dart.notNull(i) + dart.notNull(JSNumber._shru(i, 4));
      i = dart.notNull(i) + dart.notNull(JSNumber._shru(i, 8));
      i = dart.notNull(i) + dart.notNull(JSNumber._shru(i, 16));
      return dart.notNull(i) & 63;
    }
    static _shru(value, shift) {
      return value >>> shift;
    }
    static _shrs(value, shift) {
      return value >> shift;
    }
    static _ors(a, b) {
      return a | b;
    }
    static _spread(i) {
      i = JSNumber._ors(i, JSNumber._shrs(i, 1));
      i = JSNumber._ors(i, JSNumber._shrs(i, 2));
      i = JSNumber._ors(i, JSNumber._shrs(i, 4));
      i = JSNumber._ors(i, JSNumber._shrs(i, 8));
      i = JSNumber._shru(JSNumber._ors(i, JSNumber._shrs(i, 16)), 0);
      return i;
    }
    [dartx['~']]() {
      return ~this >>> 0;
    }
  }
  JSNumber[dart.implements] = () => [core.int, core.double];
  dart.setSignature(JSNumber, {
    constructors: () => ({JSNumber: [JSNumber, []]}),
    methods: () => ({
      [dartx.compareTo]: [core.int, [core.num]],
      [dartx.remainder]: [JSNumber, [core.num]],
      [dartx.abs]: [JSNumber, []],
      [dartx.toInt]: [core.int, []],
      [dartx.truncate]: [core.int, []],
      [dartx.ceil]: [core.int, []],
      [dartx.floor]: [core.int, []],
      [dartx.round]: [core.int, []],
      [dartx.ceilToDouble]: [core.double, []],
      [dartx.floorToDouble]: [core.double, []],
      [dartx.roundToDouble]: [core.double, []],
      [dartx.truncateToDouble]: [core.double, []],
      [dartx.clamp]: [core.num, [core.num, core.num]],
      [dartx.toDouble]: [core.double, []],
      [dartx.toStringAsFixed]: [core.String, [core.int]],
      [dartx.toStringAsExponential]: [core.String, [], [core.int]],
      [dartx.toStringAsPrecision]: [core.String, [core.int]],
      [dartx.toRadixString]: [core.String, [core.int]],
      [dartx['unary-']]: [JSNumber, []],
      [dartx['+']]: [JSNumber, [core.num]],
      [dartx['-']]: [JSNumber, [core.num]],
      [dartx['/']]: [core.double, [core.num]],
      [dartx['*']]: [JSNumber, [core.num]],
      [dartx['%']]: [JSNumber, [core.num]],
      [_isInt32]: [core.bool, [dart.dynamic]],
      [dartx['~/']]: [core.int, [core.num]],
      [_tdivSlow]: [core.int, [core.num]],
      [dartx['<<']]: [core.int, [core.num]],
      [_shlPositive]: [core.int, [core.num]],
      [dartx['>>']]: [core.int, [core.num]],
      [_shrOtherPositive]: [core.int, [core.num]],
      [_shrBothPositive]: [core.int, [core.num]],
      [dartx['&']]: [core.int, [core.num]],
      [dartx['|']]: [core.int, [core.num]],
      [dartx['^']]: [core.int, [core.num]],
      [dartx['<']]: [core.bool, [core.num]],
      [dartx['>']]: [core.bool, [core.num]],
      [dartx['<=']]: [core.bool, [core.num]],
      [dartx['>=']]: [core.bool, [core.num]],
      [dartx.toUnsigned]: [core.int, [core.int]],
      [dartx.toSigned]: [core.int, [core.int]],
      [dartx['~']]: [core.int, []]
    }),
    statics: () => ({
      _handleIEtoString: [core.String, [core.String]],
      _bitCount: [core.int, [core.int]],
      _shru: [core.int, [core.int, core.int]],
      _shrs: [core.int, [core.int, core.int]],
      _ors: [core.int, [core.int, core.int]],
      _spread: [core.int, [core.int]]
    }),
    names: ['_handleIEtoString', '_bitCount', '_shru', '_shrs', '_ors', '_spread']
  });
  JSNumber[dart.metadata] = () => [dart.const(new _js_helper.JsPeerInterface({name: 'Number'}))];
  dart.registerExtension(dart.global.Number, JSNumber);
  JSNumber._MIN_INT32 = -2147483648;
  JSNumber._MAX_INT32 = 2147483647;
  const _defaultSplit = Symbol('_defaultSplit');
  dart.defineExtensionNames([
    'codeUnitAt',
    'allMatches',
    'matchAsPrefix',
    '+',
    'endsWith',
    'replaceAll',
    'replaceAllMapped',
    'splitMapJoin',
    'replaceFirst',
    'split',
    'startsWith',
    'substring',
    'toLowerCase',
    'toUpperCase',
    'trim',
    'trimLeft',
    'trimRight',
    '*',
    'padLeft',
    'padRight',
    'codeUnits',
    'runes',
    'indexOf',
    'lastIndexOf',
    'contains',
    'isEmpty',
    'isNotEmpty',
    'compareTo',
    'toString',
    'hashCode',
    'runtimeType',
    'length',
    'get'
  ]);
  class JSString extends Interceptor {
    JSString() {
      super.Interceptor();
    }
    [dartx.codeUnitAt](index) {
      if (!(typeof index == 'number'))
        dart.throw(new core.ArgumentError(index));
      if (dart.notNull(index) < 0)
        dart.throw(new core.RangeError.value(index));
      if (dart.notNull(index) >= dart.notNull(this[dartx.length]))
        dart.throw(new core.RangeError.value(index));
      return this.charCodeAt(index);
    }
    [dartx.allMatches](string, start) {
      if (start === void 0)
        start = 0;
      _js_helper.checkString(string);
      _js_helper.checkInt(start);
      if (0 > dart.notNull(start) || dart.notNull(start) > dart.notNull(string[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, string[dartx.length]));
      }
      return _js_helper.allMatchesInStringUnchecked(this, string, start);
    }
    [dartx.matchAsPrefix](string, start) {
      if (start === void 0)
        start = 0;
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(string[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, string[dartx.length]));
      }
      if (dart.notNull(start) + dart.notNull(this[dartx.length]) > dart.notNull(string[dartx.length]))
        return null;
      for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
        if (string[dartx.codeUnitAt](dart.notNull(start) + dart.notNull(i)) != this[dartx.codeUnitAt](i)) {
          return null;
        }
      }
      return new _js_helper.StringMatch(start, string, this);
    }
    [dartx['+']](other) {
      if (!(typeof other == 'string'))
        dart.throw(new core.ArgumentError(other));
      return this + other;
    }
    [dartx.endsWith](other) {
      _js_helper.checkString(other);
      let otherLength = other[dartx.length];
      if (dart.notNull(otherLength) > dart.notNull(this[dartx.length]))
        return false;
      return other == this[dartx.substring](dart.notNull(this[dartx.length]) - dart.notNull(otherLength));
    }
    [dartx.replaceAll](from, to) {
      _js_helper.checkString(to);
      return dart.as(_js_helper.stringReplaceAllUnchecked(this, from, to), core.String);
    }
    [dartx.replaceAllMapped](from, convert) {
      return this[dartx.splitMapJoin](from, {onMatch: convert});
    }
    [dartx.splitMapJoin](from, opts) {
      let onMatch = opts && 'onMatch' in opts ? opts.onMatch : null;
      let onNonMatch = opts && 'onNonMatch' in opts ? opts.onNonMatch : null;
      return dart.as(_js_helper.stringReplaceAllFuncUnchecked(this, from, onMatch, onNonMatch), core.String);
    }
    [dartx.replaceFirst](from, to, startIndex) {
      if (startIndex === void 0)
        startIndex = 0;
      _js_helper.checkString(to);
      _js_helper.checkInt(startIndex);
      if (dart.notNull(startIndex) < 0 || dart.notNull(startIndex) > dart.notNull(this[dartx.length])) {
        dart.throw(new core.RangeError.range(startIndex, 0, this[dartx.length]));
      }
      return dart.as(_js_helper.stringReplaceFirstUnchecked(this, from, to, startIndex), core.String);
    }
    [dartx.split](pattern) {
      _js_helper.checkNull(pattern);
      if (typeof pattern == 'string') {
        return dart.as(this.split(pattern), core.List$(core.String));
      } else if (dart.is(pattern, _js_helper.JSSyntaxRegExp) && _js_helper.regExpCaptureCount(pattern) == 0) {
        let re = _js_helper.regExpGetNative(pattern);
        return dart.as(this.split(re), core.List$(core.String));
      } else {
        return this[_defaultSplit](pattern);
      }
    }
    [_defaultSplit](pattern) {
      let result = dart.list([], core.String);
      let start = 0;
      let length = 1;
      for (let match of pattern[dartx.allMatches](this)) {
        let matchStart = match.start;
        let matchEnd = match.end;
        length = dart.notNull(matchEnd) - dart.notNull(matchStart);
        if (length == 0 && start == matchStart) {
          continue;
        }
        let end = matchStart;
        result[dartx.add](this[dartx.substring](start, end));
        start = matchEnd;
      }
      if (dart.notNull(start) < dart.notNull(this[dartx.length]) || dart.notNull(length) > 0) {
        result[dartx.add](this[dartx.substring](start));
      }
      return result;
    }
    [dartx.startsWith](pattern, index) {
      if (index === void 0)
        index = 0;
      _js_helper.checkInt(index);
      if (dart.notNull(index) < 0 || dart.notNull(index) > dart.notNull(this[dartx.length])) {
        dart.throw(new core.RangeError.range(index, 0, this[dartx.length]));
      }
      if (typeof pattern == 'string') {
        let other = pattern;
        let otherLength = other[dartx.length];
        let endIndex = dart.notNull(index) + dart.notNull(otherLength);
        if (dart.notNull(endIndex) > dart.notNull(this[dartx.length]))
          return false;
        return other == this.substring(index, endIndex);
      }
      return pattern[dartx.matchAsPrefix](this, index) != null;
    }
    [dartx.substring](startIndex, endIndex) {
      if (endIndex === void 0)
        endIndex = null;
      _js_helper.checkInt(startIndex);
      if (endIndex == null)
        endIndex = this[dartx.length];
      _js_helper.checkInt(endIndex);
      if (dart.notNull(startIndex) < 0)
        dart.throw(new core.RangeError.value(startIndex));
      if (dart.notNull(startIndex) > dart.notNull(endIndex))
        dart.throw(new core.RangeError.value(startIndex));
      if (dart.notNull(endIndex) > dart.notNull(this[dartx.length]))
        dart.throw(new core.RangeError.value(endIndex));
      return this.substring(startIndex, endIndex);
    }
    [dartx.toLowerCase]() {
      return this.toLowerCase();
    }
    [dartx.toUpperCase]() {
      return this.toUpperCase();
    }
    static _isWhitespace(codeUnit) {
      if (dart.notNull(codeUnit) < 256) {
        switch (codeUnit) {
          case 9:
          case 10:
          case 11:
          case 12:
          case 13:
          case 32:
          case 133:
          case 160:
          {
            return true;
          }
          default:
          {
            return false;
          }
        }
      }
      switch (codeUnit) {
        case 5760:
        case 6158:
        case 8192:
        case 8193:
        case 8194:
        case 8195:
        case 8196:
        case 8197:
        case 8198:
        case 8199:
        case 8200:
        case 8201:
        case 8202:
        case 8232:
        case 8233:
        case 8239:
        case 8287:
        case 12288:
        case 65279:
        {
          return true;
        }
        default:
        {
          return false;
        }
      }
    }
    static _skipLeadingWhitespace(string, index) {
      let SPACE = 32;
      let CARRIAGE_RETURN = 13;
      while (dart.notNull(index) < dart.notNull(string[dartx.length])) {
        let codeUnit = string[dartx.codeUnitAt](index);
        if (codeUnit != SPACE && codeUnit != CARRIAGE_RETURN && !dart.notNull(JSString._isWhitespace(codeUnit))) {
          break;
        }
        index = dart.notNull(index) + 1;
      }
      return index;
    }
    static _skipTrailingWhitespace(string, index) {
      let SPACE = 32;
      let CARRIAGE_RETURN = 13;
      while (dart.notNull(index) > 0) {
        let codeUnit = string[dartx.codeUnitAt](dart.notNull(index) - 1);
        if (codeUnit != SPACE && codeUnit != CARRIAGE_RETURN && !dart.notNull(JSString._isWhitespace(codeUnit))) {
          break;
        }
        index = dart.notNull(index) - 1;
      }
      return index;
    }
    [dartx.trim]() {
      let NEL = 133;
      let result = this.trim();
      if (result[dartx.length] == 0)
        return result;
      let firstCode = result[dartx.codeUnitAt](0);
      let startIndex = 0;
      if (firstCode == NEL) {
        startIndex = JSString._skipLeadingWhitespace(result, 1);
        if (startIndex == result[dartx.length])
          return "";
      }
      let endIndex = result[dartx.length];
      let lastCode = result[dartx.codeUnitAt](dart.notNull(endIndex) - 1);
      if (lastCode == NEL) {
        endIndex = JSString._skipTrailingWhitespace(result, dart.notNull(endIndex) - 1);
      }
      if (startIndex == 0 && endIndex == result[dartx.length])
        return result;
      return result.substring(startIndex, endIndex);
    }
    [dartx.trimLeft]() {
      let NEL = 133;
      let result = null;
      let startIndex = 0;
      if (typeof this.trimLeft != "undefined") {
        result = this.trimLeft();
        if (result[dartx.length] == 0)
          return result;
        let firstCode = result[dartx.codeUnitAt](0);
        if (firstCode == NEL) {
          startIndex = JSString._skipLeadingWhitespace(result, 1);
        }
      } else {
        result = this;
        startIndex = JSString._skipLeadingWhitespace(this, 0);
      }
      if (startIndex == 0)
        return result;
      if (startIndex == result[dartx.length])
        return "";
      return result.substring(startIndex);
    }
    [dartx.trimRight]() {
      let NEL = 133;
      let result = null;
      let endIndex = null;
      if (typeof this.trimRight != "undefined") {
        result = this.trimRight();
        endIndex = result[dartx.length];
        if (endIndex == 0)
          return result;
        let lastCode = result[dartx.codeUnitAt](dart.notNull(endIndex) - 1);
        if (lastCode == NEL) {
          endIndex = JSString._skipTrailingWhitespace(result, dart.notNull(endIndex) - 1);
        }
      } else {
        result = this;
        endIndex = JSString._skipTrailingWhitespace(this, this[dartx.length]);
      }
      if (endIndex == result[dartx.length])
        return result;
      if (endIndex == 0)
        return "";
      return result.substring(0, endIndex);
    }
    [dartx['*']](times) {
      if (0 >= dart.notNull(times))
        return '';
      if (times == 1 || this[dartx.length] == 0)
        return this;
      if (times != times >>> 0) {
        dart.throw(dart.const(new core.OutOfMemoryError()));
      }
      let result = '';
      let s = this;
      while (true) {
        if ((dart.notNull(times) & 1) == 1)
          result = dart.notNull(s) + dart.notNull(result);
        times = times >>> 1;
        if (times == 0)
          break;
        s = dart.notNull(s) + dart.notNull(s);
      }
      return result;
    }
    [dartx.padLeft](width, padding) {
      if (padding === void 0)
        padding = ' ';
      let delta = dart.notNull(width) - dart.notNull(this[dartx.length]);
      if (dart.notNull(delta) <= 0)
        return this;
      return padding[dartx['*']](delta) + this;
    }
    [dartx.padRight](width, padding) {
      if (padding === void 0)
        padding = ' ';
      let delta = dart.notNull(width) - dart.notNull(this[dartx.length]);
      if (dart.notNull(delta) <= 0)
        return this;
      return this[dartx['+']](padding[dartx['*']](delta));
    }
    get [dartx.codeUnits]() {
      return new _CodeUnits(this);
    }
    get [dartx.runes]() {
      return new core.Runes(this);
    }
    [dartx.indexOf](pattern, start) {
      if (start === void 0)
        start = 0;
      _js_helper.checkNull(pattern);
      if (!(typeof start == 'number'))
        dart.throw(new core.ArgumentError(start));
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(this[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, this[dartx.length]));
      }
      if (typeof pattern == 'string') {
        return this.indexOf(pattern, start);
      }
      if (dart.is(pattern, _js_helper.JSSyntaxRegExp)) {
        let re = pattern;
        let match = _js_helper.firstMatchAfter(re, this, start);
        return match == null ? -1 : match.start;
      }
      for (let i = start; dart.notNull(i) <= dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
        if (pattern[dartx.matchAsPrefix](this, i) != null)
          return i;
      }
      return -1;
    }
    [dartx.lastIndexOf](pattern, start) {
      if (start === void 0)
        start = null;
      _js_helper.checkNull(pattern);
      if (start == null) {
        start = this[dartx.length];
      } else if (!(typeof start == 'number')) {
        dart.throw(new core.ArgumentError(start));
      } else if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(this[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, this[dartx.length]));
      }
      if (typeof pattern == 'string') {
        let other = pattern;
        if (dart.notNull(start) + dart.notNull(other[dartx.length]) > dart.notNull(this[dartx.length])) {
          start = dart.notNull(this[dartx.length]) - dart.notNull(other[dartx.length]);
        }
        return dart.as(_js_helper.stringLastIndexOfUnchecked(this, other, start), core.int);
      }
      for (let i = start; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
        if (pattern[dartx.matchAsPrefix](this, i) != null)
          return i;
      }
      return -1;
    }
    [dartx.contains](other, startIndex) {
      if (startIndex === void 0)
        startIndex = 0;
      _js_helper.checkNull(other);
      if (dart.notNull(startIndex) < 0 || dart.notNull(startIndex) > dart.notNull(this[dartx.length])) {
        dart.throw(new core.RangeError.range(startIndex, 0, this[dartx.length]));
      }
      return dart.as(_js_helper.stringContainsUnchecked(this, other, startIndex), core.bool);
    }
    get [dartx.isEmpty]() {
      return this[dartx.length] == 0;
    }
    get [dartx.isNotEmpty]() {
      return !dart.notNull(this[dartx.isEmpty]);
    }
    [dartx.compareTo](other) {
      if (!(typeof other == 'string'))
        dart.throw(new core.ArgumentError(other));
      return dart.equals(this, other) ? 0 : this < other ? -1 : 1;
    }
    toString() {
      return this;
    }
    get hashCode() {
      let hash = 0;
      for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
        hash = 536870911 & dart.notNull(hash) + this.charCodeAt(i);
        hash = 536870911 & dart.notNull(hash) + ((524287 & dart.notNull(hash)) << 10);
        hash = hash ^ hash >> 6;
      }
      hash = 536870911 & dart.notNull(hash) + ((67108863 & dart.notNull(hash)) << 3);
      hash = hash ^ hash >> 11;
      return 536870911 & dart.notNull(hash) + ((16383 & dart.notNull(hash)) << 15);
    }
    get runtimeType() {
      return core.String;
    }
    get [dartx.length]() {
      return this.length;
    }
    [dartx.get](index) {
      if (!(typeof index == 'number'))
        dart.throw(new core.ArgumentError(index));
      if (dart.notNull(index) >= dart.notNull(this[dartx.length]) || dart.notNull(index) < 0)
        dart.throw(new core.RangeError.value(index));
      return this[index];
    }
  }
  JSString[dart.implements] = () => [core.String, JSIndexable];
  dart.setSignature(JSString, {
    constructors: () => ({JSString: [JSString, []]}),
    methods: () => ({
      [dartx.codeUnitAt]: [core.int, [core.int]],
      [dartx.allMatches]: [core.Iterable$(core.Match), [core.String], [core.int]],
      [dartx.matchAsPrefix]: [core.Match, [core.String], [core.int]],
      [dartx['+']]: [core.String, [core.String]],
      [dartx.endsWith]: [core.bool, [core.String]],
      [dartx.replaceAll]: [core.String, [core.Pattern, core.String]],
      [dartx.replaceAllMapped]: [core.String, [core.Pattern, dart.functionType(core.String, [core.Match])]],
      [dartx.splitMapJoin]: [core.String, [core.Pattern], {onMatch: dart.functionType(core.String, [core.Match]), onNonMatch: dart.functionType(core.String, [core.String])}],
      [dartx.replaceFirst]: [core.String, [core.Pattern, core.String], [core.int]],
      [dartx.split]: [core.List$(core.String), [core.Pattern]],
      [_defaultSplit]: [core.List$(core.String), [core.Pattern]],
      [dartx.startsWith]: [core.bool, [core.Pattern], [core.int]],
      [dartx.substring]: [core.String, [core.int], [core.int]],
      [dartx.toLowerCase]: [core.String, []],
      [dartx.toUpperCase]: [core.String, []],
      [dartx.trim]: [core.String, []],
      [dartx.trimLeft]: [core.String, []],
      [dartx.trimRight]: [core.String, []],
      [dartx['*']]: [core.String, [core.int]],
      [dartx.padLeft]: [core.String, [core.int], [core.String]],
      [dartx.padRight]: [core.String, [core.int], [core.String]],
      [dartx.indexOf]: [core.int, [core.Pattern], [core.int]],
      [dartx.lastIndexOf]: [core.int, [core.Pattern], [core.int]],
      [dartx.contains]: [core.bool, [core.Pattern], [core.int]],
      [dartx.compareTo]: [core.int, [core.String]],
      [dartx.get]: [core.String, [core.int]]
    }),
    statics: () => ({
      _isWhitespace: [core.bool, [core.int]],
      _skipLeadingWhitespace: [core.int, [core.String, core.int]],
      _skipTrailingWhitespace: [core.int, [core.String, core.int]]
    }),
    names: ['_isWhitespace', '_skipLeadingWhitespace', '_skipTrailingWhitespace']
  });
  JSString[dart.metadata] = () => [dart.const(new _js_helper.JsPeerInterface({name: 'String'}))];
  dart.registerExtension(dart.global.String, JSString);
  const _string = Symbol('_string');
  class _CodeUnits extends _internal.UnmodifiableListBase$(core.int) {
    _CodeUnits(string) {
      this[_string] = string;
    }
    get length() {
      return this[_string][dartx.length];
    }
    get(i) {
      return this[_string][dartx.codeUnitAt](i);
    }
  }
  dart.setSignature(_CodeUnits, {
    constructors: () => ({_CodeUnits: [_CodeUnits, [core.String]]}),
    methods: () => ({get: [core.int, [core.int]]})
  });
  dart.defineExtensionMembers(_CodeUnits, ['get', 'length']);
  function getInterceptor(obj) {
    return obj;
  }
  dart.fn(getInterceptor);
  dart.defineExtensionNames([
    'toString',
    'hashCode',
    'runtimeType'
  ]);
  class JSBool extends Interceptor {
    JSBool() {
      super.Interceptor();
    }
    toString() {
      return String(this);
    }
    get hashCode() {
      return this ? 2 * 3 * 23 * 3761 : 269 * 811;
    }
    get runtimeType() {
      return core.bool;
    }
  }
  JSBool[dart.implements] = () => [core.bool];
  dart.setSignature(JSBool, {
    constructors: () => ({JSBool: [JSBool, []]})
  });
  JSBool[dart.metadata] = () => [dart.const(new _js_helper.JsPeerInterface({name: 'Boolean'}))];
  dart.registerExtension(dart.global.Boolean, JSBool);
  class JSIndexable extends core.Object {}
  class JSMutableIndexable extends JSIndexable {}
  class JSObject extends core.Object {}
  class JavaScriptObject extends Interceptor {
    JavaScriptObject() {
      super.Interceptor();
    }
    get hashCode() {
      return 0;
    }
    get runtimeType() {
      return JSObject;
    }
  }
  JavaScriptObject[dart.implements] = () => [JSObject];
  dart.setSignature(JavaScriptObject, {
    constructors: () => ({JavaScriptObject: [JavaScriptObject, []]})
  });
  class PlainJavaScriptObject extends JavaScriptObject {
    PlainJavaScriptObject() {
      super.JavaScriptObject();
    }
  }
  dart.setSignature(PlainJavaScriptObject, {
    constructors: () => ({PlainJavaScriptObject: [PlainJavaScriptObject, []]})
  });
  class UnknownJavaScriptObject extends JavaScriptObject {
    UnknownJavaScriptObject() {
      super.JavaScriptObject();
    }
    toString() {
      return String(this);
    }
  }
  dart.setSignature(UnknownJavaScriptObject, {
    constructors: () => ({UnknownJavaScriptObject: [UnknownJavaScriptObject, []]})
  });
  // Exports:
  exports.JSArray$ = JSArray$;
  exports.JSArray = JSArray;
  exports.JSMutableArray$ = JSMutableArray$;
  exports.JSMutableArray = JSMutableArray;
  exports.JSFixedArray$ = JSFixedArray$;
  exports.JSFixedArray = JSFixedArray;
  exports.JSExtendableArray$ = JSExtendableArray$;
  exports.JSExtendableArray = JSExtendableArray;
  exports.Interceptor = Interceptor;
  exports.JSNumber = JSNumber;
  exports.JSString = JSString;
  exports.getInterceptor = getInterceptor;
  exports.JSBool = JSBool;
  exports.JSIndexable = JSIndexable;
  exports.JSMutableIndexable = JSMutableIndexable;
  exports.JSObject = JSObject;
  exports.JavaScriptObject = JavaScriptObject;
  exports.PlainJavaScriptObject = PlainJavaScriptObject;
  exports.UnknownJavaScriptObject = UnknownJavaScriptObject;
});
dart_library.library('dart/_internal', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/collection'
], /* Lazy imports */[
  'dart/math',
  'dart/_interceptors',
  'dart/_js_primitives'
], function(exports, dart, core, collection, math, _interceptors, _js_primitives) {
  'use strict';
  let dartx = dart.dartx;
  class EfficientLength extends core.Object {}
  const ListIterable$ = dart.generic(function(E) {
    class ListIterable extends collection.IterableBase$(E) {
      ListIterable() {
        super.IterableBase();
      }
      get iterator() {
        return new (ListIterator$(E))(this);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          action(this.elementAt(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
      }
      get isEmpty() {
        return this.length == 0;
      }
      get first() {
        if (this.length == 0)
          dart.throw(IterableElementError.noElement());
        return this.elementAt(0);
      }
      get last() {
        if (this.length == 0)
          dart.throw(IterableElementError.noElement());
        return this.elementAt(dart.notNull(this.length) - 1);
      }
      get single() {
        if (this.length == 0)
          dart.throw(IterableElementError.noElement());
        if (dart.notNull(this.length) > 1)
          dart.throw(IterableElementError.tooMany());
        return this.elementAt(0);
      }
      contains(element) {
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.equals(this.elementAt(i), element))
            return true;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return false;
      }
      every(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (!dart.notNull(test(this.elementAt(i))))
            return false;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return true;
      }
      any(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.notNull(test(this.elementAt(i))))
            return true;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return false;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = this.elementAt(i);
          if (dart.notNull(test(element)))
            return element;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let length = this.length;
        for (let i = dart.notNull(length) - 1; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
          let element = this.elementAt(i);
          if (dart.notNull(test(element)))
            return element;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        let match = null;
        let matchFound = false;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = this.elementAt(i);
          if (dart.notNull(test(element))) {
            if (dart.notNull(matchFound)) {
              dart.throw(IterableElementError.tooMany());
            }
            matchFound = true;
            match = element;
          }
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (dart.notNull(matchFound))
          return match;
        dart.throw(IterableElementError.noElement());
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        let length = this.length;
        if (!dart.notNull(separator[dartx.isEmpty])) {
          if (length == 0)
            return "";
          let first = `${this.elementAt(0)}`;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          let buffer = new core.StringBuffer(first);
          for (let i = 1; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
            buffer.write(separator);
            buffer.write(this.elementAt(i));
            if (length != this.length) {
              dart.throw(new core.ConcurrentModificationError(this));
            }
          }
          return dart.toString(buffer);
        } else {
          let buffer = new core.StringBuffer();
          for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
            buffer.write(this.elementAt(i));
            if (length != this.length) {
              dart.throw(new core.ConcurrentModificationError(this));
            }
          }
          return dart.toString(buffer);
        }
      }
      where(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return super.where(test);
      }
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return new MappedListIterable(this, f);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [dart.dynamic, E]));
        let length = this.length;
        if (length == 0)
          dart.throw(IterableElementError.noElement());
        let value = this.elementAt(0);
        for (let i = 1; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          value = dart.dcall(combine, value, this.elementAt(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return value;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        let value = initialValue;
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          value = dart.dcall(combine, value, this.elementAt(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return value;
      }
      skip(count) {
        return new (SubListIterable$(E))(this, count, null);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return super.skipWhile(test);
      }
      take(count) {
        return new (SubListIterable$(E))(this, 0, count);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return super.takeWhile(test);
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let result = null;
        if (dart.notNull(growable)) {
          result = core.List$(E).new();
          result[dartx.length] = this.length;
        } else {
          result = core.List$(E).new(this.length);
        }
        for (let i = 0; dart.notNull(i) < dart.notNull(this.length); i = dart.notNull(i) + 1) {
          result[dartx.set](i, this.elementAt(i));
        }
        return result;
      }
      toSet() {
        let result = core.Set$(E).new();
        for (let i = 0; dart.notNull(i) < dart.notNull(this.length); i = dart.notNull(i) + 1) {
          result.add(this.elementAt(i));
        }
        return result;
      }
    }
    ListIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(ListIterable, {
      constructors: () => ({ListIterable: [ListIterable$(E), []]}),
      methods: () => ({
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])]],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        reduce: [E, [dart.functionType(E, [dart.dynamic, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        toList: [core.List$(E), [], {growable: core.bool}],
        toSet: [core.Set$(E), []]
      })
    });
    dart.defineExtensionMembers(ListIterable, [
      'forEach',
      'contains',
      'every',
      'any',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'join',
      'where',
      'map',
      'reduce',
      'fold',
      'skip',
      'skipWhile',
      'take',
      'takeWhile',
      'toList',
      'toSet',
      'iterator',
      'isEmpty',
      'first',
      'last',
      'single'
    ]);
    return ListIterable;
  });
  let ListIterable = ListIterable$();
  const _iterable = dart.JsSymbol('_iterable');
  const _start = dart.JsSymbol('_start');
  const _endOrLength = dart.JsSymbol('_endOrLength');
  const _endIndex = dart.JsSymbol('_endIndex');
  const _startIndex = dart.JsSymbol('_startIndex');
  const SubListIterable$ = dart.generic(function(E) {
    class SubListIterable extends ListIterable$(E) {
      SubListIterable(iterable, start, endOrLength) {
        this[_iterable] = iterable;
        this[_start] = start;
        this[_endOrLength] = endOrLength;
        super.ListIterable();
        core.RangeError.checkNotNegative(this[_start], "start");
        if (this[_endOrLength] != null) {
          core.RangeError.checkNotNegative(this[_endOrLength], "end");
          if (dart.notNull(this[_start]) > dart.notNull(this[_endOrLength])) {
            dart.throw(new core.RangeError.range(this[_start], 0, this[_endOrLength], "start"));
          }
        }
      }
      get [_endIndex]() {
        let length = this[_iterable][dartx.length];
        if (this[_endOrLength] == null || dart.notNull(this[_endOrLength]) > dart.notNull(length))
          return length;
        return this[_endOrLength];
      }
      get [_startIndex]() {
        let length = this[_iterable][dartx.length];
        if (dart.notNull(this[_start]) > dart.notNull(length))
          return length;
        return this[_start];
      }
      get length() {
        let length = this[_iterable][dartx.length];
        if (dart.notNull(this[_start]) >= dart.notNull(length))
          return 0;
        if (this[_endOrLength] == null || dart.notNull(this[_endOrLength]) >= dart.notNull(length)) {
          return dart.notNull(length) - dart.notNull(this[_start]);
        }
        return dart.notNull(this[_endOrLength]) - dart.notNull(this[_start]);
      }
      elementAt(index) {
        let realIndex = dart.notNull(this[_startIndex]) + dart.notNull(index);
        if (dart.notNull(index) < 0 || dart.notNull(realIndex) >= dart.notNull(this[_endIndex])) {
          dart.throw(core.RangeError.index(index, this, "index"));
        }
        return this[_iterable][dartx.elementAt](realIndex);
      }
      skip(count) {
        core.RangeError.checkNotNegative(count, "count");
        let newStart = dart.notNull(this[_start]) + dart.notNull(count);
        if (this[_endOrLength] != null && dart.notNull(newStart) >= dart.notNull(this[_endOrLength])) {
          return new (EmptyIterable$(E))();
        }
        return new (SubListIterable$(E))(this[_iterable], newStart, this[_endOrLength]);
      }
      take(count) {
        core.RangeError.checkNotNegative(count, "count");
        if (this[_endOrLength] == null) {
          return new (SubListIterable$(E))(this[_iterable], this[_start], dart.notNull(this[_start]) + dart.notNull(count));
        } else {
          let newEnd = dart.notNull(this[_start]) + dart.notNull(count);
          if (dart.notNull(this[_endOrLength]) < dart.notNull(newEnd))
            return this;
          return new (SubListIterable$(E))(this[_iterable], this[_start], newEnd);
        }
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let start = this[_start];
        let end = this[_iterable][dartx.length];
        if (this[_endOrLength] != null && dart.notNull(this[_endOrLength]) < dart.notNull(end))
          end = this[_endOrLength];
        let length = dart.notNull(end) - dart.notNull(start);
        if (dart.notNull(length) < 0)
          length = 0;
        let result = dart.notNull(growable) ? (() => {
          let _ = core.List$(E).new();
          _[dartx.length] = length;
          return _;
        })() : core.List$(E).new(length);
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          result[dartx.set](i, this[_iterable][dartx.elementAt](dart.notNull(start) + dart.notNull(i)));
          if (dart.notNull(this[_iterable][dartx.length]) < dart.notNull(end))
            dart.throw(new core.ConcurrentModificationError(this));
        }
        return dart.as(result, core.List$(E));
      }
    }
    dart.setSignature(SubListIterable, {
      constructors: () => ({SubListIterable: [SubListIterable$(E), [core.Iterable$(E), core.int, core.int]]}),
      methods: () => ({
        elementAt: [E, [core.int]],
        skip: [core.Iterable$(E), [core.int]],
        take: [core.Iterable$(E), [core.int]],
        toList: [core.List$(E), [], {growable: core.bool}]
      })
    });
    dart.defineExtensionMembers(SubListIterable, [
      'elementAt',
      'skip',
      'take',
      'toList',
      'length'
    ]);
    return SubListIterable;
  });
  let SubListIterable = SubListIterable$();
  const _length = dart.JsSymbol('_length');
  const _index = dart.JsSymbol('_index');
  const _current = dart.JsSymbol('_current');
  const ListIterator$ = dart.generic(function(E) {
    class ListIterator extends core.Object {
      ListIterator(iterable) {
        this[_iterable] = iterable;
        this[_length] = iterable[dartx.length];
        this[_index] = 0;
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        let length = this[_iterable][dartx.length];
        if (this[_length] != length) {
          dart.throw(new core.ConcurrentModificationError(this[_iterable]));
        }
        if (dart.notNull(this[_index]) >= dart.notNull(length)) {
          this[_current] = null;
          return false;
        }
        this[_current] = this[_iterable][dartx.elementAt](this[_index]);
        this[_index] = dart.notNull(this[_index]) + 1;
        return true;
      }
    }
    ListIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(ListIterator, {
      constructors: () => ({ListIterator: [ListIterator$(E), [core.Iterable$(E)]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return ListIterator;
  });
  let ListIterator = ListIterator$();
  const _Transformation$ = dart.generic(function(S, T) {
    const _Transformation = dart.typedef('_Transformation', () => dart.functionType(T, [S]));
    return _Transformation;
  });
  let _Transformation = _Transformation$();
  const _f = dart.JsSymbol('_f');
  const MappedIterable$ = dart.generic(function(S, T) {
    class MappedIterable extends collection.IterableBase$(T) {
      static new(iterable, func) {
        if (dart.is(iterable, EfficientLength)) {
          return new (EfficientLengthMappedIterable$(S, T))(iterable, func);
        }
        return new (MappedIterable$(S, T))._(dart.as(iterable, core.Iterable$(S)), func);
      }
      _(iterable, f) {
        this[_iterable] = iterable;
        this[_f] = f;
        super.IterableBase();
      }
      get iterator() {
        return new (MappedIterator$(S, T))(this[_iterable][dartx.iterator], this[_f]);
      }
      get length() {
        return this[_iterable][dartx.length];
      }
      get isEmpty() {
        return this[_iterable][dartx.isEmpty];
      }
      get first() {
        return this[_f](this[_iterable][dartx.first]);
      }
      get last() {
        return this[_f](this[_iterable][dartx.last]);
      }
      get single() {
        return this[_f](this[_iterable][dartx.single]);
      }
      elementAt(index) {
        return this[_f](this[_iterable][dartx.elementAt](index));
      }
    }
    dart.defineNamedConstructor(MappedIterable, '_');
    dart.setSignature(MappedIterable, {
      constructors: () => ({
        new: [MappedIterable$(S, T), [core.Iterable, dart.functionType(T, [S])]],
        _: [MappedIterable$(S, T), [core.Iterable$(S), dart.functionType(T, [S])]]
      }),
      methods: () => ({elementAt: [T, [core.int]]})
    });
    dart.defineExtensionMembers(MappedIterable, [
      'elementAt',
      'iterator',
      'length',
      'isEmpty',
      'first',
      'last',
      'single'
    ]);
    return MappedIterable;
  });
  let MappedIterable = MappedIterable$();
  const EfficientLengthMappedIterable$ = dart.generic(function(S, T) {
    class EfficientLengthMappedIterable extends MappedIterable$(S, T) {
      EfficientLengthMappedIterable(iterable, func) {
        super._(dart.as(iterable, core.Iterable$(S)), func);
      }
    }
    EfficientLengthMappedIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(EfficientLengthMappedIterable, {
      constructors: () => ({EfficientLengthMappedIterable: [EfficientLengthMappedIterable$(S, T), [core.Iterable, dart.functionType(T, [S])]]})
    });
    return EfficientLengthMappedIterable;
  });
  let EfficientLengthMappedIterable = EfficientLengthMappedIterable$();
  const _iterator = dart.JsSymbol('_iterator');
  const MappedIterator$ = dart.generic(function(S, T) {
    class MappedIterator extends core.Iterator$(T) {
      MappedIterator(iterator, f) {
        this[_iterator] = iterator;
        this[_f] = f;
        this[_current] = null;
      }
      moveNext() {
        if (dart.notNull(this[_iterator].moveNext())) {
          this[_current] = this[_f](this[_iterator].current);
          return true;
        }
        this[_current] = null;
        return false;
      }
      get current() {
        return this[_current];
      }
    }
    dart.setSignature(MappedIterator, {
      constructors: () => ({MappedIterator: [MappedIterator$(S, T), [core.Iterator$(S), dart.functionType(T, [S])]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return MappedIterator;
  });
  let MappedIterator = MappedIterator$();
  const _source = dart.JsSymbol('_source');
  const MappedListIterable$ = dart.generic(function(S, T) {
    class MappedListIterable extends ListIterable$(T) {
      MappedListIterable(source, f) {
        this[_source] = source;
        this[_f] = f;
        super.ListIterable();
      }
      get length() {
        return this[_source][dartx.length];
      }
      elementAt(index) {
        return this[_f](this[_source][dartx.elementAt](index));
      }
    }
    MappedListIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(MappedListIterable, {
      constructors: () => ({MappedListIterable: [MappedListIterable$(S, T), [core.Iterable$(S), dart.functionType(T, [S])]]}),
      methods: () => ({elementAt: [T, [core.int]]})
    });
    dart.defineExtensionMembers(MappedListIterable, ['elementAt', 'length']);
    return MappedListIterable;
  });
  let MappedListIterable = MappedListIterable$();
  const _ElementPredicate$ = dart.generic(function(E) {
    const _ElementPredicate = dart.typedef('_ElementPredicate', () => dart.functionType(core.bool, [E]));
    return _ElementPredicate;
  });
  let _ElementPredicate = _ElementPredicate$();
  const WhereIterable$ = dart.generic(function(E) {
    class WhereIterable extends collection.IterableBase$(E) {
      WhereIterable(iterable, f) {
        this[_iterable] = iterable;
        this[_f] = f;
        super.IterableBase();
      }
      get iterator() {
        return new (WhereIterator$(E))(this[_iterable][dartx.iterator], this[_f]);
      }
    }
    dart.setSignature(WhereIterable, {
      constructors: () => ({WhereIterable: [WhereIterable$(E), [core.Iterable$(E), dart.functionType(core.bool, [E])]]})
    });
    dart.defineExtensionMembers(WhereIterable, ['iterator']);
    return WhereIterable;
  });
  let WhereIterable = WhereIterable$();
  const WhereIterator$ = dart.generic(function(E) {
    class WhereIterator extends core.Iterator$(E) {
      WhereIterator(iterator, f) {
        this[_iterator] = iterator;
        this[_f] = f;
      }
      moveNext() {
        while (dart.notNull(this[_iterator].moveNext())) {
          if (dart.notNull(this[_f](this[_iterator].current))) {
            return true;
          }
        }
        return false;
      }
      get current() {
        return this[_iterator].current;
      }
    }
    dart.setSignature(WhereIterator, {
      constructors: () => ({WhereIterator: [WhereIterator$(E), [core.Iterator$(E), dart.functionType(core.bool, [E])]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return WhereIterator;
  });
  let WhereIterator = WhereIterator$();
  const _ExpandFunction$ = dart.generic(function(S, T) {
    const _ExpandFunction = dart.typedef('_ExpandFunction', () => dart.functionType(core.Iterable$(T), [S]));
    return _ExpandFunction;
  });
  let _ExpandFunction = _ExpandFunction$();
  const ExpandIterable$ = dart.generic(function(S, T) {
    class ExpandIterable extends collection.IterableBase$(T) {
      ExpandIterable(iterable, f) {
        this[_iterable] = iterable;
        this[_f] = f;
        super.IterableBase();
      }
      get iterator() {
        return new (ExpandIterator$(S, T))(this[_iterable][dartx.iterator], dart.as(this[_f], __CastType0$(S, T)));
      }
    }
    dart.setSignature(ExpandIterable, {
      constructors: () => ({ExpandIterable: [ExpandIterable$(S, T), [core.Iterable$(S), dart.functionType(core.Iterable$(T), [S])]]})
    });
    dart.defineExtensionMembers(ExpandIterable, ['iterator']);
    return ExpandIterable;
  });
  let ExpandIterable = ExpandIterable$();
  const _currentExpansion = dart.JsSymbol('_currentExpansion');
  const _nextExpansion = dart.JsSymbol('_nextExpansion');
  const ExpandIterator$ = dart.generic(function(S, T) {
    class ExpandIterator extends core.Object {
      ExpandIterator(iterator, f) {
        this[_iterator] = iterator;
        this[_f] = f;
        this[_currentExpansion] = dart.const(new (EmptyIterator$(T))());
        this[_current] = null;
      }
      [_nextExpansion]() {}
      get current() {
        return this[_current];
      }
      moveNext() {
        if (this[_currentExpansion] == null)
          return false;
        while (!dart.notNull(this[_currentExpansion].moveNext())) {
          this[_current] = null;
          if (dart.notNull(this[_iterator].moveNext())) {
            this[_currentExpansion] = null;
            this[_currentExpansion] = dart.as(dart.dcall(this[_f], this[_iterator].current)[dartx.iterator], core.Iterator$(T));
          } else {
            return false;
          }
        }
        this[_current] = this[_currentExpansion].current;
        return true;
      }
    }
    ExpandIterator[dart.implements] = () => [core.Iterator$(T)];
    dart.setSignature(ExpandIterator, {
      constructors: () => ({ExpandIterator: [ExpandIterator$(S, T), [core.Iterator$(S), dart.functionType(core.Iterable$(T), [S])]]}),
      methods: () => ({
        [_nextExpansion]: [dart.void, []],
        moveNext: [core.bool, []]
      })
    });
    return ExpandIterator;
  });
  let ExpandIterator = ExpandIterator$();
  const _takeCount = dart.JsSymbol('_takeCount');
  const TakeIterable$ = dart.generic(function(E) {
    class TakeIterable extends collection.IterableBase$(E) {
      static new(iterable, takeCount) {
        if (!(typeof takeCount == 'number') || dart.notNull(takeCount) < 0) {
          dart.throw(new core.ArgumentError(takeCount));
        }
        if (dart.is(iterable, EfficientLength)) {
          return new (EfficientLengthTakeIterable$(E))(iterable, takeCount);
        }
        return new (TakeIterable$(E))._(iterable, takeCount);
      }
      _(iterable, takeCount) {
        this[_iterable] = iterable;
        this[_takeCount] = takeCount;
        super.IterableBase();
      }
      get iterator() {
        return new (TakeIterator$(E))(this[_iterable][dartx.iterator], this[_takeCount]);
      }
    }
    dart.defineNamedConstructor(TakeIterable, '_');
    dart.setSignature(TakeIterable, {
      constructors: () => ({
        new: [TakeIterable$(E), [core.Iterable$(E), core.int]],
        _: [TakeIterable$(E), [core.Iterable$(E), core.int]]
      })
    });
    dart.defineExtensionMembers(TakeIterable, ['iterator']);
    return TakeIterable;
  });
  let TakeIterable = TakeIterable$();
  const EfficientLengthTakeIterable$ = dart.generic(function(E) {
    class EfficientLengthTakeIterable extends TakeIterable$(E) {
      EfficientLengthTakeIterable(iterable, takeCount) {
        super._(iterable, takeCount);
      }
      get length() {
        let iterableLength = this[_iterable][dartx.length];
        if (dart.notNull(iterableLength) > dart.notNull(this[_takeCount]))
          return this[_takeCount];
        return iterableLength;
      }
    }
    EfficientLengthTakeIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(EfficientLengthTakeIterable, {
      constructors: () => ({EfficientLengthTakeIterable: [EfficientLengthTakeIterable$(E), [core.Iterable$(E), core.int]]})
    });
    dart.defineExtensionMembers(EfficientLengthTakeIterable, ['length']);
    return EfficientLengthTakeIterable;
  });
  let EfficientLengthTakeIterable = EfficientLengthTakeIterable$();
  const _remaining = dart.JsSymbol('_remaining');
  const TakeIterator$ = dart.generic(function(E) {
    class TakeIterator extends core.Iterator$(E) {
      TakeIterator(iterator, remaining) {
        this[_iterator] = iterator;
        this[_remaining] = remaining;
        dart.assert(typeof this[_remaining] == 'number' && dart.notNull(this[_remaining]) >= 0);
      }
      moveNext() {
        this[_remaining] = dart.notNull(this[_remaining]) - 1;
        if (dart.notNull(this[_remaining]) >= 0) {
          return this[_iterator].moveNext();
        }
        this[_remaining] = -1;
        return false;
      }
      get current() {
        if (dart.notNull(this[_remaining]) < 0)
          return null;
        return this[_iterator].current;
      }
    }
    dart.setSignature(TakeIterator, {
      constructors: () => ({TakeIterator: [TakeIterator$(E), [core.Iterator$(E), core.int]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return TakeIterator;
  });
  let TakeIterator = TakeIterator$();
  const TakeWhileIterable$ = dart.generic(function(E) {
    class TakeWhileIterable extends collection.IterableBase$(E) {
      TakeWhileIterable(iterable, f) {
        this[_iterable] = iterable;
        this[_f] = f;
        super.IterableBase();
      }
      get iterator() {
        return new (TakeWhileIterator$(E))(this[_iterable][dartx.iterator], this[_f]);
      }
    }
    dart.setSignature(TakeWhileIterable, {
      constructors: () => ({TakeWhileIterable: [TakeWhileIterable$(E), [core.Iterable$(E), dart.functionType(core.bool, [E])]]})
    });
    dart.defineExtensionMembers(TakeWhileIterable, ['iterator']);
    return TakeWhileIterable;
  });
  let TakeWhileIterable = TakeWhileIterable$();
  const _isFinished = dart.JsSymbol('_isFinished');
  const TakeWhileIterator$ = dart.generic(function(E) {
    class TakeWhileIterator extends core.Iterator$(E) {
      TakeWhileIterator(iterator, f) {
        this[_iterator] = iterator;
        this[_f] = f;
        this[_isFinished] = false;
      }
      moveNext() {
        if (dart.notNull(this[_isFinished]))
          return false;
        if (!dart.notNull(this[_iterator].moveNext()) || !dart.notNull(this[_f](this[_iterator].current))) {
          this[_isFinished] = true;
          return false;
        }
        return true;
      }
      get current() {
        if (dart.notNull(this[_isFinished]))
          return null;
        return this[_iterator].current;
      }
    }
    dart.setSignature(TakeWhileIterator, {
      constructors: () => ({TakeWhileIterator: [TakeWhileIterator$(E), [core.Iterator$(E), dart.functionType(core.bool, [E])]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return TakeWhileIterator;
  });
  let TakeWhileIterator = TakeWhileIterator$();
  const _skipCount = dart.JsSymbol('_skipCount');
  const SkipIterable$ = dart.generic(function(E) {
    class SkipIterable extends collection.IterableBase$(E) {
      static new(iterable, count) {
        if (dart.is(iterable, EfficientLength)) {
          return new (EfficientLengthSkipIterable$(E))(iterable, count);
        }
        return new (SkipIterable$(E))._(iterable, count);
      }
      _(iterable, skipCount) {
        this[_iterable] = iterable;
        this[_skipCount] = skipCount;
        super.IterableBase();
        if (!(typeof this[_skipCount] == 'number')) {
          dart.throw(new core.ArgumentError.value(this[_skipCount], "count is not an integer"));
        }
        core.RangeError.checkNotNegative(this[_skipCount], "count");
      }
      skip(count) {
        if (!(typeof this[_skipCount] == 'number')) {
          dart.throw(new core.ArgumentError.value(this[_skipCount], "count is not an integer"));
        }
        core.RangeError.checkNotNegative(this[_skipCount], "count");
        return new (SkipIterable$(E))._(this[_iterable], dart.notNull(this[_skipCount]) + dart.notNull(count));
      }
      get iterator() {
        return new (SkipIterator$(E))(this[_iterable][dartx.iterator], this[_skipCount]);
      }
    }
    dart.defineNamedConstructor(SkipIterable, '_');
    dart.setSignature(SkipIterable, {
      constructors: () => ({
        new: [SkipIterable$(E), [core.Iterable$(E), core.int]],
        _: [SkipIterable$(E), [core.Iterable$(E), core.int]]
      }),
      methods: () => ({skip: [core.Iterable$(E), [core.int]]})
    });
    dart.defineExtensionMembers(SkipIterable, ['skip', 'iterator']);
    return SkipIterable;
  });
  let SkipIterable = SkipIterable$();
  const EfficientLengthSkipIterable$ = dart.generic(function(E) {
    class EfficientLengthSkipIterable extends SkipIterable$(E) {
      EfficientLengthSkipIterable(iterable, skipCount) {
        super._(iterable, skipCount);
      }
      get length() {
        let length = dart.notNull(this[_iterable][dartx.length]) - dart.notNull(this[_skipCount]);
        if (dart.notNull(length) >= 0)
          return length;
        return 0;
      }
    }
    EfficientLengthSkipIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(EfficientLengthSkipIterable, {
      constructors: () => ({EfficientLengthSkipIterable: [EfficientLengthSkipIterable$(E), [core.Iterable$(E), core.int]]})
    });
    dart.defineExtensionMembers(EfficientLengthSkipIterable, ['length']);
    return EfficientLengthSkipIterable;
  });
  let EfficientLengthSkipIterable = EfficientLengthSkipIterable$();
  const SkipIterator$ = dart.generic(function(E) {
    class SkipIterator extends core.Iterator$(E) {
      SkipIterator(iterator, skipCount) {
        this[_iterator] = iterator;
        this[_skipCount] = skipCount;
        dart.assert(typeof this[_skipCount] == 'number' && dart.notNull(this[_skipCount]) >= 0);
      }
      moveNext() {
        for (let i = 0; dart.notNull(i) < dart.notNull(this[_skipCount]); i = dart.notNull(i) + 1)
          this[_iterator].moveNext();
        this[_skipCount] = 0;
        return this[_iterator].moveNext();
      }
      get current() {
        return this[_iterator].current;
      }
    }
    dart.setSignature(SkipIterator, {
      constructors: () => ({SkipIterator: [SkipIterator$(E), [core.Iterator$(E), core.int]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return SkipIterator;
  });
  let SkipIterator = SkipIterator$();
  const SkipWhileIterable$ = dart.generic(function(E) {
    class SkipWhileIterable extends collection.IterableBase$(E) {
      SkipWhileIterable(iterable, f) {
        this[_iterable] = iterable;
        this[_f] = f;
        super.IterableBase();
      }
      get iterator() {
        return new (SkipWhileIterator$(E))(this[_iterable][dartx.iterator], this[_f]);
      }
    }
    dart.setSignature(SkipWhileIterable, {
      constructors: () => ({SkipWhileIterable: [SkipWhileIterable$(E), [core.Iterable$(E), dart.functionType(core.bool, [E])]]})
    });
    dart.defineExtensionMembers(SkipWhileIterable, ['iterator']);
    return SkipWhileIterable;
  });
  let SkipWhileIterable = SkipWhileIterable$();
  const _hasSkipped = dart.JsSymbol('_hasSkipped');
  const SkipWhileIterator$ = dart.generic(function(E) {
    class SkipWhileIterator extends core.Iterator$(E) {
      SkipWhileIterator(iterator, f) {
        this[_iterator] = iterator;
        this[_f] = f;
        this[_hasSkipped] = false;
      }
      moveNext() {
        if (!dart.notNull(this[_hasSkipped])) {
          this[_hasSkipped] = true;
          while (dart.notNull(this[_iterator].moveNext())) {
            if (!dart.notNull(this[_f](this[_iterator].current)))
              return true;
          }
        }
        return this[_iterator].moveNext();
      }
      get current() {
        return this[_iterator].current;
      }
    }
    dart.setSignature(SkipWhileIterator, {
      constructors: () => ({SkipWhileIterator: [SkipWhileIterator$(E), [core.Iterator$(E), dart.functionType(core.bool, [E])]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return SkipWhileIterator;
  });
  let SkipWhileIterator = SkipWhileIterator$();
  const EmptyIterable$ = dart.generic(function(E) {
    class EmptyIterable extends collection.IterableBase$(E) {
      EmptyIterable() {
        super.IterableBase();
      }
      get iterator() {
        return dart.const(new (EmptyIterator$(E))());
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
      }
      get isEmpty() {
        return true;
      }
      get length() {
        return 0;
      }
      get first() {
        dart.throw(IterableElementError.noElement());
      }
      get last() {
        dart.throw(IterableElementError.noElement());
      }
      get single() {
        dart.throw(IterableElementError.noElement());
      }
      elementAt(index) {
        dart.throw(new core.RangeError.range(index, 0, 0, "index"));
      }
      contains(element) {
        return false;
      }
      every(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return true;
      }
      any(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return false;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      singleWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        return "";
      }
      where(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return this;
      }
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return dart.const(new (EmptyIterable$())());
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        dart.throw(IterableElementError.noElement());
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        return initialValue;
      }
      skip(count) {
        core.RangeError.checkNotNegative(count, "count");
        return this;
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return this;
      }
      take(count) {
        core.RangeError.checkNotNegative(count, "count");
        return this;
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return this;
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        return dart.notNull(growable) ? dart.list([], E) : core.List$(E).new(0);
      }
      toSet() {
        return core.Set$(E).new();
      }
    }
    EmptyIterable[dart.implements] = () => [EfficientLength];
    dart.setSignature(EmptyIterable, {
      constructors: () => ({EmptyIterable: [EmptyIterable$(E), []]}),
      methods: () => ({
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        elementAt: [E, [core.int]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        reduce: [E, [dart.functionType(E, [E, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        toList: [core.List$(E), [], {growable: core.bool}],
        toSet: [core.Set$(E), []]
      })
    });
    dart.defineExtensionMembers(EmptyIterable, [
      'forEach',
      'elementAt',
      'contains',
      'every',
      'any',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'join',
      'where',
      'map',
      'reduce',
      'fold',
      'skip',
      'skipWhile',
      'take',
      'takeWhile',
      'toList',
      'toSet',
      'iterator',
      'isEmpty',
      'length',
      'first',
      'last',
      'single'
    ]);
    return EmptyIterable;
  });
  let EmptyIterable = EmptyIterable$();
  const EmptyIterator$ = dart.generic(function(E) {
    class EmptyIterator extends core.Object {
      EmptyIterator() {
      }
      moveNext() {
        return false;
      }
      get current() {
        return null;
      }
    }
    EmptyIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(EmptyIterator, {
      constructors: () => ({EmptyIterator: [EmptyIterator$(E), []]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return EmptyIterator;
  });
  let EmptyIterator = EmptyIterator$();
  const BidirectionalIterator$ = dart.generic(function(T) {
    class BidirectionalIterator extends core.Object {}
    BidirectionalIterator[dart.implements] = () => [core.Iterator$(T)];
    return BidirectionalIterator;
  });
  let BidirectionalIterator = BidirectionalIterator$();
  const IterableMixinWorkaround$ = dart.generic(function(T) {
    class IterableMixinWorkaround extends core.Object {
      static contains(iterable, element) {
        for (let e of iterable) {
          if (dart.equals(e, element))
            return true;
        }
        return false;
      }
      static forEach(iterable, f) {
        dart.as(f, dart.functionType(dart.void, [dart.dynamic]));
        for (let e of iterable) {
          dart.dcall(f, e);
        }
      }
      static any(iterable, f) {
        dart.as(f, dart.functionType(core.bool, [dart.dynamic]));
        for (let e of iterable) {
          if (dart.notNull(dart.dcall(f, e)))
            return true;
        }
        return false;
      }
      static every(iterable, f) {
        dart.as(f, dart.functionType(core.bool, [dart.dynamic]));
        for (let e of iterable) {
          if (!dart.notNull(dart.dcall(f, e)))
            return false;
        }
        return true;
      }
      static reduce(iterable, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
        let iterator = iterable[dartx.iterator];
        if (!dart.notNull(iterator.moveNext()))
          dart.throw(IterableElementError.noElement());
        let value = iterator.current;
        while (dart.notNull(iterator.moveNext())) {
          value = dart.dcall(combine, value, iterator.current);
        }
        return value;
      }
      static fold(iterable, initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
        for (let element of iterable) {
          initialValue = dart.dcall(combine, initialValue, element);
        }
        return initialValue;
      }
      static removeWhereList(list, test) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        let retained = [];
        let length = list[dartx.length];
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = list[dartx.get](i);
          if (!dart.notNull(dart.dcall(test, element))) {
            retained[dartx.add](element);
          }
          if (length != list[dartx.length]) {
            dart.throw(new core.ConcurrentModificationError(list));
          }
        }
        if (retained[dartx.length] == length)
          return;
        list[dartx.length] = retained[dartx.length];
        for (let i = 0; dart.notNull(i) < dart.notNull(retained[dartx.length]); i = dart.notNull(i) + 1) {
          list[dartx.set](i, retained[dartx.get](i));
        }
      }
      static isEmpty(iterable) {
        return !dart.notNull(iterable[dartx.iterator].moveNext());
      }
      static first(iterable) {
        let it = iterable[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(IterableElementError.noElement());
        }
        return it.current;
      }
      static last(iterable) {
        let it = iterable[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(IterableElementError.noElement());
        }
        let result = null;
        do {
          result = it.current;
        } while (dart.notNull(it.moveNext()));
        return result;
      }
      static single(iterable) {
        let it = iterable[dartx.iterator];
        if (!dart.notNull(it.moveNext()))
          dart.throw(IterableElementError.noElement());
        let result = it.current;
        if (dart.notNull(it.moveNext()))
          dart.throw(IterableElementError.tooMany());
        return result;
      }
      static firstWhere(iterable, test, orElse) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        dart.as(orElse, dart.functionType(dart.dynamic, []));
        for (let element of iterable) {
          if (dart.notNull(dart.dcall(test, element)))
            return element;
        }
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      static lastWhere(iterable, test, orElse) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        dart.as(orElse, dart.functionType(dart.dynamic, []));
        let result = null;
        let foundMatching = false;
        for (let element of iterable) {
          if (dart.notNull(dart.dcall(test, element))) {
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      static lastWhereList(list, test, orElse) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        dart.as(orElse, dart.functionType(dart.dynamic, []));
        for (let i = dart.notNull(list[dartx.length]) - 1; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
          let element = list[dartx.get](i);
          if (dart.notNull(dart.dcall(test, element)))
            return element;
        }
        if (orElse != null)
          return orElse();
        dart.throw(IterableElementError.noElement());
      }
      static singleWhere(iterable, test) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        let result = null;
        let foundMatching = false;
        for (let element of iterable) {
          if (dart.notNull(dart.dcall(test, element))) {
            if (dart.notNull(foundMatching)) {
              dart.throw(IterableElementError.tooMany());
            }
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        dart.throw(IterableElementError.noElement());
      }
      static elementAt(iterable, index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError.notNull("index"));
        core.RangeError.checkNotNegative(index, "index");
        let elementIndex = 0;
        for (let element of iterable) {
          if (index == elementIndex)
            return element;
          elementIndex = dart.notNull(elementIndex) + 1;
        }
        dart.throw(core.RangeError.index(index, iterable, "index", null, elementIndex));
      }
      static join(iterable, separator) {
        if (separator === void 0)
          separator = null;
        let buffer = new core.StringBuffer();
        buffer.writeAll(iterable, separator);
        return dart.toString(buffer);
      }
      static joinList(list, separator) {
        if (separator === void 0)
          separator = null;
        if (dart.notNull(list[dartx.isEmpty]))
          return "";
        if (list[dartx.length] == 1)
          return `${list[dartx.get](0)}`;
        let buffer = new core.StringBuffer();
        if (dart.notNull(separator[dartx.isEmpty])) {
          for (let i = 0; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
            buffer.write(list[dartx.get](i));
          }
        } else {
          buffer.write(list[dartx.get](0));
          for (let i = 1; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
            buffer.write(separator);
            buffer.write(list[dartx.get](i));
          }
        }
        return dart.toString(buffer);
      }
      where(iterable, f) {
        dart.as(f, dart.functionType(core.bool, [dart.dynamic]));
        return new (WhereIterable$(T))(dart.as(iterable, core.Iterable$(T)), dart.as(f, __CastType2$(T)));
      }
      static map(iterable, f) {
        dart.as(f, dart.functionType(dart.dynamic, [dart.dynamic]));
        return MappedIterable.new(iterable, f);
      }
      static mapList(list, f) {
        dart.as(f, dart.functionType(dart.dynamic, [dart.dynamic]));
        return new MappedListIterable(list, f);
      }
      static expand(iterable, f) {
        dart.as(f, dart.functionType(core.Iterable, [dart.dynamic]));
        return new ExpandIterable(iterable, f);
      }
      takeList(list, n) {
        return new (SubListIterable$(T))(dart.as(list, core.Iterable$(T)), 0, n);
      }
      takeWhile(iterable, test) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        return new (TakeWhileIterable$(T))(dart.as(iterable, core.Iterable$(T)), dart.as(test, dart.functionType(core.bool, [T])));
      }
      skipList(list, n) {
        return new (SubListIterable$(T))(dart.as(list, core.Iterable$(T)), n, null);
      }
      skipWhile(iterable, test) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        return new (SkipWhileIterable$(T))(dart.as(iterable, core.Iterable$(T)), dart.as(test, dart.functionType(core.bool, [T])));
      }
      reversedList(list) {
        return new (ReversedListIterable$(T))(dart.as(list, core.Iterable$(T)));
      }
      static sortList(list, compare) {
        dart.as(compare, dart.functionType(core.int, [dart.dynamic, dart.dynamic]));
        if (compare == null)
          compare = core.Comparable.compare;
        Sort.sort(list, compare);
      }
      static shuffleList(list, random) {
        if (random == null)
          random = math.Random.new();
        let length = list[dartx.length];
        while (dart.notNull(length) > 1) {
          let pos = random.nextInt(length);
          length = dart.notNull(length) - 1;
          let tmp = list[dartx.get](length);
          list[dartx.set](length, list[dartx.get](pos));
          list[dartx.set](pos, tmp);
        }
      }
      static indexOfList(list, element, start) {
        return Lists.indexOf(list, element, start, list[dartx.length]);
      }
      static lastIndexOfList(list, element, start) {
        if (start == null)
          start = dart.notNull(list[dartx.length]) - 1;
        return Lists.lastIndexOf(list, element, start);
      }
      static _rangeCheck(list, start, end) {
        core.RangeError.checkValidRange(start, end, list[dartx.length]);
      }
      getRangeList(list, start, end) {
        IterableMixinWorkaround$()._rangeCheck(list, start, end);
        return new (SubListIterable$(T))(dart.as(list, core.Iterable$(T)), start, end);
      }
      static setRangeList(list, start, end, from, skipCount) {
        IterableMixinWorkaround$()._rangeCheck(list, start, end);
        let length = dart.notNull(end) - dart.notNull(start);
        if (length == 0)
          return;
        if (dart.notNull(skipCount) < 0)
          dart.throw(new core.ArgumentError(skipCount));
        let otherList = null;
        let otherStart = null;
        if (dart.is(from, core.List)) {
          otherList = from;
          otherStart = skipCount;
        } else {
          otherList = from[dartx.skip](skipCount)[dartx.toList]({growable: false});
          otherStart = 0;
        }
        if (dart.notNull(otherStart) + dart.notNull(length) > dart.notNull(otherList[dartx.length])) {
          dart.throw(IterableElementError.tooFew());
        }
        Lists.copy(otherList, otherStart, list, start, length);
      }
      static replaceRangeList(list, start, end, iterable) {
        IterableMixinWorkaround$()._rangeCheck(list, start, end);
        if (!dart.is(iterable, EfficientLength)) {
          iterable = iterable[dartx.toList]();
        }
        let removeLength = dart.notNull(end) - dart.notNull(start);
        let insertLength = iterable[dartx.length];
        if (dart.notNull(removeLength) >= dart.notNull(insertLength)) {
          let delta = dart.notNull(removeLength) - dart.notNull(insertLength);
          let insertEnd = dart.notNull(start) + dart.notNull(insertLength);
          let newEnd = dart.notNull(list[dartx.length]) - dart.notNull(delta);
          list[dartx.setRange](start, insertEnd, iterable);
          if (delta != 0) {
            list[dartx.setRange](insertEnd, newEnd, list, end);
            list[dartx.length] = newEnd;
          }
        } else {
          let delta = dart.notNull(insertLength) - dart.notNull(removeLength);
          let newLength = dart.notNull(list[dartx.length]) + dart.notNull(delta);
          let insertEnd = dart.notNull(start) + dart.notNull(insertLength);
          list[dartx.length] = newLength;
          list[dartx.setRange](insertEnd, newLength, list, end);
          list[dartx.setRange](start, insertEnd, iterable);
        }
      }
      static fillRangeList(list, start, end, fillValue) {
        IterableMixinWorkaround$()._rangeCheck(list, start, end);
        for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
          list[dartx.set](i, fillValue);
        }
      }
      static insertAllList(list, index, iterable) {
        core.RangeError.checkValueInInterval(index, 0, list[dartx.length], "index");
        if (!dart.is(iterable, EfficientLength)) {
          iterable = iterable[dartx.toList]({growable: false});
        }
        let insertionLength = iterable[dartx.length];
        list[dartx.length] = dart.notNull(list[dartx.length]) + dart.notNull(insertionLength);
        list[dartx.setRange](dart.notNull(index) + dart.notNull(insertionLength), list[dartx.length], list, index);
        for (let element of iterable) {
          list[dartx.set]((() => {
            let x = index;
            index = dart.notNull(x) + 1;
            return x;
          })(), element);
        }
      }
      static setAllList(list, index, iterable) {
        core.RangeError.checkValueInInterval(index, 0, list[dartx.length], "index");
        for (let element of iterable) {
          list[dartx.set]((() => {
            let x = index;
            index = dart.notNull(x) + 1;
            return x;
          })(), element);
        }
      }
      asMapList(l) {
        return new (ListMapView$(T))(dart.as(l, core.List$(T)));
      }
      static setContainsAll(set, other) {
        for (let element of other) {
          if (!dart.notNull(set.contains(element)))
            return false;
        }
        return true;
      }
      static setIntersection(set, other, result) {
        let smaller = null;
        let larger = null;
        if (dart.notNull(set.length) < dart.notNull(other.length)) {
          smaller = set;
          larger = other;
        } else {
          smaller = other;
          larger = set;
        }
        for (let element of smaller) {
          if (dart.notNull(larger.contains(element))) {
            result.add(element);
          }
        }
        return result;
      }
      static setUnion(set, other, result) {
        result.addAll(set);
        result.addAll(other);
        return result;
      }
      static setDifference(set, other, result) {
        for (let element of set) {
          if (!dart.notNull(other.contains(element))) {
            result.add(element);
          }
        }
        return result;
      }
    }
    dart.setSignature(IterableMixinWorkaround, {
      methods: () => ({
        where: [core.Iterable$(T), [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        takeList: [core.Iterable$(T), [core.List, core.int]],
        takeWhile: [core.Iterable$(T), [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        skipList: [core.Iterable$(T), [core.List, core.int]],
        skipWhile: [core.Iterable$(T), [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        reversedList: [core.Iterable$(T), [core.List]],
        getRangeList: [core.Iterable$(T), [core.List, core.int, core.int]],
        asMapList: [core.Map$(core.int, T), [core.List]]
      }),
      statics: () => ({
        contains: [core.bool, [core.Iterable, dart.dynamic]],
        forEach: [dart.void, [core.Iterable, dart.functionType(dart.void, [dart.dynamic])]],
        any: [core.bool, [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        every: [core.bool, [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        reduce: [dart.dynamic, [core.Iterable, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]],
        fold: [dart.dynamic, [core.Iterable, dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]],
        removeWhereList: [dart.void, [core.List, dart.functionType(core.bool, [dart.dynamic])]],
        isEmpty: [core.bool, [core.Iterable]],
        first: [dart.dynamic, [core.Iterable]],
        last: [dart.dynamic, [core.Iterable]],
        single: [dart.dynamic, [core.Iterable]],
        firstWhere: [dart.dynamic, [core.Iterable, dart.functionType(core.bool, [dart.dynamic]), dart.functionType(dart.dynamic, [])]],
        lastWhere: [dart.dynamic, [core.Iterable, dart.functionType(core.bool, [dart.dynamic]), dart.functionType(dart.dynamic, [])]],
        lastWhereList: [dart.dynamic, [core.List, dart.functionType(core.bool, [dart.dynamic]), dart.functionType(dart.dynamic, [])]],
        singleWhere: [dart.dynamic, [core.Iterable, dart.functionType(core.bool, [dart.dynamic])]],
        elementAt: [dart.dynamic, [core.Iterable, core.int]],
        join: [core.String, [core.Iterable], [core.String]],
        joinList: [core.String, [core.List], [core.String]],
        map: [core.Iterable, [core.Iterable, dart.functionType(dart.dynamic, [dart.dynamic])]],
        mapList: [core.Iterable, [core.List, dart.functionType(dart.dynamic, [dart.dynamic])]],
        expand: [core.Iterable, [core.Iterable, dart.functionType(core.Iterable, [dart.dynamic])]],
        sortList: [dart.void, [core.List, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]],
        shuffleList: [dart.void, [core.List, math.Random]],
        indexOfList: [core.int, [core.List, dart.dynamic, core.int]],
        lastIndexOfList: [core.int, [core.List, dart.dynamic, core.int]],
        _rangeCheck: [dart.void, [core.List, core.int, core.int]],
        setRangeList: [dart.void, [core.List, core.int, core.int, core.Iterable, core.int]],
        replaceRangeList: [dart.void, [core.List, core.int, core.int, core.Iterable]],
        fillRangeList: [dart.void, [core.List, core.int, core.int, dart.dynamic]],
        insertAllList: [dart.void, [core.List, core.int, core.Iterable]],
        setAllList: [dart.void, [core.List, core.int, core.Iterable]],
        setContainsAll: [core.bool, [core.Set, core.Iterable]],
        setIntersection: [core.Set, [core.Set, core.Set, core.Set]],
        setUnion: [core.Set, [core.Set, core.Set, core.Set]],
        setDifference: [core.Set, [core.Set, core.Set, core.Set]]
      }),
      names: ['contains', 'forEach', 'any', 'every', 'reduce', 'fold', 'removeWhereList', 'isEmpty', 'first', 'last', 'single', 'firstWhere', 'lastWhere', 'lastWhereList', 'singleWhere', 'elementAt', 'join', 'joinList', 'map', 'mapList', 'expand', 'sortList', 'shuffleList', 'indexOfList', 'lastIndexOfList', '_rangeCheck', 'setRangeList', 'replaceRangeList', 'fillRangeList', 'insertAllList', 'setAllList', 'setContainsAll', 'setIntersection', 'setUnion', 'setDifference']
    });
    return IterableMixinWorkaround;
  });
  let IterableMixinWorkaround = IterableMixinWorkaround$();
  class IterableElementError extends core.Object {
    static noElement() {
      return new core.StateError("No element");
    }
    static tooMany() {
      return new core.StateError("Too many elements");
    }
    static tooFew() {
      return new core.StateError("Too few elements");
    }
  }
  dart.setSignature(IterableElementError, {
    statics: () => ({
      noElement: [core.StateError, []],
      tooMany: [core.StateError, []],
      tooFew: [core.StateError, []]
    }),
    names: ['noElement', 'tooMany', 'tooFew']
  });
  const __CastType0$ = dart.generic(function(S, T) {
    const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(core.Iterable$(T), [S]));
    return __CastType0;
  });
  let __CastType0 = __CastType0$();
  const __CastType2$ = dart.generic(function(T) {
    const __CastType2 = dart.typedef('__CastType2', () => dart.functionType(core.bool, [T]));
    return __CastType2;
  });
  let __CastType2 = __CastType2$();
  const FixedLengthListMixin$ = dart.generic(function(E) {
    class FixedLengthListMixin extends core.Object {
      set length(newLength) {
        dart.throw(new core.UnsupportedError("Cannot change the length of a fixed-length list"));
      }
      add(value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot add to a fixed-length list"));
      }
      insert(index, value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot add to a fixed-length list"));
      }
      insertAll(at, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to a fixed-length list"));
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to a fixed-length list"));
      }
      remove(element) {
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      clear() {
        dart.throw(new core.UnsupportedError("Cannot clear a fixed-length list"));
      }
      removeAt(index) {
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      removeLast() {
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      removeRange(start, end) {
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
      replaceRange(start, end, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot remove from a fixed-length list"));
      }
    }
    dart.setSignature(FixedLengthListMixin, {
      methods: () => ({
        add: [dart.void, [E]],
        insert: [dart.void, [core.int, E]],
        insertAll: [dart.void, [core.int, core.Iterable$(E)]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [core.Object]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        clear: [dart.void, []],
        removeAt: [E, [core.int]],
        removeLast: [E, []],
        removeRange: [dart.void, [core.int, core.int]],
        replaceRange: [dart.void, [core.int, core.int, core.Iterable$(E)]]
      })
    });
    return FixedLengthListMixin;
  });
  let FixedLengthListMixin = FixedLengthListMixin$();
  const UnmodifiableListMixin$ = dart.generic(function(E) {
    class UnmodifiableListMixin extends core.Object {
      set(index, value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
        return value;
      }
      set length(newLength) {
        dart.throw(new core.UnsupportedError("Cannot change the length of an unmodifiable list"));
      }
      setAll(at, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
      }
      add(value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot add to an unmodifiable list"));
      }
      insert(index, value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot add to an unmodifiable list"));
      }
      insertAll(at, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to an unmodifiable list"));
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to an unmodifiable list"));
      }
      remove(element) {
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      sort(compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, core.Comparator$(E));
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
      }
      shuffle(random) {
        if (random === void 0)
          random = null;
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
      }
      clear() {
        dart.throw(new core.UnsupportedError("Cannot clear an unmodifiable list"));
      }
      removeAt(index) {
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      removeLast() {
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      setRange(start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
      }
      removeRange(start, end) {
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      replaceRange(start, end, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot remove from an unmodifiable list"));
      }
      fillRange(start, end, fillValue) {
        if (fillValue === void 0)
          fillValue = null;
        dart.as(fillValue, E);
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable list"));
      }
    }
    UnmodifiableListMixin[dart.implements] = () => [core.List$(E)];
    dart.setSignature(UnmodifiableListMixin, {
      methods: () => ({
        set: [dart.void, [core.int, E]],
        setAll: [dart.void, [core.int, core.Iterable$(E)]],
        add: [dart.void, [E]],
        insert: [E, [core.int, E]],
        insertAll: [dart.void, [core.int, core.Iterable$(E)]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [core.Object]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        sort: [dart.void, [], [core.Comparator$(E)]],
        shuffle: [dart.void, [], [math.Random]],
        clear: [dart.void, []],
        removeAt: [E, [core.int]],
        removeLast: [E, []],
        setRange: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        removeRange: [dart.void, [core.int, core.int]],
        replaceRange: [dart.void, [core.int, core.int, core.Iterable$(E)]],
        fillRange: [dart.void, [core.int, core.int], [E]]
      })
    });
    dart.defineExtensionMembers(UnmodifiableListMixin, [
      'set',
      'setAll',
      'add',
      'insert',
      'insertAll',
      'addAll',
      'remove',
      'removeWhere',
      'retainWhere',
      'sort',
      'shuffle',
      'clear',
      'removeAt',
      'removeLast',
      'setRange',
      'removeRange',
      'replaceRange',
      'fillRange',
      'length'
    ]);
    return UnmodifiableListMixin;
  });
  let UnmodifiableListMixin = UnmodifiableListMixin$();
  const FixedLengthListBase$ = dart.generic(function(E) {
    class FixedLengthListBase extends dart.mixin(collection.ListBase$(E), FixedLengthListMixin$(E)) {
      FixedLengthListBase() {
        super.ListBase(...arguments);
      }
    }
    return FixedLengthListBase;
  });
  let FixedLengthListBase = FixedLengthListBase$();
  const UnmodifiableListBase$ = dart.generic(function(E) {
    class UnmodifiableListBase extends dart.mixin(collection.ListBase$(E), UnmodifiableListMixin$(E)) {
      UnmodifiableListBase() {
        super.ListBase(...arguments);
      }
    }
    return UnmodifiableListBase;
  });
  let UnmodifiableListBase = UnmodifiableListBase$();
  const _backedList = dart.JsSymbol('_backedList');
  class _ListIndicesIterable extends ListIterable$(core.int) {
    _ListIndicesIterable(backedList) {
      this[_backedList] = backedList;
      super.ListIterable();
    }
    get length() {
      return this[_backedList][dartx.length];
    }
    elementAt(index) {
      core.RangeError.checkValidIndex(index, this);
      return index;
    }
  }
  dart.setSignature(_ListIndicesIterable, {
    constructors: () => ({_ListIndicesIterable: [_ListIndicesIterable, [core.List]]}),
    methods: () => ({elementAt: [core.int, [core.int]]})
  });
  dart.defineExtensionMembers(_ListIndicesIterable, ['elementAt', 'length']);
  const _values = dart.JsSymbol('_values');
  const ListMapView$ = dart.generic(function(E) {
    class ListMapView extends core.Object {
      ListMapView(values) {
        this[_values] = values;
      }
      get(key) {
        return dart.notNull(this.containsKey(key)) ? this[_values][dartx.get](dart.as(key, core.int)) : null;
      }
      get length() {
        return this[_values][dartx.length];
      }
      get values() {
        return new (SubListIterable$(E))(this[_values], 0, null);
      }
      get keys() {
        return new _ListIndicesIterable(this[_values]);
      }
      get isEmpty() {
        return this[_values][dartx.isEmpty];
      }
      get isNotEmpty() {
        return this[_values][dartx.isNotEmpty];
      }
      containsValue(value) {
        return this[_values][dartx.contains](value);
      }
      containsKey(key) {
        return typeof key == 'number' && dart.notNull(key) >= 0 && dart.notNull(key) < dart.notNull(this.length);
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [core.int, E]));
        let length = this[_values][dartx.length];
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          f(i, this[_values][dartx.get](i));
          if (length != this[_values][dartx.length]) {
            dart.throw(new core.ConcurrentModificationError(this[_values]));
          }
        }
      }
      set(key, value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable map"));
        return value;
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(ifAbsent, dart.functionType(E, []));
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable map"));
      }
      remove(key) {
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable map"));
      }
      clear() {
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable map"));
      }
      addAll(other) {
        dart.as(other, core.Map$(core.int, E));
        dart.throw(new core.UnsupportedError("Cannot modify an unmodifiable map"));
      }
      toString() {
        return collection.Maps.mapToString(this);
      }
    }
    ListMapView[dart.implements] = () => [core.Map$(core.int, E)];
    dart.setSignature(ListMapView, {
      constructors: () => ({ListMapView: [ListMapView$(E), [core.List$(E)]]}),
      methods: () => ({
        get: [E, [core.Object]],
        containsValue: [core.bool, [core.Object]],
        containsKey: [core.bool, [core.Object]],
        forEach: [dart.void, [dart.functionType(dart.void, [core.int, E])]],
        set: [dart.void, [core.int, E]],
        putIfAbsent: [E, [core.int, dart.functionType(E, [])]],
        remove: [E, [core.Object]],
        clear: [dart.void, []],
        addAll: [dart.void, [core.Map$(core.int, E)]]
      })
    });
    return ListMapView;
  });
  let ListMapView = ListMapView$();
  const ReversedListIterable$ = dart.generic(function(E) {
    class ReversedListIterable extends ListIterable$(E) {
      ReversedListIterable(source) {
        this[_source] = source;
        super.ListIterable();
      }
      get length() {
        return this[_source][dartx.length];
      }
      elementAt(index) {
        return this[_source][dartx.elementAt](dart.notNull(this[_source][dartx.length]) - 1 - dart.notNull(index));
      }
    }
    dart.setSignature(ReversedListIterable, {
      constructors: () => ({ReversedListIterable: [ReversedListIterable$(E), [core.Iterable$(E)]]}),
      methods: () => ({elementAt: [E, [core.int]]})
    });
    dart.defineExtensionMembers(ReversedListIterable, ['elementAt', 'length']);
    return ReversedListIterable;
  });
  let ReversedListIterable = ReversedListIterable$();
  class UnmodifiableListError extends core.Object {
    static add() {
      return new core.UnsupportedError("Cannot add to unmodifiable List");
    }
    static change() {
      return new core.UnsupportedError("Cannot change the content of an unmodifiable List");
    }
    static length() {
      return new core.UnsupportedError("Cannot change length of unmodifiable List");
    }
    static remove() {
      return new core.UnsupportedError("Cannot remove from unmodifiable List");
    }
  }
  dart.setSignature(UnmodifiableListError, {
    statics: () => ({
      add: [core.UnsupportedError, []],
      change: [core.UnsupportedError, []],
      length: [core.UnsupportedError, []],
      remove: [core.UnsupportedError, []]
    }),
    names: ['add', 'change', 'length', 'remove']
  });
  class NonGrowableListError extends core.Object {
    static add() {
      return new core.UnsupportedError("Cannot add to non-growable List");
    }
    static length() {
      return new core.UnsupportedError("Cannot change length of non-growable List");
    }
    static remove() {
      return new core.UnsupportedError("Cannot remove from non-growable List");
    }
  }
  dart.setSignature(NonGrowableListError, {
    statics: () => ({
      add: [core.UnsupportedError, []],
      length: [core.UnsupportedError, []],
      remove: [core.UnsupportedError, []]
    }),
    names: ['add', 'length', 'remove']
  });
  function makeListFixedLength(growableList) {
    _interceptors.JSArray.markFixedList(growableList);
    return growableList;
  }
  dart.fn(makeListFixedLength, core.List, [core.List]);
  class Lists extends core.Object {
    static copy(src, srcStart, dst, dstStart, count) {
      if (dart.notNull(srcStart) < dart.notNull(dstStart)) {
        for (let i = dart.notNull(srcStart) + dart.notNull(count) - 1, j = dart.notNull(dstStart) + dart.notNull(count) - 1; dart.notNull(i) >= dart.notNull(srcStart); i = dart.notNull(i) - 1, j = dart.notNull(j) - 1) {
          dst[dartx.set](j, src[dartx.get](i));
        }
      } else {
        for (let i = srcStart, j = dstStart; dart.notNull(i) < dart.notNull(srcStart) + dart.notNull(count); i = dart.notNull(i) + 1, j = dart.notNull(j) + 1) {
          dst[dartx.set](j, src[dartx.get](i));
        }
      }
    }
    static areEqual(a, b) {
      if (dart.notNull(core.identical(a, b)))
        return true;
      if (!dart.is(b, core.List))
        return false;
      let length = a[dartx.length];
      if (!dart.equals(length, dart.dload(b, 'length')))
        return false;
      for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        if (!dart.notNull(core.identical(a[dartx.get](i), dart.dindex(b, i))))
          return false;
      }
      return true;
    }
    static indexOf(a, element, startIndex, endIndex) {
      if (dart.notNull(startIndex) >= dart.notNull(a[dartx.length])) {
        return -1;
      }
      if (dart.notNull(startIndex) < 0) {
        startIndex = 0;
      }
      for (let i = startIndex; dart.notNull(i) < dart.notNull(endIndex); i = dart.notNull(i) + 1) {
        if (dart.equals(a[dartx.get](i), element)) {
          return i;
        }
      }
      return -1;
    }
    static lastIndexOf(a, element, startIndex) {
      if (dart.notNull(startIndex) < 0) {
        return -1;
      }
      if (dart.notNull(startIndex) >= dart.notNull(a[dartx.length])) {
        startIndex = dart.notNull(a[dartx.length]) - 1;
      }
      for (let i = startIndex; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
        if (dart.equals(a[dartx.get](i), element)) {
          return i;
        }
      }
      return -1;
    }
    static indicesCheck(a, start, end) {
      core.RangeError.checkValidRange(start, end, a[dartx.length]);
    }
    static rangeCheck(a, start, length) {
      core.RangeError.checkNotNegative(length);
      core.RangeError.checkNotNegative(start);
      if (dart.notNull(start) + dart.notNull(length) > dart.notNull(a[dartx.length])) {
        let message = `${start} + ${length} must be in the range [0..${a[dartx.length]}]`;
        dart.throw(new core.RangeError.range(length, 0, dart.notNull(a[dartx.length]) - dart.notNull(start), "length", message));
      }
    }
  }
  dart.setSignature(Lists, {
    statics: () => ({
      copy: [dart.void, [core.List, core.int, core.List, core.int, core.int]],
      areEqual: [core.bool, [core.List, dart.dynamic]],
      indexOf: [core.int, [core.List, core.Object, core.int, core.int]],
      lastIndexOf: [core.int, [core.List, core.Object, core.int]],
      indicesCheck: [dart.void, [core.List, core.int, core.int]],
      rangeCheck: [dart.void, [core.List, core.int, core.int]]
    }),
    names: ['copy', 'areEqual', 'indexOf', 'lastIndexOf', 'indicesCheck', 'rangeCheck']
  });
  exports.printToZone = null;
  function printToConsole(line) {
    _js_primitives.printString(`${line}`);
  }
  dart.fn(printToConsole, dart.void, [core.String]);
  class Sort extends core.Object {
    static sort(a, compare) {
      Sort._doSort(a, 0, dart.notNull(a[dartx.length]) - 1, compare);
    }
    static sortRange(a, from, to, compare) {
      if (dart.notNull(from) < 0 || dart.notNull(to) > dart.notNull(a[dartx.length]) || dart.notNull(to) < dart.notNull(from)) {
        dart.throw("OutOfRange");
      }
      Sort._doSort(a, from, dart.notNull(to) - 1, compare);
    }
    static _doSort(a, left, right, compare) {
      if (dart.notNull(right) - dart.notNull(left) <= dart.notNull(Sort._INSERTION_SORT_THRESHOLD)) {
        Sort._insertionSort(a, left, right, compare);
      } else {
        Sort._dualPivotQuicksort(a, left, right, compare);
      }
    }
    static _insertionSort(a, left, right, compare) {
      for (let i = dart.notNull(left) + 1; dart.notNull(i) <= dart.notNull(right); i = dart.notNull(i) + 1) {
        let el = a[dartx.get](i);
        let j = i;
        while (dart.notNull(j) > dart.notNull(left) && dart.notNull(dart.dcall(compare, a[dartx.get](dart.notNull(j) - 1), el)) > 0) {
          a[dartx.set](j, a[dartx.get](dart.notNull(j) - 1));
          j = dart.notNull(j) - 1;
        }
        a[dartx.set](j, el);
      }
    }
    static _dualPivotQuicksort(a, left, right, compare) {
      dart.assert(dart.notNull(right) - dart.notNull(left) > dart.notNull(Sort._INSERTION_SORT_THRESHOLD));
      let sixth = ((dart.notNull(right) - dart.notNull(left) + 1) / 6)[dartx.truncate]();
      let index1 = dart.notNull(left) + dart.notNull(sixth);
      let index5 = dart.notNull(right) - dart.notNull(sixth);
      let index3 = ((dart.notNull(left) + dart.notNull(right)) / 2)[dartx.truncate]();
      let index2 = dart.notNull(index3) - dart.notNull(sixth);
      let index4 = dart.notNull(index3) + dart.notNull(sixth);
      let el1 = a[dartx.get](index1);
      let el2 = a[dartx.get](index2);
      let el3 = a[dartx.get](index3);
      let el4 = a[dartx.get](index4);
      let el5 = a[dartx.get](index5);
      if (dart.notNull(dart.dcall(compare, el1, el2)) > 0) {
        let t = el1;
        el1 = el2;
        el2 = t;
      }
      if (dart.notNull(dart.dcall(compare, el4, el5)) > 0) {
        let t = el4;
        el4 = el5;
        el5 = t;
      }
      if (dart.notNull(dart.dcall(compare, el1, el3)) > 0) {
        let t = el1;
        el1 = el3;
        el3 = t;
      }
      if (dart.notNull(dart.dcall(compare, el2, el3)) > 0) {
        let t = el2;
        el2 = el3;
        el3 = t;
      }
      if (dart.notNull(dart.dcall(compare, el1, el4)) > 0) {
        let t = el1;
        el1 = el4;
        el4 = t;
      }
      if (dart.notNull(dart.dcall(compare, el3, el4)) > 0) {
        let t = el3;
        el3 = el4;
        el4 = t;
      }
      if (dart.notNull(dart.dcall(compare, el2, el5)) > 0) {
        let t = el2;
        el2 = el5;
        el5 = t;
      }
      if (dart.notNull(dart.dcall(compare, el2, el3)) > 0) {
        let t = el2;
        el2 = el3;
        el3 = t;
      }
      if (dart.notNull(dart.dcall(compare, el4, el5)) > 0) {
        let t = el4;
        el4 = el5;
        el5 = t;
      }
      let pivot1 = el2;
      let pivot2 = el4;
      a[dartx.set](index1, el1);
      a[dartx.set](index3, el3);
      a[dartx.set](index5, el5);
      a[dartx.set](index2, a[dartx.get](left));
      a[dartx.set](index4, a[dartx.get](right));
      let less = dart.notNull(left) + 1;
      let great = dart.notNull(right) - 1;
      let pivots_are_equal = dart.dcall(compare, pivot1, pivot2) == 0;
      if (dart.notNull(pivots_are_equal)) {
        let pivot = pivot1;
        for (let k = less; dart.notNull(k) <= dart.notNull(great); k = dart.notNull(k) + 1) {
          let ak = a[dartx.get](k);
          let comp = dart.dcall(compare, ak, pivot);
          if (comp == 0)
            continue;
          if (dart.notNull(comp) < 0) {
            if (k != less) {
              a[dartx.set](k, a[dartx.get](less));
              a[dartx.set](less, ak);
            }
            less = dart.notNull(less) + 1;
          } else {
            while (true) {
              comp = dart.dcall(compare, a[dartx.get](great), pivot);
              if (dart.notNull(comp) > 0) {
                great = dart.notNull(great) - 1;
                continue;
              } else if (dart.notNull(comp) < 0) {
                a[dartx.set](k, a[dartx.get](less));
                a[dartx.set]((() => {
                  let x = less;
                  less = dart.notNull(x) + 1;
                  return x;
                })(), a[dartx.get](great));
                a[dartx.set]((() => {
                  let x = great;
                  great = dart.notNull(x) - 1;
                  return x;
                })(), ak);
                break;
              } else {
                a[dartx.set](k, a[dartx.get](great));
                a[dartx.set]((() => {
                  let x = great;
                  great = dart.notNull(x) - 1;
                  return x;
                })(), ak);
                break;
              }
            }
          }
        }
      } else {
        for (let k = less; dart.notNull(k) <= dart.notNull(great); k = dart.notNull(k) + 1) {
          let ak = a[dartx.get](k);
          let comp_pivot1 = dart.dcall(compare, ak, pivot1);
          if (dart.notNull(comp_pivot1) < 0) {
            if (k != less) {
              a[dartx.set](k, a[dartx.get](less));
              a[dartx.set](less, ak);
            }
            less = dart.notNull(less) + 1;
          } else {
            let comp_pivot2 = dart.dcall(compare, ak, pivot2);
            if (dart.notNull(comp_pivot2) > 0) {
              while (true) {
                let comp = dart.dcall(compare, a[dartx.get](great), pivot2);
                if (dart.notNull(comp) > 0) {
                  great = dart.notNull(great) - 1;
                  if (dart.notNull(great) < dart.notNull(k))
                    break;
                  continue;
                } else {
                  comp = dart.dcall(compare, a[dartx.get](great), pivot1);
                  if (dart.notNull(comp) < 0) {
                    a[dartx.set](k, a[dartx.get](less));
                    a[dartx.set]((() => {
                      let x = less;
                      less = dart.notNull(x) + 1;
                      return x;
                    })(), a[dartx.get](great));
                    a[dartx.set]((() => {
                      let x = great;
                      great = dart.notNull(x) - 1;
                      return x;
                    })(), ak);
                  } else {
                    a[dartx.set](k, a[dartx.get](great));
                    a[dartx.set]((() => {
                      let x = great;
                      great = dart.notNull(x) - 1;
                      return x;
                    })(), ak);
                  }
                  break;
                }
              }
            }
          }
        }
      }
      a[dartx.set](left, a[dartx.get](dart.notNull(less) - 1));
      a[dartx.set](dart.notNull(less) - 1, pivot1);
      a[dartx.set](right, a[dartx.get](dart.notNull(great) + 1));
      a[dartx.set](dart.notNull(great) + 1, pivot2);
      Sort._doSort(a, left, dart.notNull(less) - 2, compare);
      Sort._doSort(a, dart.notNull(great) + 2, right, compare);
      if (dart.notNull(pivots_are_equal)) {
        return;
      }
      if (dart.notNull(less) < dart.notNull(index1) && dart.notNull(great) > dart.notNull(index5)) {
        while (dart.dcall(compare, a[dartx.get](less), pivot1) == 0) {
          less = dart.notNull(less) + 1;
        }
        while (dart.dcall(compare, a[dartx.get](great), pivot2) == 0) {
          great = dart.notNull(great) - 1;
        }
        for (let k = less; dart.notNull(k) <= dart.notNull(great); k = dart.notNull(k) + 1) {
          let ak = a[dartx.get](k);
          let comp_pivot1 = dart.dcall(compare, ak, pivot1);
          if (comp_pivot1 == 0) {
            if (k != less) {
              a[dartx.set](k, a[dartx.get](less));
              a[dartx.set](less, ak);
            }
            less = dart.notNull(less) + 1;
          } else {
            let comp_pivot2 = dart.dcall(compare, ak, pivot2);
            if (comp_pivot2 == 0) {
              while (true) {
                let comp = dart.dcall(compare, a[dartx.get](great), pivot2);
                if (comp == 0) {
                  great = dart.notNull(great) - 1;
                  if (dart.notNull(great) < dart.notNull(k))
                    break;
                  continue;
                } else {
                  comp = dart.dcall(compare, a[dartx.get](great), pivot1);
                  if (dart.notNull(comp) < 0) {
                    a[dartx.set](k, a[dartx.get](less));
                    a[dartx.set]((() => {
                      let x = less;
                      less = dart.notNull(x) + 1;
                      return x;
                    })(), a[dartx.get](great));
                    a[dartx.set]((() => {
                      let x = great;
                      great = dart.notNull(x) - 1;
                      return x;
                    })(), ak);
                  } else {
                    a[dartx.set](k, a[dartx.get](great));
                    a[dartx.set]((() => {
                      let x = great;
                      great = dart.notNull(x) - 1;
                      return x;
                    })(), ak);
                  }
                  break;
                }
              }
            }
          }
        }
        Sort._doSort(a, less, great, compare);
      } else {
        Sort._doSort(a, less, great, compare);
      }
    }
  }
  dart.setSignature(Sort, {
    statics: () => ({
      sort: [dart.void, [core.List, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]],
      sortRange: [dart.void, [core.List, core.int, core.int, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]],
      _doSort: [dart.void, [core.List, core.int, core.int, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]],
      _insertionSort: [dart.void, [core.List, core.int, core.int, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]],
      _dualPivotQuicksort: [dart.void, [core.List, core.int, core.int, dart.functionType(core.int, [dart.dynamic, dart.dynamic])]]
    }),
    names: ['sort', 'sortRange', '_doSort', '_insertionSort', '_dualPivotQuicksort']
  });
  Sort._INSERTION_SORT_THRESHOLD = 32;
  const _name = dart.JsSymbol('_name');
  class Symbol extends core.Object {
    Symbol(name) {
      this[_name] = name;
    }
    unvalidated(name) {
      this[_name] = name;
    }
    validated(name) {
      this[_name] = Symbol.validatePublicSymbol(name);
    }
    ['=='](other) {
      return dart.is(other, Symbol) && this[_name] == other[_name];
    }
    get hashCode() {
      let arbitraryPrime = 664597;
      return 536870911 & dart.notNull(arbitraryPrime) * dart.notNull(dart.hashCode(this[_name]));
    }
    toString() {
      return `Symbol("${this[_name]}")`;
    }
    static getName(symbol) {
      return symbol[_name];
    }
    static validatePublicSymbol(name) {
      if (dart.notNull(name[dartx.isEmpty]) || dart.notNull(Symbol.publicSymbolPattern.hasMatch(name)))
        return name;
      if (dart.notNull(name[dartx.startsWith]('_'))) {
        dart.throw(new core.ArgumentError(`"${name}" is a private identifier`));
      }
      dart.throw(new core.ArgumentError(`"${name}" is not a valid (qualified) symbol name`));
    }
    static isValidSymbol(name) {
      return dart.notNull(name[dartx.isEmpty]) || dart.notNull(Symbol.symbolPattern.hasMatch(name));
    }
  }
  Symbol[dart.implements] = () => [core.Symbol];
  dart.defineNamedConstructor(Symbol, 'unvalidated');
  dart.defineNamedConstructor(Symbol, 'validated');
  dart.setSignature(Symbol, {
    constructors: () => ({
      Symbol: [Symbol, [core.String]],
      unvalidated: [Symbol, [core.String]],
      validated: [Symbol, [core.String]]
    }),
    methods: () => ({'==': [core.bool, [core.Object]]}),
    statics: () => ({
      getName: [core.String, [Symbol]],
      validatePublicSymbol: [core.String, [core.String]],
      isValidSymbol: [core.bool, [core.String]]
    }),
    names: ['getName', 'validatePublicSymbol', 'isValidSymbol']
  });
  Symbol.reservedWordRE = '(?:assert|break|c(?:a(?:se|tch)|lass|on(?:st|tinue))|d(?:efault|o)|' + 'e(?:lse|num|xtends)|f(?:alse|inal(?:ly)?|or)|i[fns]|n(?:ew|ull)|' + 'ret(?:hrow|urn)|s(?:uper|witch)|t(?:h(?:is|row)|r(?:ue|y))|' + 'v(?:ar|oid)|w(?:hile|ith))';
  Symbol.publicIdentifierRE = '(?!' + `${Symbol.reservedWordRE}` + '\\b(?!\\$))[a-zA-Z$][\\w$]*';
  Symbol.identifierRE = '(?!' + `${Symbol.reservedWordRE}` + '\\b(?!\\$))[a-zA-Z$_][\\w$]*';
  Symbol.operatorRE = '(?:[\\-+*/%&|^]|\\[\\]=?|==|~/?|<[<=]?|>[>=]?|unary-)';
  const POWERS_OF_TEN = dart.const([1.0, 10.0, 100.0, 1000.0, 10000.0, 100000.0, 1000000.0, 10000000.0, 100000000.0, 1000000000.0, 10000000000.0, 100000000000.0, 1000000000000.0, 10000000000000.0, 100000000000000.0, 1000000000000000.0, 10000000000000000.0, 100000000000000000.0, 1000000000000000000.0, 10000000000000000000.0, 100000000000000000000.0, 1e+21, 1e+22]);
  dart.defineLazyProperties(Symbol, {
    get publicSymbolPattern() {
      return core.RegExp.new(`^(?:${Symbol.operatorRE}\$|${Symbol.publicIdentifierRE}(?:=?\$|[.](?!\$)))+?\$`);
    },
    get symbolPattern() {
      return core.RegExp.new(`^(?:${Symbol.operatorRE}\$|${Symbol.identifierRE}(?:=?\$|[.](?!\$)))+?\$`);
    }
  });
  // Exports:
  exports.EfficientLength = EfficientLength;
  exports.ListIterable$ = ListIterable$;
  exports.ListIterable = ListIterable;
  exports.SubListIterable$ = SubListIterable$;
  exports.SubListIterable = SubListIterable;
  exports.ListIterator$ = ListIterator$;
  exports.ListIterator = ListIterator;
  exports.MappedIterable$ = MappedIterable$;
  exports.MappedIterable = MappedIterable;
  exports.EfficientLengthMappedIterable$ = EfficientLengthMappedIterable$;
  exports.EfficientLengthMappedIterable = EfficientLengthMappedIterable;
  exports.MappedIterator$ = MappedIterator$;
  exports.MappedIterator = MappedIterator;
  exports.MappedListIterable$ = MappedListIterable$;
  exports.MappedListIterable = MappedListIterable;
  exports.WhereIterable$ = WhereIterable$;
  exports.WhereIterable = WhereIterable;
  exports.WhereIterator$ = WhereIterator$;
  exports.WhereIterator = WhereIterator;
  exports.ExpandIterable$ = ExpandIterable$;
  exports.ExpandIterable = ExpandIterable;
  exports.ExpandIterator$ = ExpandIterator$;
  exports.ExpandIterator = ExpandIterator;
  exports.TakeIterable$ = TakeIterable$;
  exports.TakeIterable = TakeIterable;
  exports.EfficientLengthTakeIterable$ = EfficientLengthTakeIterable$;
  exports.EfficientLengthTakeIterable = EfficientLengthTakeIterable;
  exports.TakeIterator$ = TakeIterator$;
  exports.TakeIterator = TakeIterator;
  exports.TakeWhileIterable$ = TakeWhileIterable$;
  exports.TakeWhileIterable = TakeWhileIterable;
  exports.TakeWhileIterator$ = TakeWhileIterator$;
  exports.TakeWhileIterator = TakeWhileIterator;
  exports.SkipIterable$ = SkipIterable$;
  exports.SkipIterable = SkipIterable;
  exports.EfficientLengthSkipIterable$ = EfficientLengthSkipIterable$;
  exports.EfficientLengthSkipIterable = EfficientLengthSkipIterable;
  exports.SkipIterator$ = SkipIterator$;
  exports.SkipIterator = SkipIterator;
  exports.SkipWhileIterable$ = SkipWhileIterable$;
  exports.SkipWhileIterable = SkipWhileIterable;
  exports.SkipWhileIterator$ = SkipWhileIterator$;
  exports.SkipWhileIterator = SkipWhileIterator;
  exports.EmptyIterable$ = EmptyIterable$;
  exports.EmptyIterable = EmptyIterable;
  exports.EmptyIterator$ = EmptyIterator$;
  exports.EmptyIterator = EmptyIterator;
  exports.BidirectionalIterator$ = BidirectionalIterator$;
  exports.BidirectionalIterator = BidirectionalIterator;
  exports.IterableMixinWorkaround$ = IterableMixinWorkaround$;
  exports.IterableMixinWorkaround = IterableMixinWorkaround;
  exports.IterableElementError = IterableElementError;
  exports.FixedLengthListMixin$ = FixedLengthListMixin$;
  exports.FixedLengthListMixin = FixedLengthListMixin;
  exports.UnmodifiableListMixin$ = UnmodifiableListMixin$;
  exports.UnmodifiableListMixin = UnmodifiableListMixin;
  exports.FixedLengthListBase$ = FixedLengthListBase$;
  exports.FixedLengthListBase = FixedLengthListBase;
  exports.UnmodifiableListBase$ = UnmodifiableListBase$;
  exports.UnmodifiableListBase = UnmodifiableListBase;
  exports.ListMapView$ = ListMapView$;
  exports.ListMapView = ListMapView;
  exports.ReversedListIterable$ = ReversedListIterable$;
  exports.ReversedListIterable = ReversedListIterable;
  exports.UnmodifiableListError = UnmodifiableListError;
  exports.NonGrowableListError = NonGrowableListError;
  exports.makeListFixedLength = makeListFixedLength;
  exports.Lists = Lists;
  exports.printToConsole = printToConsole;
  exports.Sort = Sort;
  exports.Symbol = Symbol;
  exports.POWERS_OF_TEN = POWERS_OF_TEN;
});
dart_library.library('dart/_isolate_helper', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/_interceptors',
  'dart/_js_helper',
  'dart/isolate',
  'dart/_foreign_helper',
  'dart/collection',
  'dart/async'
], /* Lazy imports */[
  'dart/_native_typed_data',
  'dart/_js_embedded_names'
], function(exports, dart, core, _interceptors, _js_helper, isolate, _foreign_helper, collection, async, _native_typed_data, _js_embedded_names) {
  'use strict';
  let dartx = dart.dartx;
  function _serializeMessage(message) {
    return new _Serializer().serialize(message);
  }
  dart.fn(_serializeMessage);
  function _deserializeMessage(message) {
    return new _Deserializer().deserialize(message);
  }
  dart.fn(_deserializeMessage);
  function _clone(message) {
    let serializer = new _Serializer({serializeSendPorts: false});
    let deserializer = new _Deserializer();
    return deserializer.deserialize(serializer.serialize(message));
  }
  dart.fn(_clone);
  const _serializeSendPorts = Symbol('_serializeSendPorts');
  const _workerId = Symbol('_workerId');
  const _isolateId = Symbol('_isolateId');
  const _receivePortId = Symbol('_receivePortId');
  const _id = Symbol('_id');
  const _receivePort = Symbol('_receivePort');
  class _Serializer extends core.Object {
    _Serializer(opts) {
      let serializeSendPorts = opts && 'serializeSendPorts' in opts ? opts.serializeSendPorts : true;
      this.serializedObjectIds = core.Map$(dart.dynamic, core.int).identity();
      this[_serializeSendPorts] = dart.as(serializeSendPorts, core.bool);
    }
    serialize(x) {
      if (dart.notNull(this.isPrimitive(x)))
        return this.serializePrimitive(x);
      let serializationId = this.serializedObjectIds.get(x);
      if (serializationId != null)
        return this.makeRef(serializationId);
      serializationId = this.serializedObjectIds.length;
      this.serializedObjectIds.set(x, serializationId);
      if (dart.is(x, _native_typed_data.NativeByteBuffer))
        return this.serializeByteBuffer(dart.as(x, _native_typed_data.NativeByteBuffer));
      if (dart.is(x, _native_typed_data.NativeTypedData))
        return this.serializeTypedData(dart.as(x, _native_typed_data.NativeTypedData));
      if (dart.is(x, _interceptors.JSIndexable))
        return this.serializeJSIndexable(dart.as(x, _interceptors.JSIndexable));
      if (dart.is(x, _js_helper.InternalMap))
        return this.serializeMap(dart.as(x, core.Map));
      if (dart.is(x, _interceptors.JSObject))
        return this.serializeJSObject(dart.as(x, _interceptors.JSObject));
      if (dart.is(x, _interceptors.Interceptor))
        this.unsupported(x);
      if (dart.is(x, isolate.RawReceivePort)) {
        this.unsupported(x, "RawReceivePorts can't be transmitted:");
      }
      if (dart.is(x, _NativeJsSendPort))
        return this.serializeJsSendPort(dart.as(x, _NativeJsSendPort));
      if (dart.is(x, _WorkerSendPort))
        return this.serializeWorkerSendPort(dart.as(x, _WorkerSendPort));
      if (dart.is(x, core.Function))
        return this.serializeClosure(dart.as(x, core.Function));
      return this.serializeDartObject(x);
    }
    unsupported(x, message) {
      if (message === void 0)
        message = null;
      if (message == null)
        message = "Can't transmit:";
      dart.throw(new core.UnsupportedError(`${message} ${x}`));
    }
    makeRef(serializationId) {
      return ["ref", serializationId];
    }
    isPrimitive(x) {
      return x == null || typeof x == 'string' || typeof x == 'number' || typeof x == 'boolean';
    }
    serializePrimitive(primitive) {
      return primitive;
    }
    serializeByteBuffer(buffer) {
      return ["buffer", buffer];
    }
    serializeTypedData(data) {
      return ["typed", data];
    }
    serializeJSIndexable(indexable) {
      dart.assert(!(typeof indexable == 'string'));
      let serialized = dart.as(this.serializeArray(dart.as(indexable, _interceptors.JSArray)), core.List);
      if (dart.is(indexable, _interceptors.JSFixedArray))
        return ["fixed", serialized];
      if (dart.is(indexable, _interceptors.JSExtendableArray))
        return ["extendable", serialized];
      if (dart.is(indexable, _interceptors.JSMutableArray))
        return ["mutable", serialized];
      if (dart.is(indexable, _interceptors.JSArray))
        return ["const", serialized];
      this.unsupported(indexable, "Can't serialize indexable: ");
      return null;
    }
    serializeArray(x) {
      let serialized = [];
      serialized[dartx.length] = x[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(x[dartx.length]); i = dart.notNull(i) + 1) {
        serialized[dartx.set](i, this.serialize(x[dartx.get](i)));
      }
      return serialized;
    }
    serializeArrayInPlace(x) {
      for (let i = 0; dart.notNull(i) < dart.notNull(x[dartx.length]); i = dart.notNull(i) + 1) {
        x[dartx.set](i, this.serialize(x[dartx.get](i)));
      }
      return x;
    }
    serializeMap(x) {
      let serializeTearOff = dart.bind(this, 'serialize');
      return ['map', x.keys[dartx.map](dart.as(serializeTearOff, __CastType0))[dartx.toList](), x.values[dartx.map](dart.as(serializeTearOff, dart.functionType(dart.dynamic, [dart.dynamic])))[dartx.toList]()];
    }
    serializeJSObject(x) {
      if (!!x.constructor && x.constructor !== Object) {
        this.unsupported(x, "Only plain JS Objects are supported:");
      }
      let keys = dart.as(Object.keys(x), core.List);
      let values = [];
      values[dartx.length] = keys[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        values[dartx.set](i, this.serialize(x[keys[dartx.get](i)]));
      }
      return ['js-object', keys, values];
    }
    serializeWorkerSendPort(x) {
      if (dart.notNull(this[_serializeSendPorts])) {
        return ['sendport', x[_workerId], x[_isolateId], x[_receivePortId]];
      }
      return ['raw sendport', x];
    }
    serializeJsSendPort(x) {
      if (dart.notNull(this[_serializeSendPorts])) {
        let workerId = exports._globalState.currentManagerId;
        return ['sendport', workerId, x[_isolateId], x[_receivePort][_id]];
      }
      return ['raw sendport', x];
    }
    serializeCapability(x) {
      return ['capability', x[_id]];
    }
    serializeClosure(x) {
      let name = IsolateNatives._getJSFunctionName(x);
      if (name == null) {
        this.unsupported(x, "Closures can't be transmitted:");
      }
      return ['function', name];
    }
    serializeDartObject(x) {
      let classExtractor = _foreign_helper.JS_EMBEDDED_GLOBAL('', _js_embedded_names.CLASS_ID_EXTRACTOR);
      let fieldsExtractor = _foreign_helper.JS_EMBEDDED_GLOBAL('', _js_embedded_names.CLASS_FIELDS_EXTRACTOR);
      let classId = classExtractor(x);
      let fields = dart.as(fieldsExtractor(x), core.List);
      return ['dart', classId, this.serializeArrayInPlace(dart.as(fields, _interceptors.JSArray))];
    }
  }
  dart.setSignature(_Serializer, {
    constructors: () => ({_Serializer: [_Serializer, [], {serializeSendPorts: dart.dynamic}]}),
    methods: () => ({
      serialize: [dart.dynamic, [dart.dynamic]],
      unsupported: [dart.void, [dart.dynamic], [core.String]],
      makeRef: [dart.dynamic, [core.int]],
      isPrimitive: [core.bool, [dart.dynamic]],
      serializePrimitive: [dart.dynamic, [dart.dynamic]],
      serializeByteBuffer: [dart.dynamic, [_native_typed_data.NativeByteBuffer]],
      serializeTypedData: [dart.dynamic, [_native_typed_data.NativeTypedData]],
      serializeJSIndexable: [dart.dynamic, [_interceptors.JSIndexable]],
      serializeArray: [dart.dynamic, [_interceptors.JSArray]],
      serializeArrayInPlace: [dart.dynamic, [_interceptors.JSArray]],
      serializeMap: [dart.dynamic, [core.Map]],
      serializeJSObject: [dart.dynamic, [_interceptors.JSObject]],
      serializeWorkerSendPort: [dart.dynamic, [_WorkerSendPort]],
      serializeJsSendPort: [dart.dynamic, [_NativeJsSendPort]],
      serializeCapability: [dart.dynamic, [CapabilityImpl]],
      serializeClosure: [dart.dynamic, [core.Function]],
      serializeDartObject: [dart.dynamic, [dart.dynamic]]
    })
  });
  const _adjustSendPorts = Symbol('_adjustSendPorts');
  class _Deserializer extends core.Object {
    _Deserializer(opts) {
      let adjustSendPorts = opts && 'adjustSendPorts' in opts ? opts.adjustSendPorts : true;
      this.deserializedObjects = core.List.new();
      this[_adjustSendPorts] = dart.as(adjustSendPorts, core.bool);
    }
    deserialize(x) {
      if (dart.notNull(this.isPrimitive(x)))
        return this.deserializePrimitive(x);
      if (!dart.is(x, _interceptors.JSArray))
        dart.throw(new core.ArgumentError(`Bad serialized message: ${x}`));
      switch (dart.dload(x, 'first')) {
        case "ref":
        {
          return this.deserializeRef(x);
        }
        case "buffer":
        {
          return this.deserializeByteBuffer(x);
        }
        case "typed":
        {
          return this.deserializeTypedData(x);
        }
        case "fixed":
        {
          return this.deserializeFixed(x);
        }
        case "extendable":
        {
          return this.deserializeExtendable(x);
        }
        case "mutable":
        {
          return this.deserializeMutable(x);
        }
        case "const":
        {
          return this.deserializeConst(x);
        }
        case "map":
        {
          return this.deserializeMap(x);
        }
        case "sendport":
        {
          return this.deserializeSendPort(x);
        }
        case "raw sendport":
        {
          return this.deserializeRawSendPort(x);
        }
        case "js-object":
        {
          return this.deserializeJSObject(x);
        }
        case "function":
        {
          return this.deserializeClosure(x);
        }
        case "dart":
        {
          return this.deserializeDartObject(x);
        }
        default:
        {
          dart.throw(`couldn't deserialize: ${x}`);
        }
      }
    }
    isPrimitive(x) {
      return x == null || typeof x == 'string' || typeof x == 'number' || typeof x == 'boolean';
    }
    deserializePrimitive(x) {
      return x;
    }
    deserializeRef(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'ref'));
      let serializationId = dart.as(dart.dindex(x, 1), core.int);
      return this.deserializedObjects[dartx.get](serializationId);
    }
    deserializeByteBuffer(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'buffer'));
      let result = dart.as(dart.dindex(x, 1), _native_typed_data.NativeByteBuffer);
      this.deserializedObjects[dartx.add](result);
      return result;
    }
    deserializeTypedData(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'typed'));
      let result = dart.as(dart.dindex(x, 1), _native_typed_data.NativeTypedData);
      this.deserializedObjects[dartx.add](result);
      return result;
    }
    deserializeArrayInPlace(x) {
      for (let i = 0; dart.notNull(i) < dart.notNull(x[dartx.length]); i = dart.notNull(i) + 1) {
        x[dartx.set](i, this.deserialize(x[dartx.get](i)));
      }
      return x;
    }
    deserializeFixed(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'fixed'));
      let result = dart.as(dart.dindex(x, 1), core.List);
      this.deserializedObjects[dartx.add](result);
      return _interceptors.JSArray.markFixed(this.deserializeArrayInPlace(dart.as(result, _interceptors.JSArray)));
    }
    deserializeExtendable(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'extendable'));
      let result = dart.as(dart.dindex(x, 1), core.List);
      this.deserializedObjects[dartx.add](result);
      return _interceptors.JSArray.markGrowable(this.deserializeArrayInPlace(dart.as(result, _interceptors.JSArray)));
    }
    deserializeMutable(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'mutable'));
      let result = dart.as(dart.dindex(x, 1), core.List);
      this.deserializedObjects[dartx.add](result);
      return this.deserializeArrayInPlace(dart.as(result, _interceptors.JSArray));
    }
    deserializeConst(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'const'));
      let result = dart.as(dart.dindex(x, 1), core.List);
      this.deserializedObjects[dartx.add](result);
      return _interceptors.JSArray.markFixed(this.deserializeArrayInPlace(dart.as(result, _interceptors.JSArray)));
    }
    deserializeMap(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'map'));
      let keys = dart.as(dart.dindex(x, 1), core.List);
      let values = dart.as(dart.dindex(x, 2), core.List);
      let result = dart.map();
      this.deserializedObjects[dartx.add](result);
      keys = keys[dartx.map](dart.bind(this, 'deserialize'))[dartx.toList]();
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        result.set(keys[dartx.get](i), this.deserialize(values[dartx.get](i)));
      }
      return result;
    }
    deserializeSendPort(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'sendport'));
      let managerId = dart.as(dart.dindex(x, 1), core.int);
      let isolateId = dart.as(dart.dindex(x, 2), core.int);
      let receivePortId = dart.as(dart.dindex(x, 3), core.int);
      let result = null;
      if (managerId == exports._globalState.currentManagerId) {
        let isolate = exports._globalState.isolates.get(isolateId);
        if (isolate == null)
          return null;
        let receivePort = isolate.lookup(receivePortId);
        if (receivePort == null)
          return null;
        result = new _NativeJsSendPort(receivePort, isolateId);
      } else {
        result = new _WorkerSendPort(managerId, isolateId, receivePortId);
      }
      this.deserializedObjects[dartx.add](result);
      return result;
    }
    deserializeRawSendPort(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'raw sendport'));
      let result = dart.as(dart.dindex(x, 1), isolate.SendPort);
      this.deserializedObjects[dartx.add](result);
      return result;
    }
    deserializeJSObject(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'js-object'));
      let keys = dart.as(dart.dindex(x, 1), core.List);
      let values = dart.as(dart.dindex(x, 2), core.List);
      let o = {};
      this.deserializedObjects[dartx.add](o);
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        o[keys[dartx.get](i)] = this.deserialize(values[dartx.get](i));
      }
      return o;
    }
    deserializeClosure(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'function'));
      let name = dart.as(dart.dindex(x, 1), core.String);
      let result = dart.as(IsolateNatives._getJSFunctionFromName(name), core.Function);
      this.deserializedObjects[dartx.add](result);
      return result;
    }
    deserializeDartObject(x) {
      dart.assert(dart.equals(dart.dindex(x, 0), 'dart'));
      let classId = dart.as(dart.dindex(x, 1), core.String);
      let fields = dart.as(dart.dindex(x, 2), core.List);
      let instanceFromClassId = _foreign_helper.JS_EMBEDDED_GLOBAL('', _js_embedded_names.INSTANCE_FROM_CLASS_ID);
      let initializeObject = _foreign_helper.JS_EMBEDDED_GLOBAL('', _js_embedded_names.INITIALIZE_EMPTY_INSTANCE);
      let emptyInstance = instanceFromClassId(classId);
      this.deserializedObjects[dartx.add](emptyInstance);
      this.deserializeArrayInPlace(dart.as(fields, _interceptors.JSArray));
      return initializeObject(classId, emptyInstance, fields);
    }
  }
  dart.setSignature(_Deserializer, {
    constructors: () => ({_Deserializer: [_Deserializer, [], {adjustSendPorts: dart.dynamic}]}),
    methods: () => ({
      deserialize: [dart.dynamic, [dart.dynamic]],
      isPrimitive: [core.bool, [dart.dynamic]],
      deserializePrimitive: [dart.dynamic, [dart.dynamic]],
      deserializeRef: [dart.dynamic, [dart.dynamic]],
      deserializeByteBuffer: [_native_typed_data.NativeByteBuffer, [dart.dynamic]],
      deserializeTypedData: [_native_typed_data.NativeTypedData, [dart.dynamic]],
      deserializeArrayInPlace: [core.List, [_interceptors.JSArray]],
      deserializeFixed: [core.List, [dart.dynamic]],
      deserializeExtendable: [core.List, [dart.dynamic]],
      deserializeMutable: [core.List, [dart.dynamic]],
      deserializeConst: [core.List, [dart.dynamic]],
      deserializeMap: [core.Map, [dart.dynamic]],
      deserializeSendPort: [isolate.SendPort, [dart.dynamic]],
      deserializeRawSendPort: [isolate.SendPort, [dart.dynamic]],
      deserializeJSObject: [dart.dynamic, [dart.dynamic]],
      deserializeClosure: [core.Function, [dart.dynamic]],
      deserializeDartObject: [dart.dynamic, [dart.dynamic]]
    })
  });
  const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  function _callInIsolate(isolate, func) {
    let result = isolate.eval(func);
    exports._globalState.topEventLoop.run();
    return result;
  }
  dart.fn(_callInIsolate, () => dart.definiteFunctionType(dart.dynamic, [_IsolateContext, core.Function]));
  const _activeJsAsyncCount = Symbol('_activeJsAsyncCount');
  function enterJsAsync() {
    let o = exports._globalState.topEventLoop;
    o[_activeJsAsyncCount] = dart.notNull(o[_activeJsAsyncCount]) + 1;
  }
  dart.fn(enterJsAsync);
  function leaveJsAsync() {
    let o = exports._globalState.topEventLoop;
    o[_activeJsAsyncCount] = dart.notNull(o[_activeJsAsyncCount]) - 1;
    dart.assert(dart.notNull(exports._globalState.topEventLoop[_activeJsAsyncCount]) >= 0);
  }
  dart.fn(leaveJsAsync);
  function isWorker() {
    return exports._globalState.isWorker;
  }
  dart.fn(isWorker, core.bool, []);
  function _currentIsolate() {
    return exports._globalState.currentContext;
  }
  dart.fn(_currentIsolate, () => dart.definiteFunctionType(_IsolateContext, []));
  function startRootIsolate(entry, args) {
    args = args;
    if (args == null)
      args = [];
    if (!dart.is(args, core.List)) {
      dart.throw(new core.ArgumentError(`Arguments to main must be a List: ${args}`));
    }
    exports._globalState = new _Manager(dart.as(entry, core.Function));
    if (dart.notNull(exports._globalState.isWorker))
      return;
    let rootContext = new _IsolateContext();
    exports._globalState.rootContext = rootContext;
    exports._globalState.currentContext = rootContext;
    if (dart.is(entry, _MainFunctionArgs)) {
      rootContext.eval(dart.fn(() => {
        dart.dcall(entry, args);
      }));
    } else if (dart.is(entry, _MainFunctionArgsMessage)) {
      rootContext.eval(dart.fn(() => {
        dart.dcall(entry, args, null);
      }));
    } else {
      rootContext.eval(dart.as(entry, core.Function));
    }
    exports._globalState.topEventLoop.run();
  }
  dart.fn(startRootIsolate, dart.void, [dart.dynamic, dart.dynamic]);
  dart.copyProperties(exports, {
    get _globalState() {
      return dart.as(dart.globalState, _Manager);
    },
    set _globalState(val) {
      dart.globalState = val;
    }
  });
  const _nativeDetectEnvironment = Symbol('_nativeDetectEnvironment');
  const _nativeInitWorkerMessageHandler = Symbol('_nativeInitWorkerMessageHandler');
  class _Manager extends core.Object {
    get useWorkers() {
      return this.supportsWorkers;
    }
    _Manager(entry) {
      this.entry = entry;
      this.nextIsolateId = 0;
      this.currentManagerId = 0;
      this.nextManagerId = 1;
      this.currentContext = null;
      this.rootContext = null;
      this.topEventLoop = null;
      this.fromCommandLine = null;
      this.isWorker = null;
      this.supportsWorkers = null;
      this.isolates = null;
      this.mainManager = null;
      this.managers = null;
      this[_nativeDetectEnvironment]();
      this.topEventLoop = new _EventLoop();
      this.isolates = core.Map$(core.int, _IsolateContext).new();
      this.managers = core.Map$(core.int, dart.dynamic).new();
      if (dart.notNull(this.isWorker)) {
        this.mainManager = new _MainManagerStub();
        this[_nativeInitWorkerMessageHandler]();
      }
    }
    [_nativeDetectEnvironment]() {
      let isWindowDefined = exports.globalWindow != null;
      let isWorkerDefined = exports.globalWorker != null;
      this.isWorker = !dart.notNull(isWindowDefined) && dart.notNull(exports.globalPostMessageDefined);
      this.supportsWorkers = dart.notNull(this.isWorker) || dart.notNull(isWorkerDefined) && IsolateNatives.thisScript != null;
      this.fromCommandLine = !dart.notNull(isWindowDefined) && !dart.notNull(this.isWorker);
    }
    [_nativeInitWorkerMessageHandler]() {
      let func = (function(f, a) {
        return function(e) {
          f(a, e);
        };
      })(IsolateNatives._processWorkerMessage, this.mainManager);
      self.onmessage = func;
      self.dartPrint = self.dartPrint || (function(serialize) {
        return function(object) {
          if (self.console && self.console.log) {
            self.console.log(object);
          } else {
            self.postMessage(serialize(object));
          }
        };
      })(_Manager._serializePrintMessage);
    }
    static _serializePrintMessage(object) {
      return _serializeMessage(dart.map({command: "print", msg: object}));
    }
    maybeCloseWorker() {
      if (dart.notNull(this.isWorker) && dart.notNull(this.isolates.isEmpty) && this.topEventLoop[_activeJsAsyncCount] == 0) {
        this.mainManager.postMessage(_serializeMessage(dart.map({command: 'close'})));
      }
    }
  }
  dart.setSignature(_Manager, {
    constructors: () => ({_Manager: [_Manager, [core.Function]]}),
    methods: () => ({
      [_nativeDetectEnvironment]: [dart.void, []],
      [_nativeInitWorkerMessageHandler]: [dart.void, []],
      maybeCloseWorker: [dart.void, []]
    }),
    statics: () => ({_serializePrintMessage: [dart.dynamic, [dart.dynamic]]}),
    names: ['_serializePrintMessage']
  });
  const _scheduledControlEvents = Symbol('_scheduledControlEvents');
  const _isExecutingEvent = Symbol('_isExecutingEvent');
  const _updateGlobalState = Symbol('_updateGlobalState');
  const _setGlobals = Symbol('_setGlobals');
  const _addRegistration = Symbol('_addRegistration');
  const _close = Symbol('_close');
  class _IsolateContext extends core.Object {
    _IsolateContext() {
      this.id = (() => {
        let o = exports._globalState, x = o.nextIsolateId;
        o.nextIsolateId = dart.notNull(x) + 1;
        return x;
      })();
      this.ports = core.Map$(core.int, RawReceivePortImpl).new();
      this.weakPorts = core.Set$(core.int).new();
      this.isolateStatics = _foreign_helper.JS_CREATE_ISOLATE();
      this.controlPort = new RawReceivePortImpl._controlPort();
      this.pauseCapability = isolate.Capability.new();
      this.terminateCapability = isolate.Capability.new();
      this.delayedEvents = dart.list([], _IsolateEvent);
      this.pauseTokens = core.Set$(isolate.Capability).new();
      this.errorPorts = core.Set$(isolate.SendPort).new();
      this.initialized = false;
      this.isPaused = false;
      this.doneHandlers = null;
      this[_scheduledControlEvents] = null;
      this[_isExecutingEvent] = false;
      this.errorsAreFatal = true;
      this.registerWeak(this.controlPort[_id], this.controlPort);
    }
    addPause(authentification, resume) {
      if (!dart.equals(this.pauseCapability, authentification))
        return;
      if (dart.notNull(this.pauseTokens.add(resume)) && !dart.notNull(this.isPaused)) {
        this.isPaused = true;
      }
      this[_updateGlobalState]();
    }
    removePause(resume) {
      if (!dart.notNull(this.isPaused))
        return;
      this.pauseTokens.remove(resume);
      if (dart.notNull(this.pauseTokens.isEmpty)) {
        while (dart.notNull(this.delayedEvents[dartx.isNotEmpty])) {
          let event = this.delayedEvents[dartx.removeLast]();
          exports._globalState.topEventLoop.prequeue(event);
        }
        this.isPaused = false;
      }
      this[_updateGlobalState]();
    }
    addDoneListener(responsePort) {
      if (this.doneHandlers == null) {
        this.doneHandlers = [];
      }
      if (dart.notNull(dart.as(dart.dsend(this.doneHandlers, 'contains', responsePort), core.bool)))
        return;
      dart.dsend(this.doneHandlers, 'add', responsePort);
    }
    removeDoneListener(responsePort) {
      if (this.doneHandlers == null)
        return;
      dart.dsend(this.doneHandlers, 'remove', responsePort);
    }
    setErrorsFatal(authentification, errorsAreFatal) {
      if (!dart.equals(this.terminateCapability, authentification))
        return;
      this.errorsAreFatal = errorsAreFatal;
    }
    handlePing(responsePort, pingType) {
      if (pingType == isolate.Isolate.IMMEDIATE || pingType == isolate.Isolate.BEFORE_NEXT_EVENT && !dart.notNull(this[_isExecutingEvent])) {
        responsePort.send(null);
        return;
      }
      function respond() {
        responsePort.send(null);
      }
      dart.fn(respond, dart.void, []);
      if (pingType == isolate.Isolate.AS_EVENT) {
        exports._globalState.topEventLoop.enqueue(this, respond, "ping");
        return;
      }
      dart.assert(pingType == isolate.Isolate.BEFORE_NEXT_EVENT);
      if (this[_scheduledControlEvents] == null) {
        this[_scheduledControlEvents] = collection.Queue.new();
      }
      dart.dsend(this[_scheduledControlEvents], 'addLast', respond);
    }
    handleKill(authentification, priority) {
      if (!dart.equals(this.terminateCapability, authentification))
        return;
      if (priority == isolate.Isolate.IMMEDIATE || priority == isolate.Isolate.BEFORE_NEXT_EVENT && !dart.notNull(this[_isExecutingEvent])) {
        this.kill();
        return;
      }
      if (priority == isolate.Isolate.AS_EVENT) {
        exports._globalState.topEventLoop.enqueue(this, dart.bind(this, 'kill'), "kill");
        return;
      }
      dart.assert(priority == isolate.Isolate.BEFORE_NEXT_EVENT);
      if (this[_scheduledControlEvents] == null) {
        this[_scheduledControlEvents] = collection.Queue.new();
      }
      dart.dsend(this[_scheduledControlEvents], 'addLast', dart.bind(this, 'kill'));
    }
    addErrorListener(port) {
      this.errorPorts.add(port);
    }
    removeErrorListener(port) {
      this.errorPorts.remove(port);
    }
    handleUncaughtError(error, stackTrace) {
      if (dart.notNull(this.errorPorts.isEmpty)) {
        if (dart.notNull(this.errorsAreFatal) && dart.notNull(core.identical(this, exports._globalState.rootContext))) {
          return;
        }
        if (self.console && self.console.error) {
          self.console.error(error, stackTrace);
        } else {
          core.print(error);
          if (stackTrace != null)
            core.print(stackTrace);
        }
        return;
      }
      let message = core.List.new(2);
      message[dartx.set](0, dart.toString(error));
      message[dartx.set](1, stackTrace == null ? null : dart.toString(stackTrace));
      for (let port of this.errorPorts)
        port.send(message);
    }
    eval(code) {
      let old = exports._globalState.currentContext;
      exports._globalState.currentContext = this;
      this[_setGlobals]();
      let result = null;
      this[_isExecutingEvent] = true;
      try {
        result = dart.dcall(code);
      } catch (e) {
        let s = dart.stackTrace(e);
        this.handleUncaughtError(e, s);
        if (dart.notNull(this.errorsAreFatal)) {
          this.kill();
          if (dart.notNull(core.identical(this, exports._globalState.rootContext))) {
            throw e;
          }
        }
      }
 finally {
        this[_isExecutingEvent] = false;
        exports._globalState.currentContext = old;
        if (old != null)
          old[_setGlobals]();
        if (this[_scheduledControlEvents] != null) {
          while (dart.notNull(dart.as(dart.dload(this[_scheduledControlEvents], 'isNotEmpty'), core.bool))) {
            dart.dcall(dart.dsend(this[_scheduledControlEvents], 'removeFirst'));
          }
        }
      }
      return result;
    }
    [_setGlobals]() {
      _foreign_helper.JS_SET_CURRENT_ISOLATE(this.isolateStatics);
    }
    handleControlMessage(message) {
      switch (dart.dindex(message, 0)) {
        case "pause":
        {
          this.addPause(dart.as(dart.dindex(message, 1), isolate.Capability), dart.as(dart.dindex(message, 2), isolate.Capability));
          break;
        }
        case "resume":
        {
          this.removePause(dart.as(dart.dindex(message, 1), isolate.Capability));
          break;
        }
        case 'add-ondone':
        {
          this.addDoneListener(dart.as(dart.dindex(message, 1), isolate.SendPort));
          break;
        }
        case 'remove-ondone':
        {
          this.removeDoneListener(dart.as(dart.dindex(message, 1), isolate.SendPort));
          break;
        }
        case 'set-errors-fatal':
        {
          this.setErrorsFatal(dart.as(dart.dindex(message, 1), isolate.Capability), dart.as(dart.dindex(message, 2), core.bool));
          break;
        }
        case "ping":
        {
          this.handlePing(dart.as(dart.dindex(message, 1), isolate.SendPort), dart.as(dart.dindex(message, 2), core.int));
          break;
        }
        case "kill":
        {
          this.handleKill(dart.as(dart.dindex(message, 1), isolate.Capability), dart.as(dart.dindex(message, 2), core.int));
          break;
        }
        case "getErrors":
        {
          this.addErrorListener(dart.as(dart.dindex(message, 1), isolate.SendPort));
          break;
        }
        case "stopErrors":
        {
          this.removeErrorListener(dart.as(dart.dindex(message, 1), isolate.SendPort));
          break;
        }
        default:
      }
    }
    lookup(portId) {
      return this.ports.get(portId);
    }
    [_addRegistration](portId, port) {
      if (dart.notNull(this.ports.containsKey(portId))) {
        dart.throw(core.Exception.new("Registry: ports must be registered only once."));
      }
      this.ports.set(portId, port);
    }
    register(portId, port) {
      this[_addRegistration](portId, port);
      this[_updateGlobalState]();
    }
    registerWeak(portId, port) {
      this.weakPorts.add(portId);
      this[_addRegistration](portId, port);
    }
    [_updateGlobalState]() {
      if (dart.notNull(this.ports.length) - dart.notNull(this.weakPorts.length) > 0 || dart.notNull(this.isPaused) || !dart.notNull(this.initialized)) {
        exports._globalState.isolates.set(this.id, this);
      } else {
        this.kill();
      }
    }
    kill() {
      if (this[_scheduledControlEvents] != null) {
        dart.dsend(this[_scheduledControlEvents], 'clear');
      }
      for (let port of this.ports.values) {
        port[_close]();
      }
      this.ports.clear();
      this.weakPorts.clear();
      exports._globalState.isolates.remove(this.id);
      this.errorPorts.clear();
      if (this.doneHandlers != null) {
        for (let port of dart.as(this.doneHandlers, core.Iterable$(isolate.SendPort))) {
          port.send(null);
        }
        this.doneHandlers = null;
      }
    }
    unregister(portId) {
      this.ports.remove(portId);
      this.weakPorts.remove(portId);
      this[_updateGlobalState]();
    }
  }
  _IsolateContext[dart.implements] = () => [_foreign_helper.IsolateContext];
  dart.setSignature(_IsolateContext, {
    constructors: () => ({_IsolateContext: [_IsolateContext, []]}),
    methods: () => ({
      addPause: [dart.void, [isolate.Capability, isolate.Capability]],
      removePause: [dart.void, [isolate.Capability]],
      addDoneListener: [dart.void, [isolate.SendPort]],
      removeDoneListener: [dart.void, [isolate.SendPort]],
      setErrorsFatal: [dart.void, [isolate.Capability, core.bool]],
      handlePing: [dart.void, [isolate.SendPort, core.int]],
      handleKill: [dart.void, [isolate.Capability, core.int]],
      addErrorListener: [dart.void, [isolate.SendPort]],
      removeErrorListener: [dart.void, [isolate.SendPort]],
      handleUncaughtError: [dart.void, [dart.dynamic, core.StackTrace]],
      eval: [dart.dynamic, [core.Function]],
      [_setGlobals]: [dart.void, []],
      handleControlMessage: [dart.void, [dart.dynamic]],
      lookup: [RawReceivePortImpl, [core.int]],
      [_addRegistration]: [dart.void, [core.int, RawReceivePortImpl]],
      register: [dart.void, [core.int, RawReceivePortImpl]],
      registerWeak: [dart.void, [core.int, RawReceivePortImpl]],
      [_updateGlobalState]: [dart.void, []],
      kill: [dart.void, []],
      unregister: [dart.void, [core.int]]
    })
  });
  const _runHelper = Symbol('_runHelper');
  class _EventLoop extends core.Object {
    _EventLoop() {
      this.events = collection.Queue$(_IsolateEvent).new();
      this[_activeJsAsyncCount] = 0;
    }
    enqueue(isolate, fn, msg) {
      this.events.addLast(new _IsolateEvent(dart.as(isolate, _IsolateContext), dart.as(fn, core.Function), dart.as(msg, core.String)));
    }
    prequeue(event) {
      this.events.addFirst(event);
    }
    dequeue() {
      if (dart.notNull(this.events.isEmpty))
        return null;
      return this.events.removeFirst();
    }
    checkOpenReceivePortsFromCommandLine() {
      if (exports._globalState.rootContext != null && dart.notNull(exports._globalState.isolates.containsKey(exports._globalState.rootContext.id)) && dart.notNull(exports._globalState.fromCommandLine) && dart.notNull(exports._globalState.rootContext.ports.isEmpty)) {
        dart.throw(core.Exception.new("Program exited with open ReceivePorts."));
      }
    }
    runIteration() {
      let event = this.dequeue();
      if (event == null) {
        this.checkOpenReceivePortsFromCommandLine();
        exports._globalState.maybeCloseWorker();
        return false;
      }
      event.process();
      return true;
    }
    [_runHelper]() {
      if (exports.globalWindow != null) {
        const next = (function() {
          if (!dart.notNull(this.runIteration()))
            return;
          async.Timer.run(next);
        }).bind(this);
        dart.fn(next);
        next();
      } else {
        while (dart.notNull(this.runIteration())) {
        }
      }
    }
    run() {
      if (!dart.notNull(exports._globalState.isWorker)) {
        this[_runHelper]();
      } else {
        try {
          this[_runHelper]();
        } catch (e) {
          let trace = dart.stackTrace(e);
          exports._globalState.mainManager.postMessage(_serializeMessage(dart.map({command: 'error', msg: `${e}\n${trace}`})));
        }

      }
    }
  }
  dart.setSignature(_EventLoop, {
    constructors: () => ({_EventLoop: [_EventLoop, []]}),
    methods: () => ({
      enqueue: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      prequeue: [dart.void, [_IsolateEvent]],
      dequeue: [_IsolateEvent, []],
      checkOpenReceivePortsFromCommandLine: [dart.void, []],
      runIteration: [core.bool, []],
      [_runHelper]: [dart.void, []],
      run: [dart.void, []]
    })
  });
  class _IsolateEvent extends core.Object {
    _IsolateEvent(isolate, fn, message) {
      this.isolate = isolate;
      this.fn = fn;
      this.message = message;
    }
    process() {
      if (dart.notNull(this.isolate.isPaused)) {
        this.isolate.delayedEvents[dartx.add](this);
        return;
      }
      this.isolate.eval(this.fn);
    }
  }
  dart.setSignature(_IsolateEvent, {
    constructors: () => ({_IsolateEvent: [_IsolateEvent, [_IsolateContext, core.Function, core.String]]}),
    methods: () => ({process: [dart.void, []]})
  });
  dart.defineLazyProperties(exports, {
    get _global() {
      return typeof global == 'undefined' ? self : global;
    }
  });
  class _MainManagerStub extends core.Object {
    postMessage(msg) {
      exports._global.postMessage(msg);
    }
  }
  dart.setSignature(_MainManagerStub, {
    methods: () => ({postMessage: [dart.void, [dart.dynamic]]})
  });
  const _SPAWNED_SIGNAL = "spawned";
  const _SPAWN_FAILED_SIGNAL = "spawn failed";
  dart.copyProperties(exports, {
    get globalWindow() {
      return exports._global.window;
    },
    get globalWorker() {
      return exports._global.Worker;
    },
    get globalPostMessageDefined() {
      return !!exports._global.postMessage;
    }
  });
  const _MainFunction = dart.typedef('_MainFunction', () => dart.functionType(dart.dynamic, []));
  const _MainFunctionArgs = dart.typedef('_MainFunctionArgs', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const _MainFunctionArgsMessage = dart.typedef('_MainFunctionArgsMessage', () => dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
  class IsolateNatives extends core.Object {
    static computeThisScript() {
      let currentScript = document.currentScript;
      if (currentScript != null) {
        return String(currentScript.src);
      }
      if (dart.notNull(_js_helper.Primitives.isD8))
        return IsolateNatives.computeThisScriptD8();
      if (dart.notNull(_js_helper.Primitives.isJsshell))
        return IsolateNatives.computeThisScriptJsshell();
      if (exports._globalState != null && dart.notNull(exports._globalState.isWorker)) {
        return IsolateNatives.computeThisScriptFromTrace();
      }
      return null;
    }
    static computeThisScriptJsshell() {
      return dart.as(thisFilename(), core.String);
    }
    static computeThisScriptD8() {
      return IsolateNatives.computeThisScriptFromTrace();
    }
    static computeThisScriptFromTrace() {
      let stack = new Error().stack;
      if (stack == null) {
        stack = (function() {
          try {
            throw new Error();
          } catch (e) {
            return e.stack;
          }

        })();
        if (stack == null)
          dart.throw(new core.UnsupportedError('No stack trace'));
      }
      let pattern = null, matches = null;
      pattern = new RegExp("^ *at [^(]*\\((.*):[0-9]*:[0-9]*\\)$", "m");
      matches = stack.match(pattern);
      if (matches != null)
        return matches[1];
      pattern = new RegExp("^[^@]*@(.*):[0-9]*$", "m");
      matches = stack.match(pattern);
      if (matches != null)
        return matches[1];
      dart.throw(new core.UnsupportedError(`Cannot extract URI from "${stack}"`));
    }
    static _getEventData(e) {
      return e.data;
    }
    static _processWorkerMessage(sender, e) {
      let msg = _deserializeMessage(IsolateNatives._getEventData(e));
      switch (dart.dindex(msg, 'command')) {
        case 'start':
        {
          exports._globalState.currentManagerId = dart.as(dart.dindex(msg, 'id'), core.int);
          let functionName = dart.as(dart.dindex(msg, 'functionName'), core.String);
          let entryPoint = functionName == null ? exports._globalState.entry : dart.as(IsolateNatives._getJSFunctionFromName(functionName), core.Function);
          let args = dart.dindex(msg, 'args');
          let message = _deserializeMessage(dart.dindex(msg, 'msg'));
          let isSpawnUri = dart.dindex(msg, 'isSpawnUri');
          let startPaused = dart.dindex(msg, 'startPaused');
          let replyTo = _deserializeMessage(dart.dindex(msg, 'replyTo'));
          let context = new _IsolateContext();
          exports._globalState.topEventLoop.enqueue(context, dart.fn(() => {
            IsolateNatives._startIsolate(entryPoint, dart.as(args, core.List$(core.String)), message, dart.as(isSpawnUri, core.bool), dart.as(startPaused, core.bool), dart.as(replyTo, isolate.SendPort));
          }), 'worker-start');
          exports._globalState.currentContext = context;
          exports._globalState.topEventLoop.run();
          break;
        }
        case 'spawn-worker':
        {
          if (IsolateNatives.enableSpawnWorker != null)
            IsolateNatives.handleSpawnWorkerRequest(msg);
          break;
        }
        case 'message':
        {
          let port = dart.as(dart.dindex(msg, 'port'), isolate.SendPort);
          if (port != null) {
            dart.dsend(dart.dindex(msg, 'port'), 'send', dart.dindex(msg, 'msg'));
          }
          exports._globalState.topEventLoop.run();
          break;
        }
        case 'close':
        {
          exports._globalState.managers.remove(IsolateNatives.workerIds.get(sender));
          sender.terminate();
          exports._globalState.topEventLoop.run();
          break;
        }
        case 'log':
        {
          IsolateNatives._log(dart.dindex(msg, 'msg'));
          break;
        }
        case 'print':
        {
          if (dart.notNull(exports._globalState.isWorker)) {
            exports._globalState.mainManager.postMessage(_serializeMessage(dart.map({command: 'print', msg: msg})));
          } else {
            core.print(dart.dindex(msg, 'msg'));
          }
          break;
        }
        case 'error':
        {
          dart.throw(dart.dindex(msg, 'msg'));
        }
      }
    }
    static handleSpawnWorkerRequest(msg) {
      let replyPort = dart.dindex(msg, 'replyPort');
      IsolateNatives.spawn(dart.as(dart.dindex(msg, 'functionName'), core.String), dart.as(dart.dindex(msg, 'uri'), core.String), dart.as(dart.dindex(msg, 'args'), core.List$(core.String)), dart.dindex(msg, 'msg'), false, dart.as(dart.dindex(msg, 'isSpawnUri'), core.bool), dart.as(dart.dindex(msg, 'startPaused'), core.bool)).then(dart.fn(msg => {
        dart.dsend(replyPort, 'send', msg);
      }), {onError: dart.fn(errorMessage => {
          dart.dsend(replyPort, 'send', [_SPAWN_FAILED_SIGNAL, errorMessage]);
        }, dart.dynamic, [core.String])});
    }
    static _log(msg) {
      if (dart.notNull(exports._globalState.isWorker)) {
        exports._globalState.mainManager.postMessage(_serializeMessage(dart.map({command: 'log', msg: msg})));
      } else {
        try {
          IsolateNatives._consoleLog(msg);
        } catch (e) {
          let trace = dart.stackTrace(e);
          dart.throw(core.Exception.new(trace));
        }

      }
    }
    static _consoleLog(msg) {
      self.console.log(msg);
    }
    static _getJSFunctionFromName(functionName) {
      let globalFunctionsContainer = _foreign_helper.JS_EMBEDDED_GLOBAL("", _js_embedded_names.GLOBAL_FUNCTIONS);
      return globalFunctionsContainer[functionName]();
    }
    static _getJSFunctionName(f) {
      return dart.as(f.$name, core.String);
    }
    static _allocate(ctor) {
      return new ctor();
    }
    static spawnFunction(topLevelFunction, message, startPaused) {
      IsolateNatives.enableSpawnWorker = true;
      let name = IsolateNatives._getJSFunctionName(topLevelFunction);
      if (name == null) {
        dart.throw(new core.UnsupportedError("only top-level functions can be spawned."));
      }
      let isLight = false;
      let isSpawnUri = false;
      return IsolateNatives.spawn(name, null, null, message, isLight, isSpawnUri, startPaused);
    }
    static spawnUri(uri, args, message, startPaused) {
      IsolateNatives.enableSpawnWorker = true;
      let isLight = false;
      let isSpawnUri = true;
      return IsolateNatives.spawn(null, dart.toString(uri), args, message, isLight, isSpawnUri, startPaused);
    }
    static spawn(functionName, uri, args, message, isLight, isSpawnUri, startPaused) {
      if (uri != null && dart.notNull(uri[dartx.endsWith](".dart"))) {
        uri = dart.notNull(uri) + ".js";
      }
      let port = isolate.ReceivePort.new();
      let completer = async.Completer$(core.List).new();
      port.first.then(dart.fn(msg => {
        if (dart.equals(dart.dindex(msg, 0), _SPAWNED_SIGNAL)) {
          completer.complete(msg);
        } else {
          dart.assert(dart.equals(dart.dindex(msg, 0), _SPAWN_FAILED_SIGNAL));
          completer.completeError(dart.dindex(msg, 1));
        }
      }));
      let signalReply = port.sendPort;
      if (dart.notNull(exports._globalState.useWorkers) && !dart.notNull(isLight)) {
        IsolateNatives._startWorker(functionName, uri, args, message, isSpawnUri, startPaused, signalReply, dart.fn(message => completer.completeError(message), dart.void, [core.String]));
      } else {
        IsolateNatives._startNonWorker(functionName, uri, args, message, isSpawnUri, startPaused, signalReply);
      }
      return completer.future;
    }
    static _startWorker(functionName, uri, args, message, isSpawnUri, startPaused, replyPort, onError) {
      if (args != null)
        args = core.List$(core.String).from(args);
      if (dart.notNull(exports._globalState.isWorker)) {
        exports._globalState.mainManager.postMessage(_serializeMessage(dart.map({command: 'spawn-worker', functionName: functionName, args: args, msg: message, uri: uri, isSpawnUri: isSpawnUri, startPaused: startPaused, replyPort: replyPort})));
      } else {
        IsolateNatives._spawnWorker(functionName, uri, args, message, isSpawnUri, startPaused, replyPort, onError);
      }
    }
    static _startNonWorker(functionName, uri, args, message, isSpawnUri, startPaused, replyPort) {
      if (uri != null) {
        dart.throw(new core.UnsupportedError("Currently spawnUri is not supported without web workers."));
      }
      message = _clone(message);
      if (args != null)
        args = core.List$(core.String).from(args);
      exports._globalState.topEventLoop.enqueue(new _IsolateContext(), dart.fn(() => {
        let func = IsolateNatives._getJSFunctionFromName(functionName);
        IsolateNatives._startIsolate(dart.as(func, core.Function), args, message, isSpawnUri, startPaused, replyPort);
      }), 'nonworker start');
    }
    static get currentIsolate() {
      let context = dart.as(_foreign_helper.JS_CURRENT_ISOLATE_CONTEXT(), _IsolateContext);
      return new isolate.Isolate(context.controlPort.sendPort, {pauseCapability: context.pauseCapability, terminateCapability: context.terminateCapability});
    }
    static _startIsolate(topLevel, args, message, isSpawnUri, startPaused, replyTo) {
      let context = dart.as(_foreign_helper.JS_CURRENT_ISOLATE_CONTEXT(), _IsolateContext);
      _js_helper.Primitives.initializeStatics(context.id);
      replyTo.send([_SPAWNED_SIGNAL, context.controlPort.sendPort, context.pauseCapability, context.terminateCapability]);
      function runStartFunction() {
        context.initialized = true;
        if (!dart.notNull(isSpawnUri)) {
          dart.dcall(topLevel, message);
        } else if (dart.is(topLevel, _MainFunctionArgsMessage)) {
          dart.dcall(topLevel, args, message);
        } else if (dart.is(topLevel, _MainFunctionArgs)) {
          dart.dcall(topLevel, args);
        } else {
          dart.dcall(topLevel);
        }
      }
      dart.fn(runStartFunction, dart.void, []);
      if (dart.notNull(startPaused)) {
        context.addPause(context.pauseCapability, context.pauseCapability);
        exports._globalState.topEventLoop.enqueue(context, runStartFunction, 'start isolate');
      } else {
        runStartFunction();
      }
    }
    static _spawnWorker(functionName, uri, args, message, isSpawnUri, startPaused, replyPort, onError) {
      if (uri == null)
        uri = IsolateNatives.thisScript;
      let worker = new Worker(uri);
      let onerrorTrampoline = (function(f, u, c) {
        return function(e) {
          return f(e, u, c);
        };
      })(IsolateNatives.workerOnError, uri, onError);
      worker.onerror = onerrorTrampoline;
      let processWorkerMessageTrampoline = (function(f, a) {
        return function(e) {
          e.onerror = null;
          return f(a, e);
        };
      })(IsolateNatives._processWorkerMessage, worker);
      worker.onmessage = processWorkerMessageTrampoline;
      let o = exports._globalState;
      let workerId = o.nextManagerId;
      o.nextManagerId = dart.notNull(workerId) + 1;
      IsolateNatives.workerIds.set(worker, workerId);
      exports._globalState.managers.set(workerId, worker);
      worker.postMessage(_serializeMessage(dart.map({command: 'start', id: workerId, replyTo: _serializeMessage(replyPort), args: args, msg: _serializeMessage(message), isSpawnUri: isSpawnUri, startPaused: startPaused, functionName: functionName})));
    }
    static workerOnError(event, uri, onError) {
      event.preventDefault();
      let message = dart.as(event.message, core.String);
      if (message == null) {
        message = `Error spawning worker for ${uri}`;
      } else {
        message = `Error spawning worker for ${uri} (${message})`;
      }
      onError(message);
      return true;
    }
  }
  dart.setSignature(IsolateNatives, {
    statics: () => ({
      computeThisScript: [core.String, []],
      computeThisScriptJsshell: [core.String, []],
      computeThisScriptD8: [core.String, []],
      computeThisScriptFromTrace: [core.String, []],
      _getEventData: [dart.dynamic, [dart.dynamic]],
      _processWorkerMessage: [dart.void, [dart.dynamic, dart.dynamic]],
      handleSpawnWorkerRequest: [dart.dynamic, [dart.dynamic]],
      _log: [dart.dynamic, [dart.dynamic]],
      _consoleLog: [dart.void, [dart.dynamic]],
      _getJSFunctionFromName: [dart.dynamic, [core.String]],
      _getJSFunctionName: [core.String, [core.Function]],
      _allocate: [dart.dynamic, [dart.dynamic]],
      spawnFunction: [async.Future$(core.List), [dart.functionType(dart.void, [dart.dynamic]), dart.dynamic, core.bool]],
      spawnUri: [async.Future$(core.List), [core.Uri, core.List$(core.String), dart.dynamic, core.bool]],
      spawn: [async.Future$(core.List), [core.String, core.String, core.List$(core.String), dart.dynamic, core.bool, core.bool, core.bool]],
      _startWorker: [dart.void, [core.String, core.String, core.List$(core.String), dart.dynamic, core.bool, core.bool, isolate.SendPort, dart.functionType(dart.void, [core.String])]],
      _startNonWorker: [dart.void, [core.String, core.String, core.List$(core.String), dart.dynamic, core.bool, core.bool, isolate.SendPort]],
      _startIsolate: [dart.void, [core.Function, core.List$(core.String), dart.dynamic, core.bool, core.bool, isolate.SendPort]],
      _spawnWorker: [dart.void, [dart.dynamic, core.String, core.List$(core.String), dart.dynamic, core.bool, core.bool, isolate.SendPort, dart.functionType(dart.void, [core.String])]],
      workerOnError: [core.bool, [dart.dynamic, core.String, dart.functionType(dart.void, [core.String])]]
    }),
    names: ['computeThisScript', 'computeThisScriptJsshell', 'computeThisScriptD8', 'computeThisScriptFromTrace', '_getEventData', '_processWorkerMessage', 'handleSpawnWorkerRequest', '_log', '_consoleLog', '_getJSFunctionFromName', '_getJSFunctionName', '_allocate', 'spawnFunction', 'spawnUri', 'spawn', '_startWorker', '_startNonWorker', '_startIsolate', '_spawnWorker', 'workerOnError']
  });
  IsolateNatives.enableSpawnWorker = null;
  dart.defineLazyProperties(IsolateNatives, {
    get thisScript() {
      return IsolateNatives.computeThisScript();
    },
    set thisScript(_) {},
    get workerIds() {
      return new (core.Expando$(core.int))();
    }
  });
  const _checkReplyTo = Symbol('_checkReplyTo');
  class _BaseSendPort extends core.Object {
    _BaseSendPort(isolateId) {
      this[_isolateId] = isolateId;
    }
    [_checkReplyTo](replyTo) {
      if (replyTo != null && !dart.is(replyTo, _NativeJsSendPort) && !dart.is(replyTo, _WorkerSendPort)) {
        dart.throw(core.Exception.new("SendPort.send: Illegal replyTo port type"));
      }
    }
  }
  _BaseSendPort[dart.implements] = () => [isolate.SendPort];
  dart.setSignature(_BaseSendPort, {
    constructors: () => ({_BaseSendPort: [_BaseSendPort, [core.int]]}),
    methods: () => ({[_checkReplyTo]: [dart.void, [isolate.SendPort]]})
  });
  const _isClosed = Symbol('_isClosed');
  const _add = Symbol('_add');
  class _NativeJsSendPort extends _BaseSendPort {
    _NativeJsSendPort(receivePort, isolateId) {
      this[_receivePort] = receivePort;
      super._BaseSendPort(isolateId);
    }
    send(message) {
      let isolate = exports._globalState.isolates.get(this[_isolateId]);
      if (isolate == null)
        return;
      if (dart.notNull(this[_receivePort][_isClosed]))
        return;
      let msg = _clone(message);
      if (dart.equals(isolate.controlPort, this[_receivePort])) {
        isolate.handleControlMessage(msg);
        return;
      }
      exports._globalState.topEventLoop.enqueue(isolate, dart.fn((() => {
        if (!dart.notNull(this[_receivePort][_isClosed])) {
          this[_receivePort][_add](msg);
        }
      }).bind(this)), `receive ${message}`);
    }
    ['=='](other) {
      return dart.is(other, _NativeJsSendPort) && dart.equals(this[_receivePort], dart.dload(other, _receivePort));
    }
    get hashCode() {
      return this[_receivePort][_id];
    }
  }
  _NativeJsSendPort[dart.implements] = () => [isolate.SendPort];
  dart.setSignature(_NativeJsSendPort, {
    constructors: () => ({_NativeJsSendPort: [_NativeJsSendPort, [RawReceivePortImpl, core.int]]}),
    methods: () => ({send: [dart.void, [dart.dynamic]]})
  });
  class _WorkerSendPort extends _BaseSendPort {
    _WorkerSendPort(workerId, isolateId, receivePortId) {
      this[_workerId] = workerId;
      this[_receivePortId] = receivePortId;
      super._BaseSendPort(isolateId);
    }
    send(message) {
      let workerMessage = _serializeMessage(dart.map({command: 'message', port: this, msg: message}));
      if (dart.notNull(exports._globalState.isWorker)) {
        exports._globalState.mainManager.postMessage(workerMessage);
      } else {
        let manager = exports._globalState.managers.get(this[_workerId]);
        if (manager != null) {
          manager.postMessage(workerMessage);
        }
      }
    }
    ['=='](other) {
      return dart.is(other, _WorkerSendPort) && dart.equals(this[_workerId], dart.dload(other, _workerId)) && dart.equals(this[_isolateId], dart.dload(other, _isolateId)) && dart.equals(this[_receivePortId], dart.dload(other, _receivePortId));
    }
    get hashCode() {
      return dart.notNull(this[_workerId]) << 16 ^ dart.notNull(this[_isolateId]) << 8 ^ dart.notNull(this[_receivePortId]);
    }
  }
  _WorkerSendPort[dart.implements] = () => [isolate.SendPort];
  dart.setSignature(_WorkerSendPort, {
    constructors: () => ({_WorkerSendPort: [_WorkerSendPort, [core.int, core.int, core.int]]}),
    methods: () => ({send: [dart.void, [dart.dynamic]]})
  });
  const _handler = Symbol('_handler');
  class RawReceivePortImpl extends core.Object {
    RawReceivePortImpl(handler) {
      this[_handler] = handler;
      this[_id] = (() => {
        let x = RawReceivePortImpl._nextFreeId;
        RawReceivePortImpl._nextFreeId = dart.notNull(x) + 1;
        return x;
      })();
      this[_isClosed] = false;
      exports._globalState.currentContext.register(this[_id], this);
    }
    weak(handler) {
      this[_handler] = handler;
      this[_id] = (() => {
        let x = RawReceivePortImpl._nextFreeId;
        RawReceivePortImpl._nextFreeId = dart.notNull(x) + 1;
        return x;
      })();
      this[_isClosed] = false;
      exports._globalState.currentContext.registerWeak(this[_id], this);
    }
    _controlPort() {
      this[_handler] = null;
      this[_id] = 0;
      this[_isClosed] = false;
    }
    set handler(newHandler) {
      this[_handler] = newHandler;
    }
    [_close]() {
      this[_isClosed] = true;
      this[_handler] = null;
    }
    close() {
      if (dart.notNull(this[_isClosed]))
        return;
      this[_isClosed] = true;
      this[_handler] = null;
      exports._globalState.currentContext.unregister(this[_id]);
    }
    [_add](dataEvent) {
      if (dart.notNull(this[_isClosed]))
        return;
      dart.dcall(this[_handler], dataEvent);
    }
    get sendPort() {
      return new _NativeJsSendPort(this, exports._globalState.currentContext.id);
    }
  }
  RawReceivePortImpl[dart.implements] = () => [isolate.RawReceivePort];
  dart.defineNamedConstructor(RawReceivePortImpl, 'weak');
  dart.defineNamedConstructor(RawReceivePortImpl, '_controlPort');
  dart.setSignature(RawReceivePortImpl, {
    constructors: () => ({
      RawReceivePortImpl: [RawReceivePortImpl, [core.Function]],
      weak: [RawReceivePortImpl, [core.Function]],
      _controlPort: [RawReceivePortImpl, []]
    }),
    methods: () => ({
      [_close]: [dart.void, []],
      close: [dart.void, []],
      [_add]: [dart.void, [dart.dynamic]]
    })
  });
  RawReceivePortImpl._nextFreeId = 1;
  const _rawPort = Symbol('_rawPort');
  const _controller = Symbol('_controller');
  class ReceivePortImpl extends async.Stream {
    ReceivePortImpl() {
      this.fromRawReceivePort(new RawReceivePortImpl(null));
    }
    weak() {
      this.fromRawReceivePort(new RawReceivePortImpl.weak(null));
    }
    fromRawReceivePort(rawPort) {
      this[_rawPort] = rawPort;
      this[_controller] = null;
      super.Stream();
      this[_controller] = async.StreamController.new({onCancel: dart.bind(this, 'close'), sync: true});
      this[_rawPort].handler = dart.bind(this[_controller], 'add');
    }
    listen(onData, opts) {
      let onError = opts && 'onError' in opts ? opts.onError : null;
      let onDone = opts && 'onDone' in opts ? opts.onDone : null;
      let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
      return this[_controller].stream.listen(onData, {onError: onError, onDone: onDone, cancelOnError: cancelOnError});
    }
    close() {
      this[_rawPort].close();
      this[_controller].close();
    }
    get sendPort() {
      return this[_rawPort].sendPort;
    }
  }
  ReceivePortImpl[dart.implements] = () => [isolate.ReceivePort];
  dart.defineNamedConstructor(ReceivePortImpl, 'weak');
  dart.defineNamedConstructor(ReceivePortImpl, 'fromRawReceivePort');
  dart.setSignature(ReceivePortImpl, {
    constructors: () => ({
      ReceivePortImpl: [ReceivePortImpl, []],
      weak: [ReceivePortImpl, []],
      fromRawReceivePort: [ReceivePortImpl, [isolate.RawReceivePort]]
    }),
    methods: () => ({
      listen: [async.StreamSubscription, [dart.functionType(dart.void, [dart.dynamic])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
      close: [dart.void, []]
    })
  });
  const _once = Symbol('_once');
  const _inEventLoop = Symbol('_inEventLoop');
  const _handle = Symbol('_handle');
  class TimerImpl extends core.Object {
    TimerImpl(milliseconds, callback) {
      this[_once] = true;
      this[_inEventLoop] = false;
      this[_handle] = null;
      if (milliseconds == 0 && (!dart.notNull(hasTimer()) || dart.notNull(exports._globalState.isWorker))) {
        const internalCallback = (function() {
          this[_handle] = null;
          callback();
        }).bind(this);
        dart.fn(internalCallback, dart.void, []);
        this[_handle] = 1;
        exports._globalState.topEventLoop.enqueue(exports._globalState.currentContext, internalCallback, 'timer');
        this[_inEventLoop] = true;
      } else if (dart.notNull(hasTimer())) {
        const internalCallback = (function() {
          this[_handle] = null;
          leaveJsAsync();
          callback();
        }).bind(this);
        dart.fn(internalCallback, dart.void, []);
        enterJsAsync();
        this[_handle] = self.setTimeout(internalCallback, milliseconds);
      } else {
        dart.assert(dart.notNull(milliseconds) > 0);
        dart.throw(new core.UnsupportedError("Timer greater than 0."));
      }
    }
    periodic(milliseconds, callback) {
      this[_once] = false;
      this[_inEventLoop] = false;
      this[_handle] = null;
      if (dart.notNull(hasTimer())) {
        enterJsAsync();
        this[_handle] = self.setInterval(dart.fn((() => {
          callback(this);
        }).bind(this)), milliseconds);
      } else {
        dart.throw(new core.UnsupportedError("Periodic timer."));
      }
    }
    cancel() {
      if (dart.notNull(hasTimer())) {
        if (dart.notNull(this[_inEventLoop])) {
          dart.throw(new core.UnsupportedError("Timer in event loop cannot be canceled."));
        }
        if (this[_handle] == null)
          return;
        leaveJsAsync();
        if (dart.notNull(this[_once])) {
          self.clearTimeout(this[_handle]);
        } else {
          self.clearInterval(this[_handle]);
        }
        this[_handle] = null;
      } else {
        dart.throw(new core.UnsupportedError("Canceling a timer."));
      }
    }
    get isActive() {
      return this[_handle] != null;
    }
  }
  TimerImpl[dart.implements] = () => [async.Timer];
  dart.defineNamedConstructor(TimerImpl, 'periodic');
  dart.setSignature(TimerImpl, {
    constructors: () => ({
      TimerImpl: [TimerImpl, [core.int, dart.functionType(dart.void, [])]],
      periodic: [TimerImpl, [core.int, dart.functionType(dart.void, [async.Timer])]]
    }),
    methods: () => ({cancel: [dart.void, []]})
  });
  function hasTimer() {
    return self.setTimeout != null;
  }
  dart.fn(hasTimer, core.bool, []);
  class CapabilityImpl extends core.Object {
    CapabilityImpl() {
      this._internal(_js_helper.random64());
    }
    _internal(id) {
      this[_id] = id;
    }
    get hashCode() {
      let hash = this[_id];
      hash = dart.notNull(hash) >> 0 ^ (dart.notNull(hash) / 4294967296)[dartx.truncate]();
      hash = ~dart.notNull(hash) + (dart.notNull(hash) << 15) & 4294967295;
      hash = dart.notNull(hash) ^ dart.notNull(hash) >> 12;
      hash = dart.notNull(hash) * 5 & 4294967295;
      hash = dart.notNull(hash) ^ dart.notNull(hash) >> 4;
      hash = dart.notNull(hash) * 2057 & 4294967295;
      hash = dart.notNull(hash) ^ dart.notNull(hash) >> 16;
      return hash;
    }
    ['=='](other) {
      if (dart.notNull(core.identical(other, this)))
        return true;
      if (dart.is(other, CapabilityImpl)) {
        return core.identical(this[_id], other[_id]);
      }
      return false;
    }
  }
  CapabilityImpl[dart.implements] = () => [isolate.Capability];
  dart.defineNamedConstructor(CapabilityImpl, '_internal');
  dart.setSignature(CapabilityImpl, {
    constructors: () => ({
      CapabilityImpl: [CapabilityImpl, []],
      _internal: [CapabilityImpl, [core.int]]
    }),
    methods: () => ({'==': [core.bool, [core.Object]]})
  });
  // Exports:
  exports.enterJsAsync = enterJsAsync;
  exports.leaveJsAsync = leaveJsAsync;
  exports.isWorker = isWorker;
  exports.startRootIsolate = startRootIsolate;
  exports.IsolateNatives = IsolateNatives;
  exports.RawReceivePortImpl = RawReceivePortImpl;
  exports.ReceivePortImpl = ReceivePortImpl;
  exports.TimerImpl = TimerImpl;
  exports.hasTimer = hasTimer;
  exports.CapabilityImpl = CapabilityImpl;
});
dart_library.library('dart/_js_embedded_names', null, /* Imports */[
  "dart/_runtime"
], /* Lazy imports */[
], function(exports, dart) {
  'use strict';
  let dartx = dart.dartx;
  const DISPATCH_PROPERTY_NAME = "dispatchPropertyName";
  const TYPE_INFORMATION = 'typeInformation';
  const GLOBAL_FUNCTIONS = 'globalFunctions';
  const STATICS = 'statics';
  const INTERCEPTED_NAMES = 'interceptedNames';
  const MANGLED_GLOBAL_NAMES = 'mangledGlobalNames';
  const MANGLED_NAMES = 'mangledNames';
  const LIBRARIES = 'libraries';
  const FINISHED_CLASSES = 'finishedClasses';
  const ALL_CLASSES = 'allClasses';
  const METADATA = 'metadata';
  const INTERCEPTORS_BY_TAG = 'interceptorsByTag';
  const LEAF_TAGS = 'leafTags';
  const LAZIES = 'lazies';
  const GET_ISOLATE_TAG = 'getIsolateTag';
  const ISOLATE_TAG = 'isolateTag';
  const CURRENT_SCRIPT = 'currentScript';
  const DEFERRED_LIBRARY_URIS = 'deferredLibraryUris';
  const DEFERRED_LIBRARY_HASHES = 'deferredLibraryHashes';
  const INITIALIZE_LOADED_HUNK = 'initializeLoadedHunk';
  const IS_HUNK_LOADED = 'isHunkLoaded';
  const IS_HUNK_INITIALIZED = 'isHunkInitialized';
  const DEFERRED_INITIALIZED = 'deferredInitialized';
  const CLASS_ID_EXTRACTOR = 'classIdExtractor';
  const CLASS_FIELDS_EXTRACTOR = 'classFieldsExtractor';
  const INSTANCE_FROM_CLASS_ID = "instanceFromClassId";
  const INITIALIZE_EMPTY_INSTANCE = "initializeEmptyInstance";
  const TYPEDEF_TYPE_PROPERTY_NAME = "$typedefType";
  const TYPEDEF_PREDICATE_PROPERTY_NAME = "$$isTypedef";
  const NATIVE_SUPERCLASS_TAG_NAME = "$nativeSuperclassTag";
  const MAP_TYPE_TO_INTERCEPTOR = "mapTypeToInterceptor";
  // Exports:
  exports.DISPATCH_PROPERTY_NAME = DISPATCH_PROPERTY_NAME;
  exports.TYPE_INFORMATION = TYPE_INFORMATION;
  exports.GLOBAL_FUNCTIONS = GLOBAL_FUNCTIONS;
  exports.STATICS = STATICS;
  exports.INTERCEPTED_NAMES = INTERCEPTED_NAMES;
  exports.MANGLED_GLOBAL_NAMES = MANGLED_GLOBAL_NAMES;
  exports.MANGLED_NAMES = MANGLED_NAMES;
  exports.LIBRARIES = LIBRARIES;
  exports.FINISHED_CLASSES = FINISHED_CLASSES;
  exports.ALL_CLASSES = ALL_CLASSES;
  exports.METADATA = METADATA;
  exports.INTERCEPTORS_BY_TAG = INTERCEPTORS_BY_TAG;
  exports.LEAF_TAGS = LEAF_TAGS;
  exports.LAZIES = LAZIES;
  exports.GET_ISOLATE_TAG = GET_ISOLATE_TAG;
  exports.ISOLATE_TAG = ISOLATE_TAG;
  exports.CURRENT_SCRIPT = CURRENT_SCRIPT;
  exports.DEFERRED_LIBRARY_URIS = DEFERRED_LIBRARY_URIS;
  exports.DEFERRED_LIBRARY_HASHES = DEFERRED_LIBRARY_HASHES;
  exports.INITIALIZE_LOADED_HUNK = INITIALIZE_LOADED_HUNK;
  exports.IS_HUNK_LOADED = IS_HUNK_LOADED;
  exports.IS_HUNK_INITIALIZED = IS_HUNK_INITIALIZED;
  exports.DEFERRED_INITIALIZED = DEFERRED_INITIALIZED;
  exports.CLASS_ID_EXTRACTOR = CLASS_ID_EXTRACTOR;
  exports.CLASS_FIELDS_EXTRACTOR = CLASS_FIELDS_EXTRACTOR;
  exports.INSTANCE_FROM_CLASS_ID = INSTANCE_FROM_CLASS_ID;
  exports.INITIALIZE_EMPTY_INSTANCE = INITIALIZE_EMPTY_INSTANCE;
  exports.TYPEDEF_TYPE_PROPERTY_NAME = TYPEDEF_TYPE_PROPERTY_NAME;
  exports.TYPEDEF_PREDICATE_PROPERTY_NAME = TYPEDEF_PREDICATE_PROPERTY_NAME;
  exports.NATIVE_SUPERCLASS_TAG_NAME = NATIVE_SUPERCLASS_TAG_NAME;
  exports.MAP_TYPE_TO_INTERCEPTOR = MAP_TYPE_TO_INTERCEPTOR;
});
dart_library.library('dart/_js_helper', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/collection',
  'dart/_interceptors',
  'dart/_foreign_helper'
], /* Lazy imports */[
], function(exports, dart, core, collection, _interceptors, _foreign_helper) {
  'use strict';
  let dartx = dart.dartx;
  class NoThrows extends core.Object {
    NoThrows() {
    }
  }
  dart.setSignature(NoThrows, {
    constructors: () => ({NoThrows: [NoThrows, []]})
  });
  class NoInline extends core.Object {
    NoInline() {
    }
  }
  dart.setSignature(NoInline, {
    constructors: () => ({NoInline: [NoInline, []]})
  });
  class Native extends core.Object {
    Native(name) {
      this.name = name;
    }
  }
  dart.setSignature(Native, {
    constructors: () => ({Native: [Native, [core.String]]})
  });
  class JsPeerInterface extends core.Object {
    JsPeerInterface(opts) {
      let name = opts && 'name' in opts ? opts.name : null;
      this.name = name;
    }
  }
  dart.setSignature(JsPeerInterface, {
    constructors: () => ({JsPeerInterface: [JsPeerInterface, [], {name: core.String}]})
  });
  class SupportJsExtensionMethods extends core.Object {
    SupportJsExtensionMethods() {
    }
  }
  dart.setSignature(SupportJsExtensionMethods, {
    constructors: () => ({SupportJsExtensionMethods: [SupportJsExtensionMethods, []]})
  });
  function defineProperty(obj, property, value) {
    Object.defineProperty(obj, property, {value: value, enumerable: false, writable: true, configurable: true});
  }
  dart.fn(defineProperty, dart.void, [dart.dynamic, core.String, dart.dynamic]);
  const _nativeRegExp = Symbol('_nativeRegExp');
  function regExpGetNative(regexp) {
    return regexp[_nativeRegExp];
  }
  dart.fn(regExpGetNative, () => dart.definiteFunctionType(dart.dynamic, [JSSyntaxRegExp]));
  const _nativeGlobalVersion = Symbol('_nativeGlobalVersion');
  function regExpGetGlobalNative(regexp) {
    let nativeRegexp = regexp[_nativeGlobalVersion];
    nativeRegexp.lastIndex = 0;
    return nativeRegexp;
  }
  dart.fn(regExpGetGlobalNative, () => dart.definiteFunctionType(dart.dynamic, [JSSyntaxRegExp]));
  const _nativeAnchoredVersion = Symbol('_nativeAnchoredVersion');
  function regExpCaptureCount(regexp) {
    let nativeAnchoredRegExp = regexp[_nativeAnchoredVersion];
    let match = nativeAnchoredRegExp.exec('');
    return dart.as(dart.dsend(dart.dload(match, 'length'), '-', 2), core.int);
  }
  dart.fn(regExpCaptureCount, () => dart.definiteFunctionType(core.int, [JSSyntaxRegExp]));
  const _nativeGlobalRegExp = Symbol('_nativeGlobalRegExp');
  const _nativeAnchoredRegExp = Symbol('_nativeAnchoredRegExp');
  const _isMultiLine = Symbol('_isMultiLine');
  const _isCaseSensitive = Symbol('_isCaseSensitive');
  const _execGlobal = Symbol('_execGlobal');
  const _execAnchored = Symbol('_execAnchored');
  class JSSyntaxRegExp extends core.Object {
    toString() {
      return `RegExp/${this.pattern}/`;
    }
    JSSyntaxRegExp(source, opts) {
      let multiLine = opts && 'multiLine' in opts ? opts.multiLine : false;
      let caseSensitive = opts && 'caseSensitive' in opts ? opts.caseSensitive : true;
      this.pattern = source;
      this[_nativeRegExp] = JSSyntaxRegExp.makeNative(source, multiLine, caseSensitive, false);
      this[_nativeGlobalRegExp] = null;
      this[_nativeAnchoredRegExp] = null;
    }
    get [_nativeGlobalVersion]() {
      if (this[_nativeGlobalRegExp] != null)
        return this[_nativeGlobalRegExp];
      return this[_nativeGlobalRegExp] = JSSyntaxRegExp.makeNative(this.pattern, this[_isMultiLine], this[_isCaseSensitive], true);
    }
    get [_nativeAnchoredVersion]() {
      if (this[_nativeAnchoredRegExp] != null)
        return this[_nativeAnchoredRegExp];
      return this[_nativeAnchoredRegExp] = JSSyntaxRegExp.makeNative(`${this.pattern}|()`, this[_isMultiLine], this[_isCaseSensitive], true);
    }
    get [_isMultiLine]() {
      return this[_nativeRegExp].multiline;
    }
    get [_isCaseSensitive]() {
      return !this[_nativeRegExp].ignoreCase;
    }
    static makeNative(source, multiLine, caseSensitive, global) {
      checkString(source);
      let m = dart.notNull(multiLine) ? 'm' : '';
      let i = dart.notNull(caseSensitive) ? '' : 'i';
      let g = dart.notNull(global) ? 'g' : '';
      let regexp = (function() {
        try {
          return new RegExp(source, m + i + g);
        } catch (e) {
          return e;
        }

      })();
      if (regexp instanceof RegExp)
        return regexp;
      let errorMessage = String(regexp);
      dart.throw(new core.FormatException(`Illegal RegExp pattern: ${source}, ${errorMessage}`));
    }
    firstMatch(string) {
      let m = dart.as(this[_nativeRegExp].exec(checkString(string)), core.List$(core.String));
      if (m == null)
        return null;
      return new _MatchImplementation(this, m);
    }
    hasMatch(string) {
      return this[_nativeRegExp].test(checkString(string));
    }
    stringMatch(string) {
      let match = this.firstMatch(string);
      if (match != null)
        return match.group(0);
      return null;
    }
    allMatches(string, start) {
      if (start === void 0)
        start = 0;
      checkString(string);
      checkInt(start);
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(string[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, string[dartx.length]));
      }
      return new _AllMatchesIterable(this, string, start);
    }
    [_execGlobal](string, start) {
      let regexp = this[_nativeGlobalVersion];
      regexp.lastIndex = start;
      let match = dart.as(regexp.exec(string), core.List);
      if (match == null)
        return null;
      return new _MatchImplementation(this, dart.as(match, core.List$(core.String)));
    }
    [_execAnchored](string, start) {
      let regexp = this[_nativeAnchoredVersion];
      regexp.lastIndex = start;
      let match = dart.as(regexp.exec(string), core.List);
      if (match == null)
        return null;
      if (match[dartx.get](dart.notNull(match[dartx.length]) - 1) != null)
        return null;
      match[dartx.length] = dart.notNull(match[dartx.length]) - 1;
      return new _MatchImplementation(this, dart.as(match, core.List$(core.String)));
    }
    matchAsPrefix(string, start) {
      if (start === void 0)
        start = 0;
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(string[dartx.length])) {
        dart.throw(new core.RangeError.range(start, 0, string[dartx.length]));
      }
      return this[_execAnchored](string, start);
    }
    get isMultiLine() {
      return this[_isMultiLine];
    }
    get isCaseSensitive() {
      return this[_isCaseSensitive];
    }
  }
  JSSyntaxRegExp[dart.implements] = () => [core.RegExp];
  dart.setSignature(JSSyntaxRegExp, {
    constructors: () => ({JSSyntaxRegExp: [JSSyntaxRegExp, [core.String], {multiLine: core.bool, caseSensitive: core.bool}]}),
    methods: () => ({
      firstMatch: [core.Match, [core.String]],
      hasMatch: [core.bool, [core.String]],
      stringMatch: [core.String, [core.String]],
      allMatches: [core.Iterable$(core.Match), [core.String], [core.int]],
      [_execGlobal]: [core.Match, [core.String, core.int]],
      [_execAnchored]: [core.Match, [core.String, core.int]],
      matchAsPrefix: [core.Match, [core.String], [core.int]]
    }),
    statics: () => ({makeNative: [dart.dynamic, [core.String, core.bool, core.bool, core.bool]]}),
    names: ['makeNative']
  });
  dart.defineExtensionMembers(JSSyntaxRegExp, ['allMatches', 'matchAsPrefix']);
  const _match = Symbol('_match');
  class _MatchImplementation extends core.Object {
    _MatchImplementation(pattern, match) {
      this.pattern = pattern;
      this[_match] = match;
      dart.assert(typeof this[_match].input == 'string');
      dart.assert(typeof this[_match].index == 'number');
    }
    get input() {
      return this[_match].input;
    }
    get start() {
      return this[_match].index;
    }
    get end() {
      return dart.notNull(this.start) + dart.notNull(this[_match][dartx.get](0)[dartx.length]);
    }
    group(index) {
      return this[_match][dartx.get](index);
    }
    get(index) {
      return this.group(index);
    }
    get groupCount() {
      return dart.notNull(this[_match][dartx.length]) - 1;
    }
    groups(groups) {
      let out = dart.list([], core.String);
      for (let i of groups) {
        out[dartx.add](this.group(i));
      }
      return out;
    }
  }
  _MatchImplementation[dart.implements] = () => [core.Match];
  dart.setSignature(_MatchImplementation, {
    constructors: () => ({_MatchImplementation: [_MatchImplementation, [core.Pattern, core.List$(core.String)]]}),
    methods: () => ({
      group: [core.String, [core.int]],
      get: [core.String, [core.int]],
      groups: [core.List$(core.String), [core.List$(core.int)]]
    })
  });
  const _re = Symbol('_re');
  const _string = Symbol('_string');
  const _start = Symbol('_start');
  class _AllMatchesIterable extends collection.IterableBase$(core.Match) {
    _AllMatchesIterable(re, string, start) {
      this[_re] = re;
      this[_string] = string;
      this[_start] = start;
      super.IterableBase();
    }
    get iterator() {
      return new _AllMatchesIterator(this[_re], this[_string], this[_start]);
    }
  }
  dart.setSignature(_AllMatchesIterable, {
    constructors: () => ({_AllMatchesIterable: [_AllMatchesIterable, [JSSyntaxRegExp, core.String, core.int]]})
  });
  dart.defineExtensionMembers(_AllMatchesIterable, ['iterator']);
  const _regExp = Symbol('_regExp');
  const _nextIndex = Symbol('_nextIndex');
  const _current = Symbol('_current');
  class _AllMatchesIterator extends core.Object {
    _AllMatchesIterator(regExp, string, nextIndex) {
      this[_regExp] = regExp;
      this[_string] = string;
      this[_nextIndex] = nextIndex;
      this[_current] = null;
    }
    get current() {
      return this[_current];
    }
    moveNext() {
      if (this[_string] == null)
        return false;
      if (dart.notNull(this[_nextIndex]) <= dart.notNull(this[_string][dartx.length])) {
        let match = this[_regExp][_execGlobal](this[_string], this[_nextIndex]);
        if (match != null) {
          this[_current] = match;
          let nextIndex = match.end;
          if (match.start == nextIndex) {
            nextIndex = dart.notNull(nextIndex) + 1;
          }
          this[_nextIndex] = nextIndex;
          return true;
        }
      }
      this[_current] = null;
      this[_string] = null;
      return false;
    }
  }
  _AllMatchesIterator[dart.implements] = () => [core.Iterator$(core.Match)];
  dart.setSignature(_AllMatchesIterator, {
    constructors: () => ({_AllMatchesIterator: [_AllMatchesIterator, [JSSyntaxRegExp, core.String, core.int]]}),
    methods: () => ({moveNext: [core.bool, []]})
  });
  function firstMatchAfter(regExp, string, start) {
    return regExp[_execGlobal](string, start);
  }
  dart.fn(firstMatchAfter, core.Match, [JSSyntaxRegExp, core.String, core.int]);
  class StringMatch extends core.Object {
    StringMatch(start, input, pattern) {
      this.start = start;
      this.input = input;
      this.pattern = pattern;
    }
    get end() {
      return dart.notNull(this.start) + dart.notNull(this.pattern[dartx.length]);
    }
    get(g) {
      return this.group(g);
    }
    get groupCount() {
      return 0;
    }
    group(group_) {
      if (group_ != 0) {
        dart.throw(new core.RangeError.value(group_));
      }
      return this.pattern;
    }
    groups(groups_) {
      let result = core.List$(core.String).new();
      for (let g of groups_) {
        result[dartx.add](this.group(g));
      }
      return result;
    }
  }
  StringMatch[dart.implements] = () => [core.Match];
  dart.setSignature(StringMatch, {
    constructors: () => ({StringMatch: [StringMatch, [core.int, core.String, core.String]]}),
    methods: () => ({
      get: [core.String, [core.int]],
      group: [core.String, [core.int]],
      groups: [core.List$(core.String), [core.List$(core.int)]]
    })
  });
  function allMatchesInStringUnchecked(needle, haystack, startIndex) {
    let result = core.List$(core.Match).new();
    let length = haystack[dartx.length];
    let patternLength = needle[dartx.length];
    while (true) {
      let position = haystack[dartx.indexOf](needle, startIndex);
      if (position == -1) {
        break;
      }
      result[dartx.add](new StringMatch(position, haystack, needle));
      let endIndex = dart.notNull(position) + dart.notNull(patternLength);
      if (endIndex == length) {
        break;
      } else if (position == endIndex) {
        startIndex = dart.notNull(startIndex) + 1;
      } else {
        startIndex = endIndex;
      }
    }
    return result;
  }
  dart.fn(allMatchesInStringUnchecked, core.List$(core.Match), [core.String, core.String, core.int]);
  function stringContainsUnchecked(receiver, other, startIndex) {
    if (typeof other == 'string') {
      return !dart.equals(dart.dsend(receiver, 'indexOf', other, startIndex), -1);
    } else if (dart.is(other, JSSyntaxRegExp)) {
      return dart.dsend(other, 'hasMatch', dart.dsend(receiver, 'substring', startIndex));
    } else {
      let substr = dart.dsend(receiver, 'substring', startIndex);
      return dart.dload(dart.dsend(other, 'allMatches', substr), 'isNotEmpty');
    }
  }
  dart.fn(stringContainsUnchecked);
  function stringReplaceJS(receiver, replacer, to) {
    to = to.replace(/\$/g, "$$$$");
    return receiver.replace(replacer, to);
  }
  dart.fn(stringReplaceJS);
  function stringReplaceFirstRE(receiver, regexp, to, startIndex) {
    let match = dart.dsend(regexp, _execGlobal, receiver, startIndex);
    if (match == null)
      return receiver;
    let start = dart.dload(match, 'start');
    let end = dart.dload(match, 'end');
    return `${dart.dsend(receiver, 'substring', 0, start)}${to}${dart.dsend(receiver, 'substring', end)}`;
  }
  dart.fn(stringReplaceFirstRE);
  const ESCAPE_REGEXP = '[[\\]{}()*+?.\\\\^$|]';
  function stringReplaceAllUnchecked(receiver, from, to) {
    checkString(to);
    if (typeof from == 'string') {
      if (dart.equals(from, "")) {
        if (dart.equals(receiver, "")) {
          return to;
        } else {
          let result = new core.StringBuffer();
          let length = dart.as(dart.dload(receiver, 'length'), core.int);
          result.write(to);
          for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
            result.write(dart.dindex(receiver, i));
            result.write(to);
          }
          return dart.toString(result);
        }
      } else {
        let quoter = new RegExp(ESCAPE_REGEXP, 'g');
        let quoted = from.replace(quoter, "\\$&");
        let replacer = new RegExp(quoted, 'g');
        return stringReplaceJS(receiver, replacer, to);
      }
    } else if (dart.is(from, JSSyntaxRegExp)) {
      let re = regExpGetGlobalNative(dart.as(from, JSSyntaxRegExp));
      return stringReplaceJS(receiver, re, to);
    } else {
      checkNull(from);
      dart.throw("String.replaceAll(Pattern) UNIMPLEMENTED");
    }
  }
  dart.fn(stringReplaceAllUnchecked);
  function _matchString(match) {
    return match.get(0);
  }
  dart.fn(_matchString, core.String, [core.Match]);
  function _stringIdentity(string) {
    return string;
  }
  dart.fn(_stringIdentity, core.String, [core.String]);
  function stringReplaceAllFuncUnchecked(receiver, pattern, onMatch, onNonMatch) {
    if (!dart.is(pattern, core.Pattern)) {
      dart.throw(new core.ArgumentError(`${pattern} is not a Pattern`));
    }
    if (onMatch == null)
      onMatch = _matchString;
    if (onNonMatch == null)
      onNonMatch = _stringIdentity;
    if (typeof pattern == 'string') {
      return stringReplaceAllStringFuncUnchecked(receiver, pattern, onMatch, onNonMatch);
    }
    let buffer = new core.StringBuffer();
    let startIndex = 0;
    for (let match of dart.as(dart.dsend(pattern, 'allMatches', receiver), core.Iterable$(core.Match))) {
      buffer.write(dart.dcall(onNonMatch, dart.dsend(receiver, 'substring', startIndex, match.start)));
      buffer.write(dart.dcall(onMatch, match));
      startIndex = match.end;
    }
    buffer.write(dart.dcall(onNonMatch, dart.dsend(receiver, 'substring', startIndex)));
    return dart.toString(buffer);
  }
  dart.fn(stringReplaceAllFuncUnchecked);
  function stringReplaceAllEmptyFuncUnchecked(receiver, onMatch, onNonMatch) {
    let buffer = new core.StringBuffer();
    let length = dart.as(dart.dload(receiver, 'length'), core.int);
    let i = 0;
    buffer.write(dart.dcall(onNonMatch, ""));
    while (dart.notNull(i) < dart.notNull(length)) {
      buffer.write(dart.dcall(onMatch, new StringMatch(i, dart.as(receiver, core.String), "")));
      let code = dart.as(dart.dsend(receiver, 'codeUnitAt', i), core.int);
      if ((dart.notNull(code) & ~1023) == 55296 && dart.notNull(length) > dart.notNull(i) + 1) {
        code = dart.as(dart.dsend(receiver, 'codeUnitAt', dart.notNull(i) + 1), core.int);
        if ((dart.notNull(code) & ~1023) == 56320) {
          buffer.write(dart.dcall(onNonMatch, dart.dsend(receiver, 'substring', i, dart.notNull(i) + 2)));
          i = dart.notNull(i) + 2;
          continue;
        }
      }
      buffer.write(dart.dcall(onNonMatch, dart.dindex(receiver, i)));
      i = dart.notNull(i) + 1;
    }
    buffer.write(dart.dcall(onMatch, new StringMatch(i, dart.as(receiver, core.String), "")));
    buffer.write(dart.dcall(onNonMatch, ""));
    return dart.toString(buffer);
  }
  dart.fn(stringReplaceAllEmptyFuncUnchecked);
  function stringReplaceAllStringFuncUnchecked(receiver, pattern, onMatch, onNonMatch) {
    let patternLength = dart.as(dart.dload(pattern, 'length'), core.int);
    if (patternLength == 0) {
      return stringReplaceAllEmptyFuncUnchecked(receiver, onMatch, onNonMatch);
    }
    let length = dart.as(dart.dload(receiver, 'length'), core.int);
    let buffer = new core.StringBuffer();
    let startIndex = 0;
    while (dart.notNull(startIndex) < dart.notNull(length)) {
      let position = dart.as(dart.dsend(receiver, 'indexOf', pattern, startIndex), core.int);
      if (position == -1) {
        break;
      }
      buffer.write(dart.dcall(onNonMatch, dart.dsend(receiver, 'substring', startIndex, position)));
      buffer.write(dart.dcall(onMatch, new StringMatch(position, dart.as(receiver, core.String), dart.as(pattern, core.String))));
      startIndex = dart.notNull(position) + dart.notNull(patternLength);
    }
    buffer.write(dart.dcall(onNonMatch, dart.dsend(receiver, 'substring', startIndex)));
    return dart.toString(buffer);
  }
  dart.fn(stringReplaceAllStringFuncUnchecked);
  function stringReplaceFirstUnchecked(receiver, from, to, startIndex) {
    if (startIndex === void 0)
      startIndex = 0;
    if (typeof from == 'string') {
      let index = dart.dsend(receiver, 'indexOf', from, startIndex);
      if (dart.notNull(dart.as(dart.dsend(index, '<', 0), core.bool)))
        return receiver;
      return `${dart.dsend(receiver, 'substring', 0, index)}${to}` + `${dart.dsend(receiver, 'substring', dart.dsend(index, '+', dart.dload(from, 'length')))}`;
    } else if (dart.is(from, JSSyntaxRegExp)) {
      return startIndex == 0 ? stringReplaceJS(receiver, regExpGetNative(dart.as(from, JSSyntaxRegExp)), to) : stringReplaceFirstRE(receiver, from, to, startIndex);
    } else {
      checkNull(from);
      dart.throw("String.replace(Pattern) UNIMPLEMENTED");
    }
  }
  dart.fn(stringReplaceFirstUnchecked, dart.dynamic, [dart.dynamic, dart.dynamic, dart.dynamic], [core.int]);
  function stringJoinUnchecked(array, separator) {
    return array.join(separator);
  }
  dart.fn(stringJoinUnchecked);
  function getRuntimeType(object) {
    return dart.as(dart.realRuntimeType(object), core.Type);
  }
  dart.fn(getRuntimeType, core.Type, [dart.dynamic]);
  function getIndex(array, index) {
    dart.assert(isJsArray(array));
    return array[index];
  }
  dart.fn(getIndex, dart.dynamic, [dart.dynamic, core.int]);
  function getLength(array) {
    dart.assert(isJsArray(array));
    return array.length;
  }
  dart.fn(getLength, core.int, [dart.dynamic]);
  function isJsArray(value) {
    return dart.is(value, _interceptors.JSArray);
  }
  dart.fn(isJsArray, core.bool, [dart.dynamic]);
  class _Patch extends core.Object {
    _Patch() {
    }
  }
  dart.setSignature(_Patch, {
    constructors: () => ({_Patch: [_Patch, []]})
  });
  const patch = dart.const(new _Patch());
  class InternalMap extends core.Object {}
  class Primitives extends core.Object {
    static initializeStatics(id) {
      Primitives.mirrorFunctionCacheName = dart.notNull(Primitives.mirrorFunctionCacheName) + `_${id}`;
      Primitives.mirrorInvokeCacheName = dart.notNull(Primitives.mirrorInvokeCacheName) + `_${id}`;
    }
    static objectHashCode(object) {
      let hash = dart.as(object.$identityHash, core.int);
      if (hash == null) {
        hash = Math.random() * 0x3fffffff | 0;
        object.$identityHash = hash;
      }
      return hash;
    }
    static _throwFormatException(string) {
      dart.throw(new core.FormatException(string));
    }
    static parseInt(source, radix, handleError) {
      if (handleError == null)
        handleError = dart.fn(s => dart.as(Primitives._throwFormatException(dart.as(s, core.String)), core.int), core.int, [dart.dynamic]);
      checkString(source);
      let match = /^\s*[+-]?((0x[a-f0-9]+)|(\d+)|([a-z0-9]+))\s*$/i.exec(source);
      let digitsIndex = 1;
      let hexIndex = 2;
      let decimalIndex = 3;
      let nonDecimalHexIndex = 4;
      if (radix == null) {
        radix = 10;
        if (match != null) {
          if (dart.dindex(match, hexIndex) != null) {
            return parseInt(source, 16);
          }
          if (dart.dindex(match, decimalIndex) != null) {
            return parseInt(source, 10);
          }
          return handleError(source);
        }
      } else {
        if (!(typeof radix == 'number'))
          dart.throw(new core.ArgumentError("Radix is not an integer"));
        if (dart.notNull(radix) < 2 || dart.notNull(radix) > 36) {
          dart.throw(new core.RangeError(`Radix ${radix} not in range 2..36`));
        }
        if (match != null) {
          if (radix == 10 && dart.dindex(match, decimalIndex) != null) {
            return parseInt(source, 10);
          }
          if (dart.notNull(radix) < 10 || dart.dindex(match, decimalIndex) == null) {
            let maxCharCode = null;
            if (dart.notNull(radix) <= 10) {
              maxCharCode = 48 + dart.notNull(radix) - 1;
            } else {
              maxCharCode = 97 + dart.notNull(radix) - 10 - 1;
            }
            let digitsPart = dart.as(dart.dindex(match, digitsIndex), core.String);
            for (let i = 0; dart.notNull(i) < dart.notNull(digitsPart[dartx.length]); i = dart.notNull(i) + 1) {
              let characterCode = dart.notNull(digitsPart[dartx.codeUnitAt](0)) | 32;
              if (dart.notNull(digitsPart[dartx.codeUnitAt](i)) > dart.notNull(maxCharCode)) {
                return handleError(source);
              }
            }
          }
        }
      }
      if (match == null)
        return handleError(source);
      return parseInt(source, radix);
    }
    static parseDouble(source, handleError) {
      checkString(source);
      if (handleError == null)
        handleError = dart.fn(s => dart.as(Primitives._throwFormatException(dart.as(s, core.String)), core.double), core.double, [dart.dynamic]);
      if (!/^\s*[+-]?(?:Infinity|NaN|(?:\.\d+|\d+(?:\.\d*)?)(?:[eE][+-]?\d+)?)\s*$/.test(source)) {
        return handleError(source);
      }
      let result = parseFloat(source);
      if (dart.notNull(result[dartx.isNaN])) {
        let trimmed = source[dartx.trim]();
        if (trimmed == 'NaN' || trimmed == '+NaN' || trimmed == '-NaN') {
          return result;
        }
        return handleError(source);
      }
      return result;
    }
    static objectTypeName(object) {
      return dart.toString(getRuntimeType(object));
    }
    static objectToString(object) {
      let name = dart.typeName(dart.realRuntimeType(object));
      return `Instance of '${name}'`;
    }
    static dateNow() {
      return Date.now();
    }
    static initTicker() {
      if (Primitives.timerFrequency != null)
        return;
      Primitives.timerFrequency = 1000;
      Primitives.timerTicks = Primitives.dateNow;
      if (typeof window == "undefined")
        return;
      let jsWindow = window;
      if (jsWindow == null)
        return;
      let performance = jsWindow.performance;
      if (performance == null)
        return;
      if (typeof performance.now != "function")
        return;
      Primitives.timerFrequency = 1000000;
      Primitives.timerTicks = dart.fn(() => (1000 * performance.now())[dartx.floor](), core.int, []);
    }
    static get isD8() {
      return typeof version == "function" && typeof os == "object" && "system" in os;
    }
    static get isJsshell() {
      return typeof version == "function" && typeof system == "function";
    }
    static currentUri() {
      if (!!self.location) {
        return self.location.href;
      }
      return null;
    }
    static _fromCharCodeApply(array) {
      let result = "";
      let kMaxApply = 500;
      let end = array[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + dart.notNull(kMaxApply)) {
        let subarray = null;
        if (dart.notNull(end) <= dart.notNull(kMaxApply)) {
          subarray = array;
        } else {
          subarray = array.slice(i, dart.notNull(i) + dart.notNull(kMaxApply) < dart.notNull(end) ? dart.notNull(i) + dart.notNull(kMaxApply) : end);
        }
        result = result + String.fromCharCode.apply(null, subarray);
      }
      return result;
    }
    static stringFromCodePoints(codePoints) {
      let a = dart.list([], core.int);
      for (let i of dart.as(codePoints, core.Iterable)) {
        if (!(typeof i == 'number'))
          dart.throw(new core.ArgumentError(i));
        if (dart.notNull(dart.as(dart.dsend(i, '<=', 65535), core.bool))) {
          a[dartx.add](dart.as(i, core.int));
        } else if (dart.notNull(dart.as(dart.dsend(i, '<=', 1114111), core.bool))) {
          a[dartx.add](dart.asInt((55296)[dartx['+']](dart.as(dart.dsend(dart.dsend(dart.dsend(i, '-', 65536), '>>', 10), '&', 1023), core.num))));
          a[dartx.add](dart.asInt((56320)[dartx['+']](dart.as(dart.dsend(i, '&', 1023), core.num))));
        } else {
          dart.throw(new core.ArgumentError(i));
        }
      }
      return Primitives._fromCharCodeApply(a);
    }
    static stringFromCharCodes(charCodes) {
      for (let i of dart.as(charCodes, core.Iterable)) {
        if (!(typeof i == 'number'))
          dart.throw(new core.ArgumentError(i));
        if (dart.notNull(dart.as(dart.dsend(i, '<', 0), core.bool)))
          dart.throw(new core.ArgumentError(i));
        if (dart.notNull(dart.as(dart.dsend(i, '>', 65535), core.bool)))
          return Primitives.stringFromCodePoints(charCodes);
      }
      return Primitives._fromCharCodeApply(dart.as(charCodes, core.List$(core.int)));
    }
    static stringFromCharCode(charCode) {
      if (0 <= dart.notNull(dart.as(charCode, core.num))) {
        if (dart.notNull(dart.as(dart.dsend(charCode, '<=', 65535), core.bool))) {
          return String.fromCharCode(charCode);
        }
        if (dart.notNull(dart.as(dart.dsend(charCode, '<=', 1114111), core.bool))) {
          let bits = dart.dsend(charCode, '-', 65536);
          let low = (56320)[dartx['|']](dart.as(dart.dsend(bits, '&', 1023), core.int));
          let high = (55296)[dartx['|']](dart.as(dart.dsend(bits, '>>', 10), core.int));
          return String.fromCharCode(high, low);
        }
      }
      dart.throw(new core.RangeError.range(dart.as(charCode, core.num), 0, 1114111));
    }
    static stringConcatUnchecked(string1, string2) {
      return _foreign_helper.JS_STRING_CONCAT(string1, string2);
    }
    static flattenString(str) {
      return str.charCodeAt(0) == 0 ? str : str;
    }
    static getTimeZoneName(receiver) {
      let d = Primitives.lazyAsJsDate(receiver);
      let match = dart.as(/\((.*)\)/.exec(d.toString()), core.List);
      if (match != null)
        return dart.as(match[dartx.get](1), core.String);
      match = dart.as(/^[A-Z,a-z]{3}\s[A-Z,a-z]{3}\s\d+\s\d{2}:\d{2}:\d{2}\s([A-Z]{3,5})\s\d{4}$/.exec(d.toString()), core.List);
      if (match != null)
        return dart.as(match[dartx.get](1), core.String);
      match = dart.as(/(?:GMT|UTC)[+-]\d{4}/.exec(d.toString()), core.List);
      if (match != null)
        return dart.as(match[dartx.get](0), core.String);
      return "";
    }
    static getTimeZoneOffsetInMinutes(receiver) {
      return -Primitives.lazyAsJsDate(receiver).getTimezoneOffset();
    }
    static valueFromDecomposedDate(years, month, day, hours, minutes, seconds, milliseconds, isUtc) {
      let MAX_MILLISECONDS_SINCE_EPOCH = 8640000000000000;
      checkInt(years);
      checkInt(month);
      checkInt(day);
      checkInt(hours);
      checkInt(minutes);
      checkInt(seconds);
      checkInt(milliseconds);
      checkBool(isUtc);
      let jsMonth = dart.dsend(month, '-', 1);
      let value = null;
      if (dart.notNull(dart.as(isUtc, core.bool))) {
        value = Date.UTC(years, jsMonth, day, hours, minutes, seconds, milliseconds);
      } else {
        value = new Date(years, jsMonth, day, hours, minutes, seconds, milliseconds).valueOf();
      }
      if (dart.notNull(dart.as(dart.dload(value, 'isNaN'), core.bool)) || dart.notNull(dart.as(dart.dsend(value, '<', -dart.notNull(MAX_MILLISECONDS_SINCE_EPOCH)), core.bool)) || dart.notNull(dart.as(dart.dsend(value, '>', MAX_MILLISECONDS_SINCE_EPOCH), core.bool))) {
        return null;
      }
      if (dart.notNull(dart.as(dart.dsend(years, '<=', 0), core.bool)) || dart.notNull(dart.as(dart.dsend(years, '<', 100), core.bool)))
        return Primitives.patchUpY2K(value, years, isUtc);
      return value;
    }
    static patchUpY2K(value, years, isUtc) {
      let date = new Date(value);
      if (dart.notNull(dart.as(isUtc, core.bool))) {
        date.setUTCFullYear(years);
      } else {
        date.setFullYear(years);
      }
      return date.valueOf();
    }
    static lazyAsJsDate(receiver) {
      if (receiver.date === void 0) {
        receiver.date = new Date(dart.dload(receiver, 'millisecondsSinceEpoch'));
      }
      return receiver.date;
    }
    static getYear(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCFullYear() + 0 : Primitives.lazyAsJsDate(receiver).getFullYear() + 0;
    }
    static getMonth(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCMonth() + 1 : Primitives.lazyAsJsDate(receiver).getMonth() + 1;
    }
    static getDay(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCDate() + 0 : Primitives.lazyAsJsDate(receiver).getDate() + 0;
    }
    static getHours(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCHours() + 0 : Primitives.lazyAsJsDate(receiver).getHours() + 0;
    }
    static getMinutes(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCMinutes() + 0 : Primitives.lazyAsJsDate(receiver).getMinutes() + 0;
    }
    static getSeconds(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCSeconds() + 0 : Primitives.lazyAsJsDate(receiver).getSeconds() + 0;
    }
    static getMilliseconds(receiver) {
      return dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCMilliseconds() + 0 : Primitives.lazyAsJsDate(receiver).getMilliseconds() + 0;
    }
    static getWeekday(receiver) {
      let weekday = dart.notNull(dart.as(dart.dload(receiver, 'isUtc'), core.bool)) ? Primitives.lazyAsJsDate(receiver).getUTCDay() + 0 : Primitives.lazyAsJsDate(receiver).getDay() + 0;
      return (dart.notNull(weekday) + 6) % 7 + 1;
    }
    static valueFromDateString(str) {
      if (!(typeof str == 'string'))
        dart.throw(new core.ArgumentError(str));
      let value = Date.parse(str);
      if (dart.notNull(value[dartx.isNaN]))
        dart.throw(new core.ArgumentError(str));
      return value;
    }
    static getProperty(object, key) {
      if (object == null || typeof object == 'boolean' || typeof object == 'number' || typeof object == 'string') {
        dart.throw(new core.ArgumentError(object));
      }
      return object[key];
    }
    static setProperty(object, key, value) {
      if (object == null || typeof object == 'boolean' || typeof object == 'number' || typeof object == 'string') {
        dart.throw(new core.ArgumentError(object));
      }
      object[key] = value;
    }
    static identicalImplementation(a, b) {
      return a == null ? b == null : a === b;
    }
    static extractStackTrace(error) {
      return getTraceFromException(error.$thrownJsError);
    }
  }
  dart.setSignature(Primitives, {
    statics: () => ({
      initializeStatics: [dart.void, [core.int]],
      objectHashCode: [core.int, [dart.dynamic]],
      _throwFormatException: [dart.dynamic, [core.String]],
      parseInt: [core.int, [core.String, core.int, dart.functionType(core.int, [core.String])]],
      parseDouble: [core.double, [core.String, dart.functionType(core.double, [core.String])]],
      objectTypeName: [core.String, [core.Object]],
      objectToString: [core.String, [core.Object]],
      dateNow: [core.int, []],
      initTicker: [dart.void, []],
      currentUri: [core.String, []],
      _fromCharCodeApply: [core.String, [core.List$(core.int)]],
      stringFromCodePoints: [core.String, [dart.dynamic]],
      stringFromCharCodes: [core.String, [dart.dynamic]],
      stringFromCharCode: [core.String, [dart.dynamic]],
      stringConcatUnchecked: [core.String, [core.String, core.String]],
      flattenString: [core.String, [core.String]],
      getTimeZoneName: [core.String, [dart.dynamic]],
      getTimeZoneOffsetInMinutes: [core.int, [dart.dynamic]],
      valueFromDecomposedDate: [dart.dynamic, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      patchUpY2K: [dart.dynamic, [dart.dynamic, dart.dynamic, dart.dynamic]],
      lazyAsJsDate: [dart.dynamic, [dart.dynamic]],
      getYear: [dart.dynamic, [dart.dynamic]],
      getMonth: [dart.dynamic, [dart.dynamic]],
      getDay: [dart.dynamic, [dart.dynamic]],
      getHours: [dart.dynamic, [dart.dynamic]],
      getMinutes: [dart.dynamic, [dart.dynamic]],
      getSeconds: [dart.dynamic, [dart.dynamic]],
      getMilliseconds: [dart.dynamic, [dart.dynamic]],
      getWeekday: [dart.dynamic, [dart.dynamic]],
      valueFromDateString: [dart.dynamic, [dart.dynamic]],
      getProperty: [dart.dynamic, [dart.dynamic, dart.dynamic]],
      setProperty: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      identicalImplementation: [core.bool, [dart.dynamic, dart.dynamic]],
      extractStackTrace: [core.StackTrace, [core.Error]]
    }),
    names: ['initializeStatics', 'objectHashCode', '_throwFormatException', 'parseInt', 'parseDouble', 'objectTypeName', 'objectToString', 'dateNow', 'initTicker', 'currentUri', '_fromCharCodeApply', 'stringFromCodePoints', 'stringFromCharCodes', 'stringFromCharCode', 'stringConcatUnchecked', 'flattenString', 'getTimeZoneName', 'getTimeZoneOffsetInMinutes', 'valueFromDecomposedDate', 'patchUpY2K', 'lazyAsJsDate', 'getYear', 'getMonth', 'getDay', 'getHours', 'getMinutes', 'getSeconds', 'getMilliseconds', 'getWeekday', 'valueFromDateString', 'getProperty', 'setProperty', 'identicalImplementation', 'extractStackTrace']
  });
  Primitives.mirrorFunctionCacheName = '$cachedFunction';
  Primitives.mirrorInvokeCacheName = '$cachedInvocation';
  Primitives.DOLLAR_CHAR_VALUE = 36;
  Primitives.timerFrequency = null;
  Primitives.timerTicks = null;
  function stringLastIndexOfUnchecked(receiver, element, start) {
    return receiver.lastIndexOf(element, start);
  }
  dart.fn(stringLastIndexOfUnchecked);
  function checkNull(object) {
    if (object == null)
      dart.throw(new core.ArgumentError(null));
    return object;
  }
  dart.fn(checkNull);
  function checkNum(value) {
    if (!(typeof value == 'number')) {
      dart.throw(new core.ArgumentError(value));
    }
    return value;
  }
  dart.fn(checkNum);
  function checkInt(value) {
    if (!(typeof value == 'number')) {
      dart.throw(new core.ArgumentError(value));
    }
    return value;
  }
  dart.fn(checkInt);
  function checkBool(value) {
    if (!(typeof value == 'boolean')) {
      dart.throw(new core.ArgumentError(value));
    }
    return value;
  }
  dart.fn(checkBool);
  function checkString(value) {
    if (!(typeof value == 'string')) {
      dart.throw(new core.ArgumentError(value));
    }
    return value;
  }
  dart.fn(checkString);
  function throwRuntimeError(message) {
    dart.throw(new RuntimeError(message));
  }
  dart.fn(throwRuntimeError);
  function throwAbstractClassInstantiationError(className) {
    dart.throw(new core.AbstractClassInstantiationError(dart.as(className, core.String)));
  }
  dart.fn(throwAbstractClassInstantiationError);
  const _message = Symbol('_message');
  const _method = Symbol('_method');
  class NullError extends core.Error {
    NullError(message, match) {
      this[_message] = message;
      this[_method] = match == null ? null : dart.as(match.method, core.String);
      super.Error();
    }
    toString() {
      if (this[_method] == null)
        return `NullError: ${this[_message]}`;
      return `NullError: Cannot call "${this[_method]}" on null`;
    }
  }
  NullError[dart.implements] = () => [core.NoSuchMethodError];
  dart.setSignature(NullError, {
    constructors: () => ({NullError: [NullError, [core.String, dart.dynamic]]})
  });
  const _receiver = Symbol('_receiver');
  class JsNoSuchMethodError extends core.Error {
    JsNoSuchMethodError(message, match) {
      this[_message] = message;
      this[_method] = match == null ? null : dart.as(match.method, core.String);
      this[_receiver] = match == null ? null : dart.as(match.receiver, core.String);
      super.Error();
    }
    toString() {
      if (this[_method] == null)
        return `NoSuchMethodError: ${this[_message]}`;
      if (this[_receiver] == null) {
        return `NoSuchMethodError: Cannot call "${this[_method]}" (${this[_message]})`;
      }
      return `NoSuchMethodError: Cannot call "${this[_method]}" on "${this[_receiver]}" ` + `(${this[_message]})`;
    }
  }
  JsNoSuchMethodError[dart.implements] = () => [core.NoSuchMethodError];
  dart.setSignature(JsNoSuchMethodError, {
    constructors: () => ({JsNoSuchMethodError: [JsNoSuchMethodError, [core.String, dart.dynamic]]})
  });
  class UnknownJsTypeError extends core.Error {
    UnknownJsTypeError(message) {
      this[_message] = message;
      super.Error();
    }
    toString() {
      return dart.notNull(this[_message][dartx.isEmpty]) ? 'Error' : `Error: ${this[_message]}`;
    }
  }
  dart.setSignature(UnknownJsTypeError, {
    constructors: () => ({UnknownJsTypeError: [UnknownJsTypeError, [core.String]]})
  });
  function getTraceFromException(exception) {
    return new _StackTrace(exception);
  }
  dart.fn(getTraceFromException, core.StackTrace, [dart.dynamic]);
  const _exception = Symbol('_exception');
  const _trace = Symbol('_trace');
  class _StackTrace extends core.Object {
    _StackTrace(exception) {
      this[_exception] = exception;
      this[_trace] = null;
    }
    toString() {
      if (this[_trace] != null)
        return this[_trace];
      let trace = null;
      if (typeof this[_exception] === "object") {
        trace = dart.as(this[_exception].stack, core.String);
      }
      return this[_trace] = trace == null ? '' : trace;
    }
  }
  _StackTrace[dart.implements] = () => [core.StackTrace];
  dart.setSignature(_StackTrace, {
    constructors: () => ({_StackTrace: [_StackTrace, [dart.dynamic]]})
  });
  function objectHashCode(object) {
    if (object == null || typeof object != 'object') {
      return dart.hashCode(object);
    } else {
      return Primitives.objectHashCode(object);
    }
  }
  dart.fn(objectHashCode, core.int, [dart.dynamic]);
  function fillLiteralMap(keyValuePairs, result) {
    let index = 0;
    let length = getLength(keyValuePairs);
    while (dart.notNull(index) < dart.notNull(length)) {
      let key = getIndex(keyValuePairs, (() => {
        let x = index;
        index = dart.notNull(x) + 1;
        return x;
      })());
      let value = getIndex(keyValuePairs, (() => {
        let x = index;
        index = dart.notNull(x) + 1;
        return x;
      })());
      result.set(key, value);
    }
    return result;
  }
  dart.fn(fillLiteralMap, dart.dynamic, [dart.dynamic, core.Map]);
  function jsHasOwnProperty(jsObject, property) {
    return jsObject.hasOwnProperty(property);
  }
  dart.fn(jsHasOwnProperty, core.bool, [dart.dynamic, core.String]);
  function jsPropertyAccess(jsObject, property) {
    return jsObject[property];
  }
  dart.fn(jsPropertyAccess, dart.dynamic, [dart.dynamic, core.String]);
  function getFallThroughError() {
    return new FallThroughErrorImplementation();
  }
  dart.fn(getFallThroughError);
  class Creates extends core.Object {
    Creates(types) {
      this.types = types;
    }
  }
  dart.setSignature(Creates, {
    constructors: () => ({Creates: [Creates, [core.String]]})
  });
  class Returns extends core.Object {
    Returns(types) {
      this.types = types;
    }
  }
  dart.setSignature(Returns, {
    constructors: () => ({Returns: [Returns, [core.String]]})
  });
  class JSName extends core.Object {
    JSName(name) {
      this.name = name;
    }
  }
  dart.setSignature(JSName, {
    constructors: () => ({JSName: [JSName, [core.String]]})
  });
  class JavaScriptIndexingBehavior extends _interceptors.JSMutableIndexable {}
  class TypeErrorImplementation extends core.Error {
    TypeErrorImplementation(value, type) {
      this.message = `type '${Primitives.objectTypeName(value)}' is not a subtype ` + `of type '${type}'`;
      super.Error();
    }
    fromMessage(message) {
      this.message = message;
      super.Error();
    }
    toString() {
      return this.message;
    }
  }
  TypeErrorImplementation[dart.implements] = () => [core.TypeError];
  dart.defineNamedConstructor(TypeErrorImplementation, 'fromMessage');
  dart.setSignature(TypeErrorImplementation, {
    constructors: () => ({
      TypeErrorImplementation: [TypeErrorImplementation, [core.Object, core.String]],
      fromMessage: [TypeErrorImplementation, [core.String]]
    })
  });
  class CastErrorImplementation extends core.Error {
    CastErrorImplementation(actualType, expectedType) {
      this.message = `CastError: Casting value of type ${actualType} to` + ` incompatible type ${expectedType}`;
      super.Error();
    }
    toString() {
      return this.message;
    }
  }
  CastErrorImplementation[dart.implements] = () => [core.CastError];
  dart.setSignature(CastErrorImplementation, {
    constructors: () => ({CastErrorImplementation: [CastErrorImplementation, [core.Object, core.Object]]})
  });
  class FallThroughErrorImplementation extends core.FallThroughError {
    FallThroughErrorImplementation() {
      super.FallThroughError();
    }
    toString() {
      return "Switch case fall-through.";
    }
  }
  dart.setSignature(FallThroughErrorImplementation, {
    constructors: () => ({FallThroughErrorImplementation: [FallThroughErrorImplementation, []]})
  });
  class RuntimeError extends core.Error {
    RuntimeError(message) {
      this.message = message;
      super.Error();
    }
    toString() {
      return `RuntimeError: ${this.message}`;
    }
  }
  dart.setSignature(RuntimeError, {
    constructors: () => ({RuntimeError: [RuntimeError, [dart.dynamic]]})
  });
  function random64() {
    let int32a = Math.random() * 0x100000000 >>> 0;
    let int32b = Math.random() * 0x100000000 >>> 0;
    return dart.notNull(int32a) + dart.notNull(int32b) * 4294967296;
  }
  dart.fn(random64, core.int, []);
  function jsonEncodeNative(string) {
    return JSON.stringify(string);
  }
  dart.fn(jsonEncodeNative, core.String, [core.String]);
  const _jsIterator = Symbol('_jsIterator');
  const SyncIterator$ = dart.generic(function(E) {
    class SyncIterator extends core.Object {
      SyncIterator(jsIterator) {
        this[_jsIterator] = jsIterator;
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        let ret = this[_jsIterator].next();
        this[_current] = dart.as(ret.value, E);
        return !ret.done;
      }
    }
    SyncIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(SyncIterator, {
      constructors: () => ({SyncIterator: [SyncIterator$(E), [dart.dynamic]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return SyncIterator;
  });
  let SyncIterator = SyncIterator$();
  const _generator = Symbol('_generator');
  const _args = Symbol('_args');
  const SyncIterable$ = dart.generic(function(E) {
    class SyncIterable extends collection.IterableBase$(E) {
      SyncIterable(generator, args) {
        this[_generator] = generator;
        this[_args] = args;
        super.IterableBase();
      }
      [_jsIterator]() {
        return this[_generator](...this[_args]);
      }
      get iterator() {
        return new (SyncIterator$(E))(this[_jsIterator]());
      }
    }
    dart.setSignature(SyncIterable, {
      constructors: () => ({SyncIterable: [SyncIterable$(E), [dart.dynamic, dart.dynamic]]}),
      methods: () => ({[_jsIterator]: [dart.dynamic, []]})
    });
    dart.defineExtensionMembers(SyncIterable, ['iterator']);
    return SyncIterable;
  });
  let SyncIterable = SyncIterable$();
  // Exports:
  exports.NoThrows = NoThrows;
  exports.NoInline = NoInline;
  exports.Native = Native;
  exports.JsPeerInterface = JsPeerInterface;
  exports.SupportJsExtensionMethods = SupportJsExtensionMethods;
  exports.defineProperty = defineProperty;
  exports.regExpGetNative = regExpGetNative;
  exports.regExpGetGlobalNative = regExpGetGlobalNative;
  exports.regExpCaptureCount = regExpCaptureCount;
  exports.JSSyntaxRegExp = JSSyntaxRegExp;
  exports.firstMatchAfter = firstMatchAfter;
  exports.StringMatch = StringMatch;
  exports.allMatchesInStringUnchecked = allMatchesInStringUnchecked;
  exports.stringContainsUnchecked = stringContainsUnchecked;
  exports.stringReplaceJS = stringReplaceJS;
  exports.stringReplaceFirstRE = stringReplaceFirstRE;
  exports.ESCAPE_REGEXP = ESCAPE_REGEXP;
  exports.stringReplaceAllUnchecked = stringReplaceAllUnchecked;
  exports.stringReplaceAllFuncUnchecked = stringReplaceAllFuncUnchecked;
  exports.stringReplaceAllEmptyFuncUnchecked = stringReplaceAllEmptyFuncUnchecked;
  exports.stringReplaceAllStringFuncUnchecked = stringReplaceAllStringFuncUnchecked;
  exports.stringReplaceFirstUnchecked = stringReplaceFirstUnchecked;
  exports.stringJoinUnchecked = stringJoinUnchecked;
  exports.getRuntimeType = getRuntimeType;
  exports.getIndex = getIndex;
  exports.getLength = getLength;
  exports.isJsArray = isJsArray;
  exports.patch = patch;
  exports.InternalMap = InternalMap;
  exports.Primitives = Primitives;
  exports.stringLastIndexOfUnchecked = stringLastIndexOfUnchecked;
  exports.checkNull = checkNull;
  exports.checkNum = checkNum;
  exports.checkInt = checkInt;
  exports.checkBool = checkBool;
  exports.checkString = checkString;
  exports.throwRuntimeError = throwRuntimeError;
  exports.throwAbstractClassInstantiationError = throwAbstractClassInstantiationError;
  exports.NullError = NullError;
  exports.JsNoSuchMethodError = JsNoSuchMethodError;
  exports.UnknownJsTypeError = UnknownJsTypeError;
  exports.getTraceFromException = getTraceFromException;
  exports.objectHashCode = objectHashCode;
  exports.fillLiteralMap = fillLiteralMap;
  exports.jsHasOwnProperty = jsHasOwnProperty;
  exports.jsPropertyAccess = jsPropertyAccess;
  exports.getFallThroughError = getFallThroughError;
  exports.Creates = Creates;
  exports.Returns = Returns;
  exports.JSName = JSName;
  exports.JavaScriptIndexingBehavior = JavaScriptIndexingBehavior;
  exports.TypeErrorImplementation = TypeErrorImplementation;
  exports.CastErrorImplementation = CastErrorImplementation;
  exports.FallThroughErrorImplementation = FallThroughErrorImplementation;
  exports.RuntimeError = RuntimeError;
  exports.random64 = random64;
  exports.jsonEncodeNative = jsonEncodeNative;
  exports.SyncIterator$ = SyncIterator$;
  exports.SyncIterator = SyncIterator;
  exports.SyncIterable$ = SyncIterable$;
  exports.SyncIterable = SyncIterable;
});
dart_library.library('dart/_js_mirrors', null, /* Imports */[
  "dart/_runtime",
  'dart/_internal',
  'dart/core',
  'dart/mirrors'
], /* Lazy imports */[
], function(exports, dart, _internal, core, mirrors) {
  'use strict';
  let dartx = dart.dartx;
  function getName(symbol) {
    return _internal.Symbol.getName(dart.as(symbol, _internal.Symbol));
  }
  dart.fn(getName, core.String, [core.Symbol]);
  function getSymbol(name, library) {
    return dart.throw(new core.UnimplementedError("MirrorSystem.getSymbol unimplemented"));
  }
  dart.fn(getSymbol, core.Symbol, [dart.dynamic, dart.dynamic]);
  dart.defineLazyProperties(exports, {
    get currentJsMirrorSystem() {
      return dart.throw(new core.UnimplementedError("MirrorSystem.currentJsMirrorSystem unimplemented"));
    }
  });
  function reflect(reflectee) {
    return new JsInstanceMirror._(reflectee);
  }
  dart.fn(reflect, mirrors.InstanceMirror, [dart.dynamic]);
  function reflectType(key) {
    return new JsClassMirror._(key);
  }
  dart.fn(reflectType, mirrors.TypeMirror, [core.Type]);
  dart.defineLazyProperties(exports, {
    get _dart() {
      return dart;
    },
    get _metadata() {
      return exports._dart.metadata;
    }
  });
  function _dload(obj, name) {
    return exports._dart.dload(obj, name);
  }
  dart.fn(_dload, dart.dynamic, [dart.dynamic, core.String]);
  function _dput(obj, name, val) {
    exports._dart.dput(obj, name, val);
  }
  dart.fn(_dput, dart.void, [dart.dynamic, core.String, dart.dynamic]);
  function _dsend(obj, name, args) {
    return exports._dart.dsend(obj, name, ...args);
  }
  dart.fn(_dsend, dart.dynamic, [dart.dynamic, core.String, core.List]);
  const _toJsMap = Symbol('_toJsMap');
  class JsInstanceMirror extends core.Object {
    _(reflectee) {
      this.reflectee = reflectee;
    }
    get type() {
      return dart.throw(new core.UnimplementedError("ClassMirror.type unimplemented"));
    }
    get hasReflectee() {
      return dart.throw(new core.UnimplementedError("ClassMirror.hasReflectee unimplemented"));
    }
    delegate(invocation) {
      return dart.throw(new core.UnimplementedError("ClassMirror.delegate unimplemented"));
    }
    getField(symbol) {
      let name = getName(symbol);
      let field = _dload(this.reflectee, name);
      return new JsInstanceMirror._(field);
    }
    setField(symbol, value) {
      let name = getName(symbol);
      _dput(this.reflectee, name, value);
      return new JsInstanceMirror._(value);
    }
    invoke(symbol, args, namedArgs) {
      if (namedArgs === void 0)
        namedArgs = null;
      let name = getName(symbol);
      if (namedArgs != null) {
        args = core.List.from(args);
        args[dartx.add](this[_toJsMap](namedArgs));
      }
      let result = _dsend(this.reflectee, name, args);
      return new JsInstanceMirror._(result);
    }
    [_toJsMap](map) {
      let obj = {};
      map.forEach(dart.fn((key, value) => {
        obj[getName(key)] = value;
      }, dart.dynamic, [core.Symbol, dart.dynamic]));
      return obj;
    }
  }
  JsInstanceMirror[dart.implements] = () => [mirrors.InstanceMirror];
  dart.defineNamedConstructor(JsInstanceMirror, '_');
  dart.setSignature(JsInstanceMirror, {
    constructors: () => ({_: [JsInstanceMirror, [core.Object]]}),
    methods: () => ({
      delegate: [dart.dynamic, [core.Invocation]],
      getField: [mirrors.InstanceMirror, [core.Symbol]],
      setField: [mirrors.InstanceMirror, [core.Symbol, core.Object]],
      invoke: [mirrors.InstanceMirror, [core.Symbol, core.List], [core.Map$(core.Symbol, dart.dynamic)]],
      [_toJsMap]: [dart.dynamic, [core.Map$(core.Symbol, dart.dynamic)]]
    })
  });
  const _metadata = Symbol('_metadata');
  const _declarations = Symbol('_declarations');
  const _cls = Symbol('_cls');
  class JsClassMirror extends core.Object {
    get metadata() {
      return this[_metadata];
    }
    get declarations() {
      return this[_declarations];
    }
    _(cls) {
      this[_cls] = cls;
      this.simpleName = core.Symbol.new(cls.name);
      this[_metadata] = null;
      this[_declarations] = null;
      let fn = this[_cls][dart.metadata];
      this[_metadata] = fn == null ? dart.list([], mirrors.InstanceMirror) : core.List$(mirrors.InstanceMirror).from(dart.as(dart.dsend(dart.dcall(fn), 'map', dart.fn(i => new JsInstanceMirror._(i), JsInstanceMirror, [dart.dynamic])), core.Iterable));
      this[_declarations] = core.Map$(core.Symbol, mirrors.MethodMirror).new();
      this[_declarations].set(this.simpleName, new JsMethodMirror._(this, this[_cls]));
    }
    newInstance(constructorName, args, namedArgs) {
      if (namedArgs === void 0)
        namedArgs = null;
      dart.assert(getName(constructorName) == "");
      dart.assert(namedArgs == null || dart.notNull(namedArgs.isEmpty));
      let instance = new this[_cls](...args);
      return new JsInstanceMirror._(instance);
    }
    get superinterfaces() {
      let interfaceThunk = this[_cls][dart.implements];
      if (interfaceThunk == null) {
        return dart.list([], mirrors.ClassMirror);
      } else {
        let interfaces = dart.as(dart.dcall(interfaceThunk), core.List$(core.Type));
        return interfaces[dartx.map](dart.fn(t => new JsClassMirror._(dart.as(t, core.Type)), JsClassMirror, [dart.dynamic]))[dartx.toList]();
      }
    }
    getField(fieldName) {
      return dart.throw(new core.UnimplementedError("ClassMirror.getField unimplemented"));
    }
    invoke(memberName, positionalArguments, namedArguments) {
      if (namedArguments === void 0)
        namedArguments = null;
      return dart.throw(new core.UnimplementedError("ClassMirror.invoke unimplemented"));
    }
    isAssignableTo(other) {
      return dart.throw(new core.UnimplementedError("ClassMirror.isAssignable unimplemented"));
    }
    isSubclassOf(other) {
      return dart.throw(new core.UnimplementedError("ClassMirror.isSubclassOf unimplemented"));
    }
    isSubtypeOf(other) {
      return dart.throw(new core.UnimplementedError("ClassMirror.isSubtypeOf unimplemented"));
    }
    setField(fieldName, value) {
      return dart.throw(new core.UnimplementedError("ClassMirror.setField unimplemented"));
    }
    get hasReflectedType() {
      return dart.throw(new core.UnimplementedError("ClassMirror.hasReflectedType unimplemented"));
    }
    get instanceMembers() {
      return dart.throw(new core.UnimplementedError("ClassMirror.instanceMembers unimplemented"));
    }
    get isAbstract() {
      return dart.throw(new core.UnimplementedError("ClassMirror.isAbstract unimplemented"));
    }
    get isEnum() {
      return dart.throw(new core.UnimplementedError("ClassMirror.isEnum unimplemented"));
    }
    get isOriginalDeclaration() {
      return dart.throw(new core.UnimplementedError("ClassMirror.isOriginalDeclaration unimplemented"));
    }
    get isPrivate() {
      return dart.throw(new core.UnimplementedError("ClassMirror.isPrivate unimplemented"));
    }
    get isTopLevel() {
      return dart.throw(new core.UnimplementedError("ClassMirror.isTopLevel unimplemented"));
    }
    get location() {
      return dart.throw(new core.UnimplementedError("ClassMirror.location unimplemented"));
    }
    get mixin() {
      return dart.throw(new core.UnimplementedError("ClassMirror.mixin unimplemented"));
    }
    get originalDeclaration() {
      return this;
    }
    get owner() {
      return dart.throw(new core.UnimplementedError("ClassMirror.owner unimplemented"));
    }
    get qualifiedName() {
      return dart.throw(new core.UnimplementedError("ClassMirror.qualifiedName unimplemented"));
    }
    get reflectedType() {
      return this[_cls];
    }
    get staticMembers() {
      return dart.throw(new core.UnimplementedError("ClassMirror.staticMembers unimplemented"));
    }
    get superclass() {
      if (dart.equals(this[_cls], core.Object)) {
        return null;
      } else {
        return new JsClassMirror._(this[_cls].__proto__);
      }
    }
    get typeArguments() {
      return dart.throw(new core.UnimplementedError("ClassMirror.typeArguments unimplemented"));
    }
    get typeVariables() {
      return dart.throw(new core.UnimplementedError("ClassMirror.typeVariables unimplemented"));
    }
  }
  JsClassMirror[dart.implements] = () => [mirrors.ClassMirror];
  dart.defineNamedConstructor(JsClassMirror, '_');
  dart.setSignature(JsClassMirror, {
    constructors: () => ({_: [JsClassMirror, [core.Type]]}),
    methods: () => ({
      newInstance: [mirrors.InstanceMirror, [core.Symbol, core.List], [core.Map$(core.Symbol, dart.dynamic)]],
      getField: [mirrors.InstanceMirror, [core.Symbol]],
      invoke: [mirrors.InstanceMirror, [core.Symbol, core.List], [core.Map$(core.Symbol, dart.dynamic)]],
      isAssignableTo: [core.bool, [mirrors.TypeMirror]],
      isSubclassOf: [core.bool, [mirrors.ClassMirror]],
      isSubtypeOf: [core.bool, [mirrors.TypeMirror]],
      setField: [mirrors.InstanceMirror, [core.Symbol, core.Object]]
    })
  });
  class JsTypeMirror extends core.Object {
    _(reflectedType) {
      this.reflectedType = reflectedType;
      this.hasReflectedType = true;
    }
    isAssignableTo(other) {
      return dart.throw(new core.UnimplementedError("TypeMirror.isAssignable unimplemented"));
    }
    isSubtypeOf(other) {
      return dart.throw(new core.UnimplementedError("TypeMirror.isSubtypeOf unimplemented"));
    }
    get isOriginalDeclaration() {
      return dart.throw(new core.UnimplementedError("TypeMirror.isOriginalDeclaration unimplemented"));
    }
    get isPrivate() {
      return dart.throw(new core.UnimplementedError("TypeMirror.isPrivate unimplemented"));
    }
    get isTopLevel() {
      return dart.throw(new core.UnimplementedError("TypeMirror.isTopLevel unimplemented"));
    }
    get location() {
      return dart.throw(new core.UnimplementedError("TypeMirror.location unimplemented"));
    }
    get metadata() {
      return dart.throw(new core.UnimplementedError("TypeMirror.metadata unimplemented"));
    }
    get originalDeclaration() {
      return dart.throw(new core.UnimplementedError("TypeMirror.originalDeclaration unimplemented"));
    }
    get owner() {
      return dart.throw(new core.UnimplementedError("TypeMirror.owner unimplemented"));
    }
    get qualifiedName() {
      return dart.throw(new core.UnimplementedError("TypeMirror.qualifiedName unimplemented"));
    }
    get simpleName() {
      return dart.throw(new core.UnimplementedError("TypeMirror.simpleName unimplemented"));
    }
    get typeArguments() {
      return dart.throw(new core.UnimplementedError("TypeMirror.typeArguments unimplemented"));
    }
    get typeVariables() {
      return dart.throw(new core.UnimplementedError("TypeMirror.typeVariables unimplemented"));
    }
  }
  JsTypeMirror[dart.implements] = () => [mirrors.TypeMirror];
  dart.defineNamedConstructor(JsTypeMirror, '_');
  dart.setSignature(JsTypeMirror, {
    constructors: () => ({_: [JsTypeMirror, [core.Type]]}),
    methods: () => ({
      isAssignableTo: [core.bool, [mirrors.TypeMirror]],
      isSubtypeOf: [core.bool, [mirrors.TypeMirror]]
    })
  });
  const _name = Symbol('_name');
  class JsParameterMirror extends core.Object {
    _(name, t, annotations) {
      this[_name] = name;
      this.type = new JsTypeMirror._(t);
      this.metadata = core.List$(mirrors.InstanceMirror).from(annotations[dartx.map](dart.fn(a => new JsInstanceMirror._(a), JsInstanceMirror, [dart.dynamic])));
    }
    get defaultValue() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.defaultValues unimplemented"));
    }
    get hasDefaultValue() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.hasDefaultValue unimplemented"));
    }
    get isConst() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isConst unimplemented"));
    }
    get isFinal() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isFinal unimplemented"));
    }
    get isNamed() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isNamed unimplemented"));
    }
    get isOptional() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isOptional unimplemented"));
    }
    get isPrivate() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isPrivate unimplemented"));
    }
    get isStatic() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isStatic unimplemented"));
    }
    get isTopLevel() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.isTopLevel unimplemented"));
    }
    get location() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.location unimplemented"));
    }
    get owner() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.owner unimplemented"));
    }
    get qualifiedName() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.qualifiedName unimplemented"));
    }
    get simpleName() {
      return dart.throw(new core.UnimplementedError("ParameterMirror.simpleName unimplemented"));
    }
  }
  JsParameterMirror[dart.implements] = () => [mirrors.ParameterMirror];
  dart.defineNamedConstructor(JsParameterMirror, '_');
  dart.setSignature(JsParameterMirror, {
    constructors: () => ({_: [JsParameterMirror, [core.String, core.Type, core.List]]})
  });
  const _method = Symbol('_method');
  const _params = Symbol('_params');
  const _createParameterMirrorList = Symbol('_createParameterMirrorList');
  class JsMethodMirror extends core.Object {
    _(cls, method) {
      this[_method] = method;
      this[_name] = getName(cls.simpleName);
      this[_params] = null;
      let ftype = exports._dart.classGetConstructorType(cls[_cls]);
      this[_params] = this[_createParameterMirrorList](ftype);
    }
    get constructorName() {
      return core.Symbol.new('');
    }
    get parameters() {
      return this[_params];
    }
    [_createParameterMirrorList](ftype) {
      if (ftype == null) {
        return dart.list([], mirrors.ParameterMirror);
      }
      let args = dart.as(dart.dload(ftype, 'args'), core.List);
      let opts = dart.as(dart.dload(ftype, 'optionals'), core.List);
      let params = core.List$(mirrors.ParameterMirror).new(dart.notNull(args[dartx.length]) + dart.notNull(opts[dartx.length]));
      for (let i = 0; dart.notNull(i) < dart.notNull(args[dartx.length]); i = dart.notNull(i) + 1) {
        let type = args[dartx.get](i);
        let metadata = dart.dindex(dart.dload(ftype, 'metadata'), i);
        let param = new JsParameterMirror._('', dart.as(type, core.Type), dart.as(metadata, core.List));
        params[dartx.set](i, param);
      }
      for (let i = 0; dart.notNull(i) < dart.notNull(opts[dartx.length]); i = dart.notNull(i) + 1) {
        let type = opts[dartx.get](i);
        let metadata = dart.dindex(dart.dload(ftype, 'metadata'), dart.notNull(args[dartx.length]) + dart.notNull(i));
        let param = new JsParameterMirror._('', dart.as(type, core.Type), dart.as(metadata, core.List));
        params[dartx.set](dart.notNull(i) + dart.notNull(args[dartx.length]), param);
      }
      return params;
    }
    get isAbstract() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isAbstract unimplemented"));
    }
    get isConstConstructor() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isConstConstructor unimplemented"));
    }
    get isConstructor() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isConstructor unimplemented"));
    }
    get isFactoryConstructor() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isFactoryConstructor unimplemented"));
    }
    get isGenerativeConstructor() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isGenerativeConstructor unimplemented"));
    }
    get isGetter() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isGetter unimplemented"));
    }
    get isOperator() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isOperator unimplemented"));
    }
    get isPrivate() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isPrivate unimplemented"));
    }
    get isRedirectingConstructor() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isRedirectingConstructor unimplemented"));
    }
    get isRegularMethod() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isRegularMethod unimplemented"));
    }
    get isSetter() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isSetter unimplemented"));
    }
    get isStatic() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isStatic unimplemented"));
    }
    get isSynthetic() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isSynthetic unimplemented"));
    }
    get isTopLevel() {
      return dart.throw(new core.UnimplementedError("MethodMirror.isTopLevel unimplemented"));
    }
    get location() {
      return dart.throw(new core.UnimplementedError("MethodMirror.location unimplemented"));
    }
    get metadata() {
      return dart.list([], mirrors.InstanceMirror);
    }
    get owner() {
      return dart.throw(new core.UnimplementedError("MethodMirror.owner unimplemented"));
    }
    get qualifiedName() {
      return dart.throw(new core.UnimplementedError("MethodMirror.qualifiedName unimplemented"));
    }
    get returnType() {
      return dart.throw(new core.UnimplementedError("MethodMirror.returnType unimplemented"));
    }
    get simpleName() {
      return dart.throw(new core.UnimplementedError("MethodMirror.simpleName unimplemented"));
    }
    get source() {
      return dart.throw(new core.UnimplementedError("MethodMirror.source unimplemented"));
    }
  }
  JsMethodMirror[dart.implements] = () => [mirrors.MethodMirror];
  dart.defineNamedConstructor(JsMethodMirror, '_');
  dart.setSignature(JsMethodMirror, {
    constructors: () => ({_: [JsMethodMirror, [JsClassMirror, dart.dynamic]]}),
    methods: () => ({[_createParameterMirrorList]: [core.List$(mirrors.ParameterMirror), [dart.dynamic]]})
  });
  // Exports:
  exports.getName = getName;
  exports.getSymbol = getSymbol;
  exports.reflect = reflect;
  exports.reflectType = reflectType;
  exports.JsInstanceMirror = JsInstanceMirror;
  exports.JsClassMirror = JsClassMirror;
  exports.JsTypeMirror = JsTypeMirror;
  exports.JsParameterMirror = JsParameterMirror;
  exports.JsMethodMirror = JsMethodMirror;
});
dart_library.library('dart/_js_primitives', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
], function(exports, dart, core) {
  'use strict';
  let dartx = dart.dartx;
  function printString(string) {
    if (typeof dartPrint == "function") {
      dartPrint(string);
      return;
    }
    if (typeof console == "object" && typeof console.log != "undefined") {
      console.log(string);
      return;
    }
    if (typeof window == "object") {
      return;
    }
    if (typeof print == "function") {
      print(string);
      return;
    }
    throw "Unable to print message: " + String(string);
  }
  dart.fn(printString, dart.void, [core.String]);
  // Exports:
  exports.printString = printString;
});
dart_library.library('dart/_metadata', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
], function(exports, dart, core) {
  'use strict';
  let dartx = dart.dartx;
  class SupportedBrowser extends core.Object {
    SupportedBrowser(browserName, minimumVersion) {
      if (minimumVersion === void 0)
        minimumVersion = null;
      this.browserName = browserName;
      this.minimumVersion = minimumVersion;
    }
  }
  dart.setSignature(SupportedBrowser, {
    constructors: () => ({SupportedBrowser: [SupportedBrowser, [core.String], [core.String]]})
  });
  SupportedBrowser.CHROME = "Chrome";
  SupportedBrowser.FIREFOX = "Firefox";
  SupportedBrowser.IE = "Internet Explorer";
  SupportedBrowser.OPERA = "Opera";
  SupportedBrowser.SAFARI = "Safari";
  class Experimental extends core.Object {
    Experimental() {
    }
  }
  dart.setSignature(Experimental, {
    constructors: () => ({Experimental: [Experimental, []]})
  });
  class DomName extends core.Object {
    DomName(name) {
      this.name = name;
    }
  }
  dart.setSignature(DomName, {
    constructors: () => ({DomName: [DomName, [core.String]]})
  });
  class DocsEditable extends core.Object {
    DocsEditable() {
    }
  }
  dart.setSignature(DocsEditable, {
    constructors: () => ({DocsEditable: [DocsEditable, []]})
  });
  class Unstable extends core.Object {
    Unstable() {
    }
  }
  dart.setSignature(Unstable, {
    constructors: () => ({Unstable: [Unstable, []]})
  });
  // Exports:
  exports.SupportedBrowser = SupportedBrowser;
  exports.Experimental = Experimental;
  exports.DomName = DomName;
  exports.DocsEditable = DocsEditable;
  exports.Unstable = Unstable;
});
dart_library.library('dart/_native_typed_data', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/typed_data',
  'dart/_js_helper',
  'dart/collection',
  'dart/_internal',
  'dart/_interceptors',
  'dart/math'
], /* Lazy imports */[
], function(exports, dart, core, typed_data, _js_helper, collection, _internal, _interceptors, math) {
  'use strict';
  let dartx = dart.dartx;
  class NativeByteBuffer extends core.Object {
    get runtimeType() {
      return typed_data.ByteBuffer;
    }
    asUint8List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeUint8List.view(this, offsetInBytes, length);
    }
    asInt8List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeInt8List.view(this, offsetInBytes, length);
    }
    asUint8ClampedList(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeUint8ClampedList.view(this, offsetInBytes, length);
    }
    asUint16List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeUint16List.view(this, offsetInBytes, length);
    }
    asInt16List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeInt16List.view(this, offsetInBytes, length);
    }
    asUint32List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeUint32List.view(this, offsetInBytes, length);
    }
    asInt32List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeInt32List.view(this, offsetInBytes, length);
    }
    asUint64List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      dart.throw(new core.UnsupportedError("Uint64List not supported by dart2js."));
    }
    asInt64List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      dart.throw(new core.UnsupportedError("Int64List not supported by dart2js."));
    }
    asInt32x4List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      let storage = dart.as(this.asInt32List(offsetInBytes, length != null ? dart.notNull(length) * 4 : null), NativeInt32List);
      return new NativeInt32x4List._externalStorage(storage);
    }
    asFloat32List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeFloat32List.view(this, offsetInBytes, length);
    }
    asFloat64List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeFloat64List.view(this, offsetInBytes, length);
    }
    asFloat32x4List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      let storage = dart.as(this.asFloat32List(offsetInBytes, length != null ? dart.notNull(length) * 4 : null), NativeFloat32List);
      return new NativeFloat32x4List._externalStorage(storage);
    }
    asFloat64x2List(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      let storage = dart.as(this.asFloat64List(offsetInBytes, length != null ? dart.notNull(length) * 2 : null), NativeFloat64List);
      return new NativeFloat64x2List._externalStorage(storage);
    }
    asByteData(offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return NativeByteData.view(this, offsetInBytes, length);
    }
  }
  NativeByteBuffer[dart.implements] = () => [typed_data.ByteBuffer];
  dart.setSignature(NativeByteBuffer, {
    methods: () => ({
      asUint8List: [typed_data.Uint8List, [], [core.int, core.int]],
      asInt8List: [typed_data.Int8List, [], [core.int, core.int]],
      asUint8ClampedList: [typed_data.Uint8ClampedList, [], [core.int, core.int]],
      asUint16List: [typed_data.Uint16List, [], [core.int, core.int]],
      asInt16List: [typed_data.Int16List, [], [core.int, core.int]],
      asUint32List: [typed_data.Uint32List, [], [core.int, core.int]],
      asInt32List: [typed_data.Int32List, [], [core.int, core.int]],
      asUint64List: [typed_data.Uint64List, [], [core.int, core.int]],
      asInt64List: [typed_data.Int64List, [], [core.int, core.int]],
      asInt32x4List: [typed_data.Int32x4List, [], [core.int, core.int]],
      asFloat32List: [typed_data.Float32List, [], [core.int, core.int]],
      asFloat64List: [typed_data.Float64List, [], [core.int, core.int]],
      asFloat32x4List: [typed_data.Float32x4List, [], [core.int, core.int]],
      asFloat64x2List: [typed_data.Float64x2List, [], [core.int, core.int]],
      asByteData: [typed_data.ByteData, [], [core.int, core.int]]
    })
  });
  NativeByteBuffer[dart.metadata] = () => [dart.const(new _js_helper.Native("ArrayBuffer"))];
  const _storage = Symbol('_storage');
  const _invalidIndex = Symbol('_invalidIndex');
  const _checkIndex = Symbol('_checkIndex');
  const _checkSublistArguments = Symbol('_checkSublistArguments');
  class NativeFloat32x4List extends dart.mixin(core.Object, collection.ListMixin$(typed_data.Float32x4), _internal.FixedLengthListMixin$(typed_data.Float32x4)) {
    NativeFloat32x4List(length) {
      this[_storage] = NativeFloat32List.new(dart.notNull(length) * 4);
    }
    _externalStorage(storage) {
      this[_storage] = storage;
    }
    _slowFromList(list) {
      this[_storage] = NativeFloat32List.new(dart.notNull(list[dartx.length]) * 4);
      for (let i = 0; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
        let e = list[dartx.get](i);
        this[_storage].set(dart.notNull(i) * 4 + 0, e.x);
        this[_storage].set(dart.notNull(i) * 4 + 1, e.y);
        this[_storage].set(dart.notNull(i) * 4 + 2, e.z);
        this[_storage].set(dart.notNull(i) * 4 + 3, e.w);
      }
    }
    get runtimeType() {
      return typed_data.Float32x4List;
    }
    static fromList(list) {
      if (dart.is(list, NativeFloat32x4List)) {
        return new NativeFloat32x4List._externalStorage(NativeFloat32List.fromList(list[_storage]));
      } else {
        return new NativeFloat32x4List._slowFromList(list);
      }
    }
    get buffer() {
      return this[_storage].buffer;
    }
    get lengthInBytes() {
      return this[_storage].lengthInBytes;
    }
    get offsetInBytes() {
      return this[_storage].offsetInBytes;
    }
    get elementSizeInBytes() {
      return typed_data.Float32x4List.BYTES_PER_ELEMENT;
    }
    [_invalidIndex](index, length) {
      if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(length)) {
        if (length == this.length) {
          dart.throw(core.RangeError.index(index, this));
        }
        dart.throw(new core.RangeError.range(index, 0, dart.notNull(length) - 1));
      } else {
        dart.throw(new core.ArgumentError(`Invalid list index ${index}`));
      }
    }
    [_checkIndex](index, length) {
      if (index >>> 0 != index || dart.notNull(index) >= dart.notNull(length)) {
        this[_invalidIndex](index, length);
      }
    }
    [_checkSublistArguments](start, end, length) {
      this[_checkIndex](start, dart.notNull(length) + 1);
      if (end == null)
        return length;
      this[_checkIndex](end, dart.notNull(length) + 1);
      if (dart.notNull(start) > dart.notNull(end))
        dart.throw(new core.RangeError.range(start, 0, end));
      return end;
    }
    get length() {
      return (dart.notNull(this[_storage].length) / 4)[dartx.truncate]();
    }
    get(index) {
      this[_checkIndex](index, this.length);
      let _x = this[_storage].get(dart.notNull(index) * 4 + 0);
      let _y = this[_storage].get(dart.notNull(index) * 4 + 1);
      let _z = this[_storage].get(dart.notNull(index) * 4 + 2);
      let _w = this[_storage].get(dart.notNull(index) * 4 + 3);
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    set(index, value) {
      this[_checkIndex](index, this.length);
      this[_storage].set(dart.notNull(index) * 4 + 0, value.x);
      this[_storage].set(dart.notNull(index) * 4 + 1, value.y);
      this[_storage].set(dart.notNull(index) * 4 + 2, value.z);
      this[_storage].set(dart.notNull(index) * 4 + 3, value.w);
      return value;
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      return new NativeFloat32x4List._externalStorage(dart.as(this[_storage].sublist(dart.notNull(start) * 4, dart.notNull(end) * 4), NativeFloat32List));
    }
  }
  NativeFloat32x4List[dart.implements] = () => [typed_data.Float32x4List];
  dart.defineNamedConstructor(NativeFloat32x4List, '_externalStorage');
  dart.defineNamedConstructor(NativeFloat32x4List, '_slowFromList');
  dart.setSignature(NativeFloat32x4List, {
    constructors: () => ({
      NativeFloat32x4List: [NativeFloat32x4List, [core.int]],
      _externalStorage: [NativeFloat32x4List, [NativeFloat32List]],
      _slowFromList: [NativeFloat32x4List, [core.List$(typed_data.Float32x4)]],
      fromList: [NativeFloat32x4List, [core.List$(typed_data.Float32x4)]]
    }),
    methods: () => ({
      [_invalidIndex]: [dart.void, [core.int, core.int]],
      [_checkIndex]: [dart.void, [core.int, core.int]],
      [_checkSublistArguments]: [core.int, [core.int, core.int, core.int]],
      get: [typed_data.Float32x4, [core.int]],
      set: [dart.void, [core.int, typed_data.Float32x4]],
      sublist: [core.List$(typed_data.Float32x4), [core.int], [core.int]]
    })
  });
  dart.defineExtensionMembers(NativeFloat32x4List, ['get', 'set', 'sublist', 'length']);
  class NativeInt32x4List extends dart.mixin(core.Object, collection.ListMixin$(typed_data.Int32x4), _internal.FixedLengthListMixin$(typed_data.Int32x4)) {
    NativeInt32x4List(length) {
      this[_storage] = NativeInt32List.new(dart.notNull(length) * 4);
    }
    _externalStorage(storage) {
      this[_storage] = storage;
    }
    _slowFromList(list) {
      this[_storage] = NativeInt32List.new(dart.notNull(list[dartx.length]) * 4);
      for (let i = 0; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
        let e = list[dartx.get](i);
        this[_storage].set(dart.notNull(i) * 4 + 0, e.x);
        this[_storage].set(dart.notNull(i) * 4 + 1, e.y);
        this[_storage].set(dart.notNull(i) * 4 + 2, e.z);
        this[_storage].set(dart.notNull(i) * 4 + 3, e.w);
      }
    }
    get runtimeType() {
      return typed_data.Int32x4List;
    }
    static fromList(list) {
      if (dart.is(list, NativeInt32x4List)) {
        return new NativeInt32x4List._externalStorage(NativeInt32List.fromList(list[_storage]));
      } else {
        return new NativeInt32x4List._slowFromList(list);
      }
    }
    get buffer() {
      return this[_storage].buffer;
    }
    get lengthInBytes() {
      return this[_storage].lengthInBytes;
    }
    get offsetInBytes() {
      return this[_storage].offsetInBytes;
    }
    get elementSizeInBytes() {
      return typed_data.Int32x4List.BYTES_PER_ELEMENT;
    }
    [_invalidIndex](index, length) {
      if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(length)) {
        if (length == this.length) {
          dart.throw(core.RangeError.index(index, this));
        }
        dart.throw(new core.RangeError.range(index, 0, dart.notNull(length) - 1));
      } else {
        dart.throw(new core.ArgumentError(`Invalid list index ${index}`));
      }
    }
    [_checkIndex](index, length) {
      if (index >>> 0 != index || index >= length) {
        this[_invalidIndex](index, length);
      }
    }
    [_checkSublistArguments](start, end, length) {
      this[_checkIndex](start, dart.notNull(length) + 1);
      if (end == null)
        return length;
      this[_checkIndex](end, dart.notNull(length) + 1);
      if (dart.notNull(start) > dart.notNull(end))
        dart.throw(new core.RangeError.range(start, 0, end));
      return end;
    }
    get length() {
      return (dart.notNull(this[_storage].length) / 4)[dartx.truncate]();
    }
    get(index) {
      this[_checkIndex](index, this.length);
      let _x = this[_storage].get(dart.notNull(index) * 4 + 0);
      let _y = this[_storage].get(dart.notNull(index) * 4 + 1);
      let _z = this[_storage].get(dart.notNull(index) * 4 + 2);
      let _w = this[_storage].get(dart.notNull(index) * 4 + 3);
      return new NativeInt32x4._truncated(_x, _y, _z, _w);
    }
    set(index, value) {
      this[_checkIndex](index, this.length);
      this[_storage].set(dart.notNull(index) * 4 + 0, value.x);
      this[_storage].set(dart.notNull(index) * 4 + 1, value.y);
      this[_storage].set(dart.notNull(index) * 4 + 2, value.z);
      this[_storage].set(dart.notNull(index) * 4 + 3, value.w);
      return value;
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      return new NativeInt32x4List._externalStorage(dart.as(this[_storage].sublist(dart.notNull(start) * 4, dart.notNull(end) * 4), typed_data.Int32List));
    }
  }
  NativeInt32x4List[dart.implements] = () => [typed_data.Int32x4List];
  dart.defineNamedConstructor(NativeInt32x4List, '_externalStorage');
  dart.defineNamedConstructor(NativeInt32x4List, '_slowFromList');
  dart.setSignature(NativeInt32x4List, {
    constructors: () => ({
      NativeInt32x4List: [NativeInt32x4List, [core.int]],
      _externalStorage: [NativeInt32x4List, [typed_data.Int32List]],
      _slowFromList: [NativeInt32x4List, [core.List$(typed_data.Int32x4)]],
      fromList: [NativeInt32x4List, [core.List$(typed_data.Int32x4)]]
    }),
    methods: () => ({
      [_invalidIndex]: [dart.void, [core.int, core.int]],
      [_checkIndex]: [dart.void, [core.int, core.int]],
      [_checkSublistArguments]: [core.int, [core.int, core.int, core.int]],
      get: [typed_data.Int32x4, [core.int]],
      set: [dart.void, [core.int, typed_data.Int32x4]],
      sublist: [core.List$(typed_data.Int32x4), [core.int], [core.int]]
    })
  });
  dart.defineExtensionMembers(NativeInt32x4List, ['get', 'set', 'sublist', 'length']);
  class NativeFloat64x2List extends dart.mixin(core.Object, collection.ListMixin$(typed_data.Float64x2), _internal.FixedLengthListMixin$(typed_data.Float64x2)) {
    NativeFloat64x2List(length) {
      this[_storage] = NativeFloat64List.new(dart.notNull(length) * 2);
    }
    _externalStorage(storage) {
      this[_storage] = storage;
    }
    _slowFromList(list) {
      this[_storage] = NativeFloat64List.new(dart.notNull(list[dartx.length]) * 2);
      for (let i = 0; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
        let e = list[dartx.get](i);
        this[_storage].set(dart.notNull(i) * 2 + 0, e.x);
        this[_storage].set(dart.notNull(i) * 2 + 1, e.y);
      }
    }
    static fromList(list) {
      if (dart.is(list, NativeFloat64x2List)) {
        return new NativeFloat64x2List._externalStorage(NativeFloat64List.fromList(list[_storage]));
      } else {
        return new NativeFloat64x2List._slowFromList(list);
      }
    }
    get runtimeType() {
      return typed_data.Float64x2List;
    }
    get buffer() {
      return this[_storage].buffer;
    }
    get lengthInBytes() {
      return this[_storage].lengthInBytes;
    }
    get offsetInBytes() {
      return this[_storage].offsetInBytes;
    }
    get elementSizeInBytes() {
      return typed_data.Float64x2List.BYTES_PER_ELEMENT;
    }
    [_invalidIndex](index, length) {
      if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(length)) {
        if (length == this.length) {
          dart.throw(core.RangeError.index(index, this));
        }
        dart.throw(new core.RangeError.range(index, 0, dart.notNull(length) - 1));
      } else {
        dart.throw(new core.ArgumentError(`Invalid list index ${index}`));
      }
    }
    [_checkIndex](index, length) {
      if (index >>> 0 != index || dart.notNull(index) >= dart.notNull(length)) {
        this[_invalidIndex](index, length);
      }
    }
    [_checkSublistArguments](start, end, length) {
      this[_checkIndex](start, dart.notNull(length) + 1);
      if (end == null)
        return length;
      this[_checkIndex](end, dart.notNull(length) + 1);
      if (dart.notNull(start) > dart.notNull(end))
        dart.throw(new core.RangeError.range(start, 0, end));
      return end;
    }
    get length() {
      return (dart.notNull(this[_storage].length) / 2)[dartx.truncate]();
    }
    get(index) {
      this[_checkIndex](index, this.length);
      let _x = this[_storage].get(dart.notNull(index) * 2 + 0);
      let _y = this[_storage].get(dart.notNull(index) * 2 + 1);
      return typed_data.Float64x2.new(_x, _y);
    }
    set(index, value) {
      this[_checkIndex](index, this.length);
      this[_storage].set(dart.notNull(index) * 2 + 0, value.x);
      this[_storage].set(dart.notNull(index) * 2 + 1, value.y);
      return value;
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      return new NativeFloat64x2List._externalStorage(dart.as(this[_storage].sublist(dart.notNull(start) * 2, dart.notNull(end) * 2), NativeFloat64List));
    }
  }
  NativeFloat64x2List[dart.implements] = () => [typed_data.Float64x2List];
  dart.defineNamedConstructor(NativeFloat64x2List, '_externalStorage');
  dart.defineNamedConstructor(NativeFloat64x2List, '_slowFromList');
  dart.setSignature(NativeFloat64x2List, {
    constructors: () => ({
      NativeFloat64x2List: [NativeFloat64x2List, [core.int]],
      _externalStorage: [NativeFloat64x2List, [NativeFloat64List]],
      _slowFromList: [NativeFloat64x2List, [core.List$(typed_data.Float64x2)]],
      fromList: [NativeFloat64x2List, [core.List$(typed_data.Float64x2)]]
    }),
    methods: () => ({
      [_invalidIndex]: [dart.void, [core.int, core.int]],
      [_checkIndex]: [dart.void, [core.int, core.int]],
      [_checkSublistArguments]: [core.int, [core.int, core.int, core.int]],
      get: [typed_data.Float64x2, [core.int]],
      set: [dart.void, [core.int, typed_data.Float64x2]],
      sublist: [core.List$(typed_data.Float64x2), [core.int], [core.int]]
    })
  });
  dart.defineExtensionMembers(NativeFloat64x2List, ['get', 'set', 'sublist', 'length']);
  class NativeTypedData extends core.Object {
    [_invalidIndex](index, length) {
      if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(length)) {
        if (dart.is(this, core.List)) {
          if (dart.equals(length, dart.dload(this, 'length'))) {
            dart.throw(core.RangeError.index(index, this));
          }
        }
        dart.throw(new core.RangeError.range(index, 0, dart.notNull(length) - 1));
      } else {
        dart.throw(new core.ArgumentError(`Invalid list index ${index}`));
      }
    }
    [_checkIndex](index, length) {
      if (index >>> 0 !== index || index >= dart.notNull(length)) {
        this[_invalidIndex](index, length);
      }
    }
    [_checkSublistArguments](start, end, length) {
      this[_checkIndex](start, dart.notNull(length) + 1);
      if (end == null)
        return length;
      this[_checkIndex](end, dart.notNull(length) + 1);
      if (dart.notNull(start) > dart.notNull(end))
        dart.throw(new core.RangeError.range(start, 0, end));
      return end;
    }
  }
  NativeTypedData[dart.implements] = () => [typed_data.TypedData];
  dart.setSignature(NativeTypedData, {
    methods: () => ({
      [_invalidIndex]: [dart.void, [core.int, core.int]],
      [_checkIndex]: [dart.void, [core.int, core.int]],
      [_checkSublistArguments]: [core.int, [core.int, core.int, core.int]]
    })
  });
  NativeTypedData[dart.metadata] = () => [dart.const(new _js_helper.Native("ArrayBufferView"))];
  function _checkLength(length) {
    if (!(typeof length == 'number'))
      dart.throw(new core.ArgumentError(`Invalid length ${length}`));
    return dart.as(length, core.int);
  }
  dart.fn(_checkLength, core.int, [dart.dynamic]);
  function _checkViewArguments(buffer, offsetInBytes, length) {
    if (!dart.is(buffer, NativeByteBuffer)) {
      dart.throw(new core.ArgumentError('Invalid view buffer'));
    }
    if (!(typeof offsetInBytes == 'number')) {
      dart.throw(new core.ArgumentError(`Invalid view offsetInBytes ${offsetInBytes}`));
    }
    if (length != null && !(typeof length == 'number')) {
      dart.throw(new core.ArgumentError(`Invalid view length ${length}`));
    }
  }
  dart.fn(_checkViewArguments, dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]);
  function _ensureNativeList(list) {
    if (dart.is(list, _interceptors.JSIndexable))
      return list;
    let result = core.List.new(list[dartx.length]);
    for (let i = 0; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
      result[dartx.set](i, list[dartx.get](i));
    }
    return result;
  }
  dart.fn(_ensureNativeList, core.List, [core.List]);
  const _getFloat32 = Symbol('_getFloat32');
  const _getFloat64 = Symbol('_getFloat64');
  const _getInt16 = Symbol('_getInt16');
  const _getInt32 = Symbol('_getInt32');
  const _getUint16 = Symbol('_getUint16');
  const _getUint32 = Symbol('_getUint32');
  const _setFloat32 = Symbol('_setFloat32');
  const _setFloat64 = Symbol('_setFloat64');
  const _setInt16 = Symbol('_setInt16');
  const _setInt32 = Symbol('_setInt32');
  const _setUint16 = Symbol('_setUint16');
  const _setUint32 = Symbol('_setUint32');
  class NativeByteData extends NativeTypedData {
    static new(length) {
      return NativeByteData._create1(_checkLength(length));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeByteData._create2(buffer, offsetInBytes) : NativeByteData._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.ByteData;
    }
    get elementSizeInBytes() {
      return 1;
    }
    getFloat32(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getFloat32](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getFloat64(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getFloat64](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getInt16(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getInt16](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getInt32(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getInt32](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getInt64(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      dart.throw(new core.UnsupportedError('Int64 accessor not supported by dart2js.'));
    }
    getUint16(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getUint16](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getUint32(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_getUint32](byteOffset, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    getUint64(byteOffset, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      dart.throw(new core.UnsupportedError('Uint64 accessor not supported by dart2js.'));
    }
    setFloat32(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setFloat32](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setFloat64(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setFloat64](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setInt16(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setInt16](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setInt32(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setInt32](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setInt64(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      dart.throw(new core.UnsupportedError('Int64 accessor not supported by dart2js.'));
    }
    setUint16(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setUint16](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setUint32(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      return this[_setUint32](byteOffset, value, dart.equals(typed_data.Endianness.LITTLE_ENDIAN, endian));
    }
    setUint64(byteOffset, value, endian) {
      if (endian === void 0)
        endian = typed_data.Endianness.BIG_ENDIAN;
      dart.throw(new core.UnsupportedError('Uint64 accessor not supported by dart2js.'));
    }
    static _create1(arg) {
      return dart.as(new DataView(new ArrayBuffer(arg)), NativeByteData);
    }
    static _create2(arg1, arg2) {
      return dart.as(new DataView(arg1, arg2), NativeByteData);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new DataView(arg1, arg2, arg3), NativeByteData);
    }
  }
  NativeByteData[dart.implements] = () => [typed_data.ByteData];
  dart.setSignature(NativeByteData, {
    constructors: () => ({
      new: [NativeByteData, [core.int]],
      view: [NativeByteData, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      getFloat32: [core.double, [core.int], [typed_data.Endianness]],
      [_getFloat32]: [core.double, [core.int], [core.bool]],
      getFloat64: [core.double, [core.int], [typed_data.Endianness]],
      [_getFloat64]: [core.double, [core.int], [core.bool]],
      getInt16: [core.int, [core.int], [typed_data.Endianness]],
      [_getInt16]: [core.int, [core.int], [core.bool]],
      getInt32: [core.int, [core.int], [typed_data.Endianness]],
      [_getInt32]: [core.int, [core.int], [core.bool]],
      getInt64: [core.int, [core.int], [typed_data.Endianness]],
      getInt8: [core.int, [core.int]],
      getUint16: [core.int, [core.int], [typed_data.Endianness]],
      [_getUint16]: [core.int, [core.int], [core.bool]],
      getUint32: [core.int, [core.int], [typed_data.Endianness]],
      [_getUint32]: [core.int, [core.int], [core.bool]],
      getUint64: [core.int, [core.int], [typed_data.Endianness]],
      getUint8: [core.int, [core.int]],
      setFloat32: [dart.void, [core.int, core.num], [typed_data.Endianness]],
      [_setFloat32]: [dart.void, [core.int, core.num], [core.bool]],
      setFloat64: [dart.void, [core.int, core.num], [typed_data.Endianness]],
      [_setFloat64]: [dart.void, [core.int, core.num], [core.bool]],
      setInt16: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      [_setInt16]: [dart.void, [core.int, core.int], [core.bool]],
      setInt32: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      [_setInt32]: [dart.void, [core.int, core.int], [core.bool]],
      setInt64: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      setInt8: [dart.void, [core.int, core.int]],
      setUint16: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      [_setUint16]: [dart.void, [core.int, core.int], [core.bool]],
      setUint32: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      [_setUint32]: [dart.void, [core.int, core.int], [core.bool]],
      setUint64: [dart.void, [core.int, core.int], [typed_data.Endianness]],
      setUint8: [dart.void, [core.int, core.int]]
    }),
    statics: () => ({
      _create1: [NativeByteData, [dart.dynamic]],
      _create2: [NativeByteData, [dart.dynamic, dart.dynamic]],
      _create3: [NativeByteData, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  NativeByteData[dart.metadata] = () => [dart.const(new _js_helper.Native("DataView"))];
  const _setRangeFast = Symbol('_setRangeFast');
  class NativeTypedArray extends NativeTypedData {
    [_setRangeFast](start, end, source, skipCount) {
      let targetLength = this.length;
      this[_checkIndex](start, dart.notNull(targetLength) + 1);
      this[_checkIndex](end, dart.notNull(targetLength) + 1);
      if (dart.notNull(start) > dart.notNull(end))
        dart.throw(new core.RangeError.range(start, 0, end));
      let count = dart.notNull(end) - dart.notNull(start);
      if (dart.notNull(skipCount) < 0)
        dart.throw(new core.ArgumentError(skipCount));
      let sourceLength = source.length;
      if (dart.notNull(sourceLength) - dart.notNull(skipCount) < dart.notNull(count)) {
        dart.throw(new core.StateError('Not enough elements'));
      }
      if (skipCount != 0 || sourceLength != count) {
        source = dart.as(source.subarray(skipCount, dart.notNull(skipCount) + dart.notNull(count)), NativeTypedArray);
      }
      this.set(source, start);
    }
  }
  NativeTypedArray[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior];
  dart.setSignature(NativeTypedArray, {
    methods: () => ({[_setRangeFast]: [dart.void, [core.int, core.int, NativeTypedArray, core.int]]})
  });
  class NativeTypedArrayOfDouble extends dart.mixin(NativeTypedArray, collection.ListMixin$(core.double), _internal.FixedLengthListMixin$(core.double)) {
    get length() {
      return this.length;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    set(index, value) {
      this[_checkIndex](index, this.length);
      this[index] = value;
      return value;
    }
    setRange(start, end, iterable, skipCount) {
      if (skipCount === void 0)
        skipCount = 0;
      if (dart.is(iterable, NativeTypedArrayOfDouble)) {
        this[_setRangeFast](start, end, iterable, skipCount);
        return;
      }
      super.setRange(start, end, iterable, skipCount);
    }
  }
  dart.setSignature(NativeTypedArrayOfDouble, {
    methods: () => ({
      get: [core.double, [core.int]],
      set: [dart.void, [core.int, core.num]],
      setRange: [dart.void, [core.int, core.int, core.Iterable$(core.double)], [core.int]]
    })
  });
  dart.defineExtensionMembers(NativeTypedArrayOfDouble, ['get', 'set', 'setRange', 'length']);
  class NativeTypedArrayOfInt extends dart.mixin(NativeTypedArray, collection.ListMixin$(core.int), _internal.FixedLengthListMixin$(core.int)) {
    get length() {
      return this.length;
    }
    set(index, value) {
      this[_checkIndex](index, this.length);
      this[index] = value;
      return value;
    }
    setRange(start, end, iterable, skipCount) {
      if (skipCount === void 0)
        skipCount = 0;
      if (dart.is(iterable, NativeTypedArrayOfInt)) {
        this[_setRangeFast](start, end, iterable, skipCount);
        return;
      }
      super.setRange(start, end, iterable, skipCount);
    }
  }
  NativeTypedArrayOfInt[dart.implements] = () => [core.List$(core.int)];
  dart.setSignature(NativeTypedArrayOfInt, {
    methods: () => ({
      set: [dart.void, [core.int, core.int]],
      setRange: [dart.void, [core.int, core.int, core.Iterable$(core.int)], [core.int]]
    })
  });
  dart.defineExtensionMembers(NativeTypedArrayOfInt, ['set', 'setRange', 'length']);
  class NativeFloat32List extends NativeTypedArrayOfDouble {
    static new(length) {
      return NativeFloat32List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeFloat32List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeFloat32List._create2(buffer, offsetInBytes) : NativeFloat32List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Float32List;
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeFloat32List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Float32Array(arg), NativeFloat32List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Float32Array(arg1, arg2), NativeFloat32List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Float32Array(arg1, arg2, arg3), NativeFloat32List);
    }
  }
  NativeFloat32List[dart.implements] = () => [typed_data.Float32List];
  dart.setSignature(NativeFloat32List, {
    constructors: () => ({
      new: [NativeFloat32List, [core.int]],
      fromList: [NativeFloat32List, [core.List$(core.double)]],
      view: [NativeFloat32List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({sublist: [core.List$(core.double), [core.int], [core.int]]}),
    statics: () => ({
      _create1: [NativeFloat32List, [dart.dynamic]],
      _create2: [NativeFloat32List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeFloat32List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeFloat32List, ['sublist']);
  NativeFloat32List[dart.metadata] = () => [dart.const(new _js_helper.Native("Float32Array"))];
  class NativeFloat64List extends NativeTypedArrayOfDouble {
    static new(length) {
      return NativeFloat64List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeFloat64List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeFloat64List._create2(buffer, offsetInBytes) : NativeFloat64List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Float64List;
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeFloat64List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Float64Array(arg), NativeFloat64List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Float64Array(arg1, arg2), NativeFloat64List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Float64Array(arg1, arg2, arg3), NativeFloat64List);
    }
  }
  NativeFloat64List[dart.implements] = () => [typed_data.Float64List];
  dart.setSignature(NativeFloat64List, {
    constructors: () => ({
      new: [NativeFloat64List, [core.int]],
      fromList: [NativeFloat64List, [core.List$(core.double)]],
      view: [NativeFloat64List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({sublist: [core.List$(core.double), [core.int], [core.int]]}),
    statics: () => ({
      _create1: [NativeFloat64List, [dart.dynamic]],
      _create2: [NativeFloat64List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeFloat64List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeFloat64List, ['sublist']);
  NativeFloat64List[dart.metadata] = () => [dart.const(new _js_helper.Native("Float64Array"))];
  class NativeInt16List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeInt16List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeInt16List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeInt16List._create2(buffer, offsetInBytes) : NativeInt16List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Int16List;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeInt16List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Int16Array(arg), NativeInt16List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Int16Array(arg1, arg2), NativeInt16List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Int16Array(arg1, arg2, arg3), NativeInt16List);
    }
  }
  NativeInt16List[dart.implements] = () => [typed_data.Int16List];
  dart.setSignature(NativeInt16List, {
    constructors: () => ({
      new: [NativeInt16List, [core.int]],
      fromList: [NativeInt16List, [core.List$(core.int)]],
      view: [NativeInt16List, [NativeByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeInt16List, [dart.dynamic]],
      _create2: [NativeInt16List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeInt16List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeInt16List, ['get', 'sublist']);
  NativeInt16List[dart.metadata] = () => [dart.const(new _js_helper.Native("Int16Array"))];
  class NativeInt32List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeInt32List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeInt32List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeInt32List._create2(buffer, offsetInBytes) : NativeInt32List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Int32List;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeInt32List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Int32Array(arg), NativeInt32List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Int32Array(arg1, arg2), NativeInt32List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Int32Array(arg1, arg2, arg3), NativeInt32List);
    }
  }
  NativeInt32List[dart.implements] = () => [typed_data.Int32List];
  dart.setSignature(NativeInt32List, {
    constructors: () => ({
      new: [NativeInt32List, [core.int]],
      fromList: [NativeInt32List, [core.List$(core.int)]],
      view: [NativeInt32List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeInt32List, [dart.dynamic]],
      _create2: [NativeInt32List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeInt32List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeInt32List, ['get', 'sublist']);
  NativeInt32List[dart.metadata] = () => [dart.const(new _js_helper.Native("Int32Array"))];
  class NativeInt8List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeInt8List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeInt8List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return dart.as(length == null ? NativeInt8List._create2(buffer, offsetInBytes) : NativeInt8List._create3(buffer, offsetInBytes, length), NativeInt8List);
    }
    get runtimeType() {
      return typed_data.Int8List;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeInt8List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Int8Array(arg), NativeInt8List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Int8Array(arg1, arg2), NativeInt8List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Int8Array(arg1, arg2, arg3), typed_data.Int8List);
    }
  }
  NativeInt8List[dart.implements] = () => [typed_data.Int8List];
  dart.setSignature(NativeInt8List, {
    constructors: () => ({
      new: [NativeInt8List, [core.int]],
      fromList: [NativeInt8List, [core.List$(core.int)]],
      view: [NativeInt8List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeInt8List, [dart.dynamic]],
      _create2: [NativeInt8List, [dart.dynamic, dart.dynamic]],
      _create3: [typed_data.Int8List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeInt8List, ['get', 'sublist']);
  NativeInt8List[dart.metadata] = () => [dart.const(new _js_helper.Native("Int8Array"))];
  class NativeUint16List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeUint16List._create1(_checkLength(length));
    }
    static fromList(list) {
      return NativeUint16List._create1(_ensureNativeList(list));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeUint16List._create2(buffer, offsetInBytes) : NativeUint16List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Uint16List;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeUint16List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Uint16Array(arg), NativeUint16List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Uint16Array(arg1, arg2), NativeUint16List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Uint16Array(arg1, arg2, arg3), NativeUint16List);
    }
  }
  NativeUint16List[dart.implements] = () => [typed_data.Uint16List];
  dart.setSignature(NativeUint16List, {
    constructors: () => ({
      new: [NativeUint16List, [core.int]],
      fromList: [NativeUint16List, [core.List$(core.int)]],
      view: [NativeUint16List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeUint16List, [dart.dynamic]],
      _create2: [NativeUint16List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeUint16List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeUint16List, ['get', 'sublist']);
  NativeUint16List[dart.metadata] = () => [dart.const(new _js_helper.Native("Uint16Array"))];
  class NativeUint32List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeUint32List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeUint32List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeUint32List._create2(buffer, offsetInBytes) : NativeUint32List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Uint32List;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeUint32List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Uint32Array(arg), NativeUint32List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Uint32Array(arg1, arg2), NativeUint32List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Uint32Array(arg1, arg2, arg3), NativeUint32List);
    }
  }
  NativeUint32List[dart.implements] = () => [typed_data.Uint32List];
  dart.setSignature(NativeUint32List, {
    constructors: () => ({
      new: [NativeUint32List, [core.int]],
      fromList: [NativeUint32List, [core.List$(core.int)]],
      view: [NativeUint32List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeUint32List, [dart.dynamic]],
      _create2: [NativeUint32List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeUint32List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeUint32List, ['get', 'sublist']);
  NativeUint32List[dart.metadata] = () => [dart.const(new _js_helper.Native("Uint32Array"))];
  class NativeUint8ClampedList extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeUint8ClampedList._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeUint8ClampedList._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeUint8ClampedList._create2(buffer, offsetInBytes) : NativeUint8ClampedList._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Uint8ClampedList;
    }
    get length() {
      return this.length;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeUint8ClampedList._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Uint8ClampedArray(arg), NativeUint8ClampedList);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Uint8ClampedArray(arg1, arg2), NativeUint8ClampedList);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Uint8ClampedArray(arg1, arg2, arg3), NativeUint8ClampedList);
    }
  }
  NativeUint8ClampedList[dart.implements] = () => [typed_data.Uint8ClampedList];
  dart.setSignature(NativeUint8ClampedList, {
    constructors: () => ({
      new: [NativeUint8ClampedList, [core.int]],
      fromList: [NativeUint8ClampedList, [core.List$(core.int)]],
      view: [NativeUint8ClampedList, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeUint8ClampedList, [dart.dynamic]],
      _create2: [NativeUint8ClampedList, [dart.dynamic, dart.dynamic]],
      _create3: [NativeUint8ClampedList, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeUint8ClampedList, ['get', 'sublist', 'length']);
  NativeUint8ClampedList[dart.metadata] = () => [dart.const(new _js_helper.Native("Uint8ClampedArray,CanvasPixelArray"))];
  class NativeUint8List extends NativeTypedArrayOfInt {
    static new(length) {
      return NativeUint8List._create1(_checkLength(length));
    }
    static fromList(elements) {
      return NativeUint8List._create1(_ensureNativeList(elements));
    }
    static view(buffer, offsetInBytes, length) {
      _checkViewArguments(buffer, offsetInBytes, length);
      return length == null ? NativeUint8List._create2(buffer, offsetInBytes) : NativeUint8List._create3(buffer, offsetInBytes, length);
    }
    get runtimeType() {
      return typed_data.Uint8List;
    }
    get length() {
      return this.length;
    }
    get(index) {
      this[_checkIndex](index, this.length);
      return this[index];
    }
    sublist(start, end) {
      if (end === void 0)
        end = null;
      end = this[_checkSublistArguments](start, end, this.length);
      let source = this.subarray(start, end);
      return NativeUint8List._create1(source);
    }
    static _create1(arg) {
      return dart.as(new Uint8Array(arg), NativeUint8List);
    }
    static _create2(arg1, arg2) {
      return dart.as(new Uint8Array(arg1, arg2), NativeUint8List);
    }
    static _create3(arg1, arg2, arg3) {
      return dart.as(new Uint8Array(arg1, arg2, arg3), NativeUint8List);
    }
  }
  NativeUint8List[dart.implements] = () => [typed_data.Uint8List];
  dart.setSignature(NativeUint8List, {
    constructors: () => ({
      new: [NativeUint8List, [core.int]],
      fromList: [NativeUint8List, [core.List$(core.int)]],
      view: [NativeUint8List, [typed_data.ByteBuffer, core.int, core.int]]
    }),
    methods: () => ({
      get: [core.int, [core.int]],
      sublist: [core.List$(core.int), [core.int], [core.int]]
    }),
    statics: () => ({
      _create1: [NativeUint8List, [dart.dynamic]],
      _create2: [NativeUint8List, [dart.dynamic, dart.dynamic]],
      _create3: [NativeUint8List, [dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['_create1', '_create2', '_create3']
  });
  dart.defineExtensionMembers(NativeUint8List, ['get', 'sublist', 'length']);
  NativeUint8List[dart.metadata] = () => [dart.const(new _js_helper.Native("Uint8Array,!nonleaf"))];
  class NativeFloat32x4 extends core.Object {
    static _truncate(x) {
      NativeFloat32x4._list.set(0, dart.as(x, core.num));
      return NativeFloat32x4._list.get(0);
    }
    NativeFloat32x4(x, y, z, w) {
      this.x = dart.as(NativeFloat32x4._truncate(x), core.double);
      this.y = dart.as(NativeFloat32x4._truncate(y), core.double);
      this.z = dart.as(NativeFloat32x4._truncate(z), core.double);
      this.w = dart.as(NativeFloat32x4._truncate(w), core.double);
      if (!(typeof x == 'number'))
        dart.throw(new core.ArgumentError(x));
      if (!(typeof y == 'number'))
        dart.throw(new core.ArgumentError(y));
      if (!(typeof z == 'number'))
        dart.throw(new core.ArgumentError(z));
      if (!(typeof w == 'number'))
        dart.throw(new core.ArgumentError(w));
    }
    splat(v) {
      this.NativeFloat32x4(v, v, v, v);
    }
    zero() {
      this._truncated(0.0, 0.0, 0.0, 0.0);
    }
    static fromInt32x4Bits(i) {
      NativeFloat32x4._uint32view.set(0, i.x);
      NativeFloat32x4._uint32view.set(1, i.y);
      NativeFloat32x4._uint32view.set(2, i.z);
      NativeFloat32x4._uint32view.set(3, i.w);
      return new NativeFloat32x4._truncated(NativeFloat32x4._list.get(0), NativeFloat32x4._list.get(1), NativeFloat32x4._list.get(2), NativeFloat32x4._list.get(3));
    }
    fromFloat64x2(v) {
      this._truncated(dart.as(NativeFloat32x4._truncate(v.x), core.double), dart.as(NativeFloat32x4._truncate(v.y), core.double), 0.0, 0.0);
    }
    _doubles(x, y, z, w) {
      this.x = dart.as(NativeFloat32x4._truncate(x), core.double);
      this.y = dart.as(NativeFloat32x4._truncate(y), core.double);
      this.z = dart.as(NativeFloat32x4._truncate(z), core.double);
      this.w = dart.as(NativeFloat32x4._truncate(w), core.double);
    }
    _truncated(x, y, z, w) {
      this.x = x;
      this.y = y;
      this.z = z;
      this.w = w;
    }
    toString() {
      return `[${this.x}, ${this.y}, ${this.z}, ${this.w}]`;
    }
    ['+'](other) {
      let _x = dart.notNull(this.x) + dart.notNull(other.x);
      let _y = dart.notNull(this.y) + dart.notNull(other.y);
      let _z = dart.notNull(this.z) + dart.notNull(other.z);
      let _w = dart.notNull(this.w) + dart.notNull(other.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    ['unary-']() {
      return new NativeFloat32x4._truncated(-dart.notNull(this.x), -dart.notNull(this.y), -dart.notNull(this.z), -dart.notNull(this.w));
    }
    ['-'](other) {
      let _x = dart.notNull(this.x) - dart.notNull(other.x);
      let _y = dart.notNull(this.y) - dart.notNull(other.y);
      let _z = dart.notNull(this.z) - dart.notNull(other.z);
      let _w = dart.notNull(this.w) - dart.notNull(other.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    ['*'](other) {
      let _x = dart.notNull(this.x) * dart.notNull(other.x);
      let _y = dart.notNull(this.y) * dart.notNull(other.y);
      let _z = dart.notNull(this.z) * dart.notNull(other.z);
      let _w = dart.notNull(this.w) * dart.notNull(other.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    ['/'](other) {
      let _x = dart.notNull(this.x) / dart.notNull(other.x);
      let _y = dart.notNull(this.y) / dart.notNull(other.y);
      let _z = dart.notNull(this.z) / dart.notNull(other.z);
      let _w = dart.notNull(this.w) / dart.notNull(other.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    lessThan(other) {
      let _cx = dart.notNull(this.x) < dart.notNull(other.x);
      let _cy = dart.notNull(this.y) < dart.notNull(other.y);
      let _cz = dart.notNull(this.z) < dart.notNull(other.z);
      let _cw = dart.notNull(this.w) < dart.notNull(other.w);
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    lessThanOrEqual(other) {
      let _cx = dart.notNull(this.x) <= dart.notNull(other.x);
      let _cy = dart.notNull(this.y) <= dart.notNull(other.y);
      let _cz = dart.notNull(this.z) <= dart.notNull(other.z);
      let _cw = dart.notNull(this.w) <= dart.notNull(other.w);
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    greaterThan(other) {
      let _cx = dart.notNull(this.x) > dart.notNull(other.x);
      let _cy = dart.notNull(this.y) > dart.notNull(other.y);
      let _cz = dart.notNull(this.z) > dart.notNull(other.z);
      let _cw = dart.notNull(this.w) > dart.notNull(other.w);
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    greaterThanOrEqual(other) {
      let _cx = dart.notNull(this.x) >= dart.notNull(other.x);
      let _cy = dart.notNull(this.y) >= dart.notNull(other.y);
      let _cz = dart.notNull(this.z) >= dart.notNull(other.z);
      let _cw = dart.notNull(this.w) >= dart.notNull(other.w);
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    equal(other) {
      let _cx = this.x == other.x;
      let _cy = this.y == other.y;
      let _cz = this.z == other.z;
      let _cw = this.w == other.w;
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    notEqual(other) {
      let _cx = this.x != other.x;
      let _cy = this.y != other.y;
      let _cz = this.z != other.z;
      let _cw = this.w != other.w;
      return new NativeInt32x4._truncated(dart.notNull(_cx) ? -1 : 0, dart.notNull(_cy) ? -1 : 0, dart.notNull(_cz) ? -1 : 0, dart.notNull(_cw) ? -1 : 0);
    }
    scale(s) {
      let _x = dart.notNull(s) * dart.notNull(this.x);
      let _y = dart.notNull(s) * dart.notNull(this.y);
      let _z = dart.notNull(s) * dart.notNull(this.z);
      let _w = dart.notNull(s) * dart.notNull(this.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    abs() {
      let _x = this.x[dartx.abs]();
      let _y = this.y[dartx.abs]();
      let _z = this.z[dartx.abs]();
      let _w = this.w[dartx.abs]();
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    clamp(lowerLimit, upperLimit) {
      let _lx = lowerLimit.x;
      let _ly = lowerLimit.y;
      let _lz = lowerLimit.z;
      let _lw = lowerLimit.w;
      let _ux = upperLimit.x;
      let _uy = upperLimit.y;
      let _uz = upperLimit.z;
      let _uw = upperLimit.w;
      let _x = this.x;
      let _y = this.y;
      let _z = this.z;
      let _w = this.w;
      _x = dart.notNull(_x) > dart.notNull(_ux) ? _ux : _x;
      _y = dart.notNull(_y) > dart.notNull(_uy) ? _uy : _y;
      _z = dart.notNull(_z) > dart.notNull(_uz) ? _uz : _z;
      _w = dart.notNull(_w) > dart.notNull(_uw) ? _uw : _w;
      _x = dart.notNull(_x) < dart.notNull(_lx) ? _lx : _x;
      _y = dart.notNull(_y) < dart.notNull(_ly) ? _ly : _y;
      _z = dart.notNull(_z) < dart.notNull(_lz) ? _lz : _z;
      _w = dart.notNull(_w) < dart.notNull(_lw) ? _lw : _w;
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    get signMask() {
      let view = NativeFloat32x4._uint32view;
      let mx = null, my = null, mz = null, mw = null;
      NativeFloat32x4._list.set(0, this.x);
      NativeFloat32x4._list.set(1, this.y);
      NativeFloat32x4._list.set(2, this.z);
      NativeFloat32x4._list.set(3, this.w);
      mx = (dart.notNull(view.get(0)) & 2147483648) >> 31;
      my = (dart.notNull(view.get(1)) & 2147483648) >> 30;
      mz = (dart.notNull(view.get(2)) & 2147483648) >> 29;
      mw = (dart.notNull(view.get(3)) & 2147483648) >> 28;
      return dart.as(dart.dsend(dart.dsend(dart.dsend(mx, '|', my), '|', mz), '|', mw), core.int);
    }
    shuffle(m) {
      if (dart.notNull(m) < 0 || dart.notNull(m) > 255) {
        dart.throw(new core.RangeError(`mask ${m} must be in the range [0..256)`));
      }
      NativeFloat32x4._list.set(0, this.x);
      NativeFloat32x4._list.set(1, this.y);
      NativeFloat32x4._list.set(2, this.z);
      NativeFloat32x4._list.set(3, this.w);
      let _x = NativeFloat32x4._list.get(dart.notNull(m) & 3);
      let _y = NativeFloat32x4._list.get(dart.notNull(m) >> 2 & 3);
      let _z = NativeFloat32x4._list.get(dart.notNull(m) >> 4 & 3);
      let _w = NativeFloat32x4._list.get(dart.notNull(m) >> 6 & 3);
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    shuffleMix(other, m) {
      if (dart.notNull(m) < 0 || dart.notNull(m) > 255) {
        dart.throw(new core.RangeError(`mask ${m} must be in the range [0..256)`));
      }
      NativeFloat32x4._list.set(0, this.x);
      NativeFloat32x4._list.set(1, this.y);
      NativeFloat32x4._list.set(2, this.z);
      NativeFloat32x4._list.set(3, this.w);
      let _x = NativeFloat32x4._list.get(dart.notNull(m) & 3);
      let _y = NativeFloat32x4._list.get(dart.notNull(m) >> 2 & 3);
      NativeFloat32x4._list.set(0, other.x);
      NativeFloat32x4._list.set(1, other.y);
      NativeFloat32x4._list.set(2, other.z);
      NativeFloat32x4._list.set(3, other.w);
      let _z = NativeFloat32x4._list.get(dart.notNull(m) >> 4 & 3);
      let _w = NativeFloat32x4._list.get(dart.notNull(m) >> 6 & 3);
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    withX(newX) {
      return new NativeFloat32x4._truncated(dart.as(NativeFloat32x4._truncate(newX), core.double), this.y, this.z, this.w);
    }
    withY(newY) {
      return new NativeFloat32x4._truncated(this.x, dart.as(NativeFloat32x4._truncate(newY), core.double), this.z, this.w);
    }
    withZ(newZ) {
      return new NativeFloat32x4._truncated(this.x, this.y, dart.as(NativeFloat32x4._truncate(newZ), core.double), this.w);
    }
    withW(newW) {
      return new NativeFloat32x4._truncated(this.x, this.y, this.z, dart.as(NativeFloat32x4._truncate(newW), core.double));
    }
    min(other) {
      let _x = dart.notNull(this.x) < dart.notNull(other.x) ? this.x : other.x;
      let _y = dart.notNull(this.y) < dart.notNull(other.y) ? this.y : other.y;
      let _z = dart.notNull(this.z) < dart.notNull(other.z) ? this.z : other.z;
      let _w = dart.notNull(this.w) < dart.notNull(other.w) ? this.w : other.w;
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    max(other) {
      let _x = dart.notNull(this.x) > dart.notNull(other.x) ? this.x : other.x;
      let _y = dart.notNull(this.y) > dart.notNull(other.y) ? this.y : other.y;
      let _z = dart.notNull(this.z) > dart.notNull(other.z) ? this.z : other.z;
      let _w = dart.notNull(this.w) > dart.notNull(other.w) ? this.w : other.w;
      return new NativeFloat32x4._truncated(_x, _y, _z, _w);
    }
    sqrt() {
      let _x = math.sqrt(this.x);
      let _y = math.sqrt(this.y);
      let _z = math.sqrt(this.z);
      let _w = math.sqrt(this.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    reciprocal() {
      let _x = 1.0 / dart.notNull(this.x);
      let _y = 1.0 / dart.notNull(this.y);
      let _z = 1.0 / dart.notNull(this.z);
      let _w = 1.0 / dart.notNull(this.w);
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
    reciprocalSqrt() {
      let _x = math.sqrt(1.0 / dart.notNull(this.x));
      let _y = math.sqrt(1.0 / dart.notNull(this.y));
      let _z = math.sqrt(1.0 / dart.notNull(this.z));
      let _w = math.sqrt(1.0 / dart.notNull(this.w));
      return new NativeFloat32x4._doubles(_x, _y, _z, _w);
    }
  }
  NativeFloat32x4[dart.implements] = () => [typed_data.Float32x4];
  dart.defineNamedConstructor(NativeFloat32x4, 'splat');
  dart.defineNamedConstructor(NativeFloat32x4, 'zero');
  dart.defineNamedConstructor(NativeFloat32x4, 'fromFloat64x2');
  dart.defineNamedConstructor(NativeFloat32x4, '_doubles');
  dart.defineNamedConstructor(NativeFloat32x4, '_truncated');
  dart.setSignature(NativeFloat32x4, {
    constructors: () => ({
      NativeFloat32x4: [NativeFloat32x4, [core.double, core.double, core.double, core.double]],
      splat: [NativeFloat32x4, [core.double]],
      zero: [NativeFloat32x4, []],
      fromInt32x4Bits: [NativeFloat32x4, [typed_data.Int32x4]],
      fromFloat64x2: [NativeFloat32x4, [typed_data.Float64x2]],
      _doubles: [NativeFloat32x4, [core.double, core.double, core.double, core.double]],
      _truncated: [NativeFloat32x4, [core.double, core.double, core.double, core.double]]
    }),
    methods: () => ({
      '+': [typed_data.Float32x4, [typed_data.Float32x4]],
      'unary-': [typed_data.Float32x4, []],
      '-': [typed_data.Float32x4, [typed_data.Float32x4]],
      '*': [typed_data.Float32x4, [typed_data.Float32x4]],
      '/': [typed_data.Float32x4, [typed_data.Float32x4]],
      lessThan: [typed_data.Int32x4, [typed_data.Float32x4]],
      lessThanOrEqual: [typed_data.Int32x4, [typed_data.Float32x4]],
      greaterThan: [typed_data.Int32x4, [typed_data.Float32x4]],
      greaterThanOrEqual: [typed_data.Int32x4, [typed_data.Float32x4]],
      equal: [typed_data.Int32x4, [typed_data.Float32x4]],
      notEqual: [typed_data.Int32x4, [typed_data.Float32x4]],
      scale: [typed_data.Float32x4, [core.double]],
      abs: [typed_data.Float32x4, []],
      clamp: [typed_data.Float32x4, [typed_data.Float32x4, typed_data.Float32x4]],
      shuffle: [typed_data.Float32x4, [core.int]],
      shuffleMix: [typed_data.Float32x4, [typed_data.Float32x4, core.int]],
      withX: [typed_data.Float32x4, [core.double]],
      withY: [typed_data.Float32x4, [core.double]],
      withZ: [typed_data.Float32x4, [core.double]],
      withW: [typed_data.Float32x4, [core.double]],
      min: [typed_data.Float32x4, [typed_data.Float32x4]],
      max: [typed_data.Float32x4, [typed_data.Float32x4]],
      sqrt: [typed_data.Float32x4, []],
      reciprocal: [typed_data.Float32x4, []],
      reciprocalSqrt: [typed_data.Float32x4, []]
    }),
    statics: () => ({_truncate: [dart.dynamic, [dart.dynamic]]}),
    names: ['_truncate']
  });
  dart.defineLazyProperties(NativeFloat32x4, {
    get _list() {
      return NativeFloat32List.new(4);
    },
    get _uint32view() {
      return NativeFloat32x4._list.buffer.asUint32List();
    }
  });
  class NativeInt32x4 extends core.Object {
    static _truncate(x) {
      NativeInt32x4._list.set(0, dart.as(x, core.int));
      return NativeInt32x4._list.get(0);
    }
    NativeInt32x4(x, y, z, w) {
      this.x = dart.as(NativeInt32x4._truncate(x), core.int);
      this.y = dart.as(NativeInt32x4._truncate(y), core.int);
      this.z = dart.as(NativeInt32x4._truncate(z), core.int);
      this.w = dart.as(NativeInt32x4._truncate(w), core.int);
      if (x != this.x && !(typeof x == 'number'))
        dart.throw(new core.ArgumentError(x));
      if (y != this.y && !(typeof y == 'number'))
        dart.throw(new core.ArgumentError(y));
      if (z != this.z && !(typeof z == 'number'))
        dart.throw(new core.ArgumentError(z));
      if (w != this.w && !(typeof w == 'number'))
        dart.throw(new core.ArgumentError(w));
    }
    bool(x, y, z, w) {
      this.x = dart.notNull(x) ? -1 : 0;
      this.y = dart.notNull(y) ? -1 : 0;
      this.z = dart.notNull(z) ? -1 : 0;
      this.w = dart.notNull(w) ? -1 : 0;
    }
    static fromFloat32x4Bits(f) {
      let floatList = NativeFloat32x4._list;
      floatList.set(0, f.x);
      floatList.set(1, f.y);
      floatList.set(2, f.z);
      floatList.set(3, f.w);
      let view = dart.as(floatList.buffer.asInt32List(), NativeInt32List);
      return new NativeInt32x4._truncated(view.get(0), view.get(1), view.get(2), view.get(3));
    }
    _truncated(x, y, z, w) {
      this.x = x;
      this.y = y;
      this.z = z;
      this.w = w;
    }
    toString() {
      return `[${this.x}, ${this.y}, ${this.z}, ${this.w}]`;
    }
    ['|'](other) {
      return new NativeInt32x4._truncated(this.x | other.x, this.y | other.y, this.z | other.z, this.w | other.w);
    }
    ['&'](other) {
      return new NativeInt32x4._truncated(this.x & other.x, this.y & other.y, this.z & other.z, this.w & other.w);
    }
    ['^'](other) {
      return new NativeInt32x4._truncated(this.x ^ other.x, this.y ^ other.y, this.z ^ other.z, this.w ^ other.w);
    }
    ['+'](other) {
      return new NativeInt32x4._truncated(this.x + other.x | 0, this.y + other.y | 0, this.z + other.z | 0, this.w + other.w | 0);
    }
    ['-'](other) {
      return new NativeInt32x4._truncated(this.x - other.x | 0, this.y - other.y | 0, this.z - other.z | 0, this.w - other.w | 0);
    }
    ['unary-']() {
      return new NativeInt32x4._truncated(-this.x | 0, -this.y | 0, -this.z | 0, -this.w | 0);
    }
    get signMask() {
      let mx = (dart.notNull(this.x) & 2147483648) >> 31;
      let my = (dart.notNull(this.y) & 2147483648) >> 31;
      let mz = (dart.notNull(this.z) & 2147483648) >> 31;
      let mw = (dart.notNull(this.w) & 2147483648) >> 31;
      return dart.notNull(mx) | dart.notNull(my) << 1 | dart.notNull(mz) << 2 | dart.notNull(mw) << 3;
    }
    shuffle(mask) {
      if (dart.notNull(mask) < 0 || dart.notNull(mask) > 255) {
        dart.throw(new core.RangeError(`mask ${mask} must be in the range [0..256)`));
      }
      NativeInt32x4._list.set(0, this.x);
      NativeInt32x4._list.set(1, this.y);
      NativeInt32x4._list.set(2, this.z);
      NativeInt32x4._list.set(3, this.w);
      let _x = NativeInt32x4._list.get(dart.notNull(mask) & 3);
      let _y = NativeInt32x4._list.get(dart.notNull(mask) >> 2 & 3);
      let _z = NativeInt32x4._list.get(dart.notNull(mask) >> 4 & 3);
      let _w = NativeInt32x4._list.get(dart.notNull(mask) >> 6 & 3);
      return new NativeInt32x4._truncated(_x, _y, _z, _w);
    }
    shuffleMix(other, mask) {
      if (dart.notNull(mask) < 0 || dart.notNull(mask) > 255) {
        dart.throw(new core.RangeError(`mask ${mask} must be in the range [0..256)`));
      }
      NativeInt32x4._list.set(0, this.x);
      NativeInt32x4._list.set(1, this.y);
      NativeInt32x4._list.set(2, this.z);
      NativeInt32x4._list.set(3, this.w);
      let _x = NativeInt32x4._list.get(dart.notNull(mask) & 3);
      let _y = NativeInt32x4._list.get(dart.notNull(mask) >> 2 & 3);
      NativeInt32x4._list.set(0, other.x);
      NativeInt32x4._list.set(1, other.y);
      NativeInt32x4._list.set(2, other.z);
      NativeInt32x4._list.set(3, other.w);
      let _z = NativeInt32x4._list.get(dart.notNull(mask) >> 4 & 3);
      let _w = NativeInt32x4._list.get(dart.notNull(mask) >> 6 & 3);
      return new NativeInt32x4._truncated(_x, _y, _z, _w);
    }
    withX(x) {
      let _x = dart.as(NativeInt32x4._truncate(x), core.int);
      return new NativeInt32x4._truncated(_x, this.y, this.z, this.w);
    }
    withY(y) {
      let _y = dart.as(NativeInt32x4._truncate(y), core.int);
      return new NativeInt32x4._truncated(this.x, _y, this.z, this.w);
    }
    withZ(z) {
      let _z = dart.as(NativeInt32x4._truncate(z), core.int);
      return new NativeInt32x4._truncated(this.x, this.y, _z, this.w);
    }
    withW(w) {
      let _w = dart.as(NativeInt32x4._truncate(w), core.int);
      return new NativeInt32x4._truncated(this.x, this.y, this.z, _w);
    }
    get flagX() {
      return this.x != 0;
    }
    get flagY() {
      return this.y != 0;
    }
    get flagZ() {
      return this.z != 0;
    }
    get flagW() {
      return this.w != 0;
    }
    withFlagX(flagX) {
      let _x = dart.notNull(flagX) ? -1 : 0;
      return new NativeInt32x4._truncated(_x, this.y, this.z, this.w);
    }
    withFlagY(flagY) {
      let _y = dart.notNull(flagY) ? -1 : 0;
      return new NativeInt32x4._truncated(this.x, _y, this.z, this.w);
    }
    withFlagZ(flagZ) {
      let _z = dart.notNull(flagZ) ? -1 : 0;
      return new NativeInt32x4._truncated(this.x, this.y, _z, this.w);
    }
    withFlagW(flagW) {
      let _w = dart.notNull(flagW) ? -1 : 0;
      return new NativeInt32x4._truncated(this.x, this.y, this.z, _w);
    }
    select(trueValue, falseValue) {
      let floatList = NativeFloat32x4._list;
      let intView = NativeFloat32x4._uint32view;
      floatList.set(0, trueValue.x);
      floatList.set(1, trueValue.y);
      floatList.set(2, trueValue.z);
      floatList.set(3, trueValue.w);
      let stx = intView.get(0);
      let sty = intView.get(1);
      let stz = intView.get(2);
      let stw = intView.get(3);
      floatList.set(0, falseValue.x);
      floatList.set(1, falseValue.y);
      floatList.set(2, falseValue.z);
      floatList.set(3, falseValue.w);
      let sfx = intView.get(0);
      let sfy = intView.get(1);
      let sfz = intView.get(2);
      let sfw = intView.get(3);
      let _x = dart.notNull(this.x) & dart.notNull(stx) | ~dart.notNull(this.x) & dart.notNull(sfx);
      let _y = dart.notNull(this.y) & dart.notNull(sty) | ~dart.notNull(this.y) & dart.notNull(sfy);
      let _z = dart.notNull(this.z) & dart.notNull(stz) | ~dart.notNull(this.z) & dart.notNull(sfz);
      let _w = dart.notNull(this.w) & dart.notNull(stw) | ~dart.notNull(this.w) & dart.notNull(sfw);
      intView.set(0, _x);
      intView.set(1, _y);
      intView.set(2, _z);
      intView.set(3, _w);
      return new NativeFloat32x4._truncated(floatList.get(0), floatList.get(1), floatList.get(2), floatList.get(3));
    }
  }
  NativeInt32x4[dart.implements] = () => [typed_data.Int32x4];
  dart.defineNamedConstructor(NativeInt32x4, 'bool');
  dart.defineNamedConstructor(NativeInt32x4, '_truncated');
  dart.setSignature(NativeInt32x4, {
    constructors: () => ({
      NativeInt32x4: [NativeInt32x4, [core.int, core.int, core.int, core.int]],
      bool: [NativeInt32x4, [core.bool, core.bool, core.bool, core.bool]],
      fromFloat32x4Bits: [NativeInt32x4, [typed_data.Float32x4]],
      _truncated: [NativeInt32x4, [core.int, core.int, core.int, core.int]]
    }),
    methods: () => ({
      '|': [typed_data.Int32x4, [typed_data.Int32x4]],
      '&': [typed_data.Int32x4, [typed_data.Int32x4]],
      '^': [typed_data.Int32x4, [typed_data.Int32x4]],
      '+': [typed_data.Int32x4, [typed_data.Int32x4]],
      '-': [typed_data.Int32x4, [typed_data.Int32x4]],
      'unary-': [typed_data.Int32x4, []],
      shuffle: [typed_data.Int32x4, [core.int]],
      shuffleMix: [typed_data.Int32x4, [typed_data.Int32x4, core.int]],
      withX: [typed_data.Int32x4, [core.int]],
      withY: [typed_data.Int32x4, [core.int]],
      withZ: [typed_data.Int32x4, [core.int]],
      withW: [typed_data.Int32x4, [core.int]],
      withFlagX: [typed_data.Int32x4, [core.bool]],
      withFlagY: [typed_data.Int32x4, [core.bool]],
      withFlagZ: [typed_data.Int32x4, [core.bool]],
      withFlagW: [typed_data.Int32x4, [core.bool]],
      select: [typed_data.Float32x4, [typed_data.Float32x4, typed_data.Float32x4]]
    }),
    statics: () => ({_truncate: [dart.dynamic, [dart.dynamic]]}),
    names: ['_truncate']
  });
  dart.defineLazyProperties(NativeInt32x4, {
    get _list() {
      return NativeInt32List.new(4);
    }
  });
  class NativeFloat64x2 extends core.Object {
    NativeFloat64x2(x, y) {
      this.x = x;
      this.y = y;
      if (!(typeof this.x == 'number'))
        dart.throw(new core.ArgumentError(this.x));
      if (!(typeof this.y == 'number'))
        dart.throw(new core.ArgumentError(this.y));
    }
    splat(v) {
      this.NativeFloat64x2(v, v);
    }
    zero() {
      this.splat(0.0);
    }
    fromFloat32x4(v) {
      this.NativeFloat64x2(v.x, v.y);
    }
    _doubles(x, y) {
      this.x = x;
      this.y = y;
    }
    toString() {
      return `[${this.x}, ${this.y}]`;
    }
    ['+'](other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) + dart.notNull(other.x), dart.notNull(this.y) + dart.notNull(other.y));
    }
    ['unary-']() {
      return new NativeFloat64x2._doubles(-dart.notNull(this.x), -dart.notNull(this.y));
    }
    ['-'](other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) - dart.notNull(other.x), dart.notNull(this.y) - dart.notNull(other.y));
    }
    ['*'](other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) * dart.notNull(other.x), dart.notNull(this.y) * dart.notNull(other.y));
    }
    ['/'](other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) / dart.notNull(other.x), dart.notNull(this.y) / dart.notNull(other.y));
    }
    scale(s) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) * dart.notNull(s), dart.notNull(this.y) * dart.notNull(s));
    }
    abs() {
      return new NativeFloat64x2._doubles(this.x[dartx.abs](), this.y[dartx.abs]());
    }
    clamp(lowerLimit, upperLimit) {
      let _lx = lowerLimit.x;
      let _ly = lowerLimit.y;
      let _ux = upperLimit.x;
      let _uy = upperLimit.y;
      let _x = this.x;
      let _y = this.y;
      _x = dart.notNull(_x) > dart.notNull(_ux) ? _ux : _x;
      _y = dart.notNull(_y) > dart.notNull(_uy) ? _uy : _y;
      _x = dart.notNull(_x) < dart.notNull(_lx) ? _lx : _x;
      _y = dart.notNull(_y) < dart.notNull(_ly) ? _ly : _y;
      return new NativeFloat64x2._doubles(_x, _y);
    }
    get signMask() {
      let view = NativeFloat64x2._uint32View;
      NativeFloat64x2._list.set(0, this.x);
      NativeFloat64x2._list.set(1, this.y);
      let mx = (dart.notNull(view.get(1)) & 2147483648) >> 31;
      let my = (dart.notNull(view.get(3)) & 2147483648) >> 31;
      return dart.notNull(mx) | dart.notNull(my) << 1;
    }
    withX(x) {
      if (!(typeof x == 'number'))
        dart.throw(new core.ArgumentError(x));
      return new NativeFloat64x2._doubles(x, this.y);
    }
    withY(y) {
      if (!(typeof y == 'number'))
        dart.throw(new core.ArgumentError(y));
      return new NativeFloat64x2._doubles(this.x, y);
    }
    min(other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) < dart.notNull(other.x) ? this.x : other.x, dart.notNull(this.y) < dart.notNull(other.y) ? this.y : other.y);
    }
    max(other) {
      return new NativeFloat64x2._doubles(dart.notNull(this.x) > dart.notNull(other.x) ? this.x : other.x, dart.notNull(this.y) > dart.notNull(other.y) ? this.y : other.y);
    }
    sqrt() {
      return new NativeFloat64x2._doubles(math.sqrt(this.x), math.sqrt(this.y));
    }
  }
  NativeFloat64x2[dart.implements] = () => [typed_data.Float64x2];
  dart.defineNamedConstructor(NativeFloat64x2, 'splat');
  dart.defineNamedConstructor(NativeFloat64x2, 'zero');
  dart.defineNamedConstructor(NativeFloat64x2, 'fromFloat32x4');
  dart.defineNamedConstructor(NativeFloat64x2, '_doubles');
  dart.setSignature(NativeFloat64x2, {
    constructors: () => ({
      NativeFloat64x2: [NativeFloat64x2, [core.double, core.double]],
      splat: [NativeFloat64x2, [core.double]],
      zero: [NativeFloat64x2, []],
      fromFloat32x4: [NativeFloat64x2, [typed_data.Float32x4]],
      _doubles: [NativeFloat64x2, [core.double, core.double]]
    }),
    methods: () => ({
      '+': [typed_data.Float64x2, [typed_data.Float64x2]],
      'unary-': [typed_data.Float64x2, []],
      '-': [typed_data.Float64x2, [typed_data.Float64x2]],
      '*': [typed_data.Float64x2, [typed_data.Float64x2]],
      '/': [typed_data.Float64x2, [typed_data.Float64x2]],
      scale: [typed_data.Float64x2, [core.double]],
      abs: [typed_data.Float64x2, []],
      clamp: [typed_data.Float64x2, [typed_data.Float64x2, typed_data.Float64x2]],
      withX: [typed_data.Float64x2, [core.double]],
      withY: [typed_data.Float64x2, [core.double]],
      min: [typed_data.Float64x2, [typed_data.Float64x2]],
      max: [typed_data.Float64x2, [typed_data.Float64x2]],
      sqrt: [typed_data.Float64x2, []]
    })
  });
  dart.defineLazyProperties(NativeFloat64x2, {
    get _list() {
      return NativeFloat64List.new(2);
    },
    set _list(_) {},
    get _uint32View() {
      return dart.as(NativeFloat64x2._list.buffer.asUint32List(), NativeUint32List);
    },
    set _uint32View(_) {}
  });
  // Exports:
  exports.NativeByteBuffer = NativeByteBuffer;
  exports.NativeFloat32x4List = NativeFloat32x4List;
  exports.NativeInt32x4List = NativeInt32x4List;
  exports.NativeFloat64x2List = NativeFloat64x2List;
  exports.NativeTypedData = NativeTypedData;
  exports.NativeByteData = NativeByteData;
  exports.NativeTypedArray = NativeTypedArray;
  exports.NativeTypedArrayOfDouble = NativeTypedArrayOfDouble;
  exports.NativeTypedArrayOfInt = NativeTypedArrayOfInt;
  exports.NativeFloat32List = NativeFloat32List;
  exports.NativeFloat64List = NativeFloat64List;
  exports.NativeInt16List = NativeInt16List;
  exports.NativeInt32List = NativeInt32List;
  exports.NativeInt8List = NativeInt8List;
  exports.NativeUint16List = NativeUint16List;
  exports.NativeUint32List = NativeUint32List;
  exports.NativeUint8ClampedList = NativeUint8ClampedList;
  exports.NativeUint8List = NativeUint8List;
  exports.NativeFloat32x4 = NativeFloat32x4;
  exports.NativeInt32x4 = NativeInt32x4;
  exports.NativeFloat64x2 = NativeFloat64x2;
});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/* This library defines runtime operations on objects used by the code
 * generator.
 */
dart_library.library('dart/_operations', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_utils',
  'dart/async',
  'dart/collection',
  'dart/core',
  'dart/_js_helper',
  'dart/_classes',
  'dart/_errors',
  'dart/_rtti',
  'dart/_types'
], function(exports, dart_utils, async, collection, core, _js_helper, classes, errors, rtti,
            types) {
  'use strict';

  const getOwnNamesAndSymbols = dart_utils.getOwnNamesAndSymbols;
  const throwError = dart_utils.throwError;

  const getOwnPropertyNames = Object.getOwnPropertyNames;
  const hasOwnProperty = Object.prototype.hasOwnProperty;

  function _canonicalFieldName(obj, name, args, displayName) {
    name = classes.canonicalMember(obj, name);
    if (name) return name;
    // TODO(jmesserly): in the future we might have types that "overlay" Dart
    // methods while also exposing the full native API, e.g. dart:html vs
    // dart:dom. To support that we'd need to fall back to the normal name
    // if an extension method wasn't found.
    errors.throwNoSuchMethod(obj, displayName, args);
  }

  function dload(obj, field) {
    field = _canonicalFieldName(obj, field, [], field);
    if (classes.hasMethod(obj, field)) {
      return classes.bind(obj, field);
    }
    // TODO(vsm): Implement NSM robustly.  An 'in' check breaks on certain
    // types.  hasOwnProperty doesn't chase the proto chain.
    // Also, do we want an NSM on regular JS objects?
    // See: https://github.com/dart-lang/dev_compiler/issues/169
    let result = obj[field];
    return result;
  }
  exports.dload = dload;

  function dput(obj, field, value) {
    field = _canonicalFieldName(obj, field, [value], field);
    // TODO(vsm): Implement NSM and type checks.
    // See: https://github.com/dart-lang/dev_compiler/issues/170
    obj[field] = value;
    return value;
  }
  exports.dput = dput;


  /// Check that a function of a given type can be applied to
  /// actuals.
  function checkApply(type, actuals) {
    if (actuals.length < type.args.length) return false;
    let index = 0;
    for(let i = 0; i < type.args.length; ++i) {
      if (!instanceOfOrNull(actuals[i], type.args[i])) return false;
      ++index;
    }
    if (actuals.length == type.args.length) return true;
    let extras = actuals.length - type.args.length;
    if (type.optionals.length > 0) {
      if (extras > type.optionals.length) return false;
      for(let i = 0, j=index; i < extras; ++i, ++j) {
        if (!instanceOfOrNull(actuals[j], type.optionals[i])) return false;
      }
      return true;
    }
    // TODO(leafp): We can't tell when someone might be calling
    // something expecting an optional argument with named arguments

    if (extras != 1) return false;
    // An empty named list means no named arguments
    if (getOwnPropertyNames(type.named).length == 0) return false;
    let opts = actuals[index];
    let names = getOwnPropertyNames(opts);
    // Type is something other than a map
    if (names.length == 0) return false;
    for (var name of names) {
      if (!(hasOwnProperty.call(type.named, name))) {
        return false;
      }
      if (!instanceOfOrNull(opts[name], type.named[name])) return false;
    }
    return true;
  }

  function throwNoSuchMethod(obj, name, args, opt_func) {
    if (obj === void 0) obj = opt_func;
    errors.throwNoSuchMethod(obj, name, args);
  }

  function checkAndCall(f, ftype, obj, args, name) {
    if (!(f instanceof Function)) {
      // We're not a function (and hence not a method either)
      // Grab the `call` method if it's not a function.
      if (f != null) {
        ftype = classes.getMethodType(f, 'call');
        f = f.call;
      }
      if (!(f instanceof Function)) {
        throwNoSuchMethod(obj, name, args);
      }
    }
    // If f is a function, but not a method (no method type)
    // then it should have been a function valued field, so
    // get the type from the function.
    if (ftype === void 0) {
      ftype = rtti.read(f);
    }

    if (!ftype) {
      // TODO(leafp): Allow JS objects to go through?
      // This includes the DOM.
      return f.apply(obj, args);
    }

    if (checkApply(ftype, args)) {
      return f.apply(obj, args);
    }

    // TODO(leafp): throw a type error (rather than NSM)
    // if the arity matches but the types are wrong.
    throwNoSuchMethod(obj, name, args, f);
  }

  function dcall(f, ...args) {
    let ftype = rtti.read(f);
    return checkAndCall(f, ftype, void 0, args, 'call');
  }
  exports.dcall = dcall;

  /** Shared code for dsend, dindex, and dsetindex. */
  function callMethod(obj, name, args, displayName) {
    let symbol = _canonicalFieldName(obj, name, args, displayName);
    let f = obj != null ? obj[symbol] : null;
    let ftype = classes.getMethodType(obj, name);
    return checkAndCall(f, ftype, obj, args, displayName);
  }

  function dsend(obj, method, ...args) {
    return callMethod(obj, method, args, method);
  }
  exports.dsend = dsend;

  function dindex(obj, index) {
    return callMethod(obj, 'get', [index], '[]');
  }
  exports.dindex = dindex;

  function dsetindex(obj, index, value) {
    callMethod(obj, 'set', [index, value], '[]=');
    return value;
  }
  exports.dsetindex = dsetindex;

  function _ignoreTypeFailure(actual, type) {
    // TODO(vsm): Remove this hack ...
    // This is primarily due to the lack of generic methods,
    // but we need to triage all the errors.
    let isSubtype = types.isSubtype;
    if (isSubtype(type, core.Iterable) && isSubtype(actual, core.Iterable) ||
        isSubtype(type, async.Future) && isSubtype(actual, async.Future) ||
        isSubtype(type, core.Map) && isSubtype(actual, core.Map) ||
        isSubtype(type, core.Function) && isSubtype(actual, core.Function) ||
        isSubtype(type, async.Stream) && isSubtype(actual, async.Stream) ||
        isSubtype(type, async.StreamSubscription) &&
            isSubtype(actual, async.StreamSubscription)) {
      console.warn('Ignoring cast fail from ' + types.typeName(actual) +
        ' to ' + types.typeName(type));
      return true;
    }
    return false;
  }

  function strongInstanceOf(obj, type, ignoreFromWhiteList) {
    let actual = rtti.realRuntimeType(obj);
    if (types.isSubtype(actual, type) || actual == types.jsobject) return true;
    if (ignoreFromWhiteList == void 0) return false;
    if (types.isGroundType(type)) return false;
    if (_ignoreTypeFailure(actual, type)) return true;
    return false;
  }
  exports.strongInstanceOf = strongInstanceOf;

  function instanceOfOrNull(obj, type) {
    if ((obj == null) || strongInstanceOf(obj, type, true)) return true;
    return false;
  }

  function instanceOf(obj, type) {
    if (strongInstanceOf(obj, type)) return true;
    // TODO(#296): This is perhaps too eager to throw a StrongModeError?
    // It will throw on <int>[] is List<String>.
    // TODO(vsm): We can statically detect many cases where this
    // check is unnecessary.
    if (types.isGroundType(type)) return false;
    let actual = rtti.realRuntimeType(obj);
    dart_utils.throwStrongModeError('Strong mode is check failure: ' +
      types.typeName(actual) + ' does not soundly subtype ' +
      types.typeName(type));
  }
  exports.instanceOf = instanceOf;

  function cast(obj, type) {
    // TODO(#296): This is perhaps too eager to throw a StrongModeError?
    // TODO(vsm): handle non-nullable types
    if (instanceOfOrNull(obj, type)) return obj;
    let actual = rtti.realRuntimeType(obj);
    if (types.isGroundType(type)) errors.throwCastError(actual, type);

    if (_ignoreTypeFailure(actual, type)) return obj;

    dart_utils.throwStrongModeError('Strong mode cast failure from ' +
      types.typeName(actual) + ' to ' + types.typeName(type));
  }
  exports.cast = cast;

  function asInt(obj) {
    if (obj == null) {
      return null;
    }
    if (Math.floor(obj) != obj) {
      // Note: null will also be caught by this check
      errors.throwCastError(rtti.realRuntimeType(obj), core.int);
    }
    return obj;
  }
  exports.asInt = asInt;

  function arity(f) {
    // TODO(jmesserly): need to parse optional params.
    // In ES6, length is the number of required arguments.
    return { min: f.length, max: f.length };
  }
  exports.arity = arity;

  function equals(x, y) {
    if (x == null || y == null) return x == y;
    let eq = x['=='];
    return eq ? eq.call(x, y) : x === y;
  }
  exports.equals = equals;

  /** Checks that `x` is not null or undefined. */
  function notNull(x) {
    if (x == null) errors.throwNullValueError();
    return x;
  }
  exports.notNull = notNull;

  /**
   * Creates a dart:collection LinkedHashMap.
   *
   * For a map with string keys an object literal can be used, for example
   * `map({'hi': 1, 'there': 2})`.
   *
   * Otherwise an array should be used, for example `map([1, 2, 3, 4])` will
   * create a map with keys [1, 3] and values [2, 4]. Each key-value pair
   * should be adjacent entries in the array.
   *
   * For a map with no keys the function can be called with no arguments, for
   * example `map()`.
   */
  // TODO(jmesserly): this could be faster
  function map(values) {
    let map = collection.LinkedHashMap.new();
    if (Array.isArray(values)) {
      for (let i = 0, end = values.length - 1; i < end; i += 2) {
        let key = values[i];
        let value = values[i + 1];
        map.set(key, value);
      }
    } else if (typeof values === 'object') {
      for (let key of getOwnPropertyNames(values)) {
        map.set(key, values[key]);
      }
    }
    return map;
  }
  exports.map = map;

  function assert(condition) {
    if (!condition) errors.throwAssertionError();
  }
  exports.assert = assert;

  let _stack = new WeakMap();
  function throw_(obj) {
    if (obj != null && (typeof obj == 'object' || typeof obj == 'function')) {
      // TODO(jmesserly): couldn't we store the most recent stack in a single
      // variable? There should only be one active stack trace. That would
      // allow it to work for things like strings and numbers.
      _stack.set(obj, new Error());
    }
    throw obj;
  }
  exports.throw = throw_;

  function getError(exception) {
    var stack = _stack.get(exception);
    return stack !== void 0 ? stack : exception;
  }

  // This is a utility function: it is only intended to be called from dev
  // tools.
  function stackPrint(exception) {
    var error = getError(exception);
    console.log(error.stack ? error.stack : 'No stack trace for: ' + error);
  }
  exports.stackPrint = stackPrint;

  function stackTrace(exception) {
    var error = getError(exception);
    return _js_helper.getTraceFromException(error);
  }
  exports.stackTrace = stackTrace;

  /**
   * Implements a sequence of .? operations.
   *
   * Will call each successive callback, unless one returns null, which stops
   * the sequence.
   */
  function nullSafe(obj, ...callbacks) {
    if (obj == null) return obj;
    for (const callback of callbacks) {
      obj = callback(obj);
      if (obj == null) break;
    }
    return obj;
  }
  exports.nullSafe = nullSafe;

  let _value = Symbol('_value');
  /**
   * Looks up a sequence of [keys] in [map], recursively, and
   * returns the result. If the value is not found, [valueFn] will be called to
   * add it. For example:
   *
   *     let map = new Map();
   *     putIfAbsent(map, [1, 2, 'hi ', 'there '], () => 'world');
   *
   * ... will create a Map with a structure like:
   *
   *     { 1: { 2: { 'hi ': { 'there ': 'world' } } } }
   */
  function multiKeyPutIfAbsent(map, keys, valueFn) {
    for (let k of keys) {
      let value = map.get(k);
      if (!value) {
        // TODO(jmesserly): most of these maps are very small (e.g. 1 item),
        // so it may be worth optimizing for that.
        map.set(k, value = new Map());
      }
      map = value;
    }
    if (map.has(_value)) return map.get(_value);
    let value = valueFn();
    map.set(_value, value);
    return value;
  }

  /** The global constant table. */
  const constants = new Map();

  /**
   * Canonicalize a constant object.
   *
   * Preconditions:
   * - `obj` is an objects or array, not a primitive.
   * - nested values of the object are themselves already canonicalized.
   */
  function constant(obj) {
    let objectKey = [rtti.realRuntimeType(obj)];
    // TODO(jmesserly): there's no guarantee in JS that names/symbols are
    // returned in the same order.
    //
    // We could probably get the same order if we're judicious about
    // initializing fields in a consistent order across all const constructors.
    // Alternatively we need a way to sort them to make consistent.
    //
    // Right now we use the (name,value) pairs in sequence, which prevents
    // an object with incorrect field values being returned, but won't
    // canonicalize correctly if key order is different.
    for (let name of getOwnNamesAndSymbols(obj)) {
      objectKey.push(name);
      objectKey.push(obj[name]);
    }
    return multiKeyPutIfAbsent(constants, objectKey, () => obj);
  }
  exports.const = constant;


  // The following are helpers for Object methods when the receiver
  // may be null or primitive.  These should only be generated by
  // the compiler.
  function hashCode(obj) {
    if (obj == null) {
      return 0;
    }
    // TODO(vsm): What should we do for primitives and non-Dart objects?
    switch (typeof obj) {
    case "number":
    case "boolean":
      return obj & 0x1FFFFFFF;
    case "string":
        // TODO(vsm): Call the JSString hashCode?
      return obj.length;
    }
    return obj.hashCode;
  }
  exports.hashCode = hashCode;

  function toString(obj) {
    if (obj == null) {
      return "null";
    }
    return obj.toString();
  }
  exports.toString = toString;

  function noSuchMethod(obj, invocation) {
    if (obj == null) {
      errors.throwNoSuchMethod(obj, invocation.memberName,
        invocation.positionalArguments, invocation.namedArguments);
    }
    switch (typeof obj) {
      case "number":
      case "boolean":
      case "string":
        errors.throwNoSuchMethod(obj, invocation.memberName,
          invocation.positionalArguments, invocation.namedArguments);
    }
    return obj.noSuchMethod(invocation);
  }
  exports.noSuchMethod = noSuchMethod;

  class JsIterator {
    constructor(dartIterator) {
      this.dartIterator = dartIterator;
    }
    next() {
      let i = this.dartIterator;
      let done = !i.moveNext();
      return { done: done, value: done ? void 0 : i.current };
    }
  }
  exports.JsIterator = JsIterator;


});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/* This library defines the association between runtime objects and
 * runtime types.
*/

dart_library.library('dart/_rtti', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_utils',
  'dart/core',
  'dart/_types'
], function(exports, dart_utils, core, types) {
  'use strict';

  const defineLazyProperty = dart_utils.defineLazyProperty;

  const defineProperty = Object.defineProperty;

  /**
   * Runtime type information.  This module defines the mapping from
   * runtime objects to their runtime type information.  See the types
   * module for the definition of how type information is represented.
   *
   * Runtime objects fall into four main categories:
   *
   *   - Things represented by javascript primitives, such as
   *     null, numbers, booleans, strings, and symbols.  For these
   *     we map directly from the javascript type (given by typeof)
   *     to the appropriate class type from core, which serves as their
   *     rtti.
   *
   *   - Functions, which are represented by javascript functions.
   *     Representations of Dart functions always have a
   *     _runtimeType property attached to them with the appropriate
   *     rtti.
   *
   *   - Objects (instances) which are represented by instances of
   *     javascript (ES6) classes.  Their types are given by their
   *     classes, and the rtti is accessed by projecting out their
   *     constructor field.
   *
   *   - Types objects, which are represented as described in the types
   *     module.  Types always have a _runtimeType property attached to
   *     them with the appropriate rtti.  The rtti for these is always
   *     core.Type.  TODO(leafp): consider the possibility that we can
   *     reliably recognize type objects and map directly to core.Type
   *     rather than attaching this property everywhere.
   *
   */

  /**
   *Tag a closure with a type, using one of three forms:
   * dart.fn(cls) marks cls has having no optional or named
   *  parameters, with all argument and return types as dynamic
   * dart.fn(cls, func) marks cls with the lazily computed
   *  runtime type as computed by func()
   * dart.fn(cls, rType, argsT, extras) marks cls as having the
   * runtime type dart.functionType(rType, argsT, extras)
   *
   * Note that since we are producing a type for a concrete function,
   * it is sound to use the definite arrow type.
   */
  function fn(closure, ...args) {
    // Closure and a lazy type constructor
    if (args.length == 1) {
      defineLazyProperty(closure, _runtimeType, {get : args[0]});
      return closure;
    }
    let t;
    if (args.length == 0) {
      // No type arguments, it's all dynamic
      t = types.definiteFunctionType(
        types.dynamic, Array(closure.length).fill(types.dynamic));
    } else {
      // We're passed the piecewise components of the function type,
      // construct it.
      t = types.definiteFunctionType.apply(null, args);
    }
    tag(closure, t);
    return closure;
  }
  exports.fn = fn;

  // TODO(vsm): How should we encode the runtime type?
  const _runtimeType = Symbol('_runtimeType');

  function checkPrimitiveType(obj) {
    switch (typeof obj) {
      case "undefined":
        return core.Null;
      case "number":
        return Math.floor(obj) == obj ? core.int : core.double;
      case "boolean":
        return core.bool;
      case "string":
        return core.String;
      case "symbol":
        return Symbol;
    }
    // Undefined is handled above. For historical reasons,
    // typeof null == "object" in JS.
    if (obj === null) return core.Null;
    return null;
  }

  function runtimeType(obj) {
    let result = checkPrimitiveType(obj);
    if (result !== null) return result;
    return obj.runtimeType;
  }
  exports.runtimeType = runtimeType;

  function getFunctionType(obj) {
    // TODO(vsm): Encode this properly on the function for Dart-generated code.
    let args = Array(obj.length).fill(types.dynamic);
    return types.definiteFunctionType(types.bottom, args);
  }

  /**
   * Returns the runtime type of obj. This is the same as `obj.realRuntimeType`
   * but will not call an overridden getter.
   *
   * Currently this will return null for non-Dart objects.
   */
  function realRuntimeType(obj) {
    let result = checkPrimitiveType(obj);
    if (result !== null) return result;
    // TODO(vsm): Should we treat Dart and JS objects differently here?
    // E.g., we can check if obj instanceof core.Object to differentiate.
    result = obj[_runtimeType];
    if (result) return result;
    result = obj.constructor;
    if (result == Function) {
      // An undecorated Function should have come from
      // JavaScript.  Treat as untyped.
      return types.jsobject;
    }
    return result;
  }
  exports.realRuntimeType = realRuntimeType;

  function LazyTagged(infoFn) {
    class _Tagged {
      get [_runtimeType]() {return infoFn();}
    }
    return _Tagged;
  }
  exports.LazyTagged = LazyTagged;

  function read(value) {
    return value[_runtimeType];
  }
  exports.read = read;

  function tag(value, info) {
    value[_runtimeType] = info;
  }
  exports.tag = tag;

  function tagComputed(value, compute) {
    defineProperty(value, _runtimeType, { get: compute });
  }
  exports.tagComputed = tagComputed;

  function tagMemoized(value, compute) {
    let cache = null;
    function getter() {
      if (compute == null) return cache;
      cache = compute();
      compute = null;
      return cache;
    }
    tagComputed(value, getter);
  }
  exports.tagMemoized = tagMemoized;
});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

dart_library.library('dart/_runtime', null, /* Imports */[
  'dart/_utils',
  'dart/_classes',
  'dart/_errors',
  'dart/_generators',
  'dart/_operations',
  'dart/_rtti',
  'dart/_types',
], /* Lazy Imports */[
  'dart/_js_helper'
], function(exports, dart_utils, classes, errors, generators, operations, rtti, types,
    _js_helper) {
  'use strict';

  function _export(value) {
    if (value) return value;
    console.log("Re-exporting null field: " + name);
    throw "Bad export";
  }

  function exportFrom(value, names) {
    for (let name of names) {
      exports[name] = _export(value[name]);
    }
  }

  exports.global = typeof window == "undefined" ? global : window;
  exports.JsSymbol = _export(Symbol);

  // TODO(vsm): This is referenced (as init.globalState) from
  // isolate_helper.dart.  Where should it go?
  // See: https://github.com/dart-lang/dev_compiler/issues/164
  exports.globalState = null;
  _js_helper.checkNum = operations.notNull;

  // Re-exports

  // From classes
  exportFrom(classes, [
    'bind',
    'classGetConstructorType',
    'dartx',
    'defineNamedConstructor',
    'defineExtensionNames',
    'defineExtensionMembers',
    'generic',
    'implements',
    'list',
    'metadata',
    'mixin',
    'registerExtension',
    'setBaseClass',
    'setSignature',
    'virtualField'
  ]);

  exportFrom(generators, [
    'syncStar',
    'async',
    'asyncStar'
  ]);

  // From dart_utils
  exportFrom(dart_utils, ['copyProperties', 'export_']);
  // Renames
  exports.defineLazyClass = _export(dart_utils.defineLazy);
  exports.defineLazyProperties = _export(dart_utils.defineLazy);
  exports.defineLazyClassGeneric = _export(dart_utils.defineLazyProperty);

  // From operations
  exportFrom(operations, [
    'JsIterator',
    'arity',
    'asInt',
    'assert',
    'const',
    'dcall',
    'dindex',
    'dload',
    'dput',
    'dsend',
    'dsetindex',
    'equals',
    'hashCode',
    'map',
    'noSuchMethod',
    'notNull',
    'nullSafe',
    'stackPrint',
    'stackTrace',
    'strongInstanceOf',
    'throw',
    'toString',
  ])
  // Renames
  exports.as = _export(operations.cast);
  exports.is = _export(operations.instanceOf);

  // From types
  exportFrom(types, [
    'bottom',
    'definiteFunctionType',
    'dynamic',
    'functionType',
    'jsobject',
    'typedef',
    'typeName',
    'void',
  ]);

  // From rtti
  exportFrom(rtti, [
    'fn',
    'realRuntimeType',
    'runtimeType',
  ]);

});
// Copyright (c) 2015, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/* This library defines the representation of runtime types.
*/

dart_library.library('dart/_types', null, /* Imports */[
], /* Lazy Imports */[
  'dart/_utils',
  'dart/core',
  'dart/_classes',
  'dart/_rtti'
], function(exports, dart_utils, core, classes, rtti) {
  'use strict';

  const getOwnPropertyNames = Object.getOwnPropertyNames;

  const assert = dart_utils.assert_;

  /**
   * Types in dart are represented at runtime as follows.
   *   - Normal nominal types, produced from classes, are represented
   *     at runtime by the JS class of which they are an instance.
   *     If the type is the result of instantiating a generic class,
   *     then the "classes" module manages the association between the
   *     instantiated class and the original class declaration
   *     and the type arguments with which it was instantiated.  This
   *     assocation can be queried via the "classes" module".
   *
   *   - All other types are represented as instances of class TypeRep,
   *     defined in this module.
   *     - Dynamic, Void, and Bottom are singleton instances of sentinal
   *       classes.
   *     - Function types are instances of subclasses of AbstractFunctionType.
   *
   * Function types are represented in one of two ways:
   *   - As an instance of FunctionType.  These are eagerly computed.
   *   - As an instance of TypeDef.  The TypeDef representation lazily
   *     computes an instance of FunctionType, and delegates to that instance.
   *
   * All types satisfy the following interface:
   *  get String name;
   *  String toString();
   *
   */
  class TypeRep extends rtti.LazyTagged(() => core.Type) {
    get name() {return this.toString();}
  }

  class Dynamic extends TypeRep {
    toString() { return "dynamic"; }
  }
  let dynamicR = new Dynamic();
  exports.dynamic = dynamicR;

  class Void extends TypeRep {
    toString() { return "void"; }
  }

  let voidR = new Void();
  exports.void = voidR;

  class Bottom extends TypeRep {
    toString() { return "bottom"; }
  }
  let bottomR = new Bottom();
  exports.bottom = bottomR;

  class JSObject extends TypeRep {
    toString() { return "NativeJavaScriptObject"; }
  }
  let jsobjectR = new JSObject();
  exports.jsobject = jsobjectR;

  class AbstractFunctionType extends TypeRep {
    constructor() {
      super();
      this._stringValue = null;
    }

    toString() { return this.name; }

    get name() {
      if (this._stringValue) return this._stringValue;

      let buffer = '(';
      for (let i = 0; i < this.args.length; ++i) {
        if (i > 0) {
          buffer += ', ';
        }
        buffer += typeName(this.args[i]);
      }
      if (this.optionals.length > 0) {
        if (this.args.length > 0) buffer += ', ';
        buffer += '[';
        for (let i = 0; i < this.optionals.length; ++i) {
          if (i > 0) {
            buffer += ', ';
          }
          buffer += typeName(this.optionals[i]);
        }
        buffer += ']';
      } else if (Object.keys(this.named).length > 0) {
        if (this.args.length > 0) buffer += ', ';
        buffer += '{';
        let names = getOwnPropertyNames(this.named).sort();
        for (let i = 0; i < names.length; ++i) {
          if (i > 0) {
            buffer += ', ';
          }
          buffer += names[i] + ': ' + typeName(this.named[names[i]]);
        }
        buffer += '}';
      }

      buffer += ') -> ' + typeName(this.returnType);
      this._stringValue = buffer;
      return buffer;
    }
  }

  class FunctionType extends AbstractFunctionType {
    /**
     * Construct a function type. There are two arrow constructors,
     * distinguished by the "definite" flag.
     *
     * The fuzzy arrow (definite is false) treats any arguments
     * of type dynamic as having type bottom, and will always be
     * called with a dynamic invoke.
     *
     * The definite arrow (definite is true) leaves arguments unchanged.
     *
     * We eagerly canonize the argument types to avoid having to deal with
     * this logic in multiple places.
     *
     * TODO(leafp): Figure out how to present this to the user.  How
     * should these be printed out?
     */
    constructor(definite, returnType, args, optionals, named) {
      super();
      this.definite = definite;
      this.returnType = returnType;
      this.args = args;
      this.optionals = optionals;
      this.named = named;

      // TODO(vsm): This is just parameter metadata for now.
      this.metadata = [];
      function process(array, metadata) {
        var result = [];
        for (var i = 0; i < array.length; ++i) {
          var arg = array[i];
          if (arg instanceof Array) {
            metadata.push(arg.slice(1));
            result.push(arg[0]);
          } else {
            metadata.push([]);
            result.push(arg);
          }
        }
        return result;
      }
      this.args = process(this.args, this.metadata);
      this.optionals = process(this.optionals, this.metadata);
      // TODO(vsm): Add named arguments.
      this._canonize();
    }
    _canonize() {
      if (this.definite) return;

      function replace(a) {
        return (a == dynamicR) ? bottomR : a;
      }

      this.args = this.args.map(replace);

      if (this.optionals.length > 0) {
        this.optionals = this.optionals.map(replace);
      }

      if (Object.keys(this.named).length > 0) {
        let r = {};
        for (let name of getOwnPropertyNames(this.named)) {
          r[name] = replace(this.named[name]);
        }
        this.named = r;
      }
    }
  }

  class Typedef extends AbstractFunctionType {
    constructor(name, closure) {
      super();
      this._name = name;
      this._closure = closure;
      this._functionType = null;
    }

    get definite() {
      return this._functionType.definite;
    }

    get name() {
      return this._name;
    }

    get functionType() {
      if (!this._functionType) {
        this._functionType = this._closure();
      }
      return this._functionType;
    }

    get returnType() {
      return this.functionType.returnType;
    }

    get args() {
      return this.functionType.args;
    }

    get optionals() {
      return this.functionType.optionals;
    }

    get named() {
      return this.functionType.named;
    }

    get metadata() {
      return this.functionType.metadata;
    }
  }

  function _functionType(definite, returnType, args, extra) {
    // TODO(vsm): Cache / memomize?
    let optionals;
    let named;
    if (extra === void 0) {
      optionals = [];
      named = {};
    } else if (extra instanceof Array) {
      optionals = extra;
      named = {};
    } else {
      optionals = [];
      named = extra;
    }
    return new FunctionType(definite, returnType, args, optionals, named);
  }

  /**
   * Create a "fuzzy" function type.  If any arguments are dynamic
   * they will be replaced with bottom.
   */
  function functionType(returnType, args, extra) {
    return _functionType(false, returnType, args, extra);
  }
  exports.functionType = functionType;

  /**
   * Create a definite function type. No substitution of dynamic for
   * bottom occurs.
   */
  function definiteFunctionType(returnType, args, extra) {
    return _functionType(true, returnType, args, extra);
  }
  exports.definiteFunctionType = definiteFunctionType;

  function typedef(name, closure) {
    return new Typedef(name, closure);
  }
  exports.typedef = typedef;

  function isDartType(type) {
    return rtti.read(type) === core.Type;
  }
  exports.isDartType = isDartType;

  function typeName(type) {
    // Non-instance types
    if (type instanceof TypeRep) return type.toString();
    // Instance types
    let tag = rtti.read(type);
    if (tag === core.Type) {
      let name = type.name;
      let args = classes.getGenericArgs(type);
      if (args) {
        name += '<';
        for (let i = 0; i < args.length; ++i) {
          if (i > 0) name += ', ';
          name += typeName(args[i]);
        }
        name += '>';
      }
      return name;
    }
    if (tag) return "Not a type: " + tag.name;
    return "JSObject<" + type.name + ">";
  }
  exports.typeName = typeName;

  function isFunctionType(type) {
    return type instanceof AbstractFunctionType || type == core.Function;
  }

  function isFunctionSubType(ft1, ft2) {
    if (ft2 == core.Function) {
      return true;
    }

    let ret1 = ft1.returnType;
    let ret2 = ft2.returnType;

    if (!isSubtype_(ret1, ret2)) {
      // Covariant return types
      // Note, void (which can only appear as a return type) is effectively
      // treated as dynamic.  If the base return type is void, we allow any
      // subtype return type.
      // E.g., we allow:
      //   () -> int <: () -> void
      if (ret2 != voidR) {
        return false;
      }
    }

    let args1 = ft1.args;
    let args2 = ft2.args;

    if (args1.length > args2.length) {
      return false;
    }

    for (let i = 0; i < args1.length; ++i) {
      if (!isSubtype_(args2[i], args1[i])) {
        return false;
      }
    }

    let optionals1 = ft1.optionals;
    let optionals2 = ft2.optionals;

    if (args1.length + optionals1.length < args2.length + optionals2.length) {
      return false;
    }

    let j = 0;
    for (let i = args1.length; i < args2.length; ++i, ++j) {
      if (!isSubtype_(args2[i], optionals1[j])) {
        return false;
      }
    }

    for (let i = 0; i < optionals2.length; ++i, ++j) {
      if (!isSubtype_(optionals2[i], optionals1[j])) {
        return false;
      }
    }

    let named1 = ft1.named;
    let named2 = ft2.named;

    let names = getOwnPropertyNames(named2);
    for (let i = 0; i < names.length; ++i) {
      let name = names[i];
      let n1 = named1[name];
      let n2 = named2[name];
      if (n1 === void 0) {
        return false;
      }
      if (!isSubtype_(n2, n1)) {
        return false;
      }
    }

    return true;
  }

  /**
   * Computes the canonical type.
   * This maps JS types onto their corresponding Dart Type.
   */
  // TODO(jmesserly): lots more needs to be done here.
  function canonicalType(t) {
    if (t === Object) return core.Object;
    if (t === Function) return core.Function;
    if (t === Array) return core.List;

    // We shouldn't normally get here with these types, unless something strange
    // happens like subclassing Number in JS and passing it to Dart.
    if (t === String) return core.String;
    if (t === Number) return core.double;
    if (t === Boolean) return core.bool;
    return t;
  }

  const subtypeMap = new Map();
  function isSubtype(t1, t2) {
    // See if we already know the answer
    // TODO(jmesserly): general purpose memoize function?
    let map = subtypeMap.get(t1);
    let result;
    if (map) {
      result = map.get(t2);
      if (result !== void 0) return result;
    } else {
      subtypeMap.set(t1, map = new Map());
    }
    result = isSubtype_(t1, t2)
    map.set(t2, result);
    return result;
  }
  exports.isSubtype = isSubtype;

  function _isBottom(type) {
    return type == bottomR;
  }

  function _isTop(type) {
    return type == core.Object || (type == dynamicR);
  }

  function isSubtype_(t1, t2) {
    t1 = canonicalType(t1);
    t2 = canonicalType(t2);
    if (t1 == t2) return true;

    // Trivially true.
    if (_isTop(t2) || _isBottom(t1)) {
      return true;
    }

    // Trivially false.
    if (_isTop(t1) || _isBottom(t2)) {
      return false;
    }

    // "Traditional" name-based subtype check.
    if (isClassSubType(t1, t2)) {
      return true;
    }

    // Function subtyping.
    // TODO(vsm): Handle Objects with call methods.  Those are functions
    // even if they do not *nominally* subtype core.Function.
    if (isFunctionType(t1) &&
        isFunctionType(t2)) {
      return isFunctionSubType(t1, t2);
    }
    return false;
  }

  function isClassSubType(t1, t2) {
    // We support Dart's covariant generics with the caveat that we do not
    // substitute bottom for dynamic in subtyping rules.
    // I.e., given T1, ..., Tn where at least one Ti != dynamic we disallow:
    // - S !<: S<T1, ..., Tn>
    // - S<dynamic, ..., dynamic> !<: S<T1, ..., Tn>
    t1 = canonicalType(t1);
    assert(t2 == canonicalType(t2));
    if (t1 == t2) return true;

    if (t1 == core.Object) return false;

    // If t1 is a JS Object, we may not hit core.Object.
    if (t1 == null) return t2 == core.Object || t2 == dynamicR;

    // Check if t1 and t2 have the same raw type.  If so, check covariance on
    // type parameters.
    let raw1 = classes.getGenericClass(t1);
    let raw2 = classes.getGenericClass(t2);
    if (raw1 != null && raw1 == raw2) {
      let typeArguments1 = classes.getGenericArgs(t1);
      let typeArguments2 = classes.getGenericArgs(t2);
      let length = typeArguments1.length;
      if (typeArguments2.length == 0) {
        // t2 is the raw form of t1
        return true;
      } else if (length == 0) {
        // t1 is raw, but t2 is not
        return false;
      }
      assert(length == typeArguments2.length);
      for (let i = 0; i < length; ++i) {
        if (!isSubtype(typeArguments1[i], typeArguments2[i])) {
          return false;
        }
      }
      return true;
    }

    // Check superclass.
    if (isClassSubType(t1.__proto__, t2)) return true;

    // Check mixins.
    let mixins = classes.getMixins(t1);
    if (mixins) {
      for (let m1 of mixins) {
        // TODO(jmesserly): remove the != null check once we can load core libs.
        if (m1 != null && isClassSubType(m1, t2)) return true;
      }
    }

    // Check interfaces.
    let getInterfaces = classes.getImplements(t1);
    if (getInterfaces) {
      for (let i1 of getInterfaces()) {
        // TODO(jmesserly): remove the != null check once we can load core libs.
        if (i1 != null && isClassSubType(i1, t2)) return true;
      }
    }

    return false;
  }

  // TODO(jmesserly): this isn't currently used, but it could be if we want
  // `obj is NonGroundType<T,S>` to be rejected at runtime instead of compile
  // time.
  function isGroundType(type) {
    // TODO(vsm): Cache this if we start using it at runtime.

    if (type instanceof AbstractFunctionType) {
      if (!_isTop(type.returnType)) return false;
      for (let i = 0; i < type.args.length; ++i) {
        if (!_isBottom(type.args[i])) return false;
      }
      for (let i = 0; i < type.optionals.length; ++i) {
        if (!_isBottom(type.optionals[i])) return false;
      }
      let names = getOwnPropertyNames(type.named);
      for (let i = 0; i < names.length; ++i) {
        if (!_isBottom(type.named[names[i]])) return false;
      }
      return true;
    }

    let typeArgs = classes.getGenericArgs(type);
    if (!typeArgs) return true;
    for (let t of typeArgs) {
      if (t != core.Object && t != dynamicR) return false;
    }
    return true;
  }
  exports.isGroundType = isGroundType;

});
dart_library.library('dart/_utils', null, /* Imports */[
], /* Lazy imports */[
], function(exports, dart) {
  'use strict';
  const defineProperty = Object.defineProperty;
  const getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor;
  const getOwnPropertyNames = Object.getOwnPropertyNames;
  const getOwnPropertySymbols = Object.getOwnPropertySymbols;
  const hasOwnProperty = Object.prototype.hasOwnProperty;
  const StrongModeError = (function() {
    function StrongModeError(message) {
      Error.call(this);
      this.message = message;
    }
    ;
    Object.setPrototypeOf(StrongModeError.prototype, Error.prototype);
    return StrongModeError;
  })();
  function throwStrongModeError(message) {
    throw new StrongModeError(message);
  }
  function throwInternalError(message) {
    throw Error(message);
  }
  function assert_(condition) {
    if (!condition)
      throwInternalError("The compiler is broken: failed assert");
  }
  function getOwnNamesAndSymbols(obj) {
    return getOwnPropertyNames(obj).concat(getOwnPropertySymbols(obj));
  }
  function safeGetOwnProperty(obj, name) {
    let desc = getOwnPropertyDescriptor(obj, name);
    if (desc)
      return desc.value;
  }
  function defineLazyProperty(to, name, desc) {
    let init = desc.get;
    let value = null;
    function lazySetter(x) {
      init = null;
      value = x;
    }
    function circularInitError() {
      throwInternalError('circular initialization for field ' + name);
    }
    function lazyGetter() {
      if (init == null)
        return value;
      let f = init;
      init = circularInitError;
      lazySetter(f());
      return value;
    }
    desc.get = lazyGetter;
    desc.configurable = true;
    if (desc.set)
      desc.set = lazySetter;
    return defineProperty(to, name, desc);
  }
  function defineLazy(to, from) {
    for (let name of getOwnNamesAndSymbols(from)) {
      defineLazyProperty(to, name, getOwnPropertyDescriptor(from, name));
    }
  }
  function defineMemoizedGetter(obj, name, getter) {
    return defineLazyProperty(obj, name, {get: getter});
  }
  function copyTheseProperties(to, from, names) {
    for (let name of names) {
      var desc = getOwnPropertyDescriptor(from, name);
      if (desc != void 0) {
        defineProperty(to, name, desc);
      } else {
        defineLazyProperty(to, name, () => from[name]);
      }
    }
    return to;
  }
  function copyProperties(to, from) {
    return copyTheseProperties(to, from, getOwnNamesAndSymbols(from));
  }
  function export_(to, from, show, hide) {
    if (show == void 0 || show.length == 0) {
      show = getOwnNamesAndSymbols(from);
    }
    if (hide != void 0) {
      var hideMap = new Set(hide);
      show = show.filter(k => !hideMap.has(k));
    }
    return copyTheseProperties(to, from, show);
  }
  // Exports:
  exports.defineProperty = defineProperty;
  exports.getOwnPropertyDescriptor = getOwnPropertyDescriptor;
  exports.getOwnPropertyNames = getOwnPropertyNames;
  exports.getOwnPropertySymbols = getOwnPropertySymbols;
  exports.hasOwnProperty = hasOwnProperty;
  exports.StrongModeError = StrongModeError;
  exports.throwStrongModeError = throwStrongModeError;
  exports.throwInternalError = throwInternalError;
  exports.assert_ = assert_;
  exports.getOwnNamesAndSymbols = getOwnNamesAndSymbols;
  exports.safeGetOwnProperty = safeGetOwnProperty;
  exports.defineLazyProperty = defineLazyProperty;
  exports.defineLazy = defineLazy;
  exports.defineMemoizedGetter = defineMemoizedGetter;
  exports.copyTheseProperties = copyTheseProperties;
  exports.copyProperties = copyProperties;
  exports.export_ = export_;
});
dart_library.library('dart/async', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/_internal',
  'dart/collection'
], /* Lazy imports */[
  'dart/_isolate_helper'
], function(exports, dart, core, _internal, collection, _isolate_helper) {
  'use strict';
  let dartx = dart.dartx;
  function _invokeErrorHandler(errorHandler, error, stackTrace) {
    if (dart.is(errorHandler, ZoneBinaryCallback)) {
      return dart.dcall(errorHandler, error, stackTrace);
    } else {
      return dart.dcall(errorHandler, error);
    }
  }
  dart.fn(_invokeErrorHandler, dart.dynamic, [core.Function, core.Object, core.StackTrace]);
  function _registerErrorHandler(errorHandler, zone) {
    if (dart.is(errorHandler, ZoneBinaryCallback)) {
      return zone.registerBinaryCallback(errorHandler);
    } else {
      return zone.registerUnaryCallback(dart.as(errorHandler, __CastType0));
    }
  }
  dart.fn(_registerErrorHandler, () => dart.definiteFunctionType(core.Function, [core.Function, Zone]));
  class AsyncError extends core.Object {
    AsyncError(error, stackTrace) {
      this.error = error;
      this.stackTrace = stackTrace;
    }
    toString() {
      return dart.toString(this.error);
    }
  }
  AsyncError[dart.implements] = () => [core.Error];
  dart.setSignature(AsyncError, {
    constructors: () => ({AsyncError: [AsyncError, [dart.dynamic, core.StackTrace]]})
  });
  class _UncaughtAsyncError extends AsyncError {
    _UncaughtAsyncError(error, stackTrace) {
      super.AsyncError(error, _UncaughtAsyncError._getBestStackTrace(error, stackTrace));
    }
    static _getBestStackTrace(error, stackTrace) {
      if (stackTrace != null)
        return stackTrace;
      if (dart.is(error, core.Error)) {
        return dart.as(dart.dload(error, 'stackTrace'), core.StackTrace);
      }
      return null;
    }
    toString() {
      let result = `Uncaught Error: ${this.error}`;
      if (this.stackTrace != null) {
        result = dart.notNull(result) + `\nStack Trace:\n${this.stackTrace}`;
      }
      return result;
    }
  }
  dart.setSignature(_UncaughtAsyncError, {
    constructors: () => ({_UncaughtAsyncError: [_UncaughtAsyncError, [dart.dynamic, core.StackTrace]]}),
    statics: () => ({_getBestStackTrace: [core.StackTrace, [dart.dynamic, core.StackTrace]]}),
    names: ['_getBestStackTrace']
  });
  const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const _add = Symbol('_add');
  const _closeUnchecked = Symbol('_closeUnchecked');
  const _addError = Symbol('_addError');
  const _completeError = Symbol('_completeError');
  const _complete = Symbol('_complete');
  const _sink = Symbol('_sink');
  const Stream$ = dart.generic(function(T) {
    class Stream extends core.Object {
      Stream() {
      }
      static fromFuture(future) {
        let controller = dart.as(StreamController$(T).new({sync: true}), _StreamController$(T));
        future.then(dart.fn(value => {
          controller[_add](dart.as(value, T));
          controller[_closeUnchecked]();
        }), {onError: dart.fn((error, stackTrace) => {
            controller[_addError](error, dart.as(stackTrace, core.StackTrace));
            controller[_closeUnchecked]();
          })});
        return controller.stream;
      }
      static fromIterable(data) {
        return new (_GeneratedStreamImpl$(T))(dart.fn(() => new (_IterablePendingEvents$(T))(data), _IterablePendingEvents$(T), []));
      }
      static periodic(period, computation) {
        if (computation === void 0)
          computation = null;
        if (computation == null)
          computation = dart.fn(i => null, dart.bottom, [dart.dynamic]);
        let timer = null;
        let computationCount = 0;
        let controller = null;
        let watch = new core.Stopwatch();
        function sendEvent() {
          watch.reset();
          let data = computation((() => {
            let x = computationCount;
            computationCount = dart.notNull(x) + 1;
            return x;
          })());
          controller.add(data);
        }
        dart.fn(sendEvent, dart.void, []);
        function startPeriodicTimer() {
          dart.assert(timer == null);
          timer = Timer.periodic(period, dart.fn(timer => {
            sendEvent();
          }, dart.dynamic, [Timer]));
        }
        dart.fn(startPeriodicTimer, dart.void, []);
        controller = StreamController$(T).new({sync: true, onListen: dart.fn(() => {
            watch.start();
            startPeriodicTimer();
          }), onPause: dart.fn(() => {
            timer.cancel();
            timer = null;
            watch.stop();
          }), onResume: dart.fn(() => {
            dart.assert(timer == null);
            let elapsed = watch.elapsed;
            watch.start();
            timer = Timer.new(period['-'](elapsed), dart.fn(() => {
              timer = null;
              startPeriodicTimer();
              sendEvent();
            }));
          }), onCancel: dart.fn(() => {
            if (timer != null)
              timer.cancel();
            timer = null;
          })});
        return controller.stream;
      }
      static eventTransformed(source, mapSink) {
        return new (_BoundSinkStream$(dart.dynamic, T))(source, dart.as(mapSink, _SinkMapper));
      }
      get isBroadcast() {
        return false;
      }
      asBroadcastStream(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        dart.as(onListen, dart.functionType(dart.void, [StreamSubscription$(T)]));
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        dart.as(onCancel, dart.functionType(dart.void, [StreamSubscription$(T)]));
        return new (_AsBroadcastStream$(T))(this, dart.as(onListen, __CastType12), dart.as(onCancel, dart.functionType(dart.void, [StreamSubscription])));
      }
      where(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        return new (_WhereStream$(T))(this, test);
      }
      map(convert) {
        dart.as(convert, dart.functionType(dart.dynamic, [T]));
        return new (_MapStream$(T, dart.dynamic))(this, convert);
      }
      asyncMap(convert) {
        dart.as(convert, dart.functionType(dart.dynamic, [T]));
        let controller = null;
        let subscription = null;
        const onListen = (function() {
          let add = dart.bind(controller, 'add');
          dart.assert(dart.is(controller, _StreamController) || dart.is(controller, _BroadcastStreamController));
          let eventSink = dart.as(controller, _EventSink$(T));
          let addError = dart.bind(eventSink, _addError);
          subscription = this.listen(dart.fn(event => {
            dart.as(event, T);
            let newValue = null;
            try {
              newValue = convert(event);
            } catch (e) {
              let s = dart.stackTrace(e);
              controller.addError(e, s);
              return;
            }

            if (dart.is(newValue, Future)) {
              subscription.pause();
              dart.dsend(dart.dsend(newValue, 'then', add, {onError: addError}), 'whenComplete', dart.bind(subscription, 'resume'));
            } else {
              controller.add(newValue);
            }
          }, dart.dynamic, [T]), {onError: addError, onDone: dart.bind(controller, 'close')});
        }).bind(this);
        dart.fn(onListen, dart.void, []);
        if (dart.notNull(this.isBroadcast)) {
          controller = StreamController.broadcast({onListen: onListen, onCancel: dart.fn(() => {
              subscription.cancel();
            }), sync: true});
        } else {
          controller = StreamController.new({onListen: onListen, onPause: dart.fn(() => {
              subscription.pause();
            }), onResume: dart.fn(() => {
              subscription.resume();
            }), onCancel: dart.fn(() => {
              subscription.cancel();
            }), sync: true});
        }
        return controller.stream;
      }
      asyncExpand(convert) {
        dart.as(convert, dart.functionType(Stream$(), [T]));
        let controller = null;
        let subscription = null;
        const onListen = (function() {
          dart.assert(dart.is(controller, _StreamController) || dart.is(controller, _BroadcastStreamController));
          let eventSink = dart.as(controller, _EventSink$(T));
          subscription = this.listen(dart.fn(event => {
            dart.as(event, T);
            let newStream = null;
            try {
              newStream = convert(event);
            } catch (e) {
              let s = dart.stackTrace(e);
              controller.addError(e, s);
              return;
            }

            if (newStream != null) {
              subscription.pause();
              controller.addStream(newStream).whenComplete(dart.bind(subscription, 'resume'));
            }
          }, dart.dynamic, [T]), {onError: dart.bind(eventSink, _addError), onDone: dart.bind(controller, 'close')});
        }).bind(this);
        dart.fn(onListen, dart.void, []);
        if (dart.notNull(this.isBroadcast)) {
          controller = StreamController.broadcast({onListen: onListen, onCancel: dart.fn(() => {
              subscription.cancel();
            }), sync: true});
        } else {
          controller = StreamController.new({onListen: onListen, onPause: dart.fn(() => {
              subscription.pause();
            }), onResume: dart.fn(() => {
              subscription.resume();
            }), onCancel: dart.fn(() => {
              subscription.cancel();
            }), sync: true});
        }
        return controller.stream;
      }
      handleError(onError, opts) {
        let test = opts && 'test' in opts ? opts.test : null;
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        return new (_HandleErrorStream$(T))(this, onError, test);
      }
      expand(convert) {
        dart.as(convert, dart.functionType(core.Iterable, [T]));
        return new (_ExpandStream$(T, dart.dynamic))(this, convert);
      }
      pipe(streamConsumer) {
        dart.as(streamConsumer, StreamConsumer$(T));
        return streamConsumer.addStream(this).then(dart.fn(_ => streamConsumer.close(), Future, [dart.dynamic]));
      }
      transform(streamTransformer) {
        dart.as(streamTransformer, StreamTransformer$(T, dart.dynamic));
        return streamTransformer.bind(this);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(T, [T, T]));
        let result = new (_Future$(T))();
        let seenFirst = false;
        let value = null;
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          if (dart.notNull(seenFirst)) {
            _runUserCode(dart.fn(() => combine(value, element), T, []), dart.fn(newValue => {
              dart.as(newValue, T);
              value = newValue;
            }, dart.dynamic, [T]), dart.as(_cancelAndErrorClosure(subscription, result), __CastType14));
          } else {
            value = element;
            seenFirst = true;
          }
        }, dart.dynamic, [T]), {onError: dart.bind(result, _completeError), onDone: dart.fn(() => {
            if (!dart.notNull(seenFirst)) {
              try {
                dart.throw(_internal.IterableElementError.noElement());
              } catch (e) {
                let s = dart.stackTrace(e);
                _completeWithErrorCallback(result, e, s);
              }

            } else {
              result[_complete](value);
            }
          }), cancelOnError: true});
        return result;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, T]));
        let result = new _Future();
        let value = initialValue;
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          _runUserCode(dart.fn(() => dart.dcall(combine, value, element)), dart.fn(newValue => {
            value = newValue;
          }), dart.as(_cancelAndErrorClosure(subscription, result), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.fn((e, st) => {
            result[_completeError](e, dart.as(st, core.StackTrace));
          }), onDone: dart.fn(() => {
            result[_complete](value);
          }), cancelOnError: true});
        return result;
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        let result = new (_Future$(core.String))();
        let buffer = new core.StringBuffer();
        let subscription = null;
        let first = true;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          if (!dart.notNull(first)) {
            buffer.write(separator);
          }
          first = false;
          try {
            buffer.write(element);
          } catch (e) {
            let s = dart.stackTrace(e);
            _cancelAndErrorWithReplacement(subscription, result, e, s);
          }

        }, dart.dynamic, [T]), {onError: dart.fn(e => {
            result[_completeError](e);
          }), onDone: dart.fn(() => {
            result[_complete](dart.toString(buffer));
          }), cancelOnError: true});
        return result;
      }
      contains(needle) {
        let future = new (_Future$(core.bool))();
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          _runUserCode(dart.fn(() => dart.equals(element, needle), core.bool, []), dart.fn(isMatch => {
            if (dart.notNull(isMatch)) {
              _cancelAndValue(subscription, future, true);
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](false);
          }), cancelOnError: true});
        return future;
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [T]));
        let future = new _Future();
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          _runUserCode(dart.fn(() => action(element), dart.void, []), dart.fn(_ => {
          }), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](null);
          }), cancelOnError: true});
        return future;
      }
      every(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        let future = new (_Future$(core.bool))();
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          _runUserCode(dart.fn(() => test(element), core.bool, []), dart.fn(isMatch => {
            if (!dart.notNull(isMatch)) {
              _cancelAndValue(subscription, future, false);
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](true);
          }), cancelOnError: true});
        return future;
      }
      any(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        let future = new (_Future$(core.bool))();
        let subscription = null;
        subscription = this.listen(dart.fn(element => {
          dart.as(element, T);
          _runUserCode(dart.fn(() => test(element), core.bool, []), dart.fn(isMatch => {
            if (dart.notNull(isMatch)) {
              _cancelAndValue(subscription, future, true);
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](false);
          }), cancelOnError: true});
        return future;
      }
      get length() {
        let future = new (_Future$(core.int))();
        let count = 0;
        this.listen(dart.fn(_ => {
          count = dart.notNull(count) + 1;
        }), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](count);
          }), cancelOnError: true});
        return future;
      }
      get isEmpty() {
        let future = new (_Future$(core.bool))();
        let subscription = null;
        subscription = this.listen(dart.fn(_ => {
          _cancelAndValue(subscription, future, false);
        }), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](true);
          }), cancelOnError: true});
        return future;
      }
      toList() {
        let result = dart.list([], T);
        let future = new (_Future$(core.List$(T)))();
        this.listen(dart.fn(data => {
          dart.as(data, T);
          result[dartx.add](data);
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](result);
          }), cancelOnError: true});
        return future;
      }
      toSet() {
        let result = core.Set$(T).new();
        let future = new (_Future$(core.Set$(T)))();
        this.listen(dart.fn(data => {
          dart.as(data, T);
          result.add(data);
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            future[_complete](result);
          }), cancelOnError: true});
        return future;
      }
      drain(futureValue) {
        if (futureValue === void 0)
          futureValue = null;
        return this.listen(null, {cancelOnError: true}).asFuture(futureValue);
      }
      take(count) {
        return new (_TakeStream$(T))(this, count);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        return new (_TakeWhileStream$(T))(this, test);
      }
      skip(count) {
        return new (_SkipStream$(T))(this, count);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        return new (_SkipWhileStream$(T))(this, test);
      }
      distinct(equals) {
        if (equals === void 0)
          equals = null;
        dart.as(equals, dart.functionType(core.bool, [T, T]));
        return new (_DistinctStream$(T))(this, equals);
      }
      get first() {
        let future = new (_Future$(T))();
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          _cancelAndValue(subscription, future, value);
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      get last() {
        let future = new (_Future$(T))();
        let result = null;
        let foundResult = false;
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          foundResult = true;
          result = value;
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            if (dart.notNull(foundResult)) {
              future[_complete](result);
              return;
            }
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      get single() {
        let future = new (_Future$(T))();
        let result = null;
        let foundResult = false;
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          if (dart.notNull(foundResult)) {
            try {
              dart.throw(_internal.IterableElementError.tooMany());
            } catch (e) {
              let s = dart.stackTrace(e);
              _cancelAndErrorWithReplacement(subscription, future, e, s);
            }

            return;
          }
          foundResult = true;
          result = value;
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            if (dart.notNull(foundResult)) {
              future[_complete](result);
              return;
            }
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [T]));
        let defaultValue = opts && 'defaultValue' in opts ? opts.defaultValue : null;
        dart.as(defaultValue, dart.functionType(core.Object, []));
        let future = new _Future();
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          _runUserCode(dart.fn(() => test(value), core.bool, []), dart.fn(isMatch => {
            if (dart.notNull(isMatch)) {
              _cancelAndValue(subscription, future, value);
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            if (defaultValue != null) {
              _runUserCode(defaultValue, dart.bind(future, _complete), dart.bind(future, _completeError));
              return;
            }
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [T]));
        let defaultValue = opts && 'defaultValue' in opts ? opts.defaultValue : null;
        dart.as(defaultValue, dart.functionType(core.Object, []));
        let future = new _Future();
        let result = null;
        let foundResult = false;
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          _runUserCode(dart.fn(() => true == test(value), core.bool, []), dart.fn(isMatch => {
            if (dart.notNull(isMatch)) {
              foundResult = true;
              result = value;
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            if (dart.notNull(foundResult)) {
              future[_complete](result);
              return;
            }
            if (defaultValue != null) {
              _runUserCode(defaultValue, dart.bind(future, _complete), dart.bind(future, _completeError));
              return;
            }
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [T]));
        let future = new (_Future$(T))();
        let result = null;
        let foundResult = false;
        let subscription = null;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          _runUserCode(dart.fn(() => true == test(value), core.bool, []), dart.fn(isMatch => {
            if (dart.notNull(isMatch)) {
              if (dart.notNull(foundResult)) {
                try {
                  dart.throw(_internal.IterableElementError.tooMany());
                } catch (e) {
                  let s = dart.stackTrace(e);
                  _cancelAndErrorWithReplacement(subscription, future, e, s);
                }

                return;
              }
              foundResult = true;
              result = value;
            }
          }, dart.dynamic, [core.bool]), dart.as(_cancelAndErrorClosure(subscription, future), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])));
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn(() => {
            if (dart.notNull(foundResult)) {
              future[_complete](result);
              return;
            }
            try {
              dart.throw(_internal.IterableElementError.noElement());
            } catch (e) {
              let s = dart.stackTrace(e);
              _completeWithErrorCallback(future, e, s);
            }

          }), cancelOnError: true});
        return future;
      }
      elementAt(index) {
        if (!(typeof index == 'number') || dart.notNull(index) < 0)
          dart.throw(new core.ArgumentError(index));
        let future = new (_Future$(T))();
        let subscription = null;
        let elementIndex = 0;
        subscription = this.listen(dart.fn(value => {
          dart.as(value, T);
          if (index == elementIndex) {
            _cancelAndValue(subscription, future, value);
            return;
          }
          elementIndex = dart.notNull(elementIndex) + 1;
        }, dart.dynamic, [T]), {onError: dart.bind(future, _completeError), onDone: dart.fn((() => {
            future[_completeError](core.RangeError.index(index, this, "index", null, elementIndex));
          }).bind(this)), cancelOnError: true});
        return future;
      }
      timeout(timeLimit, opts) {
        let onTimeout = opts && 'onTimeout' in opts ? opts.onTimeout : null;
        dart.as(onTimeout, dart.functionType(dart.void, [EventSink]));
        let controller = null;
        let subscription = null;
        let timer = null;
        let zone = null;
        let timeout2 = null;
        function onData(event) {
          dart.as(event, T);
          timer.cancel();
          controller.add(event);
          timer = zone.createTimer(timeLimit, dart.as(timeout2, __CastType17));
        }
        dart.fn(onData, dart.void, [T]);
        function onError(error, stackTrace) {
          timer.cancel();
          dart.assert(dart.is(controller, _StreamController) || dart.is(controller, _BroadcastStreamController));
          let eventSink = dart.as(controller, _EventSink$(T));
          eventSink[_addError](error, stackTrace);
          timer = zone.createTimer(timeLimit, dart.as(timeout2, dart.functionType(dart.void, [])));
        }
        dart.fn(onError, dart.void, [dart.dynamic, core.StackTrace]);
        function onDone() {
          timer.cancel();
          controller.close();
        }
        dart.fn(onDone, dart.void, []);
        const onListen = (function() {
          zone = Zone.current;
          if (onTimeout == null) {
            timeout2 = dart.fn(() => {
              controller.addError(new TimeoutException("No stream event", timeLimit), null);
            });
          } else {
            onTimeout = dart.as(zone.registerUnaryCallback(onTimeout), __CastType18);
            let wrapper = new _ControllerEventSinkWrapper(null);
            timeout2 = dart.fn(() => {
              wrapper[_sink] = controller;
              zone.runUnaryGuarded(onTimeout, wrapper);
              wrapper[_sink] = null;
            });
          }
          subscription = this.listen(onData, {onError: onError, onDone: onDone});
          timer = zone.createTimer(timeLimit, dart.as(timeout2, dart.functionType(dart.void, [])));
        }).bind(this);
        dart.fn(onListen, dart.void, []);
        function onCancel() {
          timer.cancel();
          let result = subscription.cancel();
          subscription = null;
          return result;
        }
        dart.fn(onCancel, Future, []);
        controller = dart.notNull(this.isBroadcast) ? new _SyncBroadcastStreamController(onListen, onCancel) : new _SyncStreamController(onListen, dart.fn(() => {
          timer.cancel();
          subscription.pause();
        }), dart.fn(() => {
          subscription.resume();
          timer = zone.createTimer(timeLimit, dart.as(timeout2, dart.functionType(dart.void, [])));
        }), onCancel);
        return controller.stream;
      }
    }
    dart.setSignature(Stream, {
      constructors: () => ({
        Stream: [Stream$(T), []],
        fromFuture: [Stream$(T), [Future$(T)]],
        fromIterable: [Stream$(T), [core.Iterable$(T)]],
        periodic: [Stream$(T), [core.Duration], [dart.functionType(T, [core.int])]],
        eventTransformed: [Stream$(T), [Stream$(), dart.functionType(EventSink, [EventSink$(T)])]]
      }),
      methods: () => ({
        asBroadcastStream: [Stream$(T), [], {onListen: dart.functionType(dart.void, [StreamSubscription$(T)]), onCancel: dart.functionType(dart.void, [StreamSubscription$(T)])}],
        where: [Stream$(T), [dart.functionType(core.bool, [T])]],
        map: [Stream$(), [dart.functionType(dart.dynamic, [T])]],
        asyncMap: [Stream$(), [dart.functionType(dart.dynamic, [T])]],
        asyncExpand: [Stream$(), [dart.functionType(Stream$(), [T])]],
        handleError: [Stream$(T), [core.Function], {test: dart.functionType(core.bool, [dart.dynamic])}],
        expand: [Stream$(), [dart.functionType(core.Iterable, [T])]],
        pipe: [Future, [StreamConsumer$(T)]],
        transform: [Stream$(), [StreamTransformer$(T, dart.dynamic)]],
        reduce: [Future$(T), [dart.functionType(T, [T, T])]],
        fold: [Future, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, T])]],
        join: [Future$(core.String), [], [core.String]],
        contains: [Future$(core.bool), [core.Object]],
        forEach: [Future, [dart.functionType(dart.void, [T])]],
        every: [Future$(core.bool), [dart.functionType(core.bool, [T])]],
        any: [Future$(core.bool), [dart.functionType(core.bool, [T])]],
        toList: [Future$(core.List$(T)), []],
        toSet: [Future$(core.Set$(T)), []],
        drain: [Future, [], [dart.dynamic]],
        take: [Stream$(T), [core.int]],
        takeWhile: [Stream$(T), [dart.functionType(core.bool, [T])]],
        skip: [Stream$(T), [core.int]],
        skipWhile: [Stream$(T), [dart.functionType(core.bool, [T])]],
        distinct: [Stream$(T), [], [dart.functionType(core.bool, [T, T])]],
        firstWhere: [Future, [dart.functionType(core.bool, [T])], {defaultValue: dart.functionType(core.Object, [])}],
        lastWhere: [Future, [dart.functionType(core.bool, [T])], {defaultValue: dart.functionType(core.Object, [])}],
        singleWhere: [Future$(T), [dart.functionType(core.bool, [T])]],
        elementAt: [Future$(T), [core.int]],
        timeout: [Stream$(), [core.Duration], {onTimeout: dart.functionType(dart.void, [EventSink])}]
      })
    });
    return Stream;
  });
  let Stream = Stream$();
  const _createSubscription = Symbol('_createSubscription');
  const _onListen = Symbol('_onListen');
  const _StreamImpl$ = dart.generic(function(T) {
    class _StreamImpl extends Stream$(T) {
      _StreamImpl() {
        super.Stream();
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        cancelOnError = core.identical(true, cancelOnError);
        let subscription = this[_createSubscription](onData, onError, onDone, cancelOnError);
        this[_onListen](subscription);
        return dart.as(subscription, StreamSubscription$(T));
      }
      [_createSubscription](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        return new (_BufferingStreamSubscription$(T))(onData, onError, onDone, cancelOnError);
      }
      [_onListen](subscription) {}
    }
    dart.setSignature(_StreamImpl, {
      methods: () => ({
        listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
        [_createSubscription]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]],
        [_onListen]: [dart.void, [StreamSubscription]]
      })
    });
    return _StreamImpl;
  });
  let _StreamImpl = _StreamImpl$();
  const _controller = Symbol('_controller');
  const _subscribe = Symbol('_subscribe');
  const _ControllerStream$ = dart.generic(function(T) {
    class _ControllerStream extends _StreamImpl$(T) {
      _ControllerStream(controller) {
        this[_controller] = controller;
      }
      [_createSubscription](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        return this[_controller][_subscribe](onData, onError, onDone, cancelOnError);
      }
      get hashCode() {
        return dart.notNull(dart.hashCode(this[_controller])) ^ 892482866;
      }
      ['=='](other) {
        if (dart.notNull(core.identical(this, other)))
          return true;
        if (!dart.is(other, _ControllerStream$()))
          return false;
        let otherStream = dart.as(other, _ControllerStream$());
        return core.identical(otherStream[_controller], this[_controller]);
      }
    }
    dart.setSignature(_ControllerStream, {
      constructors: () => ({_ControllerStream: [_ControllerStream$(T), [_StreamControllerLifecycle$(T)]]}),
      methods: () => ({
        [_createSubscription]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]],
        '==': [core.bool, [core.Object]]
      })
    });
    return _ControllerStream;
  });
  let _ControllerStream = _ControllerStream$();
  const _BroadcastStream$ = dart.generic(function(T) {
    class _BroadcastStream extends _ControllerStream$(T) {
      _BroadcastStream(controller) {
        super._ControllerStream(dart.as(controller, _StreamControllerLifecycle$(T)));
      }
      get isBroadcast() {
        return true;
      }
    }
    dart.setSignature(_BroadcastStream, {
      constructors: () => ({_BroadcastStream: [_BroadcastStream$(T), [_StreamControllerLifecycle]]})
    });
    return _BroadcastStream;
  });
  let _BroadcastStream = _BroadcastStream$();
  const _next = Symbol('_next');
  const _previous = Symbol('_previous');
  class _BroadcastSubscriptionLink extends core.Object {
    _BroadcastSubscriptionLink() {
      this[_next] = null;
      this[_previous] = null;
    }
  }
  const _zone = Symbol('_zone');
  const _state = Symbol('_state');
  const _onData = Symbol('_onData');
  const _onError = Symbol('_onError');
  const _onDone = Symbol('_onDone');
  const _cancelFuture = Symbol('_cancelFuture');
  const _pending = Symbol('_pending');
  const _setPendingEvents = Symbol('_setPendingEvents');
  const _isCanceled = Symbol('_isCanceled');
  const _extractPending = Symbol('_extractPending');
  const _isPaused = Symbol('_isPaused');
  const _isInputPaused = Symbol('_isInputPaused');
  const _inCallback = Symbol('_inCallback');
  const _guardCallback = Symbol('_guardCallback');
  const _onPause = Symbol('_onPause');
  const _decrementPauseCount = Symbol('_decrementPauseCount');
  const _hasPending = Symbol('_hasPending');
  const _mayResumeInput = Symbol('_mayResumeInput');
  const _onResume = Symbol('_onResume');
  const _cancel = Symbol('_cancel');
  const _isClosed = Symbol('_isClosed');
  const _waitsForCancel = Symbol('_waitsForCancel');
  const _canFire = Symbol('_canFire');
  const _cancelOnError = Symbol('_cancelOnError');
  const _onCancel = Symbol('_onCancel');
  const _incrementPauseCount = Symbol('_incrementPauseCount');
  const _sendData = Symbol('_sendData');
  const _addPending = Symbol('_addPending');
  const _sendError = Symbol('_sendError');
  const _sendDone = Symbol('_sendDone');
  const _close = Symbol('_close');
  const _checkState = Symbol('_checkState');
  const _BufferingStreamSubscription$ = dart.generic(function(T) {
    class _BufferingStreamSubscription extends core.Object {
      _BufferingStreamSubscription(onData, onError, onDone, cancelOnError) {
        this[_zone] = Zone.current;
        this[_state] = dart.notNull(cancelOnError) ? _BufferingStreamSubscription$()._STATE_CANCEL_ON_ERROR : 0;
        this[_onData] = null;
        this[_onError] = null;
        this[_onDone] = null;
        this[_cancelFuture] = null;
        this[_pending] = null;
        this.onData(onData);
        this.onError(onError);
        this.onDone(onDone);
      }
      [_setPendingEvents](pendingEvents) {
        dart.assert(this[_pending] == null);
        if (pendingEvents == null)
          return;
        this[_pending] = pendingEvents;
        if (!dart.notNull(pendingEvents.isEmpty)) {
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_HAS_PENDING);
          this[_pending].schedule(this);
        }
      }
      [_extractPending]() {
        dart.assert(this[_isCanceled]);
        let events = this[_pending];
        this[_pending] = null;
        return events;
      }
      onData(handleData) {
        dart.as(handleData, dart.functionType(dart.void, [T]));
        if (handleData == null)
          handleData = dart.as(_nullDataHandler, __CastType20$(T));
        this[_onData] = dart.as(this[_zone].registerUnaryCallback(handleData), _DataHandler$(T));
      }
      onError(handleError) {
        if (handleError == null)
          handleError = _nullErrorHandler;
        this[_onError] = _registerErrorHandler(handleError, this[_zone]);
      }
      onDone(handleDone) {
        dart.as(handleDone, dart.functionType(dart.void, []));
        if (handleDone == null)
          handleDone = _nullDoneHandler;
        this[_onDone] = this[_zone].registerCallback(handleDone);
      }
      pause(resumeSignal) {
        if (resumeSignal === void 0)
          resumeSignal = null;
        if (dart.notNull(this[_isCanceled]))
          return;
        let wasPaused = this[_isPaused];
        let wasInputPaused = this[_isInputPaused];
        this[_state] = dart.notNull(this[_state]) + dart.notNull(_BufferingStreamSubscription$()._STATE_PAUSE_COUNT) | dart.notNull(_BufferingStreamSubscription$()._STATE_INPUT_PAUSED);
        if (resumeSignal != null)
          resumeSignal.whenComplete(dart.bind(this, 'resume'));
        if (!dart.notNull(wasPaused) && this[_pending] != null)
          this[_pending].cancelSchedule();
        if (!dart.notNull(wasInputPaused) && !dart.notNull(this[_inCallback]))
          this[_guardCallback](dart.bind(this, _onPause));
      }
      resume() {
        if (dart.notNull(this[_isCanceled]))
          return;
        if (dart.notNull(this[_isPaused])) {
          this[_decrementPauseCount]();
          if (!dart.notNull(this[_isPaused])) {
            if (dart.notNull(this[_hasPending]) && !dart.notNull(this[_pending].isEmpty)) {
              this[_pending].schedule(this);
            } else {
              dart.assert(this[_mayResumeInput]);
              this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_INPUT_PAUSED);
              if (!dart.notNull(this[_inCallback]))
                this[_guardCallback](dart.bind(this, _onResume));
            }
          }
        }
      }
      cancel() {
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_WAIT_FOR_CANCEL);
        if (dart.notNull(this[_isCanceled]))
          return this[_cancelFuture];
        this[_cancel]();
        return this[_cancelFuture];
      }
      asFuture(futureValue) {
        if (futureValue === void 0)
          futureValue = null;
        let result = new (_Future$(T))();
        this[_onDone] = dart.fn(() => {
          result[_complete](futureValue);
        });
        this[_onError] = dart.fn(((error, stackTrace) => {
          this.cancel();
          result[_completeError](error, dart.as(stackTrace, core.StackTrace));
        }).bind(this));
        return result;
      }
      get [_isInputPaused]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_INPUT_PAUSED)) != 0;
      }
      get [_isClosed]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_CLOSED)) != 0;
      }
      get [_isCanceled]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_CANCELED)) != 0;
      }
      get [_waitsForCancel]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_WAIT_FOR_CANCEL)) != 0;
      }
      get [_inCallback]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK)) != 0;
      }
      get [_hasPending]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_HAS_PENDING)) != 0;
      }
      get [_isPaused]() {
        return dart.notNull(this[_state]) >= dart.notNull(_BufferingStreamSubscription$()._STATE_PAUSE_COUNT);
      }
      get [_canFire]() {
        return dart.notNull(this[_state]) < dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
      }
      get [_mayResumeInput]() {
        return !dart.notNull(this[_isPaused]) && (this[_pending] == null || dart.notNull(this[_pending].isEmpty));
      }
      get [_cancelOnError]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BufferingStreamSubscription$()._STATE_CANCEL_ON_ERROR)) != 0;
      }
      get isPaused() {
        return this[_isPaused];
      }
      [_cancel]() {
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_CANCELED);
        if (dart.notNull(this[_hasPending])) {
          this[_pending].cancelSchedule();
        }
        if (!dart.notNull(this[_inCallback]))
          this[_pending] = null;
        this[_cancelFuture] = this[_onCancel]();
      }
      [_incrementPauseCount]() {
        this[_state] = dart.notNull(this[_state]) + dart.notNull(_BufferingStreamSubscription$()._STATE_PAUSE_COUNT) | dart.notNull(_BufferingStreamSubscription$()._STATE_INPUT_PAUSED);
      }
      [_decrementPauseCount]() {
        dart.assert(this[_isPaused]);
        this[_state] = dart.notNull(this[_state]) - dart.notNull(_BufferingStreamSubscription$()._STATE_PAUSE_COUNT);
      }
      [_add](data) {
        dart.as(data, T);
        dart.assert(!dart.notNull(this[_isClosed]));
        if (dart.notNull(this[_isCanceled]))
          return;
        if (dart.notNull(this[_canFire])) {
          this[_sendData](data);
        } else {
          this[_addPending](new _DelayedData(data));
        }
      }
      [_addError](error, stackTrace) {
        if (dart.notNull(this[_isCanceled]))
          return;
        if (dart.notNull(this[_canFire])) {
          this[_sendError](error, stackTrace);
        } else {
          this[_addPending](new _DelayedError(error, stackTrace));
        }
      }
      [_close]() {
        dart.assert(!dart.notNull(this[_isClosed]));
        if (dart.notNull(this[_isCanceled]))
          return;
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_CLOSED);
        if (dart.notNull(this[_canFire])) {
          this[_sendDone]();
        } else {
          this[_addPending](dart.const(new _DelayedDone()));
        }
      }
      [_onPause]() {
        dart.assert(this[_isInputPaused]);
      }
      [_onResume]() {
        dart.assert(!dart.notNull(this[_isInputPaused]));
      }
      [_onCancel]() {
        dart.assert(this[_isCanceled]);
        return null;
      }
      [_addPending](event) {
        let pending = dart.as(this[_pending], _StreamImplEvents);
        if (this[_pending] == null)
          pending = this[_pending] = new _StreamImplEvents();
        pending.add(event);
        if (!dart.notNull(this[_hasPending])) {
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_HAS_PENDING);
          if (!dart.notNull(this[_isPaused])) {
            this[_pending].schedule(this);
          }
        }
      }
      [_sendData](data) {
        dart.as(data, T);
        dart.assert(!dart.notNull(this[_isCanceled]));
        dart.assert(!dart.notNull(this[_isPaused]));
        dart.assert(!dart.notNull(this[_inCallback]));
        let wasInputPaused = this[_isInputPaused];
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        this[_zone].runUnaryGuarded(this[_onData], data);
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        this[_checkState](wasInputPaused);
      }
      [_sendError](error, stackTrace) {
        dart.assert(!dart.notNull(this[_isCanceled]));
        dart.assert(!dart.notNull(this[_isPaused]));
        dart.assert(!dart.notNull(this[_inCallback]));
        let wasInputPaused = this[_isInputPaused];
        const sendError = (function() {
          if (dart.notNull(this[_isCanceled]) && !dart.notNull(this[_waitsForCancel]))
            return;
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
          if (dart.is(this[_onError], ZoneBinaryCallback)) {
            this[_zone].runBinaryGuarded(dart.as(this[_onError], __CastType22), error, stackTrace);
          } else {
            this[_zone].runUnaryGuarded(dart.as(this[_onError], __CastType25), error);
          }
          this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        }).bind(this);
        dart.fn(sendError, dart.void, []);
        if (dart.notNull(this[_cancelOnError])) {
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_WAIT_FOR_CANCEL);
          this[_cancel]();
          if (dart.is(this[_cancelFuture], Future)) {
            this[_cancelFuture].whenComplete(sendError);
          } else {
            sendError();
          }
        } else {
          sendError();
          this[_checkState](wasInputPaused);
        }
      }
      [_sendDone]() {
        dart.assert(!dart.notNull(this[_isCanceled]));
        dart.assert(!dart.notNull(this[_isPaused]));
        dart.assert(!dart.notNull(this[_inCallback]));
        const sendDone = (function() {
          if (!dart.notNull(this[_waitsForCancel]))
            return;
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_CANCELED) | dart.notNull(_BufferingStreamSubscription$()._STATE_CLOSED) | dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
          this[_zone].runGuarded(this[_onDone]);
          this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        }).bind(this);
        dart.fn(sendDone, dart.void, []);
        this[_cancel]();
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_WAIT_FOR_CANCEL);
        if (dart.is(this[_cancelFuture], Future)) {
          this[_cancelFuture].whenComplete(sendDone);
        } else {
          sendDone();
        }
      }
      [_guardCallback](callback) {
        dart.assert(!dart.notNull(this[_inCallback]));
        let wasInputPaused = this[_isInputPaused];
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        dart.dcall(callback);
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
        this[_checkState](wasInputPaused);
      }
      [_checkState](wasInputPaused) {
        dart.assert(!dart.notNull(this[_inCallback]));
        if (dart.notNull(this[_hasPending]) && dart.notNull(this[_pending].isEmpty)) {
          this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_HAS_PENDING);
          if (dart.notNull(this[_isInputPaused]) && dart.notNull(this[_mayResumeInput])) {
            this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_INPUT_PAUSED);
          }
        }
        while (true) {
          if (dart.notNull(this[_isCanceled])) {
            this[_pending] = null;
            return;
          }
          let isInputPaused = this[_isInputPaused];
          if (wasInputPaused == isInputPaused)
            break;
          this[_state] = dart.notNull(this[_state]) ^ dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
          if (dart.notNull(isInputPaused)) {
            this[_onPause]();
          } else {
            this[_onResume]();
          }
          this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BufferingStreamSubscription$()._STATE_IN_CALLBACK);
          wasInputPaused = isInputPaused;
        }
        if (dart.notNull(this[_hasPending]) && !dart.notNull(this[_isPaused])) {
          this[_pending].schedule(this);
        }
      }
    }
    _BufferingStreamSubscription[dart.implements] = () => [StreamSubscription$(T), _EventSink$(T), _EventDispatch$(T)];
    dart.setSignature(_BufferingStreamSubscription, {
      constructors: () => ({_BufferingStreamSubscription: [_BufferingStreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]}),
      methods: () => ({
        [_setPendingEvents]: [dart.void, [_PendingEvents]],
        [_extractPending]: [_PendingEvents, []],
        onData: [dart.void, [dart.functionType(dart.void, [T])]],
        onError: [dart.void, [core.Function]],
        onDone: [dart.void, [dart.functionType(dart.void, [])]],
        pause: [dart.void, [], [Future]],
        resume: [dart.void, []],
        cancel: [Future, []],
        asFuture: [Future, [], [dart.dynamic]],
        [_cancel]: [dart.void, []],
        [_incrementPauseCount]: [dart.void, []],
        [_decrementPauseCount]: [dart.void, []],
        [_add]: [dart.void, [T]],
        [_addError]: [dart.void, [core.Object, core.StackTrace]],
        [_close]: [dart.void, []],
        [_onPause]: [dart.void, []],
        [_onResume]: [dart.void, []],
        [_onCancel]: [Future, []],
        [_addPending]: [dart.void, [_DelayedEvent]],
        [_sendData]: [dart.void, [T]],
        [_sendError]: [dart.void, [core.Object, core.StackTrace]],
        [_sendDone]: [dart.void, []],
        [_guardCallback]: [dart.void, [dart.dynamic]],
        [_checkState]: [dart.void, [core.bool]]
      })
    });
    return _BufferingStreamSubscription;
  });
  let _BufferingStreamSubscription = _BufferingStreamSubscription$();
  const _recordCancel = Symbol('_recordCancel');
  const _recordPause = Symbol('_recordPause');
  const _recordResume = Symbol('_recordResume');
  const _ControllerSubscription$ = dart.generic(function(T) {
    class _ControllerSubscription extends _BufferingStreamSubscription$(T) {
      _ControllerSubscription(controller, onData, onError, onDone, cancelOnError) {
        this[_controller] = controller;
        super._BufferingStreamSubscription(onData, onError, onDone, cancelOnError);
      }
      [_onCancel]() {
        return this[_controller][_recordCancel](this);
      }
      [_onPause]() {
        this[_controller][_recordPause](this);
      }
      [_onResume]() {
        this[_controller][_recordResume](this);
      }
    }
    dart.setSignature(_ControllerSubscription, {
      constructors: () => ({_ControllerSubscription: [_ControllerSubscription$(T), [_StreamControllerLifecycle$(T), dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]})
    });
    return _ControllerSubscription;
  });
  let _ControllerSubscription = _ControllerSubscription$();
  const _eventState = Symbol('_eventState');
  const _expectsEvent = Symbol('_expectsEvent');
  const _toggleEventId = Symbol('_toggleEventId');
  const _isFiring = Symbol('_isFiring');
  const _setRemoveAfterFiring = Symbol('_setRemoveAfterFiring');
  const _removeAfterFiring = Symbol('_removeAfterFiring');
  const _BroadcastSubscription$ = dart.generic(function(T) {
    class _BroadcastSubscription extends _ControllerSubscription$(T) {
      _BroadcastSubscription(controller, onData, onError, onDone, cancelOnError) {
        this[_eventState] = null;
        this[_next] = null;
        this[_previous] = null;
        super._ControllerSubscription(dart.as(controller, _StreamControllerLifecycle$(T)), onData, onError, onDone, cancelOnError);
        this[_next] = this[_previous] = this;
      }
      [_expectsEvent](eventId) {
        return (dart.notNull(this[_eventState]) & dart.notNull(_BroadcastSubscription$()._STATE_EVENT_ID)) == eventId;
      }
      [_toggleEventId]() {
        this[_eventState] = dart.notNull(this[_eventState]) ^ dart.notNull(_BroadcastSubscription$()._STATE_EVENT_ID);
      }
      get [_isFiring]() {
        return (dart.notNull(this[_eventState]) & dart.notNull(_BroadcastSubscription$()._STATE_FIRING)) != 0;
      }
      [_setRemoveAfterFiring]() {
        dart.assert(this[_isFiring]);
        this[_eventState] = dart.notNull(this[_eventState]) | dart.notNull(_BroadcastSubscription$()._STATE_REMOVE_AFTER_FIRING);
      }
      get [_removeAfterFiring]() {
        return (dart.notNull(this[_eventState]) & dart.notNull(_BroadcastSubscription$()._STATE_REMOVE_AFTER_FIRING)) != 0;
      }
      [_onPause]() {}
      [_onResume]() {}
    }
    _BroadcastSubscription[dart.implements] = () => [_BroadcastSubscriptionLink];
    dart.setSignature(_BroadcastSubscription, {
      constructors: () => ({_BroadcastSubscription: [_BroadcastSubscription$(T), [_StreamControllerLifecycle, dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]}),
      methods: () => ({
        [_expectsEvent]: [core.bool, [core.int]],
        [_toggleEventId]: [dart.void, []],
        [_setRemoveAfterFiring]: [dart.void, []]
      })
    });
    return _BroadcastSubscription;
  });
  let _BroadcastSubscription = _BroadcastSubscription$();
  _BroadcastSubscription._STATE_EVENT_ID = 1;
  _BroadcastSubscription._STATE_FIRING = 2;
  _BroadcastSubscription._STATE_REMOVE_AFTER_FIRING = 4;
  const _addStreamState = Symbol('_addStreamState');
  const _doneFuture = Symbol('_doneFuture');
  const _isEmpty = Symbol('_isEmpty');
  const _hasOneListener = Symbol('_hasOneListener');
  const _isAddingStream = Symbol('_isAddingStream');
  const _mayAddEvent = Symbol('_mayAddEvent');
  const _ensureDoneFuture = Symbol('_ensureDoneFuture');
  const _addListener = Symbol('_addListener');
  const _removeListener = Symbol('_removeListener');
  const _callOnCancel = Symbol('_callOnCancel');
  const _addEventError = Symbol('_addEventError');
  const _forEachListener = Symbol('_forEachListener');
  const _mayComplete = Symbol('_mayComplete');
  const _asyncComplete = Symbol('_asyncComplete');
  const _BroadcastStreamController$ = dart.generic(function(T) {
    class _BroadcastStreamController extends core.Object {
      _BroadcastStreamController(onListen, onCancel) {
        this[_onListen] = onListen;
        this[_onCancel] = onCancel;
        this[_state] = _BroadcastStreamController$()._STATE_INITIAL;
        this[_next] = null;
        this[_previous] = null;
        this[_addStreamState] = null;
        this[_doneFuture] = null;
        this[_next] = this[_previous] = this;
      }
      get stream() {
        return new (_BroadcastStream$(T))(this);
      }
      get sink() {
        return new (_StreamSinkWrapper$(T))(this);
      }
      get isClosed() {
        return (dart.notNull(this[_state]) & dart.notNull(_BroadcastStreamController$()._STATE_CLOSED)) != 0;
      }
      get isPaused() {
        return false;
      }
      get hasListener() {
        return !dart.notNull(this[_isEmpty]);
      }
      get [_hasOneListener]() {
        dart.assert(!dart.notNull(this[_isEmpty]));
        return core.identical(this[_next][_next], this);
      }
      get [_isFiring]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BroadcastStreamController$()._STATE_FIRING)) != 0;
      }
      get [_isAddingStream]() {
        return (dart.notNull(this[_state]) & dart.notNull(_BroadcastStreamController$()._STATE_ADDSTREAM)) != 0;
      }
      get [_mayAddEvent]() {
        return dart.notNull(this[_state]) < dart.notNull(_BroadcastStreamController$()._STATE_CLOSED);
      }
      [_ensureDoneFuture]() {
        if (this[_doneFuture] != null)
          return this[_doneFuture];
        return this[_doneFuture] = new _Future();
      }
      get [_isEmpty]() {
        return core.identical(this[_next], this);
      }
      [_addListener](subscription) {
        dart.as(subscription, _BroadcastSubscription$(T));
        dart.assert(core.identical(subscription[_next], subscription));
        subscription[_previous] = this[_previous];
        subscription[_next] = this;
        this[_previous][_next] = subscription;
        this[_previous] = subscription;
        subscription[_eventState] = dart.notNull(this[_state]) & dart.notNull(_BroadcastStreamController$()._STATE_EVENT_ID);
      }
      [_removeListener](subscription) {
        dart.as(subscription, _BroadcastSubscription$(T));
        dart.assert(core.identical(subscription[_controller], this));
        dart.assert(!dart.notNull(core.identical(subscription[_next], subscription)));
        let previous = subscription[_previous];
        let next = subscription[_next];
        previous[_next] = next;
        next[_previous] = previous;
        subscription[_next] = subscription[_previous] = subscription;
      }
      [_subscribe](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        if (dart.notNull(this.isClosed)) {
          if (onDone == null)
            onDone = _nullDoneHandler;
          return new (_DoneStreamSubscription$(T))(onDone);
        }
        let subscription = new (_BroadcastSubscription$(T))(this, onData, onError, onDone, cancelOnError);
        this[_addListener](dart.as(subscription, _BroadcastSubscription$(T)));
        if (dart.notNull(core.identical(this[_next], this[_previous]))) {
          _runGuarded(this[_onListen]);
        }
        return dart.as(subscription, StreamSubscription$(T));
      }
      [_recordCancel](sub) {
        dart.as(sub, StreamSubscription$(T));
        let subscription = dart.as(sub, _BroadcastSubscription$(T));
        if (dart.notNull(core.identical(subscription[_next], subscription)))
          return null;
        dart.assert(!dart.notNull(core.identical(subscription[_next], subscription)));
        if (dart.notNull(subscription[_isFiring])) {
          subscription[_setRemoveAfterFiring]();
        } else {
          dart.assert(!dart.notNull(core.identical(subscription[_next], subscription)));
          this[_removeListener](subscription);
          if (!dart.notNull(this[_isFiring]) && dart.notNull(this[_isEmpty])) {
            this[_callOnCancel]();
          }
        }
        return null;
      }
      [_recordPause](subscription) {
        dart.as(subscription, StreamSubscription$(T));
      }
      [_recordResume](subscription) {
        dart.as(subscription, StreamSubscription$(T));
      }
      [_addEventError]() {
        if (dart.notNull(this.isClosed)) {
          return new core.StateError("Cannot add new events after calling close");
        }
        dart.assert(this[_isAddingStream]);
        return new core.StateError("Cannot add new events while doing an addStream");
      }
      add(data) {
        dart.as(data, T);
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_addEventError]());
        this[_sendData](data);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        error = _nonNullError(error);
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_addEventError]());
        let replacement = Zone.current.errorCallback(error, stackTrace);
        if (replacement != null) {
          error = _nonNullError(replacement.error);
          stackTrace = replacement.stackTrace;
        }
        this[_sendError](error, stackTrace);
      }
      close() {
        if (dart.notNull(this.isClosed)) {
          dart.assert(this[_doneFuture] != null);
          return this[_doneFuture];
        }
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_addEventError]());
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BroadcastStreamController$()._STATE_CLOSED);
        let doneFuture = this[_ensureDoneFuture]();
        this[_sendDone]();
        return doneFuture;
      }
      get done() {
        return this[_ensureDoneFuture]();
      }
      addStream(stream, opts) {
        dart.as(stream, Stream$(T));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : true;
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_addEventError]());
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_BroadcastStreamController$()._STATE_ADDSTREAM);
        this[_addStreamState] = new (_AddStreamState$(T))(this, stream, cancelOnError);
        return this[_addStreamState].addStreamFuture;
      }
      [_add](data) {
        dart.as(data, T);
        this[_sendData](data);
      }
      [_addError](error, stackTrace) {
        this[_sendError](error, stackTrace);
      }
      [_close]() {
        dart.assert(this[_isAddingStream]);
        let addState = this[_addStreamState];
        this[_addStreamState] = null;
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BroadcastStreamController$()._STATE_ADDSTREAM);
        addState.complete();
      }
      [_forEachListener](action) {
        dart.as(action, dart.functionType(dart.void, [_BufferingStreamSubscription$(T)]));
        if (dart.notNull(this[_isFiring])) {
          dart.throw(new core.StateError("Cannot fire new event. Controller is already firing an event"));
        }
        if (dart.notNull(this[_isEmpty]))
          return;
        let id = dart.notNull(this[_state]) & dart.notNull(_BroadcastStreamController$()._STATE_EVENT_ID);
        this[_state] = dart.notNull(this[_state]) ^ (dart.notNull(_BroadcastStreamController$()._STATE_EVENT_ID) | dart.notNull(_BroadcastStreamController$()._STATE_FIRING));
        let link = this[_next];
        while (!dart.notNull(core.identical(link, this))) {
          let subscription = dart.as(link, _BroadcastSubscription$(T));
          if (dart.notNull(subscription[_expectsEvent](id))) {
            subscription[_eventState] = dart.notNull(subscription[_eventState]) | dart.notNull(_BroadcastSubscription._STATE_FIRING);
            action(subscription);
            subscription[_toggleEventId]();
            link = subscription[_next];
            if (dart.notNull(subscription[_removeAfterFiring])) {
              this[_removeListener](subscription);
            }
            subscription[_eventState] = dart.notNull(subscription[_eventState]) & ~dart.notNull(_BroadcastSubscription._STATE_FIRING);
          } else {
            link = subscription[_next];
          }
        }
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BroadcastStreamController$()._STATE_FIRING);
        if (dart.notNull(this[_isEmpty])) {
          this[_callOnCancel]();
        }
      }
      [_callOnCancel]() {
        dart.assert(this[_isEmpty]);
        if (dart.notNull(this.isClosed) && dart.notNull(this[_doneFuture][_mayComplete])) {
          this[_doneFuture][_asyncComplete](null);
        }
        _runGuarded(this[_onCancel]);
      }
    }
    _BroadcastStreamController[dart.implements] = () => [StreamController$(T), _StreamControllerLifecycle$(T), _BroadcastSubscriptionLink, _EventSink$(T), _EventDispatch$(T)];
    dart.setSignature(_BroadcastStreamController, {
      constructors: () => ({_BroadcastStreamController: [_BroadcastStreamController$(T), [_NotificationHandler, _NotificationHandler]]}),
      methods: () => ({
        [_ensureDoneFuture]: [_Future, []],
        [_addListener]: [dart.void, [_BroadcastSubscription$(T)]],
        [_removeListener]: [dart.void, [_BroadcastSubscription$(T)]],
        [_subscribe]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]],
        [_recordCancel]: [Future, [StreamSubscription$(T)]],
        [_recordPause]: [dart.void, [StreamSubscription$(T)]],
        [_recordResume]: [dart.void, [StreamSubscription$(T)]],
        [_addEventError]: [core.Error, []],
        add: [dart.void, [T]],
        addError: [dart.void, [core.Object], [core.StackTrace]],
        close: [Future, []],
        addStream: [Future, [Stream$(T)], {cancelOnError: core.bool}],
        [_add]: [dart.void, [T]],
        [_addError]: [dart.void, [core.Object, core.StackTrace]],
        [_close]: [dart.void, []],
        [_forEachListener]: [dart.void, [dart.functionType(dart.void, [_BufferingStreamSubscription$(T)])]],
        [_callOnCancel]: [dart.void, []]
      })
    });
    return _BroadcastStreamController;
  });
  let _BroadcastStreamController = _BroadcastStreamController$();
  _BroadcastStreamController._STATE_INITIAL = 0;
  _BroadcastStreamController._STATE_EVENT_ID = 1;
  _BroadcastStreamController._STATE_FIRING = 2;
  _BroadcastStreamController._STATE_CLOSED = 4;
  _BroadcastStreamController._STATE_ADDSTREAM = 8;
  const _SyncBroadcastStreamController$ = dart.generic(function(T) {
    class _SyncBroadcastStreamController extends _BroadcastStreamController$(T) {
      _SyncBroadcastStreamController(onListen, onCancel) {
        super._BroadcastStreamController(onListen, onCancel);
      }
      [_sendData](data) {
        dart.as(data, T);
        if (dart.notNull(this[_isEmpty]))
          return;
        if (dart.notNull(this[_hasOneListener])) {
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BroadcastStreamController._STATE_FIRING);
          let subscription = dart.as(this[_next], _BroadcastSubscription);
          subscription[_add](data);
          this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_BroadcastStreamController._STATE_FIRING);
          if (dart.notNull(this[_isEmpty])) {
            this[_callOnCancel]();
          }
          return;
        }
        this[_forEachListener](dart.fn(subscription => {
          dart.as(subscription, _BufferingStreamSubscription$(T));
          subscription[_add](data);
        }, dart.dynamic, [_BufferingStreamSubscription$(T)]));
      }
      [_sendError](error, stackTrace) {
        if (dart.notNull(this[_isEmpty]))
          return;
        this[_forEachListener](dart.fn(subscription => {
          dart.as(subscription, _BufferingStreamSubscription$(T));
          subscription[_addError](error, stackTrace);
        }, dart.dynamic, [_BufferingStreamSubscription$(T)]));
      }
      [_sendDone]() {
        if (!dart.notNull(this[_isEmpty])) {
          this[_forEachListener](dart.as(dart.fn(subscription => {
            dart.as(subscription, _BroadcastSubscription$(T));
            subscription[_close]();
          }, dart.dynamic, [_BroadcastSubscription$(T)]), __CastType2$(T)));
        } else {
          dart.assert(this[_doneFuture] != null);
          dart.assert(this[_doneFuture][_mayComplete]);
          this[_doneFuture][_asyncComplete](null);
        }
      }
    }
    dart.setSignature(_SyncBroadcastStreamController, {
      constructors: () => ({_SyncBroadcastStreamController: [_SyncBroadcastStreamController$(T), [dart.functionType(dart.void, []), dart.functionType(dart.void, [])]]}),
      methods: () => ({
        [_sendData]: [dart.void, [T]],
        [_sendError]: [dart.void, [core.Object, core.StackTrace]],
        [_sendDone]: [dart.void, []]
      })
    });
    return _SyncBroadcastStreamController;
  });
  let _SyncBroadcastStreamController = _SyncBroadcastStreamController$();
  const _AsyncBroadcastStreamController$ = dart.generic(function(T) {
    class _AsyncBroadcastStreamController extends _BroadcastStreamController$(T) {
      _AsyncBroadcastStreamController(onListen, onCancel) {
        super._BroadcastStreamController(onListen, onCancel);
      }
      [_sendData](data) {
        dart.as(data, T);
        for (let link = this[_next]; !dart.notNull(core.identical(link, this)); link = link[_next]) {
          let subscription = dart.as(link, _BroadcastSubscription$(T));
          subscription[_addPending](new _DelayedData(data));
        }
      }
      [_sendError](error, stackTrace) {
        for (let link = this[_next]; !dart.notNull(core.identical(link, this)); link = link[_next]) {
          let subscription = dart.as(link, _BroadcastSubscription$(T));
          subscription[_addPending](new _DelayedError(error, stackTrace));
        }
      }
      [_sendDone]() {
        if (!dart.notNull(this[_isEmpty])) {
          for (let link = this[_next]; !dart.notNull(core.identical(link, this)); link = link[_next]) {
            let subscription = dart.as(link, _BroadcastSubscription$(T));
            subscription[_addPending](dart.const(new _DelayedDone()));
          }
        } else {
          dart.assert(this[_doneFuture] != null);
          dart.assert(this[_doneFuture][_mayComplete]);
          this[_doneFuture][_asyncComplete](null);
        }
      }
    }
    dart.setSignature(_AsyncBroadcastStreamController, {
      constructors: () => ({_AsyncBroadcastStreamController: [_AsyncBroadcastStreamController$(T), [dart.functionType(dart.void, []), dart.functionType(dart.void, [])]]}),
      methods: () => ({
        [_sendData]: [dart.void, [T]],
        [_sendError]: [dart.void, [core.Object, core.StackTrace]],
        [_sendDone]: [dart.void, []]
      })
    });
    return _AsyncBroadcastStreamController;
  });
  let _AsyncBroadcastStreamController = _AsyncBroadcastStreamController$();
  const _addPendingEvent = Symbol('_addPendingEvent');
  const _AsBroadcastStreamController$ = dart.generic(function(T) {
    class _AsBroadcastStreamController extends _SyncBroadcastStreamController$(T) {
      _AsBroadcastStreamController(onListen, onCancel) {
        this[_pending] = null;
        super._SyncBroadcastStreamController(onListen, onCancel);
      }
      get [_hasPending]() {
        return this[_pending] != null && !dart.notNull(this[_pending].isEmpty);
      }
      [_addPendingEvent](event) {
        if (this[_pending] == null) {
          this[_pending] = new _StreamImplEvents();
        }
        this[_pending].add(event);
      }
      add(data) {
        dart.as(data, T);
        if (!dart.notNull(this.isClosed) && dart.notNull(this[_isFiring])) {
          this[_addPendingEvent](new (_DelayedData$(T))(data));
          return;
        }
        super.add(data);
        while (dart.notNull(this[_hasPending])) {
          this[_pending].handleNext(this);
        }
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        if (!dart.notNull(this.isClosed) && dart.notNull(this[_isFiring])) {
          this[_addPendingEvent](new _DelayedError(error, stackTrace));
          return;
        }
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_addEventError]());
        this[_sendError](error, stackTrace);
        while (dart.notNull(this[_hasPending])) {
          this[_pending].handleNext(this);
        }
      }
      close() {
        if (!dart.notNull(this.isClosed) && dart.notNull(this[_isFiring])) {
          this[_addPendingEvent](dart.const(new _DelayedDone()));
          this[_state] = dart.notNull(this[_state]) | dart.notNull(_BroadcastStreamController._STATE_CLOSED);
          return super.done;
        }
        let result = super.close();
        dart.assert(!dart.notNull(this[_hasPending]));
        return result;
      }
      [_callOnCancel]() {
        if (dart.notNull(this[_hasPending])) {
          this[_pending].clear();
          this[_pending] = null;
        }
        super[_callOnCancel]();
      }
    }
    _AsBroadcastStreamController[dart.implements] = () => [_EventDispatch$(T)];
    dart.setSignature(_AsBroadcastStreamController, {
      constructors: () => ({_AsBroadcastStreamController: [_AsBroadcastStreamController$(T), [dart.functionType(dart.void, []), dart.functionType(dart.void, [])]]}),
      methods: () => ({
        [_addPendingEvent]: [dart.void, [_DelayedEvent]],
        add: [dart.void, [T]]
      })
    });
    return _AsBroadcastStreamController;
  });
  let _AsBroadcastStreamController = _AsBroadcastStreamController$();
  const _pauseCount = Symbol('_pauseCount');
  const _resume = Symbol('_resume');
  const _DoneSubscription$ = dart.generic(function(T) {
    class _DoneSubscription extends core.Object {
      _DoneSubscription() {
        this[_pauseCount] = 0;
      }
      onData(handleData) {
        dart.as(handleData, dart.functionType(dart.void, [T]));
      }
      onError(handleError) {}
      onDone(handleDone) {
        dart.as(handleDone, dart.functionType(dart.void, []));
      }
      pause(resumeSignal) {
        if (resumeSignal === void 0)
          resumeSignal = null;
        if (resumeSignal != null)
          resumeSignal.then(dart.bind(this, _resume));
        this[_pauseCount] = dart.notNull(this[_pauseCount]) + 1;
      }
      resume() {
        this[_resume](null);
      }
      [_resume](_) {
        if (dart.notNull(this[_pauseCount]) > 0) {
          this[_pauseCount] = dart.notNull(this[_pauseCount]) - 1;
        }
      }
      cancel() {
        return new _Future.immediate(null);
      }
      get isPaused() {
        return dart.notNull(this[_pauseCount]) > 0;
      }
      asFuture(value) {
        if (value === void 0)
          value = null;
        return new _Future();
      }
    }
    _DoneSubscription[dart.implements] = () => [StreamSubscription$(T)];
    dart.setSignature(_DoneSubscription, {
      methods: () => ({
        onData: [dart.void, [dart.functionType(dart.void, [T])]],
        onError: [dart.void, [core.Function]],
        onDone: [dart.void, [dart.functionType(dart.void, [])]],
        pause: [dart.void, [], [Future]],
        resume: [dart.void, []],
        [_resume]: [dart.void, [dart.dynamic]],
        cancel: [Future, []],
        asFuture: [Future, [], [core.Object]]
      })
    });
    return _DoneSubscription;
  });
  let _DoneSubscription = _DoneSubscription$();
  const __CastType2$ = dart.generic(function(T) {
    const __CastType2 = dart.typedef('__CastType2', () => dart.functionType(dart.void, [_BufferingStreamSubscription$(T)]));
    return __CastType2;
  });
  let __CastType2 = __CastType2$();
  class DeferredLibrary extends core.Object {
    DeferredLibrary(libraryName, opts) {
      let uri = opts && 'uri' in opts ? opts.uri : null;
      this.libraryName = libraryName;
      this.uri = uri;
    }
    load() {
      dart.throw('DeferredLibrary not supported. ' + 'please use the `import "lib.dart" deferred as lib` syntax.');
    }
  }
  dart.setSignature(DeferredLibrary, {
    constructors: () => ({DeferredLibrary: [DeferredLibrary, [core.String], {uri: core.String}]}),
    methods: () => ({load: [Future$(core.Null), []]})
  });
  DeferredLibrary[dart.metadata] = () => [dart.const(new core.Deprecated("Dart sdk v. 1.8"))];
  const _s = Symbol('_s');
  class DeferredLoadException extends core.Object {
    DeferredLoadException(s) {
      this[_s] = s;
    }
    toString() {
      return `DeferredLoadException: '${this[_s]}'`;
    }
  }
  DeferredLoadException[dart.implements] = () => [core.Exception];
  dart.setSignature(DeferredLoadException, {
    constructors: () => ({DeferredLoadException: [DeferredLoadException, [core.String]]})
  });
  const _completeWithValue = Symbol('_completeWithValue');
  const Future$ = dart.generic(function(T) {
    class Future extends core.Object {
      static new(computation) {
        let result = new (_Future$(T))();
        Timer.run(dart.fn(() => {
          try {
            result[_complete](computation());
          } catch (e) {
            let s = dart.stackTrace(e);
            _completeWithErrorCallback(result, e, s);
          }

        }));
        return dart.as(result, Future$(T));
      }
      static microtask(computation) {
        let result = new (_Future$(T))();
        scheduleMicrotask(dart.fn(() => {
          try {
            result[_complete](computation());
          } catch (e) {
            let s = dart.stackTrace(e);
            _completeWithErrorCallback(result, e, s);
          }

        }));
        return dart.as(result, Future$(T));
      }
      static sync(computation) {
        try {
          let result = computation();
          return Future$(T).value(result);
        } catch (error) {
          let stackTrace = dart.stackTrace(error);
          return Future$(T).error(error, stackTrace);
        }

      }
      static value(value) {
        if (value === void 0)
          value = null;
        return new (_Future$(T)).immediate(value);
      }
      static error(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        error = _nonNullError(error);
        if (!dart.notNull(core.identical(Zone.current, _ROOT_ZONE))) {
          let replacement = Zone.current.errorCallback(error, stackTrace);
          if (replacement != null) {
            error = _nonNullError(replacement.error);
            stackTrace = replacement.stackTrace;
          }
        }
        return new (_Future$(T)).immediateError(error, stackTrace);
      }
      static delayed(duration, computation) {
        if (computation === void 0)
          computation = null;
        let result = new (_Future$(T))();
        Timer.new(duration, dart.fn(() => {
          try {
            result[_complete](computation == null ? null : computation());
          } catch (e) {
            let s = dart.stackTrace(e);
            _completeWithErrorCallback(result, e, s);
          }

        }));
        return dart.as(result, Future$(T));
      }
      static wait(futures, opts) {
        let eagerError = opts && 'eagerError' in opts ? opts.eagerError : false;
        let cleanUp = opts && 'cleanUp' in opts ? opts.cleanUp : null;
        dart.as(cleanUp, dart.functionType(dart.void, [dart.dynamic]));
        let result = new (_Future$(core.List))();
        let values = null;
        let remaining = 0;
        let error = null;
        let stackTrace = null;
        function handleError(theError, theStackTrace) {
          remaining = dart.notNull(remaining) - 1;
          if (values != null) {
            if (cleanUp != null) {
              for (let value2 of values) {
                if (value2 != null) {
                  Future$().sync(dart.fn(() => {
                    dart.dcall(cleanUp, value2);
                  }));
                }
              }
            }
            values = null;
            if (remaining == 0 || dart.notNull(eagerError)) {
              result[_completeError](theError, dart.as(theStackTrace, core.StackTrace));
            } else {
              error = theError;
              stackTrace = dart.as(theStackTrace, core.StackTrace);
            }
          } else if (remaining == 0 && !dart.notNull(eagerError)) {
            result[_completeError](error, stackTrace);
          }
        }
        dart.fn(handleError, dart.void, [dart.dynamic, dart.dynamic]);
        for (let future of futures) {
          let pos = remaining;
          remaining = dart.notNull(pos) + 1;
          future.then(dart.fn(value => {
            remaining = dart.notNull(remaining) - 1;
            if (values != null) {
              values[dartx.set](pos, value);
              if (remaining == 0) {
                result[_completeWithValue](values);
              }
            } else {
              if (cleanUp != null && value != null) {
                Future$().sync(dart.fn(() => {
                  dart.dcall(cleanUp, value);
                }));
              }
              if (remaining == 0 && !dart.notNull(eagerError)) {
                result[_completeError](error, stackTrace);
              }
            }
          }, dart.dynamic, [core.Object]), {onError: handleError});
        }
        if (remaining == 0) {
          return Future$(core.List).value(dart.const([]));
        }
        values = core.List.new(remaining);
        return result;
      }
      static forEach(input, f) {
        dart.as(f, dart.functionType(dart.dynamic, [dart.dynamic]));
        let iterator = input[dartx.iterator];
        return Future$().doWhile(dart.fn(() => {
          if (!dart.notNull(iterator.moveNext()))
            return false;
          return Future$().sync(dart.fn(() => dart.dcall(f, iterator.current))).then(dart.fn(_ => true, core.bool, [dart.dynamic]));
        }));
      }
      static doWhile(f) {
        dart.as(f, dart.functionType(dart.dynamic, []));
        let doneSignal = new _Future();
        let nextIteration = null;
        nextIteration = Zone.current.bindUnaryCallback(dart.fn(keepGoing => {
          if (dart.notNull(keepGoing)) {
            Future$().sync(f).then(dart.as(nextIteration, __CastType4), {onError: dart.bind(doneSignal, _completeError)});
          } else {
            doneSignal[_complete](null);
          }
        }, dart.dynamic, [core.bool]), {runGuarded: true});
        dart.dcall(nextIteration, true);
        return doneSignal;
      }
    }
    dart.setSignature(Future, {
      constructors: () => ({
        new: [Future$(T), [dart.functionType(dart.dynamic, [])]],
        microtask: [Future$(T), [dart.functionType(dart.dynamic, [])]],
        sync: [Future$(T), [dart.functionType(dart.dynamic, [])]],
        value: [Future$(T), [], [dart.dynamic]],
        error: [Future$(T), [core.Object], [core.StackTrace]],
        delayed: [Future$(T), [core.Duration], [dart.functionType(T, [])]]
      }),
      statics: () => ({
        wait: [Future$(core.List), [core.Iterable$(Future$())], {eagerError: core.bool, cleanUp: dart.functionType(dart.void, [dart.dynamic])}],
        forEach: [Future$(), [core.Iterable, dart.functionType(dart.dynamic, [dart.dynamic])]],
        doWhile: [Future$(), [dart.functionType(dart.dynamic, [])]]
      }),
      names: ['wait', 'forEach', 'doWhile']
    });
    return Future;
  });
  let Future = Future$();
  dart.defineLazyProperties(Future, {
    get _nullFuture() {
      return new _Future.immediate(null);
    }
  });
  class TimeoutException extends core.Object {
    TimeoutException(message, duration) {
      if (duration === void 0)
        duration = null;
      this.message = message;
      this.duration = duration;
    }
    toString() {
      let result = "TimeoutException";
      if (this.duration != null)
        result = `TimeoutException after ${this.duration}`;
      if (this.message != null)
        result = `${result}: ${this.message}`;
      return result;
    }
  }
  TimeoutException[dart.implements] = () => [core.Exception];
  dart.setSignature(TimeoutException, {
    constructors: () => ({TimeoutException: [TimeoutException, [core.String], [core.Duration]]})
  });
  const Completer$ = dart.generic(function(T) {
    class Completer extends core.Object {
      static new() {
        return new (_AsyncCompleter$(T))();
      }
      static sync() {
        return new (_SyncCompleter$(T))();
      }
    }
    dart.setSignature(Completer, {
      constructors: () => ({
        new: [Completer$(T), []],
        sync: [Completer$(T), []]
      })
    });
    return Completer;
  });
  let Completer = Completer$();
  function _completeWithErrorCallback(result, error, stackTrace) {
    let replacement = Zone.current.errorCallback(error, dart.as(stackTrace, core.StackTrace));
    if (replacement != null) {
      error = _nonNullError(replacement.error);
      stackTrace = replacement.stackTrace;
    }
    result[_completeError](error, dart.as(stackTrace, core.StackTrace));
  }
  dart.fn(_completeWithErrorCallback, () => dart.definiteFunctionType(dart.void, [_Future, dart.dynamic, dart.dynamic]));
  function _nonNullError(error) {
    return error != null ? error : new core.NullThrownError();
  }
  dart.fn(_nonNullError, core.Object, [core.Object]);
  const __CastType4 = dart.typedef('__CastType4', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const _FutureOnValue$ = dart.generic(function(T) {
    const _FutureOnValue = dart.typedef('_FutureOnValue', () => dart.functionType(dart.dynamic, [T]));
    return _FutureOnValue;
  });
  let _FutureOnValue = _FutureOnValue$();
  const _FutureErrorTest = dart.typedef('_FutureErrorTest', () => dart.functionType(core.bool, [dart.dynamic]));
  const _FutureAction = dart.typedef('_FutureAction', () => dart.functionType(dart.dynamic, []));
  const _Completer$ = dart.generic(function(T) {
    class _Completer extends core.Object {
      _Completer() {
        this.future = new (_Future$(T))();
      }
      completeError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        error = _nonNullError(error);
        if (!dart.notNull(this.future[_mayComplete]))
          dart.throw(new core.StateError("Future already completed"));
        let replacement = Zone.current.errorCallback(error, stackTrace);
        if (replacement != null) {
          error = _nonNullError(replacement.error);
          stackTrace = replacement.stackTrace;
        }
        this[_completeError](error, stackTrace);
      }
      get isCompleted() {
        return !dart.notNull(this.future[_mayComplete]);
      }
    }
    _Completer[dart.implements] = () => [Completer$(T)];
    dart.setSignature(_Completer, {
      methods: () => ({completeError: [dart.void, [core.Object], [core.StackTrace]]})
    });
    return _Completer;
  });
  let _Completer = _Completer$();
  const _asyncCompleteError = Symbol('_asyncCompleteError');
  const _AsyncCompleter$ = dart.generic(function(T) {
    class _AsyncCompleter extends _Completer$(T) {
      _AsyncCompleter() {
        super._Completer();
      }
      complete(value) {
        if (value === void 0)
          value = null;
        if (!dart.notNull(this.future[_mayComplete]))
          dart.throw(new core.StateError("Future already completed"));
        this.future[_asyncComplete](value);
      }
      [_completeError](error, stackTrace) {
        this.future[_asyncCompleteError](error, stackTrace);
      }
    }
    dart.setSignature(_AsyncCompleter, {
      methods: () => ({
        complete: [dart.void, [], [dart.dynamic]],
        [_completeError]: [dart.void, [core.Object, core.StackTrace]]
      })
    });
    return _AsyncCompleter;
  });
  let _AsyncCompleter = _AsyncCompleter$();
  const _SyncCompleter$ = dart.generic(function(T) {
    class _SyncCompleter extends _Completer$(T) {
      _SyncCompleter() {
        super._Completer();
      }
      complete(value) {
        if (value === void 0)
          value = null;
        if (!dart.notNull(this.future[_mayComplete]))
          dart.throw(new core.StateError("Future already completed"));
        this.future[_complete](value);
      }
      [_completeError](error, stackTrace) {
        this.future[_completeError](error, stackTrace);
      }
    }
    dart.setSignature(_SyncCompleter, {
      methods: () => ({
        complete: [dart.void, [], [dart.dynamic]],
        [_completeError]: [dart.void, [core.Object, core.StackTrace]]
      })
    });
    return _SyncCompleter;
  });
  let _SyncCompleter = _SyncCompleter$();
  const _nextListener = Symbol('_nextListener');
  const _onValue = Symbol('_onValue');
  const _errorTest = Symbol('_errorTest');
  const _whenCompleteAction = Symbol('_whenCompleteAction');
  class _FutureListener extends core.Object {
    then(result, onValue, errorCallback) {
      this.result = result;
      this.callback = onValue;
      this.errorCallback = errorCallback;
      this.state = errorCallback == null ? _FutureListener.STATE_THEN : _FutureListener.STATE_THEN_ONERROR;
      this[_nextListener] = null;
    }
    catchError(result, errorCallback, test) {
      this.result = result;
      this.errorCallback = errorCallback;
      this.callback = test;
      this.state = test == null ? _FutureListener.STATE_CATCHERROR : _FutureListener.STATE_CATCHERROR_TEST;
      this[_nextListener] = null;
    }
    whenComplete(result, onComplete) {
      this.result = result;
      this.callback = onComplete;
      this.errorCallback = null;
      this.state = _FutureListener.STATE_WHENCOMPLETE;
      this[_nextListener] = null;
    }
    chain(result) {
      this.result = result;
      this.callback = null;
      this.errorCallback = null;
      this.state = _FutureListener.STATE_CHAIN;
      this[_nextListener] = null;
    }
    get [_zone]() {
      return this.result[_zone];
    }
    get handlesValue() {
      return (dart.notNull(this.state) & dart.notNull(_FutureListener.MASK_VALUE)) != 0;
    }
    get handlesError() {
      return (dart.notNull(this.state) & dart.notNull(_FutureListener.MASK_ERROR)) != 0;
    }
    get hasErrorTest() {
      return this.state == _FutureListener.STATE_CATCHERROR_TEST;
    }
    get handlesComplete() {
      return this.state == _FutureListener.STATE_WHENCOMPLETE;
    }
    get [_onValue]() {
      dart.assert(this.handlesValue);
      return dart.as(this.callback, _FutureOnValue);
    }
    get [_onError]() {
      return this.errorCallback;
    }
    get [_errorTest]() {
      dart.assert(this.hasErrorTest);
      return dart.as(this.callback, _FutureErrorTest);
    }
    get [_whenCompleteAction]() {
      dart.assert(this.handlesComplete);
      return dart.as(this.callback, _FutureAction);
    }
  }
  dart.defineNamedConstructor(_FutureListener, 'then');
  dart.defineNamedConstructor(_FutureListener, 'catchError');
  dart.defineNamedConstructor(_FutureListener, 'whenComplete');
  dart.defineNamedConstructor(_FutureListener, 'chain');
  dart.setSignature(_FutureListener, {
    constructors: () => ({
      then: [_FutureListener, [_Future, _FutureOnValue, core.Function]],
      catchError: [_FutureListener, [_Future, core.Function, _FutureErrorTest]],
      whenComplete: [_FutureListener, [_Future, _FutureAction]],
      chain: [_FutureListener, [_Future]]
    })
  });
  _FutureListener.MASK_VALUE = 1;
  _FutureListener.MASK_ERROR = 2;
  _FutureListener.MASK_TEST_ERROR = 4;
  _FutureListener.MASK_WHENCOMPLETE = 8;
  _FutureListener.STATE_CHAIN = 0;
  _FutureListener.STATE_THEN = _FutureListener.MASK_VALUE;
  _FutureListener.STATE_THEN_ONERROR = dart.notNull(_FutureListener.MASK_VALUE) | dart.notNull(_FutureListener.MASK_ERROR);
  _FutureListener.STATE_CATCHERROR = _FutureListener.MASK_ERROR;
  _FutureListener.STATE_CATCHERROR_TEST = dart.notNull(_FutureListener.MASK_ERROR) | dart.notNull(_FutureListener.MASK_TEST_ERROR);
  _FutureListener.STATE_WHENCOMPLETE = _FutureListener.MASK_WHENCOMPLETE;
  const _resultOrListeners = Symbol('_resultOrListeners');
  const _isChained = Symbol('_isChained');
  const _isComplete = Symbol('_isComplete');
  const _hasValue = Symbol('_hasValue');
  const _hasError = Symbol('_hasError');
  const _markPendingCompletion = Symbol('_markPendingCompletion');
  const _value = Symbol('_value');
  const _error = Symbol('_error');
  const _setValue = Symbol('_setValue');
  const _setErrorObject = Symbol('_setErrorObject');
  const _setError = Symbol('_setError');
  const _removeListeners = Symbol('_removeListeners');
  const _Future$ = dart.generic(function(T) {
    class _Future extends core.Object {
      _Future() {
        this[_zone] = Zone.current;
        this[_state] = _Future$()._INCOMPLETE;
        this[_resultOrListeners] = null;
      }
      immediate(value) {
        this[_zone] = Zone.current;
        this[_state] = _Future$()._INCOMPLETE;
        this[_resultOrListeners] = null;
        this[_asyncComplete](value);
      }
      immediateError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        this[_zone] = Zone.current;
        this[_state] = _Future$()._INCOMPLETE;
        this[_resultOrListeners] = null;
        this[_asyncCompleteError](error, stackTrace);
      }
      get [_mayComplete]() {
        return this[_state] == _Future$()._INCOMPLETE;
      }
      get [_isChained]() {
        return this[_state] == _Future$()._CHAINED;
      }
      get [_isComplete]() {
        return dart.notNull(this[_state]) >= dart.notNull(_Future$()._VALUE);
      }
      get [_hasValue]() {
        return this[_state] == _Future$()._VALUE;
      }
      get [_hasError]() {
        return this[_state] == _Future$()._ERROR;
      }
      set [_isChained](value) {
        if (dart.notNull(value)) {
          dart.assert(!dart.notNull(this[_isComplete]));
          this[_state] = _Future$()._CHAINED;
        } else {
          dart.assert(this[_isChained]);
          this[_state] = _Future$()._INCOMPLETE;
        }
      }
      then(f, opts) {
        dart.as(f, dart.functionType(dart.dynamic, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let result = new (_Future$())();
        if (!dart.notNull(core.identical(result[_zone], _ROOT_ZONE))) {
          f = dart.as(result[_zone].registerUnaryCallback(f), __CastType6$(T));
          if (onError != null) {
            onError = _registerErrorHandler(onError, result[_zone]);
          }
        }
        this[_addListener](new _FutureListener.then(result, f, onError));
        return result;
      }
      catchError(onError, opts) {
        let test = opts && 'test' in opts ? opts.test : null;
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        let result = new (_Future$())();
        if (!dart.notNull(core.identical(result[_zone], _ROOT_ZONE))) {
          onError = _registerErrorHandler(onError, result[_zone]);
          if (test != null)
            test = dart.as(result[_zone].registerUnaryCallback(test), __CastType8);
        }
        this[_addListener](new _FutureListener.catchError(result, onError, test));
        return result;
      }
      whenComplete(action) {
        dart.as(action, dart.functionType(dart.dynamic, []));
        let result = new (_Future$(T))();
        if (!dart.notNull(core.identical(result[_zone], _ROOT_ZONE))) {
          action = result[_zone].registerCallback(action);
        }
        this[_addListener](new _FutureListener.whenComplete(result, action));
        return dart.as(result, Future$(T));
      }
      asStream() {
        return Stream$(T).fromFuture(this);
      }
      [_markPendingCompletion]() {
        if (!dart.notNull(this[_mayComplete]))
          dart.throw(new core.StateError("Future already completed"));
        this[_state] = _Future$()._PENDING_COMPLETE;
      }
      get [_value]() {
        dart.assert(dart.notNull(this[_isComplete]) && dart.notNull(this[_hasValue]));
        return dart.as(this[_resultOrListeners], T);
      }
      get [_error]() {
        dart.assert(dart.notNull(this[_isComplete]) && dart.notNull(this[_hasError]));
        return dart.as(this[_resultOrListeners], AsyncError);
      }
      [_setValue](value) {
        dart.as(value, T);
        dart.assert(!dart.notNull(this[_isComplete]));
        this[_state] = _Future$()._VALUE;
        this[_resultOrListeners] = value;
      }
      [_setErrorObject](error) {
        dart.assert(!dart.notNull(this[_isComplete]));
        this[_state] = _Future$()._ERROR;
        this[_resultOrListeners] = error;
      }
      [_setError](error, stackTrace) {
        this[_setErrorObject](new AsyncError(error, stackTrace));
      }
      [_addListener](listener) {
        dart.assert(listener[_nextListener] == null);
        if (dart.notNull(this[_isComplete])) {
          this[_zone].scheduleMicrotask(dart.fn((() => {
            _Future$()._propagateToListeners(this, listener);
          }).bind(this)));
        } else {
          listener[_nextListener] = dart.as(this[_resultOrListeners], _FutureListener);
          this[_resultOrListeners] = listener;
        }
      }
      [_removeListeners]() {
        dart.assert(!dart.notNull(this[_isComplete]));
        let current = dart.as(this[_resultOrListeners], _FutureListener);
        this[_resultOrListeners] = null;
        let prev = null;
        while (current != null) {
          let next = current[_nextListener];
          current[_nextListener] = prev;
          prev = current;
          current = next;
        }
        return prev;
      }
      static _chainForeignFuture(source, target) {
        dart.assert(!dart.notNull(target[_isComplete]));
        dart.assert(!dart.is(source, _Future$()));
        target[_isChained] = true;
        source.then(dart.fn(value => {
          dart.assert(target[_isChained]);
          target[_completeWithValue](value);
        }), {onError: dart.fn((error, stackTrace) => {
            if (stackTrace === void 0)
              stackTrace = null;
            dart.assert(target[_isChained]);
            target[_completeError](error, dart.as(stackTrace, core.StackTrace));
          }, dart.dynamic, [dart.dynamic], [dart.dynamic])});
      }
      static _chainCoreFuture(source, target) {
        dart.assert(!dart.notNull(target[_isComplete]));
        dart.assert(dart.is(source, _Future$()));
        target[_isChained] = true;
        let listener = new _FutureListener.chain(target);
        if (dart.notNull(source[_isComplete])) {
          _Future$()._propagateToListeners(source, listener);
        } else {
          source[_addListener](listener);
        }
      }
      [_complete](value) {
        dart.assert(!dart.notNull(this[_isComplete]));
        if (dart.is(value, Future)) {
          if (dart.is(value, _Future$())) {
            _Future$()._chainCoreFuture(dart.as(value, _Future$()), this);
          } else {
            _Future$()._chainForeignFuture(dart.as(value, Future), this);
          }
        } else {
          let listeners = this[_removeListeners]();
          this[_setValue](dart.as(value, T));
          _Future$()._propagateToListeners(this, listeners);
        }
      }
      [_completeWithValue](value) {
        dart.assert(!dart.notNull(this[_isComplete]));
        dart.assert(!dart.is(value, Future));
        let listeners = this[_removeListeners]();
        this[_setValue](dart.as(value, T));
        _Future$()._propagateToListeners(this, listeners);
      }
      [_completeError](error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        dart.assert(!dart.notNull(this[_isComplete]));
        let listeners = this[_removeListeners]();
        this[_setError](error, stackTrace);
        _Future$()._propagateToListeners(this, listeners);
      }
      [_asyncComplete](value) {
        dart.assert(!dart.notNull(this[_isComplete]));
        if (value == null) {
        } else if (dart.is(value, Future)) {
          let typedFuture = dart.as(value, Future$(T));
          if (dart.is(typedFuture, _Future$())) {
            let coreFuture = dart.as(typedFuture, _Future$(T));
            if (dart.notNull(coreFuture[_isComplete]) && dart.notNull(coreFuture[_hasError])) {
              this[_markPendingCompletion]();
              this[_zone].scheduleMicrotask(dart.fn((() => {
                _Future$()._chainCoreFuture(coreFuture, this);
              }).bind(this)));
            } else {
              _Future$()._chainCoreFuture(coreFuture, this);
            }
          } else {
            _Future$()._chainForeignFuture(typedFuture, this);
          }
          return;
        } else {
          let typedValue = dart.as(value, T);
        }
        this[_markPendingCompletion]();
        this[_zone].scheduleMicrotask(dart.fn((() => {
          this[_completeWithValue](value);
        }).bind(this)));
      }
      [_asyncCompleteError](error, stackTrace) {
        dart.assert(!dart.notNull(this[_isComplete]));
        this[_markPendingCompletion]();
        this[_zone].scheduleMicrotask(dart.fn((() => {
          this[_completeError](error, stackTrace);
        }).bind(this)));
      }
      static _propagateToListeners(source, listeners) {
        while (true) {
          dart.assert(source[_isComplete]);
          let hasError = source[_hasError];
          if (listeners == null) {
            if (dart.notNull(hasError)) {
              let asyncError = source[_error];
              source[_zone].handleUncaughtError(asyncError.error, asyncError.stackTrace);
            }
            return;
          }
          while (listeners[_nextListener] != null) {
            let listener = listeners;
            listeners = listener[_nextListener];
            listener[_nextListener] = null;
            _Future$()._propagateToListeners(source, listener);
          }
          let listener = listeners;
          let listenerHasValue = true;
          let sourceValue = dart.notNull(hasError) ? null : source[_value];
          let listenerValueOrError = sourceValue;
          let isPropagationAborted = false;
          if (dart.notNull(hasError) || dart.notNull(listener.handlesValue) || dart.notNull(listener.handlesComplete)) {
            let zone = listener[_zone];
            if (dart.notNull(hasError) && !dart.notNull(source[_zone].inSameErrorZone(zone))) {
              let asyncError = source[_error];
              source[_zone].handleUncaughtError(asyncError.error, asyncError.stackTrace);
              return;
            }
            let oldZone = null;
            if (!dart.notNull(core.identical(Zone.current, zone))) {
              oldZone = Zone._enter(zone);
            }
            function handleValueCallback() {
              try {
                listenerValueOrError = zone.runUnary(listener[_onValue], sourceValue);
                return true;
              } catch (e) {
                let s = dart.stackTrace(e);
                listenerValueOrError = new AsyncError(e, s);
                return false;
              }

            }
            dart.fn(handleValueCallback, core.bool, []);
            function handleError() {
              let asyncError = source[_error];
              let matchesTest = true;
              if (dart.notNull(listener.hasErrorTest)) {
                let test = listener[_errorTest];
                try {
                  matchesTest = dart.as(zone.runUnary(test, asyncError.error), core.bool);
                } catch (e) {
                  let s = dart.stackTrace(e);
                  listenerValueOrError = dart.notNull(core.identical(asyncError.error, e)) ? asyncError : new AsyncError(e, s);
                  listenerHasValue = false;
                  return;
                }

              }
              let errorCallback = listener[_onError];
              if (dart.notNull(matchesTest) && errorCallback != null) {
                try {
                  if (dart.is(errorCallback, ZoneBinaryCallback)) {
                    listenerValueOrError = zone.runBinary(errorCallback, asyncError.error, asyncError.stackTrace);
                  } else {
                    listenerValueOrError = zone.runUnary(dart.as(errorCallback, __CastType10), asyncError.error);
                  }
                } catch (e) {
                  let s = dart.stackTrace(e);
                  listenerValueOrError = dart.notNull(core.identical(asyncError.error, e)) ? asyncError : new AsyncError(e, s);
                  listenerHasValue = false;
                  return;
                }

                listenerHasValue = true;
              } else {
                listenerValueOrError = asyncError;
                listenerHasValue = false;
              }
            }
            dart.fn(handleError, dart.void, []);
            function handleWhenCompleteCallback() {
              let completeResult = null;
              try {
                completeResult = zone.run(listener[_whenCompleteAction]);
              } catch (e) {
                let s = dart.stackTrace(e);
                if (dart.notNull(hasError) && dart.notNull(core.identical(source[_error].error, e))) {
                  listenerValueOrError = source[_error];
                } else {
                  listenerValueOrError = new AsyncError(e, s);
                }
                listenerHasValue = false;
                return;
              }

              if (dart.is(completeResult, Future)) {
                let result = listener.result;
                result[_isChained] = true;
                isPropagationAborted = true;
                dart.dsend(completeResult, 'then', dart.fn(ignored => {
                  _Future$()._propagateToListeners(source, new _FutureListener.chain(result));
                }), {onError: dart.fn((error, stackTrace) => {
                    if (stackTrace === void 0)
                      stackTrace = null;
                    if (!dart.is(completeResult, _Future$())) {
                      completeResult = new (_Future$())();
                      dart.dsend(completeResult, _setError, error, stackTrace);
                    }
                    _Future$()._propagateToListeners(dart.as(completeResult, _Future$()), new _FutureListener.chain(result));
                  }, dart.dynamic, [dart.dynamic], [dart.dynamic])});
              }
            }
            dart.fn(handleWhenCompleteCallback, dart.void, []);
            if (!dart.notNull(hasError)) {
              if (dart.notNull(listener.handlesValue)) {
                listenerHasValue = handleValueCallback();
              }
            } else {
              handleError();
            }
            if (dart.notNull(listener.handlesComplete)) {
              handleWhenCompleteCallback();
            }
            if (oldZone != null)
              Zone._leave(oldZone);
            if (dart.notNull(isPropagationAborted))
              return;
            if (dart.notNull(listenerHasValue) && !dart.notNull(core.identical(sourceValue, listenerValueOrError)) && dart.is(listenerValueOrError, Future)) {
              let chainSource = dart.as(listenerValueOrError, Future);
              let result = listener.result;
              if (dart.is(chainSource, _Future$())) {
                if (dart.notNull(chainSource[_isComplete])) {
                  result[_isChained] = true;
                  source = chainSource;
                  listeners = new _FutureListener.chain(result);
                  continue;
                } else {
                  _Future$()._chainCoreFuture(chainSource, result);
                }
              } else {
                _Future$()._chainForeignFuture(chainSource, result);
              }
              return;
            }
          }
          let result = listener.result;
          listeners = result[_removeListeners]();
          if (dart.notNull(listenerHasValue)) {
            result[_setValue](listenerValueOrError);
          } else {
            let asyncError = dart.as(listenerValueOrError, AsyncError);
            result[_setErrorObject](asyncError);
          }
          source = result;
        }
      }
      timeout(timeLimit, opts) {
        let onTimeout = opts && 'onTimeout' in opts ? opts.onTimeout : null;
        dart.as(onTimeout, dart.functionType(dart.dynamic, []));
        if (dart.notNull(this[_isComplete]))
          return new (_Future$()).immediate(this);
        let result = new (_Future$())();
        let timer = null;
        if (onTimeout == null) {
          timer = Timer.new(timeLimit, dart.fn(() => {
            result[_completeError](new TimeoutException("Future not completed", timeLimit));
          }));
        } else {
          let zone = Zone.current;
          onTimeout = zone.registerCallback(onTimeout);
          timer = Timer.new(timeLimit, dart.fn(() => {
            try {
              result[_complete](zone.run(onTimeout));
            } catch (e) {
              let s = dart.stackTrace(e);
              result[_completeError](e, s);
            }

          }));
        }
        this.then(dart.fn(v => {
          dart.as(v, T);
          if (dart.notNull(timer.isActive)) {
            timer.cancel();
            result[_completeWithValue](v);
          }
        }, dart.dynamic, [T]), {onError: dart.fn((e, s) => {
            if (dart.notNull(timer.isActive)) {
              timer.cancel();
              result[_completeError](e, dart.as(s, core.StackTrace));
            }
          })});
        return result;
      }
    }
    _Future[dart.implements] = () => [Future$(T)];
    dart.defineNamedConstructor(_Future, 'immediate');
    dart.defineNamedConstructor(_Future, 'immediateError');
    dart.setSignature(_Future, {
      constructors: () => ({
        _Future: [_Future$(T), []],
        immediate: [_Future$(T), [dart.dynamic]],
        immediateError: [_Future$(T), [dart.dynamic], [core.StackTrace]]
      }),
      methods: () => ({
        then: [Future, [dart.functionType(dart.dynamic, [T])], {onError: core.Function}],
        catchError: [Future, [core.Function], {test: dart.functionType(core.bool, [dart.dynamic])}],
        whenComplete: [Future$(T), [dart.functionType(dart.dynamic, [])]],
        asStream: [Stream$(T), []],
        [_markPendingCompletion]: [dart.void, []],
        [_setValue]: [dart.void, [T]],
        [_setErrorObject]: [dart.void, [AsyncError]],
        [_setError]: [dart.void, [core.Object, core.StackTrace]],
        [_addListener]: [dart.void, [_FutureListener]],
        [_removeListeners]: [_FutureListener, []],
        [_complete]: [dart.void, [dart.dynamic]],
        [_completeWithValue]: [dart.void, [dart.dynamic]],
        [_completeError]: [dart.void, [dart.dynamic], [core.StackTrace]],
        [_asyncComplete]: [dart.void, [dart.dynamic]],
        [_asyncCompleteError]: [dart.void, [dart.dynamic, core.StackTrace]],
        timeout: [Future, [core.Duration], {onTimeout: dart.functionType(dart.dynamic, [])}]
      }),
      statics: () => ({
        _chainForeignFuture: [dart.void, [Future, _Future$()]],
        _chainCoreFuture: [dart.void, [_Future$(), _Future$()]],
        _propagateToListeners: [dart.void, [_Future$(), _FutureListener]]
      }),
      names: ['_chainForeignFuture', '_chainCoreFuture', '_propagateToListeners']
    });
    return _Future;
  });
  let _Future = _Future$();
  _Future._INCOMPLETE = 0;
  _Future._PENDING_COMPLETE = 1;
  _Future._CHAINED = 2;
  _Future._VALUE = 4;
  _Future._ERROR = 8;
  const __CastType6$ = dart.generic(function(T) {
    const __CastType6 = dart.typedef('__CastType6', () => dart.functionType(dart.dynamic, [T]));
    return __CastType6;
  });
  let __CastType6 = __CastType6$();
  const __CastType8 = dart.typedef('__CastType8', () => dart.functionType(core.bool, [dart.dynamic]));
  const __CastType10 = dart.typedef('__CastType10', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const _AsyncCallback = dart.typedef('_AsyncCallback', () => dart.functionType(dart.void, []));
  class _AsyncCallbackEntry extends core.Object {
    _AsyncCallbackEntry(callback) {
      this.callback = callback;
      this.next = null;
    }
  }
  dart.setSignature(_AsyncCallbackEntry, {
    constructors: () => ({_AsyncCallbackEntry: [_AsyncCallbackEntry, [_AsyncCallback]]})
  });
  exports._nextCallback = null;
  exports._lastCallback = null;
  exports._lastPriorityCallback = null;
  exports._isInCallbackLoop = false;
  function _asyncRunCallbackLoop() {
    while (exports._nextCallback != null) {
      exports._lastPriorityCallback = null;
      let entry = exports._nextCallback;
      exports._nextCallback = entry.next;
      if (exports._nextCallback == null)
        exports._lastCallback = null;
      entry.callback();
    }
  }
  dart.fn(_asyncRunCallbackLoop, dart.void, []);
  function _asyncRunCallback() {
    exports._isInCallbackLoop = true;
    try {
      _asyncRunCallbackLoop();
    } finally {
      exports._lastPriorityCallback = null;
      exports._isInCallbackLoop = false;
      if (exports._nextCallback != null)
        _AsyncRun._scheduleImmediate(_asyncRunCallback);
    }
  }
  dart.fn(_asyncRunCallback, dart.void, []);
  function _scheduleAsyncCallback(callback) {
    if (exports._nextCallback == null) {
      exports._nextCallback = exports._lastCallback = new _AsyncCallbackEntry(dart.as(callback, _AsyncCallback));
      if (!dart.notNull(exports._isInCallbackLoop)) {
        _AsyncRun._scheduleImmediate(_asyncRunCallback);
      }
    } else {
      let newEntry = new _AsyncCallbackEntry(dart.as(callback, _AsyncCallback));
      exports._lastCallback.next = newEntry;
      exports._lastCallback = newEntry;
    }
  }
  dart.fn(_scheduleAsyncCallback, dart.void, [dart.dynamic]);
  function _schedulePriorityAsyncCallback(callback) {
    let entry = new _AsyncCallbackEntry(dart.as(callback, _AsyncCallback));
    if (exports._nextCallback == null) {
      _scheduleAsyncCallback(callback);
      exports._lastPriorityCallback = exports._lastCallback;
    } else if (exports._lastPriorityCallback == null) {
      entry.next = exports._nextCallback;
      exports._nextCallback = exports._lastPriorityCallback = entry;
    } else {
      entry.next = exports._lastPriorityCallback.next;
      exports._lastPriorityCallback.next = entry;
      exports._lastPriorityCallback = entry;
      if (entry.next == null) {
        exports._lastCallback = entry;
      }
    }
  }
  dart.fn(_schedulePriorityAsyncCallback, dart.void, [dart.dynamic]);
  function scheduleMicrotask(callback) {
    if (dart.notNull(core.identical(_ROOT_ZONE, Zone.current))) {
      _rootScheduleMicrotask(null, null, _ROOT_ZONE, callback);
      return;
    }
    Zone.current.scheduleMicrotask(Zone.current.bindCallback(callback, {runGuarded: true}));
  }
  dart.fn(scheduleMicrotask, dart.void, [dart.functionType(dart.void, [])]);
  class _AsyncRun extends core.Object {
    static _scheduleImmediate(callback) {
      dart.dcall(_AsyncRun.scheduleImmediateClosure, callback);
    }
    static _initializeScheduleImmediate() {
      if (self.scheduleImmediate != null) {
        return _AsyncRun._scheduleImmediateJsOverride;
      }
      if (self.MutationObserver != null && self.document != null) {
        let div = self.document.createElement("div");
        let span = self.document.createElement("span");
        let storedCallback = null;
        function internalCallback(_) {
          _isolate_helper.leaveJsAsync();
          let f = storedCallback;
          storedCallback = null;
          dart.dcall(f);
        }
        dart.fn(internalCallback);
        ;
        let observer = new self.MutationObserver(internalCallback);
        observer.observe(div, {childList: true});
        return dart.fn(callback => {
          dart.assert(storedCallback == null);
          _isolate_helper.enterJsAsync();
          storedCallback = callback;
          div.firstChild ? div.removeChild(span) : div.appendChild(span);
        }, dart.dynamic, [dart.functionType(dart.void, [])]);
      } else if (self.setImmediate != null) {
        return _AsyncRun._scheduleImmediateWithSetImmediate;
      }
      return _AsyncRun._scheduleImmediateWithTimer;
    }
    static _scheduleImmediateJsOverride(callback) {
      function internalCallback() {
        _isolate_helper.leaveJsAsync();
        callback();
      }
      dart.fn(internalCallback);
      ;
      _isolate_helper.enterJsAsync();
      self.scheduleImmediate(internalCallback);
    }
    static _scheduleImmediateWithSetImmediate(callback) {
      function internalCallback() {
        _isolate_helper.leaveJsAsync();
        callback();
      }
      dart.fn(internalCallback);
      ;
      _isolate_helper.enterJsAsync();
      self.setImmediate(internalCallback);
    }
    static _scheduleImmediateWithTimer(callback) {
      Timer._createTimer(core.Duration.ZERO, callback);
    }
  }
  dart.setSignature(_AsyncRun, {
    statics: () => ({
      _scheduleImmediate: [dart.void, [dart.functionType(dart.void, [])]],
      _initializeScheduleImmediate: [core.Function, []],
      _scheduleImmediateJsOverride: [dart.void, [dart.functionType(dart.void, [])]],
      _scheduleImmediateWithSetImmediate: [dart.void, [dart.functionType(dart.void, [])]],
      _scheduleImmediateWithTimer: [dart.void, [dart.functionType(dart.void, [])]]
    }),
    names: ['_scheduleImmediate', '_initializeScheduleImmediate', '_scheduleImmediateJsOverride', '_scheduleImmediateWithSetImmediate', '_scheduleImmediateWithTimer']
  });
  dart.defineLazyProperties(_AsyncRun, {
    get scheduleImmediateClosure() {
      return _AsyncRun._initializeScheduleImmediate();
    }
  });
  const StreamSubscription$ = dart.generic(function(T) {
    class StreamSubscription extends core.Object {}
    return StreamSubscription;
  });
  let StreamSubscription = StreamSubscription$();
  const EventSink$ = dart.generic(function(T) {
    class EventSink extends core.Object {}
    EventSink[dart.implements] = () => [core.Sink$(T)];
    return EventSink;
  });
  let EventSink = EventSink$();
  const _stream = Symbol('_stream');
  const StreamView$ = dart.generic(function(T) {
    class StreamView extends Stream$(T) {
      StreamView(stream) {
        this[_stream] = stream;
        super.Stream();
      }
      get isBroadcast() {
        return this[_stream].isBroadcast;
      }
      asBroadcastStream(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        dart.as(onListen, dart.functionType(dart.void, [StreamSubscription$(T)]));
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        dart.as(onCancel, dart.functionType(dart.void, [StreamSubscription$(T)]));
        return this[_stream].asBroadcastStream({onListen: onListen, onCancel: onCancel});
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        return this[_stream].listen(onData, {onError: onError, onDone: onDone, cancelOnError: cancelOnError});
      }
    }
    dart.setSignature(StreamView, {
      constructors: () => ({StreamView: [StreamView$(T), [Stream$(T)]]}),
      methods: () => ({
        asBroadcastStream: [Stream$(T), [], {onListen: dart.functionType(dart.void, [StreamSubscription$(T)]), onCancel: dart.functionType(dart.void, [StreamSubscription$(T)])}],
        listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}]
      })
    });
    return StreamView;
  });
  let StreamView = StreamView$();
  const StreamConsumer$ = dart.generic(function(S) {
    class StreamConsumer extends core.Object {}
    return StreamConsumer;
  });
  let StreamConsumer = StreamConsumer$();
  const StreamSink$ = dart.generic(function(S) {
    class StreamSink extends core.Object {}
    StreamSink[dart.implements] = () => [StreamConsumer$(S), EventSink$(S)];
    return StreamSink;
  });
  let StreamSink = StreamSink$();
  const StreamTransformer$ = dart.generic(function(S, T) {
    class StreamTransformer extends core.Object {
      static new(transformer) {
        return new _StreamSubscriptionTransformer(transformer);
      }
      static fromHandlers(opts) {
        return new _StreamHandlerTransformer(opts);
      }
    }
    dart.setSignature(StreamTransformer, {
      constructors: () => ({
        new: [StreamTransformer$(S, T), [dart.functionType(StreamSubscription$(T), [Stream$(S), core.bool])]],
        fromHandlers: [StreamTransformer$(S, T), [], {handleData: dart.functionType(dart.void, [S, EventSink$(T)]), handleError: dart.functionType(dart.void, [core.Object, core.StackTrace, EventSink$(T)]), handleDone: dart.functionType(dart.void, [EventSink$(T)])}]
      })
    });
    return StreamTransformer;
  });
  let StreamTransformer = StreamTransformer$();
  const StreamIterator$ = dart.generic(function(T) {
    class StreamIterator extends core.Object {
      static new(stream) {
        return new (_StreamIteratorImpl$(T))(stream);
      }
    }
    dart.setSignature(StreamIterator, {
      constructors: () => ({new: [StreamIterator$(T), [Stream$(T)]]})
    });
    return StreamIterator;
  });
  let StreamIterator = StreamIterator$();
  const _ControllerEventSinkWrapper$ = dart.generic(function(T) {
    class _ControllerEventSinkWrapper extends core.Object {
      _ControllerEventSinkWrapper(sink) {
        this[_sink] = sink;
      }
      add(data) {
        dart.as(data, T);
        this[_sink].add(data);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        this[_sink].addError(error, stackTrace);
      }
      close() {
        this[_sink].close();
      }
    }
    _ControllerEventSinkWrapper[dart.implements] = () => [EventSink$(T)];
    dart.setSignature(_ControllerEventSinkWrapper, {
      constructors: () => ({_ControllerEventSinkWrapper: [_ControllerEventSinkWrapper$(T), [EventSink]]}),
      methods: () => ({
        add: [dart.void, [T]],
        addError: [dart.void, [dart.dynamic], [core.StackTrace]],
        close: [dart.void, []]
      })
    });
    return _ControllerEventSinkWrapper;
  });
  let _ControllerEventSinkWrapper = _ControllerEventSinkWrapper$();
  const __CastType12 = dart.typedef('__CastType12', () => dart.functionType(dart.void, [StreamSubscription]));
  const __CastType14 = dart.typedef('__CastType14', () => dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace]));
  const __CastType17 = dart.typedef('__CastType17', () => dart.functionType(dart.void, []));
  const __CastType18 = dart.typedef('__CastType18', () => dart.functionType(dart.void, [EventSink]));
  const StreamController$ = dart.generic(function(T) {
    class StreamController extends core.Object {
      static new(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        let onPause = opts && 'onPause' in opts ? opts.onPause : null;
        let onResume = opts && 'onResume' in opts ? opts.onResume : null;
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        let sync = opts && 'sync' in opts ? opts.sync : false;
        if (onListen == null && onPause == null && onResume == null && onCancel == null) {
          return dart.as(dart.notNull(sync) ? new _NoCallbackSyncStreamController() : new _NoCallbackAsyncStreamController(), StreamController$(T));
        }
        return dart.notNull(sync) ? new (_SyncStreamController$(T))(onListen, onPause, onResume, onCancel) : new (_AsyncStreamController$(T))(onListen, onPause, onResume, onCancel);
      }
      static broadcast(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        let sync = opts && 'sync' in opts ? opts.sync : false;
        return dart.notNull(sync) ? new (_SyncBroadcastStreamController$(T))(onListen, onCancel) : new (_AsyncBroadcastStreamController$(T))(onListen, onCancel);
      }
    }
    StreamController[dart.implements] = () => [StreamSink$(T)];
    dart.setSignature(StreamController, {
      constructors: () => ({
        new: [StreamController$(T), [], {onListen: dart.functionType(dart.void, []), onPause: dart.functionType(dart.void, []), onResume: dart.functionType(dart.void, []), onCancel: dart.functionType(dart.dynamic, []), sync: core.bool}],
        broadcast: [StreamController$(T), [], {onListen: dart.functionType(dart.void, []), onCancel: dart.functionType(dart.void, []), sync: core.bool}]
      })
    });
    return StreamController;
  });
  let StreamController = StreamController$();
  const _StreamControllerLifecycle$ = dart.generic(function(T) {
    class _StreamControllerLifecycle extends core.Object {
      [_recordPause](subscription) {
        dart.as(subscription, StreamSubscription$(T));
      }
      [_recordResume](subscription) {
        dart.as(subscription, StreamSubscription$(T));
      }
      [_recordCancel](subscription) {
        dart.as(subscription, StreamSubscription$(T));
        return null;
      }
    }
    dart.setSignature(_StreamControllerLifecycle, {
      methods: () => ({
        [_recordPause]: [dart.void, [StreamSubscription$(T)]],
        [_recordResume]: [dart.void, [StreamSubscription$(T)]],
        [_recordCancel]: [Future, [StreamSubscription$(T)]]
      })
    });
    return _StreamControllerLifecycle;
  });
  let _StreamControllerLifecycle = _StreamControllerLifecycle$();
  const _varData = Symbol('_varData');
  const _isInitialState = Symbol('_isInitialState');
  const _subscription = Symbol('_subscription');
  const _pendingEvents = Symbol('_pendingEvents');
  const _ensurePendingEvents = Symbol('_ensurePendingEvents');
  const _badEventState = Symbol('_badEventState');
  const _StreamController$ = dart.generic(function(T) {
    class _StreamController extends core.Object {
      _StreamController() {
        this[_varData] = null;
        this[_state] = _StreamController$()._STATE_INITIAL;
        this[_doneFuture] = null;
      }
      get stream() {
        return new (_ControllerStream$(T))(this);
      }
      get sink() {
        return new (_StreamSinkWrapper$(T))(this);
      }
      get [_isCanceled]() {
        return (dart.notNull(this[_state]) & dart.notNull(_StreamController$()._STATE_CANCELED)) != 0;
      }
      get hasListener() {
        return (dart.notNull(this[_state]) & dart.notNull(_StreamController$()._STATE_SUBSCRIBED)) != 0;
      }
      get [_isInitialState]() {
        return (dart.notNull(this[_state]) & dart.notNull(_StreamController$()._STATE_SUBSCRIPTION_MASK)) == _StreamController$()._STATE_INITIAL;
      }
      get isClosed() {
        return (dart.notNull(this[_state]) & dart.notNull(_StreamController$()._STATE_CLOSED)) != 0;
      }
      get isPaused() {
        return dart.notNull(this.hasListener) ? this[_subscription][_isInputPaused] : !dart.notNull(this[_isCanceled]);
      }
      get [_isAddingStream]() {
        return (dart.notNull(this[_state]) & dart.notNull(_StreamController$()._STATE_ADDSTREAM)) != 0;
      }
      get [_mayAddEvent]() {
        return dart.notNull(this[_state]) < dart.notNull(_StreamController$()._STATE_CLOSED);
      }
      get [_pendingEvents]() {
        dart.assert(this[_isInitialState]);
        if (!dart.notNull(this[_isAddingStream])) {
          return dart.as(this[_varData], _PendingEvents);
        }
        let state = dart.as(this[_varData], _StreamControllerAddStreamState);
        return dart.as(state.varData, _PendingEvents);
      }
      [_ensurePendingEvents]() {
        dart.assert(this[_isInitialState]);
        if (!dart.notNull(this[_isAddingStream])) {
          if (this[_varData] == null)
            this[_varData] = new _StreamImplEvents();
          return dart.as(this[_varData], _StreamImplEvents);
        }
        let state = dart.as(this[_varData], _StreamControllerAddStreamState);
        if (state.varData == null)
          state.varData = new _StreamImplEvents();
        return dart.as(state.varData, _StreamImplEvents);
      }
      get [_subscription]() {
        dart.assert(this.hasListener);
        if (dart.notNull(this[_isAddingStream])) {
          let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
          return dart.as(addState.varData, _ControllerSubscription);
        }
        return dart.as(this[_varData], _ControllerSubscription);
      }
      [_badEventState]() {
        if (dart.notNull(this.isClosed)) {
          return new core.StateError("Cannot add event after closing");
        }
        dart.assert(this[_isAddingStream]);
        return new core.StateError("Cannot add event while adding a stream");
      }
      addStream(source, opts) {
        dart.as(source, Stream$(T));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : true;
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_badEventState]());
        if (dart.notNull(this[_isCanceled]))
          return new _Future.immediate(null);
        let addState = new _StreamControllerAddStreamState(this, this[_varData], source, cancelOnError);
        this[_varData] = addState;
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_StreamController$()._STATE_ADDSTREAM);
        return addState.addStreamFuture;
      }
      get done() {
        return this[_ensureDoneFuture]();
      }
      [_ensureDoneFuture]() {
        if (this[_doneFuture] == null) {
          this[_doneFuture] = dart.notNull(this[_isCanceled]) ? Future._nullFuture : new _Future();
        }
        return this[_doneFuture];
      }
      add(value) {
        dart.as(value, T);
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_badEventState]());
        this[_add](value);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        error = _nonNullError(error);
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_badEventState]());
        let replacement = Zone.current.errorCallback(error, stackTrace);
        if (replacement != null) {
          error = _nonNullError(replacement.error);
          stackTrace = replacement.stackTrace;
        }
        this[_addError](error, stackTrace);
      }
      close() {
        if (dart.notNull(this.isClosed)) {
          return this[_ensureDoneFuture]();
        }
        if (!dart.notNull(this[_mayAddEvent]))
          dart.throw(this[_badEventState]());
        this[_closeUnchecked]();
        return this[_ensureDoneFuture]();
      }
      [_closeUnchecked]() {
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_StreamController$()._STATE_CLOSED);
        if (dart.notNull(this.hasListener)) {
          this[_sendDone]();
        } else if (dart.notNull(this[_isInitialState])) {
          this[_ensurePendingEvents]().add(dart.const(new _DelayedDone()));
        }
      }
      [_add](value) {
        dart.as(value, T);
        if (dart.notNull(this.hasListener)) {
          this[_sendData](value);
        } else if (dart.notNull(this[_isInitialState])) {
          this[_ensurePendingEvents]().add(new (_DelayedData$(T))(value));
        }
      }
      [_addError](error, stackTrace) {
        if (dart.notNull(this.hasListener)) {
          this[_sendError](error, stackTrace);
        } else if (dart.notNull(this[_isInitialState])) {
          this[_ensurePendingEvents]().add(new _DelayedError(error, stackTrace));
        }
      }
      [_close]() {
        dart.assert(this[_isAddingStream]);
        let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
        this[_varData] = addState.varData;
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_StreamController$()._STATE_ADDSTREAM);
        addState.complete();
      }
      [_subscribe](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        if (!dart.notNull(this[_isInitialState])) {
          dart.throw(new core.StateError("Stream has already been listened to."));
        }
        let subscription = new _ControllerSubscription(this, onData, onError, onDone, cancelOnError);
        let pendingEvents = this[_pendingEvents];
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_StreamController$()._STATE_SUBSCRIBED);
        if (dart.notNull(this[_isAddingStream])) {
          let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
          addState.varData = subscription;
          addState.resume();
        } else {
          this[_varData] = subscription;
        }
        subscription[_setPendingEvents](pendingEvents);
        subscription[_guardCallback](dart.fn((() => {
          _runGuarded(this[_onListen]);
        }).bind(this)));
        return dart.as(subscription, StreamSubscription$(T));
      }
      [_recordCancel](subscription) {
        dart.as(subscription, StreamSubscription$(T));
        let result = null;
        if (dart.notNull(this[_isAddingStream])) {
          let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
          result = addState.cancel();
        }
        this[_varData] = null;
        this[_state] = dart.notNull(this[_state]) & ~(dart.notNull(_StreamController$()._STATE_SUBSCRIBED) | dart.notNull(_StreamController$()._STATE_ADDSTREAM)) | dart.notNull(_StreamController$()._STATE_CANCELED);
        if (this[_onCancel] != null) {
          if (result == null) {
            try {
              result = dart.as(this[_onCancel](), Future);
            } catch (e) {
              let s = dart.stackTrace(e);
              result = new _Future();
              result[_asyncCompleteError](e, s);
            }

          } else {
            result = result.whenComplete(this[_onCancel]);
          }
        }
        const complete = (function() {
          if (this[_doneFuture] != null && dart.notNull(this[_doneFuture][_mayComplete])) {
            this[_doneFuture][_asyncComplete](null);
          }
        }).bind(this);
        dart.fn(complete, dart.void, []);
        if (result != null) {
          result = result.whenComplete(complete);
        } else {
          complete();
        }
        return result;
      }
      [_recordPause](subscription) {
        dart.as(subscription, StreamSubscription$(T));
        if (dart.notNull(this[_isAddingStream])) {
          let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
          addState.pause();
        }
        _runGuarded(this[_onPause]);
      }
      [_recordResume](subscription) {
        dart.as(subscription, StreamSubscription$(T));
        if (dart.notNull(this[_isAddingStream])) {
          let addState = dart.as(this[_varData], _StreamControllerAddStreamState);
          addState.resume();
        }
        _runGuarded(this[_onResume]);
      }
    }
    _StreamController[dart.implements] = () => [StreamController$(T), _StreamControllerLifecycle$(T), _EventSink$(T), _EventDispatch$(T)];
    dart.setSignature(_StreamController, {
      constructors: () => ({_StreamController: [_StreamController$(T), []]}),
      methods: () => ({
        [_ensurePendingEvents]: [_StreamImplEvents, []],
        [_badEventState]: [core.Error, []],
        addStream: [Future, [Stream$(T)], {cancelOnError: core.bool}],
        [_ensureDoneFuture]: [Future, []],
        add: [dart.void, [T]],
        addError: [dart.void, [core.Object], [core.StackTrace]],
        close: [Future, []],
        [_closeUnchecked]: [dart.void, []],
        [_add]: [dart.void, [T]],
        [_addError]: [dart.void, [core.Object, core.StackTrace]],
        [_close]: [dart.void, []],
        [_subscribe]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]],
        [_recordCancel]: [Future, [StreamSubscription$(T)]],
        [_recordPause]: [dart.void, [StreamSubscription$(T)]],
        [_recordResume]: [dart.void, [StreamSubscription$(T)]]
      })
    });
    return _StreamController;
  });
  let _StreamController = _StreamController$();
  _StreamController._STATE_INITIAL = 0;
  _StreamController._STATE_SUBSCRIBED = 1;
  _StreamController._STATE_CANCELED = 2;
  _StreamController._STATE_SUBSCRIPTION_MASK = 3;
  _StreamController._STATE_CLOSED = 4;
  _StreamController._STATE_ADDSTREAM = 8;
  const _SyncStreamControllerDispatch$ = dart.generic(function(T) {
    class _SyncStreamControllerDispatch extends core.Object {
      [_sendData](data) {
        dart.as(data, T);
        this[_subscription][_add](data);
      }
      [_sendError](error, stackTrace) {
        this[_subscription][_addError](error, stackTrace);
      }
      [_sendDone]() {
        this[_subscription][_close]();
      }
    }
    _SyncStreamControllerDispatch[dart.implements] = () => [_StreamController$(T)];
    dart.setSignature(_SyncStreamControllerDispatch, {
      methods: () => ({
        [_sendData]: [dart.void, [T]],
        [_sendError]: [dart.void, [core.Object, core.StackTrace]],
        [_sendDone]: [dart.void, []]
      })
    });
    return _SyncStreamControllerDispatch;
  });
  let _SyncStreamControllerDispatch = _SyncStreamControllerDispatch$();
  const _AsyncStreamControllerDispatch$ = dart.generic(function(T) {
    class _AsyncStreamControllerDispatch extends core.Object {
      [_sendData](data) {
        dart.as(data, T);
        this[_subscription][_addPending](new _DelayedData(data));
      }
      [_sendError](error, stackTrace) {
        this[_subscription][_addPending](new _DelayedError(error, stackTrace));
      }
      [_sendDone]() {
        this[_subscription][_addPending](dart.const(new _DelayedDone()));
      }
    }
    _AsyncStreamControllerDispatch[dart.implements] = () => [_StreamController$(T)];
    dart.setSignature(_AsyncStreamControllerDispatch, {
      methods: () => ({
        [_sendData]: [dart.void, [T]],
        [_sendError]: [dart.void, [core.Object, core.StackTrace]],
        [_sendDone]: [dart.void, []]
      })
    });
    return _AsyncStreamControllerDispatch;
  });
  let _AsyncStreamControllerDispatch = _AsyncStreamControllerDispatch$();
  const _AsyncStreamController$ = dart.generic(function(T) {
    class _AsyncStreamController extends dart.mixin(_StreamController$(T), _AsyncStreamControllerDispatch$(T)) {
      _AsyncStreamController(onListen, onPause, onResume, onCancel) {
        this[_onListen] = onListen;
        this[_onPause] = onPause;
        this[_onResume] = onResume;
        this[_onCancel] = onCancel;
        super._StreamController();
      }
    }
    dart.setSignature(_AsyncStreamController, {
      constructors: () => ({_AsyncStreamController: [_AsyncStreamController$(T), [dart.functionType(dart.void, []), dart.functionType(dart.void, []), dart.functionType(dart.void, []), dart.functionType(dart.dynamic, [])]]})
    });
    return _AsyncStreamController;
  });
  let _AsyncStreamController = _AsyncStreamController$();
  const _SyncStreamController$ = dart.generic(function(T) {
    class _SyncStreamController extends dart.mixin(_StreamController$(T), _SyncStreamControllerDispatch$(T)) {
      _SyncStreamController(onListen, onPause, onResume, onCancel) {
        this[_onListen] = onListen;
        this[_onPause] = onPause;
        this[_onResume] = onResume;
        this[_onCancel] = onCancel;
        super._StreamController();
      }
    }
    dart.setSignature(_SyncStreamController, {
      constructors: () => ({_SyncStreamController: [_SyncStreamController$(T), [dart.functionType(dart.void, []), dart.functionType(dart.void, []), dart.functionType(dart.void, []), dart.functionType(dart.dynamic, [])]]})
    });
    return _SyncStreamController;
  });
  let _SyncStreamController = _SyncStreamController$();
  class _NoCallbacks extends core.Object {
    get [_onListen]() {
      return null;
    }
    get [_onPause]() {
      return null;
    }
    get [_onResume]() {
      return null;
    }
    get [_onCancel]() {
      return null;
    }
  }
  class _NoCallbackAsyncStreamController extends dart.mixin(_StreamController$(dart.dynamic), _AsyncStreamControllerDispatch$(dart.dynamic), _NoCallbacks) {
    _NoCallbackAsyncStreamController() {
      super._StreamController(...arguments);
    }
  }
  class _NoCallbackSyncStreamController extends dart.mixin(_StreamController$(dart.dynamic), _SyncStreamControllerDispatch$(dart.dynamic), _NoCallbacks) {
    _NoCallbackSyncStreamController() {
      super._StreamController(...arguments);
    }
  }
  const _NotificationHandler = dart.typedef('_NotificationHandler', () => dart.functionType(dart.dynamic, []));
  function _runGuarded(notificationHandler) {
    if (notificationHandler == null)
      return null;
    try {
      let result = notificationHandler();
      if (dart.is(result, Future))
        return dart.as(result, Future);
      return null;
    } catch (e) {
      let s = dart.stackTrace(e);
      Zone.current.handleUncaughtError(e, s);
    }

  }
  dart.fn(_runGuarded, Future, [_NotificationHandler]);
  const _target = Symbol('_target');
  const _StreamSinkWrapper$ = dart.generic(function(T) {
    class _StreamSinkWrapper extends core.Object {
      _StreamSinkWrapper(target) {
        this[_target] = target;
      }
      add(data) {
        dart.as(data, T);
        this[_target].add(data);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        this[_target].addError(error, stackTrace);
      }
      close() {
        return this[_target].close();
      }
      addStream(source, opts) {
        dart.as(source, Stream$(T));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : true;
        return this[_target].addStream(source, {cancelOnError: cancelOnError});
      }
      get done() {
        return this[_target].done;
      }
    }
    _StreamSinkWrapper[dart.implements] = () => [StreamSink$(T)];
    dart.setSignature(_StreamSinkWrapper, {
      constructors: () => ({_StreamSinkWrapper: [_StreamSinkWrapper$(T), [StreamController]]}),
      methods: () => ({
        add: [dart.void, [T]],
        addError: [dart.void, [core.Object], [core.StackTrace]],
        close: [Future, []],
        addStream: [Future, [Stream$(T)], {cancelOnError: core.bool}]
      })
    });
    return _StreamSinkWrapper;
  });
  let _StreamSinkWrapper = _StreamSinkWrapper$();
  const _AddStreamState$ = dart.generic(function(T) {
    class _AddStreamState extends core.Object {
      _AddStreamState(controller, source, cancelOnError) {
        this.addStreamFuture = new _Future();
        this.addSubscription = source.listen(dart.bind(controller, _add), {onError: dart.notNull(cancelOnError) ? dart.as(_AddStreamState$().makeErrorHandler(controller), core.Function) : dart.bind(controller, _addError), onDone: dart.bind(controller, _close), cancelOnError: cancelOnError});
      }
      static makeErrorHandler(controller) {
        return dart.fn((e, s) => {
          controller[_addError](e, s);
          controller[_close]();
        }, dart.dynamic, [dart.dynamic, core.StackTrace]);
      }
      pause() {
        this.addSubscription.pause();
      }
      resume() {
        this.addSubscription.resume();
      }
      cancel() {
        let cancel2 = this.addSubscription.cancel();
        if (cancel2 == null) {
          this.addStreamFuture[_asyncComplete](null);
          return null;
        }
        return cancel2.whenComplete(dart.fn((() => {
          this.addStreamFuture[_asyncComplete](null);
        }).bind(this)));
      }
      complete() {
        this.addStreamFuture[_asyncComplete](null);
      }
    }
    dart.setSignature(_AddStreamState, {
      constructors: () => ({_AddStreamState: [_AddStreamState$(T), [_EventSink$(T), Stream, core.bool]]}),
      methods: () => ({
        pause: [dart.void, []],
        resume: [dart.void, []],
        cancel: [Future, []],
        complete: [dart.void, []]
      }),
      statics: () => ({makeErrorHandler: [dart.dynamic, [_EventSink]]}),
      names: ['makeErrorHandler']
    });
    return _AddStreamState;
  });
  let _AddStreamState = _AddStreamState$();
  const _StreamControllerAddStreamState$ = dart.generic(function(T) {
    class _StreamControllerAddStreamState extends _AddStreamState$(T) {
      _StreamControllerAddStreamState(controller, varData, source, cancelOnError) {
        this.varData = varData;
        super._AddStreamState(dart.as(controller, _EventSink$(T)), source, cancelOnError);
        if (dart.notNull(controller.isPaused)) {
          this.addSubscription.pause();
        }
      }
    }
    dart.setSignature(_StreamControllerAddStreamState, {
      constructors: () => ({_StreamControllerAddStreamState: [_StreamControllerAddStreamState$(T), [_StreamController, dart.dynamic, Stream, core.bool]]})
    });
    return _StreamControllerAddStreamState;
  });
  let _StreamControllerAddStreamState = _StreamControllerAddStreamState$();
  const _EventSink$ = dart.generic(function(T) {
    class _EventSink extends core.Object {}
    return _EventSink;
  });
  let _EventSink = _EventSink$();
  const _EventDispatch$ = dart.generic(function(T) {
    class _EventDispatch extends core.Object {}
    return _EventDispatch;
  });
  let _EventDispatch = _EventDispatch$();
  _BufferingStreamSubscription._STATE_CANCEL_ON_ERROR = 1;
  _BufferingStreamSubscription._STATE_CLOSED = 2;
  _BufferingStreamSubscription._STATE_INPUT_PAUSED = 4;
  _BufferingStreamSubscription._STATE_CANCELED = 8;
  _BufferingStreamSubscription._STATE_WAIT_FOR_CANCEL = 16;
  _BufferingStreamSubscription._STATE_IN_CALLBACK = 32;
  _BufferingStreamSubscription._STATE_HAS_PENDING = 64;
  _BufferingStreamSubscription._STATE_PAUSE_COUNT = 128;
  _BufferingStreamSubscription._STATE_PAUSE_COUNT_SHIFT = 7;
  const _EventGenerator = dart.typedef('_EventGenerator', () => dart.functionType(_PendingEvents, []));
  const _isUsed = Symbol('_isUsed');
  const _GeneratedStreamImpl$ = dart.generic(function(T) {
    class _GeneratedStreamImpl extends _StreamImpl$(T) {
      _GeneratedStreamImpl(pending) {
        this[_pending] = pending;
        this[_isUsed] = false;
      }
      [_createSubscription](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        if (dart.notNull(this[_isUsed]))
          dart.throw(new core.StateError("Stream has already been listened to."));
        this[_isUsed] = true;
        return dart.as((() => {
          let _ = new _BufferingStreamSubscription(onData, onError, onDone, cancelOnError);
          _[_setPendingEvents](this[_pending]());
          return _;
        }).bind(this)(), StreamSubscription$(T));
      }
    }
    dart.setSignature(_GeneratedStreamImpl, {
      constructors: () => ({_GeneratedStreamImpl: [_GeneratedStreamImpl$(T), [_EventGenerator]]}),
      methods: () => ({[_createSubscription]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]})
    });
    return _GeneratedStreamImpl;
  });
  let _GeneratedStreamImpl = _GeneratedStreamImpl$();
  const _eventScheduled = Symbol('_eventScheduled');
  class _PendingEvents extends core.Object {
    _PendingEvents() {
      this[_state] = _PendingEvents._STATE_UNSCHEDULED;
    }
    get isScheduled() {
      return this[_state] == _PendingEvents._STATE_SCHEDULED;
    }
    get [_eventScheduled]() {
      return dart.notNull(this[_state]) >= dart.notNull(_PendingEvents._STATE_SCHEDULED);
    }
    schedule(dispatch) {
      if (dart.notNull(this.isScheduled))
        return;
      dart.assert(!dart.notNull(this.isEmpty));
      if (dart.notNull(this[_eventScheduled])) {
        dart.assert(this[_state] == _PendingEvents._STATE_CANCELED);
        this[_state] = _PendingEvents._STATE_SCHEDULED;
        return;
      }
      scheduleMicrotask(dart.fn((() => {
        let oldState = this[_state];
        this[_state] = _PendingEvents._STATE_UNSCHEDULED;
        if (oldState == _PendingEvents._STATE_CANCELED)
          return;
        this.handleNext(dispatch);
      }).bind(this)));
      this[_state] = _PendingEvents._STATE_SCHEDULED;
    }
    cancelSchedule() {
      if (dart.notNull(this.isScheduled))
        this[_state] = _PendingEvents._STATE_CANCELED;
    }
  }
  dart.setSignature(_PendingEvents, {
    methods: () => ({
      schedule: [dart.void, [_EventDispatch]],
      cancelSchedule: [dart.void, []]
    })
  });
  const _iterator = Symbol('_iterator');
  const _IterablePendingEvents$ = dart.generic(function(T) {
    class _IterablePendingEvents extends _PendingEvents {
      _IterablePendingEvents(data) {
        this[_iterator] = data[dartx.iterator];
        super._PendingEvents();
      }
      get isEmpty() {
        return this[_iterator] == null;
      }
      handleNext(dispatch) {
        if (this[_iterator] == null) {
          dart.throw(new core.StateError("No events pending."));
        }
        let isDone = null;
        try {
          isDone = !dart.notNull(this[_iterator].moveNext());
        } catch (e) {
          let s = dart.stackTrace(e);
          this[_iterator] = null;
          dispatch[_sendError](e, s);
          return;
        }

        if (!dart.notNull(isDone)) {
          dispatch[_sendData](this[_iterator].current);
        } else {
          this[_iterator] = null;
          dispatch[_sendDone]();
        }
      }
      clear() {
        if (dart.notNull(this.isScheduled))
          this.cancelSchedule();
        this[_iterator] = null;
      }
    }
    dart.setSignature(_IterablePendingEvents, {
      constructors: () => ({_IterablePendingEvents: [_IterablePendingEvents$(T), [core.Iterable$(T)]]}),
      methods: () => ({
        handleNext: [dart.void, [_EventDispatch]],
        clear: [dart.void, []]
      })
    });
    return _IterablePendingEvents;
  });
  let _IterablePendingEvents = _IterablePendingEvents$();
  const _DataHandler$ = dart.generic(function(T) {
    const _DataHandler = dart.typedef('_DataHandler', () => dart.functionType(dart.void, [T]));
    return _DataHandler;
  });
  let _DataHandler = _DataHandler$();
  const _DoneHandler = dart.typedef('_DoneHandler', () => dart.functionType(dart.void, []));
  function _nullDataHandler(value) {
  }
  dart.fn(_nullDataHandler, dart.void, [dart.dynamic]);
  function _nullErrorHandler(error, stackTrace) {
    if (stackTrace === void 0)
      stackTrace = null;
    Zone.current.handleUncaughtError(error, stackTrace);
  }
  dart.fn(_nullErrorHandler, dart.void, [dart.dynamic], [core.StackTrace]);
  function _nullDoneHandler() {
  }
  dart.fn(_nullDoneHandler, dart.void, []);
  const _DelayedEvent$ = dart.generic(function(T) {
    class _DelayedEvent extends core.Object {
      _DelayedEvent() {
        this.next = null;
      }
    }
    return _DelayedEvent;
  });
  let _DelayedEvent = _DelayedEvent$();
  const _DelayedData$ = dart.generic(function(T) {
    class _DelayedData extends _DelayedEvent$(T) {
      _DelayedData(value) {
        this.value = value;
        super._DelayedEvent();
      }
      perform(dispatch) {
        dart.as(dispatch, _EventDispatch$(T));
        dispatch[_sendData](this.value);
      }
    }
    dart.setSignature(_DelayedData, {
      constructors: () => ({_DelayedData: [_DelayedData$(T), [T]]}),
      methods: () => ({perform: [dart.void, [_EventDispatch$(T)]]})
    });
    return _DelayedData;
  });
  let _DelayedData = _DelayedData$();
  class _DelayedError extends _DelayedEvent {
    _DelayedError(error, stackTrace) {
      this.error = error;
      this.stackTrace = stackTrace;
      super._DelayedEvent();
    }
    perform(dispatch) {
      dispatch[_sendError](this.error, this.stackTrace);
    }
  }
  dart.setSignature(_DelayedError, {
    constructors: () => ({_DelayedError: [_DelayedError, [dart.dynamic, core.StackTrace]]}),
    methods: () => ({perform: [dart.void, [_EventDispatch]]})
  });
  class _DelayedDone extends core.Object {
    _DelayedDone() {
    }
    perform(dispatch) {
      dispatch[_sendDone]();
    }
    get next() {
      return null;
    }
    set next(_) {
      dart.throw(new core.StateError("No events after a done."));
    }
  }
  _DelayedDone[dart.implements] = () => [_DelayedEvent];
  dart.setSignature(_DelayedDone, {
    constructors: () => ({_DelayedDone: [_DelayedDone, []]}),
    methods: () => ({perform: [dart.void, [_EventDispatch]]})
  });
  _PendingEvents._STATE_UNSCHEDULED = 0;
  _PendingEvents._STATE_SCHEDULED = 1;
  _PendingEvents._STATE_CANCELED = 3;
  class _StreamImplEvents extends _PendingEvents {
    _StreamImplEvents() {
      this.firstPendingEvent = null;
      this.lastPendingEvent = null;
      super._PendingEvents();
    }
    get isEmpty() {
      return this.lastPendingEvent == null;
    }
    add(event) {
      if (this.lastPendingEvent == null) {
        this.firstPendingEvent = this.lastPendingEvent = event;
      } else {
        this.lastPendingEvent = this.lastPendingEvent.next = event;
      }
    }
    handleNext(dispatch) {
      dart.assert(!dart.notNull(this.isScheduled));
      let event = this.firstPendingEvent;
      this.firstPendingEvent = event.next;
      if (this.firstPendingEvent == null) {
        this.lastPendingEvent = null;
      }
      event.perform(dispatch);
    }
    clear() {
      if (dart.notNull(this.isScheduled))
        this.cancelSchedule();
      this.firstPendingEvent = this.lastPendingEvent = null;
    }
  }
  dart.setSignature(_StreamImplEvents, {
    methods: () => ({
      add: [dart.void, [_DelayedEvent]],
      handleNext: [dart.void, [_EventDispatch]],
      clear: [dart.void, []]
    })
  });
  const _unlink = Symbol('_unlink');
  const _insertBefore = Symbol('_insertBefore');
  class _BroadcastLinkedList extends core.Object {
    _BroadcastLinkedList() {
      this[_next] = null;
      this[_previous] = null;
    }
    [_unlink]() {
      this[_previous][_next] = this[_next];
      this[_next][_previous] = this[_previous];
      this[_next] = this[_previous] = this;
    }
    [_insertBefore](newNext) {
      let newPrevious = newNext[_previous];
      newPrevious[_next] = this;
      newNext[_previous] = this[_previous];
      this[_previous][_next] = newNext;
      this[_previous] = newPrevious;
    }
  }
  dart.setSignature(_BroadcastLinkedList, {
    methods: () => ({
      [_unlink]: [dart.void, []],
      [_insertBefore]: [dart.void, [_BroadcastLinkedList]]
    })
  });
  const _broadcastCallback = dart.typedef('_broadcastCallback', () => dart.functionType(dart.void, [StreamSubscription]));
  const _schedule = Symbol('_schedule');
  const _isSent = Symbol('_isSent');
  const _isScheduled = Symbol('_isScheduled');
  const _DoneStreamSubscription$ = dart.generic(function(T) {
    class _DoneStreamSubscription extends core.Object {
      _DoneStreamSubscription(onDone) {
        this[_onDone] = onDone;
        this[_zone] = Zone.current;
        this[_state] = 0;
        this[_schedule]();
      }
      get [_isSent]() {
        return (dart.notNull(this[_state]) & dart.notNull(_DoneStreamSubscription$()._DONE_SENT)) != 0;
      }
      get [_isScheduled]() {
        return (dart.notNull(this[_state]) & dart.notNull(_DoneStreamSubscription$()._SCHEDULED)) != 0;
      }
      get isPaused() {
        return dart.notNull(this[_state]) >= dart.notNull(_DoneStreamSubscription$()._PAUSED);
      }
      [_schedule]() {
        if (dart.notNull(this[_isScheduled]))
          return;
        this[_zone].scheduleMicrotask(dart.bind(this, _sendDone));
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_DoneStreamSubscription$()._SCHEDULED);
      }
      onData(handleData) {
        dart.as(handleData, dart.functionType(dart.void, [T]));
      }
      onError(handleError) {}
      onDone(handleDone) {
        dart.as(handleDone, dart.functionType(dart.void, []));
        this[_onDone] = handleDone;
      }
      pause(resumeSignal) {
        if (resumeSignal === void 0)
          resumeSignal = null;
        this[_state] = dart.notNull(this[_state]) + dart.notNull(_DoneStreamSubscription$()._PAUSED);
        if (resumeSignal != null)
          resumeSignal.whenComplete(dart.bind(this, 'resume'));
      }
      resume() {
        if (dart.notNull(this.isPaused)) {
          this[_state] = dart.notNull(this[_state]) - dart.notNull(_DoneStreamSubscription$()._PAUSED);
          if (!dart.notNull(this.isPaused) && !dart.notNull(this[_isSent])) {
            this[_schedule]();
          }
        }
      }
      cancel() {
        return null;
      }
      asFuture(futureValue) {
        if (futureValue === void 0)
          futureValue = null;
        let result = new _Future();
        this[_onDone] = dart.fn(() => {
          result[_completeWithValue](null);
        });
        return result;
      }
      [_sendDone]() {
        this[_state] = dart.notNull(this[_state]) & ~dart.notNull(_DoneStreamSubscription$()._SCHEDULED);
        if (dart.notNull(this.isPaused))
          return;
        this[_state] = dart.notNull(this[_state]) | dart.notNull(_DoneStreamSubscription$()._DONE_SENT);
        if (this[_onDone] != null)
          this[_zone].runGuarded(this[_onDone]);
      }
    }
    _DoneStreamSubscription[dart.implements] = () => [StreamSubscription$(T)];
    dart.setSignature(_DoneStreamSubscription, {
      constructors: () => ({_DoneStreamSubscription: [_DoneStreamSubscription$(T), [_DoneHandler]]}),
      methods: () => ({
        [_schedule]: [dart.void, []],
        onData: [dart.void, [dart.functionType(dart.void, [T])]],
        onError: [dart.void, [core.Function]],
        onDone: [dart.void, [dart.functionType(dart.void, [])]],
        pause: [dart.void, [], [Future]],
        resume: [dart.void, []],
        cancel: [Future, []],
        asFuture: [Future, [], [dart.dynamic]],
        [_sendDone]: [dart.void, []]
      })
    });
    return _DoneStreamSubscription;
  });
  let _DoneStreamSubscription = _DoneStreamSubscription$();
  _DoneStreamSubscription._DONE_SENT = 1;
  _DoneStreamSubscription._SCHEDULED = 2;
  _DoneStreamSubscription._PAUSED = 4;
  const _source = Symbol('_source');
  const _onListenHandler = Symbol('_onListenHandler');
  const _onCancelHandler = Symbol('_onCancelHandler');
  const _cancelSubscription = Symbol('_cancelSubscription');
  const _pauseSubscription = Symbol('_pauseSubscription');
  const _resumeSubscription = Symbol('_resumeSubscription');
  const _isSubscriptionPaused = Symbol('_isSubscriptionPaused');
  const _AsBroadcastStream$ = dart.generic(function(T) {
    class _AsBroadcastStream extends Stream$(T) {
      _AsBroadcastStream(source, onListenHandler, onCancelHandler) {
        this[_source] = source;
        this[_onListenHandler] = dart.as(Zone.current.registerUnaryCallback(onListenHandler), _broadcastCallback);
        this[_onCancelHandler] = dart.as(Zone.current.registerUnaryCallback(onCancelHandler), _broadcastCallback);
        this[_zone] = Zone.current;
        this[_controller] = null;
        this[_subscription] = null;
        super.Stream();
        this[_controller] = new (_AsBroadcastStreamController$(T))(dart.bind(this, _onListen), dart.bind(this, _onCancel));
      }
      get isBroadcast() {
        return true;
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        if (this[_controller] == null || dart.notNull(this[_controller].isClosed)) {
          return new (_DoneStreamSubscription$(T))(onDone);
        }
        if (this[_subscription] == null) {
          this[_subscription] = this[_source].listen(dart.bind(this[_controller], 'add'), {onError: dart.bind(this[_controller], 'addError'), onDone: dart.bind(this[_controller], 'close')});
        }
        cancelOnError = core.identical(true, cancelOnError);
        return this[_controller][_subscribe](onData, onError, onDone, cancelOnError);
      }
      [_onCancel]() {
        let shutdown = this[_controller] == null || dart.notNull(this[_controller].isClosed);
        if (this[_onCancelHandler] != null) {
          this[_zone].runUnary(this[_onCancelHandler], new _BroadcastSubscriptionWrapper(this));
        }
        if (dart.notNull(shutdown)) {
          if (this[_subscription] != null) {
            this[_subscription].cancel();
            this[_subscription] = null;
          }
        }
      }
      [_onListen]() {
        if (this[_onListenHandler] != null) {
          this[_zone].runUnary(this[_onListenHandler], new _BroadcastSubscriptionWrapper(this));
        }
      }
      [_cancelSubscription]() {
        if (this[_subscription] == null)
          return;
        let subscription = this[_subscription];
        this[_subscription] = null;
        this[_controller] = null;
        subscription.cancel();
      }
      [_pauseSubscription](resumeSignal) {
        if (this[_subscription] == null)
          return;
        this[_subscription].pause(resumeSignal);
      }
      [_resumeSubscription]() {
        if (this[_subscription] == null)
          return;
        this[_subscription].resume();
      }
      get [_isSubscriptionPaused]() {
        if (this[_subscription] == null)
          return false;
        return this[_subscription].isPaused;
      }
    }
    dart.setSignature(_AsBroadcastStream, {
      constructors: () => ({_AsBroadcastStream: [_AsBroadcastStream$(T), [Stream$(T), dart.functionType(dart.void, [StreamSubscription]), dart.functionType(dart.void, [StreamSubscription])]]}),
      methods: () => ({
        listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
        [_onCancel]: [dart.void, []],
        [_onListen]: [dart.void, []],
        [_cancelSubscription]: [dart.void, []],
        [_pauseSubscription]: [dart.void, [Future]],
        [_resumeSubscription]: [dart.void, []]
      })
    });
    return _AsBroadcastStream;
  });
  let _AsBroadcastStream = _AsBroadcastStream$();
  const _BroadcastSubscriptionWrapper$ = dart.generic(function(T) {
    class _BroadcastSubscriptionWrapper extends core.Object {
      _BroadcastSubscriptionWrapper(stream) {
        this[_stream] = stream;
      }
      onData(handleData) {
        dart.as(handleData, dart.functionType(dart.void, [T]));
        dart.throw(new core.UnsupportedError("Cannot change handlers of asBroadcastStream source subscription."));
      }
      onError(handleError) {
        dart.throw(new core.UnsupportedError("Cannot change handlers of asBroadcastStream source subscription."));
      }
      onDone(handleDone) {
        dart.as(handleDone, dart.functionType(dart.void, []));
        dart.throw(new core.UnsupportedError("Cannot change handlers of asBroadcastStream source subscription."));
      }
      pause(resumeSignal) {
        if (resumeSignal === void 0)
          resumeSignal = null;
        this[_stream][_pauseSubscription](resumeSignal);
      }
      resume() {
        this[_stream][_resumeSubscription]();
      }
      cancel() {
        this[_stream][_cancelSubscription]();
        return null;
      }
      get isPaused() {
        return this[_stream][_isSubscriptionPaused];
      }
      asFuture(futureValue) {
        if (futureValue === void 0)
          futureValue = null;
        dart.throw(new core.UnsupportedError("Cannot change handlers of asBroadcastStream source subscription."));
      }
    }
    _BroadcastSubscriptionWrapper[dart.implements] = () => [StreamSubscription$(T)];
    dart.setSignature(_BroadcastSubscriptionWrapper, {
      constructors: () => ({_BroadcastSubscriptionWrapper: [_BroadcastSubscriptionWrapper$(T), [_AsBroadcastStream]]}),
      methods: () => ({
        onData: [dart.void, [dart.functionType(dart.void, [T])]],
        onError: [dart.void, [core.Function]],
        onDone: [dart.void, [dart.functionType(dart.void, [])]],
        pause: [dart.void, [], [Future]],
        resume: [dart.void, []],
        cancel: [Future, []],
        asFuture: [Future, [], [dart.dynamic]]
      })
    });
    return _BroadcastSubscriptionWrapper;
  });
  let _BroadcastSubscriptionWrapper = _BroadcastSubscriptionWrapper$();
  const _current = Symbol('_current');
  const _futureOrPrefetch = Symbol('_futureOrPrefetch');
  const _clear = Symbol('_clear');
  const _StreamIteratorImpl$ = dart.generic(function(T) {
    class _StreamIteratorImpl extends core.Object {
      _StreamIteratorImpl(stream) {
        this[_subscription] = null;
        this[_current] = null;
        this[_futureOrPrefetch] = null;
        this[_state] = _StreamIteratorImpl$()._STATE_FOUND;
        this[_subscription] = stream.listen(dart.bind(this, _onData), {onError: dart.bind(this, _onError), onDone: dart.bind(this, _onDone), cancelOnError: true});
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        if (this[_state] == _StreamIteratorImpl$()._STATE_DONE) {
          return new (_Future$(core.bool)).immediate(false);
        }
        if (this[_state] == _StreamIteratorImpl$()._STATE_MOVING) {
          dart.throw(new core.StateError("Already waiting for next."));
        }
        if (this[_state] == _StreamIteratorImpl$()._STATE_FOUND) {
          this[_state] = _StreamIteratorImpl$()._STATE_MOVING;
          this[_current] = null;
          this[_futureOrPrefetch] = new (_Future$(core.bool))();
          return dart.as(this[_futureOrPrefetch], Future$(core.bool));
        } else {
          dart.assert(dart.notNull(this[_state]) >= dart.notNull(_StreamIteratorImpl$()._STATE_EXTRA_DATA));
          switch (this[_state]) {
            case _StreamIteratorImpl$()._STATE_EXTRA_DATA:
            {
              this[_state] = _StreamIteratorImpl$()._STATE_FOUND;
              this[_current] = dart.as(this[_futureOrPrefetch], T);
              this[_futureOrPrefetch] = null;
              this[_subscription].resume();
              return new (_Future$(core.bool)).immediate(true);
            }
            case _StreamIteratorImpl$()._STATE_EXTRA_ERROR:
            {
              let prefetch = dart.as(this[_futureOrPrefetch], AsyncError);
              this[_clear]();
              return new (_Future$(core.bool)).immediateError(prefetch.error, prefetch.stackTrace);
            }
            case _StreamIteratorImpl$()._STATE_EXTRA_DONE:
            {
              this[_clear]();
              return new (_Future$(core.bool)).immediate(false);
            }
          }
        }
      }
      [_clear]() {
        this[_subscription] = null;
        this[_futureOrPrefetch] = null;
        this[_current] = null;
        this[_state] = _StreamIteratorImpl$()._STATE_DONE;
      }
      cancel() {
        let subscription = this[_subscription];
        if (subscription == null)
          return null;
        if (this[_state] == _StreamIteratorImpl$()._STATE_MOVING) {
          let hasNext = dart.as(this[_futureOrPrefetch], _Future$(core.bool));
          this[_clear]();
          hasNext[_complete](false);
        } else {
          this[_clear]();
        }
        return subscription.cancel();
      }
      [_onData](data) {
        dart.as(data, T);
        if (this[_state] == _StreamIteratorImpl$()._STATE_MOVING) {
          this[_current] = data;
          let hasNext = dart.as(this[_futureOrPrefetch], _Future$(core.bool));
          this[_futureOrPrefetch] = null;
          this[_state] = _StreamIteratorImpl$()._STATE_FOUND;
          hasNext[_complete](true);
          return;
        }
        this[_subscription].pause();
        dart.assert(this[_futureOrPrefetch] == null);
        this[_futureOrPrefetch] = data;
        this[_state] = _StreamIteratorImpl$()._STATE_EXTRA_DATA;
      }
      [_onError](error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        if (this[_state] == _StreamIteratorImpl$()._STATE_MOVING) {
          let hasNext = dart.as(this[_futureOrPrefetch], _Future$(core.bool));
          this[_clear]();
          hasNext[_completeError](error, stackTrace);
          return;
        }
        this[_subscription].pause();
        dart.assert(this[_futureOrPrefetch] == null);
        this[_futureOrPrefetch] = new AsyncError(error, stackTrace);
        this[_state] = _StreamIteratorImpl$()._STATE_EXTRA_ERROR;
      }
      [_onDone]() {
        if (this[_state] == _StreamIteratorImpl$()._STATE_MOVING) {
          let hasNext = dart.as(this[_futureOrPrefetch], _Future$(core.bool));
          this[_clear]();
          hasNext[_complete](false);
          return;
        }
        this[_subscription].pause();
        this[_futureOrPrefetch] = null;
        this[_state] = _StreamIteratorImpl$()._STATE_EXTRA_DONE;
      }
    }
    _StreamIteratorImpl[dart.implements] = () => [StreamIterator$(T)];
    dart.setSignature(_StreamIteratorImpl, {
      constructors: () => ({_StreamIteratorImpl: [_StreamIteratorImpl$(T), [Stream$(T)]]}),
      methods: () => ({
        moveNext: [Future$(core.bool), []],
        [_clear]: [dart.void, []],
        cancel: [Future, []],
        [_onData]: [dart.void, [T]],
        [_onError]: [dart.void, [core.Object], [core.StackTrace]],
        [_onDone]: [dart.void, []]
      })
    });
    return _StreamIteratorImpl;
  });
  let _StreamIteratorImpl = _StreamIteratorImpl$();
  _StreamIteratorImpl._STATE_FOUND = 0;
  _StreamIteratorImpl._STATE_DONE = 1;
  _StreamIteratorImpl._STATE_MOVING = 2;
  _StreamIteratorImpl._STATE_EXTRA_DATA = 3;
  _StreamIteratorImpl._STATE_EXTRA_ERROR = 4;
  _StreamIteratorImpl._STATE_EXTRA_DONE = 5;
  const __CastType20$ = dart.generic(function(T) {
    const __CastType20 = dart.typedef('__CastType20', () => dart.functionType(dart.void, [T]));
    return __CastType20;
  });
  let __CastType20 = __CastType20$();
  const __CastType22 = dart.typedef('__CastType22', () => dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
  const __CastType25 = dart.typedef('__CastType25', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  function _runUserCode(userCode, onSuccess, onError) {
    try {
      dart.dcall(onSuccess, userCode());
    } catch (e) {
      let s = dart.stackTrace(e);
      let replacement = Zone.current.errorCallback(e, s);
      if (replacement == null) {
        dart.dcall(onError, e, s);
      } else {
        let error = _nonNullError(replacement.error);
        let stackTrace = replacement.stackTrace;
        dart.dcall(onError, error, stackTrace);
      }
    }

  }
  dart.fn(_runUserCode, dart.dynamic, [dart.functionType(dart.dynamic, []), dart.functionType(dart.dynamic, [dart.dynamic]), dart.functionType(dart.dynamic, [dart.dynamic, core.StackTrace])]);
  function _cancelAndError(subscription, future, error, stackTrace) {
    let cancelFuture = subscription.cancel();
    if (dart.is(cancelFuture, Future)) {
      cancelFuture.whenComplete(dart.fn(() => future[_completeError](error, stackTrace), dart.void, []));
    } else {
      future[_completeError](error, stackTrace);
    }
  }
  dart.fn(_cancelAndError, dart.void, [StreamSubscription, _Future, dart.dynamic, core.StackTrace]);
  function _cancelAndErrorWithReplacement(subscription, future, error, stackTrace) {
    let replacement = Zone.current.errorCallback(error, stackTrace);
    if (replacement != null) {
      error = _nonNullError(replacement.error);
      stackTrace = replacement.stackTrace;
    }
    _cancelAndError(subscription, future, error, stackTrace);
  }
  dart.fn(_cancelAndErrorWithReplacement, dart.void, [StreamSubscription, _Future, dart.dynamic, core.StackTrace]);
  function _cancelAndErrorClosure(subscription, future) {
    return dart.fn((error, stackTrace) => _cancelAndError(subscription, future, error, stackTrace), dart.void, [dart.dynamic, core.StackTrace]);
  }
  dart.fn(_cancelAndErrorClosure, dart.dynamic, [StreamSubscription, _Future]);
  function _cancelAndValue(subscription, future, value) {
    let cancelFuture = subscription.cancel();
    if (dart.is(cancelFuture, Future)) {
      cancelFuture.whenComplete(dart.fn(() => future[_complete](value), dart.void, []));
    } else {
      future[_complete](value);
    }
  }
  dart.fn(_cancelAndValue, dart.void, [StreamSubscription, _Future, dart.dynamic]);
  const _handleData = Symbol('_handleData');
  const _handleError = Symbol('_handleError');
  const _handleDone = Symbol('_handleDone');
  const _ForwardingStream$ = dart.generic(function(S, T) {
    class _ForwardingStream extends Stream$(T) {
      _ForwardingStream(source) {
        this[_source] = source;
        super.Stream();
      }
      get isBroadcast() {
        return this[_source].isBroadcast;
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        cancelOnError = core.identical(true, cancelOnError);
        return this[_createSubscription](onData, onError, onDone, cancelOnError);
      }
      [_createSubscription](onData, onError, onDone, cancelOnError) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        dart.as(onDone, dart.functionType(dart.void, []));
        return new (_ForwardingStreamSubscription$(S, T))(this, onData, onError, onDone, cancelOnError);
      }
      [_handleData](data, sink) {
        dart.as(data, S);
        dart.as(sink, _EventSink$(T));
        let outputData = data;
        sink[_add](dart.as(outputData, T));
      }
      [_handleError](error, stackTrace, sink) {
        dart.as(sink, _EventSink$(T));
        sink[_addError](error, stackTrace);
      }
      [_handleDone](sink) {
        dart.as(sink, _EventSink$(T));
        sink[_close]();
      }
    }
    dart.setSignature(_ForwardingStream, {
      constructors: () => ({_ForwardingStream: [_ForwardingStream$(S, T), [Stream$(S)]]}),
      methods: () => ({
        listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
        [_createSubscription]: [StreamSubscription$(T), [dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]],
        [_handleData]: [dart.void, [S, _EventSink$(T)]],
        [_handleError]: [dart.void, [dart.dynamic, core.StackTrace, _EventSink$(T)]],
        [_handleDone]: [dart.void, [_EventSink$(T)]]
      })
    });
    return _ForwardingStream;
  });
  let _ForwardingStream = _ForwardingStream$();
  const _ForwardingStreamSubscription$ = dart.generic(function(S, T) {
    class _ForwardingStreamSubscription extends _BufferingStreamSubscription$(T) {
      _ForwardingStreamSubscription(stream, onData, onError, onDone, cancelOnError) {
        this[_stream] = stream;
        this[_subscription] = null;
        super._BufferingStreamSubscription(onData, onError, onDone, cancelOnError);
        this[_subscription] = this[_stream][_source].listen(dart.bind(this, _handleData), {onError: dart.bind(this, _handleError), onDone: dart.bind(this, _handleDone)});
      }
      [_add](data) {
        dart.as(data, T);
        if (dart.notNull(this[_isClosed]))
          return;
        super[_add](data);
      }
      [_addError](error, stackTrace) {
        if (dart.notNull(this[_isClosed]))
          return;
        super[_addError](error, stackTrace);
      }
      [_onPause]() {
        if (this[_subscription] == null)
          return;
        this[_subscription].pause();
      }
      [_onResume]() {
        if (this[_subscription] == null)
          return;
        this[_subscription].resume();
      }
      [_onCancel]() {
        if (this[_subscription] != null) {
          let subscription = this[_subscription];
          this[_subscription] = null;
          subscription.cancel();
        }
        return null;
      }
      [_handleData](data) {
        dart.as(data, S);
        this[_stream][_handleData](data, this);
      }
      [_handleError](error, stackTrace) {
        this[_stream][_handleError](error, stackTrace, this);
      }
      [_handleDone]() {
        this[_stream][_handleDone](this);
      }
    }
    dart.setSignature(_ForwardingStreamSubscription, {
      constructors: () => ({_ForwardingStreamSubscription: [_ForwardingStreamSubscription$(S, T), [_ForwardingStream$(S, T), dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]}),
      methods: () => ({
        [_add]: [dart.void, [T]],
        [_handleData]: [dart.void, [S]],
        [_handleError]: [dart.void, [dart.dynamic, core.StackTrace]],
        [_handleDone]: [dart.void, []]
      })
    });
    return _ForwardingStreamSubscription;
  });
  let _ForwardingStreamSubscription = _ForwardingStreamSubscription$();
  const _Predicate$ = dart.generic(function(T) {
    const _Predicate = dart.typedef('_Predicate', () => dart.functionType(core.bool, [T]));
    return _Predicate;
  });
  let _Predicate = _Predicate$();
  function _addErrorWithReplacement(sink, error, stackTrace) {
    let replacement = Zone.current.errorCallback(error, dart.as(stackTrace, core.StackTrace));
    if (replacement != null) {
      error = _nonNullError(replacement.error);
      stackTrace = replacement.stackTrace;
    }
    sink[_addError](error, dart.as(stackTrace, core.StackTrace));
  }
  dart.fn(_addErrorWithReplacement, dart.void, [_EventSink, dart.dynamic, dart.dynamic]);
  const _test = Symbol('_test');
  const _WhereStream$ = dart.generic(function(T) {
    class _WhereStream extends _ForwardingStream$(T, T) {
      _WhereStream(source, test) {
        this[_test] = test;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        let satisfies = null;
        try {
          satisfies = this[_test](inputEvent);
        } catch (e) {
          let s = dart.stackTrace(e);
          _addErrorWithReplacement(sink, e, s);
          return;
        }

        if (dart.notNull(satisfies)) {
          sink[_add](inputEvent);
        }
      }
    }
    dart.setSignature(_WhereStream, {
      constructors: () => ({_WhereStream: [_WhereStream$(T), [Stream$(T), dart.functionType(core.bool, [T])]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _WhereStream;
  });
  let _WhereStream = _WhereStream$();
  const _Transformation$ = dart.generic(function(S, T) {
    const _Transformation = dart.typedef('_Transformation', () => dart.functionType(T, [S]));
    return _Transformation;
  });
  let _Transformation = _Transformation$();
  const _transform = Symbol('_transform');
  const _MapStream$ = dart.generic(function(S, T) {
    class _MapStream extends _ForwardingStream$(S, T) {
      _MapStream(source, transform) {
        this[_transform] = transform;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, S);
        dart.as(sink, _EventSink$(T));
        let outputEvent = null;
        try {
          outputEvent = dart.as(dart.dcall(this[_transform], inputEvent), T);
        } catch (e) {
          let s = dart.stackTrace(e);
          _addErrorWithReplacement(sink, e, s);
          return;
        }

        sink[_add](outputEvent);
      }
    }
    dart.setSignature(_MapStream, {
      constructors: () => ({_MapStream: [_MapStream$(S, T), [Stream$(S), dart.functionType(T, [S])]]}),
      methods: () => ({[_handleData]: [dart.void, [S, _EventSink$(T)]]})
    });
    return _MapStream;
  });
  let _MapStream = _MapStream$();
  const _expand = Symbol('_expand');
  const _ExpandStream$ = dart.generic(function(S, T) {
    class _ExpandStream extends _ForwardingStream$(S, T) {
      _ExpandStream(source, expand) {
        this[_expand] = expand;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, S);
        dart.as(sink, _EventSink$(T));
        try {
          for (let value of this[_expand](inputEvent)) {
            sink[_add](value);
          }
        } catch (e) {
          let s = dart.stackTrace(e);
          _addErrorWithReplacement(sink, e, s);
        }

      }
    }
    dart.setSignature(_ExpandStream, {
      constructors: () => ({_ExpandStream: [_ExpandStream$(S, T), [Stream$(S), dart.functionType(core.Iterable$(T), [S])]]}),
      methods: () => ({[_handleData]: [dart.void, [S, _EventSink$(T)]]})
    });
    return _ExpandStream;
  });
  let _ExpandStream = _ExpandStream$();
  const _ErrorTest = dart.typedef('_ErrorTest', () => dart.functionType(core.bool, [dart.dynamic]));
  const _HandleErrorStream$ = dart.generic(function(T) {
    class _HandleErrorStream extends _ForwardingStream$(T, T) {
      _HandleErrorStream(source, onError, test) {
        this[_transform] = onError;
        this[_test] = test;
        super._ForwardingStream(source);
      }
      [_handleError](error, stackTrace, sink) {
        dart.as(sink, _EventSink$(T));
        let matches = true;
        if (this[_test] != null) {
          try {
            matches = dart.dcall(this[_test], error);
          } catch (e) {
            let s = dart.stackTrace(e);
            _addErrorWithReplacement(sink, e, s);
            return;
          }

        }
        if (dart.notNull(matches)) {
          try {
            _invokeErrorHandler(this[_transform], error, stackTrace);
          } catch (e) {
            let s = dart.stackTrace(e);
            if (dart.notNull(core.identical(e, error))) {
              sink[_addError](error, stackTrace);
            } else {
              _addErrorWithReplacement(sink, e, s);
            }
            return;
          }

        } else {
          sink[_addError](error, stackTrace);
        }
      }
    }
    dart.setSignature(_HandleErrorStream, {
      constructors: () => ({_HandleErrorStream: [_HandleErrorStream$(T), [Stream$(T), core.Function, dart.functionType(core.bool, [dart.dynamic])]]}),
      methods: () => ({[_handleError]: [dart.void, [core.Object, core.StackTrace, _EventSink$(T)]]})
    });
    return _HandleErrorStream;
  });
  let _HandleErrorStream = _HandleErrorStream$();
  const _remaining = Symbol('_remaining');
  const _TakeStream$ = dart.generic(function(T) {
    class _TakeStream extends _ForwardingStream$(T, T) {
      _TakeStream(source, count) {
        this[_remaining] = count;
        super._ForwardingStream(source);
        if (!(typeof count == 'number'))
          dart.throw(new core.ArgumentError(count));
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        if (dart.notNull(this[_remaining]) > 0) {
          sink[_add](inputEvent);
          this[_remaining] = dart.notNull(this[_remaining]) - 1;
          if (this[_remaining] == 0) {
            sink[_close]();
          }
        }
      }
    }
    dart.setSignature(_TakeStream, {
      constructors: () => ({_TakeStream: [_TakeStream$(T), [Stream$(T), core.int]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _TakeStream;
  });
  let _TakeStream = _TakeStream$();
  const _TakeWhileStream$ = dart.generic(function(T) {
    class _TakeWhileStream extends _ForwardingStream$(T, T) {
      _TakeWhileStream(source, test) {
        this[_test] = test;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        let satisfies = null;
        try {
          satisfies = this[_test](inputEvent);
        } catch (e) {
          let s = dart.stackTrace(e);
          _addErrorWithReplacement(sink, e, s);
          sink[_close]();
          return;
        }

        if (dart.notNull(satisfies)) {
          sink[_add](inputEvent);
        } else {
          sink[_close]();
        }
      }
    }
    dart.setSignature(_TakeWhileStream, {
      constructors: () => ({_TakeWhileStream: [_TakeWhileStream$(T), [Stream$(T), dart.functionType(core.bool, [T])]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _TakeWhileStream;
  });
  let _TakeWhileStream = _TakeWhileStream$();
  const _SkipStream$ = dart.generic(function(T) {
    class _SkipStream extends _ForwardingStream$(T, T) {
      _SkipStream(source, count) {
        this[_remaining] = count;
        super._ForwardingStream(source);
        if (!(typeof count == 'number') || dart.notNull(count) < 0)
          dart.throw(new core.ArgumentError(count));
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        if (dart.notNull(this[_remaining]) > 0) {
          this[_remaining] = dart.notNull(this[_remaining]) - 1;
          return;
        }
        sink[_add](inputEvent);
      }
    }
    dart.setSignature(_SkipStream, {
      constructors: () => ({_SkipStream: [_SkipStream$(T), [Stream$(T), core.int]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _SkipStream;
  });
  let _SkipStream = _SkipStream$();
  const _hasFailed = Symbol('_hasFailed');
  const _SkipWhileStream$ = dart.generic(function(T) {
    class _SkipWhileStream extends _ForwardingStream$(T, T) {
      _SkipWhileStream(source, test) {
        this[_test] = test;
        this[_hasFailed] = false;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        if (dart.notNull(this[_hasFailed])) {
          sink[_add](inputEvent);
          return;
        }
        let satisfies = null;
        try {
          satisfies = this[_test](inputEvent);
        } catch (e) {
          let s = dart.stackTrace(e);
          _addErrorWithReplacement(sink, e, s);
          this[_hasFailed] = true;
          return;
        }

        if (!dart.notNull(satisfies)) {
          this[_hasFailed] = true;
          sink[_add](inputEvent);
        }
      }
    }
    dart.setSignature(_SkipWhileStream, {
      constructors: () => ({_SkipWhileStream: [_SkipWhileStream$(T), [Stream$(T), dart.functionType(core.bool, [T])]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _SkipWhileStream;
  });
  let _SkipWhileStream = _SkipWhileStream$();
  const _Equality$ = dart.generic(function(T) {
    const _Equality = dart.typedef('_Equality', () => dart.functionType(core.bool, [T, T]));
    return _Equality;
  });
  let _Equality = _Equality$();
  const _equals = Symbol('_equals');
  const _DistinctStream$ = dart.generic(function(T) {
    class _DistinctStream extends _ForwardingStream$(T, T) {
      _DistinctStream(source, equals) {
        this[_previous] = _DistinctStream$()._SENTINEL;
        this[_equals] = equals;
        super._ForwardingStream(source);
      }
      [_handleData](inputEvent, sink) {
        dart.as(inputEvent, T);
        dart.as(sink, _EventSink$(T));
        if (dart.notNull(core.identical(this[_previous], _DistinctStream$()._SENTINEL))) {
          this[_previous] = inputEvent;
          return sink[_add](inputEvent);
        } else {
          let isEqual = null;
          try {
            if (this[_equals] == null) {
              isEqual = dart.equals(this[_previous], inputEvent);
            } else {
              isEqual = this[_equals](dart.as(this[_previous], T), inputEvent);
            }
          } catch (e) {
            let s = dart.stackTrace(e);
            _addErrorWithReplacement(sink, e, s);
            return null;
          }

          if (!dart.notNull(isEqual)) {
            sink[_add](inputEvent);
            this[_previous] = inputEvent;
          }
        }
      }
    }
    dart.setSignature(_DistinctStream, {
      constructors: () => ({_DistinctStream: [_DistinctStream$(T), [Stream$(T), dart.functionType(core.bool, [T, T])]]}),
      methods: () => ({[_handleData]: [dart.void, [T, _EventSink$(T)]]})
    });
    return _DistinctStream;
  });
  let _DistinctStream = _DistinctStream$();
  dart.defineLazyProperties(_DistinctStream, {
    get _SENTINEL() {
      return new core.Object();
    },
    set _SENTINEL(_) {}
  });
  const _EventSinkWrapper$ = dart.generic(function(T) {
    class _EventSinkWrapper extends core.Object {
      _EventSinkWrapper(sink) {
        this[_sink] = sink;
      }
      add(data) {
        dart.as(data, T);
        this[_sink][_add](data);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        this[_sink][_addError](error, stackTrace);
      }
      close() {
        this[_sink][_close]();
      }
    }
    _EventSinkWrapper[dart.implements] = () => [EventSink$(T)];
    dart.setSignature(_EventSinkWrapper, {
      constructors: () => ({_EventSinkWrapper: [_EventSinkWrapper$(T), [_EventSink]]}),
      methods: () => ({
        add: [dart.void, [T]],
        addError: [dart.void, [dart.dynamic], [core.StackTrace]],
        close: [dart.void, []]
      })
    });
    return _EventSinkWrapper;
  });
  let _EventSinkWrapper = _EventSinkWrapper$();
  const _transformerSink = Symbol('_transformerSink');
  const _isSubscribed = Symbol('_isSubscribed');
  const _SinkTransformerStreamSubscription$ = dart.generic(function(S, T) {
    class _SinkTransformerStreamSubscription extends _BufferingStreamSubscription$(T) {
      _SinkTransformerStreamSubscription(source, mapper, onData, onError, onDone, cancelOnError) {
        this[_transformerSink] = null;
        this[_subscription] = null;
        super._BufferingStreamSubscription(onData, onError, onDone, cancelOnError);
        let eventSink = new (_EventSinkWrapper$(T))(this);
        this[_transformerSink] = mapper(eventSink);
        this[_subscription] = source.listen(dart.bind(this, _handleData), {onError: dart.bind(this, _handleError), onDone: dart.bind(this, _handleDone)});
      }
      get [_isSubscribed]() {
        return this[_subscription] != null;
      }
      [_add](data) {
        dart.as(data, T);
        if (dart.notNull(this[_isClosed])) {
          dart.throw(new core.StateError("Stream is already closed"));
        }
        super[_add](data);
      }
      [_addError](error, stackTrace) {
        if (dart.notNull(this[_isClosed])) {
          dart.throw(new core.StateError("Stream is already closed"));
        }
        super[_addError](error, stackTrace);
      }
      [_close]() {
        if (dart.notNull(this[_isClosed])) {
          dart.throw(new core.StateError("Stream is already closed"));
        }
        super[_close]();
      }
      [_onPause]() {
        if (dart.notNull(this[_isSubscribed]))
          this[_subscription].pause();
      }
      [_onResume]() {
        if (dart.notNull(this[_isSubscribed]))
          this[_subscription].resume();
      }
      [_onCancel]() {
        if (dart.notNull(this[_isSubscribed])) {
          let subscription = this[_subscription];
          this[_subscription] = null;
          subscription.cancel();
        }
        return null;
      }
      [_handleData](data) {
        dart.as(data, S);
        try {
          this[_transformerSink].add(data);
        } catch (e) {
          let s = dart.stackTrace(e);
          this[_addError](e, s);
        }

      }
      [_handleError](error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        try {
          this[_transformerSink].addError(error, dart.as(stackTrace, core.StackTrace));
        } catch (e) {
          let s = dart.stackTrace(e);
          if (dart.notNull(core.identical(e, error))) {
            this[_addError](error, dart.as(stackTrace, core.StackTrace));
          } else {
            this[_addError](e, s);
          }
        }

      }
      [_handleDone]() {
        try {
          this[_subscription] = null;
          this[_transformerSink].close();
        } catch (e) {
          let s = dart.stackTrace(e);
          this[_addError](e, s);
        }

      }
    }
    dart.setSignature(_SinkTransformerStreamSubscription, {
      constructors: () => ({_SinkTransformerStreamSubscription: [_SinkTransformerStreamSubscription$(S, T), [Stream$(S), _SinkMapper$(S, T), dart.functionType(dart.void, [T]), core.Function, dart.functionType(dart.void, []), core.bool]]}),
      methods: () => ({
        [_add]: [dart.void, [T]],
        [_handleData]: [dart.void, [S]],
        [_handleError]: [dart.void, [dart.dynamic], [dart.dynamic]],
        [_handleDone]: [dart.void, []]
      })
    });
    return _SinkTransformerStreamSubscription;
  });
  let _SinkTransformerStreamSubscription = _SinkTransformerStreamSubscription$();
  const _SinkMapper$ = dart.generic(function(S, T) {
    const _SinkMapper = dart.typedef('_SinkMapper', () => dart.functionType(EventSink$(S), [EventSink$(T)]));
    return _SinkMapper;
  });
  let _SinkMapper = _SinkMapper$();
  const _sinkMapper = Symbol('_sinkMapper');
  const _StreamSinkTransformer$ = dart.generic(function(S, T) {
    class _StreamSinkTransformer extends core.Object {
      _StreamSinkTransformer(sinkMapper) {
        this[_sinkMapper] = sinkMapper;
      }
      bind(stream) {
        dart.as(stream, Stream$(S));
        return new (_BoundSinkStream$(S, T))(stream, this[_sinkMapper]);
      }
    }
    _StreamSinkTransformer[dart.implements] = () => [StreamTransformer$(S, T)];
    dart.setSignature(_StreamSinkTransformer, {
      constructors: () => ({_StreamSinkTransformer: [_StreamSinkTransformer$(S, T), [_SinkMapper$(S, T)]]}),
      methods: () => ({bind: [Stream$(T), [Stream$(S)]]})
    });
    return _StreamSinkTransformer;
  });
  let _StreamSinkTransformer = _StreamSinkTransformer$();
  const _BoundSinkStream$ = dart.generic(function(S, T) {
    class _BoundSinkStream extends Stream$(T) {
      get isBroadcast() {
        return this[_stream].isBroadcast;
      }
      _BoundSinkStream(stream, sinkMapper) {
        this[_stream] = stream;
        this[_sinkMapper] = sinkMapper;
        super.Stream();
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        cancelOnError = core.identical(true, cancelOnError);
        let subscription = new (_SinkTransformerStreamSubscription$(dart.dynamic, T))(this[_stream], dart.as(this[_sinkMapper], _SinkMapper), onData, onError, onDone, cancelOnError);
        return subscription;
      }
    }
    dart.setSignature(_BoundSinkStream, {
      constructors: () => ({_BoundSinkStream: [_BoundSinkStream$(S, T), [Stream$(S), _SinkMapper$(S, T)]]}),
      methods: () => ({listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}]})
    });
    return _BoundSinkStream;
  });
  let _BoundSinkStream = _BoundSinkStream$();
  const _TransformDataHandler$ = dart.generic(function(S, T) {
    const _TransformDataHandler = dart.typedef('_TransformDataHandler', () => dart.functionType(dart.void, [S, EventSink$(T)]));
    return _TransformDataHandler;
  });
  let _TransformDataHandler = _TransformDataHandler$();
  const _TransformErrorHandler$ = dart.generic(function(T) {
    const _TransformErrorHandler = dart.typedef('_TransformErrorHandler', () => dart.functionType(dart.void, [core.Object, core.StackTrace, EventSink$(T)]));
    return _TransformErrorHandler;
  });
  let _TransformErrorHandler = _TransformErrorHandler$();
  const _TransformDoneHandler$ = dart.generic(function(T) {
    const _TransformDoneHandler = dart.typedef('_TransformDoneHandler', () => dart.functionType(dart.void, [EventSink$(T)]));
    return _TransformDoneHandler;
  });
  let _TransformDoneHandler = _TransformDoneHandler$();
  const _HandlerEventSink$ = dart.generic(function(S, T) {
    class _HandlerEventSink extends core.Object {
      _HandlerEventSink(handleData, handleError, handleDone, sink) {
        this[_handleData] = handleData;
        this[_handleError] = handleError;
        this[_handleDone] = handleDone;
        this[_sink] = sink;
      }
      add(data) {
        dart.as(data, S);
        return this[_handleData](data, this[_sink]);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        return this[_handleError](error, stackTrace, this[_sink]);
      }
      close() {
        return this[_handleDone](this[_sink]);
      }
    }
    _HandlerEventSink[dart.implements] = () => [EventSink$(S)];
    dart.setSignature(_HandlerEventSink, {
      constructors: () => ({_HandlerEventSink: [_HandlerEventSink$(S, T), [_TransformDataHandler$(S, T), _TransformErrorHandler$(T), _TransformDoneHandler$(T), EventSink$(T)]]}),
      methods: () => ({
        add: [dart.void, [S]],
        addError: [dart.void, [core.Object], [core.StackTrace]],
        close: [dart.void, []]
      })
    });
    return _HandlerEventSink;
  });
  let _HandlerEventSink = _HandlerEventSink$();
  const _StreamHandlerTransformer$ = dart.generic(function(S, T) {
    class _StreamHandlerTransformer extends _StreamSinkTransformer$(S, T) {
      _StreamHandlerTransformer(opts) {
        let handleData = opts && 'handleData' in opts ? opts.handleData : null;
        let handleError = opts && 'handleError' in opts ? opts.handleError : null;
        let handleDone = opts && 'handleDone' in opts ? opts.handleDone : null;
        super._StreamSinkTransformer(dart.as(dart.fn(outputSink => {
          dart.as(outputSink, EventSink$(T));
          if (handleData == null)
            handleData = dart.as(_StreamHandlerTransformer$()._defaultHandleData, __CastType27$(S, T));
          if (handleError == null)
            handleError = dart.as(_StreamHandlerTransformer$()._defaultHandleError, __CastType30$(T));
          if (handleDone == null)
            handleDone = _StreamHandlerTransformer$()._defaultHandleDone;
          return new (_HandlerEventSink$(S, T))(handleData, handleError, handleDone, outputSink);
        }, dart.dynamic, [EventSink$(T)]), _SinkMapper$(S, T)));
      }
      bind(stream) {
        dart.as(stream, Stream$(S));
        return super.bind(stream);
      }
      static _defaultHandleData(data, sink) {
        sink.add(data);
      }
      static _defaultHandleError(error, stackTrace, sink) {
        sink.addError(error);
      }
      static _defaultHandleDone(sink) {
        sink.close();
      }
    }
    dart.setSignature(_StreamHandlerTransformer, {
      constructors: () => ({_StreamHandlerTransformer: [_StreamHandlerTransformer$(S, T), [], {handleData: dart.functionType(dart.void, [S, EventSink$(T)]), handleError: dart.functionType(dart.void, [core.Object, core.StackTrace, EventSink$(T)]), handleDone: dart.functionType(dart.void, [EventSink$(T)])}]}),
      methods: () => ({bind: [Stream$(T), [Stream$(S)]]}),
      statics: () => ({
        _defaultHandleData: [dart.void, [dart.dynamic, EventSink]],
        _defaultHandleError: [dart.void, [dart.dynamic, core.StackTrace, EventSink]],
        _defaultHandleDone: [dart.void, [EventSink]]
      }),
      names: ['_defaultHandleData', '_defaultHandleError', '_defaultHandleDone']
    });
    return _StreamHandlerTransformer;
  });
  let _StreamHandlerTransformer = _StreamHandlerTransformer$();
  const _SubscriptionTransformer$ = dart.generic(function(S, T) {
    const _SubscriptionTransformer = dart.typedef('_SubscriptionTransformer', () => dart.functionType(StreamSubscription$(T), [Stream$(S), core.bool]));
    return _SubscriptionTransformer;
  });
  let _SubscriptionTransformer = _SubscriptionTransformer$();
  const _transformer = Symbol('_transformer');
  const _StreamSubscriptionTransformer$ = dart.generic(function(S, T) {
    class _StreamSubscriptionTransformer extends core.Object {
      _StreamSubscriptionTransformer(transformer) {
        this[_transformer] = transformer;
      }
      bind(stream) {
        dart.as(stream, Stream$(S));
        return new (_BoundSubscriptionStream$(S, T))(stream, this[_transformer]);
      }
    }
    _StreamSubscriptionTransformer[dart.implements] = () => [StreamTransformer$(S, T)];
    dart.setSignature(_StreamSubscriptionTransformer, {
      constructors: () => ({_StreamSubscriptionTransformer: [_StreamSubscriptionTransformer$(S, T), [_SubscriptionTransformer$(S, T)]]}),
      methods: () => ({bind: [Stream$(T), [Stream$(S)]]})
    });
    return _StreamSubscriptionTransformer;
  });
  let _StreamSubscriptionTransformer = _StreamSubscriptionTransformer$();
  const _BoundSubscriptionStream$ = dart.generic(function(S, T) {
    class _BoundSubscriptionStream extends Stream$(T) {
      _BoundSubscriptionStream(stream, transformer) {
        this[_stream] = stream;
        this[_transformer] = transformer;
        super.Stream();
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        cancelOnError = core.identical(true, cancelOnError);
        let result = this[_transformer](this[_stream], cancelOnError);
        result.onData(onData);
        result.onError(onError);
        result.onDone(onDone);
        return result;
      }
    }
    dart.setSignature(_BoundSubscriptionStream, {
      constructors: () => ({_BoundSubscriptionStream: [_BoundSubscriptionStream$(S, T), [Stream$(S), _SubscriptionTransformer$(S, T)]]}),
      methods: () => ({listen: [StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}]})
    });
    return _BoundSubscriptionStream;
  });
  let _BoundSubscriptionStream = _BoundSubscriptionStream$();
  const __CastType27$ = dart.generic(function(S, T) {
    const __CastType27 = dart.typedef('__CastType27', () => dart.functionType(dart.void, [S, EventSink$(T)]));
    return __CastType27;
  });
  let __CastType27 = __CastType27$();
  const __CastType30$ = dart.generic(function(T) {
    const __CastType30 = dart.typedef('__CastType30', () => dart.functionType(dart.void, [core.Object, core.StackTrace, EventSink$(T)]));
    return __CastType30;
  });
  let __CastType30 = __CastType30$();
  class Timer extends core.Object {
    static new(duration, callback) {
      if (dart.equals(Zone.current, Zone.ROOT)) {
        return Zone.current.createTimer(duration, callback);
      }
      return Zone.current.createTimer(duration, Zone.current.bindCallback(callback, {runGuarded: true}));
    }
    static periodic(duration, callback) {
      if (dart.equals(Zone.current, Zone.ROOT)) {
        return Zone.current.createPeriodicTimer(duration, callback);
      }
      return Zone.current.createPeriodicTimer(duration, dart.as(Zone.current.bindUnaryCallback(callback, {runGuarded: true}), __CastType34));
    }
    static run(callback) {
      Timer.new(core.Duration.ZERO, callback);
    }
    static _createTimer(duration, callback) {
      let milliseconds = duration.inMilliseconds;
      if (dart.notNull(milliseconds) < 0)
        milliseconds = 0;
      return new _isolate_helper.TimerImpl(milliseconds, callback);
    }
    static _createPeriodicTimer(duration, callback) {
      let milliseconds = duration.inMilliseconds;
      if (dart.notNull(milliseconds) < 0)
        milliseconds = 0;
      return new _isolate_helper.TimerImpl.periodic(milliseconds, callback);
    }
  }
  dart.setSignature(Timer, {
    constructors: () => ({
      new: [Timer, [core.Duration, dart.functionType(dart.void, [])]],
      periodic: [Timer, [core.Duration, dart.functionType(dart.void, [Timer])]]
    }),
    statics: () => ({
      run: [dart.void, [dart.functionType(dart.void, [])]],
      _createTimer: [Timer, [core.Duration, dart.functionType(dart.void, [])]],
      _createPeriodicTimer: [Timer, [core.Duration, dart.functionType(dart.void, [Timer])]]
    }),
    names: ['run', '_createTimer', '_createPeriodicTimer']
  });
  const __CastType34 = dart.typedef('__CastType34', () => dart.functionType(dart.void, [Timer]));
  const ZoneCallback = dart.typedef('ZoneCallback', () => dart.functionType(dart.dynamic, []));
  const ZoneUnaryCallback = dart.typedef('ZoneUnaryCallback', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const ZoneBinaryCallback = dart.typedef('ZoneBinaryCallback', () => dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
  const HandleUncaughtErrorHandler = dart.typedef('HandleUncaughtErrorHandler', () => dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.dynamic, core.StackTrace]));
  const RunHandler = dart.typedef('RunHandler', () => dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]));
  const RunUnaryHandler = dart.typedef('RunUnaryHandler', () => dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]));
  const RunBinaryHandler = dart.typedef('RunBinaryHandler', () => dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]));
  const RegisterCallbackHandler = dart.typedef('RegisterCallbackHandler', () => dart.functionType(ZoneCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]));
  const RegisterUnaryCallbackHandler = dart.typedef('RegisterUnaryCallbackHandler', () => dart.functionType(ZoneUnaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic])]));
  const RegisterBinaryCallbackHandler = dart.typedef('RegisterBinaryCallbackHandler', () => dart.functionType(ZoneBinaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]));
  const ErrorCallbackHandler = dart.typedef('ErrorCallbackHandler', () => dart.functionType(AsyncError, [Zone, ZoneDelegate, Zone, core.Object, core.StackTrace]));
  const ScheduleMicrotaskHandler = dart.typedef('ScheduleMicrotaskHandler', () => dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]));
  const CreateTimerHandler = dart.typedef('CreateTimerHandler', () => dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [])]));
  const CreatePeriodicTimerHandler = dart.typedef('CreatePeriodicTimerHandler', () => dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [Timer])]));
  const PrintHandler = dart.typedef('PrintHandler', () => dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, core.String]));
  const ForkHandler = dart.typedef('ForkHandler', () => dart.functionType(Zone, [Zone, ZoneDelegate, Zone, ZoneSpecification, core.Map]));
  class _ZoneFunction extends core.Object {
    _ZoneFunction(zone, func) {
      this.zone = zone;
      this.function = func;
    }
  }
  dart.setSignature(_ZoneFunction, {
    constructors: () => ({_ZoneFunction: [_ZoneFunction, [_Zone, core.Function]]})
  });
  class ZoneSpecification extends core.Object {
    static new(opts) {
      return new _ZoneSpecification(opts);
    }
    static from(other, opts) {
      let handleUncaughtError = opts && 'handleUncaughtError' in opts ? opts.handleUncaughtError : null;
      let run = opts && 'run' in opts ? opts.run : null;
      let runUnary = opts && 'runUnary' in opts ? opts.runUnary : null;
      let runBinary = opts && 'runBinary' in opts ? opts.runBinary : null;
      let registerCallback = opts && 'registerCallback' in opts ? opts.registerCallback : null;
      let registerUnaryCallback = opts && 'registerUnaryCallback' in opts ? opts.registerUnaryCallback : null;
      let registerBinaryCallback = opts && 'registerBinaryCallback' in opts ? opts.registerBinaryCallback : null;
      let errorCallback = opts && 'errorCallback' in opts ? opts.errorCallback : null;
      let scheduleMicrotask = opts && 'scheduleMicrotask' in opts ? opts.scheduleMicrotask : null;
      let createTimer = opts && 'createTimer' in opts ? opts.createTimer : null;
      let createPeriodicTimer = opts && 'createPeriodicTimer' in opts ? opts.createPeriodicTimer : null;
      let print = opts && 'print' in opts ? opts.print : null;
      let fork = opts && 'fork' in opts ? opts.fork : null;
      return ZoneSpecification.new({handleUncaughtError: handleUncaughtError != null ? handleUncaughtError : other.handleUncaughtError, run: run != null ? run : other.run, runUnary: runUnary != null ? runUnary : other.runUnary, runBinary: runBinary != null ? runBinary : other.runBinary, registerCallback: registerCallback != null ? registerCallback : other.registerCallback, registerUnaryCallback: registerUnaryCallback != null ? registerUnaryCallback : other.registerUnaryCallback, registerBinaryCallback: registerBinaryCallback != null ? registerBinaryCallback : other.registerBinaryCallback, errorCallback: errorCallback != null ? errorCallback : other.errorCallback, scheduleMicrotask: scheduleMicrotask != null ? scheduleMicrotask : other.scheduleMicrotask, createTimer: createTimer != null ? createTimer : other.createTimer, createPeriodicTimer: createPeriodicTimer != null ? createPeriodicTimer : other.createPeriodicTimer, print: print != null ? print : other.print, fork: fork != null ? fork : other.fork});
    }
  }
  dart.setSignature(ZoneSpecification, {
    constructors: () => ({
      new: [ZoneSpecification, [], {handleUncaughtError: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.dynamic, core.StackTrace]), run: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), runUnary: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]), runBinary: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]), registerCallback: dart.functionType(ZoneCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), registerUnaryCallback: dart.functionType(ZoneUnaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic])]), registerBinaryCallback: dart.functionType(ZoneBinaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]), errorCallback: dart.functionType(AsyncError, [Zone, ZoneDelegate, Zone, core.Object, core.StackTrace]), scheduleMicrotask: dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), createTimer: dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [])]), createPeriodicTimer: dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [Timer])]), print: dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, core.String]), fork: dart.functionType(Zone, [Zone, ZoneDelegate, Zone, ZoneSpecification, core.Map])}],
      from: [ZoneSpecification, [ZoneSpecification], {handleUncaughtError: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.dynamic, core.StackTrace]), run: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), runUnary: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]), runBinary: dart.functionType(dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]), registerCallback: dart.functionType(ZoneCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), registerUnaryCallback: dart.functionType(ZoneUnaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic])]), registerBinaryCallback: dart.functionType(ZoneBinaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]), errorCallback: dart.functionType(AsyncError, [Zone, ZoneDelegate, Zone, core.Object, core.StackTrace]), scheduleMicrotask: dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]), createTimer: dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [])]), createPeriodicTimer: dart.functionType(Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [Timer])]), print: dart.functionType(dart.void, [Zone, ZoneDelegate, Zone, core.String]), fork: dart.functionType(Zone, [Zone, ZoneDelegate, Zone, ZoneSpecification, core.Map])}]
    })
  });
  class _ZoneSpecification extends core.Object {
    _ZoneSpecification(opts) {
      let handleUncaughtError = opts && 'handleUncaughtError' in opts ? opts.handleUncaughtError : null;
      let run = opts && 'run' in opts ? opts.run : null;
      let runUnary = opts && 'runUnary' in opts ? opts.runUnary : null;
      let runBinary = opts && 'runBinary' in opts ? opts.runBinary : null;
      let registerCallback = opts && 'registerCallback' in opts ? opts.registerCallback : null;
      let registerUnaryCallback = opts && 'registerUnaryCallback' in opts ? opts.registerUnaryCallback : null;
      let registerBinaryCallback = opts && 'registerBinaryCallback' in opts ? opts.registerBinaryCallback : null;
      let errorCallback = opts && 'errorCallback' in opts ? opts.errorCallback : null;
      let scheduleMicrotask = opts && 'scheduleMicrotask' in opts ? opts.scheduleMicrotask : null;
      let createTimer = opts && 'createTimer' in opts ? opts.createTimer : null;
      let createPeriodicTimer = opts && 'createPeriodicTimer' in opts ? opts.createPeriodicTimer : null;
      let print = opts && 'print' in opts ? opts.print : null;
      let fork = opts && 'fork' in opts ? opts.fork : null;
      this.handleUncaughtError = handleUncaughtError;
      this.run = run;
      this.runUnary = runUnary;
      this.runBinary = runBinary;
      this.registerCallback = registerCallback;
      this.registerUnaryCallback = registerUnaryCallback;
      this.registerBinaryCallback = registerBinaryCallback;
      this.errorCallback = errorCallback;
      this.scheduleMicrotask = scheduleMicrotask;
      this.createTimer = createTimer;
      this.createPeriodicTimer = createPeriodicTimer;
      this.print = print;
      this.fork = fork;
    }
  }
  _ZoneSpecification[dart.implements] = () => [ZoneSpecification];
  dart.setSignature(_ZoneSpecification, {
    constructors: () => ({_ZoneSpecification: [_ZoneSpecification, [], {handleUncaughtError: HandleUncaughtErrorHandler, run: RunHandler, runUnary: RunUnaryHandler, runBinary: RunBinaryHandler, registerCallback: RegisterCallbackHandler, registerUnaryCallback: RegisterUnaryCallbackHandler, registerBinaryCallback: RegisterBinaryCallbackHandler, errorCallback: ErrorCallbackHandler, scheduleMicrotask: ScheduleMicrotaskHandler, createTimer: CreateTimerHandler, createPeriodicTimer: CreatePeriodicTimerHandler, print: PrintHandler, fork: ForkHandler}]})
  });
  class ZoneDelegate extends core.Object {}
  class Zone extends core.Object {
    _() {
    }
    static get current() {
      return Zone._current;
    }
    static _enter(zone) {
      dart.assert(zone != null);
      dart.assert(!dart.notNull(core.identical(zone, Zone._current)));
      let previous = Zone._current;
      Zone._current = zone;
      return previous;
    }
    static _leave(previous) {
      dart.assert(previous != null);
      Zone._current = previous;
    }
  }
  dart.defineNamedConstructor(Zone, '_');
  dart.setSignature(Zone, {
    constructors: () => ({_: [Zone, []]}),
    statics: () => ({
      _enter: [Zone, [Zone]],
      _leave: [dart.void, [Zone]]
    }),
    names: ['_enter', '_leave']
  });
  class _Zone extends core.Object {
    _Zone() {
    }
    inSameErrorZone(otherZone) {
      return dart.notNull(core.identical(this, otherZone)) || dart.notNull(core.identical(this.errorZone, otherZone.errorZone));
    }
  }
  _Zone[dart.implements] = () => [Zone];
  dart.setSignature(_Zone, {
    constructors: () => ({_Zone: [_Zone, []]}),
    methods: () => ({inSameErrorZone: [core.bool, [Zone]]})
  });
  const _run = Symbol('_run');
  const _runUnary = Symbol('_runUnary');
  const _runBinary = Symbol('_runBinary');
  const _registerCallback = Symbol('_registerCallback');
  const _registerUnaryCallback = Symbol('_registerUnaryCallback');
  const _registerBinaryCallback = Symbol('_registerBinaryCallback');
  const _errorCallback = Symbol('_errorCallback');
  const _scheduleMicrotask = Symbol('_scheduleMicrotask');
  const _createTimer = Symbol('_createTimer');
  const _createPeriodicTimer = Symbol('_createPeriodicTimer');
  const _print = Symbol('_print');
  const _fork = Symbol('_fork');
  const _handleUncaughtError = Symbol('_handleUncaughtError');
  const _map = Symbol('_map');
  const _delegate = Symbol('_delegate');
  class _RootZone extends _Zone {
    _RootZone() {
      super._Zone();
    }
    get [_run]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRun));
    }
    get [_runUnary]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRunUnary));
    }
    get [_runBinary]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRunBinary));
    }
    get [_registerCallback]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRegisterCallback));
    }
    get [_registerUnaryCallback]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRegisterUnaryCallback));
    }
    get [_registerBinaryCallback]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootRegisterBinaryCallback));
    }
    get [_errorCallback]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootErrorCallback));
    }
    get [_scheduleMicrotask]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootScheduleMicrotask));
    }
    get [_createTimer]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootCreateTimer));
    }
    get [_createPeriodicTimer]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootCreatePeriodicTimer));
    }
    get [_print]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootPrint));
    }
    get [_fork]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootFork));
    }
    get [_handleUncaughtError]() {
      return dart.const(new _ZoneFunction(_ROOT_ZONE, _rootHandleUncaughtError));
    }
    get parent() {
      return null;
    }
    get [_map]() {
      return _RootZone._rootMap;
    }
    get [_delegate]() {
      if (_RootZone._rootDelegate != null)
        return _RootZone._rootDelegate;
      return _RootZone._rootDelegate = new _ZoneDelegate(this);
    }
    get errorZone() {
      return this;
    }
    runGuarded(f) {
      try {
        if (dart.notNull(core.identical(_ROOT_ZONE, Zone._current))) {
          return f();
        }
        return _rootRun(null, null, this, f);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    runUnaryGuarded(f, arg) {
      try {
        if (dart.notNull(core.identical(_ROOT_ZONE, Zone._current))) {
          return dart.dcall(f, arg);
        }
        return _rootRunUnary(null, null, this, f, arg);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    runBinaryGuarded(f, arg1, arg2) {
      try {
        if (dart.notNull(core.identical(_ROOT_ZONE, Zone._current))) {
          return dart.dcall(f, arg1, arg2);
        }
        return _rootRunBinary(null, null, this, f, arg1, arg2);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    bindCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      if (dart.notNull(runGuarded)) {
        return dart.fn((() => this.runGuarded(f)).bind(this));
      } else {
        return dart.fn((() => this.run(f)).bind(this));
      }
    }
    bindUnaryCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      if (dart.notNull(runGuarded)) {
        return dart.fn((arg => this.runUnaryGuarded(f, arg)).bind(this));
      } else {
        return dart.fn((arg => this.runUnary(f, arg)).bind(this));
      }
    }
    bindBinaryCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      if (dart.notNull(runGuarded)) {
        return dart.fn(((arg1, arg2) => this.runBinaryGuarded(f, arg1, arg2)).bind(this));
      } else {
        return dart.fn(((arg1, arg2) => this.runBinary(f, arg1, arg2)).bind(this));
      }
    }
    get(key) {
      return null;
    }
    handleUncaughtError(error, stackTrace) {
      return _rootHandleUncaughtError(null, null, this, error, stackTrace);
    }
    fork(opts) {
      let specification = opts && 'specification' in opts ? opts.specification : null;
      let zoneValues = opts && 'zoneValues' in opts ? opts.zoneValues : null;
      return _rootFork(null, null, this, specification, zoneValues);
    }
    run(f) {
      if (dart.notNull(core.identical(Zone._current, _ROOT_ZONE)))
        return f();
      return _rootRun(null, null, this, f);
    }
    runUnary(f, arg) {
      if (dart.notNull(core.identical(Zone._current, _ROOT_ZONE)))
        return dart.dcall(f, arg);
      return _rootRunUnary(null, null, this, f, arg);
    }
    runBinary(f, arg1, arg2) {
      if (dart.notNull(core.identical(Zone._current, _ROOT_ZONE)))
        return dart.dcall(f, arg1, arg2);
      return _rootRunBinary(null, null, this, f, arg1, arg2);
    }
    registerCallback(f) {
      return f;
    }
    registerUnaryCallback(f) {
      return f;
    }
    registerBinaryCallback(f) {
      return f;
    }
    errorCallback(error, stackTrace) {
      return null;
    }
    scheduleMicrotask(f) {
      _rootScheduleMicrotask(null, null, this, f);
    }
    createTimer(duration, f) {
      return Timer._createTimer(duration, f);
    }
    createPeriodicTimer(duration, f) {
      return Timer._createPeriodicTimer(duration, f);
    }
    print(line) {
      _internal.printToConsole(line);
    }
  }
  dart.setSignature(_RootZone, {
    constructors: () => ({_RootZone: [_RootZone, []]}),
    methods: () => ({
      runGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [])]],
      runUnaryGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]],
      runBinaryGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]],
      bindCallback: [ZoneCallback, [dart.functionType(dart.dynamic, [])], {runGuarded: core.bool}],
      bindUnaryCallback: [ZoneUnaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic])], {runGuarded: core.bool}],
      bindBinaryCallback: [ZoneBinaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])], {runGuarded: core.bool}],
      get: [dart.dynamic, [core.Object]],
      handleUncaughtError: [dart.dynamic, [dart.dynamic, core.StackTrace]],
      fork: [Zone, [], {specification: ZoneSpecification, zoneValues: core.Map}],
      run: [dart.dynamic, [dart.functionType(dart.dynamic, [])]],
      runUnary: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]],
      runBinary: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]],
      registerCallback: [ZoneCallback, [dart.functionType(dart.dynamic, [])]],
      registerUnaryCallback: [ZoneUnaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic])]],
      registerBinaryCallback: [ZoneBinaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]],
      errorCallback: [AsyncError, [core.Object, core.StackTrace]],
      scheduleMicrotask: [dart.void, [dart.functionType(dart.void, [])]],
      createTimer: [Timer, [core.Duration, dart.functionType(dart.void, [])]],
      createPeriodicTimer: [Timer, [core.Duration, dart.functionType(dart.void, [Timer])]],
      print: [dart.void, [core.String]]
    })
  });
  const _ROOT_ZONE = dart.const(new _RootZone());
  Zone.ROOT = _ROOT_ZONE;
  Zone._current = _ROOT_ZONE;
  function _parentDelegate(zone) {
    if (zone.parent == null)
      return null;
    return zone.parent[_delegate];
  }
  dart.fn(_parentDelegate, ZoneDelegate, [_Zone]);
  const _delegationTarget = Symbol('_delegationTarget');
  class _ZoneDelegate extends core.Object {
    _ZoneDelegate(delegationTarget) {
      this[_delegationTarget] = delegationTarget;
    }
    handleUncaughtError(zone, error, stackTrace) {
      let implementation = this[_delegationTarget][_handleUncaughtError];
      let implZone = implementation.zone;
      return dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, error, stackTrace);
    }
    run(zone, f) {
      let implementation = this[_delegationTarget][_run];
      let implZone = implementation.zone;
      return dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f);
    }
    runUnary(zone, f, arg) {
      let implementation = this[_delegationTarget][_runUnary];
      let implZone = implementation.zone;
      return dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f, arg);
    }
    runBinary(zone, f, arg1, arg2) {
      let implementation = this[_delegationTarget][_runBinary];
      let implZone = implementation.zone;
      return dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f, arg1, arg2);
    }
    registerCallback(zone, f) {
      let implementation = this[_delegationTarget][_registerCallback];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f), ZoneCallback);
    }
    registerUnaryCallback(zone, f) {
      let implementation = this[_delegationTarget][_registerUnaryCallback];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f), ZoneUnaryCallback);
    }
    registerBinaryCallback(zone, f) {
      let implementation = this[_delegationTarget][_registerBinaryCallback];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f), ZoneBinaryCallback);
    }
    errorCallback(zone, error, stackTrace) {
      let implementation = this[_delegationTarget][_errorCallback];
      let implZone = implementation.zone;
      if (dart.notNull(core.identical(implZone, _ROOT_ZONE)))
        return null;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, error, stackTrace), AsyncError);
    }
    scheduleMicrotask(zone, f) {
      let implementation = this[_delegationTarget][_scheduleMicrotask];
      let implZone = implementation.zone;
      dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, f);
    }
    createTimer(zone, duration, f) {
      let implementation = this[_delegationTarget][_createTimer];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, duration, f), Timer);
    }
    createPeriodicTimer(zone, period, f) {
      let implementation = this[_delegationTarget][_createPeriodicTimer];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, period, f), Timer);
    }
    print(zone, line) {
      let implementation = this[_delegationTarget][_print];
      let implZone = implementation.zone;
      dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, line);
    }
    fork(zone, specification, zoneValues) {
      let implementation = this[_delegationTarget][_fork];
      let implZone = implementation.zone;
      return dart.as(dart.dcall(implementation.function, implZone, _parentDelegate(implZone), zone, specification, zoneValues), Zone);
    }
  }
  _ZoneDelegate[dart.implements] = () => [ZoneDelegate];
  dart.setSignature(_ZoneDelegate, {
    constructors: () => ({_ZoneDelegate: [_ZoneDelegate, [_Zone]]}),
    methods: () => ({
      handleUncaughtError: [dart.dynamic, [Zone, dart.dynamic, core.StackTrace]],
      run: [dart.dynamic, [Zone, dart.functionType(dart.dynamic, [])]],
      runUnary: [dart.dynamic, [Zone, dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]],
      runBinary: [dart.dynamic, [Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]],
      registerCallback: [ZoneCallback, [Zone, dart.functionType(dart.dynamic, [])]],
      registerUnaryCallback: [ZoneUnaryCallback, [Zone, dart.functionType(dart.dynamic, [dart.dynamic])]],
      registerBinaryCallback: [ZoneBinaryCallback, [Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]],
      errorCallback: [AsyncError, [Zone, core.Object, core.StackTrace]],
      scheduleMicrotask: [dart.void, [Zone, dart.functionType(dart.dynamic, [])]],
      createTimer: [Timer, [Zone, core.Duration, dart.functionType(dart.void, [])]],
      createPeriodicTimer: [Timer, [Zone, core.Duration, dart.functionType(dart.void, [Timer])]],
      print: [dart.void, [Zone, core.String]],
      fork: [Zone, [Zone, ZoneSpecification, core.Map]]
    })
  });
  const _delegateCache = Symbol('_delegateCache');
  class _CustomZone extends _Zone {
    get [_delegate]() {
      if (this[_delegateCache] != null)
        return this[_delegateCache];
      this[_delegateCache] = new _ZoneDelegate(this);
      return this[_delegateCache];
    }
    _CustomZone(parent, specification, map) {
      this.parent = parent;
      this[_map] = map;
      this[_runUnary] = null;
      this[_run] = null;
      this[_runBinary] = null;
      this[_registerCallback] = null;
      this[_registerUnaryCallback] = null;
      this[_registerBinaryCallback] = null;
      this[_errorCallback] = null;
      this[_scheduleMicrotask] = null;
      this[_createTimer] = null;
      this[_createPeriodicTimer] = null;
      this[_print] = null;
      this[_fork] = null;
      this[_handleUncaughtError] = null;
      this[_delegateCache] = null;
      super._Zone();
      this[_run] = specification.run != null ? new _ZoneFunction(this, specification.run) : this.parent[_run];
      this[_runUnary] = specification.runUnary != null ? new _ZoneFunction(this, specification.runUnary) : this.parent[_runUnary];
      this[_runBinary] = specification.runBinary != null ? new _ZoneFunction(this, specification.runBinary) : this.parent[_runBinary];
      this[_registerCallback] = specification.registerCallback != null ? new _ZoneFunction(this, specification.registerCallback) : this.parent[_registerCallback];
      this[_registerUnaryCallback] = specification.registerUnaryCallback != null ? new _ZoneFunction(this, specification.registerUnaryCallback) : this.parent[_registerUnaryCallback];
      this[_registerBinaryCallback] = specification.registerBinaryCallback != null ? new _ZoneFunction(this, specification.registerBinaryCallback) : this.parent[_registerBinaryCallback];
      this[_errorCallback] = specification.errorCallback != null ? new _ZoneFunction(this, specification.errorCallback) : this.parent[_errorCallback];
      this[_scheduleMicrotask] = specification.scheduleMicrotask != null ? new _ZoneFunction(this, specification.scheduleMicrotask) : this.parent[_scheduleMicrotask];
      this[_createTimer] = specification.createTimer != null ? new _ZoneFunction(this, specification.createTimer) : this.parent[_createTimer];
      this[_createPeriodicTimer] = specification.createPeriodicTimer != null ? new _ZoneFunction(this, specification.createPeriodicTimer) : this.parent[_createPeriodicTimer];
      this[_print] = specification.print != null ? new _ZoneFunction(this, specification.print) : this.parent[_print];
      this[_fork] = specification.fork != null ? new _ZoneFunction(this, specification.fork) : this.parent[_fork];
      this[_handleUncaughtError] = specification.handleUncaughtError != null ? new _ZoneFunction(this, specification.handleUncaughtError) : this.parent[_handleUncaughtError];
    }
    get errorZone() {
      return this[_handleUncaughtError].zone;
    }
    runGuarded(f) {
      try {
        return this.run(f);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    runUnaryGuarded(f, arg) {
      try {
        return this.runUnary(f, arg);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    runBinaryGuarded(f, arg1, arg2) {
      try {
        return this.runBinary(f, arg1, arg2);
      } catch (e) {
        let s = dart.stackTrace(e);
        return this.handleUncaughtError(e, s);
      }

    }
    bindCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      let registered = this.registerCallback(f);
      if (dart.notNull(runGuarded)) {
        return dart.fn((() => this.runGuarded(registered)).bind(this));
      } else {
        return dart.fn((() => this.run(registered)).bind(this));
      }
    }
    bindUnaryCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      let registered = this.registerUnaryCallback(f);
      if (dart.notNull(runGuarded)) {
        return dart.fn((arg => this.runUnaryGuarded(registered, arg)).bind(this));
      } else {
        return dart.fn((arg => this.runUnary(registered, arg)).bind(this));
      }
    }
    bindBinaryCallback(f, opts) {
      let runGuarded = opts && 'runGuarded' in opts ? opts.runGuarded : true;
      let registered = this.registerBinaryCallback(f);
      if (dart.notNull(runGuarded)) {
        return dart.fn(((arg1, arg2) => this.runBinaryGuarded(registered, arg1, arg2)).bind(this));
      } else {
        return dart.fn(((arg1, arg2) => this.runBinary(registered, arg1, arg2)).bind(this));
      }
    }
    get(key) {
      let result = this[_map].get(key);
      if (result != null || dart.notNull(this[_map].containsKey(key)))
        return result;
      if (this.parent != null) {
        let value = this.parent.get(key);
        if (value != null) {
          this[_map].set(key, value);
        }
        return value;
      }
      dart.assert(dart.equals(this, _ROOT_ZONE));
      return null;
    }
    handleUncaughtError(error, stackTrace) {
      let implementation = this[_handleUncaughtError];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, error, stackTrace);
    }
    fork(opts) {
      let specification = opts && 'specification' in opts ? opts.specification : null;
      let zoneValues = opts && 'zoneValues' in opts ? opts.zoneValues : null;
      let implementation = this[_fork];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, specification, zoneValues), Zone);
    }
    run(f) {
      let implementation = this[_run];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f);
    }
    runUnary(f, arg) {
      let implementation = this[_runUnary];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f, arg);
    }
    runBinary(f, arg1, arg2) {
      let implementation = this[_runBinary];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f, arg1, arg2);
    }
    registerCallback(f) {
      let implementation = this[_registerCallback];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f), ZoneCallback);
    }
    registerUnaryCallback(f) {
      let implementation = this[_registerUnaryCallback];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f), ZoneUnaryCallback);
    }
    registerBinaryCallback(f) {
      let implementation = this[_registerBinaryCallback];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f), ZoneBinaryCallback);
    }
    errorCallback(error, stackTrace) {
      let implementation = this[_errorCallback];
      dart.assert(implementation != null);
      let implementationZone = implementation.zone;
      if (dart.notNull(core.identical(implementationZone, _ROOT_ZONE)))
        return null;
      let parentDelegate = _parentDelegate(dart.as(implementationZone, _Zone));
      return dart.as(dart.dcall(implementation.function, implementationZone, parentDelegate, this, error, stackTrace), AsyncError);
    }
    scheduleMicrotask(f) {
      let implementation = this[_scheduleMicrotask];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, f);
    }
    createTimer(duration, f) {
      let implementation = this[_createTimer];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, duration, f), Timer);
    }
    createPeriodicTimer(duration, f) {
      let implementation = this[_createPeriodicTimer];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.as(dart.dcall(implementation.function, implementation.zone, parentDelegate, this, duration, f), Timer);
    }
    print(line) {
      let implementation = this[_print];
      dart.assert(implementation != null);
      let parentDelegate = _parentDelegate(implementation.zone);
      return dart.dcall(implementation.function, implementation.zone, parentDelegate, this, line);
    }
  }
  dart.setSignature(_CustomZone, {
    constructors: () => ({_CustomZone: [_CustomZone, [_Zone, ZoneSpecification, core.Map]]}),
    methods: () => ({
      runGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [])]],
      runUnaryGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]],
      runBinaryGuarded: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]],
      bindCallback: [ZoneCallback, [dart.functionType(dart.dynamic, [])], {runGuarded: core.bool}],
      bindUnaryCallback: [ZoneUnaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic])], {runGuarded: core.bool}],
      bindBinaryCallback: [ZoneBinaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])], {runGuarded: core.bool}],
      get: [dart.dynamic, [core.Object]],
      handleUncaughtError: [dart.dynamic, [dart.dynamic, core.StackTrace]],
      fork: [Zone, [], {specification: ZoneSpecification, zoneValues: core.Map}],
      run: [dart.dynamic, [dart.functionType(dart.dynamic, [])]],
      runUnary: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]],
      runBinary: [dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]],
      registerCallback: [ZoneCallback, [dart.functionType(dart.dynamic, [])]],
      registerUnaryCallback: [ZoneUnaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic])]],
      registerBinaryCallback: [ZoneBinaryCallback, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]],
      errorCallback: [AsyncError, [core.Object, core.StackTrace]],
      scheduleMicrotask: [dart.void, [dart.functionType(dart.void, [])]],
      createTimer: [Timer, [core.Duration, dart.functionType(dart.void, [])]],
      createPeriodicTimer: [Timer, [core.Duration, dart.functionType(dart.void, [Timer])]],
      print: [dart.void, [core.String]]
    })
  });
  function _rootHandleUncaughtError(self, parent, zone, error, stackTrace) {
    _schedulePriorityAsyncCallback(dart.fn(() => {
      dart.throw(new _UncaughtAsyncError(error, stackTrace));
    }));
  }
  dart.fn(_rootHandleUncaughtError, dart.void, [Zone, ZoneDelegate, Zone, dart.dynamic, core.StackTrace]);
  function _rootRun(self, parent, zone, f) {
    if (dart.equals(Zone._current, zone))
      return f();
    let old = Zone._enter(zone);
    try {
      return f();
    } finally {
      Zone._leave(old);
    }
  }
  dart.fn(_rootRun, dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]);
  function _rootRunUnary(self, parent, zone, f, arg) {
    if (dart.equals(Zone._current, zone))
      return dart.dcall(f, arg);
    let old = Zone._enter(zone);
    try {
      return dart.dcall(f, arg);
    } finally {
      Zone._leave(old);
    }
  }
  dart.fn(_rootRunUnary, dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic]), dart.dynamic]);
  function _rootRunBinary(self, parent, zone, f, arg1, arg2) {
    if (dart.equals(Zone._current, zone))
      return dart.dcall(f, arg1, arg2);
    let old = Zone._enter(zone);
    try {
      return dart.dcall(f, arg1, arg2);
    } finally {
      Zone._leave(old);
    }
  }
  dart.fn(_rootRunBinary, dart.dynamic, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), dart.dynamic, dart.dynamic]);
  function _rootRegisterCallback(self, parent, zone, f) {
    return f;
  }
  dart.fn(_rootRegisterCallback, ZoneCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]);
  function _rootRegisterUnaryCallback(self, parent, zone, f) {
    return f;
  }
  dart.fn(_rootRegisterUnaryCallback, ZoneUnaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic])]);
  function _rootRegisterBinaryCallback(self, parent, zone, f) {
    return f;
  }
  dart.fn(_rootRegisterBinaryCallback, ZoneBinaryCallback, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]);
  function _rootErrorCallback(self, parent, zone, error, stackTrace) {
    return null;
  }
  dart.fn(_rootErrorCallback, AsyncError, [Zone, ZoneDelegate, Zone, core.Object, core.StackTrace]);
  function _rootScheduleMicrotask(self, parent, zone, f) {
    if (!dart.notNull(core.identical(_ROOT_ZONE, zone))) {
      let hasErrorHandler = !dart.notNull(_ROOT_ZONE.inSameErrorZone(zone));
      f = zone.bindCallback(f, {runGuarded: hasErrorHandler});
    }
    _scheduleAsyncCallback(f);
  }
  dart.fn(_rootScheduleMicrotask, dart.void, [Zone, ZoneDelegate, Zone, dart.functionType(dart.dynamic, [])]);
  function _rootCreateTimer(self, parent, zone, duration, callback) {
    if (!dart.notNull(core.identical(_ROOT_ZONE, zone))) {
      callback = zone.bindCallback(callback);
    }
    return Timer._createTimer(duration, callback);
  }
  dart.fn(_rootCreateTimer, Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [])]);
  function _rootCreatePeriodicTimer(self, parent, zone, duration, callback) {
    if (!dart.notNull(core.identical(_ROOT_ZONE, zone))) {
      callback = dart.as(zone.bindUnaryCallback(callback), __CastType36);
    }
    return Timer._createPeriodicTimer(duration, callback);
  }
  dart.fn(_rootCreatePeriodicTimer, Timer, [Zone, ZoneDelegate, Zone, core.Duration, dart.functionType(dart.void, [Timer])]);
  function _rootPrint(self, parent, zone, line) {
    _internal.printToConsole(line);
  }
  dart.fn(_rootPrint, dart.void, [Zone, ZoneDelegate, Zone, core.String]);
  function _printToZone(line) {
    Zone.current.print(line);
  }
  dart.fn(_printToZone, dart.void, [core.String]);
  function _rootFork(self, parent, zone, specification, zoneValues) {
    _internal.printToZone = _printToZone;
    if (specification == null) {
      specification = dart.const(ZoneSpecification.new());
    } else if (!dart.is(specification, _ZoneSpecification)) {
      dart.throw(new core.ArgumentError("ZoneSpecifications must be instantiated" + " with the provided constructor."));
    }
    let valueMap = null;
    if (zoneValues == null) {
      if (dart.is(zone, _Zone)) {
        valueMap = zone[_map];
      } else {
        valueMap = collection.HashMap.new();
      }
    } else {
      valueMap = collection.HashMap.from(zoneValues);
    }
    return new _CustomZone(dart.as(zone, _Zone), specification, valueMap);
  }
  dart.fn(_rootFork, Zone, [Zone, ZoneDelegate, Zone, ZoneSpecification, core.Map]);
  class _RootZoneSpecification extends core.Object {
    get handleUncaughtError() {
      return _rootHandleUncaughtError;
    }
    get run() {
      return _rootRun;
    }
    get runUnary() {
      return _rootRunUnary;
    }
    get runBinary() {
      return _rootRunBinary;
    }
    get registerCallback() {
      return _rootRegisterCallback;
    }
    get registerUnaryCallback() {
      return _rootRegisterUnaryCallback;
    }
    get registerBinaryCallback() {
      return _rootRegisterBinaryCallback;
    }
    get errorCallback() {
      return _rootErrorCallback;
    }
    get scheduleMicrotask() {
      return _rootScheduleMicrotask;
    }
    get createTimer() {
      return _rootCreateTimer;
    }
    get createPeriodicTimer() {
      return _rootCreatePeriodicTimer;
    }
    get print() {
      return _rootPrint;
    }
    get fork() {
      return _rootFork;
    }
  }
  _RootZoneSpecification[dart.implements] = () => [ZoneSpecification];
  _RootZone._rootDelegate = null;
  dart.defineLazyProperties(_RootZone, {
    get _rootMap() {
      return collection.HashMap.new();
    },
    set _rootMap(_) {}
  });
  function runZoned(body, opts) {
    let zoneValues = opts && 'zoneValues' in opts ? opts.zoneValues : null;
    let zoneSpecification = opts && 'zoneSpecification' in opts ? opts.zoneSpecification : null;
    let onError = opts && 'onError' in opts ? opts.onError : null;
    let errorHandler = null;
    if (onError != null) {
      errorHandler = dart.fn((self, parent, zone, error, stackTrace) => {
        try {
          if (dart.is(onError, ZoneBinaryCallback)) {
            return self.parent.runBinary(onError, error, stackTrace);
          }
          return self.parent.runUnary(dart.as(onError, __CastType38), error);
        } catch (e) {
          let s = dart.stackTrace(e);
          if (dart.notNull(core.identical(e, error))) {
            return parent.handleUncaughtError(zone, error, stackTrace);
          } else {
            return parent.handleUncaughtError(zone, e, s);
          }
        }

      }, dart.dynamic, [Zone, ZoneDelegate, Zone, dart.dynamic, core.StackTrace]);
    }
    if (zoneSpecification == null) {
      zoneSpecification = ZoneSpecification.new({handleUncaughtError: errorHandler});
    } else if (errorHandler != null) {
      zoneSpecification = ZoneSpecification.from(zoneSpecification, {handleUncaughtError: errorHandler});
    }
    let zone = Zone.current.fork({specification: zoneSpecification, zoneValues: zoneValues});
    if (onError != null) {
      return zone.runGuarded(body);
    } else {
      return zone.run(body);
    }
  }
  dart.fn(runZoned, dart.dynamic, [dart.functionType(dart.dynamic, [])], {zoneValues: core.Map, zoneSpecification: ZoneSpecification, onError: core.Function});
  const __CastType36 = dart.typedef('__CastType36', () => dart.functionType(dart.void, [Timer]));
  const __CastType38 = dart.typedef('__CastType38', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  dart.copyProperties(exports, {
    get _hasDocument() {
      return typeof document == 'object';
    }
  });
  // Exports:
  exports.AsyncError = AsyncError;
  exports.Stream$ = Stream$;
  exports.Stream = Stream;
  exports.DeferredLibrary = DeferredLibrary;
  exports.DeferredLoadException = DeferredLoadException;
  exports.Future$ = Future$;
  exports.Future = Future;
  exports.TimeoutException = TimeoutException;
  exports.Completer$ = Completer$;
  exports.Completer = Completer;
  exports.scheduleMicrotask = scheduleMicrotask;
  exports.StreamSubscription$ = StreamSubscription$;
  exports.StreamSubscription = StreamSubscription;
  exports.EventSink$ = EventSink$;
  exports.EventSink = EventSink;
  exports.StreamView$ = StreamView$;
  exports.StreamView = StreamView;
  exports.StreamConsumer$ = StreamConsumer$;
  exports.StreamConsumer = StreamConsumer;
  exports.StreamSink$ = StreamSink$;
  exports.StreamSink = StreamSink;
  exports.StreamTransformer$ = StreamTransformer$;
  exports.StreamTransformer = StreamTransformer;
  exports.StreamIterator$ = StreamIterator$;
  exports.StreamIterator = StreamIterator;
  exports.StreamController$ = StreamController$;
  exports.StreamController = StreamController;
  exports.Timer = Timer;
  exports.ZoneCallback = ZoneCallback;
  exports.ZoneUnaryCallback = ZoneUnaryCallback;
  exports.ZoneBinaryCallback = ZoneBinaryCallback;
  exports.HandleUncaughtErrorHandler = HandleUncaughtErrorHandler;
  exports.RunHandler = RunHandler;
  exports.RunUnaryHandler = RunUnaryHandler;
  exports.RunBinaryHandler = RunBinaryHandler;
  exports.RegisterCallbackHandler = RegisterCallbackHandler;
  exports.RegisterUnaryCallbackHandler = RegisterUnaryCallbackHandler;
  exports.RegisterBinaryCallbackHandler = RegisterBinaryCallbackHandler;
  exports.ErrorCallbackHandler = ErrorCallbackHandler;
  exports.ScheduleMicrotaskHandler = ScheduleMicrotaskHandler;
  exports.CreateTimerHandler = CreateTimerHandler;
  exports.CreatePeriodicTimerHandler = CreatePeriodicTimerHandler;
  exports.PrintHandler = PrintHandler;
  exports.ForkHandler = ForkHandler;
  exports.ZoneSpecification = ZoneSpecification;
  exports.ZoneDelegate = ZoneDelegate;
  exports.Zone = Zone;
  exports.runZoned = runZoned;
});
dart_library.library('dart/collection', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
  'dart/_internal',
  'dart/_js_helper',
  'dart/math'
], function(exports, dart, core, _internal, _js_helper, math) {
  'use strict';
  let dartx = dart.dartx;
  const _source = Symbol('_source');
  const UnmodifiableListView$ = dart.generic(function(E) {
    class UnmodifiableListView extends _internal.UnmodifiableListBase$(E) {
      UnmodifiableListView(source) {
        this[_source] = source;
      }
      get length() {
        return this[_source][dartx.length];
      }
      get(index) {
        return this[_source][dartx.elementAt](index);
      }
    }
    dart.setSignature(UnmodifiableListView, {
      constructors: () => ({UnmodifiableListView: [exports.UnmodifiableListView$(E), [core.Iterable$(E)]]}),
      methods: () => ({get: [E, [core.int]]})
    });
    dart.defineExtensionMembers(UnmodifiableListView, ['get', 'length']);
    return UnmodifiableListView;
  });
  dart.defineLazyClassGeneric(exports, 'UnmodifiableListView', {get: UnmodifiableListView$});
  function _defaultEquals(a, b) {
    return dart.equals(a, b);
  }
  dart.fn(_defaultEquals, core.bool, [core.Object, core.Object]);
  function _defaultHashCode(a) {
    return dart.hashCode(a);
  }
  dart.fn(_defaultHashCode, core.int, [core.Object]);
  const _Equality$ = dart.generic(function(K) {
    const _Equality = dart.typedef('_Equality', () => dart.functionType(core.bool, [K, K]));
    return _Equality;
  });
  let _Equality = _Equality$();
  const _Hasher$ = dart.generic(function(K) {
    const _Hasher = dart.typedef('_Hasher', () => dart.functionType(core.int, [K]));
    return _Hasher;
  });
  let _Hasher = _Hasher$();
  const HashMap$ = dart.generic(function(K, V) {
    class HashMap extends core.Object {
      static new(opts) {
        let equals = opts && 'equals' in opts ? opts.equals : null;
        let hashCode = opts && 'hashCode' in opts ? opts.hashCode : null;
        let isValidKey = opts && 'isValidKey' in opts ? opts.isValidKey : null;
        if (isValidKey == null) {
          if (hashCode == null) {
            if (equals == null) {
              return new (_HashMap$(K, V))();
            }
            hashCode = _defaultHashCode;
          } else {
            if (dart.notNull(core.identical(core.identityHashCode, hashCode)) && dart.notNull(core.identical(core.identical, equals))) {
              return new (_IdentityHashMap$(K, V))();
            }
            if (equals == null) {
              equals = _defaultEquals;
            }
          }
        } else {
          if (hashCode == null) {
            hashCode = _defaultHashCode;
          }
          if (equals == null) {
            equals = _defaultEquals;
          }
        }
        return new (_CustomHashMap$(K, V))(equals, hashCode, isValidKey);
      }
      static identity() {
        return new (_IdentityHashMap$(K, V))();
      }
      static from(other) {
        let result = HashMap$(K, V).new();
        other.forEach(dart.fn((k, v) => {
          result.set(dart.as(k, K), dart.as(v, V));
        }));
        return result;
      }
      static fromIterable(iterable, opts) {
        let key = opts && 'key' in opts ? opts.key : null;
        let value = opts && 'value' in opts ? opts.value : null;
        let map = HashMap$(K, V).new();
        Maps._fillMapWithMappedIterable(map, iterable, key, value);
        return map;
      }
      static fromIterables(keys, values) {
        let map = HashMap$(K, V).new();
        Maps._fillMapWithIterables(map, keys, values);
        return map;
      }
    }
    HashMap[dart.implements] = () => [core.Map$(K, V)];
    dart.setSignature(HashMap, {
      constructors: () => ({
        new: [HashMap$(K, V), [], {equals: dart.functionType(core.bool, [K, K]), hashCode: dart.functionType(core.int, [K]), isValidKey: dart.functionType(core.bool, [core.Object])}],
        identity: [HashMap$(K, V), []],
        from: [HashMap$(K, V), [core.Map]],
        fromIterable: [HashMap$(K, V), [core.Iterable], {key: dart.functionType(K, [dart.dynamic]), value: dart.functionType(V, [dart.dynamic])}],
        fromIterables: [HashMap$(K, V), [core.Iterable$(K), core.Iterable$(V)]]
      })
    });
    return HashMap;
  });
  let HashMap = HashMap$();
  const SetMixin$ = dart.generic(function(E) {
    class SetMixin extends core.Object {
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
      get isEmpty() {
        return this.length == 0;
      }
      get isNotEmpty() {
        return this.length != 0;
      }
      clear() {
        this.removeAll(this.toList());
      }
      addAll(elements) {
        dart.as(elements, core.Iterable$(E));
        for (let element of elements)
          this.add(element);
      }
      removeAll(elements) {
        for (let element of elements)
          this.remove(element);
      }
      retainAll(elements) {
        let toRemove = this.toSet();
        for (let o of elements) {
          toRemove.remove(o);
        }
        this.removeAll(toRemove);
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let toRemove = [];
        for (let element of this) {
          if (dart.notNull(test(element)))
            toRemove[dartx.add](element);
        }
        this.removeAll(toRemove);
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let toRemove = [];
        for (let element of this) {
          if (!dart.notNull(test(element)))
            toRemove[dartx.add](element);
        }
        this.removeAll(toRemove);
      }
      containsAll(other) {
        for (let o of other) {
          if (!dart.notNull(this.contains(o)))
            return false;
        }
        return true;
      }
      union(other) {
        dart.as(other, core.Set$(E));
        let _ = this.toSet();
        _.addAll(other);
        return _;
      }
      intersection(other) {
        let result = this.toSet();
        for (let element of this) {
          if (!dart.notNull(other.contains(element)))
            result.remove(element);
        }
        return result;
      }
      difference(other) {
        let result = this.toSet();
        for (let element of this) {
          if (dart.notNull(other.contains(element)))
            result.remove(element);
        }
        return result;
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let result = dart.notNull(growable) ? (() => {
          let _ = core.List$(E).new();
          _[dartx.length] = this.length;
          return _;
        }).bind(this)() : core.List$(E).new(this.length);
        let i = 0;
        for (let element of this)
          result[dartx.set]((() => {
            let x = i;
            i = dart.notNull(x) + 1;
            return x;
          })(), element);
        return result;
      }
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return new (_internal.EfficientLengthMappedIterable$(E, dart.dynamic))(this, f);
      }
      get single() {
        if (dart.notNull(this.length) > 1)
          dart.throw(_internal.IterableElementError.tooMany());
        let it = this.iterator;
        if (!dart.notNull(it.moveNext()))
          dart.throw(_internal.IterableElementError.noElement());
        let result = it.current;
        return result;
      }
      toString() {
        return IterableBase.iterableToFullString(this, '{', '}');
      }
      where(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return new (_internal.WhereIterable$(E))(this, f);
      }
      expand(f) {
        dart.as(f, dart.functionType(core.Iterable, [E]));
        return new (_internal.ExpandIterable$(E, dart.dynamic))(this, f);
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        for (let element of this)
          f(element);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let value = iterator.current;
        while (dart.notNull(iterator.moveNext())) {
          value = combine(value, iterator.current);
        }
        return value;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        let value = initialValue;
        for (let element of this)
          value = dart.dcall(combine, value, element);
        return value;
      }
      every(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (!dart.notNull(f(element)))
            return false;
        }
        return true;
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext()))
          return "";
        let buffer = new core.StringBuffer();
        if (separator == null || separator == "") {
          do {
            buffer.write(`${iterator.current}`);
          } while (dart.notNull(iterator.moveNext()));
        } else {
          buffer.write(`${iterator.current}`);
          while (dart.notNull(iterator.moveNext())) {
            buffer.write(separator);
            buffer.write(`${iterator.current}`);
          }
        }
        return dart.toString(buffer);
      }
      any(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (dart.notNull(test(element)))
            return true;
        }
        return false;
      }
      take(n) {
        return _internal.TakeIterable$(E).new(this, n);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.TakeWhileIterable$(E))(this, test);
      }
      skip(n) {
        return _internal.SkipIterable$(E).new(this, n);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.SkipWhileIterable$(E))(this, test);
      }
      get first() {
        let it = this.iterator;
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        return it.current;
      }
      get last() {
        let it = this.iterator;
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let result = null;
        do {
          result = it.current;
        } while (dart.notNull(it.moveNext()));
        return result;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        for (let element of this) {
          if (dart.notNull(test(element)))
            return element;
        }
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            if (dart.notNull(foundMatching)) {
              dart.throw(_internal.IterableElementError.tooMany());
            }
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        dart.throw(_internal.IterableElementError.noElement());
      }
      elementAt(index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError.notNull("index"));
        core.RangeError.checkNotNegative(index, "index");
        let elementIndex = 0;
        for (let element of this) {
          if (index == elementIndex)
            return element;
          elementIndex = dart.notNull(elementIndex) + 1;
        }
        dart.throw(core.RangeError.index(index, this, "index", null, elementIndex));
      }
    }
    SetMixin[dart.implements] = () => [core.Set$(E)];
    dart.setSignature(SetMixin, {
      methods: () => ({
        clear: [dart.void, []],
        addAll: [dart.void, [core.Iterable$(E)]],
        removeAll: [dart.void, [core.Iterable$(core.Object)]],
        retainAll: [dart.void, [core.Iterable$(core.Object)]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        containsAll: [core.bool, [core.Iterable$(core.Object)]],
        union: [core.Set$(E), [core.Set$(E)]],
        intersection: [core.Set$(E), [core.Set$(core.Object)]],
        difference: [core.Set$(E), [core.Set$(core.Object)]],
        toList: [core.List$(E), [], {growable: core.bool}],
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        expand: [core.Iterable, [dart.functionType(core.Iterable, [E])]],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        reduce: [E, [dart.functionType(E, [E, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        join: [core.String, [], [core.String]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])]],
        elementAt: [E, [core.int]]
      })
    });
    dart.defineExtensionMembers(SetMixin, [
      'toList',
      'map',
      'where',
      'expand',
      'forEach',
      'reduce',
      'fold',
      'every',
      'join',
      'any',
      'take',
      'takeWhile',
      'skip',
      'skipWhile',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'elementAt',
      'isEmpty',
      'isNotEmpty',
      'single',
      'first',
      'last'
    ]);
    return SetMixin;
  });
  let SetMixin = SetMixin$();
  const SetBase$ = dart.generic(function(E) {
    class SetBase extends SetMixin$(E) {
      static setToString(set) {
        return IterableBase.iterableToFullString(set, '{', '}');
      }
    }
    dart.setSignature(SetBase, {
      statics: () => ({setToString: [core.String, [core.Set]]}),
      names: ['setToString']
    });
    return SetBase;
  });
  let SetBase = SetBase$();
  const _newSet = Symbol('_newSet');
  const _HashSetBase$ = dart.generic(function(E) {
    class _HashSetBase extends SetBase$(E) {
      difference(other) {
        let result = this[_newSet]();
        for (let element of this) {
          if (!dart.notNull(other.contains(element)))
            result.add(element);
        }
        return result;
      }
      intersection(other) {
        let result = this[_newSet]();
        for (let element of this) {
          if (dart.notNull(other.contains(element)))
            result.add(element);
        }
        return result;
      }
      toSet() {
        return (() => {
          let _ = this[_newSet]();
          _.addAll(this);
          return _;
        }).bind(this)();
      }
    }
    dart.setSignature(_HashSetBase, {
      methods: () => ({
        difference: [core.Set$(E), [core.Set$(core.Object)]],
        intersection: [core.Set$(E), [core.Set$(core.Object)]],
        toSet: [core.Set$(E), []]
      })
    });
    dart.defineExtensionMembers(_HashSetBase, ['toSet']);
    return _HashSetBase;
  });
  let _HashSetBase = _HashSetBase$();
  const HashSet$ = dart.generic(function(E) {
    class HashSet extends core.Object {
      static new(opts) {
        let equals = opts && 'equals' in opts ? opts.equals : null;
        let hashCode = opts && 'hashCode' in opts ? opts.hashCode : null;
        let isValidKey = opts && 'isValidKey' in opts ? opts.isValidKey : null;
        if (isValidKey == null) {
          if (hashCode == null) {
            if (equals == null) {
              return new (_HashSet$(E))();
            }
            hashCode = _defaultHashCode;
          } else {
            if (dart.notNull(core.identical(core.identityHashCode, hashCode)) && dart.notNull(core.identical(core.identical, equals))) {
              return new (_IdentityHashSet$(E))();
            }
            if (equals == null) {
              equals = _defaultEquals;
            }
          }
        } else {
          if (hashCode == null) {
            hashCode = _defaultHashCode;
          }
          if (equals == null) {
            equals = _defaultEquals;
          }
        }
        return new (_CustomHashSet$(E))(equals, hashCode, isValidKey);
      }
      static identity() {
        return new (_IdentityHashSet$(E))();
      }
      static from(elements) {
        let result = HashSet$(E).new();
        for (let e of dart.as(elements, core.Iterable$(E)))
          result.add(e);
        return result;
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
    }
    HashSet[dart.implements] = () => [core.Set$(E)];
    dart.setSignature(HashSet, {
      constructors: () => ({
        new: [HashSet$(E), [], {equals: dart.functionType(core.bool, [E, E]), hashCode: dart.functionType(core.int, [E]), isValidKey: dart.functionType(core.bool, [core.Object])}],
        identity: [HashSet$(E), []],
        from: [HashSet$(E), [core.Iterable]]
      })
    });
    return HashSet;
  });
  let HashSet = HashSet$();
  const IterableMixin$ = dart.generic(function(E) {
    class IterableMixin extends core.Object {
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return _internal.MappedIterable$(E, dart.dynamic).new(this, f);
      }
      where(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return new (_internal.WhereIterable$(E))(this, f);
      }
      expand(f) {
        dart.as(f, dart.functionType(core.Iterable, [E]));
        return new (_internal.ExpandIterable$(E, dart.dynamic))(this, f);
      }
      contains(element) {
        for (let e of this) {
          if (dart.equals(e, element))
            return true;
        }
        return false;
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        for (let element of this)
          f(element);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let value = iterator.current;
        while (dart.notNull(iterator.moveNext())) {
          value = combine(value, iterator.current);
        }
        return value;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        let value = initialValue;
        for (let element of this)
          value = dart.dcall(combine, value, element);
        return value;
      }
      every(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (!dart.notNull(f(element)))
            return false;
        }
        return true;
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext()))
          return "";
        let buffer = new core.StringBuffer();
        if (separator == null || separator == "") {
          do {
            buffer.write(`${iterator.current}`);
          } while (dart.notNull(iterator.moveNext()));
        } else {
          buffer.write(`${iterator.current}`);
          while (dart.notNull(iterator.moveNext())) {
            buffer.write(separator);
            buffer.write(`${iterator.current}`);
          }
        }
        return dart.toString(buffer);
      }
      any(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (dart.notNull(f(element)))
            return true;
        }
        return false;
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        return core.List$(E).from(this, {growable: growable});
      }
      toSet() {
        return core.Set$(E).from(this);
      }
      get length() {
        dart.assert(!dart.is(this, _internal.EfficientLength));
        let count = 0;
        let it = this[dartx.iterator];
        while (dart.notNull(it.moveNext())) {
          count = dart.notNull(count) + 1;
        }
        return count;
      }
      get isEmpty() {
        return !dart.notNull(this[dartx.iterator].moveNext());
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      take(n) {
        return _internal.TakeIterable$(E).new(this, n);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.TakeWhileIterable$(E))(this, test);
      }
      skip(n) {
        return _internal.SkipIterable$(E).new(this, n);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.SkipWhileIterable$(E))(this, test);
      }
      get first() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        return it.current;
      }
      get last() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let result = null;
        do {
          result = it.current;
        } while (dart.notNull(it.moveNext()));
        return result;
      }
      get single() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext()))
          dart.throw(_internal.IterableElementError.noElement());
        let result = it.current;
        if (dart.notNull(it.moveNext()))
          dart.throw(_internal.IterableElementError.tooMany());
        return result;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        for (let element of this) {
          if (dart.notNull(test(element)))
            return element;
        }
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            if (dart.notNull(foundMatching)) {
              dart.throw(_internal.IterableElementError.tooMany());
            }
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        dart.throw(_internal.IterableElementError.noElement());
      }
      elementAt(index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError.notNull("index"));
        core.RangeError.checkNotNegative(index, "index");
        let elementIndex = 0;
        for (let element of this) {
          if (index == elementIndex)
            return element;
          elementIndex = dart.notNull(elementIndex) + 1;
        }
        dart.throw(core.RangeError.index(index, this, "index", null, elementIndex));
      }
      toString() {
        return IterableBase.iterableToShortString(this, '(', ')');
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
    }
    IterableMixin[dart.implements] = () => [core.Iterable$(E)];
    dart.setSignature(IterableMixin, {
      methods: () => ({
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        expand: [core.Iterable, [dart.functionType(core.Iterable, [E])]],
        contains: [core.bool, [core.Object]],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        reduce: [E, [dart.functionType(E, [E, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        join: [core.String, [], [core.String]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        toList: [core.List$(E), [], {growable: core.bool}],
        toSet: [core.Set$(E), []],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])]],
        elementAt: [E, [core.int]]
      })
    });
    dart.defineExtensionMembers(IterableMixin, [
      'map',
      'where',
      'expand',
      'contains',
      'forEach',
      'reduce',
      'fold',
      'every',
      'join',
      'any',
      'toList',
      'toSet',
      'take',
      'takeWhile',
      'skip',
      'skipWhile',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'elementAt',
      'length',
      'isEmpty',
      'isNotEmpty',
      'first',
      'last',
      'single'
    ]);
    return IterableMixin;
  });
  let IterableMixin = IterableMixin$();
  const IterableBase$ = dart.generic(function(E) {
    class IterableBase extends core.Object {
      IterableBase() {
      }
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return _internal.MappedIterable$(E, dart.dynamic).new(this, f);
      }
      where(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        return new (_internal.WhereIterable$(E))(this, f);
      }
      expand(f) {
        dart.as(f, dart.functionType(core.Iterable, [E]));
        return new (_internal.ExpandIterable$(E, dart.dynamic))(this, f);
      }
      contains(element) {
        for (let e of this) {
          if (dart.equals(e, element))
            return true;
        }
        return false;
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        for (let element of this)
          f(element);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let value = iterator.current;
        while (dart.notNull(iterator.moveNext())) {
          value = combine(value, iterator.current);
        }
        return value;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        let value = initialValue;
        for (let element of this)
          value = dart.dcall(combine, value, element);
        return value;
      }
      every(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (!dart.notNull(f(element)))
            return false;
        }
        return true;
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        let iterator = this.iterator;
        if (!dart.notNull(iterator.moveNext()))
          return "";
        let buffer = new core.StringBuffer();
        if (separator == null || separator == "") {
          do {
            buffer.write(`${iterator.current}`);
          } while (dart.notNull(iterator.moveNext()));
        } else {
          buffer.write(`${iterator.current}`);
          while (dart.notNull(iterator.moveNext())) {
            buffer.write(separator);
            buffer.write(`${iterator.current}`);
          }
        }
        return dart.toString(buffer);
      }
      any(f) {
        dart.as(f, dart.functionType(core.bool, [E]));
        for (let element of this) {
          if (dart.notNull(f(element)))
            return true;
        }
        return false;
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        return core.List$(E).from(this, {growable: growable});
      }
      toSet() {
        return core.Set$(E).from(this);
      }
      get length() {
        dart.assert(!dart.is(this, _internal.EfficientLength));
        let count = 0;
        let it = this[dartx.iterator];
        while (dart.notNull(it.moveNext())) {
          count = dart.notNull(count) + 1;
        }
        return count;
      }
      get isEmpty() {
        return !dart.notNull(this[dartx.iterator].moveNext());
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      take(n) {
        return _internal.TakeIterable$(E).new(this, n);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.TakeWhileIterable$(E))(this, test);
      }
      skip(n) {
        return _internal.SkipIterable$(E).new(this, n);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.SkipWhileIterable$(E))(this, test);
      }
      get first() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        return it.current;
      }
      get last() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext())) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let result = null;
        do {
          result = it.current;
        } while (dart.notNull(it.moveNext()));
        return result;
      }
      get single() {
        let it = this[dartx.iterator];
        if (!dart.notNull(it.moveNext()))
          dart.throw(_internal.IterableElementError.noElement());
        let result = it.current;
        if (dart.notNull(it.moveNext()))
          dart.throw(_internal.IterableElementError.tooMany());
        return result;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        for (let element of this) {
          if (dart.notNull(test(element)))
            return element;
        }
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let result = null;
        let foundMatching = false;
        for (let element of this) {
          if (dart.notNull(test(element))) {
            if (dart.notNull(foundMatching)) {
              dart.throw(_internal.IterableElementError.tooMany());
            }
            result = element;
            foundMatching = true;
          }
        }
        if (dart.notNull(foundMatching))
          return result;
        dart.throw(_internal.IterableElementError.noElement());
      }
      elementAt(index) {
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError.notNull("index"));
        core.RangeError.checkNotNegative(index, "index");
        let elementIndex = 0;
        for (let element of this) {
          if (index == elementIndex)
            return element;
          elementIndex = dart.notNull(elementIndex) + 1;
        }
        dart.throw(core.RangeError.index(index, this, "index", null, elementIndex));
      }
      toString() {
        return IterableBase$().iterableToShortString(this, '(', ')');
      }
      static iterableToShortString(iterable, leftDelimiter, rightDelimiter) {
        if (leftDelimiter === void 0)
          leftDelimiter = '(';
        if (rightDelimiter === void 0)
          rightDelimiter = ')';
        if (dart.notNull(IterableBase$()._isToStringVisiting(iterable))) {
          if (leftDelimiter == "(" && rightDelimiter == ")") {
            return "(...)";
          }
          return `${leftDelimiter}...${rightDelimiter}`;
        }
        let parts = [];
        IterableBase$()._toStringVisiting[dartx.add](iterable);
        try {
          IterableBase$()._iterablePartsToStrings(iterable, parts);
        } finally {
          dart.assert(core.identical(IterableBase$()._toStringVisiting[dartx.last], iterable));
          IterableBase$()._toStringVisiting[dartx.removeLast]();
        }
        return dart.toString((() => {
          let _ = new core.StringBuffer(leftDelimiter);
          _.writeAll(parts, ", ");
          _.write(rightDelimiter);
          return _;
        })());
      }
      static iterableToFullString(iterable, leftDelimiter, rightDelimiter) {
        if (leftDelimiter === void 0)
          leftDelimiter = '(';
        if (rightDelimiter === void 0)
          rightDelimiter = ')';
        if (dart.notNull(IterableBase$()._isToStringVisiting(iterable))) {
          return `${leftDelimiter}...${rightDelimiter}`;
        }
        let buffer = new core.StringBuffer(leftDelimiter);
        IterableBase$()._toStringVisiting[dartx.add](iterable);
        try {
          buffer.writeAll(iterable, ", ");
        } finally {
          dart.assert(core.identical(IterableBase$()._toStringVisiting[dartx.last], iterable));
          IterableBase$()._toStringVisiting[dartx.removeLast]();
        }
        buffer.write(rightDelimiter);
        return dart.toString(buffer);
      }
      static _isToStringVisiting(o) {
        for (let i = 0; dart.notNull(i) < dart.notNull(IterableBase$()._toStringVisiting[dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.notNull(core.identical(o, IterableBase$()._toStringVisiting[dartx.get](i))))
            return true;
        }
        return false;
      }
      static _iterablePartsToStrings(iterable, parts) {
        let LENGTH_LIMIT = 80;
        let HEAD_COUNT = 3;
        let TAIL_COUNT = 2;
        let MAX_COUNT = 100;
        let OVERHEAD = 2;
        let ELLIPSIS_SIZE = 3;
        let length = 0;
        let count = 0;
        let it = iterable[dartx.iterator];
        while (dart.notNull(length) < dart.notNull(LENGTH_LIMIT) || dart.notNull(count) < dart.notNull(HEAD_COUNT)) {
          if (!dart.notNull(it.moveNext()))
            return;
          let next = `${it.current}`;
          parts[dartx.add](next);
          length = dart.notNull(length) + (dart.notNull(next[dartx.length]) + dart.notNull(OVERHEAD));
          count = dart.notNull(count) + 1;
        }
        let penultimateString = null;
        let ultimateString = null;
        let penultimate = null;
        let ultimate = null;
        if (!dart.notNull(it.moveNext())) {
          if (dart.notNull(count) <= dart.notNull(HEAD_COUNT) + dart.notNull(TAIL_COUNT))
            return;
          ultimateString = dart.as(parts[dartx.removeLast](), core.String);
          penultimateString = dart.as(parts[dartx.removeLast](), core.String);
        } else {
          penultimate = it.current;
          count = dart.notNull(count) + 1;
          if (!dart.notNull(it.moveNext())) {
            if (dart.notNull(count) <= dart.notNull(HEAD_COUNT) + 1) {
              parts[dartx.add](`${penultimate}`);
              return;
            }
            ultimateString = `${penultimate}`;
            penultimateString = dart.as(parts[dartx.removeLast](), core.String);
            length = dart.notNull(length) + (dart.notNull(ultimateString[dartx.length]) + dart.notNull(OVERHEAD));
          } else {
            ultimate = it.current;
            count = dart.notNull(count) + 1;
            dart.assert(dart.notNull(count) < dart.notNull(MAX_COUNT));
            while (dart.notNull(it.moveNext())) {
              penultimate = ultimate;
              ultimate = it.current;
              count = dart.notNull(count) + 1;
              if (dart.notNull(count) > dart.notNull(MAX_COUNT)) {
                while (dart.notNull(length) > dart.notNull(LENGTH_LIMIT) - dart.notNull(ELLIPSIS_SIZE) - dart.notNull(OVERHEAD) && dart.notNull(count) > dart.notNull(HEAD_COUNT)) {
                  length = dart.notNull(length) - dart.notNull(dart.as(dart.dsend(dart.dload(parts[dartx.removeLast](), 'length'), '+', OVERHEAD), core.int));
                  count = dart.notNull(count) - 1;
                }
                parts[dartx.add]("...");
                return;
              }
            }
            penultimateString = `${penultimate}`;
            ultimateString = `${ultimate}`;
            length = dart.notNull(length) + (dart.notNull(ultimateString[dartx.length]) + dart.notNull(penultimateString[dartx.length]) + 2 * dart.notNull(OVERHEAD));
          }
        }
        let elision = null;
        if (dart.notNull(count) > dart.notNull(parts[dartx.length]) + dart.notNull(TAIL_COUNT)) {
          elision = "...";
          length = dart.notNull(length) + (dart.notNull(ELLIPSIS_SIZE) + dart.notNull(OVERHEAD));
        }
        while (dart.notNull(length) > dart.notNull(LENGTH_LIMIT) && dart.notNull(parts[dartx.length]) > dart.notNull(HEAD_COUNT)) {
          length = dart.notNull(length) - dart.notNull(dart.as(dart.dsend(dart.dload(parts[dartx.removeLast](), 'length'), '+', OVERHEAD), core.int));
          if (elision == null) {
            elision = "...";
            length = dart.notNull(length) + (dart.notNull(ELLIPSIS_SIZE) + dart.notNull(OVERHEAD));
          }
        }
        if (elision != null) {
          parts[dartx.add](elision);
        }
        parts[dartx.add](penultimateString);
        parts[dartx.add](ultimateString);
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
    }
    IterableBase[dart.implements] = () => [core.Iterable$(E)];
    dart.setSignature(IterableBase, {
      constructors: () => ({IterableBase: [IterableBase$(E), []]}),
      methods: () => ({
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        expand: [core.Iterable, [dart.functionType(core.Iterable, [E])]],
        contains: [core.bool, [core.Object]],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        reduce: [E, [dart.functionType(E, [E, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        join: [core.String, [], [core.String]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        toList: [core.List$(E), [], {growable: core.bool}],
        toSet: [core.Set$(E), []],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])]],
        elementAt: [E, [core.int]]
      }),
      statics: () => ({
        iterableToShortString: [core.String, [core.Iterable], [core.String, core.String]],
        iterableToFullString: [core.String, [core.Iterable], [core.String, core.String]],
        _isToStringVisiting: [core.bool, [core.Object]],
        _iterablePartsToStrings: [dart.void, [core.Iterable, core.List]]
      }),
      names: ['iterableToShortString', 'iterableToFullString', '_isToStringVisiting', '_iterablePartsToStrings']
    });
    dart.defineExtensionMembers(IterableBase, [
      'map',
      'where',
      'expand',
      'contains',
      'forEach',
      'reduce',
      'fold',
      'every',
      'join',
      'any',
      'toList',
      'toSet',
      'take',
      'takeWhile',
      'skip',
      'skipWhile',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'elementAt',
      'length',
      'isEmpty',
      'isNotEmpty',
      'first',
      'last',
      'single'
    ]);
    return IterableBase;
  });
  let IterableBase = IterableBase$();
  dart.defineLazyProperties(IterableBase, {
    get _toStringVisiting() {
      return [];
    }
  });
  const _iterator = Symbol('_iterator');
  const _state = Symbol('_state');
  const _move = Symbol('_move');
  const HasNextIterator$ = dart.generic(function(E) {
    class HasNextIterator extends core.Object {
      HasNextIterator(iterator) {
        this[_iterator] = iterator;
        this[_state] = HasNextIterator$()._NOT_MOVED_YET;
      }
      get hasNext() {
        if (this[_state] == HasNextIterator$()._NOT_MOVED_YET)
          this[_move]();
        return this[_state] == HasNextIterator$()._HAS_NEXT_AND_NEXT_IN_CURRENT;
      }
      next() {
        if (!dart.notNull(this.hasNext))
          dart.throw(new core.StateError("No more elements"));
        dart.assert(this[_state] == HasNextIterator$()._HAS_NEXT_AND_NEXT_IN_CURRENT);
        let result = dart.as(this[_iterator].current, E);
        this[_move]();
        return result;
      }
      [_move]() {
        if (dart.notNull(this[_iterator].moveNext())) {
          this[_state] = HasNextIterator$()._HAS_NEXT_AND_NEXT_IN_CURRENT;
        } else {
          this[_state] = HasNextIterator$()._NO_NEXT;
        }
      }
    }
    dart.setSignature(HasNextIterator, {
      constructors: () => ({HasNextIterator: [HasNextIterator$(E), [core.Iterator]]}),
      methods: () => ({
        next: [E, []],
        [_move]: [dart.void, []]
      })
    });
    return HasNextIterator;
  });
  let HasNextIterator = HasNextIterator$();
  HasNextIterator._HAS_NEXT_AND_NEXT_IN_CURRENT = 0;
  HasNextIterator._NO_NEXT = 1;
  HasNextIterator._NOT_MOVED_YET = 2;
  const LinkedHashMap$ = dart.generic(function(K, V) {
    class LinkedHashMap extends core.Object {
      static new(opts) {
        let equals = opts && 'equals' in opts ? opts.equals : null;
        let hashCode = opts && 'hashCode' in opts ? opts.hashCode : null;
        let isValidKey = opts && 'isValidKey' in opts ? opts.isValidKey : null;
        if (isValidKey == null) {
          if (hashCode == null) {
            if (equals == null) {
              return new (_LinkedHashMap$(K, V))();
            }
            hashCode = _defaultHashCode;
          } else {
            if (dart.notNull(core.identical(core.identityHashCode, hashCode)) && dart.notNull(core.identical(core.identical, equals))) {
              return new (_LinkedIdentityHashMap$(K, V))();
            }
            if (equals == null) {
              equals = _defaultEquals;
            }
          }
        } else {
          if (hashCode == null) {
            hashCode = _defaultHashCode;
          }
          if (equals == null) {
            equals = _defaultEquals;
          }
        }
        return new (_LinkedCustomHashMap$(K, V))(equals, hashCode, isValidKey);
      }
      static identity() {
        return new (_LinkedIdentityHashMap$(K, V))();
      }
      static from(other) {
        let result = LinkedHashMap$(K, V).new();
        other.forEach(dart.fn((k, v) => {
          result.set(dart.as(k, K), dart.as(v, V));
        }));
        return result;
      }
      static fromIterable(iterable, opts) {
        let key = opts && 'key' in opts ? opts.key : null;
        let value = opts && 'value' in opts ? opts.value : null;
        let map = LinkedHashMap$(K, V).new();
        Maps._fillMapWithMappedIterable(map, iterable, key, value);
        return map;
      }
      static fromIterables(keys, values) {
        let map = LinkedHashMap$(K, V).new();
        Maps._fillMapWithIterables(map, keys, values);
        return map;
      }
      static _literal(keyValuePairs) {
        return dart.as(_js_helper.fillLiteralMap(keyValuePairs, new (_LinkedHashMap$(K, V))()), LinkedHashMap$(K, V));
      }
      static _empty() {
        return new (_LinkedHashMap$(K, V))();
      }
    }
    LinkedHashMap[dart.implements] = () => [HashMap$(K, V)];
    dart.setSignature(LinkedHashMap, {
      constructors: () => ({
        new: [LinkedHashMap$(K, V), [], {equals: dart.functionType(core.bool, [K, K]), hashCode: dart.functionType(core.int, [K]), isValidKey: dart.functionType(core.bool, [core.Object])}],
        identity: [LinkedHashMap$(K, V), []],
        from: [LinkedHashMap$(K, V), [core.Map]],
        fromIterable: [LinkedHashMap$(K, V), [core.Iterable], {key: dart.functionType(K, [dart.dynamic]), value: dart.functionType(V, [dart.dynamic])}],
        fromIterables: [LinkedHashMap$(K, V), [core.Iterable$(K), core.Iterable$(V)]],
        _literal: [LinkedHashMap$(K, V), [core.List]],
        _empty: [LinkedHashMap$(K, V), []]
      })
    });
    return LinkedHashMap;
  });
  let LinkedHashMap = LinkedHashMap$();
  const LinkedHashSet$ = dart.generic(function(E) {
    class LinkedHashSet extends core.Object {
      static new(opts) {
        let equals = opts && 'equals' in opts ? opts.equals : null;
        let hashCode = opts && 'hashCode' in opts ? opts.hashCode : null;
        let isValidKey = opts && 'isValidKey' in opts ? opts.isValidKey : null;
        if (isValidKey == null) {
          if (hashCode == null) {
            if (equals == null) {
              return new (_LinkedHashSet$(E))();
            }
            hashCode = _defaultHashCode;
          } else {
            if (dart.notNull(core.identical(core.identityHashCode, hashCode)) && dart.notNull(core.identical(core.identical, equals))) {
              return new (_LinkedIdentityHashSet$(E))();
            }
            if (equals == null) {
              equals = _defaultEquals;
            }
          }
        } else {
          if (hashCode == null) {
            hashCode = _defaultHashCode;
          }
          if (equals == null) {
            equals = _defaultEquals;
          }
        }
        return new (_LinkedCustomHashSet$(E))(equals, hashCode, isValidKey);
      }
      static identity() {
        return new (_LinkedIdentityHashSet$(E))();
      }
      static from(elements) {
        let result = LinkedHashSet$(E).new();
        for (let element of elements) {
          result.add(element);
        }
        return result;
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
    }
    LinkedHashSet[dart.implements] = () => [HashSet$(E)];
    dart.setSignature(LinkedHashSet, {
      constructors: () => ({
        new: [LinkedHashSet$(E), [], {equals: dart.functionType(core.bool, [E, E]), hashCode: dart.functionType(core.int, [E]), isValidKey: dart.functionType(core.bool, [core.Object])}],
        identity: [LinkedHashSet$(E), []],
        from: [LinkedHashSet$(E), [core.Iterable$(E)]]
      })
    });
    return LinkedHashSet;
  });
  let LinkedHashSet = LinkedHashSet$();
  const _modificationCount = Symbol('_modificationCount');
  const _length = Symbol('_length');
  const _next = Symbol('_next');
  const _previous = Symbol('_previous');
  const _insertAfter = Symbol('_insertAfter');
  const _list = Symbol('_list');
  const _unlink = Symbol('_unlink');
  const LinkedList$ = dart.generic(function(E) {
    class LinkedList extends IterableBase$(E) {
      LinkedList() {
        this[_modificationCount] = 0;
        this[_length] = 0;
        this[_next] = null;
        this[_previous] = null;
        super.IterableBase();
        this[_next] = this[_previous] = this;
      }
      addFirst(entry) {
        dart.as(entry, E);
        this[_insertAfter](this, entry);
      }
      add(entry) {
        dart.as(entry, E);
        this[_insertAfter](this[_previous], entry);
      }
      addAll(entries) {
        dart.as(entries, core.Iterable$(E));
        entries[dartx.forEach](dart.fn((entry => this[_insertAfter](this[_previous], dart.as(entry, E))).bind(this), dart.void, [dart.dynamic]));
      }
      remove(entry) {
        dart.as(entry, E);
        if (!dart.equals(entry[_list], this))
          return false;
        this[_unlink](entry);
        return true;
      }
      get iterator() {
        return new (_LinkedListIterator$(E))(this);
      }
      get length() {
        return this[_length];
      }
      clear() {
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        let next = this[_next];
        while (!dart.notNull(core.identical(next, this))) {
          let entry = dart.as(next, E);
          next = entry[_next];
          entry[_next] = entry[_previous] = entry[_list] = null;
        }
        this[_next] = this[_previous] = this;
        this[_length] = 0;
      }
      get first() {
        if (dart.notNull(core.identical(this[_next], this))) {
          dart.throw(new core.StateError('No such element'));
        }
        return dart.as(this[_next], E);
      }
      get last() {
        if (dart.notNull(core.identical(this[_previous], this))) {
          dart.throw(new core.StateError('No such element'));
        }
        return dart.as(this[_previous], E);
      }
      get single() {
        if (dart.notNull(core.identical(this[_previous], this))) {
          dart.throw(new core.StateError('No such element'));
        }
        if (!dart.notNull(core.identical(this[_previous], this[_next]))) {
          dart.throw(new core.StateError('Too many elements'));
        }
        return dart.as(this[_next], E);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
        let modificationCount = this[_modificationCount];
        let current = this[_next];
        while (!dart.notNull(core.identical(current, this))) {
          action(dart.as(current, E));
          if (modificationCount != this[_modificationCount]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          current = current[_next];
        }
      }
      get isEmpty() {
        return this[_length] == 0;
      }
      [_insertAfter](entry, newEntry) {
        dart.as(newEntry, E);
        if (newEntry.list != null) {
          dart.throw(new core.StateError('LinkedListEntry is already in a LinkedList'));
        }
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        newEntry[_list] = this;
        let predecessor = entry;
        let successor = entry[_next];
        successor[_previous] = newEntry;
        newEntry[_previous] = predecessor;
        newEntry[_next] = successor;
        predecessor[_next] = newEntry;
        this[_length] = dart.notNull(this[_length]) + 1;
      }
      [_unlink](entry) {
        dart.as(entry, LinkedListEntry$(E));
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        entry[_next][_previous] = entry[_previous];
        entry[_previous][_next] = entry[_next];
        this[_length] = dart.notNull(this[_length]) - 1;
        entry[_list] = entry[_next] = entry[_previous] = null;
      }
    }
    LinkedList[dart.implements] = () => [_LinkedListLink];
    dart.setSignature(LinkedList, {
      constructors: () => ({LinkedList: [LinkedList$(E), []]}),
      methods: () => ({
        addFirst: [dart.void, [E]],
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [E]],
        clear: [dart.void, []],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        [_insertAfter]: [dart.void, [_LinkedListLink, E]],
        [_unlink]: [dart.void, [LinkedListEntry$(E)]]
      })
    });
    dart.defineExtensionMembers(LinkedList, [
      'forEach',
      'iterator',
      'length',
      'first',
      'last',
      'single',
      'isEmpty'
    ]);
    return LinkedList;
  });
  let LinkedList = LinkedList$();
  const _current = Symbol('_current');
  const _LinkedListIterator$ = dart.generic(function(E) {
    class _LinkedListIterator extends core.Object {
      _LinkedListIterator(list) {
        this[_list] = list;
        this[_modificationCount] = list[_modificationCount];
        this[_next] = list[_next];
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        if (dart.notNull(core.identical(this[_next], this[_list]))) {
          this[_current] = null;
          return false;
        }
        if (this[_modificationCount] != this[_list][_modificationCount]) {
          dart.throw(new core.ConcurrentModificationError(this));
        }
        this[_current] = dart.as(this[_next], E);
        this[_next] = this[_next][_next];
        return true;
      }
    }
    _LinkedListIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(_LinkedListIterator, {
      constructors: () => ({_LinkedListIterator: [_LinkedListIterator$(E), [LinkedList$(E)]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _LinkedListIterator;
  });
  let _LinkedListIterator = _LinkedListIterator$();
  class _LinkedListLink extends core.Object {
    _LinkedListLink() {
      this[_next] = null;
      this[_previous] = null;
    }
  }
  const LinkedListEntry$ = dart.generic(function(E) {
    class LinkedListEntry extends core.Object {
      LinkedListEntry() {
        this[_list] = null;
        this[_next] = null;
        this[_previous] = null;
      }
      get list() {
        return this[_list];
      }
      unlink() {
        this[_list][_unlink](this);
      }
      get next() {
        if (dart.notNull(core.identical(this[_next], this[_list])))
          return null;
        let result = dart.as(this[_next], E);
        return result;
      }
      get previous() {
        if (dart.notNull(core.identical(this[_previous], this[_list])))
          return null;
        return dart.as(this[_previous], E);
      }
      insertAfter(entry) {
        dart.as(entry, E);
        this[_list][_insertAfter](this, entry);
      }
      insertBefore(entry) {
        dart.as(entry, E);
        this[_list][_insertAfter](this[_previous], entry);
      }
    }
    LinkedListEntry[dart.implements] = () => [_LinkedListLink];
    dart.setSignature(LinkedListEntry, {
      methods: () => ({
        unlink: [dart.void, []],
        insertAfter: [dart.void, [E]],
        insertBefore: [dart.void, [E]]
      })
    });
    return LinkedListEntry;
  });
  let LinkedListEntry = LinkedListEntry$();
  const ListMixin$ = dart.generic(function(E) {
    class ListMixin extends core.Object {
      get iterator() {
        return new (_internal.ListIterator$(E))(this);
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
      elementAt(index) {
        return this.get(index);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          action(this.get(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
      }
      get isEmpty() {
        return this[dartx.length] == 0;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      get first() {
        if (this[dartx.length] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        return this.get(0);
      }
      get last() {
        if (this[dartx.length] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        return this.get(dart.notNull(this[dartx.length]) - 1);
      }
      get single() {
        if (this[dartx.length] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        if (dart.notNull(this[dartx.length]) > 1)
          dart.throw(_internal.IterableElementError.tooMany());
        return this.get(0);
      }
      contains(element) {
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(this.length); i = dart.notNull(i) + 1) {
          if (dart.equals(this.get(i), element))
            return true;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return false;
      }
      every(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (!dart.notNull(test(this.get(i))))
            return false;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return true;
      }
      any(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.notNull(test(this.get(i))))
            return true;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return false;
      }
      firstWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = this.get(i);
          if (dart.notNull(test(element)))
            return element;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      lastWhere(test, opts) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let orElse = opts && 'orElse' in opts ? opts.orElse : null;
        dart.as(orElse, dart.functionType(E, []));
        let length = this.length;
        for (let i = dart.notNull(length) - 1; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
          let element = this.get(i);
          if (dart.notNull(test(element)))
            return element;
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (orElse != null)
          return orElse();
        dart.throw(_internal.IterableElementError.noElement());
      }
      singleWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let length = this.length;
        let match = null;
        let matchFound = false;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = this.get(i);
          if (dart.notNull(test(element))) {
            if (dart.notNull(matchFound)) {
              dart.throw(_internal.IterableElementError.tooMany());
            }
            matchFound = true;
            match = element;
          }
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        if (dart.notNull(matchFound))
          return match;
        dart.throw(_internal.IterableElementError.noElement());
      }
      join(separator) {
        if (separator === void 0)
          separator = "";
        if (this[dartx.length] == 0)
          return "";
        let buffer = new core.StringBuffer();
        buffer.writeAll(this, separator);
        return dart.toString(buffer);
      }
      where(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.WhereIterable$(E))(this, test);
      }
      map(f) {
        dart.as(f, dart.functionType(dart.dynamic, [E]));
        return new _internal.MappedListIterable(this, f);
      }
      expand(f) {
        dart.as(f, dart.functionType(core.Iterable, [E]));
        return new (_internal.ExpandIterable$(E, dart.dynamic))(this, f);
      }
      reduce(combine) {
        dart.as(combine, dart.functionType(E, [E, E]));
        let length = this.length;
        if (length == 0)
          dart.throw(_internal.IterableElementError.noElement());
        let value = this.get(0);
        for (let i = 1; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          value = combine(value, this.get(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return value;
      }
      fold(initialValue, combine) {
        dart.as(combine, dart.functionType(dart.dynamic, [dart.dynamic, E]));
        let value = initialValue;
        let length = this.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          value = dart.dcall(combine, value, this.get(i));
          if (length != this.length) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
        return value;
      }
      skip(count) {
        return new (_internal.SubListIterable$(E))(this, count, null);
      }
      skipWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.SkipWhileIterable$(E))(this, test);
      }
      take(count) {
        return new (_internal.SubListIterable$(E))(this, 0, count);
      }
      takeWhile(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        return new (_internal.TakeWhileIterable$(E))(this, test);
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let result = null;
        if (dart.notNull(growable)) {
          result = core.List$(E).new();
          result[dartx.length] = this[dartx.length];
        } else {
          result = core.List$(E).new(this[dartx.length]);
        }
        for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
          result[dartx.set](i, this.get(i));
        }
        return result;
      }
      toSet() {
        let result = core.Set$(E).new();
        for (let i = 0; dart.notNull(i) < dart.notNull(this[dartx.length]); i = dart.notNull(i) + 1) {
          result.add(this.get(i));
        }
        return result;
      }
      add(element) {
        dart.as(element, E);
        this.set((() => {
          let x = this.length;
          this.length = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), element);
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        for (let element of iterable) {
          this.set((() => {
            let x = this.length;
            this.length = dart.notNull(x) + 1;
            return x;
          }).bind(this)(), element);
        }
      }
      remove(element) {
        for (let i = 0; dart.notNull(i) < dart.notNull(this.length); i = dart.notNull(i) + 1) {
          if (dart.equals(this.get(i), element)) {
            this.setRange(i, dart.notNull(this.length) - 1, this, dart.notNull(i) + 1);
            this.length = dart.notNull(this.length) - 1;
            return true;
          }
        }
        return false;
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        ListMixin$()._filter(this, test, false);
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        ListMixin$()._filter(this, test, true);
      }
      static _filter(source, test, retainMatching) {
        dart.as(test, dart.functionType(core.bool, [dart.dynamic]));
        let retained = [];
        let length = source[dartx.length];
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let element = source[dartx.get](i);
          if (dart.dcall(test, element) == retainMatching) {
            retained[dartx.add](element);
          }
          if (length != source[dartx.length]) {
            dart.throw(new core.ConcurrentModificationError(source));
          }
        }
        if (retained[dartx.length] != source[dartx.length]) {
          source[dartx.setRange](0, retained[dartx.length], retained);
          source[dartx.length] = retained[dartx.length];
        }
      }
      clear() {
        this.length = 0;
      }
      removeLast() {
        if (this[dartx.length] == 0) {
          dart.throw(_internal.IterableElementError.noElement());
        }
        let result = this.get(dart.notNull(this[dartx.length]) - 1);
        this[dartx.length] = dart.notNull(this[dartx.length]) - 1;
        return result;
      }
      sort(compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, dart.functionType(core.int, [E, E]));
        _internal.Sort.sort(this, compare == null ? core.Comparable.compare : compare);
      }
      shuffle(random) {
        if (random === void 0)
          random = null;
        if (random == null)
          random = math.Random.new();
        let length = this.length;
        while (dart.notNull(length) > 1) {
          let pos = random.nextInt(length);
          length = dart.notNull(length) - 1;
          let tmp = this.get(length);
          this.set(length, this.get(pos));
          this.set(pos, tmp);
        }
      }
      asMap() {
        return new (_internal.ListMapView$(E))(this);
      }
      sublist(start, end) {
        if (end === void 0)
          end = null;
        let listLength = this.length;
        if (end == null)
          end = listLength;
        core.RangeError.checkValidRange(start, end, listLength);
        let length = dart.notNull(end) - dart.notNull(start);
        let result = core.List$(E).new();
        result[dartx.length] = length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          result[dartx.set](i, this.get(dart.notNull(start) + dart.notNull(i)));
        }
        return result;
      }
      getRange(start, end) {
        core.RangeError.checkValidRange(start, end, this.length);
        return new (_internal.SubListIterable$(E))(this, start, end);
      }
      removeRange(start, end) {
        core.RangeError.checkValidRange(start, end, this.length);
        let length = dart.notNull(end) - dart.notNull(start);
        this.setRange(start, dart.notNull(this.length) - dart.notNull(length), this, end);
        this.length = dart.notNull(this.length) - dart.notNull(length);
      }
      fillRange(start, end, fill) {
        if (fill === void 0)
          fill = null;
        dart.as(fill, E);
        core.RangeError.checkValidRange(start, end, this.length);
        for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
          this.set(i, fill);
        }
      }
      setRange(start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        core.RangeError.checkValidRange(start, end, this.length);
        let length = dart.notNull(end) - dart.notNull(start);
        if (length == 0)
          return;
        core.RangeError.checkNotNegative(skipCount, "skipCount");
        let otherList = null;
        let otherStart = null;
        if (dart.is(iterable, core.List)) {
          otherList = dart.as(iterable, core.List);
          otherStart = skipCount;
        } else {
          otherList = iterable[dartx.skip](skipCount)[dartx.toList]({growable: false});
          otherStart = 0;
        }
        if (dart.notNull(otherStart) + dart.notNull(length) > dart.notNull(otherList[dartx.length])) {
          dart.throw(_internal.IterableElementError.tooFew());
        }
        if (dart.notNull(otherStart) < dart.notNull(start)) {
          for (let i = dart.notNull(length) - 1; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
            this.set(dart.notNull(start) + dart.notNull(i), dart.as(otherList[dartx.get](dart.notNull(otherStart) + dart.notNull(i)), E));
          }
        } else {
          for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
            this.set(dart.notNull(start) + dart.notNull(i), dart.as(otherList[dartx.get](dart.notNull(otherStart) + dart.notNull(i)), E));
          }
        }
      }
      replaceRange(start, end, newContents) {
        dart.as(newContents, core.Iterable$(E));
        core.RangeError.checkValidRange(start, end, this.length);
        if (!dart.is(newContents, _internal.EfficientLength)) {
          newContents = newContents[dartx.toList]();
        }
        let removeLength = dart.notNull(end) - dart.notNull(start);
        let insertLength = newContents[dartx.length];
        if (dart.notNull(removeLength) >= dart.notNull(insertLength)) {
          let delta = dart.notNull(removeLength) - dart.notNull(insertLength);
          let insertEnd = dart.notNull(start) + dart.notNull(insertLength);
          let newLength = dart.notNull(this.length) - dart.notNull(delta);
          this.setRange(start, insertEnd, newContents);
          if (delta != 0) {
            this.setRange(insertEnd, newLength, this, end);
            this.length = newLength;
          }
        } else {
          let delta = dart.notNull(insertLength) - dart.notNull(removeLength);
          let newLength = dart.notNull(this.length) + dart.notNull(delta);
          let insertEnd = dart.notNull(start) + dart.notNull(insertLength);
          this.length = newLength;
          this.setRange(insertEnd, newLength, this, end);
          this.setRange(start, insertEnd, newContents);
        }
      }
      indexOf(element, startIndex) {
        if (startIndex === void 0)
          startIndex = 0;
        if (dart.notNull(startIndex) >= dart.notNull(this.length)) {
          return -1;
        }
        if (dart.notNull(startIndex) < 0) {
          startIndex = 0;
        }
        for (let i = startIndex; dart.notNull(i) < dart.notNull(this.length); i = dart.notNull(i) + 1) {
          if (dart.equals(this.get(i), element)) {
            return i;
          }
        }
        return -1;
      }
      lastIndexOf(element, startIndex) {
        if (startIndex === void 0)
          startIndex = null;
        if (startIndex == null) {
          startIndex = dart.notNull(this.length) - 1;
        } else {
          if (dart.notNull(startIndex) < 0) {
            return -1;
          }
          if (dart.notNull(startIndex) >= dart.notNull(this.length)) {
            startIndex = dart.notNull(this.length) - 1;
          }
        }
        for (let i = startIndex; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
          if (dart.equals(this.get(i), element)) {
            return i;
          }
        }
        return -1;
      }
      insert(index, element) {
        dart.as(element, E);
        core.RangeError.checkValueInInterval(index, 0, this[dartx.length], "index");
        if (index == this.length) {
          this.add(element);
          return;
        }
        if (!(typeof index == 'number'))
          dart.throw(new core.ArgumentError(index));
        this.length = dart.notNull(this.length) + 1;
        this.setRange(dart.notNull(index) + 1, this.length, this, index);
        this.set(index, element);
      }
      removeAt(index) {
        let result = this.get(index);
        this.setRange(index, dart.notNull(this.length) - 1, this, dart.notNull(index) + 1);
        this[dartx.length] = dart.notNull(this[dartx.length]) - 1;
        return result;
      }
      insertAll(index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        core.RangeError.checkValueInInterval(index, 0, this[dartx.length], "index");
        if (dart.is(iterable, _internal.EfficientLength)) {
          iterable = iterable[dartx.toList]();
        }
        let insertionLength = iterable[dartx.length];
        this.length = dart.notNull(this.length) + dart.notNull(insertionLength);
        this.setRange(dart.notNull(index) + dart.notNull(insertionLength), this.length, this, index);
        this.setAll(index, iterable);
      }
      setAll(index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        if (dart.is(iterable, core.List)) {
          this.setRange(index, dart.notNull(index) + dart.notNull(iterable[dartx.length]), iterable);
        } else {
          for (let element of iterable) {
            this.set((() => {
              let x = index;
              index = dart.notNull(x) + 1;
              return x;
            })(), element);
          }
        }
      }
      get reversed() {
        return new (_internal.ReversedListIterable$(E))(this);
      }
      toString() {
        return IterableBase.iterableToFullString(this, '[', ']');
      }
    }
    ListMixin[dart.implements] = () => [core.List$(E)];
    dart.setSignature(ListMixin, {
      methods: () => ({
        elementAt: [E, [core.int]],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        contains: [core.bool, [core.Object]],
        every: [core.bool, [dart.functionType(core.bool, [E])]],
        any: [core.bool, [dart.functionType(core.bool, [E])]],
        firstWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        lastWhere: [E, [dart.functionType(core.bool, [E])], {orElse: dart.functionType(E, [])}],
        singleWhere: [E, [dart.functionType(core.bool, [E])]],
        join: [core.String, [], [core.String]],
        where: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        map: [core.Iterable, [dart.functionType(dart.dynamic, [E])]],
        expand: [core.Iterable, [dart.functionType(core.Iterable, [E])]],
        reduce: [E, [dart.functionType(E, [E, E])]],
        fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, E])]],
        skip: [core.Iterable$(E), [core.int]],
        skipWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        take: [core.Iterable$(E), [core.int]],
        takeWhile: [core.Iterable$(E), [dart.functionType(core.bool, [E])]],
        toList: [core.List$(E), [], {growable: core.bool}],
        toSet: [core.Set$(E), []],
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [core.Object]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        clear: [dart.void, []],
        removeLast: [E, []],
        sort: [dart.void, [], [dart.functionType(core.int, [E, E])]],
        shuffle: [dart.void, [], [math.Random]],
        asMap: [core.Map$(core.int, E), []],
        sublist: [core.List$(E), [core.int], [core.int]],
        getRange: [core.Iterable$(E), [core.int, core.int]],
        removeRange: [dart.void, [core.int, core.int]],
        fillRange: [dart.void, [core.int, core.int], [E]],
        setRange: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        replaceRange: [dart.void, [core.int, core.int, core.Iterable$(E)]],
        indexOf: [core.int, [core.Object], [core.int]],
        lastIndexOf: [core.int, [core.Object], [core.int]],
        insert: [dart.void, [core.int, E]],
        removeAt: [E, [core.int]],
        insertAll: [dart.void, [core.int, core.Iterable$(E)]],
        setAll: [dart.void, [core.int, core.Iterable$(E)]]
      }),
      statics: () => ({_filter: [dart.void, [core.List, dart.functionType(core.bool, [dart.dynamic]), core.bool]]}),
      names: ['_filter']
    });
    dart.defineExtensionMembers(ListMixin, [
      'elementAt',
      'forEach',
      'contains',
      'every',
      'any',
      'firstWhere',
      'lastWhere',
      'singleWhere',
      'join',
      'where',
      'map',
      'expand',
      'reduce',
      'fold',
      'skip',
      'skipWhile',
      'take',
      'takeWhile',
      'toList',
      'toSet',
      'add',
      'addAll',
      'remove',
      'removeWhere',
      'retainWhere',
      'clear',
      'removeLast',
      'sort',
      'shuffle',
      'asMap',
      'sublist',
      'getRange',
      'removeRange',
      'fillRange',
      'setRange',
      'replaceRange',
      'indexOf',
      'lastIndexOf',
      'insert',
      'removeAt',
      'insertAll',
      'setAll',
      'iterator',
      'isEmpty',
      'isNotEmpty',
      'first',
      'last',
      'single',
      'reversed'
    ]);
    return ListMixin;
  });
  let ListMixin = ListMixin$();
  const ListBase$ = dart.generic(function(E) {
    class ListBase extends dart.mixin(core.Object, ListMixin$(E)) {
      static listToString(list) {
        return IterableBase.iterableToFullString(list, '[', ']');
      }
    }
    dart.setSignature(ListBase, {
      statics: () => ({listToString: [core.String, [core.List]]}),
      names: ['listToString']
    });
    return ListBase;
  });
  let ListBase = ListBase$();
  const MapMixin$ = dart.generic(function(K, V) {
    class MapMixin extends core.Object {
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [K, V]));
        for (let key of this.keys) {
          action(key, this.get(key));
        }
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        for (let key of other.keys) {
          this.set(key, other.get(key));
        }
      }
      containsValue(value) {
        for (let key of this.keys) {
          if (dart.equals(this.get(key), value))
            return true;
        }
        return false;
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        if (dart.notNull(this.keys[dartx.contains](key))) {
          return this.get(key);
        }
        return this.set(key, ifAbsent());
      }
      containsKey(key) {
        return this.keys[dartx.contains](key);
      }
      get length() {
        return this.keys[dartx.length];
      }
      get isEmpty() {
        return this.keys[dartx.isEmpty];
      }
      get isNotEmpty() {
        return this.keys[dartx.isNotEmpty];
      }
      get values() {
        return new (_MapBaseValueIterable$(V))(this);
      }
      toString() {
        return Maps.mapToString(this);
      }
    }
    MapMixin[dart.implements] = () => [core.Map$(K, V)];
    dart.setSignature(MapMixin, {
      methods: () => ({
        forEach: [dart.void, [dart.functionType(dart.void, [K, V])]],
        addAll: [dart.void, [core.Map$(K, V)]],
        containsValue: [core.bool, [core.Object]],
        putIfAbsent: [V, [K, dart.functionType(V, [])]],
        containsKey: [core.bool, [core.Object]]
      })
    });
    return MapMixin;
  });
  let MapMixin = MapMixin$();
  const MapBase$ = dart.generic(function(K, V) {
    class MapBase extends dart.mixin(core.Object, MapMixin$(K, V)) {}
    return MapBase;
  });
  let MapBase = MapBase$();
  const _UnmodifiableMapMixin$ = dart.generic(function(K, V) {
    class _UnmodifiableMapMixin extends core.Object {
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        dart.throw(new core.UnsupportedError("Cannot modify unmodifiable map"));
        return value;
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        dart.throw(new core.UnsupportedError("Cannot modify unmodifiable map"));
      }
      clear() {
        dart.throw(new core.UnsupportedError("Cannot modify unmodifiable map"));
      }
      remove(key) {
        dart.throw(new core.UnsupportedError("Cannot modify unmodifiable map"));
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        dart.throw(new core.UnsupportedError("Cannot modify unmodifiable map"));
      }
    }
    _UnmodifiableMapMixin[dart.implements] = () => [core.Map$(K, V)];
    dart.setSignature(_UnmodifiableMapMixin, {
      methods: () => ({
        set: [dart.void, [K, V]],
        addAll: [dart.void, [core.Map$(K, V)]],
        clear: [dart.void, []],
        remove: [V, [core.Object]],
        putIfAbsent: [V, [K, dart.functionType(V, [])]]
      })
    });
    return _UnmodifiableMapMixin;
  });
  let _UnmodifiableMapMixin = _UnmodifiableMapMixin$();
  const UnmodifiableMapBase$ = dart.generic(function(K, V) {
    class UnmodifiableMapBase extends dart.mixin(MapBase$(K, V), _UnmodifiableMapMixin$(K, V)) {
      UnmodifiableMapBase() {
        super.MapBase(...arguments);
      }
    }
    return UnmodifiableMapBase;
  });
  let UnmodifiableMapBase = UnmodifiableMapBase$();
  const _map = Symbol('_map');
  const _MapBaseValueIterable$ = dart.generic(function(V) {
    class _MapBaseValueIterable extends IterableBase$(V) {
      _MapBaseValueIterable(map) {
        this[_map] = map;
        super.IterableBase();
      }
      get length() {
        return this[_map].length;
      }
      get isEmpty() {
        return this[_map].isEmpty;
      }
      get isNotEmpty() {
        return this[_map].isNotEmpty;
      }
      get first() {
        return dart.as(this[_map].get(this[_map].keys[dartx.first]), V);
      }
      get single() {
        return dart.as(this[_map].get(this[_map].keys[dartx.single]), V);
      }
      get last() {
        return dart.as(this[_map].get(this[_map].keys[dartx.last]), V);
      }
      get iterator() {
        return new (_MapBaseValueIterator$(V))(this[_map]);
      }
    }
    _MapBaseValueIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(_MapBaseValueIterable, {
      constructors: () => ({_MapBaseValueIterable: [_MapBaseValueIterable$(V), [core.Map]]})
    });
    dart.defineExtensionMembers(_MapBaseValueIterable, [
      'length',
      'isEmpty',
      'isNotEmpty',
      'first',
      'single',
      'last',
      'iterator'
    ]);
    return _MapBaseValueIterable;
  });
  let _MapBaseValueIterable = _MapBaseValueIterable$();
  const _keys = Symbol('_keys');
  const _MapBaseValueIterator$ = dart.generic(function(V) {
    class _MapBaseValueIterator extends core.Object {
      _MapBaseValueIterator(map) {
        this[_map] = map;
        this[_keys] = map.keys[dartx.iterator];
        this[_current] = null;
      }
      moveNext() {
        if (dart.notNull(this[_keys].moveNext())) {
          this[_current] = dart.as(this[_map].get(this[_keys].current), V);
          return true;
        }
        this[_current] = null;
        return false;
      }
      get current() {
        return this[_current];
      }
    }
    _MapBaseValueIterator[dart.implements] = () => [core.Iterator$(V)];
    dart.setSignature(_MapBaseValueIterator, {
      constructors: () => ({_MapBaseValueIterator: [_MapBaseValueIterator$(V), [core.Map]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _MapBaseValueIterator;
  });
  let _MapBaseValueIterator = _MapBaseValueIterator$();
  const MapView$ = dart.generic(function(K, V) {
    class MapView extends core.Object {
      MapView(map) {
        this[_map] = map;
      }
      get(key) {
        return this[_map].get(key);
      }
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        this[_map].set(key, value);
        return value;
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        this[_map].addAll(other);
      }
      clear() {
        this[_map].clear();
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        return this[_map].putIfAbsent(key, ifAbsent);
      }
      containsKey(key) {
        return this[_map].containsKey(key);
      }
      containsValue(value) {
        return this[_map].containsValue(value);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [K, V]));
        this[_map].forEach(action);
      }
      get isEmpty() {
        return this[_map].isEmpty;
      }
      get isNotEmpty() {
        return this[_map].isNotEmpty;
      }
      get length() {
        return this[_map].length;
      }
      get keys() {
        return this[_map].keys;
      }
      remove(key) {
        return this[_map].remove(key);
      }
      toString() {
        return dart.toString(this[_map]);
      }
      get values() {
        return this[_map].values;
      }
    }
    MapView[dart.implements] = () => [core.Map$(K, V)];
    dart.setSignature(MapView, {
      constructors: () => ({MapView: [MapView$(K, V), [core.Map$(K, V)]]}),
      methods: () => ({
        get: [V, [core.Object]],
        set: [dart.void, [K, V]],
        addAll: [dart.void, [core.Map$(K, V)]],
        clear: [dart.void, []],
        putIfAbsent: [V, [K, dart.functionType(V, [])]],
        containsKey: [core.bool, [core.Object]],
        containsValue: [core.bool, [core.Object]],
        forEach: [dart.void, [dart.functionType(dart.void, [K, V])]],
        remove: [V, [core.Object]]
      })
    });
    return MapView;
  });
  let MapView = MapView$();
  const UnmodifiableMapView$ = dart.generic(function(K, V) {
    class UnmodifiableMapView extends dart.mixin(MapView$(K, V), _UnmodifiableMapMixin$(K, V)) {
      UnmodifiableMapView() {
        super.MapView(...arguments);
      }
    }
    return UnmodifiableMapView;
  });
  let UnmodifiableMapView = UnmodifiableMapView$();
  class Maps extends core.Object {
    static containsValue(map, value) {
      for (let v of map.values) {
        if (dart.equals(value, v)) {
          return true;
        }
      }
      return false;
    }
    static containsKey(map, key) {
      for (let k of map.keys) {
        if (dart.equals(key, k)) {
          return true;
        }
      }
      return false;
    }
    static putIfAbsent(map, key, ifAbsent) {
      if (dart.notNull(map.containsKey(key))) {
        return map.get(key);
      }
      let v = ifAbsent();
      map.set(key, v);
      return v;
    }
    static clear(map) {
      for (let k of map.keys[dartx.toList]()) {
        map.remove(k);
      }
    }
    static forEach(map, f) {
      for (let k of map.keys) {
        dart.dcall(f, k, map.get(k));
      }
    }
    static getValues(map) {
      return map.keys[dartx.map](dart.fn(key => map.get(key)));
    }
    static length(map) {
      return map.keys[dartx.length];
    }
    static isEmpty(map) {
      return map.keys[dartx.isEmpty];
    }
    static isNotEmpty(map) {
      return map.keys[dartx.isNotEmpty];
    }
    static mapToString(m) {
      if (dart.notNull(IterableBase._isToStringVisiting(m))) {
        return '{...}';
      }
      let result = new core.StringBuffer();
      try {
        IterableBase._toStringVisiting[dartx.add](m);
        result.write('{');
        let first = true;
        m.forEach(dart.fn((k, v) => {
          if (!dart.notNull(first)) {
            result.write(', ');
          }
          first = false;
          result.write(k);
          result.write(': ');
          result.write(v);
        }));
        result.write('}');
      } finally {
        dart.assert(core.identical(IterableBase._toStringVisiting[dartx.last], m));
        IterableBase._toStringVisiting[dartx.removeLast]();
      }
      return dart.toString(result);
    }
    static _id(x) {
      return x;
    }
    static _fillMapWithMappedIterable(map, iterable, key, value) {
      if (key == null)
        key = Maps._id;
      if (value == null)
        value = Maps._id;
      for (let element of iterable) {
        map.set(dart.dcall(key, element), dart.dcall(value, element));
      }
    }
    static _fillMapWithIterables(map, keys, values) {
      let keyIterator = keys[dartx.iterator];
      let valueIterator = values[dartx.iterator];
      let hasNextKey = keyIterator.moveNext();
      let hasNextValue = valueIterator.moveNext();
      while (dart.notNull(hasNextKey) && dart.notNull(hasNextValue)) {
        map.set(keyIterator.current, valueIterator.current);
        hasNextKey = keyIterator.moveNext();
        hasNextValue = valueIterator.moveNext();
      }
      if (dart.notNull(hasNextKey) || dart.notNull(hasNextValue)) {
        dart.throw(new core.ArgumentError("Iterables do not have same length."));
      }
    }
  }
  dart.setSignature(Maps, {
    statics: () => ({
      containsValue: [core.bool, [core.Map, dart.dynamic]],
      containsKey: [core.bool, [core.Map, dart.dynamic]],
      putIfAbsent: [dart.dynamic, [core.Map, dart.dynamic, dart.functionType(dart.dynamic, [])]],
      clear: [dart.dynamic, [core.Map]],
      forEach: [dart.dynamic, [core.Map, dart.functionType(dart.void, [dart.dynamic, dart.dynamic])]],
      getValues: [core.Iterable, [core.Map]],
      length: [core.int, [core.Map]],
      isEmpty: [core.bool, [core.Map]],
      isNotEmpty: [core.bool, [core.Map]],
      mapToString: [core.String, [core.Map]],
      _id: [dart.dynamic, [dart.dynamic]],
      _fillMapWithMappedIterable: [dart.void, [core.Map, core.Iterable, dart.functionType(dart.dynamic, [dart.dynamic]), dart.functionType(dart.dynamic, [dart.dynamic])]],
      _fillMapWithIterables: [dart.void, [core.Map, core.Iterable, core.Iterable]]
    }),
    names: ['containsValue', 'containsKey', 'putIfAbsent', 'clear', 'forEach', 'getValues', 'length', 'isEmpty', 'isNotEmpty', 'mapToString', '_id', '_fillMapWithMappedIterable', '_fillMapWithIterables']
  });
  const Queue$ = dart.generic(function(E) {
    class Queue extends core.Object {
      static new() {
        return new (ListQueue$(E))();
      }
      static from(elements) {
        return ListQueue$(E).from(elements);
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
    }
    Queue[dart.implements] = () => [core.Iterable$(E), _internal.EfficientLength];
    dart.setSignature(Queue, {
      constructors: () => ({
        new: [Queue$(E), []],
        from: [Queue$(E), [core.Iterable]]
      })
    });
    return Queue;
  });
  let Queue = Queue$();
  const _element = Symbol('_element');
  const _link = Symbol('_link');
  const _asNonSentinelEntry = Symbol('_asNonSentinelEntry');
  const DoubleLinkedQueueEntry$ = dart.generic(function(E) {
    class DoubleLinkedQueueEntry extends core.Object {
      DoubleLinkedQueueEntry(e) {
        this[_element] = e;
        this[_previous] = null;
        this[_next] = null;
      }
      [_link](previous, next) {
        dart.as(previous, DoubleLinkedQueueEntry$(E));
        dart.as(next, DoubleLinkedQueueEntry$(E));
        this[_next] = next;
        this[_previous] = previous;
        previous[_next] = this;
        next[_previous] = this;
      }
      append(e) {
        dart.as(e, E);
        new (DoubleLinkedQueueEntry$(E))(e)[_link](this, this[_next]);
      }
      prepend(e) {
        dart.as(e, E);
        new (DoubleLinkedQueueEntry$(E))(e)[_link](this[_previous], this);
      }
      remove() {
        this[_previous][_next] = this[_next];
        this[_next][_previous] = this[_previous];
        this[_next] = null;
        this[_previous] = null;
        return this[_element];
      }
      [_asNonSentinelEntry]() {
        return this;
      }
      previousEntry() {
        return this[_previous][_asNonSentinelEntry]();
      }
      nextEntry() {
        return this[_next][_asNonSentinelEntry]();
      }
      get element() {
        return this[_element];
      }
      set element(e) {
        dart.as(e, E);
        this[_element] = e;
      }
    }
    dart.setSignature(DoubleLinkedQueueEntry, {
      constructors: () => ({DoubleLinkedQueueEntry: [DoubleLinkedQueueEntry$(E), [E]]}),
      methods: () => ({
        [_link]: [dart.void, [DoubleLinkedQueueEntry$(E), DoubleLinkedQueueEntry$(E)]],
        append: [dart.void, [E]],
        prepend: [dart.void, [E]],
        remove: [E, []],
        [_asNonSentinelEntry]: [DoubleLinkedQueueEntry$(E), []],
        previousEntry: [DoubleLinkedQueueEntry$(E), []],
        nextEntry: [DoubleLinkedQueueEntry$(E), []]
      })
    });
    return DoubleLinkedQueueEntry;
  });
  let DoubleLinkedQueueEntry = DoubleLinkedQueueEntry$();
  const _DoubleLinkedQueueEntrySentinel$ = dart.generic(function(E) {
    class _DoubleLinkedQueueEntrySentinel extends DoubleLinkedQueueEntry$(E) {
      _DoubleLinkedQueueEntrySentinel() {
        super.DoubleLinkedQueueEntry(null);
        this[_link](this, this);
      }
      remove() {
        dart.throw(_internal.IterableElementError.noElement());
      }
      [_asNonSentinelEntry]() {
        return null;
      }
      set element(e) {
        dart.as(e, E);
        dart.assert(false);
      }
      get element() {
        dart.throw(_internal.IterableElementError.noElement());
      }
    }
    dart.setSignature(_DoubleLinkedQueueEntrySentinel, {
      constructors: () => ({_DoubleLinkedQueueEntrySentinel: [_DoubleLinkedQueueEntrySentinel$(E), []]}),
      methods: () => ({
        remove: [E, []],
        [_asNonSentinelEntry]: [DoubleLinkedQueueEntry$(E), []]
      })
    });
    return _DoubleLinkedQueueEntrySentinel;
  });
  let _DoubleLinkedQueueEntrySentinel = _DoubleLinkedQueueEntrySentinel$();
  const _sentinel = Symbol('_sentinel');
  const _elementCount = Symbol('_elementCount');
  const _filter = Symbol('_filter');
  const DoubleLinkedQueue$ = dart.generic(function(E) {
    class DoubleLinkedQueue extends IterableBase$(E) {
      DoubleLinkedQueue() {
        this[_sentinel] = null;
        this[_elementCount] = 0;
        super.IterableBase();
        this[_sentinel] = new (_DoubleLinkedQueueEntrySentinel$(E))();
      }
      static from(elements) {
        let list = new (DoubleLinkedQueue$(E))();
        for (let e of dart.as(elements, core.Iterable$(E))) {
          list.addLast(e);
        }
        return dart.as(list, DoubleLinkedQueue$(E));
      }
      get length() {
        return this[_elementCount];
      }
      addLast(value) {
        dart.as(value, E);
        this[_sentinel].prepend(value);
        this[_elementCount] = dart.notNull(this[_elementCount]) + 1;
      }
      addFirst(value) {
        dart.as(value, E);
        this[_sentinel].append(value);
        this[_elementCount] = dart.notNull(this[_elementCount]) + 1;
      }
      add(value) {
        dart.as(value, E);
        this[_sentinel].prepend(value);
        this[_elementCount] = dart.notNull(this[_elementCount]) + 1;
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        for (let value of iterable) {
          this[_sentinel].prepend(value);
          this[_elementCount] = dart.notNull(this[_elementCount]) + 1;
        }
      }
      removeLast() {
        let result = this[_sentinel][_previous].remove();
        this[_elementCount] = dart.notNull(this[_elementCount]) - 1;
        return result;
      }
      removeFirst() {
        let result = this[_sentinel][_next].remove();
        this[_elementCount] = dart.notNull(this[_elementCount]) - 1;
        return result;
      }
      remove(o) {
        let entry = this[_sentinel][_next];
        while (!dart.notNull(core.identical(entry, this[_sentinel]))) {
          if (dart.equals(entry.element, o)) {
            entry.remove();
            this[_elementCount] = dart.notNull(this[_elementCount]) - 1;
            return true;
          }
          entry = entry[_next];
        }
        return false;
      }
      [_filter](test, removeMatching) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let entry = this[_sentinel][_next];
        while (!dart.notNull(core.identical(entry, this[_sentinel]))) {
          let next = entry[_next];
          if (dart.notNull(core.identical(removeMatching, test(entry.element)))) {
            entry.remove();
            this[_elementCount] = dart.notNull(this[_elementCount]) - 1;
          }
          entry = next;
        }
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filter](test, true);
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filter](test, false);
      }
      get first() {
        return this[_sentinel][_next].element;
      }
      get last() {
        return this[_sentinel][_previous].element;
      }
      get single() {
        if (dart.notNull(core.identical(this[_sentinel][_next], this[_sentinel][_previous]))) {
          return this[_sentinel][_next].element;
        }
        dart.throw(_internal.IterableElementError.tooMany());
      }
      lastEntry() {
        return this[_sentinel].previousEntry();
      }
      firstEntry() {
        return this[_sentinel].nextEntry();
      }
      get isEmpty() {
        return core.identical(this[_sentinel][_next], this[_sentinel]);
      }
      clear() {
        this[_sentinel][_next] = this[_sentinel];
        this[_sentinel][_previous] = this[_sentinel];
        this[_elementCount] = 0;
      }
      forEachEntry(f) {
        dart.as(f, dart.functionType(dart.void, [DoubleLinkedQueueEntry$(E)]));
        let entry = this[_sentinel][_next];
        while (!dart.notNull(core.identical(entry, this[_sentinel]))) {
          let nextEntry = entry[_next];
          f(entry);
          entry = nextEntry;
        }
      }
      get iterator() {
        return new (_DoubleLinkedQueueIterator$(E))(this[_sentinel]);
      }
      toString() {
        return IterableBase.iterableToFullString(this, '{', '}');
      }
    }
    DoubleLinkedQueue[dart.implements] = () => [Queue$(E)];
    dart.setSignature(DoubleLinkedQueue, {
      constructors: () => ({
        DoubleLinkedQueue: [DoubleLinkedQueue$(E), []],
        from: [DoubleLinkedQueue$(E), [core.Iterable]]
      }),
      methods: () => ({
        addLast: [dart.void, [E]],
        addFirst: [dart.void, [E]],
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        removeLast: [E, []],
        removeFirst: [E, []],
        remove: [core.bool, [core.Object]],
        [_filter]: [dart.void, [dart.functionType(core.bool, [E]), core.bool]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        lastEntry: [DoubleLinkedQueueEntry$(E), []],
        firstEntry: [DoubleLinkedQueueEntry$(E), []],
        clear: [dart.void, []],
        forEachEntry: [dart.void, [dart.functionType(dart.void, [DoubleLinkedQueueEntry$(E)])]]
      })
    });
    dart.defineExtensionMembers(DoubleLinkedQueue, [
      'length',
      'first',
      'last',
      'single',
      'isEmpty',
      'iterator'
    ]);
    return DoubleLinkedQueue;
  });
  let DoubleLinkedQueue = DoubleLinkedQueue$();
  const _nextEntry = Symbol('_nextEntry');
  const _DoubleLinkedQueueIterator$ = dart.generic(function(E) {
    class _DoubleLinkedQueueIterator extends core.Object {
      _DoubleLinkedQueueIterator(sentinel) {
        this[_sentinel] = sentinel;
        this[_nextEntry] = sentinel[_next];
        this[_current] = null;
      }
      moveNext() {
        if (!dart.notNull(core.identical(this[_nextEntry], this[_sentinel]))) {
          this[_current] = this[_nextEntry][_element];
          this[_nextEntry] = this[_nextEntry][_next];
          return true;
        }
        this[_current] = null;
        this[_nextEntry] = this[_sentinel] = null;
        return false;
      }
      get current() {
        return this[_current];
      }
    }
    _DoubleLinkedQueueIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(_DoubleLinkedQueueIterator, {
      constructors: () => ({_DoubleLinkedQueueIterator: [_DoubleLinkedQueueIterator$(E), [_DoubleLinkedQueueEntrySentinel$(E)]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _DoubleLinkedQueueIterator;
  });
  let _DoubleLinkedQueueIterator = _DoubleLinkedQueueIterator$();
  const _head = Symbol('_head');
  const _tail = Symbol('_tail');
  const _table = Symbol('_table');
  const _checkModification = Symbol('_checkModification');
  const _writeToList = Symbol('_writeToList');
  const _add = Symbol('_add');
  const _preGrow = Symbol('_preGrow');
  const _remove = Symbol('_remove');
  const _filterWhere = Symbol('_filterWhere');
  const _grow = Symbol('_grow');
  const ListQueue$ = dart.generic(function(E) {
    class ListQueue extends IterableBase$(E) {
      ListQueue(initialCapacity) {
        if (initialCapacity === void 0)
          initialCapacity = null;
        this[_head] = 0;
        this[_tail] = 0;
        this[_table] = null;
        this[_modificationCount] = 0;
        super.IterableBase();
        if (initialCapacity == null || dart.notNull(initialCapacity) < dart.notNull(ListQueue$()._INITIAL_CAPACITY)) {
          initialCapacity = ListQueue$()._INITIAL_CAPACITY;
        } else if (!dart.notNull(ListQueue$()._isPowerOf2(initialCapacity))) {
          initialCapacity = ListQueue$()._nextPowerOf2(initialCapacity);
        }
        dart.assert(ListQueue$()._isPowerOf2(initialCapacity));
        this[_table] = core.List$(E).new(initialCapacity);
      }
      static from(elements) {
        if (dart.is(elements, core.List)) {
          let length = elements[dartx.length];
          let queue = new (ListQueue$(E))(dart.notNull(length) + 1);
          dart.assert(dart.notNull(queue[_table][dartx.length]) > dart.notNull(length));
          let sourceList = elements;
          queue[_table][dartx.setRange](0, length, dart.as(sourceList, core.Iterable$(E)), 0);
          queue[_tail] = length;
          return queue;
        } else {
          let capacity = ListQueue$()._INITIAL_CAPACITY;
          if (dart.is(elements, _internal.EfficientLength)) {
            capacity = elements[dartx.length];
          }
          let result = new (ListQueue$(E))(capacity);
          for (let element of dart.as(elements, core.Iterable$(E))) {
            result.addLast(element);
          }
          return result;
        }
      }
      get iterator() {
        return new (_ListQueueIterator$(E))(this);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
        let modificationCount = this[_modificationCount];
        for (let i = this[_head]; i != this[_tail]; i = dart.notNull(i) + 1 & dart.notNull(this[_table][dartx.length]) - 1) {
          action(this[_table][dartx.get](i));
          this[_checkModification](modificationCount);
        }
      }
      get isEmpty() {
        return this[_head] == this[_tail];
      }
      get length() {
        return dart.notNull(this[_tail]) - dart.notNull(this[_head]) & dart.notNull(this[_table][dartx.length]) - 1;
      }
      get first() {
        if (this[_head] == this[_tail])
          dart.throw(_internal.IterableElementError.noElement());
        return this[_table][dartx.get](this[_head]);
      }
      get last() {
        if (this[_head] == this[_tail])
          dart.throw(_internal.IterableElementError.noElement());
        return this[_table][dartx.get](dart.notNull(this[_tail]) - 1 & dart.notNull(this[_table][dartx.length]) - 1);
      }
      get single() {
        if (this[_head] == this[_tail])
          dart.throw(_internal.IterableElementError.noElement());
        if (dart.notNull(this.length) > 1)
          dart.throw(_internal.IterableElementError.tooMany());
        return this[_table][dartx.get](this[_head]);
      }
      elementAt(index) {
        core.RangeError.checkValidIndex(index, this);
        return this[_table][dartx.get](dart.notNull(this[_head]) + dart.notNull(index) & dart.notNull(this[_table][dartx.length]) - 1);
      }
      toList(opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let list = null;
        if (dart.notNull(growable)) {
          list = core.List$(E).new();
          list[dartx.length] = this.length;
        } else {
          list = core.List$(E).new(this.length);
        }
        this[_writeToList](list);
        return list;
      }
      add(element) {
        dart.as(element, E);
        this[_add](element);
      }
      addAll(elements) {
        dart.as(elements, core.Iterable$(E));
        if (dart.is(elements, core.List)) {
          let list = dart.as(elements, core.List);
          let addCount = list[dartx.length];
          let length = this.length;
          if (dart.notNull(length) + dart.notNull(addCount) >= dart.notNull(this[_table][dartx.length])) {
            this[_preGrow](dart.notNull(length) + dart.notNull(addCount));
            this[_table][dartx.setRange](length, dart.notNull(length) + dart.notNull(addCount), dart.as(list, core.Iterable$(E)), 0);
            this[_tail] = dart.notNull(this[_tail]) + dart.notNull(addCount);
          } else {
            let endSpace = dart.notNull(this[_table][dartx.length]) - dart.notNull(this[_tail]);
            if (dart.notNull(addCount) < dart.notNull(endSpace)) {
              this[_table][dartx.setRange](this[_tail], dart.notNull(this[_tail]) + dart.notNull(addCount), dart.as(list, core.Iterable$(E)), 0);
              this[_tail] = dart.notNull(this[_tail]) + dart.notNull(addCount);
            } else {
              let preSpace = dart.notNull(addCount) - dart.notNull(endSpace);
              this[_table][dartx.setRange](this[_tail], dart.notNull(this[_tail]) + dart.notNull(endSpace), dart.as(list, core.Iterable$(E)), 0);
              this[_table][dartx.setRange](0, preSpace, dart.as(list, core.Iterable$(E)), endSpace);
              this[_tail] = preSpace;
            }
          }
          this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        } else {
          for (let element of elements)
            this[_add](element);
        }
      }
      remove(object) {
        for (let i = this[_head]; i != this[_tail]; i = dart.notNull(i) + 1 & dart.notNull(this[_table][dartx.length]) - 1) {
          let element = this[_table][dartx.get](i);
          if (dart.equals(element, object)) {
            this[_remove](i);
            this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
            return true;
          }
        }
        return false;
      }
      [_filterWhere](test, removeMatching) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let index = this[_head];
        let modificationCount = this[_modificationCount];
        let i = this[_head];
        while (i != this[_tail]) {
          let element = this[_table][dartx.get](i);
          let remove = core.identical(removeMatching, test(element));
          this[_checkModification](modificationCount);
          if (dart.notNull(remove)) {
            i = this[_remove](i);
            modificationCount = this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
          } else {
            i = dart.notNull(i) + 1 & dart.notNull(this[_table][dartx.length]) - 1;
          }
        }
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filterWhere](test, true);
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filterWhere](test, false);
      }
      clear() {
        if (this[_head] != this[_tail]) {
          for (let i = this[_head]; i != this[_tail]; i = dart.notNull(i) + 1 & dart.notNull(this[_table][dartx.length]) - 1) {
            this[_table][dartx.set](i, null);
          }
          this[_head] = this[_tail] = 0;
          this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        }
      }
      toString() {
        return IterableBase.iterableToFullString(this, "{", "}");
      }
      addLast(element) {
        dart.as(element, E);
        this[_add](element);
      }
      addFirst(element) {
        dart.as(element, E);
        this[_head] = dart.notNull(this[_head]) - 1 & dart.notNull(this[_table][dartx.length]) - 1;
        this[_table][dartx.set](this[_head], element);
        if (this[_head] == this[_tail])
          this[_grow]();
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
      }
      removeFirst() {
        if (this[_head] == this[_tail])
          dart.throw(_internal.IterableElementError.noElement());
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        let result = this[_table][dartx.get](this[_head]);
        this[_table][dartx.set](this[_head], null);
        this[_head] = dart.notNull(this[_head]) + 1 & dart.notNull(this[_table][dartx.length]) - 1;
        return result;
      }
      removeLast() {
        if (this[_head] == this[_tail])
          dart.throw(_internal.IterableElementError.noElement());
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        this[_tail] = dart.notNull(this[_tail]) - 1 & dart.notNull(this[_table][dartx.length]) - 1;
        let result = this[_table][dartx.get](this[_tail]);
        this[_table][dartx.set](this[_tail], null);
        return result;
      }
      static _isPowerOf2(number) {
        return (dart.notNull(number) & dart.notNull(number) - 1) == 0;
      }
      static _nextPowerOf2(number) {
        dart.assert(dart.notNull(number) > 0);
        number = (dart.notNull(number) << 1) - 1;
        for (;;) {
          let nextNumber = dart.notNull(number) & dart.notNull(number) - 1;
          if (nextNumber == 0)
            return number;
          number = nextNumber;
        }
      }
      [_checkModification](expectedModificationCount) {
        if (expectedModificationCount != this[_modificationCount]) {
          dart.throw(new core.ConcurrentModificationError(this));
        }
      }
      [_add](element) {
        dart.as(element, E);
        this[_table][dartx.set](this[_tail], element);
        this[_tail] = dart.notNull(this[_tail]) + 1 & dart.notNull(this[_table][dartx.length]) - 1;
        if (this[_head] == this[_tail])
          this[_grow]();
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
      }
      [_remove](offset) {
        let mask = dart.notNull(this[_table][dartx.length]) - 1;
        let startDistance = dart.notNull(offset) - dart.notNull(this[_head]) & dart.notNull(mask);
        let endDistance = dart.notNull(this[_tail]) - dart.notNull(offset) & dart.notNull(mask);
        if (dart.notNull(startDistance) < dart.notNull(endDistance)) {
          let i = offset;
          while (i != this[_head]) {
            let prevOffset = dart.notNull(i) - 1 & dart.notNull(mask);
            this[_table][dartx.set](i, this[_table][dartx.get](prevOffset));
            i = prevOffset;
          }
          this[_table][dartx.set](this[_head], null);
          this[_head] = dart.notNull(this[_head]) + 1 & dart.notNull(mask);
          return dart.notNull(offset) + 1 & dart.notNull(mask);
        } else {
          this[_tail] = dart.notNull(this[_tail]) - 1 & dart.notNull(mask);
          let i = offset;
          while (i != this[_tail]) {
            let nextOffset = dart.notNull(i) + 1 & dart.notNull(mask);
            this[_table][dartx.set](i, this[_table][dartx.get](nextOffset));
            i = nextOffset;
          }
          this[_table][dartx.set](this[_tail], null);
          return offset;
        }
      }
      [_grow]() {
        let newTable = core.List$(E).new(dart.notNull(this[_table][dartx.length]) * 2);
        let split = dart.notNull(this[_table][dartx.length]) - dart.notNull(this[_head]);
        newTable[dartx.setRange](0, split, this[_table], this[_head]);
        newTable[dartx.setRange](split, dart.notNull(split) + dart.notNull(this[_head]), this[_table], 0);
        this[_head] = 0;
        this[_tail] = this[_table][dartx.length];
        this[_table] = newTable;
      }
      [_writeToList](target) {
        dart.as(target, core.List$(E));
        dart.assert(dart.notNull(target[dartx.length]) >= dart.notNull(this.length));
        if (dart.notNull(this[_head]) <= dart.notNull(this[_tail])) {
          let length = dart.notNull(this[_tail]) - dart.notNull(this[_head]);
          target[dartx.setRange](0, length, this[_table], this[_head]);
          return length;
        } else {
          let firstPartSize = dart.notNull(this[_table][dartx.length]) - dart.notNull(this[_head]);
          target[dartx.setRange](0, firstPartSize, this[_table], this[_head]);
          target[dartx.setRange](firstPartSize, dart.notNull(firstPartSize) + dart.notNull(this[_tail]), this[_table], 0);
          return dart.notNull(this[_tail]) + dart.notNull(firstPartSize);
        }
      }
      [_preGrow](newElementCount) {
        dart.assert(dart.notNull(newElementCount) >= dart.notNull(this.length));
        newElementCount = dart.notNull(newElementCount) + (dart.notNull(newElementCount) >> 1);
        let newCapacity = ListQueue$()._nextPowerOf2(newElementCount);
        let newTable = core.List$(E).new(newCapacity);
        this[_tail] = this[_writeToList](newTable);
        this[_table] = newTable;
        this[_head] = 0;
      }
    }
    ListQueue[dart.implements] = () => [Queue$(E)];
    dart.setSignature(ListQueue, {
      constructors: () => ({
        ListQueue: [ListQueue$(E), [], [core.int]],
        from: [ListQueue$(E), [core.Iterable]]
      }),
      methods: () => ({
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        elementAt: [E, [core.int]],
        toList: [core.List$(E), [], {growable: core.bool}],
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [core.Object]],
        [_filterWhere]: [dart.void, [dart.functionType(core.bool, [E]), core.bool]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        clear: [dart.void, []],
        addLast: [dart.void, [E]],
        addFirst: [dart.void, [E]],
        removeFirst: [E, []],
        removeLast: [E, []],
        [_checkModification]: [dart.void, [core.int]],
        [_add]: [dart.void, [E]],
        [_remove]: [core.int, [core.int]],
        [_grow]: [dart.void, []],
        [_writeToList]: [core.int, [core.List$(E)]],
        [_preGrow]: [dart.void, [core.int]]
      }),
      statics: () => ({
        _isPowerOf2: [core.bool, [core.int]],
        _nextPowerOf2: [core.int, [core.int]]
      }),
      names: ['_isPowerOf2', '_nextPowerOf2']
    });
    dart.defineExtensionMembers(ListQueue, [
      'forEach',
      'elementAt',
      'toList',
      'iterator',
      'isEmpty',
      'length',
      'first',
      'last',
      'single'
    ]);
    return ListQueue;
  });
  let ListQueue = ListQueue$();
  ListQueue._INITIAL_CAPACITY = 8;
  const _queue = Symbol('_queue');
  const _end = Symbol('_end');
  const _position = Symbol('_position');
  const _ListQueueIterator$ = dart.generic(function(E) {
    class _ListQueueIterator extends core.Object {
      _ListQueueIterator(queue) {
        this[_queue] = queue;
        this[_end] = queue[_tail];
        this[_modificationCount] = queue[_modificationCount];
        this[_position] = queue[_head];
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        this[_queue][_checkModification](this[_modificationCount]);
        if (this[_position] == this[_end]) {
          this[_current] = null;
          return false;
        }
        this[_current] = dart.as(this[_queue][_table][dartx.get](this[_position]), E);
        this[_position] = dart.notNull(this[_position]) + 1 & dart.notNull(this[_queue][_table][dartx.length]) - 1;
        return true;
      }
    }
    _ListQueueIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(_ListQueueIterator, {
      constructors: () => ({_ListQueueIterator: [_ListQueueIterator$(E), [ListQueue]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _ListQueueIterator;
  });
  let _ListQueueIterator = _ListQueueIterator$();
  const _Predicate$ = dart.generic(function(T) {
    const _Predicate = dart.typedef('_Predicate', () => dart.functionType(core.bool, [T]));
    return _Predicate;
  });
  let _Predicate = _Predicate$();
  const _SplayTreeNode$ = dart.generic(function(K) {
    class _SplayTreeNode extends core.Object {
      _SplayTreeNode(key) {
        this.key = key;
        this.left = null;
        this.right = null;
      }
    }
    dart.setSignature(_SplayTreeNode, {
      constructors: () => ({_SplayTreeNode: [_SplayTreeNode$(K), [K]]})
    });
    return _SplayTreeNode;
  });
  let _SplayTreeNode = _SplayTreeNode$();
  const _SplayTreeMapNode$ = dart.generic(function(K, V) {
    class _SplayTreeMapNode extends _SplayTreeNode$(K) {
      _SplayTreeMapNode(key, value) {
        this.value = value;
        super._SplayTreeNode(key);
      }
    }
    dart.setSignature(_SplayTreeMapNode, {
      constructors: () => ({_SplayTreeMapNode: [_SplayTreeMapNode$(K, V), [K, V]]})
    });
    return _SplayTreeMapNode;
  });
  let _SplayTreeMapNode = _SplayTreeMapNode$();
  const _dummy = Symbol('_dummy');
  const _root = Symbol('_root');
  const _count = Symbol('_count');
  const _splayCount = Symbol('_splayCount');
  const _compare = Symbol('_compare');
  const _splay = Symbol('_splay');
  const _splayMin = Symbol('_splayMin');
  const _splayMax = Symbol('_splayMax');
  const _addNewRoot = Symbol('_addNewRoot');
  const _first = Symbol('_first');
  const _last = Symbol('_last');
  const _clear = Symbol('_clear');
  const _SplayTree$ = dart.generic(function(K) {
    class _SplayTree extends core.Object {
      _SplayTree() {
        this[_dummy] = new (_SplayTreeNode$(K))(null);
        this[_root] = null;
        this[_count] = 0;
        this[_modificationCount] = 0;
        this[_splayCount] = 0;
      }
      [_splay](key) {
        dart.as(key, K);
        if (this[_root] == null)
          return -1;
        let left = this[_dummy];
        let right = this[_dummy];
        let current = this[_root];
        let comp = null;
        while (true) {
          comp = this[_compare](current.key, key);
          if (dart.notNull(comp) > 0) {
            if (current.left == null)
              break;
            comp = this[_compare](current.left.key, key);
            if (dart.notNull(comp) > 0) {
              let tmp = current.left;
              current.left = tmp.right;
              tmp.right = current;
              current = tmp;
              if (current.left == null)
                break;
            }
            right.left = current;
            right = current;
            current = current.left;
          } else if (dart.notNull(comp) < 0) {
            if (current.right == null)
              break;
            comp = this[_compare](current.right.key, key);
            if (dart.notNull(comp) < 0) {
              let tmp = current.right;
              current.right = tmp.left;
              tmp.left = current;
              current = tmp;
              if (current.right == null)
                break;
            }
            left.right = current;
            left = current;
            current = current.right;
          } else {
            break;
          }
        }
        left.right = current.left;
        right.left = current.right;
        current.left = this[_dummy].right;
        current.right = this[_dummy].left;
        this[_root] = current;
        this[_dummy].right = null;
        this[_dummy].left = null;
        this[_splayCount] = dart.notNull(this[_splayCount]) + 1;
        return comp;
      }
      [_splayMin](node) {
        dart.as(node, _SplayTreeNode$(K));
        let current = node;
        while (current.left != null) {
          let left = current.left;
          current.left = left.right;
          left.right = current;
          current = left;
        }
        return dart.as(current, _SplayTreeNode$(K));
      }
      [_splayMax](node) {
        dart.as(node, _SplayTreeNode$(K));
        let current = node;
        while (current.right != null) {
          let right = current.right;
          current.right = right.left;
          right.left = current;
          current = right;
        }
        return dart.as(current, _SplayTreeNode$(K));
      }
      [_remove](key) {
        dart.as(key, K);
        if (this[_root] == null)
          return null;
        let comp = this[_splay](key);
        if (comp != 0)
          return null;
        let result = this[_root];
        this[_count] = dart.notNull(this[_count]) - 1;
        if (this[_root].left == null) {
          this[_root] = this[_root].right;
        } else {
          let right = this[_root].right;
          this[_root] = this[_splayMax](this[_root].left);
          this[_root].right = right;
        }
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        return result;
      }
      [_addNewRoot](node, comp) {
        dart.as(node, _SplayTreeNode$(K));
        this[_count] = dart.notNull(this[_count]) + 1;
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        if (this[_root] == null) {
          this[_root] = node;
          return;
        }
        if (dart.notNull(comp) < 0) {
          node.left = this[_root];
          node.right = this[_root].right;
          this[_root].right = null;
        } else {
          node.right = this[_root];
          node.left = this[_root].left;
          this[_root].left = null;
        }
        this[_root] = node;
      }
      get [_first]() {
        if (this[_root] == null)
          return null;
        this[_root] = this[_splayMin](this[_root]);
        return this[_root];
      }
      get [_last]() {
        if (this[_root] == null)
          return null;
        this[_root] = this[_splayMax](this[_root]);
        return this[_root];
      }
      [_clear]() {
        this[_root] = null;
        this[_count] = 0;
        this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
      }
    }
    dart.setSignature(_SplayTree, {
      methods: () => ({
        [_splay]: [core.int, [K]],
        [_splayMin]: [_SplayTreeNode$(K), [_SplayTreeNode$(K)]],
        [_splayMax]: [_SplayTreeNode$(K), [_SplayTreeNode$(K)]],
        [_remove]: [_SplayTreeNode, [K]],
        [_addNewRoot]: [dart.void, [_SplayTreeNode$(K), core.int]],
        [_clear]: [dart.void, []]
      })
    });
    return _SplayTree;
  });
  let _SplayTree = _SplayTree$();
  const _comparator = Symbol('_comparator');
  const _validKey = Symbol('_validKey');
  const SplayTreeMap$ = dart.generic(function(K, V) {
    class SplayTreeMap extends _SplayTree$(K) {
      SplayTreeMap(compare, isValidKey) {
        if (compare === void 0)
          compare = null;
        if (isValidKey === void 0)
          isValidKey = null;
        this[_comparator] = dart.as(compare == null ? core.Comparable.compare : compare, core.Comparator$(K));
        this[_validKey] = isValidKey != null ? isValidKey : dart.fn(v => dart.is(v, K), core.bool, [dart.dynamic]);
        super._SplayTree();
      }
      static from(other, compare, isValidKey) {
        if (compare === void 0)
          compare = null;
        if (isValidKey === void 0)
          isValidKey = null;
        let result = new (SplayTreeMap$(K, V))();
        other.forEach(dart.fn((k, v) => {
          result.set(dart.as(k, K), dart.as(v, V));
        }));
        return result;
      }
      static fromIterable(iterable, opts) {
        let key = opts && 'key' in opts ? opts.key : null;
        let value = opts && 'value' in opts ? opts.value : null;
        let compare = opts && 'compare' in opts ? opts.compare : null;
        let isValidKey = opts && 'isValidKey' in opts ? opts.isValidKey : null;
        let map = new (SplayTreeMap$(K, V))(compare, isValidKey);
        Maps._fillMapWithMappedIterable(map, iterable, key, value);
        return map;
      }
      static fromIterables(keys, values, compare, isValidKey) {
        if (compare === void 0)
          compare = null;
        if (isValidKey === void 0)
          isValidKey = null;
        let map = new (SplayTreeMap$(K, V))(compare, isValidKey);
        Maps._fillMapWithIterables(map, keys, values);
        return map;
      }
      [_compare](key1, key2) {
        dart.as(key1, K);
        dart.as(key2, K);
        return this[_comparator](key1, key2);
      }
      _internal() {
        this[_comparator] = null;
        this[_validKey] = null;
        super._SplayTree();
      }
      get(key) {
        if (key == null)
          dart.throw(new core.ArgumentError(key));
        if (!dart.notNull(this[_validKey](key)))
          return null;
        if (this[_root] != null) {
          let comp = this[_splay](dart.as(key, K));
          if (comp == 0) {
            let mapRoot = dart.as(this[_root], _SplayTreeMapNode);
            return dart.as(mapRoot.value, V);
          }
        }
        return null;
      }
      remove(key) {
        if (!dart.notNull(this[_validKey](key)))
          return null;
        let mapRoot = dart.as(this[_remove](dart.as(key, K)), _SplayTreeMapNode);
        if (mapRoot != null)
          return dart.as(mapRoot.value, V);
        return null;
      }
      set(key, value) {
        ((() => {
          dart.as(key, K);
          dart.as(value, V);
          if (key == null)
            dart.throw(new core.ArgumentError(key));
          let comp = this[_splay](key);
          if (comp == 0) {
            let mapRoot = dart.as(this[_root], _SplayTreeMapNode);
            mapRoot.value = value;
            return;
          }
          this[_addNewRoot](new (_SplayTreeMapNode$(K, dart.dynamic))(key, value), comp);
        }).bind(this))();
        return value;
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        if (key == null)
          dart.throw(new core.ArgumentError(key));
        let comp = this[_splay](key);
        if (comp == 0) {
          let mapRoot = dart.as(this[_root], _SplayTreeMapNode);
          return dart.as(mapRoot.value, V);
        }
        let modificationCount = this[_modificationCount];
        let splayCount = this[_splayCount];
        let value = ifAbsent();
        if (modificationCount != this[_modificationCount]) {
          dart.throw(new core.ConcurrentModificationError(this));
        }
        if (splayCount != this[_splayCount]) {
          comp = this[_splay](key);
          dart.assert(comp != 0);
        }
        this[_addNewRoot](new (_SplayTreeMapNode$(K, dart.dynamic))(key, value), comp);
        return value;
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        other.forEach(dart.fn(((key, value) => {
          dart.as(key, K);
          dart.as(value, V);
          this.set(key, value);
        }).bind(this), dart.dynamic, [K, V]));
      }
      get isEmpty() {
        return this[_root] == null;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [K, V]));
        let nodes = new (_SplayTreeNodeIterator$(K))(this);
        while (dart.notNull(nodes.moveNext())) {
          let node = dart.as(nodes.current, _SplayTreeMapNode$(K, V));
          f(node.key, node.value);
        }
      }
      get length() {
        return this[_count];
      }
      clear() {
        this[_clear]();
      }
      containsKey(key) {
        return dart.notNull(this[_validKey](key)) && this[_splay](dart.as(key, K)) == 0;
      }
      containsValue(value) {
        let found = false;
        let initialSplayCount = this[_splayCount];
        const visit = (function(node) {
          while (node != null) {
            if (dart.equals(node.value, value))
              return true;
            if (initialSplayCount != this[_splayCount]) {
              dart.throw(new core.ConcurrentModificationError(this));
            }
            if (node.right != null && dart.notNull(visit(dart.as(node.right, _SplayTreeMapNode))))
              return true;
            node = dart.as(node.left, _SplayTreeMapNode);
          }
          return false;
        }).bind(this);
        dart.fn(visit, core.bool, [_SplayTreeMapNode]);
        return visit(dart.as(this[_root], _SplayTreeMapNode));
      }
      get keys() {
        return new (_SplayTreeKeyIterable$(K))(this);
      }
      get values() {
        return new (_SplayTreeValueIterable$(K, V))(this);
      }
      toString() {
        return Maps.mapToString(this);
      }
      firstKey() {
        if (this[_root] == null)
          return null;
        return dart.as(this[_first].key, K);
      }
      lastKey() {
        if (this[_root] == null)
          return null;
        return dart.as(this[_last].key, K);
      }
      lastKeyBefore(key) {
        dart.as(key, K);
        if (key == null)
          dart.throw(new core.ArgumentError(key));
        if (this[_root] == null)
          return null;
        let comp = this[_splay](key);
        if (dart.notNull(comp) < 0)
          return this[_root].key;
        let node = this[_root].left;
        if (node == null)
          return null;
        while (node.right != null) {
          node = node.right;
        }
        return node.key;
      }
      firstKeyAfter(key) {
        dart.as(key, K);
        if (key == null)
          dart.throw(new core.ArgumentError(key));
        if (this[_root] == null)
          return null;
        let comp = this[_splay](key);
        if (dart.notNull(comp) > 0)
          return this[_root].key;
        let node = this[_root].right;
        if (node == null)
          return null;
        while (node.left != null) {
          node = node.left;
        }
        return node.key;
      }
    }
    SplayTreeMap[dart.implements] = () => [core.Map$(K, V)];
    dart.defineNamedConstructor(SplayTreeMap, '_internal');
    dart.setSignature(SplayTreeMap, {
      constructors: () => ({
        SplayTreeMap: [SplayTreeMap$(K, V), [], [dart.functionType(core.int, [K, K]), dart.functionType(core.bool, [core.Object])]],
        from: [SplayTreeMap$(K, V), [core.Map], [dart.functionType(core.int, [K, K]), dart.functionType(core.bool, [core.Object])]],
        fromIterable: [SplayTreeMap$(K, V), [core.Iterable], {key: dart.functionType(K, [dart.dynamic]), value: dart.functionType(V, [dart.dynamic]), compare: dart.functionType(core.int, [K, K]), isValidKey: dart.functionType(core.bool, [core.Object])}],
        fromIterables: [SplayTreeMap$(K, V), [core.Iterable$(K), core.Iterable$(V)], [dart.functionType(core.int, [K, K]), dart.functionType(core.bool, [core.Object])]],
        _internal: [SplayTreeMap$(K, V), []]
      }),
      methods: () => ({
        [_compare]: [core.int, [K, K]],
        get: [V, [core.Object]],
        remove: [V, [core.Object]],
        set: [dart.void, [K, V]],
        putIfAbsent: [V, [K, dart.functionType(V, [])]],
        addAll: [dart.void, [core.Map$(K, V)]],
        forEach: [dart.void, [dart.functionType(dart.void, [K, V])]],
        clear: [dart.void, []],
        containsKey: [core.bool, [core.Object]],
        containsValue: [core.bool, [core.Object]],
        firstKey: [K, []],
        lastKey: [K, []],
        lastKeyBefore: [K, [K]],
        firstKeyAfter: [K, [K]]
      })
    });
    return SplayTreeMap;
  });
  let SplayTreeMap = SplayTreeMap$();
  const _workList = Symbol('_workList');
  const _tree = Symbol('_tree');
  const _currentNode = Symbol('_currentNode');
  const _findLeftMostDescendent = Symbol('_findLeftMostDescendent');
  const _getValue = Symbol('_getValue');
  const _rebuildWorkList = Symbol('_rebuildWorkList');
  const _SplayTreeIterator$ = dart.generic(function(T) {
    class _SplayTreeIterator extends core.Object {
      _SplayTreeIterator(tree) {
        this[_workList] = dart.list([], _SplayTreeNode);
        this[_tree] = tree;
        this[_modificationCount] = tree[_modificationCount];
        this[_splayCount] = tree[_splayCount];
        this[_currentNode] = null;
        this[_findLeftMostDescendent](tree[_root]);
      }
      startAt(tree, startKey) {
        this[_workList] = dart.list([], _SplayTreeNode);
        this[_tree] = tree;
        this[_modificationCount] = tree[_modificationCount];
        this[_splayCount] = null;
        this[_currentNode] = null;
        if (tree[_root] == null)
          return;
        let compare = tree[_splay](startKey);
        this[_splayCount] = tree[_splayCount];
        if (dart.notNull(compare) < 0) {
          this[_findLeftMostDescendent](tree[_root].right);
        } else {
          this[_workList][dartx.add](tree[_root]);
        }
      }
      get current() {
        if (this[_currentNode] == null)
          return null;
        return this[_getValue](dart.as(this[_currentNode], _SplayTreeMapNode));
      }
      [_findLeftMostDescendent](node) {
        while (node != null) {
          this[_workList][dartx.add](node);
          node = node.left;
        }
      }
      [_rebuildWorkList](currentNode) {
        dart.assert(!dart.notNull(this[_workList][dartx.isEmpty]));
        this[_workList][dartx.clear]();
        if (currentNode == null) {
          this[_findLeftMostDescendent](this[_tree][_root]);
        } else {
          this[_tree][_splay](currentNode.key);
          this[_findLeftMostDescendent](this[_tree][_root].right);
          dart.assert(!dart.notNull(this[_workList][dartx.isEmpty]));
        }
      }
      moveNext() {
        if (this[_modificationCount] != this[_tree][_modificationCount]) {
          dart.throw(new core.ConcurrentModificationError(this[_tree]));
        }
        if (dart.notNull(this[_workList][dartx.isEmpty])) {
          this[_currentNode] = null;
          return false;
        }
        if (this[_tree][_splayCount] != this[_splayCount] && this[_currentNode] != null) {
          this[_rebuildWorkList](this[_currentNode]);
        }
        this[_currentNode] = this[_workList][dartx.removeLast]();
        this[_findLeftMostDescendent](this[_currentNode].right);
        return true;
      }
    }
    _SplayTreeIterator[dart.implements] = () => [core.Iterator$(T)];
    dart.defineNamedConstructor(_SplayTreeIterator, 'startAt');
    dart.setSignature(_SplayTreeIterator, {
      constructors: () => ({
        _SplayTreeIterator: [_SplayTreeIterator$(T), [_SplayTree]],
        startAt: [_SplayTreeIterator$(T), [_SplayTree, dart.dynamic]]
      }),
      methods: () => ({
        [_findLeftMostDescendent]: [dart.void, [_SplayTreeNode]],
        [_rebuildWorkList]: [dart.void, [_SplayTreeNode]],
        moveNext: [core.bool, []]
      })
    });
    return _SplayTreeIterator;
  });
  let _SplayTreeIterator = _SplayTreeIterator$();
  const _copyNode = Symbol('_copyNode');
  const _SplayTreeKeyIterable$ = dart.generic(function(K) {
    class _SplayTreeKeyIterable extends IterableBase$(K) {
      _SplayTreeKeyIterable(tree) {
        this[_tree] = tree;
        super.IterableBase();
      }
      get length() {
        return this[_tree][_count];
      }
      get isEmpty() {
        return this[_tree][_count] == 0;
      }
      get iterator() {
        return new (_SplayTreeKeyIterator$(K))(this[_tree]);
      }
      toSet() {
        let setOrMap = this[_tree];
        let set = new (SplayTreeSet$(K))(setOrMap[_comparator], setOrMap[_validKey]);
        set[_count] = this[_tree][_count];
        set[_root] = set[_copyNode](this[_tree][_root]);
        return set;
      }
    }
    _SplayTreeKeyIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(_SplayTreeKeyIterable, {
      constructors: () => ({_SplayTreeKeyIterable: [_SplayTreeKeyIterable$(K), [_SplayTree$(K)]]}),
      methods: () => ({toSet: [core.Set$(K), []]})
    });
    dart.defineExtensionMembers(_SplayTreeKeyIterable, ['toSet', 'length', 'isEmpty', 'iterator']);
    return _SplayTreeKeyIterable;
  });
  let _SplayTreeKeyIterable = _SplayTreeKeyIterable$();
  const _SplayTreeValueIterable$ = dart.generic(function(K, V) {
    class _SplayTreeValueIterable extends IterableBase$(V) {
      _SplayTreeValueIterable(map) {
        this[_map] = map;
        super.IterableBase();
      }
      get length() {
        return this[_map][_count];
      }
      get isEmpty() {
        return this[_map][_count] == 0;
      }
      get iterator() {
        return new (_SplayTreeValueIterator$(K, V))(this[_map]);
      }
    }
    _SplayTreeValueIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(_SplayTreeValueIterable, {
      constructors: () => ({_SplayTreeValueIterable: [_SplayTreeValueIterable$(K, V), [SplayTreeMap$(K, V)]]})
    });
    dart.defineExtensionMembers(_SplayTreeValueIterable, ['length', 'isEmpty', 'iterator']);
    return _SplayTreeValueIterable;
  });
  let _SplayTreeValueIterable = _SplayTreeValueIterable$();
  const _SplayTreeKeyIterator$ = dart.generic(function(K) {
    class _SplayTreeKeyIterator extends _SplayTreeIterator$(K) {
      _SplayTreeKeyIterator(map) {
        super._SplayTreeIterator(map);
      }
      [_getValue](node) {
        return dart.as(node.key, K);
      }
    }
    dart.setSignature(_SplayTreeKeyIterator, {
      constructors: () => ({_SplayTreeKeyIterator: [_SplayTreeKeyIterator$(K), [_SplayTree$(K)]]}),
      methods: () => ({[_getValue]: [K, [_SplayTreeNode]]})
    });
    return _SplayTreeKeyIterator;
  });
  let _SplayTreeKeyIterator = _SplayTreeKeyIterator$();
  const _SplayTreeValueIterator$ = dart.generic(function(K, V) {
    class _SplayTreeValueIterator extends _SplayTreeIterator$(V) {
      _SplayTreeValueIterator(map) {
        super._SplayTreeIterator(map);
      }
      [_getValue](node) {
        return dart.as(node.value, V);
      }
    }
    dart.setSignature(_SplayTreeValueIterator, {
      constructors: () => ({_SplayTreeValueIterator: [_SplayTreeValueIterator$(K, V), [SplayTreeMap$(K, V)]]}),
      methods: () => ({[_getValue]: [V, [_SplayTreeMapNode]]})
    });
    return _SplayTreeValueIterator;
  });
  let _SplayTreeValueIterator = _SplayTreeValueIterator$();
  const _SplayTreeNodeIterator$ = dart.generic(function(K) {
    class _SplayTreeNodeIterator extends _SplayTreeIterator$(_SplayTreeNode$(K)) {
      _SplayTreeNodeIterator(tree) {
        super._SplayTreeIterator(tree);
      }
      startAt(tree, startKey) {
        super.startAt(tree, startKey);
      }
      [_getValue](node) {
        return dart.as(node, _SplayTreeNode$(K));
      }
    }
    dart.defineNamedConstructor(_SplayTreeNodeIterator, 'startAt');
    dart.setSignature(_SplayTreeNodeIterator, {
      constructors: () => ({
        _SplayTreeNodeIterator: [_SplayTreeNodeIterator$(K), [_SplayTree$(K)]],
        startAt: [_SplayTreeNodeIterator$(K), [_SplayTree$(K), dart.dynamic]]
      }),
      methods: () => ({[_getValue]: [_SplayTreeNode$(K), [_SplayTreeNode]]})
    });
    return _SplayTreeNodeIterator;
  });
  let _SplayTreeNodeIterator = _SplayTreeNodeIterator$();
  const _clone = Symbol('_clone');
  const SplayTreeSet$ = dart.generic(function(E) {
    class SplayTreeSet extends dart.mixin(_SplayTree$(E), IterableMixin$(E), SetMixin$(E)) {
      SplayTreeSet(compare, isValidKey) {
        if (compare === void 0)
          compare = null;
        if (isValidKey === void 0)
          isValidKey = null;
        this[_comparator] = dart.as(compare == null ? core.Comparable.compare : compare, core.Comparator$(E));
        this[_validKey] = isValidKey != null ? isValidKey : dart.fn(v => dart.is(v, E), core.bool, [dart.dynamic]);
        super._SplayTree();
      }
      static from(elements, compare, isValidKey) {
        if (compare === void 0)
          compare = null;
        if (isValidKey === void 0)
          isValidKey = null;
        let result = new (SplayTreeSet$(E))(compare, isValidKey);
        for (let element of dart.as(elements, core.Iterable$(E))) {
          result.add(element);
        }
        return result;
      }
      [_compare](e1, e2) {
        dart.as(e1, E);
        dart.as(e2, E);
        return this[_comparator](e1, e2);
      }
      get iterator() {
        return new (_SplayTreeKeyIterator$(E))(this);
      }
      get length() {
        return this[_count];
      }
      get isEmpty() {
        return this[_root] == null;
      }
      get isNotEmpty() {
        return this[_root] != null;
      }
      get first() {
        if (this[_count] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        return dart.as(this[_first].key, E);
      }
      get last() {
        if (this[_count] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        return dart.as(this[_last].key, E);
      }
      get single() {
        if (this[_count] == 0)
          dart.throw(_internal.IterableElementError.noElement());
        if (dart.notNull(this[_count]) > 1)
          dart.throw(_internal.IterableElementError.tooMany());
        return this[_root].key;
      }
      contains(object) {
        return dart.notNull(this[_validKey](object)) && this[_splay](dart.as(object, E)) == 0;
      }
      add(element) {
        dart.as(element, E);
        let compare = this[_splay](element);
        if (compare == 0)
          return false;
        this[_addNewRoot](new (_SplayTreeNode$(E))(element), compare);
        return true;
      }
      remove(object) {
        if (!dart.notNull(this[_validKey](object)))
          return false;
        return this[_remove](dart.as(object, E)) != null;
      }
      addAll(elements) {
        dart.as(elements, core.Iterable$(E));
        for (let element of elements) {
          let compare = this[_splay](element);
          if (compare != 0) {
            this[_addNewRoot](new (_SplayTreeNode$(E))(element), compare);
          }
        }
      }
      removeAll(elements) {
        for (let element of elements) {
          if (dart.notNull(this[_validKey](element)))
            this[_remove](dart.as(element, E));
        }
      }
      retainAll(elements) {
        let retainSet = new (SplayTreeSet$(E))(this[_comparator], this[_validKey]);
        let modificationCount = this[_modificationCount];
        for (let object of elements) {
          if (modificationCount != this[_modificationCount]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          if (dart.notNull(this[_validKey](object)) && this[_splay](dart.as(object, E)) == 0)
            retainSet.add(this[_root].key);
        }
        if (retainSet[_count] != this[_count]) {
          this[_root] = retainSet[_root];
          this[_count] = retainSet[_count];
          this[_modificationCount] = dart.notNull(this[_modificationCount]) + 1;
        }
      }
      lookup(object) {
        if (!dart.notNull(this[_validKey](object)))
          return null;
        let comp = this[_splay](dart.as(object, E));
        if (comp != 0)
          return null;
        return this[_root].key;
      }
      intersection(other) {
        let result = new (SplayTreeSet$(E))(this[_comparator], this[_validKey]);
        for (let element of this) {
          if (dart.notNull(other.contains(element)))
            result.add(element);
        }
        return result;
      }
      difference(other) {
        let result = new (SplayTreeSet$(E))(this[_comparator], this[_validKey]);
        for (let element of this) {
          if (!dart.notNull(other.contains(element)))
            result.add(element);
        }
        return result;
      }
      union(other) {
        dart.as(other, core.Set$(E));
        let _ = this[_clone]();
        _.addAll(other);
        return _;
      }
      [_clone]() {
        let set = new (SplayTreeSet$(E))(this[_comparator], this[_validKey]);
        set[_count] = this[_count];
        set[_root] = this[_copyNode](this[_root]);
        return set;
      }
      [_copyNode](node) {
        dart.as(node, _SplayTreeNode$(E));
        if (node == null)
          return null;
        let _ = new (_SplayTreeNode$(E))(node.key);
        _.left = this[_copyNode](node.left);
        _.right = this[_copyNode](node.right);
        return _;
      }
      clear() {
        this[_clear]();
      }
      toSet() {
        return this[_clone]();
      }
      toString() {
        return IterableBase.iterableToFullString(this, '{', '}');
      }
    }
    dart.setSignature(SplayTreeSet, {
      constructors: () => ({
        SplayTreeSet: [SplayTreeSet$(E), [], [dart.functionType(core.int, [E, E]), dart.functionType(core.bool, [core.Object])]],
        from: [SplayTreeSet$(E), [core.Iterable], [dart.functionType(core.int, [E, E]), dart.functionType(core.bool, [core.Object])]]
      }),
      methods: () => ({
        [_compare]: [core.int, [E, E]],
        contains: [core.bool, [core.Object]],
        add: [core.bool, [E]],
        remove: [core.bool, [core.Object]],
        addAll: [dart.void, [core.Iterable$(E)]],
        lookup: [E, [core.Object]],
        intersection: [core.Set$(E), [core.Set$(core.Object)]],
        difference: [core.Set$(E), [core.Set$(core.Object)]],
        union: [core.Set$(E), [core.Set$(E)]],
        [_clone]: [SplayTreeSet$(E), []],
        [_copyNode]: [_SplayTreeNode$(E), [_SplayTreeNode$(E)]],
        toSet: [core.Set$(E), []]
      })
    });
    dart.defineExtensionMembers(SplayTreeSet, [
      'contains',
      'toSet',
      'iterator',
      'length',
      'isEmpty',
      'isNotEmpty',
      'first',
      'last',
      'single'
    ]);
    return SplayTreeSet;
  });
  let SplayTreeSet = SplayTreeSet$();
  const _strings = Symbol('_strings');
  const _nums = Symbol('_nums');
  const _rest = Symbol('_rest');
  const _containsKey = Symbol('_containsKey');
  const _getBucket = Symbol('_getBucket');
  const _findBucketIndex = Symbol('_findBucketIndex');
  const _computeKeys = Symbol('_computeKeys');
  const _get = Symbol('_get');
  const _addHashTableEntry = Symbol('_addHashTableEntry');
  const _set = Symbol('_set');
  const _computeHashCode = Symbol('_computeHashCode');
  const _removeHashTableEntry = Symbol('_removeHashTableEntry');
  const _HashMap$ = dart.generic(function(K, V) {
    class _HashMap extends core.Object {
      _HashMap() {
        this[_length] = 0;
        this[_strings] = null;
        this[_nums] = null;
        this[_rest] = null;
        this[_keys] = null;
      }
      get length() {
        return this[_length];
      }
      get isEmpty() {
        return this[_length] == 0;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      get keys() {
        return new (HashMapKeyIterable$(K))(this);
      }
      get values() {
        return _internal.MappedIterable$(K, V).new(this.keys, dart.fn((each => this.get(each)).bind(this), V, [dart.dynamic]));
      }
      containsKey(key) {
        if (dart.notNull(_HashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          return strings == null ? false : _HashMap$()._hasTableEntry(strings, key);
        } else if (dart.notNull(_HashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          return nums == null ? false : _HashMap$()._hasTableEntry(nums, key);
        } else {
          return this[_containsKey](key);
        }
      }
      [_containsKey](key) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, key);
        return dart.notNull(this[_findBucketIndex](bucket, key)) >= 0;
      }
      containsValue(value) {
        return this[_computeKeys]()[dartx.any](dart.fn((each => dart.equals(this.get(each), value)).bind(this), core.bool, [dart.dynamic]));
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        other.forEach(dart.fn(((key, value) => {
          dart.as(key, K);
          dart.as(value, V);
          this.set(key, value);
        }).bind(this), dart.dynamic, [K, V]));
      }
      get(key) {
        if (dart.notNull(_HashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          return strings == null ? null : dart.as(_HashMap$()._getTableEntry(strings, key), V);
        } else if (dart.notNull(_HashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          return nums == null ? null : dart.as(_HashMap$()._getTableEntry(nums, key), V);
        } else {
          return this[_get](key);
        }
      }
      [_get](key) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, key);
        let index = this[_findBucketIndex](bucket, key);
        return dart.notNull(index) < 0 ? null : dart.as(bucket[dart.notNull(index) + 1], V);
      }
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        if (dart.notNull(_HashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          if (strings == null)
            this[_strings] = strings = _HashMap$()._newHashTable();
          this[_addHashTableEntry](strings, key, value);
        } else if (dart.notNull(_HashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          if (nums == null)
            this[_nums] = nums = _HashMap$()._newHashTable();
          this[_addHashTableEntry](nums, key, value);
        } else {
          this[_set](key, value);
        }
        return value;
      }
      [_set](key, value) {
        dart.as(key, K);
        dart.as(value, V);
        let rest = this[_rest];
        if (rest == null)
          this[_rest] = rest = _HashMap$()._newHashTable();
        let hash = this[_computeHashCode](key);
        let bucket = rest[hash];
        if (bucket == null) {
          _HashMap$()._setTableEntry(rest, hash, [key, value]);
          this[_length] = dart.notNull(this[_length]) + 1;
          this[_keys] = null;
        } else {
          let index = this[_findBucketIndex](bucket, key);
          if (dart.notNull(index) >= 0) {
            bucket[dart.notNull(index) + 1] = value;
          } else {
            bucket.push(key, value);
            this[_length] = dart.notNull(this[_length]) + 1;
            this[_keys] = null;
          }
        }
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        if (dart.notNull(this.containsKey(key)))
          return this.get(key);
        let value = ifAbsent();
        this.set(key, value);
        return value;
      }
      remove(key) {
        if (dart.notNull(_HashMap$()._isStringKey(key))) {
          return this[_removeHashTableEntry](this[_strings], key);
        } else if (dart.notNull(_HashMap$()._isNumericKey(key))) {
          return this[_removeHashTableEntry](this[_nums], key);
        } else {
          return this[_remove](key);
        }
      }
      [_remove](key) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, key);
        let index = this[_findBucketIndex](bucket, key);
        if (dart.notNull(index) < 0)
          return null;
        this[_length] = dart.notNull(this[_length]) - 1;
        this[_keys] = null;
        return dart.as(bucket.splice(index, 2)[1], V);
      }
      clear() {
        if (dart.notNull(this[_length]) > 0) {
          this[_strings] = this[_nums] = this[_rest] = this[_keys] = null;
          this[_length] = 0;
        }
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [K, V]));
        let keys = this[_computeKeys]();
        for (let i = 0, length = keys[dartx.length]; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let key = keys[i];
          action(dart.as(key, K), this.get(key));
          if (keys !== this[_keys]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
        }
      }
      [_computeKeys]() {
        if (this[_keys] != null)
          return this[_keys];
        let result = core.List.new(this[_length]);
        let index = 0;
        let strings = this[_strings];
        if (strings != null) {
          let names = Object.getOwnPropertyNames(strings);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let key = names[i];
            result[index] = key;
            index = dart.notNull(index) + 1;
          }
        }
        let nums = this[_nums];
        if (nums != null) {
          let names = Object.getOwnPropertyNames(nums);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let key = +names[i];
            result[index] = key;
            index = dart.notNull(index) + 1;
          }
        }
        let rest = this[_rest];
        if (rest != null) {
          let names = Object.getOwnPropertyNames(rest);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let key = names[i];
            let bucket = rest[key];
            let length = bucket.length;
            for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 2) {
              let key = bucket[i];
              result[index] = key;
              index = dart.notNull(index) + 1;
            }
          }
        }
        dart.assert(index == this[_length]);
        return this[_keys] = result;
      }
      [_addHashTableEntry](table, key, value) {
        dart.as(key, K);
        dart.as(value, V);
        if (!dart.notNull(_HashMap$()._hasTableEntry(table, key))) {
          this[_length] = dart.notNull(this[_length]) + 1;
          this[_keys] = null;
        }
        _HashMap$()._setTableEntry(table, key, value);
      }
      [_removeHashTableEntry](table, key) {
        if (table != null && dart.notNull(_HashMap$()._hasTableEntry(table, key))) {
          let value = dart.as(_HashMap$()._getTableEntry(table, key), V);
          _HashMap$()._deleteTableEntry(table, key);
          this[_length] = dart.notNull(this[_length]) - 1;
          this[_keys] = null;
          return value;
        } else {
          return null;
        }
      }
      static _isStringKey(key) {
        return typeof key == 'string' && !dart.equals(key, '__proto__');
      }
      static _isNumericKey(key) {
        return typeof key == 'number' && (key & 0x3ffffff) === key;
      }
      [_computeHashCode](key) {
        return dart.hashCode(key) & 0x3ffffff;
      }
      static _hasTableEntry(table, key) {
        let entry = table[key];
        return entry != null;
      }
      static _getTableEntry(table, key) {
        let entry = table[key];
        return entry === table ? null : entry;
      }
      static _setTableEntry(table, key, value) {
        if (value == null) {
          table[key] = table;
        } else {
          table[key] = value;
        }
      }
      static _deleteTableEntry(table, key) {
        delete table[key];
      }
      [_getBucket](table, key) {
        let hash = this[_computeHashCode](key);
        return dart.as(table[hash], core.List);
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 2) {
          if (dart.equals(bucket[i], key))
            return i;
        }
        return -1;
      }
      static _newHashTable() {
        let table = Object.create(null);
        let temporaryKey = '<non-identifier-key>';
        _HashMap$()._setTableEntry(table, temporaryKey, table);
        _HashMap$()._deleteTableEntry(table, temporaryKey);
        return table;
      }
    }
    _HashMap[dart.implements] = () => [HashMap$(K, V)];
    dart.setSignature(_HashMap, {
      constructors: () => ({_HashMap: [_HashMap$(K, V), []]}),
      methods: () => ({
        containsKey: [core.bool, [core.Object]],
        [_containsKey]: [core.bool, [core.Object]],
        containsValue: [core.bool, [core.Object]],
        addAll: [dart.void, [core.Map$(K, V)]],
        get: [V, [core.Object]],
        [_get]: [V, [core.Object]],
        set: [dart.void, [K, V]],
        [_set]: [dart.void, [K, V]],
        putIfAbsent: [V, [K, dart.functionType(V, [])]],
        remove: [V, [core.Object]],
        [_remove]: [V, [core.Object]],
        clear: [dart.void, []],
        forEach: [dart.void, [dart.functionType(dart.void, [K, V])]],
        [_computeKeys]: [core.List, []],
        [_addHashTableEntry]: [dart.void, [dart.dynamic, K, V]],
        [_removeHashTableEntry]: [V, [dart.dynamic, core.Object]],
        [_computeHashCode]: [core.int, [dart.dynamic]],
        [_getBucket]: [core.List, [dart.dynamic, dart.dynamic]],
        [_findBucketIndex]: [core.int, [dart.dynamic, dart.dynamic]]
      }),
      statics: () => ({
        _isStringKey: [core.bool, [dart.dynamic]],
        _isNumericKey: [core.bool, [dart.dynamic]],
        _hasTableEntry: [core.bool, [dart.dynamic, dart.dynamic]],
        _getTableEntry: [dart.dynamic, [dart.dynamic, dart.dynamic]],
        _setTableEntry: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
        _deleteTableEntry: [dart.void, [dart.dynamic, dart.dynamic]],
        _newHashTable: [dart.dynamic, []]
      }),
      names: ['_isStringKey', '_isNumericKey', '_hasTableEntry', '_getTableEntry', '_setTableEntry', '_deleteTableEntry', '_newHashTable']
    });
    return _HashMap;
  });
  let _HashMap = _HashMap$();
  const _IdentityHashMap$ = dart.generic(function(K, V) {
    class _IdentityHashMap extends _HashMap$(K, V) {
      _IdentityHashMap() {
        super._HashMap();
      }
      [_computeHashCode](key) {
        return core.identityHashCode(key) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 2) {
          if (dart.notNull(core.identical(bucket[i], key)))
            return i;
        }
        return -1;
      }
    }
    return _IdentityHashMap;
  });
  let _IdentityHashMap = _IdentityHashMap$();
  const _equals = Symbol('_equals');
  const _hashCode = Symbol('_hashCode');
  const _CustomHashMap$ = dart.generic(function(K, V) {
    class _CustomHashMap extends _HashMap$(K, V) {
      _CustomHashMap(equals, hashCode, validKey) {
        this[_equals] = equals;
        this[_hashCode] = hashCode;
        this[_validKey] = validKey != null ? validKey : dart.fn(v => dart.is(v, K), core.bool, [dart.dynamic]);
        super._HashMap();
      }
      get(key) {
        if (!dart.notNull(this[_validKey](key)))
          return null;
        return super[_get](key);
      }
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        super[_set](key, value);
        return value;
      }
      containsKey(key) {
        if (!dart.notNull(this[_validKey](key)))
          return false;
        return super[_containsKey](key);
      }
      remove(key) {
        if (!dart.notNull(this[_validKey](key)))
          return null;
        return super[_remove](key);
      }
      [_computeHashCode](key) {
        return this[_hashCode](dart.as(key, K)) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 2) {
          if (dart.notNull(this[_equals](dart.as(bucket[i], K), dart.as(key, K))))
            return i;
        }
        return -1;
      }
      toString() {
        return Maps.mapToString(this);
      }
    }
    dart.setSignature(_CustomHashMap, {
      constructors: () => ({_CustomHashMap: [_CustomHashMap$(K, V), [_Equality$(K), _Hasher$(K), dart.functionType(core.bool, [core.Object])]]}),
      methods: () => ({
        get: [V, [core.Object]],
        set: [dart.void, [K, V]],
        remove: [V, [core.Object]]
      })
    });
    return _CustomHashMap;
  });
  let _CustomHashMap = _CustomHashMap$();
  const HashMapKeyIterable$ = dart.generic(function(E) {
    class HashMapKeyIterable extends IterableBase$(E) {
      HashMapKeyIterable(map) {
        this[_map] = map;
        super.IterableBase();
      }
      get length() {
        return dart.as(dart.dload(this[_map], _length), core.int);
      }
      get isEmpty() {
        return dart.equals(dart.dload(this[_map], _length), 0);
      }
      get iterator() {
        return new (HashMapKeyIterator$(E))(this[_map], dart.as(dart.dsend(this[_map], _computeKeys), core.List));
      }
      contains(element) {
        return dart.as(dart.dsend(this[_map], 'containsKey', element), core.bool);
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        let keys = dart.as(dart.dsend(this[_map], _computeKeys), core.List);
        for (let i = 0, length = keys.length; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          f(dart.as(keys[i], E));
          if (keys !== dart.dload(this[_map], _keys)) {
            dart.throw(new core.ConcurrentModificationError(this[_map]));
          }
        }
      }
    }
    HashMapKeyIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(HashMapKeyIterable, {
      constructors: () => ({HashMapKeyIterable: [HashMapKeyIterable$(E), [dart.dynamic]]}),
      methods: () => ({forEach: [dart.void, [dart.functionType(dart.void, [E])]]})
    });
    dart.defineExtensionMembers(HashMapKeyIterable, [
      'contains',
      'forEach',
      'length',
      'isEmpty',
      'iterator'
    ]);
    return HashMapKeyIterable;
  });
  let HashMapKeyIterable = HashMapKeyIterable$();
  const _offset = Symbol('_offset');
  const HashMapKeyIterator$ = dart.generic(function(E) {
    class HashMapKeyIterator extends core.Object {
      HashMapKeyIterator(map, keys) {
        this[_map] = map;
        this[_keys] = keys;
        this[_offset] = 0;
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        let keys = this[_keys];
        let offset = this[_offset];
        if (keys !== dart.dload(this[_map], _keys)) {
          dart.throw(new core.ConcurrentModificationError(this[_map]));
        } else if (dart.notNull(offset) >= keys.length) {
          this[_current] = null;
          return false;
        } else {
          this[_current] = dart.as(keys[offset], E);
          this[_offset] = dart.notNull(offset) + 1;
          return true;
        }
      }
    }
    HashMapKeyIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(HashMapKeyIterator, {
      constructors: () => ({HashMapKeyIterator: [HashMapKeyIterator$(E), [dart.dynamic, core.List]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return HashMapKeyIterator;
  });
  let HashMapKeyIterator = HashMapKeyIterator$();
  const _modifications = Symbol('_modifications');
  const _value = Symbol('_value');
  const _newLinkedCell = Symbol('_newLinkedCell');
  const _unlinkCell = Symbol('_unlinkCell');
  const _modified = Symbol('_modified');
  const _key = Symbol('_key');
  const _LinkedHashMap$ = dart.generic(function(K, V) {
    class _LinkedHashMap extends core.Object {
      _LinkedHashMap() {
        this[_length] = 0;
        this[_strings] = null;
        this[_nums] = null;
        this[_rest] = null;
        this[_first] = null;
        this[_last] = null;
        this[_modifications] = 0;
      }
      get length() {
        return this[_length];
      }
      get isEmpty() {
        return this[_length] == 0;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      get keys() {
        return new (LinkedHashMapKeyIterable$(K))(this);
      }
      get values() {
        return _internal.MappedIterable$(K, V).new(this.keys, dart.fn((each => this.get(each)).bind(this), V, [dart.dynamic]));
      }
      containsKey(key) {
        if (dart.notNull(_LinkedHashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          if (strings == null)
            return false;
          let cell = dart.as(_LinkedHashMap$()._getTableEntry(strings, key), LinkedHashMapCell);
          return cell != null;
        } else if (dart.notNull(_LinkedHashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          if (nums == null)
            return false;
          let cell = dart.as(_LinkedHashMap$()._getTableEntry(nums, key), LinkedHashMapCell);
          return cell != null;
        } else {
          return this[_containsKey](key);
        }
      }
      [_containsKey](key) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, key);
        return dart.notNull(this[_findBucketIndex](bucket, key)) >= 0;
      }
      containsValue(value) {
        return this.keys[dartx.any](dart.fn((each => dart.equals(this.get(each), value)).bind(this), core.bool, [dart.dynamic]));
      }
      addAll(other) {
        dart.as(other, core.Map$(K, V));
        other.forEach(dart.fn(((key, value) => {
          dart.as(key, K);
          dart.as(value, V);
          this.set(key, value);
        }).bind(this), dart.dynamic, [K, V]));
      }
      get(key) {
        if (dart.notNull(_LinkedHashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          if (strings == null)
            return null;
          let cell = dart.as(_LinkedHashMap$()._getTableEntry(strings, key), LinkedHashMapCell);
          return cell == null ? null : dart.as(cell[_value], V);
        } else if (dart.notNull(_LinkedHashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          if (nums == null)
            return null;
          let cell = dart.as(_LinkedHashMap$()._getTableEntry(nums, key), LinkedHashMapCell);
          return cell == null ? null : dart.as(cell[_value], V);
        } else {
          return this[_get](key);
        }
      }
      [_get](key) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, key);
        let index = this[_findBucketIndex](bucket, key);
        if (dart.notNull(index) < 0)
          return null;
        let cell = dart.as(bucket[index], LinkedHashMapCell);
        return dart.as(cell[_value], V);
      }
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        if (dart.notNull(_LinkedHashMap$()._isStringKey(key))) {
          let strings = this[_strings];
          if (strings == null)
            this[_strings] = strings = _LinkedHashMap$()._newHashTable();
          this[_addHashTableEntry](strings, key, value);
        } else if (dart.notNull(_LinkedHashMap$()._isNumericKey(key))) {
          let nums = this[_nums];
          if (nums == null)
            this[_nums] = nums = _LinkedHashMap$()._newHashTable();
          this[_addHashTableEntry](nums, key, value);
        } else {
          this[_set](key, value);
        }
        return value;
      }
      [_set](key, value) {
        dart.as(key, K);
        dart.as(value, V);
        let rest = this[_rest];
        if (rest == null)
          this[_rest] = rest = _LinkedHashMap$()._newHashTable();
        let hash = this[_computeHashCode](key);
        let bucket = rest[hash];
        if (bucket == null) {
          let cell = this[_newLinkedCell](key, value);
          _LinkedHashMap$()._setTableEntry(rest, hash, [cell]);
        } else {
          let index = this[_findBucketIndex](bucket, key);
          if (dart.notNull(index) >= 0) {
            let cell = dart.as(bucket[index], LinkedHashMapCell);
            cell[_value] = value;
          } else {
            let cell = this[_newLinkedCell](key, value);
            bucket.push(cell);
          }
        }
      }
      putIfAbsent(key, ifAbsent) {
        dart.as(key, K);
        dart.as(ifAbsent, dart.functionType(V, []));
        if (dart.notNull(this.containsKey(key)))
          return this.get(key);
        let value = ifAbsent();
        this.set(key, value);
        return value;
      }
      remove(key) {
        if (dart.notNull(_LinkedHashMap$()._isStringKey(key))) {
          return this[_removeHashTableEntry](this[_strings], key);
        } else if (dart.notNull(_LinkedHashMap$()._isNumericKey(key))) {
          return this[_removeHashTableEntry](this[_nums], key);
        } else {
          return this[_remove](key);
        }
      }
      [_remove](key) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, key);
        let index = this[_findBucketIndex](bucket, key);
        if (dart.notNull(index) < 0)
          return null;
        let cell = dart.as(bucket.splice(index, 1)[0], LinkedHashMapCell);
        this[_unlinkCell](cell);
        return dart.as(cell[_value], V);
      }
      clear() {
        if (dart.notNull(this[_length]) > 0) {
          this[_strings] = this[_nums] = this[_rest] = this[_first] = this[_last] = null;
          this[_length] = 0;
          this[_modified]();
        }
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [K, V]));
        let cell = this[_first];
        let modifications = this[_modifications];
        while (cell != null) {
          action(dart.as(cell[_key], K), dart.as(cell[_value], V));
          if (modifications != this[_modifications]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          cell = cell[_next];
        }
      }
      [_addHashTableEntry](table, key, value) {
        dart.as(key, K);
        dart.as(value, V);
        let cell = dart.as(_LinkedHashMap$()._getTableEntry(table, key), LinkedHashMapCell);
        if (cell == null) {
          _LinkedHashMap$()._setTableEntry(table, key, this[_newLinkedCell](key, value));
        } else {
          cell[_value] = value;
        }
      }
      [_removeHashTableEntry](table, key) {
        if (table == null)
          return null;
        let cell = dart.as(_LinkedHashMap$()._getTableEntry(table, key), LinkedHashMapCell);
        if (cell == null)
          return null;
        this[_unlinkCell](cell);
        _LinkedHashMap$()._deleteTableEntry(table, key);
        return dart.as(cell[_value], V);
      }
      [_modified]() {
        this[_modifications] = dart.notNull(this[_modifications]) + 1 & 67108863;
      }
      [_newLinkedCell](key, value) {
        dart.as(key, K);
        dart.as(value, V);
        let cell = new LinkedHashMapCell(key, value);
        if (this[_first] == null) {
          this[_first] = this[_last] = cell;
        } else {
          let last = this[_last];
          cell[_previous] = last;
          this[_last] = last[_next] = cell;
        }
        this[_length] = dart.notNull(this[_length]) + 1;
        this[_modified]();
        return cell;
      }
      [_unlinkCell](cell) {
        let previous = cell[_previous];
        let next = cell[_next];
        if (previous == null) {
          dart.assert(dart.equals(cell, this[_first]));
          this[_first] = next;
        } else {
          previous[_next] = next;
        }
        if (next == null) {
          dart.assert(dart.equals(cell, this[_last]));
          this[_last] = previous;
        } else {
          next[_previous] = previous;
        }
        this[_length] = dart.notNull(this[_length]) - 1;
        this[_modified]();
      }
      static _isStringKey(key) {
        return typeof key == 'string' && !dart.equals(key, '__proto__');
      }
      static _isNumericKey(key) {
        return typeof key == 'number' && (key & 0x3ffffff) === key;
      }
      [_computeHashCode](key) {
        return dart.hashCode(key) & 0x3ffffff;
      }
      static _getTableEntry(table, key) {
        return table[key];
      }
      static _setTableEntry(table, key, value) {
        dart.assert(value != null);
        table[key] = value;
      }
      static _deleteTableEntry(table, key) {
        delete table[key];
      }
      [_getBucket](table, key) {
        let hash = this[_computeHashCode](key);
        return dart.as(table[hash], core.List);
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashMapCell);
          if (dart.equals(cell[_key], key))
            return i;
        }
        return -1;
      }
      static _newHashTable() {
        let table = Object.create(null);
        let temporaryKey = '<non-identifier-key>';
        _LinkedHashMap$()._setTableEntry(table, temporaryKey, table);
        _LinkedHashMap$()._deleteTableEntry(table, temporaryKey);
        return table;
      }
      toString() {
        return Maps.mapToString(this);
      }
    }
    _LinkedHashMap[dart.implements] = () => [LinkedHashMap$(K, V), _js_helper.InternalMap];
    dart.setSignature(_LinkedHashMap, {
      constructors: () => ({_LinkedHashMap: [_LinkedHashMap$(K, V), []]}),
      methods: () => ({
        containsKey: [core.bool, [core.Object]],
        [_containsKey]: [core.bool, [core.Object]],
        containsValue: [core.bool, [core.Object]],
        addAll: [dart.void, [core.Map$(K, V)]],
        get: [V, [core.Object]],
        [_get]: [V, [core.Object]],
        set: [dart.void, [K, V]],
        [_set]: [dart.void, [K, V]],
        putIfAbsent: [V, [K, dart.functionType(V, [])]],
        remove: [V, [core.Object]],
        [_remove]: [V, [core.Object]],
        clear: [dart.void, []],
        forEach: [dart.void, [dart.functionType(dart.void, [K, V])]],
        [_addHashTableEntry]: [dart.void, [dart.dynamic, K, V]],
        [_removeHashTableEntry]: [V, [dart.dynamic, core.Object]],
        [_modified]: [dart.void, []],
        [_newLinkedCell]: [LinkedHashMapCell, [K, V]],
        [_unlinkCell]: [dart.void, [LinkedHashMapCell]],
        [_computeHashCode]: [core.int, [dart.dynamic]],
        [_getBucket]: [core.List, [dart.dynamic, dart.dynamic]],
        [_findBucketIndex]: [core.int, [dart.dynamic, dart.dynamic]]
      }),
      statics: () => ({
        _isStringKey: [core.bool, [dart.dynamic]],
        _isNumericKey: [core.bool, [dart.dynamic]],
        _getTableEntry: [dart.dynamic, [dart.dynamic, dart.dynamic]],
        _setTableEntry: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
        _deleteTableEntry: [dart.void, [dart.dynamic, dart.dynamic]],
        _newHashTable: [dart.dynamic, []]
      }),
      names: ['_isStringKey', '_isNumericKey', '_getTableEntry', '_setTableEntry', '_deleteTableEntry', '_newHashTable']
    });
    return _LinkedHashMap;
  });
  let _LinkedHashMap = _LinkedHashMap$();
  const _LinkedIdentityHashMap$ = dart.generic(function(K, V) {
    class _LinkedIdentityHashMap extends _LinkedHashMap$(K, V) {
      _LinkedIdentityHashMap() {
        super._LinkedHashMap();
      }
      [_computeHashCode](key) {
        return core.identityHashCode(key) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashMapCell);
          if (dart.notNull(core.identical(cell[_key], key)))
            return i;
        }
        return -1;
      }
    }
    return _LinkedIdentityHashMap;
  });
  let _LinkedIdentityHashMap = _LinkedIdentityHashMap$();
  const _LinkedCustomHashMap$ = dart.generic(function(K, V) {
    class _LinkedCustomHashMap extends _LinkedHashMap$(K, V) {
      _LinkedCustomHashMap(equals, hashCode, validKey) {
        this[_equals] = equals;
        this[_hashCode] = hashCode;
        this[_validKey] = validKey != null ? validKey : dart.fn(v => dart.is(v, K), core.bool, [dart.dynamic]);
        super._LinkedHashMap();
      }
      get(key) {
        if (!dart.notNull(this[_validKey](key)))
          return null;
        return super[_get](key);
      }
      set(key, value) {
        dart.as(key, K);
        dart.as(value, V);
        super[_set](key, value);
        return value;
      }
      containsKey(key) {
        if (!dart.notNull(this[_validKey](key)))
          return false;
        return super[_containsKey](key);
      }
      remove(key) {
        if (!dart.notNull(this[_validKey](key)))
          return null;
        return super[_remove](key);
      }
      [_computeHashCode](key) {
        return this[_hashCode](dart.as(key, K)) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, key) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashMapCell);
          if (dart.notNull(this[_equals](dart.as(cell[_key], K), dart.as(key, K))))
            return i;
        }
        return -1;
      }
    }
    dart.setSignature(_LinkedCustomHashMap, {
      constructors: () => ({_LinkedCustomHashMap: [_LinkedCustomHashMap$(K, V), [_Equality$(K), _Hasher$(K), dart.functionType(core.bool, [core.Object])]]}),
      methods: () => ({
        get: [V, [core.Object]],
        set: [dart.void, [K, V]],
        remove: [V, [core.Object]]
      })
    });
    return _LinkedCustomHashMap;
  });
  let _LinkedCustomHashMap = _LinkedCustomHashMap$();
  class LinkedHashMapCell extends core.Object {
    LinkedHashMapCell(key, value) {
      this[_key] = key;
      this[_value] = value;
      this[_next] = null;
      this[_previous] = null;
    }
  }
  dart.setSignature(LinkedHashMapCell, {
    constructors: () => ({LinkedHashMapCell: [LinkedHashMapCell, [dart.dynamic, dart.dynamic]]})
  });
  const LinkedHashMapKeyIterable$ = dart.generic(function(E) {
    class LinkedHashMapKeyIterable extends IterableBase$(E) {
      LinkedHashMapKeyIterable(map) {
        this[_map] = map;
        super.IterableBase();
      }
      get length() {
        return dart.as(dart.dload(this[_map], _length), core.int);
      }
      get isEmpty() {
        return dart.equals(dart.dload(this[_map], _length), 0);
      }
      get iterator() {
        return new (LinkedHashMapKeyIterator$(E))(this[_map], dart.as(dart.dload(this[_map], _modifications), core.int));
      }
      contains(element) {
        return dart.as(dart.dsend(this[_map], 'containsKey', element), core.bool);
      }
      forEach(f) {
        dart.as(f, dart.functionType(dart.void, [E]));
        let cell = dart.as(dart.dload(this[_map], _first), LinkedHashMapCell);
        let modifications = dart.as(dart.dload(this[_map], _modifications), core.int);
        while (cell != null) {
          f(dart.as(cell[_key], E));
          if (!dart.equals(modifications, dart.dload(this[_map], _modifications))) {
            dart.throw(new core.ConcurrentModificationError(this[_map]));
          }
          cell = cell[_next];
        }
      }
    }
    LinkedHashMapKeyIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(LinkedHashMapKeyIterable, {
      constructors: () => ({LinkedHashMapKeyIterable: [LinkedHashMapKeyIterable$(E), [dart.dynamic]]}),
      methods: () => ({forEach: [dart.void, [dart.functionType(dart.void, [E])]]})
    });
    dart.defineExtensionMembers(LinkedHashMapKeyIterable, [
      'contains',
      'forEach',
      'length',
      'isEmpty',
      'iterator'
    ]);
    return LinkedHashMapKeyIterable;
  });
  let LinkedHashMapKeyIterable = LinkedHashMapKeyIterable$();
  const _cell = Symbol('_cell');
  const LinkedHashMapKeyIterator$ = dart.generic(function(E) {
    class LinkedHashMapKeyIterator extends core.Object {
      LinkedHashMapKeyIterator(map, modifications) {
        this[_map] = map;
        this[_modifications] = modifications;
        this[_cell] = null;
        this[_current] = null;
        this[_cell] = dart.as(dart.dload(this[_map], _first), LinkedHashMapCell);
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        if (!dart.equals(this[_modifications], dart.dload(this[_map], _modifications))) {
          dart.throw(new core.ConcurrentModificationError(this[_map]));
        } else if (this[_cell] == null) {
          this[_current] = null;
          return false;
        } else {
          this[_current] = dart.as(this[_cell][_key], E);
          this[_cell] = this[_cell][_next];
          return true;
        }
      }
    }
    LinkedHashMapKeyIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(LinkedHashMapKeyIterator, {
      constructors: () => ({LinkedHashMapKeyIterator: [LinkedHashMapKeyIterator$(E), [dart.dynamic, core.int]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return LinkedHashMapKeyIterator;
  });
  let LinkedHashMapKeyIterator = LinkedHashMapKeyIterator$();
  const _elements = Symbol('_elements');
  const _computeElements = Symbol('_computeElements');
  const _contains = Symbol('_contains');
  const _lookup = Symbol('_lookup');
  const _HashSet$ = dart.generic(function(E) {
    class _HashSet extends _HashSetBase$(E) {
      _HashSet() {
        this[_length] = 0;
        this[_strings] = null;
        this[_nums] = null;
        this[_rest] = null;
        this[_elements] = null;
      }
      [_newSet]() {
        return new (_HashSet$(E))();
      }
      get iterator() {
        return new (HashSetIterator$(E))(this, this[_computeElements]());
      }
      get length() {
        return this[_length];
      }
      get isEmpty() {
        return this[_length] == 0;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      contains(object) {
        if (dart.notNull(_HashSet$()._isStringElement(object))) {
          let strings = this[_strings];
          return strings == null ? false : _HashSet$()._hasTableEntry(strings, object);
        } else if (dart.notNull(_HashSet$()._isNumericElement(object))) {
          let nums = this[_nums];
          return nums == null ? false : _HashSet$()._hasTableEntry(nums, object);
        } else {
          return this[_contains](object);
        }
      }
      [_contains](object) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, object);
        return dart.notNull(this[_findBucketIndex](bucket, object)) >= 0;
      }
      lookup(object) {
        if (dart.notNull(_HashSet$()._isStringElement(object)) || dart.notNull(_HashSet$()._isNumericElement(object))) {
          return dart.as(dart.notNull(this.contains(object)) ? object : null, E);
        }
        return this[_lookup](object);
      }
      [_lookup](object) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, object);
        let index = this[_findBucketIndex](bucket, object);
        if (dart.notNull(index) < 0)
          return null;
        return dart.as(bucket[dartx.get](index), E);
      }
      add(element) {
        dart.as(element, E);
        if (dart.notNull(_HashSet$()._isStringElement(element))) {
          let strings = this[_strings];
          if (strings == null)
            this[_strings] = strings = _HashSet$()._newHashTable();
          return this[_addHashTableEntry](strings, element);
        } else if (dart.notNull(_HashSet$()._isNumericElement(element))) {
          let nums = this[_nums];
          if (nums == null)
            this[_nums] = nums = _HashSet$()._newHashTable();
          return this[_addHashTableEntry](nums, element);
        } else {
          return this[_add](element);
        }
      }
      [_add](element) {
        dart.as(element, E);
        let rest = this[_rest];
        if (rest == null)
          this[_rest] = rest = _HashSet$()._newHashTable();
        let hash = this[_computeHashCode](element);
        let bucket = rest[hash];
        if (bucket == null) {
          _HashSet$()._setTableEntry(rest, hash, [element]);
        } else {
          let index = this[_findBucketIndex](bucket, element);
          if (dart.notNull(index) >= 0)
            return false;
          bucket.push(element);
        }
        this[_length] = dart.notNull(this[_length]) + 1;
        this[_elements] = null;
        return true;
      }
      addAll(objects) {
        dart.as(objects, core.Iterable$(E));
        for (let each of objects) {
          this.add(each);
        }
      }
      remove(object) {
        if (dart.notNull(_HashSet$()._isStringElement(object))) {
          return this[_removeHashTableEntry](this[_strings], object);
        } else if (dart.notNull(_HashSet$()._isNumericElement(object))) {
          return this[_removeHashTableEntry](this[_nums], object);
        } else {
          return this[_remove](object);
        }
      }
      [_remove](object) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, object);
        let index = this[_findBucketIndex](bucket, object);
        if (dart.notNull(index) < 0)
          return false;
        this[_length] = dart.notNull(this[_length]) - 1;
        this[_elements] = null;
        bucket.splice(index, 1);
        return true;
      }
      clear() {
        if (dart.notNull(this[_length]) > 0) {
          this[_strings] = this[_nums] = this[_rest] = this[_elements] = null;
          this[_length] = 0;
        }
      }
      [_computeElements]() {
        if (this[_elements] != null)
          return this[_elements];
        let result = core.List.new(this[_length]);
        let index = 0;
        let strings = this[_strings];
        if (strings != null) {
          let names = Object.getOwnPropertyNames(strings);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let element = names[i];
            result[index] = element;
            index = dart.notNull(index) + 1;
          }
        }
        let nums = this[_nums];
        if (nums != null) {
          let names = Object.getOwnPropertyNames(nums);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let element = +names[i];
            result[index] = element;
            index = dart.notNull(index) + 1;
          }
        }
        let rest = this[_rest];
        if (rest != null) {
          let names = Object.getOwnPropertyNames(rest);
          let entries = names.length;
          for (let i = 0; dart.notNull(i) < dart.notNull(entries); i = dart.notNull(i) + 1) {
            let entry = names[i];
            let bucket = rest[entry];
            let length = bucket.length;
            for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
              result[index] = bucket[i];
              index = dart.notNull(index) + 1;
            }
          }
        }
        dart.assert(index == this[_length]);
        return this[_elements] = result;
      }
      [_addHashTableEntry](table, element) {
        dart.as(element, E);
        if (dart.notNull(_HashSet$()._hasTableEntry(table, element)))
          return false;
        _HashSet$()._setTableEntry(table, element, 0);
        this[_length] = dart.notNull(this[_length]) + 1;
        this[_elements] = null;
        return true;
      }
      [_removeHashTableEntry](table, element) {
        if (table != null && dart.notNull(_HashSet$()._hasTableEntry(table, element))) {
          _HashSet$()._deleteTableEntry(table, element);
          this[_length] = dart.notNull(this[_length]) - 1;
          this[_elements] = null;
          return true;
        } else {
          return false;
        }
      }
      static _isStringElement(element) {
        return typeof element == 'string' && !dart.equals(element, '__proto__');
      }
      static _isNumericElement(element) {
        return typeof element == 'number' && (element & 0x3ffffff) === element;
      }
      [_computeHashCode](element) {
        return dart.hashCode(element) & 0x3ffffff;
      }
      static _hasTableEntry(table, key) {
        let entry = table[key];
        return entry != null;
      }
      static _setTableEntry(table, key, value) {
        dart.assert(value != null);
        table[key] = value;
      }
      static _deleteTableEntry(table, key) {
        delete table[key];
      }
      [_getBucket](table, element) {
        let hash = this[_computeHashCode](element);
        return dart.as(table[hash], core.List);
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.equals(bucket[i], element))
            return i;
        }
        return -1;
      }
      static _newHashTable() {
        let table = Object.create(null);
        let temporaryKey = '<non-identifier-key>';
        _HashSet$()._setTableEntry(table, temporaryKey, table);
        _HashSet$()._deleteTableEntry(table, temporaryKey);
        return table;
      }
    }
    _HashSet[dart.implements] = () => [HashSet$(E)];
    dart.setSignature(_HashSet, {
      constructors: () => ({_HashSet: [_HashSet$(E), []]}),
      methods: () => ({
        [_newSet]: [core.Set$(E), []],
        contains: [core.bool, [core.Object]],
        [_contains]: [core.bool, [core.Object]],
        lookup: [E, [core.Object]],
        [_lookup]: [E, [core.Object]],
        add: [core.bool, [E]],
        [_add]: [core.bool, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        remove: [core.bool, [core.Object]],
        [_remove]: [core.bool, [core.Object]],
        [_computeElements]: [core.List, []],
        [_addHashTableEntry]: [core.bool, [dart.dynamic, E]],
        [_removeHashTableEntry]: [core.bool, [dart.dynamic, core.Object]],
        [_computeHashCode]: [core.int, [dart.dynamic]],
        [_getBucket]: [core.List, [dart.dynamic, dart.dynamic]],
        [_findBucketIndex]: [core.int, [dart.dynamic, dart.dynamic]]
      }),
      statics: () => ({
        _isStringElement: [core.bool, [dart.dynamic]],
        _isNumericElement: [core.bool, [dart.dynamic]],
        _hasTableEntry: [core.bool, [dart.dynamic, dart.dynamic]],
        _setTableEntry: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
        _deleteTableEntry: [dart.void, [dart.dynamic, dart.dynamic]],
        _newHashTable: [dart.dynamic, []]
      }),
      names: ['_isStringElement', '_isNumericElement', '_hasTableEntry', '_setTableEntry', '_deleteTableEntry', '_newHashTable']
    });
    dart.defineExtensionMembers(_HashSet, [
      'contains',
      'iterator',
      'length',
      'isEmpty',
      'isNotEmpty'
    ]);
    return _HashSet;
  });
  let _HashSet = _HashSet$();
  const _IdentityHashSet$ = dart.generic(function(E) {
    class _IdentityHashSet extends _HashSet$(E) {
      _IdentityHashSet() {
        super._HashSet();
      }
      [_newSet]() {
        return new (_IdentityHashSet$(E))();
      }
      [_computeHashCode](key) {
        return core.identityHashCode(key) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.notNull(core.identical(bucket[i], element)))
            return i;
        }
        return -1;
      }
    }
    dart.setSignature(_IdentityHashSet, {
      methods: () => ({[_newSet]: [core.Set$(E), []]})
    });
    return _IdentityHashSet;
  });
  let _IdentityHashSet = _IdentityHashSet$();
  const _equality = Symbol('_equality');
  const _hasher = Symbol('_hasher');
  const _CustomHashSet$ = dart.generic(function(E) {
    class _CustomHashSet extends _HashSet$(E) {
      _CustomHashSet(equality, hasher, validKey) {
        this[_equality] = equality;
        this[_hasher] = hasher;
        this[_validKey] = validKey != null ? validKey : dart.fn(x => dart.is(x, E), core.bool, [dart.dynamic]);
        super._HashSet();
      }
      [_newSet]() {
        return new (_CustomHashSet$(E))(this[_equality], this[_hasher], this[_validKey]);
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          if (dart.notNull(this[_equality](dart.as(bucket[i], E), dart.as(element, E))))
            return i;
        }
        return -1;
      }
      [_computeHashCode](element) {
        return this[_hasher](dart.as(element, E)) & 0x3ffffff;
      }
      add(object) {
        dart.as(object, E);
        return super[_add](object);
      }
      contains(object) {
        if (!dart.notNull(this[_validKey](object)))
          return false;
        return super[_contains](object);
      }
      lookup(object) {
        if (!dart.notNull(this[_validKey](object)))
          return null;
        return super[_lookup](object);
      }
      remove(object) {
        if (!dart.notNull(this[_validKey](object)))
          return false;
        return super[_remove](object);
      }
    }
    dart.setSignature(_CustomHashSet, {
      constructors: () => ({_CustomHashSet: [_CustomHashSet$(E), [_Equality$(E), _Hasher$(E), dart.functionType(core.bool, [core.Object])]]}),
      methods: () => ({
        [_newSet]: [core.Set$(E), []],
        add: [core.bool, [E]],
        lookup: [E, [core.Object]]
      })
    });
    dart.defineExtensionMembers(_CustomHashSet, ['contains']);
    return _CustomHashSet;
  });
  let _CustomHashSet = _CustomHashSet$();
  const HashSetIterator$ = dart.generic(function(E) {
    class HashSetIterator extends core.Object {
      HashSetIterator(set, elements) {
        this[_set] = set;
        this[_elements] = elements;
        this[_offset] = 0;
        this[_current] = null;
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        let elements = this[_elements];
        let offset = this[_offset];
        if (elements !== dart.dload(this[_set], _elements)) {
          dart.throw(new core.ConcurrentModificationError(this[_set]));
        } else if (dart.notNull(offset) >= elements.length) {
          this[_current] = null;
          return false;
        } else {
          this[_current] = dart.as(elements[offset], E);
          this[_offset] = dart.notNull(offset) + 1;
          return true;
        }
      }
    }
    HashSetIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(HashSetIterator, {
      constructors: () => ({HashSetIterator: [HashSetIterator$(E), [dart.dynamic, core.List]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return HashSetIterator;
  });
  let HashSetIterator = HashSetIterator$();
  const _unsupported = Symbol('_unsupported');
  const _LinkedHashSet$ = dart.generic(function(E) {
    class _LinkedHashSet extends _HashSetBase$(E) {
      _LinkedHashSet() {
        this[_length] = 0;
        this[_strings] = null;
        this[_nums] = null;
        this[_rest] = null;
        this[_first] = null;
        this[_last] = null;
        this[_modifications] = 0;
      }
      [_newSet]() {
        return new (_LinkedHashSet$(E))();
      }
      [_unsupported](operation) {
        dart.throw(`LinkedHashSet: unsupported ${operation}`);
      }
      get iterator() {
        return new (LinkedHashSetIterator$(E))(this, this[_modifications]);
      }
      get length() {
        return this[_length];
      }
      get isEmpty() {
        return this[_length] == 0;
      }
      get isNotEmpty() {
        return !dart.notNull(this.isEmpty);
      }
      contains(object) {
        if (dart.notNull(_LinkedHashSet$()._isStringElement(object))) {
          let strings = this[_strings];
          if (strings == null)
            return false;
          let cell = dart.as(_LinkedHashSet$()._getTableEntry(strings, object), LinkedHashSetCell);
          return cell != null;
        } else if (dart.notNull(_LinkedHashSet$()._isNumericElement(object))) {
          let nums = this[_nums];
          if (nums == null)
            return false;
          let cell = dart.as(_LinkedHashSet$()._getTableEntry(nums, object), LinkedHashSetCell);
          return cell != null;
        } else {
          return this[_contains](object);
        }
      }
      [_contains](object) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, object);
        return dart.notNull(this[_findBucketIndex](bucket, object)) >= 0;
      }
      lookup(object) {
        if (dart.notNull(_LinkedHashSet$()._isStringElement(object)) || dart.notNull(_LinkedHashSet$()._isNumericElement(object))) {
          return dart.as(dart.notNull(this.contains(object)) ? object : null, E);
        } else {
          return this[_lookup](object);
        }
      }
      [_lookup](object) {
        let rest = this[_rest];
        if (rest == null)
          return null;
        let bucket = this[_getBucket](rest, object);
        let index = this[_findBucketIndex](bucket, object);
        if (dart.notNull(index) < 0)
          return null;
        return dart.as(dart.dload(bucket[dartx.get](index), _element), E);
      }
      forEach(action) {
        dart.as(action, dart.functionType(dart.void, [E]));
        let cell = this[_first];
        let modifications = this[_modifications];
        while (cell != null) {
          action(dart.as(cell[_element], E));
          if (modifications != this[_modifications]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          cell = cell[_next];
        }
      }
      get first() {
        if (this[_first] == null)
          dart.throw(new core.StateError("No elements"));
        return dart.as(this[_first][_element], E);
      }
      get last() {
        if (this[_last] == null)
          dart.throw(new core.StateError("No elements"));
        return dart.as(this[_last][_element], E);
      }
      add(element) {
        dart.as(element, E);
        if (dart.notNull(_LinkedHashSet$()._isStringElement(element))) {
          let strings = this[_strings];
          if (strings == null)
            this[_strings] = strings = _LinkedHashSet$()._newHashTable();
          return this[_addHashTableEntry](strings, element);
        } else if (dart.notNull(_LinkedHashSet$()._isNumericElement(element))) {
          let nums = this[_nums];
          if (nums == null)
            this[_nums] = nums = _LinkedHashSet$()._newHashTable();
          return this[_addHashTableEntry](nums, element);
        } else {
          return this[_add](element);
        }
      }
      [_add](element) {
        dart.as(element, E);
        let rest = this[_rest];
        if (rest == null)
          this[_rest] = rest = _LinkedHashSet$()._newHashTable();
        let hash = this[_computeHashCode](element);
        let bucket = rest[hash];
        if (bucket == null) {
          let cell = this[_newLinkedCell](element);
          _LinkedHashSet$()._setTableEntry(rest, hash, [cell]);
        } else {
          let index = this[_findBucketIndex](bucket, element);
          if (dart.notNull(index) >= 0)
            return false;
          let cell = this[_newLinkedCell](element);
          bucket.push(cell);
        }
        return true;
      }
      remove(object) {
        if (dart.notNull(_LinkedHashSet$()._isStringElement(object))) {
          return this[_removeHashTableEntry](this[_strings], object);
        } else if (dart.notNull(_LinkedHashSet$()._isNumericElement(object))) {
          return this[_removeHashTableEntry](this[_nums], object);
        } else {
          return this[_remove](object);
        }
      }
      [_remove](object) {
        let rest = this[_rest];
        if (rest == null)
          return false;
        let bucket = this[_getBucket](rest, object);
        let index = this[_findBucketIndex](bucket, object);
        if (dart.notNull(index) < 0)
          return false;
        let cell = dart.as(bucket.splice(index, 1)[0], LinkedHashSetCell);
        this[_unlinkCell](cell);
        return true;
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filterWhere](test, true);
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        this[_filterWhere](test, false);
      }
      [_filterWhere](test, removeMatching) {
        dart.as(test, dart.functionType(core.bool, [E]));
        let cell = this[_first];
        while (cell != null) {
          let element = dart.as(cell[_element], E);
          let next = cell[_next];
          let modifications = this[_modifications];
          let shouldRemove = removeMatching == test(element);
          if (modifications != this[_modifications]) {
            dart.throw(new core.ConcurrentModificationError(this));
          }
          if (dart.notNull(shouldRemove))
            this.remove(element);
          cell = next;
        }
      }
      clear() {
        if (dart.notNull(this[_length]) > 0) {
          this[_strings] = this[_nums] = this[_rest] = this[_first] = this[_last] = null;
          this[_length] = 0;
          this[_modified]();
        }
      }
      [_addHashTableEntry](table, element) {
        dart.as(element, E);
        let cell = dart.as(_LinkedHashSet$()._getTableEntry(table, element), LinkedHashSetCell);
        if (cell != null)
          return false;
        _LinkedHashSet$()._setTableEntry(table, element, this[_newLinkedCell](element));
        return true;
      }
      [_removeHashTableEntry](table, element) {
        if (table == null)
          return false;
        let cell = dart.as(_LinkedHashSet$()._getTableEntry(table, element), LinkedHashSetCell);
        if (cell == null)
          return false;
        this[_unlinkCell](cell);
        _LinkedHashSet$()._deleteTableEntry(table, element);
        return true;
      }
      [_modified]() {
        this[_modifications] = dart.notNull(this[_modifications]) + 1 & 67108863;
      }
      [_newLinkedCell](element) {
        dart.as(element, E);
        let cell = new LinkedHashSetCell(element);
        if (this[_first] == null) {
          this[_first] = this[_last] = cell;
        } else {
          let last = this[_last];
          cell[_previous] = last;
          this[_last] = last[_next] = cell;
        }
        this[_length] = dart.notNull(this[_length]) + 1;
        this[_modified]();
        return cell;
      }
      [_unlinkCell](cell) {
        let previous = cell[_previous];
        let next = cell[_next];
        if (previous == null) {
          dart.assert(dart.equals(cell, this[_first]));
          this[_first] = next;
        } else {
          previous[_next] = next;
        }
        if (next == null) {
          dart.assert(dart.equals(cell, this[_last]));
          this[_last] = previous;
        } else {
          next[_previous] = previous;
        }
        this[_length] = dart.notNull(this[_length]) - 1;
        this[_modified]();
      }
      static _isStringElement(element) {
        return typeof element == 'string' && !dart.equals(element, '__proto__');
      }
      static _isNumericElement(element) {
        return typeof element == 'number' && (element & 0x3ffffff) === element;
      }
      [_computeHashCode](element) {
        return dart.hashCode(element) & 0x3ffffff;
      }
      static _getTableEntry(table, key) {
        return table[key];
      }
      static _setTableEntry(table, key, value) {
        dart.assert(value != null);
        table[key] = value;
      }
      static _deleteTableEntry(table, key) {
        delete table[key];
      }
      [_getBucket](table, element) {
        let hash = this[_computeHashCode](element);
        return dart.as(table[hash], core.List);
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashSetCell);
          if (dart.equals(cell[_element], element))
            return i;
        }
        return -1;
      }
      static _newHashTable() {
        let table = Object.create(null);
        let temporaryKey = '<non-identifier-key>';
        _LinkedHashSet$()._setTableEntry(table, temporaryKey, table);
        _LinkedHashSet$()._deleteTableEntry(table, temporaryKey);
        return table;
      }
    }
    _LinkedHashSet[dart.implements] = () => [LinkedHashSet$(E)];
    dart.setSignature(_LinkedHashSet, {
      constructors: () => ({_LinkedHashSet: [_LinkedHashSet$(E), []]}),
      methods: () => ({
        [_newSet]: [core.Set$(E), []],
        [_unsupported]: [dart.void, [core.String]],
        contains: [core.bool, [core.Object]],
        [_contains]: [core.bool, [core.Object]],
        lookup: [E, [core.Object]],
        [_lookup]: [E, [core.Object]],
        forEach: [dart.void, [dart.functionType(dart.void, [E])]],
        add: [core.bool, [E]],
        [_add]: [core.bool, [E]],
        remove: [core.bool, [core.Object]],
        [_remove]: [core.bool, [core.Object]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        [_filterWhere]: [dart.void, [dart.functionType(core.bool, [E]), core.bool]],
        [_addHashTableEntry]: [core.bool, [dart.dynamic, E]],
        [_removeHashTableEntry]: [core.bool, [dart.dynamic, core.Object]],
        [_modified]: [dart.void, []],
        [_newLinkedCell]: [LinkedHashSetCell, [E]],
        [_unlinkCell]: [dart.void, [LinkedHashSetCell]],
        [_computeHashCode]: [core.int, [dart.dynamic]],
        [_getBucket]: [core.List, [dart.dynamic, dart.dynamic]],
        [_findBucketIndex]: [core.int, [dart.dynamic, dart.dynamic]]
      }),
      statics: () => ({
        _isStringElement: [core.bool, [dart.dynamic]],
        _isNumericElement: [core.bool, [dart.dynamic]],
        _getTableEntry: [dart.dynamic, [dart.dynamic, dart.dynamic]],
        _setTableEntry: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
        _deleteTableEntry: [dart.void, [dart.dynamic, dart.dynamic]],
        _newHashTable: [dart.dynamic, []]
      }),
      names: ['_isStringElement', '_isNumericElement', '_getTableEntry', '_setTableEntry', '_deleteTableEntry', '_newHashTable']
    });
    dart.defineExtensionMembers(_LinkedHashSet, [
      'contains',
      'forEach',
      'iterator',
      'length',
      'isEmpty',
      'isNotEmpty',
      'first',
      'last'
    ]);
    return _LinkedHashSet;
  });
  let _LinkedHashSet = _LinkedHashSet$();
  const _LinkedIdentityHashSet$ = dart.generic(function(E) {
    class _LinkedIdentityHashSet extends _LinkedHashSet$(E) {
      _LinkedIdentityHashSet() {
        super._LinkedHashSet();
      }
      [_newSet]() {
        return new (_LinkedIdentityHashSet$(E))();
      }
      [_computeHashCode](key) {
        return core.identityHashCode(key) & 0x3ffffff;
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashSetCell);
          if (dart.notNull(core.identical(cell[_element], element)))
            return i;
        }
        return -1;
      }
    }
    dart.setSignature(_LinkedIdentityHashSet, {
      methods: () => ({[_newSet]: [core.Set$(E), []]})
    });
    return _LinkedIdentityHashSet;
  });
  let _LinkedIdentityHashSet = _LinkedIdentityHashSet$();
  const _LinkedCustomHashSet$ = dart.generic(function(E) {
    class _LinkedCustomHashSet extends _LinkedHashSet$(E) {
      _LinkedCustomHashSet(equality, hasher, validKey) {
        this[_equality] = equality;
        this[_hasher] = hasher;
        this[_validKey] = validKey != null ? validKey : dart.fn(x => dart.is(x, E), core.bool, [dart.dynamic]);
        super._LinkedHashSet();
      }
      [_newSet]() {
        return new (_LinkedCustomHashSet$(E))(this[_equality], this[_hasher], this[_validKey]);
      }
      [_findBucketIndex](bucket, element) {
        if (bucket == null)
          return -1;
        let length = bucket.length;
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          let cell = dart.as(bucket[i], LinkedHashSetCell);
          if (dart.notNull(this[_equality](dart.as(cell[_element], E), dart.as(element, E))))
            return i;
        }
        return -1;
      }
      [_computeHashCode](element) {
        return this[_hasher](dart.as(element, E)) & 0x3ffffff;
      }
      add(element) {
        dart.as(element, E);
        return super[_add](element);
      }
      contains(object) {
        if (!dart.notNull(this[_validKey](object)))
          return false;
        return super[_contains](object);
      }
      lookup(object) {
        if (!dart.notNull(this[_validKey](object)))
          return null;
        return super[_lookup](object);
      }
      remove(object) {
        if (!dart.notNull(this[_validKey](object)))
          return false;
        return super[_remove](object);
      }
      containsAll(elements) {
        for (let element of elements) {
          if (!dart.notNull(this[_validKey](element)) || !dart.notNull(this.contains(element)))
            return false;
        }
        return true;
      }
      removeAll(elements) {
        for (let element of elements) {
          if (dart.notNull(this[_validKey](element))) {
            super[_remove](element);
          }
        }
      }
    }
    dart.setSignature(_LinkedCustomHashSet, {
      constructors: () => ({_LinkedCustomHashSet: [_LinkedCustomHashSet$(E), [_Equality$(E), _Hasher$(E), dart.functionType(core.bool, [core.Object])]]}),
      methods: () => ({
        [_newSet]: [core.Set$(E), []],
        add: [core.bool, [E]],
        lookup: [E, [core.Object]]
      })
    });
    dart.defineExtensionMembers(_LinkedCustomHashSet, ['contains']);
    return _LinkedCustomHashSet;
  });
  let _LinkedCustomHashSet = _LinkedCustomHashSet$();
  class LinkedHashSetCell extends core.Object {
    LinkedHashSetCell(element) {
      this[_element] = element;
      this[_next] = null;
      this[_previous] = null;
    }
  }
  dart.setSignature(LinkedHashSetCell, {
    constructors: () => ({LinkedHashSetCell: [LinkedHashSetCell, [dart.dynamic]]})
  });
  const LinkedHashSetIterator$ = dart.generic(function(E) {
    class LinkedHashSetIterator extends core.Object {
      LinkedHashSetIterator(set, modifications) {
        this[_set] = set;
        this[_modifications] = modifications;
        this[_cell] = null;
        this[_current] = null;
        this[_cell] = dart.as(dart.dload(this[_set], _first), LinkedHashSetCell);
      }
      get current() {
        return this[_current];
      }
      moveNext() {
        if (!dart.equals(this[_modifications], dart.dload(this[_set], _modifications))) {
          dart.throw(new core.ConcurrentModificationError(this[_set]));
        } else if (this[_cell] == null) {
          this[_current] = null;
          return false;
        } else {
          this[_current] = dart.as(this[_cell][_element], E);
          this[_cell] = this[_cell][_next];
          return true;
        }
      }
    }
    LinkedHashSetIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(LinkedHashSetIterator, {
      constructors: () => ({LinkedHashSetIterator: [LinkedHashSetIterator$(E), [dart.dynamic, core.int]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return LinkedHashSetIterator;
  });
  let LinkedHashSetIterator = LinkedHashSetIterator$();
  // Exports:
  exports.UnmodifiableListView$ = UnmodifiableListView$;
  exports.HashMap$ = HashMap$;
  exports.HashMap = HashMap;
  exports.SetMixin$ = SetMixin$;
  exports.SetMixin = SetMixin;
  exports.SetBase$ = SetBase$;
  exports.SetBase = SetBase;
  exports.HashSet$ = HashSet$;
  exports.HashSet = HashSet;
  exports.IterableMixin$ = IterableMixin$;
  exports.IterableMixin = IterableMixin;
  exports.IterableBase$ = IterableBase$;
  exports.IterableBase = IterableBase;
  exports.HasNextIterator$ = HasNextIterator$;
  exports.HasNextIterator = HasNextIterator;
  exports.LinkedHashMap$ = LinkedHashMap$;
  exports.LinkedHashMap = LinkedHashMap;
  exports.LinkedHashSet$ = LinkedHashSet$;
  exports.LinkedHashSet = LinkedHashSet;
  exports.LinkedList$ = LinkedList$;
  exports.LinkedList = LinkedList;
  exports.LinkedListEntry$ = LinkedListEntry$;
  exports.LinkedListEntry = LinkedListEntry;
  exports.ListMixin$ = ListMixin$;
  exports.ListMixin = ListMixin;
  exports.ListBase$ = ListBase$;
  exports.ListBase = ListBase;
  exports.MapMixin$ = MapMixin$;
  exports.MapMixin = MapMixin;
  exports.MapBase$ = MapBase$;
  exports.MapBase = MapBase;
  exports.UnmodifiableMapBase$ = UnmodifiableMapBase$;
  exports.UnmodifiableMapBase = UnmodifiableMapBase;
  exports.MapView$ = MapView$;
  exports.MapView = MapView;
  exports.UnmodifiableMapView$ = UnmodifiableMapView$;
  exports.UnmodifiableMapView = UnmodifiableMapView;
  exports.Maps = Maps;
  exports.Queue$ = Queue$;
  exports.Queue = Queue;
  exports.DoubleLinkedQueueEntry$ = DoubleLinkedQueueEntry$;
  exports.DoubleLinkedQueueEntry = DoubleLinkedQueueEntry;
  exports.DoubleLinkedQueue$ = DoubleLinkedQueue$;
  exports.DoubleLinkedQueue = DoubleLinkedQueue;
  exports.ListQueue$ = ListQueue$;
  exports.ListQueue = ListQueue;
  exports.SplayTreeMap$ = SplayTreeMap$;
  exports.SplayTreeMap = SplayTreeMap;
  exports.SplayTreeSet$ = SplayTreeSet$;
  exports.SplayTreeSet = SplayTreeSet;
  exports.HashMapKeyIterable$ = HashMapKeyIterable$;
  exports.HashMapKeyIterable = HashMapKeyIterable;
  exports.HashMapKeyIterator$ = HashMapKeyIterator$;
  exports.HashMapKeyIterator = HashMapKeyIterator;
  exports.LinkedHashMapCell = LinkedHashMapCell;
  exports.LinkedHashMapKeyIterable$ = LinkedHashMapKeyIterable$;
  exports.LinkedHashMapKeyIterable = LinkedHashMapKeyIterable;
  exports.LinkedHashMapKeyIterator$ = LinkedHashMapKeyIterator$;
  exports.LinkedHashMapKeyIterator = LinkedHashMapKeyIterator;
  exports.HashSetIterator$ = HashSetIterator$;
  exports.HashSetIterator = HashSetIterator;
  exports.LinkedHashSetCell = LinkedHashSetCell;
  exports.LinkedHashSetIterator$ = LinkedHashSetIterator$;
  exports.LinkedHashSetIterator = LinkedHashSetIterator;
});
dart_library.library('dart/convert', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/async',
  'dart/typed_data',
  'dart/_internal',
  'dart/collection'
], /* Lazy imports */[
], function(exports, dart, core, async, typed_data, _internal, collection) {
  'use strict';
  let dartx = dart.dartx;
  const Codec$ = dart.generic(function(S, T) {
    class Codec extends core.Object {
      Codec() {
      }
      encode(input) {
        dart.as(input, S);
        return this.encoder.convert(input);
      }
      decode(encoded) {
        dart.as(encoded, T);
        return this.decoder.convert(encoded);
      }
      fuse(other) {
        dart.as(other, Codec$(T, dart.dynamic));
        return new (_FusedCodec$(S, T, dart.dynamic))(this, other);
      }
      get inverted() {
        return new (_InvertedCodec$(T, S))(this);
      }
    }
    dart.setSignature(Codec, {
      constructors: () => ({Codec: [Codec$(S, T), []]}),
      methods: () => ({
        encode: [T, [S]],
        decode: [S, [T]],
        fuse: [Codec$(S, dart.dynamic), [Codec$(T, dart.dynamic)]]
      })
    });
    return Codec;
  });
  let Codec = Codec$();
  class Encoding extends Codec$(core.String, core.List$(core.int)) {
    Encoding() {
      super.Codec();
    }
    decodeStream(byteStream) {
      return byteStream.transform(this.decoder).fold(new core.StringBuffer(), dart.fn((buffer, string) => ((() => {
        dart.dsend(buffer, 'write', string);
        return buffer;
      })()))).then(dart.fn(buffer => dart.toString(buffer), core.String, [dart.dynamic]));
    }
    static getByName(name) {
      if (name == null)
        return null;
      name = name[dartx.toLowerCase]();
      return Encoding._nameToEncoding.get(name);
    }
  }
  dart.setSignature(Encoding, {
    constructors: () => ({Encoding: [Encoding, []]}),
    methods: () => ({decodeStream: [async.Future$(core.String), [async.Stream$(core.List$(core.int))]]}),
    statics: () => ({getByName: [Encoding, [core.String]]}),
    names: ['getByName']
  });
  const _allowInvalid = Symbol('_allowInvalid');
  class AsciiCodec extends Encoding {
    AsciiCodec(opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : false;
      this[_allowInvalid] = allowInvalid;
      super.Encoding();
    }
    get name() {
      return "us-ascii";
    }
    decode(bytes, opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : null;
      if (allowInvalid == null)
        allowInvalid = this[_allowInvalid];
      if (dart.notNull(allowInvalid)) {
        return dart.const(new AsciiDecoder({allowInvalid: true})).convert(bytes);
      } else {
        return dart.const(new AsciiDecoder({allowInvalid: false})).convert(bytes);
      }
    }
    get encoder() {
      return dart.const(new AsciiEncoder());
    }
    get decoder() {
      return dart.notNull(this[_allowInvalid]) ? dart.const(new AsciiDecoder({allowInvalid: true})) : dart.const(new AsciiDecoder({allowInvalid: false}));
    }
  }
  dart.setSignature(AsciiCodec, {
    constructors: () => ({AsciiCodec: [AsciiCodec, [], {allowInvalid: core.bool}]}),
    methods: () => ({decode: [core.String, [core.List$(core.int)], {allowInvalid: core.bool}]})
  });
  const ASCII = dart.const(new AsciiCodec());
  const _ASCII_MASK = 127;
  const Converter$ = dart.generic(function(S, T) {
    class Converter extends core.Object {
      Converter() {
      }
      fuse(other) {
        dart.as(other, Converter$(T, dart.dynamic));
        return new (_FusedConverter$(S, T, dart.dynamic))(this, other);
      }
      startChunkedConversion(sink) {
        dart.as(sink, core.Sink$(T));
        dart.throw(new core.UnsupportedError(`This converter does not support chunked conversions: ${this}`));
      }
      bind(source) {
        dart.as(source, async.Stream$(S));
        return async.Stream$(T).eventTransformed(source, dart.fn((sink => new _ConverterStreamEventSink(this, sink)).bind(this), _ConverterStreamEventSink, [async.EventSink]));
      }
    }
    Converter[dart.implements] = () => [async.StreamTransformer$(S, T)];
    dart.setSignature(Converter, {
      constructors: () => ({Converter: [Converter$(S, T), []]}),
      methods: () => ({
        fuse: [Converter$(S, dart.dynamic), [Converter$(T, dart.dynamic)]],
        startChunkedConversion: [ChunkedConversionSink, [core.Sink$(T)]],
        bind: [async.Stream$(T), [async.Stream$(S)]]
      })
    });
    return Converter;
  });
  let Converter = Converter$();
  const _subsetMask = Symbol('_subsetMask');
  class _UnicodeSubsetEncoder extends Converter$(core.String, core.List$(core.int)) {
    _UnicodeSubsetEncoder(subsetMask) {
      this[_subsetMask] = subsetMask;
      super.Converter();
    }
    convert(string, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      let stringLength = string[dartx.length];
      core.RangeError.checkValidRange(start, end, stringLength);
      if (end == null)
        end = stringLength;
      let length = dart.notNull(end) - dart.notNull(start);
      let result = typed_data.Uint8List.new(length);
      for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        let codeUnit = string[dartx.codeUnitAt](dart.notNull(start) + dart.notNull(i));
        if ((dart.notNull(codeUnit) & ~dart.notNull(this[_subsetMask])) != 0) {
          dart.throw(new core.ArgumentError("String contains invalid characters."));
        }
        result[dartx.set](i, codeUnit);
      }
      return dart.as(result, core.List$(core.int));
    }
    startChunkedConversion(sink) {
      if (!dart.is(sink, ByteConversionSink)) {
        sink = ByteConversionSink.from(sink);
      }
      return new _UnicodeSubsetEncoderSink(this[_subsetMask], dart.as(sink, ByteConversionSink));
    }
    bind(stream) {
      return super.bind(stream);
    }
  }
  dart.setSignature(_UnicodeSubsetEncoder, {
    constructors: () => ({_UnicodeSubsetEncoder: [_UnicodeSubsetEncoder, [core.int]]}),
    methods: () => ({
      convert: [core.List$(core.int), [core.String], [core.int, core.int]],
      startChunkedConversion: [StringConversionSink, [core.Sink$(core.List$(core.int))]],
      bind: [async.Stream$(core.List$(core.int)), [async.Stream$(core.String)]]
    })
  });
  class AsciiEncoder extends _UnicodeSubsetEncoder {
    AsciiEncoder() {
      super._UnicodeSubsetEncoder(_ASCII_MASK);
    }
  }
  dart.setSignature(AsciiEncoder, {
    constructors: () => ({AsciiEncoder: [AsciiEncoder, []]})
  });
  class StringConversionSinkMixin extends core.Object {
    add(str) {
      return this.addSlice(str, 0, str[dartx.length], false);
    }
    asUtf8Sink(allowMalformed) {
      return new _Utf8ConversionSink(this, allowMalformed);
    }
    asStringSink() {
      return new _StringConversionSinkAsStringSinkAdapter(this);
    }
  }
  StringConversionSinkMixin[dart.implements] = () => [StringConversionSink];
  dart.setSignature(StringConversionSinkMixin, {
    methods: () => ({
      add: [dart.void, [core.String]],
      asUtf8Sink: [ByteConversionSink, [core.bool]],
      asStringSink: [ClosableStringSink, []]
    })
  });
  class StringConversionSinkBase extends StringConversionSinkMixin {}
  const _sink = Symbol('_sink');
  class _UnicodeSubsetEncoderSink extends StringConversionSinkBase {
    _UnicodeSubsetEncoderSink(subsetMask, sink) {
      this[_subsetMask] = subsetMask;
      this[_sink] = sink;
    }
    close() {
      this[_sink].close();
    }
    addSlice(source, start, end, isLast) {
      core.RangeError.checkValidRange(start, end, source[dartx.length]);
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let codeUnit = source[dartx.codeUnitAt](i);
        if ((dart.notNull(codeUnit) & ~dart.notNull(this[_subsetMask])) != 0) {
          dart.throw(new core.ArgumentError(`Source contains invalid character with code point: ${codeUnit}.`));
        }
      }
      this[_sink].add(source[dartx.codeUnits][dartx.sublist](start, end));
      if (dart.notNull(isLast)) {
        this.close();
      }
    }
  }
  dart.setSignature(_UnicodeSubsetEncoderSink, {
    constructors: () => ({_UnicodeSubsetEncoderSink: [_UnicodeSubsetEncoderSink, [core.int, ByteConversionSink]]}),
    methods: () => ({
      close: [dart.void, []],
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]]
    })
  });
  const _convertInvalid = Symbol('_convertInvalid');
  class _UnicodeSubsetDecoder extends Converter$(core.List$(core.int), core.String) {
    _UnicodeSubsetDecoder(allowInvalid, subsetMask) {
      this[_allowInvalid] = allowInvalid;
      this[_subsetMask] = subsetMask;
      super.Converter();
    }
    convert(bytes, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      let byteCount = bytes[dartx.length];
      core.RangeError.checkValidRange(start, end, byteCount);
      if (end == null)
        end = byteCount;
      let length = dart.notNull(end) - dart.notNull(start);
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let byte = bytes[dartx.get](i);
        if ((dart.notNull(byte) & ~dart.notNull(this[_subsetMask])) != 0) {
          if (!dart.notNull(this[_allowInvalid])) {
            dart.throw(new core.FormatException(`Invalid value in input: ${byte}`));
          }
          return this[_convertInvalid](bytes, start, end);
        }
      }
      return core.String.fromCharCodes(bytes, start, end);
    }
    [_convertInvalid](bytes, start, end) {
      let buffer = new core.StringBuffer();
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let value = bytes[dartx.get](i);
        if ((dart.notNull(value) & ~dart.notNull(this[_subsetMask])) != 0)
          value = 65533;
        buffer.writeCharCode(value);
      }
      return dart.toString(buffer);
    }
    bind(stream) {
      return super.bind(stream);
    }
  }
  dart.setSignature(_UnicodeSubsetDecoder, {
    constructors: () => ({_UnicodeSubsetDecoder: [_UnicodeSubsetDecoder, [core.bool, core.int]]}),
    methods: () => ({
      convert: [core.String, [core.List$(core.int)], [core.int, core.int]],
      [_convertInvalid]: [core.String, [core.List$(core.int), core.int, core.int]],
      bind: [async.Stream$(core.String), [async.Stream$(core.List$(core.int))]]
    })
  });
  class AsciiDecoder extends _UnicodeSubsetDecoder {
    AsciiDecoder(opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : false;
      super._UnicodeSubsetDecoder(allowInvalid, _ASCII_MASK);
    }
    startChunkedConversion(sink) {
      let stringSink = null;
      if (dart.is(sink, StringConversionSink)) {
        stringSink = sink;
      } else {
        stringSink = StringConversionSink.from(sink);
      }
      if (dart.notNull(this[_allowInvalid])) {
        return new _ErrorHandlingAsciiDecoderSink(stringSink.asUtf8Sink(false));
      } else {
        return new _SimpleAsciiDecoderSink(stringSink);
      }
    }
  }
  dart.setSignature(AsciiDecoder, {
    constructors: () => ({AsciiDecoder: [AsciiDecoder, [], {allowInvalid: core.bool}]}),
    methods: () => ({startChunkedConversion: [ByteConversionSink, [core.Sink$(core.String)]]})
  });
  const ChunkedConversionSink$ = dart.generic(function(T) {
    class ChunkedConversionSink extends core.Object {
      ChunkedConversionSink() {
      }
      static withCallback(callback) {
        return new _SimpleCallbackSink(callback);
      }
    }
    ChunkedConversionSink[dart.implements] = () => [core.Sink$(T)];
    dart.setSignature(ChunkedConversionSink, {
      constructors: () => ({
        ChunkedConversionSink: [ChunkedConversionSink$(T), []],
        withCallback: [ChunkedConversionSink$(T), [dart.functionType(dart.void, [core.List$(T)])]]
      })
    });
    return ChunkedConversionSink;
  });
  let ChunkedConversionSink = ChunkedConversionSink$();
  class ByteConversionSink extends ChunkedConversionSink$(core.List$(core.int)) {
    ByteConversionSink() {
      super.ChunkedConversionSink();
    }
    static withCallback(callback) {
      return new _ByteCallbackSink(callback);
    }
    static from(sink) {
      return new _ByteAdapterSink(sink);
    }
  }
  dart.setSignature(ByteConversionSink, {
    constructors: () => ({
      ByteConversionSink: [ByteConversionSink, []],
      withCallback: [ByteConversionSink, [dart.functionType(dart.void, [core.List$(core.int)])]],
      from: [ByteConversionSink, [core.Sink$(core.List$(core.int))]]
    })
  });
  class ByteConversionSinkBase extends ByteConversionSink {
    ByteConversionSinkBase() {
      super.ByteConversionSink();
    }
    addSlice(chunk, start, end, isLast) {
      this.add(chunk[dartx.sublist](start, end));
      if (dart.notNull(isLast))
        this.close();
    }
  }
  dart.setSignature(ByteConversionSinkBase, {
    methods: () => ({addSlice: [dart.void, [core.List$(core.int), core.int, core.int, core.bool]]})
  });
  const _utf8Sink = Symbol('_utf8Sink');
  class _ErrorHandlingAsciiDecoderSink extends ByteConversionSinkBase {
    _ErrorHandlingAsciiDecoderSink(utf8Sink) {
      this[_utf8Sink] = utf8Sink;
    }
    close() {
      this[_utf8Sink].close();
    }
    add(source) {
      this.addSlice(source, 0, source[dartx.length], false);
    }
    addSlice(source, start, end, isLast) {
      core.RangeError.checkValidRange(start, end, source[dartx.length]);
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        if ((dart.notNull(source[dartx.get](i)) & ~dart.notNull(_ASCII_MASK)) != 0) {
          if (dart.notNull(i) > dart.notNull(start))
            this[_utf8Sink].addSlice(source, start, i, false);
          this[_utf8Sink].add(dart.const(dart.list([239, 191, 189], core.int)));
          start = dart.notNull(i) + 1;
        }
      }
      if (dart.notNull(start) < dart.notNull(end)) {
        this[_utf8Sink].addSlice(source, start, end, isLast);
      } else if (dart.notNull(isLast)) {
        this.close();
      }
    }
  }
  dart.setSignature(_ErrorHandlingAsciiDecoderSink, {
    constructors: () => ({_ErrorHandlingAsciiDecoderSink: [_ErrorHandlingAsciiDecoderSink, [ByteConversionSink]]}),
    methods: () => ({
      close: [dart.void, []],
      add: [dart.void, [core.List$(core.int)]]
    })
  });
  class _SimpleAsciiDecoderSink extends ByteConversionSinkBase {
    _SimpleAsciiDecoderSink(sink) {
      this[_sink] = sink;
    }
    close() {
      this[_sink].close();
    }
    add(source) {
      for (let i = 0; dart.notNull(i) < dart.notNull(source[dartx.length]); i = dart.notNull(i) + 1) {
        if ((dart.notNull(source[dartx.get](i)) & ~dart.notNull(_ASCII_MASK)) != 0) {
          dart.throw(new core.FormatException("Source contains non-ASCII bytes."));
        }
      }
      this[_sink].add(core.String.fromCharCodes(source));
    }
    addSlice(source, start, end, isLast) {
      let length = source[dartx.length];
      core.RangeError.checkValidRange(start, end, length);
      if (dart.notNull(start) < dart.notNull(end)) {
        if (start != 0 || end != length) {
          source = source[dartx.sublist](start, end);
        }
        this.add(source);
      }
      if (dart.notNull(isLast))
        this.close();
    }
  }
  dart.setSignature(_SimpleAsciiDecoderSink, {
    constructors: () => ({_SimpleAsciiDecoderSink: [_SimpleAsciiDecoderSink, [core.Sink]]}),
    methods: () => ({
      close: [dart.void, []],
      add: [dart.void, [core.List$(core.int)]]
    })
  });
  class _ByteAdapterSink extends ByteConversionSinkBase {
    _ByteAdapterSink(sink) {
      this[_sink] = sink;
    }
    add(chunk) {
      return this[_sink].add(chunk);
    }
    close() {
      return this[_sink].close();
    }
  }
  dart.setSignature(_ByteAdapterSink, {
    constructors: () => ({_ByteAdapterSink: [_ByteAdapterSink, [core.Sink$(core.List$(core.int))]]}),
    methods: () => ({
      add: [dart.void, [core.List$(core.int)]],
      close: [dart.void, []]
    })
  });
  const _buffer = Symbol('_buffer');
  const _callback = Symbol('_callback');
  const _bufferIndex = Symbol('_bufferIndex');
  class _ByteCallbackSink extends ByteConversionSinkBase {
    _ByteCallbackSink(callback) {
      this[_buffer] = typed_data.Uint8List.new(_ByteCallbackSink._INITIAL_BUFFER_SIZE);
      this[_callback] = callback;
      this[_bufferIndex] = 0;
    }
    add(chunk) {
      let freeCount = dart.notNull(this[_buffer][dartx.length]) - dart.notNull(this[_bufferIndex]);
      if (dart.notNull(chunk[dartx.length]) > dart.notNull(freeCount)) {
        let oldLength = this[_buffer][dartx.length];
        let newLength = dart.notNull(_ByteCallbackSink._roundToPowerOf2(dart.notNull(chunk[dartx.length]) + dart.notNull(oldLength))) * 2;
        let grown = typed_data.Uint8List.new(newLength);
        grown[dartx.setRange](0, this[_buffer][dartx.length], this[_buffer]);
        this[_buffer] = grown;
      }
      this[_buffer][dartx.setRange](this[_bufferIndex], dart.notNull(this[_bufferIndex]) + dart.notNull(chunk[dartx.length]), chunk);
      this[_bufferIndex] = dart.notNull(this[_bufferIndex]) + dart.notNull(chunk[dartx.length]);
    }
    static _roundToPowerOf2(v) {
      dart.assert(dart.notNull(v) > 0);
      v = dart.notNull(v) - 1;
      v = dart.notNull(v) | dart.notNull(v) >> 1;
      v = dart.notNull(v) | dart.notNull(v) >> 2;
      v = dart.notNull(v) | dart.notNull(v) >> 4;
      v = dart.notNull(v) | dart.notNull(v) >> 8;
      v = dart.notNull(v) | dart.notNull(v) >> 16;
      v = dart.notNull(v) + 1;
      return v;
    }
    close() {
      this[_callback](this[_buffer][dartx.sublist](0, this[_bufferIndex]));
    }
  }
  dart.setSignature(_ByteCallbackSink, {
    constructors: () => ({_ByteCallbackSink: [_ByteCallbackSink, [dart.functionType(dart.void, [core.List$(core.int)])]]}),
    methods: () => ({
      add: [dart.void, [core.Iterable$(core.int)]],
      close: [dart.void, []]
    }),
    statics: () => ({_roundToPowerOf2: [core.int, [core.int]]}),
    names: ['_roundToPowerOf2']
  });
  _ByteCallbackSink._INITIAL_BUFFER_SIZE = 1024;
  const _ChunkedConversionCallback$ = dart.generic(function(T) {
    const _ChunkedConversionCallback = dart.typedef('_ChunkedConversionCallback', () => dart.functionType(dart.void, [T]));
    return _ChunkedConversionCallback;
  });
  let _ChunkedConversionCallback = _ChunkedConversionCallback$();
  const _accumulated = Symbol('_accumulated');
  const _SimpleCallbackSink$ = dart.generic(function(T) {
    class _SimpleCallbackSink extends ChunkedConversionSink$(T) {
      _SimpleCallbackSink(callback) {
        this[_accumulated] = dart.list([], T);
        this[_callback] = callback;
        super.ChunkedConversionSink();
      }
      add(chunk) {
        dart.as(chunk, T);
        this[_accumulated][dartx.add](chunk);
      }
      close() {
        this[_callback](this[_accumulated]);
      }
    }
    dart.setSignature(_SimpleCallbackSink, {
      constructors: () => ({_SimpleCallbackSink: [_SimpleCallbackSink$(T), [_ChunkedConversionCallback$(core.List$(T))]]}),
      methods: () => ({
        add: [dart.void, [T]],
        close: [dart.void, []]
      })
    });
    return _SimpleCallbackSink;
  });
  let _SimpleCallbackSink = _SimpleCallbackSink$();
  const _EventSinkAdapter$ = dart.generic(function(T) {
    class _EventSinkAdapter extends core.Object {
      _EventSinkAdapter(sink) {
        this[_sink] = sink;
      }
      add(data) {
        dart.as(data, T);
        return this[_sink].add(data);
      }
      close() {
        return this[_sink].close();
      }
    }
    _EventSinkAdapter[dart.implements] = () => [ChunkedConversionSink$(T)];
    dart.setSignature(_EventSinkAdapter, {
      constructors: () => ({_EventSinkAdapter: [_EventSinkAdapter$(T), [async.EventSink$(T)]]}),
      methods: () => ({
        add: [dart.void, [T]],
        close: [dart.void, []]
      })
    });
    return _EventSinkAdapter;
  });
  let _EventSinkAdapter = _EventSinkAdapter$();
  const _eventSink = Symbol('_eventSink');
  const _chunkedSink = Symbol('_chunkedSink');
  const _ConverterStreamEventSink$ = dart.generic(function(S, T) {
    class _ConverterStreamEventSink extends core.Object {
      _ConverterStreamEventSink(converter, sink) {
        this[_eventSink] = sink;
        this[_chunkedSink] = converter.startChunkedConversion(sink);
      }
      add(o) {
        dart.as(o, S);
        return this[_chunkedSink].add(o);
      }
      addError(error, stackTrace) {
        if (stackTrace === void 0)
          stackTrace = null;
        this[_eventSink].addError(error, stackTrace);
      }
      close() {
        return this[_chunkedSink].close();
      }
    }
    _ConverterStreamEventSink[dart.implements] = () => [async.EventSink$(S)];
    dart.setSignature(_ConverterStreamEventSink, {
      constructors: () => ({_ConverterStreamEventSink: [_ConverterStreamEventSink$(S, T), [Converter, async.EventSink$(T)]]}),
      methods: () => ({
        add: [dart.void, [S]],
        addError: [dart.void, [core.Object], [core.StackTrace]],
        close: [dart.void, []]
      })
    });
    return _ConverterStreamEventSink;
  });
  let _ConverterStreamEventSink = _ConverterStreamEventSink$();
  const _first = Symbol('_first');
  const _second = Symbol('_second');
  const _FusedCodec$ = dart.generic(function(S, M, T) {
    class _FusedCodec extends Codec$(S, T) {
      get encoder() {
        return dart.as(this[_first].encoder.fuse(this[_second].encoder), Converter$(S, T));
      }
      get decoder() {
        return dart.as(this[_second].decoder.fuse(this[_first].decoder), Converter$(T, S));
      }
      _FusedCodec(first, second) {
        this[_first] = first;
        this[_second] = second;
        super.Codec();
      }
    }
    dart.setSignature(_FusedCodec, {
      constructors: () => ({_FusedCodec: [_FusedCodec$(S, M, T), [Codec$(S, M), Codec$(M, T)]]})
    });
    return _FusedCodec;
  });
  let _FusedCodec = _FusedCodec$();
  const _codec = Symbol('_codec');
  const _InvertedCodec$ = dart.generic(function(T, S) {
    class _InvertedCodec extends Codec$(T, S) {
      _InvertedCodec(codec) {
        this[_codec] = codec;
        super.Codec();
      }
      get encoder() {
        return this[_codec].decoder;
      }
      get decoder() {
        return this[_codec].encoder;
      }
      get inverted() {
        return this[_codec];
      }
    }
    dart.setSignature(_InvertedCodec, {
      constructors: () => ({_InvertedCodec: [_InvertedCodec$(T, S), [Codec$(S, T)]]})
    });
    return _InvertedCodec;
  });
  let _InvertedCodec = _InvertedCodec$();
  const _FusedConverter$ = dart.generic(function(S, M, T) {
    class _FusedConverter extends Converter$(S, T) {
      _FusedConverter(first, second) {
        this[_first] = first;
        this[_second] = second;
        super.Converter();
      }
      convert(input) {
        dart.as(input, S);
        return dart.as(this[_second].convert(this[_first].convert(input)), T);
      }
      startChunkedConversion(sink) {
        dart.as(sink, core.Sink$(T));
        return this[_first].startChunkedConversion(this[_second].startChunkedConversion(sink));
      }
    }
    dart.setSignature(_FusedConverter, {
      constructors: () => ({_FusedConverter: [_FusedConverter$(S, M, T), [Converter, Converter]]}),
      methods: () => ({
        convert: [T, [S]],
        startChunkedConversion: [ChunkedConversionSink, [core.Sink$(T)]]
      })
    });
    return _FusedConverter;
  });
  let _FusedConverter = _FusedConverter$();
  dart.defineLazyProperties(Encoding, {
    get _nameToEncoding() {
      return dart.map({"iso_8859-1:1987": LATIN1, "iso-ir-100": LATIN1, "iso_8859-1": LATIN1, "iso-8859-1": LATIN1, latin1: LATIN1, l1: LATIN1, ibm819: LATIN1, cp819: LATIN1, csisolatin1: LATIN1, "iso-ir-6": ASCII, "ansi_x3.4-1968": ASCII, "ansi_x3.4-1986": ASCII, "iso_646.irv:1991": ASCII, "iso646-us": ASCII, "us-ascii": ASCII, us: ASCII, ibm367: ASCII, cp367: ASCII, csascii: ASCII, ascii: ASCII, csutf8: UTF8, "utf-8": UTF8});
    },
    set _nameToEncoding(_) {}
  });
  const _name = Symbol('_name');
  class HtmlEscapeMode extends core.Object {
    _(name, escapeLtGt, escapeQuot, escapeApos, escapeSlash) {
      this[_name] = name;
      this.escapeLtGt = escapeLtGt;
      this.escapeQuot = escapeQuot;
      this.escapeApos = escapeApos;
      this.escapeSlash = escapeSlash;
    }
    toString() {
      return this[_name];
    }
  }
  dart.defineNamedConstructor(HtmlEscapeMode, '_');
  dart.setSignature(HtmlEscapeMode, {
    constructors: () => ({_: [HtmlEscapeMode, [core.String, core.bool, core.bool, core.bool, core.bool]]})
  });
  HtmlEscapeMode.UNKNOWN = dart.const(new HtmlEscapeMode._('unknown', true, true, true, true));
  const _convert = Symbol('_convert');
  class HtmlEscape extends Converter$(core.String, core.String) {
    HtmlEscape(mode) {
      if (mode === void 0)
        mode = HtmlEscapeMode.UNKNOWN;
      this.mode = mode;
      super.Converter();
    }
    convert(text) {
      let val = this[_convert](text, 0, text[dartx.length]);
      return val == null ? text : val;
    }
    [_convert](text, start, end) {
      let result = null;
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let ch = text[dartx.get](i);
        let replace = null;
        switch (ch) {
          case '&':
          {
            replace = '&amp;';
            break;
          }
          case ' ':
          {
            replace = '&nbsp;';
            break;
          }
          case '"':
          {
            if (dart.notNull(this.mode.escapeQuot))
              replace = '&quot;';
            break;
          }
          case "'":
          {
            if (dart.notNull(this.mode.escapeApos))
              replace = '&#x27;';
            break;
          }
          case '<':
          {
            if (dart.notNull(this.mode.escapeLtGt))
              replace = '&lt;';
            break;
          }
          case '>':
          {
            if (dart.notNull(this.mode.escapeLtGt))
              replace = '&gt;';
            break;
          }
          case '/':
          {
            if (dart.notNull(this.mode.escapeSlash))
              replace = '&#x2F;';
            break;
          }
        }
        if (replace != null) {
          if (result == null)
            result = new core.StringBuffer(text[dartx.substring](start, i));
          result.write(replace);
        } else if (result != null) {
          result.write(ch);
        }
      }
      return result != null ? dart.toString(result) : null;
    }
    startChunkedConversion(sink) {
      if (!dart.is(sink, StringConversionSink)) {
        sink = StringConversionSink.from(sink);
      }
      return new _HtmlEscapeSink(this, dart.as(sink, StringConversionSink));
    }
  }
  dart.setSignature(HtmlEscape, {
    constructors: () => ({HtmlEscape: [HtmlEscape, [], [HtmlEscapeMode]]}),
    methods: () => ({
      convert: [core.String, [core.String]],
      [_convert]: [core.String, [core.String, core.int, core.int]],
      startChunkedConversion: [StringConversionSink, [core.Sink$(core.String)]]
    })
  });
  const HTML_ESCAPE = dart.const(new HtmlEscape());
  HtmlEscapeMode.ATTRIBUTE = dart.const(new HtmlEscapeMode._('attribute', false, true, false, false));
  HtmlEscapeMode.ELEMENT = dart.const(new HtmlEscapeMode._('element', true, false, false, true));
  const _escape = Symbol('_escape');
  class _HtmlEscapeSink extends StringConversionSinkBase {
    _HtmlEscapeSink(escape, sink) {
      this[_escape] = escape;
      this[_sink] = sink;
    }
    addSlice(chunk, start, end, isLast) {
      let val = this[_escape][_convert](chunk, start, end);
      if (val == null) {
        this[_sink].addSlice(chunk, start, end, isLast);
      } else {
        this[_sink].add(val);
        if (dart.notNull(isLast))
          this[_sink].close();
      }
    }
    close() {
      return this[_sink].close();
    }
  }
  dart.setSignature(_HtmlEscapeSink, {
    constructors: () => ({_HtmlEscapeSink: [_HtmlEscapeSink, [HtmlEscape, StringConversionSink]]}),
    methods: () => ({
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]],
      close: [dart.void, []]
    })
  });
  class JsonUnsupportedObjectError extends core.Error {
    JsonUnsupportedObjectError(unsupportedObject, opts) {
      let cause = opts && 'cause' in opts ? opts.cause : null;
      this.unsupportedObject = unsupportedObject;
      this.cause = cause;
      super.Error();
    }
    toString() {
      if (this.cause != null) {
        return "Converting object to an encodable object failed.";
      } else {
        return "Converting object did not return an encodable object.";
      }
    }
  }
  dart.setSignature(JsonUnsupportedObjectError, {
    constructors: () => ({JsonUnsupportedObjectError: [JsonUnsupportedObjectError, [dart.dynamic], {cause: dart.dynamic}]})
  });
  class JsonCyclicError extends JsonUnsupportedObjectError {
    JsonCyclicError(object) {
      super.JsonUnsupportedObjectError(object);
    }
    toString() {
      return "Cyclic error in JSON stringify";
    }
  }
  dart.setSignature(JsonCyclicError, {
    constructors: () => ({JsonCyclicError: [JsonCyclicError, [core.Object]]})
  });
  const _reviver = Symbol('_reviver');
  const _toEncodable$ = Symbol('_toEncodable');
  class JsonCodec extends Codec$(core.Object, core.String) {
    JsonCodec(opts) {
      let reviver = opts && 'reviver' in opts ? opts.reviver : null;
      let toEncodable = opts && 'toEncodable' in opts ? opts.toEncodable : null;
      this[_reviver] = reviver;
      this[_toEncodable$] = toEncodable;
      super.Codec();
    }
    withReviver(reviver) {
      this.JsonCodec({reviver: reviver});
    }
    decode(source, opts) {
      let reviver = opts && 'reviver' in opts ? opts.reviver : null;
      if (reviver == null)
        reviver = this[_reviver];
      if (reviver == null)
        return this.decoder.convert(source);
      return new JsonDecoder(reviver).convert(source);
    }
    encode(value, opts) {
      let toEncodable = opts && 'toEncodable' in opts ? opts.toEncodable : null;
      if (toEncodable == null)
        toEncodable = this[_toEncodable$];
      if (toEncodable == null)
        return this.encoder.convert(value);
      return new JsonEncoder(dart.as(toEncodable, __CastType0)).convert(value);
    }
    get encoder() {
      if (this[_toEncodable$] == null)
        return dart.const(new JsonEncoder());
      return new JsonEncoder(dart.as(this[_toEncodable$], dart.functionType(core.Object, [core.Object])));
    }
    get decoder() {
      if (this[_reviver] == null)
        return dart.const(new JsonDecoder());
      return new JsonDecoder(this[_reviver]);
    }
  }
  dart.defineNamedConstructor(JsonCodec, 'withReviver');
  dart.setSignature(JsonCodec, {
    constructors: () => ({
      JsonCodec: [JsonCodec, [], {reviver: dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]), toEncodable: dart.functionType(dart.dynamic, [dart.dynamic])}],
      withReviver: [JsonCodec, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]]
    }),
    methods: () => ({
      decode: [dart.dynamic, [core.String], {reviver: dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])}],
      encode: [core.String, [core.Object], {toEncodable: dart.functionType(dart.dynamic, [dart.dynamic])}]
    })
  });
  const JSON = dart.const(new JsonCodec());
  const _Reviver = dart.typedef('_Reviver', () => dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic]));
  const _ToEncodable = dart.typedef('_ToEncodable', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  class JsonEncoder extends Converter$(core.Object, core.String) {
    JsonEncoder(toEncodable) {
      if (toEncodable === void 0)
        toEncodable = null;
      this.indent = null;
      this[_toEncodable$] = toEncodable;
      super.Converter();
    }
    withIndent(indent, toEncodable) {
      if (toEncodable === void 0)
        toEncodable = null;
      this.indent = indent;
      this[_toEncodable$] = toEncodable;
      super.Converter();
    }
    convert(object) {
      return _JsonStringStringifier.stringify(object, dart.as(this[_toEncodable$], __CastType2), this.indent);
    }
    startChunkedConversion(sink) {
      if (!dart.is(sink, StringConversionSink)) {
        sink = StringConversionSink.from(sink);
      } else if (dart.is(sink, _Utf8EncoderSink)) {
        return new _JsonUtf8EncoderSink(sink[_sink], this[_toEncodable$], JsonUtf8Encoder._utf8Encode(this.indent), JsonUtf8Encoder.DEFAULT_BUFFER_SIZE);
      }
      return new _JsonEncoderSink(dart.as(sink, StringConversionSink), this[_toEncodable$], this.indent);
    }
    bind(stream) {
      return super.bind(stream);
    }
    fuse(other) {
      if (dart.is(other, Utf8Encoder)) {
        return new JsonUtf8Encoder(this.indent, dart.as(this[_toEncodable$], __CastType4));
      }
      return super.fuse(other);
    }
  }
  dart.defineNamedConstructor(JsonEncoder, 'withIndent');
  dart.setSignature(JsonEncoder, {
    constructors: () => ({
      JsonEncoder: [JsonEncoder, [], [dart.functionType(core.Object, [core.Object])]],
      withIndent: [JsonEncoder, [core.String], [dart.functionType(core.Object, [core.Object])]]
    }),
    methods: () => ({
      convert: [core.String, [core.Object]],
      startChunkedConversion: [ChunkedConversionSink$(core.Object), [core.Sink$(core.String)]],
      bind: [async.Stream$(core.String), [async.Stream$(core.Object)]],
      fuse: [Converter$(core.Object, dart.dynamic), [Converter$(core.String, dart.dynamic)]]
    })
  });
  const _indent = Symbol('_indent');
  const _bufferSize = Symbol('_bufferSize');
  class JsonUtf8Encoder extends Converter$(core.Object, core.List$(core.int)) {
    JsonUtf8Encoder(indent, toEncodable, bufferSize) {
      if (indent === void 0)
        indent = null;
      if (toEncodable === void 0)
        toEncodable = null;
      if (bufferSize === void 0)
        bufferSize = JsonUtf8Encoder.DEFAULT_BUFFER_SIZE;
      this[_indent] = JsonUtf8Encoder._utf8Encode(indent);
      this[_toEncodable$] = toEncodable;
      this[_bufferSize] = bufferSize;
      super.Converter();
    }
    static _utf8Encode(string) {
      if (string == null)
        return null;
      if (dart.notNull(string[dartx.isEmpty]))
        return typed_data.Uint8List.new(0);
      checkAscii: {
        for (let i = 0; dart.notNull(i) < dart.notNull(string[dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.notNull(string[dartx.codeUnitAt](i)) >= 128)
            break checkAscii;
        }
        return string[dartx.codeUnits];
      }
      return UTF8.encode(string);
    }
    convert(object) {
      let bytes = dart.list([], core.List$(core.int));
      function addChunk(chunk, start, end) {
        if (dart.notNull(start) > 0 || dart.notNull(end) < dart.notNull(chunk.length)) {
          let length = dart.notNull(end) - dart.notNull(start);
          chunk = typed_data.Uint8List.view(chunk.buffer, dart.notNull(chunk.offsetInBytes) + dart.notNull(start), length);
        }
        bytes[dartx.add](chunk);
      }
      dart.fn(addChunk, dart.void, [typed_data.Uint8List, core.int, core.int]);
      _JsonUtf8Stringifier.stringify(object, this[_indent], dart.as(this[_toEncodable$], dart.functionType(dart.dynamic, [core.Object])), this[_bufferSize], addChunk);
      if (bytes[dartx.length] == 1)
        return bytes[dartx.get](0);
      let length = 0;
      for (let i = 0; dart.notNull(i) < dart.notNull(bytes[dartx.length]); i = dart.notNull(i) + 1) {
        length = dart.notNull(length) + dart.notNull(bytes[dartx.get](i)[dartx.length]);
      }
      let result = typed_data.Uint8List.new(length);
      for (let i = 0, offset = 0; dart.notNull(i) < dart.notNull(bytes[dartx.length]); i = dart.notNull(i) + 1) {
        let byteList = bytes[dartx.get](i);
        let end = dart.notNull(offset) + dart.notNull(byteList[dartx.length]);
        result.setRange(offset, end, byteList);
        offset = end;
      }
      return result;
    }
    startChunkedConversion(sink) {
      let byteSink = null;
      if (dart.is(sink, ByteConversionSink)) {
        byteSink = sink;
      } else {
        byteSink = ByteConversionSink.from(sink);
      }
      return new _JsonUtf8EncoderSink(byteSink, this[_toEncodable$], this[_indent], this[_bufferSize]);
    }
    bind(stream) {
      return super.bind(stream);
    }
    fuse(other) {
      return super.fuse(other);
    }
  }
  dart.setSignature(JsonUtf8Encoder, {
    constructors: () => ({JsonUtf8Encoder: [JsonUtf8Encoder, [], [core.String, dart.functionType(dart.dynamic, [core.Object]), core.int]]}),
    methods: () => ({
      convert: [core.List$(core.int), [core.Object]],
      startChunkedConversion: [ChunkedConversionSink$(core.Object), [core.Sink$(core.List$(core.int))]],
      bind: [async.Stream$(core.List$(core.int)), [async.Stream$(core.Object)]],
      fuse: [Converter$(core.Object, dart.dynamic), [Converter$(core.List$(core.int), dart.dynamic)]]
    }),
    statics: () => ({_utf8Encode: [core.List$(core.int), [core.String]]}),
    names: ['_utf8Encode']
  });
  JsonUtf8Encoder.DEFAULT_BUFFER_SIZE = 256;
  const _isDone = Symbol('_isDone');
  class _JsonEncoderSink extends ChunkedConversionSink$(core.Object) {
    _JsonEncoderSink(sink, toEncodable, indent) {
      this[_sink] = sink;
      this[_toEncodable$] = toEncodable;
      this[_indent] = indent;
      this[_isDone] = false;
      super.ChunkedConversionSink();
    }
    add(o) {
      if (dart.notNull(this[_isDone])) {
        dart.throw(new core.StateError("Only one call to add allowed"));
      }
      this[_isDone] = true;
      let stringSink = this[_sink].asStringSink();
      _JsonStringStringifier.printOn(o, stringSink, dart.as(this[_toEncodable$], dart.functionType(dart.dynamic, [dart.dynamic])), this[_indent]);
      stringSink.close();
    }
    close() {}
  }
  dart.setSignature(_JsonEncoderSink, {
    constructors: () => ({_JsonEncoderSink: [_JsonEncoderSink, [StringConversionSink, core.Function, core.String]]}),
    methods: () => ({
      add: [dart.void, [core.Object]],
      close: [dart.void, []]
    })
  });
  const _addChunk = Symbol('_addChunk');
  class _JsonUtf8EncoderSink extends ChunkedConversionSink$(core.Object) {
    _JsonUtf8EncoderSink(sink, toEncodable, indent, bufferSize) {
      this[_sink] = sink;
      this[_toEncodable$] = toEncodable;
      this[_indent] = indent;
      this[_bufferSize] = bufferSize;
      this[_isDone] = false;
      super.ChunkedConversionSink();
    }
    [_addChunk](chunk, start, end) {
      this[_sink].addSlice(chunk, start, end, false);
    }
    add(object) {
      if (dart.notNull(this[_isDone])) {
        dart.throw(new core.StateError("Only one call to add allowed"));
      }
      this[_isDone] = true;
      _JsonUtf8Stringifier.stringify(object, this[_indent], dart.as(this[_toEncodable$], dart.functionType(dart.dynamic, [core.Object])), this[_bufferSize], dart.bind(this, _addChunk));
      this[_sink].close();
    }
    close() {
      if (!dart.notNull(this[_isDone])) {
        this[_isDone] = true;
        this[_sink].close();
      }
    }
  }
  dart.setSignature(_JsonUtf8EncoderSink, {
    constructors: () => ({_JsonUtf8EncoderSink: [_JsonUtf8EncoderSink, [ByteConversionSink, core.Function, core.List$(core.int), core.int]]}),
    methods: () => ({
      [_addChunk]: [dart.void, [typed_data.Uint8List, core.int, core.int]],
      add: [dart.void, [core.Object]],
      close: [dart.void, []]
    })
  });
  class JsonDecoder extends Converter$(core.String, core.Object) {
    JsonDecoder(reviver) {
      if (reviver === void 0)
        reviver = null;
      this[_reviver] = reviver;
      super.Converter();
    }
    convert(input) {
      return _parseJson(input, this[_reviver]);
    }
    startChunkedConversion(sink) {
      return new _JsonDecoderSink(this[_reviver], sink);
    }
    bind(stream) {
      return super.bind(stream);
    }
  }
  dart.setSignature(JsonDecoder, {
    constructors: () => ({JsonDecoder: [JsonDecoder, [], [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]]}),
    methods: () => ({
      convert: [dart.dynamic, [core.String]],
      startChunkedConversion: [StringConversionSink, [core.Sink$(core.Object)]],
      bind: [async.Stream$(core.Object), [async.Stream$(core.String)]]
    })
  });
  function _parseJson(source, reviver) {
    if (!(typeof source == 'string'))
      dart.throw(new core.ArgumentError(source));
    let parsed = null;
    try {
      parsed = JSON.parse(source);
    } catch (e) {
      dart.throw(new core.FormatException(String(e)));
    }

    if (reviver == null) {
      return _convertJsonToDartLazy(parsed);
    } else {
      return _convertJsonToDart(parsed, reviver);
    }
  }
  dart.fn(_parseJson, dart.dynamic, [core.String, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]);
  function _defaultToEncodable(object) {
    return dart.dsend(object, 'toJson');
  }
  dart.fn(_defaultToEncodable, core.Object, [dart.dynamic]);
  const _seen = Symbol('_seen');
  const _checkCycle = Symbol('_checkCycle');
  const _removeSeen = Symbol('_removeSeen');
  class _JsonStringifier extends core.Object {
    _JsonStringifier(_toEncodable) {
      this[_seen] = core.List.new();
      this[_toEncodable$] = _toEncodable != null ? _toEncodable : _defaultToEncodable;
    }
    static hexDigit(x) {
      return dart.notNull(x) < 10 ? 48 + dart.notNull(x) : 87 + dart.notNull(x);
    }
    writeStringContent(s) {
      let offset = 0;
      let length = s[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        let charCode = s[dartx.codeUnitAt](i);
        if (dart.notNull(charCode) > dart.notNull(_JsonStringifier.BACKSLASH))
          continue;
        if (dart.notNull(charCode) < 32) {
          if (dart.notNull(i) > dart.notNull(offset))
            this.writeStringSlice(s, offset, i);
          offset = dart.notNull(i) + 1;
          this.writeCharCode(_JsonStringifier.BACKSLASH);
          switch (charCode) {
            case _JsonStringifier.BACKSPACE:
            {
              this.writeCharCode(_JsonStringifier.CHAR_b);
              break;
            }
            case _JsonStringifier.TAB:
            {
              this.writeCharCode(_JsonStringifier.CHAR_t);
              break;
            }
            case _JsonStringifier.NEWLINE:
            {
              this.writeCharCode(_JsonStringifier.CHAR_n);
              break;
            }
            case _JsonStringifier.FORM_FEED:
            {
              this.writeCharCode(_JsonStringifier.CHAR_f);
              break;
            }
            case _JsonStringifier.CARRIAGE_RETURN:
            {
              this.writeCharCode(_JsonStringifier.CHAR_r);
              break;
            }
            default:
            {
              this.writeCharCode(_JsonStringifier.CHAR_u);
              this.writeCharCode(_JsonStringifier.CHAR_0);
              this.writeCharCode(_JsonStringifier.CHAR_0);
              this.writeCharCode(_JsonStringifier.hexDigit(dart.notNull(charCode) >> 4 & 15));
              this.writeCharCode(_JsonStringifier.hexDigit(dart.notNull(charCode) & 15));
              break;
            }
          }
        } else if (charCode == _JsonStringifier.QUOTE || charCode == _JsonStringifier.BACKSLASH) {
          if (dart.notNull(i) > dart.notNull(offset))
            this.writeStringSlice(s, offset, i);
          offset = dart.notNull(i) + 1;
          this.writeCharCode(_JsonStringifier.BACKSLASH);
          this.writeCharCode(charCode);
        }
      }
      if (offset == 0) {
        this.writeString(s);
      } else if (dart.notNull(offset) < dart.notNull(length)) {
        this.writeStringSlice(s, offset, length);
      }
    }
    [_checkCycle](object) {
      for (let i = 0; dart.notNull(i) < dart.notNull(this[_seen][dartx.length]); i = dart.notNull(i) + 1) {
        if (dart.notNull(core.identical(object, this[_seen][dartx.get](i)))) {
          dart.throw(new JsonCyclicError(object));
        }
      }
      this[_seen][dartx.add](object);
    }
    [_removeSeen](object) {
      dart.assert(!dart.notNull(this[_seen][dartx.isEmpty]));
      dart.assert(core.identical(this[_seen][dartx.last], object));
      this[_seen][dartx.removeLast]();
    }
    writeObject(object) {
      if (dart.notNull(this.writeJsonValue(object)))
        return;
      this[_checkCycle](object);
      try {
        let customJson = dart.dcall(this[_toEncodable$], object);
        if (!dart.notNull(this.writeJsonValue(customJson))) {
          dart.throw(new JsonUnsupportedObjectError(object));
        }
        this[_removeSeen](object);
      } catch (e) {
        dart.throw(new JsonUnsupportedObjectError(object, {cause: e}));
      }

    }
    writeJsonValue(object) {
      if (typeof object == 'number') {
        if (!dart.notNull(dart.as(dart.dload(object, 'isFinite'), core.bool)))
          return false;
        this.writeNumber(dart.as(object, core.num));
        return true;
      } else if (dart.notNull(core.identical(object, true))) {
        this.writeString('true');
        return true;
      } else if (dart.notNull(core.identical(object, false))) {
        this.writeString('false');
        return true;
      } else if (object == null) {
        this.writeString('null');
        return true;
      } else if (typeof object == 'string') {
        this.writeString('"');
        this.writeStringContent(dart.as(object, core.String));
        this.writeString('"');
        return true;
      } else if (dart.is(object, core.List)) {
        this[_checkCycle](object);
        this.writeList(dart.as(object, core.List));
        this[_removeSeen](object);
        return true;
      } else if (dart.is(object, core.Map)) {
        this[_checkCycle](object);
        this.writeMap(dart.as(object, core.Map$(core.String, core.Object)));
        this[_removeSeen](object);
        return true;
      } else {
        return false;
      }
    }
    writeList(list) {
      this.writeString('[');
      if (dart.notNull(list[dartx.length]) > 0) {
        this.writeObject(list[dartx.get](0));
        for (let i = 1; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
          this.writeString(',');
          this.writeObject(list[dartx.get](i));
        }
      }
      this.writeString(']');
    }
    writeMap(map) {
      this.writeString('{');
      let separator = '"';
      map.forEach(dart.fn(((key, value) => {
        this.writeString(separator);
        separator = ',"';
        this.writeStringContent(key);
        this.writeString('":');
        this.writeObject(value);
      }).bind(this), dart.dynamic, [core.String, dart.dynamic]));
      this.writeString('}');
    }
  }
  dart.setSignature(_JsonStringifier, {
    constructors: () => ({_JsonStringifier: [_JsonStringifier, [dart.functionType(core.Object, [core.Object])]]}),
    methods: () => ({
      writeStringContent: [dart.void, [core.String]],
      [_checkCycle]: [dart.void, [dart.dynamic]],
      [_removeSeen]: [dart.void, [dart.dynamic]],
      writeObject: [dart.void, [dart.dynamic]],
      writeJsonValue: [core.bool, [dart.dynamic]],
      writeList: [dart.void, [core.List]],
      writeMap: [dart.void, [core.Map$(core.String, core.Object)]]
    }),
    statics: () => ({hexDigit: [core.int, [core.int]]}),
    names: ['hexDigit']
  });
  _JsonStringifier.BACKSPACE = 8;
  _JsonStringifier.TAB = 9;
  _JsonStringifier.NEWLINE = 10;
  _JsonStringifier.CARRIAGE_RETURN = 13;
  _JsonStringifier.FORM_FEED = 12;
  _JsonStringifier.QUOTE = 34;
  _JsonStringifier.CHAR_0 = 48;
  _JsonStringifier.BACKSLASH = 92;
  _JsonStringifier.CHAR_b = 98;
  _JsonStringifier.CHAR_f = 102;
  _JsonStringifier.CHAR_n = 110;
  _JsonStringifier.CHAR_r = 114;
  _JsonStringifier.CHAR_t = 116;
  _JsonStringifier.CHAR_u = 117;
  const _indentLevel = Symbol('_indentLevel');
  class _JsonPrettyPrintMixin extends core.Object {
    _JsonPrettyPrintMixin() {
      this[_indentLevel] = 0;
    }
    writeList(list) {
      if (dart.notNull(list[dartx.isEmpty])) {
        this.writeString('[]');
      } else {
        this.writeString('[\n');
        this[_indentLevel] = dart.notNull(this[_indentLevel]) + 1;
        this.writeIndentation(this[_indentLevel]);
        this.writeObject(list[dartx.get](0));
        for (let i = 1; dart.notNull(i) < dart.notNull(list[dartx.length]); i = dart.notNull(i) + 1) {
          this.writeString(',\n');
          this.writeIndentation(this[_indentLevel]);
          this.writeObject(list[dartx.get](i));
        }
        this.writeString('\n');
        this[_indentLevel] = dart.notNull(this[_indentLevel]) - 1;
        this.writeIndentation(this[_indentLevel]);
        this.writeString(']');
      }
    }
    writeMap(map) {
      if (dart.notNull(map.isEmpty)) {
        this.writeString('{}');
      } else {
        this.writeString('{\n');
        this[_indentLevel] = dart.notNull(this[_indentLevel]) + 1;
        let first = true;
        map.forEach(dart.fn(((key, value) => {
          if (!dart.notNull(first)) {
            this.writeString(",\n");
          }
          this.writeIndentation(this[_indentLevel]);
          this.writeString('"');
          this.writeStringContent(key);
          this.writeString('": ');
          this.writeObject(value);
          first = false;
        }).bind(this), dart.dynamic, [core.String, core.Object]));
        this.writeString('\n');
        this[_indentLevel] = dart.notNull(this[_indentLevel]) - 1;
        this.writeIndentation(this[_indentLevel]);
        this.writeString('}');
      }
    }
  }
  _JsonPrettyPrintMixin[dart.implements] = () => [_JsonStringifier];
  dart.setSignature(_JsonPrettyPrintMixin, {
    methods: () => ({
      writeList: [dart.void, [core.List]],
      writeMap: [dart.void, [core.Map]]
    })
  });
  class _JsonStringStringifier extends _JsonStringifier {
    _JsonStringStringifier(sink, _toEncodable) {
      this[_sink] = sink;
      super._JsonStringifier(dart.as(_toEncodable, dart.functionType(core.Object, [core.Object])));
    }
    static stringify(object, toEncodable, indent) {
      let output = new core.StringBuffer();
      _JsonStringStringifier.printOn(object, output, toEncodable, indent);
      return dart.toString(output);
    }
    static printOn(object, output, toEncodable, indent) {
      let stringifier = null;
      if (indent == null) {
        stringifier = new _JsonStringStringifier(output, toEncodable);
      } else {
        stringifier = new _JsonStringStringifierPretty(output, toEncodable, indent);
      }
      dart.dsend(stringifier, 'writeObject', object);
    }
    writeNumber(number) {
      this[_sink].write(dart.toString(number));
    }
    writeString(string) {
      this[_sink].write(string);
    }
    writeStringSlice(string, start, end) {
      this[_sink].write(string[dartx.substring](start, end));
    }
    writeCharCode(charCode) {
      this[_sink].writeCharCode(charCode);
    }
  }
  dart.setSignature(_JsonStringStringifier, {
    constructors: () => ({_JsonStringStringifier: [_JsonStringStringifier, [core.StringSink, dart.dynamic]]}),
    methods: () => ({
      writeNumber: [dart.void, [core.num]],
      writeString: [dart.void, [core.String]],
      writeStringSlice: [dart.void, [core.String, core.int, core.int]],
      writeCharCode: [dart.void, [core.int]]
    }),
    statics: () => ({
      stringify: [core.String, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic]), core.String]],
      printOn: [dart.void, [dart.dynamic, core.StringSink, dart.functionType(dart.dynamic, [dart.dynamic]), core.String]]
    }),
    names: ['stringify', 'printOn']
  });
  class _JsonStringStringifierPretty extends dart.mixin(_JsonStringStringifier, _JsonPrettyPrintMixin) {
    _JsonStringStringifierPretty(sink, toEncodable, indent) {
      this[_indent] = indent;
      super._JsonStringStringifier(sink, toEncodable);
    }
    writeIndentation(count) {
      for (let i = 0; dart.notNull(i) < dart.notNull(count); i = dart.notNull(i) + 1)
        this.writeString(this[_indent]);
    }
  }
  dart.setSignature(_JsonStringStringifierPretty, {
    constructors: () => ({_JsonStringStringifierPretty: [_JsonStringStringifierPretty, [core.StringSink, core.Function, core.String]]}),
    methods: () => ({writeIndentation: [dart.void, [core.int]]})
  });
  class _JsonUtf8Stringifier extends _JsonStringifier {
    _JsonUtf8Stringifier(toEncodable, bufferSize, addChunk) {
      this.addChunk = addChunk;
      this.bufferSize = bufferSize;
      this.buffer = typed_data.Uint8List.new(bufferSize);
      this.index = 0;
      super._JsonStringifier(dart.as(toEncodable, dart.functionType(core.Object, [core.Object])));
    }
    static stringify(object, indent, toEncodableFunction, bufferSize, addChunk) {
      let stringifier = null;
      if (indent != null) {
        stringifier = new _JsonUtf8StringifierPretty(toEncodableFunction, indent, bufferSize, addChunk);
      } else {
        stringifier = new _JsonUtf8Stringifier(toEncodableFunction, bufferSize, addChunk);
      }
      stringifier.writeObject(object);
      stringifier.flush();
    }
    flush() {
      if (dart.notNull(this.index) > 0) {
        dart.dcall(this.addChunk, this.buffer, 0, this.index);
      }
      this.buffer = null;
      this.index = 0;
    }
    writeNumber(number) {
      this.writeAsciiString(dart.toString(number));
    }
    writeAsciiString(string) {
      for (let i = 0; dart.notNull(i) < dart.notNull(string[dartx.length]); i = dart.notNull(i) + 1) {
        let char = string[dartx.codeUnitAt](i);
        dart.assert(dart.notNull(char) <= 127);
        this.writeByte(char);
      }
    }
    writeString(string) {
      this.writeStringSlice(string, 0, string[dartx.length]);
    }
    writeStringSlice(string, start, end) {
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let char = string[dartx.codeUnitAt](i);
        if (dart.notNull(char) <= 127) {
          this.writeByte(char);
        } else {
          if ((dart.notNull(char) & 64512) == 55296 && dart.notNull(i) + 1 < dart.notNull(end)) {
            let nextChar = string[dartx.codeUnitAt](dart.notNull(i) + 1);
            if ((dart.notNull(nextChar) & 64512) == 56320) {
              char = 65536 + ((dart.notNull(char) & 1023) << 10) + (dart.notNull(nextChar) & 1023);
              this.writeFourByteCharCode(char);
              i = dart.notNull(i) + 1;
              continue;
            }
          }
          this.writeMultiByteCharCode(char);
        }
      }
    }
    writeCharCode(charCode) {
      if (dart.notNull(charCode) <= 127) {
        this.writeByte(charCode);
        return;
      }
      this.writeMultiByteCharCode(charCode);
    }
    writeMultiByteCharCode(charCode) {
      if (dart.notNull(charCode) <= 2047) {
        this.writeByte(192 | dart.notNull(charCode) >> 6);
        this.writeByte(128 | dart.notNull(charCode) & 63);
        return;
      }
      if (dart.notNull(charCode) <= 65535) {
        this.writeByte(224 | dart.notNull(charCode) >> 12);
        this.writeByte(128 | dart.notNull(charCode) >> 6 & 63);
        this.writeByte(128 | dart.notNull(charCode) & 63);
        return;
      }
      this.writeFourByteCharCode(charCode);
    }
    writeFourByteCharCode(charCode) {
      dart.assert(dart.notNull(charCode) <= 1114111);
      this.writeByte(240 | dart.notNull(charCode) >> 18);
      this.writeByte(128 | dart.notNull(charCode) >> 12 & 63);
      this.writeByte(128 | dart.notNull(charCode) >> 6 & 63);
      this.writeByte(128 | dart.notNull(charCode) & 63);
    }
    writeByte(byte) {
      dart.assert(dart.notNull(byte) <= 255);
      if (this.index == this.buffer.length) {
        dart.dcall(this.addChunk, this.buffer, 0, this.index);
        this.buffer = typed_data.Uint8List.new(this.bufferSize);
        this.index = 0;
      }
      this.buffer.set((() => {
        let x = this.index;
        this.index = dart.notNull(x) + 1;
        return x;
      }).bind(this)(), byte);
    }
  }
  dart.setSignature(_JsonUtf8Stringifier, {
    constructors: () => ({_JsonUtf8Stringifier: [_JsonUtf8Stringifier, [dart.dynamic, core.int, core.Function]]}),
    methods: () => ({
      flush: [dart.void, []],
      writeNumber: [dart.void, [core.num]],
      writeAsciiString: [dart.void, [core.String]],
      writeString: [dart.void, [core.String]],
      writeStringSlice: [dart.void, [core.String, core.int, core.int]],
      writeCharCode: [dart.void, [core.int]],
      writeMultiByteCharCode: [dart.void, [core.int]],
      writeFourByteCharCode: [dart.void, [core.int]],
      writeByte: [dart.void, [core.int]]
    }),
    statics: () => ({stringify: [dart.void, [core.Object, core.List$(core.int), dart.functionType(dart.dynamic, [core.Object]), core.int, dart.functionType(dart.void, [typed_data.Uint8List, core.int, core.int])]]}),
    names: ['stringify']
  });
  class _JsonUtf8StringifierPretty extends dart.mixin(_JsonUtf8Stringifier, _JsonPrettyPrintMixin) {
    _JsonUtf8StringifierPretty(toEncodableFunction, indent, bufferSize, addChunk) {
      this.indent = indent;
      super._JsonUtf8Stringifier(toEncodableFunction, dart.as(bufferSize, core.int), dart.as(addChunk, core.Function));
    }
    writeIndentation(count) {
      let indent = this.indent;
      let indentLength = indent[dartx.length];
      if (indentLength == 1) {
        let char = indent[dartx.get](0);
        while (dart.notNull(count) > 0) {
          this.writeByte(char);
          count = dart.notNull(count) - 1;
        }
        return;
      }
      while (dart.notNull(count) > 0) {
        count = dart.notNull(count) - 1;
        let end = dart.notNull(this.index) + dart.notNull(indentLength);
        if (dart.notNull(end) <= dart.notNull(this.buffer.length)) {
          this.buffer.setRange(this.index, end, indent);
          this.index = end;
        } else {
          for (let i = 0; dart.notNull(i) < dart.notNull(indentLength); i = dart.notNull(i) + 1) {
            this.writeByte(indent[dartx.get](i));
          }
        }
      }
    }
  }
  dart.setSignature(_JsonUtf8StringifierPretty, {
    constructors: () => ({_JsonUtf8StringifierPretty: [_JsonUtf8StringifierPretty, [dart.dynamic, core.List$(core.int), dart.dynamic, dart.dynamic]]}),
    methods: () => ({writeIndentation: [dart.void, [core.int]]})
  });
  const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(core.Object, [core.Object]));
  const __CastType2 = dart.typedef('__CastType2', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  const __CastType4 = dart.typedef('__CastType4', () => dart.functionType(dart.dynamic, [core.Object]));
  class Latin1Codec extends Encoding {
    Latin1Codec(opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : false;
      this[_allowInvalid] = allowInvalid;
      super.Encoding();
    }
    get name() {
      return "iso-8859-1";
    }
    decode(bytes, opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : null;
      if (allowInvalid == null)
        allowInvalid = this[_allowInvalid];
      if (dart.notNull(allowInvalid)) {
        return dart.const(new Latin1Decoder({allowInvalid: true})).convert(bytes);
      } else {
        return dart.const(new Latin1Decoder({allowInvalid: false})).convert(bytes);
      }
    }
    get encoder() {
      return dart.const(new Latin1Encoder());
    }
    get decoder() {
      return dart.notNull(this[_allowInvalid]) ? dart.const(new Latin1Decoder({allowInvalid: true})) : dart.const(new Latin1Decoder({allowInvalid: false}));
    }
  }
  dart.setSignature(Latin1Codec, {
    constructors: () => ({Latin1Codec: [Latin1Codec, [], {allowInvalid: core.bool}]}),
    methods: () => ({decode: [core.String, [core.List$(core.int)], {allowInvalid: core.bool}]})
  });
  const LATIN1 = dart.const(new Latin1Codec());
  const _LATIN1_MASK = 255;
  class Latin1Encoder extends _UnicodeSubsetEncoder {
    Latin1Encoder() {
      super._UnicodeSubsetEncoder(_LATIN1_MASK);
    }
  }
  dart.setSignature(Latin1Encoder, {
    constructors: () => ({Latin1Encoder: [Latin1Encoder, []]})
  });
  class Latin1Decoder extends _UnicodeSubsetDecoder {
    Latin1Decoder(opts) {
      let allowInvalid = opts && 'allowInvalid' in opts ? opts.allowInvalid : false;
      super._UnicodeSubsetDecoder(allowInvalid, _LATIN1_MASK);
    }
    startChunkedConversion(sink) {
      let stringSink = null;
      if (dart.is(sink, StringConversionSink)) {
        stringSink = sink;
      } else {
        stringSink = StringConversionSink.from(sink);
      }
      if (!dart.notNull(this[_allowInvalid]))
        return new _Latin1DecoderSink(stringSink);
      return new _Latin1AllowInvalidDecoderSink(stringSink);
    }
  }
  dart.setSignature(Latin1Decoder, {
    constructors: () => ({Latin1Decoder: [Latin1Decoder, [], {allowInvalid: core.bool}]}),
    methods: () => ({startChunkedConversion: [ByteConversionSink, [core.Sink$(core.String)]]})
  });
  const _addSliceToSink = Symbol('_addSliceToSink');
  class _Latin1DecoderSink extends ByteConversionSinkBase {
    _Latin1DecoderSink(sink) {
      this[_sink] = sink;
    }
    close() {
      this[_sink].close();
    }
    add(source) {
      this.addSlice(source, 0, source[dartx.length], false);
    }
    [_addSliceToSink](source, start, end, isLast) {
      this[_sink].add(core.String.fromCharCodes(source, start, end));
      if (dart.notNull(isLast))
        this.close();
    }
    addSlice(source, start, end, isLast) {
      core.RangeError.checkValidRange(start, end, source[dartx.length]);
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let char = source[dartx.get](i);
        if (dart.notNull(char) > dart.notNull(_LATIN1_MASK) || dart.notNull(char) < 0) {
          dart.throw(new core.FormatException("Source contains non-Latin-1 characters."));
        }
      }
      if (dart.notNull(start) < dart.notNull(end)) {
        this[_addSliceToSink](source, start, end, isLast);
      }
      if (dart.notNull(isLast)) {
        this.close();
      }
    }
  }
  dart.setSignature(_Latin1DecoderSink, {
    constructors: () => ({_Latin1DecoderSink: [_Latin1DecoderSink, [StringConversionSink]]}),
    methods: () => ({
      close: [dart.void, []],
      add: [dart.void, [core.List$(core.int)]],
      [_addSliceToSink]: [dart.void, [core.List$(core.int), core.int, core.int, core.bool]]
    })
  });
  class _Latin1AllowInvalidDecoderSink extends _Latin1DecoderSink {
    _Latin1AllowInvalidDecoderSink(sink) {
      super._Latin1DecoderSink(sink);
    }
    addSlice(source, start, end, isLast) {
      core.RangeError.checkValidRange(start, end, source[dartx.length]);
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let char = source[dartx.get](i);
        if (dart.notNull(char) > dart.notNull(_LATIN1_MASK) || dart.notNull(char) < 0) {
          if (dart.notNull(i) > dart.notNull(start))
            this[_addSliceToSink](source, start, i, false);
          this[_addSliceToSink](dart.const(dart.list([65533], core.int)), 0, 1, false);
          start = dart.notNull(i) + 1;
        }
      }
      if (dart.notNull(start) < dart.notNull(end)) {
        this[_addSliceToSink](source, start, end, isLast);
      }
      if (dart.notNull(isLast)) {
        this.close();
      }
    }
  }
  dart.setSignature(_Latin1AllowInvalidDecoderSink, {
    constructors: () => ({_Latin1AllowInvalidDecoderSink: [_Latin1AllowInvalidDecoderSink, [StringConversionSink]]})
  });
  class LineSplitter extends Converter$(core.String, core.List$(core.String)) {
    LineSplitter() {
      super.Converter();
    }
    convert(data) {
      let lines = core.List$(core.String).new();
      _LineSplitterSink._addSlice(data, 0, data[dartx.length], true, dart.bind(lines, dartx.add));
      return lines;
    }
    startChunkedConversion(sink) {
      if (!dart.is(sink, StringConversionSink)) {
        sink = StringConversionSink.from(dart.as(sink, core.Sink$(core.String)));
      }
      return new _LineSplitterSink(dart.as(sink, StringConversionSink));
    }
  }
  dart.setSignature(LineSplitter, {
    constructors: () => ({LineSplitter: [LineSplitter, []]}),
    methods: () => ({
      convert: [core.List$(core.String), [core.String]],
      startChunkedConversion: [StringConversionSink, [core.Sink]]
    })
  });
  const _carry = Symbol('_carry');
  class _LineSplitterSink extends StringConversionSinkBase {
    _LineSplitterSink(sink) {
      this[_sink] = sink;
      this[_carry] = null;
    }
    addSlice(chunk, start, end, isLast) {
      if (this[_carry] != null) {
        chunk = dart.notNull(this[_carry]) + dart.notNull(chunk[dartx.substring](start, end));
        start = 0;
        end = chunk[dartx.length];
        this[_carry] = null;
      }
      this[_carry] = _LineSplitterSink._addSlice(chunk, start, end, isLast, dart.bind(this[_sink], 'add'));
      if (dart.notNull(isLast))
        this[_sink].close();
    }
    close() {
      this.addSlice('', 0, 0, true);
    }
    static _addSlice(chunk, start, end, isLast, adder) {
      let pos = start;
      while (dart.notNull(pos) < dart.notNull(end)) {
        let skip = 0;
        let char = chunk[dartx.codeUnitAt](pos);
        if (char == _LineSplitterSink._LF) {
          skip = 1;
        } else if (char == _LineSplitterSink._CR) {
          skip = 1;
          if (dart.notNull(pos) + 1 < dart.notNull(end)) {
            if (chunk[dartx.codeUnitAt](dart.notNull(pos) + 1) == _LineSplitterSink._LF) {
              skip = 2;
            }
          } else if (!dart.notNull(isLast)) {
            return chunk[dartx.substring](start, end);
          }
        }
        if (dart.notNull(skip) > 0) {
          adder(chunk[dartx.substring](start, pos));
          start = pos = dart.notNull(pos) + dart.notNull(skip);
        } else {
          pos = dart.notNull(pos) + 1;
        }
      }
      if (pos != start) {
        let carry = chunk[dartx.substring](start, pos);
        if (dart.notNull(isLast)) {
          adder(carry);
        } else {
          return carry;
        }
      }
      return null;
    }
  }
  dart.setSignature(_LineSplitterSink, {
    constructors: () => ({_LineSplitterSink: [_LineSplitterSink, [StringConversionSink]]}),
    methods: () => ({
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]],
      close: [dart.void, []]
    }),
    statics: () => ({_addSlice: [core.String, [core.String, core.int, core.int, core.bool, dart.functionType(dart.void, [core.String])]]}),
    names: ['_addSlice']
  });
  _LineSplitterSink._LF = 10;
  _LineSplitterSink._CR = 13;
  class StringConversionSink extends ChunkedConversionSink$(core.String) {
    StringConversionSink() {
      super.ChunkedConversionSink();
    }
    static withCallback(callback) {
      return new _StringCallbackSink(callback);
    }
    static from(sink) {
      return new _StringAdapterSink(sink);
    }
    static fromStringSink(sink) {
      return new _StringSinkConversionSink(sink);
    }
  }
  dart.setSignature(StringConversionSink, {
    constructors: () => ({
      StringConversionSink: [StringConversionSink, []],
      withCallback: [StringConversionSink, [dart.functionType(dart.void, [core.String])]],
      from: [StringConversionSink, [core.Sink$(core.String)]],
      fromStringSink: [StringConversionSink, [core.StringSink]]
    })
  });
  class ClosableStringSink extends core.StringSink {
    static fromStringSink(sink, onClose) {
      return new _ClosableStringSink(sink, onClose);
    }
  }
  dart.setSignature(ClosableStringSink, {
    constructors: () => ({fromStringSink: [ClosableStringSink, [core.StringSink, dart.functionType(dart.void, [])]]})
  });
  const _StringSinkCloseCallback = dart.typedef('_StringSinkCloseCallback', () => dart.functionType(dart.void, []));
  class _ClosableStringSink extends core.Object {
    _ClosableStringSink(sink, callback) {
      this[_sink] = sink;
      this[_callback] = callback;
    }
    close() {
      return this[_callback]();
    }
    writeCharCode(charCode) {
      return this[_sink].writeCharCode(charCode);
    }
    write(o) {
      return this[_sink].write(o);
    }
    writeln(o) {
      if (o === void 0)
        o = "";
      return this[_sink].writeln(o);
    }
    writeAll(objects, separator) {
      if (separator === void 0)
        separator = "";
      return this[_sink].writeAll(objects, separator);
    }
  }
  _ClosableStringSink[dart.implements] = () => [ClosableStringSink];
  dart.setSignature(_ClosableStringSink, {
    constructors: () => ({_ClosableStringSink: [_ClosableStringSink, [core.StringSink, _StringSinkCloseCallback]]}),
    methods: () => ({
      close: [dart.void, []],
      writeCharCode: [dart.void, [core.int]],
      write: [dart.void, [core.Object]],
      writeln: [dart.void, [], [core.Object]],
      writeAll: [dart.void, [core.Iterable], [core.String]]
    })
  });
  const _flush = Symbol('_flush');
  class _StringConversionSinkAsStringSinkAdapter extends core.Object {
    _StringConversionSinkAsStringSinkAdapter(chunkedSink) {
      this[_chunkedSink] = chunkedSink;
      this[_buffer] = new core.StringBuffer();
    }
    close() {
      if (dart.notNull(this[_buffer].isNotEmpty))
        this[_flush]();
      this[_chunkedSink].close();
    }
    writeCharCode(charCode) {
      this[_buffer].writeCharCode(charCode);
      if (dart.notNull(this[_buffer].length) > dart.notNull(_StringConversionSinkAsStringSinkAdapter._MIN_STRING_SIZE))
        this[_flush]();
    }
    write(o) {
      if (dart.notNull(this[_buffer].isNotEmpty))
        this[_flush]();
      let str = dart.toString(o);
      this[_chunkedSink].add(dart.toString(o));
    }
    writeln(o) {
      if (o === void 0)
        o = "";
      this[_buffer].writeln(o);
      if (dart.notNull(this[_buffer].length) > dart.notNull(_StringConversionSinkAsStringSinkAdapter._MIN_STRING_SIZE))
        this[_flush]();
    }
    writeAll(objects, separator) {
      if (separator === void 0)
        separator = "";
      if (dart.notNull(this[_buffer].isNotEmpty))
        this[_flush]();
      let iterator = objects[dartx.iterator];
      if (!dart.notNull(iterator.moveNext()))
        return;
      if (dart.notNull(separator[dartx.isEmpty])) {
        do {
          this[_chunkedSink].add(dart.toString(iterator.current));
        } while (dart.notNull(iterator.moveNext()));
      } else {
        this[_chunkedSink].add(dart.toString(iterator.current));
        while (dart.notNull(iterator.moveNext())) {
          this.write(separator);
          this[_chunkedSink].add(dart.toString(iterator.current));
        }
      }
    }
    [_flush]() {
      let accumulated = dart.toString(this[_buffer]);
      this[_buffer].clear();
      this[_chunkedSink].add(accumulated);
    }
  }
  _StringConversionSinkAsStringSinkAdapter[dart.implements] = () => [ClosableStringSink];
  dart.setSignature(_StringConversionSinkAsStringSinkAdapter, {
    constructors: () => ({_StringConversionSinkAsStringSinkAdapter: [_StringConversionSinkAsStringSinkAdapter, [StringConversionSink]]}),
    methods: () => ({
      close: [dart.void, []],
      writeCharCode: [dart.void, [core.int]],
      write: [dart.void, [core.Object]],
      writeln: [dart.void, [], [core.Object]],
      writeAll: [dart.void, [core.Iterable], [core.String]],
      [_flush]: [dart.void, []]
    })
  });
  _StringConversionSinkAsStringSinkAdapter._MIN_STRING_SIZE = 16;
  const _stringSink = Symbol('_stringSink');
  class _StringSinkConversionSink extends StringConversionSinkBase {
    _StringSinkConversionSink(stringSink) {
      this[_stringSink] = stringSink;
    }
    close() {}
    addSlice(str, start, end, isLast) {
      if (start != 0 || end != str[dartx.length]) {
        for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
          this[_stringSink].writeCharCode(str[dartx.codeUnitAt](i));
        }
      } else {
        this[_stringSink].write(str);
      }
      if (dart.notNull(isLast))
        this.close();
    }
    add(str) {
      return this[_stringSink].write(str);
    }
    asUtf8Sink(allowMalformed) {
      return new _Utf8StringSinkAdapter(this, this[_stringSink], allowMalformed);
    }
    asStringSink() {
      return ClosableStringSink.fromStringSink(this[_stringSink], dart.bind(this, 'close'));
    }
  }
  dart.setSignature(_StringSinkConversionSink, {
    constructors: () => ({_StringSinkConversionSink: [_StringSinkConversionSink, [core.StringSink]]}),
    methods: () => ({
      close: [dart.void, []],
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]]
    })
  });
  class _StringCallbackSink extends _StringSinkConversionSink {
    _StringCallbackSink(callback) {
      this[_callback] = callback;
      super._StringSinkConversionSink(new core.StringBuffer());
    }
    close() {
      let buffer = dart.as(this[_stringSink], core.StringBuffer);
      let accumulated = dart.toString(buffer);
      buffer.clear();
      this[_callback](accumulated);
    }
    asUtf8Sink(allowMalformed) {
      return new _Utf8StringSinkAdapter(this, this[_stringSink], allowMalformed);
    }
  }
  dart.setSignature(_StringCallbackSink, {
    constructors: () => ({_StringCallbackSink: [_StringCallbackSink, [_ChunkedConversionCallback$(core.String)]]})
  });
  class _StringAdapterSink extends StringConversionSinkBase {
    _StringAdapterSink(sink) {
      this[_sink] = sink;
    }
    add(str) {
      return this[_sink].add(str);
    }
    addSlice(str, start, end, isLast) {
      if (start == 0 && end == str[dartx.length]) {
        this.add(str);
      } else {
        this.add(str[dartx.substring](start, end));
      }
      if (dart.notNull(isLast))
        this.close();
    }
    close() {
      return this[_sink].close();
    }
  }
  dart.setSignature(_StringAdapterSink, {
    constructors: () => ({_StringAdapterSink: [_StringAdapterSink, [core.Sink$(core.String)]]}),
    methods: () => ({
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]],
      close: [dart.void, []]
    })
  });
  const _decoder = Symbol('_decoder');
  class _Utf8StringSinkAdapter extends ByteConversionSink {
    _Utf8StringSinkAdapter(sink, stringSink, allowMalformed) {
      this[_sink] = sink;
      this[_decoder] = new _Utf8Decoder(stringSink, allowMalformed);
      super.ByteConversionSink();
    }
    close() {
      this[_decoder].close();
      if (this[_sink] != null)
        this[_sink].close();
    }
    add(chunk) {
      this.addSlice(chunk, 0, chunk[dartx.length], false);
    }
    addSlice(codeUnits, startIndex, endIndex, isLast) {
      this[_decoder].convert(codeUnits, startIndex, endIndex);
      if (dart.notNull(isLast))
        this.close();
    }
  }
  dart.setSignature(_Utf8StringSinkAdapter, {
    constructors: () => ({_Utf8StringSinkAdapter: [_Utf8StringSinkAdapter, [core.Sink, core.StringSink, core.bool]]}),
    methods: () => ({
      close: [dart.void, []],
      add: [dart.void, [core.List$(core.int)]],
      addSlice: [dart.void, [core.List$(core.int), core.int, core.int, core.bool]]
    })
  });
  class _Utf8ConversionSink extends ByteConversionSink {
    _Utf8ConversionSink(sink, allowMalformed) {
      this._(sink, new core.StringBuffer(), allowMalformed);
    }
    _(chunkedSink, stringBuffer, allowMalformed) {
      this[_chunkedSink] = chunkedSink;
      this[_decoder] = new _Utf8Decoder(stringBuffer, allowMalformed);
      this[_buffer] = stringBuffer;
      super.ByteConversionSink();
    }
    close() {
      this[_decoder].close();
      if (dart.notNull(this[_buffer].isNotEmpty)) {
        let accumulated = dart.toString(this[_buffer]);
        this[_buffer].clear();
        this[_chunkedSink].addSlice(accumulated, 0, accumulated[dartx.length], true);
      } else {
        this[_chunkedSink].close();
      }
    }
    add(chunk) {
      this.addSlice(chunk, 0, chunk[dartx.length], false);
    }
    addSlice(chunk, startIndex, endIndex, isLast) {
      this[_decoder].convert(chunk, startIndex, endIndex);
      if (dart.notNull(this[_buffer].isNotEmpty)) {
        let accumulated = dart.toString(this[_buffer]);
        this[_chunkedSink].addSlice(accumulated, 0, accumulated[dartx.length], isLast);
        this[_buffer].clear();
        return;
      }
      if (dart.notNull(isLast))
        this.close();
    }
  }
  dart.defineNamedConstructor(_Utf8ConversionSink, '_');
  dart.setSignature(_Utf8ConversionSink, {
    constructors: () => ({
      _Utf8ConversionSink: [_Utf8ConversionSink, [StringConversionSink, core.bool]],
      _: [_Utf8ConversionSink, [StringConversionSink, core.StringBuffer, core.bool]]
    }),
    methods: () => ({
      close: [dart.void, []],
      add: [dart.void, [core.List$(core.int)]],
      addSlice: [dart.void, [core.List$(core.int), core.int, core.int, core.bool]]
    })
  });
  const UNICODE_REPLACEMENT_CHARACTER_RUNE = 65533;
  const UNICODE_BOM_CHARACTER_RUNE = 65279;
  const _allowMalformed = Symbol('_allowMalformed');
  class Utf8Codec extends Encoding {
    Utf8Codec(opts) {
      let allowMalformed = opts && 'allowMalformed' in opts ? opts.allowMalformed : false;
      this[_allowMalformed] = allowMalformed;
      super.Encoding();
    }
    get name() {
      return "utf-8";
    }
    decode(codeUnits, opts) {
      let allowMalformed = opts && 'allowMalformed' in opts ? opts.allowMalformed : null;
      if (allowMalformed == null)
        allowMalformed = this[_allowMalformed];
      return new Utf8Decoder({allowMalformed: allowMalformed}).convert(codeUnits);
    }
    get encoder() {
      return new Utf8Encoder();
    }
    get decoder() {
      return new Utf8Decoder({allowMalformed: this[_allowMalformed]});
    }
  }
  dart.setSignature(Utf8Codec, {
    constructors: () => ({Utf8Codec: [Utf8Codec, [], {allowMalformed: core.bool}]}),
    methods: () => ({decode: [core.String, [core.List$(core.int)], {allowMalformed: core.bool}]})
  });
  const UTF8 = dart.const(new Utf8Codec());
  const _fillBuffer = Symbol('_fillBuffer');
  const _writeSurrogate = Symbol('_writeSurrogate');
  class Utf8Encoder extends Converter$(core.String, core.List$(core.int)) {
    Utf8Encoder() {
      super.Converter();
    }
    convert(string, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      let stringLength = string[dartx.length];
      core.RangeError.checkValidRange(start, end, stringLength);
      if (end == null)
        end = stringLength;
      let length = dart.notNull(end) - dart.notNull(start);
      if (length == 0)
        return typed_data.Uint8List.new(0);
      let encoder = new _Utf8Encoder.withBufferSize(dart.notNull(length) * 3);
      let endPosition = encoder[_fillBuffer](string, start, end);
      dart.assert(dart.notNull(endPosition) >= dart.notNull(end) - 1);
      if (endPosition != end) {
        let lastCodeUnit = string[dartx.codeUnitAt](dart.notNull(end) - 1);
        dart.assert(_isLeadSurrogate(lastCodeUnit));
        let wasCombined = encoder[_writeSurrogate](lastCodeUnit, 0);
        dart.assert(!dart.notNull(wasCombined));
      }
      return encoder[_buffer][dartx.sublist](0, encoder[_bufferIndex]);
    }
    startChunkedConversion(sink) {
      if (!dart.is(sink, ByteConversionSink)) {
        sink = ByteConversionSink.from(sink);
      }
      return new _Utf8EncoderSink(dart.as(sink, ByteConversionSink));
    }
    bind(stream) {
      return super.bind(stream);
    }
  }
  dart.setSignature(Utf8Encoder, {
    constructors: () => ({Utf8Encoder: [Utf8Encoder, []]}),
    methods: () => ({
      convert: [core.List$(core.int), [core.String], [core.int, core.int]],
      startChunkedConversion: [StringConversionSink, [core.Sink$(core.List$(core.int))]],
      bind: [async.Stream$(core.List$(core.int)), [async.Stream$(core.String)]]
    })
  });
  class _Utf8Encoder extends core.Object {
    _Utf8Encoder() {
      this.withBufferSize(_Utf8Encoder._DEFAULT_BYTE_BUFFER_SIZE);
    }
    withBufferSize(bufferSize) {
      this[_buffer] = _Utf8Encoder._createBuffer(bufferSize);
      this[_carry] = 0;
      this[_bufferIndex] = 0;
    }
    static _createBuffer(size) {
      return typed_data.Uint8List.new(size);
    }
    [_writeSurrogate](leadingSurrogate, nextCodeUnit) {
      if (dart.notNull(_isTailSurrogate(nextCodeUnit))) {
        let rune = _combineSurrogatePair(leadingSurrogate, nextCodeUnit);
        dart.assert(dart.notNull(rune) > dart.notNull(_THREE_BYTE_LIMIT));
        dart.assert(dart.notNull(rune) <= dart.notNull(_FOUR_BYTE_LIMIT));
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 240 | dart.notNull(rune) >> 18);
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 128 | dart.notNull(rune) >> 12 & 63);
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 128 | dart.notNull(rune) >> 6 & 63);
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 128 | dart.notNull(rune) & 63);
        return true;
      } else {
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 224 | dart.notNull(leadingSurrogate) >> 12);
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 128 | dart.notNull(leadingSurrogate) >> 6 & 63);
        this[_buffer][dartx.set]((() => {
          let x = this[_bufferIndex];
          this[_bufferIndex] = dart.notNull(x) + 1;
          return x;
        }).bind(this)(), 128 | dart.notNull(leadingSurrogate) & 63);
        return false;
      }
    }
    [_fillBuffer](str, start, end) {
      if (start != end && dart.notNull(_isLeadSurrogate(str[dartx.codeUnitAt](dart.notNull(end) - 1)))) {
        end = dart.notNull(end) - 1;
      }
      let stringIndex = null;
      for (stringIndex = start; dart.notNull(stringIndex) < dart.notNull(end); stringIndex = dart.notNull(stringIndex) + 1) {
        let codeUnit = str[dartx.codeUnitAt](stringIndex);
        if (dart.notNull(codeUnit) <= dart.notNull(_ONE_BYTE_LIMIT)) {
          if (dart.notNull(this[_bufferIndex]) >= dart.notNull(this[_buffer][dartx.length]))
            break;
          this[_buffer][dartx.set]((() => {
            let x = this[_bufferIndex];
            this[_bufferIndex] = dart.notNull(x) + 1;
            return x;
          }).bind(this)(), codeUnit);
        } else if (dart.notNull(_isLeadSurrogate(codeUnit))) {
          if (dart.notNull(this[_bufferIndex]) + 3 >= dart.notNull(this[_buffer][dartx.length]))
            break;
          let nextCodeUnit = str[dartx.codeUnitAt](dart.notNull(stringIndex) + 1);
          let wasCombined = this[_writeSurrogate](codeUnit, nextCodeUnit);
          if (dart.notNull(wasCombined)) {
            stringIndex = dart.notNull(stringIndex) + 1;
          }
        } else {
          let rune = codeUnit;
          if (dart.notNull(rune) <= dart.notNull(_TWO_BYTE_LIMIT)) {
            if (dart.notNull(this[_bufferIndex]) + 1 >= dart.notNull(this[_buffer][dartx.length]))
              break;
            this[_buffer][dartx.set]((() => {
              let x = this[_bufferIndex];
              this[_bufferIndex] = dart.notNull(x) + 1;
              return x;
            }).bind(this)(), 192 | dart.notNull(rune) >> 6);
            this[_buffer][dartx.set]((() => {
              let x = this[_bufferIndex];
              this[_bufferIndex] = dart.notNull(x) + 1;
              return x;
            }).bind(this)(), 128 | dart.notNull(rune) & 63);
          } else {
            dart.assert(dart.notNull(rune) <= dart.notNull(_THREE_BYTE_LIMIT));
            if (dart.notNull(this[_bufferIndex]) + 2 >= dart.notNull(this[_buffer][dartx.length]))
              break;
            this[_buffer][dartx.set]((() => {
              let x = this[_bufferIndex];
              this[_bufferIndex] = dart.notNull(x) + 1;
              return x;
            }).bind(this)(), 224 | dart.notNull(rune) >> 12);
            this[_buffer][dartx.set]((() => {
              let x = this[_bufferIndex];
              this[_bufferIndex] = dart.notNull(x) + 1;
              return x;
            }).bind(this)(), 128 | dart.notNull(rune) >> 6 & 63);
            this[_buffer][dartx.set]((() => {
              let x = this[_bufferIndex];
              this[_bufferIndex] = dart.notNull(x) + 1;
              return x;
            }).bind(this)(), 128 | dart.notNull(rune) & 63);
          }
        }
      }
      return stringIndex;
    }
  }
  dart.defineNamedConstructor(_Utf8Encoder, 'withBufferSize');
  dart.setSignature(_Utf8Encoder, {
    constructors: () => ({
      _Utf8Encoder: [_Utf8Encoder, []],
      withBufferSize: [_Utf8Encoder, [core.int]]
    }),
    methods: () => ({
      [_writeSurrogate]: [core.bool, [core.int, core.int]],
      [_fillBuffer]: [core.int, [core.String, core.int, core.int]]
    }),
    statics: () => ({_createBuffer: [core.List$(core.int), [core.int]]}),
    names: ['_createBuffer']
  });
  _Utf8Encoder._DEFAULT_BYTE_BUFFER_SIZE = 1024;
  class _Utf8EncoderSink extends dart.mixin(_Utf8Encoder, StringConversionSinkMixin) {
    _Utf8EncoderSink(sink) {
      this[_sink] = sink;
      super._Utf8Encoder();
    }
    close() {
      if (this[_carry] != 0) {
        this.addSlice("", 0, 0, true);
        return;
      }
      this[_sink].close();
    }
    addSlice(str, start, end, isLast) {
      this[_bufferIndex] = 0;
      if (start == end && !dart.notNull(isLast)) {
        return;
      }
      if (this[_carry] != 0) {
        let nextCodeUnit = 0;
        if (start != end) {
          nextCodeUnit = str[dartx.codeUnitAt](start);
        } else {
          dart.assert(isLast);
        }
        let wasCombined = this[_writeSurrogate](this[_carry], nextCodeUnit);
        dart.assert(!dart.notNull(wasCombined) || start != end);
        if (dart.notNull(wasCombined)) {
          start = dart.notNull(start) + 1;
        }
        this[_carry] = 0;
      }
      do {
        start = this[_fillBuffer](str, start, end);
        let isLastSlice = dart.notNull(isLast) && start == end;
        if (start == dart.notNull(end) - 1 && dart.notNull(_isLeadSurrogate(str[dartx.codeUnitAt](start)))) {
          if (dart.notNull(isLast) && dart.notNull(this[_bufferIndex]) < dart.notNull(this[_buffer][dartx.length]) - 3) {
            let hasBeenCombined = this[_writeSurrogate](str[dartx.codeUnitAt](start), 0);
            dart.assert(!dart.notNull(hasBeenCombined));
          } else {
            this[_carry] = str[dartx.codeUnitAt](start);
          }
          start = dart.notNull(start) + 1;
        }
        this[_sink].addSlice(this[_buffer], 0, this[_bufferIndex], isLastSlice);
        this[_bufferIndex] = 0;
      } while (dart.notNull(start) < dart.notNull(end));
      if (dart.notNull(isLast))
        this.close();
    }
  }
  dart.setSignature(_Utf8EncoderSink, {
    constructors: () => ({_Utf8EncoderSink: [_Utf8EncoderSink, [ByteConversionSink]]}),
    methods: () => ({
      close: [dart.void, []],
      addSlice: [dart.void, [core.String, core.int, core.int, core.bool]]
    })
  });
  class Utf8Decoder extends Converter$(core.List$(core.int), core.String) {
    Utf8Decoder(opts) {
      let allowMalformed = opts && 'allowMalformed' in opts ? opts.allowMalformed : false;
      this[_allowMalformed] = allowMalformed;
      super.Converter();
    }
    convert(codeUnits, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      let length = codeUnits[dartx.length];
      core.RangeError.checkValidRange(start, end, length);
      if (end == null)
        end = length;
      let buffer = new core.StringBuffer();
      let decoder = new _Utf8Decoder(buffer, this[_allowMalformed]);
      decoder.convert(codeUnits, start, end);
      decoder.close();
      return dart.toString(buffer);
    }
    startChunkedConversion(sink) {
      let stringSink = null;
      if (dart.is(sink, StringConversionSink)) {
        stringSink = sink;
      } else {
        stringSink = StringConversionSink.from(sink);
      }
      return stringSink.asUtf8Sink(this[_allowMalformed]);
    }
    bind(stream) {
      return super.bind(stream);
    }
    fuse(next) {
      return super.fuse(next);
    }
  }
  dart.setSignature(Utf8Decoder, {
    constructors: () => ({Utf8Decoder: [Utf8Decoder, [], {allowMalformed: core.bool}]}),
    methods: () => ({
      convert: [core.String, [core.List$(core.int)], [core.int, core.int]],
      startChunkedConversion: [ByteConversionSink, [core.Sink$(core.String)]],
      bind: [async.Stream$(core.String), [async.Stream$(core.List$(core.int))]],
      fuse: [Converter$(core.List$(core.int), dart.dynamic), [Converter$(core.String, dart.dynamic)]]
    })
  });
  const _ONE_BYTE_LIMIT = 127;
  const _TWO_BYTE_LIMIT = 2047;
  const _THREE_BYTE_LIMIT = 65535;
  const _FOUR_BYTE_LIMIT = 1114111;
  const _SURROGATE_MASK = 63488;
  const _SURROGATE_TAG_MASK = 64512;
  const _SURROGATE_VALUE_MASK = 1023;
  const _LEAD_SURROGATE_MIN = 55296;
  const _TAIL_SURROGATE_MIN = 56320;
  function _isSurrogate(codeUnit) {
    return (dart.notNull(codeUnit) & dart.notNull(_SURROGATE_MASK)) == _LEAD_SURROGATE_MIN;
  }
  dart.fn(_isSurrogate, core.bool, [core.int]);
  function _isLeadSurrogate(codeUnit) {
    return (dart.notNull(codeUnit) & dart.notNull(_SURROGATE_TAG_MASK)) == _LEAD_SURROGATE_MIN;
  }
  dart.fn(_isLeadSurrogate, core.bool, [core.int]);
  function _isTailSurrogate(codeUnit) {
    return (dart.notNull(codeUnit) & dart.notNull(_SURROGATE_TAG_MASK)) == _TAIL_SURROGATE_MIN;
  }
  dart.fn(_isTailSurrogate, core.bool, [core.int]);
  function _combineSurrogatePair(lead, tail) {
    return 65536 + ((dart.notNull(lead) & dart.notNull(_SURROGATE_VALUE_MASK)) << 10) | dart.notNull(tail) & dart.notNull(_SURROGATE_VALUE_MASK);
  }
  dart.fn(_combineSurrogatePair, core.int, [core.int, core.int]);
  const _isFirstCharacter = Symbol('_isFirstCharacter');
  const _value = Symbol('_value');
  const _expectedUnits = Symbol('_expectedUnits');
  const _extraUnits = Symbol('_extraUnits');
  class _Utf8Decoder extends core.Object {
    _Utf8Decoder(stringSink, allowMalformed) {
      this[_stringSink] = stringSink;
      this[_allowMalformed] = allowMalformed;
      this[_isFirstCharacter] = true;
      this[_value] = 0;
      this[_expectedUnits] = 0;
      this[_extraUnits] = 0;
    }
    get hasPartialInput() {
      return dart.notNull(this[_expectedUnits]) > 0;
    }
    close() {
      this.flush();
    }
    flush() {
      if (dart.notNull(this.hasPartialInput)) {
        if (!dart.notNull(this[_allowMalformed])) {
          dart.throw(new core.FormatException("Unfinished UTF-8 octet sequence"));
        }
        this[_stringSink].writeCharCode(UNICODE_REPLACEMENT_CHARACTER_RUNE);
        this[_value] = 0;
        this[_expectedUnits] = 0;
        this[_extraUnits] = 0;
      }
    }
    convert(codeUnits, startIndex, endIndex) {
      let value = this[_value];
      let expectedUnits = this[_expectedUnits];
      let extraUnits = this[_extraUnits];
      this[_value] = 0;
      this[_expectedUnits] = 0;
      this[_extraUnits] = 0;
      function scanOneByteCharacters(units, from) {
        let to = endIndex;
        let mask = _ONE_BYTE_LIMIT;
        for (let i = from; dart.notNull(i) < dart.notNull(to); i = dart.notNull(i) + 1) {
          let unit = dart.dindex(units, i);
          if (!dart.equals(dart.dsend(unit, '&', mask), unit))
            return dart.notNull(i) - dart.notNull(from);
        }
        return dart.notNull(to) - dart.notNull(from);
      }
      dart.fn(scanOneByteCharacters, core.int, [dart.dynamic, core.int]);
      const addSingleBytes = (function(from, to) {
        dart.assert(dart.notNull(from) >= dart.notNull(startIndex) && dart.notNull(from) <= dart.notNull(endIndex));
        dart.assert(dart.notNull(to) >= dart.notNull(startIndex) && dart.notNull(to) <= dart.notNull(endIndex));
        this[_stringSink].write(core.String.fromCharCodes(codeUnits, from, to));
      }).bind(this);
      dart.fn(addSingleBytes, dart.void, [core.int, core.int]);
      let i = startIndex;
      loop:
        while (true) {
          multibyte:
            if (dart.notNull(expectedUnits) > 0) {
              do {
                if (i == endIndex) {
                  break loop;
                }
                let unit = codeUnits[dartx.get](i);
                if ((dart.notNull(unit) & 192) != 128) {
                  expectedUnits = 0;
                  if (!dart.notNull(this[_allowMalformed])) {
                    dart.throw(new core.FormatException(`Bad UTF-8 encoding 0x${unit[dartx.toRadixString](16)}`));
                  }
                  this[_isFirstCharacter] = false;
                  this[_stringSink].writeCharCode(UNICODE_REPLACEMENT_CHARACTER_RUNE);
                  break multibyte;
                } else {
                  value = dart.notNull(value) << 6 | dart.notNull(unit) & 63;
                  expectedUnits = dart.notNull(expectedUnits) - 1;
                  i = dart.notNull(i) + 1;
                }
              } while (dart.notNull(expectedUnits) > 0);
              if (dart.notNull(value) <= dart.notNull(_Utf8Decoder._LIMITS[dartx.get](dart.notNull(extraUnits) - 1))) {
                if (!dart.notNull(this[_allowMalformed])) {
                  dart.throw(new core.FormatException(`Overlong encoding of 0x${value[dartx.toRadixString](16)}`));
                }
                expectedUnits = extraUnits = 0;
                value = UNICODE_REPLACEMENT_CHARACTER_RUNE;
              }
              if (dart.notNull(value) > dart.notNull(_FOUR_BYTE_LIMIT)) {
                if (!dart.notNull(this[_allowMalformed])) {
                  dart.throw(new core.FormatException("Character outside valid Unicode range: " + `0x${value[dartx.toRadixString](16)}`));
                }
                value = UNICODE_REPLACEMENT_CHARACTER_RUNE;
              }
              if (!dart.notNull(this[_isFirstCharacter]) || value != UNICODE_BOM_CHARACTER_RUNE) {
                this[_stringSink].writeCharCode(value);
              }
              this[_isFirstCharacter] = false;
            }
          while (dart.notNull(i) < dart.notNull(endIndex)) {
            let oneBytes = scanOneByteCharacters(codeUnits, i);
            if (dart.notNull(oneBytes) > 0) {
              this[_isFirstCharacter] = false;
              addSingleBytes(i, dart.notNull(i) + dart.notNull(oneBytes));
              i = dart.notNull(i) + dart.notNull(oneBytes);
              if (i == endIndex)
                break;
            }
            let unit = codeUnits[dartx.get]((() => {
              let x = i;
              i = dart.notNull(x) + 1;
              return x;
            })());
            if (dart.notNull(unit) < 0) {
              if (!dart.notNull(this[_allowMalformed])) {
                dart.throw(new core.FormatException(`Negative UTF-8 code unit: -0x${(-dart.notNull(unit))[dartx.toRadixString](16)}`));
              }
              this[_stringSink].writeCharCode(UNICODE_REPLACEMENT_CHARACTER_RUNE);
            } else {
              dart.assert(dart.notNull(unit) > dart.notNull(_ONE_BYTE_LIMIT));
              if ((dart.notNull(unit) & 224) == 192) {
                value = dart.notNull(unit) & 31;
                expectedUnits = extraUnits = 1;
                continue loop;
              }
              if ((dart.notNull(unit) & 240) == 224) {
                value = dart.notNull(unit) & 15;
                expectedUnits = extraUnits = 2;
                continue loop;
              }
              if ((dart.notNull(unit) & 248) == 240 && dart.notNull(unit) < 245) {
                value = dart.notNull(unit) & 7;
                expectedUnits = extraUnits = 3;
                continue loop;
              }
              if (!dart.notNull(this[_allowMalformed])) {
                dart.throw(new core.FormatException(`Bad UTF-8 encoding 0x${unit[dartx.toRadixString](16)}`));
              }
              value = UNICODE_REPLACEMENT_CHARACTER_RUNE;
              expectedUnits = extraUnits = 0;
              this[_isFirstCharacter] = false;
              this[_stringSink].writeCharCode(value);
            }
          }
          break loop;
        }
      if (dart.notNull(expectedUnits) > 0) {
        this[_value] = value;
        this[_expectedUnits] = expectedUnits;
        this[_extraUnits] = extraUnits;
      }
    }
  }
  dart.setSignature(_Utf8Decoder, {
    constructors: () => ({_Utf8Decoder: [_Utf8Decoder, [core.StringSink, core.bool]]}),
    methods: () => ({
      close: [dart.void, []],
      flush: [dart.void, []],
      convert: [dart.void, [core.List$(core.int), core.int, core.int]]
    })
  });
  _Utf8Decoder._LIMITS = dart.const(dart.list([_ONE_BYTE_LIMIT, _TWO_BYTE_LIMIT, _THREE_BYTE_LIMIT, _FOUR_BYTE_LIMIT], core.int));
  const _processed = Symbol('_processed');
  const _computeKeys = Symbol('_computeKeys');
  const _original = Symbol('_original');
  function _convertJsonToDart(json, reviver) {
    dart.assert(reviver != null);
    function walk(e) {
      if (e == null || typeof e != "object") {
        return e;
      }
      if (Object.getPrototypeOf(e) === Array.prototype) {
        for (let i = 0; dart.notNull(i) < e.length; i = dart.notNull(i) + 1) {
          let item = e[i];
          e[i] = dart.dcall(reviver, i, walk(item));
        }
        return e;
      }
      let map = new _JsonMap(e);
      let processed = map[_processed];
      let keys = map[_computeKeys]();
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        let key = keys[dartx.get](i);
        let revived = dart.dcall(reviver, key, walk(e[key]));
        processed[key] = revived;
      }
      map[_original] = processed;
      return map;
    }
    dart.fn(walk);
    return dart.dcall(reviver, null, walk(json));
  }
  dart.fn(_convertJsonToDart, dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]);
  function _convertJsonToDartLazy(object) {
    if (object == null)
      return null;
    if (typeof object != "object") {
      return object;
    }
    if (Object.getPrototypeOf(object) !== Array.prototype) {
      return new _JsonMap(object);
    }
    for (let i = 0; dart.notNull(i) < object.length; i = dart.notNull(i) + 1) {
      let item = object[i];
      object[i] = _convertJsonToDartLazy(item);
    }
    return object;
  }
  dart.fn(_convertJsonToDartLazy);
  const _data = Symbol('_data');
  const _isUpgraded = Symbol('_isUpgraded');
  const _upgradedMap = Symbol('_upgradedMap');
  const _process = Symbol('_process');
  const _upgrade = Symbol('_upgrade');
  class _JsonMap extends core.Object {
    _JsonMap(original) {
      this[_processed] = _JsonMap._newJavaScriptObject();
      this[_original] = original;
      this[_data] = null;
    }
    get(key) {
      if (dart.notNull(this[_isUpgraded])) {
        return this[_upgradedMap].get(key);
      } else if (!(typeof key == 'string')) {
        return null;
      } else {
        let result = _JsonMap._getProperty(this[_processed], dart.as(key, core.String));
        if (dart.notNull(_JsonMap._isUnprocessed(result)))
          result = this[_process](dart.as(key, core.String));
        return result;
      }
    }
    get length() {
      return dart.notNull(this[_isUpgraded]) ? this[_upgradedMap].length : this[_computeKeys]()[dartx.length];
    }
    get isEmpty() {
      return this.length == 0;
    }
    get isNotEmpty() {
      return dart.notNull(this.length) > 0;
    }
    get keys() {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap].keys;
      return new _JsonMapKeyIterable(this);
    }
    get values() {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap].values;
      return _internal.MappedIterable.new(this[_computeKeys](), dart.fn((each => this.get(each)).bind(this)));
    }
    set(key, value) {
      if (dart.notNull(this[_isUpgraded])) {
        this[_upgradedMap].set(key, value);
      } else if (dart.notNull(this.containsKey(key))) {
        let processed = this[_processed];
        _JsonMap._setProperty(processed, dart.as(key, core.String), value);
        let original = this[_original];
        if (!dart.notNull(core.identical(original, processed))) {
          _JsonMap._setProperty(original, dart.as(key, core.String), null);
        }
      } else {
        this[_upgrade]().set(key, value);
      }
      return value;
    }
    addAll(other) {
      other.forEach(dart.fn(((key, value) => {
        this.set(key, value);
      }).bind(this)));
    }
    containsValue(value) {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap].containsValue(value);
      let keys = this[_computeKeys]();
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        let key = keys[dartx.get](i);
        if (dart.equals(this.get(key), value))
          return true;
      }
      return false;
    }
    containsKey(key) {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap].containsKey(key);
      if (!(typeof key == 'string'))
        return false;
      return _JsonMap._hasProperty(this[_original], dart.as(key, core.String));
    }
    putIfAbsent(key, ifAbsent) {
      if (dart.notNull(this.containsKey(key)))
        return this.get(key);
      let value = ifAbsent();
      this.set(key, value);
      return value;
    }
    remove(key) {
      if (!dart.notNull(this[_isUpgraded]) && !dart.notNull(this.containsKey(key)))
        return null;
      return this[_upgrade]().remove(key);
    }
    clear() {
      if (dart.notNull(this[_isUpgraded])) {
        this[_upgradedMap].clear();
      } else {
        if (this[_data] != null) {
          dart.dsend(this[_data], 'clear');
        }
        this[_original] = this[_processed] = null;
        this[_data] = dart.map();
      }
    }
    forEach(f) {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap].forEach(f);
      let keys = this[_computeKeys]();
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        let key = keys[dartx.get](i);
        let value = _JsonMap._getProperty(this[_processed], key);
        if (dart.notNull(_JsonMap._isUnprocessed(value))) {
          value = _convertJsonToDartLazy(_JsonMap._getProperty(this[_original], key));
          _JsonMap._setProperty(this[_processed], key, value);
        }
        dart.dcall(f, key, value);
        if (!dart.notNull(core.identical(keys, this[_data]))) {
          dart.throw(new core.ConcurrentModificationError(this));
        }
      }
    }
    toString() {
      return collection.Maps.mapToString(this);
    }
    get [_isUpgraded]() {
      return this[_processed] == null;
    }
    get [_upgradedMap]() {
      dart.assert(this[_isUpgraded]);
      return dart.as(this[_data], core.Map);
    }
    [_computeKeys]() {
      dart.assert(!dart.notNull(this[_isUpgraded]));
      let keys = dart.as(this[_data], core.List);
      if (keys == null) {
        keys = this[_data] = _JsonMap._getPropertyNames(this[_original]);
      }
      return dart.as(keys, core.List$(core.String));
    }
    [_upgrade]() {
      if (dart.notNull(this[_isUpgraded]))
        return this[_upgradedMap];
      let result = dart.map();
      let keys = this[_computeKeys]();
      for (let i = 0; dart.notNull(i) < dart.notNull(keys[dartx.length]); i = dart.notNull(i) + 1) {
        let key = keys[dartx.get](i);
        result.set(key, this.get(key));
      }
      if (dart.notNull(keys[dartx.isEmpty])) {
        keys[dartx.add](null);
      } else {
        keys[dartx.clear]();
      }
      this[_original] = this[_processed] = null;
      this[_data] = result;
      dart.assert(this[_isUpgraded]);
      return result;
    }
    [_process](key) {
      if (!dart.notNull(_JsonMap._hasProperty(this[_original], key)))
        return null;
      let result = _convertJsonToDartLazy(_JsonMap._getProperty(this[_original], key));
      return _JsonMap._setProperty(this[_processed], key, result);
    }
    static _hasProperty(object, key) {
      return Object.prototype.hasOwnProperty.call(object, key);
    }
    static _getProperty(object, key) {
      return object[key];
    }
    static _setProperty(object, key, value) {
      return object[key] = value;
    }
    static _getPropertyNames(object) {
      return dart.as(Object.keys(object), core.List);
    }
    static _isUnprocessed(object) {
      return typeof object == "undefined";
    }
    static _newJavaScriptObject() {
      return Object.create(null);
    }
  }
  _JsonMap[dart.implements] = () => [collection.LinkedHashMap];
  dart.setSignature(_JsonMap, {
    constructors: () => ({_JsonMap: [_JsonMap, [dart.dynamic]]}),
    methods: () => ({
      get: [dart.dynamic, [core.Object]],
      set: [dart.void, [dart.dynamic, dart.dynamic]],
      addAll: [dart.void, [core.Map]],
      containsValue: [core.bool, [core.Object]],
      containsKey: [core.bool, [core.Object]],
      putIfAbsent: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [])]],
      remove: [dart.dynamic, [core.Object]],
      clear: [dart.void, []],
      forEach: [dart.void, [dart.functionType(dart.void, [dart.dynamic, dart.dynamic])]],
      [_computeKeys]: [core.List$(core.String), []],
      [_upgrade]: [core.Map, []],
      [_process]: [dart.dynamic, [core.String]]
    }),
    statics: () => ({
      _hasProperty: [core.bool, [dart.dynamic, core.String]],
      _getProperty: [dart.dynamic, [dart.dynamic, core.String]],
      _setProperty: [dart.dynamic, [dart.dynamic, core.String, dart.dynamic]],
      _getPropertyNames: [core.List, [dart.dynamic]],
      _isUnprocessed: [core.bool, [dart.dynamic]],
      _newJavaScriptObject: [dart.dynamic, []]
    }),
    names: ['_hasProperty', '_getProperty', '_setProperty', '_getPropertyNames', '_isUnprocessed', '_newJavaScriptObject']
  });
  const _parent = Symbol('_parent');
  class _JsonMapKeyIterable extends _internal.ListIterable {
    _JsonMapKeyIterable(parent) {
      this[_parent] = parent;
      super.ListIterable();
    }
    get length() {
      return this[_parent].length;
    }
    elementAt(index) {
      return dart.notNull(this[_parent][_isUpgraded]) ? dart.as(this[_parent].keys[dartx.elementAt](index), core.String) : this[_parent][_computeKeys]()[dartx.get](index);
    }
    get iterator() {
      return dart.notNull(this[_parent][_isUpgraded]) ? this[_parent].keys[dartx.iterator] : this[_parent][_computeKeys]()[dartx.iterator];
    }
    contains(key) {
      return this[_parent].containsKey(key);
    }
  }
  dart.setSignature(_JsonMapKeyIterable, {
    constructors: () => ({_JsonMapKeyIterable: [_JsonMapKeyIterable, [_JsonMap]]}),
    methods: () => ({elementAt: [core.String, [core.int]]})
  });
  dart.defineExtensionMembers(_JsonMapKeyIterable, ['elementAt', 'contains', 'length', 'iterator']);
  class _JsonDecoderSink extends _StringSinkConversionSink {
    _JsonDecoderSink(reviver, sink) {
      this[_reviver] = reviver;
      this[_sink] = sink;
      super._StringSinkConversionSink(new core.StringBuffer());
    }
    close() {
      super.close();
      let buffer = dart.as(this[_stringSink], core.StringBuffer);
      let accumulated = dart.toString(buffer);
      buffer.clear();
      let decoded = _parseJson(accumulated, this[_reviver]);
      this[_sink].add(decoded);
      this[_sink].close();
    }
  }
  dart.setSignature(_JsonDecoderSink, {
    constructors: () => ({_JsonDecoderSink: [_JsonDecoderSink, [_Reviver, core.Sink$(core.Object)]]})
  });
  // Exports:
  exports.Codec$ = Codec$;
  exports.Codec = Codec;
  exports.Encoding = Encoding;
  exports.AsciiCodec = AsciiCodec;
  exports.ASCII = ASCII;
  exports.Converter$ = Converter$;
  exports.Converter = Converter;
  exports.AsciiEncoder = AsciiEncoder;
  exports.StringConversionSinkMixin = StringConversionSinkMixin;
  exports.StringConversionSinkBase = StringConversionSinkBase;
  exports.AsciiDecoder = AsciiDecoder;
  exports.ChunkedConversionSink$ = ChunkedConversionSink$;
  exports.ChunkedConversionSink = ChunkedConversionSink;
  exports.ByteConversionSink = ByteConversionSink;
  exports.ByteConversionSinkBase = ByteConversionSinkBase;
  exports.HtmlEscapeMode = HtmlEscapeMode;
  exports.HtmlEscape = HtmlEscape;
  exports.HTML_ESCAPE = HTML_ESCAPE;
  exports.JsonUnsupportedObjectError = JsonUnsupportedObjectError;
  exports.JsonCyclicError = JsonCyclicError;
  exports.JsonCodec = JsonCodec;
  exports.JSON = JSON;
  exports.JsonEncoder = JsonEncoder;
  exports.JsonUtf8Encoder = JsonUtf8Encoder;
  exports.JsonDecoder = JsonDecoder;
  exports.Latin1Codec = Latin1Codec;
  exports.LATIN1 = LATIN1;
  exports.Latin1Encoder = Latin1Encoder;
  exports.Latin1Decoder = Latin1Decoder;
  exports.LineSplitter = LineSplitter;
  exports.StringConversionSink = StringConversionSink;
  exports.ClosableStringSink = ClosableStringSink;
  exports.UNICODE_REPLACEMENT_CHARACTER_RUNE = UNICODE_REPLACEMENT_CHARACTER_RUNE;
  exports.UNICODE_BOM_CHARACTER_RUNE = UNICODE_BOM_CHARACTER_RUNE;
  exports.Utf8Codec = Utf8Codec;
  exports.UTF8 = UTF8;
  exports.Utf8Encoder = Utf8Encoder;
  exports.Utf8Decoder = Utf8Decoder;
});
dart_library.library('dart/core', null, /* Imports */[
  "dart/_runtime"
], /* Lazy imports */[
  'dart/_js_helper',
  'dart/_internal',
  'dart/collection',
  'dart/_interceptors',
  'dart/convert'
], function(exports, dart, _js_helper, _internal, collection, _interceptors, convert) {
  'use strict';
  let dartx = dart.dartx;
  class Object {
    constructor() {
      let name = this.constructor.name;
      let result = void 0;
      if (name in this)
        result = this[name](...arguments);
      return result === void 0 ? this : result;
    }
    ['=='](other) {
      return identical(this, other);
    }
    get hashCode() {
      return _js_helper.Primitives.objectHashCode(this);
    }
    toString() {
      return _js_helper.Primitives.objectToString(this);
    }
    noSuchMethod(invocation) {
      dart.throw(new NoSuchMethodError(this, invocation.memberName, invocation.positionalArguments, invocation.namedArguments));
    }
    get runtimeType() {
      return dart.realRuntimeType(this);
    }
  }
  dart.setSignature(Object, {
    constructors: () => ({Object: [Object, []]}),
    methods: () => ({
      '==': [bool, [dart.dynamic]],
      toString: [String, []],
      noSuchMethod: [dart.dynamic, [Invocation]]
    })
  });
  class Deprecated extends Object {
    Deprecated(expires) {
      this.expires = expires;
    }
    toString() {
      return `Deprecated feature. Will be removed ${this.expires}`;
    }
  }
  dart.setSignature(Deprecated, {
    constructors: () => ({Deprecated: [Deprecated, [String]]})
  });
  class _Override extends Object {
    _Override() {
    }
  }
  dart.setSignature(_Override, {
    constructors: () => ({_Override: [_Override, []]})
  });
  const deprecated = dart.const(new Deprecated("next release"));
  const override = dart.const(new _Override());
  class _Proxy extends Object {
    _Proxy() {
    }
  }
  dart.setSignature(_Proxy, {
    constructors: () => ({_Proxy: [_Proxy, []]})
  });
  const proxy = dart.const(new _Proxy());
  dart.defineExtensionNames([
    'toString'
  ]);
  class bool extends Object {
    static fromEnvironment(name, opts) {
      let defaultValue = opts && 'defaultValue' in opts ? opts.defaultValue : false;
      dart.throw(new UnsupportedError('bool.fromEnvironment can only be used as a const constructor'));
    }
    toString() {
      return this ? "true" : "false";
    }
  }
  dart.setSignature(bool, {
    constructors: () => ({fromEnvironment: [bool, [String], {defaultValue: bool}]})
  });
  const Comparator$ = dart.generic(function(T) {
    const Comparator = dart.typedef('Comparator', () => dart.functionType(int, [T, T]));
    return Comparator;
  });
  let Comparator = Comparator$();
  const Comparable$ = dart.generic(function(T) {
    class Comparable extends Object {
      static compare(a, b) {
        return a[dartx.compareTo](b);
      }
    }
    dart.setSignature(Comparable, {
      statics: () => ({compare: [int, [Comparable$(), Comparable$()]]}),
      names: ['compare']
    });
    return Comparable;
  });
  let Comparable = Comparable$();
  class DateTime extends Object {
    DateTime(year, month, day, hour, minute, second, millisecond) {
      if (month === void 0)
        month = 1;
      if (day === void 0)
        day = 1;
      if (hour === void 0)
        hour = 0;
      if (minute === void 0)
        minute = 0;
      if (second === void 0)
        second = 0;
      if (millisecond === void 0)
        millisecond = 0;
      this._internal(year, month, day, hour, minute, second, millisecond, false);
    }
    utc(year, month, day, hour, minute, second, millisecond) {
      if (month === void 0)
        month = 1;
      if (day === void 0)
        day = 1;
      if (hour === void 0)
        hour = 0;
      if (minute === void 0)
        minute = 0;
      if (second === void 0)
        second = 0;
      if (millisecond === void 0)
        millisecond = 0;
      this._internal(year, month, day, hour, minute, second, millisecond, true);
    }
    now() {
      this._now();
    }
    static parse(formattedString) {
      let re = RegExp.new('^([+-]?\\d{4,6})-?(\\d\\d)-?(\\d\\d)' + '(?:[ T](\\d\\d)(?::?(\\d\\d)(?::?(\\d\\d)(.\\d{1,6})?)?)?' + '( ?[zZ]| ?([-+])(\\d\\d)(?::?(\\d\\d))?)?)?$');
      let match = re.firstMatch(formattedString);
      if (match != null) {
        function parseIntOrZero(matched) {
          if (matched == null)
            return 0;
          return int.parse(matched);
        }
        dart.fn(parseIntOrZero, int, [String]);
        function parseDoubleOrZero(matched) {
          if (matched == null)
            return 0.0;
          return double.parse(matched);
        }
        dart.fn(parseDoubleOrZero, double, [String]);
        let years = int.parse(match.get(1));
        let month = int.parse(match.get(2));
        let day = int.parse(match.get(3));
        let hour = parseIntOrZero(match.get(4));
        let minute = parseIntOrZero(match.get(5));
        let second = parseIntOrZero(match.get(6));
        let addOneMillisecond = false;
        let millisecond = (dart.notNull(parseDoubleOrZero(match.get(7))) * 1000)[dartx.round]();
        if (millisecond == 1000) {
          addOneMillisecond = true;
          millisecond = 999;
        }
        let isUtc = false;
        if (match.get(8) != null) {
          isUtc = true;
          if (match.get(9) != null) {
            let sign = match.get(9) == '-' ? -1 : 1;
            let hourDifference = int.parse(match.get(10));
            let minuteDifference = parseIntOrZero(match.get(11));
            minuteDifference = dart.notNull(minuteDifference) + 60 * dart.notNull(hourDifference);
            minute = dart.notNull(minute) - dart.notNull(sign) * dart.notNull(minuteDifference);
          }
        }
        let millisecondsSinceEpoch = DateTime._brokenDownDateToMillisecondsSinceEpoch(years, month, day, hour, minute, second, millisecond, isUtc);
        if (millisecondsSinceEpoch == null) {
          dart.throw(new FormatException("Time out of range", formattedString));
        }
        if (dart.notNull(addOneMillisecond)) {
          millisecondsSinceEpoch = dart.notNull(millisecondsSinceEpoch) + 1;
        }
        return new DateTime.fromMillisecondsSinceEpoch(millisecondsSinceEpoch, {isUtc: isUtc});
      } else {
        dart.throw(new FormatException("Invalid date format", formattedString));
      }
    }
    fromMillisecondsSinceEpoch(millisecondsSinceEpoch, opts) {
      let isUtc = opts && 'isUtc' in opts ? opts.isUtc : false;
      this.millisecondsSinceEpoch = millisecondsSinceEpoch;
      this.isUtc = isUtc;
      if (dart.notNull(millisecondsSinceEpoch[dartx.abs]()) > dart.notNull(DateTime._MAX_MILLISECONDS_SINCE_EPOCH)) {
        dart.throw(new ArgumentError(millisecondsSinceEpoch));
      }
      if (isUtc == null)
        dart.throw(new ArgumentError(isUtc));
    }
    ['=='](other) {
      if (!dart.is(other, DateTime))
        return false;
      return dart.equals(this.millisecondsSinceEpoch, dart.dload(other, 'millisecondsSinceEpoch')) && dart.equals(this.isUtc, dart.dload(other, 'isUtc'));
    }
    isBefore(other) {
      return dart.notNull(this.millisecondsSinceEpoch) < dart.notNull(other.millisecondsSinceEpoch);
    }
    isAfter(other) {
      return dart.notNull(this.millisecondsSinceEpoch) > dart.notNull(other.millisecondsSinceEpoch);
    }
    isAtSameMomentAs(other) {
      return this.millisecondsSinceEpoch == other.millisecondsSinceEpoch;
    }
    compareTo(other) {
      return this.millisecondsSinceEpoch[dartx.compareTo](other.millisecondsSinceEpoch);
    }
    get hashCode() {
      return this.millisecondsSinceEpoch;
    }
    toLocal() {
      if (dart.notNull(this.isUtc)) {
        return new DateTime.fromMillisecondsSinceEpoch(this.millisecondsSinceEpoch, {isUtc: false});
      }
      return this;
    }
    toUtc() {
      if (dart.notNull(this.isUtc))
        return this;
      return new DateTime.fromMillisecondsSinceEpoch(this.millisecondsSinceEpoch, {isUtc: true});
    }
    static _fourDigits(n) {
      let absN = n[dartx.abs]();
      let sign = dart.notNull(n) < 0 ? "-" : "";
      if (dart.notNull(absN) >= 1000)
        return `${n}`;
      if (dart.notNull(absN) >= 100)
        return `${sign}0${absN}`;
      if (dart.notNull(absN) >= 10)
        return `${sign}00${absN}`;
      return `${sign}000${absN}`;
    }
    static _sixDigits(n) {
      dart.assert(dart.notNull(n) < -9999 || dart.notNull(n) > 9999);
      let absN = n[dartx.abs]();
      let sign = dart.notNull(n) < 0 ? "-" : "+";
      if (dart.notNull(absN) >= 100000)
        return `${sign}${absN}`;
      return `${sign}0${absN}`;
    }
    static _threeDigits(n) {
      if (dart.notNull(n) >= 100)
        return `${n}`;
      if (dart.notNull(n) >= 10)
        return `0${n}`;
      return `00${n}`;
    }
    static _twoDigits(n) {
      if (dart.notNull(n) >= 10)
        return `${n}`;
      return `0${n}`;
    }
    toString() {
      let y = DateTime._fourDigits(this.year);
      let m = DateTime._twoDigits(this.month);
      let d = DateTime._twoDigits(this.day);
      let h = DateTime._twoDigits(this.hour);
      let min = DateTime._twoDigits(this.minute);
      let sec = DateTime._twoDigits(this.second);
      let ms = DateTime._threeDigits(this.millisecond);
      if (dart.notNull(this.isUtc)) {
        return `${y}-${m}-${d} ${h}:${min}:${sec}.${ms}Z`;
      } else {
        return `${y}-${m}-${d} ${h}:${min}:${sec}.${ms}`;
      }
    }
    toIso8601String() {
      let y = dart.notNull(this.year) >= -9999 && dart.notNull(this.year) <= 9999 ? DateTime._fourDigits(this.year) : DateTime._sixDigits(this.year);
      let m = DateTime._twoDigits(this.month);
      let d = DateTime._twoDigits(this.day);
      let h = DateTime._twoDigits(this.hour);
      let min = DateTime._twoDigits(this.minute);
      let sec = DateTime._twoDigits(this.second);
      let ms = DateTime._threeDigits(this.millisecond);
      if (dart.notNull(this.isUtc)) {
        return `${y}-${m}-${d}T${h}:${min}:${sec}.${ms}Z`;
      } else {
        return `${y}-${m}-${d}T${h}:${min}:${sec}.${ms}`;
      }
    }
    add(duration) {
      let ms = this.millisecondsSinceEpoch;
      return new DateTime.fromMillisecondsSinceEpoch(dart.notNull(ms) + dart.notNull(duration.inMilliseconds), {isUtc: this.isUtc});
    }
    subtract(duration) {
      let ms = this.millisecondsSinceEpoch;
      return new DateTime.fromMillisecondsSinceEpoch(dart.notNull(ms) - dart.notNull(duration.inMilliseconds), {isUtc: this.isUtc});
    }
    difference(other) {
      let ms = this.millisecondsSinceEpoch;
      let otherMs = other.millisecondsSinceEpoch;
      return new Duration({milliseconds: dart.notNull(ms) - dart.notNull(otherMs)});
    }
    _internal(year, month, day, hour, minute, second, millisecond, isUtc) {
      this.isUtc = typeof isUtc == 'boolean' ? isUtc : dart.throw(new ArgumentError(isUtc));
      this.millisecondsSinceEpoch = dart.as(_js_helper.checkInt(_js_helper.Primitives.valueFromDecomposedDate(year, month, day, hour, minute, second, millisecond, isUtc)), int);
    }
    _now() {
      this.isUtc = false;
      this.millisecondsSinceEpoch = _js_helper.Primitives.dateNow();
    }
    static _brokenDownDateToMillisecondsSinceEpoch(year, month, day, hour, minute, second, millisecond, isUtc) {
      return dart.as(_js_helper.Primitives.valueFromDecomposedDate(year, month, day, hour, minute, second, millisecond, isUtc), int);
    }
    get timeZoneName() {
      if (dart.notNull(this.isUtc))
        return "UTC";
      return _js_helper.Primitives.getTimeZoneName(this);
    }
    get timeZoneOffset() {
      if (dart.notNull(this.isUtc))
        return new Duration();
      return new Duration({minutes: _js_helper.Primitives.getTimeZoneOffsetInMinutes(this)});
    }
    get year() {
      return dart.as(_js_helper.Primitives.getYear(this), int);
    }
    get month() {
      return dart.as(_js_helper.Primitives.getMonth(this), int);
    }
    get day() {
      return dart.as(_js_helper.Primitives.getDay(this), int);
    }
    get hour() {
      return dart.as(_js_helper.Primitives.getHours(this), int);
    }
    get minute() {
      return dart.as(_js_helper.Primitives.getMinutes(this), int);
    }
    get second() {
      return dart.as(_js_helper.Primitives.getSeconds(this), int);
    }
    get millisecond() {
      return dart.as(_js_helper.Primitives.getMilliseconds(this), int);
    }
    get weekday() {
      return dart.as(_js_helper.Primitives.getWeekday(this), int);
    }
  }
  DateTime[dart.implements] = () => [Comparable];
  dart.defineNamedConstructor(DateTime, 'utc');
  dart.defineNamedConstructor(DateTime, 'now');
  dart.defineNamedConstructor(DateTime, 'fromMillisecondsSinceEpoch');
  dart.defineNamedConstructor(DateTime, '_internal');
  dart.defineNamedConstructor(DateTime, '_now');
  dart.setSignature(DateTime, {
    constructors: () => ({
      DateTime: [DateTime, [int], [int, int, int, int, int, int]],
      utc: [DateTime, [int], [int, int, int, int, int, int]],
      now: [DateTime, []],
      fromMillisecondsSinceEpoch: [DateTime, [int], {isUtc: bool}],
      _internal: [DateTime, [int, int, int, int, int, int, int, bool]],
      _now: [DateTime, []]
    }),
    methods: () => ({
      isBefore: [bool, [DateTime]],
      isAfter: [bool, [DateTime]],
      isAtSameMomentAs: [bool, [DateTime]],
      compareTo: [int, [DateTime]],
      toLocal: [DateTime, []],
      toUtc: [DateTime, []],
      toIso8601String: [String, []],
      add: [DateTime, [Duration]],
      subtract: [DateTime, [Duration]],
      difference: [Duration, [DateTime]]
    }),
    statics: () => ({
      parse: [DateTime, [String]],
      _fourDigits: [String, [int]],
      _sixDigits: [String, [int]],
      _threeDigits: [String, [int]],
      _twoDigits: [String, [int]],
      _brokenDownDateToMillisecondsSinceEpoch: [int, [int, int, int, int, int, int, int, bool]]
    }),
    names: ['parse', '_fourDigits', '_sixDigits', '_threeDigits', '_twoDigits', '_brokenDownDateToMillisecondsSinceEpoch']
  });
  dart.defineExtensionMembers(DateTime, ['compareTo']);
  DateTime.MONDAY = 1;
  DateTime.TUESDAY = 2;
  DateTime.WEDNESDAY = 3;
  DateTime.THURSDAY = 4;
  DateTime.FRIDAY = 5;
  DateTime.SATURDAY = 6;
  DateTime.SUNDAY = 7;
  DateTime.DAYS_PER_WEEK = 7;
  DateTime.JANUARY = 1;
  DateTime.FEBRUARY = 2;
  DateTime.MARCH = 3;
  DateTime.APRIL = 4;
  DateTime.MAY = 5;
  DateTime.JUNE = 6;
  DateTime.JULY = 7;
  DateTime.AUGUST = 8;
  DateTime.SEPTEMBER = 9;
  DateTime.OCTOBER = 10;
  DateTime.NOVEMBER = 11;
  DateTime.DECEMBER = 12;
  DateTime.MONTHS_PER_YEAR = 12;
  DateTime._MAX_MILLISECONDS_SINCE_EPOCH = 8640000000000000;
  class num extends Object {
    static parse(input, onError) {
      if (onError === void 0)
        onError = null;
      let source = input[dartx.trim]();
      num._parseError = false;
      let result = int.parse(source, {onError: num._onParseErrorInt});
      if (!dart.notNull(num._parseError))
        return result;
      num._parseError = false;
      result = double.parse(source, num._onParseErrorDouble);
      if (!dart.notNull(num._parseError))
        return result;
      if (onError == null)
        dart.throw(new FormatException(input));
      return onError(input);
    }
    static _onParseErrorInt(_) {
      num._parseError = true;
      return 0;
    }
    static _onParseErrorDouble(_) {
      num._parseError = true;
      return 0.0;
    }
  }
  num[dart.implements] = () => [Comparable$(num)];
  dart.setSignature(num, {
    statics: () => ({
      parse: [num, [String], [dart.functionType(num, [String])]],
      _onParseErrorInt: [int, [String]],
      _onParseErrorDouble: [double, [String]]
    }),
    names: ['parse', '_onParseErrorInt', '_onParseErrorDouble']
  });
  class double extends num {
    static parse(source, onError) {
      if (onError === void 0)
        onError = null;
      return _js_helper.Primitives.parseDouble(source, onError);
    }
  }
  dart.setSignature(double, {
    statics: () => ({parse: [double, [String], [dart.functionType(double, [String])]]}),
    names: ['parse']
  });
  double.NAN = 0.0 / 0.0;
  double.INFINITY = 1.0 / 0.0;
  double.NEGATIVE_INFINITY = -dart.notNull(double.INFINITY);
  double.MIN_POSITIVE = 5e-324;
  double.MAX_FINITE = 1.7976931348623157e+308;
  const _duration = dart.JsSymbol('_duration');
  class Duration extends Object {
    Duration(opts) {
      let days = opts && 'days' in opts ? opts.days : 0;
      let hours = opts && 'hours' in opts ? opts.hours : 0;
      let minutes = opts && 'minutes' in opts ? opts.minutes : 0;
      let seconds = opts && 'seconds' in opts ? opts.seconds : 0;
      let milliseconds = opts && 'milliseconds' in opts ? opts.milliseconds : 0;
      let microseconds = opts && 'microseconds' in opts ? opts.microseconds : 0;
      this._microseconds(dart.notNull(days) * dart.notNull(Duration.MICROSECONDS_PER_DAY) + dart.notNull(hours) * dart.notNull(Duration.MICROSECONDS_PER_HOUR) + dart.notNull(minutes) * dart.notNull(Duration.MICROSECONDS_PER_MINUTE) + dart.notNull(seconds) * dart.notNull(Duration.MICROSECONDS_PER_SECOND) + dart.notNull(milliseconds) * dart.notNull(Duration.MICROSECONDS_PER_MILLISECOND) + dart.notNull(microseconds));
    }
    _microseconds(duration) {
      this[_duration] = duration;
    }
    ['+'](other) {
      return new Duration._microseconds(dart.notNull(this[_duration]) + dart.notNull(other[_duration]));
    }
    ['-'](other) {
      return new Duration._microseconds(dart.notNull(this[_duration]) - dart.notNull(other[_duration]));
    }
    ['*'](factor) {
      return new Duration._microseconds((dart.notNull(this[_duration]) * dart.notNull(factor))[dartx.round]());
    }
    ['~/'](quotient) {
      if (quotient == 0)
        dart.throw(new IntegerDivisionByZeroException());
      return new Duration._microseconds((dart.notNull(this[_duration]) / dart.notNull(quotient))[dartx.truncate]());
    }
    ['<'](other) {
      return dart.notNull(this[_duration]) < dart.notNull(other[_duration]);
    }
    ['>'](other) {
      return dart.notNull(this[_duration]) > dart.notNull(other[_duration]);
    }
    ['<='](other) {
      return dart.notNull(this[_duration]) <= dart.notNull(other[_duration]);
    }
    ['>='](other) {
      return dart.notNull(this[_duration]) >= dart.notNull(other[_duration]);
    }
    get inDays() {
      return (dart.notNull(this[_duration]) / dart.notNull(Duration.MICROSECONDS_PER_DAY))[dartx.truncate]();
    }
    get inHours() {
      return (dart.notNull(this[_duration]) / dart.notNull(Duration.MICROSECONDS_PER_HOUR))[dartx.truncate]();
    }
    get inMinutes() {
      return (dart.notNull(this[_duration]) / dart.notNull(Duration.MICROSECONDS_PER_MINUTE))[dartx.truncate]();
    }
    get inSeconds() {
      return (dart.notNull(this[_duration]) / dart.notNull(Duration.MICROSECONDS_PER_SECOND))[dartx.truncate]();
    }
    get inMilliseconds() {
      return (dart.notNull(this[_duration]) / dart.notNull(Duration.MICROSECONDS_PER_MILLISECOND))[dartx.truncate]();
    }
    get inMicroseconds() {
      return this[_duration];
    }
    ['=='](other) {
      if (!dart.is(other, Duration))
        return false;
      return dart.equals(this[_duration], dart.dload(other, _duration));
    }
    get hashCode() {
      return dart.hashCode(this[_duration]);
    }
    compareTo(other) {
      return this[_duration][dartx.compareTo](other[_duration]);
    }
    toString() {
      function sixDigits(n) {
        if (dart.notNull(n) >= 100000)
          return `${n}`;
        if (dart.notNull(n) >= 10000)
          return `0${n}`;
        if (dart.notNull(n) >= 1000)
          return `00${n}`;
        if (dart.notNull(n) >= 100)
          return `000${n}`;
        if (dart.notNull(n) >= 10)
          return `0000${n}`;
        return `00000${n}`;
      }
      dart.fn(sixDigits, String, [int]);
      function twoDigits(n) {
        if (dart.notNull(n) >= 10)
          return `${n}`;
        return `0${n}`;
      }
      dart.fn(twoDigits, String, [int]);
      if (dart.notNull(this.inMicroseconds) < 0) {
        return `-${this['unary-']()}`;
      }
      let twoDigitMinutes = twoDigits(dart.asInt(this.inMinutes[dartx.remainder](Duration.MINUTES_PER_HOUR)));
      let twoDigitSeconds = twoDigits(dart.asInt(this.inSeconds[dartx.remainder](Duration.SECONDS_PER_MINUTE)));
      let sixDigitUs = sixDigits(dart.asInt(this.inMicroseconds[dartx.remainder](Duration.MICROSECONDS_PER_SECOND)));
      return `${this.inHours}:${twoDigitMinutes}:${twoDigitSeconds}.${sixDigitUs}`;
    }
    get isNegative() {
      return dart.notNull(this[_duration]) < 0;
    }
    abs() {
      return new Duration._microseconds(this[_duration][dartx.abs]());
    }
    ['unary-']() {
      return new Duration._microseconds(-dart.notNull(this[_duration]));
    }
  }
  Duration[dart.implements] = () => [Comparable$(Duration)];
  dart.defineNamedConstructor(Duration, '_microseconds');
  dart.setSignature(Duration, {
    constructors: () => ({
      Duration: [Duration, [], {days: int, hours: int, minutes: int, seconds: int, milliseconds: int, microseconds: int}],
      _microseconds: [Duration, [int]]
    }),
    methods: () => ({
      '+': [Duration, [Duration]],
      '-': [Duration, [Duration]],
      '*': [Duration, [num]],
      '~/': [Duration, [int]],
      '<': [bool, [Duration]],
      '>': [bool, [Duration]],
      '<=': [bool, [Duration]],
      '>=': [bool, [Duration]],
      compareTo: [int, [Duration]],
      abs: [Duration, []],
      'unary-': [Duration, []]
    })
  });
  dart.defineExtensionMembers(Duration, ['compareTo']);
  Duration.MICROSECONDS_PER_MILLISECOND = 1000;
  Duration.MILLISECONDS_PER_SECOND = 1000;
  Duration.SECONDS_PER_MINUTE = 60;
  Duration.MINUTES_PER_HOUR = 60;
  Duration.HOURS_PER_DAY = 24;
  Duration.MICROSECONDS_PER_SECOND = dart.notNull(Duration.MICROSECONDS_PER_MILLISECOND) * dart.notNull(Duration.MILLISECONDS_PER_SECOND);
  Duration.MICROSECONDS_PER_MINUTE = dart.notNull(Duration.MICROSECONDS_PER_SECOND) * dart.notNull(Duration.SECONDS_PER_MINUTE);
  Duration.MICROSECONDS_PER_HOUR = dart.notNull(Duration.MICROSECONDS_PER_MINUTE) * dart.notNull(Duration.MINUTES_PER_HOUR);
  Duration.MICROSECONDS_PER_DAY = dart.notNull(Duration.MICROSECONDS_PER_HOUR) * dart.notNull(Duration.HOURS_PER_DAY);
  Duration.MILLISECONDS_PER_MINUTE = dart.notNull(Duration.MILLISECONDS_PER_SECOND) * dart.notNull(Duration.SECONDS_PER_MINUTE);
  Duration.MILLISECONDS_PER_HOUR = dart.notNull(Duration.MILLISECONDS_PER_MINUTE) * dart.notNull(Duration.MINUTES_PER_HOUR);
  Duration.MILLISECONDS_PER_DAY = dart.notNull(Duration.MILLISECONDS_PER_HOUR) * dart.notNull(Duration.HOURS_PER_DAY);
  Duration.SECONDS_PER_HOUR = dart.notNull(Duration.SECONDS_PER_MINUTE) * dart.notNull(Duration.MINUTES_PER_HOUR);
  Duration.SECONDS_PER_DAY = dart.notNull(Duration.SECONDS_PER_HOUR) * dart.notNull(Duration.HOURS_PER_DAY);
  Duration.MINUTES_PER_DAY = dart.notNull(Duration.MINUTES_PER_HOUR) * dart.notNull(Duration.HOURS_PER_DAY);
  Duration.ZERO = dart.const(new Duration({seconds: 0}));
  class Error extends Object {
    Error() {
    }
    static safeToString(object) {
      if (typeof object == 'number' || typeof object == 'boolean' || null == object) {
        return dart.toString(object);
      }
      if (typeof object == 'string') {
        return Error._stringToSafeString(object);
      }
      return Error._objectToString(object);
    }
    static _stringToSafeString(string) {
      return _js_helper.jsonEncodeNative(string);
    }
    static _objectToString(object) {
      return _js_helper.Primitives.objectToString(object);
    }
    get stackTrace() {
      return _js_helper.Primitives.extractStackTrace(this);
    }
  }
  dart.setSignature(Error, {
    constructors: () => ({Error: [Error, []]}),
    statics: () => ({
      safeToString: [String, [Object]],
      _stringToSafeString: [String, [String]],
      _objectToString: [String, [Object]]
    }),
    names: ['safeToString', '_stringToSafeString', '_objectToString']
  });
  class AssertionError extends Error {
    AssertionError() {
      super.Error();
    }
  }
  class TypeError extends AssertionError {}
  class CastError extends Error {
    CastError() {
      super.Error();
    }
  }
  class NullThrownError extends Error {
    NullThrownError() {
      super.Error();
    }
    toString() {
      return "Throw of null.";
    }
  }
  const _hasValue = dart.JsSymbol('_hasValue');
  class ArgumentError extends Error {
    ArgumentError(message) {
      if (message === void 0)
        message = null;
      this.message = message;
      this.invalidValue = null;
      this[_hasValue] = false;
      this.name = null;
      super.Error();
    }
    value(value, name, message) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = "Invalid argument";
      this.name = name;
      this.message = message;
      this.invalidValue = value;
      this[_hasValue] = true;
      super.Error();
    }
    notNull(name) {
      if (name === void 0)
        name = null;
      this.value(null, name, "Must not be null");
    }
    toString() {
      if (!dart.notNull(this[_hasValue])) {
        let result = "Invalid arguments(s)";
        if (this.message != null) {
          result = `${result}: ${this.message}`;
        }
        return result;
      }
      let nameString = "";
      if (this.name != null) {
        nameString = ` (${this.name})`;
      }
      return `${this.message}${nameString}: ${Error.safeToString(this.invalidValue)}`;
    }
  }
  dart.defineNamedConstructor(ArgumentError, 'value');
  dart.defineNamedConstructor(ArgumentError, 'notNull');
  dart.setSignature(ArgumentError, {
    constructors: () => ({
      ArgumentError: [ArgumentError, [], [dart.dynamic]],
      value: [ArgumentError, [dart.dynamic], [String, String]],
      notNull: [ArgumentError, [], [String]]
    })
  });
  class RangeError extends ArgumentError {
    RangeError(message) {
      this.start = null;
      this.end = null;
      super.ArgumentError(message);
    }
    value(value, name, message) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = null;
      this.start = null;
      this.end = null;
      super.value(value, name, message != null ? message : "Value not in range");
    }
    range(invalidValue, minValue, maxValue, name, message) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = null;
      this.start = minValue;
      this.end = maxValue;
      super.value(invalidValue, name, message != null ? message : "Invalid value");
    }
    static index(index, indexable, name, message, length) {
      return new IndexError(index, indexable, name, message, length);
    }
    static checkValueInInterval(value, minValue, maxValue, name, message) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = null;
      if (dart.notNull(value) < dart.notNull(minValue) || dart.notNull(value) > dart.notNull(maxValue)) {
        dart.throw(new RangeError.range(value, minValue, maxValue, name, message));
      }
    }
    static checkValidIndex(index, indexable, name, length, message) {
      if (name === void 0)
        name = null;
      if (length === void 0)
        length = null;
      if (message === void 0)
        message = null;
      if (length == null)
        length = dart.as(dart.dload(indexable, 'length'), int);
      if (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(length)) {
        if (name == null)
          name = "index";
        dart.throw(RangeError.index(index, indexable, name, message, length));
      }
    }
    static checkValidRange(start, end, length, startName, endName, message) {
      if (startName === void 0)
        startName = null;
      if (endName === void 0)
        endName = null;
      if (message === void 0)
        message = null;
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(length)) {
        if (startName == null)
          startName = "start";
        dart.throw(new RangeError.range(start, 0, length, startName, message));
      }
      if (end != null && (dart.notNull(end) < dart.notNull(start) || dart.notNull(end) > dart.notNull(length))) {
        if (endName == null)
          endName = "end";
        dart.throw(new RangeError.range(end, start, length, endName, message));
      }
    }
    static checkNotNegative(value, name, message) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = null;
      if (dart.notNull(value) < 0)
        dart.throw(new RangeError.range(value, 0, null, name, message));
    }
    toString() {
      if (!dart.notNull(this[_hasValue]))
        return `RangeError: ${this.message}`;
      let value = Error.safeToString(this.invalidValue);
      let explanation = "";
      if (this.start == null) {
        if (this.end != null) {
          explanation = `: Not less than or equal to ${this.end}`;
        }
      } else if (this.end == null) {
        explanation = `: Not greater than or equal to ${this.start}`;
      } else if (dart.notNull(this.end) > dart.notNull(this.start)) {
        explanation = `: Not in range ${this.start}..${this.end}, inclusive.`;
      } else if (dart.notNull(this.end) < dart.notNull(this.start)) {
        explanation = ": Valid value range is empty";
      } else {
        explanation = `: Only valid value is ${this.start}`;
      }
      return `RangeError: ${this.message} (${value})${explanation}`;
    }
  }
  dart.defineNamedConstructor(RangeError, 'value');
  dart.defineNamedConstructor(RangeError, 'range');
  dart.setSignature(RangeError, {
    constructors: () => ({
      RangeError: [RangeError, [dart.dynamic]],
      value: [RangeError, [num], [String, String]],
      range: [RangeError, [num, int, int], [String, String]],
      index: [RangeError, [int, dart.dynamic], [String, String, int]]
    }),
    statics: () => ({
      checkValueInInterval: [dart.void, [int, int, int], [String, String]],
      checkValidIndex: [dart.void, [int, dart.dynamic], [String, int, String]],
      checkValidRange: [dart.void, [int, int, int], [String, String, String]],
      checkNotNegative: [dart.void, [int], [String, String]]
    }),
    names: ['checkValueInInterval', 'checkValidIndex', 'checkValidRange', 'checkNotNegative']
  });
  class IndexError extends ArgumentError {
    IndexError(invalidValue, indexable, name, message, length) {
      if (name === void 0)
        name = null;
      if (message === void 0)
        message = null;
      if (length === void 0)
        length = null;
      this.indexable = indexable;
      this.length = length != null ? length : dart.as(dart.dload(indexable, 'length'), int);
      super.value(invalidValue, name, message != null ? message : "Index out of range");
    }
    get start() {
      return 0;
    }
    get end() {
      return dart.notNull(this.length) - 1;
    }
    toString() {
      dart.assert(this[_hasValue]);
      let target = Error.safeToString(this.indexable);
      let explanation = `index should be less than ${this.length}`;
      if (dart.notNull(dart.as(dart.dsend(this.invalidValue, '<', 0), bool))) {
        explanation = "index must not be negative";
      }
      return `RangeError: ${this.message} (${target}[${this.invalidValue}]): ${explanation}`;
    }
  }
  IndexError[dart.implements] = () => [RangeError];
  dart.setSignature(IndexError, {
    constructors: () => ({IndexError: [IndexError, [int, dart.dynamic], [String, String, int]]})
  });
  class FallThroughError extends Error {
    FallThroughError() {
      super.Error();
    }
  }
  dart.setSignature(FallThroughError, {
    constructors: () => ({FallThroughError: [FallThroughError, []]})
  });
  const _className = dart.JsSymbol('_className');
  class AbstractClassInstantiationError extends Error {
    AbstractClassInstantiationError(className) {
      this[_className] = className;
      super.Error();
    }
    toString() {
      return `Cannot instantiate abstract class: '${this[_className]}'`;
    }
  }
  dart.setSignature(AbstractClassInstantiationError, {
    constructors: () => ({AbstractClassInstantiationError: [AbstractClassInstantiationError, [String]]})
  });
  const _receiver = dart.JsSymbol('_receiver');
  const _memberName = dart.JsSymbol('_memberName');
  const _arguments = dart.JsSymbol('_arguments');
  const _namedArguments = dart.JsSymbol('_namedArguments');
  const _existingArgumentNames = dart.JsSymbol('_existingArgumentNames');
  class NoSuchMethodError extends Error {
    NoSuchMethodError(receiver, memberName, positionalArguments, namedArguments, existingArgumentNames) {
      if (existingArgumentNames === void 0)
        existingArgumentNames = null;
      this[_receiver] = receiver;
      this[_memberName] = memberName;
      this[_arguments] = positionalArguments;
      this[_namedArguments] = namedArguments;
      this[_existingArgumentNames] = existingArgumentNames;
      super.Error();
    }
    toString() {
      let sb = new StringBuffer();
      let i = 0;
      if (this[_arguments] != null) {
        for (; dart.notNull(i) < dart.notNull(this[_arguments][dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.notNull(i) > 0) {
            sb.write(", ");
          }
          sb.write(Error.safeToString(this[_arguments][dartx.get](i)));
        }
      }
      if (this[_namedArguments] != null) {
        this[_namedArguments].forEach(dart.fn((key, value) => {
          if (dart.notNull(i) > 0) {
            sb.write(", ");
          }
          sb.write(_symbolToString(key));
          sb.write(": ");
          sb.write(Error.safeToString(value));
          i = dart.notNull(i) + 1;
        }, dart.dynamic, [Symbol, dart.dynamic]));
      }
      if (this[_existingArgumentNames] == null) {
        return `NoSuchMethodError : method not found: '${this[_memberName]}'\n` + `Receiver: ${Error.safeToString(this[_receiver])}\n` + `Arguments: [${sb}]`;
      } else {
        let actualParameters = dart.toString(sb);
        sb = new StringBuffer();
        for (let i = 0; dart.notNull(i) < dart.notNull(this[_existingArgumentNames][dartx.length]); i = dart.notNull(i) + 1) {
          if (dart.notNull(i) > 0) {
            sb.write(", ");
          }
          sb.write(this[_existingArgumentNames][dartx.get](i));
        }
        let formalParameters = dart.toString(sb);
        return "NoSuchMethodError: incorrect number of arguments passed to " + `method named '${this[_memberName]}'\n` + `Receiver: ${Error.safeToString(this[_receiver])}\n` + `Tried calling: ${this[_memberName]}(${actualParameters})\n` + `Found: ${this[_memberName]}(${formalParameters})`;
      }
    }
  }
  dart.setSignature(NoSuchMethodError, {
    constructors: () => ({NoSuchMethodError: [NoSuchMethodError, [Object, Symbol, List, Map$(Symbol, dart.dynamic)], [List]]})
  });
  class UnsupportedError extends Error {
    UnsupportedError(message) {
      this.message = message;
      super.Error();
    }
    toString() {
      return `Unsupported operation: ${this.message}`;
    }
  }
  dart.setSignature(UnsupportedError, {
    constructors: () => ({UnsupportedError: [UnsupportedError, [String]]})
  });
  class UnimplementedError extends Error {
    UnimplementedError(message) {
      if (message === void 0)
        message = null;
      this.message = message;
      super.Error();
    }
    toString() {
      return this.message != null ? `UnimplementedError: ${this.message}` : "UnimplementedError";
    }
  }
  UnimplementedError[dart.implements] = () => [UnsupportedError];
  dart.setSignature(UnimplementedError, {
    constructors: () => ({UnimplementedError: [UnimplementedError, [], [String]]})
  });
  class StateError extends Error {
    StateError(message) {
      this.message = message;
      super.Error();
    }
    toString() {
      return `Bad state: ${this.message}`;
    }
  }
  dart.setSignature(StateError, {
    constructors: () => ({StateError: [StateError, [String]]})
  });
  class ConcurrentModificationError extends Error {
    ConcurrentModificationError(modifiedObject) {
      if (modifiedObject === void 0)
        modifiedObject = null;
      this.modifiedObject = modifiedObject;
      super.Error();
    }
    toString() {
      if (this.modifiedObject == null) {
        return "Concurrent modification during iteration.";
      }
      return "Concurrent modification during iteration: " + `${Error.safeToString(this.modifiedObject)}.`;
    }
  }
  dart.setSignature(ConcurrentModificationError, {
    constructors: () => ({ConcurrentModificationError: [ConcurrentModificationError, [], [Object]]})
  });
  class OutOfMemoryError extends Object {
    OutOfMemoryError() {
    }
    toString() {
      return "Out of Memory";
    }
    get stackTrace() {
      return null;
    }
  }
  OutOfMemoryError[dart.implements] = () => [Error];
  dart.setSignature(OutOfMemoryError, {
    constructors: () => ({OutOfMemoryError: [OutOfMemoryError, []]})
  });
  class StackOverflowError extends Object {
    StackOverflowError() {
    }
    toString() {
      return "Stack Overflow";
    }
    get stackTrace() {
      return null;
    }
  }
  StackOverflowError[dart.implements] = () => [Error];
  dart.setSignature(StackOverflowError, {
    constructors: () => ({StackOverflowError: [StackOverflowError, []]})
  });
  class CyclicInitializationError extends Error {
    CyclicInitializationError(variableName) {
      if (variableName === void 0)
        variableName = null;
      this.variableName = variableName;
      super.Error();
    }
    toString() {
      return this.variableName == null ? "Reading static variable during its initialization" : `Reading static variable '${this.variableName}' during its initialization`;
    }
  }
  dart.setSignature(CyclicInitializationError, {
    constructors: () => ({CyclicInitializationError: [CyclicInitializationError, [], [String]]})
  });
  class Exception extends Object {
    static new(message) {
      if (message === void 0)
        message = null;
      return new _ExceptionImplementation(message);
    }
  }
  dart.setSignature(Exception, {
    constructors: () => ({new: [Exception, [], [dart.dynamic]]})
  });
  class _ExceptionImplementation extends Object {
    _ExceptionImplementation(message) {
      if (message === void 0)
        message = null;
      this.message = message;
    }
    toString() {
      if (this.message == null)
        return "Exception";
      return `Exception: ${this.message}`;
    }
  }
  _ExceptionImplementation[dart.implements] = () => [Exception];
  dart.setSignature(_ExceptionImplementation, {
    constructors: () => ({_ExceptionImplementation: [_ExceptionImplementation, [], [dart.dynamic]]})
  });
  class FormatException extends Object {
    FormatException(message, source, offset) {
      if (message === void 0)
        message = "";
      if (source === void 0)
        source = null;
      if (offset === void 0)
        offset = -1;
      this.message = message;
      this.source = source;
      this.offset = offset;
    }
    toString() {
      let report = "FormatException";
      if (this.message != null && "" != this.message) {
        report = `${report}: ${this.message}`;
      }
      let offset = this.offset;
      if (!(typeof this.source == 'string')) {
        if (offset != -1) {
          report = dart.notNull(report) + ` (at offset ${offset})`;
        }
        return report;
      }
      if (offset != -1 && (dart.notNull(offset) < 0 || dart.notNull(offset) > dart.notNull(dart.as(dart.dload(this.source, 'length'), num)))) {
        offset = -1;
      }
      if (offset == -1) {
        let source = dart.as(this.source, String);
        if (dart.notNull(source[dartx.length]) > 78) {
          source = dart.notNull(source[dartx.substring](0, 75)) + "...";
        }
        return `${report}\n${source}`;
      }
      let lineNum = 1;
      let lineStart = 0;
      let lastWasCR = null;
      for (let i = 0; dart.notNull(i) < dart.notNull(offset); i = dart.notNull(i) + 1) {
        let char = dart.as(dart.dsend(this.source, 'codeUnitAt', i), int);
        if (char == 10) {
          if (lineStart != i || !dart.notNull(lastWasCR)) {
            lineNum = dart.notNull(lineNum) + 1;
          }
          lineStart = dart.notNull(i) + 1;
          lastWasCR = false;
        } else if (char == 13) {
          lineNum = dart.notNull(lineNum) + 1;
          lineStart = dart.notNull(i) + 1;
          lastWasCR = true;
        }
      }
      if (dart.notNull(lineNum) > 1) {
        report = dart.notNull(report) + ` (at line ${lineNum}, character ${dart.notNull(offset) - dart.notNull(lineStart) + 1})\n`;
      } else {
        report = dart.notNull(report) + ` (at character ${dart.notNull(offset) + 1})\n`;
      }
      let lineEnd = dart.as(dart.dload(this.source, 'length'), int);
      for (let i = offset; dart.notNull(i) < dart.notNull(dart.as(dart.dload(this.source, 'length'), num)); i = dart.notNull(i) + 1) {
        let char = dart.as(dart.dsend(this.source, 'codeUnitAt', i), int);
        if (char == 10 || char == 13) {
          lineEnd = i;
          break;
        }
      }
      let length = dart.notNull(lineEnd) - dart.notNull(lineStart);
      let start = lineStart;
      let end = lineEnd;
      let prefix = "";
      let postfix = "";
      if (dart.notNull(length) > 78) {
        let index = dart.notNull(offset) - dart.notNull(lineStart);
        if (dart.notNull(index) < 75) {
          end = dart.notNull(start) + 75;
          postfix = "...";
        } else if (dart.notNull(end) - dart.notNull(offset) < 75) {
          start = dart.notNull(end) - 75;
          prefix = "...";
        } else {
          start = dart.notNull(offset) - 36;
          end = dart.notNull(offset) + 36;
          prefix = postfix = "...";
        }
      }
      let slice = dart.as(dart.dsend(this.source, 'substring', start, end), String);
      let markOffset = dart.notNull(offset) - dart.notNull(start) + dart.notNull(prefix[dartx.length]);
      return `${report}${prefix}${slice}${postfix}\n${" "[dartx['*']](markOffset)}^\n`;
    }
  }
  FormatException[dart.implements] = () => [Exception];
  dart.setSignature(FormatException, {
    constructors: () => ({FormatException: [FormatException, [], [String, dart.dynamic, int]]})
  });
  class IntegerDivisionByZeroException extends Object {
    IntegerDivisionByZeroException() {
    }
    toString() {
      return "IntegerDivisionByZeroException";
    }
  }
  IntegerDivisionByZeroException[dart.implements] = () => [Exception];
  dart.setSignature(IntegerDivisionByZeroException, {
    constructors: () => ({IntegerDivisionByZeroException: [IntegerDivisionByZeroException, []]})
  });
  const _getKey = dart.JsSymbol('_getKey');
  const Expando$ = dart.generic(function(T) {
    class Expando extends Object {
      Expando(name) {
        if (name === void 0)
          name = null;
        this.name = name;
      }
      toString() {
        return `Expando:${this.name}`;
      }
      get(object) {
        let values = _js_helper.Primitives.getProperty(object, Expando$()._EXPANDO_PROPERTY_NAME);
        return values == null ? null : dart.as(_js_helper.Primitives.getProperty(values, this[_getKey]()), T);
      }
      set(object, value) {
        dart.as(value, T);
        let values = _js_helper.Primitives.getProperty(object, Expando$()._EXPANDO_PROPERTY_NAME);
        if (values == null) {
          values = new Object();
          _js_helper.Primitives.setProperty(object, Expando$()._EXPANDO_PROPERTY_NAME, values);
        }
        _js_helper.Primitives.setProperty(values, this[_getKey](), value);
        return value;
      }
      [_getKey]() {
        let key = dart.as(_js_helper.Primitives.getProperty(this, Expando$()._KEY_PROPERTY_NAME), String);
        if (key == null) {
          key = `expando\$key\$${(() => {
            let x = Expando$()._keyCount;
            Expando$()._keyCount = dart.notNull(x) + 1;
            return x;
          })()}`;
          _js_helper.Primitives.setProperty(this, Expando$()._KEY_PROPERTY_NAME, key);
        }
        return key;
      }
    }
    dart.setSignature(Expando, {
      constructors: () => ({Expando: [Expando$(T), [], [String]]}),
      methods: () => ({
        get: [T, [Object]],
        set: [dart.void, [Object, T]],
        [_getKey]: [String, []]
      })
    });
    return Expando;
  });
  let Expando = Expando$();
  Expando._KEY_PROPERTY_NAME = 'expando$key';
  Expando._EXPANDO_PROPERTY_NAME = 'expando$values';
  Expando._keyCount = 0;
  class Function extends Object {
    static apply(f, positionalArguments, namedArguments) {
      if (namedArguments === void 0)
        namedArguments = null;
      return dart.dcall.apply(null, [f].concat(positionalArguments));
    }
    static _toMangledNames(namedArguments) {
      let result = dart.map();
      namedArguments.forEach(dart.fn((symbol, value) => {
        result.set(_symbolToString(dart.as(symbol, Symbol)), value);
      }));
      return result;
    }
  }
  dart.setSignature(Function, {
    statics: () => ({
      apply: [dart.dynamic, [Function, List], [Map$(Symbol, dart.dynamic)]],
      _toMangledNames: [Map$(String, dart.dynamic), [Map$(Symbol, dart.dynamic)]]
    }),
    names: ['apply', '_toMangledNames']
  });
  function identical(a, b) {
    return _js_helper.Primitives.identicalImplementation(a, b);
  }
  dart.fn(identical, bool, [Object, Object]);
  function identityHashCode(object) {
    return _js_helper.objectHashCode(object);
  }
  dart.fn(identityHashCode, () => dart.definiteFunctionType(int, [Object]));
  class int extends num {
    static fromEnvironment(name, opts) {
      let defaultValue = opts && 'defaultValue' in opts ? opts.defaultValue : null;
      dart.throw(new UnsupportedError('int.fromEnvironment can only be used as a const constructor'));
    }
    static parse(source, opts) {
      let radix = opts && 'radix' in opts ? opts.radix : null;
      let onError = opts && 'onError' in opts ? opts.onError : null;
      return _js_helper.Primitives.parseInt(source, radix, onError);
    }
  }
  dart.setSignature(int, {
    constructors: () => ({fromEnvironment: [int, [String], {defaultValue: int}]}),
    statics: () => ({parse: [int, [String], {radix: int, onError: dart.functionType(int, [String])}]}),
    names: ['parse']
  });
  class Invocation extends Object {
    get isAccessor() {
      return dart.notNull(this.isGetter) || dart.notNull(this.isSetter);
    }
  }
  const Iterable$ = dart.generic(function(E) {
    dart.defineExtensionNames([
      'join'
    ]);
    class Iterable extends Object {
      Iterable() {
      }
      static generate(count, generator) {
        if (generator === void 0)
          generator = null;
        if (dart.notNull(count) <= 0)
          return new (_internal.EmptyIterable$(E))();
        return new (exports._GeneratorIterable$(E))(count, generator);
      }
      [dart.JsSymbol.iterator]() {
        return new dart.JsIterator(this[dartx.iterator]);
      }
      [dartx.join](separator) {
        if (separator === void 0)
          separator = "";
        let buffer = new StringBuffer();
        buffer.writeAll(this, separator);
        return dart.toString(buffer);
      }
    }
    dart.setSignature(Iterable, {
      constructors: () => ({
        Iterable: [Iterable$(E), []],
        generate: [Iterable$(E), [int], [dart.functionType(E, [int])]]
      }),
      methods: () => ({[dartx.join]: [String, [], [String]]})
    });
    return Iterable;
  });
  let Iterable = Iterable$();
  const _Generator$ = dart.generic(function(E) {
    const _Generator = dart.typedef('_Generator', () => dart.functionType(E, [int]));
    return _Generator;
  });
  let _Generator = _Generator$();
  const _end = dart.JsSymbol('_end');
  const _start = dart.JsSymbol('_start');
  const _generator = dart.JsSymbol('_generator');
  const _GeneratorIterable$ = dart.generic(function(E) {
    class _GeneratorIterable extends collection.IterableBase$(E) {
      _GeneratorIterable(end, generator) {
        this[_end] = end;
        this[_start] = 0;
        this[_generator] = dart.as(generator != null ? generator : exports._GeneratorIterable$()._id, _Generator$(E));
        super.IterableBase();
      }
      slice(start, end, generator) {
        this[_start] = start;
        this[_end] = end;
        this[_generator] = generator;
        super.IterableBase();
      }
      get iterator() {
        return new (_GeneratorIterator$(E))(this[_start], this[_end], this[_generator]);
      }
      get length() {
        return dart.notNull(this[_end]) - dart.notNull(this[_start]);
      }
      skip(count) {
        RangeError.checkNotNegative(count, "count");
        if (count == 0)
          return this;
        let newStart = dart.notNull(this[_start]) + dart.notNull(count);
        if (dart.notNull(newStart) >= dart.notNull(this[_end]))
          return new (_internal.EmptyIterable$(E))();
        return new (exports._GeneratorIterable$(E)).slice(newStart, this[_end], this[_generator]);
      }
      take(count) {
        RangeError.checkNotNegative(count, "count");
        if (count == 0)
          return new (_internal.EmptyIterable$(E))();
        let newEnd = dart.notNull(this[_start]) + dart.notNull(count);
        if (dart.notNull(newEnd) >= dart.notNull(this[_end]))
          return this;
        return new (exports._GeneratorIterable$(E)).slice(this[_start], newEnd, this[_generator]);
      }
      static _id(n) {
        return n;
      }
    }
    _GeneratorIterable[dart.implements] = () => [_internal.EfficientLength];
    dart.defineNamedConstructor(_GeneratorIterable, 'slice');
    dart.setSignature(_GeneratorIterable, {
      constructors: () => ({
        _GeneratorIterable: [exports._GeneratorIterable$(E), [int, dart.functionType(E, [int])]],
        slice: [exports._GeneratorIterable$(E), [int, int, _Generator$(E)]]
      }),
      methods: () => ({
        skip: [Iterable$(E), [int]],
        take: [Iterable$(E), [int]]
      }),
      statics: () => ({_id: [int, [int]]}),
      names: ['_id']
    });
    dart.defineExtensionMembers(_GeneratorIterable, ['skip', 'take', 'iterator', 'length']);
    return _GeneratorIterable;
  });
  dart.defineLazyClassGeneric(exports, '_GeneratorIterable', {get: _GeneratorIterable$});
  const _index = dart.JsSymbol('_index');
  const _current = dart.JsSymbol('_current');
  const _GeneratorIterator$ = dart.generic(function(E) {
    class _GeneratorIterator extends Object {
      _GeneratorIterator(index, end, generator) {
        this[_index] = index;
        this[_end] = end;
        this[_generator] = generator;
        this[_current] = null;
      }
      moveNext() {
        if (dart.notNull(this[_index]) < dart.notNull(this[_end])) {
          this[_current] = this[_generator](this[_index]);
          this[_index] = dart.notNull(this[_index]) + 1;
          return true;
        } else {
          this[_current] = null;
          return false;
        }
      }
      get current() {
        return this[_current];
      }
    }
    _GeneratorIterator[dart.implements] = () => [Iterator$(E)];
    dart.setSignature(_GeneratorIterator, {
      constructors: () => ({_GeneratorIterator: [_GeneratorIterator$(E), [int, int, _Generator$(E)]]}),
      methods: () => ({moveNext: [bool, []]})
    });
    return _GeneratorIterator;
  });
  let _GeneratorIterator = _GeneratorIterator$();
  const BidirectionalIterator$ = dart.generic(function(E) {
    class BidirectionalIterator extends Object {}
    BidirectionalIterator[dart.implements] = () => [Iterator$(E)];
    return BidirectionalIterator;
  });
  let BidirectionalIterator = BidirectionalIterator$();
  const Iterator$ = dart.generic(function(E) {
    class Iterator extends Object {}
    return Iterator;
  });
  let Iterator = Iterator$();
  const List$ = dart.generic(function(E) {
    class List extends Object {
      static new(length) {
        if (length === void 0)
          length = null;
        let list = null;
        if (length == null) {
          list = [];
        } else {
          if (!(typeof length == 'number') || dart.notNull(length) < 0) {
            dart.throw(new ArgumentError(`Length must be a non-negative integer: ${length}`));
          }
          list = _interceptors.JSArray.markFixedList(dart.as(new Array(length), List$()));
        }
        return _interceptors.JSArray$(E).typed(list);
      }
      static filled(length, fill) {
        let result = List$(E).new(length);
        if (length != 0 && fill != null) {
          for (let i = 0; dart.notNull(i) < dart.notNull(result[dartx.length]); i = dart.notNull(i) + 1) {
            result[dartx.set](i, fill);
          }
        }
        return result;
      }
      static from(elements, opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let list = List$(E).new();
        for (let e of elements) {
          list[dartx.add](dart.as(e, E));
        }
        if (dart.notNull(growable))
          return list;
        return dart.as(_internal.makeListFixedLength(list), List$(E));
      }
      static generate(length, generator, opts) {
        let growable = opts && 'growable' in opts ? opts.growable : true;
        let result = null;
        if (dart.notNull(growable)) {
          result = dart.list([], E);
          result[dartx.length] = length;
        } else {
          result = List$(E).new(length);
        }
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          result[dartx.set](i, generator(i));
        }
        return result;
      }
      [dart.JsSymbol.iterator]() {
        return new dart.JsIterator(this[dartx.iterator]);
      }
    }
    List[dart.implements] = () => [Iterable$(E)];
    dart.setSignature(List, {
      constructors: () => ({
        new: [List$(E), [], [int]],
        filled: [List$(E), [int, E]],
        from: [List$(E), [Iterable], {growable: bool}],
        generate: [List$(E), [int, dart.functionType(E, [int])], {growable: bool}]
      })
    });
    return List;
  });
  let List = List$();
  const Map$ = dart.generic(function(K, V) {
    class Map extends Object {
      static new() {
        return collection.LinkedHashMap$(K, V).new();
      }
      static from(other) {
        return collection.LinkedHashMap$(K, V).from(other);
      }
      static identity() {
        return collection.LinkedHashMap$(K, V).identity();
      }
      static fromIterable(iterable, opts) {
        return collection.LinkedHashMap$(K, V).fromIterable(iterable, opts);
      }
      static fromIterables(keys, values) {
        return collection.LinkedHashMap$(K, V).fromIterables(keys, values);
      }
    }
    dart.setSignature(Map, {
      constructors: () => ({
        new: [Map$(K, V), []],
        from: [Map$(K, V), [Map$()]],
        identity: [Map$(K, V), []],
        fromIterable: [Map$(K, V), [Iterable], {key: dart.functionType(K, [dart.dynamic]), value: dart.functionType(V, [dart.dynamic])}],
        fromIterables: [Map$(K, V), [Iterable$(K), Iterable$(V)]]
      })
    });
    return Map;
  });
  let Map = Map$();
  class Null extends Object {
    static _uninstantiable() {
      dart.throw(new UnsupportedError('class Null cannot be instantiated'));
    }
    toString() {
      return "null";
    }
  }
  dart.setSignature(Null, {
    constructors: () => ({_uninstantiable: [Null, []]})
  });
  num._parseError = false;
  class Pattern extends Object {}
  function print(object) {
    let line = `${object}`;
    if (_internal.printToZone == null) {
      _internal.printToConsole(line);
    } else {
      dart.dcall(_internal.printToZone, line);
    }
  }
  dart.fn(print, dart.void, [Object]);
  class Match extends Object {}
  class RegExp extends Object {
    static new(source, opts) {
      let multiLine = opts && 'multiLine' in opts ? opts.multiLine : false;
      let caseSensitive = opts && 'caseSensitive' in opts ? opts.caseSensitive : true;
      return new _js_helper.JSSyntaxRegExp(source, {multiLine: multiLine, caseSensitive: caseSensitive});
    }
  }
  RegExp[dart.implements] = () => [Pattern];
  dart.setSignature(RegExp, {
    constructors: () => ({new: [RegExp, [String], {multiLine: bool, caseSensitive: bool}]})
  });
  const Set$ = dart.generic(function(E) {
    class Set extends collection.IterableBase$(E) {
      static new() {
        return collection.LinkedHashSet$(E).new();
      }
      static identity() {
        return collection.LinkedHashSet$(E).identity();
      }
      static from(elements) {
        return collection.LinkedHashSet$(E).from(elements);
      }
    }
    Set[dart.implements] = () => [_internal.EfficientLength];
    dart.setSignature(Set, {
      constructors: () => ({
        new: [exports.Set$(E), []],
        identity: [exports.Set$(E), []],
        from: [exports.Set$(E), [Iterable]]
      })
    });
    return Set;
  });
  dart.defineLazyClassGeneric(exports, 'Set', {get: Set$});
  const Sink$ = dart.generic(function(T) {
    class Sink extends Object {}
    return Sink;
  });
  let Sink = Sink$();
  class StackTrace extends Object {}
  const _stop = dart.JsSymbol('_stop');
  class Stopwatch extends Object {
    get frequency() {
      return Stopwatch._frequency;
    }
    Stopwatch() {
      this[_start] = null;
      this[_stop] = null;
      Stopwatch._initTicker();
    }
    start() {
      if (dart.notNull(this.isRunning))
        return;
      if (this[_start] == null) {
        this[_start] = Stopwatch._now();
      } else {
        this[_start] = dart.notNull(Stopwatch._now()) - (dart.notNull(this[_stop]) - dart.notNull(this[_start]));
        this[_stop] = null;
      }
    }
    stop() {
      if (!dart.notNull(this.isRunning))
        return;
      this[_stop] = Stopwatch._now();
    }
    reset() {
      if (this[_start] == null)
        return;
      this[_start] = Stopwatch._now();
      if (this[_stop] != null) {
        this[_stop] = this[_start];
      }
    }
    get elapsedTicks() {
      if (this[_start] == null) {
        return 0;
      }
      return dart.asInt(this[_stop] == null ? dart.notNull(Stopwatch._now()) - dart.notNull(this[_start]) : dart.notNull(this[_stop]) - dart.notNull(this[_start]));
    }
    get elapsed() {
      return new Duration({microseconds: this.elapsedMicroseconds});
    }
    get elapsedMicroseconds() {
      return (dart.notNull(this.elapsedTicks) * 1000000 / dart.notNull(this.frequency))[dartx.truncate]();
    }
    get elapsedMilliseconds() {
      return (dart.notNull(this.elapsedTicks) * 1000 / dart.notNull(this.frequency))[dartx.truncate]();
    }
    get isRunning() {
      return this[_start] != null && this[_stop] == null;
    }
    static _initTicker() {
      _js_helper.Primitives.initTicker();
      Stopwatch._frequency = _js_helper.Primitives.timerFrequency;
    }
    static _now() {
      return dart.as(dart.dcall(_js_helper.Primitives.timerTicks), int);
    }
  }
  dart.setSignature(Stopwatch, {
    constructors: () => ({Stopwatch: [Stopwatch, []]}),
    methods: () => ({
      start: [dart.void, []],
      stop: [dart.void, []],
      reset: [dart.void, []]
    }),
    statics: () => ({
      _initTicker: [dart.void, []],
      _now: [int, []]
    }),
    names: ['_initTicker', '_now']
  });
  Stopwatch._frequency = null;
  class String extends Object {
    static fromCharCodes(charCodes, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      if (!dart.is(charCodes, _interceptors.JSArray)) {
        return String._stringFromIterable(charCodes, start, end);
      }
      let list = dart.as(charCodes, List);
      let len = list[dartx.length];
      if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(len)) {
        dart.throw(new RangeError.range(start, 0, len));
      }
      if (end == null) {
        end = len;
      } else if (dart.notNull(end) < dart.notNull(start) || dart.notNull(end) > dart.notNull(len)) {
        dart.throw(new RangeError.range(end, start, len));
      }
      if (dart.notNull(start) > 0 || dart.notNull(end) < dart.notNull(len)) {
        list = list[dartx.sublist](start, end);
      }
      return _js_helper.Primitives.stringFromCharCodes(list);
    }
    static fromCharCode(charCode) {
      return _js_helper.Primitives.stringFromCharCode(charCode);
    }
    static fromEnvironment(name, opts) {
      let defaultValue = opts && 'defaultValue' in opts ? opts.defaultValue : null;
      dart.throw(new UnsupportedError('String.fromEnvironment can only be used as a const constructor'));
    }
    static _stringFromIterable(charCodes, start, end) {
      if (dart.notNull(start) < 0)
        dart.throw(new RangeError.range(start, 0, charCodes[dartx.length]));
      if (end != null && dart.notNull(end) < dart.notNull(start)) {
        dart.throw(new RangeError.range(end, start, charCodes[dartx.length]));
      }
      let it = charCodes[dartx.iterator];
      for (let i = 0; dart.notNull(i) < dart.notNull(start); i = dart.notNull(i) + 1) {
        if (!dart.notNull(it.moveNext())) {
          dart.throw(new RangeError.range(start, 0, i));
        }
      }
      let list = [];
      if (end == null) {
        while (dart.notNull(it.moveNext()))
          list[dartx.add](it.current);
      } else {
        for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
          if (!dart.notNull(it.moveNext())) {
            dart.throw(new RangeError.range(end, start, i));
          }
          list[dartx.add](it.current);
        }
      }
      return _js_helper.Primitives.stringFromCharCodes(list);
    }
  }
  String[dart.implements] = () => [Comparable$(String), Pattern];
  dart.setSignature(String, {
    constructors: () => ({
      fromCharCodes: [String, [Iterable$(int)], [int, int]],
      fromCharCode: [String, [int]],
      fromEnvironment: [String, [String], {defaultValue: String}]
    }),
    statics: () => ({_stringFromIterable: [String, [Iterable$(int), int, int]]}),
    names: ['_stringFromIterable']
  });
  dart.defineLazyClass(exports, {
    get Runes() {
      class Runes extends collection.IterableBase$(int) {
        Runes(string) {
          this.string = string;
          super.IterableBase();
        }
        get iterator() {
          return new RuneIterator(this.string);
        }
        get last() {
          if (this.string[dartx.length] == 0) {
            dart.throw(new StateError('No elements.'));
          }
          let length = this.string[dartx.length];
          let code = this.string[dartx.codeUnitAt](dart.notNull(length) - 1);
          if (dart.notNull(_isTrailSurrogate(code)) && dart.notNull(this.string[dartx.length]) > 1) {
            let previousCode = this.string[dartx.codeUnitAt](dart.notNull(length) - 2);
            if (dart.notNull(_isLeadSurrogate(previousCode))) {
              return _combineSurrogatePair(previousCode, code);
            }
          }
          return code;
        }
      }
      dart.setSignature(Runes, {
        constructors: () => ({Runes: [exports.Runes, [String]]})
      });
      dart.defineExtensionMembers(Runes, ['iterator', 'last']);
      return Runes;
    }
  });
  function _isLeadSurrogate(code) {
    return (dart.notNull(code) & 64512) == 55296;
  }
  dart.fn(_isLeadSurrogate, bool, [int]);
  function _isTrailSurrogate(code) {
    return (dart.notNull(code) & 64512) == 56320;
  }
  dart.fn(_isTrailSurrogate, bool, [int]);
  function _combineSurrogatePair(start, end) {
    return 65536 + ((dart.notNull(start) & 1023) << 10) + (dart.notNull(end) & 1023);
  }
  dart.fn(_combineSurrogatePair, int, [int, int]);
  const _position = dart.JsSymbol('_position');
  const _nextPosition = dart.JsSymbol('_nextPosition');
  const _currentCodePoint = dart.JsSymbol('_currentCodePoint');
  const _checkSplitSurrogate = dart.JsSymbol('_checkSplitSurrogate');
  class RuneIterator extends Object {
    RuneIterator(string) {
      this.string = string;
      this[_position] = 0;
      this[_nextPosition] = 0;
      this[_currentCodePoint] = null;
    }
    at(string, index) {
      this.string = string;
      this[_position] = index;
      this[_nextPosition] = index;
      this[_currentCodePoint] = null;
      RangeError.checkValueInInterval(index, 0, string[dartx.length]);
      this[_checkSplitSurrogate](index);
    }
    [_checkSplitSurrogate](index) {
      if (dart.notNull(index) > 0 && dart.notNull(index) < dart.notNull(this.string[dartx.length]) && dart.notNull(_isLeadSurrogate(this.string[dartx.codeUnitAt](dart.notNull(index) - 1))) && dart.notNull(_isTrailSurrogate(this.string[dartx.codeUnitAt](index)))) {
        dart.throw(new ArgumentError(`Index inside surrogate pair: ${index}`));
      }
    }
    get rawIndex() {
      return this[_position] != this[_nextPosition] ? this[_position] : null;
    }
    set rawIndex(rawIndex) {
      RangeError.checkValidIndex(rawIndex, this.string, "rawIndex");
      this.reset(rawIndex);
      this.moveNext();
    }
    reset(rawIndex) {
      if (rawIndex === void 0)
        rawIndex = 0;
      RangeError.checkValueInInterval(rawIndex, 0, this.string[dartx.length], "rawIndex");
      this[_checkSplitSurrogate](rawIndex);
      this[_position] = this[_nextPosition] = rawIndex;
      this[_currentCodePoint] = null;
    }
    get current() {
      return dart.asInt(this[_currentCodePoint]);
    }
    get currentSize() {
      return dart.notNull(this[_nextPosition]) - dart.notNull(this[_position]);
    }
    get currentAsString() {
      if (this[_position] == this[_nextPosition])
        return null;
      if (dart.notNull(this[_position]) + 1 == this[_nextPosition])
        return this.string[dartx.get](this[_position]);
      return this.string[dartx.substring](this[_position], this[_nextPosition]);
    }
    moveNext() {
      this[_position] = this[_nextPosition];
      if (this[_position] == this.string[dartx.length]) {
        this[_currentCodePoint] = null;
        return false;
      }
      let codeUnit = this.string[dartx.codeUnitAt](this[_position]);
      let nextPosition = dart.notNull(this[_position]) + 1;
      if (dart.notNull(_isLeadSurrogate(codeUnit)) && dart.notNull(nextPosition) < dart.notNull(this.string[dartx.length])) {
        let nextCodeUnit = this.string[dartx.codeUnitAt](nextPosition);
        if (dart.notNull(_isTrailSurrogate(nextCodeUnit))) {
          this[_nextPosition] = dart.notNull(nextPosition) + 1;
          this[_currentCodePoint] = _combineSurrogatePair(codeUnit, nextCodeUnit);
          return true;
        }
      }
      this[_nextPosition] = nextPosition;
      this[_currentCodePoint] = codeUnit;
      return true;
    }
    movePrevious() {
      this[_nextPosition] = this[_position];
      if (this[_position] == 0) {
        this[_currentCodePoint] = null;
        return false;
      }
      let position = dart.notNull(this[_position]) - 1;
      let codeUnit = this.string[dartx.codeUnitAt](position);
      if (dart.notNull(_isTrailSurrogate(codeUnit)) && dart.notNull(position) > 0) {
        let prevCodeUnit = this.string[dartx.codeUnitAt](dart.notNull(position) - 1);
        if (dart.notNull(_isLeadSurrogate(prevCodeUnit))) {
          this[_position] = dart.notNull(position) - 1;
          this[_currentCodePoint] = _combineSurrogatePair(prevCodeUnit, codeUnit);
          return true;
        }
      }
      this[_position] = position;
      this[_currentCodePoint] = codeUnit;
      return true;
    }
  }
  RuneIterator[dart.implements] = () => [BidirectionalIterator$(int)];
  dart.defineNamedConstructor(RuneIterator, 'at');
  dart.setSignature(RuneIterator, {
    constructors: () => ({
      RuneIterator: [RuneIterator, [String]],
      at: [RuneIterator, [String, int]]
    }),
    methods: () => ({
      [_checkSplitSurrogate]: [dart.void, [int]],
      reset: [dart.void, [], [int]],
      moveNext: [bool, []],
      movePrevious: [bool, []]
    })
  });
  const _contents = dart.JsSymbol('_contents');
  const _writeString = dart.JsSymbol('_writeString');
  class StringBuffer extends Object {
    StringBuffer(content) {
      if (content === void 0)
        content = "";
      this[_contents] = `${content}`;
    }
    get length() {
      return this[_contents][dartx.length];
    }
    get isEmpty() {
      return this.length == 0;
    }
    get isNotEmpty() {
      return !dart.notNull(this.isEmpty);
    }
    write(obj) {
      this[_writeString](`${obj}`);
    }
    writeCharCode(charCode) {
      this[_writeString](String.fromCharCode(charCode));
    }
    writeAll(objects, separator) {
      if (separator === void 0)
        separator = "";
      let iterator = objects[dartx.iterator];
      if (!dart.notNull(iterator.moveNext()))
        return;
      if (dart.notNull(separator[dartx.isEmpty])) {
        do {
          this.write(iterator.current);
        } while (dart.notNull(iterator.moveNext()));
      } else {
        this.write(iterator.current);
        while (dart.notNull(iterator.moveNext())) {
          this.write(separator);
          this.write(iterator.current);
        }
      }
    }
    writeln(obj) {
      if (obj === void 0)
        obj = "";
      this.write(obj);
      this.write("\n");
    }
    clear() {
      this[_contents] = "";
    }
    toString() {
      return _js_helper.Primitives.flattenString(this[_contents]);
    }
    [_writeString](str) {
      this[_contents] = _js_helper.Primitives.stringConcatUnchecked(this[_contents], dart.as(str, String));
    }
  }
  StringBuffer[dart.implements] = () => [StringSink];
  dart.setSignature(StringBuffer, {
    constructors: () => ({StringBuffer: [StringBuffer, [], [Object]]}),
    methods: () => ({
      write: [dart.void, [Object]],
      writeCharCode: [dart.void, [int]],
      writeAll: [dart.void, [Iterable], [String]],
      writeln: [dart.void, [], [Object]],
      clear: [dart.void, []],
      [_writeString]: [dart.void, [dart.dynamic]]
    })
  });
  class StringSink extends Object {}
  class Symbol extends Object {
    static new(name) {
      return new _internal.Symbol(name);
    }
  }
  dart.setSignature(Symbol, {
    constructors: () => ({new: [Symbol, [String]]})
  });
  class Type extends Object {}
  const _writeAuthority = dart.JsSymbol('_writeAuthority');
  const _userInfo = dart.JsSymbol('_userInfo');
  const _host = dart.JsSymbol('_host');
  const _port = dart.JsSymbol('_port');
  const _path = dart.JsSymbol('_path');
  const _query = dart.JsSymbol('_query');
  const _fragment = dart.JsSymbol('_fragment');
  const _pathSegments = dart.JsSymbol('_pathSegments');
  const _queryParameters = dart.JsSymbol('_queryParameters');
  const _merge = dart.JsSymbol('_merge');
  const _hasDotSegments = dart.JsSymbol('_hasDotSegments');
  const _removeDotSegments = dart.JsSymbol('_removeDotSegments');
  const _toWindowsFilePath = dart.JsSymbol('_toWindowsFilePath');
  const _toFilePath = dart.JsSymbol('_toFilePath');
  const _isPathAbsolute = dart.JsSymbol('_isPathAbsolute');
  class Uri extends Object {
    get authority() {
      if (!dart.notNull(this.hasAuthority))
        return "";
      let sb = new StringBuffer();
      this[_writeAuthority](sb);
      return dart.toString(sb);
    }
    get userInfo() {
      return this[_userInfo];
    }
    get host() {
      if (this[_host] == null)
        return "";
      if (dart.notNull(this[_host][dartx.startsWith]('['))) {
        return this[_host][dartx.substring](1, dart.notNull(this[_host][dartx.length]) - 1);
      }
      return this[_host];
    }
    get port() {
      if (this[_port] == null)
        return Uri._defaultPort(this.scheme);
      return dart.asInt(this[_port]);
    }
    static _defaultPort(scheme) {
      if (scheme == "http")
        return 80;
      if (scheme == "https")
        return 443;
      return 0;
    }
    get path() {
      return this[_path];
    }
    get query() {
      return this[_query] == null ? "" : this[_query];
    }
    get fragment() {
      return this[_fragment] == null ? "" : this[_fragment];
    }
    static parse(uri) {
      function isRegName(ch) {
        return dart.notNull(ch) < 128 && !dart.equals(dart.dsend(Uri._regNameTable[dartx.get](dart.notNull(ch) >> 4), '&', 1 << (dart.notNull(ch) & 15)), 0);
      }
      dart.fn(isRegName, bool, [int]);
      let EOI = -1;
      let scheme = "";
      let userinfo = "";
      let host = null;
      let port = null;
      let path = null;
      let query = null;
      let fragment = null;
      let index = 0;
      let pathStart = 0;
      let char = EOI;
      function parseAuth() {
        if (index == uri[dartx.length]) {
          char = EOI;
          return;
        }
        let authStart = index;
        let lastColon = -1;
        let lastAt = -1;
        char = uri[dartx.codeUnitAt](index);
        while (dart.notNull(index) < dart.notNull(uri[dartx.length])) {
          char = uri[dartx.codeUnitAt](index);
          if (char == Uri._SLASH || char == Uri._QUESTION || char == Uri._NUMBER_SIGN) {
            break;
          }
          if (char == Uri._AT_SIGN) {
            lastAt = index;
            lastColon = -1;
          } else if (char == Uri._COLON) {
            lastColon = index;
          } else if (char == Uri._LEFT_BRACKET) {
            lastColon = -1;
            let endBracket = uri[dartx.indexOf](']', dart.notNull(index) + 1);
            if (endBracket == -1) {
              index = uri[dartx.length];
              char = EOI;
              break;
            } else {
              index = endBracket;
            }
          }
          index = dart.notNull(index) + 1;
          char = EOI;
        }
        let hostStart = authStart;
        let hostEnd = index;
        if (dart.notNull(lastAt) >= 0) {
          userinfo = Uri._makeUserInfo(uri, authStart, lastAt);
          hostStart = dart.notNull(lastAt) + 1;
        }
        if (dart.notNull(lastColon) >= 0) {
          let portNumber = null;
          if (dart.notNull(lastColon) + 1 < dart.notNull(index)) {
            portNumber = 0;
            for (let i = dart.notNull(lastColon) + 1; dart.notNull(i) < dart.notNull(index); i = dart.notNull(i) + 1) {
              let digit = uri[dartx.codeUnitAt](i);
              if (dart.notNull(Uri._ZERO) > dart.notNull(digit) || dart.notNull(Uri._NINE) < dart.notNull(digit)) {
                Uri._fail(uri, i, "Invalid port number");
              }
              portNumber = dart.notNull(portNumber) * 10 + (dart.notNull(digit) - dart.notNull(Uri._ZERO));
            }
          }
          port = Uri._makePort(portNumber, scheme);
          hostEnd = lastColon;
        }
        host = Uri._makeHost(uri, hostStart, hostEnd, true);
        if (dart.notNull(index) < dart.notNull(uri[dartx.length])) {
          char = uri[dartx.codeUnitAt](index);
        }
      }
      dart.fn(parseAuth, dart.void, []);
      let NOT_IN_PATH = 0;
      let IN_PATH = 1;
      let ALLOW_AUTH = 2;
      let state = NOT_IN_PATH;
      let i = index;
      while (dart.notNull(i) < dart.notNull(uri[dartx.length])) {
        char = uri[dartx.codeUnitAt](i);
        if (char == Uri._QUESTION || char == Uri._NUMBER_SIGN) {
          state = NOT_IN_PATH;
          break;
        }
        if (char == Uri._SLASH) {
          state = i == 0 ? ALLOW_AUTH : IN_PATH;
          break;
        }
        if (char == Uri._COLON) {
          if (i == 0)
            Uri._fail(uri, 0, "Invalid empty scheme");
          scheme = Uri._makeScheme(uri, i);
          i = dart.notNull(i) + 1;
          pathStart = i;
          if (i == uri[dartx.length]) {
            char = EOI;
            state = NOT_IN_PATH;
          } else {
            char = uri[dartx.codeUnitAt](i);
            if (char == Uri._QUESTION || char == Uri._NUMBER_SIGN) {
              state = NOT_IN_PATH;
            } else if (char == Uri._SLASH) {
              state = ALLOW_AUTH;
            } else {
              state = IN_PATH;
            }
          }
          break;
        }
        i = dart.notNull(i) + 1;
        char = EOI;
      }
      index = i;
      if (state == ALLOW_AUTH) {
        dart.assert(char == Uri._SLASH);
        index = dart.notNull(index) + 1;
        if (index == uri[dartx.length]) {
          char = EOI;
          state = NOT_IN_PATH;
        } else {
          char = uri[dartx.codeUnitAt](index);
          if (char == Uri._SLASH) {
            index = dart.notNull(index) + 1;
            parseAuth();
            pathStart = index;
          }
          if (char == Uri._QUESTION || char == Uri._NUMBER_SIGN || char == EOI) {
            state = NOT_IN_PATH;
          } else {
            state = IN_PATH;
          }
        }
      }
      dart.assert(state == IN_PATH || state == NOT_IN_PATH);
      if (state == IN_PATH) {
        while ((index = dart.notNull(index) + 1) < dart.notNull(uri[dartx.length])) {
          char = uri[dartx.codeUnitAt](index);
          if (char == Uri._QUESTION || char == Uri._NUMBER_SIGN) {
            break;
          }
          char = EOI;
        }
        state = NOT_IN_PATH;
      }
      dart.assert(state == NOT_IN_PATH);
      let isFile = scheme == "file";
      let ensureLeadingSlash = host != null;
      path = Uri._makePath(uri, pathStart, index, null, ensureLeadingSlash, isFile);
      if (char == Uri._QUESTION) {
        let numberSignIndex = uri[dartx.indexOf]('#', dart.notNull(index) + 1);
        if (dart.notNull(numberSignIndex) < 0) {
          query = Uri._makeQuery(uri, dart.notNull(index) + 1, uri[dartx.length], null);
        } else {
          query = Uri._makeQuery(uri, dart.notNull(index) + 1, numberSignIndex, null);
          fragment = Uri._makeFragment(uri, dart.notNull(numberSignIndex) + 1, uri[dartx.length]);
        }
      } else if (char == Uri._NUMBER_SIGN) {
        fragment = Uri._makeFragment(uri, dart.notNull(index) + 1, uri[dartx.length]);
      }
      return new Uri._internal(scheme, userinfo, host, port, path, query, fragment);
    }
    static _fail(uri, index, message) {
      dart.throw(new FormatException(message, uri, index));
    }
    _internal(scheme, userInfo, host, port, path, query, fragment) {
      this.scheme = scheme;
      this[_userInfo] = userInfo;
      this[_host] = host;
      this[_port] = port;
      this[_path] = path;
      this[_query] = query;
      this[_fragment] = fragment;
      this[_pathSegments] = null;
      this[_queryParameters] = null;
    }
    static new(opts) {
      let scheme = opts && 'scheme' in opts ? opts.scheme : "";
      let userInfo = opts && 'userInfo' in opts ? opts.userInfo : "";
      let host = opts && 'host' in opts ? opts.host : null;
      let port = opts && 'port' in opts ? opts.port : null;
      let path = opts && 'path' in opts ? opts.path : null;
      let pathSegments = opts && 'pathSegments' in opts ? opts.pathSegments : null;
      let query = opts && 'query' in opts ? opts.query : null;
      let queryParameters = opts && 'queryParameters' in opts ? opts.queryParameters : null;
      let fragment = opts && 'fragment' in opts ? opts.fragment : null;
      scheme = Uri._makeScheme(scheme, Uri._stringOrNullLength(scheme));
      userInfo = Uri._makeUserInfo(userInfo, 0, Uri._stringOrNullLength(userInfo));
      host = Uri._makeHost(host, 0, Uri._stringOrNullLength(host), false);
      if (query == "")
        query = null;
      query = Uri._makeQuery(query, 0, Uri._stringOrNullLength(query), queryParameters);
      fragment = Uri._makeFragment(fragment, 0, Uri._stringOrNullLength(fragment));
      port = Uri._makePort(port, scheme);
      let isFile = scheme == "file";
      if (host == null && (dart.notNull(userInfo[dartx.isNotEmpty]) || port != null || dart.notNull(isFile))) {
        host = "";
      }
      let ensureLeadingSlash = host != null;
      path = Uri._makePath(path, 0, Uri._stringOrNullLength(path), pathSegments, ensureLeadingSlash, isFile);
      return new Uri._internal(scheme, userInfo, host, port, path, query, fragment);
    }
    static http(authority, unencodedPath, queryParameters) {
      if (queryParameters === void 0)
        queryParameters = null;
      return Uri._makeHttpUri("http", authority, unencodedPath, queryParameters);
    }
    static https(authority, unencodedPath, queryParameters) {
      if (queryParameters === void 0)
        queryParameters = null;
      return Uri._makeHttpUri("https", authority, unencodedPath, queryParameters);
    }
    static _makeHttpUri(scheme, authority, unencodedPath, queryParameters) {
      let userInfo = "";
      let host = null;
      let port = null;
      if (authority != null && dart.notNull(authority[dartx.isNotEmpty])) {
        let hostStart = 0;
        let hasUserInfo = false;
        for (let i = 0; dart.notNull(i) < dart.notNull(authority[dartx.length]); i = dart.notNull(i) + 1) {
          if (authority[dartx.codeUnitAt](i) == Uri._AT_SIGN) {
            hasUserInfo = true;
            userInfo = authority[dartx.substring](0, i);
            hostStart = dart.notNull(i) + 1;
            break;
          }
        }
        let hostEnd = hostStart;
        if (dart.notNull(hostStart) < dart.notNull(authority[dartx.length]) && authority[dartx.codeUnitAt](hostStart) == Uri._LEFT_BRACKET) {
          for (; dart.notNull(hostEnd) < dart.notNull(authority[dartx.length]); hostEnd = dart.notNull(hostEnd) + 1) {
            if (authority[dartx.codeUnitAt](hostEnd) == Uri._RIGHT_BRACKET)
              break;
          }
          if (hostEnd == authority[dartx.length]) {
            dart.throw(new FormatException("Invalid IPv6 host entry.", authority, hostStart));
          }
          Uri.parseIPv6Address(authority, dart.notNull(hostStart) + 1, hostEnd);
          hostEnd = dart.notNull(hostEnd) + 1;
          if (hostEnd != authority[dartx.length] && authority[dartx.codeUnitAt](hostEnd) != Uri._COLON) {
            dart.throw(new FormatException("Invalid end of authority", authority, hostEnd));
          }
        }
        let hasPort = false;
        for (; dart.notNull(hostEnd) < dart.notNull(authority[dartx.length]); hostEnd = dart.notNull(hostEnd) + 1) {
          if (authority[dartx.codeUnitAt](hostEnd) == Uri._COLON) {
            let portString = authority[dartx.substring](dart.notNull(hostEnd) + 1);
            if (dart.notNull(portString[dartx.isNotEmpty]))
              port = int.parse(portString);
            break;
          }
        }
        host = authority[dartx.substring](hostStart, hostEnd);
      }
      return Uri.new({scheme: scheme, userInfo: userInfo, host: dart.as(host, String), port: dart.as(port, int), pathSegments: unencodedPath[dartx.split]("/"), queryParameters: queryParameters});
    }
    static file(path, opts) {
      let windows = opts && 'windows' in opts ? opts.windows : null;
      windows = windows == null ? Uri._isWindows : windows;
      return dart.notNull(windows) ? dart.as(Uri._makeWindowsFileUrl(path), Uri) : dart.as(Uri._makeFileUri(path), Uri);
    }
    static get base() {
      let uri = _js_helper.Primitives.currentUri();
      if (uri != null)
        return Uri.parse(uri);
      dart.throw(new UnsupportedError("'Uri.base' is not supported"));
    }
    static get _isWindows() {
      return false;
    }
    static _checkNonWindowsPathReservedCharacters(segments, argumentError) {
      segments[dartx.forEach](dart.fn(segment => {
        if (dart.notNull(dart.as(dart.dsend(segment, 'contains', "/"), bool))) {
          if (dart.notNull(argumentError)) {
            dart.throw(new ArgumentError(`Illegal path character ${segment}`));
          } else {
            dart.throw(new UnsupportedError(`Illegal path character ${segment}`));
          }
        }
      }));
    }
    static _checkWindowsPathReservedCharacters(segments, argumentError, firstSegment) {
      if (firstSegment === void 0)
        firstSegment = 0;
      segments[dartx.skip](firstSegment)[dartx.forEach](dart.fn(segment => {
        if (dart.notNull(dart.as(dart.dsend(segment, 'contains', RegExp.new('["*/:<>?\\\\|]')), bool))) {
          if (dart.notNull(argumentError)) {
            dart.throw(new ArgumentError("Illegal character in path"));
          } else {
            dart.throw(new UnsupportedError("Illegal character in path"));
          }
        }
      }));
    }
    static _checkWindowsDriveLetter(charCode, argumentError) {
      if (dart.notNull(Uri._UPPER_CASE_A) <= dart.notNull(charCode) && dart.notNull(charCode) <= dart.notNull(Uri._UPPER_CASE_Z) || dart.notNull(Uri._LOWER_CASE_A) <= dart.notNull(charCode) && dart.notNull(charCode) <= dart.notNull(Uri._LOWER_CASE_Z)) {
        return;
      }
      if (dart.notNull(argumentError)) {
        dart.throw(new ArgumentError("Illegal drive letter " + dart.notNull(String.fromCharCode(charCode))));
      } else {
        dart.throw(new UnsupportedError("Illegal drive letter " + dart.notNull(String.fromCharCode(charCode))));
      }
    }
    static _makeFileUri(path) {
      let sep = "/";
      if (dart.notNull(path[dartx.startsWith](sep))) {
        return Uri.new({scheme: "file", pathSegments: path[dartx.split](sep)});
      } else {
        return Uri.new({pathSegments: path[dartx.split](sep)});
      }
    }
    static _makeWindowsFileUrl(path) {
      if (dart.notNull(path[dartx.startsWith]("\\\\?\\"))) {
        if (dart.notNull(path[dartx.startsWith]("\\\\?\\UNC\\"))) {
          path = `\\${path[dartx.substring](7)}`;
        } else {
          path = path[dartx.substring](4);
          if (dart.notNull(path[dartx.length]) < 3 || path[dartx.codeUnitAt](1) != Uri._COLON || path[dartx.codeUnitAt](2) != Uri._BACKSLASH) {
            dart.throw(new ArgumentError("Windows paths with \\\\?\\ prefix must be absolute"));
          }
        }
      } else {
        path = path[dartx.replaceAll]("/", "\\");
      }
      let sep = "\\";
      if (dart.notNull(path[dartx.length]) > 1 && path[dartx.get](1) == ":") {
        Uri._checkWindowsDriveLetter(path[dartx.codeUnitAt](0), true);
        if (path[dartx.length] == 2 || path[dartx.codeUnitAt](2) != Uri._BACKSLASH) {
          dart.throw(new ArgumentError("Windows paths with drive letter must be absolute"));
        }
        let pathSegments = path[dartx.split](sep);
        Uri._checkWindowsPathReservedCharacters(pathSegments, true, 1);
        return Uri.new({scheme: "file", pathSegments: pathSegments});
      }
      if (dart.notNull(path[dartx.length]) > 0 && path[dartx.get](0) == sep) {
        if (dart.notNull(path[dartx.length]) > 1 && path[dartx.get](1) == sep) {
          let pathStart = path[dartx.indexOf]("\\", 2);
          let hostPart = pathStart == -1 ? path[dartx.substring](2) : path[dartx.substring](2, pathStart);
          let pathPart = pathStart == -1 ? "" : path[dartx.substring](dart.notNull(pathStart) + 1);
          let pathSegments = pathPart[dartx.split](sep);
          Uri._checkWindowsPathReservedCharacters(pathSegments, true);
          return Uri.new({scheme: "file", host: hostPart, pathSegments: pathSegments});
        } else {
          let pathSegments = path[dartx.split](sep);
          Uri._checkWindowsPathReservedCharacters(pathSegments, true);
          return Uri.new({scheme: "file", pathSegments: pathSegments});
        }
      } else {
        let pathSegments = path[dartx.split](sep);
        Uri._checkWindowsPathReservedCharacters(pathSegments, true);
        return Uri.new({pathSegments: pathSegments});
      }
    }
    replace(opts) {
      let scheme = opts && 'scheme' in opts ? opts.scheme : null;
      let userInfo = opts && 'userInfo' in opts ? opts.userInfo : null;
      let host = opts && 'host' in opts ? opts.host : null;
      let port = opts && 'port' in opts ? opts.port : null;
      let path = opts && 'path' in opts ? opts.path : null;
      let pathSegments = opts && 'pathSegments' in opts ? opts.pathSegments : null;
      let query = opts && 'query' in opts ? opts.query : null;
      let queryParameters = opts && 'queryParameters' in opts ? opts.queryParameters : null;
      let fragment = opts && 'fragment' in opts ? opts.fragment : null;
      let schemeChanged = false;
      if (scheme != null) {
        scheme = Uri._makeScheme(scheme, scheme[dartx.length]);
        schemeChanged = true;
      } else {
        scheme = this.scheme;
      }
      let isFile = scheme == "file";
      if (userInfo != null) {
        userInfo = Uri._makeUserInfo(userInfo, 0, userInfo[dartx.length]);
      } else {
        userInfo = this.userInfo;
      }
      if (port != null) {
        port = Uri._makePort(port, scheme);
      } else {
        port = dart.asInt(this[_port]);
        if (dart.notNull(schemeChanged)) {
          port = Uri._makePort(port, scheme);
        }
      }
      if (host != null) {
        host = Uri._makeHost(host, 0, host[dartx.length], false);
      } else if (dart.notNull(this.hasAuthority)) {
        host = this.host;
      } else if (dart.notNull(userInfo[dartx.isNotEmpty]) || port != null || dart.notNull(isFile)) {
        host = "";
      }
      let ensureLeadingSlash = host != null;
      if (path != null || pathSegments != null) {
        path = Uri._makePath(path, 0, Uri._stringOrNullLength(path), pathSegments, ensureLeadingSlash, isFile);
      } else {
        path = this.path;
        if ((dart.notNull(isFile) || dart.notNull(ensureLeadingSlash) && !dart.notNull(path[dartx.isEmpty])) && !dart.notNull(path[dartx.startsWith]('/'))) {
          path = `/${path}`;
        }
      }
      if (query != null || queryParameters != null) {
        query = Uri._makeQuery(query, 0, Uri._stringOrNullLength(query), queryParameters);
      } else if (dart.notNull(this.hasQuery)) {
        query = this.query;
      }
      if (fragment != null) {
        fragment = Uri._makeFragment(fragment, 0, fragment[dartx.length]);
      } else if (dart.notNull(this.hasFragment)) {
        fragment = this.fragment;
      }
      return new Uri._internal(scheme, userInfo, host, port, path, query, fragment);
    }
    get pathSegments() {
      if (this[_pathSegments] == null) {
        let pathToSplit = !dart.notNull(this.path[dartx.isEmpty]) && this.path[dartx.codeUnitAt](0) == Uri._SLASH ? this.path[dartx.substring](1) : this.path;
        this[_pathSegments] = new (collection.UnmodifiableListView$(String))(pathToSplit == "" ? dart.const(dart.list([], String)) : List$(String).from(pathToSplit[dartx.split]("/")[dartx.map](Uri.decodeComponent), {growable: false}));
      }
      return this[_pathSegments];
    }
    get queryParameters() {
      if (this[_queryParameters] == null) {
        this[_queryParameters] = new (collection.UnmodifiableMapView$(String, String))(Uri.splitQueryString(this.query));
      }
      return this[_queryParameters];
    }
    static _makePort(port, scheme) {
      if (port != null && port == Uri._defaultPort(scheme))
        return null;
      return port;
    }
    static _makeHost(host, start, end, strictIPv6) {
      if (host == null)
        return null;
      if (start == end)
        return "";
      if (host[dartx.codeUnitAt](start) == Uri._LEFT_BRACKET) {
        if (host[dartx.codeUnitAt](dart.notNull(end) - 1) != Uri._RIGHT_BRACKET) {
          Uri._fail(host, start, 'Missing end `]` to match `[` in host');
        }
        Uri.parseIPv6Address(host, dart.notNull(start) + 1, dart.notNull(end) - 1);
        return host[dartx.substring](start, end)[dartx.toLowerCase]();
      }
      if (!dart.notNull(strictIPv6)) {
        for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
          if (host[dartx.codeUnitAt](i) == Uri._COLON) {
            Uri.parseIPv6Address(host, start, end);
            return `[${host}]`;
          }
        }
      }
      return Uri._normalizeRegName(host, start, end);
    }
    static _isRegNameChar(char) {
      return dart.notNull(char) < 127 && !dart.equals(dart.dsend(Uri._regNameTable[dartx.get](dart.notNull(char) >> 4), '&', 1 << (dart.notNull(char) & 15)), 0);
    }
    static _normalizeRegName(host, start, end) {
      let buffer = null;
      let sectionStart = start;
      let index = start;
      let isNormalized = true;
      while (dart.notNull(index) < dart.notNull(end)) {
        let char = host[dartx.codeUnitAt](index);
        if (char == Uri._PERCENT) {
          let replacement = Uri._normalizeEscape(host, index, true);
          if (replacement == null && dart.notNull(isNormalized)) {
            index = dart.notNull(index) + 3;
            continue;
          }
          if (buffer == null)
            buffer = new StringBuffer();
          let slice = host[dartx.substring](sectionStart, index);
          if (!dart.notNull(isNormalized))
            slice = slice[dartx.toLowerCase]();
          buffer.write(slice);
          let sourceLength = 3;
          if (replacement == null) {
            replacement = host[dartx.substring](index, dart.notNull(index) + 3);
          } else if (replacement == "%") {
            replacement = "%25";
            sourceLength = 1;
          }
          buffer.write(replacement);
          index = dart.notNull(index) + dart.notNull(sourceLength);
          sectionStart = index;
          isNormalized = true;
        } else if (dart.notNull(Uri._isRegNameChar(char))) {
          if (dart.notNull(isNormalized) && dart.notNull(Uri._UPPER_CASE_A) <= dart.notNull(char) && dart.notNull(Uri._UPPER_CASE_Z) >= dart.notNull(char)) {
            if (buffer == null)
              buffer = new StringBuffer();
            if (dart.notNull(sectionStart) < dart.notNull(index)) {
              buffer.write(host[dartx.substring](sectionStart, index));
              sectionStart = index;
            }
            isNormalized = false;
          }
          index = dart.notNull(index) + 1;
        } else if (dart.notNull(Uri._isGeneralDelimiter(char))) {
          Uri._fail(host, index, "Invalid character");
        } else {
          let sourceLength = 1;
          if ((dart.notNull(char) & 64512) == 55296 && dart.notNull(index) + 1 < dart.notNull(end)) {
            let tail = host[dartx.codeUnitAt](dart.notNull(index) + 1);
            if ((dart.notNull(tail) & 64512) == 56320) {
              char = 65536 | (dart.notNull(char) & 1023) << 10 | dart.notNull(tail) & 1023;
              sourceLength = 2;
            }
          }
          if (buffer == null)
            buffer = new StringBuffer();
          let slice = host[dartx.substring](sectionStart, index);
          if (!dart.notNull(isNormalized))
            slice = slice[dartx.toLowerCase]();
          buffer.write(slice);
          buffer.write(Uri._escapeChar(char));
          index = dart.notNull(index) + dart.notNull(sourceLength);
          sectionStart = index;
        }
      }
      if (buffer == null)
        return host[dartx.substring](start, end);
      if (dart.notNull(sectionStart) < dart.notNull(end)) {
        let slice = host[dartx.substring](sectionStart, end);
        if (!dart.notNull(isNormalized))
          slice = slice[dartx.toLowerCase]();
        buffer.write(slice);
      }
      return dart.toString(buffer);
    }
    static _makeScheme(scheme, end) {
      if (end == 0)
        return "";
      let firstCodeUnit = scheme[dartx.codeUnitAt](0);
      if (!dart.notNull(Uri._isAlphabeticCharacter(firstCodeUnit))) {
        Uri._fail(scheme, 0, "Scheme not starting with alphabetic character");
      }
      let allLowercase = dart.notNull(firstCodeUnit) >= dart.notNull(Uri._LOWER_CASE_A);
      for (let i = 0; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        let codeUnit = scheme[dartx.codeUnitAt](i);
        if (!dart.notNull(Uri._isSchemeCharacter(codeUnit))) {
          Uri._fail(scheme, i, "Illegal scheme character");
        }
        if (dart.notNull(codeUnit) < dart.notNull(Uri._LOWER_CASE_A) || dart.notNull(codeUnit) > dart.notNull(Uri._LOWER_CASE_Z)) {
          allLowercase = false;
        }
      }
      scheme = scheme[dartx.substring](0, end);
      if (!dart.notNull(allLowercase))
        scheme = scheme[dartx.toLowerCase]();
      return scheme;
    }
    static _makeUserInfo(userInfo, start, end) {
      if (userInfo == null)
        return "";
      return Uri._normalize(userInfo, start, end, dart.as(Uri._userinfoTable, List$(int)));
    }
    static _makePath(path, start, end, pathSegments, ensureLeadingSlash, isFile) {
      if (path == null && pathSegments == null)
        return dart.notNull(isFile) ? "/" : "";
      if (path != null && pathSegments != null) {
        dart.throw(new ArgumentError('Both path and pathSegments specified'));
      }
      let result = null;
      if (path != null) {
        result = Uri._normalize(path, start, end, dart.as(Uri._pathCharOrSlashTable, List$(int)));
      } else {
        result = pathSegments[dartx.map](dart.fn(s => Uri._uriEncode(dart.as(Uri._pathCharTable, List$(int)), dart.as(s, String)), String, [dart.dynamic]))[dartx.join]("/");
      }
      if (dart.notNull(dart.as(dart.dload(result, 'isEmpty'), bool))) {
        if (dart.notNull(isFile))
          return "/";
      } else if ((dart.notNull(isFile) || dart.notNull(ensureLeadingSlash)) && !dart.equals(dart.dsend(result, 'codeUnitAt', 0), Uri._SLASH)) {
        return `/${result}`;
      }
      return dart.as(result, String);
    }
    static _makeQuery(query, start, end, queryParameters) {
      if (query == null && queryParameters == null)
        return null;
      if (query != null && queryParameters != null) {
        dart.throw(new ArgumentError('Both query and queryParameters specified'));
      }
      if (query != null)
        return Uri._normalize(query, start, end, dart.as(Uri._queryCharTable, List$(int)));
      let result = new StringBuffer();
      let first = true;
      queryParameters.forEach(dart.fn((key, value) => {
        if (!dart.notNull(first)) {
          result.write("&");
        }
        first = false;
        result.write(Uri.encodeQueryComponent(dart.as(key, String)));
        if (value != null && !dart.notNull(dart.as(dart.dload(value, 'isEmpty'), bool))) {
          result.write("=");
          result.write(Uri.encodeQueryComponent(dart.as(value, String)));
        }
      }));
      return dart.toString(result);
    }
    static _makeFragment(fragment, start, end) {
      if (fragment == null)
        return null;
      return Uri._normalize(fragment, start, end, dart.as(Uri._queryCharTable, List$(int)));
    }
    static _stringOrNullLength(s) {
      return s == null ? 0 : s[dartx.length];
    }
    static _isHexDigit(char) {
      if (dart.notNull(Uri._NINE) >= dart.notNull(char))
        return dart.notNull(Uri._ZERO) <= dart.notNull(char);
      char = dart.notNull(char) | 32;
      return dart.notNull(Uri._LOWER_CASE_A) <= dart.notNull(char) && dart.notNull(Uri._LOWER_CASE_F) >= dart.notNull(char);
    }
    static _hexValue(char) {
      dart.assert(Uri._isHexDigit(char));
      if (dart.notNull(Uri._NINE) >= dart.notNull(char))
        return dart.notNull(char) - dart.notNull(Uri._ZERO);
      char = dart.notNull(char) | 32;
      return dart.notNull(char) - (dart.notNull(Uri._LOWER_CASE_A) - 10);
    }
    static _normalizeEscape(source, index, lowerCase) {
      dart.assert(source[dartx.codeUnitAt](index) == Uri._PERCENT);
      if (dart.notNull(index) + 2 >= dart.notNull(source[dartx.length])) {
        return "%";
      }
      let firstDigit = source[dartx.codeUnitAt](dart.notNull(index) + 1);
      let secondDigit = source[dartx.codeUnitAt](dart.notNull(index) + 2);
      if (!dart.notNull(Uri._isHexDigit(firstDigit)) || !dart.notNull(Uri._isHexDigit(secondDigit))) {
        return "%";
      }
      let value = dart.notNull(Uri._hexValue(firstDigit)) * 16 + dart.notNull(Uri._hexValue(secondDigit));
      if (dart.notNull(Uri._isUnreservedChar(value))) {
        if (dart.notNull(lowerCase) && dart.notNull(Uri._UPPER_CASE_A) <= dart.notNull(value) && dart.notNull(Uri._UPPER_CASE_Z) >= dart.notNull(value)) {
          value = dart.notNull(value) | 32;
        }
        return String.fromCharCode(value);
      }
      if (dart.notNull(firstDigit) >= dart.notNull(Uri._LOWER_CASE_A) || dart.notNull(secondDigit) >= dart.notNull(Uri._LOWER_CASE_A)) {
        return source[dartx.substring](index, dart.notNull(index) + 3)[dartx.toUpperCase]();
      }
      return null;
    }
    static _isUnreservedChar(ch) {
      return dart.notNull(ch) < 127 && !dart.equals(dart.dsend(Uri._unreservedTable[dartx.get](dart.notNull(ch) >> 4), '&', 1 << (dart.notNull(ch) & 15)), 0);
    }
    static _escapeChar(char) {
      dart.assert(dart.dsend(char, '<=', 1114111));
      let hexDigits = "0123456789ABCDEF";
      let codeUnits = null;
      if (dart.notNull(dart.as(dart.dsend(char, '<', 128), bool))) {
        codeUnits = List.new(3);
        codeUnits[dartx.set](0, Uri._PERCENT);
        codeUnits[dartx.set](1, hexDigits[dartx.codeUnitAt](dart.as(dart.dsend(char, '>>', 4), int)));
        codeUnits[dartx.set](2, hexDigits[dartx.codeUnitAt](dart.as(dart.dsend(char, '&', 15), int)));
      } else {
        let flag = 192;
        let encodedBytes = 2;
        if (dart.notNull(dart.as(dart.dsend(char, '>', 2047), bool))) {
          flag = 224;
          encodedBytes = 3;
          if (dart.notNull(dart.as(dart.dsend(char, '>', 65535), bool))) {
            encodedBytes = 4;
            flag = 240;
          }
        }
        codeUnits = List.new(3 * dart.notNull(encodedBytes));
        let index = 0;
        while ((encodedBytes = dart.notNull(encodedBytes) - 1) >= 0) {
          let byte = dart.as(dart.dsend(dart.dsend(dart.dsend(char, '>>', 6 * dart.notNull(encodedBytes)), '&', 63), '|', flag), int);
          codeUnits[dartx.set](index, Uri._PERCENT);
          codeUnits[dartx.set](dart.notNull(index) + 1, hexDigits[dartx.codeUnitAt](dart.notNull(byte) >> 4));
          codeUnits[dartx.set](dart.notNull(index) + 2, hexDigits[dartx.codeUnitAt](dart.notNull(byte) & 15));
          index = dart.notNull(index) + 3;
          flag = 128;
        }
      }
      return String.fromCharCodes(dart.as(codeUnits, Iterable$(int)));
    }
    static _normalize(component, start, end, charTable) {
      let buffer = null;
      let sectionStart = start;
      let index = start;
      while (dart.notNull(index) < dart.notNull(end)) {
        let char = component[dartx.codeUnitAt](index);
        if (dart.notNull(char) < 127 && (dart.notNull(charTable[dartx.get](dart.notNull(char) >> 4)) & 1 << (dart.notNull(char) & 15)) != 0) {
          index = dart.notNull(index) + 1;
        } else {
          let replacement = null;
          let sourceLength = null;
          if (char == Uri._PERCENT) {
            replacement = Uri._normalizeEscape(component, index, false);
            if (replacement == null) {
              index = dart.notNull(index) + 3;
              continue;
            }
            if ("%" == replacement) {
              replacement = "%25";
              sourceLength = 1;
            } else {
              sourceLength = 3;
            }
          } else if (dart.notNull(Uri._isGeneralDelimiter(char))) {
            Uri._fail(component, index, "Invalid character");
          } else {
            sourceLength = 1;
            if ((dart.notNull(char) & 64512) == 55296) {
              if (dart.notNull(index) + 1 < dart.notNull(end)) {
                let tail = component[dartx.codeUnitAt](dart.notNull(index) + 1);
                if ((dart.notNull(tail) & 64512) == 56320) {
                  sourceLength = 2;
                  char = 65536 | (dart.notNull(char) & 1023) << 10 | dart.notNull(tail) & 1023;
                }
              }
            }
            replacement = Uri._escapeChar(char);
          }
          if (buffer == null)
            buffer = new StringBuffer();
          buffer.write(component[dartx.substring](sectionStart, index));
          buffer.write(replacement);
          index = dart.notNull(index) + dart.notNull(sourceLength);
          sectionStart = index;
        }
      }
      if (buffer == null) {
        return component[dartx.substring](start, end);
      }
      if (dart.notNull(sectionStart) < dart.notNull(end)) {
        buffer.write(component[dartx.substring](sectionStart, end));
      }
      return dart.toString(buffer);
    }
    static _isSchemeCharacter(ch) {
      return dart.notNull(ch) < 128 && !dart.equals(dart.dsend(Uri._schemeTable[dartx.get](dart.notNull(ch) >> 4), '&', 1 << (dart.notNull(ch) & 15)), 0);
    }
    static _isGeneralDelimiter(ch) {
      return dart.notNull(ch) <= dart.notNull(Uri._RIGHT_BRACKET) && !dart.equals(dart.dsend(Uri._genDelimitersTable[dartx.get](dart.notNull(ch) >> 4), '&', 1 << (dart.notNull(ch) & 15)), 0);
    }
    get isAbsolute() {
      return this.scheme != "" && this.fragment == "";
    }
    [_merge](base, reference) {
      if (dart.notNull(base[dartx.isEmpty]))
        return `/${reference}`;
      let backCount = 0;
      let refStart = 0;
      while (dart.notNull(reference[dartx.startsWith]("../", refStart))) {
        refStart = dart.notNull(refStart) + 3;
        backCount = dart.notNull(backCount) + 1;
      }
      let baseEnd = base[dartx.lastIndexOf]('/');
      while (dart.notNull(baseEnd) > 0 && dart.notNull(backCount) > 0) {
        let newEnd = base[dartx.lastIndexOf]('/', dart.notNull(baseEnd) - 1);
        if (dart.notNull(newEnd) < 0) {
          break;
        }
        let delta = dart.notNull(baseEnd) - dart.notNull(newEnd);
        if ((delta == 2 || delta == 3) && base[dartx.codeUnitAt](dart.notNull(newEnd) + 1) == Uri._DOT && (delta == 2 || base[dartx.codeUnitAt](dart.notNull(newEnd) + 2) == Uri._DOT)) {
          break;
        }
        baseEnd = newEnd;
        backCount = dart.notNull(backCount) - 1;
      }
      return dart.notNull(base[dartx.substring](0, dart.notNull(baseEnd) + 1)) + dart.notNull(reference[dartx.substring](dart.notNull(refStart) - 3 * dart.notNull(backCount)));
    }
    [_hasDotSegments](path) {
      if (dart.notNull(path[dartx.length]) > 0 && path[dartx.codeUnitAt](0) == Uri._DOT)
        return true;
      let index = path[dartx.indexOf]("/.");
      return index != -1;
    }
    [_removeDotSegments](path) {
      if (!dart.notNull(this[_hasDotSegments](path)))
        return path;
      let output = dart.list([], String);
      let appendSlash = false;
      for (let segment of path[dartx.split]("/")) {
        appendSlash = false;
        if (segment == "..") {
          if (!dart.notNull(output[dartx.isEmpty]) && (output[dartx.length] != 1 || output[dartx.get](0) != ""))
            output[dartx.removeLast]();
          appendSlash = true;
        } else if ("." == segment) {
          appendSlash = true;
        } else {
          output[dartx.add](segment);
        }
      }
      if (dart.notNull(appendSlash))
        output[dartx.add]("");
      return output[dartx.join]("/");
    }
    resolve(reference) {
      return this.resolveUri(Uri.parse(reference));
    }
    resolveUri(reference) {
      let targetScheme = null;
      let targetUserInfo = "";
      let targetHost = null;
      let targetPort = null;
      let targetPath = null;
      let targetQuery = null;
      if (dart.notNull(reference.scheme[dartx.isNotEmpty])) {
        targetScheme = reference.scheme;
        if (dart.notNull(reference.hasAuthority)) {
          targetUserInfo = reference.userInfo;
          targetHost = reference.host;
          targetPort = dart.notNull(reference.hasPort) ? reference.port : null;
        }
        targetPath = this[_removeDotSegments](reference.path);
        if (dart.notNull(reference.hasQuery)) {
          targetQuery = reference.query;
        }
      } else {
        targetScheme = this.scheme;
        if (dart.notNull(reference.hasAuthority)) {
          targetUserInfo = reference.userInfo;
          targetHost = reference.host;
          targetPort = Uri._makePort(dart.notNull(reference.hasPort) ? reference.port : null, targetScheme);
          targetPath = this[_removeDotSegments](reference.path);
          if (dart.notNull(reference.hasQuery))
            targetQuery = reference.query;
        } else {
          if (reference.path == "") {
            targetPath = this[_path];
            if (dart.notNull(reference.hasQuery)) {
              targetQuery = reference.query;
            } else {
              targetQuery = this[_query];
            }
          } else {
            if (dart.notNull(reference.path[dartx.startsWith]("/"))) {
              targetPath = this[_removeDotSegments](reference.path);
            } else {
              targetPath = this[_removeDotSegments](this[_merge](this[_path], reference.path));
            }
            if (dart.notNull(reference.hasQuery))
              targetQuery = reference.query;
          }
          targetUserInfo = this[_userInfo];
          targetHost = this[_host];
          targetPort = dart.asInt(this[_port]);
        }
      }
      let fragment = dart.notNull(reference.hasFragment) ? reference.fragment : null;
      return new Uri._internal(targetScheme, targetUserInfo, targetHost, targetPort, targetPath, targetQuery, fragment);
    }
    get hasAuthority() {
      return this[_host] != null;
    }
    get hasPort() {
      return this[_port] != null;
    }
    get hasQuery() {
      return this[_query] != null;
    }
    get hasFragment() {
      return this[_fragment] != null;
    }
    get origin() {
      if (this.scheme == "" || this[_host] == null || this[_host] == "") {
        dart.throw(new StateError(`Cannot use origin without a scheme: ${this}`));
      }
      if (this.scheme != "http" && this.scheme != "https") {
        dart.throw(new StateError(`Origin is only applicable schemes http and https: ${this}`));
      }
      if (this[_port] == null)
        return `${this.scheme}://${this[_host]}`;
      return `${this.scheme}://${this[_host]}:${this[_port]}`;
    }
    toFilePath(opts) {
      let windows = opts && 'windows' in opts ? opts.windows : null;
      if (this.scheme != "" && this.scheme != "file") {
        dart.throw(new UnsupportedError(`Cannot extract a file path from a ${this.scheme} URI`));
      }
      if (this.query != "") {
        dart.throw(new UnsupportedError("Cannot extract a file path from a URI with a query component"));
      }
      if (this.fragment != "") {
        dart.throw(new UnsupportedError("Cannot extract a file path from a URI with a fragment component"));
      }
      if (windows == null)
        windows = Uri._isWindows;
      return dart.notNull(windows) ? this[_toWindowsFilePath]() : this[_toFilePath]();
    }
    [_toFilePath]() {
      if (this.host != "") {
        dart.throw(new UnsupportedError("Cannot extract a non-Windows file path from a file URI " + "with an authority"));
      }
      Uri._checkNonWindowsPathReservedCharacters(this.pathSegments, false);
      let result = new StringBuffer();
      if (dart.notNull(this[_isPathAbsolute]))
        result.write("/");
      result.writeAll(this.pathSegments, "/");
      return dart.toString(result);
    }
    [_toWindowsFilePath]() {
      let hasDriveLetter = false;
      let segments = this.pathSegments;
      if (dart.notNull(segments[dartx.length]) > 0 && segments[dartx.get](0)[dartx.length] == 2 && segments[dartx.get](0)[dartx.codeUnitAt](1) == Uri._COLON) {
        Uri._checkWindowsDriveLetter(segments[dartx.get](0)[dartx.codeUnitAt](0), false);
        Uri._checkWindowsPathReservedCharacters(segments, false, 1);
        hasDriveLetter = true;
      } else {
        Uri._checkWindowsPathReservedCharacters(segments, false);
      }
      let result = new StringBuffer();
      if (dart.notNull(this[_isPathAbsolute]) && !dart.notNull(hasDriveLetter))
        result.write("\\");
      if (this.host != "") {
        result.write("\\");
        result.write(this.host);
        result.write("\\");
      }
      result.writeAll(segments, "\\");
      if (dart.notNull(hasDriveLetter) && segments[dartx.length] == 1)
        result.write("\\");
      return dart.toString(result);
    }
    get [_isPathAbsolute]() {
      if (this.path == null || dart.notNull(this.path[dartx.isEmpty]))
        return false;
      return this.path[dartx.startsWith]('/');
    }
    [_writeAuthority](ss) {
      if (dart.notNull(this[_userInfo][dartx.isNotEmpty])) {
        ss.write(this[_userInfo]);
        ss.write("@");
      }
      if (this[_host] != null)
        ss.write(this[_host]);
      if (this[_port] != null) {
        ss.write(":");
        ss.write(this[_port]);
      }
    }
    toString() {
      let sb = new StringBuffer();
      Uri._addIfNonEmpty(sb, this.scheme, this.scheme, ':');
      if (dart.notNull(this.hasAuthority) || dart.notNull(this.path[dartx.startsWith]("//")) || this.scheme == "file") {
        sb.write("//");
        this[_writeAuthority](sb);
      }
      sb.write(this.path);
      if (this[_query] != null) {
        sb.write("?");
        sb.write(this[_query]);
      }
      if (this[_fragment] != null) {
        sb.write("#");
        sb.write(this[_fragment]);
      }
      return dart.toString(sb);
    }
    ['=='](other) {
      if (!dart.is(other, Uri))
        return false;
      let uri = dart.as(other, Uri);
      return this.scheme == uri.scheme && this.hasAuthority == uri.hasAuthority && this.userInfo == uri.userInfo && this.host == uri.host && this.port == uri.port && this.path == uri.path && this.hasQuery == uri.hasQuery && this.query == uri.query && this.hasFragment == uri.hasFragment && this.fragment == uri.fragment;
    }
    get hashCode() {
      function combine(part, current) {
        return dart.as(dart.dsend(dart.dsend(dart.dsend(current, '*', 31), '+', dart.hashCode(part)), '&', 1073741823), int);
      }
      dart.fn(combine, int, [dart.dynamic, dart.dynamic]);
      return combine(this.scheme, combine(this.userInfo, combine(this.host, combine(this.port, combine(this.path, combine(this.query, combine(this.fragment, 1)))))));
    }
    static _addIfNonEmpty(sb, test, first, second) {
      if ("" != test) {
        sb.write(first);
        sb.write(second);
      }
    }
    static encodeComponent(component) {
      return Uri._uriEncode(dart.as(Uri._unreserved2396Table, List$(int)), component);
    }
    static encodeQueryComponent(component, opts) {
      let encoding = opts && 'encoding' in opts ? opts.encoding : convert.UTF8;
      return Uri._uriEncode(dart.as(Uri._unreservedTable, List$(int)), component, {encoding: encoding, spaceToPlus: true});
    }
    static decodeComponent(encodedComponent) {
      return Uri._uriDecode(encodedComponent);
    }
    static decodeQueryComponent(encodedComponent, opts) {
      let encoding = opts && 'encoding' in opts ? opts.encoding : convert.UTF8;
      return Uri._uriDecode(encodedComponent, {plusToSpace: true, encoding: encoding});
    }
    static encodeFull(uri) {
      return Uri._uriEncode(dart.as(Uri._encodeFullTable, List$(int)), uri);
    }
    static decodeFull(uri) {
      return Uri._uriDecode(uri);
    }
    static splitQueryString(query, opts) {
      let encoding = opts && 'encoding' in opts ? opts.encoding : convert.UTF8;
      return dart.as(query[dartx.split]("&")[dartx.fold](dart.map(), dart.fn((map, element) => {
        let index = dart.as(dart.dsend(element, 'indexOf', "="), int);
        if (index == -1) {
          if (!dart.equals(element, "")) {
            dart.dsetindex(map, Uri.decodeQueryComponent(dart.as(element, String), {encoding: encoding}), "");
          }
        } else if (index != 0) {
          let key = dart.dsend(element, 'substring', 0, index);
          let value = dart.dsend(element, 'substring', dart.notNull(index) + 1);
          dart.dsetindex(map, Uri.decodeQueryComponent(dart.as(key, String), {encoding: encoding}), Uri.decodeQueryComponent(dart.as(value, String), {encoding: encoding}));
        }
        return map;
      })), Map$(String, String));
    }
    static parseIPv4Address(host) {
      function error(msg) {
        dart.throw(new FormatException(`Illegal IPv4 address, ${msg}`));
      }
      dart.fn(error, dart.void, [String]);
      let bytes = host[dartx.split]('.');
      if (bytes[dartx.length] != 4) {
        error('IPv4 address should contain exactly 4 parts');
      }
      return dart.as(bytes[dartx.map](dart.fn(byteString => {
        let byte = int.parse(dart.as(byteString, String));
        if (dart.notNull(byte) < 0 || dart.notNull(byte) > 255) {
          error('each part must be in the range of `0..255`');
        }
        return byte;
      }))[dartx.toList](), List$(int));
    }
    static parseIPv6Address(host, start, end) {
      if (start === void 0)
        start = 0;
      if (end === void 0)
        end = null;
      if (end == null)
        end = host[dartx.length];
      function error(msg, position) {
        if (position === void 0)
          position = null;
        dart.throw(new FormatException(`Illegal IPv6 address, ${msg}`, host, dart.as(position, int)));
      }
      dart.fn(error, dart.void, [String], [dart.dynamic]);
      function parseHex(start, end) {
        if (dart.notNull(end) - dart.notNull(start) > 4) {
          error('an IPv6 part can only contain a maximum of 4 hex digits', start);
        }
        let value = int.parse(host[dartx.substring](start, end), {radix: 16});
        if (dart.notNull(value) < 0 || dart.notNull(value) > (1 << 16) - 1) {
          error('each part must be in the range of `0x0..0xFFFF`', start);
        }
        return value;
      }
      dart.fn(parseHex, int, [int, int]);
      if (dart.notNull(host[dartx.length]) < 2)
        error('address is too short');
      let parts = dart.list([], int);
      let wildcardSeen = false;
      let partStart = start;
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        if (host[dartx.codeUnitAt](i) == Uri._COLON) {
          if (i == start) {
            i = dart.notNull(i) + 1;
            if (host[dartx.codeUnitAt](i) != Uri._COLON) {
              error('invalid start colon.', i);
            }
            partStart = i;
          }
          if (i == partStart) {
            if (dart.notNull(wildcardSeen)) {
              error('only one wildcard `::` is allowed', i);
            }
            wildcardSeen = true;
            parts[dartx.add](-1);
          } else {
            parts[dartx.add](parseHex(partStart, i));
          }
          partStart = dart.notNull(i) + 1;
        }
      }
      if (parts[dartx.length] == 0)
        error('too few parts');
      let atEnd = partStart == end;
      let isLastWildcard = parts[dartx.last] == -1;
      if (dart.notNull(atEnd) && !dart.notNull(isLastWildcard)) {
        error('expected a part after last `:`', end);
      }
      if (!dart.notNull(atEnd)) {
        try {
          parts[dartx.add](parseHex(partStart, end));
        } catch (e) {
          try {
            let last = Uri.parseIPv4Address(host[dartx.substring](partStart, end));
            parts[dartx.add](dart.notNull(last[dartx.get](0)) << 8 | dart.notNull(last[dartx.get](1)));
            parts[dartx.add](dart.notNull(last[dartx.get](2)) << 8 | dart.notNull(last[dartx.get](3)));
          } catch (e) {
            error('invalid end of IPv6 address.', partStart);
          }

        }

      }
      if (dart.notNull(wildcardSeen)) {
        if (dart.notNull(parts[dartx.length]) > 7) {
          error('an address with a wildcard must have less than 7 parts');
        }
      } else if (parts[dartx.length] != 8) {
        error('an address without a wildcard must contain exactly 8 parts');
      }
      let bytes = List$(int).new(16);
      for (let i = 0, index = 0; dart.notNull(i) < dart.notNull(parts[dartx.length]); i = dart.notNull(i) + 1) {
        let value = parts[dartx.get](i);
        if (value == -1) {
          let wildCardLength = 9 - dart.notNull(parts[dartx.length]);
          for (let j = 0; dart.notNull(j) < dart.notNull(wildCardLength); j = dart.notNull(j) + 1) {
            bytes[dartx.set](index, 0);
            bytes[dartx.set](dart.notNull(index) + 1, 0);
            index = dart.notNull(index) + 2;
          }
        } else {
          bytes[dartx.set](index, dart.notNull(value) >> 8);
          bytes[dartx.set](dart.notNull(index) + 1, dart.notNull(value) & 255);
          index = dart.notNull(index) + 2;
        }
      }
      return dart.as(bytes, List$(int));
    }
    static _uriEncode(canonicalTable, text, opts) {
      let encoding = opts && 'encoding' in opts ? opts.encoding : convert.UTF8;
      let spaceToPlus = opts && 'spaceToPlus' in opts ? opts.spaceToPlus : false;
      function byteToHex(byte, buffer) {
        let hex = '0123456789ABCDEF';
        dart.dsend(buffer, 'writeCharCode', hex[dartx.codeUnitAt](dart.as(dart.dsend(byte, '>>', 4), int)));
        dart.dsend(buffer, 'writeCharCode', hex[dartx.codeUnitAt](dart.as(dart.dsend(byte, '&', 15), int)));
      }
      dart.fn(byteToHex);
      let result = new StringBuffer();
      let bytes = encoding.encode(text);
      for (let i = 0; dart.notNull(i) < dart.notNull(bytes[dartx.length]); i = dart.notNull(i) + 1) {
        let byte = bytes[dartx.get](i);
        if (dart.notNull(byte) < 128 && (dart.notNull(canonicalTable[dartx.get](dart.notNull(byte) >> 4)) & 1 << (dart.notNull(byte) & 15)) != 0) {
          result.writeCharCode(byte);
        } else if (dart.notNull(spaceToPlus) && byte == Uri._SPACE) {
          result.writeCharCode(Uri._PLUS);
        } else {
          result.writeCharCode(Uri._PERCENT);
          byteToHex(byte, result);
        }
      }
      return dart.toString(result);
    }
    static _hexCharPairToByte(s, pos) {
      let byte = 0;
      for (let i = 0; dart.notNull(i) < 2; i = dart.notNull(i) + 1) {
        let charCode = s[dartx.codeUnitAt](dart.notNull(pos) + dart.notNull(i));
        if (48 <= dart.notNull(charCode) && dart.notNull(charCode) <= 57) {
          byte = dart.notNull(byte) * 16 + dart.notNull(charCode) - 48;
        } else {
          charCode = dart.notNull(charCode) | 32;
          if (97 <= dart.notNull(charCode) && dart.notNull(charCode) <= 102) {
            byte = dart.notNull(byte) * 16 + dart.notNull(charCode) - 87;
          } else {
            dart.throw(new ArgumentError("Invalid URL encoding"));
          }
        }
      }
      return byte;
    }
    static _uriDecode(text, opts) {
      let plusToSpace = opts && 'plusToSpace' in opts ? opts.plusToSpace : false;
      let encoding = opts && 'encoding' in opts ? opts.encoding : convert.UTF8;
      let simple = true;
      for (let i = 0; dart.notNull(i) < dart.notNull(text[dartx.length]) && dart.notNull(simple); i = dart.notNull(i) + 1) {
        let codeUnit = text[dartx.codeUnitAt](i);
        simple = codeUnit != Uri._PERCENT && codeUnit != Uri._PLUS;
      }
      let bytes = null;
      if (dart.notNull(simple)) {
        if (dart.equals(encoding, convert.UTF8) || dart.equals(encoding, convert.LATIN1)) {
          return text;
        } else {
          bytes = text[dartx.codeUnits];
        }
      } else {
        bytes = List$(int).new();
        for (let i = 0; dart.notNull(i) < dart.notNull(text[dartx.length]); i = dart.notNull(i) + 1) {
          let codeUnit = text[dartx.codeUnitAt](i);
          if (dart.notNull(codeUnit) > 127) {
            dart.throw(new ArgumentError("Illegal percent encoding in URI"));
          }
          if (codeUnit == Uri._PERCENT) {
            if (dart.notNull(i) + 3 > dart.notNull(text[dartx.length])) {
              dart.throw(new ArgumentError('Truncated URI'));
            }
            bytes[dartx.add](Uri._hexCharPairToByte(text, dart.notNull(i) + 1));
            i = dart.notNull(i) + 2;
          } else if (dart.notNull(plusToSpace) && codeUnit == Uri._PLUS) {
            bytes[dartx.add](Uri._SPACE);
          } else {
            bytes[dartx.add](codeUnit);
          }
        }
      }
      return encoding.decode(bytes);
    }
    static _isAlphabeticCharacter(codeUnit) {
      return dart.notNull(codeUnit) >= dart.notNull(Uri._LOWER_CASE_A) && dart.notNull(codeUnit) <= dart.notNull(Uri._LOWER_CASE_Z) || dart.notNull(codeUnit) >= dart.notNull(Uri._UPPER_CASE_A) && dart.notNull(codeUnit) <= dart.notNull(Uri._UPPER_CASE_Z);
    }
  }
  dart.defineNamedConstructor(Uri, '_internal');
  dart.setSignature(Uri, {
    constructors: () => ({
      _internal: [Uri, [String, String, String, num, String, String, String]],
      new: [Uri, [], {scheme: String, userInfo: String, host: String, port: int, path: String, pathSegments: Iterable$(String), query: String, queryParameters: Map$(String, String), fragment: String}],
      http: [Uri, [String, String], [Map$(String, String)]],
      https: [Uri, [String, String], [Map$(String, String)]],
      file: [Uri, [String], {windows: bool}]
    }),
    methods: () => ({
      replace: [Uri, [], {scheme: String, userInfo: String, host: String, port: int, path: String, pathSegments: Iterable$(String), query: String, queryParameters: Map$(String, String), fragment: String}],
      [_merge]: [String, [String, String]],
      [_hasDotSegments]: [bool, [String]],
      [_removeDotSegments]: [String, [String]],
      resolve: [Uri, [String]],
      resolveUri: [Uri, [Uri]],
      toFilePath: [String, [], {windows: bool}],
      [_toFilePath]: [String, []],
      [_toWindowsFilePath]: [String, []],
      [_writeAuthority]: [dart.void, [StringSink]]
    }),
    statics: () => ({
      _defaultPort: [int, [String]],
      parse: [Uri, [String]],
      _fail: [dart.void, [String, int, String]],
      _makeHttpUri: [Uri, [String, String, String, Map$(String, String)]],
      _checkNonWindowsPathReservedCharacters: [dart.dynamic, [List$(String), bool]],
      _checkWindowsPathReservedCharacters: [dart.dynamic, [List$(String), bool], [int]],
      _checkWindowsDriveLetter: [dart.dynamic, [int, bool]],
      _makeFileUri: [dart.dynamic, [String]],
      _makeWindowsFileUrl: [dart.dynamic, [String]],
      _makePort: [int, [int, String]],
      _makeHost: [String, [String, int, int, bool]],
      _isRegNameChar: [bool, [int]],
      _normalizeRegName: [String, [String, int, int]],
      _makeScheme: [String, [String, int]],
      _makeUserInfo: [String, [String, int, int]],
      _makePath: [String, [String, int, int, Iterable$(String), bool, bool]],
      _makeQuery: [String, [String, int, int, Map$(String, String)]],
      _makeFragment: [String, [String, int, int]],
      _stringOrNullLength: [int, [String]],
      _isHexDigit: [bool, [int]],
      _hexValue: [int, [int]],
      _normalizeEscape: [String, [String, int, bool]],
      _isUnreservedChar: [bool, [int]],
      _escapeChar: [String, [dart.dynamic]],
      _normalize: [String, [String, int, int, List$(int)]],
      _isSchemeCharacter: [bool, [int]],
      _isGeneralDelimiter: [bool, [int]],
      _addIfNonEmpty: [dart.void, [StringBuffer, String, String, String]],
      encodeComponent: [String, [String]],
      encodeQueryComponent: [String, [String], {encoding: convert.Encoding}],
      decodeComponent: [String, [String]],
      decodeQueryComponent: [String, [String], {encoding: convert.Encoding}],
      encodeFull: [String, [String]],
      decodeFull: [String, [String]],
      splitQueryString: [Map$(String, String), [String], {encoding: convert.Encoding}],
      parseIPv4Address: [List$(int), [String]],
      parseIPv6Address: [List$(int), [String], [int, int]],
      _uriEncode: [String, [List$(int), String], {encoding: convert.Encoding, spaceToPlus: bool}],
      _hexCharPairToByte: [int, [String, int]],
      _uriDecode: [String, [String], {plusToSpace: bool, encoding: convert.Encoding}],
      _isAlphabeticCharacter: [bool, [int]]
    }),
    names: ['_defaultPort', 'parse', '_fail', '_makeHttpUri', '_checkNonWindowsPathReservedCharacters', '_checkWindowsPathReservedCharacters', '_checkWindowsDriveLetter', '_makeFileUri', '_makeWindowsFileUrl', '_makePort', '_makeHost', '_isRegNameChar', '_normalizeRegName', '_makeScheme', '_makeUserInfo', '_makePath', '_makeQuery', '_makeFragment', '_stringOrNullLength', '_isHexDigit', '_hexValue', '_normalizeEscape', '_isUnreservedChar', '_escapeChar', '_normalize', '_isSchemeCharacter', '_isGeneralDelimiter', '_addIfNonEmpty', 'encodeComponent', 'encodeQueryComponent', 'decodeComponent', 'decodeQueryComponent', 'encodeFull', 'decodeFull', 'splitQueryString', 'parseIPv4Address', 'parseIPv6Address', '_uriEncode', '_hexCharPairToByte', '_uriDecode', '_isAlphabeticCharacter']
  });
  Uri._SPACE = 32;
  Uri._DOUBLE_QUOTE = 34;
  Uri._NUMBER_SIGN = 35;
  Uri._PERCENT = 37;
  Uri._ASTERISK = 42;
  Uri._PLUS = 43;
  Uri._DOT = 46;
  Uri._SLASH = 47;
  Uri._ZERO = 48;
  Uri._NINE = 57;
  Uri._COLON = 58;
  Uri._LESS = 60;
  Uri._GREATER = 62;
  Uri._QUESTION = 63;
  Uri._AT_SIGN = 64;
  Uri._UPPER_CASE_A = 65;
  Uri._UPPER_CASE_F = 70;
  Uri._UPPER_CASE_Z = 90;
  Uri._LEFT_BRACKET = 91;
  Uri._BACKSLASH = 92;
  Uri._RIGHT_BRACKET = 93;
  Uri._LOWER_CASE_A = 97;
  Uri._LOWER_CASE_F = 102;
  Uri._LOWER_CASE_Z = 122;
  Uri._BAR = 124;
  Uri._unreservedTable = dart.const([0, 0, 24576, 1023, 65534, 34815, 65534, 18431]);
  Uri._unreserved2396Table = dart.const([0, 0, 26498, 1023, 65534, 34815, 65534, 18431]);
  Uri._encodeFullTable = dart.const([0, 0, 65498, 45055, 65535, 34815, 65534, 18431]);
  Uri._schemeTable = dart.const([0, 0, 26624, 1023, 65534, 2047, 65534, 2047]);
  Uri._schemeLowerTable = dart.const([0, 0, 26624, 1023, 0, 0, 65534, 2047]);
  Uri._subDelimitersTable = dart.const([0, 0, 32722, 11263, 65534, 34815, 65534, 18431]);
  Uri._genDelimitersTable = dart.const([0, 0, 32776, 33792, 1, 10240, 0, 0]);
  Uri._userinfoTable = dart.const([0, 0, 32722, 12287, 65534, 34815, 65534, 18431]);
  Uri._regNameTable = dart.const([0, 0, 32754, 11263, 65534, 34815, 65534, 18431]);
  Uri._pathCharTable = dart.const([0, 0, 32722, 12287, 65535, 34815, 65534, 18431]);
  Uri._pathCharOrSlashTable = dart.const([0, 0, 65490, 12287, 65535, 34815, 65534, 18431]);
  Uri._queryCharTable = dart.const([0, 0, 65490, 45055, 65535, 34815, 65534, 18431]);
  function _symbolToString(symbol) {
    return _internal.Symbol.getName(dart.as(symbol, _internal.Symbol));
  }
  dart.fn(_symbolToString, String, [Symbol]);
  // Exports:
  exports.Object = Object;
  exports.Deprecated = Deprecated;
  exports.deprecated = deprecated;
  exports.override = override;
  exports.proxy = proxy;
  exports.bool = bool;
  exports.Comparator$ = Comparator$;
  exports.Comparator = Comparator;
  exports.Comparable$ = Comparable$;
  exports.Comparable = Comparable;
  exports.DateTime = DateTime;
  exports.num = num;
  exports.double = double;
  exports.Duration = Duration;
  exports.Error = Error;
  exports.AssertionError = AssertionError;
  exports.TypeError = TypeError;
  exports.CastError = CastError;
  exports.NullThrownError = NullThrownError;
  exports.ArgumentError = ArgumentError;
  exports.RangeError = RangeError;
  exports.IndexError = IndexError;
  exports.FallThroughError = FallThroughError;
  exports.AbstractClassInstantiationError = AbstractClassInstantiationError;
  exports.NoSuchMethodError = NoSuchMethodError;
  exports.UnsupportedError = UnsupportedError;
  exports.UnimplementedError = UnimplementedError;
  exports.StateError = StateError;
  exports.ConcurrentModificationError = ConcurrentModificationError;
  exports.OutOfMemoryError = OutOfMemoryError;
  exports.StackOverflowError = StackOverflowError;
  exports.CyclicInitializationError = CyclicInitializationError;
  exports.Exception = Exception;
  exports.FormatException = FormatException;
  exports.IntegerDivisionByZeroException = IntegerDivisionByZeroException;
  exports.Expando$ = Expando$;
  exports.Expando = Expando;
  exports.Function = Function;
  exports.identical = identical;
  exports.identityHashCode = identityHashCode;
  exports.int = int;
  exports.Invocation = Invocation;
  exports.Iterable$ = Iterable$;
  exports.Iterable = Iterable;
  exports.BidirectionalIterator$ = BidirectionalIterator$;
  exports.BidirectionalIterator = BidirectionalIterator;
  exports.Iterator$ = Iterator$;
  exports.Iterator = Iterator;
  exports.List$ = List$;
  exports.List = List;
  exports.Map$ = Map$;
  exports.Map = Map;
  exports.Null = Null;
  exports.Pattern = Pattern;
  exports.print = print;
  exports.Match = Match;
  exports.RegExp = RegExp;
  exports.Set$ = Set$;
  exports.Sink$ = Sink$;
  exports.Sink = Sink;
  exports.StackTrace = StackTrace;
  exports.Stopwatch = Stopwatch;
  exports.String = String;
  exports.RuneIterator = RuneIterator;
  exports.StringBuffer = StringBuffer;
  exports.StringSink = StringSink;
  exports.Symbol = Symbol;
  exports.Type = Type;
  exports.Uri = Uri;
});
dart_library.library('dart/html', null, /* Imports */[
  "dart/_runtime",
  'dart/math',
  'dart/core',
  'dart/_js_helper',
  'dart/async',
  'dart/collection',
  'dart/typed_data',
  'dart/_foreign_helper'
], /* Lazy imports */[
  'dart/_metadata',
  'dart/html_common'
], function(exports, dart, math, core, _js_helper, async, collection, typed_data, _foreign_helper, _metadata, html_common) {
  'use strict';
  let dartx = dart.dartx;
  dart.export_(exports, math, ['Rectangle', 'Point'], []);
  class DartHtmlDomObject extends core.Object {
    DartHtmlDomObject() {
      this.raw = null;
    }
    internal_() {
      this.raw = null;
    }
  }
  dart.defineNamedConstructor(DartHtmlDomObject, 'internal_');
  dart.setSignature(DartHtmlDomObject, {
    constructors: () => ({
      DartHtmlDomObject: [DartHtmlDomObject, []],
      internal_: [DartHtmlDomObject, []]
    })
  });
  const _addEventListener = Symbol('_addEventListener');
  const _removeEventListener = Symbol('_removeEventListener');
  const _addEventListener_1 = Symbol('_addEventListener_1');
  const _addEventListener_2 = Symbol('_addEventListener_2');
  const _addEventListener_3 = Symbol('_addEventListener_3');
  const _addEventListener_4 = Symbol('_addEventListener_4');
  const _dispatchEvent_1 = Symbol('_dispatchEvent_1');
  const _removeEventListener_1 = Symbol('_removeEventListener_1');
  const _removeEventListener_2 = Symbol('_removeEventListener_2');
  const _removeEventListener_3 = Symbol('_removeEventListener_3');
  const _removeEventListener_4 = Symbol('_removeEventListener_4');
  class EventTarget extends DartHtmlDomObject {
    _created() {
      super.DartHtmlDomObject();
    }
    get on() {
      return new Events(this);
    }
    addEventListener(type, listener, useCapture) {
      if (useCapture === void 0)
        useCapture = null;
      if (listener != null) {
        this[_addEventListener](type, listener, useCapture);
      }
    }
    removeEventListener(type, listener, useCapture) {
      if (useCapture === void 0)
        useCapture = null;
      if (listener != null) {
        this[_removeEventListener](type, listener, useCapture);
      }
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateEventTarget() {
      return EventTarget._internalWrap();
    }
    static _internalWrap() {
      return new EventTarget.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    [_addEventListener](type, listener, useCapture) {
      if (type === void 0)
        type = null;
      if (listener === void 0)
        listener = null;
      if (useCapture === void 0)
        useCapture = null;
      if (useCapture != null) {
        this[_addEventListener_1](type, listener, useCapture);
        return;
      }
      if (listener != null) {
        this[_addEventListener_2](type, listener);
        return;
      }
      if (type != null) {
        this[_addEventListener_3](type);
        return;
      }
      this[_addEventListener_4]();
      return;
    }
    [_addEventListener_1](type, listener, useCapture) {
      return wrap_jso(this.raw.addEventListener(unwrap_jso(type), unwrap_jso(listener), unwrap_jso(useCapture)));
    }
    [_addEventListener_2](type, listener) {
      return wrap_jso(this.raw.addEventListener(unwrap_jso(type), unwrap_jso(listener)));
    }
    [_addEventListener_3](type) {
      return wrap_jso(this.raw.addEventListener(unwrap_jso(type)));
    }
    [_addEventListener_4]() {
      return wrap_jso(this.raw.addEventListener());
    }
    dispatchEvent(event) {
      return this[_dispatchEvent_1](event);
    }
    [_dispatchEvent_1](event) {
      return dart.as(wrap_jso(this.raw.dispatchEvent(unwrap_jso(event))), core.bool);
    }
    [_removeEventListener](type, listener, useCapture) {
      if (type === void 0)
        type = null;
      if (listener === void 0)
        listener = null;
      if (useCapture === void 0)
        useCapture = null;
      if (useCapture != null) {
        this[_removeEventListener_1](type, listener, useCapture);
        return;
      }
      if (listener != null) {
        this[_removeEventListener_2](type, listener);
        return;
      }
      if (type != null) {
        this[_removeEventListener_3](type);
        return;
      }
      this[_removeEventListener_4]();
      return;
    }
    [_removeEventListener_1](type, listener, useCapture) {
      return wrap_jso(this.raw.removeEventListener(unwrap_jso(type), unwrap_jso(listener), unwrap_jso(useCapture)));
    }
    [_removeEventListener_2](type, listener) {
      return wrap_jso(this.raw.removeEventListener(unwrap_jso(type), unwrap_jso(listener)));
    }
    [_removeEventListener_3](type) {
      return wrap_jso(this.raw.removeEventListener(unwrap_jso(type)));
    }
    [_removeEventListener_4]() {
      return wrap_jso(this.raw.removeEventListener());
    }
  }
  dart.defineNamedConstructor(EventTarget, '_created');
  dart.defineNamedConstructor(EventTarget, 'internal_');
  dart.setSignature(EventTarget, {
    constructors: () => ({
      _created: [EventTarget, []],
      _: [EventTarget, []],
      _internalWrap: [EventTarget, []],
      internal_: [EventTarget, []]
    }),
    methods: () => ({
      addEventListener: [dart.void, [core.String, EventListener], [core.bool]],
      removeEventListener: [dart.void, [core.String, EventListener], [core.bool]],
      [_addEventListener]: [dart.void, [], [core.String, EventListener, core.bool]],
      [_addEventListener_1]: [dart.void, [dart.dynamic, EventListener, dart.dynamic]],
      [_addEventListener_2]: [dart.void, [dart.dynamic, EventListener]],
      [_addEventListener_3]: [dart.void, [dart.dynamic]],
      [_addEventListener_4]: [dart.void, []],
      dispatchEvent: [core.bool, [Event]],
      [_dispatchEvent_1]: [core.bool, [Event]],
      [_removeEventListener]: [dart.void, [], [core.String, EventListener, core.bool]],
      [_removeEventListener_1]: [dart.void, [dart.dynamic, EventListener, dart.dynamic]],
      [_removeEventListener_2]: [dart.void, [dart.dynamic, EventListener]],
      [_removeEventListener_3]: [dart.void, [dart.dynamic]],
      [_removeEventListener_4]: [dart.void, []]
    }),
    statics: () => ({internalCreateEventTarget: [EventTarget, []]}),
    names: ['internalCreateEventTarget']
  });
  EventTarget[dart.metadata] = () => [dart.const(new _metadata.DomName('EventTarget')), dart.const(new _js_helper.Native("EventTarget"))];
  const _removeChild = Symbol('_removeChild');
  const _replaceChild = Symbol('_replaceChild');
  const _this = Symbol('_this');
  const _clearChildren = Symbol('_clearChildren');
  const _localName = Symbol('_localName');
  const _namespaceUri = Symbol('_namespaceUri');
  const _append_1 = Symbol('_append_1');
  const _clone_1 = Symbol('_clone_1');
  const _contains_1 = Symbol('_contains_1');
  const _hasChildNodes_1 = Symbol('_hasChildNodes_1');
  const _insertBefore_1 = Symbol('_insertBefore_1');
  const _removeChild_1 = Symbol('_removeChild_1');
  const _replaceChild_1 = Symbol('_replaceChild_1');
  class Node extends EventTarget {
    _created() {
      super._created();
      if (dart.dload(this, 'blink_jsObject') == null) {
        dart.throw(new dart.dynamic("the created constructor cannot be called directly"));
      }
    }
    get nodes() {
      return new _ChildNodeListLazy(this);
    }
    set nodes(value) {
      let copy = core.List.from(value);
      this.text = '';
      for (let node of dart.as(copy, core.Iterable$(Node))) {
        this.append(node);
      }
    }
    remove() {
      if (this.parentNode != null) {
        let parent = this.parentNode;
        this.parentNode[_removeChild](this);
      }
    }
    replaceWith(otherNode) {
      try {
        let parent = this.parentNode;
        parent[_replaceChild](otherNode, this);
      } catch (e) {
      }

      ;
      return this;
    }
    insertAllBefore(newNodes, refChild) {
      if (dart.is(newNodes, _ChildNodeListLazy)) {
        let otherList = newNodes;
        if (dart.notNull(core.identical(otherList[_this], this))) {
          dart.throw(new core.ArgumentError(newNodes));
        }
        for (let i = 0, len = otherList.length; dart.notNull(i) < dart.notNull(len); i = dart.notNull(i) + 1) {
          this.insertBefore(otherList[_this].firstChild, refChild);
        }
      } else {
        for (let node of newNodes) {
          this.insertBefore(node, refChild);
        }
      }
    }
    [_clearChildren]() {
      while (this.firstChild != null) {
        this[_removeChild](this.firstChild);
      }
    }
    toString() {
      let value = this.nodeValue;
      return value == null ? super.toString() : value;
    }
    get childNodes() {
      return dart.as(wrap_jso(this.raw.childNodes), core.List$(Node));
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNode() {
      return Node._internalWrap();
    }
    static _internalWrap() {
      return new Node.internal_();
    }
    internal_() {
      super.internal_();
    }
    get baseUri() {
      return dart.as(wrap_jso(this.raw.baseURI), core.String);
    }
    get firstChild() {
      return dart.as(wrap_jso(this.raw.firstChild), Node);
    }
    get lastChild() {
      return dart.as(wrap_jso(this.raw.lastChild), Node);
    }
    get [_localName]() {
      return dart.as(wrap_jso(this.raw.localName), core.String);
    }
    get [_namespaceUri]() {
      return dart.as(wrap_jso(this.raw.namespaceURI), core.String);
    }
    get nextNode() {
      return dart.as(wrap_jso(this.raw.nextSibling), Node);
    }
    get nodeName() {
      return dart.as(wrap_jso(this.raw.nodeName), core.String);
    }
    get nodeType() {
      return dart.as(wrap_jso(this.raw.nodeType), core.int);
    }
    get nodeValue() {
      return dart.as(wrap_jso(this.raw.nodeValue), core.String);
    }
    get ownerDocument() {
      return dart.as(wrap_jso(this.raw.ownerDocument), Document);
    }
    get parent() {
      return dart.as(wrap_jso(this.raw.parentElement), Element);
    }
    get parentNode() {
      return dart.as(wrap_jso(this.raw.parentNode), Node);
    }
    get previousNode() {
      return dart.as(wrap_jso(this.raw.previousSibling), Node);
    }
    get text() {
      return dart.as(wrap_jso(this.raw.textContent), core.String);
    }
    set text(val) {
      return this.raw.textContent = unwrap_jso(val);
    }
    append(newChild) {
      return this[_append_1](newChild);
    }
    [_append_1](newChild) {
      return dart.as(wrap_jso(this.raw.appendChild(unwrap_jso(newChild))), Node);
    }
    clone(deep) {
      return this[_clone_1](deep);
    }
    [_clone_1](deep) {
      return dart.as(wrap_jso(this.raw.cloneNode(unwrap_jso(deep))), Node);
    }
    contains(other) {
      return this[_contains_1](other);
    }
    [_contains_1](other) {
      return dart.as(wrap_jso(this.raw.contains(unwrap_jso(other))), core.bool);
    }
    hasChildNodes() {
      return this[_hasChildNodes_1]();
    }
    [_hasChildNodes_1]() {
      return dart.as(wrap_jso(this.raw.hasChildNodes()), core.bool);
    }
    insertBefore(newChild, refChild) {
      return this[_insertBefore_1](newChild, refChild);
    }
    [_insertBefore_1](newChild, refChild) {
      return dart.as(wrap_jso(this.raw.insertBefore(unwrap_jso(newChild), unwrap_jso(refChild))), Node);
    }
    [_removeChild](oldChild) {
      return this[_removeChild_1](oldChild);
    }
    [_removeChild_1](oldChild) {
      return dart.as(wrap_jso(this.raw.removeChild(unwrap_jso(oldChild))), Node);
    }
    [_replaceChild](newChild, oldChild) {
      return this[_replaceChild_1](newChild, oldChild);
    }
    [_replaceChild_1](newChild, oldChild) {
      return dart.as(wrap_jso(this.raw.replaceChild(unwrap_jso(newChild), unwrap_jso(oldChild))), Node);
    }
  }
  dart.defineNamedConstructor(Node, '_created');
  dart.defineNamedConstructor(Node, 'internal_');
  dart.setSignature(Node, {
    constructors: () => ({
      _created: [Node, []],
      _: [Node, []],
      _internalWrap: [Node, []],
      internal_: [Node, []]
    }),
    methods: () => ({
      remove: [dart.void, []],
      replaceWith: [Node, [Node]],
      insertAllBefore: [Node, [core.Iterable$(Node), Node]],
      [_clearChildren]: [dart.void, []],
      append: [Node, [Node]],
      [_append_1]: [Node, [Node]],
      clone: [Node, [core.bool]],
      [_clone_1]: [Node, [dart.dynamic]],
      contains: [core.bool, [Node]],
      [_contains_1]: [core.bool, [Node]],
      hasChildNodes: [core.bool, []],
      [_hasChildNodes_1]: [core.bool, []],
      insertBefore: [Node, [Node, Node]],
      [_insertBefore_1]: [Node, [Node, Node]],
      [_removeChild]: [Node, [Node]],
      [_removeChild_1]: [Node, [Node]],
      [_replaceChild]: [Node, [Node, Node]],
      [_replaceChild_1]: [Node, [Node, Node]]
    }),
    statics: () => ({internalCreateNode: [Node, []]}),
    names: ['internalCreateNode']
  });
  Node[dart.metadata] = () => [dart.const(new _metadata.DomName('Node')), dart.const(new _js_helper.Native("Node"))];
  const _xtag = Symbol('_xtag');
  const _querySelectorAll = Symbol('_querySelectorAll');
  const _getComputedStyle = Symbol('_getComputedStyle');
  const _scrollIntoView = Symbol('_scrollIntoView');
  const _scrollIntoViewIfNeeded = Symbol('_scrollIntoViewIfNeeded');
  const _insertAdjacentNode = Symbol('_insertAdjacentNode');
  const _canBeUsedToCreateContextualFragment = Symbol('_canBeUsedToCreateContextualFragment');
  const _innerHtml = Symbol('_innerHtml');
  const _cannotBeUsedToCreateContextualFragment = Symbol('_cannotBeUsedToCreateContextualFragment');
  const _children = Symbol('_children');
  const _click_1 = Symbol('_click_1');
  const _attributes = Symbol('_attributes');
  const _clientHeight = Symbol('_clientHeight');
  const _clientLeft = Symbol('_clientLeft');
  const _clientTop = Symbol('_clientTop');
  const _clientWidth = Symbol('_clientWidth');
  const _offsetHeight = Symbol('_offsetHeight');
  const _offsetLeft = Symbol('_offsetLeft');
  const _offsetTop = Symbol('_offsetTop');
  const _offsetWidth = Symbol('_offsetWidth');
  const _scrollHeight = Symbol('_scrollHeight');
  const _scrollLeft = Symbol('_scrollLeft');
  const _scrollTop = Symbol('_scrollTop');
  const _scrollWidth = Symbol('_scrollWidth');
  const _blur_1 = Symbol('_blur_1');
  const _focus_1 = Symbol('_focus_1');
  const _getAnimationPlayers_1 = Symbol('_getAnimationPlayers_1');
  const _getAttribute_1 = Symbol('_getAttribute_1');
  const _getAttributeNS_1 = Symbol('_getAttributeNS_1');
  const _getBoundingClientRect_1 = Symbol('_getBoundingClientRect_1');
  const _getClientRects_1 = Symbol('_getClientRects_1');
  const _getDestinationInsertionPoints_1 = Symbol('_getDestinationInsertionPoints_1');
  const _getElementsByClassName_1 = Symbol('_getElementsByClassName_1');
  const _getElementsByTagName_1 = Symbol('_getElementsByTagName_1');
  const _getElementsByTagName = Symbol('_getElementsByTagName');
  const _hasAttribute_1 = Symbol('_hasAttribute_1');
  const _hasAttribute = Symbol('_hasAttribute');
  const _hasAttributeNS_1 = Symbol('_hasAttributeNS_1');
  const _hasAttributeNS = Symbol('_hasAttributeNS');
  const _removeAttribute_1 = Symbol('_removeAttribute_1');
  const _removeAttribute = Symbol('_removeAttribute');
  const _removeAttributeNS_1 = Symbol('_removeAttributeNS_1');
  const _removeAttributeNS = Symbol('_removeAttributeNS');
  const _requestFullscreen_1 = Symbol('_requestFullscreen_1');
  const _requestPointerLock_1 = Symbol('_requestPointerLock_1');
  const _scrollIntoView_1 = Symbol('_scrollIntoView_1');
  const _scrollIntoView_2 = Symbol('_scrollIntoView_2');
  const _scrollIntoViewIfNeeded_1 = Symbol('_scrollIntoViewIfNeeded_1');
  const _scrollIntoViewIfNeeded_2 = Symbol('_scrollIntoViewIfNeeded_2');
  const _setAttribute_1 = Symbol('_setAttribute_1');
  const _setAttributeNS_1 = Symbol('_setAttributeNS_1');
  const _childElementCount = Symbol('_childElementCount');
  const _firstElementChild = Symbol('_firstElementChild');
  const _lastElementChild = Symbol('_lastElementChild');
  const _querySelector_1 = Symbol('_querySelector_1');
  const _querySelectorAll_1 = Symbol('_querySelectorAll_1');
  class Element extends Node {
    static html(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      let fragment = exports.document.body.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer});
      return dart.as(fragment.nodes[dartx.where](dart.fn(e => dart.is(e, Element), core.bool, [dart.dynamic]))[dartx.single], Element);
    }
    created() {
      this[_xtag] = null;
      super._created();
    }
    static tag(tag, typeExtention) {
      if (typeExtention === void 0)
        typeExtention = null;
      return _ElementFactoryProvider.createElement_tag(tag, typeExtention);
    }
    static a() {
      return Element.tag('a');
    }
    static article() {
      return Element.tag('article');
    }
    static aside() {
      return Element.tag('aside');
    }
    static audio() {
      return Element.tag('audio');
    }
    static br() {
      return Element.tag('br');
    }
    static canvas() {
      return Element.tag('canvas');
    }
    static div() {
      return Element.tag('div');
    }
    static footer() {
      return Element.tag('footer');
    }
    static header() {
      return Element.tag('header');
    }
    static hr() {
      return Element.tag('hr');
    }
    static iframe() {
      return Element.tag('iframe');
    }
    static img() {
      return Element.tag('img');
    }
    static li() {
      return Element.tag('li');
    }
    static nav() {
      return Element.tag('nav');
    }
    static ol() {
      return Element.tag('ol');
    }
    static option() {
      return Element.tag('option');
    }
    static p() {
      return Element.tag('p');
    }
    static pre() {
      return Element.tag('pre');
    }
    static section() {
      return Element.tag('section');
    }
    static select() {
      return Element.tag('select');
    }
    static span() {
      return Element.tag('span');
    }
    static svg() {
      return Element.tag('svg');
    }
    static table() {
      return Element.tag('table');
    }
    static td() {
      return Element.tag('td');
    }
    static textarea() {
      return Element.tag('textarea');
    }
    static th() {
      return Element.tag('th');
    }
    static tr() {
      return Element.tag('tr');
    }
    static ul() {
      return Element.tag('ul');
    }
    static video() {
      return Element.tag('video');
    }
    get attributes() {
      return new _ElementAttributeMap(this);
    }
    set attributes(value) {
      let attributes = this.attributes;
      attributes.clear();
      for (let key of value.keys) {
        attributes.set(key, value.get(key));
      }
    }
    get children() {
      return new _ChildrenElementList._wrap(this);
    }
    set children(value) {
      let copy = core.List.from(value);
      let children = this.children;
      children[dartx.clear]();
      children[dartx.addAll](dart.as(copy, core.Iterable$(Element)));
    }
    querySelectorAll(selectors) {
      return new _FrozenElementList._wrap(this[_querySelectorAll](selectors));
    }
    query(relativeSelectors) {
      return this.querySelector(relativeSelectors);
    }
    queryAll(relativeSelectors) {
      return this.querySelectorAll(relativeSelectors);
    }
    get classes() {
      return new exports._ElementCssClassSet(this);
    }
    set classes(value) {
      let classSet = this.classes;
      classSet.clear();
      classSet.addAll(value);
    }
    get dataset() {
      return new _DataAttributeMap(this.attributes);
    }
    set dataset(value) {
      let data = this.dataset;
      data.clear();
      for (let key of value.keys) {
        data.set(key, value.get(key));
      }
    }
    getNamespacedAttributes(namespace) {
      return new _NamespacedAttributeMap(this, namespace);
    }
    getComputedStyle(pseudoElement) {
      if (pseudoElement === void 0)
        pseudoElement = null;
      if (pseudoElement == null) {
        pseudoElement = '';
      }
      return exports.window[_getComputedStyle](this, pseudoElement);
    }
    get client() {
      return new math.Rectangle(this.clientLeft, this.clientTop, this.clientWidth, this.clientHeight);
    }
    get offset() {
      return new math.Rectangle(this.offsetLeft, this.offsetTop, this.offsetWidth, this.offsetHeight);
    }
    appendText(text) {
      this.append(Text.new(text));
    }
    appendHtml(text, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      this.insertAdjacentHtml('beforeend', text, {validator: validator, treeSanitizer: treeSanitizer});
    }
    static isTagSupported(tag) {
      let e = _ElementFactoryProvider.createElement_tag(tag, null);
      return dart.is(e, Element) && !dart.is(e, UnknownElement);
    }
    attached() {
      this.enteredView();
    }
    detached() {
      this.leftView();
    }
    enteredView() {}
    leftView() {}
    animate(frames, timing) {
      if (timing === void 0)
        timing = null;
      if (!dart.is(frames, core.Iterable) || !dart.notNull(frames[dartx.every](dart.fn(x => dart.is(x, core.Map), core.bool, [dart.dynamic])))) {
        dart.throw(new core.ArgumentError("The frames parameter should be a List of Maps " + "with frame information"));
      }
      let convertedFrames = frames;
      if (dart.is(convertedFrames, core.Iterable)) {
        convertedFrames = dart.as(dart.dcall(/* Unimplemented unknown name */convertDartToNative_List, frames[dartx.map](html_common.convertDartToNative_Dictionary)[dartx.toList]()), core.Iterable$(core.Map$(core.String, dart.dynamic)));
      }
      let convertedTiming = timing;
      if (dart.is(convertedTiming, core.Map)) {
        convertedTiming = html_common.convertDartToNative_Dictionary(dart.as(convertedTiming, core.Map));
      }
      return convertedTiming == null ? dart.dcall(/* Unimplemented unknown name */_animate, convertedFrames) : dart.dcall(/* Unimplemented unknown name */_animate, convertedFrames, convertedTiming);
    }
    attributeChanged(name, oldValue, newValue) {}
    get xtag() {
      return this[_xtag] != null ? this[_xtag] : this;
    }
    set xtag(value) {
      this[_xtag] = value;
    }
    get localName() {
      return this[_localName];
    }
    get namespaceUri() {
      return this[_namespaceUri];
    }
    toString() {
      return this.localName;
    }
    scrollIntoView(alignment) {
      if (alignment === void 0)
        alignment = null;
      let hasScrollIntoViewIfNeeded = true;
      if (dart.equals(alignment, dart.dload(/* Unimplemented unknown name */ScrollAlignment, 'TOP'))) {
        this[_scrollIntoView](true);
      } else if (dart.equals(alignment, dart.dload(/* Unimplemented unknown name */ScrollAlignment, 'BOTTOM'))) {
        this[_scrollIntoView](false);
      } else if (dart.notNull(hasScrollIntoViewIfNeeded)) {
        if (dart.equals(alignment, dart.dload(/* Unimplemented unknown name */ScrollAlignment, 'CENTER'))) {
          this[_scrollIntoViewIfNeeded](true);
        } else {
          this[_scrollIntoViewIfNeeded]();
        }
      } else {
        this[_scrollIntoView]();
      }
    }
    insertAdjacentHtml(where, html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      if (dart.is(treeSanitizer, _TrustedHtmlTreeSanitizer)) {
        dart.dcall(/* Unimplemented unknown name */_insertAdjacentHtml, where, html);
      } else {
        this[_insertAdjacentNode](where, this.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer}));
      }
    }
    [_insertAdjacentNode](where, node) {
      switch (where[dartx.toLowerCase]()) {
        case 'beforebegin':
        {
          this.parentNode.insertBefore(node, this);
          break;
        }
        case 'afterbegin':
        {
          let first = dart.notNull(this.nodes[dartx.length]) > 0 ? this.nodes[dartx.get](0) : null;
          this.insertBefore(node, first);
          break;
        }
        case 'beforeend':
        {
          this.append(node);
          break;
        }
        case 'afterend':
        {
          this.parentNode.insertBefore(node, this.nextNode);
          break;
        }
        default:
        {
          dart.throw(new core.ArgumentError(`Invalid position ${where}`));
        }
      }
    }
    matchesWithAncestors(selectors) {
      let elem = this;
      do {
        if (dart.notNull(dart.as(dart.dcall(elem.matches, selectors), core.bool)))
          return true;
        elem = elem.parent;
      } while (elem != null);
      return false;
    }
    get contentEdge() {
      return new _ContentCssRect(this);
    }
    get paddingEdge() {
      return new _PaddingCssRect(this);
    }
    get borderEdge() {
      return new _BorderCssRect(this);
    }
    get marginEdge() {
      return new _MarginCssRect(this);
    }
    get documentOffset() {
      return this.offsetTo(exports.document.documentElement);
    }
    offsetTo(parent) {
      return Element._offsetToHelper(this, parent);
    }
    static _offsetToHelper(current, parent) {
      let sameAsParent = dart.equals(current, parent);
      let foundAsParent = dart.notNull(sameAsParent) || parent.tagName == 'HTML';
      if (current == null || dart.notNull(sameAsParent)) {
        if (dart.notNull(foundAsParent))
          return new math.Point(0, 0);
        dart.throw(new core.ArgumentError("Specified element is not a transitive offset " + "parent of this element."));
      }
      let parentOffset = current.offsetParent;
      let p = Element._offsetToHelper(parentOffset, parent);
      return new math.Point(dart.dsend(p.x, '+', current.offsetLeft), dart.dsend(p.y, '+', current.offsetTop));
    }
    createFragment(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      if (treeSanitizer == null) {
        if (validator == null) {
          if (Element._defaultValidator == null) {
            Element._defaultValidator = new NodeValidatorBuilder.common();
          }
          validator = Element._defaultValidator;
        }
        if (Element._defaultSanitizer == null) {
          Element._defaultSanitizer = new _ValidatingTreeSanitizer(validator);
        } else {
          Element._defaultSanitizer.validator = validator;
        }
        treeSanitizer = Element._defaultSanitizer;
      } else if (validator != null) {
        dart.throw(new core.ArgumentError('validator can only be passed if treeSanitizer is null'));
      }
      if (Element._parseDocument == null) {
        Element._parseDocument = exports.document.implementation.createHtmlDocument('');
        Element._parseRange = Element._parseDocument.createRange();
        let base = Element._parseDocument.createElement('base');
        dart.dput(base, 'href', exports.document.baseUri);
        Element._parseDocument.head.append(base);
      }
      let contextElement = null;
      if (dart.is(this, BodyElement)) {
        contextElement = Element._parseDocument.body;
      } else {
        contextElement = Element._parseDocument.createElement(this.tagName);
        Element._parseDocument.body.append(dart.as(contextElement, Node));
      }
      let fragment = null;
      if (dart.notNull(Range.supportsCreateContextualFragment) && dart.notNull(this[_canBeUsedToCreateContextualFragment])) {
        Element._parseRange.selectNodeContents(dart.as(contextElement, Node));
        fragment = Element._parseRange.createContextualFragment(html);
      } else {
        dart.dput(contextElement, _innerHtml, html);
        fragment = Element._parseDocument.createDocumentFragment();
        while (dart.dload(contextElement, 'firstChild') != null) {
          dart.dsend(fragment, 'append', dart.dload(contextElement, 'firstChild'));
        }
      }
      if (!dart.equals(contextElement, Element._parseDocument.body)) {
        dart.dsend(contextElement, 'remove');
      }
      treeSanitizer.sanitizeTree(dart.as(fragment, Node));
      exports.document.adoptNode(dart.as(fragment, Node));
      return dart.as(fragment, DocumentFragment);
    }
    get [_canBeUsedToCreateContextualFragment]() {
      return !dart.notNull(this[_cannotBeUsedToCreateContextualFragment]);
    }
    get [_cannotBeUsedToCreateContextualFragment]() {
      return Element._tagsForWhichCreateContextualFragmentIsNotSupported[dartx.contains](this.tagName);
    }
    set innerHtml(html) {
      this.setInnerHtml(html);
    }
    setInnerHtml(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      this.text = null;
      if (dart.is(treeSanitizer, _TrustedHtmlTreeSanitizer)) {
        this[_innerHtml] = html;
      } else {
        this.append(this.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer}));
      }
    }
    get innerHtml() {
      return this[_innerHtml];
    }
    get on() {
      return new ElementEvents(this);
    }
    static _hasCorruptedAttributes(element) {
      let attributes = dart.dindex(unwrap_jso(element), "attributes");
      if (!dart.notNull(dart.as(dart.dsend(attributes, 'instanceof', Element._namedNodeMap), core.bool))) {
        return true;
      }
      let childNodes = unwrap_jso(element.childNodes);
      let length = dart.dindex(childNodes, "length");
      let lastChild = unwrap_jso(element.lastChild);
      if (null != lastChild && !dart.equals(lastChild, dart.dindex(childNodes, dart.dsend(length, '-', 1)))) {
        return true;
      }
      let children = unwrap_jso(element[_children]);
      if (null != children) {
        if (!dart.notNull(dart.as(dart.dsend(children, 'instanceof', Element._htmlCollection), core.bool)) || dart.notNull(dart.as(dart.dsend(children, 'instanceof', Element._nodeList), core.bool))) {
          return true;
        }
      }
      return false;
    }
    static _hasCorruptedAttributesAdditionalCheck(element) {
      return false;
    }
    static _safeTagName(element) {
      let result = 'element tag unavailable';
      try {
        if (typeof dart.dload(element, 'tagName') == 'string') {
          result = dart.as(dart.dload(element, 'tagName'), core.String);
        }
      } catch (e) {
      }

      return result;
    }
    get isContentEditable() {
      return dart.as(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkHTMLElement'), 'instance'), 'isContentEditable_Getter_', unwrap_jso(this)), core.bool);
    }
    click() {
      return dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkHTMLElement'), 'instance'), 'click_Callback_0_', unwrap_jso(this));
    }
    get offsetHeight() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'offsetHeight_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get offsetLeft() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'offsetLeft_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get offsetTop() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'offsetTop_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get offsetWidth() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'offsetWidth_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get clientHeight() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'clientHeight_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get clientLeft() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'clientLeft_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get clientTop() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'clientTop_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get clientWidth() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'clientWidth_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get scrollHeight() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollHeight_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get scrollLeft() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollLeft_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    set scrollLeft(value) {
      return dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollLeft_Setter_', unwrap_jso(this), value[dartx.round]());
    }
    get scrollTop() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollTop_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    set scrollTop(value) {
      return dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollTop_Setter_', unwrap_jso(this), value[dartx.round]());
    }
    get scrollWidth() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkElement'), 'instance'), 'scrollWidth_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateElement() {
      return Element._internalWrap();
    }
    static _internalWrap() {
      return new Element.internal_();
    }
    internal_() {
      this[_xtag] = null;
      super.internal_();
    }
    get contentEditable() {
      return dart.as(wrap_jso(this.raw.contentEditable), core.String);
    }
    set contentEditable(val) {
      return this.raw.contentEditable = unwrap_jso(val);
    }
    get contextMenu() {
      return wrap_jso(this.raw.contextMenu);
    }
    set contextMenu(val) {
      return this.raw.contextMenu = unwrap_jso(val);
    }
    get dir() {
      return dart.as(wrap_jso(this.raw.dir), core.String);
    }
    set dir(val) {
      return this.raw.dir = unwrap_jso(val);
    }
    get draggable() {
      return dart.as(wrap_jso(this.raw.draggable), core.bool);
    }
    set draggable(val) {
      return this.raw.draggable = unwrap_jso(val);
    }
    get hidden() {
      return dart.as(wrap_jso(this.raw.hidden), core.bool);
    }
    set hidden(val) {
      return this.raw.hidden = unwrap_jso(val);
    }
    get isContentEditable() {
      return dart.as(wrap_jso(this.raw.isContentEditable), core.bool);
    }
    get lang() {
      return dart.as(wrap_jso(this.raw.lang), core.String);
    }
    set lang(val) {
      return this.raw.lang = unwrap_jso(val);
    }
    get spellcheck() {
      return dart.as(wrap_jso(this.raw.spellcheck), core.bool);
    }
    set spellcheck(val) {
      return this.raw.spellcheck = unwrap_jso(val);
    }
    get tabIndex() {
      return dart.as(wrap_jso(this.raw.tabIndex), core.int);
    }
    set tabIndex(val) {
      return this.raw.tabIndex = unwrap_jso(val);
    }
    get title() {
      return dart.as(wrap_jso(this.raw.title), core.String);
    }
    set title(val) {
      return this.raw.title = unwrap_jso(val);
    }
    get translate() {
      return dart.as(wrap_jso(this.raw.translate), core.bool);
    }
    set translate(val) {
      return this.raw.translate = unwrap_jso(val);
    }
    get dropzone() {
      return dart.as(wrap_jso(this.raw.webkitdropzone), core.String);
    }
    set dropzone(val) {
      return this.raw.webkitdropzone = unwrap_jso(val);
    }
    click() {
      this[_click_1]();
      return;
    }
    [_click_1]() {
      return wrap_jso(this.raw.click());
    }
    get [_attributes]() {
      return dart.as(wrap_jso(this.raw.attributes), _NamedNodeMap);
    }
    get className() {
      return dart.as(wrap_jso(this.raw.className), core.String);
    }
    set className(val) {
      return this.raw.className = unwrap_jso(val);
    }
    get [_clientHeight]() {
      return dart.as(wrap_jso(this.raw.clientHeight), core.int);
    }
    get [_clientLeft]() {
      return dart.as(wrap_jso(this.raw.clientLeft), core.int);
    }
    get [_clientTop]() {
      return dart.as(wrap_jso(this.raw.clientTop), core.int);
    }
    get [_clientWidth]() {
      return dart.as(wrap_jso(this.raw.clientWidth), core.int);
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    set id(val) {
      return this.raw.id = unwrap_jso(val);
    }
    get [_innerHtml]() {
      return dart.as(wrap_jso(this.raw.innerHTML), core.String);
    }
    set [_innerHtml](val) {
      return this.raw.innerHTML = unwrap_jso(val);
    }
    get [_offsetHeight]() {
      return dart.as(wrap_jso(this.raw.offsetHeight), core.int);
    }
    get [_offsetLeft]() {
      return dart.as(wrap_jso(this.raw.offsetLeft), core.int);
    }
    get offsetParent() {
      return dart.as(wrap_jso(this.raw.offsetParent), Element);
    }
    get [_offsetTop]() {
      return dart.as(wrap_jso(this.raw.offsetTop), core.int);
    }
    get [_offsetWidth]() {
      return dart.as(wrap_jso(this.raw.offsetWidth), core.int);
    }
    get outerHtml() {
      return dart.as(wrap_jso(this.raw.outerHTML), core.String);
    }
    get [_scrollHeight]() {
      return dart.as(wrap_jso(this.raw.scrollHeight), core.int);
    }
    get [_scrollLeft]() {
      return dart.as(wrap_jso(this.raw.scrollLeft), core.num);
    }
    set [_scrollLeft](val) {
      return this.raw.scrollLeft = unwrap_jso(val);
    }
    get [_scrollTop]() {
      return dart.as(wrap_jso(this.raw.scrollTop), core.num);
    }
    set [_scrollTop](val) {
      return this.raw.scrollTop = unwrap_jso(val);
    }
    get [_scrollWidth]() {
      return dart.as(wrap_jso(this.raw.scrollWidth), core.int);
    }
    get style() {
      return dart.as(wrap_jso(this.raw.style), CssStyleDeclaration);
    }
    get tagName() {
      return dart.as(wrap_jso(this.raw.tagName), core.String);
    }
    blur() {
      this[_blur_1]();
      return;
    }
    [_blur_1]() {
      return wrap_jso(this.raw.blur());
    }
    focus() {
      this[_focus_1]();
      return;
    }
    [_focus_1]() {
      return wrap_jso(this.raw.focus());
    }
    getAnimationPlayers() {
      return this[_getAnimationPlayers_1]();
    }
    [_getAnimationPlayers_1]() {
      return dart.as(wrap_jso(this.raw.getAnimationPlayers()), core.List$(dart.dynamic));
    }
    getAttribute(name) {
      return this[_getAttribute_1](name);
    }
    [_getAttribute_1](name) {
      return dart.as(wrap_jso(this.raw.getAttribute(unwrap_jso(name))), core.String);
    }
    getAttributeNS(namespaceURI, localName) {
      return this[_getAttributeNS_1](namespaceURI, localName);
    }
    [_getAttributeNS_1](namespaceURI, localName) {
      return dart.as(wrap_jso(this.raw.getAttributeNS(unwrap_jso(namespaceURI), unwrap_jso(localName))), core.String);
    }
    getBoundingClientRect() {
      return this[_getBoundingClientRect_1]();
    }
    [_getBoundingClientRect_1]() {
      return dart.as(wrap_jso(this.raw.getBoundingClientRect()), math.Rectangle);
    }
    getClientRects() {
      return this[_getClientRects_1]();
    }
    [_getClientRects_1]() {
      return dart.as(wrap_jso(this.raw.getClientRects()), _ClientRectList);
    }
    getDestinationInsertionPoints() {
      return this[_getDestinationInsertionPoints_1]();
    }
    [_getDestinationInsertionPoints_1]() {
      return dart.as(wrap_jso(this.raw.getDestinationInsertionPoints()), NodeList);
    }
    getElementsByClassName(classNames) {
      return this[_getElementsByClassName_1](classNames);
    }
    [_getElementsByClassName_1](classNames) {
      return dart.as(wrap_jso(this.raw.getElementsByClassName(unwrap_jso(classNames))), HtmlCollection);
    }
    [_getElementsByTagName](name) {
      return this[_getElementsByTagName_1](name);
    }
    [_getElementsByTagName_1](name) {
      return dart.as(wrap_jso(this.raw.getElementsByTagName(unwrap_jso(name))), HtmlCollection);
    }
    [_hasAttribute](name) {
      return this[_hasAttribute_1](name);
    }
    [_hasAttribute_1](name) {
      return dart.as(wrap_jso(this.raw.hasAttribute(unwrap_jso(name))), core.bool);
    }
    [_hasAttributeNS](namespaceURI, localName) {
      return this[_hasAttributeNS_1](namespaceURI, localName);
    }
    [_hasAttributeNS_1](namespaceURI, localName) {
      return dart.as(wrap_jso(this.raw.hasAttributeNS(unwrap_jso(namespaceURI), unwrap_jso(localName))), core.bool);
    }
    [_removeAttribute](name) {
      this[_removeAttribute_1](name);
      return;
    }
    [_removeAttribute_1](name) {
      return wrap_jso(this.raw.removeAttribute(unwrap_jso(name)));
    }
    [_removeAttributeNS](namespaceURI, localName) {
      this[_removeAttributeNS_1](namespaceURI, localName);
      return;
    }
    [_removeAttributeNS_1](namespaceURI, localName) {
      return wrap_jso(this.raw.removeAttributeNS(unwrap_jso(namespaceURI), unwrap_jso(localName)));
    }
    requestFullscreen() {
      this[_requestFullscreen_1]();
      return;
    }
    [_requestFullscreen_1]() {
      return wrap_jso(this.raw.requestFullscreen());
    }
    requestPointerLock() {
      this[_requestPointerLock_1]();
      return;
    }
    [_requestPointerLock_1]() {
      return wrap_jso(this.raw.requestPointerLock());
    }
    [_scrollIntoView](alignWithTop) {
      if (alignWithTop === void 0)
        alignWithTop = null;
      if (alignWithTop != null) {
        this[_scrollIntoView_1](alignWithTop);
        return;
      }
      this[_scrollIntoView_2]();
      return;
    }
    [_scrollIntoView_1](alignWithTop) {
      return wrap_jso(this.raw.scrollIntoView(unwrap_jso(alignWithTop)));
    }
    [_scrollIntoView_2]() {
      return wrap_jso(this.raw.scrollIntoView());
    }
    [_scrollIntoViewIfNeeded](centerIfNeeded) {
      if (centerIfNeeded === void 0)
        centerIfNeeded = null;
      if (centerIfNeeded != null) {
        this[_scrollIntoViewIfNeeded_1](centerIfNeeded);
        return;
      }
      this[_scrollIntoViewIfNeeded_2]();
      return;
    }
    [_scrollIntoViewIfNeeded_1](centerIfNeeded) {
      return wrap_jso(this.raw.scrollIntoViewIfNeeded(unwrap_jso(centerIfNeeded)));
    }
    [_scrollIntoViewIfNeeded_2]() {
      return wrap_jso(this.raw.scrollIntoViewIfNeeded());
    }
    setAttribute(name, value) {
      this[_setAttribute_1](name, value);
      return;
    }
    [_setAttribute_1](name, value) {
      return wrap_jso(this.raw.setAttribute(unwrap_jso(name), unwrap_jso(value)));
    }
    setAttributeNS(namespaceURI, qualifiedName, value) {
      this[_setAttributeNS_1](namespaceURI, qualifiedName, value);
      return;
    }
    [_setAttributeNS_1](namespaceURI, qualifiedName, value) {
      return wrap_jso(this.raw.setAttributeNS(unwrap_jso(namespaceURI), unwrap_jso(qualifiedName), unwrap_jso(value)));
    }
    get nextElementSibling() {
      return dart.as(wrap_jso(this.raw.nextElementSibling), Element);
    }
    get previousElementSibling() {
      return dart.as(wrap_jso(this.raw.previousElementSibling), Element);
    }
    get [_childElementCount]() {
      return dart.as(wrap_jso(this.raw.childElementCount), core.int);
    }
    get [_children]() {
      return dart.as(wrap_jso(this.raw.children), core.List$(Node));
    }
    get [_firstElementChild]() {
      return dart.as(wrap_jso(this.raw.firstElementChild), Element);
    }
    get [_lastElementChild]() {
      return dart.as(wrap_jso(this.raw.lastElementChild), Element);
    }
    querySelector(selectors) {
      return this[_querySelector_1](selectors);
    }
    [_querySelector_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelector(unwrap_jso(selectors))), Element);
    }
    [_querySelectorAll](selectors) {
      return this[_querySelectorAll_1](selectors);
    }
    [_querySelectorAll_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelectorAll(unwrap_jso(selectors))), NodeList);
    }
    get onAbort() {
      return Element.abortEvent.forElement(this);
    }
    get onBeforeCopy() {
      return Element.beforeCopyEvent.forElement(this);
    }
    get onBeforeCut() {
      return Element.beforeCutEvent.forElement(this);
    }
    get onBeforePaste() {
      return Element.beforePasteEvent.forElement(this);
    }
    get onBlur() {
      return Element.blurEvent.forElement(this);
    }
    get onCanPlay() {
      return Element.canPlayEvent.forElement(this);
    }
    get onCanPlayThrough() {
      return Element.canPlayThroughEvent.forElement(this);
    }
    get onChange() {
      return Element.changeEvent.forElement(this);
    }
    get onClick() {
      return Element.clickEvent.forElement(this);
    }
    get onContextMenu() {
      return Element.contextMenuEvent.forElement(this);
    }
    get onCopy() {
      return Element.copyEvent.forElement(this);
    }
    get onCut() {
      return Element.cutEvent.forElement(this);
    }
    get onDoubleClick() {
      return Element.doubleClickEvent.forElement(this);
    }
    get onDrag() {
      return Element.dragEvent.forElement(this);
    }
    get onDragEnd() {
      return Element.dragEndEvent.forElement(this);
    }
    get onDragEnter() {
      return Element.dragEnterEvent.forElement(this);
    }
    get onDragLeave() {
      return Element.dragLeaveEvent.forElement(this);
    }
    get onDragOver() {
      return Element.dragOverEvent.forElement(this);
    }
    get onDragStart() {
      return Element.dragStartEvent.forElement(this);
    }
    get onDrop() {
      return Element.dropEvent.forElement(this);
    }
    get onDurationChange() {
      return Element.durationChangeEvent.forElement(this);
    }
    get onEmptied() {
      return Element.emptiedEvent.forElement(this);
    }
    get onEnded() {
      return Element.endedEvent.forElement(this);
    }
    get onError() {
      return Element.errorEvent.forElement(this);
    }
    get onFocus() {
      return Element.focusEvent.forElement(this);
    }
    get onInput() {
      return Element.inputEvent.forElement(this);
    }
    get onInvalid() {
      return Element.invalidEvent.forElement(this);
    }
    get onKeyDown() {
      return Element.keyDownEvent.forElement(this);
    }
    get onKeyPress() {
      return Element.keyPressEvent.forElement(this);
    }
    get onKeyUp() {
      return Element.keyUpEvent.forElement(this);
    }
    get onLoad() {
      return Element.loadEvent.forElement(this);
    }
    get onLoadedData() {
      return Element.loadedDataEvent.forElement(this);
    }
    get onLoadedMetadata() {
      return Element.loadedMetadataEvent.forElement(this);
    }
    get onMouseDown() {
      return Element.mouseDownEvent.forElement(this);
    }
    get onMouseEnter() {
      return Element.mouseEnterEvent.forElement(this);
    }
    get onMouseLeave() {
      return Element.mouseLeaveEvent.forElement(this);
    }
    get onMouseMove() {
      return Element.mouseMoveEvent.forElement(this);
    }
    get onMouseOut() {
      return Element.mouseOutEvent.forElement(this);
    }
    get onMouseOver() {
      return Element.mouseOverEvent.forElement(this);
    }
    get onMouseUp() {
      return Element.mouseUpEvent.forElement(this);
    }
    get onMouseWheel() {
      return dart.as(dart.dsend(/* Unimplemented unknown name */mouseWheelEvent, 'forElement', this), ElementStream$(dart.dynamic));
    }
    get onPaste() {
      return Element.pasteEvent.forElement(this);
    }
    get onPause() {
      return Element.pauseEvent.forElement(this);
    }
    get onPlay() {
      return Element.playEvent.forElement(this);
    }
    get onPlaying() {
      return Element.playingEvent.forElement(this);
    }
    get onRateChange() {
      return Element.rateChangeEvent.forElement(this);
    }
    get onReset() {
      return Element.resetEvent.forElement(this);
    }
    get onResize() {
      return Element.resizeEvent.forElement(this);
    }
    get onScroll() {
      return Element.scrollEvent.forElement(this);
    }
    get onSearch() {
      return Element.searchEvent.forElement(this);
    }
    get onSeeked() {
      return Element.seekedEvent.forElement(this);
    }
    get onSeeking() {
      return Element.seekingEvent.forElement(this);
    }
    get onSelect() {
      return Element.selectEvent.forElement(this);
    }
    get onSelectStart() {
      return Element.selectStartEvent.forElement(this);
    }
    get onStalled() {
      return Element.stalledEvent.forElement(this);
    }
    get onSubmit() {
      return Element.submitEvent.forElement(this);
    }
    get onSuspend() {
      return Element.suspendEvent.forElement(this);
    }
    get onTimeUpdate() {
      return Element.timeUpdateEvent.forElement(this);
    }
    get onTouchCancel() {
      return Element.touchCancelEvent.forElement(this);
    }
    get onTouchEnd() {
      return Element.touchEndEvent.forElement(this);
    }
    get onTouchEnter() {
      return Element.touchEnterEvent.forElement(this);
    }
    get onTouchLeave() {
      return Element.touchLeaveEvent.forElement(this);
    }
    get onTouchMove() {
      return Element.touchMoveEvent.forElement(this);
    }
    get onTouchStart() {
      return Element.touchStartEvent.forElement(this);
    }
    get onTransitionEnd() {
      return dart.as(dart.dsend(/* Unimplemented unknown name */transitionEndEvent, 'forElement', this), ElementStream$(TransitionEvent));
    }
    get onVolumeChange() {
      return Element.volumeChangeEvent.forElement(this);
    }
    get onWaiting() {
      return Element.waitingEvent.forElement(this);
    }
    get onFullscreenChange() {
      return Element.fullscreenChangeEvent.forElement(this);
    }
    get onFullscreenError() {
      return Element.fullscreenErrorEvent.forElement(this);
    }
  }
  Element[dart.implements] = () => [ParentNode, ChildNode];
  dart.defineNamedConstructor(Element, 'created');
  dart.defineNamedConstructor(Element, 'internal_');
  dart.setSignature(Element, {
    constructors: () => ({
      html: [Element, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      created: [Element, []],
      tag: [Element, [core.String], [core.String]],
      a: [Element, []],
      article: [Element, []],
      aside: [Element, []],
      audio: [Element, []],
      br: [Element, []],
      canvas: [Element, []],
      div: [Element, []],
      footer: [Element, []],
      header: [Element, []],
      hr: [Element, []],
      iframe: [Element, []],
      img: [Element, []],
      li: [Element, []],
      nav: [Element, []],
      ol: [Element, []],
      option: [Element, []],
      p: [Element, []],
      pre: [Element, []],
      section: [Element, []],
      select: [Element, []],
      span: [Element, []],
      svg: [Element, []],
      table: [Element, []],
      td: [Element, []],
      textarea: [Element, []],
      th: [Element, []],
      tr: [Element, []],
      ul: [Element, []],
      video: [Element, []],
      _: [Element, []],
      _internalWrap: [Element, []],
      internal_: [Element, []]
    }),
    methods: () => ({
      querySelectorAll: [ElementList$(Element), [core.String]],
      query: [Element, [core.String]],
      queryAll: [ElementList$(Element), [core.String]],
      getNamespacedAttributes: [core.Map$(core.String, core.String), [core.String]],
      getComputedStyle: [CssStyleDeclaration, [], [core.String]],
      appendText: [dart.void, [core.String]],
      appendHtml: [dart.void, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      attached: [dart.void, []],
      detached: [dart.void, []],
      enteredView: [dart.void, []],
      leftView: [dart.void, []],
      animate: [dart.dynamic, [core.Iterable$(core.Map$(core.String, dart.dynamic))], [dart.dynamic]],
      attributeChanged: [dart.void, [core.String, core.String, core.String]],
      scrollIntoView: [dart.void, [], [dart.dynamic]],
      insertAdjacentHtml: [dart.void, [core.String, core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      [_insertAdjacentNode]: [dart.void, [core.String, Node]],
      matchesWithAncestors: [core.bool, [core.String]],
      offsetTo: [math.Point, [Element]],
      createFragment: [DocumentFragment, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      setInnerHtml: [dart.void, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      click: [dart.void, []],
      click: [dart.void, []],
      [_click_1]: [dart.void, []],
      blur: [dart.void, []],
      [_blur_1]: [dart.void, []],
      focus: [dart.void, []],
      [_focus_1]: [dart.void, []],
      getAnimationPlayers: [core.List$(dart.dynamic), []],
      [_getAnimationPlayers_1]: [core.List$(dart.dynamic), []],
      getAttribute: [core.String, [core.String]],
      [_getAttribute_1]: [core.String, [dart.dynamic]],
      getAttributeNS: [core.String, [core.String, core.String]],
      [_getAttributeNS_1]: [core.String, [dart.dynamic, dart.dynamic]],
      getBoundingClientRect: [math.Rectangle, []],
      [_getBoundingClientRect_1]: [math.Rectangle, []],
      getClientRects: [_ClientRectList, []],
      [_getClientRects_1]: [_ClientRectList, []],
      getDestinationInsertionPoints: [NodeList, []],
      [_getDestinationInsertionPoints_1]: [NodeList, []],
      getElementsByClassName: [HtmlCollection, [core.String]],
      [_getElementsByClassName_1]: [HtmlCollection, [dart.dynamic]],
      [_getElementsByTagName]: [HtmlCollection, [core.String]],
      [_getElementsByTagName_1]: [HtmlCollection, [dart.dynamic]],
      [_hasAttribute]: [core.bool, [core.String]],
      [_hasAttribute_1]: [core.bool, [dart.dynamic]],
      [_hasAttributeNS]: [core.bool, [core.String, core.String]],
      [_hasAttributeNS_1]: [core.bool, [dart.dynamic, dart.dynamic]],
      [_removeAttribute]: [dart.void, [core.String]],
      [_removeAttribute_1]: [dart.void, [dart.dynamic]],
      [_removeAttributeNS]: [dart.void, [core.String, core.String]],
      [_removeAttributeNS_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      requestFullscreen: [dart.void, []],
      [_requestFullscreen_1]: [dart.void, []],
      requestPointerLock: [dart.void, []],
      [_requestPointerLock_1]: [dart.void, []],
      [_scrollIntoView]: [dart.void, [], [core.bool]],
      [_scrollIntoView_1]: [dart.void, [dart.dynamic]],
      [_scrollIntoView_2]: [dart.void, []],
      [_scrollIntoViewIfNeeded]: [dart.void, [], [core.bool]],
      [_scrollIntoViewIfNeeded_1]: [dart.void, [dart.dynamic]],
      [_scrollIntoViewIfNeeded_2]: [dart.void, []],
      setAttribute: [dart.void, [core.String, core.String]],
      [_setAttribute_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      setAttributeNS: [dart.void, [core.String, core.String, core.String]],
      [_setAttributeNS_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      querySelector: [Element, [core.String]],
      [_querySelector_1]: [Element, [dart.dynamic]],
      [_querySelectorAll]: [NodeList, [core.String]],
      [_querySelectorAll_1]: [NodeList, [dart.dynamic]]
    }),
    statics: () => ({
      isTagSupported: [core.bool, [core.String]],
      _offsetToHelper: [math.Point, [Element, Element]],
      _hasCorruptedAttributes: [core.bool, [Element]],
      _hasCorruptedAttributesAdditionalCheck: [core.bool, [Element]],
      _safeTagName: [core.String, [dart.dynamic]],
      internalCreateElement: [Element, []]
    }),
    names: ['isTagSupported', '_offsetToHelper', '_hasCorruptedAttributes', '_hasCorruptedAttributesAdditionalCheck', '_safeTagName', 'internalCreateElement']
  });
  Element[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Element')), dart.const(new _js_helper.Native("Element"))];
  class HtmlElement extends Element {
    static new() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    created() {
      super.created();
    }
    internal_() {
      super.internal_();
    }
    static internalCreateHtmlElement() {
      return HtmlElement._internalWrap();
    }
    static _internalWrap() {
      return new HtmlElement.internal_();
    }
  }
  dart.defineNamedConstructor(HtmlElement, 'created');
  dart.defineNamedConstructor(HtmlElement, 'internal_');
  dart.setSignature(HtmlElement, {
    constructors: () => ({
      new: [HtmlElement, []],
      created: [HtmlElement, []],
      internal_: [HtmlElement, []],
      _internalWrap: [HtmlElement, []]
    }),
    statics: () => ({internalCreateHtmlElement: [HtmlElement, []]}),
    names: ['internalCreateHtmlElement']
  });
  HtmlElement[dart.metadata] = () => [dart.const(new _js_helper.Native("HTMLElement"))];
  class AnchorElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(opts) {
      let href = opts && 'href' in opts ? opts.href : null;
      let e = exports.document.createElement("a");
      if (href != null)
        dart.dput(e, 'href', href);
      return dart.as(e, AnchorElement);
    }
    static internalCreateAnchorElement() {
      return AnchorElement._internalWrap();
    }
    static _internalWrap() {
      return new AnchorElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get download() {
      return dart.as(wrap_jso(this.raw.download), core.String);
    }
    set download(val) {
      return this.raw.download = unwrap_jso(val);
    }
    get hreflang() {
      return dart.as(wrap_jso(this.raw.hreflang), core.String);
    }
    set hreflang(val) {
      return this.raw.hreflang = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get rel() {
      return dart.as(wrap_jso(this.raw.rel), core.String);
    }
    set rel(val) {
      return this.raw.rel = unwrap_jso(val);
    }
    get target() {
      return dart.as(wrap_jso(this.raw.target), core.String);
    }
    set target(val) {
      return this.raw.target = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
    get hash() {
      return dart.as(wrap_jso(this.raw.hash), core.String);
    }
    set hash(val) {
      return this.raw.hash = unwrap_jso(val);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), core.String);
    }
    set host(val) {
      return this.raw.host = unwrap_jso(val);
    }
    get hostname() {
      return dart.as(wrap_jso(this.raw.hostname), core.String);
    }
    set hostname(val) {
      return this.raw.hostname = unwrap_jso(val);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    set href(val) {
      return this.raw.href = unwrap_jso(val);
    }
    get origin() {
      return dart.as(wrap_jso(this.raw.origin), core.String);
    }
    get password() {
      return dart.as(wrap_jso(this.raw.password), core.String);
    }
    set password(val) {
      return this.raw.password = unwrap_jso(val);
    }
    get pathname() {
      return dart.as(wrap_jso(this.raw.pathname), core.String);
    }
    set pathname(val) {
      return this.raw.pathname = unwrap_jso(val);
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), core.String);
    }
    set port(val) {
      return this.raw.port = unwrap_jso(val);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    set protocol(val) {
      return this.raw.protocol = unwrap_jso(val);
    }
    get search() {
      return dart.as(wrap_jso(this.raw.search), core.String);
    }
    set search(val) {
      return this.raw.search = unwrap_jso(val);
    }
    get username() {
      return dart.as(wrap_jso(this.raw.username), core.String);
    }
    set username(val) {
      return this.raw.username = unwrap_jso(val);
    }
  }
  AnchorElement[dart.implements] = () => [UrlUtils];
  dart.defineNamedConstructor(AnchorElement, 'internal_');
  dart.defineNamedConstructor(AnchorElement, 'created');
  dart.setSignature(AnchorElement, {
    constructors: () => ({
      _: [AnchorElement, []],
      new: [AnchorElement, [], {href: core.String}],
      _internalWrap: [AnchorElement, []],
      internal_: [AnchorElement, []],
      created: [AnchorElement, []]
    }),
    statics: () => ({internalCreateAnchorElement: [AnchorElement, []]}),
    names: ['internalCreateAnchorElement']
  });
  AnchorElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLAnchorElement')), dart.const(new _js_helper.Native("HTMLAnchorElement"))];
  class BodyElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("body"), BodyElement);
    }
    static internalCreateBodyElement() {
      return BodyElement._internalWrap();
    }
    static _internalWrap() {
      return new BodyElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get onBlur() {
      return BodyElement.blurEvent.forElement(this);
    }
    get onError() {
      return BodyElement.errorEvent.forElement(this);
    }
    get onFocus() {
      return BodyElement.focusEvent.forElement(this);
    }
    get onHashChange() {
      return BodyElement.hashChangeEvent.forElement(this);
    }
    get onLoad() {
      return BodyElement.loadEvent.forElement(this);
    }
    get onMessage() {
      return BodyElement.messageEvent.forElement(this);
    }
    get onOffline() {
      return BodyElement.offlineEvent.forElement(this);
    }
    get onOnline() {
      return BodyElement.onlineEvent.forElement(this);
    }
    get onPopState() {
      return BodyElement.popStateEvent.forElement(this);
    }
    get onResize() {
      return BodyElement.resizeEvent.forElement(this);
    }
    get onScroll() {
      return BodyElement.scrollEvent.forElement(this);
    }
    get onStorage() {
      return BodyElement.storageEvent.forElement(this);
    }
    get onUnload() {
      return BodyElement.unloadEvent.forElement(this);
    }
  }
  BodyElement[dart.implements] = () => [WindowEventHandlers];
  dart.defineNamedConstructor(BodyElement, 'internal_');
  dart.defineNamedConstructor(BodyElement, 'created');
  dart.setSignature(BodyElement, {
    constructors: () => ({
      _: [BodyElement, []],
      new: [BodyElement, []],
      _internalWrap: [BodyElement, []],
      internal_: [BodyElement, []],
      created: [BodyElement, []]
    }),
    statics: () => ({internalCreateBodyElement: [BodyElement, []]}),
    names: ['internalCreateBodyElement']
  });
  BodyElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLBodyElement')), dart.const(new _js_helper.Native("HTMLBodyElement"))];
  const _eventType = Symbol('_eventType');
  const _forElementList = Symbol('_forElementList');
  const EventStreamProvider$ = dart.generic(function(T) {
    class EventStreamProvider extends core.Object {
      EventStreamProvider(eventType) {
        this[_eventType] = eventType;
      }
      forTarget(e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new (_EventStream$(T))(e, this[_eventType], useCapture);
      }
      forElement(e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new (_ElementEventStreamImpl$(T))(e, this[_eventType], useCapture);
      }
      [_forElementList](e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new _ElementListEventStreamImpl(dart.as(e, core.Iterable$(Element)), this[_eventType], useCapture);
      }
      getEventType(target) {
        return this[_eventType];
      }
    }
    dart.setSignature(EventStreamProvider, {
      constructors: () => ({EventStreamProvider: [EventStreamProvider$(T), [core.String]]}),
      methods: () => ({
        forTarget: [async.Stream$(T), [EventTarget], {useCapture: core.bool}],
        forElement: [ElementStream$(T), [Element], {useCapture: core.bool}],
        [_forElementList]: [ElementStream$(T), [ElementList], {useCapture: core.bool}],
        getEventType: [core.String, [EventTarget]]
      })
    });
    return EventStreamProvider;
  });
  let EventStreamProvider = EventStreamProvider$();
  const _createEvent = Symbol('_createEvent');
  const _initEvent = Symbol('_initEvent');
  const _selector = Symbol('_selector');
  const _get_currentTarget = Symbol('_get_currentTarget');
  const _get_target = Symbol('_get_target');
  const _initEvent_1 = Symbol('_initEvent_1');
  const _preventDefault_1 = Symbol('_preventDefault_1');
  const _stopImmediatePropagation_1 = Symbol('_stopImmediatePropagation_1');
  const _stopPropagation_1 = Symbol('_stopPropagation_1');
  class Event extends DartHtmlDomObject {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      return Event.eventType('Event', type, {canBubble: canBubble, cancelable: cancelable});
    }
    static eventType(type, name, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      let e = exports.document[_createEvent](type);
      e[_initEvent](name, canBubble, cancelable);
      return e;
    }
    get matchingTarget() {
      if (this[_selector] == null) {
        dart.throw(new core.UnsupportedError('Cannot call matchingTarget if this Event did' + ' not arise as a result of event delegation.'));
      }
      let currentTarget = this.currentTarget;
      let target = this.target;
      let matchedTarget = null;
      do {
        if (dart.notNull(dart.as(dart.dcall(target.matches, this[_selector]), core.bool)))
          return dart.as(target, Element);
        target = dart.as(dart.dload(target, 'parent'), EventTarget);
      } while (target != null && !dart.equals(target, dart.dload(currentTarget, 'parent')));
      dart.throw(new core.StateError('No selector matched for populating matchedTarget.'));
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateEvent() {
      return Event._internalWrap();
    }
    static _internalWrap() {
      return new Event.internal_();
    }
    internal_() {
      this[_selector] = null;
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get bubbles() {
      return dart.as(wrap_jso(this.raw.bubbles), core.bool);
    }
    get cancelable() {
      return dart.as(wrap_jso(this.raw.cancelable), core.bool);
    }
    get clipboardData() {
      return wrap_jso(this.raw.clipboardData);
    }
    get currentTarget() {
      return _convertNativeToDart_EventTarget(this[_get_currentTarget]);
    }
    get [_get_currentTarget]() {
      return wrap_jso(this.raw.currentTarget);
    }
    get defaultPrevented() {
      return dart.as(wrap_jso(this.raw.defaultPrevented), core.bool);
    }
    get eventPhase() {
      return dart.as(wrap_jso(this.raw.eventPhase), core.int);
    }
    get path() {
      return dart.as(wrap_jso(this.raw.path), core.List$(Node));
    }
    get target() {
      return _convertNativeToDart_EventTarget(this[_get_target]);
    }
    get [_get_target]() {
      return wrap_jso(this.raw.target);
    }
    get timeStamp() {
      return dart.as(wrap_jso(this.raw.timeStamp), core.int);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    [_initEvent](eventTypeArg, canBubbleArg, cancelableArg) {
      this[_initEvent_1](eventTypeArg, canBubbleArg, cancelableArg);
      return;
    }
    [_initEvent_1](eventTypeArg, canBubbleArg, cancelableArg) {
      return wrap_jso(this.raw.initEvent(unwrap_jso(eventTypeArg), unwrap_jso(canBubbleArg), unwrap_jso(cancelableArg)));
    }
    preventDefault() {
      this[_preventDefault_1]();
      return;
    }
    [_preventDefault_1]() {
      return wrap_jso(this.raw.preventDefault());
    }
    stopImmediatePropagation() {
      this[_stopImmediatePropagation_1]();
      return;
    }
    [_stopImmediatePropagation_1]() {
      return wrap_jso(this.raw.stopImmediatePropagation());
    }
    stopPropagation() {
      this[_stopPropagation_1]();
      return;
    }
    [_stopPropagation_1]() {
      return wrap_jso(this.raw.stopPropagation());
    }
  }
  dart.defineNamedConstructor(Event, 'internal_');
  dart.setSignature(Event, {
    constructors: () => ({
      new: [Event, [core.String], {canBubble: core.bool, cancelable: core.bool}],
      eventType: [Event, [core.String, core.String], {canBubble: core.bool, cancelable: core.bool}],
      _: [Event, []],
      _internalWrap: [Event, []],
      internal_: [Event, []]
    }),
    methods: () => ({
      [_initEvent]: [dart.void, [core.String, core.bool, core.bool]],
      [_initEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      preventDefault: [dart.void, []],
      [_preventDefault_1]: [dart.void, []],
      stopImmediatePropagation: [dart.void, []],
      [_stopImmediatePropagation_1]: [dart.void, []],
      stopPropagation: [dart.void, []],
      [_stopPropagation_1]: [dart.void, []]
    }),
    statics: () => ({internalCreateEvent: [Event, []]}),
    names: ['internalCreateEvent']
  });
  Event[dart.metadata] = () => [dart.const(new _metadata.DomName('Event')), dart.const(new _js_helper.Native("Event,InputEvent,ClipboardEvent"))];
  BodyElement.blurEvent = dart.const(new (EventStreamProvider$(Event))('blur'));
  BodyElement.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  BodyElement.focusEvent = dart.const(new (EventStreamProvider$(Event))('focus'));
  BodyElement.hashChangeEvent = dart.const(new (EventStreamProvider$(Event))('hashchange'));
  BodyElement.loadEvent = dart.const(new (EventStreamProvider$(Event))('load'));
  const _initMessageEvent = Symbol('_initMessageEvent');
  const _get_data = Symbol('_get_data');
  const _get_source = Symbol('_get_source');
  const _initMessageEvent_1 = Symbol('_initMessageEvent_1');
  class MessageEvent extends Event {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : false;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : false;
      let data = opts && 'data' in opts ? opts.data : null;
      let origin = opts && 'origin' in opts ? opts.origin : null;
      let lastEventId = opts && 'lastEventId' in opts ? opts.lastEventId : null;
      let source = opts && 'source' in opts ? opts.source : null;
      let messagePorts = opts && 'messagePorts' in opts ? opts.messagePorts : null;
      if (source == null) {
        source = exports.window;
      }
      let event = exports.document[_createEvent]("MessageEvent");
      dart.dcall(event[_initMessageEvent], type, canBubble, cancelable, data, origin, lastEventId, source, messagePorts);
      return dart.as(event, MessageEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMessageEvent() {
      return MessageEvent._internalWrap();
    }
    static _internalWrap() {
      return new MessageEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get data() {
      return html_common.convertNativeToDart_SerializedScriptValue(this[_get_data]);
    }
    get [_get_data]() {
      return wrap_jso(this.raw.data);
    }
    get lastEventId() {
      return dart.as(wrap_jso(this.raw.lastEventId), core.String);
    }
    get origin() {
      return dart.as(wrap_jso(this.raw.origin), core.String);
    }
    get source() {
      return _convertNativeToDart_EventTarget(this[_get_source]);
    }
    get [_get_source]() {
      return wrap_jso(this.raw.source);
    }
    [_initMessageEvent](typeArg, canBubbleArg, cancelableArg, dataArg, originArg, lastEventIdArg, sourceArg, messagePorts) {
      this[_initMessageEvent_1](typeArg, canBubbleArg, cancelableArg, dataArg, originArg, lastEventIdArg, sourceArg, messagePorts);
      return;
    }
    [_initMessageEvent_1](typeArg, canBubbleArg, cancelableArg, dataArg, originArg, lastEventIdArg, sourceArg, messagePorts) {
      return wrap_jso(this.raw.initMessageEvent(unwrap_jso(typeArg), unwrap_jso(canBubbleArg), unwrap_jso(cancelableArg), unwrap_jso(dataArg), unwrap_jso(originArg), unwrap_jso(lastEventIdArg), unwrap_jso(sourceArg), unwrap_jso(messagePorts)));
    }
  }
  dart.defineNamedConstructor(MessageEvent, 'internal_');
  dart.setSignature(MessageEvent, {
    constructors: () => ({
      new: [MessageEvent, [core.String], {canBubble: core.bool, cancelable: core.bool, data: core.Object, origin: core.String, lastEventId: core.String, source: Window, messagePorts: core.List}],
      _: [MessageEvent, []],
      _internalWrap: [MessageEvent, []],
      internal_: [MessageEvent, []]
    }),
    methods: () => ({
      [_initMessageEvent]: [dart.void, [core.String, core.bool, core.bool, core.Object, core.String, core.String, Window, core.List$(MessagePort)]],
      [_initMessageEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, Window, core.List$(MessagePort)]]
    }),
    statics: () => ({internalCreateMessageEvent: [MessageEvent, []]}),
    names: ['internalCreateMessageEvent']
  });
  MessageEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('MessageEvent')), dart.const(new _js_helper.Native("MessageEvent"))];
  BodyElement.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  BodyElement.offlineEvent = dart.const(new (EventStreamProvider$(Event))('offline'));
  BodyElement.onlineEvent = dart.const(new (EventStreamProvider$(Event))('online'));
  const _get_state = Symbol('_get_state');
  class PopStateEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePopStateEvent() {
      return PopStateEvent._internalWrap();
    }
    static _internalWrap() {
      return new PopStateEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get state() {
      return html_common.convertNativeToDart_SerializedScriptValue(this[_get_state]);
    }
    get [_get_state]() {
      return wrap_jso(this.raw.state);
    }
  }
  dart.defineNamedConstructor(PopStateEvent, 'internal_');
  dart.setSignature(PopStateEvent, {
    constructors: () => ({
      _: [PopStateEvent, []],
      _internalWrap: [PopStateEvent, []],
      internal_: [PopStateEvent, []]
    }),
    statics: () => ({internalCreatePopStateEvent: [PopStateEvent, []]}),
    names: ['internalCreatePopStateEvent']
  });
  PopStateEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PopStateEvent')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _js_helper.Native("PopStateEvent"))];
  BodyElement.popStateEvent = dart.const(new (EventStreamProvider$(PopStateEvent))('popstate'));
  BodyElement.resizeEvent = dart.const(new (EventStreamProvider$(Event))('resize'));
  BodyElement.scrollEvent = dart.const(new (EventStreamProvider$(Event))('scroll'));
  const _initStorageEvent = Symbol('_initStorageEvent');
  const _initStorageEvent_1 = Symbol('_initStorageEvent_1');
  class StorageEvent extends Event {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : false;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : false;
      let key = opts && 'key' in opts ? opts.key : null;
      let oldValue = opts && 'oldValue' in opts ? opts.oldValue : null;
      let newValue = opts && 'newValue' in opts ? opts.newValue : null;
      let url = opts && 'url' in opts ? opts.url : null;
      let storageArea = opts && 'storageArea' in opts ? opts.storageArea : null;
      let e = exports.document[_createEvent]("StorageEvent");
      dart.dcall(e[_initStorageEvent], type, canBubble, cancelable, key, oldValue, newValue, url, storageArea);
      return dart.as(e, StorageEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStorageEvent() {
      return StorageEvent._internalWrap();
    }
    static _internalWrap() {
      return new StorageEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get key() {
      return dart.as(wrap_jso(this.raw.key), core.String);
    }
    get newValue() {
      return dart.as(wrap_jso(this.raw.newValue), core.String);
    }
    get oldValue() {
      return dart.as(wrap_jso(this.raw.oldValue), core.String);
    }
    get storageArea() {
      return dart.as(wrap_jso(this.raw.storageArea), Storage);
    }
    get url() {
      return dart.as(wrap_jso(this.raw.url), core.String);
    }
    [_initStorageEvent](typeArg, canBubbleArg, cancelableArg, keyArg, oldValueArg, newValueArg, urlArg, storageAreaArg) {
      this[_initStorageEvent_1](typeArg, canBubbleArg, cancelableArg, keyArg, oldValueArg, newValueArg, urlArg, storageAreaArg);
      return;
    }
    [_initStorageEvent_1](typeArg, canBubbleArg, cancelableArg, keyArg, oldValueArg, newValueArg, urlArg, storageAreaArg) {
      return wrap_jso(this.raw.initStorageEvent(unwrap_jso(typeArg), unwrap_jso(canBubbleArg), unwrap_jso(cancelableArg), unwrap_jso(keyArg), unwrap_jso(oldValueArg), unwrap_jso(newValueArg), unwrap_jso(urlArg), unwrap_jso(storageAreaArg)));
    }
  }
  dart.defineNamedConstructor(StorageEvent, 'internal_');
  dart.setSignature(StorageEvent, {
    constructors: () => ({
      new: [StorageEvent, [core.String], {canBubble: core.bool, cancelable: core.bool, key: core.String, oldValue: core.String, newValue: core.String, url: core.String, storageArea: Storage}],
      _: [StorageEvent, []],
      _internalWrap: [StorageEvent, []],
      internal_: [StorageEvent, []]
    }),
    methods: () => ({
      [_initStorageEvent]: [dart.void, [core.String, core.bool, core.bool, core.String, core.String, core.String, core.String, Storage]],
      [_initStorageEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, Storage]]
    }),
    statics: () => ({internalCreateStorageEvent: [StorageEvent, []]}),
    names: ['internalCreateStorageEvent']
  });
  StorageEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('StorageEvent')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("StorageEvent"))];
  BodyElement.storageEvent = dart.const(new (EventStreamProvider$(StorageEvent))('storage'));
  BodyElement.unloadEvent = dart.const(new (EventStreamProvider$(Event))('unload'));
  const _checkValidity_1 = Symbol('_checkValidity_1');
  const _setCustomValidity_1 = Symbol('_setCustomValidity_1');
  class ButtonElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("button"), ButtonElement);
    }
    static internalCreateButtonElement() {
      return ButtonElement._internalWrap();
    }
    static _internalWrap() {
      return new ButtonElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get autofocus() {
      return dart.as(wrap_jso(this.raw.autofocus), core.bool);
    }
    set autofocus(val) {
      return this.raw.autofocus = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get formAction() {
      return dart.as(wrap_jso(this.raw.formAction), core.String);
    }
    set formAction(val) {
      return this.raw.formAction = unwrap_jso(val);
    }
    get formEnctype() {
      return dart.as(wrap_jso(this.raw.formEnctype), core.String);
    }
    set formEnctype(val) {
      return this.raw.formEnctype = unwrap_jso(val);
    }
    get formMethod() {
      return dart.as(wrap_jso(this.raw.formMethod), core.String);
    }
    set formMethod(val) {
      return this.raw.formMethod = unwrap_jso(val);
    }
    get formNoValidate() {
      return dart.as(wrap_jso(this.raw.formNoValidate), core.bool);
    }
    set formNoValidate(val) {
      return this.raw.formNoValidate = unwrap_jso(val);
    }
    get formTarget() {
      return dart.as(wrap_jso(this.raw.formTarget), core.String);
    }
    set formTarget(val) {
      return this.raw.formTarget = unwrap_jso(val);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
  }
  dart.defineNamedConstructor(ButtonElement, 'internal_');
  dart.defineNamedConstructor(ButtonElement, 'created');
  dart.setSignature(ButtonElement, {
    constructors: () => ({
      _: [ButtonElement, []],
      new: [ButtonElement, []],
      _internalWrap: [ButtonElement, []],
      internal_: [ButtonElement, []],
      created: [ButtonElement, []]
    }),
    methods: () => ({
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateButtonElement: [ButtonElement, []]}),
    names: ['internalCreateButtonElement']
  });
  ButtonElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLButtonElement')), dart.const(new _js_helper.Native("HTMLButtonElement"))];
  class ChildNode extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get nextElementSibling() {
      return dart.as(wrap_jso(this.raw.nextElementSibling), Element);
    }
    get previousElementSibling() {
      return dart.as(wrap_jso(this.raw.previousElementSibling), Element);
    }
    remove() {
      return wrap_jso(this.raw.remove());
    }
  }
  dart.setSignature(ChildNode, {
    constructors: () => ({_: [ChildNode, []]}),
    methods: () => ({remove: [dart.void, []]})
  });
  ChildNode[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ChildNode')), dart.const(new _metadata.Experimental())];
  const _isConsoleDefined = Symbol('_isConsoleDefined');
  class Console extends DartHtmlDomObject {
    _safe() {
      super.DartHtmlDomObject();
    }
    get [_isConsoleDefined]() {
      return typeof console != "undefined";
    }
    get memory() {
      return dart.notNull(this[_isConsoleDefined]) ? console.memory : null;
    }
    assertCondition(condition, arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.assertCondition(condition, arg) : null;
    }
    clear(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.clear(arg) : null;
    }
    count(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.count(arg) : null;
    }
    debug(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.debug(arg) : null;
    }
    dir(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.dir(arg) : null;
    }
    dirxml(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.dirxml(arg) : null;
    }
    error(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.error(arg) : null;
    }
    group(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.group(arg) : null;
    }
    groupCollapsed(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.groupCollapsed(arg) : null;
    }
    groupEnd() {
      return dart.notNull(this[_isConsoleDefined]) ? console.groupEnd() : null;
    }
    info(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.info(arg) : null;
    }
    log(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.log(arg) : null;
    }
    markTimeline(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.markTimeline(arg) : null;
    }
    profile(title) {
      return dart.notNull(this[_isConsoleDefined]) ? console.profile(title) : null;
    }
    profileEnd(title) {
      return dart.notNull(this[_isConsoleDefined]) ? console.profileEnd(title) : null;
    }
    table(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.table(arg) : null;
    }
    time(title) {
      return dart.notNull(this[_isConsoleDefined]) ? console.time(title) : null;
    }
    timeEnd(title) {
      return dart.notNull(this[_isConsoleDefined]) ? console.timeEnd(title) : null;
    }
    timeStamp(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.timeStamp(arg) : null;
    }
    trace(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.trace(arg) : null;
    }
    warn(arg) {
      return dart.notNull(this[_isConsoleDefined]) ? console.warn(arg) : null;
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateConsole() {
      return Console._internalWrap();
    }
    static _internalWrap() {
      return new Console.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(Console, '_safe');
  dart.defineNamedConstructor(Console, 'internal_');
  dart.setSignature(Console, {
    constructors: () => ({
      _safe: [Console, []],
      _: [Console, []],
      _internalWrap: [Console, []],
      internal_: [Console, []]
    }),
    methods: () => ({
      assertCondition: [dart.void, [core.bool, core.Object]],
      clear: [dart.void, [core.Object]],
      count: [dart.void, [core.Object]],
      debug: [dart.void, [core.Object]],
      dir: [dart.void, [core.Object]],
      dirxml: [dart.void, [core.Object]],
      error: [dart.void, [core.Object]],
      group: [dart.void, [core.Object]],
      groupCollapsed: [dart.void, [core.Object]],
      groupEnd: [dart.void, []],
      info: [dart.void, [core.Object]],
      log: [dart.void, [core.Object]],
      markTimeline: [dart.void, [core.Object]],
      profile: [dart.void, [core.String]],
      profileEnd: [dart.void, [core.String]],
      table: [dart.void, [core.Object]],
      time: [dart.void, [core.String]],
      timeEnd: [dart.void, [core.String]],
      timeStamp: [dart.void, [core.Object]],
      trace: [dart.void, [core.Object]],
      warn: [dart.void, [core.Object]]
    }),
    statics: () => ({internalCreateConsole: [Console, []]}),
    names: ['internalCreateConsole']
  });
  Console[dart.metadata] = () => [dart.const(new _metadata.DomName('Console'))];
  Console._safeConsole = dart.const(new Console._safe());
  class CssStyleDeclarationBase extends core.Object {
    getPropertyValue(propertyName) {
      return dart.throw(new core.StateError('getProperty not overridden in dart:html'));
    }
    setProperty(propertyName, value, priority) {
      if (priority === void 0)
        priority = null;
      return dart.throw(new core.StateError('setProperty not overridden in dart:html'));
    }
    get alignContent() {
      return this.getPropertyValue('align-content');
    }
    set alignContent(value) {
      this.setProperty('align-content', value, '');
    }
    get alignItems() {
      return this.getPropertyValue('align-items');
    }
    set alignItems(value) {
      this.setProperty('align-items', value, '');
    }
    get alignSelf() {
      return this.getPropertyValue('align-self');
    }
    set alignSelf(value) {
      this.setProperty('align-self', value, '');
    }
    get animation() {
      return this.getPropertyValue('animation');
    }
    set animation(value) {
      this.setProperty('animation', value, '');
    }
    get animationDelay() {
      return this.getPropertyValue('animation-delay');
    }
    set animationDelay(value) {
      this.setProperty('animation-delay', value, '');
    }
    get animationDirection() {
      return this.getPropertyValue('animation-direction');
    }
    set animationDirection(value) {
      this.setProperty('animation-direction', value, '');
    }
    get animationDuration() {
      return this.getPropertyValue('animation-duration');
    }
    set animationDuration(value) {
      this.setProperty('animation-duration', value, '');
    }
    get animationFillMode() {
      return this.getPropertyValue('animation-fill-mode');
    }
    set animationFillMode(value) {
      this.setProperty('animation-fill-mode', value, '');
    }
    get animationIterationCount() {
      return this.getPropertyValue('animation-iteration-count');
    }
    set animationIterationCount(value) {
      this.setProperty('animation-iteration-count', value, '');
    }
    get animationName() {
      return this.getPropertyValue('animation-name');
    }
    set animationName(value) {
      this.setProperty('animation-name', value, '');
    }
    get animationPlayState() {
      return this.getPropertyValue('animation-play-state');
    }
    set animationPlayState(value) {
      this.setProperty('animation-play-state', value, '');
    }
    get animationTimingFunction() {
      return this.getPropertyValue('animation-timing-function');
    }
    set animationTimingFunction(value) {
      this.setProperty('animation-timing-function', value, '');
    }
    get appRegion() {
      return this.getPropertyValue('app-region');
    }
    set appRegion(value) {
      this.setProperty('app-region', value, '');
    }
    get appearance() {
      return this.getPropertyValue('appearance');
    }
    set appearance(value) {
      this.setProperty('appearance', value, '');
    }
    get aspectRatio() {
      return this.getPropertyValue('aspect-ratio');
    }
    set aspectRatio(value) {
      this.setProperty('aspect-ratio', value, '');
    }
    get backfaceVisibility() {
      return this.getPropertyValue('backface-visibility');
    }
    set backfaceVisibility(value) {
      this.setProperty('backface-visibility', value, '');
    }
    get background() {
      return this.getPropertyValue('background');
    }
    set background(value) {
      this.setProperty('background', value, '');
    }
    get backgroundAttachment() {
      return this.getPropertyValue('background-attachment');
    }
    set backgroundAttachment(value) {
      this.setProperty('background-attachment', value, '');
    }
    get backgroundBlendMode() {
      return this.getPropertyValue('background-blend-mode');
    }
    set backgroundBlendMode(value) {
      this.setProperty('background-blend-mode', value, '');
    }
    get backgroundClip() {
      return this.getPropertyValue('background-clip');
    }
    set backgroundClip(value) {
      this.setProperty('background-clip', value, '');
    }
    get backgroundColor() {
      return this.getPropertyValue('background-color');
    }
    set backgroundColor(value) {
      this.setProperty('background-color', value, '');
    }
    get backgroundComposite() {
      return this.getPropertyValue('background-composite');
    }
    set backgroundComposite(value) {
      this.setProperty('background-composite', value, '');
    }
    get backgroundImage() {
      return this.getPropertyValue('background-image');
    }
    set backgroundImage(value) {
      this.setProperty('background-image', value, '');
    }
    get backgroundOrigin() {
      return this.getPropertyValue('background-origin');
    }
    set backgroundOrigin(value) {
      this.setProperty('background-origin', value, '');
    }
    get backgroundPosition() {
      return this.getPropertyValue('background-position');
    }
    set backgroundPosition(value) {
      this.setProperty('background-position', value, '');
    }
    get backgroundPositionX() {
      return this.getPropertyValue('background-position-x');
    }
    set backgroundPositionX(value) {
      this.setProperty('background-position-x', value, '');
    }
    get backgroundPositionY() {
      return this.getPropertyValue('background-position-y');
    }
    set backgroundPositionY(value) {
      this.setProperty('background-position-y', value, '');
    }
    get backgroundRepeat() {
      return this.getPropertyValue('background-repeat');
    }
    set backgroundRepeat(value) {
      this.setProperty('background-repeat', value, '');
    }
    get backgroundRepeatX() {
      return this.getPropertyValue('background-repeat-x');
    }
    set backgroundRepeatX(value) {
      this.setProperty('background-repeat-x', value, '');
    }
    get backgroundRepeatY() {
      return this.getPropertyValue('background-repeat-y');
    }
    set backgroundRepeatY(value) {
      this.setProperty('background-repeat-y', value, '');
    }
    get backgroundSize() {
      return this.getPropertyValue('background-size');
    }
    set backgroundSize(value) {
      this.setProperty('background-size', value, '');
    }
    get border() {
      return this.getPropertyValue('border');
    }
    set border(value) {
      this.setProperty('border', value, '');
    }
    get borderAfter() {
      return this.getPropertyValue('border-after');
    }
    set borderAfter(value) {
      this.setProperty('border-after', value, '');
    }
    get borderAfterColor() {
      return this.getPropertyValue('border-after-color');
    }
    set borderAfterColor(value) {
      this.setProperty('border-after-color', value, '');
    }
    get borderAfterStyle() {
      return this.getPropertyValue('border-after-style');
    }
    set borderAfterStyle(value) {
      this.setProperty('border-after-style', value, '');
    }
    get borderAfterWidth() {
      return this.getPropertyValue('border-after-width');
    }
    set borderAfterWidth(value) {
      this.setProperty('border-after-width', value, '');
    }
    get borderBefore() {
      return this.getPropertyValue('border-before');
    }
    set borderBefore(value) {
      this.setProperty('border-before', value, '');
    }
    get borderBeforeColor() {
      return this.getPropertyValue('border-before-color');
    }
    set borderBeforeColor(value) {
      this.setProperty('border-before-color', value, '');
    }
    get borderBeforeStyle() {
      return this.getPropertyValue('border-before-style');
    }
    set borderBeforeStyle(value) {
      this.setProperty('border-before-style', value, '');
    }
    get borderBeforeWidth() {
      return this.getPropertyValue('border-before-width');
    }
    set borderBeforeWidth(value) {
      this.setProperty('border-before-width', value, '');
    }
    get borderBottom() {
      return this.getPropertyValue('border-bottom');
    }
    set borderBottom(value) {
      this.setProperty('border-bottom', value, '');
    }
    get borderBottomColor() {
      return this.getPropertyValue('border-bottom-color');
    }
    set borderBottomColor(value) {
      this.setProperty('border-bottom-color', value, '');
    }
    get borderBottomLeftRadius() {
      return this.getPropertyValue('border-bottom-left-radius');
    }
    set borderBottomLeftRadius(value) {
      this.setProperty('border-bottom-left-radius', value, '');
    }
    get borderBottomRightRadius() {
      return this.getPropertyValue('border-bottom-right-radius');
    }
    set borderBottomRightRadius(value) {
      this.setProperty('border-bottom-right-radius', value, '');
    }
    get borderBottomStyle() {
      return this.getPropertyValue('border-bottom-style');
    }
    set borderBottomStyle(value) {
      this.setProperty('border-bottom-style', value, '');
    }
    get borderBottomWidth() {
      return this.getPropertyValue('border-bottom-width');
    }
    set borderBottomWidth(value) {
      this.setProperty('border-bottom-width', value, '');
    }
    get borderCollapse() {
      return this.getPropertyValue('border-collapse');
    }
    set borderCollapse(value) {
      this.setProperty('border-collapse', value, '');
    }
    get borderColor() {
      return this.getPropertyValue('border-color');
    }
    set borderColor(value) {
      this.setProperty('border-color', value, '');
    }
    get borderEnd() {
      return this.getPropertyValue('border-end');
    }
    set borderEnd(value) {
      this.setProperty('border-end', value, '');
    }
    get borderEndColor() {
      return this.getPropertyValue('border-end-color');
    }
    set borderEndColor(value) {
      this.setProperty('border-end-color', value, '');
    }
    get borderEndStyle() {
      return this.getPropertyValue('border-end-style');
    }
    set borderEndStyle(value) {
      this.setProperty('border-end-style', value, '');
    }
    get borderEndWidth() {
      return this.getPropertyValue('border-end-width');
    }
    set borderEndWidth(value) {
      this.setProperty('border-end-width', value, '');
    }
    get borderFit() {
      return this.getPropertyValue('border-fit');
    }
    set borderFit(value) {
      this.setProperty('border-fit', value, '');
    }
    get borderHorizontalSpacing() {
      return this.getPropertyValue('border-horizontal-spacing');
    }
    set borderHorizontalSpacing(value) {
      this.setProperty('border-horizontal-spacing', value, '');
    }
    get borderImage() {
      return this.getPropertyValue('border-image');
    }
    set borderImage(value) {
      this.setProperty('border-image', value, '');
    }
    get borderImageOutset() {
      return this.getPropertyValue('border-image-outset');
    }
    set borderImageOutset(value) {
      this.setProperty('border-image-outset', value, '');
    }
    get borderImageRepeat() {
      return this.getPropertyValue('border-image-repeat');
    }
    set borderImageRepeat(value) {
      this.setProperty('border-image-repeat', value, '');
    }
    get borderImageSlice() {
      return this.getPropertyValue('border-image-slice');
    }
    set borderImageSlice(value) {
      this.setProperty('border-image-slice', value, '');
    }
    get borderImageSource() {
      return this.getPropertyValue('border-image-source');
    }
    set borderImageSource(value) {
      this.setProperty('border-image-source', value, '');
    }
    get borderImageWidth() {
      return this.getPropertyValue('border-image-width');
    }
    set borderImageWidth(value) {
      this.setProperty('border-image-width', value, '');
    }
    get borderLeft() {
      return this.getPropertyValue('border-left');
    }
    set borderLeft(value) {
      this.setProperty('border-left', value, '');
    }
    get borderLeftColor() {
      return this.getPropertyValue('border-left-color');
    }
    set borderLeftColor(value) {
      this.setProperty('border-left-color', value, '');
    }
    get borderLeftStyle() {
      return this.getPropertyValue('border-left-style');
    }
    set borderLeftStyle(value) {
      this.setProperty('border-left-style', value, '');
    }
    get borderLeftWidth() {
      return this.getPropertyValue('border-left-width');
    }
    set borderLeftWidth(value) {
      this.setProperty('border-left-width', value, '');
    }
    get borderRadius() {
      return this.getPropertyValue('border-radius');
    }
    set borderRadius(value) {
      this.setProperty('border-radius', value, '');
    }
    get borderRight() {
      return this.getPropertyValue('border-right');
    }
    set borderRight(value) {
      this.setProperty('border-right', value, '');
    }
    get borderRightColor() {
      return this.getPropertyValue('border-right-color');
    }
    set borderRightColor(value) {
      this.setProperty('border-right-color', value, '');
    }
    get borderRightStyle() {
      return this.getPropertyValue('border-right-style');
    }
    set borderRightStyle(value) {
      this.setProperty('border-right-style', value, '');
    }
    get borderRightWidth() {
      return this.getPropertyValue('border-right-width');
    }
    set borderRightWidth(value) {
      this.setProperty('border-right-width', value, '');
    }
    get borderSpacing() {
      return this.getPropertyValue('border-spacing');
    }
    set borderSpacing(value) {
      this.setProperty('border-spacing', value, '');
    }
    get borderStart() {
      return this.getPropertyValue('border-start');
    }
    set borderStart(value) {
      this.setProperty('border-start', value, '');
    }
    get borderStartColor() {
      return this.getPropertyValue('border-start-color');
    }
    set borderStartColor(value) {
      this.setProperty('border-start-color', value, '');
    }
    get borderStartStyle() {
      return this.getPropertyValue('border-start-style');
    }
    set borderStartStyle(value) {
      this.setProperty('border-start-style', value, '');
    }
    get borderStartWidth() {
      return this.getPropertyValue('border-start-width');
    }
    set borderStartWidth(value) {
      this.setProperty('border-start-width', value, '');
    }
    get borderStyle() {
      return this.getPropertyValue('border-style');
    }
    set borderStyle(value) {
      this.setProperty('border-style', value, '');
    }
    get borderTop() {
      return this.getPropertyValue('border-top');
    }
    set borderTop(value) {
      this.setProperty('border-top', value, '');
    }
    get borderTopColor() {
      return this.getPropertyValue('border-top-color');
    }
    set borderTopColor(value) {
      this.setProperty('border-top-color', value, '');
    }
    get borderTopLeftRadius() {
      return this.getPropertyValue('border-top-left-radius');
    }
    set borderTopLeftRadius(value) {
      this.setProperty('border-top-left-radius', value, '');
    }
    get borderTopRightRadius() {
      return this.getPropertyValue('border-top-right-radius');
    }
    set borderTopRightRadius(value) {
      this.setProperty('border-top-right-radius', value, '');
    }
    get borderTopStyle() {
      return this.getPropertyValue('border-top-style');
    }
    set borderTopStyle(value) {
      this.setProperty('border-top-style', value, '');
    }
    get borderTopWidth() {
      return this.getPropertyValue('border-top-width');
    }
    set borderTopWidth(value) {
      this.setProperty('border-top-width', value, '');
    }
    get borderVerticalSpacing() {
      return this.getPropertyValue('border-vertical-spacing');
    }
    set borderVerticalSpacing(value) {
      this.setProperty('border-vertical-spacing', value, '');
    }
    get borderWidth() {
      return this.getPropertyValue('border-width');
    }
    set borderWidth(value) {
      this.setProperty('border-width', value, '');
    }
    get bottom() {
      return this.getPropertyValue('bottom');
    }
    set bottom(value) {
      this.setProperty('bottom', value, '');
    }
    get boxAlign() {
      return this.getPropertyValue('box-align');
    }
    set boxAlign(value) {
      this.setProperty('box-align', value, '');
    }
    get boxDecorationBreak() {
      return this.getPropertyValue('box-decoration-break');
    }
    set boxDecorationBreak(value) {
      this.setProperty('box-decoration-break', value, '');
    }
    get boxDirection() {
      return this.getPropertyValue('box-direction');
    }
    set boxDirection(value) {
      this.setProperty('box-direction', value, '');
    }
    get boxFlex() {
      return this.getPropertyValue('box-flex');
    }
    set boxFlex(value) {
      this.setProperty('box-flex', value, '');
    }
    get boxFlexGroup() {
      return this.getPropertyValue('box-flex-group');
    }
    set boxFlexGroup(value) {
      this.setProperty('box-flex-group', value, '');
    }
    get boxLines() {
      return this.getPropertyValue('box-lines');
    }
    set boxLines(value) {
      this.setProperty('box-lines', value, '');
    }
    get boxOrdinalGroup() {
      return this.getPropertyValue('box-ordinal-group');
    }
    set boxOrdinalGroup(value) {
      this.setProperty('box-ordinal-group', value, '');
    }
    get boxOrient() {
      return this.getPropertyValue('box-orient');
    }
    set boxOrient(value) {
      this.setProperty('box-orient', value, '');
    }
    get boxPack() {
      return this.getPropertyValue('box-pack');
    }
    set boxPack(value) {
      this.setProperty('box-pack', value, '');
    }
    get boxReflect() {
      return this.getPropertyValue('box-reflect');
    }
    set boxReflect(value) {
      this.setProperty('box-reflect', value, '');
    }
    get boxShadow() {
      return this.getPropertyValue('box-shadow');
    }
    set boxShadow(value) {
      this.setProperty('box-shadow', value, '');
    }
    get boxSizing() {
      return this.getPropertyValue('box-sizing');
    }
    set boxSizing(value) {
      this.setProperty('box-sizing', value, '');
    }
    get captionSide() {
      return this.getPropertyValue('caption-side');
    }
    set captionSide(value) {
      this.setProperty('caption-side', value, '');
    }
    get clear() {
      return this.getPropertyValue('clear');
    }
    set clear(value) {
      this.setProperty('clear', value, '');
    }
    get clip() {
      return this.getPropertyValue('clip');
    }
    set clip(value) {
      this.setProperty('clip', value, '');
    }
    get clipPath() {
      return this.getPropertyValue('clip-path');
    }
    set clipPath(value) {
      this.setProperty('clip-path', value, '');
    }
    get color() {
      return this.getPropertyValue('color');
    }
    set color(value) {
      this.setProperty('color', value, '');
    }
    get columnBreakAfter() {
      return this.getPropertyValue('column-break-after');
    }
    set columnBreakAfter(value) {
      this.setProperty('column-break-after', value, '');
    }
    get columnBreakBefore() {
      return this.getPropertyValue('column-break-before');
    }
    set columnBreakBefore(value) {
      this.setProperty('column-break-before', value, '');
    }
    get columnBreakInside() {
      return this.getPropertyValue('column-break-inside');
    }
    set columnBreakInside(value) {
      this.setProperty('column-break-inside', value, '');
    }
    get columnCount() {
      return this.getPropertyValue('column-count');
    }
    set columnCount(value) {
      this.setProperty('column-count', value, '');
    }
    get columnFill() {
      return this.getPropertyValue('column-fill');
    }
    set columnFill(value) {
      this.setProperty('column-fill', value, '');
    }
    get columnGap() {
      return this.getPropertyValue('column-gap');
    }
    set columnGap(value) {
      this.setProperty('column-gap', value, '');
    }
    get columnRule() {
      return this.getPropertyValue('column-rule');
    }
    set columnRule(value) {
      this.setProperty('column-rule', value, '');
    }
    get columnRuleColor() {
      return this.getPropertyValue('column-rule-color');
    }
    set columnRuleColor(value) {
      this.setProperty('column-rule-color', value, '');
    }
    get columnRuleStyle() {
      return this.getPropertyValue('column-rule-style');
    }
    set columnRuleStyle(value) {
      this.setProperty('column-rule-style', value, '');
    }
    get columnRuleWidth() {
      return this.getPropertyValue('column-rule-width');
    }
    set columnRuleWidth(value) {
      this.setProperty('column-rule-width', value, '');
    }
    get columnSpan() {
      return this.getPropertyValue('column-span');
    }
    set columnSpan(value) {
      this.setProperty('column-span', value, '');
    }
    get columnWidth() {
      return this.getPropertyValue('column-width');
    }
    set columnWidth(value) {
      this.setProperty('column-width', value, '');
    }
    get columns() {
      return this.getPropertyValue('columns');
    }
    set columns(value) {
      this.setProperty('columns', value, '');
    }
    get content() {
      return this.getPropertyValue('content');
    }
    set content(value) {
      this.setProperty('content', value, '');
    }
    get counterIncrement() {
      return this.getPropertyValue('counter-increment');
    }
    set counterIncrement(value) {
      this.setProperty('counter-increment', value, '');
    }
    get counterReset() {
      return this.getPropertyValue('counter-reset');
    }
    set counterReset(value) {
      this.setProperty('counter-reset', value, '');
    }
    get cursor() {
      return this.getPropertyValue('cursor');
    }
    set cursor(value) {
      this.setProperty('cursor', value, '');
    }
    get direction() {
      return this.getPropertyValue('direction');
    }
    set direction(value) {
      this.setProperty('direction', value, '');
    }
    get display() {
      return this.getPropertyValue('display');
    }
    set display(value) {
      this.setProperty('display', value, '');
    }
    get emptyCells() {
      return this.getPropertyValue('empty-cells');
    }
    set emptyCells(value) {
      this.setProperty('empty-cells', value, '');
    }
    get filter() {
      return this.getPropertyValue('filter');
    }
    set filter(value) {
      this.setProperty('filter', value, '');
    }
    get flex() {
      return this.getPropertyValue('flex');
    }
    set flex(value) {
      this.setProperty('flex', value, '');
    }
    get flexBasis() {
      return this.getPropertyValue('flex-basis');
    }
    set flexBasis(value) {
      this.setProperty('flex-basis', value, '');
    }
    get flexDirection() {
      return this.getPropertyValue('flex-direction');
    }
    set flexDirection(value) {
      this.setProperty('flex-direction', value, '');
    }
    get flexFlow() {
      return this.getPropertyValue('flex-flow');
    }
    set flexFlow(value) {
      this.setProperty('flex-flow', value, '');
    }
    get flexGrow() {
      return this.getPropertyValue('flex-grow');
    }
    set flexGrow(value) {
      this.setProperty('flex-grow', value, '');
    }
    get flexShrink() {
      return this.getPropertyValue('flex-shrink');
    }
    set flexShrink(value) {
      this.setProperty('flex-shrink', value, '');
    }
    get flexWrap() {
      return this.getPropertyValue('flex-wrap');
    }
    set flexWrap(value) {
      this.setProperty('flex-wrap', value, '');
    }
    get float() {
      return this.getPropertyValue('float');
    }
    set float(value) {
      this.setProperty('float', value, '');
    }
    get font() {
      return this.getPropertyValue('font');
    }
    set font(value) {
      this.setProperty('font', value, '');
    }
    get fontFamily() {
      return this.getPropertyValue('font-family');
    }
    set fontFamily(value) {
      this.setProperty('font-family', value, '');
    }
    get fontFeatureSettings() {
      return this.getPropertyValue('font-feature-settings');
    }
    set fontFeatureSettings(value) {
      this.setProperty('font-feature-settings', value, '');
    }
    get fontKerning() {
      return this.getPropertyValue('font-kerning');
    }
    set fontKerning(value) {
      this.setProperty('font-kerning', value, '');
    }
    get fontSize() {
      return this.getPropertyValue('font-size');
    }
    set fontSize(value) {
      this.setProperty('font-size', value, '');
    }
    get fontSizeDelta() {
      return this.getPropertyValue('font-size-delta');
    }
    set fontSizeDelta(value) {
      this.setProperty('font-size-delta', value, '');
    }
    get fontSmoothing() {
      return this.getPropertyValue('font-smoothing');
    }
    set fontSmoothing(value) {
      this.setProperty('font-smoothing', value, '');
    }
    get fontStretch() {
      return this.getPropertyValue('font-stretch');
    }
    set fontStretch(value) {
      this.setProperty('font-stretch', value, '');
    }
    get fontStyle() {
      return this.getPropertyValue('font-style');
    }
    set fontStyle(value) {
      this.setProperty('font-style', value, '');
    }
    get fontVariant() {
      return this.getPropertyValue('font-variant');
    }
    set fontVariant(value) {
      this.setProperty('font-variant', value, '');
    }
    get fontVariantLigatures() {
      return this.getPropertyValue('font-variant-ligatures');
    }
    set fontVariantLigatures(value) {
      this.setProperty('font-variant-ligatures', value, '');
    }
    get fontWeight() {
      return this.getPropertyValue('font-weight');
    }
    set fontWeight(value) {
      this.setProperty('font-weight', value, '');
    }
    get grid() {
      return this.getPropertyValue('grid');
    }
    set grid(value) {
      this.setProperty('grid', value, '');
    }
    get gridArea() {
      return this.getPropertyValue('grid-area');
    }
    set gridArea(value) {
      this.setProperty('grid-area', value, '');
    }
    get gridAutoColumns() {
      return this.getPropertyValue('grid-auto-columns');
    }
    set gridAutoColumns(value) {
      this.setProperty('grid-auto-columns', value, '');
    }
    get gridAutoFlow() {
      return this.getPropertyValue('grid-auto-flow');
    }
    set gridAutoFlow(value) {
      this.setProperty('grid-auto-flow', value, '');
    }
    get gridAutoRows() {
      return this.getPropertyValue('grid-auto-rows');
    }
    set gridAutoRows(value) {
      this.setProperty('grid-auto-rows', value, '');
    }
    get gridColumn() {
      return this.getPropertyValue('grid-column');
    }
    set gridColumn(value) {
      this.setProperty('grid-column', value, '');
    }
    get gridColumnEnd() {
      return this.getPropertyValue('grid-column-end');
    }
    set gridColumnEnd(value) {
      this.setProperty('grid-column-end', value, '');
    }
    get gridColumnStart() {
      return this.getPropertyValue('grid-column-start');
    }
    set gridColumnStart(value) {
      this.setProperty('grid-column-start', value, '');
    }
    get gridRow() {
      return this.getPropertyValue('grid-row');
    }
    set gridRow(value) {
      this.setProperty('grid-row', value, '');
    }
    get gridRowEnd() {
      return this.getPropertyValue('grid-row-end');
    }
    set gridRowEnd(value) {
      this.setProperty('grid-row-end', value, '');
    }
    get gridRowStart() {
      return this.getPropertyValue('grid-row-start');
    }
    set gridRowStart(value) {
      this.setProperty('grid-row-start', value, '');
    }
    get gridTemplate() {
      return this.getPropertyValue('grid-template');
    }
    set gridTemplate(value) {
      this.setProperty('grid-template', value, '');
    }
    get gridTemplateAreas() {
      return this.getPropertyValue('grid-template-areas');
    }
    set gridTemplateAreas(value) {
      this.setProperty('grid-template-areas', value, '');
    }
    get gridTemplateColumns() {
      return this.getPropertyValue('grid-template-columns');
    }
    set gridTemplateColumns(value) {
      this.setProperty('grid-template-columns', value, '');
    }
    get gridTemplateRows() {
      return this.getPropertyValue('grid-template-rows');
    }
    set gridTemplateRows(value) {
      this.setProperty('grid-template-rows', value, '');
    }
    get height() {
      return this.getPropertyValue('height');
    }
    set height(value) {
      this.setProperty('height', value, '');
    }
    get highlight() {
      return this.getPropertyValue('highlight');
    }
    set highlight(value) {
      this.setProperty('highlight', value, '');
    }
    get hyphenateCharacter() {
      return this.getPropertyValue('hyphenate-character');
    }
    set hyphenateCharacter(value) {
      this.setProperty('hyphenate-character', value, '');
    }
    get imageRendering() {
      return this.getPropertyValue('image-rendering');
    }
    set imageRendering(value) {
      this.setProperty('image-rendering', value, '');
    }
    get isolation() {
      return this.getPropertyValue('isolation');
    }
    set isolation(value) {
      this.setProperty('isolation', value, '');
    }
    get justifyContent() {
      return this.getPropertyValue('justify-content');
    }
    set justifyContent(value) {
      this.setProperty('justify-content', value, '');
    }
    get justifySelf() {
      return this.getPropertyValue('justify-self');
    }
    set justifySelf(value) {
      this.setProperty('justify-self', value, '');
    }
    get left() {
      return this.getPropertyValue('left');
    }
    set left(value) {
      this.setProperty('left', value, '');
    }
    get letterSpacing() {
      return this.getPropertyValue('letter-spacing');
    }
    set letterSpacing(value) {
      this.setProperty('letter-spacing', value, '');
    }
    get lineBoxContain() {
      return this.getPropertyValue('line-box-contain');
    }
    set lineBoxContain(value) {
      this.setProperty('line-box-contain', value, '');
    }
    get lineBreak() {
      return this.getPropertyValue('line-break');
    }
    set lineBreak(value) {
      this.setProperty('line-break', value, '');
    }
    get lineClamp() {
      return this.getPropertyValue('line-clamp');
    }
    set lineClamp(value) {
      this.setProperty('line-clamp', value, '');
    }
    get lineHeight() {
      return this.getPropertyValue('line-height');
    }
    set lineHeight(value) {
      this.setProperty('line-height', value, '');
    }
    get listStyle() {
      return this.getPropertyValue('list-style');
    }
    set listStyle(value) {
      this.setProperty('list-style', value, '');
    }
    get listStyleImage() {
      return this.getPropertyValue('list-style-image');
    }
    set listStyleImage(value) {
      this.setProperty('list-style-image', value, '');
    }
    get listStylePosition() {
      return this.getPropertyValue('list-style-position');
    }
    set listStylePosition(value) {
      this.setProperty('list-style-position', value, '');
    }
    get listStyleType() {
      return this.getPropertyValue('list-style-type');
    }
    set listStyleType(value) {
      this.setProperty('list-style-type', value, '');
    }
    get locale() {
      return this.getPropertyValue('locale');
    }
    set locale(value) {
      this.setProperty('locale', value, '');
    }
    get logicalHeight() {
      return this.getPropertyValue('logical-height');
    }
    set logicalHeight(value) {
      this.setProperty('logical-height', value, '');
    }
    get logicalWidth() {
      return this.getPropertyValue('logical-width');
    }
    set logicalWidth(value) {
      this.setProperty('logical-width', value, '');
    }
    get margin() {
      return this.getPropertyValue('margin');
    }
    set margin(value) {
      this.setProperty('margin', value, '');
    }
    get marginAfter() {
      return this.getPropertyValue('margin-after');
    }
    set marginAfter(value) {
      this.setProperty('margin-after', value, '');
    }
    get marginAfterCollapse() {
      return this.getPropertyValue('margin-after-collapse');
    }
    set marginAfterCollapse(value) {
      this.setProperty('margin-after-collapse', value, '');
    }
    get marginBefore() {
      return this.getPropertyValue('margin-before');
    }
    set marginBefore(value) {
      this.setProperty('margin-before', value, '');
    }
    get marginBeforeCollapse() {
      return this.getPropertyValue('margin-before-collapse');
    }
    set marginBeforeCollapse(value) {
      this.setProperty('margin-before-collapse', value, '');
    }
    get marginBottom() {
      return this.getPropertyValue('margin-bottom');
    }
    set marginBottom(value) {
      this.setProperty('margin-bottom', value, '');
    }
    get marginBottomCollapse() {
      return this.getPropertyValue('margin-bottom-collapse');
    }
    set marginBottomCollapse(value) {
      this.setProperty('margin-bottom-collapse', value, '');
    }
    get marginCollapse() {
      return this.getPropertyValue('margin-collapse');
    }
    set marginCollapse(value) {
      this.setProperty('margin-collapse', value, '');
    }
    get marginEnd() {
      return this.getPropertyValue('margin-end');
    }
    set marginEnd(value) {
      this.setProperty('margin-end', value, '');
    }
    get marginLeft() {
      return this.getPropertyValue('margin-left');
    }
    set marginLeft(value) {
      this.setProperty('margin-left', value, '');
    }
    get marginRight() {
      return this.getPropertyValue('margin-right');
    }
    set marginRight(value) {
      this.setProperty('margin-right', value, '');
    }
    get marginStart() {
      return this.getPropertyValue('margin-start');
    }
    set marginStart(value) {
      this.setProperty('margin-start', value, '');
    }
    get marginTop() {
      return this.getPropertyValue('margin-top');
    }
    set marginTop(value) {
      this.setProperty('margin-top', value, '');
    }
    get marginTopCollapse() {
      return this.getPropertyValue('margin-top-collapse');
    }
    set marginTopCollapse(value) {
      this.setProperty('margin-top-collapse', value, '');
    }
    get mask() {
      return this.getPropertyValue('mask');
    }
    set mask(value) {
      this.setProperty('mask', value, '');
    }
    get maskBoxImage() {
      return this.getPropertyValue('mask-box-image');
    }
    set maskBoxImage(value) {
      this.setProperty('mask-box-image', value, '');
    }
    get maskBoxImageOutset() {
      return this.getPropertyValue('mask-box-image-outset');
    }
    set maskBoxImageOutset(value) {
      this.setProperty('mask-box-image-outset', value, '');
    }
    get maskBoxImageRepeat() {
      return this.getPropertyValue('mask-box-image-repeat');
    }
    set maskBoxImageRepeat(value) {
      this.setProperty('mask-box-image-repeat', value, '');
    }
    get maskBoxImageSlice() {
      return this.getPropertyValue('mask-box-image-slice');
    }
    set maskBoxImageSlice(value) {
      this.setProperty('mask-box-image-slice', value, '');
    }
    get maskBoxImageSource() {
      return this.getPropertyValue('mask-box-image-source');
    }
    set maskBoxImageSource(value) {
      this.setProperty('mask-box-image-source', value, '');
    }
    get maskBoxImageWidth() {
      return this.getPropertyValue('mask-box-image-width');
    }
    set maskBoxImageWidth(value) {
      this.setProperty('mask-box-image-width', value, '');
    }
    get maskClip() {
      return this.getPropertyValue('mask-clip');
    }
    set maskClip(value) {
      this.setProperty('mask-clip', value, '');
    }
    get maskComposite() {
      return this.getPropertyValue('mask-composite');
    }
    set maskComposite(value) {
      this.setProperty('mask-composite', value, '');
    }
    get maskImage() {
      return this.getPropertyValue('mask-image');
    }
    set maskImage(value) {
      this.setProperty('mask-image', value, '');
    }
    get maskOrigin() {
      return this.getPropertyValue('mask-origin');
    }
    set maskOrigin(value) {
      this.setProperty('mask-origin', value, '');
    }
    get maskPosition() {
      return this.getPropertyValue('mask-position');
    }
    set maskPosition(value) {
      this.setProperty('mask-position', value, '');
    }
    get maskPositionX() {
      return this.getPropertyValue('mask-position-x');
    }
    set maskPositionX(value) {
      this.setProperty('mask-position-x', value, '');
    }
    get maskPositionY() {
      return this.getPropertyValue('mask-position-y');
    }
    set maskPositionY(value) {
      this.setProperty('mask-position-y', value, '');
    }
    get maskRepeat() {
      return this.getPropertyValue('mask-repeat');
    }
    set maskRepeat(value) {
      this.setProperty('mask-repeat', value, '');
    }
    get maskRepeatX() {
      return this.getPropertyValue('mask-repeat-x');
    }
    set maskRepeatX(value) {
      this.setProperty('mask-repeat-x', value, '');
    }
    get maskRepeatY() {
      return this.getPropertyValue('mask-repeat-y');
    }
    set maskRepeatY(value) {
      this.setProperty('mask-repeat-y', value, '');
    }
    get maskSize() {
      return this.getPropertyValue('mask-size');
    }
    set maskSize(value) {
      this.setProperty('mask-size', value, '');
    }
    get maskSourceType() {
      return this.getPropertyValue('mask-source-type');
    }
    set maskSourceType(value) {
      this.setProperty('mask-source-type', value, '');
    }
    get maxHeight() {
      return this.getPropertyValue('max-height');
    }
    set maxHeight(value) {
      this.setProperty('max-height', value, '');
    }
    get maxLogicalHeight() {
      return this.getPropertyValue('max-logical-height');
    }
    set maxLogicalHeight(value) {
      this.setProperty('max-logical-height', value, '');
    }
    get maxLogicalWidth() {
      return this.getPropertyValue('max-logical-width');
    }
    set maxLogicalWidth(value) {
      this.setProperty('max-logical-width', value, '');
    }
    get maxWidth() {
      return this.getPropertyValue('max-width');
    }
    set maxWidth(value) {
      this.setProperty('max-width', value, '');
    }
    get maxZoom() {
      return this.getPropertyValue('max-zoom');
    }
    set maxZoom(value) {
      this.setProperty('max-zoom', value, '');
    }
    get minHeight() {
      return this.getPropertyValue('min-height');
    }
    set minHeight(value) {
      this.setProperty('min-height', value, '');
    }
    get minLogicalHeight() {
      return this.getPropertyValue('min-logical-height');
    }
    set minLogicalHeight(value) {
      this.setProperty('min-logical-height', value, '');
    }
    get minLogicalWidth() {
      return this.getPropertyValue('min-logical-width');
    }
    set minLogicalWidth(value) {
      this.setProperty('min-logical-width', value, '');
    }
    get minWidth() {
      return this.getPropertyValue('min-width');
    }
    set minWidth(value) {
      this.setProperty('min-width', value, '');
    }
    get minZoom() {
      return this.getPropertyValue('min-zoom');
    }
    set minZoom(value) {
      this.setProperty('min-zoom', value, '');
    }
    get mixBlendMode() {
      return this.getPropertyValue('mix-blend-mode');
    }
    set mixBlendMode(value) {
      this.setProperty('mix-blend-mode', value, '');
    }
    get objectFit() {
      return this.getPropertyValue('object-fit');
    }
    set objectFit(value) {
      this.setProperty('object-fit', value, '');
    }
    get objectPosition() {
      return this.getPropertyValue('object-position');
    }
    set objectPosition(value) {
      this.setProperty('object-position', value, '');
    }
    get opacity() {
      return this.getPropertyValue('opacity');
    }
    set opacity(value) {
      this.setProperty('opacity', value, '');
    }
    get order() {
      return this.getPropertyValue('order');
    }
    set order(value) {
      this.setProperty('order', value, '');
    }
    get orientation() {
      return this.getPropertyValue('orientation');
    }
    set orientation(value) {
      this.setProperty('orientation', value, '');
    }
    get orphans() {
      return this.getPropertyValue('orphans');
    }
    set orphans(value) {
      this.setProperty('orphans', value, '');
    }
    get outline() {
      return this.getPropertyValue('outline');
    }
    set outline(value) {
      this.setProperty('outline', value, '');
    }
    get outlineColor() {
      return this.getPropertyValue('outline-color');
    }
    set outlineColor(value) {
      this.setProperty('outline-color', value, '');
    }
    get outlineOffset() {
      return this.getPropertyValue('outline-offset');
    }
    set outlineOffset(value) {
      this.setProperty('outline-offset', value, '');
    }
    get outlineStyle() {
      return this.getPropertyValue('outline-style');
    }
    set outlineStyle(value) {
      this.setProperty('outline-style', value, '');
    }
    get outlineWidth() {
      return this.getPropertyValue('outline-width');
    }
    set outlineWidth(value) {
      this.setProperty('outline-width', value, '');
    }
    get overflow() {
      return this.getPropertyValue('overflow');
    }
    set overflow(value) {
      this.setProperty('overflow', value, '');
    }
    get overflowWrap() {
      return this.getPropertyValue('overflow-wrap');
    }
    set overflowWrap(value) {
      this.setProperty('overflow-wrap', value, '');
    }
    get overflowX() {
      return this.getPropertyValue('overflow-x');
    }
    set overflowX(value) {
      this.setProperty('overflow-x', value, '');
    }
    get overflowY() {
      return this.getPropertyValue('overflow-y');
    }
    set overflowY(value) {
      this.setProperty('overflow-y', value, '');
    }
    get padding() {
      return this.getPropertyValue('padding');
    }
    set padding(value) {
      this.setProperty('padding', value, '');
    }
    get paddingAfter() {
      return this.getPropertyValue('padding-after');
    }
    set paddingAfter(value) {
      this.setProperty('padding-after', value, '');
    }
    get paddingBefore() {
      return this.getPropertyValue('padding-before');
    }
    set paddingBefore(value) {
      this.setProperty('padding-before', value, '');
    }
    get paddingBottom() {
      return this.getPropertyValue('padding-bottom');
    }
    set paddingBottom(value) {
      this.setProperty('padding-bottom', value, '');
    }
    get paddingEnd() {
      return this.getPropertyValue('padding-end');
    }
    set paddingEnd(value) {
      this.setProperty('padding-end', value, '');
    }
    get paddingLeft() {
      return this.getPropertyValue('padding-left');
    }
    set paddingLeft(value) {
      this.setProperty('padding-left', value, '');
    }
    get paddingRight() {
      return this.getPropertyValue('padding-right');
    }
    set paddingRight(value) {
      this.setProperty('padding-right', value, '');
    }
    get paddingStart() {
      return this.getPropertyValue('padding-start');
    }
    set paddingStart(value) {
      this.setProperty('padding-start', value, '');
    }
    get paddingTop() {
      return this.getPropertyValue('padding-top');
    }
    set paddingTop(value) {
      this.setProperty('padding-top', value, '');
    }
    get page() {
      return this.getPropertyValue('page');
    }
    set page(value) {
      this.setProperty('page', value, '');
    }
    get pageBreakAfter() {
      return this.getPropertyValue('page-break-after');
    }
    set pageBreakAfter(value) {
      this.setProperty('page-break-after', value, '');
    }
    get pageBreakBefore() {
      return this.getPropertyValue('page-break-before');
    }
    set pageBreakBefore(value) {
      this.setProperty('page-break-before', value, '');
    }
    get pageBreakInside() {
      return this.getPropertyValue('page-break-inside');
    }
    set pageBreakInside(value) {
      this.setProperty('page-break-inside', value, '');
    }
    get perspective() {
      return this.getPropertyValue('perspective');
    }
    set perspective(value) {
      this.setProperty('perspective', value, '');
    }
    get perspectiveOrigin() {
      return this.getPropertyValue('perspective-origin');
    }
    set perspectiveOrigin(value) {
      this.setProperty('perspective-origin', value, '');
    }
    get perspectiveOriginX() {
      return this.getPropertyValue('perspective-origin-x');
    }
    set perspectiveOriginX(value) {
      this.setProperty('perspective-origin-x', value, '');
    }
    get perspectiveOriginY() {
      return this.getPropertyValue('perspective-origin-y');
    }
    set perspectiveOriginY(value) {
      this.setProperty('perspective-origin-y', value, '');
    }
    get pointerEvents() {
      return this.getPropertyValue('pointer-events');
    }
    set pointerEvents(value) {
      this.setProperty('pointer-events', value, '');
    }
    get position() {
      return this.getPropertyValue('position');
    }
    set position(value) {
      this.setProperty('position', value, '');
    }
    get printColorAdjust() {
      return this.getPropertyValue('print-color-adjust');
    }
    set printColorAdjust(value) {
      this.setProperty('print-color-adjust', value, '');
    }
    get quotes() {
      return this.getPropertyValue('quotes');
    }
    set quotes(value) {
      this.setProperty('quotes', value, '');
    }
    get resize() {
      return this.getPropertyValue('resize');
    }
    set resize(value) {
      this.setProperty('resize', value, '');
    }
    get right() {
      return this.getPropertyValue('right');
    }
    set right(value) {
      this.setProperty('right', value, '');
    }
    get rtlOrdering() {
      return this.getPropertyValue('rtl-ordering');
    }
    set rtlOrdering(value) {
      this.setProperty('rtl-ordering', value, '');
    }
    get rubyPosition() {
      return this.getPropertyValue('ruby-position');
    }
    set rubyPosition(value) {
      this.setProperty('ruby-position', value, '');
    }
    get scrollBehavior() {
      return this.getPropertyValue('scroll-behavior');
    }
    set scrollBehavior(value) {
      this.setProperty('scroll-behavior', value, '');
    }
    get shapeImageThreshold() {
      return this.getPropertyValue('shape-image-threshold');
    }
    set shapeImageThreshold(value) {
      this.setProperty('shape-image-threshold', value, '');
    }
    get shapeMargin() {
      return this.getPropertyValue('shape-margin');
    }
    set shapeMargin(value) {
      this.setProperty('shape-margin', value, '');
    }
    get shapeOutside() {
      return this.getPropertyValue('shape-outside');
    }
    set shapeOutside(value) {
      this.setProperty('shape-outside', value, '');
    }
    get size() {
      return this.getPropertyValue('size');
    }
    set size(value) {
      this.setProperty('size', value, '');
    }
    get speak() {
      return this.getPropertyValue('speak');
    }
    set speak(value) {
      this.setProperty('speak', value, '');
    }
    get src() {
      return this.getPropertyValue('src');
    }
    set src(value) {
      this.setProperty('src', value, '');
    }
    get tabSize() {
      return this.getPropertyValue('tab-size');
    }
    set tabSize(value) {
      this.setProperty('tab-size', value, '');
    }
    get tableLayout() {
      return this.getPropertyValue('table-layout');
    }
    set tableLayout(value) {
      this.setProperty('table-layout', value, '');
    }
    get tapHighlightColor() {
      return this.getPropertyValue('tap-highlight-color');
    }
    set tapHighlightColor(value) {
      this.setProperty('tap-highlight-color', value, '');
    }
    get textAlign() {
      return this.getPropertyValue('text-align');
    }
    set textAlign(value) {
      this.setProperty('text-align', value, '');
    }
    get textAlignLast() {
      return this.getPropertyValue('text-align-last');
    }
    set textAlignLast(value) {
      this.setProperty('text-align-last', value, '');
    }
    get textCombine() {
      return this.getPropertyValue('text-combine');
    }
    set textCombine(value) {
      this.setProperty('text-combine', value, '');
    }
    get textDecoration() {
      return this.getPropertyValue('text-decoration');
    }
    set textDecoration(value) {
      this.setProperty('text-decoration', value, '');
    }
    get textDecorationColor() {
      return this.getPropertyValue('text-decoration-color');
    }
    set textDecorationColor(value) {
      this.setProperty('text-decoration-color', value, '');
    }
    get textDecorationLine() {
      return this.getPropertyValue('text-decoration-line');
    }
    set textDecorationLine(value) {
      this.setProperty('text-decoration-line', value, '');
    }
    get textDecorationStyle() {
      return this.getPropertyValue('text-decoration-style');
    }
    set textDecorationStyle(value) {
      this.setProperty('text-decoration-style', value, '');
    }
    get textDecorationsInEffect() {
      return this.getPropertyValue('text-decorations-in-effect');
    }
    set textDecorationsInEffect(value) {
      this.setProperty('text-decorations-in-effect', value, '');
    }
    get textEmphasis() {
      return this.getPropertyValue('text-emphasis');
    }
    set textEmphasis(value) {
      this.setProperty('text-emphasis', value, '');
    }
    get textEmphasisColor() {
      return this.getPropertyValue('text-emphasis-color');
    }
    set textEmphasisColor(value) {
      this.setProperty('text-emphasis-color', value, '');
    }
    get textEmphasisPosition() {
      return this.getPropertyValue('text-emphasis-position');
    }
    set textEmphasisPosition(value) {
      this.setProperty('text-emphasis-position', value, '');
    }
    get textEmphasisStyle() {
      return this.getPropertyValue('text-emphasis-style');
    }
    set textEmphasisStyle(value) {
      this.setProperty('text-emphasis-style', value, '');
    }
    get textFillColor() {
      return this.getPropertyValue('text-fill-color');
    }
    set textFillColor(value) {
      this.setProperty('text-fill-color', value, '');
    }
    get textIndent() {
      return this.getPropertyValue('text-indent');
    }
    set textIndent(value) {
      this.setProperty('text-indent', value, '');
    }
    get textJustify() {
      return this.getPropertyValue('text-justify');
    }
    set textJustify(value) {
      this.setProperty('text-justify', value, '');
    }
    get textLineThroughColor() {
      return this.getPropertyValue('text-line-through-color');
    }
    set textLineThroughColor(value) {
      this.setProperty('text-line-through-color', value, '');
    }
    get textLineThroughMode() {
      return this.getPropertyValue('text-line-through-mode');
    }
    set textLineThroughMode(value) {
      this.setProperty('text-line-through-mode', value, '');
    }
    get textLineThroughStyle() {
      return this.getPropertyValue('text-line-through-style');
    }
    set textLineThroughStyle(value) {
      this.setProperty('text-line-through-style', value, '');
    }
    get textLineThroughWidth() {
      return this.getPropertyValue('text-line-through-width');
    }
    set textLineThroughWidth(value) {
      this.setProperty('text-line-through-width', value, '');
    }
    get textOrientation() {
      return this.getPropertyValue('text-orientation');
    }
    set textOrientation(value) {
      this.setProperty('text-orientation', value, '');
    }
    get textOverflow() {
      return this.getPropertyValue('text-overflow');
    }
    set textOverflow(value) {
      this.setProperty('text-overflow', value, '');
    }
    get textOverlineColor() {
      return this.getPropertyValue('text-overline-color');
    }
    set textOverlineColor(value) {
      this.setProperty('text-overline-color', value, '');
    }
    get textOverlineMode() {
      return this.getPropertyValue('text-overline-mode');
    }
    set textOverlineMode(value) {
      this.setProperty('text-overline-mode', value, '');
    }
    get textOverlineStyle() {
      return this.getPropertyValue('text-overline-style');
    }
    set textOverlineStyle(value) {
      this.setProperty('text-overline-style', value, '');
    }
    get textOverlineWidth() {
      return this.getPropertyValue('text-overline-width');
    }
    set textOverlineWidth(value) {
      this.setProperty('text-overline-width', value, '');
    }
    get textRendering() {
      return this.getPropertyValue('text-rendering');
    }
    set textRendering(value) {
      this.setProperty('text-rendering', value, '');
    }
    get textSecurity() {
      return this.getPropertyValue('text-security');
    }
    set textSecurity(value) {
      this.setProperty('text-security', value, '');
    }
    get textShadow() {
      return this.getPropertyValue('text-shadow');
    }
    set textShadow(value) {
      this.setProperty('text-shadow', value, '');
    }
    get textStroke() {
      return this.getPropertyValue('text-stroke');
    }
    set textStroke(value) {
      this.setProperty('text-stroke', value, '');
    }
    get textStrokeColor() {
      return this.getPropertyValue('text-stroke-color');
    }
    set textStrokeColor(value) {
      this.setProperty('text-stroke-color', value, '');
    }
    get textStrokeWidth() {
      return this.getPropertyValue('text-stroke-width');
    }
    set textStrokeWidth(value) {
      this.setProperty('text-stroke-width', value, '');
    }
    get textTransform() {
      return this.getPropertyValue('text-transform');
    }
    set textTransform(value) {
      this.setProperty('text-transform', value, '');
    }
    get textUnderlineColor() {
      return this.getPropertyValue('text-underline-color');
    }
    set textUnderlineColor(value) {
      this.setProperty('text-underline-color', value, '');
    }
    get textUnderlineMode() {
      return this.getPropertyValue('text-underline-mode');
    }
    set textUnderlineMode(value) {
      this.setProperty('text-underline-mode', value, '');
    }
    get textUnderlinePosition() {
      return this.getPropertyValue('text-underline-position');
    }
    set textUnderlinePosition(value) {
      this.setProperty('text-underline-position', value, '');
    }
    get textUnderlineStyle() {
      return this.getPropertyValue('text-underline-style');
    }
    set textUnderlineStyle(value) {
      this.setProperty('text-underline-style', value, '');
    }
    get textUnderlineWidth() {
      return this.getPropertyValue('text-underline-width');
    }
    set textUnderlineWidth(value) {
      this.setProperty('text-underline-width', value, '');
    }
    get top() {
      return this.getPropertyValue('top');
    }
    set top(value) {
      this.setProperty('top', value, '');
    }
    get touchAction() {
      return this.getPropertyValue('touch-action');
    }
    set touchAction(value) {
      this.setProperty('touch-action', value, '');
    }
    get touchActionDelay() {
      return this.getPropertyValue('touch-action-delay');
    }
    set touchActionDelay(value) {
      this.setProperty('touch-action-delay', value, '');
    }
    get transform() {
      return this.getPropertyValue('transform');
    }
    set transform(value) {
      this.setProperty('transform', value, '');
    }
    get transformOrigin() {
      return this.getPropertyValue('transform-origin');
    }
    set transformOrigin(value) {
      this.setProperty('transform-origin', value, '');
    }
    get transformOriginX() {
      return this.getPropertyValue('transform-origin-x');
    }
    set transformOriginX(value) {
      this.setProperty('transform-origin-x', value, '');
    }
    get transformOriginY() {
      return this.getPropertyValue('transform-origin-y');
    }
    set transformOriginY(value) {
      this.setProperty('transform-origin-y', value, '');
    }
    get transformOriginZ() {
      return this.getPropertyValue('transform-origin-z');
    }
    set transformOriginZ(value) {
      this.setProperty('transform-origin-z', value, '');
    }
    get transformStyle() {
      return this.getPropertyValue('transform-style');
    }
    set transformStyle(value) {
      this.setProperty('transform-style', value, '');
    }
    get transition() {
      return this.getPropertyValue('transition');
    }
    set transition(value) {
      this.setProperty('transition', value, '');
    }
    get transitionDelay() {
      return this.getPropertyValue('transition-delay');
    }
    set transitionDelay(value) {
      this.setProperty('transition-delay', value, '');
    }
    get transitionDuration() {
      return this.getPropertyValue('transition-duration');
    }
    set transitionDuration(value) {
      this.setProperty('transition-duration', value, '');
    }
    get transitionProperty() {
      return this.getPropertyValue('transition-property');
    }
    set transitionProperty(value) {
      this.setProperty('transition-property', value, '');
    }
    get transitionTimingFunction() {
      return this.getPropertyValue('transition-timing-function');
    }
    set transitionTimingFunction(value) {
      this.setProperty('transition-timing-function', value, '');
    }
    get unicodeBidi() {
      return this.getPropertyValue('unicode-bidi');
    }
    set unicodeBidi(value) {
      this.setProperty('unicode-bidi', value, '');
    }
    get unicodeRange() {
      return this.getPropertyValue('unicode-range');
    }
    set unicodeRange(value) {
      this.setProperty('unicode-range', value, '');
    }
    get userDrag() {
      return this.getPropertyValue('user-drag');
    }
    set userDrag(value) {
      this.setProperty('user-drag', value, '');
    }
    get userModify() {
      return this.getPropertyValue('user-modify');
    }
    set userModify(value) {
      this.setProperty('user-modify', value, '');
    }
    get userSelect() {
      return this.getPropertyValue('user-select');
    }
    set userSelect(value) {
      this.setProperty('user-select', value, '');
    }
    get userZoom() {
      return this.getPropertyValue('user-zoom');
    }
    set userZoom(value) {
      this.setProperty('user-zoom', value, '');
    }
    get verticalAlign() {
      return this.getPropertyValue('vertical-align');
    }
    set verticalAlign(value) {
      this.setProperty('vertical-align', value, '');
    }
    get visibility() {
      return this.getPropertyValue('visibility');
    }
    set visibility(value) {
      this.setProperty('visibility', value, '');
    }
    get whiteSpace() {
      return this.getPropertyValue('white-space');
    }
    set whiteSpace(value) {
      this.setProperty('white-space', value, '');
    }
    get widows() {
      return this.getPropertyValue('widows');
    }
    set widows(value) {
      this.setProperty('widows', value, '');
    }
    get width() {
      return this.getPropertyValue('width');
    }
    set width(value) {
      this.setProperty('width', value, '');
    }
    get willChange() {
      return this.getPropertyValue('will-change');
    }
    set willChange(value) {
      this.setProperty('will-change', value, '');
    }
    get wordBreak() {
      return this.getPropertyValue('word-break');
    }
    set wordBreak(value) {
      this.setProperty('word-break', value, '');
    }
    get wordSpacing() {
      return this.getPropertyValue('word-spacing');
    }
    set wordSpacing(value) {
      this.setProperty('word-spacing', value, '');
    }
    get wordWrap() {
      return this.getPropertyValue('word-wrap');
    }
    set wordWrap(value) {
      this.setProperty('word-wrap', value, '');
    }
    get wrapFlow() {
      return this.getPropertyValue('wrap-flow');
    }
    set wrapFlow(value) {
      this.setProperty('wrap-flow', value, '');
    }
    get wrapThrough() {
      return this.getPropertyValue('wrap-through');
    }
    set wrapThrough(value) {
      this.setProperty('wrap-through', value, '');
    }
    get writingMode() {
      return this.getPropertyValue('writing-mode');
    }
    set writingMode(value) {
      this.setProperty('writing-mode', value, '');
    }
    get zIndex() {
      return this.getPropertyValue('z-index');
    }
    set zIndex(value) {
      this.setProperty('z-index', value, '');
    }
    get zoom() {
      return this.getPropertyValue('zoom');
    }
    set zoom(value) {
      this.setProperty('zoom', value, '');
    }
  }
  dart.setSignature(CssStyleDeclarationBase, {
    methods: () => ({
      getPropertyValue: [core.String, [core.String]],
      setProperty: [dart.void, [core.String, core.String], [core.String]]
    })
  });
  const _getPropertyValueHelper = Symbol('_getPropertyValueHelper');
  const _supportsProperty = Symbol('_supportsProperty');
  const _getPropertyValue = Symbol('_getPropertyValue');
  const _setPropertyHelper = Symbol('_setPropertyHelper');
  const _browserPropertyName = Symbol('_browserPropertyName');
  const __getter___1 = Symbol('__getter___1');
  const __getter__ = Symbol('__getter__');
  const __setter___1 = Symbol('__setter___1');
  const __setter__ = Symbol('__setter__');
  const _getPropertyPriority_1 = Symbol('_getPropertyPriority_1');
  const _getPropertyValue_1 = Symbol('_getPropertyValue_1');
  const _item_1 = Symbol('_item_1');
  const _removeProperty_1 = Symbol('_removeProperty_1');
  class CssStyleDeclaration extends dart.mixin(DartHtmlDomObject, CssStyleDeclarationBase) {
    static new() {
      return CssStyleDeclaration.css('');
    }
    static css(css) {
      let style = Element.tag('div').style;
      style.cssText = css;
      return style;
    }
    getPropertyValue(propertyName) {
      let propValue = this[_getPropertyValueHelper](propertyName);
      return propValue != null ? propValue : '';
    }
    [_getPropertyValueHelper](propertyName) {
      if (dart.notNull(this[_supportsProperty](CssStyleDeclaration._camelCase(propertyName)))) {
        return this[_getPropertyValue](propertyName);
      } else {
        return this[_getPropertyValue](dart.notNull(html_common.Device.cssPrefix) + dart.notNull(propertyName));
      }
    }
    supportsProperty(propertyName) {
      return dart.notNull(this[_supportsProperty](propertyName)) || dart.notNull(this[_supportsProperty](CssStyleDeclaration._camelCase(dart.notNull(html_common.Device.cssPrefix) + dart.notNull(propertyName))));
    }
    [_supportsProperty](propertyName) {
      return propertyName in this.raw;
    }
    setProperty(propertyName, value, priority) {
      if (priority === void 0)
        priority = null;
      return this[_setPropertyHelper](this[_browserPropertyName](propertyName), value, priority);
    }
    [_browserPropertyName](propertyName) {
      let name = CssStyleDeclaration._readCache(propertyName);
      if (typeof name == 'string')
        return name;
      if (dart.notNull(this[_supportsProperty](CssStyleDeclaration._camelCase(propertyName)))) {
        name = propertyName;
      } else {
        name = dart.notNull(html_common.Device.cssPrefix) + dart.notNull(propertyName);
      }
      CssStyleDeclaration._writeCache(propertyName, name);
      return name;
    }
    static _readCache(key) {
      return null;
    }
    static _writeCache(key, value) {}
    static _camelCase(hyphenated) {
      return hyphenated[dartx.replaceFirst](core.RegExp.new('^-ms-'), 'ms-')[dartx.replaceAllMapped](core.RegExp.new('-([a-z]+)', {caseSensitive: false}), dart.fn(match => dart.as(dart.dsend(dart.dsend(dart.dindex(dart.dindex(match, 0), 1), 'toUpperCase'), '+', dart.dsend(dart.dindex(match, 0), 'substring', 2)), core.String), core.String, [dart.dynamic]));
    }
    [_setPropertyHelper](propertyName, value, priority) {
      if (priority === void 0)
        priority = null;
      if (value == null)
        value = '';
      if (priority == null)
        priority = '';
      this.raw.setProperty(propertyName, value, priority);
    }
    static get supportsTransitions() {
      return exports.document.body.style.supportsProperty('transition');
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateCssStyleDeclaration() {
      return CssStyleDeclaration._internalWrap();
    }
    static _internalWrap() {
      return new CssStyleDeclaration.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get cssText() {
      return dart.as(wrap_jso(this.raw.cssText), core.String);
    }
    set cssText(val) {
      return this.raw.cssText = unwrap_jso(val);
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get parentRule() {
      return wrap_jso(this.raw.parentRule);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return wrap_jso(this.raw.__getter__(unwrap_jso(name)));
    }
    [__setter__](propertyName, propertyValue) {
      this[__setter___1](propertyName, propertyValue);
      return;
    }
    [__setter___1](propertyName, propertyValue) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(propertyName), unwrap_jso(propertyValue)));
    }
    getPropertyPriority(propertyName) {
      return this[_getPropertyPriority_1](propertyName);
    }
    [_getPropertyPriority_1](propertyName) {
      return dart.as(wrap_jso(this.raw.getPropertyPriority(unwrap_jso(propertyName))), core.String);
    }
    [_getPropertyValue](propertyName) {
      return this[_getPropertyValue_1](propertyName);
    }
    [_getPropertyValue_1](propertyName) {
      return dart.as(wrap_jso(this.raw.getPropertyValue(unwrap_jso(propertyName))), core.String);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), core.String);
    }
    removeProperty(propertyName) {
      return this[_removeProperty_1](propertyName);
    }
    [_removeProperty_1](propertyName) {
      return dart.as(wrap_jso(this.raw.removeProperty(unwrap_jso(propertyName))), core.String);
    }
  }
  dart.defineNamedConstructor(CssStyleDeclaration, 'internal_');
  dart.setSignature(CssStyleDeclaration, {
    constructors: () => ({
      new: [CssStyleDeclaration, []],
      css: [CssStyleDeclaration, [core.String]],
      _: [CssStyleDeclaration, []],
      _internalWrap: [CssStyleDeclaration, []],
      internal_: [CssStyleDeclaration, []]
    }),
    methods: () => ({
      [_getPropertyValueHelper]: [core.String, [core.String]],
      supportsProperty: [core.bool, [core.String]],
      [_supportsProperty]: [core.bool, [core.String]],
      [_browserPropertyName]: [core.String, [core.String]],
      [_setPropertyHelper]: [dart.void, [core.String, core.String], [core.String]],
      [__getter__]: [core.Object, [core.String]],
      [__getter___1]: [core.Object, [dart.dynamic]],
      [__setter__]: [dart.void, [core.String, core.String]],
      [__setter___1]: [dart.void, [dart.dynamic, dart.dynamic]],
      getPropertyPriority: [core.String, [core.String]],
      [_getPropertyPriority_1]: [core.String, [dart.dynamic]],
      [_getPropertyValue]: [core.String, [core.String]],
      [_getPropertyValue_1]: [core.String, [dart.dynamic]],
      item: [core.String, [core.int]],
      [_item_1]: [core.String, [dart.dynamic]],
      removeProperty: [core.String, [core.String]],
      [_removeProperty_1]: [core.String, [dart.dynamic]]
    }),
    statics: () => ({
      _readCache: [core.String, [core.String]],
      _writeCache: [dart.void, [core.String, dart.dynamic]],
      _camelCase: [core.String, [core.String]],
      internalCreateCssStyleDeclaration: [CssStyleDeclaration, []]
    }),
    names: ['_readCache', '_writeCache', '_camelCase', 'internalCreateCssStyleDeclaration']
  });
  CssStyleDeclaration[dart.metadata] = () => [dart.const(new _metadata.DomName('CSSStyleDeclaration')), dart.const(new _js_helper.Native("CSSStyleDeclaration,MSStyleCSSProperties,CSS2Properties"))];
  const _elementIterable = Symbol('_elementIterable');
  const _elementCssStyleDeclarationSetIterable = Symbol('_elementCssStyleDeclarationSetIterable');
  class _CssStyleDeclarationSet extends dart.mixin(core.Object, CssStyleDeclarationBase) {
    _CssStyleDeclarationSet(elementIterable) {
      this[_elementIterable] = elementIterable;
      this[_elementCssStyleDeclarationSetIterable] = null;
      this[_elementCssStyleDeclarationSetIterable] = dart.as(core.List.from(this[_elementIterable])[dartx.map](dart.fn(e => dart.dload(e, 'style'))), core.Iterable$(CssStyleDeclaration));
    }
    getPropertyValue(propertyName) {
      return this[_elementCssStyleDeclarationSetIterable][dartx.first].getPropertyValue(propertyName);
    }
    setProperty(propertyName, value, priority) {
      if (priority === void 0)
        priority = null;
      this[_elementCssStyleDeclarationSetIterable][dartx.forEach](dart.fn(e => dart.dsend(e, 'setProperty', propertyName, value, priority)));
    }
  }
  dart.setSignature(_CssStyleDeclarationSet, {
    constructors: () => ({_CssStyleDeclarationSet: [_CssStyleDeclarationSet, [core.Iterable$(Element)]]})
  });
  class DivElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("div"), DivElement);
    }
    static internalCreateDivElement() {
      return DivElement._internalWrap();
    }
    static _internalWrap() {
      return new DivElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(DivElement, 'internal_');
  dart.defineNamedConstructor(DivElement, 'created');
  dart.setSignature(DivElement, {
    constructors: () => ({
      _: [DivElement, []],
      new: [DivElement, []],
      _internalWrap: [DivElement, []],
      internal_: [DivElement, []],
      created: [DivElement, []]
    }),
    statics: () => ({internalCreateDivElement: [DivElement, []]}),
    names: ['internalCreateDivElement']
  });
  DivElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLDivElement')), dart.const(new _js_helper.Native("HTMLDivElement"))];
  const _body = Symbol('_body');
  const _get_window = Symbol('_get_window');
  const _head = Symbol('_head');
  const _lastModified = Symbol('_lastModified');
  const _preferredStylesheetSet = Symbol('_preferredStylesheetSet');
  const _referrer = Symbol('_referrer');
  const _selectedStylesheetSet = Symbol('_selectedStylesheetSet');
  const _styleSheets = Symbol('_styleSheets');
  const _title = Symbol('_title');
  const _webkitFullscreenElement = Symbol('_webkitFullscreenElement');
  const _webkitFullscreenEnabled = Symbol('_webkitFullscreenEnabled');
  const _webkitHidden = Symbol('_webkitHidden');
  const _webkitVisibilityState = Symbol('_webkitVisibilityState');
  const _adoptNode_1 = Symbol('_adoptNode_1');
  const _caretRangeFromPoint_1 = Symbol('_caretRangeFromPoint_1');
  const _caretRangeFromPoint = Symbol('_caretRangeFromPoint');
  const _createDocumentFragment_1 = Symbol('_createDocumentFragment_1');
  const _createElement_1 = Symbol('_createElement_1');
  const _createElement_2 = Symbol('_createElement_2');
  const _createElement = Symbol('_createElement');
  const _createElementNS_1 = Symbol('_createElementNS_1');
  const _createElementNS_2 = Symbol('_createElementNS_2');
  const _createElementNS = Symbol('_createElementNS');
  const _createEvent_1 = Symbol('_createEvent_1');
  const _createRange_1 = Symbol('_createRange_1');
  const _createTextNode_1 = Symbol('_createTextNode_1');
  const _createTextNode = Symbol('_createTextNode');
  const _createTouch_1 = Symbol('_createTouch_1');
  const _createTouch = Symbol('_createTouch');
  const _createTouchList_1 = Symbol('_createTouchList_1');
  const _createTouchList = Symbol('_createTouchList');
  const _elementFromPoint_1 = Symbol('_elementFromPoint_1');
  const _elementFromPoint = Symbol('_elementFromPoint');
  const _execCommand_1 = Symbol('_execCommand_1');
  const _exitFullscreen_1 = Symbol('_exitFullscreen_1');
  const _exitPointerLock_1 = Symbol('_exitPointerLock_1');
  const _getCssCanvasContext_1 = Symbol('_getCssCanvasContext_1');
  const _getCssCanvasContext = Symbol('_getCssCanvasContext');
  const _getElementById_1 = Symbol('_getElementById_1');
  const _getElementsByName_1 = Symbol('_getElementsByName_1');
  const _importNode_1 = Symbol('_importNode_1');
  const _importNode_2 = Symbol('_importNode_2');
  const _queryCommandEnabled_1 = Symbol('_queryCommandEnabled_1');
  const _queryCommandIndeterm_1 = Symbol('_queryCommandIndeterm_1');
  const _queryCommandState_1 = Symbol('_queryCommandState_1');
  const _queryCommandSupported_1 = Symbol('_queryCommandSupported_1');
  const _queryCommandValue_1 = Symbol('_queryCommandValue_1');
  const _transformDocumentToTreeView_1 = Symbol('_transformDocumentToTreeView_1');
  const _webkitExitFullscreen_1 = Symbol('_webkitExitFullscreen_1');
  const _webkitExitFullscreen = Symbol('_webkitExitFullscreen');
  class Document extends Node {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateDocument() {
      return Document._internalWrap();
    }
    static _internalWrap() {
      return new Document.internal_();
    }
    internal_() {
      super.internal_();
    }
    get activeElement() {
      return dart.as(wrap_jso(this.raw.activeElement), Element);
    }
    get [_body]() {
      return dart.as(wrap_jso(this.raw.body), HtmlElement);
    }
    set [_body](val) {
      return this.raw.body = unwrap_jso(val);
    }
    get contentType() {
      return dart.as(wrap_jso(this.raw.contentType), core.String);
    }
    get cookie() {
      return dart.as(wrap_jso(this.raw.cookie), core.String);
    }
    set cookie(val) {
      return this.raw.cookie = unwrap_jso(val);
    }
    get currentScript() {
      return dart.as(wrap_jso(this.raw.currentScript), ScriptElement);
    }
    get window() {
      return _convertNativeToDart_Window(this[_get_window]);
    }
    get [_get_window]() {
      return wrap_jso(this.raw.defaultView);
    }
    get documentElement() {
      return dart.as(wrap_jso(this.raw.documentElement), Element);
    }
    get domain() {
      return dart.as(wrap_jso(this.raw.domain), core.String);
    }
    get fonts() {
      return wrap_jso(this.raw.fonts);
    }
    get fullscreenElement() {
      return dart.as(wrap_jso(this.raw.fullscreenElement), Element);
    }
    get fullscreenEnabled() {
      return dart.as(wrap_jso(this.raw.fullscreenEnabled), core.bool);
    }
    get [_head]() {
      return dart.as(wrap_jso(this.raw.head), Element);
    }
    get hidden() {
      return dart.as(wrap_jso(this.raw.hidden), core.bool);
    }
    get implementation() {
      return dart.as(wrap_jso(this.raw.implementation), DomImplementation);
    }
    get [_lastModified]() {
      return dart.as(wrap_jso(this.raw.lastModified), core.String);
    }
    get pointerLockElement() {
      return dart.as(wrap_jso(this.raw.pointerLockElement), Element);
    }
    get [_preferredStylesheetSet]() {
      return dart.as(wrap_jso(this.raw.preferredStylesheetSet), core.String);
    }
    get readyState() {
      return dart.as(wrap_jso(this.raw.readyState), core.String);
    }
    get [_referrer]() {
      return dart.as(wrap_jso(this.raw.referrer), core.String);
    }
    get rootElement() {
      return wrap_jso(this.raw.rootElement);
    }
    get [_selectedStylesheetSet]() {
      return dart.as(wrap_jso(this.raw.selectedStylesheetSet), core.String);
    }
    set [_selectedStylesheetSet](val) {
      return this.raw.selectedStylesheetSet = unwrap_jso(val);
    }
    get [_styleSheets]() {
      return dart.as(wrap_jso(this.raw.styleSheets), core.List$(StyleSheet));
    }
    get timeline() {
      return wrap_jso(this.raw.timeline);
    }
    get [_title]() {
      return dart.as(wrap_jso(this.raw.title), core.String);
    }
    set [_title](val) {
      return this.raw.title = unwrap_jso(val);
    }
    get visibilityState() {
      return dart.as(wrap_jso(this.raw.visibilityState), core.String);
    }
    get [_webkitFullscreenElement]() {
      return dart.as(wrap_jso(this.raw.webkitFullscreenElement), Element);
    }
    get [_webkitFullscreenEnabled]() {
      return dart.as(wrap_jso(this.raw.webkitFullscreenEnabled), core.bool);
    }
    get [_webkitHidden]() {
      return dart.as(wrap_jso(this.raw.webkitHidden), core.bool);
    }
    get [_webkitVisibilityState]() {
      return dart.as(wrap_jso(this.raw.webkitVisibilityState), core.String);
    }
    adoptNode(node) {
      return this[_adoptNode_1](node);
    }
    [_adoptNode_1](node) {
      return dart.as(wrap_jso(this.raw.adoptNode(unwrap_jso(node))), Node);
    }
    [_caretRangeFromPoint](x, y) {
      return this[_caretRangeFromPoint_1](x, y);
    }
    [_caretRangeFromPoint_1](x, y) {
      return dart.as(wrap_jso(this.raw.caretRangeFromPoint(unwrap_jso(x), unwrap_jso(y))), Range);
    }
    createDocumentFragment() {
      return this[_createDocumentFragment_1]();
    }
    [_createDocumentFragment_1]() {
      return dart.as(wrap_jso(this.raw.createDocumentFragment()), DocumentFragment);
    }
    [_createElement](localName_OR_tagName, typeExtension) {
      if (typeExtension === void 0)
        typeExtension = null;
      if (typeExtension == null) {
        return this[_createElement_1](localName_OR_tagName);
      }
      if (typeExtension != null) {
        return this[_createElement_2](localName_OR_tagName, typeExtension);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_createElement_1](tagName) {
      return dart.as(wrap_jso(this.raw.createElement(unwrap_jso(tagName))), Element);
    }
    [_createElement_2](localName, typeExtension) {
      return dart.as(wrap_jso(this.raw.createElement(unwrap_jso(localName), unwrap_jso(typeExtension))), Element);
    }
    [_createElementNS](namespaceURI, qualifiedName, typeExtension) {
      if (typeExtension === void 0)
        typeExtension = null;
      if (typeExtension == null) {
        return this[_createElementNS_1](namespaceURI, qualifiedName);
      }
      if (typeExtension != null) {
        return this[_createElementNS_2](namespaceURI, qualifiedName, typeExtension);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_createElementNS_1](namespaceURI, qualifiedName) {
      return dart.as(wrap_jso(this.raw.createElementNS(unwrap_jso(namespaceURI), unwrap_jso(qualifiedName))), Element);
    }
    [_createElementNS_2](namespaceURI, qualifiedName, typeExtension) {
      return dart.as(wrap_jso(this.raw.createElementNS(unwrap_jso(namespaceURI), unwrap_jso(qualifiedName), unwrap_jso(typeExtension))), Element);
    }
    [_createEvent](eventType) {
      return this[_createEvent_1](eventType);
    }
    [_createEvent_1](eventType) {
      return dart.as(wrap_jso(this.raw.createEvent(unwrap_jso(eventType))), Event);
    }
    createRange() {
      return this[_createRange_1]();
    }
    [_createRange_1]() {
      return dart.as(wrap_jso(this.raw.createRange()), Range);
    }
    [_createTextNode](data) {
      return this[_createTextNode_1](data);
    }
    [_createTextNode_1](data) {
      return dart.as(wrap_jso(this.raw.createTextNode(unwrap_jso(data))), Text);
    }
    [_createTouch](window, target, identifier, pageX, pageY, screenX, screenY, webkitRadiusX, webkitRadiusY, webkitRotationAngle, webkitForce) {
      let target_1 = _convertDartToNative_EventTarget(target);
      return this[_createTouch_1](window, target_1, identifier, pageX, pageY, screenX, screenY, webkitRadiusX, webkitRadiusY, webkitRotationAngle, webkitForce);
    }
    [_createTouch_1](window, target, identifier, pageX, pageY, screenX, screenY, webkitRadiusX, webkitRadiusY, webkitRotationAngle, webkitForce) {
      return dart.as(wrap_jso(this.raw.createTouch(unwrap_jso(window), unwrap_jso(target), unwrap_jso(identifier), unwrap_jso(pageX), unwrap_jso(pageY), unwrap_jso(screenX), unwrap_jso(screenY), unwrap_jso(webkitRadiusX), unwrap_jso(webkitRadiusY), unwrap_jso(webkitRotationAngle), unwrap_jso(webkitForce))), Touch);
    }
    [_createTouchList](touches) {
      return this[_createTouchList_1](touches);
    }
    [_createTouchList_1](touches) {
      return dart.as(wrap_jso(this.raw.createTouchList(unwrap_jso(touches))), TouchList);
    }
    [_elementFromPoint](x, y) {
      return this[_elementFromPoint_1](x, y);
    }
    [_elementFromPoint_1](x, y) {
      return dart.as(wrap_jso(this.raw.elementFromPoint(unwrap_jso(x), unwrap_jso(y))), Element);
    }
    execCommand(command, userInterface, value) {
      return this[_execCommand_1](command, userInterface, value);
    }
    [_execCommand_1](command, userInterface, value) {
      return dart.as(wrap_jso(this.raw.execCommand(unwrap_jso(command), unwrap_jso(userInterface), unwrap_jso(value))), core.bool);
    }
    exitFullscreen() {
      this[_exitFullscreen_1]();
      return;
    }
    [_exitFullscreen_1]() {
      return wrap_jso(this.raw.exitFullscreen());
    }
    exitPointerLock() {
      this[_exitPointerLock_1]();
      return;
    }
    [_exitPointerLock_1]() {
      return wrap_jso(this.raw.exitPointerLock());
    }
    [_getCssCanvasContext](contextId, name, width, height) {
      return this[_getCssCanvasContext_1](contextId, name, width, height);
    }
    [_getCssCanvasContext_1](contextId, name, width, height) {
      return wrap_jso(this.raw.getCSSCanvasContext(unwrap_jso(contextId), unwrap_jso(name), unwrap_jso(width), unwrap_jso(height)));
    }
    getElementById(elementId) {
      return this[_getElementById_1](elementId);
    }
    [_getElementById_1](elementId) {
      return dart.as(wrap_jso(this.raw.getElementById(unwrap_jso(elementId))), Element);
    }
    getElementsByClassName(classNames) {
      return this[_getElementsByClassName_1](classNames);
    }
    [_getElementsByClassName_1](classNames) {
      return dart.as(wrap_jso(this.raw.getElementsByClassName(unwrap_jso(classNames))), HtmlCollection);
    }
    getElementsByName(elementName) {
      return this[_getElementsByName_1](elementName);
    }
    [_getElementsByName_1](elementName) {
      return dart.as(wrap_jso(this.raw.getElementsByName(unwrap_jso(elementName))), NodeList);
    }
    getElementsByTagName(localName) {
      return this[_getElementsByTagName_1](localName);
    }
    [_getElementsByTagName_1](localName) {
      return dart.as(wrap_jso(this.raw.getElementsByTagName(unwrap_jso(localName))), HtmlCollection);
    }
    importNode(node, deep) {
      if (deep === void 0)
        deep = null;
      if (deep != null) {
        return this[_importNode_1](node, deep);
      }
      return this[_importNode_2](node);
    }
    [_importNode_1](node, deep) {
      return dart.as(wrap_jso(this.raw.importNode(unwrap_jso(node), unwrap_jso(deep))), Node);
    }
    [_importNode_2](node) {
      return dart.as(wrap_jso(this.raw.importNode(unwrap_jso(node))), Node);
    }
    queryCommandEnabled(command) {
      return this[_queryCommandEnabled_1](command);
    }
    [_queryCommandEnabled_1](command) {
      return dart.as(wrap_jso(this.raw.queryCommandEnabled(unwrap_jso(command))), core.bool);
    }
    queryCommandIndeterm(command) {
      return this[_queryCommandIndeterm_1](command);
    }
    [_queryCommandIndeterm_1](command) {
      return dart.as(wrap_jso(this.raw.queryCommandIndeterm(unwrap_jso(command))), core.bool);
    }
    queryCommandState(command) {
      return this[_queryCommandState_1](command);
    }
    [_queryCommandState_1](command) {
      return dart.as(wrap_jso(this.raw.queryCommandState(unwrap_jso(command))), core.bool);
    }
    queryCommandSupported(command) {
      return this[_queryCommandSupported_1](command);
    }
    [_queryCommandSupported_1](command) {
      return dart.as(wrap_jso(this.raw.queryCommandSupported(unwrap_jso(command))), core.bool);
    }
    queryCommandValue(command) {
      return this[_queryCommandValue_1](command);
    }
    [_queryCommandValue_1](command) {
      return dart.as(wrap_jso(this.raw.queryCommandValue(unwrap_jso(command))), core.String);
    }
    transformDocumentToTreeView(noStyleMessage) {
      this[_transformDocumentToTreeView_1](noStyleMessage);
      return;
    }
    [_transformDocumentToTreeView_1](noStyleMessage) {
      return wrap_jso(this.raw.transformDocumentToTreeView(unwrap_jso(noStyleMessage)));
    }
    [_webkitExitFullscreen]() {
      this[_webkitExitFullscreen_1]();
      return;
    }
    [_webkitExitFullscreen_1]() {
      return wrap_jso(this.raw.webkitExitFullscreen());
    }
    get [_childElementCount]() {
      return dart.as(wrap_jso(this.raw.childElementCount), core.int);
    }
    get [_children]() {
      return dart.as(wrap_jso(this.raw.children), core.List$(Node));
    }
    get [_firstElementChild]() {
      return dart.as(wrap_jso(this.raw.firstElementChild), Element);
    }
    get [_lastElementChild]() {
      return dart.as(wrap_jso(this.raw.lastElementChild), Element);
    }
    querySelector(selectors) {
      return this[_querySelector_1](selectors);
    }
    [_querySelector_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelector(unwrap_jso(selectors))), Element);
    }
    [_querySelectorAll](selectors) {
      return this[_querySelectorAll_1](selectors);
    }
    [_querySelectorAll_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelectorAll(unwrap_jso(selectors))), NodeList);
    }
    get onAbort() {
      return Element.abortEvent.forTarget(this);
    }
    get onBeforeCopy() {
      return Element.beforeCopyEvent.forTarget(this);
    }
    get onBeforeCut() {
      return Element.beforeCutEvent.forTarget(this);
    }
    get onBeforePaste() {
      return Element.beforePasteEvent.forTarget(this);
    }
    get onBlur() {
      return Element.blurEvent.forTarget(this);
    }
    get onCanPlay() {
      return Element.canPlayEvent.forTarget(this);
    }
    get onCanPlayThrough() {
      return Element.canPlayThroughEvent.forTarget(this);
    }
    get onChange() {
      return Element.changeEvent.forTarget(this);
    }
    get onClick() {
      return Element.clickEvent.forTarget(this);
    }
    get onContextMenu() {
      return Element.contextMenuEvent.forTarget(this);
    }
    get onCopy() {
      return Element.copyEvent.forTarget(this);
    }
    get onCut() {
      return Element.cutEvent.forTarget(this);
    }
    get onDoubleClick() {
      return Element.doubleClickEvent.forTarget(this);
    }
    get onDrag() {
      return Element.dragEvent.forTarget(this);
    }
    get onDragEnd() {
      return Element.dragEndEvent.forTarget(this);
    }
    get onDragEnter() {
      return Element.dragEnterEvent.forTarget(this);
    }
    get onDragLeave() {
      return Element.dragLeaveEvent.forTarget(this);
    }
    get onDragOver() {
      return Element.dragOverEvent.forTarget(this);
    }
    get onDragStart() {
      return Element.dragStartEvent.forTarget(this);
    }
    get onDrop() {
      return Element.dropEvent.forTarget(this);
    }
    get onDurationChange() {
      return Element.durationChangeEvent.forTarget(this);
    }
    get onEmptied() {
      return Element.emptiedEvent.forTarget(this);
    }
    get onEnded() {
      return Element.endedEvent.forTarget(this);
    }
    get onError() {
      return Element.errorEvent.forTarget(this);
    }
    get onFocus() {
      return Element.focusEvent.forTarget(this);
    }
    get onInput() {
      return Element.inputEvent.forTarget(this);
    }
    get onInvalid() {
      return Element.invalidEvent.forTarget(this);
    }
    get onKeyDown() {
      return Element.keyDownEvent.forTarget(this);
    }
    get onKeyPress() {
      return Element.keyPressEvent.forTarget(this);
    }
    get onKeyUp() {
      return Element.keyUpEvent.forTarget(this);
    }
    get onLoad() {
      return Element.loadEvent.forTarget(this);
    }
    get onLoadedData() {
      return Element.loadedDataEvent.forTarget(this);
    }
    get onLoadedMetadata() {
      return Element.loadedMetadataEvent.forTarget(this);
    }
    get onMouseDown() {
      return Element.mouseDownEvent.forTarget(this);
    }
    get onMouseEnter() {
      return Element.mouseEnterEvent.forTarget(this);
    }
    get onMouseLeave() {
      return Element.mouseLeaveEvent.forTarget(this);
    }
    get onMouseMove() {
      return Element.mouseMoveEvent.forTarget(this);
    }
    get onMouseOut() {
      return Element.mouseOutEvent.forTarget(this);
    }
    get onMouseOver() {
      return Element.mouseOverEvent.forTarget(this);
    }
    get onMouseUp() {
      return Element.mouseUpEvent.forTarget(this);
    }
    get onMouseWheel() {
      return dart.as(dart.dsend(dart.dload(Element, 'mouseWheelEvent'), 'forTarget', this), async.Stream$(dart.dynamic));
    }
    get onPaste() {
      return Element.pasteEvent.forTarget(this);
    }
    get onPause() {
      return Element.pauseEvent.forTarget(this);
    }
    get onPlay() {
      return Element.playEvent.forTarget(this);
    }
    get onPlaying() {
      return Element.playingEvent.forTarget(this);
    }
    get onPointerLockChange() {
      return Document.pointerLockChangeEvent.forTarget(this);
    }
    get onPointerLockError() {
      return Document.pointerLockErrorEvent.forTarget(this);
    }
    get onRateChange() {
      return Element.rateChangeEvent.forTarget(this);
    }
    get onReadyStateChange() {
      return Document.readyStateChangeEvent.forTarget(this);
    }
    get onReset() {
      return Element.resetEvent.forTarget(this);
    }
    get onResize() {
      return Element.resizeEvent.forTarget(this);
    }
    get onScroll() {
      return Element.scrollEvent.forTarget(this);
    }
    get onSearch() {
      return Element.searchEvent.forTarget(this);
    }
    get onSecurityPolicyViolation() {
      return Document.securityPolicyViolationEvent.forTarget(this);
    }
    get onSeeked() {
      return Element.seekedEvent.forTarget(this);
    }
    get onSeeking() {
      return Element.seekingEvent.forTarget(this);
    }
    get onSelect() {
      return Element.selectEvent.forTarget(this);
    }
    get onSelectionChange() {
      return Document.selectionChangeEvent.forTarget(this);
    }
    get onSelectStart() {
      return Element.selectStartEvent.forTarget(this);
    }
    get onStalled() {
      return Element.stalledEvent.forTarget(this);
    }
    get onSubmit() {
      return Element.submitEvent.forTarget(this);
    }
    get onSuspend() {
      return Element.suspendEvent.forTarget(this);
    }
    get onTimeUpdate() {
      return Element.timeUpdateEvent.forTarget(this);
    }
    get onTouchCancel() {
      return Element.touchCancelEvent.forTarget(this);
    }
    get onTouchEnd() {
      return Element.touchEndEvent.forTarget(this);
    }
    get onTouchMove() {
      return Element.touchMoveEvent.forTarget(this);
    }
    get onTouchStart() {
      return Element.touchStartEvent.forTarget(this);
    }
    get onVolumeChange() {
      return Element.volumeChangeEvent.forTarget(this);
    }
    get onWaiting() {
      return Element.waitingEvent.forTarget(this);
    }
    get onFullscreenChange() {
      return Element.fullscreenChangeEvent.forTarget(this);
    }
    get onFullscreenError() {
      return Element.fullscreenErrorEvent.forTarget(this);
    }
    querySelectorAll(selectors) {
      return new _FrozenElementList._wrap(this[_querySelectorAll](selectors));
    }
    query(relativeSelectors) {
      return this.querySelector(relativeSelectors);
    }
    queryAll(relativeSelectors) {
      return this.querySelectorAll(relativeSelectors);
    }
    get supportsRegisterElement() {
      return true;
    }
    get supportsRegister() {
      return this.supportsRegisterElement;
    }
    createElement(tagName, typeExtension) {
      if (typeExtension === void 0)
        typeExtension = null;
      return this[_createElement](tagName, typeExtension);
    }
    createElementNS(namespaceURI, qualifiedName, typeExtension) {
      if (typeExtension === void 0)
        typeExtension = null;
      let newElement = typeExtension == null ? dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkDocument'), 'instance'), 'createElementNS_Callback_2_', unwrap_jso(this), namespaceURI, qualifiedName) : dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkDocument'), 'instance'), 'createElementNS_Callback_3_', unwrap_jso(this), namespaceURI, qualifiedName, typeExtension);
      let wrapped = null;
      wrapped = dart.dsend(/* Unimplemented unknown name */js, 'getDartHtmlWrapperFor', newElement);
      if (wrapped != null) {
        dart.dput(wrapped, 'blink_jsObject', newElement);
      } else {
        wrapped = wrap_jso(newElement);
        if (wrapped == null) {
          wrapped = dart.dcall(/* Unimplemented unknown name */wrap_jso_custom_element, newElement);
        } else {
          dart.dsend(/* Unimplemented unknown name */js, 'setDartHtmlWrapperFor', dart.dload(wrapped, 'blink_jsObject'), wrapped);
        }
      }
      return dart.as(wrapped, Element);
    }
  }
  dart.defineNamedConstructor(Document, 'internal_');
  dart.setSignature(Document, {
    constructors: () => ({
      _: [Document, []],
      _internalWrap: [Document, []],
      internal_: [Document, []]
    }),
    methods: () => ({
      adoptNode: [Node, [Node]],
      [_adoptNode_1]: [Node, [Node]],
      [_caretRangeFromPoint]: [Range, [core.int, core.int]],
      [_caretRangeFromPoint_1]: [Range, [dart.dynamic, dart.dynamic]],
      createDocumentFragment: [DocumentFragment, []],
      [_createDocumentFragment_1]: [DocumentFragment, []],
      [_createElement]: [Element, [core.String], [core.String]],
      [_createElement_1]: [Element, [dart.dynamic]],
      [_createElement_2]: [Element, [dart.dynamic, dart.dynamic]],
      [_createElementNS]: [Element, [core.String, core.String], [core.String]],
      [_createElementNS_1]: [Element, [dart.dynamic, dart.dynamic]],
      [_createElementNS_2]: [Element, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_createEvent]: [Event, [core.String]],
      [_createEvent_1]: [Event, [dart.dynamic]],
      createRange: [Range, []],
      [_createRange_1]: [Range, []],
      [_createTextNode]: [Text, [core.String]],
      [_createTextNode_1]: [Text, [dart.dynamic]],
      [_createTouch]: [Touch, [Window, EventTarget, core.int, core.num, core.num, core.num, core.num, core.num, core.num, core.num, core.num]],
      [_createTouch_1]: [Touch, [Window, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      [_createTouchList]: [TouchList, [Touch]],
      [_createTouchList_1]: [TouchList, [Touch]],
      [_elementFromPoint]: [Element, [core.int, core.int]],
      [_elementFromPoint_1]: [Element, [dart.dynamic, dart.dynamic]],
      execCommand: [core.bool, [core.String, core.bool, core.String]],
      [_execCommand_1]: [core.bool, [dart.dynamic, dart.dynamic, dart.dynamic]],
      exitFullscreen: [dart.void, []],
      [_exitFullscreen_1]: [dart.void, []],
      exitPointerLock: [dart.void, []],
      [_exitPointerLock_1]: [dart.void, []],
      [_getCssCanvasContext]: [core.Object, [core.String, core.String, core.int, core.int]],
      [_getCssCanvasContext_1]: [core.Object, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      getElementById: [Element, [core.String]],
      [_getElementById_1]: [Element, [dart.dynamic]],
      getElementsByClassName: [HtmlCollection, [core.String]],
      [_getElementsByClassName_1]: [HtmlCollection, [dart.dynamic]],
      getElementsByName: [NodeList, [core.String]],
      [_getElementsByName_1]: [NodeList, [dart.dynamic]],
      getElementsByTagName: [HtmlCollection, [core.String]],
      [_getElementsByTagName_1]: [HtmlCollection, [dart.dynamic]],
      importNode: [Node, [Node], [core.bool]],
      [_importNode_1]: [Node, [Node, dart.dynamic]],
      [_importNode_2]: [Node, [Node]],
      queryCommandEnabled: [core.bool, [core.String]],
      [_queryCommandEnabled_1]: [core.bool, [dart.dynamic]],
      queryCommandIndeterm: [core.bool, [core.String]],
      [_queryCommandIndeterm_1]: [core.bool, [dart.dynamic]],
      queryCommandState: [core.bool, [core.String]],
      [_queryCommandState_1]: [core.bool, [dart.dynamic]],
      queryCommandSupported: [core.bool, [core.String]],
      [_queryCommandSupported_1]: [core.bool, [dart.dynamic]],
      queryCommandValue: [core.String, [core.String]],
      [_queryCommandValue_1]: [core.String, [dart.dynamic]],
      transformDocumentToTreeView: [dart.void, [core.String]],
      [_transformDocumentToTreeView_1]: [dart.void, [dart.dynamic]],
      [_webkitExitFullscreen]: [dart.void, []],
      [_webkitExitFullscreen_1]: [dart.void, []],
      querySelector: [Element, [core.String]],
      [_querySelector_1]: [Element, [dart.dynamic]],
      [_querySelectorAll]: [NodeList, [core.String]],
      [_querySelectorAll_1]: [NodeList, [dart.dynamic]],
      querySelectorAll: [ElementList$(Element), [core.String]],
      query: [Element, [core.String]],
      queryAll: [ElementList$(Element), [core.String]],
      createElement: [Element, [core.String], [core.String]],
      createElementNS: [Element, [core.String, core.String], [core.String]]
    }),
    statics: () => ({internalCreateDocument: [Document, []]}),
    names: ['internalCreateDocument']
  });
  Document[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Document')), dart.const(new _js_helper.Native("Document"))];
  Document.pointerLockChangeEvent = dart.const(new (EventStreamProvider$(Event))('pointerlockchange'));
  Document.pointerLockErrorEvent = dart.const(new (EventStreamProvider$(Event))('pointerlockerror'));
  Document.readyStateChangeEvent = dart.const(new (EventStreamProvider$(Event))('readystatechange'));
  class SecurityPolicyViolationEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSecurityPolicyViolationEvent() {
      return SecurityPolicyViolationEvent._internalWrap();
    }
    static _internalWrap() {
      return new SecurityPolicyViolationEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get blockedUri() {
      return dart.as(wrap_jso(this.raw.blockedURI), core.String);
    }
    get columnNumber() {
      return dart.as(wrap_jso(this.raw.columnNumber), core.int);
    }
    get documentUri() {
      return dart.as(wrap_jso(this.raw.documentURI), core.String);
    }
    get effectiveDirective() {
      return dart.as(wrap_jso(this.raw.effectiveDirective), core.String);
    }
    get lineNumber() {
      return dart.as(wrap_jso(this.raw.lineNumber), core.int);
    }
    get originalPolicy() {
      return dart.as(wrap_jso(this.raw.originalPolicy), core.String);
    }
    get referrer() {
      return dart.as(wrap_jso(this.raw.referrer), core.String);
    }
    get sourceFile() {
      return dart.as(wrap_jso(this.raw.sourceFile), core.String);
    }
    get statusCode() {
      return dart.as(wrap_jso(this.raw.statusCode), core.int);
    }
    get violatedDirective() {
      return dart.as(wrap_jso(this.raw.violatedDirective), core.String);
    }
  }
  dart.defineNamedConstructor(SecurityPolicyViolationEvent, 'internal_');
  dart.setSignature(SecurityPolicyViolationEvent, {
    constructors: () => ({
      _: [SecurityPolicyViolationEvent, []],
      _internalWrap: [SecurityPolicyViolationEvent, []],
      internal_: [SecurityPolicyViolationEvent, []]
    }),
    statics: () => ({internalCreateSecurityPolicyViolationEvent: [SecurityPolicyViolationEvent, []]}),
    names: ['internalCreateSecurityPolicyViolationEvent']
  });
  SecurityPolicyViolationEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SecurityPolicyViolationEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SecurityPolicyViolationEvent"))];
  Document.securityPolicyViolationEvent = dart.const(new (EventStreamProvider$(SecurityPolicyViolationEvent))('securitypolicyviolation'));
  Document.selectionChangeEvent = dart.const(new (EventStreamProvider$(Event))('selectionchange'));
  const _docChildren = Symbol('_docChildren');
  class DocumentFragment extends Node {
    static new() {
      return exports.document.createDocumentFragment();
    }
    static html(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      return exports.document.body.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer});
    }
    static svg(svgContent, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      return dart.as(dart.dsend(new dart.dynamic(), 'createFragment', svgContent, {validator: validator, treeSanitizer: treeSanitizer}), DocumentFragment);
    }
    get [_children]() {
      return dart.throw(new core.UnimplementedError('Use _docChildren instead'));
    }
    get children() {
      if (this[_docChildren] == null) {
        this[_docChildren] = new html_common.FilteredElementList(this);
      }
      return this[_docChildren];
    }
    set children(value) {
      let copy = core.List.from(value);
      let children = this.children;
      children[dartx.clear]();
      children[dartx.addAll](dart.as(copy, core.Iterable$(Element)));
    }
    querySelectorAll(selectors) {
      return new _FrozenElementList._wrap(this[_querySelectorAll](selectors));
    }
    get innerHtml() {
      let e = Element.tag("div");
      e.append(this.clone(true));
      return e.innerHtml;
    }
    set innerHtml(value) {
      this.setInnerHtml(value);
    }
    setInnerHtml(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      this.nodes[dartx.clear]();
      this.append(exports.document.body.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer}));
    }
    appendText(text) {
      this.append(Text.new(text));
    }
    appendHtml(text, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let NodeTreeSanitizer = opts && 'NodeTreeSanitizer' in opts ? opts.NodeTreeSanitizer : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      this.append(DocumentFragment.html(text, {validator: validator, treeSanitizer: dart.as(treeSanitizer, NodeTreeSanitizer)}));
    }
    query(relativeSelectors) {
      return this.querySelector(relativeSelectors);
    }
    queryAll(relativeSelectors) {
      return this.querySelectorAll(relativeSelectors);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateDocumentFragment() {
      return DocumentFragment._internalWrap();
    }
    static _internalWrap() {
      return new DocumentFragment.internal_();
    }
    internal_() {
      this[_docChildren] = null;
      super.internal_();
    }
    getElementById(elementId) {
      return this[_getElementById_1](elementId);
    }
    [_getElementById_1](elementId) {
      return dart.as(wrap_jso(this.raw.getElementById(unwrap_jso(elementId))), Element);
    }
    get [_childElementCount]() {
      return dart.as(wrap_jso(this.raw.childElementCount), core.int);
    }
    get [_firstElementChild]() {
      return dart.as(wrap_jso(this.raw.firstElementChild), Element);
    }
    get [_lastElementChild]() {
      return dart.as(wrap_jso(this.raw.lastElementChild), Element);
    }
    querySelector(selectors) {
      return this[_querySelector_1](selectors);
    }
    [_querySelector_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelector(unwrap_jso(selectors))), Element);
    }
    [_querySelectorAll](selectors) {
      return this[_querySelectorAll_1](selectors);
    }
    [_querySelectorAll_1](selectors) {
      return dart.as(wrap_jso(this.raw.querySelectorAll(unwrap_jso(selectors))), NodeList);
    }
  }
  DocumentFragment[dart.implements] = () => [ParentNode];
  dart.defineNamedConstructor(DocumentFragment, 'internal_');
  dart.setSignature(DocumentFragment, {
    constructors: () => ({
      new: [DocumentFragment, []],
      html: [DocumentFragment, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      svg: [DocumentFragment, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      _: [DocumentFragment, []],
      _internalWrap: [DocumentFragment, []],
      internal_: [DocumentFragment, []]
    }),
    methods: () => ({
      querySelectorAll: [ElementList$(Element), [core.String]],
      setInnerHtml: [dart.void, [core.String], {validator: NodeValidator, treeSanitizer: NodeTreeSanitizer}],
      appendText: [dart.void, [core.String]],
      appendHtml: [dart.void, [core.String], {validator: NodeValidator, NodeTreeSanitizer: dart.dynamic, treeSanitizer: dart.dynamic}],
      query: [Element, [core.String]],
      queryAll: [ElementList$(Element), [core.String]],
      getElementById: [Element, [core.String]],
      [_getElementById_1]: [Element, [dart.dynamic]],
      querySelector: [Element, [core.String]],
      [_querySelector_1]: [Element, [dart.dynamic]],
      [_querySelectorAll]: [NodeList, [core.String]],
      [_querySelectorAll_1]: [NodeList, [dart.dynamic]]
    }),
    statics: () => ({internalCreateDocumentFragment: [DocumentFragment, []]}),
    names: ['internalCreateDocumentFragment']
  });
  DocumentFragment[dart.metadata] = () => [dart.const(new _metadata.DomName('DocumentFragment')), dart.const(new _js_helper.Native("DocumentFragment"))];
  const _createDocument_1 = Symbol('_createDocument_1');
  const _createDocumentType_1 = Symbol('_createDocumentType_1');
  const _createHtmlDocument_1 = Symbol('_createHtmlDocument_1');
  const _hasFeature_1 = Symbol('_hasFeature_1');
  class DomImplementation extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateDomImplementation() {
      return DomImplementation._internalWrap();
    }
    static _internalWrap() {
      return new DomImplementation.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    createDocument(namespaceURI, qualifiedName, doctype) {
      return this[_createDocument_1](namespaceURI, qualifiedName, doctype);
    }
    [_createDocument_1](namespaceURI, qualifiedName, doctype) {
      return dart.as(wrap_jso(this.raw.createDocument(unwrap_jso(namespaceURI), unwrap_jso(qualifiedName), unwrap_jso(doctype))), XmlDocument);
    }
    createDocumentType(qualifiedName, publicId, systemId) {
      return this[_createDocumentType_1](qualifiedName, publicId, systemId);
    }
    [_createDocumentType_1](qualifiedName, publicId, systemId) {
      return dart.as(wrap_jso(this.raw.createDocumentType(unwrap_jso(qualifiedName), unwrap_jso(publicId), unwrap_jso(systemId))), _DocumentType);
    }
    createHtmlDocument(title) {
      return this[_createHtmlDocument_1](title);
    }
    [_createHtmlDocument_1](title) {
      return dart.as(wrap_jso(this.raw.createHTMLDocument(unwrap_jso(title))), HtmlDocument);
    }
    hasFeature(feature, version) {
      return this[_hasFeature_1](feature, version);
    }
    [_hasFeature_1](feature, version) {
      return dart.as(wrap_jso(this.raw.hasFeature(unwrap_jso(feature), unwrap_jso(version))), core.bool);
    }
  }
  dart.defineNamedConstructor(DomImplementation, 'internal_');
  dart.setSignature(DomImplementation, {
    constructors: () => ({
      _: [DomImplementation, []],
      _internalWrap: [DomImplementation, []],
      internal_: [DomImplementation, []]
    }),
    methods: () => ({
      createDocument: [XmlDocument, [core.String, core.String, _DocumentType]],
      [_createDocument_1]: [XmlDocument, [dart.dynamic, dart.dynamic, _DocumentType]],
      createDocumentType: [_DocumentType, [core.String, core.String, core.String]],
      [_createDocumentType_1]: [_DocumentType, [dart.dynamic, dart.dynamic, dart.dynamic]],
      createHtmlDocument: [HtmlDocument, [core.String]],
      [_createHtmlDocument_1]: [HtmlDocument, [dart.dynamic]],
      hasFeature: [core.bool, [core.String, core.String]],
      [_hasFeature_1]: [core.bool, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateDomImplementation: [DomImplementation, []]}),
    names: ['internalCreateDomImplementation']
  });
  DomImplementation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DOMImplementation')), dart.const(new _js_helper.Native("DOMImplementation"))];
  const _add_1 = Symbol('_add_1');
  const _remove_1 = Symbol('_remove_1');
  const _toggle_1 = Symbol('_toggle_1');
  const _toggle_2 = Symbol('_toggle_2');
  class DomTokenList extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateDomTokenList() {
      return DomTokenList._internalWrap();
    }
    static _internalWrap() {
      return new DomTokenList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    add(tokens) {
      this[_add_1](tokens);
      return;
    }
    [_add_1](tokens) {
      return wrap_jso(this.raw.add(unwrap_jso(tokens)));
    }
    contains(token) {
      return this[_contains_1](token);
    }
    [_contains_1](token) {
      return dart.as(wrap_jso(this.raw.contains(unwrap_jso(token))), core.bool);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), core.String);
    }
    remove(tokens) {
      this[_remove_1](tokens);
      return;
    }
    [_remove_1](tokens) {
      return wrap_jso(this.raw.remove(unwrap_jso(tokens)));
    }
    toggle(token, force) {
      if (force === void 0)
        force = null;
      if (force != null) {
        return this[_toggle_1](token, force);
      }
      return this[_toggle_2](token);
    }
    [_toggle_1](token, force) {
      return dart.as(wrap_jso(this.raw.toggle(unwrap_jso(token), unwrap_jso(force))), core.bool);
    }
    [_toggle_2](token) {
      return dart.as(wrap_jso(this.raw.toggle(unwrap_jso(token))), core.bool);
    }
  }
  dart.defineNamedConstructor(DomTokenList, 'internal_');
  dart.setSignature(DomTokenList, {
    constructors: () => ({
      _: [DomTokenList, []],
      _internalWrap: [DomTokenList, []],
      internal_: [DomTokenList, []]
    }),
    methods: () => ({
      add: [dart.void, [core.String]],
      [_add_1]: [dart.void, [dart.dynamic]],
      contains: [core.bool, [core.String]],
      [_contains_1]: [core.bool, [dart.dynamic]],
      item: [core.String, [core.int]],
      [_item_1]: [core.String, [dart.dynamic]],
      remove: [dart.void, [core.String]],
      [_remove_1]: [dart.void, [dart.dynamic]],
      toggle: [core.bool, [core.String], [core.bool]],
      [_toggle_1]: [core.bool, [dart.dynamic, dart.dynamic]],
      [_toggle_2]: [core.bool, [dart.dynamic]]
    }),
    statics: () => ({internalCreateDomTokenList: [DomTokenList, []]}),
    names: ['internalCreateDomTokenList']
  });
  DomTokenList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DOMTokenList')), dart.const(new _js_helper.Native("DOMTokenList"))];
  const _childElements = Symbol('_childElements');
  const _element = Symbol('_element');
  const _filter = Symbol('_filter');
  class _ChildrenElementList extends collection.ListBase$(Element) {
    _wrap(element) {
      this[_childElements] = dart.as(element[_children], HtmlCollection);
      this[_element] = element;
    }
    contains(element) {
      return this[_childElements].contains(element);
    }
    get isEmpty() {
      return this[_element][_firstElementChild] == null;
    }
    get length() {
      return this[_childElements].length;
    }
    get(index) {
      return dart.as(this[_childElements].get(index), Element);
    }
    set(index, value) {
      this[_element][_replaceChild](value, this[_childElements].get(index));
      return value;
    }
    set length(newLength) {
      dart.throw(new core.UnsupportedError('Cannot resize element lists'));
    }
    add(value) {
      this[_element].append(value);
      return value;
    }
    get iterator() {
      return this.toList()[dartx.iterator];
    }
    addAll(iterable) {
      if (dart.is(iterable, _ChildNodeListLazy)) {
        iterable = core.List$(Element).from(iterable);
      }
      for (let element of iterable) {
        this[_element].append(element);
      }
    }
    sort(compare) {
      if (compare === void 0)
        compare = null;
      dart.throw(new core.UnsupportedError('Cannot sort element lists'));
    }
    shuffle(random) {
      if (random === void 0)
        random = null;
      dart.throw(new core.UnsupportedError('Cannot shuffle element lists'));
    }
    removeWhere(test) {
      this[_filter](test, false);
    }
    retainWhere(test) {
      this[_filter](test, true);
    }
    [_filter](test, retainMatching) {
      let removed = null;
      if (dart.notNull(retainMatching)) {
        removed = this[_element].children[dartx.where](dart.fn(e => !dart.notNull(dart.dcall(test, e)), core.bool, [dart.dynamic]));
      } else {
        removed = this[_element].children[dartx.where](dart.as(test, __CastType0));
      }
      for (let e of dart.as(removed, core.Iterable))
        dart.dsend(e, 'remove');
    }
    setRange(start, end, iterable, skipCount) {
      if (skipCount === void 0)
        skipCount = 0;
      dart.throw(new core.UnimplementedError());
    }
    replaceRange(start, end, iterable) {
      dart.throw(new core.UnimplementedError());
    }
    fillRange(start, end, fillValue) {
      if (fillValue === void 0)
        fillValue = null;
      dart.throw(new core.UnimplementedError());
    }
    remove(object) {
      if (dart.is(object, Element)) {
        let element = object;
        if (dart.equals(element.parentNode, this[_element])) {
          this[_element][_removeChild](element);
          return true;
        }
      }
      return false;
    }
    insert(index, element) {
      if (dart.notNull(index) < 0 || dart.notNull(index) > dart.notNull(this.length)) {
        dart.throw(new core.RangeError.range(index, 0, this.length));
      }
      if (index == this.length) {
        this[_element].append(element);
      } else {
        this[_element].insertBefore(element, this.get(index));
      }
    }
    setAll(index, iterable) {
      dart.throw(new core.UnimplementedError());
    }
    clear() {
      this[_element][_clearChildren]();
    }
    removeAt(index) {
      let result = this.get(index);
      if (result != null) {
        this[_element][_removeChild](result);
      }
      return result;
    }
    removeLast() {
      let result = this.last;
      if (result != null) {
        this[_element][_removeChild](result);
      }
      return result;
    }
    get first() {
      let result = this[_element][_firstElementChild];
      if (result == null)
        dart.throw(new core.StateError("No elements"));
      return result;
    }
    get last() {
      let result = this[_element][_lastElementChild];
      if (result == null)
        dart.throw(new core.StateError("No elements"));
      return result;
    }
    get single() {
      if (dart.notNull(this.length) > 1)
        dart.throw(new core.StateError("More than one element"));
      return this.first;
    }
    get rawList() {
      return this[_childElements];
    }
  }
  _ChildrenElementList[dart.implements] = () => [html_common.NodeListWrapper];
  dart.defineNamedConstructor(_ChildrenElementList, '_wrap');
  dart.setSignature(_ChildrenElementList, {
    constructors: () => ({_wrap: [_ChildrenElementList, [Element]]}),
    methods: () => ({
      get: [Element, [core.int]],
      set: [dart.void, [core.int, Element]],
      add: [Element, [Element]],
      addAll: [dart.void, [core.Iterable$(Element)]],
      sort: [dart.void, [], [dart.functionType(core.int, [Element, Element])]],
      removeWhere: [dart.void, [dart.functionType(core.bool, [Element])]],
      retainWhere: [dart.void, [dart.functionType(core.bool, [Element])]],
      [_filter]: [dart.void, [dart.functionType(core.bool, [dart.dynamic]), core.bool]],
      setRange: [dart.void, [core.int, core.int, core.Iterable$(Element)], [core.int]],
      replaceRange: [dart.void, [core.int, core.int, core.Iterable$(Element)]],
      fillRange: [dart.void, [core.int, core.int], [Element]],
      insert: [dart.void, [core.int, Element]],
      setAll: [dart.void, [core.int, core.Iterable$(Element)]],
      removeAt: [Element, [core.int]],
      removeLast: [Element, []]
    })
  });
  dart.defineExtensionMembers(_ChildrenElementList, [
    'contains',
    'get',
    'set',
    'add',
    'addAll',
    'sort',
    'shuffle',
    'removeWhere',
    'retainWhere',
    'setRange',
    'replaceRange',
    'fillRange',
    'remove',
    'insert',
    'setAll',
    'clear',
    'removeAt',
    'removeLast',
    'isEmpty',
    'length',
    'length',
    'iterator',
    'first',
    'last',
    'single'
  ]);
  const ElementList$ = dart.generic(function(T) {
    class ElementList extends collection.ListBase$(T) {}
    return ElementList;
  });
  let ElementList = ElementList$();
  const _nodeList = Symbol('_nodeList');
  class _FrozenElementList extends collection.ListBase {
    _wrap(nodeList) {
      this[_nodeList] = nodeList;
      this.dartClass_instance = null;
      this.dartClass_instance = this[_nodeList];
    }
    get length() {
      return this[_nodeList][dartx.length];
    }
    get(index) {
      return dart.as(this[_nodeList][dartx.get](index), Element);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError('Cannot modify list'));
      return value;
    }
    set length(newLength) {
      dart.throw(new core.UnsupportedError('Cannot modify list'));
    }
    sort(compare) {
      if (compare === void 0)
        compare = null;
      dart.throw(new core.UnsupportedError('Cannot sort list'));
    }
    shuffle(random) {
      if (random === void 0)
        random = null;
      dart.throw(new core.UnsupportedError('Cannot shuffle list'));
    }
    get first() {
      return dart.as(this[_nodeList][dartx.first], Element);
    }
    get last() {
      return dart.as(this[_nodeList][dartx.last], Element);
    }
    get single() {
      return dart.as(this[_nodeList][dartx.single], Element);
    }
    get classes() {
      return exports._MultiElementCssClassSet.new(dart.as(this, core.Iterable$(Element)));
    }
    get style() {
      return new _CssStyleDeclarationSet(dart.as(this, core.Iterable$(Element)));
    }
    set classes(value) {
      this[_nodeList][dartx.forEach](dart.fn(e => dart.dput(e, 'classes', value), core.Iterable$(core.String), [dart.dynamic]));
    }
    get contentEdge() {
      return new _ContentCssListRect(this);
    }
    get paddingEdge() {
      return this.first.paddingEdge;
    }
    get borderEdge() {
      return this.first.borderEdge;
    }
    get marginEdge() {
      return this.first.marginEdge;
    }
    get rawList() {
      return this[_nodeList];
    }
    get onAbort() {
      return Element.abortEvent[_forElementList](this);
    }
    get onBeforeCopy() {
      return Element.beforeCopyEvent[_forElementList](this);
    }
    get onBeforeCut() {
      return Element.beforeCutEvent[_forElementList](this);
    }
    get onBeforePaste() {
      return Element.beforePasteEvent[_forElementList](this);
    }
    get onBlur() {
      return Element.blurEvent[_forElementList](this);
    }
    get onCanPlay() {
      return Element.canPlayEvent[_forElementList](this);
    }
    get onCanPlayThrough() {
      return Element.canPlayThroughEvent[_forElementList](this);
    }
    get onChange() {
      return Element.changeEvent[_forElementList](this);
    }
    get onClick() {
      return Element.clickEvent[_forElementList](this);
    }
    get onContextMenu() {
      return Element.contextMenuEvent[_forElementList](this);
    }
    get onCopy() {
      return Element.copyEvent[_forElementList](this);
    }
    get onCut() {
      return Element.cutEvent[_forElementList](this);
    }
    get onDoubleClick() {
      return Element.doubleClickEvent[_forElementList](this);
    }
    get onDrag() {
      return Element.dragEvent[_forElementList](this);
    }
    get onDragEnd() {
      return Element.dragEndEvent[_forElementList](this);
    }
    get onDragEnter() {
      return Element.dragEnterEvent[_forElementList](this);
    }
    get onDragLeave() {
      return Element.dragLeaveEvent[_forElementList](this);
    }
    get onDragOver() {
      return Element.dragOverEvent[_forElementList](this);
    }
    get onDragStart() {
      return Element.dragStartEvent[_forElementList](this);
    }
    get onDrop() {
      return Element.dropEvent[_forElementList](this);
    }
    get onDurationChange() {
      return Element.durationChangeEvent[_forElementList](this);
    }
    get onEmptied() {
      return Element.emptiedEvent[_forElementList](this);
    }
    get onEnded() {
      return Element.endedEvent[_forElementList](this);
    }
    get onError() {
      return Element.errorEvent[_forElementList](this);
    }
    get onFocus() {
      return Element.focusEvent[_forElementList](this);
    }
    get onInput() {
      return Element.inputEvent[_forElementList](this);
    }
    get onInvalid() {
      return Element.invalidEvent[_forElementList](this);
    }
    get onKeyDown() {
      return Element.keyDownEvent[_forElementList](this);
    }
    get onKeyPress() {
      return Element.keyPressEvent[_forElementList](this);
    }
    get onKeyUp() {
      return Element.keyUpEvent[_forElementList](this);
    }
    get onLoad() {
      return Element.loadEvent[_forElementList](this);
    }
    get onLoadedData() {
      return Element.loadedDataEvent[_forElementList](this);
    }
    get onLoadedMetadata() {
      return Element.loadedMetadataEvent[_forElementList](this);
    }
    get onMouseDown() {
      return Element.mouseDownEvent[_forElementList](this);
    }
    get onMouseEnter() {
      return Element.mouseEnterEvent[_forElementList](this);
    }
    get onMouseLeave() {
      return Element.mouseLeaveEvent[_forElementList](this);
    }
    get onMouseMove() {
      return Element.mouseMoveEvent[_forElementList](this);
    }
    get onMouseOut() {
      return Element.mouseOutEvent[_forElementList](this);
    }
    get onMouseOver() {
      return Element.mouseOverEvent[_forElementList](this);
    }
    get onMouseUp() {
      return Element.mouseUpEvent[_forElementList](this);
    }
    get onMouseWheel() {
      return dart.as(dart.dsend(dart.dload(Element, 'mouseWheelEvent'), _forElementList, this), ElementStream$(dart.dynamic));
    }
    get onPaste() {
      return Element.pasteEvent[_forElementList](this);
    }
    get onPause() {
      return Element.pauseEvent[_forElementList](this);
    }
    get onPlay() {
      return Element.playEvent[_forElementList](this);
    }
    get onPlaying() {
      return Element.playingEvent[_forElementList](this);
    }
    get onRateChange() {
      return Element.rateChangeEvent[_forElementList](this);
    }
    get onReset() {
      return Element.resetEvent[_forElementList](this);
    }
    get onResize() {
      return Element.resizeEvent[_forElementList](this);
    }
    get onScroll() {
      return Element.scrollEvent[_forElementList](this);
    }
    get onSearch() {
      return Element.searchEvent[_forElementList](this);
    }
    get onSeeked() {
      return Element.seekedEvent[_forElementList](this);
    }
    get onSeeking() {
      return Element.seekingEvent[_forElementList](this);
    }
    get onSelect() {
      return Element.selectEvent[_forElementList](this);
    }
    get onSelectStart() {
      return Element.selectStartEvent[_forElementList](this);
    }
    get onStalled() {
      return Element.stalledEvent[_forElementList](this);
    }
    get onSubmit() {
      return Element.submitEvent[_forElementList](this);
    }
    get onSuspend() {
      return Element.suspendEvent[_forElementList](this);
    }
    get onTimeUpdate() {
      return Element.timeUpdateEvent[_forElementList](this);
    }
    get onTouchCancel() {
      return Element.touchCancelEvent[_forElementList](this);
    }
    get onTouchEnd() {
      return Element.touchEndEvent[_forElementList](this);
    }
    get onTouchEnter() {
      return Element.touchEnterEvent[_forElementList](this);
    }
    get onTouchLeave() {
      return Element.touchLeaveEvent[_forElementList](this);
    }
    get onTouchMove() {
      return Element.touchMoveEvent[_forElementList](this);
    }
    get onTouchStart() {
      return Element.touchStartEvent[_forElementList](this);
    }
    get onTransitionEnd() {
      return dart.as(dart.dsend(dart.dload(Element, 'transitionEndEvent'), _forElementList, this), ElementStream$(TransitionEvent));
    }
    get onVolumeChange() {
      return Element.volumeChangeEvent[_forElementList](this);
    }
    get onWaiting() {
      return Element.waitingEvent[_forElementList](this);
    }
    get onFullscreenChange() {
      return Element.fullscreenChangeEvent[_forElementList](this);
    }
    get onFullscreenError() {
      return Element.fullscreenErrorEvent[_forElementList](this);
    }
  }
  _FrozenElementList[dart.implements] = () => [ElementList, html_common.NodeListWrapper];
  dart.defineNamedConstructor(_FrozenElementList, '_wrap');
  dart.setSignature(_FrozenElementList, {
    constructors: () => ({_wrap: [_FrozenElementList, [core.List$(Node)]]}),
    methods: () => ({
      get: [Element, [core.int]],
      set: [dart.void, [core.int, Element]],
      sort: [dart.void, [], [core.Comparator$(Element)]]
    })
  });
  dart.defineExtensionMembers(_FrozenElementList, [
    'get',
    'set',
    'sort',
    'shuffle',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  Element._parseDocument = null;
  Element._parseRange = null;
  Element._defaultValidator = null;
  Element._defaultSanitizer = null;
  Element._tagsForWhichCreateContextualFragmentIsNotSupported = dart.const(['HEAD', 'AREA', 'BASE', 'BASEFONT', 'BR', 'COL', 'COLGROUP', 'EMBED', 'FRAME', 'FRAMESET', 'HR', 'IMAGE', 'IMG', 'INPUT', 'ISINDEX', 'LINK', 'META', 'PARAM', 'SOURCE', 'STYLE', 'TITLE', 'WBR']);
  Element.abortEvent = dart.const(new (EventStreamProvider$(Event))('abort'));
  Element.beforeCopyEvent = dart.const(new (EventStreamProvider$(Event))('beforecopy'));
  Element.beforeCutEvent = dart.const(new (EventStreamProvider$(Event))('beforecut'));
  Element.beforePasteEvent = dart.const(new (EventStreamProvider$(Event))('beforepaste'));
  Element.blurEvent = dart.const(new (EventStreamProvider$(Event))('blur'));
  Element.canPlayEvent = dart.const(new (EventStreamProvider$(Event))('canplay'));
  Element.canPlayThroughEvent = dart.const(new (EventStreamProvider$(Event))('canplaythrough'));
  Element.changeEvent = dart.const(new (EventStreamProvider$(Event))('change'));
  dart.defineLazyProperties(Element, {
    get _namedNodeMap() {
      return dart.dindex(dart.dload(/* Unimplemented unknown name */js, 'context'), "NamedNodeMap");
    },
    set _namedNodeMap(_) {},
    get _htmlCollection() {
      return dart.dindex(dart.dload(/* Unimplemented unknown name */js, 'context'), "HTMLCollection");
    },
    set _htmlCollection(_) {},
    get _nodeList() {
      return dart.dindex(dart.dload(/* Unimplemented unknown name */js, 'context'), "NodeList");
    },
    set _nodeList(_) {}
  });
  const _initUIEvent = Symbol('_initUIEvent');
  const _charCode = Symbol('_charCode');
  const _keyCode = Symbol('_keyCode');
  const _layerX = Symbol('_layerX');
  const _layerY = Symbol('_layerY');
  const _pageX = Symbol('_pageX');
  const _pageY = Symbol('_pageY');
  const _get_view = Symbol('_get_view');
  const _initUIEvent_1 = Symbol('_initUIEvent_1');
  class UIEvent extends Event {
    static new(type, opts) {
      let view = opts && 'view' in opts ? opts.view : null;
      let detail = opts && 'detail' in opts ? opts.detail : 0;
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      if (view == null) {
        view = exports.window;
      }
      let e = exports.document[_createEvent]("UIEvent");
      dart.dcall(e[_initUIEvent], type, canBubble, cancelable, view, detail);
      return dart.as(e, UIEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateUIEvent() {
      return UIEvent._internalWrap();
    }
    static _internalWrap() {
      return new UIEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get [_charCode]() {
      return dart.as(wrap_jso(this.raw.charCode), core.int);
    }
    get detail() {
      return dart.as(wrap_jso(this.raw.detail), core.int);
    }
    get [_keyCode]() {
      return dart.as(wrap_jso(this.raw.keyCode), core.int);
    }
    get [_layerX]() {
      return dart.as(wrap_jso(this.raw.layerX), core.int);
    }
    get [_layerY]() {
      return dart.as(wrap_jso(this.raw.layerY), core.int);
    }
    get [_pageX]() {
      return dart.as(wrap_jso(this.raw.pageX), core.int);
    }
    get [_pageY]() {
      return dart.as(wrap_jso(this.raw.pageY), core.int);
    }
    get view() {
      return _convertNativeToDart_Window(this[_get_view]);
    }
    get [_get_view]() {
      return wrap_jso(this.raw.view);
    }
    get which() {
      return dart.as(wrap_jso(this.raw.which), core.int);
    }
    [_initUIEvent](type, canBubble, cancelable, view, detail) {
      this[_initUIEvent_1](type, canBubble, cancelable, view, detail);
      return;
    }
    [_initUIEvent_1](type, canBubble, cancelable, view, detail) {
      return wrap_jso(this.raw.initUIEvent(unwrap_jso(type), unwrap_jso(canBubble), unwrap_jso(cancelable), unwrap_jso(view), unwrap_jso(detail)));
    }
    get layer() {
      return new math.Point(this[_layerX], this[_layerY]);
    }
    get page() {
      return new math.Point(this[_pageX], this[_pageY]);
    }
  }
  dart.defineNamedConstructor(UIEvent, 'internal_');
  dart.setSignature(UIEvent, {
    constructors: () => ({
      new: [UIEvent, [core.String], {view: Window, detail: core.int, canBubble: core.bool, cancelable: core.bool}],
      _: [UIEvent, []],
      _internalWrap: [UIEvent, []],
      internal_: [UIEvent, []]
    }),
    methods: () => ({
      [_initUIEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.int]],
      [_initUIEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, Window, dart.dynamic]]
    }),
    statics: () => ({internalCreateUIEvent: [UIEvent, []]}),
    names: ['internalCreateUIEvent']
  });
  UIEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('UIEvent')), dart.const(new _js_helper.Native("UIEvent"))];
  const _clientX = Symbol('_clientX');
  const _clientY = Symbol('_clientY');
  const _movementX = Symbol('_movementX');
  const _movementY = Symbol('_movementY');
  const _get_relatedTarget = Symbol('_get_relatedTarget');
  const _screenX = Symbol('_screenX');
  const _screenY = Symbol('_screenY');
  const _webkitMovementX = Symbol('_webkitMovementX');
  const _webkitMovementY = Symbol('_webkitMovementY');
  const _initMouseEvent_1 = Symbol('_initMouseEvent_1');
  const _initMouseEvent = Symbol('_initMouseEvent');
  class MouseEvent extends UIEvent {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMouseEvent() {
      return MouseEvent._internalWrap();
    }
    static _internalWrap() {
      return new MouseEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get altKey() {
      return dart.as(wrap_jso(this.raw.altKey), core.bool);
    }
    get button() {
      return dart.as(wrap_jso(this.raw.button), core.int);
    }
    get [_clientX]() {
      return dart.as(wrap_jso(this.raw.clientX), core.int);
    }
    get [_clientY]() {
      return dart.as(wrap_jso(this.raw.clientY), core.int);
    }
    get ctrlKey() {
      return dart.as(wrap_jso(this.raw.ctrlKey), core.bool);
    }
    get dataTransfer() {
      return wrap_jso(this.raw.dataTransfer);
    }
    get fromElement() {
      return dart.as(wrap_jso(this.raw.fromElement), Node);
    }
    get metaKey() {
      return dart.as(wrap_jso(this.raw.metaKey), core.bool);
    }
    get [_movementX]() {
      return dart.as(wrap_jso(this.raw.movementX), core.int);
    }
    get [_movementY]() {
      return dart.as(wrap_jso(this.raw.movementY), core.int);
    }
    get region() {
      return dart.as(wrap_jso(this.raw.region), core.String);
    }
    get relatedTarget() {
      return _convertNativeToDart_EventTarget(this[_get_relatedTarget]);
    }
    get [_get_relatedTarget]() {
      return wrap_jso(this.raw.relatedTarget);
    }
    get [_screenX]() {
      return dart.as(wrap_jso(this.raw.screenX), core.int);
    }
    get [_screenY]() {
      return dart.as(wrap_jso(this.raw.screenY), core.int);
    }
    get shiftKey() {
      return dart.as(wrap_jso(this.raw.shiftKey), core.bool);
    }
    get toElement() {
      return dart.as(wrap_jso(this.raw.toElement), Node);
    }
    get [_webkitMovementX]() {
      return dart.as(wrap_jso(this.raw.webkitMovementX), core.int);
    }
    get [_webkitMovementY]() {
      return dart.as(wrap_jso(this.raw.webkitMovementY), core.int);
    }
    [_initMouseEvent](type, canBubble, cancelable, view, detail, screenX, screenY, clientX, clientY, ctrlKey, altKey, shiftKey, metaKey, button, relatedTarget) {
      let relatedTarget_1 = _convertDartToNative_EventTarget(relatedTarget);
      this[_initMouseEvent_1](type, canBubble, cancelable, view, detail, screenX, screenY, clientX, clientY, ctrlKey, altKey, shiftKey, metaKey, button, relatedTarget_1);
      return;
    }
    [_initMouseEvent_1](type, canBubble, cancelable, view, detail, screenX, screenY, clientX, clientY, ctrlKey, altKey, shiftKey, metaKey, button, relatedTarget) {
      return wrap_jso(this.raw.initMouseEvent(unwrap_jso(type), unwrap_jso(canBubble), unwrap_jso(cancelable), unwrap_jso(view), unwrap_jso(detail), unwrap_jso(screenX), unwrap_jso(screenY), unwrap_jso(clientX), unwrap_jso(clientY), unwrap_jso(ctrlKey), unwrap_jso(altKey), unwrap_jso(shiftKey), unwrap_jso(metaKey), unwrap_jso(button), unwrap_jso(relatedTarget)));
    }
  }
  dart.defineNamedConstructor(MouseEvent, 'internal_');
  dart.setSignature(MouseEvent, {
    constructors: () => ({
      _: [MouseEvent, []],
      _internalWrap: [MouseEvent, []],
      internal_: [MouseEvent, []]
    }),
    methods: () => ({
      [_initMouseEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.int, core.int, core.int, core.int, core.int, core.bool, core.bool, core.bool, core.bool, core.int, EventTarget]],
      [_initMouseEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, Window, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateMouseEvent: [MouseEvent, []]}),
    names: ['internalCreateMouseEvent']
  });
  MouseEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MouseEvent')), dart.const(new _js_helper.Native("MouseEvent,DragEvent,PointerEvent,MSPointerEvent"))];
  Element.clickEvent = dart.const(new (EventStreamProvider$(MouseEvent))('click'));
  Element.contextMenuEvent = dart.const(new (EventStreamProvider$(MouseEvent))('contextmenu'));
  Element.copyEvent = dart.const(new (EventStreamProvider$(Event))('copy'));
  Element.cutEvent = dart.const(new (EventStreamProvider$(Event))('cut'));
  Element.doubleClickEvent = dart.const(new (EventStreamProvider$(Event))('dblclick'));
  Element.dragEvent = dart.const(new (EventStreamProvider$(MouseEvent))('drag'));
  Element.dragEndEvent = dart.const(new (EventStreamProvider$(MouseEvent))('dragend'));
  Element.dragEnterEvent = dart.const(new (EventStreamProvider$(MouseEvent))('dragenter'));
  Element.dragLeaveEvent = dart.const(new (EventStreamProvider$(MouseEvent))('dragleave'));
  Element.dragOverEvent = dart.const(new (EventStreamProvider$(MouseEvent))('dragover'));
  Element.dragStartEvent = dart.const(new (EventStreamProvider$(MouseEvent))('dragstart'));
  Element.dropEvent = dart.const(new (EventStreamProvider$(MouseEvent))('drop'));
  Element.durationChangeEvent = dart.const(new (EventStreamProvider$(Event))('durationchange'));
  Element.emptiedEvent = dart.const(new (EventStreamProvider$(Event))('emptied'));
  Element.endedEvent = dart.const(new (EventStreamProvider$(Event))('ended'));
  Element.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  Element.focusEvent = dart.const(new (EventStreamProvider$(Event))('focus'));
  Element.inputEvent = dart.const(new (EventStreamProvider$(Event))('input'));
  Element.invalidEvent = dart.const(new (EventStreamProvider$(Event))('invalid'));
  const _initKeyboardEvent = Symbol('_initKeyboardEvent');
  const _keyIdentifier = Symbol('_keyIdentifier');
  const _getModifierState_1 = Symbol('_getModifierState_1');
  class KeyboardEvent extends UIEvent {
    static new(type, opts) {
      let view = opts && 'view' in opts ? opts.view : null;
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      let keyLocation = opts && 'keyLocation' in opts ? opts.keyLocation : 1;
      let ctrlKey = opts && 'ctrlKey' in opts ? opts.ctrlKey : false;
      let altKey = opts && 'altKey' in opts ? opts.altKey : false;
      let shiftKey = opts && 'shiftKey' in opts ? opts.shiftKey : false;
      let metaKey = opts && 'metaKey' in opts ? opts.metaKey : false;
      if (view == null) {
        view = exports.window;
      }
      let e = exports.document[_createEvent]("KeyboardEvent");
      dart.dcall(e[_initKeyboardEvent], type, canBubble, cancelable, view, "", keyLocation, ctrlKey, altKey, shiftKey, metaKey);
      return dart.as(e, KeyboardEvent);
    }
    [_initKeyboardEvent](type, canBubble, cancelable, view, keyIdentifier, keyLocation, ctrlKey, altKey, shiftKey, metaKey) {
      if (typeof this.raw.initKeyEvent == "function") {
        this.raw.initKeyEvent(type, canBubble, cancelable, unwrap_jso(view), ctrlKey, altKey, shiftKey, metaKey, 0, 0);
      } else {
        this.raw.initKeyboardEvent(type, canBubble, cancelable, unwrap_jso(view), keyIdentifier, keyLocation, ctrlKey, altKey, shiftKey, metaKey);
      }
    }
    get keyCode() {
      return this[_keyCode];
    }
    get charCode() {
      return this[_charCode];
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateKeyboardEvent() {
      return KeyboardEvent._internalWrap();
    }
    static _internalWrap() {
      return new KeyboardEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get altKey() {
      return dart.as(wrap_jso(this.raw.altKey), core.bool);
    }
    get ctrlKey() {
      return dart.as(wrap_jso(this.raw.ctrlKey), core.bool);
    }
    get [_keyIdentifier]() {
      return dart.as(wrap_jso(this.raw.keyIdentifier), core.String);
    }
    get keyLocation() {
      return dart.as(wrap_jso(this.raw.keyLocation), core.int);
    }
    get location() {
      return dart.as(wrap_jso(this.raw.location), core.int);
    }
    get metaKey() {
      return dart.as(wrap_jso(this.raw.metaKey), core.bool);
    }
    get repeat() {
      return dart.as(wrap_jso(this.raw.repeat), core.bool);
    }
    get shiftKey() {
      return dart.as(wrap_jso(this.raw.shiftKey), core.bool);
    }
    getModifierState(keyArgument) {
      return this[_getModifierState_1](keyArgument);
    }
    [_getModifierState_1](keyArgument) {
      return dart.as(wrap_jso(this.raw.getModifierState(unwrap_jso(keyArgument))), core.bool);
    }
  }
  dart.defineNamedConstructor(KeyboardEvent, 'internal_');
  dart.setSignature(KeyboardEvent, {
    constructors: () => ({
      new: [KeyboardEvent, [core.String], {view: Window, canBubble: core.bool, cancelable: core.bool, keyLocation: core.int, ctrlKey: core.bool, altKey: core.bool, shiftKey: core.bool, metaKey: core.bool}],
      _: [KeyboardEvent, []],
      _internalWrap: [KeyboardEvent, []],
      internal_: [KeyboardEvent, []]
    }),
    methods: () => ({
      [_initKeyboardEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.String, core.int, core.bool, core.bool, core.bool, core.bool]],
      getModifierState: [core.bool, [core.String]],
      [_getModifierState_1]: [core.bool, [dart.dynamic]]
    }),
    statics: () => ({internalCreateKeyboardEvent: [KeyboardEvent, []]}),
    names: ['internalCreateKeyboardEvent']
  });
  KeyboardEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('KeyboardEvent')), dart.const(new _js_helper.Native("KeyboardEvent"))];
  Element.keyDownEvent = dart.const(new (EventStreamProvider$(KeyboardEvent))('keydown'));
  Element.keyPressEvent = dart.const(new (EventStreamProvider$(KeyboardEvent))('keypress'));
  Element.keyUpEvent = dart.const(new (EventStreamProvider$(KeyboardEvent))('keyup'));
  Element.loadEvent = dart.const(new (EventStreamProvider$(Event))('load'));
  Element.loadedDataEvent = dart.const(new (EventStreamProvider$(Event))('loadeddata'));
  Element.loadedMetadataEvent = dart.const(new (EventStreamProvider$(Event))('loadedmetadata'));
  Element.mouseDownEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mousedown'));
  Element.mouseEnterEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mouseenter'));
  Element.mouseLeaveEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mouseleave'));
  Element.mouseMoveEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mousemove'));
  Element.mouseOutEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mouseout'));
  Element.mouseOverEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mouseover'));
  Element.mouseUpEvent = dart.const(new (EventStreamProvider$(MouseEvent))('mouseup'));
  Element.pasteEvent = dart.const(new (EventStreamProvider$(Event))('paste'));
  Element.pauseEvent = dart.const(new (EventStreamProvider$(Event))('pause'));
  Element.playEvent = dart.const(new (EventStreamProvider$(Event))('play'));
  Element.playingEvent = dart.const(new (EventStreamProvider$(Event))('playing'));
  Element.rateChangeEvent = dart.const(new (EventStreamProvider$(Event))('ratechange'));
  Element.resetEvent = dart.const(new (EventStreamProvider$(Event))('reset'));
  Element.resizeEvent = dart.const(new (EventStreamProvider$(Event))('resize'));
  Element.scrollEvent = dart.const(new (EventStreamProvider$(Event))('scroll'));
  Element.searchEvent = dart.const(new (EventStreamProvider$(Event))('search'));
  Element.seekedEvent = dart.const(new (EventStreamProvider$(Event))('seeked'));
  Element.seekingEvent = dart.const(new (EventStreamProvider$(Event))('seeking'));
  Element.selectEvent = dart.const(new (EventStreamProvider$(Event))('select'));
  Element.selectStartEvent = dart.const(new (EventStreamProvider$(Event))('selectstart'));
  Element.stalledEvent = dart.const(new (EventStreamProvider$(Event))('stalled'));
  Element.submitEvent = dart.const(new (EventStreamProvider$(Event))('submit'));
  Element.suspendEvent = dart.const(new (EventStreamProvider$(Event))('suspend'));
  Element.timeUpdateEvent = dart.const(new (EventStreamProvider$(Event))('timeupdate'));
  const _initTouchEvent = Symbol('_initTouchEvent');
  const _initTouchEvent_1 = Symbol('_initTouchEvent_1');
  class TouchEvent extends UIEvent {
    static new(touches, targetTouches, changedTouches, type, opts) {
      let view = opts && 'view' in opts ? opts.view : null;
      let screenX = opts && 'screenX' in opts ? opts.screenX : 0;
      let screenY = opts && 'screenY' in opts ? opts.screenY : 0;
      let clientX = opts && 'clientX' in opts ? opts.clientX : 0;
      let clientY = opts && 'clientY' in opts ? opts.clientY : 0;
      let ctrlKey = opts && 'ctrlKey' in opts ? opts.ctrlKey : false;
      let altKey = opts && 'altKey' in opts ? opts.altKey : false;
      let shiftKey = opts && 'shiftKey' in opts ? opts.shiftKey : false;
      let metaKey = opts && 'metaKey' in opts ? opts.metaKey : false;
      if (view == null) {
        view = exports.window;
      }
      let e = exports.document[_createEvent]("TouchEvent");
      dart.dcall(e[_initTouchEvent], touches, targetTouches, changedTouches, type, view, screenX, screenY, clientX, clientY, ctrlKey, altKey, shiftKey, metaKey);
      return dart.as(e, TouchEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTouchEvent() {
      return TouchEvent._internalWrap();
    }
    static _internalWrap() {
      return new TouchEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get altKey() {
      return dart.as(wrap_jso(this.raw.altKey), core.bool);
    }
    get changedTouches() {
      return dart.as(wrap_jso(this.raw.changedTouches), TouchList);
    }
    get ctrlKey() {
      return dart.as(wrap_jso(this.raw.ctrlKey), core.bool);
    }
    get metaKey() {
      return dart.as(wrap_jso(this.raw.metaKey), core.bool);
    }
    get shiftKey() {
      return dart.as(wrap_jso(this.raw.shiftKey), core.bool);
    }
    get targetTouches() {
      return dart.as(wrap_jso(this.raw.targetTouches), TouchList);
    }
    get touches() {
      return dart.as(wrap_jso(this.raw.touches), TouchList);
    }
    [_initTouchEvent](touches, targetTouches, changedTouches, type, view, unused1, unused2, unused3, unused4, ctrlKey, altKey, shiftKey, metaKey) {
      this[_initTouchEvent_1](touches, targetTouches, changedTouches, type, view, unused1, unused2, unused3, unused4, ctrlKey, altKey, shiftKey, metaKey);
      return;
    }
    [_initTouchEvent_1](touches, targetTouches, changedTouches, type, view, unused1, unused2, unused3, unused4, ctrlKey, altKey, shiftKey, metaKey) {
      return wrap_jso(this.raw.initTouchEvent(unwrap_jso(touches), unwrap_jso(targetTouches), unwrap_jso(changedTouches), unwrap_jso(type), unwrap_jso(view), unwrap_jso(unused1), unwrap_jso(unused2), unwrap_jso(unused3), unwrap_jso(unused4), unwrap_jso(ctrlKey), unwrap_jso(altKey), unwrap_jso(shiftKey), unwrap_jso(metaKey)));
    }
    static get supported() {
      return html_common.Device.isEventTypeSupported('TouchEvent');
    }
  }
  dart.defineNamedConstructor(TouchEvent, 'internal_');
  dart.setSignature(TouchEvent, {
    constructors: () => ({
      new: [TouchEvent, [TouchList, TouchList, TouchList, core.String], {view: Window, screenX: core.int, screenY: core.int, clientX: core.int, clientY: core.int, ctrlKey: core.bool, altKey: core.bool, shiftKey: core.bool, metaKey: core.bool}],
      _: [TouchEvent, []],
      _internalWrap: [TouchEvent, []],
      internal_: [TouchEvent, []]
    }),
    methods: () => ({
      [_initTouchEvent]: [dart.void, [TouchList, TouchList, TouchList, core.String, Window, core.int, core.int, core.int, core.int, core.bool, core.bool, core.bool, core.bool]],
      [_initTouchEvent_1]: [dart.void, [TouchList, TouchList, TouchList, dart.dynamic, Window, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateTouchEvent: [TouchEvent, []]}),
    names: ['internalCreateTouchEvent']
  });
  TouchEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('TouchEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TouchEvent"))];
  Element.touchCancelEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchcancel'));
  Element.touchEndEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchend'));
  Element.touchEnterEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchenter'));
  Element.touchLeaveEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchleave'));
  Element.touchMoveEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchmove'));
  Element.touchStartEvent = dart.const(new (EventStreamProvider$(TouchEvent))('touchstart'));
  Element.volumeChangeEvent = dart.const(new (EventStreamProvider$(Event))('volumechange'));
  Element.waitingEvent = dart.const(new (EventStreamProvider$(Event))('waiting'));
  Element.fullscreenChangeEvent = dart.const(new (EventStreamProvider$(Event))('webkitfullscreenchange'));
  Element.fullscreenErrorEvent = dart.const(new (EventStreamProvider$(Event))('webkitfullscreenerror'));
  class _ElementFactoryProvider extends core.Object {
    static createElement_tag(tag, typeExtension) {
      return exports.document.createElement(tag, typeExtension);
    }
  }
  dart.setSignature(_ElementFactoryProvider, {
    statics: () => ({createElement_tag: [Element, [core.String, core.String]]}),
    names: ['createElement_tag']
  });
  Event.AT_TARGET = 2;
  Event.BUBBLING_PHASE = 3;
  Event.CAPTURING_PHASE = 1;
  const _ptr = Symbol('_ptr');
  class Events extends core.Object {
    Events(ptr) {
      this[_ptr] = ptr;
    }
    get(type) {
      return new _EventStream(this[_ptr], type, false);
    }
  }
  dart.setSignature(Events, {
    constructors: () => ({Events: [Events, [EventTarget]]}),
    methods: () => ({get: [async.Stream, [core.String]]})
  });
  class ElementEvents extends Events {
    ElementEvents(ptr) {
      super.Events(ptr);
    }
    get(type) {
      if (dart.notNull(ElementEvents.webkitEvents.keys[dartx.contains](type[dartx.toLowerCase]()))) {
        if (dart.notNull(html_common.Device.isWebKit)) {
          return new _ElementEventStreamImpl(this[_ptr], ElementEvents.webkitEvents.get(type[dartx.toLowerCase]()), false);
        }
      }
      return new _ElementEventStreamImpl(this[_ptr], type, false);
    }
  }
  dart.setSignature(ElementEvents, {
    constructors: () => ({ElementEvents: [ElementEvents, [Element]]})
  });
  dart.defineLazyProperties(ElementEvents, {
    get webkitEvents() {
      return dart.map({animationend: 'webkitAnimationEnd', animationiteration: 'webkitAnimationIteration', animationstart: 'webkitAnimationStart', fullscreenchange: 'webkitfullscreenchange', fullscreenerror: 'webkitfullscreenerror', keyadded: 'webkitkeyadded', keyerror: 'webkitkeyerror', keymessage: 'webkitkeymessage', needkey: 'webkitneedkey', pointerlockchange: 'webkitpointerlockchange', pointerlockerror: 'webkitpointerlockerror', resourcetimingbufferfull: 'webkitresourcetimingbufferfull', transitionend: 'webkitTransitionEnd', speechchange: 'webkitSpeechChange'});
    }
  });
  class HRElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("hr"), HRElement);
    }
    static internalCreateHRElement() {
      return HRElement._internalWrap();
    }
    static _internalWrap() {
      return new HRElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get color() {
      return dart.as(wrap_jso(this.raw.color), core.String);
    }
    set color(val) {
      return this.raw.color = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(HRElement, 'internal_');
  dart.defineNamedConstructor(HRElement, 'created');
  dart.setSignature(HRElement, {
    constructors: () => ({
      _: [HRElement, []],
      new: [HRElement, []],
      _internalWrap: [HRElement, []],
      internal_: [HRElement, []],
      created: [HRElement, []]
    }),
    statics: () => ({internalCreateHRElement: [HRElement, []]}),
    names: ['internalCreateHRElement']
  });
  HRElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLHRElement')), dart.const(new _js_helper.Native("HTMLHRElement"))];
  const _initHashChangeEvent = Symbol('_initHashChangeEvent');
  const _initHashChangeEvent_1 = Symbol('_initHashChangeEvent_1');
  class HashChangeEvent extends Event {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      let oldUrl = opts && 'oldUrl' in opts ? opts.oldUrl : null;
      let newUrl = opts && 'newUrl' in opts ? opts.newUrl : null;
      let event = exports.document[_createEvent]("HashChangeEvent");
      dart.dcall(event[_initHashChangeEvent], type, canBubble, cancelable, oldUrl, newUrl);
      return dart.as(event, HashChangeEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHashChangeEvent() {
      return HashChangeEvent._internalWrap();
    }
    static _internalWrap() {
      return new HashChangeEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    static get supported() {
      return html_common.Device.isEventTypeSupported('HashChangeEvent');
    }
    get newUrl() {
      return dart.as(wrap_jso(this.raw.newURL), core.String);
    }
    get oldUrl() {
      return dart.as(wrap_jso(this.raw.oldURL), core.String);
    }
    [_initHashChangeEvent](type, canBubble, cancelable, oldURL, newURL) {
      this[_initHashChangeEvent_1](type, canBubble, cancelable, oldURL, newURL);
      return;
    }
    [_initHashChangeEvent_1](type, canBubble, cancelable, oldURL, newURL) {
      return wrap_jso(this.raw.initHashChangeEvent(unwrap_jso(type), unwrap_jso(canBubble), unwrap_jso(cancelable), unwrap_jso(oldURL), unwrap_jso(newURL)));
    }
  }
  dart.defineNamedConstructor(HashChangeEvent, 'internal_');
  dart.setSignature(HashChangeEvent, {
    constructors: () => ({
      new: [HashChangeEvent, [core.String], {canBubble: core.bool, cancelable: core.bool, oldUrl: core.String, newUrl: core.String}],
      _: [HashChangeEvent, []],
      _internalWrap: [HashChangeEvent, []],
      internal_: [HashChangeEvent, []]
    }),
    methods: () => ({
      [_initHashChangeEvent]: [dart.void, [core.String, core.bool, core.bool, core.String, core.String]],
      [_initHashChangeEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateHashChangeEvent: [HashChangeEvent, []]}),
    names: ['internalCreateHashChangeEvent']
  });
  HashChangeEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('HashChangeEvent')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("HashChangeEvent"))];
  class HeadElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("head"), HeadElement);
    }
    static internalCreateHeadElement() {
      return HeadElement._internalWrap();
    }
    static _internalWrap() {
      return new HeadElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(HeadElement, 'internal_');
  dart.defineNamedConstructor(HeadElement, 'created');
  dart.setSignature(HeadElement, {
    constructors: () => ({
      _: [HeadElement, []],
      new: [HeadElement, []],
      _internalWrap: [HeadElement, []],
      internal_: [HeadElement, []],
      created: [HeadElement, []]
    }),
    statics: () => ({internalCreateHeadElement: [HeadElement, []]}),
    names: ['internalCreateHeadElement']
  });
  HeadElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLHeadElement')), dart.const(new _js_helper.Native("HTMLHeadElement"))];
  const _forEach_1 = Symbol('_forEach_1');
  const _forEach_2 = Symbol('_forEach_2');
  class Headers extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(input) {
      if (input === void 0)
        input = null;
      if (input == null) {
        return Headers._create_1();
      }
      if (dart.is(input, Headers) || input == null) {
        return Headers._create_2(input);
      }
      if (dart.is(input, core.Map) || input == null) {
        let input_1 = html_common.convertDartToNative_Dictionary(dart.as(input, core.Map));
        return Headers._create_3(input_1);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    static _create_1() {
      return dart.as(wrap_jso(new Headers()), Headers);
    }
    static _create_2(input) {
      return dart.as(wrap_jso(new Headers(input)), Headers);
    }
    static _create_3(input) {
      return dart.as(wrap_jso(new Headers(input)), Headers);
    }
    static internalCreateHeaders() {
      return Headers._internalWrap();
    }
    static _internalWrap() {
      return new Headers.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
    forEach(callback, thisArg) {
      if (thisArg === void 0)
        thisArg = null;
      if (thisArg != null) {
        this[_forEach_1](callback, thisArg);
        return;
      }
      this[_forEach_2](callback);
      return;
    }
    [_forEach_1](callback, thisArg) {
      return wrap_jso(this.raw.forEach(unwrap_jso(callback), unwrap_jso(thisArg)));
    }
    [_forEach_2](callback) {
      return wrap_jso(this.raw.forEach(unwrap_jso(callback)));
    }
  }
  dart.defineNamedConstructor(Headers, 'internal_');
  dart.setSignature(Headers, {
    constructors: () => ({
      _: [Headers, []],
      new: [Headers, [], [dart.dynamic]],
      _internalWrap: [Headers, []],
      internal_: [Headers, []]
    }),
    methods: () => ({
      forEach: [dart.void, [HeadersForEachCallback], [core.Object]],
      [_forEach_1]: [dart.void, [HeadersForEachCallback, dart.dynamic]],
      [_forEach_2]: [dart.void, [HeadersForEachCallback]]
    }),
    statics: () => ({
      _create_1: [Headers, []],
      _create_2: [Headers, [dart.dynamic]],
      _create_3: [Headers, [dart.dynamic]],
      internalCreateHeaders: [Headers, []]
    }),
    names: ['_create_1', '_create_2', '_create_3', 'internalCreateHeaders']
  });
  Headers[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Headers')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Headers"))];
  const HeadersForEachCallback = dart.typedef('HeadersForEachCallback', () => dart.functionType(dart.void, [core.String, core.String, Headers]));
  class HeadingElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static h1() {
      return dart.as(exports.document.createElement("h1"), HeadingElement);
    }
    static h2() {
      return dart.as(exports.document.createElement("h2"), HeadingElement);
    }
    static h3() {
      return dart.as(exports.document.createElement("h3"), HeadingElement);
    }
    static h4() {
      return dart.as(exports.document.createElement("h4"), HeadingElement);
    }
    static h5() {
      return dart.as(exports.document.createElement("h5"), HeadingElement);
    }
    static h6() {
      return dart.as(exports.document.createElement("h6"), HeadingElement);
    }
    static internalCreateHeadingElement() {
      return HeadingElement._internalWrap();
    }
    static _internalWrap() {
      return new HeadingElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(HeadingElement, 'internal_');
  dart.defineNamedConstructor(HeadingElement, 'created');
  dart.setSignature(HeadingElement, {
    constructors: () => ({
      _: [HeadingElement, []],
      h1: [HeadingElement, []],
      h2: [HeadingElement, []],
      h3: [HeadingElement, []],
      h4: [HeadingElement, []],
      h5: [HeadingElement, []],
      h6: [HeadingElement, []],
      _internalWrap: [HeadingElement, []],
      internal_: [HeadingElement, []],
      created: [HeadingElement, []]
    }),
    statics: () => ({internalCreateHeadingElement: [HeadingElement, []]}),
    names: ['internalCreateHeadingElement']
  });
  HeadingElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLHeadingElement')), dart.const(new _js_helper.Native("HTMLHeadingElement"))];
  const _back_1 = Symbol('_back_1');
  const _forward_1 = Symbol('_forward_1');
  const _go_1 = Symbol('_go_1');
  const _pushState_1 = Symbol('_pushState_1');
  const _pushState_2 = Symbol('_pushState_2');
  const _replaceState_1 = Symbol('_replaceState_1');
  const _replaceState_2 = Symbol('_replaceState_2');
  class History extends DartHtmlDomObject {
    static get supportsState() {
      return true;
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHistory() {
      return History._internalWrap();
    }
    static _internalWrap() {
      return new History.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get state() {
      return html_common.convertNativeToDart_SerializedScriptValue(this[_get_state]);
    }
    get [_get_state]() {
      return wrap_jso(this.raw.state);
    }
    back() {
      this[_back_1]();
      return;
    }
    [_back_1]() {
      return wrap_jso(this.raw.back());
    }
    forward() {
      this[_forward_1]();
      return;
    }
    [_forward_1]() {
      return wrap_jso(this.raw.forward());
    }
    go(distance) {
      this[_go_1](distance);
      return;
    }
    [_go_1](distance) {
      return wrap_jso(this.raw.go(unwrap_jso(distance)));
    }
    pushState(data, title, url) {
      if (url === void 0)
        url = null;
      if (url != null) {
        let data_1 = html_common.convertDartToNative_SerializedScriptValue(data);
        this[_pushState_1](data_1, title, url);
        return;
      }
      let data_1 = html_common.convertDartToNative_SerializedScriptValue(data);
      this[_pushState_2](data_1, title);
      return;
    }
    [_pushState_1](data, title, url) {
      return wrap_jso(this.raw.pushState(unwrap_jso(data), unwrap_jso(title), unwrap_jso(url)));
    }
    [_pushState_2](data, title) {
      return wrap_jso(this.raw.pushState(unwrap_jso(data), unwrap_jso(title)));
    }
    replaceState(data, title, url) {
      if (url === void 0)
        url = null;
      if (url != null) {
        let data_1 = html_common.convertDartToNative_SerializedScriptValue(data);
        this[_replaceState_1](data_1, title, url);
        return;
      }
      let data_1 = html_common.convertDartToNative_SerializedScriptValue(data);
      this[_replaceState_2](data_1, title);
      return;
    }
    [_replaceState_1](data, title, url) {
      return wrap_jso(this.raw.replaceState(unwrap_jso(data), unwrap_jso(title), unwrap_jso(url)));
    }
    [_replaceState_2](data, title) {
      return wrap_jso(this.raw.replaceState(unwrap_jso(data), unwrap_jso(title)));
    }
  }
  History[dart.implements] = () => [HistoryBase];
  dart.defineNamedConstructor(History, 'internal_');
  dart.setSignature(History, {
    constructors: () => ({
      _: [History, []],
      _internalWrap: [History, []],
      internal_: [History, []]
    }),
    methods: () => ({
      back: [dart.void, []],
      [_back_1]: [dart.void, []],
      forward: [dart.void, []],
      [_forward_1]: [dart.void, []],
      go: [dart.void, [core.int]],
      [_go_1]: [dart.void, [dart.dynamic]],
      pushState: [dart.void, [dart.dynamic, core.String], [core.String]],
      [_pushState_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_pushState_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      replaceState: [dart.void, [dart.dynamic, core.String], [core.String]],
      [_replaceState_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_replaceState_2]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateHistory: [History, []]}),
    names: ['internalCreateHistory']
  });
  History[dart.metadata] = () => [dart.const(new _metadata.DomName('History')), dart.const(new _js_helper.Native("History"))];
  const ImmutableListMixin$ = dart.generic(function(E) {
    class ImmutableListMixin extends core.Object {
      get iterator() {
        return new (FixedSizeListIterator$(E))(this);
      }
      [Symbol.iterator]() {
        return new dart.JsIterator(this.iterator);
      }
      add(value) {
        dart.as(value, E);
        dart.throw(new core.UnsupportedError("Cannot add to immutable List."));
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to immutable List."));
      }
      sort(compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, dart.functionType(core.int, [E, E]));
        dart.throw(new core.UnsupportedError("Cannot sort immutable List."));
      }
      shuffle(random) {
        if (random === void 0)
          random = null;
        dart.throw(new core.UnsupportedError("Cannot shuffle immutable List."));
      }
      insert(index, element) {
        dart.as(element, E);
        dart.throw(new core.UnsupportedError("Cannot add to immutable List."));
      }
      insertAll(index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot add to immutable List."));
      }
      setAll(index, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot modify an immutable List."));
      }
      removeAt(pos) {
        dart.throw(new core.UnsupportedError("Cannot remove from immutable List."));
      }
      removeLast() {
        dart.throw(new core.UnsupportedError("Cannot remove from immutable List."));
      }
      remove(object) {
        dart.throw(new core.UnsupportedError("Cannot remove from immutable List."));
      }
      removeWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from immutable List."));
      }
      retainWhere(test) {
        dart.as(test, dart.functionType(core.bool, [E]));
        dart.throw(new core.UnsupportedError("Cannot remove from immutable List."));
      }
      setRange(start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        dart.throw(new core.UnsupportedError("Cannot setRange on immutable List."));
      }
      removeRange(start, end) {
        dart.throw(new core.UnsupportedError("Cannot removeRange on immutable List."));
      }
      replaceRange(start, end, iterable) {
        dart.as(iterable, core.Iterable$(E));
        dart.throw(new core.UnsupportedError("Cannot modify an immutable List."));
      }
      fillRange(start, end, fillValue) {
        if (fillValue === void 0)
          fillValue = null;
        dart.as(fillValue, E);
        dart.throw(new core.UnsupportedError("Cannot modify an immutable List."));
      }
    }
    ImmutableListMixin[dart.implements] = () => [core.List$(E)];
    dart.setSignature(ImmutableListMixin, {
      methods: () => ({
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        sort: [dart.void, [], [dart.functionType(core.int, [E, E])]],
        shuffle: [dart.void, [], [math.Random]],
        insert: [dart.void, [core.int, E]],
        insertAll: [dart.void, [core.int, core.Iterable$(E)]],
        setAll: [dart.void, [core.int, core.Iterable$(E)]],
        removeAt: [E, [core.int]],
        removeLast: [E, []],
        remove: [core.bool, [core.Object]],
        removeWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        retainWhere: [dart.void, [dart.functionType(core.bool, [E])]],
        setRange: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        removeRange: [dart.void, [core.int, core.int]],
        replaceRange: [dart.void, [core.int, core.int, core.Iterable$(E)]],
        fillRange: [dart.void, [core.int, core.int], [E]]
      })
    });
    dart.defineExtensionMembers(ImmutableListMixin, [
      'add',
      'addAll',
      'sort',
      'shuffle',
      'insert',
      'insertAll',
      'setAll',
      'removeAt',
      'removeLast',
      'remove',
      'removeWhere',
      'retainWhere',
      'setRange',
      'removeRange',
      'replaceRange',
      'fillRange',
      'iterator'
    ]);
    return ImmutableListMixin;
  });
  let ImmutableListMixin = ImmutableListMixin$();
  const _namedItem_1 = Symbol('_namedItem_1');
  class HtmlCollection extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(Node), ImmutableListMixin$(Node)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHtmlCollection() {
      return HtmlCollection._internalWrap();
    }
    static _internalWrap() {
      return new HtmlCollection.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), Node);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Element);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), Element);
    }
  }
  HtmlCollection[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(Node)];
  dart.defineNamedConstructor(HtmlCollection, 'internal_');
  dart.setSignature(HtmlCollection, {
    constructors: () => ({
      _: [HtmlCollection, []],
      _internalWrap: [HtmlCollection, []],
      internal_: [HtmlCollection, []]
    }),
    methods: () => ({
      get: [Node, [core.int]],
      set: [dart.void, [core.int, Node]],
      elementAt: [Node, [core.int]],
      item: [Element, [core.int]],
      [_item_1]: [Element, [dart.dynamic]],
      namedItem: [Element, [core.String]],
      [_namedItem_1]: [Element, [dart.dynamic]]
    }),
    statics: () => ({internalCreateHtmlCollection: [HtmlCollection, []]}),
    names: ['internalCreateHtmlCollection']
  });
  dart.defineExtensionMembers(HtmlCollection, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  HtmlCollection[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLCollection')), dart.const(new _js_helper.Native("HTMLCollection"))];
  const _badUpgrade = Symbol('_badUpgrade');
  class HtmlDocument extends Document {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHtmlDocument() {
      return HtmlDocument._internalWrap();
    }
    static _internalWrap() {
      return new HtmlDocument.internal_();
    }
    internal_() {
      super.internal_();
    }
    get body() {
      return dart.as(this[_body], BodyElement);
    }
    set body(value) {
      this[_body] = value;
    }
    caretRangeFromPoint(x, y) {
      return this[_caretRangeFromPoint](x, y);
    }
    elementFromPoint(x, y) {
      return this[_elementFromPoint](x, y);
    }
    static get supportsCssCanvasContext() {
      return true;
    }
    getCssCanvasContext(contextId, name, width, height) {
      return this[_getCssCanvasContext](contextId, name, width, height);
    }
    get head() {
      return dart.as(this[_head], HeadElement);
    }
    get lastModified() {
      return this[_lastModified];
    }
    get preferredStylesheetSet() {
      return this[_preferredStylesheetSet];
    }
    get referrer() {
      return this[_referrer];
    }
    get selectedStylesheetSet() {
      return this[_selectedStylesheetSet];
    }
    set selectedStylesheetSet(value) {
      this[_selectedStylesheetSet] = value;
    }
    get styleSheets() {
      return this[_styleSheets];
    }
    get title() {
      return this[_title];
    }
    set title(value) {
      this[_title] = value;
    }
    exitFullscreen() {
      this[_webkitExitFullscreen]();
    }
    get fullscreenElement() {
      return this[_webkitFullscreenElement];
    }
    get fullscreenEnabled() {
      return this[_webkitFullscreenEnabled];
    }
    get hidden() {
      return this[_webkitHidden];
    }
    get visibilityState() {
      return this[_webkitVisibilityState];
    }
    registerElement(tag, customElementClass, opts) {
      let extendsTag = opts && 'extendsTag' in opts ? opts.extendsTag : null;
      let classMirror = dart.dcall(/* Unimplemented unknown name */reflectClass, customElementClass);
      let locationUri = dart.toString(dart.dload(dart.dload(classMirror, 'location'), 'sourceUri'));
      if (locationUri == 'dart:html' || locationUri == 'dart:svg') {
        dart.throw(new dart.dynamic("HierarchyRequestError: Cannot register an existing dart:html or dart:svg type."));
      }
      if (dart.notNull(dart.as(dart.dload(classMirror, 'isAbstract'), core.bool))) {
        dart.throw(new dart.dynamic("HierarchyRequestError: Cannot register an abstract class."));
      }
      let jsClassName = dart.dcall(/* Unimplemented unknown name */_getJSClassName, classMirror);
      if (jsClassName == null) {
        dart.throw(new dart.dynamic("HierarchyRequestError: Only HTML elements can be customized."));
      }
      let customClassType = dart.dcall(/* Unimplemented unknown name */_getDartHtmlClassName, classMirror);
      if (extendsTag != null) {
        let nativeElement = exports.document.createElement(extendsTag);
        if (!dart.equals(dart.runtimeType(nativeElement), dart.dload(customClassType, 'reflectedType'))) {
          let nativeElementClassMirror = dart.dcall(/* Unimplemented unknown name */reflectClass, dart.runtimeType(nativeElement));
          let customClassNativeElement = dart.dsend(/* Unimplemented unknown name */MirrorSystem, 'getName', dart.dload(customClassType, 'simpleName'));
          let extendsNativeElement = dart.dsend(/* Unimplemented unknown name */MirrorSystem, 'getName', dart.dload(nativeElementClassMirror, 'simpleName'));
          dart.throw(new dart.dynamic(`HierarchyRequestError: Custom class type (${customClassNativeElement}) and extendsTag class (${extendsNativeElement}) don't match .`));
        }
      } else if (!dart.equals(dart.dload(customClassType, 'reflectedType'), HtmlElement) && !dart.equals(dart.dload(customClassType, 'reflectedType'), dart.dload(/* Unimplemented unknown name */svg, 'SvgElement'))) {
        let customClassName = dart.dsend(/* Unimplemented unknown name */MirrorSystem, 'getName', dart.dload(classMirror, 'simpleName'));
        let customClassElement = dart.dsend(/* Unimplemented unknown name */MirrorSystem, 'getName', dart.dload(customClassType, 'simpleName'));
        dart.throw(new dart.dynamic(`HierarchyRequestError: Custom element ${customClassName} is a native ${customClassElement} should be derived from HtmlElement or SvgElement.`));
      }
      if (dart.notNull(dart.as(dart.dcall(/* Unimplemented unknown name */_hasCreatedConstructor, classMirror), core.bool))) {
        let baseElement = dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'getProperty', dart.dload(/* Unimplemented unknown name */js, 'context'), jsClassName);
        if (baseElement == null) {
          baseElement = dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'getProperty', dart.dload(/* Unimplemented unknown name */js, 'context'), 'HTMLElement');
        }
        let elemProto = dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'callMethod', dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'getProperty', dart.dload(/* Unimplemented unknown name */js, 'context'), 'Object'), "create", [dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'getProperty', baseElement, 'prototype')]);
        dart.dcall(/* Unimplemented unknown name */addCustomElementType, tag, customElementClass, extendsTag);
        let creating = 0;
        let oldCreatedCallback = dart.dindex(elemProto, 'createdCallback');
        let oldAttributeChangedCallback = dart.dindex(elemProto, 'attributeChangedCallback');
        let oldAttachedCallback = dart.dindex(elemProto, 'attachedCallback');
        let oldDetachedCallback = dart.dindex(elemProto, 'detachedCallback');
        dart.dsetindex(elemProto, 'createdCallback', dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'withThis', dart.fn($this => {
          if (dart.dcall(/* Unimplemented unknown name */_getJSClassName, dart.dload(dart.dcall(/* Unimplemented unknown name */reflectClass, customElementClass), 'superclass')) != null && dart.notNull(creating) < 2) {
            creating = dart.notNull(creating) + 1;
            let dartClass = null;
            try {
              if (extendsTag != null) {
                dartClass = exports.document.createElement(extendsTag);
                dart.dsend(/* Unimplemented unknown name */js, 'setDartHtmlWrapperFor', $this, dartClass);
                dart.dput(dartClass, 'blink_jsObject', $this);
              }
              dartClass = dart.dsend(dart.dload(/* Unimplemented unknown name */_blink, 'Blink_Utils'), 'constructElement', customElementClass, $this);
            } catch (e) {
              dartClass = HtmlElement.internalCreateHtmlElement();
              dart.dput(dartClass, 'blink_jsObject', $this);
              dart.dsend(dartClass, _badUpgrade);
              dart.throw(e);
            }
 finally {
              dart.dsend(/* Unimplemented unknown name */js, 'setDartHtmlWrapperFor', $this, dartClass);
              creating = dart.notNull(creating) - 1;
            }
          }
          if (oldCreatedCallback != null)
            dart.dsend(oldCreatedCallback, 'apply', [], {thisArg: unwrap_jso($this)});
        })));
        dart.dsetindex(elemProto, 'attributeChangedCallback', new dart.dynamic.withThis(dart.fn(($this, attrName, oldVal, newVal) => {
          dart.dsend($this, 'attributeChanged', attrName, oldVal, newVal);
          if (oldAttributeChangedCallback != null)
            dart.dsend(oldAttributeChangedCallback, 'apply', [], {thisArg: unwrap_jso($this)});
        })));
        dart.dsetindex(elemProto, 'attachedCallback', new dart.dynamic.withThis(dart.fn($this => {
          dart.dsend($this, 'attached');
          if (oldAttachedCallback != null)
            dart.dsend(oldAttachedCallback, 'apply', [], {thisArg: unwrap_jso($this)});
        })));
        dart.dsetindex(elemProto, 'detachedCallback', new dart.dynamic.withThis(dart.fn($this => {
          dart.dsend($this, 'detached');
          if (oldDetachedCallback != null)
            dart.dsend(oldDetachedCallback, 'apply', [], {thisArg: unwrap_jso($this)});
        })));
        let jsMap = new dart.dynamic.jsify(dart.map({prototype: elemProto, extends: extendsTag}));
        dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'callMethod', dart.dsend(dart.dload(/* Unimplemented unknown name */js, 'JsNative'), 'getProperty', dart.dload(/* Unimplemented unknown name */js, 'context'), 'document'), 'registerElement', [tag, jsMap]);
      }
    }
    register(tag, customElementClass, opts) {
      let extendsTag = opts && 'extendsTag' in opts ? opts.extendsTag : null;
      return this.registerElement(tag, customElementClass, {extendsTag: extendsTag});
    }
    static _determineVisibilityChangeEventType(e) {
      return 'webkitvisibilitychange';
    }
    get onVisibilityChange() {
      return HtmlDocument.visibilityChangeEvent.forTarget(this);
    }
    createElementUpgrader(type, opts) {
      let extendsTag = opts && 'extendsTag' in opts ? opts.extendsTag : null;
      return new dart.dynamic(this, type, extendsTag);
    }
  }
  dart.defineNamedConstructor(HtmlDocument, 'internal_');
  dart.setSignature(HtmlDocument, {
    constructors: () => ({
      _: [HtmlDocument, []],
      _internalWrap: [HtmlDocument, []],
      internal_: [HtmlDocument, []]
    }),
    methods: () => ({
      caretRangeFromPoint: [Range, [core.int, core.int]],
      elementFromPoint: [Element, [core.int, core.int]],
      getCssCanvasContext: [dart.dynamic, [core.String, core.String, core.int, core.int]],
      registerElement: [dart.void, [core.String, core.Type], {extendsTag: core.String}],
      register: [dart.void, [core.String, core.Type], {extendsTag: core.String}],
      createElementUpgrader: [ElementUpgrader, [core.Type], {extendsTag: core.String}]
    }),
    statics: () => ({
      internalCreateHtmlDocument: [HtmlDocument, []],
      _determineVisibilityChangeEventType: [core.String, [EventTarget]]
    }),
    names: ['internalCreateHtmlDocument', '_determineVisibilityChangeEventType']
  });
  HtmlDocument[dart.metadata] = () => [dart.const(new _metadata.DomName('HTMLDocument')), dart.const(new _js_helper.Native("HTMLDocument"))];
  const _eventTypeGetter = Symbol('_eventTypeGetter');
  const _CustomEventStreamProvider$ = dart.generic(function(T) {
    class _CustomEventStreamProvider extends core.Object {
      _CustomEventStreamProvider(eventTypeGetter) {
        this[_eventTypeGetter] = eventTypeGetter;
      }
      forTarget(e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new (_EventStream$(T))(e, dart.as(dart.dcall(this[_eventTypeGetter], e), core.String), useCapture);
      }
      forElement(e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new (_ElementEventStreamImpl$(T))(e, dart.dcall(this[_eventTypeGetter], e), useCapture);
      }
      [_forElementList](e, opts) {
        let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
        return new _ElementListEventStreamImpl(dart.as(e, core.Iterable$(Element)), dart.as(dart.dcall(this[_eventTypeGetter], e), core.String), useCapture);
      }
      getEventType(target) {
        return dart.as(dart.dcall(this[_eventTypeGetter], target), core.String);
      }
      get [_eventType]() {
        return dart.throw(new core.UnsupportedError('Access type through getEventType method.'));
      }
    }
    _CustomEventStreamProvider[dart.implements] = () => [EventStreamProvider$(T)];
    dart.setSignature(_CustomEventStreamProvider, {
      constructors: () => ({_CustomEventStreamProvider: [_CustomEventStreamProvider$(T), [dart.dynamic]]}),
      methods: () => ({
        forTarget: [async.Stream$(T), [EventTarget], {useCapture: core.bool}],
        forElement: [ElementStream$(T), [Element], {useCapture: core.bool}],
        [_forElementList]: [ElementStream$(T), [ElementList], {useCapture: core.bool}],
        getEventType: [core.String, [EventTarget]]
      })
    });
    return _CustomEventStreamProvider;
  });
  let _CustomEventStreamProvider = _CustomEventStreamProvider$();
  HtmlDocument.visibilityChangeEvent = dart.const(new (_CustomEventStreamProvider$(Event))(HtmlDocument._determineVisibilityChangeEventType));
  class HtmlFormControlsCollection extends HtmlCollection {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHtmlFormControlsCollection() {
      return HtmlFormControlsCollection._internalWrap();
    }
    static _internalWrap() {
      return new HtmlFormControlsCollection.internal_();
    }
    internal_() {
      super.internal_();
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return wrap_jso(this.raw.namedItem(unwrap_jso(name)));
    }
  }
  dart.defineNamedConstructor(HtmlFormControlsCollection, 'internal_');
  dart.setSignature(HtmlFormControlsCollection, {
    constructors: () => ({
      _: [HtmlFormControlsCollection, []],
      _internalWrap: [HtmlFormControlsCollection, []],
      internal_: [HtmlFormControlsCollection, []]
    }),
    methods: () => ({
      namedItem: [core.Object, [core.String]],
      [_namedItem_1]: [core.Object, [dart.dynamic]]
    }),
    statics: () => ({internalCreateHtmlFormControlsCollection: [HtmlFormControlsCollection, []]}),
    names: ['internalCreateHtmlFormControlsCollection']
  });
  HtmlFormControlsCollection[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLFormControlsCollection')), dart.const(new _js_helper.Native("HTMLFormControlsCollection"))];
  class HtmlHtmlElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("html"), HtmlHtmlElement);
    }
    static internalCreateHtmlHtmlElement() {
      return HtmlHtmlElement._internalWrap();
    }
    static _internalWrap() {
      return new HtmlHtmlElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(HtmlHtmlElement, 'internal_');
  dart.defineNamedConstructor(HtmlHtmlElement, 'created');
  dart.setSignature(HtmlHtmlElement, {
    constructors: () => ({
      _: [HtmlHtmlElement, []],
      new: [HtmlHtmlElement, []],
      _internalWrap: [HtmlHtmlElement, []],
      internal_: [HtmlHtmlElement, []],
      created: [HtmlHtmlElement, []]
    }),
    statics: () => ({internalCreateHtmlHtmlElement: [HtmlHtmlElement, []]}),
    names: ['internalCreateHtmlHtmlElement']
  });
  HtmlHtmlElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLHtmlElement')), dart.const(new _js_helper.Native("HTMLHtmlElement"))];
  class HtmlOptionsCollection extends HtmlCollection {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHtmlOptionsCollection() {
      return HtmlOptionsCollection._internalWrap();
    }
    static _internalWrap() {
      return new HtmlOptionsCollection.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(HtmlOptionsCollection, 'internal_');
  dart.setSignature(HtmlOptionsCollection, {
    constructors: () => ({
      _: [HtmlOptionsCollection, []],
      _internalWrap: [HtmlOptionsCollection, []],
      internal_: [HtmlOptionsCollection, []]
    }),
    statics: () => ({internalCreateHtmlOptionsCollection: [HtmlOptionsCollection, []]}),
    names: ['internalCreateHtmlOptionsCollection']
  });
  HtmlOptionsCollection[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLOptionsCollection')), dart.const(new _js_helper.Native("HTMLOptionsCollection"))];
  class HttpRequestEventTarget extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHttpRequestEventTarget() {
      return HttpRequestEventTarget._internalWrap();
    }
    static _internalWrap() {
      return new HttpRequestEventTarget.internal_();
    }
    internal_() {
      super.internal_();
    }
    get onAbort() {
      return HttpRequestEventTarget.abortEvent.forTarget(this);
    }
    get onError() {
      return HttpRequestEventTarget.errorEvent.forTarget(this);
    }
    get onLoad() {
      return HttpRequestEventTarget.loadEvent.forTarget(this);
    }
    get onLoadEnd() {
      return HttpRequestEventTarget.loadEndEvent.forTarget(this);
    }
    get onLoadStart() {
      return HttpRequestEventTarget.loadStartEvent.forTarget(this);
    }
    get onProgress() {
      return HttpRequestEventTarget.progressEvent.forTarget(this);
    }
    get onTimeout() {
      return HttpRequestEventTarget.timeoutEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(HttpRequestEventTarget, 'internal_');
  dart.setSignature(HttpRequestEventTarget, {
    constructors: () => ({
      _: [HttpRequestEventTarget, []],
      _internalWrap: [HttpRequestEventTarget, []],
      internal_: [HttpRequestEventTarget, []]
    }),
    statics: () => ({internalCreateHttpRequestEventTarget: [HttpRequestEventTarget, []]}),
    names: ['internalCreateHttpRequestEventTarget']
  });
  HttpRequestEventTarget[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XMLHttpRequestEventTarget')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("XMLHttpRequestEventTarget"))];
  const _get_response = Symbol('_get_response');
  const _abort_1 = Symbol('_abort_1');
  const _getAllResponseHeaders_1 = Symbol('_getAllResponseHeaders_1');
  const _getResponseHeader_1 = Symbol('_getResponseHeader_1');
  const _overrideMimeType_1 = Symbol('_overrideMimeType_1');
  const _send_1 = Symbol('_send_1');
  const _send_2 = Symbol('_send_2');
  const _send_3 = Symbol('_send_3');
  const _send_4 = Symbol('_send_4');
  const _send_5 = Symbol('_send_5');
  const _send_6 = Symbol('_send_6');
  const _send_7 = Symbol('_send_7');
  const _setRequestHeader_1 = Symbol('_setRequestHeader_1');
  class HttpRequest extends HttpRequestEventTarget {
    static getString(url, opts) {
      let withCredentials = opts && 'withCredentials' in opts ? opts.withCredentials : null;
      let onProgress = opts && 'onProgress' in opts ? opts.onProgress : null;
      return HttpRequest.request(url, {withCredentials: withCredentials, onProgress: onProgress}).then(dart.fn(xhr => xhr.responseText, core.String, [HttpRequest]));
    }
    static postFormData(url, data, opts) {
      let withCredentials = opts && 'withCredentials' in opts ? opts.withCredentials : null;
      let responseType = opts && 'responseType' in opts ? opts.responseType : null;
      let requestHeaders = opts && 'requestHeaders' in opts ? opts.requestHeaders : null;
      let onProgress = opts && 'onProgress' in opts ? opts.onProgress : null;
      let parts = [];
      data.forEach(dart.fn((key, value) => {
        parts[dartx.add](`${core.Uri.encodeQueryComponent(dart.as(key, core.String))}=` + `${core.Uri.encodeQueryComponent(dart.as(value, core.String))}`);
      }));
      let formData = parts[dartx.join]('&');
      if (requestHeaders == null) {
        requestHeaders = dart.map();
      }
      requestHeaders.putIfAbsent('Content-Type', dart.fn(() => 'application/x-www-form-urlencoded; charset=UTF-8', core.String, []));
      return HttpRequest.request(url, {method: 'POST', withCredentials: withCredentials, responseType: responseType, requestHeaders: requestHeaders, sendData: formData, onProgress: onProgress});
    }
    static request(url, opts) {
      let method = opts && 'method' in opts ? opts.method : null;
      let withCredentials = opts && 'withCredentials' in opts ? opts.withCredentials : null;
      let responseType = opts && 'responseType' in opts ? opts.responseType : null;
      let mimeType = opts && 'mimeType' in opts ? opts.mimeType : null;
      let requestHeaders = opts && 'requestHeaders' in opts ? opts.requestHeaders : null;
      let sendData = opts && 'sendData' in opts ? opts.sendData : null;
      let onProgress = opts && 'onProgress' in opts ? opts.onProgress : null;
      let completer = async.Completer$(HttpRequest).new();
      let xhr = HttpRequest.new();
      if (method == null) {
        method = 'GET';
      }
      xhr.open(method, url, {async: true});
      if (withCredentials != null) {
        xhr.withCredentials = withCredentials;
      }
      if (responseType != null) {
        xhr.responseType = responseType;
      }
      if (mimeType != null) {
        xhr.overrideMimeType(mimeType);
      }
      if (requestHeaders != null) {
        requestHeaders.forEach(dart.fn((header, value) => {
          xhr.setRequestHeader(dart.as(header, core.String), dart.as(value, core.String));
        }));
      }
      if (onProgress != null) {
        xhr.onProgress.listen(onProgress);
      }
      xhr.onLoad.listen(dart.fn(e => {
        let accepted = dart.notNull(xhr.status) >= 200 && dart.notNull(xhr.status) < 300;
        let fileUri = xhr.status == 0;
        let notModified = xhr.status == 304;
        let unknownRedirect = dart.notNull(xhr.status) > 307 && dart.notNull(xhr.status) < 400;
        if (dart.notNull(accepted) || dart.notNull(fileUri) || dart.notNull(notModified) || dart.notNull(unknownRedirect)) {
          completer.complete(xhr);
        } else {
          completer.completeError(e);
        }
      }));
      xhr.onError.listen(dart.bind(completer, 'completeError'));
      if (sendData != null) {
        xhr.send(sendData);
      } else {
        xhr.send();
      }
      return completer.future;
    }
    static get supportsProgressEvent() {
      return true;
    }
    static get supportsCrossOrigin() {
      return true;
    }
    static get supportsLoadEndEvent() {
      return true;
    }
    static get supportsOverrideMimeType() {
      return true;
    }
    static requestCrossOrigin(url, opts) {
      let method = opts && 'method' in opts ? opts.method : null;
      let sendData = opts && 'sendData' in opts ? opts.sendData : null;
      if (dart.notNull(HttpRequest.supportsCrossOrigin)) {
        return dart.as(HttpRequest.request(url, {method: method, sendData: sendData}).then(dart.fn(xhr => {
          return dart.dload(xhr, 'responseText');
        })), async.Future$(core.String));
      }
    }
    get responseHeaders() {
      let headers = dart.map();
      let headersString = this.getAllResponseHeaders();
      if (headersString == null) {
        return headers;
      }
      let headersList = headersString[dartx.split]('\r\n');
      for (let header of headersList) {
        if (dart.notNull(header[dartx.isEmpty])) {
          continue;
        }
        let splitIdx = header[dartx.indexOf](': ');
        if (splitIdx == -1) {
          continue;
        }
        let key = header[dartx.substring](0, splitIdx)[dartx.toLowerCase]();
        let value = header[dartx.substring](dart.notNull(splitIdx) + 2);
        if (dart.notNull(headers.containsKey(key))) {
          headers.set(key, `${headers.get(key)}, ${value}`);
        } else {
          headers.set(key, value);
        }
      }
      return headers;
    }
    open(method, url, opts) {
      let async = opts && 'async' in opts ? opts.async : null;
      let user = opts && 'user' in opts ? opts.user : null;
      let password = opts && 'password' in opts ? opts.password : null;
      if (async == null && user == null && password == null) {
        this.raw.open(method, url);
      } else {
        this.raw.open(method, url, async, user, password);
      }
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return HttpRequest._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new XMLHttpRequest()), HttpRequest);
    }
    static internalCreateHttpRequest() {
      return HttpRequest._internalWrap();
    }
    static _internalWrap() {
      return new HttpRequest.internal_();
    }
    internal_() {
      super.internal_();
    }
    get readyState() {
      return dart.as(wrap_jso(this.raw.readyState), core.int);
    }
    get response() {
      return _convertNativeToDart_XHR_Response(this[_get_response]);
    }
    get [_get_response]() {
      return wrap_jso(this.raw.response);
    }
    get responseText() {
      return dart.as(wrap_jso(this.raw.responseText), core.String);
    }
    get responseType() {
      return dart.as(wrap_jso(this.raw.responseType), core.String);
    }
    set responseType(val) {
      return this.raw.responseType = unwrap_jso(val);
    }
    get responseUrl() {
      return dart.as(wrap_jso(this.raw.responseURL), core.String);
    }
    get responseXml() {
      return dart.as(wrap_jso(this.raw.responseXML), Document);
    }
    get status() {
      return dart.as(wrap_jso(this.raw.status), core.int);
    }
    get statusText() {
      return dart.as(wrap_jso(this.raw.statusText), core.String);
    }
    get timeout() {
      return dart.as(wrap_jso(this.raw.timeout), core.int);
    }
    set timeout(val) {
      return this.raw.timeout = unwrap_jso(val);
    }
    get upload() {
      return dart.as(wrap_jso(this.raw.upload), HttpRequestUpload);
    }
    get withCredentials() {
      return dart.as(wrap_jso(this.raw.withCredentials), core.bool);
    }
    set withCredentials(val) {
      return this.raw.withCredentials = unwrap_jso(val);
    }
    abort() {
      this[_abort_1]();
      return;
    }
    [_abort_1]() {
      return wrap_jso(this.raw.abort());
    }
    getAllResponseHeaders() {
      return this[_getAllResponseHeaders_1]();
    }
    [_getAllResponseHeaders_1]() {
      return dart.as(wrap_jso(this.raw.getAllResponseHeaders()), core.String);
    }
    getResponseHeader(header) {
      return this[_getResponseHeader_1](header);
    }
    [_getResponseHeader_1](header) {
      return dart.as(wrap_jso(this.raw.getResponseHeader(unwrap_jso(header))), core.String);
    }
    overrideMimeType(override) {
      this[_overrideMimeType_1](override);
      return;
    }
    [_overrideMimeType_1](override) {
      return wrap_jso(this.raw.overrideMimeType(unwrap_jso(override)));
    }
    send(data) {
      if (data === void 0)
        data = null;
      if (data == null) {
        this[_send_1]();
        return;
      }
      if (dart.is(data, typed_data.TypedData) || data == null) {
        this[_send_2](dart.as(data, typed_data.TypedData));
        return;
      }
      if (dart.is(data, typed_data.ByteBuffer) || data == null) {
        this[_send_3](dart.as(data, typed_data.ByteBuffer));
        return;
      }
      if (dart.is(data, dart.dynamic) || data == null) {
        this[_send_4](data);
        return;
      }
      if (dart.is(data, Document) || data == null) {
        this[_send_5](dart.as(data, Document));
        return;
      }
      if (typeof data == 'string' || data == null) {
        this[_send_6](dart.as(data, core.String));
        return;
      }
      if (dart.is(data, dart.dynamic) || data == null) {
        this[_send_7](data);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_send_1]() {
      return wrap_jso(this.raw.send());
    }
    [_send_2](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_3](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_4](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_5](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_6](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_7](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    setRequestHeader(header, value) {
      this[_setRequestHeader_1](header, value);
      return;
    }
    [_setRequestHeader_1](header, value) {
      return wrap_jso(this.raw.setRequestHeader(unwrap_jso(header), unwrap_jso(value)));
    }
    get onReadyStateChange() {
      return HttpRequest.readyStateChangeEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(HttpRequest, 'internal_');
  dart.setSignature(HttpRequest, {
    constructors: () => ({
      _: [HttpRequest, []],
      new: [HttpRequest, []],
      _internalWrap: [HttpRequest, []],
      internal_: [HttpRequest, []]
    }),
    methods: () => ({
      open: [dart.void, [core.String, core.String], {async: core.bool, user: core.String, password: core.String}],
      abort: [dart.void, []],
      [_abort_1]: [dart.void, []],
      getAllResponseHeaders: [core.String, []],
      [_getAllResponseHeaders_1]: [core.String, []],
      getResponseHeader: [core.String, [core.String]],
      [_getResponseHeader_1]: [core.String, [dart.dynamic]],
      overrideMimeType: [dart.void, [core.String]],
      [_overrideMimeType_1]: [dart.void, [dart.dynamic]],
      send: [dart.void, [], [dart.dynamic]],
      [_send_1]: [dart.void, []],
      [_send_2]: [dart.void, [typed_data.TypedData]],
      [_send_3]: [dart.void, [typed_data.ByteBuffer]],
      [_send_4]: [dart.void, [dart.dynamic]],
      [_send_5]: [dart.void, [Document]],
      [_send_6]: [dart.void, [core.String]],
      [_send_7]: [dart.void, [dart.dynamic]],
      setRequestHeader: [dart.void, [core.String, core.String]],
      [_setRequestHeader_1]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({
      getString: [async.Future$(core.String), [core.String], {withCredentials: core.bool, onProgress: dart.functionType(dart.void, [ProgressEvent])}],
      postFormData: [async.Future$(HttpRequest), [core.String, core.Map$(core.String, core.String)], {withCredentials: core.bool, responseType: core.String, requestHeaders: core.Map$(core.String, core.String), onProgress: dart.functionType(dart.void, [ProgressEvent])}],
      request: [async.Future$(HttpRequest), [core.String], {method: core.String, withCredentials: core.bool, responseType: core.String, mimeType: core.String, requestHeaders: core.Map$(core.String, core.String), sendData: dart.dynamic, onProgress: dart.functionType(dart.void, [ProgressEvent])}],
      requestCrossOrigin: [async.Future$(core.String), [core.String], {method: core.String, sendData: core.String}],
      _create_1: [HttpRequest, []],
      internalCreateHttpRequest: [HttpRequest, []]
    }),
    names: ['getString', 'postFormData', 'request', 'requestCrossOrigin', '_create_1', 'internalCreateHttpRequest']
  });
  HttpRequest[dart.metadata] = () => [dart.const(new _metadata.DomName('XMLHttpRequest')), dart.const(new _js_helper.Native("XMLHttpRequest"))];
  class ProgressEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateProgressEvent() {
      return ProgressEvent._internalWrap();
    }
    static _internalWrap() {
      return new ProgressEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get lengthComputable() {
      return dart.as(wrap_jso(this.raw.lengthComputable), core.bool);
    }
    get loaded() {
      return dart.as(wrap_jso(this.raw.loaded), core.int);
    }
    get total() {
      return dart.as(wrap_jso(this.raw.total), core.int);
    }
  }
  dart.defineNamedConstructor(ProgressEvent, 'internal_');
  dart.setSignature(ProgressEvent, {
    constructors: () => ({
      _: [ProgressEvent, []],
      _internalWrap: [ProgressEvent, []],
      internal_: [ProgressEvent, []]
    }),
    statics: () => ({internalCreateProgressEvent: [ProgressEvent, []]}),
    names: ['internalCreateProgressEvent']
  });
  ProgressEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ProgressEvent')), dart.const(new _js_helper.Native("ProgressEvent"))];
  HttpRequest.readyStateChangeEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('readystatechange'));
  HttpRequest.DONE = 4;
  HttpRequest.HEADERS_RECEIVED = 2;
  HttpRequest.LOADING = 3;
  HttpRequest.OPENED = 1;
  HttpRequest.UNSENT = 0;
  HttpRequestEventTarget.abortEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('abort'));
  HttpRequestEventTarget.errorEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('error'));
  HttpRequestEventTarget.loadEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('load'));
  HttpRequestEventTarget.loadEndEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('loadend'));
  HttpRequestEventTarget.loadStartEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('loadstart'));
  HttpRequestEventTarget.progressEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('progress'));
  HttpRequestEventTarget.timeoutEvent = dart.const(new (EventStreamProvider$(ProgressEvent))('timeout'));
  class HttpRequestUpload extends HttpRequestEventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateHttpRequestUpload() {
      return HttpRequestUpload._internalWrap();
    }
    static _internalWrap() {
      return new HttpRequestUpload.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(HttpRequestUpload, 'internal_');
  dart.setSignature(HttpRequestUpload, {
    constructors: () => ({
      _: [HttpRequestUpload, []],
      _internalWrap: [HttpRequestUpload, []],
      internal_: [HttpRequestUpload, []]
    }),
    statics: () => ({internalCreateHttpRequestUpload: [HttpRequestUpload, []]}),
    names: ['internalCreateHttpRequestUpload']
  });
  HttpRequestUpload[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XMLHttpRequestUpload')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("XMLHttpRequestUpload"))];
  const _get_contentWindow = Symbol('_get_contentWindow');
  class IFrameElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("iframe"), IFrameElement);
    }
    static internalCreateIFrameElement() {
      return IFrameElement._internalWrap();
    }
    static _internalWrap() {
      return new IFrameElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get allowFullscreen() {
      return dart.as(wrap_jso(this.raw.allowFullscreen), core.bool);
    }
    set allowFullscreen(val) {
      return this.raw.allowFullscreen = unwrap_jso(val);
    }
    get contentWindow() {
      return _convertNativeToDart_Window(this[_get_contentWindow]);
    }
    get [_get_contentWindow]() {
      return wrap_jso(this.raw.contentWindow);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.String);
    }
    set height(val) {
      return this.raw.height = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get sandbox() {
      return dart.as(wrap_jso(this.raw.sandbox), core.String);
    }
    set sandbox(val) {
      return this.raw.sandbox = unwrap_jso(val);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get srcdoc() {
      return dart.as(wrap_jso(this.raw.srcdoc), core.String);
    }
    set srcdoc(val) {
      return this.raw.srcdoc = unwrap_jso(val);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.String);
    }
    set width(val) {
      return this.raw.width = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(IFrameElement, 'internal_');
  dart.defineNamedConstructor(IFrameElement, 'created');
  dart.setSignature(IFrameElement, {
    constructors: () => ({
      _: [IFrameElement, []],
      new: [IFrameElement, []],
      _internalWrap: [IFrameElement, []],
      internal_: [IFrameElement, []],
      created: [IFrameElement, []]
    }),
    statics: () => ({internalCreateIFrameElement: [IFrameElement, []]}),
    names: ['internalCreateIFrameElement']
  });
  IFrameElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLIFrameElement')), dart.const(new _js_helper.Native("HTMLIFrameElement"))];
  class ImageBitmap extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateImageBitmap() {
      return ImageBitmap._internalWrap();
    }
    static _internalWrap() {
      return new ImageBitmap.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.int);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.int);
    }
  }
  dart.defineNamedConstructor(ImageBitmap, 'internal_');
  dart.setSignature(ImageBitmap, {
    constructors: () => ({
      _: [ImageBitmap, []],
      _internalWrap: [ImageBitmap, []],
      internal_: [ImageBitmap, []]
    }),
    statics: () => ({internalCreateImageBitmap: [ImageBitmap, []]}),
    names: ['internalCreateImageBitmap']
  });
  ImageBitmap[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ImageBitmap')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ImageBitmap"))];
  class ImageData extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(data_OR_width, height_OR_width, height) {
      if (height === void 0)
        height = null;
      if ((typeof height_OR_width == 'number' || height_OR_width == null) && (typeof data_OR_width == 'number' || data_OR_width == null) && height == null) {
        return ImageData._create_1(data_OR_width, height_OR_width);
      }
      if ((typeof height_OR_width == 'number' || height_OR_width == null) && (dart.is(data_OR_width, typed_data.Uint8ClampedList) || data_OR_width == null) && height == null) {
        return ImageData._create_2(data_OR_width, height_OR_width);
      }
      if ((typeof height == 'number' || height == null) && (typeof height_OR_width == 'number' || height_OR_width == null) && (dart.is(data_OR_width, typed_data.Uint8ClampedList) || data_OR_width == null)) {
        return ImageData._create_3(data_OR_width, height_OR_width, height);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    static _create_1(data_OR_width, height_OR_width) {
      return dart.as(wrap_jso(new ImageData(data_OR_width, height_OR_width)), ImageData);
    }
    static _create_2(data_OR_width, height_OR_width) {
      return dart.as(wrap_jso(new ImageData(data_OR_width, height_OR_width)), ImageData);
    }
    static _create_3(data_OR_width, height_OR_width, height) {
      return dart.as(wrap_jso(new ImageData(data_OR_width, height_OR_width, height)), ImageData);
    }
    static internalCreateImageData() {
      return ImageData._internalWrap();
    }
    static _internalWrap() {
      return new ImageData.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), typed_data.Uint8ClampedList);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.int);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.int);
    }
  }
  dart.defineNamedConstructor(ImageData, 'internal_');
  dart.setSignature(ImageData, {
    constructors: () => ({
      _: [ImageData, []],
      new: [ImageData, [dart.dynamic, core.int], [core.int]],
      _internalWrap: [ImageData, []],
      internal_: [ImageData, []]
    }),
    statics: () => ({
      _create_1: [ImageData, [dart.dynamic, dart.dynamic]],
      _create_2: [ImageData, [dart.dynamic, dart.dynamic]],
      _create_3: [ImageData, [dart.dynamic, dart.dynamic, dart.dynamic]],
      internalCreateImageData: [ImageData, []]
    }),
    names: ['_create_1', '_create_2', '_create_3', 'internalCreateImageData']
  });
  ImageData[dart.metadata] = () => [dart.const(new _metadata.DomName('ImageData')), dart.const(new _js_helper.Native("ImageData"))];
  class ImageElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(opts) {
      let src = opts && 'src' in opts ? opts.src : null;
      let width = opts && 'width' in opts ? opts.width : null;
      let height = opts && 'height' in opts ? opts.height : null;
      let e = exports.document.createElement("img");
      if (src != null)
        dart.dput(e, 'src', src);
      if (width != null)
        dart.dput(e, 'width', width);
      if (height != null)
        dart.dput(e, 'height', height);
      return dart.as(e, ImageElement);
    }
    static internalCreateImageElement() {
      return ImageElement._internalWrap();
    }
    static _internalWrap() {
      return new ImageElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get alt() {
      return dart.as(wrap_jso(this.raw.alt), core.String);
    }
    set alt(val) {
      return this.raw.alt = unwrap_jso(val);
    }
    get complete() {
      return dart.as(wrap_jso(this.raw.complete), core.bool);
    }
    get crossOrigin() {
      return dart.as(wrap_jso(this.raw.crossOrigin), core.String);
    }
    set crossOrigin(val) {
      return this.raw.crossOrigin = unwrap_jso(val);
    }
    get currentSrc() {
      return dart.as(wrap_jso(this.raw.currentSrc), core.String);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.int);
    }
    set height(val) {
      return this.raw.height = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get isMap() {
      return dart.as(wrap_jso(this.raw.isMap), core.bool);
    }
    set isMap(val) {
      return this.raw.isMap = unwrap_jso(val);
    }
    get naturalHeight() {
      return dart.as(wrap_jso(this.raw.naturalHeight), core.int);
    }
    get naturalWidth() {
      return dart.as(wrap_jso(this.raw.naturalWidth), core.int);
    }
    get sizes() {
      return dart.as(wrap_jso(this.raw.sizes), core.String);
    }
    set sizes(val) {
      return this.raw.sizes = unwrap_jso(val);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get srcset() {
      return dart.as(wrap_jso(this.raw.srcset), core.String);
    }
    set srcset(val) {
      return this.raw.srcset = unwrap_jso(val);
    }
    get useMap() {
      return dart.as(wrap_jso(this.raw.useMap), core.String);
    }
    set useMap(val) {
      return this.raw.useMap = unwrap_jso(val);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.int);
    }
    set width(val) {
      return this.raw.width = unwrap_jso(val);
    }
  }
  ImageElement[dart.implements] = () => [CanvasImageSource];
  dart.defineNamedConstructor(ImageElement, 'internal_');
  dart.defineNamedConstructor(ImageElement, 'created');
  dart.setSignature(ImageElement, {
    constructors: () => ({
      _: [ImageElement, []],
      new: [ImageElement, [], {src: core.String, width: core.int, height: core.int}],
      _internalWrap: [ImageElement, []],
      internal_: [ImageElement, []],
      created: [ImageElement, []]
    }),
    statics: () => ({internalCreateImageElement: [ImageElement, []]}),
    names: ['internalCreateImageElement']
  });
  ImageElement[dart.metadata] = () => [dart.const(new _metadata.DomName('HTMLImageElement')), dart.const(new _js_helper.Native("HTMLImageElement"))];
  const _inspect_1 = Symbol('_inspect_1');
  class InjectedScriptHost extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateInjectedScriptHost() {
      return InjectedScriptHost._internalWrap();
    }
    static _internalWrap() {
      return new InjectedScriptHost.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    inspect(objectId, hints) {
      this[_inspect_1](objectId, hints);
      return;
    }
    [_inspect_1](objectId, hints) {
      return wrap_jso(this.raw.inspect(unwrap_jso(objectId), unwrap_jso(hints)));
    }
  }
  dart.defineNamedConstructor(InjectedScriptHost, 'internal_');
  dart.setSignature(InjectedScriptHost, {
    constructors: () => ({
      _: [InjectedScriptHost, []],
      _internalWrap: [InjectedScriptHost, []],
      internal_: [InjectedScriptHost, []]
    }),
    methods: () => ({
      inspect: [dart.void, [core.Object, core.Object]],
      [_inspect_1]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateInjectedScriptHost: [InjectedScriptHost, []]}),
    names: ['internalCreateInjectedScriptHost']
  });
  InjectedScriptHost[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('InjectedScriptHost')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("InjectedScriptHost"))];
  const _get_valueAsDate = Symbol('_get_valueAsDate');
  const _set_valueAsDate = Symbol('_set_valueAsDate');
  const _select_1 = Symbol('_select_1');
  const _setRangeText_1 = Symbol('_setRangeText_1');
  const _setRangeText_2 = Symbol('_setRangeText_2');
  const _setRangeText_3 = Symbol('_setRangeText_3');
  const _setSelectionRange_1 = Symbol('_setSelectionRange_1');
  const _setSelectionRange_2 = Symbol('_setSelectionRange_2');
  const _stepDown_1 = Symbol('_stepDown_1');
  const _stepDown_2 = Symbol('_stepDown_2');
  const _stepUp_1 = Symbol('_stepUp_1');
  const _stepUp_2 = Symbol('_stepUp_2');
  class InputElement extends HtmlElement {
    static new(opts) {
      let type = opts && 'type' in opts ? opts.type : null;
      let e = exports.document.createElement("input");
      if (type != null) {
        try {
          dart.dput(e, 'type', type);
        } catch (_) {
        }

      }
      return dart.as(e, InputElement);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateInputElement() {
      return InputElement._internalWrap();
    }
    static _internalWrap() {
      return new InputElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get accept() {
      return dart.as(wrap_jso(this.raw.accept), core.String);
    }
    set accept(val) {
      return this.raw.accept = unwrap_jso(val);
    }
    get alt() {
      return dart.as(wrap_jso(this.raw.alt), core.String);
    }
    set alt(val) {
      return this.raw.alt = unwrap_jso(val);
    }
    get autocomplete() {
      return dart.as(wrap_jso(this.raw.autocomplete), core.String);
    }
    set autocomplete(val) {
      return this.raw.autocomplete = unwrap_jso(val);
    }
    get autofocus() {
      return dart.as(wrap_jso(this.raw.autofocus), core.bool);
    }
    set autofocus(val) {
      return this.raw.autofocus = unwrap_jso(val);
    }
    get capture() {
      return dart.as(wrap_jso(this.raw.capture), core.bool);
    }
    set capture(val) {
      return this.raw.capture = unwrap_jso(val);
    }
    get checked() {
      return dart.as(wrap_jso(this.raw.checked), core.bool);
    }
    set checked(val) {
      return this.raw.checked = unwrap_jso(val);
    }
    get defaultChecked() {
      return dart.as(wrap_jso(this.raw.defaultChecked), core.bool);
    }
    set defaultChecked(val) {
      return this.raw.defaultChecked = unwrap_jso(val);
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.defaultValue), core.String);
    }
    set defaultValue(val) {
      return this.raw.defaultValue = unwrap_jso(val);
    }
    get dirName() {
      return dart.as(wrap_jso(this.raw.dirName), core.String);
    }
    set dirName(val) {
      return this.raw.dirName = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get files() {
      return dart.as(wrap_jso(this.raw.files), core.List$(dart.dynamic));
    }
    set files(val) {
      return this.raw.files = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get formAction() {
      return dart.as(wrap_jso(this.raw.formAction), core.String);
    }
    set formAction(val) {
      return this.raw.formAction = unwrap_jso(val);
    }
    get formEnctype() {
      return dart.as(wrap_jso(this.raw.formEnctype), core.String);
    }
    set formEnctype(val) {
      return this.raw.formEnctype = unwrap_jso(val);
    }
    get formMethod() {
      return dart.as(wrap_jso(this.raw.formMethod), core.String);
    }
    set formMethod(val) {
      return this.raw.formMethod = unwrap_jso(val);
    }
    get formNoValidate() {
      return dart.as(wrap_jso(this.raw.formNoValidate), core.bool);
    }
    set formNoValidate(val) {
      return this.raw.formNoValidate = unwrap_jso(val);
    }
    get formTarget() {
      return dart.as(wrap_jso(this.raw.formTarget), core.String);
    }
    set formTarget(val) {
      return this.raw.formTarget = unwrap_jso(val);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.int);
    }
    set height(val) {
      return this.raw.height = unwrap_jso(val);
    }
    get incremental() {
      return dart.as(wrap_jso(this.raw.incremental), core.bool);
    }
    set incremental(val) {
      return this.raw.incremental = unwrap_jso(val);
    }
    get indeterminate() {
      return dart.as(wrap_jso(this.raw.indeterminate), core.bool);
    }
    set indeterminate(val) {
      return this.raw.indeterminate = unwrap_jso(val);
    }
    get inputMode() {
      return dart.as(wrap_jso(this.raw.inputMode), core.String);
    }
    set inputMode(val) {
      return this.raw.inputMode = unwrap_jso(val);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get list() {
      return dart.as(wrap_jso(this.raw.list), HtmlElement);
    }
    get max() {
      return dart.as(wrap_jso(this.raw.max), core.String);
    }
    set max(val) {
      return this.raw.max = unwrap_jso(val);
    }
    get maxLength() {
      return dart.as(wrap_jso(this.raw.maxLength), core.int);
    }
    set maxLength(val) {
      return this.raw.maxLength = unwrap_jso(val);
    }
    get min() {
      return dart.as(wrap_jso(this.raw.min), core.String);
    }
    set min(val) {
      return this.raw.min = unwrap_jso(val);
    }
    get multiple() {
      return dart.as(wrap_jso(this.raw.multiple), core.bool);
    }
    set multiple(val) {
      return this.raw.multiple = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get pattern() {
      return dart.as(wrap_jso(this.raw.pattern), core.String);
    }
    set pattern(val) {
      return this.raw.pattern = unwrap_jso(val);
    }
    get placeholder() {
      return dart.as(wrap_jso(this.raw.placeholder), core.String);
    }
    set placeholder(val) {
      return this.raw.placeholder = unwrap_jso(val);
    }
    get readOnly() {
      return dart.as(wrap_jso(this.raw.readOnly), core.bool);
    }
    set readOnly(val) {
      return this.raw.readOnly = unwrap_jso(val);
    }
    get required() {
      return dart.as(wrap_jso(this.raw.required), core.bool);
    }
    set required(val) {
      return this.raw.required = unwrap_jso(val);
    }
    get selectionDirection() {
      return dart.as(wrap_jso(this.raw.selectionDirection), core.String);
    }
    set selectionDirection(val) {
      return this.raw.selectionDirection = unwrap_jso(val);
    }
    get selectionEnd() {
      return dart.as(wrap_jso(this.raw.selectionEnd), core.int);
    }
    set selectionEnd(val) {
      return this.raw.selectionEnd = unwrap_jso(val);
    }
    get selectionStart() {
      return dart.as(wrap_jso(this.raw.selectionStart), core.int);
    }
    set selectionStart(val) {
      return this.raw.selectionStart = unwrap_jso(val);
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
    set size(val) {
      return this.raw.size = unwrap_jso(val);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get step() {
      return dart.as(wrap_jso(this.raw.step), core.String);
    }
    set step(val) {
      return this.raw.step = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
    get valueAsDate() {
      return html_common.convertNativeToDart_DateTime(this[_get_valueAsDate]);
    }
    get [_get_valueAsDate]() {
      return wrap_jso(this.raw.valueAsDate);
    }
    set valueAsDate(value) {
      this[_set_valueAsDate] = html_common.convertDartToNative_DateTime(value);
    }
    set [_set_valueAsDate](value) {
      this.raw.valueAsDate = unwrap_jso(value);
    }
    get valueAsNumber() {
      return dart.as(wrap_jso(this.raw.valueAsNumber), core.num);
    }
    set valueAsNumber(val) {
      return this.raw.valueAsNumber = unwrap_jso(val);
    }
    get entries() {
      return dart.as(wrap_jso(this.raw.webkitEntries), core.List$(dart.dynamic));
    }
    get directory() {
      return dart.as(wrap_jso(this.raw.webkitdirectory), core.bool);
    }
    set directory(val) {
      return this.raw.webkitdirectory = unwrap_jso(val);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.int);
    }
    set width(val) {
      return this.raw.width = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    select() {
      this[_select_1]();
      return;
    }
    [_select_1]() {
      return wrap_jso(this.raw.select());
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
    setRangeText(replacement, opts) {
      let start = opts && 'start' in opts ? opts.start : null;
      let end = opts && 'end' in opts ? opts.end : null;
      let selectionMode = opts && 'selectionMode' in opts ? opts.selectionMode : null;
      if (start == null && end == null && selectionMode == null) {
        this[_setRangeText_1](replacement);
        return;
      }
      if (end != null && start != null && selectionMode == null) {
        this[_setRangeText_2](replacement, start, end);
        return;
      }
      if (selectionMode != null && end != null && start != null) {
        this[_setRangeText_3](replacement, start, end, selectionMode);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_setRangeText_1](replacement) {
      return wrap_jso(this.raw.setRangeText(unwrap_jso(replacement)));
    }
    [_setRangeText_2](replacement, start, end) {
      return wrap_jso(this.raw.setRangeText(unwrap_jso(replacement), unwrap_jso(start), unwrap_jso(end)));
    }
    [_setRangeText_3](replacement, start, end, selectionMode) {
      return wrap_jso(this.raw.setRangeText(unwrap_jso(replacement), unwrap_jso(start), unwrap_jso(end), unwrap_jso(selectionMode)));
    }
    setSelectionRange(start, end, direction) {
      if (direction === void 0)
        direction = null;
      if (direction != null) {
        this[_setSelectionRange_1](start, end, direction);
        return;
      }
      this[_setSelectionRange_2](start, end);
      return;
    }
    [_setSelectionRange_1](start, end, direction) {
      return wrap_jso(this.raw.setSelectionRange(unwrap_jso(start), unwrap_jso(end), unwrap_jso(direction)));
    }
    [_setSelectionRange_2](start, end) {
      return wrap_jso(this.raw.setSelectionRange(unwrap_jso(start), unwrap_jso(end)));
    }
    stepDown(n) {
      if (n === void 0)
        n = null;
      if (n != null) {
        this[_stepDown_1](n);
        return;
      }
      this[_stepDown_2]();
      return;
    }
    [_stepDown_1](n) {
      return wrap_jso(this.raw.stepDown(unwrap_jso(n)));
    }
    [_stepDown_2]() {
      return wrap_jso(this.raw.stepDown());
    }
    stepUp(n) {
      if (n === void 0)
        n = null;
      if (n != null) {
        this[_stepUp_1](n);
        return;
      }
      this[_stepUp_2]();
      return;
    }
    [_stepUp_1](n) {
      return wrap_jso(this.raw.stepUp(unwrap_jso(n)));
    }
    [_stepUp_2]() {
      return wrap_jso(this.raw.stepUp());
    }
  }
  InputElement[dart.implements] = () => [HiddenInputElement, SearchInputElement, TextInputElement, UrlInputElement, TelephoneInputElement, EmailInputElement, PasswordInputElement, DateInputElement, MonthInputElement, WeekInputElement, TimeInputElement, LocalDateTimeInputElement, NumberInputElement, RangeInputElement, CheckboxInputElement, RadioButtonInputElement, FileUploadInputElement, SubmitButtonInputElement, ImageButtonInputElement, ResetButtonInputElement, ButtonInputElement];
  dart.defineNamedConstructor(InputElement, 'internal_');
  dart.defineNamedConstructor(InputElement, 'created');
  dart.setSignature(InputElement, {
    constructors: () => ({
      new: [InputElement, [], {type: core.String}],
      _: [InputElement, []],
      _internalWrap: [InputElement, []],
      internal_: [InputElement, []],
      created: [InputElement, []]
    }),
    methods: () => ({
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      select: [dart.void, []],
      [_select_1]: [dart.void, []],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]],
      setRangeText: [dart.void, [core.String], {start: core.int, end: core.int, selectionMode: core.String}],
      [_setRangeText_1]: [dart.void, [dart.dynamic]],
      [_setRangeText_2]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_setRangeText_3]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      setSelectionRange: [dart.void, [core.int, core.int], [core.String]],
      [_setSelectionRange_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_setSelectionRange_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      stepDown: [dart.void, [], [core.int]],
      [_stepDown_1]: [dart.void, [dart.dynamic]],
      [_stepDown_2]: [dart.void, []],
      stepUp: [dart.void, [], [core.int]],
      [_stepUp_1]: [dart.void, [dart.dynamic]],
      [_stepUp_2]: [dart.void, []]
    }),
    statics: () => ({internalCreateInputElement: [InputElement, []]}),
    names: ['internalCreateInputElement']
  });
  InputElement[dart.metadata] = () => [dart.const(new _metadata.DomName('HTMLInputElement')), dart.const(new _js_helper.Native("HTMLInputElement"))];
  class InputElementBase extends core.Object {
    InputElementBase() {
      this.autofocus = null;
      this.disabled = null;
      this.incremental = null;
      this.indeterminate = null;
      this.name = null;
      this.value = null;
    }
  }
  InputElementBase[dart.implements] = () => [Element];
  class HiddenInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'hidden'});
    }
  }
  HiddenInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(HiddenInputElement, {
    constructors: () => ({new: [HiddenInputElement, []]})
  });
  class TextInputElementBase extends core.Object {
    TextInputElementBase() {
      this.autocomplete = null;
      this.maxLength = null;
      this.pattern = null;
      this.placeholder = null;
      this.readOnly = null;
      this.required = null;
      this.size = null;
      this.selectionDirection = null;
      this.selectionEnd = null;
      this.selectionStart = null;
    }
  }
  TextInputElementBase[dart.implements] = () => [InputElementBase];
  class SearchInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'search'});
    }
    static get supported() {
      return InputElement.new({type: 'search'}).type == 'search';
    }
  }
  SearchInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(SearchInputElement, {
    constructors: () => ({new: [SearchInputElement, []]})
  });
  SearchInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI))];
  class TextInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'text'});
    }
  }
  TextInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(TextInputElement, {
    constructors: () => ({new: [TextInputElement, []]})
  });
  class UrlInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'url'});
    }
    static get supported() {
      return InputElement.new({type: 'url'}).type == 'url';
    }
  }
  UrlInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(UrlInputElement, {
    constructors: () => ({new: [UrlInputElement, []]})
  });
  UrlInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI))];
  class TelephoneInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'tel'});
    }
    static get supported() {
      return InputElement.new({type: 'tel'}).type == 'tel';
    }
  }
  TelephoneInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(TelephoneInputElement, {
    constructors: () => ({new: [TelephoneInputElement, []]})
  });
  TelephoneInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI))];
  class EmailInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'email'});
    }
    static get supported() {
      return InputElement.new({type: 'email'}).type == 'email';
    }
  }
  EmailInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(EmailInputElement, {
    constructors: () => ({new: [EmailInputElement, []]})
  });
  EmailInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI))];
  class PasswordInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'password'});
    }
  }
  PasswordInputElement[dart.implements] = () => [TextInputElementBase];
  dart.setSignature(PasswordInputElement, {
    constructors: () => ({new: [PasswordInputElement, []]})
  });
  class RangeInputElementBase extends core.Object {
    RangeInputElementBase() {
      this.max = null;
      this.min = null;
      this.step = null;
      this.valueAsNumber = null;
    }
  }
  RangeInputElementBase[dart.implements] = () => [InputElementBase];
  class DateInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'date'});
    }
    static get supported() {
      return InputElement.new({type: 'date'}).type == 'date';
    }
  }
  DateInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(DateInputElement, {
    constructors: () => ({new: [DateInputElement, []]})
  });
  DateInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental())];
  class MonthInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'month'});
    }
    static get supported() {
      return InputElement.new({type: 'month'}).type == 'month';
    }
  }
  MonthInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(MonthInputElement, {
    constructors: () => ({new: [MonthInputElement, []]})
  });
  MonthInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental())];
  class WeekInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'week'});
    }
    static get supported() {
      return InputElement.new({type: 'week'}).type == 'week';
    }
  }
  WeekInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(WeekInputElement, {
    constructors: () => ({new: [WeekInputElement, []]})
  });
  WeekInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental())];
  class TimeInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'time'});
    }
    static get supported() {
      return InputElement.new({type: 'time'}).type == 'time';
    }
  }
  TimeInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(TimeInputElement, {
    constructors: () => ({new: [TimeInputElement, []]})
  });
  TimeInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental())];
  class LocalDateTimeInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'datetime-local'});
    }
    static get supported() {
      return InputElement.new({type: 'datetime-local'}).type == 'datetime-local';
    }
  }
  LocalDateTimeInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(LocalDateTimeInputElement, {
    constructors: () => ({new: [LocalDateTimeInputElement, []]})
  });
  LocalDateTimeInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental())];
  class NumberInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'number'});
    }
    static get supported() {
      return InputElement.new({type: 'number'}).type == 'number';
    }
  }
  NumberInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(NumberInputElement, {
    constructors: () => ({new: [NumberInputElement, []]})
  });
  NumberInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Experimental())];
  class RangeInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'range'});
    }
    static get supported() {
      return InputElement.new({type: 'range'}).type == 'range';
    }
  }
  RangeInputElement[dart.implements] = () => [RangeInputElementBase];
  dart.setSignature(RangeInputElement, {
    constructors: () => ({new: [RangeInputElement, []]})
  });
  RangeInputElement[dart.metadata] = () => [dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.Experimental())];
  class CheckboxInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'checkbox'});
    }
  }
  CheckboxInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(CheckboxInputElement, {
    constructors: () => ({new: [CheckboxInputElement, []]})
  });
  class RadioButtonInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'radio'});
    }
  }
  RadioButtonInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(RadioButtonInputElement, {
    constructors: () => ({new: [RadioButtonInputElement, []]})
  });
  class FileUploadInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'file'});
    }
  }
  FileUploadInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(FileUploadInputElement, {
    constructors: () => ({new: [FileUploadInputElement, []]})
  });
  class SubmitButtonInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'submit'});
    }
  }
  SubmitButtonInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(SubmitButtonInputElement, {
    constructors: () => ({new: [SubmitButtonInputElement, []]})
  });
  class ImageButtonInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'image'});
    }
  }
  ImageButtonInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(ImageButtonInputElement, {
    constructors: () => ({new: [ImageButtonInputElement, []]})
  });
  class ResetButtonInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'reset'});
    }
  }
  ResetButtonInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(ResetButtonInputElement, {
    constructors: () => ({new: [ResetButtonInputElement, []]})
  });
  class ButtonInputElement extends core.Object {
    static new() {
      return InputElement.new({type: 'button'});
    }
  }
  ButtonInputElement[dart.implements] = () => [InputElementBase];
  dart.setSignature(ButtonInputElement, {
    constructors: () => ({new: [ButtonInputElement, []]})
  });
  const _confirmComposition_1 = Symbol('_confirmComposition_1');
  class InputMethodContext extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateInputMethodContext() {
      return InputMethodContext._internalWrap();
    }
    static _internalWrap() {
      return new InputMethodContext.internal_();
    }
    internal_() {
      super.internal_();
    }
    get compositionEndOffset() {
      return dart.as(wrap_jso(this.raw.compositionEndOffset), core.int);
    }
    get compositionStartOffset() {
      return dart.as(wrap_jso(this.raw.compositionStartOffset), core.int);
    }
    get locale() {
      return dart.as(wrap_jso(this.raw.locale), core.String);
    }
    get target() {
      return dart.as(wrap_jso(this.raw.target), HtmlElement);
    }
    confirmComposition() {
      this[_confirmComposition_1]();
      return;
    }
    [_confirmComposition_1]() {
      return wrap_jso(this.raw.confirmComposition());
    }
  }
  dart.defineNamedConstructor(InputMethodContext, 'internal_');
  dart.setSignature(InputMethodContext, {
    constructors: () => ({
      _: [InputMethodContext, []],
      _internalWrap: [InputMethodContext, []],
      internal_: [InputMethodContext, []]
    }),
    methods: () => ({
      confirmComposition: [dart.void, []],
      [_confirmComposition_1]: [dart.void, []]
    }),
    statics: () => ({internalCreateInputMethodContext: [InputMethodContext, []]}),
    names: ['internalCreateInputMethodContext']
  });
  InputMethodContext[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('InputMethodContext')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("InputMethodContext"))];
  KeyboardEvent.DOM_KEY_LOCATION_LEFT = 1;
  KeyboardEvent.DOM_KEY_LOCATION_NUMPAD = 3;
  KeyboardEvent.DOM_KEY_LOCATION_RIGHT = 2;
  KeyboardEvent.DOM_KEY_LOCATION_STANDARD = 0;
  class LIElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("li"), LIElement);
    }
    static internalCreateLIElement() {
      return LIElement._internalWrap();
    }
    static _internalWrap() {
      return new LIElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.int);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(LIElement, 'internal_');
  dart.defineNamedConstructor(LIElement, 'created');
  dart.setSignature(LIElement, {
    constructors: () => ({
      _: [LIElement, []],
      new: [LIElement, []],
      _internalWrap: [LIElement, []],
      internal_: [LIElement, []],
      created: [LIElement, []]
    }),
    statics: () => ({internalCreateLIElement: [LIElement, []]}),
    names: ['internalCreateLIElement']
  });
  LIElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLLIElement')), dart.const(new _js_helper.Native("HTMLLIElement"))];
  class LabelElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("label"), LabelElement);
    }
    static internalCreateLabelElement() {
      return LabelElement._internalWrap();
    }
    static _internalWrap() {
      return new LabelElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get control() {
      return dart.as(wrap_jso(this.raw.control), HtmlElement);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get htmlFor() {
      return dart.as(wrap_jso(this.raw.htmlFor), core.String);
    }
    set htmlFor(val) {
      return this.raw.htmlFor = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(LabelElement, 'internal_');
  dart.defineNamedConstructor(LabelElement, 'created');
  dart.setSignature(LabelElement, {
    constructors: () => ({
      _: [LabelElement, []],
      new: [LabelElement, []],
      _internalWrap: [LabelElement, []],
      internal_: [LabelElement, []],
      created: [LabelElement, []]
    }),
    statics: () => ({internalCreateLabelElement: [LabelElement, []]}),
    names: ['internalCreateLabelElement']
  });
  LabelElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLLabelElement')), dart.const(new _js_helper.Native("HTMLLabelElement"))];
  const _assign_1 = Symbol('_assign_1');
  const _assign_2 = Symbol('_assign_2');
  const _reload_1 = Symbol('_reload_1');
  const _replace_1 = Symbol('_replace_1');
  class Location extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateLocation() {
      return Location._internalWrap();
    }
    static _internalWrap() {
      return new Location.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get ancestorOrigins() {
      return dart.as(wrap_jso(this.raw.ancestorOrigins), core.List$(core.String));
    }
    get hash() {
      return dart.as(wrap_jso(this.raw.hash), core.String);
    }
    set hash(val) {
      return this.raw.hash = unwrap_jso(val);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), core.String);
    }
    set host(val) {
      return this.raw.host = unwrap_jso(val);
    }
    get hostname() {
      return dart.as(wrap_jso(this.raw.hostname), core.String);
    }
    set hostname(val) {
      return this.raw.hostname = unwrap_jso(val);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    set href(val) {
      return this.raw.href = unwrap_jso(val);
    }
    get pathname() {
      return dart.as(wrap_jso(this.raw.pathname), core.String);
    }
    set pathname(val) {
      return this.raw.pathname = unwrap_jso(val);
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), core.String);
    }
    set port(val) {
      return this.raw.port = unwrap_jso(val);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    set protocol(val) {
      return this.raw.protocol = unwrap_jso(val);
    }
    get search() {
      return dart.as(wrap_jso(this.raw.search), core.String);
    }
    set search(val) {
      return this.raw.search = unwrap_jso(val);
    }
    assign(url) {
      if (url === void 0)
        url = null;
      if (url != null) {
        this[_assign_1](url);
        return;
      }
      this[_assign_2]();
      return;
    }
    [_assign_1](url) {
      return wrap_jso(this.raw.assign(unwrap_jso(url)));
    }
    [_assign_2]() {
      return wrap_jso(this.raw.assign());
    }
    reload() {
      this[_reload_1]();
      return;
    }
    [_reload_1]() {
      return wrap_jso(this.raw.reload());
    }
    replace(url) {
      this[_replace_1](url);
      return;
    }
    [_replace_1](url) {
      return wrap_jso(this.raw.replace(unwrap_jso(url)));
    }
  }
  Location[dart.implements] = () => [LocationBase];
  dart.defineNamedConstructor(Location, 'internal_');
  dart.setSignature(Location, {
    constructors: () => ({
      _: [Location, []],
      _internalWrap: [Location, []],
      internal_: [Location, []]
    }),
    methods: () => ({
      assign: [dart.void, [], [core.String]],
      [_assign_1]: [dart.void, [dart.dynamic]],
      [_assign_2]: [dart.void, []],
      reload: [dart.void, []],
      [_reload_1]: [dart.void, []],
      replace: [dart.void, [core.String]],
      [_replace_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateLocation: [Location, []]}),
    names: ['internalCreateLocation']
  });
  Location[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Location')), dart.const(new _js_helper.Native("Location"))];
  class MenuItemElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMenuItemElement() {
      return MenuItemElement._internalWrap();
    }
    static _internalWrap() {
      return new MenuItemElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get checked() {
      return dart.as(wrap_jso(this.raw.checked), core.bool);
    }
    set checked(val) {
      return this.raw.checked = unwrap_jso(val);
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.default), core.bool);
    }
    set defaultValue(val) {
      return this.raw.default = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    set label(val) {
      return this.raw.label = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(MenuItemElement, 'internal_');
  dart.defineNamedConstructor(MenuItemElement, 'created');
  dart.setSignature(MenuItemElement, {
    constructors: () => ({
      _: [MenuItemElement, []],
      _internalWrap: [MenuItemElement, []],
      internal_: [MenuItemElement, []],
      created: [MenuItemElement, []]
    }),
    statics: () => ({internalCreateMenuItemElement: [MenuItemElement, []]}),
    names: ['internalCreateMenuItemElement']
  });
  MenuItemElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLMenuItemElement')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("HTMLMenuItemElement"))];
  class MessageChannel extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMessageChannel() {
      return MessageChannel._internalWrap();
    }
    static _internalWrap() {
      return new MessageChannel.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get port1() {
      return dart.as(wrap_jso(this.raw.port1), MessagePort);
    }
    get port2() {
      return dart.as(wrap_jso(this.raw.port2), MessagePort);
    }
  }
  dart.defineNamedConstructor(MessageChannel, 'internal_');
  dart.setSignature(MessageChannel, {
    constructors: () => ({
      _: [MessageChannel, []],
      _internalWrap: [MessageChannel, []],
      internal_: [MessageChannel, []]
    }),
    statics: () => ({internalCreateMessageChannel: [MessageChannel, []]}),
    names: ['internalCreateMessageChannel']
  });
  MessageChannel[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MessageChannel')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("MessageChannel"))];
  const _close_1 = Symbol('_close_1');
  const _postMessage_1 = Symbol('_postMessage_1');
  const _postMessage_2 = Symbol('_postMessage_2');
  const _start_1 = Symbol('_start_1');
  class MessagePort extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMessagePort() {
      return MessagePort._internalWrap();
    }
    static _internalWrap() {
      return new MessagePort.internal_();
    }
    internal_() {
      super.internal_();
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    postMessage(message, transfer) {
      if (transfer === void 0)
        transfer = null;
      if (transfer != null) {
        let message_1 = html_common.convertDartToNative_SerializedScriptValue(message);
        this[_postMessage_1](message_1, transfer);
        return;
      }
      let message_1 = html_common.convertDartToNative_SerializedScriptValue(message);
      this[_postMessage_2](message_1);
      return;
    }
    [_postMessage_1](message, transfer) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message), unwrap_jso(transfer)));
    }
    [_postMessage_2](message) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message)));
    }
    start() {
      this[_start_1]();
      return;
    }
    [_start_1]() {
      return wrap_jso(this.raw.start());
    }
    get onMessage() {
      return MessagePort.messageEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(MessagePort, 'internal_');
  dart.setSignature(MessagePort, {
    constructors: () => ({
      _: [MessagePort, []],
      _internalWrap: [MessagePort, []],
      internal_: [MessagePort, []]
    }),
    methods: () => ({
      close: [dart.void, []],
      [_close_1]: [dart.void, []],
      postMessage: [dart.void, [dart.dynamic], [core.List$(MessagePort)]],
      [_postMessage_1]: [dart.void, [dart.dynamic, core.List$(MessagePort)]],
      [_postMessage_2]: [dart.void, [dart.dynamic]],
      start: [dart.void, []],
      [_start_1]: [dart.void, []]
    }),
    statics: () => ({internalCreateMessagePort: [MessagePort, []]}),
    names: ['internalCreateMessagePort']
  });
  MessagePort[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MessagePort')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("MessagePort"))];
  MessagePort.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  class MetaElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("meta"), MetaElement);
    }
    static internalCreateMetaElement() {
      return MetaElement._internalWrap();
    }
    static _internalWrap() {
      return new MetaElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get content() {
      return dart.as(wrap_jso(this.raw.content), core.String);
    }
    set content(val) {
      return this.raw.content = unwrap_jso(val);
    }
    get httpEquiv() {
      return dart.as(wrap_jso(this.raw.httpEquiv), core.String);
    }
    set httpEquiv(val) {
      return this.raw.httpEquiv = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(MetaElement, 'internal_');
  dart.defineNamedConstructor(MetaElement, 'created');
  dart.setSignature(MetaElement, {
    constructors: () => ({
      _: [MetaElement, []],
      new: [MetaElement, []],
      _internalWrap: [MetaElement, []],
      internal_: [MetaElement, []],
      created: [MetaElement, []]
    }),
    statics: () => ({internalCreateMetaElement: [MetaElement, []]}),
    names: ['internalCreateMetaElement']
  });
  MetaElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLMetaElement')), dart.const(new _js_helper.Native("HTMLMetaElement"))];
  const _get_modificationTime = Symbol('_get_modificationTime');
  class Metadata extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMetadata() {
      return Metadata._internalWrap();
    }
    static _internalWrap() {
      return new Metadata.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get modificationTime() {
      return html_common.convertNativeToDart_DateTime(this[_get_modificationTime]);
    }
    get [_get_modificationTime]() {
      return wrap_jso(this.raw.modificationTime);
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
  }
  dart.defineNamedConstructor(Metadata, 'internal_');
  dart.setSignature(Metadata, {
    constructors: () => ({
      _: [Metadata, []],
      _internalWrap: [Metadata, []],
      internal_: [Metadata, []]
    }),
    statics: () => ({internalCreateMetadata: [Metadata, []]}),
    names: ['internalCreateMetadata']
  });
  Metadata[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Metadata')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Metadata"))];
  const MetadataCallback = dart.typedef('MetadataCallback', () => dart.functionType(dart.void, [Metadata]));
  class MeterElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("meter"), MeterElement);
    }
    static internalCreateMeterElement() {
      return MeterElement._internalWrap();
    }
    static _internalWrap() {
      return new MeterElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('meter');
    }
    get high() {
      return dart.as(wrap_jso(this.raw.high), core.num);
    }
    set high(val) {
      return this.raw.high = unwrap_jso(val);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get low() {
      return dart.as(wrap_jso(this.raw.low), core.num);
    }
    set low(val) {
      return this.raw.low = unwrap_jso(val);
    }
    get max() {
      return dart.as(wrap_jso(this.raw.max), core.num);
    }
    set max(val) {
      return this.raw.max = unwrap_jso(val);
    }
    get min() {
      return dart.as(wrap_jso(this.raw.min), core.num);
    }
    set min(val) {
      return this.raw.min = unwrap_jso(val);
    }
    get optimum() {
      return dart.as(wrap_jso(this.raw.optimum), core.num);
    }
    set optimum(val) {
      return this.raw.optimum = unwrap_jso(val);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.num);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(MeterElement, 'internal_');
  dart.defineNamedConstructor(MeterElement, 'created');
  dart.setSignature(MeterElement, {
    constructors: () => ({
      _: [MeterElement, []],
      new: [MeterElement, []],
      _internalWrap: [MeterElement, []],
      internal_: [MeterElement, []],
      created: [MeterElement, []]
    }),
    statics: () => ({internalCreateMeterElement: [MeterElement, []]}),
    names: ['internalCreateMeterElement']
  });
  MeterElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLMeterElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("HTMLMeterElement"))];
  class MidiAccess extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiAccess() {
      return MidiAccess._internalWrap();
    }
    static _internalWrap() {
      return new MidiAccess.internal_();
    }
    internal_() {
      super.internal_();
    }
    get inputs() {
      return dart.as(wrap_jso(this.raw.inputs), MidiInputMap);
    }
    get outputs() {
      return dart.as(wrap_jso(this.raw.outputs), MidiOutputMap);
    }
    get sysexEnabled() {
      return dart.as(wrap_jso(this.raw.sysexEnabled), core.bool);
    }
    get onConnect() {
      return MidiAccess.connectEvent.forTarget(this);
    }
    get onDisconnect() {
      return MidiAccess.disconnectEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(MidiAccess, 'internal_');
  dart.setSignature(MidiAccess, {
    constructors: () => ({
      _: [MidiAccess, []],
      _internalWrap: [MidiAccess, []],
      internal_: [MidiAccess, []]
    }),
    statics: () => ({internalCreateMidiAccess: [MidiAccess, []]}),
    names: ['internalCreateMidiAccess']
  });
  MidiAccess[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIAccess')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIAccess"))];
  class MidiConnectionEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiConnectionEvent() {
      return MidiConnectionEvent._internalWrap();
    }
    static _internalWrap() {
      return new MidiConnectionEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), MidiPort);
    }
  }
  dart.defineNamedConstructor(MidiConnectionEvent, 'internal_');
  dart.setSignature(MidiConnectionEvent, {
    constructors: () => ({
      _: [MidiConnectionEvent, []],
      _internalWrap: [MidiConnectionEvent, []],
      internal_: [MidiConnectionEvent, []]
    }),
    statics: () => ({internalCreateMidiConnectionEvent: [MidiConnectionEvent, []]}),
    names: ['internalCreateMidiConnectionEvent']
  });
  MidiConnectionEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIConnectionEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIConnectionEvent"))];
  MidiAccess.connectEvent = dart.const(new (EventStreamProvider$(MidiConnectionEvent))('connect'));
  MidiAccess.disconnectEvent = dart.const(new (EventStreamProvider$(MidiConnectionEvent))('disconnect'));
  class MidiPort extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiPort() {
      return MidiPort._internalWrap();
    }
    static _internalWrap() {
      return new MidiPort.internal_();
    }
    internal_() {
      super.internal_();
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    get manufacturer() {
      return dart.as(wrap_jso(this.raw.manufacturer), core.String);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    get version() {
      return dart.as(wrap_jso(this.raw.version), core.String);
    }
    get onDisconnect() {
      return MidiPort.disconnectEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(MidiPort, 'internal_');
  dart.setSignature(MidiPort, {
    constructors: () => ({
      _: [MidiPort, []],
      _internalWrap: [MidiPort, []],
      internal_: [MidiPort, []]
    }),
    statics: () => ({internalCreateMidiPort: [MidiPort, []]}),
    names: ['internalCreateMidiPort']
  });
  MidiPort[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIPort')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIPort"))];
  class MidiInput extends MidiPort {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiInput() {
      return MidiInput._internalWrap();
    }
    static _internalWrap() {
      return new MidiInput.internal_();
    }
    internal_() {
      super.internal_();
    }
    get onMidiMessage() {
      return MidiInput.midiMessageEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(MidiInput, 'internal_');
  dart.setSignature(MidiInput, {
    constructors: () => ({
      _: [MidiInput, []],
      _internalWrap: [MidiInput, []],
      internal_: [MidiInput, []]
    }),
    statics: () => ({internalCreateMidiInput: [MidiInput, []]}),
    names: ['internalCreateMidiInput']
  });
  MidiInput[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIInput')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIInput"))];
  class MidiMessageEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiMessageEvent() {
      return MidiMessageEvent._internalWrap();
    }
    static _internalWrap() {
      return new MidiMessageEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), typed_data.Uint8List);
    }
    get receivedTime() {
      return dart.as(wrap_jso(this.raw.receivedTime), core.double);
    }
  }
  dart.defineNamedConstructor(MidiMessageEvent, 'internal_');
  dart.setSignature(MidiMessageEvent, {
    constructors: () => ({
      _: [MidiMessageEvent, []],
      _internalWrap: [MidiMessageEvent, []],
      internal_: [MidiMessageEvent, []]
    }),
    statics: () => ({internalCreateMidiMessageEvent: [MidiMessageEvent, []]}),
    names: ['internalCreateMidiMessageEvent']
  });
  MidiMessageEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIMessageEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIMessageEvent"))];
  MidiInput.midiMessageEvent = dart.const(new (EventStreamProvider$(MidiMessageEvent))('midimessage'));
  const _entries_1 = Symbol('_entries_1');
  const _get_1 = Symbol('_get_1');
  const _has_1 = Symbol('_has_1');
  const _keys_1 = Symbol('_keys_1');
  const _values_1 = Symbol('_values_1');
  class MidiInputMap extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiInputMap() {
      return MidiInputMap._internalWrap();
    }
    static _internalWrap() {
      return new MidiInputMap.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
    entries() {
      return this[_entries_1]();
    }
    [_entries_1]() {
      return wrap_jso(this.raw.entries());
    }
    get(id) {
      return this[_get_1](id);
    }
    [_get_1](id) {
      return wrap_jso(this.raw.get(unwrap_jso(id)));
    }
    has(key) {
      return this[_has_1](key);
    }
    [_has_1](key) {
      return dart.as(wrap_jso(this.raw.has(unwrap_jso(key))), core.bool);
    }
    keys() {
      return this[_keys_1]();
    }
    [_keys_1]() {
      return wrap_jso(this.raw.keys());
    }
    values() {
      return this[_values_1]();
    }
    [_values_1]() {
      return wrap_jso(this.raw.values());
    }
  }
  dart.defineNamedConstructor(MidiInputMap, 'internal_');
  dart.setSignature(MidiInputMap, {
    constructors: () => ({
      _: [MidiInputMap, []],
      _internalWrap: [MidiInputMap, []],
      internal_: [MidiInputMap, []]
    }),
    methods: () => ({
      entries: [dart.dynamic, []],
      [_entries_1]: [dart.dynamic, []],
      get: [core.Object, [core.String]],
      [_get_1]: [core.Object, [dart.dynamic]],
      has: [core.bool, [core.String]],
      [_has_1]: [core.bool, [dart.dynamic]],
      keys: [dart.dynamic, []],
      [_keys_1]: [dart.dynamic, []],
      values: [dart.dynamic, []],
      [_values_1]: [dart.dynamic, []]
    }),
    statics: () => ({internalCreateMidiInputMap: [MidiInputMap, []]}),
    names: ['internalCreateMidiInputMap']
  });
  MidiInputMap[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIInputMap')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIInputMap"))];
  class MidiOutput extends MidiPort {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiOutput() {
      return MidiOutput._internalWrap();
    }
    static _internalWrap() {
      return new MidiOutput.internal_();
    }
    internal_() {
      super.internal_();
    }
    send(data, timestamp) {
      if (timestamp === void 0)
        timestamp = null;
      if (timestamp != null) {
        this[_send_1](data, timestamp);
        return;
      }
      this[_send_2](data);
      return;
    }
    [_send_1](data, timestamp) {
      return wrap_jso(this.raw.send(unwrap_jso(data), unwrap_jso(timestamp)));
    }
    [_send_2](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
  }
  dart.defineNamedConstructor(MidiOutput, 'internal_');
  dart.setSignature(MidiOutput, {
    constructors: () => ({
      _: [MidiOutput, []],
      _internalWrap: [MidiOutput, []],
      internal_: [MidiOutput, []]
    }),
    methods: () => ({
      send: [dart.void, [typed_data.Uint8List], [core.num]],
      [_send_1]: [dart.void, [typed_data.Uint8List, dart.dynamic]],
      [_send_2]: [dart.void, [typed_data.Uint8List]]
    }),
    statics: () => ({internalCreateMidiOutput: [MidiOutput, []]}),
    names: ['internalCreateMidiOutput']
  });
  MidiOutput[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIOutput')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIOutput"))];
  class MidiOutputMap extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMidiOutputMap() {
      return MidiOutputMap._internalWrap();
    }
    static _internalWrap() {
      return new MidiOutputMap.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
    entries() {
      return this[_entries_1]();
    }
    [_entries_1]() {
      return wrap_jso(this.raw.entries());
    }
    get(id) {
      return this[_get_1](id);
    }
    [_get_1](id) {
      return wrap_jso(this.raw.get(unwrap_jso(id)));
    }
    has(key) {
      return this[_has_1](key);
    }
    [_has_1](key) {
      return dart.as(wrap_jso(this.raw.has(unwrap_jso(key))), core.bool);
    }
    keys() {
      return this[_keys_1]();
    }
    [_keys_1]() {
      return wrap_jso(this.raw.keys());
    }
    values() {
      return this[_values_1]();
    }
    [_values_1]() {
      return wrap_jso(this.raw.values());
    }
  }
  dart.defineNamedConstructor(MidiOutputMap, 'internal_');
  dart.setSignature(MidiOutputMap, {
    constructors: () => ({
      _: [MidiOutputMap, []],
      _internalWrap: [MidiOutputMap, []],
      internal_: [MidiOutputMap, []]
    }),
    methods: () => ({
      entries: [dart.dynamic, []],
      [_entries_1]: [dart.dynamic, []],
      get: [core.Object, [core.String]],
      [_get_1]: [core.Object, [dart.dynamic]],
      has: [core.bool, [core.String]],
      [_has_1]: [core.bool, [dart.dynamic]],
      keys: [dart.dynamic, []],
      [_keys_1]: [dart.dynamic, []],
      values: [dart.dynamic, []],
      [_values_1]: [dart.dynamic, []]
    }),
    statics: () => ({internalCreateMidiOutputMap: [MidiOutputMap, []]}),
    names: ['internalCreateMidiOutputMap']
  });
  MidiOutputMap[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MIDIOutputMap')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MIDIOutputMap"))];
  MidiPort.disconnectEvent = dart.const(new (EventStreamProvider$(MidiConnectionEvent))('disconnect'));
  class MimeType extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMimeType() {
      return MimeType._internalWrap();
    }
    static _internalWrap() {
      return new MimeType.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get description() {
      return dart.as(wrap_jso(this.raw.description), core.String);
    }
    get enabledPlugin() {
      return dart.as(wrap_jso(this.raw.enabledPlugin), Plugin);
    }
    get suffixes() {
      return dart.as(wrap_jso(this.raw.suffixes), core.String);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
  }
  dart.defineNamedConstructor(MimeType, 'internal_');
  dart.setSignature(MimeType, {
    constructors: () => ({
      _: [MimeType, []],
      _internalWrap: [MimeType, []],
      internal_: [MimeType, []]
    }),
    statics: () => ({internalCreateMimeType: [MimeType, []]}),
    names: ['internalCreateMimeType']
  });
  MimeType[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MimeType')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MimeType"))];
  class MimeTypeArray extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(MimeType), ImmutableListMixin$(MimeType)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMimeTypeArray() {
      return MimeTypeArray._internalWrap();
    }
    static _internalWrap() {
      return new MimeTypeArray.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), MimeType);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), MimeType);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), MimeType);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), MimeType);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), MimeType);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), MimeType);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), MimeType);
    }
  }
  MimeTypeArray[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(MimeType)];
  dart.defineNamedConstructor(MimeTypeArray, 'internal_');
  dart.setSignature(MimeTypeArray, {
    constructors: () => ({
      _: [MimeTypeArray, []],
      _internalWrap: [MimeTypeArray, []],
      internal_: [MimeTypeArray, []]
    }),
    methods: () => ({
      get: [MimeType, [core.int]],
      set: [dart.void, [core.int, MimeType]],
      elementAt: [MimeType, [core.int]],
      [__getter__]: [MimeType, [core.String]],
      [__getter___1]: [MimeType, [dart.dynamic]],
      item: [MimeType, [core.int]],
      [_item_1]: [MimeType, [dart.dynamic]],
      namedItem: [MimeType, [core.String]],
      [_namedItem_1]: [MimeType, [dart.dynamic]]
    }),
    statics: () => ({internalCreateMimeTypeArray: [MimeTypeArray, []]}),
    names: ['internalCreateMimeTypeArray']
  });
  dart.defineExtensionMembers(MimeTypeArray, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  MimeTypeArray[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MimeTypeArray')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MimeTypeArray"))];
  class ModElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateModElement() {
      return ModElement._internalWrap();
    }
    static _internalWrap() {
      return new ModElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get cite() {
      return dart.as(wrap_jso(this.raw.cite), core.String);
    }
    set cite(val) {
      return this.raw.cite = unwrap_jso(val);
    }
    get dateTime() {
      return dart.as(wrap_jso(this.raw.dateTime), core.String);
    }
    set dateTime(val) {
      return this.raw.dateTime = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(ModElement, 'internal_');
  dart.defineNamedConstructor(ModElement, 'created');
  dart.setSignature(ModElement, {
    constructors: () => ({
      _: [ModElement, []],
      _internalWrap: [ModElement, []],
      internal_: [ModElement, []],
      created: [ModElement, []]
    }),
    statics: () => ({internalCreateModElement: [ModElement, []]}),
    names: ['internalCreateModElement']
  });
  ModElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLModElement')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("HTMLModElement"))];
  const MutationCallback = dart.typedef('MutationCallback', () => dart.functionType(dart.void, [core.List$(MutationRecord), MutationObserver]));
  const _disconnect_1 = Symbol('_disconnect_1');
  const _observe_1 = Symbol('_observe_1');
  const _observe = Symbol('_observe');
  const _takeRecords_1 = Symbol('_takeRecords_1');
  class MutationObserver extends DartHtmlDomObject {
    static internalCreateMutationObserver() {
      return MutationObserver._internalWrap();
    }
    static _internalWrap() {
      return new MutationObserver.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    disconnect() {
      this[_disconnect_1]();
      return;
    }
    [_disconnect_1]() {
      return wrap_jso(this.raw.disconnect());
    }
    [_observe](target, options) {
      let options_1 = html_common.convertDartToNative_Dictionary(options);
      this[_observe_1](target, options_1);
      return;
    }
    [_observe_1](target, options) {
      return wrap_jso(this.raw.observe(unwrap_jso(target), unwrap_jso(options)));
    }
    takeRecords() {
      return this[_takeRecords_1]();
    }
    [_takeRecords_1]() {
      return dart.as(wrap_jso(this.raw.takeRecords()), core.List$(MutationRecord));
    }
    static get supported() {
      return !!(window.MutationObserver || window.WebKitMutationObserver);
    }
    observe(target, opts) {
      let childList = opts && 'childList' in opts ? opts.childList : null;
      let attributes = opts && 'attributes' in opts ? opts.attributes : null;
      let characterData = opts && 'characterData' in opts ? opts.characterData : null;
      let subtree = opts && 'subtree' in opts ? opts.subtree : null;
      let attributeOldValue = opts && 'attributeOldValue' in opts ? opts.attributeOldValue : null;
      let characterDataOldValue = opts && 'characterDataOldValue' in opts ? opts.characterDataOldValue : null;
      let attributeFilter = opts && 'attributeFilter' in opts ? opts.attributeFilter : null;
      let parsedOptions = dart.dcall(/* Unimplemented unknown name */_createDict);
      function override(key, value) {
        if (value != null)
          dart.dcall(/* Unimplemented unknown name */_add, parsedOptions, key, value);
      }
      dart.fn(override);
      override('childList', childList);
      override('attributes', attributes);
      override('characterData', characterData);
      override('subtree', subtree);
      override('attributeOldValue', attributeOldValue);
      override('characterDataOldValue', characterDataOldValue);
      if (attributeFilter != null) {
        override('attributeFilter', dart.dcall(/* Unimplemented unknown name */_fixupList, attributeFilter));
      }
      dart.dcall(/* Unimplemented unknown name */_call, target, parsedOptions);
    }
    static new(callback) {
      return new MutationObserver._(_wrapBinaryZone(callback));
    }
  }
  dart.defineNamedConstructor(MutationObserver, 'internal_');
  dart.setSignature(MutationObserver, {
    constructors: () => ({
      _internalWrap: [MutationObserver, []],
      internal_: [MutationObserver, []],
      new: [MutationObserver, [MutationCallback]]
    }),
    methods: () => ({
      disconnect: [dart.void, []],
      [_disconnect_1]: [dart.void, []],
      [_observe]: [dart.void, [Node, core.Map]],
      [_observe_1]: [dart.void, [Node, dart.dynamic]],
      takeRecords: [core.List$(MutationRecord), []],
      [_takeRecords_1]: [core.List$(MutationRecord), []],
      observe: [dart.void, [Node], {childList: core.bool, attributes: core.bool, characterData: core.bool, subtree: core.bool, attributeOldValue: core.bool, characterDataOldValue: core.bool, attributeFilter: core.List$(core.String)}]
    }),
    statics: () => ({internalCreateMutationObserver: [MutationObserver, []]}),
    names: ['internalCreateMutationObserver']
  });
  MutationObserver[dart.metadata] = () => [dart.const(new _metadata.DomName('MutationObserver')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("MutationObserver,WebKitMutationObserver"))];
  MutationObserver._boolKeys = dart.const(dart.map({childList: true, attributes: true, characterData: true, subtree: true, attributeOldValue: true, characterDataOldValue: true}));
  class MutationRecord extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateMutationRecord() {
      return MutationRecord._internalWrap();
    }
    static _internalWrap() {
      return new MutationRecord.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get addedNodes() {
      return dart.as(wrap_jso(this.raw.addedNodes), core.List$(Node));
    }
    get attributeName() {
      return dart.as(wrap_jso(this.raw.attributeName), core.String);
    }
    get attributeNamespace() {
      return dart.as(wrap_jso(this.raw.attributeNamespace), core.String);
    }
    get nextSibling() {
      return dart.as(wrap_jso(this.raw.nextSibling), Node);
    }
    get oldValue() {
      return dart.as(wrap_jso(this.raw.oldValue), core.String);
    }
    get previousSibling() {
      return dart.as(wrap_jso(this.raw.previousSibling), Node);
    }
    get removedNodes() {
      return dart.as(wrap_jso(this.raw.removedNodes), core.List$(Node));
    }
    get target() {
      return dart.as(wrap_jso(this.raw.target), Node);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
  }
  dart.defineNamedConstructor(MutationRecord, 'internal_');
  dart.setSignature(MutationRecord, {
    constructors: () => ({
      _: [MutationRecord, []],
      _internalWrap: [MutationRecord, []],
      internal_: [MutationRecord, []]
    }),
    statics: () => ({internalCreateMutationRecord: [MutationRecord, []]}),
    names: ['internalCreateMutationRecord']
  });
  MutationRecord[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('MutationRecord')), dart.const(new _js_helper.Native("MutationRecord"))];
  const _getUserMedia = Symbol('_getUserMedia');
  const _getBattery_1 = Symbol('_getBattery_1');
  const _getGamepads_1 = Symbol('_getGamepads_1');
  const _getStorageUpdates_1 = Symbol('_getStorageUpdates_1');
  const _registerProtocolHandler_1 = Symbol('_registerProtocolHandler_1');
  const _sendBeacon_1 = Symbol('_sendBeacon_1');
  const _sendBeacon_2 = Symbol('_sendBeacon_2');
  const _sendBeacon_3 = Symbol('_sendBeacon_3');
  const _sendBeacon_4 = Symbol('_sendBeacon_4');
  class Navigator extends DartHtmlDomObject {
    getUserMedia(opts) {
      let audio = opts && 'audio' in opts ? opts.audio : false;
      let video = opts && 'video' in opts ? opts.video : false;
      let completer = async.Completer$(dart.dynamic).new();
      let options = dart.map({audio: audio, video: video});
      dart.dcall(this[_getUserMedia], options, dart.fn(stream => {
        completer.complete(stream);
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNavigator() {
      return Navigator._internalWrap();
    }
    static _internalWrap() {
      return new Navigator.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get connection() {
      return dart.as(wrap_jso(this.raw.connection), NetworkInformation);
    }
    get cookieEnabled() {
      return dart.as(wrap_jso(this.raw.cookieEnabled), core.bool);
    }
    get credentials() {
      return wrap_jso(this.raw.credentials);
    }
    get doNotTrack() {
      return dart.as(wrap_jso(this.raw.doNotTrack), core.String);
    }
    get geofencing() {
      return wrap_jso(this.raw.geofencing);
    }
    get geolocation() {
      return wrap_jso(this.raw.geolocation);
    }
    get maxTouchPoints() {
      return dart.as(wrap_jso(this.raw.maxTouchPoints), core.int);
    }
    get mimeTypes() {
      return dart.as(wrap_jso(this.raw.mimeTypes), MimeTypeArray);
    }
    get presentation() {
      return dart.as(wrap_jso(this.raw.presentation), Presentation);
    }
    get productSub() {
      return dart.as(wrap_jso(this.raw.productSub), core.String);
    }
    get push() {
      return dart.as(wrap_jso(this.raw.push), PushManager);
    }
    get serviceWorker() {
      return dart.as(wrap_jso(this.raw.serviceWorker), ServiceWorkerContainer);
    }
    get storageQuota() {
      return dart.as(wrap_jso(this.raw.storageQuota), StorageQuota);
    }
    get vendor() {
      return dart.as(wrap_jso(this.raw.vendor), core.String);
    }
    get vendorSub() {
      return dart.as(wrap_jso(this.raw.vendorSub), core.String);
    }
    get persistentStorage() {
      return wrap_jso(this.raw.webkitPersistentStorage);
    }
    get temporaryStorage() {
      return wrap_jso(this.raw.webkitTemporaryStorage);
    }
    getBattery() {
      return this[_getBattery_1]();
    }
    [_getBattery_1]() {
      return dart.as(wrap_jso(this.raw.getBattery()), async.Future);
    }
    getGamepads() {
      return this[_getGamepads_1]();
    }
    [_getGamepads_1]() {
      return dart.as(wrap_jso(this.raw.getGamepads()), _GamepadList);
    }
    getStorageUpdates() {
      this[_getStorageUpdates_1]();
      return;
    }
    [_getStorageUpdates_1]() {
      return wrap_jso(this.raw.getStorageUpdates());
    }
    registerProtocolHandler(scheme, url, title) {
      this[_registerProtocolHandler_1](scheme, url, title);
      return;
    }
    [_registerProtocolHandler_1](scheme, url, title) {
      return wrap_jso(this.raw.registerProtocolHandler(unwrap_jso(scheme), unwrap_jso(url), unwrap_jso(title)));
    }
    sendBeacon(url, data) {
      if (typeof data == 'string' || data == null) {
        return this[_sendBeacon_1](url, dart.as(data, core.String));
      }
      if (dart.is(data, typed_data.TypedData) || data == null) {
        return this[_sendBeacon_2](url, dart.as(data, typed_data.TypedData));
      }
      if (dart.is(data, dart.dynamic) || data == null) {
        return this[_sendBeacon_3](url, data);
      }
      if (dart.is(data, dart.dynamic) || data == null) {
        return this[_sendBeacon_4](url, data);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_sendBeacon_1](url, data) {
      return dart.as(wrap_jso(this.raw.sendBeacon(unwrap_jso(url), unwrap_jso(data))), core.bool);
    }
    [_sendBeacon_2](url, data) {
      return dart.as(wrap_jso(this.raw.sendBeacon(unwrap_jso(url), unwrap_jso(data))), core.bool);
    }
    [_sendBeacon_3](url, data) {
      return dart.as(wrap_jso(this.raw.sendBeacon(unwrap_jso(url), unwrap_jso(data))), core.bool);
    }
    [_sendBeacon_4](url, data) {
      return dart.as(wrap_jso(this.raw.sendBeacon(unwrap_jso(url), unwrap_jso(data))), core.bool);
    }
    get hardwareConcurrency() {
      return dart.as(wrap_jso(this.raw.hardwareConcurrency), core.int);
    }
    get appCodeName() {
      return dart.as(wrap_jso(this.raw.appCodeName), core.String);
    }
    get appName() {
      return dart.as(wrap_jso(this.raw.appName), core.String);
    }
    get appVersion() {
      return dart.as(wrap_jso(this.raw.appVersion), core.String);
    }
    get dartEnabled() {
      return dart.as(wrap_jso(this.raw.dartEnabled), core.bool);
    }
    get platform() {
      return dart.as(wrap_jso(this.raw.platform), core.String);
    }
    get product() {
      return dart.as(wrap_jso(this.raw.product), core.String);
    }
    get userAgent() {
      return dart.as(wrap_jso(this.raw.userAgent), core.String);
    }
    get languages() {
      return dart.as(wrap_jso(this.raw.languages), core.List$(core.String));
    }
    get onLine() {
      return dart.as(wrap_jso(this.raw.onLine), core.bool);
    }
  }
  Navigator[dart.implements] = () => [NavigatorCpu, NavigatorLanguage, NavigatorOnLine, NavigatorID];
  dart.defineNamedConstructor(Navigator, 'internal_');
  dart.setSignature(Navigator, {
    constructors: () => ({
      _: [Navigator, []],
      _internalWrap: [Navigator, []],
      internal_: [Navigator, []]
    }),
    methods: () => ({
      getUserMedia: [async.Future$(dart.dynamic), [], {audio: dart.dynamic, video: dart.dynamic}],
      getBattery: [async.Future, []],
      [_getBattery_1]: [async.Future, []],
      getGamepads: [_GamepadList, []],
      [_getGamepads_1]: [_GamepadList, []],
      getStorageUpdates: [dart.void, []],
      [_getStorageUpdates_1]: [dart.void, []],
      registerProtocolHandler: [dart.void, [core.String, core.String, core.String]],
      [_registerProtocolHandler_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      sendBeacon: [core.bool, [core.String, dart.dynamic]],
      [_sendBeacon_1]: [core.bool, [dart.dynamic, core.String]],
      [_sendBeacon_2]: [core.bool, [dart.dynamic, typed_data.TypedData]],
      [_sendBeacon_3]: [core.bool, [dart.dynamic, dart.dynamic]],
      [_sendBeacon_4]: [core.bool, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateNavigator: [Navigator, []]}),
    names: ['internalCreateNavigator']
  });
  Navigator[dart.metadata] = () => [dart.const(new _metadata.DomName('Navigator')), dart.const(new _js_helper.Native("Navigator"))];
  class NavigatorCpu extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get hardwareConcurrency() {
      return dart.as(wrap_jso(this.raw.hardwareConcurrency), core.int);
    }
  }
  dart.setSignature(NavigatorCpu, {
    constructors: () => ({_: [NavigatorCpu, []]})
  });
  NavigatorCpu[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NavigatorCPU')), dart.const(new _metadata.Experimental())];
  class NavigatorID extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get appCodeName() {
      return dart.as(wrap_jso(this.raw.appCodeName), core.String);
    }
    get appName() {
      return dart.as(wrap_jso(this.raw.appName), core.String);
    }
    get appVersion() {
      return dart.as(wrap_jso(this.raw.appVersion), core.String);
    }
    get dartEnabled() {
      return dart.as(wrap_jso(this.raw.dartEnabled), core.bool);
    }
    get platform() {
      return dart.as(wrap_jso(this.raw.platform), core.String);
    }
    get product() {
      return dart.as(wrap_jso(this.raw.product), core.String);
    }
    get userAgent() {
      return dart.as(wrap_jso(this.raw.userAgent), core.String);
    }
  }
  dart.setSignature(NavigatorID, {
    constructors: () => ({_: [NavigatorID, []]})
  });
  NavigatorID[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NavigatorID')), dart.const(new _metadata.Experimental())];
  class NavigatorLanguage extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get language() {
      return dart.as(wrap_jso(this.raw.language), core.String);
    }
    get languages() {
      return dart.as(wrap_jso(this.raw.languages), core.List$(core.String));
    }
  }
  dart.setSignature(NavigatorLanguage, {
    constructors: () => ({_: [NavigatorLanguage, []]})
  });
  NavigatorLanguage[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NavigatorLanguage')), dart.const(new _metadata.Experimental())];
  class NavigatorOnLine extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get onLine() {
      return dart.as(wrap_jso(this.raw.onLine), core.bool);
    }
  }
  dart.setSignature(NavigatorOnLine, {
    constructors: () => ({_: [NavigatorOnLine, []]})
  });
  NavigatorOnLine[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NavigatorOnLine')), dart.const(new _metadata.Experimental())];
  class NavigatorUserMediaError extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNavigatorUserMediaError() {
      return NavigatorUserMediaError._internalWrap();
    }
    static _internalWrap() {
      return new NavigatorUserMediaError.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get constraintName() {
      return dart.as(wrap_jso(this.raw.constraintName), core.String);
    }
    get message() {
      return dart.as(wrap_jso(this.raw.message), core.String);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
  }
  dart.defineNamedConstructor(NavigatorUserMediaError, 'internal_');
  dart.setSignature(NavigatorUserMediaError, {
    constructors: () => ({
      _: [NavigatorUserMediaError, []],
      _internalWrap: [NavigatorUserMediaError, []],
      internal_: [NavigatorUserMediaError, []]
    }),
    statics: () => ({internalCreateNavigatorUserMediaError: [NavigatorUserMediaError, []]}),
    names: ['internalCreateNavigatorUserMediaError']
  });
  NavigatorUserMediaError[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NavigatorUserMediaError')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("NavigatorUserMediaError"))];
  const _NavigatorUserMediaErrorCallback = dart.typedef('_NavigatorUserMediaErrorCallback', () => dart.functionType(dart.void, [NavigatorUserMediaError]));
  const _NavigatorUserMediaSuccessCallback = dart.typedef('_NavigatorUserMediaSuccessCallback', () => dart.functionType(dart.void, [dart.dynamic]));
  class NetworkInformation extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNetworkInformation() {
      return NetworkInformation._internalWrap();
    }
    static _internalWrap() {
      return new NetworkInformation.internal_();
    }
    internal_() {
      super.internal_();
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
  }
  dart.defineNamedConstructor(NetworkInformation, 'internal_');
  dart.setSignature(NetworkInformation, {
    constructors: () => ({
      _: [NetworkInformation, []],
      _internalWrap: [NetworkInformation, []],
      internal_: [NetworkInformation, []]
    }),
    statics: () => ({internalCreateNetworkInformation: [NetworkInformation, []]}),
    names: ['internalCreateNetworkInformation']
  });
  NetworkInformation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NetworkInformation')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("NetworkInformation"))];
  class _ChildNodeListLazy extends collection.ListBase$(Node) {
    _ChildNodeListLazy(this$) {
      this[_this] = this$;
    }
    get first() {
      let result = this[_this].firstChild;
      if (result == null)
        dart.throw(new core.StateError("No elements"));
      return result;
    }
    get last() {
      let result = this[_this].lastChild;
      if (result == null)
        dart.throw(new core.StateError("No elements"));
      return result;
    }
    get single() {
      let l = this.length;
      if (l == 0)
        dart.throw(new core.StateError("No elements"));
      if (dart.notNull(l) > 1)
        dart.throw(new core.StateError("More than one element"));
      return this[_this].firstChild;
    }
    add(value) {
      this[_this].append(value);
    }
    addAll(iterable) {
      if (dart.is(iterable, _ChildNodeListLazy)) {
        let otherList = iterable;
        if (!dart.notNull(core.identical(otherList[_this], this[_this]))) {
          for (let i = 0, len = otherList.length; dart.notNull(i) < dart.notNull(len); i = dart.notNull(i) + 1) {
            this[_this].append(otherList[_this].firstChild);
          }
        }
        return;
      }
      for (let node of iterable) {
        this[_this].append(node);
      }
    }
    insert(index, node) {
      if (dart.notNull(index) < 0 || dart.notNull(index) > dart.notNull(this.length)) {
        dart.throw(new core.RangeError.range(index, 0, this.length));
      }
      if (index == this.length) {
        this[_this].append(node);
      } else {
        this[_this].insertBefore(node, this.get(index));
      }
    }
    insertAll(index, iterable) {
      if (index == this.length) {
        this.addAll(iterable);
      } else {
        let item = this.get(index);
        this[_this].insertAllBefore(iterable, item);
      }
    }
    setAll(index, iterable) {
      dart.throw(new core.UnsupportedError("Cannot setAll on Node list"));
    }
    removeLast() {
      let result = this.last;
      if (result != null) {
        this[_this][_removeChild](result);
      }
      return result;
    }
    removeAt(index) {
      let result = this.get(index);
      if (result != null) {
        this[_this][_removeChild](result);
      }
      return result;
    }
    remove(object) {
      if (!dart.is(object, Node))
        return false;
      let node = dart.as(object, Node);
      if (!dart.equals(this[_this], node.parentNode))
        return false;
      this[_this][_removeChild](node);
      return true;
    }
    [_filter](test, removeMatching) {
      let child = this[_this].firstChild;
      while (child != null) {
        let nextChild = child.nextNode;
        if (test(child) == removeMatching) {
          this[_this][_removeChild](child);
        }
        child = nextChild;
      }
    }
    removeWhere(test) {
      this[_filter](test, true);
    }
    retainWhere(test) {
      this[_filter](test, false);
    }
    clear() {
      this[_this][_clearChildren]();
    }
    set(index, value) {
      this[_this][_replaceChild](value, this.get(index));
      return value;
    }
    get iterator() {
      return this[_this].childNodes[dartx.iterator];
    }
    sort(compare) {
      if (compare === void 0)
        compare = null;
      dart.throw(new core.UnsupportedError("Cannot sort Node list"));
    }
    shuffle(random) {
      if (random === void 0)
        random = null;
      dart.throw(new core.UnsupportedError("Cannot shuffle Node list"));
    }
    setRange(start, end, iterable, skipCount) {
      if (skipCount === void 0)
        skipCount = 0;
      dart.throw(new core.UnsupportedError("Cannot setRange on Node list"));
    }
    fillRange(start, end, fill) {
      if (fill === void 0)
        fill = null;
      dart.throw(new core.UnsupportedError("Cannot fillRange on Node list"));
    }
    get length() {
      return this[_this].childNodes[dartx.length];
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot set length on immutable List."));
    }
    get(index) {
      return this[_this].childNodes[dartx.get](index);
    }
    get rawList() {
      return this[_this].childNodes;
    }
  }
  _ChildNodeListLazy[dart.implements] = () => [html_common.NodeListWrapper];
  dart.setSignature(_ChildNodeListLazy, {
    constructors: () => ({_ChildNodeListLazy: [_ChildNodeListLazy, [Node]]}),
    methods: () => ({
      add: [dart.void, [Node]],
      addAll: [dart.void, [core.Iterable$(Node)]],
      insert: [dart.void, [core.int, Node]],
      insertAll: [dart.void, [core.int, core.Iterable$(Node)]],
      setAll: [dart.void, [core.int, core.Iterable$(Node)]],
      removeLast: [Node, []],
      removeAt: [Node, [core.int]],
      [_filter]: [dart.void, [dart.functionType(core.bool, [Node]), core.bool]],
      removeWhere: [dart.void, [dart.functionType(core.bool, [Node])]],
      retainWhere: [dart.void, [dart.functionType(core.bool, [Node])]],
      set: [dart.void, [core.int, Node]],
      sort: [dart.void, [], [core.Comparator$(Node)]],
      setRange: [dart.void, [core.int, core.int, core.Iterable$(Node)], [core.int]],
      fillRange: [dart.void, [core.int, core.int], [Node]],
      get: [Node, [core.int]]
    })
  });
  dart.defineExtensionMembers(_ChildNodeListLazy, [
    'add',
    'addAll',
    'insert',
    'insertAll',
    'setAll',
    'removeLast',
    'removeAt',
    'remove',
    'removeWhere',
    'retainWhere',
    'clear',
    'set',
    'sort',
    'shuffle',
    'setRange',
    'fillRange',
    'get',
    'first',
    'last',
    'single',
    'iterator',
    'length',
    'length'
  ]);
  Node.ATTRIBUTE_NODE = 2;
  Node.CDATA_SECTION_NODE = 4;
  Node.COMMENT_NODE = 8;
  Node.DOCUMENT_FRAGMENT_NODE = 11;
  Node.DOCUMENT_NODE = 9;
  Node.DOCUMENT_TYPE_NODE = 10;
  Node.ELEMENT_NODE = 1;
  Node.ENTITY_NODE = 6;
  Node.ENTITY_REFERENCE_NODE = 5;
  Node.NOTATION_NODE = 12;
  Node.PROCESSING_INSTRUCTION_NODE = 7;
  Node.TEXT_NODE = 3;
  class NodeFilter extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNodeFilter() {
      return NodeFilter._internalWrap();
    }
    static _internalWrap() {
      return new NodeFilter.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(NodeFilter, 'internal_');
  dart.setSignature(NodeFilter, {
    constructors: () => ({
      _: [NodeFilter, []],
      _internalWrap: [NodeFilter, []],
      internal_: [NodeFilter, []]
    }),
    statics: () => ({internalCreateNodeFilter: [NodeFilter, []]}),
    names: ['internalCreateNodeFilter']
  });
  NodeFilter[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NodeFilter')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("NodeFilter"))];
  NodeFilter.FILTER_ACCEPT = 1;
  NodeFilter.FILTER_REJECT = 2;
  NodeFilter.FILTER_SKIP = 3;
  NodeFilter.SHOW_ALL = 4294967295;
  NodeFilter.SHOW_COMMENT = 128;
  NodeFilter.SHOW_DOCUMENT = 256;
  NodeFilter.SHOW_DOCUMENT_FRAGMENT = 1024;
  NodeFilter.SHOW_DOCUMENT_TYPE = 512;
  NodeFilter.SHOW_ELEMENT = 1;
  NodeFilter.SHOW_PROCESSING_INSTRUCTION = 64;
  NodeFilter.SHOW_TEXT = 4;
  const _createNodeIterator = Symbol('_createNodeIterator');
  const _detach_1 = Symbol('_detach_1');
  const _nextNode_1 = Symbol('_nextNode_1');
  const _previousNode_1 = Symbol('_previousNode_1');
  class NodeIterator extends DartHtmlDomObject {
    static new(root, whatToShow) {
      return dart.as(dart.dcall(exports.document[_createNodeIterator], root, whatToShow, null), NodeIterator);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNodeIterator() {
      return NodeIterator._internalWrap();
    }
    static _internalWrap() {
      return new NodeIterator.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get pointerBeforeReferenceNode() {
      return dart.as(wrap_jso(this.raw.pointerBeforeReferenceNode), core.bool);
    }
    get referenceNode() {
      return dart.as(wrap_jso(this.raw.referenceNode), Node);
    }
    get root() {
      return dart.as(wrap_jso(this.raw.root), Node);
    }
    get whatToShow() {
      return dart.as(wrap_jso(this.raw.whatToShow), core.int);
    }
    detach() {
      this[_detach_1]();
      return;
    }
    [_detach_1]() {
      return wrap_jso(this.raw.detach());
    }
    nextNode() {
      return this[_nextNode_1]();
    }
    [_nextNode_1]() {
      return dart.as(wrap_jso(this.raw.nextNode()), Node);
    }
    previousNode() {
      return this[_previousNode_1]();
    }
    [_previousNode_1]() {
      return dart.as(wrap_jso(this.raw.previousNode()), Node);
    }
  }
  dart.defineNamedConstructor(NodeIterator, 'internal_');
  dart.setSignature(NodeIterator, {
    constructors: () => ({
      new: [NodeIterator, [Node, core.int]],
      _: [NodeIterator, []],
      _internalWrap: [NodeIterator, []],
      internal_: [NodeIterator, []]
    }),
    methods: () => ({
      detach: [dart.void, []],
      [_detach_1]: [dart.void, []],
      nextNode: [Node, []],
      [_nextNode_1]: [Node, []],
      previousNode: [Node, []],
      [_previousNode_1]: [Node, []]
    }),
    statics: () => ({internalCreateNodeIterator: [NodeIterator, []]}),
    names: ['internalCreateNodeIterator']
  });
  NodeIterator[dart.metadata] = () => [dart.const(new _metadata.DomName('NodeIterator')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("NodeIterator"))];
  const _item = Symbol('_item');
  class NodeList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(Node), ImmutableListMixin$(Node)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateNodeList() {
      return NodeList._internalWrap();
    }
    static _internalWrap() {
      return new NodeList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), Node);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    [_item](index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Node);
    }
  }
  NodeList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(Node)];
  dart.defineNamedConstructor(NodeList, 'internal_');
  dart.setSignature(NodeList, {
    constructors: () => ({
      _: [NodeList, []],
      _internalWrap: [NodeList, []],
      internal_: [NodeList, []]
    }),
    methods: () => ({
      get: [Node, [core.int]],
      set: [dart.void, [core.int, Node]],
      elementAt: [Node, [core.int]],
      [_item]: [Node, [core.int]],
      [_item_1]: [Node, [dart.dynamic]]
    }),
    statics: () => ({internalCreateNodeList: [NodeList, []]}),
    names: ['internalCreateNodeList']
  });
  dart.defineExtensionMembers(NodeList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  NodeList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NodeList')), dart.const(new _js_helper.Native("NodeList,RadioNodeList"))];
  class Notification extends EventTarget {
    static new(title, opts) {
      let dir = opts && 'dir' in opts ? opts.dir : null;
      let body = opts && 'body' in opts ? opts.body : null;
      let lang = opts && 'lang' in opts ? opts.lang : null;
      let tag = opts && 'tag' in opts ? opts.tag : null;
      let icon = opts && 'icon' in opts ? opts.icon : null;
      let parsedOptions = dart.map();
      if (dir != null)
        parsedOptions.set('dir', dir);
      if (body != null)
        parsedOptions.set('body', body);
      if (lang != null)
        parsedOptions.set('lang', lang);
      if (tag != null)
        parsedOptions.set('tag', tag);
      if (icon != null)
        parsedOptions.set('icon', icon);
      return Notification._factoryNotification(title, parsedOptions);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static _factoryNotification(title, options) {
      if (options === void 0)
        options = null;
      if (options != null) {
        let options_1 = html_common.convertDartToNative_Dictionary(options);
        return Notification._create_1(title, options_1);
      }
      return Notification._create_2(title);
    }
    static _create_1(title, options) {
      return dart.as(wrap_jso(new Notification(title, options)), Notification);
    }
    static _create_2(title) {
      return dart.as(wrap_jso(new Notification(title)), Notification);
    }
    static internalCreateNotification() {
      return Notification._internalWrap();
    }
    static _internalWrap() {
      return new Notification.internal_();
    }
    internal_() {
      super.internal_();
    }
    static get supported() {
      return !!window.Notification;
    }
    get body() {
      return dart.as(wrap_jso(this.raw.body), core.String);
    }
    get dir() {
      return dart.as(wrap_jso(this.raw.dir), core.String);
    }
    get icon() {
      return dart.as(wrap_jso(this.raw.icon), core.String);
    }
    get lang() {
      return dart.as(wrap_jso(this.raw.lang), core.String);
    }
    get permission() {
      return dart.as(wrap_jso(this.raw.permission), core.String);
    }
    get tag() {
      return dart.as(wrap_jso(this.raw.tag), core.String);
    }
    get title() {
      return dart.as(wrap_jso(this.raw.title), core.String);
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    static _requestPermission(callback) {
      if (callback === void 0)
        callback = null;
      if (callback != null) {
        Notification._requestPermission_1(callback);
        return;
      }
      Notification._requestPermission_2();
      return;
    }
    static _requestPermission_1(callback) {
      return wrap_jso(this.raw.requestPermission(unwrap_jso(callback)));
    }
    static _requestPermission_2() {
      return wrap_jso(this.raw.requestPermission());
    }
    static requestPermission() {
      let completer = async.Completer$(core.String).new();
      Notification._requestPermission(dart.fn(value => {
        completer.complete(value);
      }));
      return completer.future;
    }
    get onClick() {
      return Notification.clickEvent.forTarget(this);
    }
    get onClose() {
      return Notification.closeEvent.forTarget(this);
    }
    get onError() {
      return Notification.errorEvent.forTarget(this);
    }
    get onShow() {
      return Notification.showEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(Notification, 'internal_');
  dart.setSignature(Notification, {
    constructors: () => ({
      new: [Notification, [core.String], {dir: core.String, body: core.String, lang: core.String, tag: core.String, icon: core.String}],
      _: [Notification, []],
      _internalWrap: [Notification, []],
      internal_: [Notification, []]
    }),
    methods: () => ({
      close: [dart.void, []],
      [_close_1]: [dart.void, []]
    }),
    statics: () => ({
      _factoryNotification: [Notification, [core.String], [core.Map]],
      _create_1: [Notification, [dart.dynamic, dart.dynamic]],
      _create_2: [Notification, [dart.dynamic]],
      internalCreateNotification: [Notification, []],
      _requestPermission: [dart.void, [], [_NotificationPermissionCallback]],
      _requestPermission_1: [dart.void, [_NotificationPermissionCallback]],
      _requestPermission_2: [dart.void, []],
      requestPermission: [async.Future$(core.String), []]
    }),
    names: ['_factoryNotification', '_create_1', '_create_2', 'internalCreateNotification', '_requestPermission', '_requestPermission_1', '_requestPermission_2', 'requestPermission']
  });
  Notification[dart.metadata] = () => [dart.const(new _metadata.DomName('Notification')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Notification"))];
  Notification.clickEvent = dart.const(new (EventStreamProvider$(Event))('click'));
  Notification.closeEvent = dart.const(new (EventStreamProvider$(Event))('close'));
  Notification.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  Notification.showEvent = dart.const(new (EventStreamProvider$(Event))('show'));
  const _NotificationPermissionCallback = dart.typedef('_NotificationPermissionCallback', () => dart.functionType(dart.void, [core.String]));
  class OListElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("ol"), OListElement);
    }
    static internalCreateOListElement() {
      return OListElement._internalWrap();
    }
    static _internalWrap() {
      return new OListElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get reversed() {
      return dart.as(wrap_jso(this.raw.reversed), core.bool);
    }
    set reversed(val) {
      return this.raw.reversed = unwrap_jso(val);
    }
    get start() {
      return dart.as(wrap_jso(this.raw.start), core.int);
    }
    set start(val) {
      return this.raw.start = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(OListElement, 'internal_');
  dart.defineNamedConstructor(OListElement, 'created');
  dart.setSignature(OListElement, {
    constructors: () => ({
      _: [OListElement, []],
      new: [OListElement, []],
      _internalWrap: [OListElement, []],
      internal_: [OListElement, []],
      created: [OListElement, []]
    }),
    statics: () => ({internalCreateOListElement: [OListElement, []]}),
    names: ['internalCreateOListElement']
  });
  OListElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLOListElement')), dart.const(new _js_helper.Native("HTMLOListElement"))];
  const __setter___2 = Symbol('__setter___2');
  class ObjectElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("object"), ObjectElement);
    }
    static internalCreateObjectElement() {
      return ObjectElement._internalWrap();
    }
    static _internalWrap() {
      return new ObjectElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('object');
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), core.String);
    }
    set data(val) {
      return this.raw.data = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.String);
    }
    set height(val) {
      return this.raw.height = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
    get useMap() {
      return dart.as(wrap_jso(this.raw.useMap), core.String);
    }
    set useMap(val) {
      return this.raw.useMap = unwrap_jso(val);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.String);
    }
    set width(val) {
      return this.raw.width = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    [__getter__](index_OR_name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(index_OR_name))), core.bool);
    }
    [__setter__](index_OR_name, value) {
      if (typeof index_OR_name == 'number' || index_OR_name == null) {
        this[__setter___1](dart.as(index_OR_name, core.int), value);
        return;
      }
      if (typeof index_OR_name == 'string' || index_OR_name == null) {
        this[__setter___2](dart.as(index_OR_name, core.String), value);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [__setter___1](index, value) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(index), unwrap_jso(value)));
    }
    [__setter___2](name, value) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(name), unwrap_jso(value)));
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
  }
  dart.defineNamedConstructor(ObjectElement, 'internal_');
  dart.defineNamedConstructor(ObjectElement, 'created');
  dart.setSignature(ObjectElement, {
    constructors: () => ({
      _: [ObjectElement, []],
      new: [ObjectElement, []],
      _internalWrap: [ObjectElement, []],
      internal_: [ObjectElement, []],
      created: [ObjectElement, []]
    }),
    methods: () => ({
      [__getter__]: [core.bool, [dart.dynamic]],
      [__setter__]: [dart.void, [dart.dynamic, Node]],
      [__setter___1]: [dart.void, [core.int, Node]],
      [__setter___2]: [dart.void, [core.String, Node]],
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateObjectElement: [ObjectElement, []]}),
    names: ['internalCreateObjectElement']
  });
  ObjectElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLObjectElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("HTMLObjectElement"))];
  class OptGroupElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("optgroup"), OptGroupElement);
    }
    static internalCreateOptGroupElement() {
      return OptGroupElement._internalWrap();
    }
    static _internalWrap() {
      return new OptGroupElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    set label(val) {
      return this.raw.label = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(OptGroupElement, 'internal_');
  dart.defineNamedConstructor(OptGroupElement, 'created');
  dart.setSignature(OptGroupElement, {
    constructors: () => ({
      _: [OptGroupElement, []],
      new: [OptGroupElement, []],
      _internalWrap: [OptGroupElement, []],
      internal_: [OptGroupElement, []],
      created: [OptGroupElement, []]
    }),
    statics: () => ({internalCreateOptGroupElement: [OptGroupElement, []]}),
    names: ['internalCreateOptGroupElement']
  });
  OptGroupElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLOptGroupElement')), dart.const(new _js_helper.Native("HTMLOptGroupElement"))];
  class OptionElement extends HtmlElement {
    static new(opts) {
      let data = opts && 'data' in opts ? opts.data : '';
      let value = opts && 'value' in opts ? opts.value : '';
      let selected = opts && 'selected' in opts ? opts.selected : false;
      return OptionElement._(data, value, null, selected);
    }
    static _(data, value, defaultSelected, selected) {
      if (data === void 0)
        data = null;
      if (value === void 0)
        value = null;
      if (defaultSelected === void 0)
        defaultSelected = null;
      if (selected === void 0)
        selected = null;
      if (selected != null) {
        return OptionElement._create_1(data, value, defaultSelected, selected);
      }
      if (defaultSelected != null) {
        return OptionElement._create_2(data, value, defaultSelected);
      }
      if (value != null) {
        return OptionElement._create_3(data, value);
      }
      if (data != null) {
        return OptionElement._create_4(data);
      }
      return OptionElement._create_5();
    }
    static _create_1(data, value, defaultSelected, selected) {
      return dart.as(wrap_jso(new Option(data, value, defaultSelected, selected)), OptionElement);
    }
    static _create_2(data, value, defaultSelected) {
      return dart.as(wrap_jso(new Option(data, value, defaultSelected)), OptionElement);
    }
    static _create_3(data, value) {
      return dart.as(wrap_jso(new Option(data, value)), OptionElement);
    }
    static _create_4(data) {
      return dart.as(wrap_jso(new Option(data)), OptionElement);
    }
    static _create_5() {
      return dart.as(wrap_jso(new Option()), OptionElement);
    }
    static internalCreateOptionElement() {
      return OptionElement._internalWrap();
    }
    static _internalWrap() {
      return new OptionElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get defaultSelected() {
      return dart.as(wrap_jso(this.raw.defaultSelected), core.bool);
    }
    set defaultSelected(val) {
      return this.raw.defaultSelected = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get index() {
      return dart.as(wrap_jso(this.raw.index), core.int);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    set label(val) {
      return this.raw.label = unwrap_jso(val);
    }
    get selected() {
      return dart.as(wrap_jso(this.raw.selected), core.bool);
    }
    set selected(val) {
      return this.raw.selected = unwrap_jso(val);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(OptionElement, 'internal_');
  dart.defineNamedConstructor(OptionElement, 'created');
  dart.setSignature(OptionElement, {
    constructors: () => ({
      new: [OptionElement, [], {data: core.String, value: core.String, selected: core.bool}],
      _: [OptionElement, [], [core.String, core.String, core.bool, core.bool]],
      _internalWrap: [OptionElement, []],
      internal_: [OptionElement, []],
      created: [OptionElement, []]
    }),
    statics: () => ({
      _create_1: [OptionElement, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      _create_2: [OptionElement, [dart.dynamic, dart.dynamic, dart.dynamic]],
      _create_3: [OptionElement, [dart.dynamic, dart.dynamic]],
      _create_4: [OptionElement, [dart.dynamic]],
      _create_5: [OptionElement, []],
      internalCreateOptionElement: [OptionElement, []]
    }),
    names: ['_create_1', '_create_2', '_create_3', '_create_4', '_create_5', 'internalCreateOptionElement']
  });
  OptionElement[dart.metadata] = () => [dart.const(new _metadata.DomName('HTMLOptionElement')), dart.const(new _js_helper.Native("HTMLOptionElement"))];
  class OutputElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("output"), OutputElement);
    }
    static internalCreateOutputElement() {
      return OutputElement._internalWrap();
    }
    static _internalWrap() {
      return new OutputElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('output');
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.defaultValue), core.String);
    }
    set defaultValue(val) {
      return this.raw.defaultValue = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get htmlFor() {
      return wrap_jso(this.raw.htmlFor);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
  }
  dart.defineNamedConstructor(OutputElement, 'internal_');
  dart.defineNamedConstructor(OutputElement, 'created');
  dart.setSignature(OutputElement, {
    constructors: () => ({
      _: [OutputElement, []],
      new: [OutputElement, []],
      _internalWrap: [OutputElement, []],
      internal_: [OutputElement, []],
      created: [OutputElement, []]
    }),
    methods: () => ({
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateOutputElement: [OutputElement, []]}),
    names: ['internalCreateOutputElement']
  });
  OutputElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLOutputElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _js_helper.Native("HTMLOutputElement"))];
  class OverflowEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateOverflowEvent() {
      return OverflowEvent._internalWrap();
    }
    static _internalWrap() {
      return new OverflowEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get horizontalOverflow() {
      return dart.as(wrap_jso(this.raw.horizontalOverflow), core.bool);
    }
    get orient() {
      return dart.as(wrap_jso(this.raw.orient), core.int);
    }
    get verticalOverflow() {
      return dart.as(wrap_jso(this.raw.verticalOverflow), core.bool);
    }
  }
  dart.defineNamedConstructor(OverflowEvent, 'internal_');
  dart.setSignature(OverflowEvent, {
    constructors: () => ({
      _: [OverflowEvent, []],
      _internalWrap: [OverflowEvent, []],
      internal_: [OverflowEvent, []]
    }),
    statics: () => ({internalCreateOverflowEvent: [OverflowEvent, []]}),
    names: ['internalCreateOverflowEvent']
  });
  OverflowEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('OverflowEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("OverflowEvent"))];
  OverflowEvent.BOTH = 2;
  OverflowEvent.HORIZONTAL = 0;
  OverflowEvent.VERTICAL = 1;
  class PageTransitionEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePageTransitionEvent() {
      return PageTransitionEvent._internalWrap();
    }
    static _internalWrap() {
      return new PageTransitionEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get persisted() {
      return dart.as(wrap_jso(this.raw.persisted), core.bool);
    }
  }
  dart.defineNamedConstructor(PageTransitionEvent, 'internal_');
  dart.setSignature(PageTransitionEvent, {
    constructors: () => ({
      _: [PageTransitionEvent, []],
      _internalWrap: [PageTransitionEvent, []],
      internal_: [PageTransitionEvent, []]
    }),
    statics: () => ({internalCreatePageTransitionEvent: [PageTransitionEvent, []]}),
    names: ['internalCreatePageTransitionEvent']
  });
  PageTransitionEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PageTransitionEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PageTransitionEvent"))];
  class ParagraphElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("p"), ParagraphElement);
    }
    static internalCreateParagraphElement() {
      return ParagraphElement._internalWrap();
    }
    static _internalWrap() {
      return new ParagraphElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(ParagraphElement, 'internal_');
  dart.defineNamedConstructor(ParagraphElement, 'created');
  dart.setSignature(ParagraphElement, {
    constructors: () => ({
      _: [ParagraphElement, []],
      new: [ParagraphElement, []],
      _internalWrap: [ParagraphElement, []],
      internal_: [ParagraphElement, []],
      created: [ParagraphElement, []]
    }),
    statics: () => ({internalCreateParagraphElement: [ParagraphElement, []]}),
    names: ['internalCreateParagraphElement']
  });
  ParagraphElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLParagraphElement')), dart.const(new _js_helper.Native("HTMLParagraphElement"))];
  class ParamElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("param"), ParamElement);
    }
    static internalCreateParamElement() {
      return ParamElement._internalWrap();
    }
    static _internalWrap() {
      return new ParamElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(ParamElement, 'internal_');
  dart.defineNamedConstructor(ParamElement, 'created');
  dart.setSignature(ParamElement, {
    constructors: () => ({
      _: [ParamElement, []],
      new: [ParamElement, []],
      _internalWrap: [ParamElement, []],
      internal_: [ParamElement, []],
      created: [ParamElement, []]
    }),
    statics: () => ({internalCreateParamElement: [ParamElement, []]}),
    names: ['internalCreateParamElement']
  });
  ParamElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLParamElement')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("HTMLParamElement"))];
  class ParentNode extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get [_childElementCount]() {
      return dart.as(wrap_jso(this.raw.childElementCount), core.int);
    }
    get [_children]() {
      return dart.as(wrap_jso(this.raw.children), core.List$(Node));
    }
    get [_firstElementChild]() {
      return dart.as(wrap_jso(this.raw.firstElementChild), Element);
    }
    get [_lastElementChild]() {
      return dart.as(wrap_jso(this.raw.lastElementChild), Element);
    }
    querySelector(selectors) {
      return dart.as(wrap_jso(this.raw.querySelector(unwrap_jso(selectors))), Element);
    }
    [_querySelectorAll](selectors) {
      return dart.as(wrap_jso(this.raw.querySelectorAll(unwrap_jso(selectors))), core.List$(Node));
    }
  }
  dart.setSignature(ParentNode, {
    constructors: () => ({_: [ParentNode, []]}),
    methods: () => ({
      querySelector: [Element, [core.String]],
      [_querySelectorAll]: [core.List$(Node), [core.String]]
    })
  });
  ParentNode[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ParentNode')), dart.const(new _metadata.Experimental())];
  const _addPath_1 = Symbol('_addPath_1');
  const _addPath_2 = Symbol('_addPath_2');
  const _arc_1 = Symbol('_arc_1');
  const _arcTo_1 = Symbol('_arcTo_1');
  const _bezierCurveTo_1 = Symbol('_bezierCurveTo_1');
  const _closePath_1 = Symbol('_closePath_1');
  const _ellipse_1 = Symbol('_ellipse_1');
  const _lineTo_1 = Symbol('_lineTo_1');
  const _moveTo_1 = Symbol('_moveTo_1');
  const _quadraticCurveTo_1 = Symbol('_quadraticCurveTo_1');
  const _rect_1 = Symbol('_rect_1');
  class Path2D extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(path_OR_text) {
      if (path_OR_text === void 0)
        path_OR_text = null;
      if (path_OR_text == null) {
        return Path2D._create_1();
      }
      if (dart.is(path_OR_text, Path2D) || path_OR_text == null) {
        return Path2D._create_2(path_OR_text);
      }
      if (typeof path_OR_text == 'string' || path_OR_text == null) {
        return Path2D._create_3(path_OR_text);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    static _create_1() {
      return dart.as(wrap_jso(new Path2D()), Path2D);
    }
    static _create_2(path_OR_text) {
      return dart.as(wrap_jso(new Path2D(path_OR_text)), Path2D);
    }
    static _create_3(path_OR_text) {
      return dart.as(wrap_jso(new Path2D(path_OR_text)), Path2D);
    }
    static internalCreatePath2D() {
      return Path2D._internalWrap();
    }
    static _internalWrap() {
      return new Path2D.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    addPath(path, transform) {
      if (transform === void 0)
        transform = null;
      if (transform != null) {
        this[_addPath_1](path, transform);
        return;
      }
      this[_addPath_2](path);
      return;
    }
    [_addPath_1](path, transform) {
      return wrap_jso(this.raw.addPath(unwrap_jso(path), unwrap_jso(transform)));
    }
    [_addPath_2](path) {
      return wrap_jso(this.raw.addPath(unwrap_jso(path)));
    }
    arc(x, y, radius, startAngle, endAngle, anticlockwise) {
      this[_arc_1](x, y, radius, startAngle, endAngle, anticlockwise);
      return;
    }
    [_arc_1](x, y, radius, startAngle, endAngle, anticlockwise) {
      return wrap_jso(this.raw.arc(unwrap_jso(x), unwrap_jso(y), unwrap_jso(radius), unwrap_jso(startAngle), unwrap_jso(endAngle), unwrap_jso(anticlockwise)));
    }
    arcTo(x1, y1, x2, y2, radius) {
      this[_arcTo_1](x1, y1, x2, y2, radius);
      return;
    }
    [_arcTo_1](x1, y1, x2, y2, radius) {
      return wrap_jso(this.raw.arcTo(unwrap_jso(x1), unwrap_jso(y1), unwrap_jso(x2), unwrap_jso(y2), unwrap_jso(radius)));
    }
    bezierCurveTo(cp1x, cp1y, cp2x, cp2y, x, y) {
      this[_bezierCurveTo_1](cp1x, cp1y, cp2x, cp2y, x, y);
      return;
    }
    [_bezierCurveTo_1](cp1x, cp1y, cp2x, cp2y, x, y) {
      return wrap_jso(this.raw.bezierCurveTo(unwrap_jso(cp1x), unwrap_jso(cp1y), unwrap_jso(cp2x), unwrap_jso(cp2y), unwrap_jso(x), unwrap_jso(y)));
    }
    closePath() {
      this[_closePath_1]();
      return;
    }
    [_closePath_1]() {
      return wrap_jso(this.raw.closePath());
    }
    ellipse(x, y, radiusX, radiusY, rotation, startAngle, endAngle, anticlockwise) {
      this[_ellipse_1](x, y, radiusX, radiusY, rotation, startAngle, endAngle, anticlockwise);
      return;
    }
    [_ellipse_1](x, y, radiusX, radiusY, rotation, startAngle, endAngle, anticlockwise) {
      return wrap_jso(this.raw.ellipse(unwrap_jso(x), unwrap_jso(y), unwrap_jso(radiusX), unwrap_jso(radiusY), unwrap_jso(rotation), unwrap_jso(startAngle), unwrap_jso(endAngle), unwrap_jso(anticlockwise)));
    }
    lineTo(x, y) {
      this[_lineTo_1](x, y);
      return;
    }
    [_lineTo_1](x, y) {
      return wrap_jso(this.raw.lineTo(unwrap_jso(x), unwrap_jso(y)));
    }
    moveTo(x, y) {
      this[_moveTo_1](x, y);
      return;
    }
    [_moveTo_1](x, y) {
      return wrap_jso(this.raw.moveTo(unwrap_jso(x), unwrap_jso(y)));
    }
    quadraticCurveTo(cpx, cpy, x, y) {
      this[_quadraticCurveTo_1](cpx, cpy, x, y);
      return;
    }
    [_quadraticCurveTo_1](cpx, cpy, x, y) {
      return wrap_jso(this.raw.quadraticCurveTo(unwrap_jso(cpx), unwrap_jso(cpy), unwrap_jso(x), unwrap_jso(y)));
    }
    rect(x, y, width, height) {
      this[_rect_1](x, y, width, height);
      return;
    }
    [_rect_1](x, y, width, height) {
      return wrap_jso(this.raw.rect(unwrap_jso(x), unwrap_jso(y), unwrap_jso(width), unwrap_jso(height)));
    }
  }
  Path2D[dart.implements] = () => [_CanvasPathMethods];
  dart.defineNamedConstructor(Path2D, 'internal_');
  dart.setSignature(Path2D, {
    constructors: () => ({
      _: [Path2D, []],
      new: [Path2D, [], [dart.dynamic]],
      _internalWrap: [Path2D, []],
      internal_: [Path2D, []]
    }),
    methods: () => ({
      addPath: [dart.void, [Path2D], [dart.dynamic]],
      [_addPath_1]: [dart.void, [Path2D, dart.dynamic]],
      [_addPath_2]: [dart.void, [Path2D]],
      arc: [dart.void, [core.num, core.num, core.num, core.num, core.num, core.bool]],
      [_arc_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      arcTo: [dart.void, [core.num, core.num, core.num, core.num, core.num]],
      [_arcTo_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      bezierCurveTo: [dart.void, [core.num, core.num, core.num, core.num, core.num, core.num]],
      [_bezierCurveTo_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      closePath: [dart.void, []],
      [_closePath_1]: [dart.void, []],
      ellipse: [dart.void, [core.num, core.num, core.num, core.num, core.num, core.num, core.num, core.bool]],
      [_ellipse_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      lineTo: [dart.void, [core.num, core.num]],
      [_lineTo_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      moveTo: [dart.void, [core.num, core.num]],
      [_moveTo_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      quadraticCurveTo: [dart.void, [core.num, core.num, core.num, core.num]],
      [_quadraticCurveTo_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      rect: [dart.void, [core.num, core.num, core.num, core.num]],
      [_rect_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({
      _create_1: [Path2D, []],
      _create_2: [Path2D, [dart.dynamic]],
      _create_3: [Path2D, [dart.dynamic]],
      internalCreatePath2D: [Path2D, []]
    }),
    names: ['_create_1', '_create_2', '_create_3', 'internalCreatePath2D']
  });
  Path2D[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Path2D')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Path2D"))];
  const _clearMarks_1 = Symbol('_clearMarks_1');
  const _clearMeasures_1 = Symbol('_clearMeasures_1');
  const _getEntries_1 = Symbol('_getEntries_1');
  const _getEntriesByName_1 = Symbol('_getEntriesByName_1');
  const _getEntriesByType_1 = Symbol('_getEntriesByType_1');
  const _mark_1 = Symbol('_mark_1');
  const _measure_1 = Symbol('_measure_1');
  const _now_1 = Symbol('_now_1');
  const _clearResourceTimings_1 = Symbol('_clearResourceTimings_1');
  const _setResourceTimingBufferSize_1 = Symbol('_setResourceTimingBufferSize_1');
  class Performance extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformance() {
      return Performance._internalWrap();
    }
    static _internalWrap() {
      return new Performance.internal_();
    }
    internal_() {
      super.internal_();
    }
    static get supported() {
      return !!window.performance;
    }
    get memory() {
      return wrap_jso(this.raw.memory);
    }
    get navigation() {
      return dart.as(wrap_jso(this.raw.navigation), PerformanceNavigation);
    }
    get timing() {
      return dart.as(wrap_jso(this.raw.timing), PerformanceTiming);
    }
    clearMarks(markName) {
      this[_clearMarks_1](markName);
      return;
    }
    [_clearMarks_1](markName) {
      return wrap_jso(this.raw.clearMarks(unwrap_jso(markName)));
    }
    clearMeasures(measureName) {
      this[_clearMeasures_1](measureName);
      return;
    }
    [_clearMeasures_1](measureName) {
      return wrap_jso(this.raw.clearMeasures(unwrap_jso(measureName)));
    }
    getEntries() {
      return this[_getEntries_1]();
    }
    [_getEntries_1]() {
      return dart.as(wrap_jso(this.raw.getEntries()), core.List$(PerformanceEntry));
    }
    getEntriesByName(name, entryType) {
      return this[_getEntriesByName_1](name, entryType);
    }
    [_getEntriesByName_1](name, entryType) {
      return dart.as(wrap_jso(this.raw.getEntriesByName(unwrap_jso(name), unwrap_jso(entryType))), core.List$(PerformanceEntry));
    }
    getEntriesByType(entryType) {
      return this[_getEntriesByType_1](entryType);
    }
    [_getEntriesByType_1](entryType) {
      return dart.as(wrap_jso(this.raw.getEntriesByType(unwrap_jso(entryType))), core.List$(PerformanceEntry));
    }
    mark(markName) {
      this[_mark_1](markName);
      return;
    }
    [_mark_1](markName) {
      return wrap_jso(this.raw.mark(unwrap_jso(markName)));
    }
    measure(measureName, startMark, endMark) {
      this[_measure_1](measureName, startMark, endMark);
      return;
    }
    [_measure_1](measureName, startMark, endMark) {
      return wrap_jso(this.raw.measure(unwrap_jso(measureName), unwrap_jso(startMark), unwrap_jso(endMark)));
    }
    now() {
      return this[_now_1]();
    }
    [_now_1]() {
      return dart.as(wrap_jso(this.raw.now()), core.num);
    }
    clearResourceTimings() {
      this[_clearResourceTimings_1]();
      return;
    }
    [_clearResourceTimings_1]() {
      return wrap_jso(this.raw.webkitClearResourceTimings());
    }
    setResourceTimingBufferSize(maxSize) {
      this[_setResourceTimingBufferSize_1](maxSize);
      return;
    }
    [_setResourceTimingBufferSize_1](maxSize) {
      return wrap_jso(this.raw.webkitSetResourceTimingBufferSize(unwrap_jso(maxSize)));
    }
    get onResourceTimingBufferFull() {
      return Performance.resourceTimingBufferFullEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(Performance, 'internal_');
  dart.setSignature(Performance, {
    constructors: () => ({
      _: [Performance, []],
      _internalWrap: [Performance, []],
      internal_: [Performance, []]
    }),
    methods: () => ({
      clearMarks: [dart.void, [core.String]],
      [_clearMarks_1]: [dart.void, [dart.dynamic]],
      clearMeasures: [dart.void, [core.String]],
      [_clearMeasures_1]: [dart.void, [dart.dynamic]],
      getEntries: [core.List$(PerformanceEntry), []],
      [_getEntries_1]: [core.List$(PerformanceEntry), []],
      getEntriesByName: [core.List$(PerformanceEntry), [core.String, core.String]],
      [_getEntriesByName_1]: [core.List$(PerformanceEntry), [dart.dynamic, dart.dynamic]],
      getEntriesByType: [core.List$(PerformanceEntry), [core.String]],
      [_getEntriesByType_1]: [core.List$(PerformanceEntry), [dart.dynamic]],
      mark: [dart.void, [core.String]],
      [_mark_1]: [dart.void, [dart.dynamic]],
      measure: [dart.void, [core.String, core.String, core.String]],
      [_measure_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      now: [core.num, []],
      [_now_1]: [core.num, []],
      clearResourceTimings: [dart.void, []],
      [_clearResourceTimings_1]: [dart.void, []],
      setResourceTimingBufferSize: [dart.void, [core.int]],
      [_setResourceTimingBufferSize_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreatePerformance: [Performance, []]}),
    names: ['internalCreatePerformance']
  });
  Performance[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Performance')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE)), dart.const(new _js_helper.Native("Performance"))];
  Performance.resourceTimingBufferFullEvent = dart.const(new (EventStreamProvider$(Event))('webkitresourcetimingbufferfull'));
  class PerformanceEntry extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceEntry() {
      return PerformanceEntry._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceEntry.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get duration() {
      return dart.as(wrap_jso(this.raw.duration), core.double);
    }
    get entryType() {
      return dart.as(wrap_jso(this.raw.entryType), core.String);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    get startTime() {
      return dart.as(wrap_jso(this.raw.startTime), core.double);
    }
  }
  dart.defineNamedConstructor(PerformanceEntry, 'internal_');
  dart.setSignature(PerformanceEntry, {
    constructors: () => ({
      _: [PerformanceEntry, []],
      _internalWrap: [PerformanceEntry, []],
      internal_: [PerformanceEntry, []]
    }),
    statics: () => ({internalCreatePerformanceEntry: [PerformanceEntry, []]}),
    names: ['internalCreatePerformanceEntry']
  });
  PerformanceEntry[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceEntry')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PerformanceEntry"))];
  class PerformanceMark extends PerformanceEntry {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceMark() {
      return PerformanceMark._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceMark.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(PerformanceMark, 'internal_');
  dart.setSignature(PerformanceMark, {
    constructors: () => ({
      _: [PerformanceMark, []],
      _internalWrap: [PerformanceMark, []],
      internal_: [PerformanceMark, []]
    }),
    statics: () => ({internalCreatePerformanceMark: [PerformanceMark, []]}),
    names: ['internalCreatePerformanceMark']
  });
  PerformanceMark[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceMark')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PerformanceMark"))];
  class PerformanceMeasure extends PerformanceEntry {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceMeasure() {
      return PerformanceMeasure._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceMeasure.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(PerformanceMeasure, 'internal_');
  dart.setSignature(PerformanceMeasure, {
    constructors: () => ({
      _: [PerformanceMeasure, []],
      _internalWrap: [PerformanceMeasure, []],
      internal_: [PerformanceMeasure, []]
    }),
    statics: () => ({internalCreatePerformanceMeasure: [PerformanceMeasure, []]}),
    names: ['internalCreatePerformanceMeasure']
  });
  PerformanceMeasure[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceMeasure')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PerformanceMeasure"))];
  class PerformanceNavigation extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceNavigation() {
      return PerformanceNavigation._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceNavigation.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get redirectCount() {
      return dart.as(wrap_jso(this.raw.redirectCount), core.int);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.int);
    }
  }
  dart.defineNamedConstructor(PerformanceNavigation, 'internal_');
  dart.setSignature(PerformanceNavigation, {
    constructors: () => ({
      _: [PerformanceNavigation, []],
      _internalWrap: [PerformanceNavigation, []],
      internal_: [PerformanceNavigation, []]
    }),
    statics: () => ({internalCreatePerformanceNavigation: [PerformanceNavigation, []]}),
    names: ['internalCreatePerformanceNavigation']
  });
  PerformanceNavigation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceNavigation')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("PerformanceNavigation"))];
  PerformanceNavigation.TYPE_BACK_FORWARD = 2;
  PerformanceNavigation.TYPE_NAVIGATE = 0;
  PerformanceNavigation.TYPE_RELOAD = 1;
  PerformanceNavigation.TYPE_RESERVED = 255;
  class PerformanceResourceTiming extends PerformanceEntry {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceResourceTiming() {
      return PerformanceResourceTiming._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceResourceTiming.internal_();
    }
    internal_() {
      super.internal_();
    }
    get connectEnd() {
      return dart.as(wrap_jso(this.raw.connectEnd), core.double);
    }
    get connectStart() {
      return dart.as(wrap_jso(this.raw.connectStart), core.double);
    }
    get domainLookupEnd() {
      return dart.as(wrap_jso(this.raw.domainLookupEnd), core.double);
    }
    get domainLookupStart() {
      return dart.as(wrap_jso(this.raw.domainLookupStart), core.double);
    }
    get fetchStart() {
      return dart.as(wrap_jso(this.raw.fetchStart), core.double);
    }
    get initiatorType() {
      return dart.as(wrap_jso(this.raw.initiatorType), core.String);
    }
    get redirectEnd() {
      return dart.as(wrap_jso(this.raw.redirectEnd), core.double);
    }
    get redirectStart() {
      return dart.as(wrap_jso(this.raw.redirectStart), core.double);
    }
    get requestStart() {
      return dart.as(wrap_jso(this.raw.requestStart), core.double);
    }
    get responseEnd() {
      return dart.as(wrap_jso(this.raw.responseEnd), core.double);
    }
    get responseStart() {
      return dart.as(wrap_jso(this.raw.responseStart), core.double);
    }
    get secureConnectionStart() {
      return dart.as(wrap_jso(this.raw.secureConnectionStart), core.double);
    }
  }
  dart.defineNamedConstructor(PerformanceResourceTiming, 'internal_');
  dart.setSignature(PerformanceResourceTiming, {
    constructors: () => ({
      _: [PerformanceResourceTiming, []],
      _internalWrap: [PerformanceResourceTiming, []],
      internal_: [PerformanceResourceTiming, []]
    }),
    statics: () => ({internalCreatePerformanceResourceTiming: [PerformanceResourceTiming, []]}),
    names: ['internalCreatePerformanceResourceTiming']
  });
  PerformanceResourceTiming[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceResourceTiming')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PerformanceResourceTiming"))];
  class PerformanceTiming extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePerformanceTiming() {
      return PerformanceTiming._internalWrap();
    }
    static _internalWrap() {
      return new PerformanceTiming.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get connectEnd() {
      return dart.as(wrap_jso(this.raw.connectEnd), core.int);
    }
    get connectStart() {
      return dart.as(wrap_jso(this.raw.connectStart), core.int);
    }
    get domComplete() {
      return dart.as(wrap_jso(this.raw.domComplete), core.int);
    }
    get domContentLoadedEventEnd() {
      return dart.as(wrap_jso(this.raw.domContentLoadedEventEnd), core.int);
    }
    get domContentLoadedEventStart() {
      return dart.as(wrap_jso(this.raw.domContentLoadedEventStart), core.int);
    }
    get domInteractive() {
      return dart.as(wrap_jso(this.raw.domInteractive), core.int);
    }
    get domLoading() {
      return dart.as(wrap_jso(this.raw.domLoading), core.int);
    }
    get domainLookupEnd() {
      return dart.as(wrap_jso(this.raw.domainLookupEnd), core.int);
    }
    get domainLookupStart() {
      return dart.as(wrap_jso(this.raw.domainLookupStart), core.int);
    }
    get fetchStart() {
      return dart.as(wrap_jso(this.raw.fetchStart), core.int);
    }
    get loadEventEnd() {
      return dart.as(wrap_jso(this.raw.loadEventEnd), core.int);
    }
    get loadEventStart() {
      return dart.as(wrap_jso(this.raw.loadEventStart), core.int);
    }
    get navigationStart() {
      return dart.as(wrap_jso(this.raw.navigationStart), core.int);
    }
    get redirectEnd() {
      return dart.as(wrap_jso(this.raw.redirectEnd), core.int);
    }
    get redirectStart() {
      return dart.as(wrap_jso(this.raw.redirectStart), core.int);
    }
    get requestStart() {
      return dart.as(wrap_jso(this.raw.requestStart), core.int);
    }
    get responseEnd() {
      return dart.as(wrap_jso(this.raw.responseEnd), core.int);
    }
    get responseStart() {
      return dart.as(wrap_jso(this.raw.responseStart), core.int);
    }
    get secureConnectionStart() {
      return dart.as(wrap_jso(this.raw.secureConnectionStart), core.int);
    }
    get unloadEventEnd() {
      return dart.as(wrap_jso(this.raw.unloadEventEnd), core.int);
    }
    get unloadEventStart() {
      return dart.as(wrap_jso(this.raw.unloadEventStart), core.int);
    }
  }
  dart.defineNamedConstructor(PerformanceTiming, 'internal_');
  dart.setSignature(PerformanceTiming, {
    constructors: () => ({
      _: [PerformanceTiming, []],
      _internalWrap: [PerformanceTiming, []],
      internal_: [PerformanceTiming, []]
    }),
    statics: () => ({internalCreatePerformanceTiming: [PerformanceTiming, []]}),
    names: ['internalCreatePerformanceTiming']
  });
  PerformanceTiming[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PerformanceTiming')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("PerformanceTiming"))];
  class PictureElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePictureElement() {
      return PictureElement._internalWrap();
    }
    static _internalWrap() {
      return new PictureElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(PictureElement, 'internal_');
  dart.defineNamedConstructor(PictureElement, 'created');
  dart.setSignature(PictureElement, {
    constructors: () => ({
      _: [PictureElement, []],
      _internalWrap: [PictureElement, []],
      internal_: [PictureElement, []],
      created: [PictureElement, []]
    }),
    statics: () => ({internalCreatePictureElement: [PictureElement, []]}),
    names: ['internalCreatePictureElement']
  });
  PictureElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLPictureElement')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("HTMLPictureElement"))];
  class Plugin extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePlugin() {
      return Plugin._internalWrap();
    }
    static _internalWrap() {
      return new Plugin.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get description() {
      return dart.as(wrap_jso(this.raw.description), core.String);
    }
    get filename() {
      return dart.as(wrap_jso(this.raw.filename), core.String);
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), MimeType);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), MimeType);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), MimeType);
    }
  }
  dart.defineNamedConstructor(Plugin, 'internal_');
  dart.setSignature(Plugin, {
    constructors: () => ({
      _: [Plugin, []],
      _internalWrap: [Plugin, []],
      internal_: [Plugin, []]
    }),
    methods: () => ({
      [__getter__]: [MimeType, [core.String]],
      [__getter___1]: [MimeType, [dart.dynamic]],
      item: [MimeType, [core.int]],
      [_item_1]: [MimeType, [dart.dynamic]],
      namedItem: [MimeType, [core.String]],
      [_namedItem_1]: [MimeType, [dart.dynamic]]
    }),
    statics: () => ({internalCreatePlugin: [Plugin, []]}),
    names: ['internalCreatePlugin']
  });
  Plugin[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Plugin')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Plugin"))];
  const _refresh_1 = Symbol('_refresh_1');
  class PluginArray extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(Plugin), ImmutableListMixin$(Plugin)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePluginArray() {
      return PluginArray._internalWrap();
    }
    static _internalWrap() {
      return new PluginArray.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), Plugin);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Plugin);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), Plugin);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Plugin);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), Plugin);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Plugin);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), Plugin);
    }
    refresh(reload) {
      this[_refresh_1](reload);
      return;
    }
    [_refresh_1](reload) {
      return wrap_jso(this.raw.refresh(unwrap_jso(reload)));
    }
  }
  PluginArray[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(Plugin)];
  dart.defineNamedConstructor(PluginArray, 'internal_');
  dart.setSignature(PluginArray, {
    constructors: () => ({
      _: [PluginArray, []],
      _internalWrap: [PluginArray, []],
      internal_: [PluginArray, []]
    }),
    methods: () => ({
      get: [Plugin, [core.int]],
      set: [dart.void, [core.int, Plugin]],
      elementAt: [Plugin, [core.int]],
      [__getter__]: [Plugin, [core.String]],
      [__getter___1]: [Plugin, [dart.dynamic]],
      item: [Plugin, [core.int]],
      [_item_1]: [Plugin, [dart.dynamic]],
      namedItem: [Plugin, [core.String]],
      [_namedItem_1]: [Plugin, [dart.dynamic]],
      refresh: [dart.void, [core.bool]],
      [_refresh_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreatePluginArray: [PluginArray, []]}),
    names: ['internalCreatePluginArray']
  });
  dart.defineExtensionMembers(PluginArray, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  PluginArray[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PluginArray')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PluginArray"))];
  const _createdCallback_1 = Symbol('_createdCallback_1');
  class PluginPlaceholderElement extends DivElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePluginPlaceholderElement() {
      return PluginPlaceholderElement._internalWrap();
    }
    static _internalWrap() {
      return new PluginPlaceholderElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get message() {
      return dart.as(wrap_jso(this.raw.message), core.String);
    }
    set message(val) {
      return this.raw.message = unwrap_jso(val);
    }
    createdCallback() {
      this[_createdCallback_1]();
      return;
    }
    [_createdCallback_1]() {
      return wrap_jso(this.raw.createdCallback());
    }
  }
  dart.defineNamedConstructor(PluginPlaceholderElement, 'internal_');
  dart.defineNamedConstructor(PluginPlaceholderElement, 'created');
  dart.setSignature(PluginPlaceholderElement, {
    constructors: () => ({
      _: [PluginPlaceholderElement, []],
      _internalWrap: [PluginPlaceholderElement, []],
      internal_: [PluginPlaceholderElement, []],
      created: [PluginPlaceholderElement, []]
    }),
    methods: () => ({
      createdCallback: [dart.void, []],
      [_createdCallback_1]: [dart.void, []]
    }),
    statics: () => ({internalCreatePluginPlaceholderElement: [PluginPlaceholderElement, []]}),
    names: ['internalCreatePluginPlaceholderElement']
  });
  PluginPlaceholderElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PluginPlaceholderElement')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PluginPlaceholderElement"))];
  const _PositionCallback = dart.typedef('_PositionCallback', () => dart.functionType(dart.void, [dart.dynamic]));
  class PositionError extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePositionError() {
      return PositionError._internalWrap();
    }
    static _internalWrap() {
      return new PositionError.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get code() {
      return dart.as(wrap_jso(this.raw.code), core.int);
    }
    get message() {
      return dart.as(wrap_jso(this.raw.message), core.String);
    }
  }
  dart.defineNamedConstructor(PositionError, 'internal_');
  dart.setSignature(PositionError, {
    constructors: () => ({
      _: [PositionError, []],
      _internalWrap: [PositionError, []],
      internal_: [PositionError, []]
    }),
    statics: () => ({internalCreatePositionError: [PositionError, []]}),
    names: ['internalCreatePositionError']
  });
  PositionError[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PositionError')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("PositionError"))];
  PositionError.PERMISSION_DENIED = 1;
  PositionError.POSITION_UNAVAILABLE = 2;
  PositionError.TIMEOUT = 3;
  const _PositionErrorCallback = dart.typedef('_PositionErrorCallback', () => dart.functionType(dart.void, [PositionError]));
  class PreElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("pre"), PreElement);
    }
    static internalCreatePreElement() {
      return PreElement._internalWrap();
    }
    static _internalWrap() {
      return new PreElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(PreElement, 'internal_');
  dart.defineNamedConstructor(PreElement, 'created');
  dart.setSignature(PreElement, {
    constructors: () => ({
      _: [PreElement, []],
      new: [PreElement, []],
      _internalWrap: [PreElement, []],
      internal_: [PreElement, []],
      created: [PreElement, []]
    }),
    statics: () => ({internalCreatePreElement: [PreElement, []]}),
    names: ['internalCreatePreElement']
  });
  PreElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLPreElement')), dart.const(new _js_helper.Native("HTMLPreElement"))];
  class Presentation extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePresentation() {
      return Presentation._internalWrap();
    }
    static _internalWrap() {
      return new Presentation.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(Presentation, 'internal_');
  dart.setSignature(Presentation, {
    constructors: () => ({
      _: [Presentation, []],
      _internalWrap: [Presentation, []],
      internal_: [Presentation, []]
    }),
    statics: () => ({internalCreatePresentation: [Presentation, []]}),
    names: ['internalCreatePresentation']
  });
  Presentation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Presentation')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Presentation"))];
  class ProgressElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("progress"), ProgressElement);
    }
    static internalCreateProgressElement() {
      return ProgressElement._internalWrap();
    }
    static _internalWrap() {
      return new ProgressElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('progress');
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get max() {
      return dart.as(wrap_jso(this.raw.max), core.num);
    }
    set max(val) {
      return this.raw.max = unwrap_jso(val);
    }
    get position() {
      return dart.as(wrap_jso(this.raw.position), core.double);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.num);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(ProgressElement, 'internal_');
  dart.defineNamedConstructor(ProgressElement, 'created');
  dart.setSignature(ProgressElement, {
    constructors: () => ({
      _: [ProgressElement, []],
      new: [ProgressElement, []],
      _internalWrap: [ProgressElement, []],
      internal_: [ProgressElement, []],
      created: [ProgressElement, []]
    }),
    statics: () => ({internalCreateProgressElement: [ProgressElement, []]}),
    names: ['internalCreateProgressElement']
  });
  ProgressElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLProgressElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _js_helper.Native("HTMLProgressElement"))];
  class PushEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePushEvent() {
      return PushEvent._internalWrap();
    }
    static _internalWrap() {
      return new PushEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), core.String);
    }
  }
  dart.defineNamedConstructor(PushEvent, 'internal_');
  dart.setSignature(PushEvent, {
    constructors: () => ({
      _: [PushEvent, []],
      _internalWrap: [PushEvent, []],
      internal_: [PushEvent, []]
    }),
    statics: () => ({internalCreatePushEvent: [PushEvent, []]}),
    names: ['internalCreatePushEvent']
  });
  PushEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PushEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PushEvent"))];
  const _register_1 = Symbol('_register_1');
  class PushManager extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePushManager() {
      return PushManager._internalWrap();
    }
    static _internalWrap() {
      return new PushManager.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    register(senderId) {
      return this[_register_1](senderId);
    }
    [_register_1](senderId) {
      return dart.as(wrap_jso(this.raw.register(unwrap_jso(senderId))), async.Future);
    }
  }
  dart.defineNamedConstructor(PushManager, 'internal_');
  dart.setSignature(PushManager, {
    constructors: () => ({
      _: [PushManager, []],
      _internalWrap: [PushManager, []],
      internal_: [PushManager, []]
    }),
    methods: () => ({
      register: [async.Future, [core.String]],
      [_register_1]: [async.Future, [dart.dynamic]]
    }),
    statics: () => ({internalCreatePushManager: [PushManager, []]}),
    names: ['internalCreatePushManager']
  });
  PushManager[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PushManager')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PushManager"))];
  class PushRegistration extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreatePushRegistration() {
      return PushRegistration._internalWrap();
    }
    static _internalWrap() {
      return new PushRegistration.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get pushEndpoint() {
      return dart.as(wrap_jso(this.raw.pushEndpoint), core.String);
    }
    get pushRegistrationId() {
      return dart.as(wrap_jso(this.raw.pushRegistrationId), core.String);
    }
  }
  dart.defineNamedConstructor(PushRegistration, 'internal_');
  dart.setSignature(PushRegistration, {
    constructors: () => ({
      _: [PushRegistration, []],
      _internalWrap: [PushRegistration, []],
      internal_: [PushRegistration, []]
    }),
    statics: () => ({internalCreatePushRegistration: [PushRegistration, []]}),
    names: ['internalCreatePushRegistration']
  });
  PushRegistration[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PushRegistration')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("PushRegistration"))];
  class QuoteElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("q"), QuoteElement);
    }
    static internalCreateQuoteElement() {
      return QuoteElement._internalWrap();
    }
    static _internalWrap() {
      return new QuoteElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get cite() {
      return dart.as(wrap_jso(this.raw.cite), core.String);
    }
    set cite(val) {
      return this.raw.cite = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(QuoteElement, 'internal_');
  dart.defineNamedConstructor(QuoteElement, 'created');
  dart.setSignature(QuoteElement, {
    constructors: () => ({
      _: [QuoteElement, []],
      new: [QuoteElement, []],
      _internalWrap: [QuoteElement, []],
      internal_: [QuoteElement, []],
      created: [QuoteElement, []]
    }),
    statics: () => ({internalCreateQuoteElement: [QuoteElement, []]}),
    names: ['internalCreateQuoteElement']
  });
  QuoteElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLQuoteElement')), dart.const(new _js_helper.Native("HTMLQuoteElement"))];
  const _RtcErrorCallback = dart.typedef('_RtcErrorCallback', () => dart.functionType(dart.void, [core.String]));
  const _RtcSessionDescriptionCallback = dart.typedef('_RtcSessionDescriptionCallback', () => dart.functionType(dart.void, [RtcSessionDescription]));
  const RtcStatsCallback = dart.typedef('RtcStatsCallback', () => dart.functionType(dart.void, [RtcStatsResponse]));
  const _cloneContents_1 = Symbol('_cloneContents_1');
  const _cloneRange_1 = Symbol('_cloneRange_1');
  const _collapse_1 = Symbol('_collapse_1');
  const _collapse_2 = Symbol('_collapse_2');
  const _compareBoundaryPoints_1 = Symbol('_compareBoundaryPoints_1');
  const _comparePoint_1 = Symbol('_comparePoint_1');
  const _createContextualFragment_1 = Symbol('_createContextualFragment_1');
  const _deleteContents_1 = Symbol('_deleteContents_1');
  const _expand_1 = Symbol('_expand_1');
  const _extractContents_1 = Symbol('_extractContents_1');
  const _insertNode_1 = Symbol('_insertNode_1');
  const _isPointInRange_1 = Symbol('_isPointInRange_1');
  const _selectNode_1 = Symbol('_selectNode_1');
  const _selectNodeContents_1 = Symbol('_selectNodeContents_1');
  const _setEnd_1 = Symbol('_setEnd_1');
  const _setEndAfter_1 = Symbol('_setEndAfter_1');
  const _setEndBefore_1 = Symbol('_setEndBefore_1');
  const _setStart_1 = Symbol('_setStart_1');
  const _setStartAfter_1 = Symbol('_setStartAfter_1');
  const _setStartBefore_1 = Symbol('_setStartBefore_1');
  const _surroundContents_1 = Symbol('_surroundContents_1');
  class Range extends DartHtmlDomObject {
    static new() {
      return exports.document.createRange();
    }
    static fromPoint(point) {
      return exports.document[_caretRangeFromPoint](dart.as(point.x, core.int), dart.as(point.y, core.int));
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRange() {
      return Range._internalWrap();
    }
    static _internalWrap() {
      return new Range.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get collapsed() {
      return dart.as(wrap_jso(this.raw.collapsed), core.bool);
    }
    get commonAncestorContainer() {
      return dart.as(wrap_jso(this.raw.commonAncestorContainer), Node);
    }
    get endContainer() {
      return dart.as(wrap_jso(this.raw.endContainer), Node);
    }
    get endOffset() {
      return dart.as(wrap_jso(this.raw.endOffset), core.int);
    }
    get startContainer() {
      return dart.as(wrap_jso(this.raw.startContainer), Node);
    }
    get startOffset() {
      return dart.as(wrap_jso(this.raw.startOffset), core.int);
    }
    cloneContents() {
      return this[_cloneContents_1]();
    }
    [_cloneContents_1]() {
      return dart.as(wrap_jso(this.raw.cloneContents()), DocumentFragment);
    }
    cloneRange() {
      return this[_cloneRange_1]();
    }
    [_cloneRange_1]() {
      return dart.as(wrap_jso(this.raw.cloneRange()), Range);
    }
    collapse(toStart) {
      if (toStart === void 0)
        toStart = null;
      if (toStart != null) {
        this[_collapse_1](toStart);
        return;
      }
      this[_collapse_2]();
      return;
    }
    [_collapse_1](toStart) {
      return wrap_jso(this.raw.collapse(unwrap_jso(toStart)));
    }
    [_collapse_2]() {
      return wrap_jso(this.raw.collapse());
    }
    compareBoundaryPoints(how, sourceRange) {
      return this[_compareBoundaryPoints_1](how, sourceRange);
    }
    [_compareBoundaryPoints_1](how, sourceRange) {
      return dart.as(wrap_jso(this.raw.compareBoundaryPoints(unwrap_jso(how), unwrap_jso(sourceRange))), core.int);
    }
    comparePoint(refNode, offset) {
      return this[_comparePoint_1](refNode, offset);
    }
    [_comparePoint_1](refNode, offset) {
      return dart.as(wrap_jso(this.raw.comparePoint(unwrap_jso(refNode), unwrap_jso(offset))), core.int);
    }
    createContextualFragment(html) {
      return this[_createContextualFragment_1](html);
    }
    [_createContextualFragment_1](html) {
      return dart.as(wrap_jso(this.raw.createContextualFragment(unwrap_jso(html))), DocumentFragment);
    }
    deleteContents() {
      this[_deleteContents_1]();
      return;
    }
    [_deleteContents_1]() {
      return wrap_jso(this.raw.deleteContents());
    }
    detach() {
      this[_detach_1]();
      return;
    }
    [_detach_1]() {
      return wrap_jso(this.raw.detach());
    }
    expand(unit) {
      this[_expand_1](unit);
      return;
    }
    [_expand_1](unit) {
      return wrap_jso(this.raw.expand(unwrap_jso(unit)));
    }
    extractContents() {
      return this[_extractContents_1]();
    }
    [_extractContents_1]() {
      return dart.as(wrap_jso(this.raw.extractContents()), DocumentFragment);
    }
    getBoundingClientRect() {
      return this[_getBoundingClientRect_1]();
    }
    [_getBoundingClientRect_1]() {
      return dart.as(wrap_jso(this.raw.getBoundingClientRect()), math.Rectangle);
    }
    getClientRects() {
      return this[_getClientRects_1]();
    }
    [_getClientRects_1]() {
      return dart.as(wrap_jso(this.raw.getClientRects()), _ClientRectList);
    }
    insertNode(newNode) {
      this[_insertNode_1](newNode);
      return;
    }
    [_insertNode_1](newNode) {
      return wrap_jso(this.raw.insertNode(unwrap_jso(newNode)));
    }
    isPointInRange(refNode, offset) {
      return this[_isPointInRange_1](refNode, offset);
    }
    [_isPointInRange_1](refNode, offset) {
      return dart.as(wrap_jso(this.raw.isPointInRange(unwrap_jso(refNode), unwrap_jso(offset))), core.bool);
    }
    selectNode(refNode) {
      this[_selectNode_1](refNode);
      return;
    }
    [_selectNode_1](refNode) {
      return wrap_jso(this.raw.selectNode(unwrap_jso(refNode)));
    }
    selectNodeContents(refNode) {
      this[_selectNodeContents_1](refNode);
      return;
    }
    [_selectNodeContents_1](refNode) {
      return wrap_jso(this.raw.selectNodeContents(unwrap_jso(refNode)));
    }
    setEnd(refNode, offset) {
      this[_setEnd_1](refNode, offset);
      return;
    }
    [_setEnd_1](refNode, offset) {
      return wrap_jso(this.raw.setEnd(unwrap_jso(refNode), unwrap_jso(offset)));
    }
    setEndAfter(refNode) {
      this[_setEndAfter_1](refNode);
      return;
    }
    [_setEndAfter_1](refNode) {
      return wrap_jso(this.raw.setEndAfter(unwrap_jso(refNode)));
    }
    setEndBefore(refNode) {
      this[_setEndBefore_1](refNode);
      return;
    }
    [_setEndBefore_1](refNode) {
      return wrap_jso(this.raw.setEndBefore(unwrap_jso(refNode)));
    }
    setStart(refNode, offset) {
      this[_setStart_1](refNode, offset);
      return;
    }
    [_setStart_1](refNode, offset) {
      return wrap_jso(this.raw.setStart(unwrap_jso(refNode), unwrap_jso(offset)));
    }
    setStartAfter(refNode) {
      this[_setStartAfter_1](refNode);
      return;
    }
    [_setStartAfter_1](refNode) {
      return wrap_jso(this.raw.setStartAfter(unwrap_jso(refNode)));
    }
    setStartBefore(refNode) {
      this[_setStartBefore_1](refNode);
      return;
    }
    [_setStartBefore_1](refNode) {
      return wrap_jso(this.raw.setStartBefore(unwrap_jso(refNode)));
    }
    surroundContents(newParent) {
      this[_surroundContents_1](newParent);
      return;
    }
    [_surroundContents_1](newParent) {
      return wrap_jso(this.raw.surroundContents(unwrap_jso(newParent)));
    }
    static get supportsCreateContextualFragment() {
      return true;
    }
  }
  dart.defineNamedConstructor(Range, 'internal_');
  dart.setSignature(Range, {
    constructors: () => ({
      new: [Range, []],
      fromPoint: [Range, [math.Point]],
      _: [Range, []],
      _internalWrap: [Range, []],
      internal_: [Range, []]
    }),
    methods: () => ({
      cloneContents: [DocumentFragment, []],
      [_cloneContents_1]: [DocumentFragment, []],
      cloneRange: [Range, []],
      [_cloneRange_1]: [Range, []],
      collapse: [dart.void, [], [core.bool]],
      [_collapse_1]: [dart.void, [dart.dynamic]],
      [_collapse_2]: [dart.void, []],
      compareBoundaryPoints: [core.int, [core.int, Range]],
      [_compareBoundaryPoints_1]: [core.int, [dart.dynamic, Range]],
      comparePoint: [core.int, [Node, core.int]],
      [_comparePoint_1]: [core.int, [Node, dart.dynamic]],
      createContextualFragment: [DocumentFragment, [core.String]],
      [_createContextualFragment_1]: [DocumentFragment, [dart.dynamic]],
      deleteContents: [dart.void, []],
      [_deleteContents_1]: [dart.void, []],
      detach: [dart.void, []],
      [_detach_1]: [dart.void, []],
      expand: [dart.void, [core.String]],
      [_expand_1]: [dart.void, [dart.dynamic]],
      extractContents: [DocumentFragment, []],
      [_extractContents_1]: [DocumentFragment, []],
      getBoundingClientRect: [math.Rectangle, []],
      [_getBoundingClientRect_1]: [math.Rectangle, []],
      getClientRects: [_ClientRectList, []],
      [_getClientRects_1]: [_ClientRectList, []],
      insertNode: [dart.void, [Node]],
      [_insertNode_1]: [dart.void, [Node]],
      isPointInRange: [core.bool, [Node, core.int]],
      [_isPointInRange_1]: [core.bool, [Node, dart.dynamic]],
      selectNode: [dart.void, [Node]],
      [_selectNode_1]: [dart.void, [Node]],
      selectNodeContents: [dart.void, [Node]],
      [_selectNodeContents_1]: [dart.void, [Node]],
      setEnd: [dart.void, [Node, core.int]],
      [_setEnd_1]: [dart.void, [Node, dart.dynamic]],
      setEndAfter: [dart.void, [Node]],
      [_setEndAfter_1]: [dart.void, [Node]],
      setEndBefore: [dart.void, [Node]],
      [_setEndBefore_1]: [dart.void, [Node]],
      setStart: [dart.void, [Node, core.int]],
      [_setStart_1]: [dart.void, [Node, dart.dynamic]],
      setStartAfter: [dart.void, [Node]],
      [_setStartAfter_1]: [dart.void, [Node]],
      setStartBefore: [dart.void, [Node]],
      [_setStartBefore_1]: [dart.void, [Node]],
      surroundContents: [dart.void, [Node]],
      [_surroundContents_1]: [dart.void, [Node]]
    }),
    statics: () => ({internalCreateRange: [Range, []]}),
    names: ['internalCreateRange']
  });
  Range[dart.metadata] = () => [dart.const(new _metadata.DomName('Range')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("Range"))];
  Range.END_TO_END = 2;
  Range.END_TO_START = 3;
  Range.NODE_AFTER = 1;
  Range.NODE_BEFORE = 0;
  Range.NODE_BEFORE_AND_AFTER = 2;
  Range.NODE_INSIDE = 3;
  Range.START_TO_END = 1;
  Range.START_TO_START = 0;
  const _cancel_1 = Symbol('_cancel_1');
  const _read_1 = Symbol('_read_1');
  const _wait_1 = Symbol('_wait_1');
  class ReadableStream extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateReadableStream() {
      return ReadableStream._internalWrap();
    }
    static _internalWrap() {
      return new ReadableStream.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get closed() {
      return dart.as(wrap_jso(this.raw.closed), async.Future);
    }
    get state() {
      return dart.as(wrap_jso(this.raw.state), core.String);
    }
    cancel(reason) {
      return this[_cancel_1](reason);
    }
    [_cancel_1](reason) {
      return dart.as(wrap_jso(this.raw.cancel(unwrap_jso(reason))), async.Future);
    }
    read() {
      return this[_read_1]();
    }
    [_read_1]() {
      return wrap_jso(this.raw.read());
    }
    wait() {
      return this[_wait_1]();
    }
    [_wait_1]() {
      return dart.as(wrap_jso(this.raw.wait()), async.Future);
    }
  }
  dart.defineNamedConstructor(ReadableStream, 'internal_');
  dart.setSignature(ReadableStream, {
    constructors: () => ({
      _: [ReadableStream, []],
      _internalWrap: [ReadableStream, []],
      internal_: [ReadableStream, []]
    }),
    methods: () => ({
      cancel: [async.Future, [core.Object]],
      [_cancel_1]: [async.Future, [dart.dynamic]],
      read: [core.Object, []],
      [_read_1]: [core.Object, []],
      wait: [async.Future, []],
      [_wait_1]: [async.Future, []]
    }),
    statics: () => ({internalCreateReadableStream: [ReadableStream, []]}),
    names: ['internalCreateReadableStream']
  });
  ReadableStream[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ReadableStream')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ReadableStream"))];
  class RelatedEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRelatedEvent() {
      return RelatedEvent._internalWrap();
    }
    static _internalWrap() {
      return new RelatedEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get relatedTarget() {
      return _convertNativeToDart_EventTarget(this[_get_relatedTarget]);
    }
    get [_get_relatedTarget]() {
      return wrap_jso(this.raw.relatedTarget);
    }
  }
  dart.defineNamedConstructor(RelatedEvent, 'internal_');
  dart.setSignature(RelatedEvent, {
    constructors: () => ({
      _: [RelatedEvent, []],
      _internalWrap: [RelatedEvent, []],
      internal_: [RelatedEvent, []]
    }),
    statics: () => ({internalCreateRelatedEvent: [RelatedEvent, []]}),
    names: ['internalCreateRelatedEvent']
  });
  RelatedEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RelatedEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RelatedEvent"))];
  const RequestAnimationFrameCallback = dart.typedef('RequestAnimationFrameCallback', () => dart.functionType(dart.void, [core.num]));
  class ResourceProgressEvent extends ProgressEvent {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateResourceProgressEvent() {
      return ResourceProgressEvent._internalWrap();
    }
    static _internalWrap() {
      return new ResourceProgressEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get url() {
      return dart.as(wrap_jso(this.raw.url), core.String);
    }
  }
  dart.defineNamedConstructor(ResourceProgressEvent, 'internal_');
  dart.setSignature(ResourceProgressEvent, {
    constructors: () => ({
      _: [ResourceProgressEvent, []],
      _internalWrap: [ResourceProgressEvent, []],
      internal_: [ResourceProgressEvent, []]
    }),
    statics: () => ({internalCreateResourceProgressEvent: [ResourceProgressEvent, []]}),
    names: ['internalCreateResourceProgressEvent']
  });
  ResourceProgressEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ResourceProgressEvent')), core.deprecated, dart.const(new _js_helper.Native("ResourceProgressEvent"))];
  const _sendBlob_1 = Symbol('_sendBlob_1');
  const _sendByteBuffer_1 = Symbol('_sendByteBuffer_1');
  const _sendString_1 = Symbol('_sendString_1');
  const _sendTypedData_1 = Symbol('_sendTypedData_1');
  class RtcDataChannel extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcDataChannel() {
      return RtcDataChannel._internalWrap();
    }
    static _internalWrap() {
      return new RtcDataChannel.internal_();
    }
    internal_() {
      super.internal_();
    }
    get binaryType() {
      return dart.as(wrap_jso(this.raw.binaryType), core.String);
    }
    set binaryType(val) {
      return this.raw.binaryType = unwrap_jso(val);
    }
    get bufferedAmount() {
      return dart.as(wrap_jso(this.raw.bufferedAmount), core.int);
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.int);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    get maxRetransmitTime() {
      return dart.as(wrap_jso(this.raw.maxRetransmitTime), core.int);
    }
    get maxRetransmits() {
      return dart.as(wrap_jso(this.raw.maxRetransmits), core.int);
    }
    get negotiated() {
      return dart.as(wrap_jso(this.raw.negotiated), core.bool);
    }
    get ordered() {
      return dart.as(wrap_jso(this.raw.ordered), core.bool);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    get readyState() {
      return dart.as(wrap_jso(this.raw.readyState), core.String);
    }
    get reliable() {
      return dart.as(wrap_jso(this.raw.reliable), core.bool);
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    send(data) {
      if (typeof data == 'string' || data == null) {
        this[_send_1](dart.as(data, core.String));
        return;
      }
      if (dart.is(data, dart.dynamic) || data == null) {
        this[_send_2](data);
        return;
      }
      if (dart.is(data, typed_data.TypedData) || data == null) {
        this[_send_3](dart.as(data, typed_data.TypedData));
        return;
      }
      if (dart.is(data, typed_data.ByteBuffer) || data == null) {
        this[_send_4](dart.as(data, typed_data.ByteBuffer));
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_send_1](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_2](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_3](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    [_send_4](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    sendBlob(data) {
      this[_sendBlob_1](data);
      return;
    }
    [_sendBlob_1](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    sendByteBuffer(data) {
      this[_sendByteBuffer_1](data);
      return;
    }
    [_sendByteBuffer_1](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    sendString(data) {
      this[_sendString_1](data);
      return;
    }
    [_sendString_1](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    sendTypedData(data) {
      this[_sendTypedData_1](data);
      return;
    }
    [_sendTypedData_1](data) {
      return wrap_jso(this.raw.send(unwrap_jso(data)));
    }
    get onClose() {
      return RtcDataChannel.closeEvent.forTarget(this);
    }
    get onError() {
      return RtcDataChannel.errorEvent.forTarget(this);
    }
    get onMessage() {
      return RtcDataChannel.messageEvent.forTarget(this);
    }
    get onOpen() {
      return RtcDataChannel.openEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(RtcDataChannel, 'internal_');
  dart.setSignature(RtcDataChannel, {
    constructors: () => ({
      _: [RtcDataChannel, []],
      _internalWrap: [RtcDataChannel, []],
      internal_: [RtcDataChannel, []]
    }),
    methods: () => ({
      close: [dart.void, []],
      [_close_1]: [dart.void, []],
      send: [dart.void, [dart.dynamic]],
      [_send_1]: [dart.void, [core.String]],
      [_send_2]: [dart.void, [dart.dynamic]],
      [_send_3]: [dart.void, [typed_data.TypedData]],
      [_send_4]: [dart.void, [typed_data.ByteBuffer]],
      sendBlob: [dart.void, [dart.dynamic]],
      [_sendBlob_1]: [dart.void, [dart.dynamic]],
      sendByteBuffer: [dart.void, [typed_data.ByteBuffer]],
      [_sendByteBuffer_1]: [dart.void, [typed_data.ByteBuffer]],
      sendString: [dart.void, [core.String]],
      [_sendString_1]: [dart.void, [dart.dynamic]],
      sendTypedData: [dart.void, [typed_data.TypedData]],
      [_sendTypedData_1]: [dart.void, [typed_data.TypedData]]
    }),
    statics: () => ({internalCreateRtcDataChannel: [RtcDataChannel, []]}),
    names: ['internalCreateRtcDataChannel']
  });
  RtcDataChannel[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCDataChannel')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCDataChannel,DataChannel"))];
  RtcDataChannel.closeEvent = dart.const(new (EventStreamProvider$(Event))('close'));
  RtcDataChannel.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  RtcDataChannel.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  RtcDataChannel.openEvent = dart.const(new (EventStreamProvider$(Event))('open'));
  class RtcDataChannelEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcDataChannelEvent() {
      return RtcDataChannelEvent._internalWrap();
    }
    static _internalWrap() {
      return new RtcDataChannelEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get channel() {
      return dart.as(wrap_jso(this.raw.channel), RtcDataChannel);
    }
  }
  dart.defineNamedConstructor(RtcDataChannelEvent, 'internal_');
  dart.setSignature(RtcDataChannelEvent, {
    constructors: () => ({
      _: [RtcDataChannelEvent, []],
      _internalWrap: [RtcDataChannelEvent, []],
      internal_: [RtcDataChannelEvent, []]
    }),
    statics: () => ({internalCreateRtcDataChannelEvent: [RtcDataChannelEvent, []]}),
    names: ['internalCreateRtcDataChannelEvent']
  });
  RtcDataChannelEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCDataChannelEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCDataChannelEvent"))];
  const _insertDtmf_1 = Symbol('_insertDtmf_1');
  const _insertDtmf_2 = Symbol('_insertDtmf_2');
  const _insertDtmf_3 = Symbol('_insertDtmf_3');
  class RtcDtmfSender extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcDtmfSender() {
      return RtcDtmfSender._internalWrap();
    }
    static _internalWrap() {
      return new RtcDtmfSender.internal_();
    }
    internal_() {
      super.internal_();
    }
    get canInsertDtmf() {
      return dart.as(wrap_jso(this.raw.canInsertDTMF), core.bool);
    }
    get duration() {
      return dart.as(wrap_jso(this.raw.duration), core.int);
    }
    get interToneGap() {
      return dart.as(wrap_jso(this.raw.interToneGap), core.int);
    }
    get toneBuffer() {
      return dart.as(wrap_jso(this.raw.toneBuffer), core.String);
    }
    get track() {
      return wrap_jso(this.raw.track);
    }
    insertDtmf(tones, duration, interToneGap) {
      if (duration === void 0)
        duration = null;
      if (interToneGap === void 0)
        interToneGap = null;
      if (interToneGap != null) {
        this[_insertDtmf_1](tones, duration, interToneGap);
        return;
      }
      if (duration != null) {
        this[_insertDtmf_2](tones, duration);
        return;
      }
      this[_insertDtmf_3](tones);
      return;
    }
    [_insertDtmf_1](tones, duration, interToneGap) {
      return wrap_jso(this.raw.insertDTMF(unwrap_jso(tones), unwrap_jso(duration), unwrap_jso(interToneGap)));
    }
    [_insertDtmf_2](tones, duration) {
      return wrap_jso(this.raw.insertDTMF(unwrap_jso(tones), unwrap_jso(duration)));
    }
    [_insertDtmf_3](tones) {
      return wrap_jso(this.raw.insertDTMF(unwrap_jso(tones)));
    }
    get onToneChange() {
      return RtcDtmfSender.toneChangeEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(RtcDtmfSender, 'internal_');
  dart.setSignature(RtcDtmfSender, {
    constructors: () => ({
      _: [RtcDtmfSender, []],
      _internalWrap: [RtcDtmfSender, []],
      internal_: [RtcDtmfSender, []]
    }),
    methods: () => ({
      insertDtmf: [dart.void, [core.String], [core.int, core.int]],
      [_insertDtmf_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_insertDtmf_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_insertDtmf_3]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateRtcDtmfSender: [RtcDtmfSender, []]}),
    names: ['internalCreateRtcDtmfSender']
  });
  RtcDtmfSender[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCDTMFSender')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCDTMFSender"))];
  class RtcDtmfToneChangeEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcDtmfToneChangeEvent() {
      return RtcDtmfToneChangeEvent._internalWrap();
    }
    static _internalWrap() {
      return new RtcDtmfToneChangeEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get tone() {
      return dart.as(wrap_jso(this.raw.tone), core.String);
    }
  }
  dart.defineNamedConstructor(RtcDtmfToneChangeEvent, 'internal_');
  dart.setSignature(RtcDtmfToneChangeEvent, {
    constructors: () => ({
      _: [RtcDtmfToneChangeEvent, []],
      _internalWrap: [RtcDtmfToneChangeEvent, []],
      internal_: [RtcDtmfToneChangeEvent, []]
    }),
    statics: () => ({internalCreateRtcDtmfToneChangeEvent: [RtcDtmfToneChangeEvent, []]}),
    names: ['internalCreateRtcDtmfToneChangeEvent']
  });
  RtcDtmfToneChangeEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCDTMFToneChangeEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCDTMFToneChangeEvent"))];
  RtcDtmfSender.toneChangeEvent = dart.const(new (EventStreamProvider$(RtcDtmfToneChangeEvent))('tonechange'));
  class RtcIceCandidate extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcIceCandidate() {
      return RtcIceCandidate._internalWrap();
    }
    static _internalWrap() {
      return new RtcIceCandidate.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get candidate() {
      return dart.as(wrap_jso(this.raw.candidate), core.String);
    }
    set candidate(val) {
      return this.raw.candidate = unwrap_jso(val);
    }
    get sdpMLineIndex() {
      return dart.as(wrap_jso(this.raw.sdpMLineIndex), core.int);
    }
    set sdpMLineIndex(val) {
      return this.raw.sdpMLineIndex = unwrap_jso(val);
    }
    get sdpMid() {
      return dart.as(wrap_jso(this.raw.sdpMid), core.String);
    }
    set sdpMid(val) {
      return this.raw.sdpMid = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(RtcIceCandidate, 'internal_');
  dart.setSignature(RtcIceCandidate, {
    constructors: () => ({
      _: [RtcIceCandidate, []],
      _internalWrap: [RtcIceCandidate, []],
      internal_: [RtcIceCandidate, []]
    }),
    statics: () => ({internalCreateRtcIceCandidate: [RtcIceCandidate, []]}),
    names: ['internalCreateRtcIceCandidate']
  });
  RtcIceCandidate[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCIceCandidate')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCIceCandidate,mozRTCIceCandidate"))];
  class RtcIceCandidateEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcIceCandidateEvent() {
      return RtcIceCandidateEvent._internalWrap();
    }
    static _internalWrap() {
      return new RtcIceCandidateEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get candidate() {
      return dart.as(wrap_jso(this.raw.candidate), RtcIceCandidate);
    }
  }
  dart.defineNamedConstructor(RtcIceCandidateEvent, 'internal_');
  dart.setSignature(RtcIceCandidateEvent, {
    constructors: () => ({
      _: [RtcIceCandidateEvent, []],
      _internalWrap: [RtcIceCandidateEvent, []],
      internal_: [RtcIceCandidateEvent, []]
    }),
    statics: () => ({internalCreateRtcIceCandidateEvent: [RtcIceCandidateEvent, []]}),
    names: ['internalCreateRtcIceCandidateEvent']
  });
  RtcIceCandidateEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCIceCandidateEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCIceCandidateEvent,RTCPeerConnectionIceEvent"))];
  const _createOffer = Symbol('_createOffer');
  const _createAnswer = Symbol('_createAnswer');
  const _getStats = Symbol('_getStats');
  const _addIceCandidate_1 = Symbol('_addIceCandidate_1');
  const _addStream_1 = Symbol('_addStream_1');
  const _addStream_2 = Symbol('_addStream_2');
  const _createAnswer_1 = Symbol('_createAnswer_1');
  const _createAnswer_2 = Symbol('_createAnswer_2');
  const _createDtmfSender_1 = Symbol('_createDtmfSender_1');
  const _createDataChannel_1 = Symbol('_createDataChannel_1');
  const _createDataChannel_2 = Symbol('_createDataChannel_2');
  const _createOffer_1 = Symbol('_createOffer_1');
  const _createOffer_2 = Symbol('_createOffer_2');
  const _getLocalStreams_1 = Symbol('_getLocalStreams_1');
  const _getRemoteStreams_1 = Symbol('_getRemoteStreams_1');
  const _getStats_1 = Symbol('_getStats_1');
  const _getStreamById_1 = Symbol('_getStreamById_1');
  const _removeStream_1 = Symbol('_removeStream_1');
  const _setLocalDescription_1 = Symbol('_setLocalDescription_1');
  const _setLocalDescription = Symbol('_setLocalDescription');
  const _setRemoteDescription_1 = Symbol('_setRemoteDescription_1');
  const _setRemoteDescription = Symbol('_setRemoteDescription');
  const _updateIce_1 = Symbol('_updateIce_1');
  const _updateIce_2 = Symbol('_updateIce_2');
  const _updateIce_3 = Symbol('_updateIce_3');
  class RtcPeerConnection extends EventTarget {
    static get supported() {
      return true;
    }
    createOffer(mediaConstraints) {
      if (mediaConstraints === void 0)
        mediaConstraints = null;
      let completer = async.Completer$(RtcSessionDescription).new();
      this[_createOffer](dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }), mediaConstraints);
      return completer.future;
    }
    createAnswer(mediaConstraints) {
      if (mediaConstraints === void 0)
        mediaConstraints = null;
      let completer = async.Completer$(RtcSessionDescription).new();
      this[_createAnswer](dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }), mediaConstraints);
      return completer.future;
    }
    getStats(selector) {
      let completer = async.Completer$(RtcStatsResponse).new();
      this[_getStats](dart.fn(value => {
        completer.complete(value);
      }), selector);
      return completer.future;
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcPeerConnection() {
      return RtcPeerConnection._internalWrap();
    }
    static _internalWrap() {
      return new RtcPeerConnection.internal_();
    }
    internal_() {
      super.internal_();
    }
    get iceConnectionState() {
      return dart.as(wrap_jso(this.raw.iceConnectionState), core.String);
    }
    get iceGatheringState() {
      return dart.as(wrap_jso(this.raw.iceGatheringState), core.String);
    }
    get localDescription() {
      return dart.as(wrap_jso(this.raw.localDescription), RtcSessionDescription);
    }
    get remoteDescription() {
      return dart.as(wrap_jso(this.raw.remoteDescription), RtcSessionDescription);
    }
    get signalingState() {
      return dart.as(wrap_jso(this.raw.signalingState), core.String);
    }
    addIceCandidate(candidate, successCallback, failureCallback) {
      this[_addIceCandidate_1](candidate, successCallback, failureCallback);
      return;
    }
    [_addIceCandidate_1](candidate, successCallback, failureCallback) {
      return wrap_jso(this.raw.addIceCandidate(unwrap_jso(candidate), unwrap_jso(successCallback), unwrap_jso(failureCallback)));
    }
    addStream(stream, mediaConstraints) {
      if (mediaConstraints === void 0)
        mediaConstraints = null;
      if (mediaConstraints != null) {
        let mediaConstraints_1 = html_common.convertDartToNative_Dictionary(mediaConstraints);
        this[_addStream_1](stream, mediaConstraints_1);
        return;
      }
      this[_addStream_2](stream);
      return;
    }
    [_addStream_1](stream, mediaConstraints) {
      return wrap_jso(this.raw.addStream(unwrap_jso(stream), unwrap_jso(mediaConstraints)));
    }
    [_addStream_2](stream) {
      return wrap_jso(this.raw.addStream(unwrap_jso(stream)));
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    [_createAnswer](successCallback, failureCallback, mediaConstraints) {
      if (failureCallback === void 0)
        failureCallback = null;
      if (mediaConstraints === void 0)
        mediaConstraints = null;
      if (mediaConstraints != null) {
        let mediaConstraints_1 = html_common.convertDartToNative_Dictionary(mediaConstraints);
        this[_createAnswer_1](successCallback, failureCallback, mediaConstraints_1);
        return;
      }
      this[_createAnswer_2](successCallback, failureCallback);
      return;
    }
    [_createAnswer_1](successCallback, failureCallback, mediaConstraints) {
      return wrap_jso(this.raw.createAnswer(unwrap_jso(successCallback), unwrap_jso(failureCallback), unwrap_jso(mediaConstraints)));
    }
    [_createAnswer_2](successCallback, failureCallback) {
      return wrap_jso(this.raw.createAnswer(unwrap_jso(successCallback), unwrap_jso(failureCallback)));
    }
    createDtmfSender(track) {
      return this[_createDtmfSender_1](track);
    }
    [_createDtmfSender_1](track) {
      return dart.as(wrap_jso(this.raw.createDTMFSender(unwrap_jso(track))), RtcDtmfSender);
    }
    createDataChannel(label, options) {
      if (options === void 0)
        options = null;
      if (options != null) {
        let options_1 = html_common.convertDartToNative_Dictionary(options);
        return this[_createDataChannel_1](label, options_1);
      }
      return this[_createDataChannel_2](label);
    }
    [_createDataChannel_1](label, options) {
      return dart.as(wrap_jso(this.raw.createDataChannel(unwrap_jso(label), unwrap_jso(options))), RtcDataChannel);
    }
    [_createDataChannel_2](label) {
      return dart.as(wrap_jso(this.raw.createDataChannel(unwrap_jso(label))), RtcDataChannel);
    }
    [_createOffer](successCallback, failureCallback, rtcOfferOptions) {
      if (failureCallback === void 0)
        failureCallback = null;
      if (rtcOfferOptions === void 0)
        rtcOfferOptions = null;
      if (rtcOfferOptions != null) {
        let rtcOfferOptions_1 = html_common.convertDartToNative_Dictionary(rtcOfferOptions);
        this[_createOffer_1](successCallback, failureCallback, rtcOfferOptions_1);
        return;
      }
      this[_createOffer_2](successCallback, failureCallback);
      return;
    }
    [_createOffer_1](successCallback, failureCallback, rtcOfferOptions) {
      return wrap_jso(this.raw.createOffer(unwrap_jso(successCallback), unwrap_jso(failureCallback), unwrap_jso(rtcOfferOptions)));
    }
    [_createOffer_2](successCallback, failureCallback) {
      return wrap_jso(this.raw.createOffer(unwrap_jso(successCallback), unwrap_jso(failureCallback)));
    }
    getLocalStreams() {
      return this[_getLocalStreams_1]();
    }
    [_getLocalStreams_1]() {
      return dart.as(wrap_jso(this.raw.getLocalStreams()), core.List$(dart.dynamic));
    }
    getRemoteStreams() {
      return this[_getRemoteStreams_1]();
    }
    [_getRemoteStreams_1]() {
      return dart.as(wrap_jso(this.raw.getRemoteStreams()), core.List$(dart.dynamic));
    }
    [_getStats](successCallback, selector) {
      this[_getStats_1](successCallback, selector);
      return;
    }
    [_getStats_1](successCallback, selector) {
      return wrap_jso(this.raw.getStats(unwrap_jso(successCallback), unwrap_jso(selector)));
    }
    getStreamById(streamId) {
      return this[_getStreamById_1](streamId);
    }
    [_getStreamById_1](streamId) {
      return wrap_jso(this.raw.getStreamById(unwrap_jso(streamId)));
    }
    removeStream(stream) {
      this[_removeStream_1](stream);
      return;
    }
    [_removeStream_1](stream) {
      return wrap_jso(this.raw.removeStream(unwrap_jso(stream)));
    }
    [_setLocalDescription](description, successCallback, failureCallback) {
      if (successCallback === void 0)
        successCallback = null;
      if (failureCallback === void 0)
        failureCallback = null;
      this[_setLocalDescription_1](description, successCallback, failureCallback);
      return;
    }
    [_setLocalDescription_1](description, successCallback, failureCallback) {
      return wrap_jso(this.raw.setLocalDescription(unwrap_jso(description), unwrap_jso(successCallback), unwrap_jso(failureCallback)));
    }
    setLocalDescription(description) {
      let completer = async.Completer.new();
      this[_setLocalDescription](description, dart.fn(() => {
        completer.complete();
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    [_setRemoteDescription](description, successCallback, failureCallback) {
      if (successCallback === void 0)
        successCallback = null;
      if (failureCallback === void 0)
        failureCallback = null;
      this[_setRemoteDescription_1](description, successCallback, failureCallback);
      return;
    }
    [_setRemoteDescription_1](description, successCallback, failureCallback) {
      return wrap_jso(this.raw.setRemoteDescription(unwrap_jso(description), unwrap_jso(successCallback), unwrap_jso(failureCallback)));
    }
    setRemoteDescription(description) {
      let completer = async.Completer.new();
      this[_setRemoteDescription](description, dart.fn(() => {
        completer.complete();
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    updateIce(configuration, mediaConstraints) {
      if (configuration === void 0)
        configuration = null;
      if (mediaConstraints === void 0)
        mediaConstraints = null;
      if (mediaConstraints != null) {
        let configuration_1 = html_common.convertDartToNative_Dictionary(configuration);
        let mediaConstraints_2 = html_common.convertDartToNative_Dictionary(mediaConstraints);
        this[_updateIce_1](configuration_1, mediaConstraints_2);
        return;
      }
      if (configuration != null) {
        let configuration_1 = html_common.convertDartToNative_Dictionary(configuration);
        this[_updateIce_2](configuration_1);
        return;
      }
      this[_updateIce_3]();
      return;
    }
    [_updateIce_1](configuration, mediaConstraints) {
      return wrap_jso(this.raw.updateIce(unwrap_jso(configuration), unwrap_jso(mediaConstraints)));
    }
    [_updateIce_2](configuration) {
      return wrap_jso(this.raw.updateIce(unwrap_jso(configuration)));
    }
    [_updateIce_3]() {
      return wrap_jso(this.raw.updateIce());
    }
    get onAddStream() {
      return RtcPeerConnection.addStreamEvent.forTarget(this);
    }
    get onDataChannel() {
      return RtcPeerConnection.dataChannelEvent.forTarget(this);
    }
    get onIceCandidate() {
      return RtcPeerConnection.iceCandidateEvent.forTarget(this);
    }
    get onIceConnectionStateChange() {
      return RtcPeerConnection.iceConnectionStateChangeEvent.forTarget(this);
    }
    get onNegotiationNeeded() {
      return RtcPeerConnection.negotiationNeededEvent.forTarget(this);
    }
    get onRemoveStream() {
      return RtcPeerConnection.removeStreamEvent.forTarget(this);
    }
    get onSignalingStateChange() {
      return RtcPeerConnection.signalingStateChangeEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(RtcPeerConnection, 'internal_');
  dart.setSignature(RtcPeerConnection, {
    constructors: () => ({
      _: [RtcPeerConnection, []],
      _internalWrap: [RtcPeerConnection, []],
      internal_: [RtcPeerConnection, []]
    }),
    methods: () => ({
      createOffer: [async.Future$(RtcSessionDescription), [], [core.Map]],
      createAnswer: [async.Future$(RtcSessionDescription), [], [core.Map]],
      getStats: [async.Future$(RtcStatsResponse), [dart.dynamic]],
      addIceCandidate: [dart.void, [RtcIceCandidate, VoidCallback, _RtcErrorCallback]],
      [_addIceCandidate_1]: [dart.void, [RtcIceCandidate, VoidCallback, _RtcErrorCallback]],
      addStream: [dart.void, [dart.dynamic], [core.Map]],
      [_addStream_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_addStream_2]: [dart.void, [dart.dynamic]],
      close: [dart.void, []],
      [_close_1]: [dart.void, []],
      [_createAnswer]: [dart.void, [_RtcSessionDescriptionCallback], [_RtcErrorCallback, core.Map]],
      [_createAnswer_1]: [dart.void, [_RtcSessionDescriptionCallback, _RtcErrorCallback, dart.dynamic]],
      [_createAnswer_2]: [dart.void, [_RtcSessionDescriptionCallback, _RtcErrorCallback]],
      createDtmfSender: [RtcDtmfSender, [dart.dynamic]],
      [_createDtmfSender_1]: [RtcDtmfSender, [dart.dynamic]],
      createDataChannel: [RtcDataChannel, [core.String], [core.Map]],
      [_createDataChannel_1]: [RtcDataChannel, [dart.dynamic, dart.dynamic]],
      [_createDataChannel_2]: [RtcDataChannel, [dart.dynamic]],
      [_createOffer]: [dart.void, [_RtcSessionDescriptionCallback], [_RtcErrorCallback, core.Map]],
      [_createOffer_1]: [dart.void, [_RtcSessionDescriptionCallback, _RtcErrorCallback, dart.dynamic]],
      [_createOffer_2]: [dart.void, [_RtcSessionDescriptionCallback, _RtcErrorCallback]],
      getLocalStreams: [core.List$(dart.dynamic), []],
      [_getLocalStreams_1]: [core.List$(dart.dynamic), []],
      getRemoteStreams: [core.List$(dart.dynamic), []],
      [_getRemoteStreams_1]: [core.List$(dart.dynamic), []],
      [_getStats]: [dart.void, [RtcStatsCallback, dart.dynamic]],
      [_getStats_1]: [dart.void, [RtcStatsCallback, dart.dynamic]],
      getStreamById: [dart.dynamic, [core.String]],
      [_getStreamById_1]: [dart.dynamic, [dart.dynamic]],
      removeStream: [dart.void, [dart.dynamic]],
      [_removeStream_1]: [dart.void, [dart.dynamic]],
      [_setLocalDescription]: [dart.void, [RtcSessionDescription], [VoidCallback, _RtcErrorCallback]],
      [_setLocalDescription_1]: [dart.void, [RtcSessionDescription, VoidCallback, _RtcErrorCallback]],
      setLocalDescription: [async.Future, [RtcSessionDescription]],
      [_setRemoteDescription]: [dart.void, [RtcSessionDescription], [VoidCallback, _RtcErrorCallback]],
      [_setRemoteDescription_1]: [dart.void, [RtcSessionDescription, VoidCallback, _RtcErrorCallback]],
      setRemoteDescription: [async.Future, [RtcSessionDescription]],
      updateIce: [dart.void, [], [core.Map, core.Map]],
      [_updateIce_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_updateIce_2]: [dart.void, [dart.dynamic]],
      [_updateIce_3]: [dart.void, []]
    }),
    statics: () => ({internalCreateRtcPeerConnection: [RtcPeerConnection, []]}),
    names: ['internalCreateRtcPeerConnection']
  });
  RtcPeerConnection[dart.metadata] = () => [dart.const(new _metadata.DomName('RTCPeerConnection')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCPeerConnection,mozRTCPeerConnection"))];
  RtcPeerConnection.addStreamEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('addstream'));
  RtcPeerConnection.dataChannelEvent = dart.const(new (EventStreamProvider$(RtcDataChannelEvent))('datachannel'));
  RtcPeerConnection.iceCandidateEvent = dart.const(new (EventStreamProvider$(RtcIceCandidateEvent))('icecandidate'));
  RtcPeerConnection.iceConnectionStateChangeEvent = dart.const(new (EventStreamProvider$(Event))('iceconnectionstatechange'));
  RtcPeerConnection.negotiationNeededEvent = dart.const(new (EventStreamProvider$(Event))('negotiationneeded'));
  RtcPeerConnection.removeStreamEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('removestream'));
  RtcPeerConnection.signalingStateChangeEvent = dart.const(new (EventStreamProvider$(Event))('signalingstatechange'));
  class RtcSessionDescription extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcSessionDescription() {
      return RtcSessionDescription._internalWrap();
    }
    static _internalWrap() {
      return new RtcSessionDescription.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get sdp() {
      return dart.as(wrap_jso(this.raw.sdp), core.String);
    }
    set sdp(val) {
      return this.raw.sdp = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(RtcSessionDescription, 'internal_');
  dart.setSignature(RtcSessionDescription, {
    constructors: () => ({
      _: [RtcSessionDescription, []],
      _internalWrap: [RtcSessionDescription, []],
      internal_: [RtcSessionDescription, []]
    }),
    statics: () => ({internalCreateRtcSessionDescription: [RtcSessionDescription, []]}),
    names: ['internalCreateRtcSessionDescription']
  });
  RtcSessionDescription[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCSessionDescription')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCSessionDescription,mozRTCSessionDescription"))];
  const _get_timestamp = Symbol('_get_timestamp');
  const _names_1 = Symbol('_names_1');
  const _stat_1 = Symbol('_stat_1');
  class RtcStatsReport extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcStatsReport() {
      return RtcStatsReport._internalWrap();
    }
    static _internalWrap() {
      return new RtcStatsReport.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    get local() {
      return dart.as(wrap_jso(this.raw.local), RtcStatsReport);
    }
    get remote() {
      return dart.as(wrap_jso(this.raw.remote), RtcStatsReport);
    }
    get timestamp() {
      return html_common.convertNativeToDart_DateTime(this[_get_timestamp]);
    }
    get [_get_timestamp]() {
      return wrap_jso(this.raw.timestamp);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    names() {
      return this[_names_1]();
    }
    [_names_1]() {
      return dart.as(wrap_jso(this.raw.names()), core.List$(core.String));
    }
    stat(name) {
      return this[_stat_1](name);
    }
    [_stat_1](name) {
      return dart.as(wrap_jso(this.raw.stat(unwrap_jso(name))), core.String);
    }
  }
  dart.defineNamedConstructor(RtcStatsReport, 'internal_');
  dart.setSignature(RtcStatsReport, {
    constructors: () => ({
      _: [RtcStatsReport, []],
      _internalWrap: [RtcStatsReport, []],
      internal_: [RtcStatsReport, []]
    }),
    methods: () => ({
      names: [core.List$(core.String), []],
      [_names_1]: [core.List$(core.String), []],
      stat: [core.String, [core.String]],
      [_stat_1]: [core.String, [dart.dynamic]]
    }),
    statics: () => ({internalCreateRtcStatsReport: [RtcStatsReport, []]}),
    names: ['internalCreateRtcStatsReport']
  });
  RtcStatsReport[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCStatsReport')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCStatsReport"))];
  const _result_1 = Symbol('_result_1');
  class RtcStatsResponse extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateRtcStatsResponse() {
      return RtcStatsResponse._internalWrap();
    }
    static _internalWrap() {
      return new RtcStatsResponse.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), RtcStatsReport);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), RtcStatsReport);
    }
    result() {
      return this[_result_1]();
    }
    [_result_1]() {
      return dart.as(wrap_jso(this.raw.result()), core.List$(RtcStatsReport));
    }
  }
  dart.defineNamedConstructor(RtcStatsResponse, 'internal_');
  dart.setSignature(RtcStatsResponse, {
    constructors: () => ({
      _: [RtcStatsResponse, []],
      _internalWrap: [RtcStatsResponse, []],
      internal_: [RtcStatsResponse, []]
    }),
    methods: () => ({
      [__getter__]: [RtcStatsReport, [core.String]],
      [__getter___1]: [RtcStatsReport, [dart.dynamic]],
      namedItem: [RtcStatsReport, [core.String]],
      [_namedItem_1]: [RtcStatsReport, [dart.dynamic]],
      result: [core.List$(RtcStatsReport), []],
      [_result_1]: [core.List$(RtcStatsReport), []]
    }),
    statics: () => ({internalCreateRtcStatsResponse: [RtcStatsResponse, []]}),
    names: ['internalCreateRtcStatsResponse']
  });
  RtcStatsResponse[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RTCStatsResponse')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("RTCStatsResponse"))];
  const _availLeft = Symbol('_availLeft');
  const _availTop = Symbol('_availTop');
  const _availWidth = Symbol('_availWidth');
  const _availHeight = Symbol('_availHeight');
  class Screen extends DartHtmlDomObject {
    get available() {
      return new math.Rectangle(this[_availLeft], this[_availTop], this[_availWidth], this[_availHeight]);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateScreen() {
      return Screen._internalWrap();
    }
    static _internalWrap() {
      return new Screen.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get [_availHeight]() {
      return dart.as(wrap_jso(this.raw.availHeight), core.int);
    }
    get [_availLeft]() {
      return dart.as(wrap_jso(this.raw.availLeft), core.int);
    }
    get [_availTop]() {
      return dart.as(wrap_jso(this.raw.availTop), core.int);
    }
    get [_availWidth]() {
      return dart.as(wrap_jso(this.raw.availWidth), core.int);
    }
    get colorDepth() {
      return dart.as(wrap_jso(this.raw.colorDepth), core.int);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.int);
    }
    get orientation() {
      return dart.as(wrap_jso(this.raw.orientation), ScreenOrientation);
    }
    get pixelDepth() {
      return dart.as(wrap_jso(this.raw.pixelDepth), core.int);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.int);
    }
  }
  dart.defineNamedConstructor(Screen, 'internal_');
  dart.setSignature(Screen, {
    constructors: () => ({
      _: [Screen, []],
      _internalWrap: [Screen, []],
      internal_: [Screen, []]
    }),
    statics: () => ({internalCreateScreen: [Screen, []]}),
    names: ['internalCreateScreen']
  });
  Screen[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Screen')), dart.const(new _js_helper.Native("Screen"))];
  const _lock_1 = Symbol('_lock_1');
  const _unlock_1 = Symbol('_unlock_1');
  class ScreenOrientation extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateScreenOrientation() {
      return ScreenOrientation._internalWrap();
    }
    static _internalWrap() {
      return new ScreenOrientation.internal_();
    }
    internal_() {
      super.internal_();
    }
    get angle() {
      return dart.as(wrap_jso(this.raw.angle), core.int);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    lock(orientation) {
      return this[_lock_1](orientation);
    }
    [_lock_1](orientation) {
      return dart.as(wrap_jso(this.raw.lock(unwrap_jso(orientation))), async.Future);
    }
    unlock() {
      this[_unlock_1]();
      return;
    }
    [_unlock_1]() {
      return wrap_jso(this.raw.unlock());
    }
    get onChange() {
      return ScreenOrientation.changeEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(ScreenOrientation, 'internal_');
  dart.setSignature(ScreenOrientation, {
    constructors: () => ({
      _: [ScreenOrientation, []],
      _internalWrap: [ScreenOrientation, []],
      internal_: [ScreenOrientation, []]
    }),
    methods: () => ({
      lock: [async.Future, [core.String]],
      [_lock_1]: [async.Future, [dart.dynamic]],
      unlock: [dart.void, []],
      [_unlock_1]: [dart.void, []]
    }),
    statics: () => ({internalCreateScreenOrientation: [ScreenOrientation, []]}),
    names: ['internalCreateScreenOrientation']
  });
  ScreenOrientation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ScreenOrientation')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ScreenOrientation"))];
  ScreenOrientation.changeEvent = dart.const(new (EventStreamProvider$(Event))('change'));
  class ScriptElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("script"), ScriptElement);
    }
    static internalCreateScriptElement() {
      return ScriptElement._internalWrap();
    }
    static _internalWrap() {
      return new ScriptElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get async() {
      return dart.as(wrap_jso(this.raw.async), core.bool);
    }
    set async(val) {
      return this.raw.async = unwrap_jso(val);
    }
    get charset() {
      return dart.as(wrap_jso(this.raw.charset), core.String);
    }
    set charset(val) {
      return this.raw.charset = unwrap_jso(val);
    }
    get crossOrigin() {
      return dart.as(wrap_jso(this.raw.crossOrigin), core.String);
    }
    set crossOrigin(val) {
      return this.raw.crossOrigin = unwrap_jso(val);
    }
    get defer() {
      return dart.as(wrap_jso(this.raw.defer), core.bool);
    }
    set defer(val) {
      return this.raw.defer = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get nonce() {
      return dart.as(wrap_jso(this.raw.nonce), core.String);
    }
    set nonce(val) {
      return this.raw.nonce = unwrap_jso(val);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(ScriptElement, 'internal_');
  dart.defineNamedConstructor(ScriptElement, 'created');
  dart.setSignature(ScriptElement, {
    constructors: () => ({
      _: [ScriptElement, []],
      new: [ScriptElement, []],
      _internalWrap: [ScriptElement, []],
      internal_: [ScriptElement, []],
      created: [ScriptElement, []]
    }),
    statics: () => ({internalCreateScriptElement: [ScriptElement, []]}),
    names: ['internalCreateScriptElement']
  });
  ScriptElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLScriptElement')), dart.const(new _js_helper.Native("HTMLScriptElement"))];
  class SelectElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("select"), SelectElement);
    }
    static internalCreateSelectElement() {
      return SelectElement._internalWrap();
    }
    static _internalWrap() {
      return new SelectElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get autofocus() {
      return dart.as(wrap_jso(this.raw.autofocus), core.bool);
    }
    set autofocus(val) {
      return this.raw.autofocus = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    set length(val) {
      return this.raw.length = unwrap_jso(val);
    }
    get multiple() {
      return dart.as(wrap_jso(this.raw.multiple), core.bool);
    }
    set multiple(val) {
      return this.raw.multiple = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get required() {
      return dart.as(wrap_jso(this.raw.required), core.bool);
    }
    set required(val) {
      return this.raw.required = unwrap_jso(val);
    }
    get selectedIndex() {
      return dart.as(wrap_jso(this.raw.selectedIndex), core.int);
    }
    set selectedIndex(val) {
      return this.raw.selectedIndex = unwrap_jso(val);
    }
    get size() {
      return dart.as(wrap_jso(this.raw.size), core.int);
    }
    set size(val) {
      return this.raw.size = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    [__setter__](index, value) {
      this[__setter___1](index, value);
      return;
    }
    [__setter___1](index, value) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(index), unwrap_jso(value)));
    }
    add(element, before) {
      this[_add_1](element, before);
      return;
    }
    [_add_1](element, before) {
      return wrap_jso(this.raw.add(unwrap_jso(element), unwrap_jso(before)));
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Element);
    }
    namedItem(name) {
      return this[_namedItem_1](name);
    }
    [_namedItem_1](name) {
      return dart.as(wrap_jso(this.raw.namedItem(unwrap_jso(name))), Element);
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
    get options() {
      let options = this.querySelectorAll('option').where(dart.fn(e => dart.is(e, OptionElement), core.bool, [dart.dynamic]))[dartx.toList]();
      return new collection.UnmodifiableListView(options);
    }
    get selectedOptions() {
      if (dart.notNull(this.multiple)) {
        let options = this.options[dartx.where](dart.fn(o => dart.as(dart.dload(o, 'selected'), core.bool), core.bool, [dart.dynamic]))[dartx.toList]();
        return new (collection.UnmodifiableListView$(OptionElement))(options);
      } else {
        return dart.list([this.options[dartx.get](this.selectedIndex)], OptionElement);
      }
    }
  }
  dart.defineNamedConstructor(SelectElement, 'internal_');
  dart.defineNamedConstructor(SelectElement, 'created');
  dart.setSignature(SelectElement, {
    constructors: () => ({
      _: [SelectElement, []],
      new: [SelectElement, []],
      _internalWrap: [SelectElement, []],
      internal_: [SelectElement, []],
      created: [SelectElement, []]
    }),
    methods: () => ({
      [__setter__]: [dart.void, [core.int, OptionElement]],
      [__setter___1]: [dart.void, [dart.dynamic, OptionElement]],
      add: [dart.void, [HtmlElement, core.int]],
      [_add_1]: [dart.void, [HtmlElement, dart.dynamic]],
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      item: [Element, [core.int]],
      [_item_1]: [Element, [dart.dynamic]],
      namedItem: [Element, [core.String]],
      [_namedItem_1]: [Element, [dart.dynamic]],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateSelectElement: [SelectElement, []]}),
    names: ['internalCreateSelectElement']
  });
  SelectElement[dart.metadata] = () => [dart.const(new _metadata.DomName('HTMLSelectElement')), dart.const(new _js_helper.Native("HTMLSelectElement"))];
  const _addRange_1 = Symbol('_addRange_1');
  const _collapseToEnd_1 = Symbol('_collapseToEnd_1');
  const _collapseToStart_1 = Symbol('_collapseToStart_1');
  const _containsNode_1 = Symbol('_containsNode_1');
  const _deleteFromDocument_1 = Symbol('_deleteFromDocument_1');
  const _empty_1 = Symbol('_empty_1');
  const _extend_1 = Symbol('_extend_1');
  const _extend_2 = Symbol('_extend_2');
  const _getRangeAt_1 = Symbol('_getRangeAt_1');
  const _modify_1 = Symbol('_modify_1');
  const _removeAllRanges_1 = Symbol('_removeAllRanges_1');
  const _selectAllChildren_1 = Symbol('_selectAllChildren_1');
  const _setBaseAndExtent_1 = Symbol('_setBaseAndExtent_1');
  const _setPosition_1 = Symbol('_setPosition_1');
  const _setPosition_2 = Symbol('_setPosition_2');
  class Selection extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSelection() {
      return Selection._internalWrap();
    }
    static _internalWrap() {
      return new Selection.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get anchorNode() {
      return dart.as(wrap_jso(this.raw.anchorNode), Node);
    }
    get anchorOffset() {
      return dart.as(wrap_jso(this.raw.anchorOffset), core.int);
    }
    get baseNode() {
      return dart.as(wrap_jso(this.raw.baseNode), Node);
    }
    get baseOffset() {
      return dart.as(wrap_jso(this.raw.baseOffset), core.int);
    }
    get extentNode() {
      return dart.as(wrap_jso(this.raw.extentNode), Node);
    }
    get extentOffset() {
      return dart.as(wrap_jso(this.raw.extentOffset), core.int);
    }
    get focusNode() {
      return dart.as(wrap_jso(this.raw.focusNode), Node);
    }
    get focusOffset() {
      return dart.as(wrap_jso(this.raw.focusOffset), core.int);
    }
    get isCollapsed() {
      return dart.as(wrap_jso(this.raw.isCollapsed), core.bool);
    }
    get rangeCount() {
      return dart.as(wrap_jso(this.raw.rangeCount), core.int);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    addRange(range) {
      this[_addRange_1](range);
      return;
    }
    [_addRange_1](range) {
      return wrap_jso(this.raw.addRange(unwrap_jso(range)));
    }
    collapse(node, offset) {
      if (offset === void 0)
        offset = null;
      if (offset != null) {
        this[_collapse_1](node, offset);
        return;
      }
      this[_collapse_2](node);
      return;
    }
    [_collapse_1](node, offset) {
      return wrap_jso(this.raw.collapse(unwrap_jso(node), unwrap_jso(offset)));
    }
    [_collapse_2](node) {
      return wrap_jso(this.raw.collapse(unwrap_jso(node)));
    }
    collapseToEnd() {
      this[_collapseToEnd_1]();
      return;
    }
    [_collapseToEnd_1]() {
      return wrap_jso(this.raw.collapseToEnd());
    }
    collapseToStart() {
      this[_collapseToStart_1]();
      return;
    }
    [_collapseToStart_1]() {
      return wrap_jso(this.raw.collapseToStart());
    }
    containsNode(node, allowPartial) {
      return this[_containsNode_1](node, allowPartial);
    }
    [_containsNode_1](node, allowPartial) {
      return dart.as(wrap_jso(this.raw.containsNode(unwrap_jso(node), unwrap_jso(allowPartial))), core.bool);
    }
    deleteFromDocument() {
      this[_deleteFromDocument_1]();
      return;
    }
    [_deleteFromDocument_1]() {
      return wrap_jso(this.raw.deleteFromDocument());
    }
    empty() {
      this[_empty_1]();
      return;
    }
    [_empty_1]() {
      return wrap_jso(this.raw.empty());
    }
    extend(node, offset) {
      if (offset === void 0)
        offset = null;
      if (offset != null) {
        this[_extend_1](node, offset);
        return;
      }
      this[_extend_2](node);
      return;
    }
    [_extend_1](node, offset) {
      return wrap_jso(this.raw.extend(unwrap_jso(node), unwrap_jso(offset)));
    }
    [_extend_2](node) {
      return wrap_jso(this.raw.extend(unwrap_jso(node)));
    }
    getRangeAt(index) {
      return this[_getRangeAt_1](index);
    }
    [_getRangeAt_1](index) {
      return dart.as(wrap_jso(this.raw.getRangeAt(unwrap_jso(index))), Range);
    }
    modify(alter, direction, granularity) {
      this[_modify_1](alter, direction, granularity);
      return;
    }
    [_modify_1](alter, direction, granularity) {
      return wrap_jso(this.raw.modify(unwrap_jso(alter), unwrap_jso(direction), unwrap_jso(granularity)));
    }
    removeAllRanges() {
      this[_removeAllRanges_1]();
      return;
    }
    [_removeAllRanges_1]() {
      return wrap_jso(this.raw.removeAllRanges());
    }
    selectAllChildren(node) {
      this[_selectAllChildren_1](node);
      return;
    }
    [_selectAllChildren_1](node) {
      return wrap_jso(this.raw.selectAllChildren(unwrap_jso(node)));
    }
    setBaseAndExtent(baseNode, baseOffset, extentNode, extentOffset) {
      this[_setBaseAndExtent_1](baseNode, baseOffset, extentNode, extentOffset);
      return;
    }
    [_setBaseAndExtent_1](baseNode, baseOffset, extentNode, extentOffset) {
      return wrap_jso(this.raw.setBaseAndExtent(unwrap_jso(baseNode), unwrap_jso(baseOffset), unwrap_jso(extentNode), unwrap_jso(extentOffset)));
    }
    setPosition(node, offset) {
      if (offset === void 0)
        offset = null;
      if (offset != null) {
        this[_setPosition_1](node, offset);
        return;
      }
      this[_setPosition_2](node);
      return;
    }
    [_setPosition_1](node, offset) {
      return wrap_jso(this.raw.setPosition(unwrap_jso(node), unwrap_jso(offset)));
    }
    [_setPosition_2](node) {
      return wrap_jso(this.raw.setPosition(unwrap_jso(node)));
    }
  }
  dart.defineNamedConstructor(Selection, 'internal_');
  dart.setSignature(Selection, {
    constructors: () => ({
      _: [Selection, []],
      _internalWrap: [Selection, []],
      internal_: [Selection, []]
    }),
    methods: () => ({
      addRange: [dart.void, [Range]],
      [_addRange_1]: [dart.void, [Range]],
      collapse: [dart.void, [Node], [core.int]],
      [_collapse_1]: [dart.void, [Node, dart.dynamic]],
      [_collapse_2]: [dart.void, [Node]],
      collapseToEnd: [dart.void, []],
      [_collapseToEnd_1]: [dart.void, []],
      collapseToStart: [dart.void, []],
      [_collapseToStart_1]: [dart.void, []],
      containsNode: [core.bool, [Node, core.bool]],
      [_containsNode_1]: [core.bool, [Node, dart.dynamic]],
      deleteFromDocument: [dart.void, []],
      [_deleteFromDocument_1]: [dart.void, []],
      empty: [dart.void, []],
      [_empty_1]: [dart.void, []],
      extend: [dart.void, [Node], [core.int]],
      [_extend_1]: [dart.void, [Node, dart.dynamic]],
      [_extend_2]: [dart.void, [Node]],
      getRangeAt: [Range, [core.int]],
      [_getRangeAt_1]: [Range, [dart.dynamic]],
      modify: [dart.void, [core.String, core.String, core.String]],
      [_modify_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      removeAllRanges: [dart.void, []],
      [_removeAllRanges_1]: [dart.void, []],
      selectAllChildren: [dart.void, [Node]],
      [_selectAllChildren_1]: [dart.void, [Node]],
      setBaseAndExtent: [dart.void, [Node, core.int, Node, core.int]],
      [_setBaseAndExtent_1]: [dart.void, [Node, dart.dynamic, Node, dart.dynamic]],
      setPosition: [dart.void, [Node], [core.int]],
      [_setPosition_1]: [dart.void, [Node, dart.dynamic]],
      [_setPosition_2]: [dart.void, [Node]]
    }),
    statics: () => ({internalCreateSelection: [Selection, []]}),
    names: ['internalCreateSelection']
  });
  Selection[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Selection')), dart.const(new _js_helper.Native("Selection"))];
  class ServiceWorkerClient extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateServiceWorkerClient() {
      return ServiceWorkerClient._internalWrap();
    }
    static _internalWrap() {
      return new ServiceWorkerClient.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.int);
    }
    postMessage(message, transfer) {
      if (transfer === void 0)
        transfer = null;
      if (transfer != null) {
        this[_postMessage_1](message, transfer);
        return;
      }
      this[_postMessage_2](message);
      return;
    }
    [_postMessage_1](message, transfer) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message), unwrap_jso(transfer)));
    }
    [_postMessage_2](message) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message)));
    }
  }
  dart.defineNamedConstructor(ServiceWorkerClient, 'internal_');
  dart.setSignature(ServiceWorkerClient, {
    constructors: () => ({
      _: [ServiceWorkerClient, []],
      _internalWrap: [ServiceWorkerClient, []],
      internal_: [ServiceWorkerClient, []]
    }),
    methods: () => ({
      postMessage: [dart.void, [dart.dynamic], [core.List$(MessagePort)]],
      [_postMessage_1]: [dart.void, [dart.dynamic, core.List$(MessagePort)]],
      [_postMessage_2]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateServiceWorkerClient: [ServiceWorkerClient, []]}),
    names: ['internalCreateServiceWorkerClient']
  });
  ServiceWorkerClient[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorkerClient')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorkerClient"))];
  const _getAll_1 = Symbol('_getAll_1');
  const _getAll_2 = Symbol('_getAll_2');
  class ServiceWorkerClients extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateServiceWorkerClients() {
      return ServiceWorkerClients._internalWrap();
    }
    static _internalWrap() {
      return new ServiceWorkerClients.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    getAll(options) {
      if (options === void 0)
        options = null;
      if (options != null) {
        let options_1 = html_common.convertDartToNative_Dictionary(options);
        return this[_getAll_1](options_1);
      }
      return this[_getAll_2]();
    }
    [_getAll_1](options) {
      return dart.as(wrap_jso(this.raw.getAll(unwrap_jso(options))), async.Future);
    }
    [_getAll_2]() {
      return dart.as(wrap_jso(this.raw.getAll()), async.Future);
    }
  }
  dart.defineNamedConstructor(ServiceWorkerClients, 'internal_');
  dart.setSignature(ServiceWorkerClients, {
    constructors: () => ({
      _: [ServiceWorkerClients, []],
      _internalWrap: [ServiceWorkerClients, []],
      internal_: [ServiceWorkerClients, []]
    }),
    methods: () => ({
      getAll: [async.Future, [], [core.Map]],
      [_getAll_1]: [async.Future, [dart.dynamic]],
      [_getAll_2]: [async.Future, []]
    }),
    statics: () => ({internalCreateServiceWorkerClients: [ServiceWorkerClients, []]}),
    names: ['internalCreateServiceWorkerClients']
  });
  ServiceWorkerClients[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorkerClients')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorkerClients"))];
  const _getRegistration_1 = Symbol('_getRegistration_1');
  const _getRegistration_2 = Symbol('_getRegistration_2');
  const _register_2 = Symbol('_register_2');
  class ServiceWorkerContainer extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateServiceWorkerContainer() {
      return ServiceWorkerContainer._internalWrap();
    }
    static _internalWrap() {
      return new ServiceWorkerContainer.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get controller() {
      return dart.as(wrap_jso(this.raw.controller), _ServiceWorker);
    }
    get ready() {
      return dart.as(wrap_jso(this.raw.ready), async.Future);
    }
    getRegistration(documentURL) {
      if (documentURL === void 0)
        documentURL = null;
      if (documentURL != null) {
        return this[_getRegistration_1](documentURL);
      }
      return this[_getRegistration_2]();
    }
    [_getRegistration_1](documentURL) {
      return dart.as(wrap_jso(this.raw.getRegistration(unwrap_jso(documentURL))), async.Future);
    }
    [_getRegistration_2]() {
      return dart.as(wrap_jso(this.raw.getRegistration()), async.Future);
    }
    register(url, options) {
      if (options === void 0)
        options = null;
      if (options != null) {
        let options_1 = html_common.convertDartToNative_Dictionary(options);
        return this[_register_1](url, options_1);
      }
      return this[_register_2](url);
    }
    [_register_1](url, options) {
      return dart.as(wrap_jso(this.raw.register(unwrap_jso(url), unwrap_jso(options))), async.Future);
    }
    [_register_2](url) {
      return dart.as(wrap_jso(this.raw.register(unwrap_jso(url))), async.Future);
    }
  }
  dart.defineNamedConstructor(ServiceWorkerContainer, 'internal_');
  dart.setSignature(ServiceWorkerContainer, {
    constructors: () => ({
      _: [ServiceWorkerContainer, []],
      _internalWrap: [ServiceWorkerContainer, []],
      internal_: [ServiceWorkerContainer, []]
    }),
    methods: () => ({
      getRegistration: [async.Future, [], [core.String]],
      [_getRegistration_1]: [async.Future, [dart.dynamic]],
      [_getRegistration_2]: [async.Future, []],
      register: [async.Future, [core.String], [core.Map]],
      [_register_1]: [async.Future, [dart.dynamic, dart.dynamic]],
      [_register_2]: [async.Future, [dart.dynamic]]
    }),
    statics: () => ({internalCreateServiceWorkerContainer: [ServiceWorkerContainer, []]}),
    names: ['internalCreateServiceWorkerContainer']
  });
  ServiceWorkerContainer[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorkerContainer')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorkerContainer"))];
  const _importScripts_1 = Symbol('_importScripts_1');
  const _webkitRequestFileSystem_1 = Symbol('_webkitRequestFileSystem_1');
  const _webkitRequestFileSystem_2 = Symbol('_webkitRequestFileSystem_2');
  const _webkitRequestFileSystem_3 = Symbol('_webkitRequestFileSystem_3');
  const _webkitRequestFileSystem = Symbol('_webkitRequestFileSystem');
  const _requestFileSystemSync_1 = Symbol('_requestFileSystemSync_1');
  const _resolveLocalFileSystemSyncUrl_1 = Symbol('_resolveLocalFileSystemSyncUrl_1');
  const _webkitResolveLocalFileSystemUrl_1 = Symbol('_webkitResolveLocalFileSystemUrl_1');
  const _webkitResolveLocalFileSystemUrl_2 = Symbol('_webkitResolveLocalFileSystemUrl_2');
  const _webkitResolveLocalFileSystemUrl = Symbol('_webkitResolveLocalFileSystemUrl');
  const _atob_1 = Symbol('_atob_1');
  const _btoa_1 = Symbol('_btoa_1');
  const _clearInterval_1 = Symbol('_clearInterval_1');
  const _clearInterval = Symbol('_clearInterval');
  const _clearTimeout_1 = Symbol('_clearTimeout_1');
  const _clearTimeout = Symbol('_clearTimeout');
  const _setInterval_1 = Symbol('_setInterval_1');
  const _setInterval = Symbol('_setInterval');
  const _setTimeout_1 = Symbol('_setTimeout_1');
  const _setTimeout = Symbol('_setTimeout');
  class WorkerGlobalScope extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateWorkerGlobalScope() {
      return WorkerGlobalScope._internalWrap();
    }
    static _internalWrap() {
      return new WorkerGlobalScope.internal_();
    }
    internal_() {
      super.internal_();
    }
    get console() {
      return wrap_jso(this.raw.console);
    }
    get crypto() {
      return wrap_jso(this.raw.crypto);
    }
    get indexedDB() {
      return wrap_jso(this.raw.indexedDB);
    }
    get location() {
      return dart.as(wrap_jso(this.raw.location), _WorkerLocation);
    }
    get navigator() {
      return dart.as(wrap_jso(this.raw.navigator), _WorkerNavigator);
    }
    get performance() {
      return dart.as(wrap_jso(this.raw.performance), WorkerPerformance);
    }
    get self() {
      return dart.as(wrap_jso(this.raw.self), WorkerGlobalScope);
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    importScripts(urls) {
      this[_importScripts_1](urls);
      return;
    }
    [_importScripts_1](urls) {
      return wrap_jso(this.raw.importScripts(unwrap_jso(urls)));
    }
    [_webkitRequestFileSystem](type, size, successCallback, errorCallback) {
      if (successCallback === void 0)
        successCallback = null;
      if (errorCallback === void 0)
        errorCallback = null;
      if (errorCallback != null) {
        this[_webkitRequestFileSystem_1](type, size, successCallback, errorCallback);
        return;
      }
      if (successCallback != null) {
        this[_webkitRequestFileSystem_2](type, size, successCallback);
        return;
      }
      this[_webkitRequestFileSystem_3](type, size);
      return;
    }
    [_webkitRequestFileSystem_1](type, size, successCallback, errorCallback) {
      return wrap_jso(this.raw.webkitRequestFileSystem(unwrap_jso(type), unwrap_jso(size), unwrap_jso(successCallback), unwrap_jso(errorCallback)));
    }
    [_webkitRequestFileSystem_2](type, size, successCallback) {
      return wrap_jso(this.raw.webkitRequestFileSystem(unwrap_jso(type), unwrap_jso(size), unwrap_jso(successCallback)));
    }
    [_webkitRequestFileSystem_3](type, size) {
      return wrap_jso(this.raw.webkitRequestFileSystem(unwrap_jso(type), unwrap_jso(size)));
    }
    webkitRequestFileSystem(type, size) {
      let completer = async.Completer$(dart.dynamic).new();
      this[_webkitRequestFileSystem](type, size, dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    requestFileSystemSync(type, size) {
      return this[_requestFileSystemSync_1](type, size);
    }
    [_requestFileSystemSync_1](type, size) {
      return dart.as(wrap_jso(this.raw.webkitRequestFileSystemSync(unwrap_jso(type), unwrap_jso(size))), _DOMFileSystemSync);
    }
    resolveLocalFileSystemSyncUrl(url) {
      return this[_resolveLocalFileSystemSyncUrl_1](url);
    }
    [_resolveLocalFileSystemSyncUrl_1](url) {
      return dart.as(wrap_jso(this.raw.webkitResolveLocalFileSystemSyncURL(unwrap_jso(url))), _EntrySync);
    }
    [_webkitResolveLocalFileSystemUrl](url, successCallback, errorCallback) {
      if (errorCallback === void 0)
        errorCallback = null;
      if (errorCallback != null) {
        this[_webkitResolveLocalFileSystemUrl_1](url, successCallback, errorCallback);
        return;
      }
      this[_webkitResolveLocalFileSystemUrl_2](url, successCallback);
      return;
    }
    [_webkitResolveLocalFileSystemUrl_1](url, successCallback, errorCallback) {
      return wrap_jso(this.raw.webkitResolveLocalFileSystemURL(unwrap_jso(url), unwrap_jso(successCallback), unwrap_jso(errorCallback)));
    }
    [_webkitResolveLocalFileSystemUrl_2](url, successCallback) {
      return wrap_jso(this.raw.webkitResolveLocalFileSystemURL(unwrap_jso(url), unwrap_jso(successCallback)));
    }
    webkitResolveLocalFileSystemUrl(url) {
      let completer = async.Completer$(dart.dynamic).new();
      this[_webkitResolveLocalFileSystemUrl](url, dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    atob(string) {
      return this[_atob_1](string);
    }
    [_atob_1](string) {
      return dart.as(wrap_jso(this.raw.atob(unwrap_jso(string))), core.String);
    }
    btoa(string) {
      return this[_btoa_1](string);
    }
    [_btoa_1](string) {
      return dart.as(wrap_jso(this.raw.btoa(unwrap_jso(string))), core.String);
    }
    [_clearInterval](handle) {
      this[_clearInterval_1](handle);
      return;
    }
    [_clearInterval_1](handle) {
      return wrap_jso(this.raw.clearInterval(unwrap_jso(handle)));
    }
    [_clearTimeout](handle) {
      this[_clearTimeout_1](handle);
      return;
    }
    [_clearTimeout_1](handle) {
      return wrap_jso(this.raw.clearTimeout(unwrap_jso(handle)));
    }
    [_setInterval](handler, timeout) {
      return this[_setInterval_1](handler, timeout);
    }
    [_setInterval_1](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setInterval(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
    [_setTimeout](handler, timeout) {
      return this[_setTimeout_1](handler, timeout);
    }
    [_setTimeout_1](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setTimeout(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
    get onError() {
      return WorkerGlobalScope.errorEvent.forTarget(this);
    }
  }
  WorkerGlobalScope[dart.implements] = () => [_WindowTimers, WindowBase64];
  dart.defineNamedConstructor(WorkerGlobalScope, 'internal_');
  dart.setSignature(WorkerGlobalScope, {
    constructors: () => ({
      _: [WorkerGlobalScope, []],
      _internalWrap: [WorkerGlobalScope, []],
      internal_: [WorkerGlobalScope, []]
    }),
    methods: () => ({
      close: [dart.void, []],
      [_close_1]: [dart.void, []],
      importScripts: [dart.void, [core.String]],
      [_importScripts_1]: [dart.void, [dart.dynamic]],
      [_webkitRequestFileSystem]: [dart.void, [core.int, core.int], [dart.dynamic, dart.dynamic]],
      [_webkitRequestFileSystem_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      [_webkitRequestFileSystem_2]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_webkitRequestFileSystem_3]: [dart.void, [dart.dynamic, dart.dynamic]],
      webkitRequestFileSystem: [async.Future$(dart.dynamic), [core.int, core.int]],
      requestFileSystemSync: [_DOMFileSystemSync, [core.int, core.int]],
      [_requestFileSystemSync_1]: [_DOMFileSystemSync, [dart.dynamic, dart.dynamic]],
      resolveLocalFileSystemSyncUrl: [_EntrySync, [core.String]],
      [_resolveLocalFileSystemSyncUrl_1]: [_EntrySync, [dart.dynamic]],
      [_webkitResolveLocalFileSystemUrl]: [dart.void, [core.String, dart.dynamic], [dart.dynamic]],
      [_webkitResolveLocalFileSystemUrl_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_webkitResolveLocalFileSystemUrl_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      webkitResolveLocalFileSystemUrl: [async.Future$(dart.dynamic), [core.String]],
      atob: [core.String, [core.String]],
      [_atob_1]: [core.String, [dart.dynamic]],
      btoa: [core.String, [core.String]],
      [_btoa_1]: [core.String, [dart.dynamic]],
      [_clearInterval]: [dart.void, [core.int]],
      [_clearInterval_1]: [dart.void, [dart.dynamic]],
      [_clearTimeout]: [dart.void, [core.int]],
      [_clearTimeout_1]: [dart.void, [dart.dynamic]],
      [_setInterval]: [core.int, [core.Object, core.int]],
      [_setInterval_1]: [core.int, [dart.dynamic, dart.dynamic]],
      [_setTimeout]: [core.int, [core.Object, core.int]],
      [_setTimeout_1]: [core.int, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateWorkerGlobalScope: [WorkerGlobalScope, []]}),
    names: ['internalCreateWorkerGlobalScope']
  });
  WorkerGlobalScope[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WorkerGlobalScope')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("WorkerGlobalScope"))];
  const _fetch_1 = Symbol('_fetch_1');
  const _fetch_2 = Symbol('_fetch_2');
  const _fetch_3 = Symbol('_fetch_3');
  const _fetch_4 = Symbol('_fetch_4');
  const _fetch = Symbol('_fetch');
  class ServiceWorkerGlobalScope extends WorkerGlobalScope {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateServiceWorkerGlobalScope() {
      return ServiceWorkerGlobalScope._internalWrap();
    }
    static _internalWrap() {
      return new ServiceWorkerGlobalScope.internal_();
    }
    internal_() {
      super.internal_();
    }
    get caches() {
      return wrap_jso(this.raw.caches);
    }
    get clients() {
      return dart.as(wrap_jso(this.raw.clients), ServiceWorkerClients);
    }
    get scope() {
      return dart.as(wrap_jso(this.raw.scope), core.String);
    }
    [_fetch](request, requestInitDict) {
      if (requestInitDict === void 0)
        requestInitDict = null;
      if ((typeof request == 'string' || request == null) && requestInitDict == null) {
        return this[_fetch_1](dart.as(request, core.String));
      }
      if (requestInitDict != null && (typeof request == 'string' || request == null)) {
        let requestInitDict_1 = html_common.convertDartToNative_Dictionary(requestInitDict);
        return this[_fetch_2](dart.as(request, core.String), requestInitDict_1);
      }
      if ((dart.is(request, dart.dynamic) || request == null) && requestInitDict == null) {
        return this[_fetch_3](request);
      }
      if (requestInitDict != null && (dart.is(request, dart.dynamic) || request == null)) {
        let requestInitDict_1 = html_common.convertDartToNative_Dictionary(requestInitDict);
        return this[_fetch_4](request, requestInitDict_1);
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_fetch_1](request) {
      return dart.as(wrap_jso(this.raw.fetch(unwrap_jso(request))), async.Future);
    }
    [_fetch_2](request, requestInitDict) {
      return dart.as(wrap_jso(this.raw.fetch(unwrap_jso(request), unwrap_jso(requestInitDict))), async.Future);
    }
    [_fetch_3](request) {
      return dart.as(wrap_jso(this.raw.fetch(unwrap_jso(request))), async.Future);
    }
    [_fetch_4](request, requestInitDict) {
      return dart.as(wrap_jso(this.raw.fetch(unwrap_jso(request), unwrap_jso(requestInitDict))), async.Future);
    }
    get onMessage() {
      return ServiceWorkerGlobalScope.messageEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(ServiceWorkerGlobalScope, 'internal_');
  dart.setSignature(ServiceWorkerGlobalScope, {
    constructors: () => ({
      _: [ServiceWorkerGlobalScope, []],
      _internalWrap: [ServiceWorkerGlobalScope, []],
      internal_: [ServiceWorkerGlobalScope, []]
    }),
    methods: () => ({
      [_fetch]: [async.Future, [dart.dynamic], [core.Map]],
      [_fetch_1]: [async.Future, [core.String]],
      [_fetch_2]: [async.Future, [core.String, dart.dynamic]],
      [_fetch_3]: [async.Future, [dart.dynamic]],
      [_fetch_4]: [async.Future, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateServiceWorkerGlobalScope: [ServiceWorkerGlobalScope, []]}),
    names: ['internalCreateServiceWorkerGlobalScope']
  });
  ServiceWorkerGlobalScope[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorkerGlobalScope')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorkerGlobalScope"))];
  ServiceWorkerGlobalScope.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  const _unregister_1 = Symbol('_unregister_1');
  class ServiceWorkerRegistration extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateServiceWorkerRegistration() {
      return ServiceWorkerRegistration._internalWrap();
    }
    static _internalWrap() {
      return new ServiceWorkerRegistration.internal_();
    }
    internal_() {
      super.internal_();
    }
    get active() {
      return dart.as(wrap_jso(this.raw.active), _ServiceWorker);
    }
    get installing() {
      return dart.as(wrap_jso(this.raw.installing), _ServiceWorker);
    }
    get scope() {
      return dart.as(wrap_jso(this.raw.scope), core.String);
    }
    get waiting() {
      return dart.as(wrap_jso(this.raw.waiting), _ServiceWorker);
    }
    unregister() {
      return this[_unregister_1]();
    }
    [_unregister_1]() {
      return dart.as(wrap_jso(this.raw.unregister()), async.Future);
    }
  }
  dart.defineNamedConstructor(ServiceWorkerRegistration, 'internal_');
  dart.setSignature(ServiceWorkerRegistration, {
    constructors: () => ({
      _: [ServiceWorkerRegistration, []],
      _internalWrap: [ServiceWorkerRegistration, []],
      internal_: [ServiceWorkerRegistration, []]
    }),
    methods: () => ({
      unregister: [async.Future, []],
      [_unregister_1]: [async.Future, []]
    }),
    statics: () => ({internalCreateServiceWorkerRegistration: [ServiceWorkerRegistration, []]}),
    names: ['internalCreateServiceWorkerRegistration']
  });
  ServiceWorkerRegistration[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorkerRegistration')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorkerRegistration"))];
  const _getDistributedNodes_1 = Symbol('_getDistributedNodes_1');
  class ShadowElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("shadow"), ShadowElement);
    }
    static internalCreateShadowElement() {
      return ShadowElement._internalWrap();
    }
    static _internalWrap() {
      return new ShadowElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('shadow');
    }
    getDistributedNodes() {
      return this[_getDistributedNodes_1]();
    }
    [_getDistributedNodes_1]() {
      return dart.as(wrap_jso(this.raw.getDistributedNodes()), NodeList);
    }
  }
  dart.defineNamedConstructor(ShadowElement, 'internal_');
  dart.defineNamedConstructor(ShadowElement, 'created');
  dart.setSignature(ShadowElement, {
    constructors: () => ({
      _: [ShadowElement, []],
      new: [ShadowElement, []],
      _internalWrap: [ShadowElement, []],
      internal_: [ShadowElement, []],
      created: [ShadowElement, []]
    }),
    methods: () => ({
      getDistributedNodes: [NodeList, []],
      [_getDistributedNodes_1]: [NodeList, []]
    }),
    statics: () => ({internalCreateShadowElement: [ShadowElement, []]}),
    names: ['internalCreateShadowElement']
  });
  ShadowElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLShadowElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '26')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("HTMLShadowElement"))];
  const _getSelection_1 = Symbol('_getSelection_1');
  class ShadowRoot extends DocumentFragment {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateShadowRoot() {
      return ShadowRoot._internalWrap();
    }
    static _internalWrap() {
      return new ShadowRoot.internal_();
    }
    internal_() {
      super.internal_();
    }
    get activeElement() {
      return dart.as(wrap_jso(this.raw.activeElement), Element);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), Element);
    }
    get innerHtml() {
      return dart.as(wrap_jso(this.raw.innerHTML), core.String);
    }
    set innerHtml(val) {
      return this.raw.innerHTML = unwrap_jso(val);
    }
    get olderShadowRoot() {
      return dart.as(wrap_jso(this.raw.olderShadowRoot), ShadowRoot);
    }
    get styleSheets() {
      return dart.as(wrap_jso(this.raw.styleSheets), core.List$(StyleSheet));
    }
    clone(deep) {
      return this[_clone_1](deep);
    }
    [_clone_1](deep) {
      return dart.as(wrap_jso(this.raw.cloneNode(unwrap_jso(deep))), Node);
    }
    elementFromPoint(x, y) {
      return this[_elementFromPoint_1](x, y);
    }
    [_elementFromPoint_1](x, y) {
      return dart.as(wrap_jso(this.raw.elementFromPoint(unwrap_jso(x), unwrap_jso(y))), Element);
    }
    getElementsByClassName(className) {
      return this[_getElementsByClassName_1](className);
    }
    [_getElementsByClassName_1](className) {
      return dart.as(wrap_jso(this.raw.getElementsByClassName(unwrap_jso(className))), HtmlCollection);
    }
    getElementsByTagName(tagName) {
      return this[_getElementsByTagName_1](tagName);
    }
    [_getElementsByTagName_1](tagName) {
      return dart.as(wrap_jso(this.raw.getElementsByTagName(unwrap_jso(tagName))), HtmlCollection);
    }
    getSelection() {
      return this[_getSelection_1]();
    }
    [_getSelection_1]() {
      return dart.as(wrap_jso(this.raw.getSelection()), Selection);
    }
    static _shadowRootDeprecationReport() {
      if (!dart.notNull(ShadowRoot._shadowRootDeprecationReported)) {
        exports.window.console.warn('ShadowRoot.resetStyleInheritance and ShadowRoot.applyAuthorStyles now deprecated in dart:html.\nPlease remove them from your code.\n');
        ShadowRoot._shadowRootDeprecationReported = true;
      }
    }
    get resetStyleInheritance() {
      ShadowRoot._shadowRootDeprecationReport();
      return false;
    }
    set resetStyleInheritance(value) {
      ShadowRoot._shadowRootDeprecationReport();
    }
    get applyAuthorStyles() {
      ShadowRoot._shadowRootDeprecationReport();
      return false;
    }
    set applyAuthorStyles(value) {
      ShadowRoot._shadowRootDeprecationReport();
    }
  }
  dart.defineNamedConstructor(ShadowRoot, 'internal_');
  dart.setSignature(ShadowRoot, {
    constructors: () => ({
      _: [ShadowRoot, []],
      _internalWrap: [ShadowRoot, []],
      internal_: [ShadowRoot, []]
    }),
    methods: () => ({
      elementFromPoint: [Element, [core.int, core.int]],
      [_elementFromPoint_1]: [Element, [dart.dynamic, dart.dynamic]],
      getElementsByClassName: [HtmlCollection, [core.String]],
      [_getElementsByClassName_1]: [HtmlCollection, [dart.dynamic]],
      getElementsByTagName: [HtmlCollection, [core.String]],
      [_getElementsByTagName_1]: [HtmlCollection, [dart.dynamic]],
      getSelection: [Selection, []],
      [_getSelection_1]: [Selection, []]
    }),
    statics: () => ({
      internalCreateShadowRoot: [ShadowRoot, []],
      _shadowRootDeprecationReport: [dart.void, []]
    }),
    names: ['internalCreateShadowRoot', '_shadowRootDeprecationReport']
  });
  ShadowRoot[dart.metadata] = () => [dart.const(new _metadata.DomName('ShadowRoot')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '26')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ShadowRoot"))];
  ShadowRoot.supported = true;
  ShadowRoot._shadowRootDeprecationReported = false;
  class SharedWorker extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(scriptURL, name) {
      if (name === void 0)
        name = null;
      if (name != null) {
        return SharedWorker._create_1(scriptURL, name);
      }
      return SharedWorker._create_2(scriptURL);
    }
    static _create_1(scriptURL, name) {
      return dart.as(wrap_jso(new SharedWorker(scriptURL, name)), SharedWorker);
    }
    static _create_2(scriptURL) {
      return dart.as(wrap_jso(new SharedWorker(scriptURL)), SharedWorker);
    }
    static internalCreateSharedWorker() {
      return SharedWorker._internalWrap();
    }
    static _internalWrap() {
      return new SharedWorker.internal_();
    }
    internal_() {
      super.internal_();
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), MessagePort);
    }
    get workerStart() {
      return dart.as(wrap_jso(this.raw.workerStart), core.double);
    }
    get onError() {
      return SharedWorker.errorEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(SharedWorker, 'internal_');
  dart.setSignature(SharedWorker, {
    constructors: () => ({
      _: [SharedWorker, []],
      new: [SharedWorker, [core.String], [core.String]],
      _internalWrap: [SharedWorker, []],
      internal_: [SharedWorker, []]
    }),
    statics: () => ({
      _create_1: [SharedWorker, [dart.dynamic, dart.dynamic]],
      _create_2: [SharedWorker, [dart.dynamic]],
      internalCreateSharedWorker: [SharedWorker, []]
    }),
    names: ['_create_1', '_create_2', 'internalCreateSharedWorker']
  });
  SharedWorker[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SharedWorker')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SharedWorker"))];
  SharedWorker.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  class SharedWorkerGlobalScope extends WorkerGlobalScope {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSharedWorkerGlobalScope() {
      return SharedWorkerGlobalScope._internalWrap();
    }
    static _internalWrap() {
      return new SharedWorkerGlobalScope.internal_();
    }
    internal_() {
      super.internal_();
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    get onConnect() {
      return SharedWorkerGlobalScope.connectEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(SharedWorkerGlobalScope, 'internal_');
  dart.setSignature(SharedWorkerGlobalScope, {
    constructors: () => ({
      _: [SharedWorkerGlobalScope, []],
      _internalWrap: [SharedWorkerGlobalScope, []],
      internal_: [SharedWorkerGlobalScope, []]
    }),
    statics: () => ({internalCreateSharedWorkerGlobalScope: [SharedWorkerGlobalScope, []]}),
    names: ['internalCreateSharedWorkerGlobalScope']
  });
  SharedWorkerGlobalScope[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SharedWorkerGlobalScope')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SharedWorkerGlobalScope"))];
  SharedWorkerGlobalScope.connectEvent = dart.const(new (EventStreamProvider$(Event))('connect'));
  const _appendBuffer_1 = Symbol('_appendBuffer_1');
  const _appendStream_1 = Symbol('_appendStream_1');
  const _appendStream_2 = Symbol('_appendStream_2');
  const _appendTypedData_1 = Symbol('_appendTypedData_1');
  class SourceBuffer extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSourceBuffer() {
      return SourceBuffer._internalWrap();
    }
    static _internalWrap() {
      return new SourceBuffer.internal_();
    }
    internal_() {
      super.internal_();
    }
    get appendWindowEnd() {
      return dart.as(wrap_jso(this.raw.appendWindowEnd), core.num);
    }
    set appendWindowEnd(val) {
      return this.raw.appendWindowEnd = unwrap_jso(val);
    }
    get appendWindowStart() {
      return dart.as(wrap_jso(this.raw.appendWindowStart), core.num);
    }
    set appendWindowStart(val) {
      return this.raw.appendWindowStart = unwrap_jso(val);
    }
    get buffered() {
      return dart.as(wrap_jso(this.raw.buffered), TimeRanges);
    }
    get mode() {
      return dart.as(wrap_jso(this.raw.mode), core.String);
    }
    set mode(val) {
      return this.raw.mode = unwrap_jso(val);
    }
    get timestampOffset() {
      return dart.as(wrap_jso(this.raw.timestampOffset), core.num);
    }
    set timestampOffset(val) {
      return this.raw.timestampOffset = unwrap_jso(val);
    }
    get updating() {
      return dart.as(wrap_jso(this.raw.updating), core.bool);
    }
    abort() {
      this[_abort_1]();
      return;
    }
    [_abort_1]() {
      return wrap_jso(this.raw.abort());
    }
    appendBuffer(data) {
      this[_appendBuffer_1](data);
      return;
    }
    [_appendBuffer_1](data) {
      return wrap_jso(this.raw.appendBuffer(unwrap_jso(data)));
    }
    appendStream(stream, maxSize) {
      if (maxSize === void 0)
        maxSize = null;
      if (maxSize != null) {
        this[_appendStream_1](stream, maxSize);
        return;
      }
      this[_appendStream_2](stream);
      return;
    }
    [_appendStream_1](stream, maxSize) {
      return wrap_jso(this.raw.appendStream(unwrap_jso(stream), unwrap_jso(maxSize)));
    }
    [_appendStream_2](stream) {
      return wrap_jso(this.raw.appendStream(unwrap_jso(stream)));
    }
    appendTypedData(data) {
      this[_appendTypedData_1](data);
      return;
    }
    [_appendTypedData_1](data) {
      return wrap_jso(this.raw.appendBuffer(unwrap_jso(data)));
    }
    remove(start, end) {
      this[_remove_1](start, end);
      return;
    }
    [_remove_1](start, end) {
      return wrap_jso(this.raw.remove(unwrap_jso(start), unwrap_jso(end)));
    }
  }
  dart.defineNamedConstructor(SourceBuffer, 'internal_');
  dart.setSignature(SourceBuffer, {
    constructors: () => ({
      _: [SourceBuffer, []],
      _internalWrap: [SourceBuffer, []],
      internal_: [SourceBuffer, []]
    }),
    methods: () => ({
      abort: [dart.void, []],
      [_abort_1]: [dart.void, []],
      appendBuffer: [dart.void, [typed_data.ByteBuffer]],
      [_appendBuffer_1]: [dart.void, [typed_data.ByteBuffer]],
      appendStream: [dart.void, [dart.dynamic], [core.int]],
      [_appendStream_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_appendStream_2]: [dart.void, [dart.dynamic]],
      appendTypedData: [dart.void, [typed_data.TypedData]],
      [_appendTypedData_1]: [dart.void, [typed_data.TypedData]],
      remove: [dart.void, [core.num, core.num]],
      [_remove_1]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateSourceBuffer: [SourceBuffer, []]}),
    names: ['internalCreateSourceBuffer']
  });
  SourceBuffer[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SourceBuffer')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SourceBuffer"))];
  class SourceBufferList extends dart.mixin(EventTarget, collection.ListMixin$(SourceBuffer), ImmutableListMixin$(SourceBuffer)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSourceBufferList() {
      return SourceBufferList._internalWrap();
    }
    static _internalWrap() {
      return new SourceBufferList.internal_();
    }
    internal_() {
      super.internal_();
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), SourceBuffer);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SourceBuffer);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), SourceBuffer);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SourceBuffer);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), SourceBuffer);
    }
  }
  SourceBufferList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(SourceBuffer)];
  dart.defineNamedConstructor(SourceBufferList, 'internal_');
  dart.setSignature(SourceBufferList, {
    constructors: () => ({
      _: [SourceBufferList, []],
      _internalWrap: [SourceBufferList, []],
      internal_: [SourceBufferList, []]
    }),
    methods: () => ({
      get: [SourceBuffer, [core.int]],
      set: [dart.void, [core.int, SourceBuffer]],
      elementAt: [SourceBuffer, [core.int]],
      item: [SourceBuffer, [core.int]],
      [_item_1]: [SourceBuffer, [dart.dynamic]]
    }),
    statics: () => ({internalCreateSourceBufferList: [SourceBufferList, []]}),
    names: ['internalCreateSourceBufferList']
  });
  dart.defineExtensionMembers(SourceBufferList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  SourceBufferList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SourceBufferList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SourceBufferList"))];
  class SourceElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("source"), SourceElement);
    }
    static internalCreateSourceElement() {
      return SourceElement._internalWrap();
    }
    static _internalWrap() {
      return new SourceElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get media() {
      return dart.as(wrap_jso(this.raw.media), core.String);
    }
    set media(val) {
      return this.raw.media = unwrap_jso(val);
    }
    get sizes() {
      return dart.as(wrap_jso(this.raw.sizes), core.String);
    }
    set sizes(val) {
      return this.raw.sizes = unwrap_jso(val);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get srcset() {
      return dart.as(wrap_jso(this.raw.srcset), core.String);
    }
    set srcset(val) {
      return this.raw.srcset = unwrap_jso(val);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(SourceElement, 'internal_');
  dart.defineNamedConstructor(SourceElement, 'created');
  dart.setSignature(SourceElement, {
    constructors: () => ({
      _: [SourceElement, []],
      new: [SourceElement, []],
      _internalWrap: [SourceElement, []],
      internal_: [SourceElement, []],
      created: [SourceElement, []]
    }),
    statics: () => ({internalCreateSourceElement: [SourceElement, []]}),
    names: ['internalCreateSourceElement']
  });
  SourceElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLSourceElement')), dart.const(new _js_helper.Native("HTMLSourceElement"))];
  class SourceInfo extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSourceInfo() {
      return SourceInfo._internalWrap();
    }
    static _internalWrap() {
      return new SourceInfo.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get facing() {
      return dart.as(wrap_jso(this.raw.facing), core.String);
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    get kind() {
      return dart.as(wrap_jso(this.raw.kind), core.String);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
  }
  dart.defineNamedConstructor(SourceInfo, 'internal_');
  dart.setSignature(SourceInfo, {
    constructors: () => ({
      _: [SourceInfo, []],
      _internalWrap: [SourceInfo, []],
      internal_: [SourceInfo, []]
    }),
    statics: () => ({internalCreateSourceInfo: [SourceInfo, []]}),
    names: ['internalCreateSourceInfo']
  });
  SourceInfo[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SourceInfo')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SourceInfo"))];
  class SpanElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("span"), SpanElement);
    }
    static internalCreateSpanElement() {
      return SpanElement._internalWrap();
    }
    static _internalWrap() {
      return new SpanElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(SpanElement, 'internal_');
  dart.defineNamedConstructor(SpanElement, 'created');
  dart.setSignature(SpanElement, {
    constructors: () => ({
      _: [SpanElement, []],
      new: [SpanElement, []],
      _internalWrap: [SpanElement, []],
      internal_: [SpanElement, []],
      created: [SpanElement, []]
    }),
    statics: () => ({internalCreateSpanElement: [SpanElement, []]}),
    names: ['internalCreateSpanElement']
  });
  SpanElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLSpanElement')), dart.const(new _js_helper.Native("HTMLSpanElement"))];
  class SpeechGrammar extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return SpeechGrammar._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new SpeechGrammar()), SpeechGrammar);
    }
    static internalCreateSpeechGrammar() {
      return SpeechGrammar._internalWrap();
    }
    static _internalWrap() {
      return new SpeechGrammar.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get weight() {
      return dart.as(wrap_jso(this.raw.weight), core.num);
    }
    set weight(val) {
      return this.raw.weight = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(SpeechGrammar, 'internal_');
  dart.setSignature(SpeechGrammar, {
    constructors: () => ({
      _: [SpeechGrammar, []],
      new: [SpeechGrammar, []],
      _internalWrap: [SpeechGrammar, []],
      internal_: [SpeechGrammar, []]
    }),
    statics: () => ({
      _create_1: [SpeechGrammar, []],
      internalCreateSpeechGrammar: [SpeechGrammar, []]
    }),
    names: ['_create_1', 'internalCreateSpeechGrammar']
  });
  SpeechGrammar[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechGrammar')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechGrammar"))];
  const _addFromString_1 = Symbol('_addFromString_1');
  const _addFromString_2 = Symbol('_addFromString_2');
  const _addFromUri_1 = Symbol('_addFromUri_1');
  const _addFromUri_2 = Symbol('_addFromUri_2');
  class SpeechGrammarList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(SpeechGrammar), ImmutableListMixin$(SpeechGrammar)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return SpeechGrammarList._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new SpeechGrammarList()), SpeechGrammarList);
    }
    static internalCreateSpeechGrammarList() {
      return SpeechGrammarList._internalWrap();
    }
    static _internalWrap() {
      return new SpeechGrammarList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), SpeechGrammar);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SpeechGrammar);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), SpeechGrammar);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SpeechGrammar);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    addFromString(string, weight) {
      if (weight === void 0)
        weight = null;
      if (weight != null) {
        this[_addFromString_1](string, weight);
        return;
      }
      this[_addFromString_2](string);
      return;
    }
    [_addFromString_1](string, weight) {
      return wrap_jso(this.raw.addFromString(unwrap_jso(string), unwrap_jso(weight)));
    }
    [_addFromString_2](string) {
      return wrap_jso(this.raw.addFromString(unwrap_jso(string)));
    }
    addFromUri(src, weight) {
      if (weight === void 0)
        weight = null;
      if (weight != null) {
        this[_addFromUri_1](src, weight);
        return;
      }
      this[_addFromUri_2](src);
      return;
    }
    [_addFromUri_1](src, weight) {
      return wrap_jso(this.raw.addFromUri(unwrap_jso(src), unwrap_jso(weight)));
    }
    [_addFromUri_2](src) {
      return wrap_jso(this.raw.addFromUri(unwrap_jso(src)));
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), SpeechGrammar);
    }
  }
  SpeechGrammarList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(SpeechGrammar)];
  dart.defineNamedConstructor(SpeechGrammarList, 'internal_');
  dart.setSignature(SpeechGrammarList, {
    constructors: () => ({
      _: [SpeechGrammarList, []],
      new: [SpeechGrammarList, []],
      _internalWrap: [SpeechGrammarList, []],
      internal_: [SpeechGrammarList, []]
    }),
    methods: () => ({
      get: [SpeechGrammar, [core.int]],
      set: [dart.void, [core.int, SpeechGrammar]],
      elementAt: [SpeechGrammar, [core.int]],
      addFromString: [dart.void, [core.String], [core.num]],
      [_addFromString_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_addFromString_2]: [dart.void, [dart.dynamic]],
      addFromUri: [dart.void, [core.String], [core.num]],
      [_addFromUri_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_addFromUri_2]: [dart.void, [dart.dynamic]],
      item: [SpeechGrammar, [core.int]],
      [_item_1]: [SpeechGrammar, [dart.dynamic]]
    }),
    statics: () => ({
      _create_1: [SpeechGrammarList, []],
      internalCreateSpeechGrammarList: [SpeechGrammarList, []]
    }),
    names: ['_create_1', 'internalCreateSpeechGrammarList']
  });
  dart.defineExtensionMembers(SpeechGrammarList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  SpeechGrammarList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechGrammarList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechGrammarList"))];
  const _stop_1 = Symbol('_stop_1');
  class SpeechRecognition extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechRecognition() {
      return SpeechRecognition._internalWrap();
    }
    static _internalWrap() {
      return new SpeechRecognition.internal_();
    }
    internal_() {
      super.internal_();
    }
    static get supported() {
      return !!(window.SpeechRecognition || window.webkitSpeechRecognition);
    }
    get continuous() {
      return dart.as(wrap_jso(this.raw.continuous), core.bool);
    }
    set continuous(val) {
      return this.raw.continuous = unwrap_jso(val);
    }
    get grammars() {
      return dart.as(wrap_jso(this.raw.grammars), SpeechGrammarList);
    }
    set grammars(val) {
      return this.raw.grammars = unwrap_jso(val);
    }
    get interimResults() {
      return dart.as(wrap_jso(this.raw.interimResults), core.bool);
    }
    set interimResults(val) {
      return this.raw.interimResults = unwrap_jso(val);
    }
    get lang() {
      return dart.as(wrap_jso(this.raw.lang), core.String);
    }
    set lang(val) {
      return this.raw.lang = unwrap_jso(val);
    }
    get maxAlternatives() {
      return dart.as(wrap_jso(this.raw.maxAlternatives), core.int);
    }
    set maxAlternatives(val) {
      return this.raw.maxAlternatives = unwrap_jso(val);
    }
    abort() {
      this[_abort_1]();
      return;
    }
    [_abort_1]() {
      return wrap_jso(this.raw.abort());
    }
    start() {
      this[_start_1]();
      return;
    }
    [_start_1]() {
      return wrap_jso(this.raw.start());
    }
    stop() {
      this[_stop_1]();
      return;
    }
    [_stop_1]() {
      return wrap_jso(this.raw.stop());
    }
    get onAudioEnd() {
      return SpeechRecognition.audioEndEvent.forTarget(this);
    }
    get onAudioStart() {
      return SpeechRecognition.audioStartEvent.forTarget(this);
    }
    get onEnd() {
      return SpeechRecognition.endEvent.forTarget(this);
    }
    get onError() {
      return SpeechRecognition.errorEvent.forTarget(this);
    }
    get onNoMatch() {
      return SpeechRecognition.noMatchEvent.forTarget(this);
    }
    get onResult() {
      return SpeechRecognition.resultEvent.forTarget(this);
    }
    get onSoundEnd() {
      return SpeechRecognition.soundEndEvent.forTarget(this);
    }
    get onSoundStart() {
      return SpeechRecognition.soundStartEvent.forTarget(this);
    }
    get onSpeechEnd() {
      return SpeechRecognition.speechEndEvent.forTarget(this);
    }
    get onSpeechStart() {
      return SpeechRecognition.speechStartEvent.forTarget(this);
    }
    get onStart() {
      return SpeechRecognition.startEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(SpeechRecognition, 'internal_');
  dart.setSignature(SpeechRecognition, {
    constructors: () => ({
      _: [SpeechRecognition, []],
      _internalWrap: [SpeechRecognition, []],
      internal_: [SpeechRecognition, []]
    }),
    methods: () => ({
      abort: [dart.void, []],
      [_abort_1]: [dart.void, []],
      start: [dart.void, []],
      [_start_1]: [dart.void, []],
      stop: [dart.void, []],
      [_stop_1]: [dart.void, []]
    }),
    statics: () => ({internalCreateSpeechRecognition: [SpeechRecognition, []]}),
    names: ['internalCreateSpeechRecognition']
  });
  SpeechRecognition[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognition')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognition"))];
  SpeechRecognition.audioEndEvent = dart.const(new (EventStreamProvider$(Event))('audioend'));
  SpeechRecognition.audioStartEvent = dart.const(new (EventStreamProvider$(Event))('audiostart'));
  SpeechRecognition.endEvent = dart.const(new (EventStreamProvider$(Event))('end'));
  class SpeechRecognitionError extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechRecognitionError() {
      return SpeechRecognitionError._internalWrap();
    }
    static _internalWrap() {
      return new SpeechRecognitionError.internal_();
    }
    internal_() {
      super.internal_();
    }
    get error() {
      return dart.as(wrap_jso(this.raw.error), core.String);
    }
    get message() {
      return dart.as(wrap_jso(this.raw.message), core.String);
    }
  }
  dart.defineNamedConstructor(SpeechRecognitionError, 'internal_');
  dart.setSignature(SpeechRecognitionError, {
    constructors: () => ({
      _: [SpeechRecognitionError, []],
      _internalWrap: [SpeechRecognitionError, []],
      internal_: [SpeechRecognitionError, []]
    }),
    statics: () => ({internalCreateSpeechRecognitionError: [SpeechRecognitionError, []]}),
    names: ['internalCreateSpeechRecognitionError']
  });
  SpeechRecognitionError[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognitionError')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognitionError"))];
  SpeechRecognition.errorEvent = dart.const(new (EventStreamProvider$(SpeechRecognitionError))('error'));
  class SpeechRecognitionEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechRecognitionEvent() {
      return SpeechRecognitionEvent._internalWrap();
    }
    static _internalWrap() {
      return new SpeechRecognitionEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get emma() {
      return dart.as(wrap_jso(this.raw.emma), Document);
    }
    get interpretation() {
      return dart.as(wrap_jso(this.raw.interpretation), Document);
    }
    get resultIndex() {
      return dart.as(wrap_jso(this.raw.resultIndex), core.int);
    }
    get results() {
      return dart.as(wrap_jso(this.raw.results), core.List$(SpeechRecognitionResult));
    }
  }
  dart.defineNamedConstructor(SpeechRecognitionEvent, 'internal_');
  dart.setSignature(SpeechRecognitionEvent, {
    constructors: () => ({
      _: [SpeechRecognitionEvent, []],
      _internalWrap: [SpeechRecognitionEvent, []],
      internal_: [SpeechRecognitionEvent, []]
    }),
    statics: () => ({internalCreateSpeechRecognitionEvent: [SpeechRecognitionEvent, []]}),
    names: ['internalCreateSpeechRecognitionEvent']
  });
  SpeechRecognitionEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognitionEvent')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognitionEvent"))];
  SpeechRecognition.noMatchEvent = dart.const(new (EventStreamProvider$(SpeechRecognitionEvent))('nomatch'));
  SpeechRecognition.resultEvent = dart.const(new (EventStreamProvider$(SpeechRecognitionEvent))('result'));
  SpeechRecognition.soundEndEvent = dart.const(new (EventStreamProvider$(Event))('soundend'));
  SpeechRecognition.soundStartEvent = dart.const(new (EventStreamProvider$(Event))('soundstart'));
  SpeechRecognition.speechEndEvent = dart.const(new (EventStreamProvider$(Event))('speechend'));
  SpeechRecognition.speechStartEvent = dart.const(new (EventStreamProvider$(Event))('speechstart'));
  SpeechRecognition.startEvent = dart.const(new (EventStreamProvider$(Event))('start'));
  class SpeechRecognitionAlternative extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechRecognitionAlternative() {
      return SpeechRecognitionAlternative._internalWrap();
    }
    static _internalWrap() {
      return new SpeechRecognitionAlternative.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get confidence() {
      return dart.as(wrap_jso(this.raw.confidence), core.double);
    }
    get transcript() {
      return dart.as(wrap_jso(this.raw.transcript), core.String);
    }
  }
  dart.defineNamedConstructor(SpeechRecognitionAlternative, 'internal_');
  dart.setSignature(SpeechRecognitionAlternative, {
    constructors: () => ({
      _: [SpeechRecognitionAlternative, []],
      _internalWrap: [SpeechRecognitionAlternative, []],
      internal_: [SpeechRecognitionAlternative, []]
    }),
    statics: () => ({internalCreateSpeechRecognitionAlternative: [SpeechRecognitionAlternative, []]}),
    names: ['internalCreateSpeechRecognitionAlternative']
  });
  SpeechRecognitionAlternative[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognitionAlternative')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognitionAlternative"))];
  class SpeechRecognitionResult extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechRecognitionResult() {
      return SpeechRecognitionResult._internalWrap();
    }
    static _internalWrap() {
      return new SpeechRecognitionResult.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get isFinal() {
      return dart.as(wrap_jso(this.raw.isFinal), core.bool);
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), SpeechRecognitionAlternative);
    }
  }
  dart.defineNamedConstructor(SpeechRecognitionResult, 'internal_');
  dart.setSignature(SpeechRecognitionResult, {
    constructors: () => ({
      _: [SpeechRecognitionResult, []],
      _internalWrap: [SpeechRecognitionResult, []],
      internal_: [SpeechRecognitionResult, []]
    }),
    methods: () => ({
      item: [SpeechRecognitionAlternative, [core.int]],
      [_item_1]: [SpeechRecognitionAlternative, [dart.dynamic]]
    }),
    statics: () => ({internalCreateSpeechRecognitionResult: [SpeechRecognitionResult, []]}),
    names: ['internalCreateSpeechRecognitionResult']
  });
  SpeechRecognitionResult[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognitionResult')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME, '25')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognitionResult"))];
  const _getVoices_1 = Symbol('_getVoices_1');
  const _pause_1 = Symbol('_pause_1');
  const _resume_1 = Symbol('_resume_1');
  const _speak_1 = Symbol('_speak_1');
  class SpeechSynthesis extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechSynthesis() {
      return SpeechSynthesis._internalWrap();
    }
    static _internalWrap() {
      return new SpeechSynthesis.internal_();
    }
    internal_() {
      super.internal_();
    }
    get paused() {
      return dart.as(wrap_jso(this.raw.paused), core.bool);
    }
    get pending() {
      return dart.as(wrap_jso(this.raw.pending), core.bool);
    }
    get speaking() {
      return dart.as(wrap_jso(this.raw.speaking), core.bool);
    }
    cancel() {
      this[_cancel_1]();
      return;
    }
    [_cancel_1]() {
      return wrap_jso(this.raw.cancel());
    }
    getVoices() {
      return this[_getVoices_1]();
    }
    [_getVoices_1]() {
      return dart.as(wrap_jso(this.raw.getVoices()), core.List$(SpeechSynthesisVoice));
    }
    pause() {
      this[_pause_1]();
      return;
    }
    [_pause_1]() {
      return wrap_jso(this.raw.pause());
    }
    resume() {
      this[_resume_1]();
      return;
    }
    [_resume_1]() {
      return wrap_jso(this.raw.resume());
    }
    speak(utterance) {
      this[_speak_1](utterance);
      return;
    }
    [_speak_1](utterance) {
      return wrap_jso(this.raw.speak(unwrap_jso(utterance)));
    }
  }
  dart.defineNamedConstructor(SpeechSynthesis, 'internal_');
  dart.setSignature(SpeechSynthesis, {
    constructors: () => ({
      _: [SpeechSynthesis, []],
      _internalWrap: [SpeechSynthesis, []],
      internal_: [SpeechSynthesis, []]
    }),
    methods: () => ({
      cancel: [dart.void, []],
      [_cancel_1]: [dart.void, []],
      getVoices: [core.List$(SpeechSynthesisVoice), []],
      [_getVoices_1]: [core.List$(SpeechSynthesisVoice), []],
      pause: [dart.void, []],
      [_pause_1]: [dart.void, []],
      resume: [dart.void, []],
      [_resume_1]: [dart.void, []],
      speak: [dart.void, [SpeechSynthesisUtterance]],
      [_speak_1]: [dart.void, [SpeechSynthesisUtterance]]
    }),
    statics: () => ({internalCreateSpeechSynthesis: [SpeechSynthesis, []]}),
    names: ['internalCreateSpeechSynthesis']
  });
  SpeechSynthesis[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechSynthesis')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechSynthesis"))];
  class SpeechSynthesisEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechSynthesisEvent() {
      return SpeechSynthesisEvent._internalWrap();
    }
    static _internalWrap() {
      return new SpeechSynthesisEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get charIndex() {
      return dart.as(wrap_jso(this.raw.charIndex), core.int);
    }
    get elapsedTime() {
      return dart.as(wrap_jso(this.raw.elapsedTime), core.double);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
  }
  dart.defineNamedConstructor(SpeechSynthesisEvent, 'internal_');
  dart.setSignature(SpeechSynthesisEvent, {
    constructors: () => ({
      _: [SpeechSynthesisEvent, []],
      _internalWrap: [SpeechSynthesisEvent, []],
      internal_: [SpeechSynthesisEvent, []]
    }),
    statics: () => ({internalCreateSpeechSynthesisEvent: [SpeechSynthesisEvent, []]}),
    names: ['internalCreateSpeechSynthesisEvent']
  });
  SpeechSynthesisEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechSynthesisEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechSynthesisEvent"))];
  class SpeechSynthesisUtterance extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(text) {
      if (text === void 0)
        text = null;
      if (text != null) {
        return SpeechSynthesisUtterance._create_1(text);
      }
      return SpeechSynthesisUtterance._create_2();
    }
    static _create_1(text) {
      return dart.as(wrap_jso(new SpeechSynthesisUtterance(text)), SpeechSynthesisUtterance);
    }
    static _create_2() {
      return dart.as(wrap_jso(new SpeechSynthesisUtterance()), SpeechSynthesisUtterance);
    }
    static internalCreateSpeechSynthesisUtterance() {
      return SpeechSynthesisUtterance._internalWrap();
    }
    static _internalWrap() {
      return new SpeechSynthesisUtterance.internal_();
    }
    internal_() {
      super.internal_();
    }
    get lang() {
      return dart.as(wrap_jso(this.raw.lang), core.String);
    }
    set lang(val) {
      return this.raw.lang = unwrap_jso(val);
    }
    get pitch() {
      return dart.as(wrap_jso(this.raw.pitch), core.num);
    }
    set pitch(val) {
      return this.raw.pitch = unwrap_jso(val);
    }
    get rate() {
      return dart.as(wrap_jso(this.raw.rate), core.num);
    }
    set rate(val) {
      return this.raw.rate = unwrap_jso(val);
    }
    get text() {
      return dart.as(wrap_jso(this.raw.text), core.String);
    }
    set text(val) {
      return this.raw.text = unwrap_jso(val);
    }
    get voice() {
      return dart.as(wrap_jso(this.raw.voice), SpeechSynthesisVoice);
    }
    set voice(val) {
      return this.raw.voice = unwrap_jso(val);
    }
    get volume() {
      return dart.as(wrap_jso(this.raw.volume), core.num);
    }
    set volume(val) {
      return this.raw.volume = unwrap_jso(val);
    }
    get onBoundary() {
      return SpeechSynthesisUtterance.boundaryEvent.forTarget(this);
    }
    get onEnd() {
      return SpeechSynthesisUtterance.endEvent.forTarget(this);
    }
    get onError() {
      return SpeechSynthesisUtterance.errorEvent.forTarget(this);
    }
    get onMark() {
      return SpeechSynthesisUtterance.markEvent.forTarget(this);
    }
    get onPause() {
      return SpeechSynthesisUtterance.pauseEvent.forTarget(this);
    }
    get onResume() {
      return SpeechSynthesisUtterance.resumeEvent.forTarget(this);
    }
    get onStart() {
      return SpeechSynthesisUtterance.startEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(SpeechSynthesisUtterance, 'internal_');
  dart.setSignature(SpeechSynthesisUtterance, {
    constructors: () => ({
      _: [SpeechSynthesisUtterance, []],
      new: [SpeechSynthesisUtterance, [], [core.String]],
      _internalWrap: [SpeechSynthesisUtterance, []],
      internal_: [SpeechSynthesisUtterance, []]
    }),
    statics: () => ({
      _create_1: [SpeechSynthesisUtterance, [dart.dynamic]],
      _create_2: [SpeechSynthesisUtterance, []],
      internalCreateSpeechSynthesisUtterance: [SpeechSynthesisUtterance, []]
    }),
    names: ['_create_1', '_create_2', 'internalCreateSpeechSynthesisUtterance']
  });
  SpeechSynthesisUtterance[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechSynthesisUtterance')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechSynthesisUtterance"))];
  SpeechSynthesisUtterance.boundaryEvent = dart.const(new (EventStreamProvider$(SpeechSynthesisEvent))('boundary'));
  SpeechSynthesisUtterance.endEvent = dart.const(new (EventStreamProvider$(SpeechSynthesisEvent))('end'));
  SpeechSynthesisUtterance.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  SpeechSynthesisUtterance.markEvent = dart.const(new (EventStreamProvider$(SpeechSynthesisEvent))('mark'));
  SpeechSynthesisUtterance.pauseEvent = dart.const(new (EventStreamProvider$(Event))('pause'));
  SpeechSynthesisUtterance.resumeEvent = dart.const(new (EventStreamProvider$(SpeechSynthesisEvent))('resume'));
  SpeechSynthesisUtterance.startEvent = dart.const(new (EventStreamProvider$(SpeechSynthesisEvent))('start'));
  class SpeechSynthesisVoice extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateSpeechSynthesisVoice() {
      return SpeechSynthesisVoice._internalWrap();
    }
    static _internalWrap() {
      return new SpeechSynthesisVoice.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.default), core.bool);
    }
    get lang() {
      return dart.as(wrap_jso(this.raw.lang), core.String);
    }
    get localService() {
      return dart.as(wrap_jso(this.raw.localService), core.bool);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    get voiceUri() {
      return dart.as(wrap_jso(this.raw.voiceURI), core.String);
    }
  }
  dart.defineNamedConstructor(SpeechSynthesisVoice, 'internal_');
  dart.setSignature(SpeechSynthesisVoice, {
    constructors: () => ({
      _: [SpeechSynthesisVoice, []],
      _internalWrap: [SpeechSynthesisVoice, []],
      internal_: [SpeechSynthesisVoice, []]
    }),
    statics: () => ({internalCreateSpeechSynthesisVoice: [SpeechSynthesisVoice, []]}),
    names: ['internalCreateSpeechSynthesisVoice']
  });
  SpeechSynthesisVoice[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechSynthesisVoice')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechSynthesisVoice"))];
  const _getItem = Symbol('_getItem');
  const _setItem = Symbol('_setItem');
  const _removeItem = Symbol('_removeItem');
  const _clear = Symbol('_clear');
  const _key = Symbol('_key');
  const _length = Symbol('_length');
  const __delete___1 = Symbol('__delete___1');
  const __delete___2 = Symbol('__delete___2');
  const __delete__ = Symbol('__delete__');
  const __getter___2 = Symbol('__getter___2');
  const _clear_1 = Symbol('_clear_1');
  const _getItem_1 = Symbol('_getItem_1');
  const _key_1 = Symbol('_key_1');
  const _removeItem_1 = Symbol('_removeItem_1');
  const _setItem_1 = Symbol('_setItem_1');
  class Storage extends DartHtmlDomObject {
    addAll(other) {
      other.forEach(dart.fn(((k, v) => {
        this.set(dart.as(k, core.String), dart.as(v, core.String));
      }).bind(this)));
    }
    containsValue(value) {
      return this.values[dartx.any](dart.fn(e => dart.equals(e, value), core.bool, [dart.dynamic]));
    }
    containsKey(key) {
      return this[_getItem](key) != null;
    }
    get(key) {
      return this[_getItem](key);
    }
    set(key, value) {
      this[_setItem](key, value);
      return value;
    }
    putIfAbsent(key, ifAbsent) {
      if (!dart.notNull(this.containsKey(key)))
        this.set(key, ifAbsent());
      return this.get(key);
    }
    remove(key) {
      let value = this.get(key);
      this[_removeItem](key);
      return value;
    }
    clear() {
      return this[_clear]();
    }
    forEach(f) {
      for (let i = 0; true; i = dart.notNull(i) + 1) {
        let key = this[_key](i);
        if (key == null)
          return;
        f(key, this.get(key));
      }
    }
    get keys() {
      let keys = [];
      this.forEach(dart.fn((k, v) => keys[dartx.add](k), dart.void, [dart.dynamic, dart.dynamic]));
      return dart.as(keys, core.Iterable$(core.String));
    }
    get values() {
      let values = [];
      this.forEach(dart.fn((k, v) => values[dartx.add](v), dart.void, [dart.dynamic, dart.dynamic]));
      return dart.as(values, core.Iterable$(core.String));
    }
    get length() {
      return this[_length];
    }
    get isEmpty() {
      return this[_key](0) == null;
    }
    get isNotEmpty() {
      return !dart.notNull(this.isEmpty);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStorage() {
      return Storage._internalWrap();
    }
    static _internalWrap() {
      return new Storage.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get [_length]() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    [__delete__](index_OR_name) {
      if (typeof index_OR_name == 'number' || index_OR_name == null) {
        return this[__delete___1](dart.as(index_OR_name, core.int));
      }
      if (typeof index_OR_name == 'string' || index_OR_name == null) {
        return this[__delete___2](dart.as(index_OR_name, core.String));
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [__delete___1](index) {
      return dart.as(wrap_jso(this.raw.__delete__(unwrap_jso(index))), core.bool);
    }
    [__delete___2](name) {
      return dart.as(wrap_jso(this.raw.__delete__(unwrap_jso(name))), core.bool);
    }
    [__getter__](index_OR_name) {
      if (typeof index_OR_name == 'number' || index_OR_name == null) {
        return this[__getter___1](dart.as(index_OR_name, core.int));
      }
      if (typeof index_OR_name == 'string' || index_OR_name == null) {
        return this[__getter___2](dart.as(index_OR_name, core.String));
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [__getter___1](index) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(index))), core.String);
    }
    [__getter___2](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), core.String);
    }
    [__setter__](index_OR_name, value) {
      if (typeof index_OR_name == 'number' || index_OR_name == null) {
        this[__setter___1](dart.as(index_OR_name, core.int), value);
        return;
      }
      if (typeof index_OR_name == 'string' || index_OR_name == null) {
        this[__setter___2](dart.as(index_OR_name, core.String), value);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [__setter___1](index, value) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(index), unwrap_jso(value)));
    }
    [__setter___2](name, value) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(name), unwrap_jso(value)));
    }
    [_clear]() {
      this[_clear_1]();
      return;
    }
    [_clear_1]() {
      return wrap_jso(this.raw.clear());
    }
    [_getItem](key) {
      return this[_getItem_1](key);
    }
    [_getItem_1](key) {
      return dart.as(wrap_jso(this.raw.getItem(unwrap_jso(key))), core.String);
    }
    [_key](index) {
      return this[_key_1](index);
    }
    [_key_1](index) {
      return dart.as(wrap_jso(this.raw.key(unwrap_jso(index))), core.String);
    }
    [_removeItem](key) {
      this[_removeItem_1](key);
      return;
    }
    [_removeItem_1](key) {
      return wrap_jso(this.raw.removeItem(unwrap_jso(key)));
    }
    [_setItem](key, data) {
      this[_setItem_1](key, data);
      return;
    }
    [_setItem_1](key, data) {
      return wrap_jso(this.raw.setItem(unwrap_jso(key), unwrap_jso(data)));
    }
  }
  Storage[dart.implements] = () => [core.Map$(core.String, core.String)];
  dart.defineNamedConstructor(Storage, 'internal_');
  dart.setSignature(Storage, {
    constructors: () => ({
      _: [Storage, []],
      _internalWrap: [Storage, []],
      internal_: [Storage, []]
    }),
    methods: () => ({
      addAll: [dart.void, [core.Map$(core.String, core.String)]],
      containsValue: [core.bool, [core.String]],
      containsKey: [core.bool, [core.String]],
      get: [core.String, [core.String]],
      set: [dart.void, [core.String, core.String]],
      putIfAbsent: [core.String, [core.String, dart.functionType(core.String, [])]],
      remove: [core.String, [core.String]],
      clear: [dart.void, []],
      forEach: [dart.void, [dart.functionType(dart.void, [core.String, core.String])]],
      [__delete__]: [core.bool, [dart.dynamic]],
      [__delete___1]: [core.bool, [core.int]],
      [__delete___2]: [core.bool, [core.String]],
      [__getter__]: [core.String, [dart.dynamic]],
      [__getter___1]: [core.String, [core.int]],
      [__getter___2]: [core.String, [core.String]],
      [__setter__]: [dart.void, [dart.dynamic, core.String]],
      [__setter___1]: [dart.void, [core.int, dart.dynamic]],
      [__setter___2]: [dart.void, [core.String, dart.dynamic]],
      [_clear]: [dart.void, []],
      [_clear_1]: [dart.void, []],
      [_getItem]: [core.String, [core.String]],
      [_getItem_1]: [core.String, [dart.dynamic]],
      [_key]: [core.String, [core.int]],
      [_key_1]: [core.String, [dart.dynamic]],
      [_removeItem]: [dart.void, [core.String]],
      [_removeItem_1]: [dart.void, [dart.dynamic]],
      [_setItem]: [dart.void, [core.String, core.String]],
      [_setItem_1]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateStorage: [Storage, []]}),
    names: ['internalCreateStorage']
  });
  Storage[dart.metadata] = () => [dart.const(new _metadata.DomName('Storage')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("Storage"))];
  const StorageErrorCallback = dart.typedef('StorageErrorCallback', () => dart.functionType(dart.void, [dart.dynamic]));
  class StorageInfo extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStorageInfo() {
      return StorageInfo._internalWrap();
    }
    static _internalWrap() {
      return new StorageInfo.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get quota() {
      return dart.as(wrap_jso(this.raw.quota), core.int);
    }
    get usage() {
      return dart.as(wrap_jso(this.raw.usage), core.int);
    }
  }
  dart.defineNamedConstructor(StorageInfo, 'internal_');
  dart.setSignature(StorageInfo, {
    constructors: () => ({
      _: [StorageInfo, []],
      _internalWrap: [StorageInfo, []],
      internal_: [StorageInfo, []]
    }),
    statics: () => ({internalCreateStorageInfo: [StorageInfo, []]}),
    names: ['internalCreateStorageInfo']
  });
  StorageInfo[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('StorageInfo')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("StorageInfo"))];
  const _queryInfo_1 = Symbol('_queryInfo_1');
  const _requestPersistentQuota_1 = Symbol('_requestPersistentQuota_1');
  class StorageQuota extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStorageQuota() {
      return StorageQuota._internalWrap();
    }
    static _internalWrap() {
      return new StorageQuota.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get supportedTypes() {
      return dart.as(wrap_jso(this.raw.supportedTypes), core.List$(core.String));
    }
    queryInfo(type) {
      return this[_queryInfo_1](type);
    }
    [_queryInfo_1](type) {
      return dart.as(wrap_jso(this.raw.queryInfo(unwrap_jso(type))), async.Future);
    }
    requestPersistentQuota(newQuota) {
      return this[_requestPersistentQuota_1](newQuota);
    }
    [_requestPersistentQuota_1](newQuota) {
      return dart.as(wrap_jso(this.raw.requestPersistentQuota(unwrap_jso(newQuota))), async.Future);
    }
  }
  dart.defineNamedConstructor(StorageQuota, 'internal_');
  dart.setSignature(StorageQuota, {
    constructors: () => ({
      _: [StorageQuota, []],
      _internalWrap: [StorageQuota, []],
      internal_: [StorageQuota, []]
    }),
    methods: () => ({
      queryInfo: [async.Future, [core.String]],
      [_queryInfo_1]: [async.Future, [dart.dynamic]],
      requestPersistentQuota: [async.Future, [core.int]],
      [_requestPersistentQuota_1]: [async.Future, [dart.dynamic]]
    }),
    statics: () => ({internalCreateStorageQuota: [StorageQuota, []]}),
    names: ['internalCreateStorageQuota']
  });
  StorageQuota[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('StorageQuota')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("StorageQuota"))];
  const StorageQuotaCallback = dart.typedef('StorageQuotaCallback', () => dart.functionType(dart.void, [core.int]));
  const StorageUsageCallback = dart.typedef('StorageUsageCallback', () => dart.functionType(dart.void, [core.int, core.int]));
  const _StringCallback = dart.typedef('_StringCallback', () => dart.functionType(dart.void, [core.String]));
  class StyleElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("style"), StyleElement);
    }
    static internalCreateStyleElement() {
      return StyleElement._internalWrap();
    }
    static _internalWrap() {
      return new StyleElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get media() {
      return dart.as(wrap_jso(this.raw.media), core.String);
    }
    set media(val) {
      return this.raw.media = unwrap_jso(val);
    }
    get sheet() {
      return dart.as(wrap_jso(this.raw.sheet), StyleSheet);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    set type(val) {
      return this.raw.type = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(StyleElement, 'internal_');
  dart.defineNamedConstructor(StyleElement, 'created');
  dart.setSignature(StyleElement, {
    constructors: () => ({
      _: [StyleElement, []],
      new: [StyleElement, []],
      _internalWrap: [StyleElement, []],
      internal_: [StyleElement, []],
      created: [StyleElement, []]
    }),
    statics: () => ({internalCreateStyleElement: [StyleElement, []]}),
    names: ['internalCreateStyleElement']
  });
  StyleElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLStyleElement')), dart.const(new _js_helper.Native("HTMLStyleElement"))];
  const _matchMedium_1 = Symbol('_matchMedium_1');
  class StyleMedia extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStyleMedia() {
      return StyleMedia._internalWrap();
    }
    static _internalWrap() {
      return new StyleMedia.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    matchMedium(mediaquery) {
      return this[_matchMedium_1](mediaquery);
    }
    [_matchMedium_1](mediaquery) {
      return dart.as(wrap_jso(this.raw.matchMedium(unwrap_jso(mediaquery))), core.bool);
    }
  }
  dart.defineNamedConstructor(StyleMedia, 'internal_');
  dart.setSignature(StyleMedia, {
    constructors: () => ({
      _: [StyleMedia, []],
      _internalWrap: [StyleMedia, []],
      internal_: [StyleMedia, []]
    }),
    methods: () => ({
      matchMedium: [core.bool, [core.String]],
      [_matchMedium_1]: [core.bool, [dart.dynamic]]
    }),
    statics: () => ({internalCreateStyleMedia: [StyleMedia, []]}),
    names: ['internalCreateStyleMedia']
  });
  StyleMedia[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('StyleMedia')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("StyleMedia"))];
  class StyleSheet extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateStyleSheet() {
      return StyleSheet._internalWrap();
    }
    static _internalWrap() {
      return new StyleSheet.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    get media() {
      return wrap_jso(this.raw.media);
    }
    get ownerNode() {
      return dart.as(wrap_jso(this.raw.ownerNode), Node);
    }
    get parentStyleSheet() {
      return dart.as(wrap_jso(this.raw.parentStyleSheet), StyleSheet);
    }
    get title() {
      return dart.as(wrap_jso(this.raw.title), core.String);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
  }
  dart.defineNamedConstructor(StyleSheet, 'internal_');
  dart.setSignature(StyleSheet, {
    constructors: () => ({
      _: [StyleSheet, []],
      _internalWrap: [StyleSheet, []],
      internal_: [StyleSheet, []]
    }),
    statics: () => ({internalCreateStyleSheet: [StyleSheet, []]}),
    names: ['internalCreateStyleSheet']
  });
  StyleSheet[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('StyleSheet')), dart.const(new _js_helper.Native("StyleSheet"))];
  class TableCaptionElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("caption"), TableCaptionElement);
    }
    static internalCreateTableCaptionElement() {
      return TableCaptionElement._internalWrap();
    }
    static _internalWrap() {
      return new TableCaptionElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(TableCaptionElement, 'internal_');
  dart.defineNamedConstructor(TableCaptionElement, 'created');
  dart.setSignature(TableCaptionElement, {
    constructors: () => ({
      _: [TableCaptionElement, []],
      new: [TableCaptionElement, []],
      _internalWrap: [TableCaptionElement, []],
      internal_: [TableCaptionElement, []],
      created: [TableCaptionElement, []]
    }),
    statics: () => ({internalCreateTableCaptionElement: [TableCaptionElement, []]}),
    names: ['internalCreateTableCaptionElement']
  });
  TableCaptionElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableCaptionElement')), dart.const(new _js_helper.Native("HTMLTableCaptionElement"))];
  class TableCellElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("td"), TableCellElement);
    }
    static internalCreateTableCellElement() {
      return TableCellElement._internalWrap();
    }
    static _internalWrap() {
      return new TableCellElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get cellIndex() {
      return dart.as(wrap_jso(this.raw.cellIndex), core.int);
    }
    get colSpan() {
      return dart.as(wrap_jso(this.raw.colSpan), core.int);
    }
    set colSpan(val) {
      return this.raw.colSpan = unwrap_jso(val);
    }
    get headers() {
      return dart.as(wrap_jso(this.raw.headers), core.String);
    }
    set headers(val) {
      return this.raw.headers = unwrap_jso(val);
    }
    get rowSpan() {
      return dart.as(wrap_jso(this.raw.rowSpan), core.int);
    }
    set rowSpan(val) {
      return this.raw.rowSpan = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(TableCellElement, 'internal_');
  dart.defineNamedConstructor(TableCellElement, 'created');
  dart.setSignature(TableCellElement, {
    constructors: () => ({
      _: [TableCellElement, []],
      new: [TableCellElement, []],
      _internalWrap: [TableCellElement, []],
      internal_: [TableCellElement, []],
      created: [TableCellElement, []]
    }),
    statics: () => ({internalCreateTableCellElement: [TableCellElement, []]}),
    names: ['internalCreateTableCellElement']
  });
  TableCellElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableCellElement')), dart.const(new _js_helper.Native("HTMLTableCellElement,HTMLTableDataCellElement,HTMLTableHeaderCellElement"))];
  class TableColElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("col"), TableColElement);
    }
    static internalCreateTableColElement() {
      return TableColElement._internalWrap();
    }
    static _internalWrap() {
      return new TableColElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get span() {
      return dart.as(wrap_jso(this.raw.span), core.int);
    }
    set span(val) {
      return this.raw.span = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(TableColElement, 'internal_');
  dart.defineNamedConstructor(TableColElement, 'created');
  dart.setSignature(TableColElement, {
    constructors: () => ({
      _: [TableColElement, []],
      new: [TableColElement, []],
      _internalWrap: [TableColElement, []],
      internal_: [TableColElement, []],
      created: [TableColElement, []]
    }),
    statics: () => ({internalCreateTableColElement: [TableColElement, []]}),
    names: ['internalCreateTableColElement']
  });
  TableColElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableColElement')), dart.const(new _js_helper.Native("HTMLTableColElement"))];
  const _tBodies = Symbol('_tBodies');
  const _rows = Symbol('_rows');
  const _createCaption = Symbol('_createCaption');
  const _createTFoot = Symbol('_createTFoot');
  const _createTHead = Symbol('_createTHead');
  const _insertRow = Symbol('_insertRow');
  const _createCaption_1 = Symbol('_createCaption_1');
  const _createTFoot_1 = Symbol('_createTFoot_1');
  const _createTHead_1 = Symbol('_createTHead_1');
  const _deleteCaption_1 = Symbol('_deleteCaption_1');
  const _deleteRow_1 = Symbol('_deleteRow_1');
  const _deleteTFoot_1 = Symbol('_deleteTFoot_1');
  const _deleteTHead_1 = Symbol('_deleteTHead_1');
  const _insertRow_1 = Symbol('_insertRow_1');
  const _insertRow_2 = Symbol('_insertRow_2');
  class TableElement extends HtmlElement {
    get tBodies() {
      return new (_WrappedList$(TableSectionElement))(this[_tBodies]);
    }
    get rows() {
      return new (_WrappedList$(TableRowElement))(this[_rows]);
    }
    addRow() {
      return this.insertRow(-1);
    }
    createCaption() {
      return dart.as(this[_createCaption](), TableCaptionElement);
    }
    createTBody() {
      return dart.as(dart.dcall(/* Unimplemented unknown name */_createTBody), TableSectionElement);
    }
    createTFoot() {
      return dart.as(this[_createTFoot](), TableSectionElement);
    }
    createTHead() {
      return dart.as(this[_createTHead](), TableSectionElement);
    }
    insertRow(index) {
      return dart.as(this[_insertRow](index), TableRowElement);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("table"), TableElement);
    }
    static internalCreateTableElement() {
      return TableElement._internalWrap();
    }
    static _internalWrap() {
      return new TableElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get caption() {
      return dart.as(wrap_jso(this.raw.caption), TableCaptionElement);
    }
    set caption(val) {
      return this.raw.caption = unwrap_jso(val);
    }
    get [_rows]() {
      return dart.as(wrap_jso(this.raw.rows), core.List$(Node));
    }
    get [_tBodies]() {
      return dart.as(wrap_jso(this.raw.tBodies), core.List$(Node));
    }
    get tFoot() {
      return dart.as(wrap_jso(this.raw.tFoot), TableSectionElement);
    }
    set tFoot(val) {
      return this.raw.tFoot = unwrap_jso(val);
    }
    get tHead() {
      return dart.as(wrap_jso(this.raw.tHead), TableSectionElement);
    }
    set tHead(val) {
      return this.raw.tHead = unwrap_jso(val);
    }
    [_createCaption]() {
      return this[_createCaption_1]();
    }
    [_createCaption_1]() {
      return dart.as(wrap_jso(this.raw.createCaption()), HtmlElement);
    }
    [_createTFoot]() {
      return this[_createTFoot_1]();
    }
    [_createTFoot_1]() {
      return dart.as(wrap_jso(this.raw.createTFoot()), HtmlElement);
    }
    [_createTHead]() {
      return this[_createTHead_1]();
    }
    [_createTHead_1]() {
      return dart.as(wrap_jso(this.raw.createTHead()), HtmlElement);
    }
    deleteCaption() {
      this[_deleteCaption_1]();
      return;
    }
    [_deleteCaption_1]() {
      return wrap_jso(this.raw.deleteCaption());
    }
    deleteRow(index) {
      this[_deleteRow_1](index);
      return;
    }
    [_deleteRow_1](index) {
      return wrap_jso(this.raw.deleteRow(unwrap_jso(index)));
    }
    deleteTFoot() {
      this[_deleteTFoot_1]();
      return;
    }
    [_deleteTFoot_1]() {
      return wrap_jso(this.raw.deleteTFoot());
    }
    deleteTHead() {
      this[_deleteTHead_1]();
      return;
    }
    [_deleteTHead_1]() {
      return wrap_jso(this.raw.deleteTHead());
    }
    [_insertRow](index) {
      if (index === void 0)
        index = null;
      if (index != null) {
        return this[_insertRow_1](index);
      }
      return this[_insertRow_2]();
    }
    [_insertRow_1](index) {
      return dart.as(wrap_jso(this.raw.insertRow(unwrap_jso(index))), HtmlElement);
    }
    [_insertRow_2]() {
      return dart.as(wrap_jso(this.raw.insertRow()), HtmlElement);
    }
  }
  dart.defineNamedConstructor(TableElement, 'internal_');
  dart.defineNamedConstructor(TableElement, 'created');
  dart.setSignature(TableElement, {
    constructors: () => ({
      _: [TableElement, []],
      new: [TableElement, []],
      _internalWrap: [TableElement, []],
      internal_: [TableElement, []],
      created: [TableElement, []]
    }),
    methods: () => ({
      addRow: [TableRowElement, []],
      createCaption: [TableCaptionElement, []],
      createTBody: [TableSectionElement, []],
      createTFoot: [TableSectionElement, []],
      createTHead: [TableSectionElement, []],
      insertRow: [TableRowElement, [core.int]],
      [_createCaption]: [HtmlElement, []],
      [_createCaption_1]: [HtmlElement, []],
      [_createTFoot]: [HtmlElement, []],
      [_createTFoot_1]: [HtmlElement, []],
      [_createTHead]: [HtmlElement, []],
      [_createTHead_1]: [HtmlElement, []],
      deleteCaption: [dart.void, []],
      [_deleteCaption_1]: [dart.void, []],
      deleteRow: [dart.void, [core.int]],
      [_deleteRow_1]: [dart.void, [dart.dynamic]],
      deleteTFoot: [dart.void, []],
      [_deleteTFoot_1]: [dart.void, []],
      deleteTHead: [dart.void, []],
      [_deleteTHead_1]: [dart.void, []],
      [_insertRow]: [HtmlElement, [], [core.int]],
      [_insertRow_1]: [HtmlElement, [dart.dynamic]],
      [_insertRow_2]: [HtmlElement, []]
    }),
    statics: () => ({internalCreateTableElement: [TableElement, []]}),
    names: ['internalCreateTableElement']
  });
  TableElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableElement')), dart.const(new _js_helper.Native("HTMLTableElement"))];
  const _cells = Symbol('_cells');
  const _insertCell = Symbol('_insertCell');
  const _deleteCell_1 = Symbol('_deleteCell_1');
  const _insertCell_1 = Symbol('_insertCell_1');
  const _insertCell_2 = Symbol('_insertCell_2');
  class TableRowElement extends HtmlElement {
    get cells() {
      return new (_WrappedList$(TableCellElement))(this[_cells]);
    }
    addCell() {
      return this.insertCell(-1);
    }
    insertCell(index) {
      return dart.as(this[_insertCell](index), TableCellElement);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("tr"), TableRowElement);
    }
    static internalCreateTableRowElement() {
      return TableRowElement._internalWrap();
    }
    static _internalWrap() {
      return new TableRowElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get [_cells]() {
      return dart.as(wrap_jso(this.raw.cells), core.List$(Node));
    }
    get rowIndex() {
      return dart.as(wrap_jso(this.raw.rowIndex), core.int);
    }
    get sectionRowIndex() {
      return dart.as(wrap_jso(this.raw.sectionRowIndex), core.int);
    }
    deleteCell(index) {
      this[_deleteCell_1](index);
      return;
    }
    [_deleteCell_1](index) {
      return wrap_jso(this.raw.deleteCell(unwrap_jso(index)));
    }
    [_insertCell](index) {
      if (index === void 0)
        index = null;
      if (index != null) {
        return this[_insertCell_1](index);
      }
      return this[_insertCell_2]();
    }
    [_insertCell_1](index) {
      return dart.as(wrap_jso(this.raw.insertCell(unwrap_jso(index))), HtmlElement);
    }
    [_insertCell_2]() {
      return dart.as(wrap_jso(this.raw.insertCell()), HtmlElement);
    }
  }
  dart.defineNamedConstructor(TableRowElement, 'internal_');
  dart.defineNamedConstructor(TableRowElement, 'created');
  dart.setSignature(TableRowElement, {
    constructors: () => ({
      _: [TableRowElement, []],
      new: [TableRowElement, []],
      _internalWrap: [TableRowElement, []],
      internal_: [TableRowElement, []],
      created: [TableRowElement, []]
    }),
    methods: () => ({
      addCell: [TableCellElement, []],
      insertCell: [TableCellElement, [core.int]],
      deleteCell: [dart.void, [core.int]],
      [_deleteCell_1]: [dart.void, [dart.dynamic]],
      [_insertCell]: [HtmlElement, [], [core.int]],
      [_insertCell_1]: [HtmlElement, [dart.dynamic]],
      [_insertCell_2]: [HtmlElement, []]
    }),
    statics: () => ({internalCreateTableRowElement: [TableRowElement, []]}),
    names: ['internalCreateTableRowElement']
  });
  TableRowElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableRowElement')), dart.const(new _js_helper.Native("HTMLTableRowElement"))];
  class TableSectionElement extends HtmlElement {
    get rows() {
      return new (_WrappedList$(TableRowElement))(this[_rows]);
    }
    addRow() {
      return this.insertRow(-1);
    }
    insertRow(index) {
      return dart.as(this[_insertRow](index), TableRowElement);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTableSectionElement() {
      return TableSectionElement._internalWrap();
    }
    static _internalWrap() {
      return new TableSectionElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get [_rows]() {
      return dart.as(wrap_jso(this.raw.rows), core.List$(Node));
    }
    deleteRow(index) {
      this[_deleteRow_1](index);
      return;
    }
    [_deleteRow_1](index) {
      return wrap_jso(this.raw.deleteRow(unwrap_jso(index)));
    }
    [_insertRow](index) {
      if (index === void 0)
        index = null;
      if (index != null) {
        return this[_insertRow_1](index);
      }
      return this[_insertRow_2]();
    }
    [_insertRow_1](index) {
      return dart.as(wrap_jso(this.raw.insertRow(unwrap_jso(index))), HtmlElement);
    }
    [_insertRow_2]() {
      return dart.as(wrap_jso(this.raw.insertRow()), HtmlElement);
    }
  }
  dart.defineNamedConstructor(TableSectionElement, 'internal_');
  dart.defineNamedConstructor(TableSectionElement, 'created');
  dart.setSignature(TableSectionElement, {
    constructors: () => ({
      _: [TableSectionElement, []],
      _internalWrap: [TableSectionElement, []],
      internal_: [TableSectionElement, []],
      created: [TableSectionElement, []]
    }),
    methods: () => ({
      addRow: [TableRowElement, []],
      insertRow: [TableRowElement, [core.int]],
      deleteRow: [dart.void, [core.int]],
      [_deleteRow_1]: [dart.void, [dart.dynamic]],
      [_insertRow]: [HtmlElement, [], [core.int]],
      [_insertRow_1]: [HtmlElement, [dart.dynamic]],
      [_insertRow_2]: [HtmlElement, []]
    }),
    statics: () => ({internalCreateTableSectionElement: [TableSectionElement, []]}),
    names: ['internalCreateTableSectionElement']
  });
  TableSectionElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTableSectionElement')), dart.const(new _js_helper.Native("HTMLTableSectionElement"))];
  class TemplateElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("template"), TemplateElement);
    }
    static internalCreateTemplateElement() {
      return TemplateElement._internalWrap();
    }
    static _internalWrap() {
      return new TemplateElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('template');
    }
    get content() {
      return dart.as(wrap_jso(this.raw.content), DocumentFragment);
    }
    setInnerHtml(html, opts) {
      let validator = opts && 'validator' in opts ? opts.validator : null;
      let treeSanitizer = opts && 'treeSanitizer' in opts ? opts.treeSanitizer : null;
      this.text = null;
      let fragment = this.createFragment(html, {validator: validator, treeSanitizer: treeSanitizer});
      this.content.append(fragment);
    }
  }
  dart.defineNamedConstructor(TemplateElement, 'internal_');
  dart.defineNamedConstructor(TemplateElement, 'created');
  dart.setSignature(TemplateElement, {
    constructors: () => ({
      _: [TemplateElement, []],
      new: [TemplateElement, []],
      _internalWrap: [TemplateElement, []],
      internal_: [TemplateElement, []],
      created: [TemplateElement, []]
    }),
    statics: () => ({internalCreateTemplateElement: [TemplateElement, []]}),
    names: ['internalCreateTemplateElement']
  });
  TemplateElement[dart.metadata] = () => [dart.const(new _metadata.Experimental()), dart.const(new _metadata.DomName('HTMLTemplateElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("HTMLTemplateElement"))];
  const _appendData_1 = Symbol('_appendData_1');
  const _deleteData_1 = Symbol('_deleteData_1');
  const _insertData_1 = Symbol('_insertData_1');
  const _replaceData_1 = Symbol('_replaceData_1');
  const _substringData_1 = Symbol('_substringData_1');
  class CharacterData extends Node {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateCharacterData() {
      return CharacterData._internalWrap();
    }
    static _internalWrap() {
      return new CharacterData.internal_();
    }
    internal_() {
      super.internal_();
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), core.String);
    }
    set data(val) {
      return this.raw.data = unwrap_jso(val);
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    appendData(data) {
      this[_appendData_1](data);
      return;
    }
    [_appendData_1](data) {
      return wrap_jso(this.raw.appendData(unwrap_jso(data)));
    }
    deleteData(offset, length) {
      this[_deleteData_1](offset, length);
      return;
    }
    [_deleteData_1](offset, length) {
      return wrap_jso(this.raw.deleteData(unwrap_jso(offset), unwrap_jso(length)));
    }
    insertData(offset, data) {
      this[_insertData_1](offset, data);
      return;
    }
    [_insertData_1](offset, data) {
      return wrap_jso(this.raw.insertData(unwrap_jso(offset), unwrap_jso(data)));
    }
    replaceData(offset, length, data) {
      this[_replaceData_1](offset, length, data);
      return;
    }
    [_replaceData_1](offset, length, data) {
      return wrap_jso(this.raw.replaceData(unwrap_jso(offset), unwrap_jso(length), unwrap_jso(data)));
    }
    substringData(offset, length) {
      return this[_substringData_1](offset, length);
    }
    [_substringData_1](offset, length) {
      return dart.as(wrap_jso(this.raw.substringData(unwrap_jso(offset), unwrap_jso(length))), core.String);
    }
    get nextElementSibling() {
      return dart.as(wrap_jso(this.raw.nextElementSibling), Element);
    }
    get previousElementSibling() {
      return dart.as(wrap_jso(this.raw.previousElementSibling), Element);
    }
  }
  CharacterData[dart.implements] = () => [ChildNode];
  dart.defineNamedConstructor(CharacterData, 'internal_');
  dart.setSignature(CharacterData, {
    constructors: () => ({
      _: [CharacterData, []],
      _internalWrap: [CharacterData, []],
      internal_: [CharacterData, []]
    }),
    methods: () => ({
      appendData: [dart.void, [core.String]],
      [_appendData_1]: [dart.void, [dart.dynamic]],
      deleteData: [dart.void, [core.int, core.int]],
      [_deleteData_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      insertData: [dart.void, [core.int, core.String]],
      [_insertData_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      replaceData: [dart.void, [core.int, core.int, core.String]],
      [_replaceData_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      substringData: [core.String, [core.int, core.int]],
      [_substringData_1]: [core.String, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateCharacterData: [CharacterData, []]}),
    names: ['internalCreateCharacterData']
  });
  const _splitText_1 = Symbol('_splitText_1');
  class Text extends CharacterData {
    static new(data) {
      return exports.document[_createTextNode](data);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateText() {
      return Text._internalWrap();
    }
    static _internalWrap() {
      return new Text.internal_();
    }
    internal_() {
      super.internal_();
    }
    get wholeText() {
      return dart.as(wrap_jso(this.raw.wholeText), core.String);
    }
    getDestinationInsertionPoints() {
      return this[_getDestinationInsertionPoints_1]();
    }
    [_getDestinationInsertionPoints_1]() {
      return dart.as(wrap_jso(this.raw.getDestinationInsertionPoints()), NodeList);
    }
    splitText(offset) {
      return this[_splitText_1](offset);
    }
    [_splitText_1](offset) {
      return dart.as(wrap_jso(this.raw.splitText(unwrap_jso(offset))), Text);
    }
  }
  dart.defineNamedConstructor(Text, 'internal_');
  dart.setSignature(Text, {
    constructors: () => ({
      new: [Text, [core.String]],
      _: [Text, []],
      _internalWrap: [Text, []],
      internal_: [Text, []]
    }),
    methods: () => ({
      getDestinationInsertionPoints: [NodeList, []],
      [_getDestinationInsertionPoints_1]: [NodeList, []],
      splitText: [Text, [core.int]],
      [_splitText_1]: [Text, [dart.dynamic]]
    }),
    statics: () => ({internalCreateText: [Text, []]}),
    names: ['internalCreateText']
  });
  Text[dart.metadata] = () => [dart.const(new _metadata.DomName('Text')), dart.const(new _js_helper.Native("Text"))];
  class TextAreaElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("textarea"), TextAreaElement);
    }
    static internalCreateTextAreaElement() {
      return TextAreaElement._internalWrap();
    }
    static _internalWrap() {
      return new TextAreaElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    get autofocus() {
      return dart.as(wrap_jso(this.raw.autofocus), core.bool);
    }
    set autofocus(val) {
      return this.raw.autofocus = unwrap_jso(val);
    }
    get cols() {
      return dart.as(wrap_jso(this.raw.cols), core.int);
    }
    set cols(val) {
      return this.raw.cols = unwrap_jso(val);
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.defaultValue), core.String);
    }
    set defaultValue(val) {
      return this.raw.defaultValue = unwrap_jso(val);
    }
    get dirName() {
      return dart.as(wrap_jso(this.raw.dirName), core.String);
    }
    set dirName(val) {
      return this.raw.dirName = unwrap_jso(val);
    }
    get disabled() {
      return dart.as(wrap_jso(this.raw.disabled), core.bool);
    }
    set disabled(val) {
      return this.raw.disabled = unwrap_jso(val);
    }
    get form() {
      return wrap_jso(this.raw.form);
    }
    get inputMode() {
      return dart.as(wrap_jso(this.raw.inputMode), core.String);
    }
    set inputMode(val) {
      return this.raw.inputMode = unwrap_jso(val);
    }
    get labels() {
      return dart.as(wrap_jso(this.raw.labels), core.List$(Node));
    }
    get maxLength() {
      return dart.as(wrap_jso(this.raw.maxLength), core.int);
    }
    set maxLength(val) {
      return this.raw.maxLength = unwrap_jso(val);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get placeholder() {
      return dart.as(wrap_jso(this.raw.placeholder), core.String);
    }
    set placeholder(val) {
      return this.raw.placeholder = unwrap_jso(val);
    }
    get readOnly() {
      return dart.as(wrap_jso(this.raw.readOnly), core.bool);
    }
    set readOnly(val) {
      return this.raw.readOnly = unwrap_jso(val);
    }
    get required() {
      return dart.as(wrap_jso(this.raw.required), core.bool);
    }
    set required(val) {
      return this.raw.required = unwrap_jso(val);
    }
    get rows() {
      return dart.as(wrap_jso(this.raw.rows), core.int);
    }
    set rows(val) {
      return this.raw.rows = unwrap_jso(val);
    }
    get selectionDirection() {
      return dart.as(wrap_jso(this.raw.selectionDirection), core.String);
    }
    set selectionDirection(val) {
      return this.raw.selectionDirection = unwrap_jso(val);
    }
    get selectionEnd() {
      return dart.as(wrap_jso(this.raw.selectionEnd), core.int);
    }
    set selectionEnd(val) {
      return this.raw.selectionEnd = unwrap_jso(val);
    }
    get selectionStart() {
      return dart.as(wrap_jso(this.raw.selectionStart), core.int);
    }
    set selectionStart(val) {
      return this.raw.selectionStart = unwrap_jso(val);
    }
    get textLength() {
      return dart.as(wrap_jso(this.raw.textLength), core.int);
    }
    get type() {
      return dart.as(wrap_jso(this.raw.type), core.String);
    }
    get validationMessage() {
      return dart.as(wrap_jso(this.raw.validationMessage), core.String);
    }
    get validity() {
      return dart.as(wrap_jso(this.raw.validity), ValidityState);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
    get willValidate() {
      return dart.as(wrap_jso(this.raw.willValidate), core.bool);
    }
    get wrap() {
      return dart.as(wrap_jso(this.raw.wrap), core.String);
    }
    set wrap(val) {
      return this.raw.wrap = unwrap_jso(val);
    }
    checkValidity() {
      return this[_checkValidity_1]();
    }
    [_checkValidity_1]() {
      return dart.as(wrap_jso(this.raw.checkValidity()), core.bool);
    }
    select() {
      this[_select_1]();
      return;
    }
    [_select_1]() {
      return wrap_jso(this.raw.select());
    }
    setCustomValidity(error) {
      this[_setCustomValidity_1](error);
      return;
    }
    [_setCustomValidity_1](error) {
      return wrap_jso(this.raw.setCustomValidity(unwrap_jso(error)));
    }
    setRangeText(replacement, opts) {
      let start = opts && 'start' in opts ? opts.start : null;
      let end = opts && 'end' in opts ? opts.end : null;
      let selectionMode = opts && 'selectionMode' in opts ? opts.selectionMode : null;
      if (start == null && end == null && selectionMode == null) {
        this[_setRangeText_1](replacement);
        return;
      }
      if (selectionMode != null && end != null && start != null) {
        this[_setRangeText_2](replacement, start, end, selectionMode);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_setRangeText_1](replacement) {
      return wrap_jso(this.raw.setRangeText(unwrap_jso(replacement)));
    }
    [_setRangeText_2](replacement, start, end, selectionMode) {
      return wrap_jso(this.raw.setRangeText(unwrap_jso(replacement), unwrap_jso(start), unwrap_jso(end), unwrap_jso(selectionMode)));
    }
    setSelectionRange(start, end, direction) {
      if (direction === void 0)
        direction = null;
      if (direction != null) {
        this[_setSelectionRange_1](start, end, direction);
        return;
      }
      this[_setSelectionRange_2](start, end);
      return;
    }
    [_setSelectionRange_1](start, end, direction) {
      return wrap_jso(this.raw.setSelectionRange(unwrap_jso(start), unwrap_jso(end), unwrap_jso(direction)));
    }
    [_setSelectionRange_2](start, end) {
      return wrap_jso(this.raw.setSelectionRange(unwrap_jso(start), unwrap_jso(end)));
    }
  }
  dart.defineNamedConstructor(TextAreaElement, 'internal_');
  dart.defineNamedConstructor(TextAreaElement, 'created');
  dart.setSignature(TextAreaElement, {
    constructors: () => ({
      _: [TextAreaElement, []],
      new: [TextAreaElement, []],
      _internalWrap: [TextAreaElement, []],
      internal_: [TextAreaElement, []],
      created: [TextAreaElement, []]
    }),
    methods: () => ({
      checkValidity: [core.bool, []],
      [_checkValidity_1]: [core.bool, []],
      select: [dart.void, []],
      [_select_1]: [dart.void, []],
      setCustomValidity: [dart.void, [core.String]],
      [_setCustomValidity_1]: [dart.void, [dart.dynamic]],
      setRangeText: [dart.void, [core.String], {start: core.int, end: core.int, selectionMode: core.String}],
      [_setRangeText_1]: [dart.void, [dart.dynamic]],
      [_setRangeText_2]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      setSelectionRange: [dart.void, [core.int, core.int], [core.String]],
      [_setSelectionRange_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_setSelectionRange_2]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateTextAreaElement: [TextAreaElement, []]}),
    names: ['internalCreateTextAreaElement']
  });
  TextAreaElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTextAreaElement')), dart.const(new _js_helper.Native("HTMLTextAreaElement"))];
  const _initTextEvent = Symbol('_initTextEvent');
  const _initTextEvent_1 = Symbol('_initTextEvent_1');
  class TextEvent extends UIEvent {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : false;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : false;
      let view = opts && 'view' in opts ? opts.view : null;
      let data = opts && 'data' in opts ? opts.data : null;
      if (view == null) {
        view = exports.window;
      }
      let e = exports.document[_createEvent]("TextEvent");
      dart.dcall(e[_initTextEvent], type, canBubble, cancelable, view, data);
      return dart.as(e, TextEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextEvent() {
      return TextEvent._internalWrap();
    }
    static _internalWrap() {
      return new TextEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get data() {
      return dart.as(wrap_jso(this.raw.data), core.String);
    }
    [_initTextEvent](typeArg, canBubbleArg, cancelableArg, viewArg, dataArg) {
      this[_initTextEvent_1](typeArg, canBubbleArg, cancelableArg, viewArg, dataArg);
      return;
    }
    [_initTextEvent_1](typeArg, canBubbleArg, cancelableArg, viewArg, dataArg) {
      return wrap_jso(this.raw.initTextEvent(unwrap_jso(typeArg), unwrap_jso(canBubbleArg), unwrap_jso(cancelableArg), unwrap_jso(viewArg), unwrap_jso(dataArg)));
    }
  }
  dart.defineNamedConstructor(TextEvent, 'internal_');
  dart.setSignature(TextEvent, {
    constructors: () => ({
      new: [TextEvent, [core.String], {canBubble: core.bool, cancelable: core.bool, view: Window, data: core.String}],
      _: [TextEvent, []],
      _internalWrap: [TextEvent, []],
      internal_: [TextEvent, []]
    }),
    methods: () => ({
      [_initTextEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.String]],
      [_initTextEvent_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, Window, dart.dynamic]]
    }),
    statics: () => ({internalCreateTextEvent: [TextEvent, []]}),
    names: ['internalCreateTextEvent']
  });
  TextEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('TextEvent')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("TextEvent"))];
  class TextMetrics extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextMetrics() {
      return TextMetrics._internalWrap();
    }
    static _internalWrap() {
      return new TextMetrics.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get actualBoundingBoxAscent() {
      return dart.as(wrap_jso(this.raw.actualBoundingBoxAscent), core.double);
    }
    get actualBoundingBoxDescent() {
      return dart.as(wrap_jso(this.raw.actualBoundingBoxDescent), core.double);
    }
    get actualBoundingBoxLeft() {
      return dart.as(wrap_jso(this.raw.actualBoundingBoxLeft), core.double);
    }
    get actualBoundingBoxRight() {
      return dart.as(wrap_jso(this.raw.actualBoundingBoxRight), core.double);
    }
    get alphabeticBaseline() {
      return dart.as(wrap_jso(this.raw.alphabeticBaseline), core.double);
    }
    get emHeightAscent() {
      return dart.as(wrap_jso(this.raw.emHeightAscent), core.double);
    }
    get emHeightDescent() {
      return dart.as(wrap_jso(this.raw.emHeightDescent), core.double);
    }
    get fontBoundingBoxAscent() {
      return dart.as(wrap_jso(this.raw.fontBoundingBoxAscent), core.double);
    }
    get fontBoundingBoxDescent() {
      return dart.as(wrap_jso(this.raw.fontBoundingBoxDescent), core.double);
    }
    get hangingBaseline() {
      return dart.as(wrap_jso(this.raw.hangingBaseline), core.double);
    }
    get ideographicBaseline() {
      return dart.as(wrap_jso(this.raw.ideographicBaseline), core.double);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.double);
    }
  }
  dart.defineNamedConstructor(TextMetrics, 'internal_');
  dart.setSignature(TextMetrics, {
    constructors: () => ({
      _: [TextMetrics, []],
      _internalWrap: [TextMetrics, []],
      internal_: [TextMetrics, []]
    }),
    statics: () => ({internalCreateTextMetrics: [TextMetrics, []]}),
    names: ['internalCreateTextMetrics']
  });
  TextMetrics[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TextMetrics')), dart.const(new _js_helper.Native("TextMetrics"))];
  const _addCue_1 = Symbol('_addCue_1');
  const _addRegion_1 = Symbol('_addRegion_1');
  const _removeCue_1 = Symbol('_removeCue_1');
  const _removeRegion_1 = Symbol('_removeRegion_1');
  class TextTrack extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextTrack() {
      return TextTrack._internalWrap();
    }
    static _internalWrap() {
      return new TextTrack.internal_();
    }
    internal_() {
      super.internal_();
    }
    get activeCues() {
      return dart.as(wrap_jso(this.raw.activeCues), TextTrackCueList);
    }
    get cues() {
      return dart.as(wrap_jso(this.raw.cues), TextTrackCueList);
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    get kind() {
      return dart.as(wrap_jso(this.raw.kind), core.String);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    get language() {
      return dart.as(wrap_jso(this.raw.language), core.String);
    }
    get mode() {
      return dart.as(wrap_jso(this.raw.mode), core.String);
    }
    set mode(val) {
      return this.raw.mode = unwrap_jso(val);
    }
    get regions() {
      return wrap_jso(this.raw.regions);
    }
    addCue(cue) {
      this[_addCue_1](cue);
      return;
    }
    [_addCue_1](cue) {
      return wrap_jso(this.raw.addCue(unwrap_jso(cue)));
    }
    addRegion(region) {
      this[_addRegion_1](region);
      return;
    }
    [_addRegion_1](region) {
      return wrap_jso(this.raw.addRegion(unwrap_jso(region)));
    }
    removeCue(cue) {
      this[_removeCue_1](cue);
      return;
    }
    [_removeCue_1](cue) {
      return wrap_jso(this.raw.removeCue(unwrap_jso(cue)));
    }
    removeRegion(region) {
      this[_removeRegion_1](region);
      return;
    }
    [_removeRegion_1](region) {
      return wrap_jso(this.raw.removeRegion(unwrap_jso(region)));
    }
    get onCueChange() {
      return TextTrack.cueChangeEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(TextTrack, 'internal_');
  dart.setSignature(TextTrack, {
    constructors: () => ({
      _: [TextTrack, []],
      _internalWrap: [TextTrack, []],
      internal_: [TextTrack, []]
    }),
    methods: () => ({
      addCue: [dart.void, [TextTrackCue]],
      [_addCue_1]: [dart.void, [TextTrackCue]],
      addRegion: [dart.void, [dart.dynamic]],
      [_addRegion_1]: [dart.void, [dart.dynamic]],
      removeCue: [dart.void, [TextTrackCue]],
      [_removeCue_1]: [dart.void, [TextTrackCue]],
      removeRegion: [dart.void, [dart.dynamic]],
      [_removeRegion_1]: [dart.void, [dart.dynamic]]
    }),
    statics: () => ({internalCreateTextTrack: [TextTrack, []]}),
    names: ['internalCreateTextTrack']
  });
  TextTrack[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TextTrack')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TextTrack"))];
  TextTrack.cueChangeEvent = dart.const(new (EventStreamProvider$(Event))('cuechange'));
  class TextTrackCue extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextTrackCue() {
      return TextTrackCue._internalWrap();
    }
    static _internalWrap() {
      return new TextTrackCue.internal_();
    }
    internal_() {
      super.internal_();
    }
    get endTime() {
      return dart.as(wrap_jso(this.raw.endTime), core.num);
    }
    set endTime(val) {
      return this.raw.endTime = unwrap_jso(val);
    }
    get id() {
      return dart.as(wrap_jso(this.raw.id), core.String);
    }
    set id(val) {
      return this.raw.id = unwrap_jso(val);
    }
    get pauseOnExit() {
      return dart.as(wrap_jso(this.raw.pauseOnExit), core.bool);
    }
    set pauseOnExit(val) {
      return this.raw.pauseOnExit = unwrap_jso(val);
    }
    get startTime() {
      return dart.as(wrap_jso(this.raw.startTime), core.num);
    }
    set startTime(val) {
      return this.raw.startTime = unwrap_jso(val);
    }
    get track() {
      return dart.as(wrap_jso(this.raw.track), TextTrack);
    }
    get onEnter() {
      return TextTrackCue.enterEvent.forTarget(this);
    }
    get onExit() {
      return TextTrackCue.exitEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(TextTrackCue, 'internal_');
  dart.setSignature(TextTrackCue, {
    constructors: () => ({
      _: [TextTrackCue, []],
      _internalWrap: [TextTrackCue, []],
      internal_: [TextTrackCue, []]
    }),
    statics: () => ({internalCreateTextTrackCue: [TextTrackCue, []]}),
    names: ['internalCreateTextTrackCue']
  });
  TextTrackCue[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TextTrackCue')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TextTrackCue"))];
  TextTrackCue.enterEvent = dart.const(new (EventStreamProvider$(Event))('enter'));
  TextTrackCue.exitEvent = dart.const(new (EventStreamProvider$(Event))('exit'));
  const _getCueById_1 = Symbol('_getCueById_1');
  class TextTrackCueList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(TextTrackCue), ImmutableListMixin$(TextTrackCue)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextTrackCueList() {
      return TextTrackCueList._internalWrap();
    }
    static _internalWrap() {
      return new TextTrackCueList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), TextTrackCue);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), TextTrackCue);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), TextTrackCue);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), TextTrackCue);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    getCueById(id) {
      return this[_getCueById_1](id);
    }
    [_getCueById_1](id) {
      return dart.as(wrap_jso(this.raw.getCueById(unwrap_jso(id))), TextTrackCue);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), TextTrackCue);
    }
  }
  TextTrackCueList[dart.implements] = () => [core.List$(TextTrackCue), _js_helper.JavaScriptIndexingBehavior];
  dart.defineNamedConstructor(TextTrackCueList, 'internal_');
  dart.setSignature(TextTrackCueList, {
    constructors: () => ({
      _: [TextTrackCueList, []],
      _internalWrap: [TextTrackCueList, []],
      internal_: [TextTrackCueList, []]
    }),
    methods: () => ({
      get: [TextTrackCue, [core.int]],
      set: [dart.void, [core.int, TextTrackCue]],
      elementAt: [TextTrackCue, [core.int]],
      getCueById: [TextTrackCue, [core.String]],
      [_getCueById_1]: [TextTrackCue, [dart.dynamic]],
      item: [TextTrackCue, [core.int]],
      [_item_1]: [TextTrackCue, [dart.dynamic]]
    }),
    statics: () => ({internalCreateTextTrackCueList: [TextTrackCueList, []]}),
    names: ['internalCreateTextTrackCueList']
  });
  dart.defineExtensionMembers(TextTrackCueList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  TextTrackCueList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TextTrackCueList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TextTrackCueList"))];
  const _getTrackById_1 = Symbol('_getTrackById_1');
  class TextTrackList extends dart.mixin(EventTarget, collection.ListMixin$(TextTrack), ImmutableListMixin$(TextTrack)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTextTrackList() {
      return TextTrackList._internalWrap();
    }
    static _internalWrap() {
      return new TextTrackList.internal_();
    }
    internal_() {
      super.internal_();
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), TextTrack);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), TextTrack);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), TextTrack);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), TextTrack);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    getTrackById(id) {
      return this[_getTrackById_1](id);
    }
    [_getTrackById_1](id) {
      return dart.as(wrap_jso(this.raw.getTrackById(unwrap_jso(id))), TextTrack);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), TextTrack);
    }
    get onAddTrack() {
      return TextTrackList.addTrackEvent.forTarget(this);
    }
    get onChange() {
      return TextTrackList.changeEvent.forTarget(this);
    }
  }
  TextTrackList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(TextTrack)];
  dart.defineNamedConstructor(TextTrackList, 'internal_');
  dart.setSignature(TextTrackList, {
    constructors: () => ({
      _: [TextTrackList, []],
      _internalWrap: [TextTrackList, []],
      internal_: [TextTrackList, []]
    }),
    methods: () => ({
      get: [TextTrack, [core.int]],
      set: [dart.void, [core.int, TextTrack]],
      elementAt: [TextTrack, [core.int]],
      getTrackById: [TextTrack, [core.String]],
      [_getTrackById_1]: [TextTrack, [dart.dynamic]],
      item: [TextTrack, [core.int]],
      [_item_1]: [TextTrack, [dart.dynamic]]
    }),
    statics: () => ({internalCreateTextTrackList: [TextTrackList, []]}),
    names: ['internalCreateTextTrackList']
  });
  dart.defineExtensionMembers(TextTrackList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  TextTrackList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TextTrackList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TextTrackList"))];
  class TrackEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTrackEvent() {
      return TrackEvent._internalWrap();
    }
    static _internalWrap() {
      return new TrackEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get track() {
      return wrap_jso(this.raw.track);
    }
  }
  dart.defineNamedConstructor(TrackEvent, 'internal_');
  dart.setSignature(TrackEvent, {
    constructors: () => ({
      _: [TrackEvent, []],
      _internalWrap: [TrackEvent, []],
      internal_: [TrackEvent, []]
    }),
    statics: () => ({internalCreateTrackEvent: [TrackEvent, []]}),
    names: ['internalCreateTrackEvent']
  });
  TrackEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TrackEvent')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("TrackEvent"))];
  TextTrackList.addTrackEvent = dart.const(new (EventStreamProvider$(TrackEvent))('addtrack'));
  TextTrackList.changeEvent = dart.const(new (EventStreamProvider$(Event))('change'));
  const _end_1 = Symbol('_end_1');
  class TimeRanges extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTimeRanges() {
      return TimeRanges._internalWrap();
    }
    static _internalWrap() {
      return new TimeRanges.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    end(index) {
      return this[_end_1](index);
    }
    [_end_1](index) {
      return dart.as(wrap_jso(this.raw.end(unwrap_jso(index))), core.num);
    }
    start(index) {
      return this[_start_1](index);
    }
    [_start_1](index) {
      return dart.as(wrap_jso(this.raw.start(unwrap_jso(index))), core.num);
    }
  }
  dart.defineNamedConstructor(TimeRanges, 'internal_');
  dart.setSignature(TimeRanges, {
    constructors: () => ({
      _: [TimeRanges, []],
      _internalWrap: [TimeRanges, []],
      internal_: [TimeRanges, []]
    }),
    methods: () => ({
      end: [core.num, [core.int]],
      [_end_1]: [core.num, [dart.dynamic]],
      start: [core.num, [core.int]],
      [_start_1]: [core.num, [dart.dynamic]]
    }),
    statics: () => ({internalCreateTimeRanges: [TimeRanges, []]}),
    names: ['internalCreateTimeRanges']
  });
  TimeRanges[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TimeRanges')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("TimeRanges"))];
  const TimeoutHandler = dart.typedef('TimeoutHandler', () => dart.functionType(dart.void, []));
  class Timing extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTiming() {
      return Timing._internalWrap();
    }
    static _internalWrap() {
      return new Timing.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get delay() {
      return dart.as(wrap_jso(this.raw.delay), core.num);
    }
    set delay(val) {
      return this.raw.delay = unwrap_jso(val);
    }
    get direction() {
      return dart.as(wrap_jso(this.raw.direction), core.String);
    }
    set direction(val) {
      return this.raw.direction = unwrap_jso(val);
    }
    get easing() {
      return dart.as(wrap_jso(this.raw.easing), core.String);
    }
    set easing(val) {
      return this.raw.easing = unwrap_jso(val);
    }
    get endDelay() {
      return dart.as(wrap_jso(this.raw.endDelay), core.num);
    }
    set endDelay(val) {
      return this.raw.endDelay = unwrap_jso(val);
    }
    get fill() {
      return dart.as(wrap_jso(this.raw.fill), core.String);
    }
    set fill(val) {
      return this.raw.fill = unwrap_jso(val);
    }
    get iterationStart() {
      return dart.as(wrap_jso(this.raw.iterationStart), core.num);
    }
    set iterationStart(val) {
      return this.raw.iterationStart = unwrap_jso(val);
    }
    get iterations() {
      return dart.as(wrap_jso(this.raw.iterations), core.num);
    }
    set iterations(val) {
      return this.raw.iterations = unwrap_jso(val);
    }
    get playbackRate() {
      return dart.as(wrap_jso(this.raw.playbackRate), core.num);
    }
    set playbackRate(val) {
      return this.raw.playbackRate = unwrap_jso(val);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return wrap_jso(this.raw.__getter__(unwrap_jso(name)));
    }
    [__setter__](name, duration) {
      this[__setter___1](name, duration);
      return;
    }
    [__setter___1](name, duration) {
      return wrap_jso(this.raw.__setter__(unwrap_jso(name), unwrap_jso(duration)));
    }
  }
  dart.defineNamedConstructor(Timing, 'internal_');
  dart.setSignature(Timing, {
    constructors: () => ({
      _: [Timing, []],
      _internalWrap: [Timing, []],
      internal_: [Timing, []]
    }),
    methods: () => ({
      [__getter__]: [core.Object, [core.String]],
      [__getter___1]: [core.Object, [dart.dynamic]],
      [__setter__]: [dart.void, [core.String, core.num]],
      [__setter___1]: [dart.void, [dart.dynamic, dart.dynamic]]
    }),
    statics: () => ({internalCreateTiming: [Timing, []]}),
    names: ['internalCreateTiming']
  });
  Timing[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Timing')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Timing"))];
  class TitleElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("title"), TitleElement);
    }
    static internalCreateTitleElement() {
      return TitleElement._internalWrap();
    }
    static _internalWrap() {
      return new TitleElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(TitleElement, 'internal_');
  dart.defineNamedConstructor(TitleElement, 'created');
  dart.setSignature(TitleElement, {
    constructors: () => ({
      _: [TitleElement, []],
      new: [TitleElement, []],
      _internalWrap: [TitleElement, []],
      internal_: [TitleElement, []],
      created: [TitleElement, []]
    }),
    statics: () => ({internalCreateTitleElement: [TitleElement, []]}),
    names: ['internalCreateTitleElement']
  });
  TitleElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTitleElement')), dart.const(new _js_helper.Native("HTMLTitleElement"))];
  const _radiusX = Symbol('_radiusX');
  const _radiusY = Symbol('_radiusY');
  const __clientX = Symbol('__clientX');
  const __clientY = Symbol('__clientY');
  const __screenX = Symbol('__screenX');
  const __screenY = Symbol('__screenY');
  const __pageX = Symbol('__pageX');
  const __pageY = Symbol('__pageY');
  const __radiusX = Symbol('__radiusX');
  const __radiusY = Symbol('__radiusY');
  class Touch extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTouch() {
      return Touch._internalWrap();
    }
    static _internalWrap() {
      return new Touch.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get [_clientX]() {
      return dart.as(wrap_jso(this.raw.clientX), core.double);
    }
    get [_clientY]() {
      return dart.as(wrap_jso(this.raw.clientY), core.double);
    }
    get force() {
      return dart.as(wrap_jso(this.raw.force), core.double);
    }
    get identifier() {
      return dart.as(wrap_jso(this.raw.identifier), core.int);
    }
    get [_pageX]() {
      return dart.as(wrap_jso(this.raw.pageX), core.double);
    }
    get [_pageY]() {
      return dart.as(wrap_jso(this.raw.pageY), core.double);
    }
    get [_radiusX]() {
      return dart.as(wrap_jso(this.raw.radiusX), core.double);
    }
    get [_radiusY]() {
      return dart.as(wrap_jso(this.raw.radiusY), core.double);
    }
    get [_screenX]() {
      return dart.as(wrap_jso(this.raw.screenX), core.double);
    }
    get [_screenY]() {
      return dart.as(wrap_jso(this.raw.screenY), core.double);
    }
    get target() {
      return _convertNativeToDart_EventTarget(this[_get_target]);
    }
    get [_get_target]() {
      return wrap_jso(this.raw.target);
    }
    get rotationAngle() {
      return dart.as(wrap_jso(this.raw.webkitRotationAngle), core.double);
    }
    get [__clientX]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'clientX_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__clientY]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'clientY_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__screenX]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'screenX_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__screenY]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'screenY_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__pageX]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'pageX_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__pageY]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'pageY_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__radiusX]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'radiusX_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get [__radiusY]() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkTouch'), 'instance'), 'radiusY_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get client() {
      return new math.Point(this[__clientX], this[__clientY]);
    }
    get page() {
      return new math.Point(this[__pageX], this[__pageY]);
    }
    get screen() {
      return new math.Point(this[__screenX], this[__screenY]);
    }
    get radiusX() {
      return this[__radiusX];
    }
    get radiusY() {
      return this[__radiusY];
    }
  }
  dart.defineNamedConstructor(Touch, 'internal_');
  dart.setSignature(Touch, {
    constructors: () => ({
      _: [Touch, []],
      _internalWrap: [Touch, []],
      internal_: [Touch, []]
    }),
    statics: () => ({internalCreateTouch: [Touch, []]}),
    names: ['internalCreateTouch']
  });
  Touch[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Touch')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Touch"))];
  class TouchList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(Touch), ImmutableListMixin$(Touch)) {
    static new() {
      return null;
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTouchList() {
      return TouchList._internalWrap();
    }
    static _internalWrap() {
      return new TouchList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    static get supported() {
      return !!document.createTouchList;
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), Touch);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Touch);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), Touch);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Touch);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Touch);
    }
  }
  TouchList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(Touch)];
  dart.defineNamedConstructor(TouchList, 'internal_');
  dart.setSignature(TouchList, {
    constructors: () => ({
      new: [TouchList, []],
      _: [TouchList, []],
      _internalWrap: [TouchList, []],
      internal_: [TouchList, []]
    }),
    methods: () => ({
      get: [Touch, [core.int]],
      set: [dart.void, [core.int, Touch]],
      elementAt: [Touch, [core.int]],
      item: [Touch, [core.int]],
      [_item_1]: [Touch, [dart.dynamic]]
    }),
    statics: () => ({internalCreateTouchList: [TouchList, []]}),
    names: ['internalCreateTouchList']
  });
  dart.defineExtensionMembers(TouchList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  TouchList[dart.metadata] = () => [dart.const(new _metadata.DomName('TouchList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("TouchList"))];
  class TrackElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("track"), TrackElement);
    }
    static internalCreateTrackElement() {
      return TrackElement._internalWrap();
    }
    static _internalWrap() {
      return new TrackElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
    static get supported() {
      return Element.isTagSupported('track');
    }
    get defaultValue() {
      return dart.as(wrap_jso(this.raw.default), core.bool);
    }
    set defaultValue(val) {
      return this.raw.default = unwrap_jso(val);
    }
    get integrity() {
      return dart.as(wrap_jso(this.raw.integrity), core.String);
    }
    set integrity(val) {
      return this.raw.integrity = unwrap_jso(val);
    }
    get kind() {
      return dart.as(wrap_jso(this.raw.kind), core.String);
    }
    set kind(val) {
      return this.raw.kind = unwrap_jso(val);
    }
    get label() {
      return dart.as(wrap_jso(this.raw.label), core.String);
    }
    set label(val) {
      return this.raw.label = unwrap_jso(val);
    }
    get readyState() {
      return dart.as(wrap_jso(this.raw.readyState), core.int);
    }
    get src() {
      return dart.as(wrap_jso(this.raw.src), core.String);
    }
    set src(val) {
      return this.raw.src = unwrap_jso(val);
    }
    get srclang() {
      return dart.as(wrap_jso(this.raw.srclang), core.String);
    }
    set srclang(val) {
      return this.raw.srclang = unwrap_jso(val);
    }
    get track() {
      return dart.as(wrap_jso(this.raw.track), TextTrack);
    }
  }
  dart.defineNamedConstructor(TrackElement, 'internal_');
  dart.defineNamedConstructor(TrackElement, 'created');
  dart.setSignature(TrackElement, {
    constructors: () => ({
      _: [TrackElement, []],
      new: [TrackElement, []],
      _internalWrap: [TrackElement, []],
      internal_: [TrackElement, []],
      created: [TrackElement, []]
    }),
    statics: () => ({internalCreateTrackElement: [TrackElement, []]}),
    names: ['internalCreateTrackElement']
  });
  TrackElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLTrackElement')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("HTMLTrackElement"))];
  TrackElement.ERROR = 3;
  TrackElement.LOADED = 2;
  TrackElement.LOADING = 1;
  TrackElement.NONE = 0;
  class TransitionEvent extends Event {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTransitionEvent() {
      return TransitionEvent._internalWrap();
    }
    static _internalWrap() {
      return new TransitionEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
    get elapsedTime() {
      return dart.as(wrap_jso(this.raw.elapsedTime), core.double);
    }
    get propertyName() {
      return dart.as(wrap_jso(this.raw.propertyName), core.String);
    }
    get pseudoElement() {
      return dart.as(wrap_jso(this.raw.pseudoElement), core.String);
    }
  }
  dart.defineNamedConstructor(TransitionEvent, 'internal_');
  dart.setSignature(TransitionEvent, {
    constructors: () => ({
      _: [TransitionEvent, []],
      _internalWrap: [TransitionEvent, []],
      internal_: [TransitionEvent, []]
    }),
    statics: () => ({internalCreateTransitionEvent: [TransitionEvent, []]}),
    names: ['internalCreateTransitionEvent']
  });
  TransitionEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('TransitionEvent')), dart.const(new _js_helper.Native("TransitionEvent,WebKitTransitionEvent"))];
  const _createTreeWalker = Symbol('_createTreeWalker');
  const _firstChild_1 = Symbol('_firstChild_1');
  const _lastChild_1 = Symbol('_lastChild_1');
  const _nextSibling_1 = Symbol('_nextSibling_1');
  const _parentNode_1 = Symbol('_parentNode_1');
  const _previousSibling_1 = Symbol('_previousSibling_1');
  class TreeWalker extends DartHtmlDomObject {
    static new(root, whatToShow) {
      return dart.as(dart.dcall(exports.document[_createTreeWalker], root, whatToShow, null), TreeWalker);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateTreeWalker() {
      return TreeWalker._internalWrap();
    }
    static _internalWrap() {
      return new TreeWalker.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get currentNode() {
      return dart.as(wrap_jso(this.raw.currentNode), Node);
    }
    set currentNode(val) {
      return this.raw.currentNode = unwrap_jso(val);
    }
    get filter() {
      return dart.as(wrap_jso(this.raw.filter), NodeFilter);
    }
    get root() {
      return dart.as(wrap_jso(this.raw.root), Node);
    }
    get whatToShow() {
      return dart.as(wrap_jso(this.raw.whatToShow), core.int);
    }
    firstChild() {
      return this[_firstChild_1]();
    }
    [_firstChild_1]() {
      return dart.as(wrap_jso(this.raw.firstChild()), Node);
    }
    lastChild() {
      return this[_lastChild_1]();
    }
    [_lastChild_1]() {
      return dart.as(wrap_jso(this.raw.lastChild()), Node);
    }
    nextNode() {
      return this[_nextNode_1]();
    }
    [_nextNode_1]() {
      return dart.as(wrap_jso(this.raw.nextNode()), Node);
    }
    nextSibling() {
      return this[_nextSibling_1]();
    }
    [_nextSibling_1]() {
      return dart.as(wrap_jso(this.raw.nextSibling()), Node);
    }
    parentNode() {
      return this[_parentNode_1]();
    }
    [_parentNode_1]() {
      return dart.as(wrap_jso(this.raw.parentNode()), Node);
    }
    previousNode() {
      return this[_previousNode_1]();
    }
    [_previousNode_1]() {
      return dart.as(wrap_jso(this.raw.previousNode()), Node);
    }
    previousSibling() {
      return this[_previousSibling_1]();
    }
    [_previousSibling_1]() {
      return dart.as(wrap_jso(this.raw.previousSibling()), Node);
    }
  }
  dart.defineNamedConstructor(TreeWalker, 'internal_');
  dart.setSignature(TreeWalker, {
    constructors: () => ({
      new: [TreeWalker, [Node, core.int]],
      _: [TreeWalker, []],
      _internalWrap: [TreeWalker, []],
      internal_: [TreeWalker, []]
    }),
    methods: () => ({
      firstChild: [Node, []],
      [_firstChild_1]: [Node, []],
      lastChild: [Node, []],
      [_lastChild_1]: [Node, []],
      nextNode: [Node, []],
      [_nextNode_1]: [Node, []],
      nextSibling: [Node, []],
      [_nextSibling_1]: [Node, []],
      parentNode: [Node, []],
      [_parentNode_1]: [Node, []],
      previousNode: [Node, []],
      [_previousNode_1]: [Node, []],
      previousSibling: [Node, []],
      [_previousSibling_1]: [Node, []]
    }),
    statics: () => ({internalCreateTreeWalker: [TreeWalker, []]}),
    names: ['internalCreateTreeWalker']
  });
  TreeWalker[dart.metadata] = () => [dart.const(new _metadata.DomName('TreeWalker')), dart.const(new _metadata.Unstable()), dart.const(new _js_helper.Native("TreeWalker"))];
  class UListElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return dart.as(exports.document.createElement("ul"), UListElement);
    }
    static internalCreateUListElement() {
      return UListElement._internalWrap();
    }
    static _internalWrap() {
      return new UListElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(UListElement, 'internal_');
  dart.defineNamedConstructor(UListElement, 'created');
  dart.setSignature(UListElement, {
    constructors: () => ({
      _: [UListElement, []],
      new: [UListElement, []],
      _internalWrap: [UListElement, []],
      internal_: [UListElement, []],
      created: [UListElement, []]
    }),
    statics: () => ({internalCreateUListElement: [UListElement, []]}),
    names: ['internalCreateUListElement']
  });
  UListElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLUListElement')), dart.const(new _js_helper.Native("HTMLUListElement"))];
  class UnknownElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateUnknownElement() {
      return UnknownElement._internalWrap();
    }
    static _internalWrap() {
      return new UnknownElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(UnknownElement, 'internal_');
  dart.defineNamedConstructor(UnknownElement, 'created');
  dart.setSignature(UnknownElement, {
    constructors: () => ({
      _: [UnknownElement, []],
      _internalWrap: [UnknownElement, []],
      internal_: [UnknownElement, []],
      created: [UnknownElement, []]
    }),
    statics: () => ({internalCreateUnknownElement: [UnknownElement, []]}),
    names: ['internalCreateUnknownElement']
  });
  UnknownElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLUnknownElement')), dart.const(new _js_helper.Native("HTMLUnknownElement"))];
  class Url extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateUrl() {
      return Url._internalWrap();
    }
    static _internalWrap() {
      return new Url.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get hash() {
      return dart.as(wrap_jso(this.raw.hash), core.String);
    }
    set hash(val) {
      return this.raw.hash = unwrap_jso(val);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), core.String);
    }
    set host(val) {
      return this.raw.host = unwrap_jso(val);
    }
    get hostname() {
      return dart.as(wrap_jso(this.raw.hostname), core.String);
    }
    set hostname(val) {
      return this.raw.hostname = unwrap_jso(val);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    set href(val) {
      return this.raw.href = unwrap_jso(val);
    }
    get origin() {
      return dart.as(wrap_jso(this.raw.origin), core.String);
    }
    get password() {
      return dart.as(wrap_jso(this.raw.password), core.String);
    }
    set password(val) {
      return this.raw.password = unwrap_jso(val);
    }
    get pathname() {
      return dart.as(wrap_jso(this.raw.pathname), core.String);
    }
    set pathname(val) {
      return this.raw.pathname = unwrap_jso(val);
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), core.String);
    }
    set port(val) {
      return this.raw.port = unwrap_jso(val);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    set protocol(val) {
      return this.raw.protocol = unwrap_jso(val);
    }
    get search() {
      return dart.as(wrap_jso(this.raw.search), core.String);
    }
    set search(val) {
      return this.raw.search = unwrap_jso(val);
    }
    get username() {
      return dart.as(wrap_jso(this.raw.username), core.String);
    }
    set username(val) {
      return this.raw.username = unwrap_jso(val);
    }
  }
  Url[dart.implements] = () => [UrlUtils];
  dart.defineNamedConstructor(Url, 'internal_');
  dart.setSignature(Url, {
    constructors: () => ({
      _: [Url, []],
      _internalWrap: [Url, []],
      internal_: [Url, []]
    }),
    statics: () => ({internalCreateUrl: [Url, []]}),
    names: ['internalCreateUrl']
  });
  Url[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('URL')), dart.const(new _js_helper.Native("URL"))];
  class UrlUtils extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get hash() {
      return dart.as(wrap_jso(this.raw.hash), core.String);
    }
    set hash(val) {
      return this.raw.hash = unwrap_jso(val);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), core.String);
    }
    set host(val) {
      return this.raw.host = unwrap_jso(val);
    }
    get hostname() {
      return dart.as(wrap_jso(this.raw.hostname), core.String);
    }
    set hostname(val) {
      return this.raw.hostname = unwrap_jso(val);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    set href(val) {
      return this.raw.href = unwrap_jso(val);
    }
    get origin() {
      return dart.as(wrap_jso(this.raw.origin), core.String);
    }
    get password() {
      return dart.as(wrap_jso(this.raw.password), core.String);
    }
    set password(val) {
      return this.raw.password = unwrap_jso(val);
    }
    get pathname() {
      return dart.as(wrap_jso(this.raw.pathname), core.String);
    }
    set pathname(val) {
      return this.raw.pathname = unwrap_jso(val);
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), core.String);
    }
    set port(val) {
      return this.raw.port = unwrap_jso(val);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    set protocol(val) {
      return this.raw.protocol = unwrap_jso(val);
    }
    get search() {
      return dart.as(wrap_jso(this.raw.search), core.String);
    }
    set search(val) {
      return this.raw.search = unwrap_jso(val);
    }
    get username() {
      return dart.as(wrap_jso(this.raw.username), core.String);
    }
    set username(val) {
      return this.raw.username = unwrap_jso(val);
    }
    toString() {
      return dart.as(wrap_jso(this.raw.toString()), core.String);
    }
  }
  dart.setSignature(UrlUtils, {
    constructors: () => ({_: [UrlUtils, []]})
  });
  UrlUtils[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('URLUtils')), dart.const(new _metadata.Experimental())];
  class UrlUtilsReadOnly extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get hash() {
      return dart.as(wrap_jso(this.raw.hash), core.String);
    }
    get host() {
      return dart.as(wrap_jso(this.raw.host), core.String);
    }
    get hostname() {
      return dart.as(wrap_jso(this.raw.hostname), core.String);
    }
    get href() {
      return dart.as(wrap_jso(this.raw.href), core.String);
    }
    get origin() {
      return dart.as(wrap_jso(this.raw.origin), core.String);
    }
    get pathname() {
      return dart.as(wrap_jso(this.raw.pathname), core.String);
    }
    get port() {
      return dart.as(wrap_jso(this.raw.port), core.String);
    }
    get protocol() {
      return dart.as(wrap_jso(this.raw.protocol), core.String);
    }
    get search() {
      return dart.as(wrap_jso(this.raw.search), core.String);
    }
    toString() {
      return dart.as(wrap_jso(this.raw.toString()), core.String);
    }
  }
  dart.setSignature(UrlUtilsReadOnly, {
    constructors: () => ({_: [UrlUtilsReadOnly, []]})
  });
  UrlUtilsReadOnly[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('URLUtilsReadOnly')), dart.const(new _metadata.Experimental())];
  class ValidityState extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateValidityState() {
      return ValidityState._internalWrap();
    }
    static _internalWrap() {
      return new ValidityState.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get badInput() {
      return dart.as(wrap_jso(this.raw.badInput), core.bool);
    }
    get customError() {
      return dart.as(wrap_jso(this.raw.customError), core.bool);
    }
    get patternMismatch() {
      return dart.as(wrap_jso(this.raw.patternMismatch), core.bool);
    }
    get rangeOverflow() {
      return dart.as(wrap_jso(this.raw.rangeOverflow), core.bool);
    }
    get rangeUnderflow() {
      return dart.as(wrap_jso(this.raw.rangeUnderflow), core.bool);
    }
    get stepMismatch() {
      return dart.as(wrap_jso(this.raw.stepMismatch), core.bool);
    }
    get tooLong() {
      return dart.as(wrap_jso(this.raw.tooLong), core.bool);
    }
    get typeMismatch() {
      return dart.as(wrap_jso(this.raw.typeMismatch), core.bool);
    }
    get valid() {
      return dart.as(wrap_jso(this.raw.valid), core.bool);
    }
    get valueMissing() {
      return dart.as(wrap_jso(this.raw.valueMissing), core.bool);
    }
  }
  dart.defineNamedConstructor(ValidityState, 'internal_');
  dart.setSignature(ValidityState, {
    constructors: () => ({
      _: [ValidityState, []],
      _internalWrap: [ValidityState, []],
      internal_: [ValidityState, []]
    }),
    statics: () => ({internalCreateValidityState: [ValidityState, []]}),
    names: ['internalCreateValidityState']
  });
  ValidityState[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ValidityState')), dart.const(new _js_helper.Native("ValidityState"))];
  const VoidCallback = dart.typedef('VoidCallback', () => dart.functionType(dart.void, []));
  const _open2 = Symbol('_open2');
  const _open3 = Symbol('_open3');
  const _location = Symbol('_location');
  const _ensureRequestAnimationFrame = Symbol('_ensureRequestAnimationFrame');
  const _requestAnimationFrame = Symbol('_requestAnimationFrame');
  const _cancelAnimationFrame = Symbol('_cancelAnimationFrame');
  const _requestFileSystem = Symbol('_requestFileSystem');
  const _get_opener = Symbol('_get_opener');
  const _pageXOffset = Symbol('_pageXOffset');
  const _pageYOffset = Symbol('_pageYOffset');
  const _get_parent = Symbol('_get_parent');
  const _get_self = Symbol('_get_self');
  const _get_top = Symbol('_get_top');
  const _alert_1 = Symbol('_alert_1');
  const _alert_2 = Symbol('_alert_2');
  const _confirm_1 = Symbol('_confirm_1');
  const _confirm_2 = Symbol('_confirm_2');
  const _find_1 = Symbol('_find_1');
  const _getComputedStyle_1 = Symbol('_getComputedStyle_1');
  const _getMatchedCssRules_1 = Symbol('_getMatchedCssRules_1');
  const _matchMedia_1 = Symbol('_matchMedia_1');
  const _moveBy_1 = Symbol('_moveBy_1');
  const _moveTo = Symbol('_moveTo');
  const _openDatabase_1 = Symbol('_openDatabase_1');
  const _openDatabase_2 = Symbol('_openDatabase_2');
  const _print_1 = Symbol('_print_1');
  const _resizeBy_1 = Symbol('_resizeBy_1');
  const _resizeTo_1 = Symbol('_resizeTo_1');
  const _scroll_1 = Symbol('_scroll_1');
  const _scroll_2 = Symbol('_scroll_2');
  const _scroll_3 = Symbol('_scroll_3');
  const _scroll_4 = Symbol('_scroll_4');
  const _scrollBy_1 = Symbol('_scrollBy_1');
  const _scrollBy_2 = Symbol('_scrollBy_2');
  const _scrollBy_3 = Symbol('_scrollBy_3');
  const _scrollBy_4 = Symbol('_scrollBy_4');
  const _scrollTo_1 = Symbol('_scrollTo_1');
  const _scrollTo_2 = Symbol('_scrollTo_2');
  const _scrollTo_3 = Symbol('_scrollTo_3');
  const _scrollTo_4 = Symbol('_scrollTo_4');
  const _showModalDialog_1 = Symbol('_showModalDialog_1');
  const _showModalDialog_2 = Symbol('_showModalDialog_2');
  const _showModalDialog_3 = Symbol('_showModalDialog_3');
  const __requestFileSystem_1 = Symbol('__requestFileSystem_1');
  const __requestFileSystem_2 = Symbol('__requestFileSystem_2');
  const __requestFileSystem = Symbol('__requestFileSystem');
  const _resolveLocalFileSystemUrl_1 = Symbol('_resolveLocalFileSystemUrl_1');
  const _resolveLocalFileSystemUrl_2 = Symbol('_resolveLocalFileSystemUrl_2');
  const _resolveLocalFileSystemUrl = Symbol('_resolveLocalFileSystemUrl');
  class Window extends EventTarget {
    get animationFrame() {
      let completer = async.Completer$(core.num).sync();
      this.requestAnimationFrame(dart.fn(time => {
        completer.complete(time);
      }));
      return completer.future;
    }
    get document() {
      return dart.as(wrap_jso(this.raw.document), Document);
    }
    [_open2](url, name) {
      return dart.as(wrap_jso(this.raw.open(url, name)), WindowBase);
    }
    [_open3](url, name, options) {
      return dart.as(wrap_jso(this.raw.open(url, name, options)), WindowBase);
    }
    open(url, name, options) {
      if (options === void 0)
        options = null;
      if (options == null) {
        return _DOMWindowCrossFrame._createSafe(this[_open2](url, name));
      } else {
        return _DOMWindowCrossFrame._createSafe(this[_open3](url, name, options));
      }
    }
    get location() {
      return dart.as(this[_location], Location);
    }
    set location(value) {
      this[_location] = value;
    }
    get [_location]() {
      return wrap_jso(this.raw.location);
    }
    set [_location](value) {
      this.raw.location = unwrap_jso(value);
    }
    requestAnimationFrame(callback) {
      this[_ensureRequestAnimationFrame]();
      return this[_requestAnimationFrame](dart.as(_wrapZone(callback), RequestAnimationFrameCallback));
    }
    cancelAnimationFrame(id) {
      this[_ensureRequestAnimationFrame]();
      this[_cancelAnimationFrame](id);
    }
    [_requestAnimationFrame](callback) {
      return this.raw.requestAnimationFrame;
    }
    [_cancelAnimationFrame](id) {
      this.raw.cancelAnimationFrame(id);
    }
    [_ensureRequestAnimationFrame]() {
      if (!!(this.raw.requestAnimationFrame && this.raw.cancelAnimationFrame))
        return;
      (function($this) {
        var vendors = ['ms', 'moz', 'webkit', 'o'];
        for (var i = 0; i < vendors.length && !$this.requestAnimationFrame; ++i) {
          $this.requestAnimationFrame = $this[vendors[i] + 'RequestAnimationFrame'];
          $this.cancelAnimationFrame = $this[vendors[i] + 'CancelAnimationFrame'] || $this[vendors[i] + 'CancelRequestAnimationFrame'];
        }
        if ($this.requestAnimationFrame && $this.cancelAnimationFrame)
          return;
        $this.requestAnimationFrame = function(callback) {
          return window.setTimeout(function() {
            callback(Date.now());
          }, 16);
        };
        $this.cancelAnimationFrame = function(id) {
          clearTimeout(id);
        };
      })(this.raw);
    }
    get indexedDB() {
      return wrap_jso(this.indexedDB || this.webkitIndexedDB || this.mozIndexedDB);
    }
    get console() {
      return Console._safeConsole;
    }
    requestFileSystem(size, opts) {
      let persistent = opts && 'persistent' in opts ? opts.persistent : false;
      return this[_requestFileSystem](dart.notNull(persistent) ? 1 : 0, size);
    }
    static get supportsPointConversions() {
      return dart.as(dart.dload(/* Unimplemented unknown name */DomPoint, 'supported'), core.bool);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateWindow() {
      return Window._internalWrap();
    }
    static _internalWrap() {
      return new Window.internal_();
    }
    internal_() {
      super.internal_();
    }
    get css() {
      return wrap_jso(this.raw.CSS);
    }
    get applicationCache() {
      return wrap_jso(this.raw.applicationCache);
    }
    get closed() {
      return dart.as(wrap_jso(this.raw.closed), core.bool);
    }
    get crypto() {
      return wrap_jso(this.raw.crypto);
    }
    get defaultStatus() {
      return dart.as(wrap_jso(this.raw.defaultStatus), core.String);
    }
    set defaultStatus(val) {
      return this.raw.defaultStatus = unwrap_jso(val);
    }
    get defaultstatus() {
      return dart.as(wrap_jso(this.raw.defaultstatus), core.String);
    }
    set defaultstatus(val) {
      return this.raw.defaultstatus = unwrap_jso(val);
    }
    get devicePixelRatio() {
      return dart.as(wrap_jso(this.raw.devicePixelRatio), core.double);
    }
    get history() {
      return dart.as(wrap_jso(this.raw.history), History);
    }
    get innerHeight() {
      return dart.as(wrap_jso(this.raw.innerHeight), core.int);
    }
    get innerWidth() {
      return dart.as(wrap_jso(this.raw.innerWidth), core.int);
    }
    get localStorage() {
      return dart.as(wrap_jso(this.raw.localStorage), Storage);
    }
    get locationbar() {
      return wrap_jso(this.raw.locationbar);
    }
    get menubar() {
      return wrap_jso(this.raw.menubar);
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    set name(val) {
      return this.raw.name = unwrap_jso(val);
    }
    get navigator() {
      return dart.as(wrap_jso(this.raw.navigator), Navigator);
    }
    get offscreenBuffering() {
      return dart.as(wrap_jso(this.raw.offscreenBuffering), core.bool);
    }
    get opener() {
      return _convertNativeToDart_Window(this[_get_opener]);
    }
    get [_get_opener]() {
      return wrap_jso(this.raw.opener);
    }
    set opener(value) {
      this.raw.opener = unwrap_jso(value);
    }
    get orientation() {
      return dart.as(wrap_jso(this.raw.orientation), core.int);
    }
    get outerHeight() {
      return dart.as(wrap_jso(this.raw.outerHeight), core.int);
    }
    get outerWidth() {
      return dart.as(wrap_jso(this.raw.outerWidth), core.int);
    }
    get [_pageXOffset]() {
      return dart.as(wrap_jso(this.raw.pageXOffset), core.double);
    }
    get [_pageYOffset]() {
      return dart.as(wrap_jso(this.raw.pageYOffset), core.double);
    }
    get parent() {
      return _convertNativeToDart_Window(this[_get_parent]);
    }
    get [_get_parent]() {
      return wrap_jso(this.raw.parent);
    }
    get performance() {
      return dart.as(wrap_jso(this.raw.performance), Performance);
    }
    get screen() {
      return dart.as(wrap_jso(this.raw.screen), Screen);
    }
    get screenLeft() {
      return dart.as(wrap_jso(this.raw.screenLeft), core.int);
    }
    get screenTop() {
      return dart.as(wrap_jso(this.raw.screenTop), core.int);
    }
    get screenX() {
      return dart.as(wrap_jso(this.raw.screenX), core.int);
    }
    get screenY() {
      return dart.as(wrap_jso(this.raw.screenY), core.int);
    }
    get scrollbars() {
      return wrap_jso(this.raw.scrollbars);
    }
    get self() {
      return _convertNativeToDart_Window(this[_get_self]);
    }
    get [_get_self]() {
      return wrap_jso(this.raw.self);
    }
    get sessionStorage() {
      return dart.as(wrap_jso(this.raw.sessionStorage), Storage);
    }
    get speechSynthesis() {
      return dart.as(wrap_jso(this.raw.speechSynthesis), SpeechSynthesis);
    }
    get status() {
      return dart.as(wrap_jso(this.raw.status), core.String);
    }
    set status(val) {
      return this.raw.status = unwrap_jso(val);
    }
    get statusbar() {
      return wrap_jso(this.raw.statusbar);
    }
    get styleMedia() {
      return dart.as(wrap_jso(this.raw.styleMedia), StyleMedia);
    }
    get toolbar() {
      return wrap_jso(this.raw.toolbar);
    }
    get top() {
      return _convertNativeToDart_Window(this[_get_top]);
    }
    get [_get_top]() {
      return wrap_jso(this.raw.top);
    }
    get window() {
      return _convertNativeToDart_Window(this[_get_window]);
    }
    get [_get_window]() {
      return wrap_jso(this.raw.window);
    }
    [__getter__](index_OR_name) {
      if (typeof index_OR_name == 'number') {
        return _convertNativeToDart_Window(this[__getter___1](dart.as(index_OR_name, core.int)));
      }
      if (typeof index_OR_name == 'string') {
        return _convertNativeToDart_Window(this[__getter___2](dart.as(index_OR_name, core.String)));
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [__getter___1](index) {
      return wrap_jso(this.raw.__getter__(unwrap_jso(index)));
    }
    [__getter___2](name) {
      return wrap_jso(this.raw.__getter__(unwrap_jso(name)));
    }
    alert(message) {
      if (message === void 0)
        message = null;
      if (message != null) {
        this[_alert_1](message);
        return;
      }
      this[_alert_2]();
      return;
    }
    [_alert_1](message) {
      return wrap_jso(this.raw.alert(unwrap_jso(message)));
    }
    [_alert_2]() {
      return wrap_jso(this.raw.alert());
    }
    close() {
      this[_close_1]();
      return;
    }
    [_close_1]() {
      return wrap_jso(this.raw.close());
    }
    confirm(message) {
      if (message === void 0)
        message = null;
      if (message != null) {
        return this[_confirm_1](message);
      }
      return this[_confirm_2]();
    }
    [_confirm_1](message) {
      return dart.as(wrap_jso(this.raw.confirm(unwrap_jso(message))), core.bool);
    }
    [_confirm_2]() {
      return dart.as(wrap_jso(this.raw.confirm()), core.bool);
    }
    find(string, caseSensitive, backwards, wrap, wholeWord, searchInFrames, showDialog) {
      return this[_find_1](string, caseSensitive, backwards, wrap, wholeWord, searchInFrames, showDialog);
    }
    [_find_1](string, caseSensitive, backwards, wrap, wholeWord, searchInFrames, showDialog) {
      return dart.as(wrap_jso(this.raw.find(unwrap_jso(string), unwrap_jso(caseSensitive), unwrap_jso(backwards), unwrap_jso(wrap), unwrap_jso(wholeWord), unwrap_jso(searchInFrames), unwrap_jso(showDialog))), core.bool);
    }
    [_getComputedStyle](element, pseudoElement) {
      return this[_getComputedStyle_1](element, pseudoElement);
    }
    [_getComputedStyle_1](element, pseudoElement) {
      return dart.as(wrap_jso(this.raw.getComputedStyle(unwrap_jso(element), unwrap_jso(pseudoElement))), CssStyleDeclaration);
    }
    getMatchedCssRules(element, pseudoElement) {
      return this[_getMatchedCssRules_1](element, pseudoElement);
    }
    [_getMatchedCssRules_1](element, pseudoElement) {
      return dart.as(wrap_jso(this.raw.getMatchedCSSRules(unwrap_jso(element), unwrap_jso(pseudoElement))), _CssRuleList);
    }
    getSelection() {
      return this[_getSelection_1]();
    }
    [_getSelection_1]() {
      return dart.as(wrap_jso(this.raw.getSelection()), Selection);
    }
    matchMedia(query) {
      return this[_matchMedia_1](query);
    }
    [_matchMedia_1](query) {
      return wrap_jso(this.raw.matchMedia(unwrap_jso(query)));
    }
    moveBy(x, y) {
      this[_moveBy_1](x, y);
      return;
    }
    [_moveBy_1](x, y) {
      return wrap_jso(this.raw.moveBy(unwrap_jso(x), unwrap_jso(y)));
    }
    [_moveTo](x, y) {
      this[_moveTo_1](x, y);
      return;
    }
    [_moveTo_1](x, y) {
      return wrap_jso(this.raw.moveTo(unwrap_jso(x), unwrap_jso(y)));
    }
    openDatabase(name, version, displayName, estimatedSize, creationCallback) {
      if (creationCallback === void 0)
        creationCallback = null;
      if (creationCallback != null) {
        return this[_openDatabase_1](name, version, displayName, estimatedSize, creationCallback);
      }
      return this[_openDatabase_2](name, version, displayName, estimatedSize);
    }
    [_openDatabase_1](name, version, displayName, estimatedSize, creationCallback) {
      return wrap_jso(this.raw.openDatabase(unwrap_jso(name), unwrap_jso(version), unwrap_jso(displayName), unwrap_jso(estimatedSize), unwrap_jso(creationCallback)));
    }
    [_openDatabase_2](name, version, displayName, estimatedSize) {
      return wrap_jso(this.raw.openDatabase(unwrap_jso(name), unwrap_jso(version), unwrap_jso(displayName), unwrap_jso(estimatedSize)));
    }
    postMessage(message, targetOrigin, transfer) {
      if (transfer === void 0)
        transfer = null;
      if (transfer != null) {
        let message_1 = html_common.convertDartToNative_SerializedScriptValue(message);
        this[_postMessage_1](message_1, targetOrigin, transfer);
        return;
      }
      let message_1 = html_common.convertDartToNative_SerializedScriptValue(message);
      this[_postMessage_2](message_1, targetOrigin);
      return;
    }
    [_postMessage_1](message, targetOrigin, transfer) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message), unwrap_jso(targetOrigin), unwrap_jso(transfer)));
    }
    [_postMessage_2](message, targetOrigin) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message), unwrap_jso(targetOrigin)));
    }
    print() {
      this[_print_1]();
      return;
    }
    [_print_1]() {
      return wrap_jso(this.raw.print());
    }
    resizeBy(x, y) {
      this[_resizeBy_1](x, y);
      return;
    }
    [_resizeBy_1](x, y) {
      return wrap_jso(this.raw.resizeBy(unwrap_jso(x), unwrap_jso(y)));
    }
    resizeTo(width, height) {
      this[_resizeTo_1](width, height);
      return;
    }
    [_resizeTo_1](width, height) {
      return wrap_jso(this.raw.resizeTo(unwrap_jso(width), unwrap_jso(height)));
    }
    scroll(x, y, scrollOptions) {
      if (scrollOptions === void 0)
        scrollOptions = null;
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scroll_1](dart.as(x, core.num), dart.as(y, core.num));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scroll_2](dart.as(x, core.num), dart.as(y, core.num), scrollOptions_1);
        return;
      }
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scroll_3](dart.as(x, core.int), dart.as(y, core.int));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scroll_4](dart.as(x, core.int), dart.as(y, core.int), scrollOptions_1);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_scroll_1](x, y) {
      return wrap_jso(this.raw.scroll(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scroll_2](x, y, scrollOptions) {
      return wrap_jso(this.raw.scroll(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    [_scroll_3](x, y) {
      return wrap_jso(this.raw.scroll(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scroll_4](x, y, scrollOptions) {
      return wrap_jso(this.raw.scroll(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    scrollBy(x, y, scrollOptions) {
      if (scrollOptions === void 0)
        scrollOptions = null;
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scrollBy_1](dart.as(x, core.num), dart.as(y, core.num));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scrollBy_2](dart.as(x, core.num), dart.as(y, core.num), scrollOptions_1);
        return;
      }
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scrollBy_3](dart.as(x, core.int), dart.as(y, core.int));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scrollBy_4](dart.as(x, core.int), dart.as(y, core.int), scrollOptions_1);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_scrollBy_1](x, y) {
      return wrap_jso(this.raw.scrollBy(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scrollBy_2](x, y, scrollOptions) {
      return wrap_jso(this.raw.scrollBy(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    [_scrollBy_3](x, y) {
      return wrap_jso(this.raw.scrollBy(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scrollBy_4](x, y, scrollOptions) {
      return wrap_jso(this.raw.scrollBy(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    scrollTo(x, y, scrollOptions) {
      if (scrollOptions === void 0)
        scrollOptions = null;
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scrollTo_1](dart.as(x, core.num), dart.as(y, core.num));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scrollTo_2](dart.as(x, core.num), dart.as(y, core.num), scrollOptions_1);
        return;
      }
      if (typeof y == 'number' && typeof x == 'number' && scrollOptions == null) {
        this[_scrollTo_3](dart.as(x, core.int), dart.as(y, core.int));
        return;
      }
      if (scrollOptions != null && typeof y == 'number' && typeof x == 'number') {
        let scrollOptions_1 = html_common.convertDartToNative_Dictionary(scrollOptions);
        this[_scrollTo_4](dart.as(x, core.int), dart.as(y, core.int), scrollOptions_1);
        return;
      }
      dart.throw(new core.ArgumentError("Incorrect number or type of arguments"));
    }
    [_scrollTo_1](x, y) {
      return wrap_jso(this.raw.scrollTo(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scrollTo_2](x, y, scrollOptions) {
      return wrap_jso(this.raw.scrollTo(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    [_scrollTo_3](x, y) {
      return wrap_jso(this.raw.scrollTo(unwrap_jso(x), unwrap_jso(y)));
    }
    [_scrollTo_4](x, y, scrollOptions) {
      return wrap_jso(this.raw.scrollTo(unwrap_jso(x), unwrap_jso(y), unwrap_jso(scrollOptions)));
    }
    showModalDialog(url, dialogArgs, featureArgs) {
      if (dialogArgs === void 0)
        dialogArgs = null;
      if (featureArgs === void 0)
        featureArgs = null;
      if (featureArgs != null) {
        return this[_showModalDialog_1](url, dialogArgs, featureArgs);
      }
      if (dialogArgs != null) {
        return this[_showModalDialog_2](url, dialogArgs);
      }
      return this[_showModalDialog_3](url);
    }
    [_showModalDialog_1](url, dialogArgs, featureArgs) {
      return wrap_jso(this.raw.showModalDialog(unwrap_jso(url), unwrap_jso(dialogArgs), unwrap_jso(featureArgs)));
    }
    [_showModalDialog_2](url, dialogArgs) {
      return wrap_jso(this.raw.showModalDialog(unwrap_jso(url), unwrap_jso(dialogArgs)));
    }
    [_showModalDialog_3](url) {
      return wrap_jso(this.raw.showModalDialog(unwrap_jso(url)));
    }
    stop() {
      this[_stop_1]();
      return;
    }
    [_stop_1]() {
      return wrap_jso(this.raw.stop());
    }
    [__requestFileSystem](type, size, successCallback, errorCallback) {
      if (errorCallback === void 0)
        errorCallback = null;
      if (errorCallback != null) {
        this[__requestFileSystem_1](type, size, successCallback, errorCallback);
        return;
      }
      this[__requestFileSystem_2](type, size, successCallback);
      return;
    }
    [__requestFileSystem_1](type, size, successCallback, errorCallback) {
      return wrap_jso(this.raw.webkitRequestFileSystem(unwrap_jso(type), unwrap_jso(size), unwrap_jso(successCallback), unwrap_jso(errorCallback)));
    }
    [__requestFileSystem_2](type, size, successCallback) {
      return wrap_jso(this.raw.webkitRequestFileSystem(unwrap_jso(type), unwrap_jso(size), unwrap_jso(successCallback)));
    }
    [_requestFileSystem](type, size) {
      let completer = async.Completer$(dart.dynamic).new();
      this[__requestFileSystem](type, size, dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    [_resolveLocalFileSystemUrl](url, successCallback, errorCallback) {
      if (errorCallback === void 0)
        errorCallback = null;
      if (errorCallback != null) {
        this[_resolveLocalFileSystemUrl_1](url, successCallback, errorCallback);
        return;
      }
      this[_resolveLocalFileSystemUrl_2](url, successCallback);
      return;
    }
    [_resolveLocalFileSystemUrl_1](url, successCallback, errorCallback) {
      return wrap_jso(this.raw.webkitResolveLocalFileSystemURL(unwrap_jso(url), unwrap_jso(successCallback), unwrap_jso(errorCallback)));
    }
    [_resolveLocalFileSystemUrl_2](url, successCallback) {
      return wrap_jso(this.raw.webkitResolveLocalFileSystemURL(unwrap_jso(url), unwrap_jso(successCallback)));
    }
    resolveLocalFileSystemUrl(url) {
      let completer = async.Completer$(dart.dynamic).new();
      this[_resolveLocalFileSystemUrl](url, dart.fn(value => {
        completer.complete(value);
      }), dart.fn(error => {
        completer.completeError(error);
      }));
      return completer.future;
    }
    atob(string) {
      return this[_atob_1](string);
    }
    [_atob_1](string) {
      return dart.as(wrap_jso(this.raw.atob(unwrap_jso(string))), core.String);
    }
    btoa(string) {
      return this[_btoa_1](string);
    }
    [_btoa_1](string) {
      return dart.as(wrap_jso(this.raw.btoa(unwrap_jso(string))), core.String);
    }
    [_clearInterval](handle) {
      this[_clearInterval_1](handle);
      return;
    }
    [_clearInterval_1](handle) {
      return wrap_jso(this.raw.clearInterval(unwrap_jso(handle)));
    }
    [_clearTimeout](handle) {
      this[_clearTimeout_1](handle);
      return;
    }
    [_clearTimeout_1](handle) {
      return wrap_jso(this.raw.clearTimeout(unwrap_jso(handle)));
    }
    [_setInterval](handler, timeout) {
      return this[_setInterval_1](handler, timeout);
    }
    [_setInterval_1](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setInterval(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
    [_setTimeout](handler, timeout) {
      return this[_setTimeout_1](handler, timeout);
    }
    [_setTimeout_1](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setTimeout(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
    get onContentLoaded() {
      return Window.contentLoadedEvent.forTarget(this);
    }
    get onAbort() {
      return Element.abortEvent.forTarget(this);
    }
    get onBlur() {
      return Element.blurEvent.forTarget(this);
    }
    get onCanPlay() {
      return Element.canPlayEvent.forTarget(this);
    }
    get onCanPlayThrough() {
      return Element.canPlayThroughEvent.forTarget(this);
    }
    get onChange() {
      return Element.changeEvent.forTarget(this);
    }
    get onClick() {
      return Element.clickEvent.forTarget(this);
    }
    get onContextMenu() {
      return Element.contextMenuEvent.forTarget(this);
    }
    get onDoubleClick() {
      return Element.doubleClickEvent.forTarget(this);
    }
    get onDeviceMotion() {
      return Window.deviceMotionEvent.forTarget(this);
    }
    get onDeviceOrientation() {
      return Window.deviceOrientationEvent.forTarget(this);
    }
    get onDrag() {
      return Element.dragEvent.forTarget(this);
    }
    get onDragEnd() {
      return Element.dragEndEvent.forTarget(this);
    }
    get onDragEnter() {
      return Element.dragEnterEvent.forTarget(this);
    }
    get onDragLeave() {
      return Element.dragLeaveEvent.forTarget(this);
    }
    get onDragOver() {
      return Element.dragOverEvent.forTarget(this);
    }
    get onDragStart() {
      return Element.dragStartEvent.forTarget(this);
    }
    get onDrop() {
      return Element.dropEvent.forTarget(this);
    }
    get onDurationChange() {
      return Element.durationChangeEvent.forTarget(this);
    }
    get onEmptied() {
      return Element.emptiedEvent.forTarget(this);
    }
    get onEnded() {
      return Element.endedEvent.forTarget(this);
    }
    get onError() {
      return Element.errorEvent.forTarget(this);
    }
    get onFocus() {
      return Element.focusEvent.forTarget(this);
    }
    get onHashChange() {
      return Window.hashChangeEvent.forTarget(this);
    }
    get onInput() {
      return Element.inputEvent.forTarget(this);
    }
    get onInvalid() {
      return Element.invalidEvent.forTarget(this);
    }
    get onKeyDown() {
      return Element.keyDownEvent.forTarget(this);
    }
    get onKeyPress() {
      return Element.keyPressEvent.forTarget(this);
    }
    get onKeyUp() {
      return Element.keyUpEvent.forTarget(this);
    }
    get onLoad() {
      return Element.loadEvent.forTarget(this);
    }
    get onLoadedData() {
      return Element.loadedDataEvent.forTarget(this);
    }
    get onLoadedMetadata() {
      return Element.loadedMetadataEvent.forTarget(this);
    }
    get onLoadStart() {
      return Window.loadStartEvent.forTarget(this);
    }
    get onMessage() {
      return Window.messageEvent.forTarget(this);
    }
    get onMouseDown() {
      return Element.mouseDownEvent.forTarget(this);
    }
    get onMouseEnter() {
      return Element.mouseEnterEvent.forTarget(this);
    }
    get onMouseLeave() {
      return Element.mouseLeaveEvent.forTarget(this);
    }
    get onMouseMove() {
      return Element.mouseMoveEvent.forTarget(this);
    }
    get onMouseOut() {
      return Element.mouseOutEvent.forTarget(this);
    }
    get onMouseOver() {
      return Element.mouseOverEvent.forTarget(this);
    }
    get onMouseUp() {
      return Element.mouseUpEvent.forTarget(this);
    }
    get onMouseWheel() {
      return dart.as(dart.dsend(dart.dload(Element, 'mouseWheelEvent'), 'forTarget', this), async.Stream$(dart.dynamic));
    }
    get onOffline() {
      return Window.offlineEvent.forTarget(this);
    }
    get onOnline() {
      return Window.onlineEvent.forTarget(this);
    }
    get onPageHide() {
      return Window.pageHideEvent.forTarget(this);
    }
    get onPageShow() {
      return Window.pageShowEvent.forTarget(this);
    }
    get onPause() {
      return Element.pauseEvent.forTarget(this);
    }
    get onPlay() {
      return Element.playEvent.forTarget(this);
    }
    get onPlaying() {
      return Element.playingEvent.forTarget(this);
    }
    get onPopState() {
      return Window.popStateEvent.forTarget(this);
    }
    get onProgress() {
      return Window.progressEvent.forTarget(this);
    }
    get onRateChange() {
      return Element.rateChangeEvent.forTarget(this);
    }
    get onReset() {
      return Element.resetEvent.forTarget(this);
    }
    get onResize() {
      return Element.resizeEvent.forTarget(this);
    }
    get onScroll() {
      return Element.scrollEvent.forTarget(this);
    }
    get onSearch() {
      return Element.searchEvent.forTarget(this);
    }
    get onSeeked() {
      return Element.seekedEvent.forTarget(this);
    }
    get onSeeking() {
      return Element.seekingEvent.forTarget(this);
    }
    get onSelect() {
      return Element.selectEvent.forTarget(this);
    }
    get onStalled() {
      return Element.stalledEvent.forTarget(this);
    }
    get onStorage() {
      return Window.storageEvent.forTarget(this);
    }
    get onSubmit() {
      return Element.submitEvent.forTarget(this);
    }
    get onSuspend() {
      return Element.suspendEvent.forTarget(this);
    }
    get onTimeUpdate() {
      return Element.timeUpdateEvent.forTarget(this);
    }
    get onTouchCancel() {
      return Element.touchCancelEvent.forTarget(this);
    }
    get onTouchEnd() {
      return Element.touchEndEvent.forTarget(this);
    }
    get onTouchMove() {
      return Element.touchMoveEvent.forTarget(this);
    }
    get onTouchStart() {
      return Element.touchStartEvent.forTarget(this);
    }
    get onTransitionEnd() {
      return dart.as(dart.dsend(dart.dload(Element, 'transitionEndEvent'), 'forTarget', this), async.Stream$(TransitionEvent));
    }
    get onUnload() {
      return Window.unloadEvent.forTarget(this);
    }
    get onVolumeChange() {
      return Element.volumeChangeEvent.forTarget(this);
    }
    get onWaiting() {
      return Element.waitingEvent.forTarget(this);
    }
    get onAnimationEnd() {
      return Window.animationEndEvent.forTarget(this);
    }
    get onAnimationIteration() {
      return Window.animationIterationEvent.forTarget(this);
    }
    get onAnimationStart() {
      return Window.animationStartEvent.forTarget(this);
    }
    get onBeforeUnload() {
      return dart.as(Window.beforeUnloadEvent.forTarget(this), async.Stream$(Event));
    }
    moveTo(p) {
      this[_moveTo](dart.as(p.x, core.num), dart.as(p.y, core.num));
    }
    get pageXOffset() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkWindow'), 'instance'), 'pageXOffset_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get pageYOffset() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkWindow'), 'instance'), 'pageYOffset_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get scrollX() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkWindow'), 'instance'), 'scrollX_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
    get scrollY() {
      return dart.as(dart.dsend(dart.dsend(dart.dload(dart.dload(/* Unimplemented unknown name */_blink, 'BlinkWindow'), 'instance'), 'scrollY_Getter_', unwrap_jso(this)), 'round'), core.int);
    }
  }
  Window[dart.implements] = () => [WindowEventHandlers, WindowBase, _WindowTimers, WindowBase64];
  dart.defineNamedConstructor(Window, 'internal_');
  dart.setSignature(Window, {
    constructors: () => ({
      _: [Window, []],
      _internalWrap: [Window, []],
      internal_: [Window, []]
    }),
    methods: () => ({
      [_open2]: [WindowBase, [dart.dynamic, dart.dynamic]],
      [_open3]: [WindowBase, [dart.dynamic, dart.dynamic, dart.dynamic]],
      open: [WindowBase, [core.String, core.String], [core.String]],
      requestAnimationFrame: [core.int, [RequestAnimationFrameCallback]],
      cancelAnimationFrame: [dart.void, [core.int]],
      [_requestAnimationFrame]: [core.int, [RequestAnimationFrameCallback]],
      [_cancelAnimationFrame]: [dart.void, [core.int]],
      [_ensureRequestAnimationFrame]: [dart.dynamic, []],
      requestFileSystem: [async.Future$(dart.dynamic), [core.int], {persistent: core.bool}],
      [__getter__]: [WindowBase, [dart.dynamic]],
      [__getter___1]: [dart.dynamic, [core.int]],
      [__getter___2]: [dart.dynamic, [core.String]],
      alert: [dart.void, [], [core.String]],
      [_alert_1]: [dart.void, [dart.dynamic]],
      [_alert_2]: [dart.void, []],
      close: [dart.void, []],
      [_close_1]: [dart.void, []],
      confirm: [core.bool, [], [core.String]],
      [_confirm_1]: [core.bool, [dart.dynamic]],
      [_confirm_2]: [core.bool, []],
      find: [core.bool, [core.String, core.bool, core.bool, core.bool, core.bool, core.bool, core.bool]],
      [_find_1]: [core.bool, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      [_getComputedStyle]: [CssStyleDeclaration, [Element, core.String]],
      [_getComputedStyle_1]: [CssStyleDeclaration, [Element, dart.dynamic]],
      getMatchedCssRules: [_CssRuleList, [Element, core.String]],
      [_getMatchedCssRules_1]: [_CssRuleList, [Element, dart.dynamic]],
      getSelection: [Selection, []],
      [_getSelection_1]: [Selection, []],
      matchMedia: [dart.dynamic, [core.String]],
      [_matchMedia_1]: [dart.dynamic, [dart.dynamic]],
      moveBy: [dart.void, [core.num, core.num]],
      [_moveBy_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      [_moveTo]: [dart.void, [core.num, core.num]],
      [_moveTo_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      openDatabase: [dart.dynamic, [core.String, core.String, core.String, core.int], [dart.dynamic]],
      [_openDatabase_1]: [dart.dynamic, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      [_openDatabase_2]: [dart.dynamic, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      postMessage: [dart.void, [dart.dynamic, core.String], [core.List$(MessagePort)]],
      [_postMessage_1]: [dart.void, [dart.dynamic, dart.dynamic, core.List$(MessagePort)]],
      [_postMessage_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      print: [dart.void, []],
      [_print_1]: [dart.void, []],
      resizeBy: [dart.void, [core.num, core.num]],
      [_resizeBy_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      resizeTo: [dart.void, [core.num, core.num]],
      [_resizeTo_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      scroll: [dart.void, [dart.dynamic, dart.dynamic], [core.Map]],
      [_scroll_1]: [dart.void, [core.num, core.num]],
      [_scroll_2]: [dart.void, [core.num, core.num, dart.dynamic]],
      [_scroll_3]: [dart.void, [core.int, core.int]],
      [_scroll_4]: [dart.void, [core.int, core.int, dart.dynamic]],
      scrollBy: [dart.void, [dart.dynamic, dart.dynamic], [core.Map]],
      [_scrollBy_1]: [dart.void, [core.num, core.num]],
      [_scrollBy_2]: [dart.void, [core.num, core.num, dart.dynamic]],
      [_scrollBy_3]: [dart.void, [core.int, core.int]],
      [_scrollBy_4]: [dart.void, [core.int, core.int, dart.dynamic]],
      scrollTo: [dart.void, [dart.dynamic, dart.dynamic], [core.Map]],
      [_scrollTo_1]: [dart.void, [core.num, core.num]],
      [_scrollTo_2]: [dart.void, [core.num, core.num, dart.dynamic]],
      [_scrollTo_3]: [dart.void, [core.int, core.int]],
      [_scrollTo_4]: [dart.void, [core.int, core.int, dart.dynamic]],
      showModalDialog: [core.Object, [core.String], [core.Object, core.String]],
      [_showModalDialog_1]: [core.Object, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_showModalDialog_2]: [core.Object, [dart.dynamic, dart.dynamic]],
      [_showModalDialog_3]: [core.Object, [dart.dynamic]],
      stop: [dart.void, []],
      [_stop_1]: [dart.void, []],
      [__requestFileSystem]: [dart.void, [core.int, core.int, dart.dynamic], [dart.dynamic]],
      [__requestFileSystem_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]],
      [__requestFileSystem_2]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_requestFileSystem]: [async.Future$(dart.dynamic), [core.int, core.int]],
      [_resolveLocalFileSystemUrl]: [dart.void, [core.String, dart.dynamic], [dart.dynamic]],
      [_resolveLocalFileSystemUrl_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      [_resolveLocalFileSystemUrl_2]: [dart.void, [dart.dynamic, dart.dynamic]],
      resolveLocalFileSystemUrl: [async.Future$(dart.dynamic), [core.String]],
      atob: [core.String, [core.String]],
      [_atob_1]: [core.String, [dart.dynamic]],
      btoa: [core.String, [core.String]],
      [_btoa_1]: [core.String, [dart.dynamic]],
      [_clearInterval]: [dart.void, [core.int]],
      [_clearInterval_1]: [dart.void, [dart.dynamic]],
      [_clearTimeout]: [dart.void, [core.int]],
      [_clearTimeout_1]: [dart.void, [dart.dynamic]],
      [_setInterval]: [core.int, [core.Object, core.int]],
      [_setInterval_1]: [core.int, [dart.dynamic, dart.dynamic]],
      [_setTimeout]: [core.int, [core.Object, core.int]],
      [_setTimeout_1]: [core.int, [dart.dynamic, dart.dynamic]],
      moveTo: [dart.void, [math.Point]]
    }),
    statics: () => ({internalCreateWindow: [Window, []]}),
    names: ['internalCreateWindow']
  });
  Window[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Window')), dart.const(new _js_helper.Native("Window"))];
  Window.contentLoadedEvent = dart.const(new (EventStreamProvider$(Event))('DOMContentLoaded'));
  Window.deviceMotionEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('devicemotion'));
  Window.deviceOrientationEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('deviceorientation'));
  Window.hashChangeEvent = dart.const(new (EventStreamProvider$(Event))('hashchange'));
  Window.loadStartEvent = dart.const(new (EventStreamProvider$(Event))('loadstart'));
  Window.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  Window.offlineEvent = dart.const(new (EventStreamProvider$(Event))('offline'));
  Window.onlineEvent = dart.const(new (EventStreamProvider$(Event))('online'));
  Window.pageHideEvent = dart.const(new (EventStreamProvider$(Event))('pagehide'));
  Window.pageShowEvent = dart.const(new (EventStreamProvider$(Event))('pageshow'));
  Window.popStateEvent = dart.const(new (EventStreamProvider$(PopStateEvent))('popstate'));
  Window.progressEvent = dart.const(new (EventStreamProvider$(Event))('progress'));
  Window.storageEvent = dart.const(new (EventStreamProvider$(StorageEvent))('storage'));
  Window.unloadEvent = dart.const(new (EventStreamProvider$(Event))('unload'));
  Window.animationEndEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('webkitAnimationEnd'));
  Window.animationIterationEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('webkitAnimationIteration'));
  Window.animationStartEvent = dart.const(new (EventStreamProvider$(dart.dynamic))('webkitAnimationStart'));
  Window.PERSISTENT = 1;
  Window.TEMPORARY = 0;
  class _BeforeUnloadEventStreamProvider extends core.Object {
    _BeforeUnloadEventStreamProvider(eventType) {
      this[_eventType] = eventType;
    }
    forTarget(e, opts) {
      let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
      let stream = new _EventStream(e, this[_eventType], useCapture);
      return stream;
    }
    getEventType(target) {
      return this[_eventType];
    }
    forElement(e, opts) {
      let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
      return new _ElementEventStreamImpl(e, this[_eventType], useCapture);
    }
    [_forElementList](e, opts) {
      let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
      return new _ElementListEventStreamImpl(dart.as(e, core.Iterable$(Element)), this[_eventType], useCapture);
    }
  }
  _BeforeUnloadEventStreamProvider[dart.implements] = () => [EventStreamProvider$(dart.dynamic)];
  dart.setSignature(_BeforeUnloadEventStreamProvider, {
    constructors: () => ({_BeforeUnloadEventStreamProvider: [_BeforeUnloadEventStreamProvider, [core.String]]}),
    methods: () => ({
      forTarget: [async.Stream$(dart.dynamic), [EventTarget], {useCapture: core.bool}],
      getEventType: [core.String, [EventTarget]],
      forElement: [ElementStream$(dart.dynamic), [Element], {useCapture: core.bool}],
      [_forElementList]: [ElementStream$(dart.dynamic), [ElementList], {useCapture: core.bool}]
    })
  });
  Window.beforeUnloadEvent = dart.const(new _BeforeUnloadEventStreamProvider('beforeunload'));
  class WindowBase64 extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    atob(string) {
      return dart.as(wrap_jso(this.raw.atob(unwrap_jso(string))), core.String);
    }
    btoa(string) {
      return dart.as(wrap_jso(this.raw.btoa(unwrap_jso(string))), core.String);
    }
  }
  dart.setSignature(WindowBase64, {
    constructors: () => ({_: [WindowBase64, []]}),
    methods: () => ({
      atob: [core.String, [core.String]],
      btoa: [core.String, [core.String]]
    })
  });
  WindowBase64[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WindowBase64')), dart.const(new _metadata.Experimental())];
  class WindowEventHandlers extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    get onHashChange() {
      return WindowEventHandlers.hashChangeEvent.forTarget(this);
    }
    get onMessage() {
      return WindowEventHandlers.messageEvent.forTarget(this);
    }
    get onOffline() {
      return WindowEventHandlers.offlineEvent.forTarget(this);
    }
    get onOnline() {
      return WindowEventHandlers.onlineEvent.forTarget(this);
    }
    get onPopState() {
      return WindowEventHandlers.popStateEvent.forTarget(this);
    }
    get onStorage() {
      return WindowEventHandlers.storageEvent.forTarget(this);
    }
    get onUnload() {
      return WindowEventHandlers.unloadEvent.forTarget(this);
    }
  }
  dart.setSignature(WindowEventHandlers, {
    constructors: () => ({_: [WindowEventHandlers, []]})
  });
  WindowEventHandlers[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WindowEventHandlers')), dart.const(new _metadata.Experimental())];
  WindowEventHandlers.hashChangeEvent = dart.const(new (EventStreamProvider$(Event))('hashchange'));
  WindowEventHandlers.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  WindowEventHandlers.offlineEvent = dart.const(new (EventStreamProvider$(Event))('offline'));
  WindowEventHandlers.onlineEvent = dart.const(new (EventStreamProvider$(Event))('online'));
  WindowEventHandlers.popStateEvent = dart.const(new (EventStreamProvider$(PopStateEvent))('popstate'));
  WindowEventHandlers.storageEvent = dart.const(new (EventStreamProvider$(StorageEvent))('storage'));
  WindowEventHandlers.unloadEvent = dart.const(new (EventStreamProvider$(Event))('unload'));
  const _terminate_1 = Symbol('_terminate_1');
  class Worker extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(scriptUrl) {
      return Worker._create_1(scriptUrl);
    }
    static _create_1(scriptUrl) {
      return dart.as(wrap_jso(new Worker(scriptUrl)), Worker);
    }
    static internalCreateWorker() {
      return Worker._internalWrap();
    }
    static _internalWrap() {
      return new Worker.internal_();
    }
    internal_() {
      super.internal_();
    }
    static get supported() {
      return typeof window.Worker != "undefined";
    }
    postMessage(message, transfer) {
      if (transfer === void 0)
        transfer = null;
      if (transfer != null) {
        this[_postMessage_1](message, transfer);
        return;
      }
      this[_postMessage_2](message);
      return;
    }
    [_postMessage_1](message, transfer) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message), unwrap_jso(transfer)));
    }
    [_postMessage_2](message) {
      return wrap_jso(this.raw.postMessage(unwrap_jso(message)));
    }
    terminate() {
      this[_terminate_1]();
      return;
    }
    [_terminate_1]() {
      return wrap_jso(this.raw.terminate());
    }
    get onError() {
      return Worker.errorEvent.forTarget(this);
    }
    get onMessage() {
      return Worker.messageEvent.forTarget(this);
    }
  }
  dart.defineNamedConstructor(Worker, 'internal_');
  dart.setSignature(Worker, {
    constructors: () => ({
      _: [Worker, []],
      new: [Worker, [core.String]],
      _internalWrap: [Worker, []],
      internal_: [Worker, []]
    }),
    methods: () => ({
      postMessage: [dart.void, [dart.dynamic], [core.List$(MessagePort)]],
      [_postMessage_1]: [dart.void, [dart.dynamic, core.List$(MessagePort)]],
      [_postMessage_2]: [dart.void, [dart.dynamic]],
      terminate: [dart.void, []],
      [_terminate_1]: [dart.void, []]
    }),
    statics: () => ({
      _create_1: [Worker, [dart.dynamic]],
      internalCreateWorker: [Worker, []]
    }),
    names: ['_create_1', 'internalCreateWorker']
  });
  Worker[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Worker')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.IE, '10')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Worker"))];
  Worker.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  Worker.messageEvent = dart.const(new (EventStreamProvider$(MessageEvent))('message'));
  WorkerGlobalScope.errorEvent = dart.const(new (EventStreamProvider$(Event))('error'));
  WorkerGlobalScope.PERSISTENT = 1;
  WorkerGlobalScope.TEMPORARY = 0;
  class WorkerPerformance extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateWorkerPerformance() {
      return WorkerPerformance._internalWrap();
    }
    static _internalWrap() {
      return new WorkerPerformance.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get memory() {
      return wrap_jso(this.raw.memory);
    }
    now() {
      return this[_now_1]();
    }
    [_now_1]() {
      return dart.as(wrap_jso(this.raw.now()), core.num);
    }
  }
  dart.defineNamedConstructor(WorkerPerformance, 'internal_');
  dart.setSignature(WorkerPerformance, {
    constructors: () => ({
      _: [WorkerPerformance, []],
      _internalWrap: [WorkerPerformance, []],
      internal_: [WorkerPerformance, []]
    }),
    methods: () => ({
      now: [core.num, []],
      [_now_1]: [core.num, []]
    }),
    statics: () => ({internalCreateWorkerPerformance: [WorkerPerformance, []]}),
    names: ['internalCreateWorkerPerformance']
  });
  WorkerPerformance[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WorkerPerformance')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("WorkerPerformance"))];
  const _createExpression_1 = Symbol('_createExpression_1');
  const _createNSResolver_1 = Symbol('_createNSResolver_1');
  const _evaluate_1 = Symbol('_evaluate_1');
  class XPathEvaluator extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return XPathEvaluator._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new XPathEvaluator()), XPathEvaluator);
    }
    static internalCreateXPathEvaluator() {
      return XPathEvaluator._internalWrap();
    }
    static _internalWrap() {
      return new XPathEvaluator.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    createExpression(expression, resolver) {
      return this[_createExpression_1](expression, resolver);
    }
    [_createExpression_1](expression, resolver) {
      return dart.as(wrap_jso(this.raw.createExpression(unwrap_jso(expression), unwrap_jso(resolver))), XPathExpression);
    }
    createNSResolver(nodeResolver) {
      return this[_createNSResolver_1](nodeResolver);
    }
    [_createNSResolver_1](nodeResolver) {
      return dart.as(wrap_jso(this.raw.createNSResolver(unwrap_jso(nodeResolver))), XPathNSResolver);
    }
    evaluate(expression, contextNode, resolver, type, inResult) {
      return this[_evaluate_1](expression, contextNode, resolver, type, inResult);
    }
    [_evaluate_1](expression, contextNode, resolver, type, inResult) {
      return dart.as(wrap_jso(this.raw.evaluate(unwrap_jso(expression), unwrap_jso(contextNode), unwrap_jso(resolver), unwrap_jso(type), unwrap_jso(inResult))), XPathResult);
    }
  }
  dart.defineNamedConstructor(XPathEvaluator, 'internal_');
  dart.setSignature(XPathEvaluator, {
    constructors: () => ({
      _: [XPathEvaluator, []],
      new: [XPathEvaluator, []],
      _internalWrap: [XPathEvaluator, []],
      internal_: [XPathEvaluator, []]
    }),
    methods: () => ({
      createExpression: [XPathExpression, [core.String, XPathNSResolver]],
      [_createExpression_1]: [XPathExpression, [dart.dynamic, XPathNSResolver]],
      createNSResolver: [XPathNSResolver, [Node]],
      [_createNSResolver_1]: [XPathNSResolver, [Node]],
      evaluate: [XPathResult, [core.String, Node, XPathNSResolver, core.int, XPathResult]],
      [_evaluate_1]: [XPathResult, [dart.dynamic, Node, XPathNSResolver, dart.dynamic, XPathResult]]
    }),
    statics: () => ({
      _create_1: [XPathEvaluator, []],
      internalCreateXPathEvaluator: [XPathEvaluator, []]
    }),
    names: ['_create_1', 'internalCreateXPathEvaluator']
  });
  XPathEvaluator[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XPathEvaluator')), core.deprecated, dart.const(new _js_helper.Native("XPathEvaluator"))];
  class XPathExpression extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateXPathExpression() {
      return XPathExpression._internalWrap();
    }
    static _internalWrap() {
      return new XPathExpression.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    evaluate(contextNode, type, inResult) {
      return this[_evaluate_1](contextNode, type, inResult);
    }
    [_evaluate_1](contextNode, type, inResult) {
      return dart.as(wrap_jso(this.raw.evaluate(unwrap_jso(contextNode), unwrap_jso(type), unwrap_jso(inResult))), XPathResult);
    }
  }
  dart.defineNamedConstructor(XPathExpression, 'internal_');
  dart.setSignature(XPathExpression, {
    constructors: () => ({
      _: [XPathExpression, []],
      _internalWrap: [XPathExpression, []],
      internal_: [XPathExpression, []]
    }),
    methods: () => ({
      evaluate: [XPathResult, [Node, core.int, XPathResult]],
      [_evaluate_1]: [XPathResult, [Node, dart.dynamic, XPathResult]]
    }),
    statics: () => ({internalCreateXPathExpression: [XPathExpression, []]}),
    names: ['internalCreateXPathExpression']
  });
  XPathExpression[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XPathExpression')), core.deprecated, dart.const(new _js_helper.Native("XPathExpression"))];
  const _lookupNamespaceUri_1 = Symbol('_lookupNamespaceUri_1');
  class XPathNSResolver extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateXPathNSResolver() {
      return XPathNSResolver._internalWrap();
    }
    static _internalWrap() {
      return new XPathNSResolver.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    lookupNamespaceUri(prefix) {
      return this[_lookupNamespaceUri_1](prefix);
    }
    [_lookupNamespaceUri_1](prefix) {
      return dart.as(wrap_jso(this.raw.lookupNamespaceURI(unwrap_jso(prefix))), core.String);
    }
  }
  dart.defineNamedConstructor(XPathNSResolver, 'internal_');
  dart.setSignature(XPathNSResolver, {
    constructors: () => ({
      _: [XPathNSResolver, []],
      _internalWrap: [XPathNSResolver, []],
      internal_: [XPathNSResolver, []]
    }),
    methods: () => ({
      lookupNamespaceUri: [core.String, [core.String]],
      [_lookupNamespaceUri_1]: [core.String, [dart.dynamic]]
    }),
    statics: () => ({internalCreateXPathNSResolver: [XPathNSResolver, []]}),
    names: ['internalCreateXPathNSResolver']
  });
  XPathNSResolver[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XPathNSResolver')), core.deprecated, dart.const(new _js_helper.Native("XPathNSResolver"))];
  const _iterateNext_1 = Symbol('_iterateNext_1');
  const _snapshotItem_1 = Symbol('_snapshotItem_1');
  class XPathResult extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateXPathResult() {
      return XPathResult._internalWrap();
    }
    static _internalWrap() {
      return new XPathResult.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get booleanValue() {
      return dart.as(wrap_jso(this.raw.booleanValue), core.bool);
    }
    get invalidIteratorState() {
      return dart.as(wrap_jso(this.raw.invalidIteratorState), core.bool);
    }
    get numberValue() {
      return dart.as(wrap_jso(this.raw.numberValue), core.double);
    }
    get resultType() {
      return dart.as(wrap_jso(this.raw.resultType), core.int);
    }
    get singleNodeValue() {
      return dart.as(wrap_jso(this.raw.singleNodeValue), Node);
    }
    get snapshotLength() {
      return dart.as(wrap_jso(this.raw.snapshotLength), core.int);
    }
    get stringValue() {
      return dart.as(wrap_jso(this.raw.stringValue), core.String);
    }
    iterateNext() {
      return this[_iterateNext_1]();
    }
    [_iterateNext_1]() {
      return dart.as(wrap_jso(this.raw.iterateNext()), Node);
    }
    snapshotItem(index) {
      return this[_snapshotItem_1](index);
    }
    [_snapshotItem_1](index) {
      return dart.as(wrap_jso(this.raw.snapshotItem(unwrap_jso(index))), Node);
    }
  }
  dart.defineNamedConstructor(XPathResult, 'internal_');
  dart.setSignature(XPathResult, {
    constructors: () => ({
      _: [XPathResult, []],
      _internalWrap: [XPathResult, []],
      internal_: [XPathResult, []]
    }),
    methods: () => ({
      iterateNext: [Node, []],
      [_iterateNext_1]: [Node, []],
      snapshotItem: [Node, [core.int]],
      [_snapshotItem_1]: [Node, [dart.dynamic]]
    }),
    statics: () => ({internalCreateXPathResult: [XPathResult, []]}),
    names: ['internalCreateXPathResult']
  });
  XPathResult[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XPathResult')), core.deprecated, dart.const(new _js_helper.Native("XPathResult"))];
  XPathResult.ANY_TYPE = 0;
  XPathResult.ANY_UNORDERED_NODE_TYPE = 8;
  XPathResult.BOOLEAN_TYPE = 3;
  XPathResult.FIRST_ORDERED_NODE_TYPE = 9;
  XPathResult.NUMBER_TYPE = 1;
  XPathResult.ORDERED_NODE_ITERATOR_TYPE = 5;
  XPathResult.ORDERED_NODE_SNAPSHOT_TYPE = 7;
  XPathResult.STRING_TYPE = 2;
  XPathResult.UNORDERED_NODE_ITERATOR_TYPE = 4;
  XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE = 6;
  class XmlDocument extends Document {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreateXmlDocument() {
      return XmlDocument._internalWrap();
    }
    static _internalWrap() {
      return new XmlDocument.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(XmlDocument, 'internal_');
  dart.setSignature(XmlDocument, {
    constructors: () => ({
      _: [XmlDocument, []],
      _internalWrap: [XmlDocument, []],
      internal_: [XmlDocument, []]
    }),
    statics: () => ({internalCreateXmlDocument: [XmlDocument, []]}),
    names: ['internalCreateXmlDocument']
  });
  XmlDocument[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XMLDocument')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("XMLDocument"))];
  const _serializeToString_1 = Symbol('_serializeToString_1');
  class XmlSerializer extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return XmlSerializer._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new XMLSerializer()), XmlSerializer);
    }
    static internalCreateXmlSerializer() {
      return XmlSerializer._internalWrap();
    }
    static _internalWrap() {
      return new XmlSerializer.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    serializeToString(node) {
      return this[_serializeToString_1](node);
    }
    [_serializeToString_1](node) {
      return dart.as(wrap_jso(this.raw.serializeToString(unwrap_jso(node))), core.String);
    }
  }
  dart.defineNamedConstructor(XmlSerializer, 'internal_');
  dart.setSignature(XmlSerializer, {
    constructors: () => ({
      _: [XmlSerializer, []],
      new: [XmlSerializer, []],
      _internalWrap: [XmlSerializer, []],
      internal_: [XmlSerializer, []]
    }),
    methods: () => ({
      serializeToString: [core.String, [Node]],
      [_serializeToString_1]: [core.String, [Node]]
    }),
    statics: () => ({
      _create_1: [XmlSerializer, []],
      internalCreateXmlSerializer: [XmlSerializer, []]
    }),
    names: ['_create_1', 'internalCreateXmlSerializer']
  });
  XmlSerializer[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XMLSerializer')), core.deprecated, dart.const(new _js_helper.Native("XMLSerializer"))];
  const _clearParameters_1 = Symbol('_clearParameters_1');
  const _getParameter_1 = Symbol('_getParameter_1');
  const _importStylesheet_1 = Symbol('_importStylesheet_1');
  const _removeParameter_1 = Symbol('_removeParameter_1');
  const _reset_1 = Symbol('_reset_1');
  const _setParameter_1 = Symbol('_setParameter_1');
  const _transformToDocument_1 = Symbol('_transformToDocument_1');
  const _transformToFragment_1 = Symbol('_transformToFragment_1');
  class XsltProcessor extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return XsltProcessor._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new XSLTProcessor()), XsltProcessor);
    }
    static internalCreateXsltProcessor() {
      return XsltProcessor._internalWrap();
    }
    static _internalWrap() {
      return new XsltProcessor.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    static get supported() {
      return !!window.XSLTProcessor;
    }
    clearParameters() {
      this[_clearParameters_1]();
      return;
    }
    [_clearParameters_1]() {
      return wrap_jso(this.raw.clearParameters());
    }
    getParameter(namespaceURI, localName) {
      return this[_getParameter_1](namespaceURI, localName);
    }
    [_getParameter_1](namespaceURI, localName) {
      return dart.as(wrap_jso(this.raw.getParameter(unwrap_jso(namespaceURI), unwrap_jso(localName))), core.String);
    }
    importStylesheet(stylesheet) {
      this[_importStylesheet_1](stylesheet);
      return;
    }
    [_importStylesheet_1](stylesheet) {
      return wrap_jso(this.raw.importStylesheet(unwrap_jso(stylesheet)));
    }
    removeParameter(namespaceURI, localName) {
      this[_removeParameter_1](namespaceURI, localName);
      return;
    }
    [_removeParameter_1](namespaceURI, localName) {
      return wrap_jso(this.raw.removeParameter(unwrap_jso(namespaceURI), unwrap_jso(localName)));
    }
    reset() {
      this[_reset_1]();
      return;
    }
    [_reset_1]() {
      return wrap_jso(this.raw.reset());
    }
    setParameter(namespaceURI, localName, value) {
      this[_setParameter_1](namespaceURI, localName, value);
      return;
    }
    [_setParameter_1](namespaceURI, localName, value) {
      return wrap_jso(this.raw.setParameter(unwrap_jso(namespaceURI), unwrap_jso(localName), unwrap_jso(value)));
    }
    transformToDocument(source) {
      return this[_transformToDocument_1](source);
    }
    [_transformToDocument_1](source) {
      return dart.as(wrap_jso(this.raw.transformToDocument(unwrap_jso(source))), Document);
    }
    transformToFragment(source, docVal) {
      return this[_transformToFragment_1](source, docVal);
    }
    [_transformToFragment_1](source, docVal) {
      return dart.as(wrap_jso(this.raw.transformToFragment(unwrap_jso(source), unwrap_jso(docVal))), DocumentFragment);
    }
  }
  dart.defineNamedConstructor(XsltProcessor, 'internal_');
  dart.setSignature(XsltProcessor, {
    constructors: () => ({
      _: [XsltProcessor, []],
      new: [XsltProcessor, []],
      _internalWrap: [XsltProcessor, []],
      internal_: [XsltProcessor, []]
    }),
    methods: () => ({
      clearParameters: [dart.void, []],
      [_clearParameters_1]: [dart.void, []],
      getParameter: [core.String, [core.String, core.String]],
      [_getParameter_1]: [core.String, [dart.dynamic, dart.dynamic]],
      importStylesheet: [dart.void, [Node]],
      [_importStylesheet_1]: [dart.void, [Node]],
      removeParameter: [dart.void, [core.String, core.String]],
      [_removeParameter_1]: [dart.void, [dart.dynamic, dart.dynamic]],
      reset: [dart.void, []],
      [_reset_1]: [dart.void, []],
      setParameter: [dart.void, [core.String, core.String, core.String]],
      [_setParameter_1]: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      transformToDocument: [Document, [Node]],
      [_transformToDocument_1]: [Document, [Node]],
      transformToFragment: [DocumentFragment, [Node, Document]],
      [_transformToFragment_1]: [DocumentFragment, [Node, Document]]
    }),
    statics: () => ({
      _create_1: [XsltProcessor, []],
      internalCreateXsltProcessor: [XsltProcessor, []]
    }),
    names: ['_create_1', 'internalCreateXsltProcessor']
  });
  XsltProcessor[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XSLTProcessor')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.FIREFOX)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), core.deprecated, dart.const(new _js_helper.Native("XSLTProcessor"))];
  class _Attr extends Node {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_Attr() {
      return _Attr._internalWrap();
    }
    static _internalWrap() {
      return new _Attr.internal_();
    }
    internal_() {
      super.internal_();
    }
    get name() {
      return dart.as(wrap_jso(this.raw.name), core.String);
    }
    get text() {
      return dart.as(wrap_jso(this.raw.textContent), core.String);
    }
    set text(value) {
      this.raw.textContent = unwrap_jso(value);
    }
    get value() {
      return dart.as(wrap_jso(this.raw.value), core.String);
    }
    set value(val) {
      return this.raw.value = unwrap_jso(val);
    }
  }
  dart.defineNamedConstructor(_Attr, 'internal_');
  dart.setSignature(_Attr, {
    constructors: () => ({
      _: [_Attr, []],
      _internalWrap: [_Attr, []],
      internal_: [_Attr, []]
    }),
    statics: () => ({internalCreate_Attr: [_Attr, []]}),
    names: ['internalCreate_Attr']
  });
  _Attr[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Attr')), dart.const(new _js_helper.Native("Attr"))];
  class _CSSValue extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_CSSValue() {
      return _CSSValue._internalWrap();
    }
    static _internalWrap() {
      return new _CSSValue.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_CSSValue, 'internal_');
  dart.setSignature(_CSSValue, {
    constructors: () => ({
      _: [_CSSValue, []],
      _internalWrap: [_CSSValue, []],
      internal_: [_CSSValue, []]
    }),
    statics: () => ({internalCreate_CSSValue: [_CSSValue, []]}),
    names: ['internalCreate_CSSValue']
  });
  _CSSValue[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('CSSValue')), core.deprecated, dart.const(new _js_helper.Native("CSSValue"))];
  class _CSSPrimitiveValue extends _CSSValue {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_CSSPrimitiveValue() {
      return _CSSPrimitiveValue._internalWrap();
    }
    static _internalWrap() {
      return new _CSSPrimitiveValue.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_CSSPrimitiveValue, 'internal_');
  dart.setSignature(_CSSPrimitiveValue, {
    constructors: () => ({
      _: [_CSSPrimitiveValue, []],
      _internalWrap: [_CSSPrimitiveValue, []],
      internal_: [_CSSPrimitiveValue, []]
    }),
    statics: () => ({internalCreate_CSSPrimitiveValue: [_CSSPrimitiveValue, []]}),
    names: ['internalCreate_CSSPrimitiveValue']
  });
  _CSSPrimitiveValue[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('CSSPrimitiveValue')), core.deprecated, dart.const(new _js_helper.Native("CSSPrimitiveValue"))];
  class _Cache extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_Cache() {
      return _Cache._internalWrap();
    }
    static _internalWrap() {
      return new _Cache.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_Cache, 'internal_');
  dart.setSignature(_Cache, {
    constructors: () => ({
      _: [_Cache, []],
      _internalWrap: [_Cache, []],
      internal_: [_Cache, []]
    }),
    statics: () => ({internalCreate_Cache: [_Cache, []]}),
    names: ['internalCreate_Cache']
  });
  _Cache[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Cache')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("Cache"))];
  class _CanvasPathMethods extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
  }
  dart.setSignature(_CanvasPathMethods, {
    constructors: () => ({_: [_CanvasPathMethods, []]})
  });
  _CanvasPathMethods[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('CanvasPathMethods')), dart.const(new _metadata.Experimental())];
  class _ClientRect extends DartHtmlDomObject {
    toString() {
      return `Rectangle (${this.left}, ${this.top}) ${this.width} x ${this.height}`;
    }
    ['=='](other) {
      if (!dart.is(other, math.Rectangle))
        return false;
      return dart.equals(this.left, dart.dload(other, 'left')) && dart.equals(this.top, dart.dload(other, 'top')) && dart.equals(this.width, dart.dload(other, 'width')) && dart.equals(this.height, dart.dload(other, 'height'));
    }
    get hashCode() {
      return _JenkinsSmiHash.hash4(dart.hashCode(this.left), dart.hashCode(this.top), dart.hashCode(this.width), dart.hashCode(this.height));
    }
    intersection(other) {
      let x0 = math.max(this.left, dart.as(other.left, core.num));
      let x1 = math.min(dart.notNull(this.left) + dart.notNull(this.width), dart.as(dart.dsend(other.left, '+', other.width), core.num));
      if (dart.notNull(x0) <= dart.notNull(x1)) {
        let y0 = math.max(this.top, dart.as(other.top, core.num));
        let y1 = math.min(dart.notNull(this.top) + dart.notNull(this.height), dart.as(dart.dsend(other.top, '+', other.height), core.num));
        if (dart.notNull(y0) <= dart.notNull(y1)) {
          return new math.Rectangle(x0, y0, dart.notNull(x1) - dart.notNull(x0), dart.notNull(y1) - dart.notNull(y0));
        }
      }
      return null;
    }
    intersects(other) {
      return dart.notNull(this.left) <= dart.notNull(other.left) + dart.notNull(other.width) && dart.notNull(other.left) <= dart.notNull(this.left) + dart.notNull(this.width) && dart.notNull(this.top) <= dart.notNull(other.top) + dart.notNull(other.height) && dart.notNull(other.top) <= dart.notNull(this.top) + dart.notNull(this.height);
    }
    boundingBox(other) {
      let right = math.max(dart.notNull(this.left) + dart.notNull(this.width), dart.as(dart.dsend(other.left, '+', other.width), core.num));
      let bottom = math.max(dart.notNull(this.top) + dart.notNull(this.height), dart.as(dart.dsend(other.top, '+', other.height), core.num));
      let left = math.min(this.left, dart.as(other.left, core.num));
      let top = math.min(this.top, dart.as(other.top, core.num));
      return new math.Rectangle(left, top, dart.notNull(right) - dart.notNull(left), dart.notNull(bottom) - dart.notNull(top));
    }
    containsRectangle(another) {
      return dart.notNull(this.left) <= dart.notNull(another.left) && dart.notNull(this.left) + dart.notNull(this.width) >= dart.notNull(another.left) + dart.notNull(another.width) && dart.notNull(this.top) <= dart.notNull(another.top) && dart.notNull(this.top) + dart.notNull(this.height) >= dart.notNull(another.top) + dart.notNull(another.height);
    }
    containsPoint(another) {
      return dart.notNull(another.x) >= dart.notNull(this.left) && dart.notNull(another.x) <= dart.notNull(this.left) + dart.notNull(this.width) && dart.notNull(another.y) >= dart.notNull(this.top) && dart.notNull(another.y) <= dart.notNull(this.top) + dart.notNull(this.height);
    }
    get topLeft() {
      return new math.Point(this.left, this.top);
    }
    get topRight() {
      return new math.Point(dart.notNull(this.left) + dart.notNull(this.width), this.top);
    }
    get bottomRight() {
      return new math.Point(dart.notNull(this.left) + dart.notNull(this.width), dart.notNull(this.top) + dart.notNull(this.height));
    }
    get bottomLeft() {
      return new math.Point(this.left, dart.notNull(this.top) + dart.notNull(this.height));
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_ClientRect() {
      return _ClientRect._internalWrap();
    }
    static _internalWrap() {
      return new _ClientRect.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    get bottom() {
      return dart.as(wrap_jso(this.raw.bottom), core.double);
    }
    get height() {
      return dart.as(wrap_jso(this.raw.height), core.double);
    }
    get left() {
      return dart.as(wrap_jso(this.raw.left), core.double);
    }
    get right() {
      return dart.as(wrap_jso(this.raw.right), core.double);
    }
    get top() {
      return dart.as(wrap_jso(this.raw.top), core.double);
    }
    get width() {
      return dart.as(wrap_jso(this.raw.width), core.double);
    }
  }
  _ClientRect[dart.implements] = () => [math.Rectangle];
  dart.defineNamedConstructor(_ClientRect, 'internal_');
  dart.setSignature(_ClientRect, {
    constructors: () => ({
      _: [_ClientRect, []],
      _internalWrap: [_ClientRect, []],
      internal_: [_ClientRect, []]
    }),
    methods: () => ({
      intersection: [math.Rectangle, [math.Rectangle]],
      intersects: [core.bool, [math.Rectangle$(core.num)]],
      boundingBox: [math.Rectangle, [math.Rectangle]],
      containsRectangle: [core.bool, [math.Rectangle$(core.num)]],
      containsPoint: [core.bool, [math.Point$(core.num)]]
    }),
    statics: () => ({internalCreate_ClientRect: [_ClientRect, []]}),
    names: ['internalCreate_ClientRect']
  });
  _ClientRect[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ClientRect')), dart.const(new _js_helper.Native("ClientRect"))];
  class _JenkinsSmiHash extends core.Object {
    static combine(hash, value) {
      hash = 536870911 & dart.notNull(hash) + dart.notNull(value);
      hash = 536870911 & dart.notNull(hash) + ((524287 & dart.notNull(hash)) << 10);
      return dart.notNull(hash) ^ dart.notNull(hash) >> 6;
    }
    static finish(hash) {
      hash = 536870911 & dart.notNull(hash) + ((67108863 & dart.notNull(hash)) << 3);
      hash = dart.notNull(hash) ^ dart.notNull(hash) >> 11;
      return 536870911 & dart.notNull(hash) + ((16383 & dart.notNull(hash)) << 15);
    }
    static hash2(a, b) {
      return _JenkinsSmiHash.finish(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(0, dart.as(a, core.int)), dart.as(b, core.int)));
    }
    static hash4(a, b, c, d) {
      return _JenkinsSmiHash.finish(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(0, dart.as(a, core.int)), dart.as(b, core.int)), dart.as(c, core.int)), dart.as(d, core.int)));
    }
  }
  dart.setSignature(_JenkinsSmiHash, {
    statics: () => ({
      combine: [core.int, [core.int, core.int]],
      finish: [core.int, [core.int]],
      hash2: [core.int, [dart.dynamic, dart.dynamic]],
      hash4: [core.int, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['combine', 'finish', 'hash2', 'hash4']
  });
  class _ClientRectList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(math.Rectangle), ImmutableListMixin$(math.Rectangle)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_ClientRectList() {
      return _ClientRectList._internalWrap();
    }
    static _internalWrap() {
      return new _ClientRectList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), math.Rectangle);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), math.Rectangle);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), math.Rectangle);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), math.Rectangle);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), math.Rectangle);
    }
  }
  _ClientRectList[dart.implements] = () => [core.List$(math.Rectangle), _js_helper.JavaScriptIndexingBehavior];
  dart.defineNamedConstructor(_ClientRectList, 'internal_');
  dart.setSignature(_ClientRectList, {
    constructors: () => ({
      _: [_ClientRectList, []],
      _internalWrap: [_ClientRectList, []],
      internal_: [_ClientRectList, []]
    }),
    methods: () => ({
      get: [math.Rectangle, [core.int]],
      set: [dart.void, [core.int, math.Rectangle]],
      elementAt: [math.Rectangle, [core.int]],
      item: [math.Rectangle, [core.int]],
      [_item_1]: [math.Rectangle, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_ClientRectList: [_ClientRectList, []]}),
    names: ['internalCreate_ClientRectList']
  });
  dart.defineExtensionMembers(_ClientRectList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _ClientRectList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ClientRectList')), dart.const(new _js_helper.Native("ClientRectList,DOMRectList"))];
  class _Counter extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_Counter() {
      return _Counter._internalWrap();
    }
    static _internalWrap() {
      return new _Counter.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_Counter, 'internal_');
  dart.setSignature(_Counter, {
    constructors: () => ({
      _: [_Counter, []],
      _internalWrap: [_Counter, []],
      internal_: [_Counter, []]
    }),
    statics: () => ({internalCreate_Counter: [_Counter, []]}),
    names: ['internalCreate_Counter']
  });
  _Counter[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Counter')), core.deprecated, dart.const(new _js_helper.Native("Counter"))];
  class _CssRuleList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(dart.dynamic), ImmutableListMixin$(dart.dynamic)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_CssRuleList() {
      return _CssRuleList._internalWrap();
    }
    static _internalWrap() {
      return new _CssRuleList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return wrap_jso(this.raw[index]);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return wrap_jso(this.raw.item(unwrap_jso(index)));
    }
  }
  _CssRuleList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(dart.dynamic)];
  dart.defineNamedConstructor(_CssRuleList, 'internal_');
  dart.setSignature(_CssRuleList, {
    constructors: () => ({
      _: [_CssRuleList, []],
      _internalWrap: [_CssRuleList, []],
      internal_: [_CssRuleList, []]
    }),
    methods: () => ({
      get: [dart.dynamic, [core.int]],
      set: [dart.void, [core.int, dart.dynamic]],
      elementAt: [dart.dynamic, [core.int]],
      item: [dart.dynamic, [core.int]],
      [_item_1]: [dart.dynamic, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_CssRuleList: [_CssRuleList, []]}),
    names: ['internalCreate_CssRuleList']
  });
  dart.defineExtensionMembers(_CssRuleList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _CssRuleList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('CSSRuleList')), dart.const(new _js_helper.Native("CSSRuleList"))];
  class _CssValueList extends dart.mixin(_CSSValue, collection.ListMixin$(_CSSValue), ImmutableListMixin$(_CSSValue)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_CssValueList() {
      return _CssValueList._internalWrap();
    }
    static _internalWrap() {
      return new _CssValueList.internal_();
    }
    internal_() {
      super.internal_();
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), _CSSValue);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), _CSSValue);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), _CSSValue);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), _CSSValue);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), _CSSValue);
    }
  }
  _CssValueList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(_CSSValue)];
  dart.defineNamedConstructor(_CssValueList, 'internal_');
  dart.setSignature(_CssValueList, {
    constructors: () => ({
      _: [_CssValueList, []],
      _internalWrap: [_CssValueList, []],
      internal_: [_CssValueList, []]
    }),
    methods: () => ({
      get: [_CSSValue, [core.int]],
      set: [dart.void, [core.int, _CSSValue]],
      elementAt: [_CSSValue, [core.int]],
      item: [_CSSValue, [core.int]],
      [_item_1]: [_CSSValue, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_CssValueList: [_CssValueList, []]}),
    names: ['internalCreate_CssValueList']
  });
  dart.defineExtensionMembers(_CssValueList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _CssValueList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('CSSValueList')), core.deprecated, dart.const(new _js_helper.Native("CSSValueList"))];
  class _DOMFileSystemSync extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_DOMFileSystemSync() {
      return _DOMFileSystemSync._internalWrap();
    }
    static _internalWrap() {
      return new _DOMFileSystemSync.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_DOMFileSystemSync, 'internal_');
  dart.setSignature(_DOMFileSystemSync, {
    constructors: () => ({
      _: [_DOMFileSystemSync, []],
      _internalWrap: [_DOMFileSystemSync, []],
      internal_: [_DOMFileSystemSync, []]
    }),
    statics: () => ({internalCreate_DOMFileSystemSync: [_DOMFileSystemSync, []]}),
    names: ['internalCreate_DOMFileSystemSync']
  });
  _DOMFileSystemSync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DOMFileSystemSync')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("DOMFileSystemSync"))];
  class _EntrySync extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_EntrySync() {
      return _EntrySync._internalWrap();
    }
    static _internalWrap() {
      return new _EntrySync.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_EntrySync, 'internal_');
  dart.setSignature(_EntrySync, {
    constructors: () => ({
      _: [_EntrySync, []],
      _internalWrap: [_EntrySync, []],
      internal_: [_EntrySync, []]
    }),
    statics: () => ({internalCreate_EntrySync: [_EntrySync, []]}),
    names: ['internalCreate_EntrySync']
  });
  _EntrySync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('EntrySync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("EntrySync"))];
  class _DirectoryEntrySync extends _EntrySync {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_DirectoryEntrySync() {
      return _DirectoryEntrySync._internalWrap();
    }
    static _internalWrap() {
      return new _DirectoryEntrySync.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_DirectoryEntrySync, 'internal_');
  dart.setSignature(_DirectoryEntrySync, {
    constructors: () => ({
      _: [_DirectoryEntrySync, []],
      _internalWrap: [_DirectoryEntrySync, []],
      internal_: [_DirectoryEntrySync, []]
    }),
    statics: () => ({internalCreate_DirectoryEntrySync: [_DirectoryEntrySync, []]}),
    names: ['internalCreate_DirectoryEntrySync']
  });
  _DirectoryEntrySync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DirectoryEntrySync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("DirectoryEntrySync"))];
  class _DirectoryReaderSync extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_DirectoryReaderSync() {
      return _DirectoryReaderSync._internalWrap();
    }
    static _internalWrap() {
      return new _DirectoryReaderSync.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_DirectoryReaderSync, 'internal_');
  dart.setSignature(_DirectoryReaderSync, {
    constructors: () => ({
      _: [_DirectoryReaderSync, []],
      _internalWrap: [_DirectoryReaderSync, []],
      internal_: [_DirectoryReaderSync, []]
    }),
    statics: () => ({internalCreate_DirectoryReaderSync: [_DirectoryReaderSync, []]}),
    names: ['internalCreate_DirectoryReaderSync']
  });
  _DirectoryReaderSync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DirectoryReaderSync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("DirectoryReaderSync"))];
  class _DocumentType extends Node {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_DocumentType() {
      return _DocumentType._internalWrap();
    }
    static _internalWrap() {
      return new _DocumentType.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  _DocumentType[dart.implements] = () => [ChildNode];
  dart.defineNamedConstructor(_DocumentType, 'internal_');
  dart.setSignature(_DocumentType, {
    constructors: () => ({
      _: [_DocumentType, []],
      _internalWrap: [_DocumentType, []],
      internal_: [_DocumentType, []]
    }),
    statics: () => ({internalCreate_DocumentType: [_DocumentType, []]}),
    names: ['internalCreate_DocumentType']
  });
  _DocumentType[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('DocumentType')), core.deprecated, dart.const(new _js_helper.Native("DocumentType"))];
  class _FileEntrySync extends _EntrySync {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_FileEntrySync() {
      return _FileEntrySync._internalWrap();
    }
    static _internalWrap() {
      return new _FileEntrySync.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_FileEntrySync, 'internal_');
  dart.setSignature(_FileEntrySync, {
    constructors: () => ({
      _: [_FileEntrySync, []],
      _internalWrap: [_FileEntrySync, []],
      internal_: [_FileEntrySync, []]
    }),
    statics: () => ({internalCreate_FileEntrySync: [_FileEntrySync, []]}),
    names: ['internalCreate_FileEntrySync']
  });
  _FileEntrySync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('FileEntrySync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("FileEntrySync"))];
  class _FileReaderSync extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new() {
      return _FileReaderSync._create_1();
    }
    static _create_1() {
      return dart.as(wrap_jso(new FileReaderSync()), _FileReaderSync);
    }
    static internalCreate_FileReaderSync() {
      return _FileReaderSync._internalWrap();
    }
    static _internalWrap() {
      return new _FileReaderSync.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_FileReaderSync, 'internal_');
  dart.setSignature(_FileReaderSync, {
    constructors: () => ({
      _: [_FileReaderSync, []],
      new: [_FileReaderSync, []],
      _internalWrap: [_FileReaderSync, []],
      internal_: [_FileReaderSync, []]
    }),
    statics: () => ({
      _create_1: [_FileReaderSync, []],
      internalCreate_FileReaderSync: [_FileReaderSync, []]
    }),
    names: ['_create_1', 'internalCreate_FileReaderSync']
  });
  _FileReaderSync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('FileReaderSync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("FileReaderSync"))];
  class _FileWriterSync extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_FileWriterSync() {
      return _FileWriterSync._internalWrap();
    }
    static _internalWrap() {
      return new _FileWriterSync.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_FileWriterSync, 'internal_');
  dart.setSignature(_FileWriterSync, {
    constructors: () => ({
      _: [_FileWriterSync, []],
      _internalWrap: [_FileWriterSync, []],
      internal_: [_FileWriterSync, []]
    }),
    statics: () => ({internalCreate_FileWriterSync: [_FileWriterSync, []]}),
    names: ['internalCreate_FileWriterSync']
  });
  _FileWriterSync[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('FileWriterSync')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("FileWriterSync"))];
  class _GamepadList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(dart.dynamic), ImmutableListMixin$(dart.dynamic)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_GamepadList() {
      return _GamepadList._internalWrap();
    }
    static _internalWrap() {
      return new _GamepadList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return wrap_jso(this.raw[index]);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return wrap_jso(this.raw.item(unwrap_jso(index)));
    }
  }
  _GamepadList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(dart.dynamic)];
  dart.defineNamedConstructor(_GamepadList, 'internal_');
  dart.setSignature(_GamepadList, {
    constructors: () => ({
      _: [_GamepadList, []],
      _internalWrap: [_GamepadList, []],
      internal_: [_GamepadList, []]
    }),
    methods: () => ({
      get: [dart.dynamic, [core.int]],
      set: [dart.void, [core.int, dart.dynamic]],
      elementAt: [dart.dynamic, [core.int]],
      item: [dart.dynamic, [core.int]],
      [_item_1]: [dart.dynamic, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_GamepadList: [_GamepadList, []]}),
    names: ['internalCreate_GamepadList']
  });
  dart.defineExtensionMembers(_GamepadList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _GamepadList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('GamepadList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("GamepadList"))];
  class _HTMLAllCollection extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLAllCollection() {
      return _HTMLAllCollection._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLAllCollection.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    [_item](index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Element);
    }
  }
  dart.defineNamedConstructor(_HTMLAllCollection, 'internal_');
  dart.setSignature(_HTMLAllCollection, {
    constructors: () => ({
      _: [_HTMLAllCollection, []],
      _internalWrap: [_HTMLAllCollection, []],
      internal_: [_HTMLAllCollection, []]
    }),
    methods: () => ({
      [_item]: [Element, [core.int]],
      [_item_1]: [Element, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_HTMLAllCollection: [_HTMLAllCollection, []]}),
    names: ['internalCreate_HTMLAllCollection']
  });
  _HTMLAllCollection[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLAllCollection')), core.deprecated, dart.const(new _js_helper.Native("HTMLAllCollection"))];
  class _HTMLAppletElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLAppletElement() {
      return _HTMLAppletElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLAppletElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(_HTMLAppletElement, 'internal_');
  dart.defineNamedConstructor(_HTMLAppletElement, 'created');
  dart.setSignature(_HTMLAppletElement, {
    constructors: () => ({
      _: [_HTMLAppletElement, []],
      _internalWrap: [_HTMLAppletElement, []],
      internal_: [_HTMLAppletElement, []],
      created: [_HTMLAppletElement, []]
    }),
    statics: () => ({internalCreate_HTMLAppletElement: [_HTMLAppletElement, []]}),
    names: ['internalCreate_HTMLAppletElement']
  });
  _HTMLAppletElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLAppletElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLAppletElement"))];
  class _HTMLDirectoryElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLDirectoryElement() {
      return _HTMLDirectoryElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLDirectoryElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(_HTMLDirectoryElement, 'internal_');
  dart.defineNamedConstructor(_HTMLDirectoryElement, 'created');
  dart.setSignature(_HTMLDirectoryElement, {
    constructors: () => ({
      _: [_HTMLDirectoryElement, []],
      _internalWrap: [_HTMLDirectoryElement, []],
      internal_: [_HTMLDirectoryElement, []],
      created: [_HTMLDirectoryElement, []]
    }),
    statics: () => ({internalCreate_HTMLDirectoryElement: [_HTMLDirectoryElement, []]}),
    names: ['internalCreate_HTMLDirectoryElement']
  });
  _HTMLDirectoryElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLDirectoryElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLDirectoryElement"))];
  class _HTMLFontElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLFontElement() {
      return _HTMLFontElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLFontElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(_HTMLFontElement, 'internal_');
  dart.defineNamedConstructor(_HTMLFontElement, 'created');
  dart.setSignature(_HTMLFontElement, {
    constructors: () => ({
      _: [_HTMLFontElement, []],
      _internalWrap: [_HTMLFontElement, []],
      internal_: [_HTMLFontElement, []],
      created: [_HTMLFontElement, []]
    }),
    statics: () => ({internalCreate_HTMLFontElement: [_HTMLFontElement, []]}),
    names: ['internalCreate_HTMLFontElement']
  });
  _HTMLFontElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLFontElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLFontElement"))];
  class _HTMLFrameElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLFrameElement() {
      return _HTMLFrameElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLFrameElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(_HTMLFrameElement, 'internal_');
  dart.defineNamedConstructor(_HTMLFrameElement, 'created');
  dart.setSignature(_HTMLFrameElement, {
    constructors: () => ({
      _: [_HTMLFrameElement, []],
      _internalWrap: [_HTMLFrameElement, []],
      internal_: [_HTMLFrameElement, []],
      created: [_HTMLFrameElement, []]
    }),
    statics: () => ({internalCreate_HTMLFrameElement: [_HTMLFrameElement, []]}),
    names: ['internalCreate_HTMLFrameElement']
  });
  _HTMLFrameElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLFrameElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLFrameElement"))];
  class _HTMLFrameSetElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLFrameSetElement() {
      return _HTMLFrameSetElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLFrameSetElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  _HTMLFrameSetElement[dart.implements] = () => [WindowEventHandlers];
  dart.defineNamedConstructor(_HTMLFrameSetElement, 'internal_');
  dart.defineNamedConstructor(_HTMLFrameSetElement, 'created');
  dart.setSignature(_HTMLFrameSetElement, {
    constructors: () => ({
      _: [_HTMLFrameSetElement, []],
      _internalWrap: [_HTMLFrameSetElement, []],
      internal_: [_HTMLFrameSetElement, []],
      created: [_HTMLFrameSetElement, []]
    }),
    statics: () => ({internalCreate_HTMLFrameSetElement: [_HTMLFrameSetElement, []]}),
    names: ['internalCreate_HTMLFrameSetElement']
  });
  _HTMLFrameSetElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLFrameSetElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLFrameSetElement"))];
  class _HTMLMarqueeElement extends HtmlElement {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_HTMLMarqueeElement() {
      return _HTMLMarqueeElement._internalWrap();
    }
    static _internalWrap() {
      return new _HTMLMarqueeElement.internal_();
    }
    internal_() {
      super.internal_();
    }
    created() {
      super.created();
    }
  }
  dart.defineNamedConstructor(_HTMLMarqueeElement, 'internal_');
  dart.defineNamedConstructor(_HTMLMarqueeElement, 'created');
  dart.setSignature(_HTMLMarqueeElement, {
    constructors: () => ({
      _: [_HTMLMarqueeElement, []],
      _internalWrap: [_HTMLMarqueeElement, []],
      internal_: [_HTMLMarqueeElement, []],
      created: [_HTMLMarqueeElement, []]
    }),
    statics: () => ({internalCreate_HTMLMarqueeElement: [_HTMLMarqueeElement, []]}),
    names: ['internalCreate_HTMLMarqueeElement']
  });
  _HTMLMarqueeElement[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('HTMLMarqueeElement')), core.deprecated, dart.const(new _js_helper.Native("HTMLMarqueeElement"))];
  const _initMutationEvent = Symbol('_initMutationEvent');
  class _MutationEvent extends Event {
    static new(type, opts) {
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : false;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : false;
      let relatedNode = opts && 'relatedNode' in opts ? opts.relatedNode : null;
      let prevValue = opts && 'prevValue' in opts ? opts.prevValue : null;
      let newValue = opts && 'newValue' in opts ? opts.newValue : null;
      let attrName = opts && 'attrName' in opts ? opts.attrName : null;
      let attrChange = opts && 'attrChange' in opts ? opts.attrChange : 0;
      let event = exports.document[_createEvent]('MutationEvent');
      dart.dcall(event[_initMutationEvent], type, canBubble, cancelable, relatedNode, prevValue, newValue, attrName, attrChange);
      return dart.as(event, _MutationEvent);
    }
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_MutationEvent() {
      return _MutationEvent._internalWrap();
    }
    static _internalWrap() {
      return new _MutationEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_MutationEvent, 'internal_');
  dart.setSignature(_MutationEvent, {
    constructors: () => ({
      new: [_MutationEvent, [core.String], {canBubble: core.bool, cancelable: core.bool, relatedNode: Node, prevValue: core.String, newValue: core.String, attrName: core.String, attrChange: core.int}],
      _: [_MutationEvent, []],
      _internalWrap: [_MutationEvent, []],
      internal_: [_MutationEvent, []]
    }),
    statics: () => ({internalCreate_MutationEvent: [_MutationEvent, []]}),
    names: ['internalCreate_MutationEvent']
  });
  _MutationEvent[dart.metadata] = () => [dart.const(new _metadata.DomName('MutationEvent')), core.deprecated, dart.const(new _js_helper.Native("MutationEvent"))];
  const _getNamedItem_1 = Symbol('_getNamedItem_1');
  const _getNamedItemNS_1 = Symbol('_getNamedItemNS_1');
  const _removeNamedItem_1 = Symbol('_removeNamedItem_1');
  const _removeNamedItemNS_1 = Symbol('_removeNamedItemNS_1');
  const _setNamedItem_1 = Symbol('_setNamedItem_1');
  const _setNamedItemNS_1 = Symbol('_setNamedItemNS_1');
  class _NamedNodeMap extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(Node), ImmutableListMixin$(Node)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_NamedNodeMap() {
      return _NamedNodeMap._internalWrap();
    }
    static _internalWrap() {
      return new _NamedNodeMap.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), Node);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), Node);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), Node);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return dart.as(wrap_jso(this.raw.__getter__(unwrap_jso(name))), Node);
    }
    getNamedItem(name) {
      return this[_getNamedItem_1](name);
    }
    [_getNamedItem_1](name) {
      return dart.as(wrap_jso(this.raw.getNamedItem(unwrap_jso(name))), Node);
    }
    getNamedItemNS(namespaceURI, localName) {
      return this[_getNamedItemNS_1](namespaceURI, localName);
    }
    [_getNamedItemNS_1](namespaceURI, localName) {
      return dart.as(wrap_jso(this.raw.getNamedItemNS(unwrap_jso(namespaceURI), unwrap_jso(localName))), Node);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), Node);
    }
    removeNamedItem(name) {
      return this[_removeNamedItem_1](name);
    }
    [_removeNamedItem_1](name) {
      return dart.as(wrap_jso(this.raw.removeNamedItem(unwrap_jso(name))), Node);
    }
    removeNamedItemNS(namespaceURI, localName) {
      return this[_removeNamedItemNS_1](namespaceURI, localName);
    }
    [_removeNamedItemNS_1](namespaceURI, localName) {
      return dart.as(wrap_jso(this.raw.removeNamedItemNS(unwrap_jso(namespaceURI), unwrap_jso(localName))), Node);
    }
    setNamedItem(node) {
      return this[_setNamedItem_1](node);
    }
    [_setNamedItem_1](node) {
      return dart.as(wrap_jso(this.raw.setNamedItem(unwrap_jso(node))), Node);
    }
    setNamedItemNS(node) {
      return this[_setNamedItemNS_1](node);
    }
    [_setNamedItemNS_1](node) {
      return dart.as(wrap_jso(this.raw.setNamedItemNS(unwrap_jso(node))), Node);
    }
  }
  _NamedNodeMap[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(Node)];
  dart.defineNamedConstructor(_NamedNodeMap, 'internal_');
  dart.setSignature(_NamedNodeMap, {
    constructors: () => ({
      _: [_NamedNodeMap, []],
      _internalWrap: [_NamedNodeMap, []],
      internal_: [_NamedNodeMap, []]
    }),
    methods: () => ({
      get: [Node, [core.int]],
      set: [dart.void, [core.int, Node]],
      elementAt: [Node, [core.int]],
      [__getter__]: [Node, [core.String]],
      [__getter___1]: [Node, [dart.dynamic]],
      getNamedItem: [Node, [core.String]],
      [_getNamedItem_1]: [Node, [dart.dynamic]],
      getNamedItemNS: [Node, [core.String, core.String]],
      [_getNamedItemNS_1]: [Node, [dart.dynamic, dart.dynamic]],
      item: [Node, [core.int]],
      [_item_1]: [Node, [dart.dynamic]],
      removeNamedItem: [Node, [core.String]],
      [_removeNamedItem_1]: [Node, [dart.dynamic]],
      removeNamedItemNS: [Node, [core.String, core.String]],
      [_removeNamedItemNS_1]: [Node, [dart.dynamic, dart.dynamic]],
      setNamedItem: [Node, [Node]],
      [_setNamedItem_1]: [Node, [Node]],
      setNamedItemNS: [Node, [Node]],
      [_setNamedItemNS_1]: [Node, [Node]]
    }),
    statics: () => ({internalCreate_NamedNodeMap: [_NamedNodeMap, []]}),
    names: ['internalCreate_NamedNodeMap']
  });
  dart.defineExtensionMembers(_NamedNodeMap, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _NamedNodeMap[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('NamedNodeMap')), core.deprecated, dart.const(new _js_helper.Native("NamedNodeMap,MozNamedAttrMap"))];
  class _PagePopupController extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_PagePopupController() {
      return _PagePopupController._internalWrap();
    }
    static _internalWrap() {
      return new _PagePopupController.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_PagePopupController, 'internal_');
  dart.setSignature(_PagePopupController, {
    constructors: () => ({
      _: [_PagePopupController, []],
      _internalWrap: [_PagePopupController, []],
      internal_: [_PagePopupController, []]
    }),
    statics: () => ({internalCreate_PagePopupController: [_PagePopupController, []]}),
    names: ['internalCreate_PagePopupController']
  });
  _PagePopupController[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('PagePopupController')), core.deprecated, dart.const(new _js_helper.Native("PagePopupController"))];
  class _RGBColor extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_RGBColor() {
      return _RGBColor._internalWrap();
    }
    static _internalWrap() {
      return new _RGBColor.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_RGBColor, 'internal_');
  dart.setSignature(_RGBColor, {
    constructors: () => ({
      _: [_RGBColor, []],
      _internalWrap: [_RGBColor, []],
      internal_: [_RGBColor, []]
    }),
    statics: () => ({internalCreate_RGBColor: [_RGBColor, []]}),
    names: ['internalCreate_RGBColor']
  });
  _RGBColor[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('RGBColor')), core.deprecated, dart.const(new _js_helper.Native("RGBColor"))];
  class _RadioNodeList extends NodeList {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_RadioNodeList() {
      return _RadioNodeList._internalWrap();
    }
    static _internalWrap() {
      return new _RadioNodeList.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_RadioNodeList, 'internal_');
  dart.setSignature(_RadioNodeList, {
    constructors: () => ({
      _: [_RadioNodeList, []],
      _internalWrap: [_RadioNodeList, []],
      internal_: [_RadioNodeList, []]
    }),
    statics: () => ({internalCreate_RadioNodeList: [_RadioNodeList, []]}),
    names: ['internalCreate_RadioNodeList']
  });
  _RadioNodeList[dart.metadata] = () => [dart.const(new _metadata.DomName('RadioNodeList')), dart.const(new _js_helper.Native("RadioNodeList"))];
  class _Rect extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_Rect() {
      return _Rect._internalWrap();
    }
    static _internalWrap() {
      return new _Rect.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_Rect, 'internal_');
  dart.setSignature(_Rect, {
    constructors: () => ({
      _: [_Rect, []],
      _internalWrap: [_Rect, []],
      internal_: [_Rect, []]
    }),
    statics: () => ({internalCreate_Rect: [_Rect, []]}),
    names: ['internalCreate_Rect']
  });
  _Rect[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('Rect')), core.deprecated, dart.const(new _js_helper.Native("Rect"))];
  class _ServiceWorker extends EventTarget {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_ServiceWorker() {
      return _ServiceWorker._internalWrap();
    }
    static _internalWrap() {
      return new _ServiceWorker.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_ServiceWorker, 'internal_');
  dart.setSignature(_ServiceWorker, {
    constructors: () => ({
      _: [_ServiceWorker, []],
      _internalWrap: [_ServiceWorker, []],
      internal_: [_ServiceWorker, []]
    }),
    statics: () => ({internalCreate_ServiceWorker: [_ServiceWorker, []]}),
    names: ['internalCreate_ServiceWorker']
  });
  _ServiceWorker[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('ServiceWorker')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("ServiceWorker"))];
  class _SpeechRecognitionResultList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(SpeechRecognitionResult), ImmutableListMixin$(SpeechRecognitionResult)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_SpeechRecognitionResultList() {
      return _SpeechRecognitionResultList._internalWrap();
    }
    static _internalWrap() {
      return new _SpeechRecognitionResultList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), SpeechRecognitionResult);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SpeechRecognitionResult);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), SpeechRecognitionResult);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), SpeechRecognitionResult);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), SpeechRecognitionResult);
    }
  }
  _SpeechRecognitionResultList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(SpeechRecognitionResult)];
  dart.defineNamedConstructor(_SpeechRecognitionResultList, 'internal_');
  dart.setSignature(_SpeechRecognitionResultList, {
    constructors: () => ({
      _: [_SpeechRecognitionResultList, []],
      _internalWrap: [_SpeechRecognitionResultList, []],
      internal_: [_SpeechRecognitionResultList, []]
    }),
    methods: () => ({
      get: [SpeechRecognitionResult, [core.int]],
      set: [dart.void, [core.int, SpeechRecognitionResult]],
      elementAt: [SpeechRecognitionResult, [core.int]],
      item: [SpeechRecognitionResult, [core.int]],
      [_item_1]: [SpeechRecognitionResult, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_SpeechRecognitionResultList: [_SpeechRecognitionResultList, []]}),
    names: ['internalCreate_SpeechRecognitionResultList']
  });
  dart.defineExtensionMembers(_SpeechRecognitionResultList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _SpeechRecognitionResultList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SpeechRecognitionResultList')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SpeechRecognitionResultList"))];
  class _StyleSheetList extends dart.mixin(DartHtmlDomObject, collection.ListMixin$(StyleSheet), ImmutableListMixin$(StyleSheet)) {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_StyleSheetList() {
      return _StyleSheetList._internalWrap();
    }
    static _internalWrap() {
      return new _StyleSheetList.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
    get length() {
      return dart.as(wrap_jso(this.raw.length), core.int);
    }
    get(index) {
      if (index >>> 0 !== index || index >= this.length)
        dart.throw(core.RangeError.index(index, this));
      return dart.as(wrap_jso(this.raw[index]), StyleSheet);
    }
    set(index, value) {
      dart.throw(new core.UnsupportedError("Cannot assign element of immutable List."));
      return value;
    }
    set length(value) {
      dart.throw(new core.UnsupportedError("Cannot resize immutable List."));
    }
    get first() {
      if (dart.notNull(this.length) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), StyleSheet);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get last() {
      let len = this.length;
      if (dart.notNull(len) > 0) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, dart.notNull(len) - 1), StyleSheet);
      }
      dart.throw(new core.StateError("No elements"));
    }
    get single() {
      let len = this.length;
      if (len == 1) {
        return dart.as(dart.dcall(/* Unimplemented unknown name */_nativeIndexedGetter, 0), StyleSheet);
      }
      if (len == 0)
        dart.throw(new core.StateError("No elements"));
      dart.throw(new core.StateError("More than one element"));
    }
    elementAt(index) {
      return this.get(index);
    }
    [__getter__](name) {
      return this[__getter___1](name);
    }
    [__getter___1](name) {
      return wrap_jso(this.raw.__getter__(unwrap_jso(name)));
    }
    item(index) {
      return this[_item_1](index);
    }
    [_item_1](index) {
      return dart.as(wrap_jso(this.raw.item(unwrap_jso(index))), StyleSheet);
    }
  }
  _StyleSheetList[dart.implements] = () => [_js_helper.JavaScriptIndexingBehavior, core.List$(StyleSheet)];
  dart.defineNamedConstructor(_StyleSheetList, 'internal_');
  dart.setSignature(_StyleSheetList, {
    constructors: () => ({
      _: [_StyleSheetList, []],
      _internalWrap: [_StyleSheetList, []],
      internal_: [_StyleSheetList, []]
    }),
    methods: () => ({
      get: [StyleSheet, [core.int]],
      set: [dart.void, [core.int, StyleSheet]],
      elementAt: [StyleSheet, [core.int]],
      [__getter__]: [dart.dynamic, [core.String]],
      [__getter___1]: [dart.dynamic, [dart.dynamic]],
      item: [StyleSheet, [core.int]],
      [_item_1]: [StyleSheet, [dart.dynamic]]
    }),
    statics: () => ({internalCreate_StyleSheetList: [_StyleSheetList, []]}),
    names: ['internalCreate_StyleSheetList']
  });
  dart.defineExtensionMembers(_StyleSheetList, [
    'get',
    'set',
    'elementAt',
    'length',
    'length',
    'first',
    'last',
    'single'
  ]);
  _StyleSheetList[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('StyleSheetList')), dart.const(new _js_helper.Native("StyleSheetList"))];
  class _SubtleCrypto extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_SubtleCrypto() {
      return _SubtleCrypto._internalWrap();
    }
    static _internalWrap() {
      return new _SubtleCrypto.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_SubtleCrypto, 'internal_');
  dart.setSignature(_SubtleCrypto, {
    constructors: () => ({
      _: [_SubtleCrypto, []],
      _internalWrap: [_SubtleCrypto, []],
      internal_: [_SubtleCrypto, []]
    }),
    statics: () => ({internalCreate_SubtleCrypto: [_SubtleCrypto, []]}),
    names: ['internalCreate_SubtleCrypto']
  });
  _SubtleCrypto[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('SubtleCrypto')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("SubtleCrypto"))];
  class _WebKitCSSFilterValue extends _CssValueList {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_WebKitCSSFilterValue() {
      return _WebKitCSSFilterValue._internalWrap();
    }
    static _internalWrap() {
      return new _WebKitCSSFilterValue.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_WebKitCSSFilterValue, 'internal_');
  dart.setSignature(_WebKitCSSFilterValue, {
    constructors: () => ({
      _: [_WebKitCSSFilterValue, []],
      _internalWrap: [_WebKitCSSFilterValue, []],
      internal_: [_WebKitCSSFilterValue, []]
    }),
    statics: () => ({internalCreate_WebKitCSSFilterValue: [_WebKitCSSFilterValue, []]}),
    names: ['internalCreate_WebKitCSSFilterValue']
  });
  _WebKitCSSFilterValue[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WebKitCSSFilterValue')), core.deprecated, dart.const(new _js_helper.Native("WebKitCSSFilterValue"))];
  class _WebKitCSSMatrix extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static new(cssValue) {
      if (cssValue === void 0)
        cssValue = null;
      if (cssValue != null) {
        return _WebKitCSSMatrix._create_1(cssValue);
      }
      return _WebKitCSSMatrix._create_2();
    }
    static _create_1(cssValue) {
      return dart.as(wrap_jso(new WebKitCSSMatrix(cssValue)), _WebKitCSSMatrix);
    }
    static _create_2() {
      return dart.as(wrap_jso(new WebKitCSSMatrix()), _WebKitCSSMatrix);
    }
    static internalCreate_WebKitCSSMatrix() {
      return _WebKitCSSMatrix._internalWrap();
    }
    static _internalWrap() {
      return new _WebKitCSSMatrix.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  dart.defineNamedConstructor(_WebKitCSSMatrix, 'internal_');
  dart.setSignature(_WebKitCSSMatrix, {
    constructors: () => ({
      _: [_WebKitCSSMatrix, []],
      new: [_WebKitCSSMatrix, [], [core.String]],
      _internalWrap: [_WebKitCSSMatrix, []],
      internal_: [_WebKitCSSMatrix, []]
    }),
    statics: () => ({
      _create_1: [_WebKitCSSMatrix, [dart.dynamic]],
      _create_2: [_WebKitCSSMatrix, []],
      internalCreate_WebKitCSSMatrix: [_WebKitCSSMatrix, []]
    }),
    names: ['_create_1', '_create_2', 'internalCreate_WebKitCSSMatrix']
  });
  _WebKitCSSMatrix[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WebKitCSSMatrix')), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.CHROME)), dart.const(new _metadata.SupportedBrowser(_metadata.SupportedBrowser.SAFARI)), dart.const(new _metadata.Experimental()), core.deprecated, dart.const(new _js_helper.Native("WebKitCSSMatrix"))];
  class _WebKitCSSTransformValue extends _CssValueList {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_WebKitCSSTransformValue() {
      return _WebKitCSSTransformValue._internalWrap();
    }
    static _internalWrap() {
      return new _WebKitCSSTransformValue.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_WebKitCSSTransformValue, 'internal_');
  dart.setSignature(_WebKitCSSTransformValue, {
    constructors: () => ({
      _: [_WebKitCSSTransformValue, []],
      _internalWrap: [_WebKitCSSTransformValue, []],
      internal_: [_WebKitCSSTransformValue, []]
    }),
    statics: () => ({internalCreate_WebKitCSSTransformValue: [_WebKitCSSTransformValue, []]}),
    names: ['internalCreate_WebKitCSSTransformValue']
  });
  _WebKitCSSTransformValue[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WebKitCSSTransformValue')), core.deprecated, dart.const(new _js_helper.Native("WebKitCSSTransformValue"))];
  class _WindowTimers extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    [_clearInterval](handle) {
      return wrap_jso(this.raw.clearInterval(unwrap_jso(handle)));
    }
    [_clearTimeout](handle) {
      return wrap_jso(this.raw.clearTimeout(unwrap_jso(handle)));
    }
    [_setInterval](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setInterval(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
    [_setTimeout](handler, timeout) {
      return dart.as(wrap_jso(this.raw.setTimeout(unwrap_jso(handler), unwrap_jso(timeout))), core.int);
    }
  }
  dart.setSignature(_WindowTimers, {
    constructors: () => ({_: [_WindowTimers, []]}),
    methods: () => ({
      [_clearInterval]: [dart.void, [core.int]],
      [_clearTimeout]: [dart.void, [core.int]],
      [_setInterval]: [core.int, [core.Object, core.int]],
      [_setTimeout]: [core.int, [core.Object, core.int]]
    })
  });
  _WindowTimers[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WindowTimers')), dart.const(new _metadata.Experimental())];
  class _WorkerLocation extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_WorkerLocation() {
      return _WorkerLocation._internalWrap();
    }
    static _internalWrap() {
      return new _WorkerLocation.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  _WorkerLocation[dart.implements] = () => [UrlUtilsReadOnly];
  dart.defineNamedConstructor(_WorkerLocation, 'internal_');
  dart.setSignature(_WorkerLocation, {
    constructors: () => ({
      _: [_WorkerLocation, []],
      _internalWrap: [_WorkerLocation, []],
      internal_: [_WorkerLocation, []]
    }),
    statics: () => ({internalCreate_WorkerLocation: [_WorkerLocation, []]}),
    names: ['internalCreate_WorkerLocation']
  });
  _WorkerLocation[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WorkerLocation')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("WorkerLocation"))];
  class _WorkerNavigator extends DartHtmlDomObject {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_WorkerNavigator() {
      return _WorkerNavigator._internalWrap();
    }
    static _internalWrap() {
      return new _WorkerNavigator.internal_();
    }
    internal_() {
      super.DartHtmlDomObject();
    }
    ['=='](other) {
      return dart.equals(unwrap_jso(other), unwrap_jso(this)) || dart.notNull(core.identical(this, other));
    }
    get hashCode() {
      return dart.hashCode(unwrap_jso(this));
    }
  }
  _WorkerNavigator[dart.implements] = () => [NavigatorCpu, NavigatorOnLine, NavigatorID];
  dart.defineNamedConstructor(_WorkerNavigator, 'internal_');
  dart.setSignature(_WorkerNavigator, {
    constructors: () => ({
      _: [_WorkerNavigator, []],
      _internalWrap: [_WorkerNavigator, []],
      internal_: [_WorkerNavigator, []]
    }),
    statics: () => ({internalCreate_WorkerNavigator: [_WorkerNavigator, []]}),
    names: ['internalCreate_WorkerNavigator']
  });
  _WorkerNavigator[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('WorkerNavigator')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("WorkerNavigator"))];
  class _XMLHttpRequestProgressEvent extends ProgressEvent {
    static _() {
      dart.throw(new core.UnsupportedError("Not supported"));
    }
    static internalCreate_XMLHttpRequestProgressEvent() {
      return _XMLHttpRequestProgressEvent._internalWrap();
    }
    static _internalWrap() {
      return new _XMLHttpRequestProgressEvent.internal_();
    }
    internal_() {
      super.internal_();
    }
  }
  dart.defineNamedConstructor(_XMLHttpRequestProgressEvent, 'internal_');
  dart.setSignature(_XMLHttpRequestProgressEvent, {
    constructors: () => ({
      _: [_XMLHttpRequestProgressEvent, []],
      _internalWrap: [_XMLHttpRequestProgressEvent, []],
      internal_: [_XMLHttpRequestProgressEvent, []]
    }),
    statics: () => ({internalCreate_XMLHttpRequestProgressEvent: [_XMLHttpRequestProgressEvent, []]}),
    names: ['internalCreate_XMLHttpRequestProgressEvent']
  });
  _XMLHttpRequestProgressEvent[dart.metadata] = () => [dart.const(new _metadata.DocsEditable()), dart.const(new _metadata.DomName('XMLHttpRequestProgressEvent')), dart.const(new _metadata.Experimental()), dart.const(new _js_helper.Native("XMLHttpRequestProgressEvent"))];
  const _matches = Symbol('_matches');
  class _AttributeMap extends core.Object {
    _AttributeMap(element) {
      this[_element] = element;
    }
    addAll(other) {
      other.forEach(dart.fn(((k, v) => {
        this.set(dart.as(k, core.String), dart.as(v, core.String));
      }).bind(this)));
    }
    containsValue(value) {
      for (let v of this.values) {
        if (value == v) {
          return true;
        }
      }
      return false;
    }
    putIfAbsent(key, ifAbsent) {
      if (!dart.notNull(this.containsKey(key))) {
        this.set(key, ifAbsent());
      }
      return this.get(key);
    }
    clear() {
      for (let key of this.keys) {
        this.remove(key);
      }
    }
    forEach(f) {
      for (let key of this.keys) {
        let value = this.get(key);
        f(key, value);
      }
    }
    get keys() {
      let attributes = this[_element][_attributes];
      let keys = core.List$(core.String).new();
      for (let i = 0, len = attributes.length; dart.notNull(i) < dart.notNull(len); i = dart.notNull(i) + 1) {
        if (dart.notNull(this[_matches](attributes.get(i)))) {
          keys[dartx.add](dart.as(dart.dload(attributes.get(i), 'name'), core.String));
        }
      }
      return keys;
    }
    get values() {
      let attributes = this[_element][_attributes];
      let values = core.List$(core.String).new();
      for (let i = 0, len = attributes.length; dart.notNull(i) < dart.notNull(len); i = dart.notNull(i) + 1) {
        if (dart.notNull(this[_matches](attributes.get(i)))) {
          values[dartx.add](dart.as(dart.dload(attributes.get(i), 'value'), core.String));
        }
      }
      return values;
    }
    get isEmpty() {
      return this.length == 0;
    }
    get isNotEmpty() {
      return !dart.notNull(this.isEmpty);
    }
  }
  _AttributeMap[dart.implements] = () => [core.Map$(core.String, core.String)];
  dart.setSignature(_AttributeMap, {
    constructors: () => ({_AttributeMap: [_AttributeMap, [Element]]}),
    methods: () => ({
      addAll: [dart.void, [core.Map$(core.String, core.String)]],
      containsValue: [core.bool, [core.String]],
      putIfAbsent: [core.String, [core.String, dart.functionType(core.String, [])]],
      clear: [dart.void, []],
      forEach: [dart.void, [dart.functionType(dart.void, [core.String, core.String])]]
    })
  });
  class _ElementAttributeMap extends _AttributeMap {
    _ElementAttributeMap(element) {
      super._AttributeMap(element);
    }
    containsKey(key) {
      return this[_element][_hasAttribute](key);
    }
    get(key) {
      return this[_element].getAttribute(key);
    }
    set(key, value) {
      this[_element].setAttribute(key, value);
      return value;
    }
    remove(key) {
      let value = this[_element].getAttribute(key);
      this[_element][_removeAttribute](key);
      return value;
    }
    get length() {
      return this.keys[dartx.length];
    }
    [_matches](node) {
      return node[_namespaceUri] == null;
    }
  }
  dart.setSignature(_ElementAttributeMap, {
    constructors: () => ({_ElementAttributeMap: [_ElementAttributeMap, [Element]]}),
    methods: () => ({
      containsKey: [core.bool, [core.String]],
      get: [core.String, [core.String]],
      set: [dart.void, [core.String, core.String]],
      remove: [core.String, [core.String]],
      [_matches]: [core.bool, [Node]]
    })
  });
  const _namespace = Symbol('_namespace');
  class _NamespacedAttributeMap extends _AttributeMap {
    _NamespacedAttributeMap(element, namespace) {
      this[_namespace] = namespace;
      super._AttributeMap(element);
    }
    containsKey(key) {
      return this[_element][_hasAttributeNS](this[_namespace], key);
    }
    get(key) {
      return this[_element].getAttributeNS(this[_namespace], key);
    }
    set(key, value) {
      this[_element].setAttributeNS(this[_namespace], key, value);
      return value;
    }
    remove(key) {
      let value = this.get(key);
      this[_element][_removeAttributeNS](this[_namespace], key);
      return value;
    }
    get length() {
      return this.keys[dartx.length];
    }
    [_matches](node) {
      return node[_namespaceUri] == this[_namespace];
    }
  }
  dart.setSignature(_NamespacedAttributeMap, {
    constructors: () => ({_NamespacedAttributeMap: [_NamespacedAttributeMap, [Element, core.String]]}),
    methods: () => ({
      containsKey: [core.bool, [core.String]],
      get: [core.String, [core.String]],
      set: [dart.void, [core.String, core.String]],
      remove: [core.String, [core.String]],
      [_matches]: [core.bool, [Node]]
    })
  });
  const _attr = Symbol('_attr');
  const _strip = Symbol('_strip');
  const _toHyphenedName = Symbol('_toHyphenedName');
  const _toCamelCase = Symbol('_toCamelCase');
  class _DataAttributeMap extends core.Object {
    _DataAttributeMap(attributes) {
      this[_attributes] = attributes;
    }
    addAll(other) {
      other.forEach(dart.fn(((k, v) => {
        this.set(dart.as(k, core.String), dart.as(v, core.String));
      }).bind(this)));
    }
    containsValue(value) {
      return this.values[dartx.any](dart.fn(v => dart.equals(v, value), core.bool, [dart.dynamic]));
    }
    containsKey(key) {
      return this[_attributes].containsKey(this[_attr](key));
    }
    get(key) {
      return this[_attributes].get(this[_attr](key));
    }
    set(key, value) {
      this[_attributes].set(this[_attr](key), value);
      return value;
    }
    putIfAbsent(key, ifAbsent) {
      return this[_attributes].putIfAbsent(this[_attr](key), ifAbsent);
    }
    remove(key) {
      return this[_attributes].remove(this[_attr](key));
    }
    clear() {
      for (let key of this.keys) {
        this.remove(key);
      }
    }
    forEach(f) {
      this[_attributes].forEach(dart.fn(((key, value) => {
        if (dart.notNull(this[_matches](key))) {
          f(this[_strip](key), value);
        }
      }).bind(this), dart.dynamic, [core.String, core.String]));
    }
    get keys() {
      let keys = core.List$(core.String).new();
      this[_attributes].forEach(dart.fn(((key, value) => {
        if (dart.notNull(this[_matches](key))) {
          keys[dartx.add](this[_strip](key));
        }
      }).bind(this), dart.dynamic, [core.String, core.String]));
      return keys;
    }
    get values() {
      let values = core.List$(core.String).new();
      this[_attributes].forEach(dart.fn(((key, value) => {
        if (dart.notNull(this[_matches](key))) {
          values[dartx.add](value);
        }
      }).bind(this), dart.dynamic, [core.String, core.String]));
      return values;
    }
    get length() {
      return this.keys[dartx.length];
    }
    get isEmpty() {
      return this.length == 0;
    }
    get isNotEmpty() {
      return !dart.notNull(this.isEmpty);
    }
    [_attr](key) {
      return `data-${this[_toHyphenedName](key)}`;
    }
    [_matches](key) {
      return key[dartx.startsWith]('data-');
    }
    [_strip](key) {
      return this[_toCamelCase](key[dartx.substring](5));
    }
    [_toCamelCase](hyphenedName, opts) {
      let startUppercase = opts && 'startUppercase' in opts ? opts.startUppercase : false;
      let segments = hyphenedName[dartx.split]('-');
      let start = dart.notNull(startUppercase) ? 0 : 1;
      for (let i = start; dart.notNull(i) < dart.notNull(segments[dartx.length]); i = dart.notNull(i) + 1) {
        let segment = segments[dartx.get](i);
        if (dart.notNull(segment[dartx.length]) > 0) {
          segments[dartx.set](i, `${segment[dartx.get](0)[dartx.toUpperCase]()}${segment[dartx.substring](1)}`);
        }
      }
      return segments[dartx.join]('');
    }
    [_toHyphenedName](word) {
      let sb = new core.StringBuffer();
      for (let i = 0; dart.notNull(i) < dart.notNull(word[dartx.length]); i = dart.notNull(i) + 1) {
        let lower = word[dartx.get](i)[dartx.toLowerCase]();
        if (word[dartx.get](i) != lower && dart.notNull(i) > 0)
          sb.write('-');
        sb.write(lower);
      }
      return dart.toString(sb);
    }
  }
  _DataAttributeMap[dart.implements] = () => [core.Map$(core.String, core.String)];
  dart.setSignature(_DataAttributeMap, {
    constructors: () => ({_DataAttributeMap: [_DataAttributeMap, [core.Map$(core.String, core.String)]]}),
    methods: () => ({
      addAll: [dart.void, [core.Map$(core.String, core.String)]],
      containsValue: [core.bool, [core.String]],
      containsKey: [core.bool, [core.String]],
      get: [core.String, [core.String]],
      set: [dart.void, [core.String, core.String]],
      putIfAbsent: [core.String, [core.String, dart.functionType(core.String, [])]],
      remove: [core.String, [core.String]],
      clear: [dart.void, []],
      forEach: [dart.void, [dart.functionType(dart.void, [core.String, core.String])]],
      [_attr]: [core.String, [core.String]],
      [_matches]: [core.bool, [core.String]],
      [_strip]: [core.String, [core.String]],
      [_toCamelCase]: [core.String, [core.String], {startUppercase: core.bool}],
      [_toHyphenedName]: [core.String, [core.String]]
    })
  });
  class CanvasImageSource extends core.Object {}
  class WindowBase extends core.Object {}
  WindowBase[dart.implements] = () => [EventTarget];
  class LocationBase extends core.Object {}
  class HistoryBase extends core.Object {}
  class CssClassSet extends core.Object {}
  CssClassSet[dart.implements] = () => [core.Set$(core.String)];
  const _addOrSubtractToBoxModel = Symbol('_addOrSubtractToBoxModel');
  class CssRect extends math.MutableRectangle$(core.num) {
    CssRect(element) {
      this[_element] = element;
      super.MutableRectangle(0, 0, 0, 0);
    }
    set height(newHeight) {
      dart.throw(new core.UnsupportedError("Can only set height for content rect."));
    }
    set width(newWidth) {
      dart.throw(new core.UnsupportedError("Can only set width for content rect."));
    }
    [_addOrSubtractToBoxModel](dimensions, augmentingMeasurement) {
      let styles = this[_element].getComputedStyle();
      let val = 0;
      for (let measurement of dimensions) {
        if (augmentingMeasurement == exports._MARGIN) {
          val = dart.notNull(val) + dart.notNull(dart.asInt(new Dimension.css(styles.getPropertyValue(`${augmentingMeasurement}-${measurement}`)).value));
        }
        if (augmentingMeasurement == exports._CONTENT) {
          val = dart.notNull(val) - dart.notNull(dart.asInt(new Dimension.css(styles.getPropertyValue(`${exports._PADDING}-${measurement}`)).value));
        }
        if (augmentingMeasurement != exports._MARGIN) {
          val = dart.notNull(val) - dart.notNull(dart.asInt(new Dimension.css(styles.getPropertyValue(`border-${measurement}-width`)).value));
        }
      }
      return val;
    }
  }
  dart.setSignature(CssRect, {
    constructors: () => ({CssRect: [CssRect, [Element]]}),
    methods: () => ({[_addOrSubtractToBoxModel]: [core.num, [core.List$(core.String), core.String]]})
  });
  class _ContentCssRect extends CssRect {
    _ContentCssRect(element) {
      super.CssRect(dart.as(element, Element));
    }
    get height() {
      return dart.notNull(this[_element].offsetHeight) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._HEIGHT, core.List$(core.String)), exports._CONTENT));
    }
    get width() {
      return dart.notNull(this[_element].offsetWidth) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._WIDTH, core.List$(core.String)), exports._CONTENT));
    }
    set height(newHeight) {
      if (dart.is(newHeight, Dimension)) {
        if (dart.notNull(dart.as(dart.dsend(dart.dload(newHeight, dartx.value), '<', 0), core.bool)))
          newHeight = new Dimension.px(0);
        this[_element].style.height = dart.toString(newHeight);
      } else {
        if (dart.notNull(newHeight) < 0)
          newHeight = 0;
        this[_element].style.height = `${newHeight}px`;
      }
    }
    set width(newWidth) {
      if (dart.is(newWidth, Dimension)) {
        if (dart.notNull(dart.as(dart.dsend(dart.dload(newWidth, dartx.value), '<', 0), core.bool)))
          newWidth = new Dimension.px(0);
        this[_element].style.width = dart.toString(newWidth);
      } else {
        if (dart.notNull(newWidth) < 0)
          newWidth = 0;
        this[_element].style.width = `${newWidth}px`;
      }
    }
    get left() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().left, '-', this[_addOrSubtractToBoxModel](dart.list(['left'], core.String), exports._CONTENT)), core.num);
    }
    get top() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().top, '-', this[_addOrSubtractToBoxModel](dart.list(['top'], core.String), exports._CONTENT)), core.num);
    }
  }
  dart.setSignature(_ContentCssRect, {
    constructors: () => ({_ContentCssRect: [_ContentCssRect, [dart.dynamic]]})
  });
  const _elementList = Symbol('_elementList');
  class _ContentCssListRect extends _ContentCssRect {
    _ContentCssListRect(elementList) {
      this[_elementList] = null;
      super._ContentCssRect(dart.dload(elementList, 'first'));
      this[_elementList] = dart.as(elementList, core.List$(Element));
    }
    set height(newHeight) {
      this[_elementList][dartx.forEach](dart.fn(e => dart.dput(dart.dload(e, 'contentEdge'), 'height', newHeight), core.num, [dart.dynamic]));
    }
    set width(newWidth) {
      this[_elementList][dartx.forEach](dart.fn(e => dart.dput(dart.dload(e, 'contentEdge'), 'width', newWidth), core.num, [dart.dynamic]));
    }
  }
  dart.setSignature(_ContentCssListRect, {
    constructors: () => ({_ContentCssListRect: [_ContentCssListRect, [dart.dynamic]]})
  });
  class _PaddingCssRect extends CssRect {
    _PaddingCssRect(element) {
      super.CssRect(dart.as(element, Element));
    }
    get height() {
      return dart.notNull(this[_element].offsetHeight) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._HEIGHT, core.List$(core.String)), exports._PADDING));
    }
    get width() {
      return dart.notNull(this[_element].offsetWidth) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._WIDTH, core.List$(core.String)), exports._PADDING));
    }
    get left() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().left, '-', this[_addOrSubtractToBoxModel](dart.list(['left'], core.String), exports._PADDING)), core.num);
    }
    get top() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().top, '-', this[_addOrSubtractToBoxModel](dart.list(['top'], core.String), exports._PADDING)), core.num);
    }
  }
  dart.setSignature(_PaddingCssRect, {
    constructors: () => ({_PaddingCssRect: [_PaddingCssRect, [dart.dynamic]]})
  });
  class _BorderCssRect extends CssRect {
    _BorderCssRect(element) {
      super.CssRect(dart.as(element, Element));
    }
    get height() {
      return this[_element].offsetHeight;
    }
    get width() {
      return this[_element].offsetWidth;
    }
    get left() {
      return dart.as(this[_element].getBoundingClientRect().left, core.num);
    }
    get top() {
      return dart.as(this[_element].getBoundingClientRect().top, core.num);
    }
  }
  dart.setSignature(_BorderCssRect, {
    constructors: () => ({_BorderCssRect: [_BorderCssRect, [dart.dynamic]]})
  });
  class _MarginCssRect extends CssRect {
    _MarginCssRect(element) {
      super.CssRect(dart.as(element, Element));
    }
    get height() {
      return dart.notNull(this[_element].offsetHeight) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._HEIGHT, core.List$(core.String)), exports._MARGIN));
    }
    get width() {
      return dart.notNull(this[_element].offsetWidth) + dart.notNull(this[_addOrSubtractToBoxModel](dart.as(exports._WIDTH, core.List$(core.String)), exports._MARGIN));
    }
    get left() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().left, '-', this[_addOrSubtractToBoxModel](dart.list(['left'], core.String), exports._MARGIN)), core.num);
    }
    get top() {
      return dart.as(dart.dsend(this[_element].getBoundingClientRect().top, '-', this[_addOrSubtractToBoxModel](dart.list(['top'], core.String), exports._MARGIN)), core.num);
    }
  }
  dart.setSignature(_MarginCssRect, {
    constructors: () => ({_MarginCssRect: [_MarginCssRect, [dart.dynamic]]})
  });
  exports._CONTENT = 'content';
  exports._PADDING = 'padding';
  exports._MARGIN = 'margin';
  dart.defineLazyProperties(exports, {
    get _HEIGHT() {
      return ['top', 'bottom'];
    },
    get _WIDTH() {
      return ['right', 'left'];
    }
  });
  const _sets = Symbol('_sets');
  dart.defineLazyClass(exports, {
    get _MultiElementCssClassSet() {
      class _MultiElementCssClassSet extends html_common.CssClassSetImpl {
        static new(elements) {
          return new exports._MultiElementCssClassSet._(elements, dart.as(elements[dartx.map](dart.fn(e => e.classes, CssClassSet, [Element]))[dartx.toList](), core.List$(html_common.CssClassSetImpl)));
        }
        _(elementIterable, sets) {
          this[_elementIterable] = elementIterable;
          this[_sets] = sets;
        }
        readClasses() {
          let s = collection.LinkedHashSet$(core.String).new();
          this[_sets][dartx.forEach](dart.fn(e => s.addAll(e.readClasses()), dart.void, [html_common.CssClassSetImpl]));
          return s;
        }
        writeClasses(s) {
          let classes = s.join(' ');
          for (let e of this[_elementIterable]) {
            e.className = classes;
          }
        }
        modify(f) {
          this[_sets][dartx.forEach](dart.fn(e => e.modify(f), dart.dynamic, [html_common.CssClassSetImpl]));
        }
        toggle(value, shouldAdd) {
          if (shouldAdd === void 0)
            shouldAdd = null;
          return this[_sets][dartx.fold](false, dart.fn((changed, e) => dart.notNull(e.toggle(value, shouldAdd)) || dart.notNull(changed), core.bool, [core.bool, html_common.CssClassSetImpl]));
        }
        remove(value) {
          return this[_sets][dartx.fold](false, dart.fn((changed, e) => dart.notNull(e.remove(value)) || dart.notNull(changed), core.bool, [core.bool, html_common.CssClassSetImpl]));
        }
      }
      dart.defineNamedConstructor(_MultiElementCssClassSet, '_');
      dart.setSignature(_MultiElementCssClassSet, {
        constructors: () => ({
          new: [exports._MultiElementCssClassSet, [core.Iterable$(Element)]],
          _: [exports._MultiElementCssClassSet, [core.Iterable$(Element), core.List$(html_common.CssClassSetImpl)]]
        }),
        methods: () => ({
          readClasses: [core.Set$(core.String), []],
          writeClasses: [dart.void, [core.Set$(core.String)]]
        })
      });
      return _MultiElementCssClassSet;
    }
  });
  dart.defineLazyClass(exports, {
    get _ElementCssClassSet() {
      class _ElementCssClassSet extends html_common.CssClassSetImpl {
        _ElementCssClassSet(element) {
          this[_element] = element;
        }
        readClasses() {
          let s = collection.LinkedHashSet$(core.String).new();
          let classname = this[_element].className;
          for (let name of classname[dartx.split](' ')) {
            let trimmed = name[dartx.trim]();
            if (!dart.notNull(trimmed[dartx.isEmpty])) {
              s.add(trimmed);
            }
          }
          return s;
        }
        writeClasses(s) {
          this[_element].className = s.join(' ');
        }
        get length() {
          return exports._ElementCssClassSet._classListLength(exports._ElementCssClassSet._classListOf(this[_element]));
        }
        get isEmpty() {
          return this.length == 0;
        }
        get isNotEmpty() {
          return this.length != 0;
        }
        clear() {
          this[_element].className = '';
        }
        contains(value) {
          return exports._ElementCssClassSet._contains(this[_element], value);
        }
        add(value) {
          return exports._ElementCssClassSet._add(this[_element], value);
        }
        remove(value) {
          return typeof value == 'string' && dart.notNull(exports._ElementCssClassSet._remove(this[_element], value));
        }
        toggle(value, shouldAdd) {
          if (shouldAdd === void 0)
            shouldAdd = null;
          return exports._ElementCssClassSet._toggle(this[_element], value, shouldAdd);
        }
        addAll(iterable) {
          exports._ElementCssClassSet._addAll(this[_element], iterable);
        }
        removeAll(iterable) {
          exports._ElementCssClassSet._removeAll(this[_element], iterable);
        }
        retainAll(iterable) {
          exports._ElementCssClassSet._removeWhere(this[_element], dart.bind(iterable[dartx.toSet](), 'contains'), false);
        }
        removeWhere(test) {
          exports._ElementCssClassSet._removeWhere(this[_element], test, true);
        }
        retainWhere(test) {
          exports._ElementCssClassSet._removeWhere(this[_element], test, false);
        }
        static _contains(_element, value) {
          return typeof value == 'string' && dart.notNull(exports._ElementCssClassSet._classListContains(exports._ElementCssClassSet._classListOf(_element), value));
        }
        static _add(_element, value) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          let added = !dart.notNull(exports._ElementCssClassSet._classListContainsBeforeAddOrRemove(list, value));
          exports._ElementCssClassSet._classListAdd(list, value);
          return added;
        }
        static _remove(_element, value) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          let removed = exports._ElementCssClassSet._classListContainsBeforeAddOrRemove(list, value);
          exports._ElementCssClassSet._classListRemove(list, value);
          return removed;
        }
        static _toggle(_element, value, shouldAdd) {
          return shouldAdd == null ? exports._ElementCssClassSet._toggleDefault(_element, value) : exports._ElementCssClassSet._toggleOnOff(_element, value, shouldAdd);
        }
        static _toggleDefault(_element, value) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          return exports._ElementCssClassSet._classListToggle1(list, value);
        }
        static _toggleOnOff(_element, value, shouldAdd) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          if (dart.notNull(shouldAdd)) {
            exports._ElementCssClassSet._classListAdd(list, value);
            return true;
          } else {
            exports._ElementCssClassSet._classListRemove(list, value);
            return false;
          }
        }
        static _addAll(_element, iterable) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          for (let value of iterable) {
            exports._ElementCssClassSet._classListAdd(list, value);
          }
        }
        static _removeAll(_element, iterable) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          for (let value of iterable) {
            exports._ElementCssClassSet._classListRemove(list, value);
          }
        }
        static _removeWhere(_element, test, doRemove) {
          let list = exports._ElementCssClassSet._classListOf(_element);
          let i = 0;
          while (dart.notNull(i) < dart.notNull(exports._ElementCssClassSet._classListLength(list))) {
            let item = list.item(i);
            if (doRemove == test(item)) {
              exports._ElementCssClassSet._classListRemove(list, item);
            } else {
              i = dart.notNull(i) + 1;
            }
          }
        }
        static _classListOf(e) {
          return dart.as(wrap_jso(e.raw.classList), DomTokenList);
        }
        static _classListLength(list) {
          return dart.as(list.raw.length, core.int);
        }
        static _classListContains(list, value) {
          return dart.as(list.raw.contains(value), core.bool);
        }
        static _classListContainsBeforeAddOrRemove(list, value) {
          return dart.as(list.raw.contains(value), core.bool);
        }
        static _classListAdd(list, value) {
          list.raw.add(value);
        }
        static _classListRemove(list, value) {
          list.raw.remove(value);
        }
        static _classListToggle1(list, value) {
          return list.raw.toggle(value);
        }
        static _classListToggle2(list, value, shouldAdd) {
          return list.raw.toggle(value, shouldAdd);
        }
      }
      dart.setSignature(_ElementCssClassSet, {
        constructors: () => ({_ElementCssClassSet: [exports._ElementCssClassSet, [Element]]}),
        methods: () => ({
          readClasses: [core.Set$(core.String), []],
          writeClasses: [dart.void, [core.Set$(core.String)]],
          removeAll: [dart.void, [core.Iterable$(core.String)]],
          retainAll: [dart.void, [core.Iterable$(core.String)]]
        }),
        statics: () => ({
          _contains: [core.bool, [Element, core.Object]],
          _add: [core.bool, [Element, core.String]],
          _remove: [core.bool, [Element, core.String]],
          _toggle: [core.bool, [Element, core.String, core.bool]],
          _toggleDefault: [core.bool, [Element, core.String]],
          _toggleOnOff: [core.bool, [Element, core.String, core.bool]],
          _addAll: [dart.void, [Element, core.Iterable$(core.String)]],
          _removeAll: [dart.void, [Element, core.Iterable$(core.String)]],
          _removeWhere: [dart.void, [Element, dart.functionType(core.bool, [core.String]), core.bool]],
          _classListOf: [DomTokenList, [Element]],
          _classListLength: [core.int, [DomTokenList]],
          _classListContains: [core.bool, [DomTokenList, core.String]],
          _classListContainsBeforeAddOrRemove: [core.bool, [DomTokenList, core.String]],
          _classListAdd: [dart.void, [DomTokenList, core.String]],
          _classListRemove: [dart.void, [DomTokenList, core.String]],
          _classListToggle1: [core.bool, [DomTokenList, core.String]],
          _classListToggle2: [core.bool, [DomTokenList, core.String, core.bool]]
        }),
        names: ['_contains', '_add', '_remove', '_toggle', '_toggleDefault', '_toggleOnOff', '_addAll', '_removeAll', '_removeWhere', '_classListOf', '_classListLength', '_classListContains', '_classListContainsBeforeAddOrRemove', '_classListAdd', '_classListRemove', '_classListToggle1', '_classListToggle2']
      });
      dart.defineExtensionMembers(_ElementCssClassSet, ['contains', 'length', 'isEmpty', 'isNotEmpty']);
      return _ElementCssClassSet;
    }
  });
  const _value = Symbol('_value');
  const _unit = Symbol('_unit');
  class Dimension extends core.Object {
    percent(value) {
      this[_value] = value;
      this[_unit] = '%';
    }
    px(value) {
      this[_value] = value;
      this[_unit] = 'px';
    }
    pc(value) {
      this[_value] = value;
      this[_unit] = 'pc';
    }
    pt(value) {
      this[_value] = value;
      this[_unit] = 'pt';
    }
    inch(value) {
      this[_value] = value;
      this[_unit] = 'in';
    }
    cm(value) {
      this[_value] = value;
      this[_unit] = 'cm';
    }
    mm(value) {
      this[_value] = value;
      this[_unit] = 'mm';
    }
    em(value) {
      this[_value] = value;
      this[_unit] = 'em';
    }
    ex(value) {
      this[_value] = value;
      this[_unit] = 'ex';
    }
    css(cssValue) {
      this[_value] = null;
      this[_unit] = null;
      if (cssValue == '')
        cssValue = '0px';
      if (dart.notNull(cssValue[dartx.endsWith]('%'))) {
        this[_unit] = '%';
      } else {
        this[_unit] = cssValue[dartx.substring](dart.notNull(cssValue[dartx.length]) - 2);
      }
      if (dart.notNull(cssValue[dartx.contains]('.'))) {
        this[_value] = core.double.parse(cssValue[dartx.substring](0, dart.notNull(cssValue[dartx.length]) - dart.notNull(this[_unit][dartx.length])));
      } else {
        this[_value] = core.int.parse(cssValue[dartx.substring](0, dart.notNull(cssValue[dartx.length]) - dart.notNull(this[_unit][dartx.length])));
      }
    }
    toString() {
      return `${this[_value]}${this[_unit]}`;
    }
    get value() {
      return this[_value];
    }
  }
  dart.defineNamedConstructor(Dimension, 'percent');
  dart.defineNamedConstructor(Dimension, 'px');
  dart.defineNamedConstructor(Dimension, 'pc');
  dart.defineNamedConstructor(Dimension, 'pt');
  dart.defineNamedConstructor(Dimension, 'inch');
  dart.defineNamedConstructor(Dimension, 'cm');
  dart.defineNamedConstructor(Dimension, 'mm');
  dart.defineNamedConstructor(Dimension, 'em');
  dart.defineNamedConstructor(Dimension, 'ex');
  dart.defineNamedConstructor(Dimension, 'css');
  dart.setSignature(Dimension, {
    constructors: () => ({
      percent: [Dimension, [core.num]],
      px: [Dimension, [core.num]],
      pc: [Dimension, [core.num]],
      pt: [Dimension, [core.num]],
      inch: [Dimension, [core.num]],
      cm: [Dimension, [core.num]],
      mm: [Dimension, [core.num]],
      em: [Dimension, [core.num]],
      ex: [Dimension, [core.num]],
      css: [Dimension, [core.String]]
    })
  });
  Dimension[dart.metadata] = () => [dart.const(new _metadata.Experimental())];
  const EventListener = dart.typedef('EventListener', () => dart.functionType(dart.dynamic, [Event]));
  const ElementStream$ = dart.generic(function(T) {
    class ElementStream extends core.Object {}
    ElementStream[dart.implements] = () => [async.Stream$(T)];
    return ElementStream;
  });
  let ElementStream = ElementStream$();
  const _target = Symbol('_target');
  const _useCapture = Symbol('_useCapture');
  const _EventStream$ = dart.generic(function(T) {
    class _EventStream extends async.Stream$(T) {
      _EventStream(target, eventType, useCapture) {
        this[_target] = target;
        this[_eventType] = eventType;
        this[_useCapture] = useCapture;
        super.Stream();
      }
      asBroadcastStream(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        dart.as(onListen, dart.functionType(dart.void, [async.StreamSubscription]));
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        dart.as(onCancel, dart.functionType(dart.void, [async.StreamSubscription]));
        return this;
      }
      get isBroadcast() {
        return true;
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        return new (_EventStreamSubscription$(T))(this[_target], this[_eventType], onData, this[_useCapture]);
      }
    }
    dart.setSignature(_EventStream, {
      constructors: () => ({_EventStream: [_EventStream$(T), [EventTarget, core.String, core.bool]]}),
      methods: () => ({
        asBroadcastStream: [async.Stream$(T), [], {onListen: dart.functionType(dart.void, [async.StreamSubscription]), onCancel: dart.functionType(dart.void, [async.StreamSubscription])}],
        listen: [async.StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}]
      })
    });
    return _EventStream;
  });
  let _EventStream = _EventStream$();
  const _ElementEventStreamImpl$ = dart.generic(function(T) {
    class _ElementEventStreamImpl extends _EventStream$(T) {
      _ElementEventStreamImpl(target, eventType, useCapture) {
        super._EventStream(dart.as(target, EventTarget), dart.as(eventType, core.String), dart.as(useCapture, core.bool));
      }
      matches(selector) {
        return dart.as(this.where(dart.fn(event => dart.as(dart.dsend(dart.dload(event, 'target'), 'matchesWithAncestors', selector), core.bool), core.bool, [dart.dynamic])).map(dart.fn(e => {
          dart.dput(e, _selector, selector);
          return e;
        })), async.Stream$(T));
      }
      capture(onData) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        return new (_EventStreamSubscription$(T))(this[_target], this[_eventType], onData, true);
      }
    }
    _ElementEventStreamImpl[dart.implements] = () => [ElementStream$(T)];
    dart.setSignature(_ElementEventStreamImpl, {
      constructors: () => ({_ElementEventStreamImpl: [_ElementEventStreamImpl$(T), [dart.dynamic, dart.dynamic, dart.dynamic]]}),
      methods: () => ({
        matches: [async.Stream$(T), [core.String]],
        capture: [async.StreamSubscription$(T), [dart.functionType(dart.void, [T])]]
      })
    });
    return _ElementEventStreamImpl;
  });
  let _ElementEventStreamImpl = _ElementEventStreamImpl$();
  const _targetList = Symbol('_targetList');
  const _ElementListEventStreamImpl$ = dart.generic(function(T) {
    class _ElementListEventStreamImpl extends async.Stream$(T) {
      _ElementListEventStreamImpl(targetList, eventType, useCapture) {
        this[_targetList] = targetList;
        this[_eventType] = eventType;
        this[_useCapture] = useCapture;
        super.Stream();
      }
      matches(selector) {
        return dart.as(this.where(dart.fn(event => dart.as(dart.dsend(dart.dload(event, 'target'), 'matchesWithAncestors', selector), core.bool), core.bool, [dart.dynamic])).map(dart.fn(e => {
          dart.dput(e, _selector, selector);
          return e;
        })), async.Stream$(T));
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        let pool = new _StreamPool.broadcast();
        for (let target of this[_targetList]) {
          pool.add(new _EventStream(target, this[_eventType], this[_useCapture]));
        }
        return dart.as(pool.stream.listen(onData, {onError: onError, onDone: onDone, cancelOnError: cancelOnError}), async.StreamSubscription$(T));
      }
      capture(onData) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let pool = new _StreamPool.broadcast();
        for (let target of this[_targetList]) {
          pool.add(new _EventStream(target, this[_eventType], true));
        }
        return dart.as(pool.stream.listen(onData), async.StreamSubscription$(T));
      }
      asBroadcastStream(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        dart.as(onListen, dart.functionType(dart.void, [async.StreamSubscription]));
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        dart.as(onCancel, dart.functionType(dart.void, [async.StreamSubscription]));
        return this;
      }
      get isBroadcast() {
        return true;
      }
    }
    _ElementListEventStreamImpl[dart.implements] = () => [ElementStream$(T)];
    dart.setSignature(_ElementListEventStreamImpl, {
      constructors: () => ({_ElementListEventStreamImpl: [_ElementListEventStreamImpl$(T), [core.Iterable$(Element), core.String, core.bool]]}),
      methods: () => ({
        matches: [async.Stream$(T), [core.String]],
        listen: [async.StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
        capture: [async.StreamSubscription$(T), [dart.functionType(dart.void, [T])]],
        asBroadcastStream: [async.Stream$(T), [], {onListen: dart.functionType(dart.void, [async.StreamSubscription]), onCancel: dart.functionType(dart.void, [async.StreamSubscription])}]
      })
    });
    return _ElementListEventStreamImpl;
  });
  let _ElementListEventStreamImpl = _ElementListEventStreamImpl$();
  const _onData = Symbol('_onData');
  const _pauseCount = Symbol('_pauseCount');
  const _tryResume = Symbol('_tryResume');
  const _canceled = Symbol('_canceled');
  const _unlisten = Symbol('_unlisten');
  const _EventStreamSubscription$ = dart.generic(function(T) {
    class _EventStreamSubscription extends async.StreamSubscription$(T) {
      _EventStreamSubscription(target, eventType, onData, useCapture) {
        this[_target] = target;
        this[_eventType] = eventType;
        this[_useCapture] = useCapture;
        this[_onData] = _wrapZone(dart.as(onData, __CastType2));
        this[_pauseCount] = 0;
        this[_tryResume]();
      }
      cancel() {
        if (dart.notNull(this[_canceled]))
          return null;
        this[_unlisten]();
        this[_target] = null;
        this[_onData] = null;
        return null;
      }
      get [_canceled]() {
        return this[_target] == null;
      }
      onData(handleData) {
        dart.as(handleData, dart.functionType(dart.void, [T]));
        if (dart.notNull(this[_canceled])) {
          dart.throw(new core.StateError("Subscription has been canceled."));
        }
        this[_unlisten]();
        this[_onData] = _wrapZone(handleData);
        this[_tryResume]();
      }
      onError(handleError) {}
      onDone(handleDone) {
        dart.as(handleDone, dart.functionType(dart.void, []));
      }
      pause(resumeSignal) {
        if (resumeSignal === void 0)
          resumeSignal = null;
        if (dart.notNull(this[_canceled]))
          return;
        this[_pauseCount] = dart.notNull(this[_pauseCount]) + 1;
        this[_unlisten]();
        if (resumeSignal != null) {
          resumeSignal.whenComplete(dart.bind(this, 'resume'));
        }
      }
      get isPaused() {
        return dart.notNull(this[_pauseCount]) > 0;
      }
      resume() {
        if (dart.notNull(this[_canceled]) || !dart.notNull(this.isPaused))
          return;
        this[_pauseCount] = dart.notNull(this[_pauseCount]) - 1;
        this[_tryResume]();
      }
      [_tryResume]() {
        if (this[_onData] != null && !dart.notNull(this.isPaused)) {
          this[_target].addEventListener(this[_eventType], dart.as(this[_onData], EventListener), this[_useCapture]);
        }
      }
      [_unlisten]() {
        if (this[_onData] != null) {
          this[_target].removeEventListener(this[_eventType], dart.as(this[_onData], EventListener), this[_useCapture]);
        }
      }
      asFuture(futureValue) {
        if (futureValue === void 0)
          futureValue = null;
        let completer = async.Completer.new();
        return completer.future;
      }
    }
    dart.setSignature(_EventStreamSubscription, {
      constructors: () => ({_EventStreamSubscription: [_EventStreamSubscription$(T), [EventTarget, core.String, dart.dynamic, core.bool]]}),
      methods: () => ({
        cancel: [async.Future, []],
        onData: [dart.void, [dart.functionType(dart.void, [T])]],
        onError: [dart.void, [core.Function]],
        onDone: [dart.void, [dart.functionType(dart.void, [])]],
        pause: [dart.void, [], [async.Future]],
        resume: [dart.void, []],
        [_tryResume]: [dart.void, []],
        [_unlisten]: [dart.void, []],
        asFuture: [async.Future, [], [dart.dynamic]]
      })
    });
    return _EventStreamSubscription;
  });
  let _EventStreamSubscription = _EventStreamSubscription$();
  const CustomStream$ = dart.generic(function(T) {
    class CustomStream extends core.Object {}
    CustomStream[dart.implements] = () => [async.Stream$(T)];
    return CustomStream;
  });
  let CustomStream = CustomStream$();
  const _streamController = Symbol('_streamController');
  const _type = Symbol('_type');
  const _CustomEventStreamImpl$ = dart.generic(function(T) {
    class _CustomEventStreamImpl extends async.Stream$(T) {
      _CustomEventStreamImpl(type) {
        this[_streamController] = null;
        this[_type] = null;
        super.Stream();
        this[_type] = type;
        this[_streamController] = async.StreamController$(T).broadcast({sync: true});
      }
      listen(onData, opts) {
        dart.as(onData, dart.functionType(dart.void, [T]));
        let onError = opts && 'onError' in opts ? opts.onError : null;
        let onDone = opts && 'onDone' in opts ? opts.onDone : null;
        dart.as(onDone, dart.functionType(dart.void, []));
        let cancelOnError = opts && 'cancelOnError' in opts ? opts.cancelOnError : null;
        return this[_streamController].stream.listen(onData, {onError: onError, onDone: onDone, cancelOnError: cancelOnError});
      }
      asBroadcastStream(opts) {
        let onListen = opts && 'onListen' in opts ? opts.onListen : null;
        dart.as(onListen, dart.functionType(dart.void, [async.StreamSubscription]));
        let onCancel = opts && 'onCancel' in opts ? opts.onCancel : null;
        dart.as(onCancel, dart.functionType(dart.void, [async.StreamSubscription]));
        return this[_streamController].stream;
      }
      get isBroadcast() {
        return true;
      }
      add(event) {
        dart.as(event, T);
        if (event.type == this[_type])
          this[_streamController].add(event);
      }
    }
    _CustomEventStreamImpl[dart.implements] = () => [CustomStream$(T)];
    dart.setSignature(_CustomEventStreamImpl, {
      constructors: () => ({_CustomEventStreamImpl: [_CustomEventStreamImpl$(T), [core.String]]}),
      methods: () => ({
        listen: [async.StreamSubscription$(T), [dart.functionType(dart.void, [T])], {onError: core.Function, onDone: dart.functionType(dart.void, []), cancelOnError: core.bool}],
        asBroadcastStream: [async.Stream$(T), [], {onListen: dart.functionType(dart.void, [async.StreamSubscription]), onCancel: dart.functionType(dart.void, [async.StreamSubscription])}],
        add: [dart.void, [T]]
      })
    });
    return _CustomEventStreamImpl;
  });
  let _CustomEventStreamImpl = _CustomEventStreamImpl$();
  class _WrappedEvent extends core.Object {
    _WrappedEvent(wrapped) {
      this.wrapped = wrapped;
      this[_selector] = null;
    }
    get bubbles() {
      return this.wrapped.bubbles;
    }
    get cancelable() {
      return this.wrapped.cancelable;
    }
    get clipboardData() {
      return this.wrapped.clipboardData;
    }
    get currentTarget() {
      return this.wrapped.currentTarget;
    }
    get defaultPrevented() {
      return this.wrapped.defaultPrevented;
    }
    get eventPhase() {
      return this.wrapped.eventPhase;
    }
    get target() {
      return this.wrapped.target;
    }
    get timeStamp() {
      return this.wrapped.timeStamp;
    }
    get type() {
      return this.wrapped.type;
    }
    [_initEvent](eventTypeArg, canBubbleArg, cancelableArg) {
      dart.throw(new core.UnsupportedError('Cannot initialize this Event.'));
    }
    preventDefault() {
      this.wrapped.preventDefault();
    }
    stopImmediatePropagation() {
      this.wrapped.stopImmediatePropagation();
    }
    stopPropagation() {
      this.wrapped.stopPropagation();
    }
    get matchingTarget() {
      if (this[_selector] == null) {
        dart.throw(new core.UnsupportedError('Cannot call matchingTarget if this Event did' + ' not arise as a result of event delegation.'));
      }
      let currentTarget = this.currentTarget;
      let target = this.target;
      let matchedTarget = null;
      do {
        if (dart.notNull(dart.as(dart.dcall(target.matches, this[_selector]), core.bool)))
          return dart.as(target, Element);
        target = dart.as(dart.dload(target, 'parent'), EventTarget);
      } while (target != null && !dart.equals(target, dart.dload(currentTarget, 'parent')));
      dart.throw(new core.StateError('No selector matched for populating matchedTarget.'));
    }
    get path() {
      return this.wrapped.path;
    }
    get [_get_currentTarget]() {
      return this.wrapped[_get_currentTarget];
    }
    get [_get_target]() {
      return this.wrapped[_get_target];
    }
  }
  _WrappedEvent[dart.implements] = () => [Event];
  dart.setSignature(_WrappedEvent, {
    constructors: () => ({_WrappedEvent: [_WrappedEvent, [Event]]}),
    methods: () => ({
      [_initEvent]: [dart.void, [core.String, core.bool, core.bool]],
      preventDefault: [dart.void, []],
      stopImmediatePropagation: [dart.void, []],
      stopPropagation: [dart.void, []]
    })
  });
  const _shadowKeyCode = Symbol('_shadowKeyCode');
  const _shadowCharCode = Symbol('_shadowCharCode');
  const _shadowAltKey = Symbol('_shadowAltKey');
  const _parent = Symbol('_parent');
  const _realKeyCode = Symbol('_realKeyCode');
  const _realCharCode = Symbol('_realCharCode');
  const _realAltKey = Symbol('_realAltKey');
  const _currentTarget = Symbol('_currentTarget');
  const _shadowKeyIdentifier = Symbol('_shadowKeyIdentifier');
  class KeyEvent extends _WrappedEvent {
    get keyCode() {
      return this[_shadowKeyCode];
    }
    get charCode() {
      return this.type == 'keypress' ? this[_shadowCharCode] : 0;
    }
    get altKey() {
      return this[_shadowAltKey];
    }
    get which() {
      return this.keyCode;
    }
    get [_realKeyCode]() {
      return this[_parent].keyCode;
    }
    get [_realCharCode]() {
      return this[_parent].charCode;
    }
    get [_realAltKey]() {
      return this[_parent].altKey;
    }
    static _makeRecord() {
      let interceptor = _foreign_helper.JS_INTERCEPTOR_CONSTANT(KeyboardEvent);
      return dart.dcall(/* Unimplemented unknown name */makeLeafDispatchRecord, interceptor);
    }
    wrap(parent) {
      this[_parent] = null;
      this[_shadowAltKey] = null;
      this[_shadowCharCode] = null;
      this[_shadowKeyCode] = null;
      this[_currentTarget] = null;
      super._WrappedEvent(parent);
      this[_parent] = parent;
      this[_shadowAltKey] = this[_realAltKey];
      this[_shadowCharCode] = this[_realCharCode];
      this[_shadowKeyCode] = this[_realKeyCode];
      this[_currentTarget] = this[_parent].currentTarget;
    }
    static new(type, opts) {
      let view = opts && 'view' in opts ? opts.view : null;
      let canBubble = opts && 'canBubble' in opts ? opts.canBubble : true;
      let cancelable = opts && 'cancelable' in opts ? opts.cancelable : true;
      let keyCode = opts && 'keyCode' in opts ? opts.keyCode : 0;
      let charCode = opts && 'charCode' in opts ? opts.charCode : 0;
      let keyLocation = opts && 'keyLocation' in opts ? opts.keyLocation : 1;
      let ctrlKey = opts && 'ctrlKey' in opts ? opts.ctrlKey : false;
      let altKey = opts && 'altKey' in opts ? opts.altKey : false;
      let shiftKey = opts && 'shiftKey' in opts ? opts.shiftKey : false;
      let metaKey = opts && 'metaKey' in opts ? opts.metaKey : false;
      let currentTarget = opts && 'currentTarget' in opts ? opts.currentTarget : null;
      if (view == null) {
        view = exports.window;
      }
      let eventObj = null;
      if (dart.notNull(KeyEvent.canUseDispatchEvent)) {
        eventObj = Event.eventType('Event', type, {canBubble: canBubble, cancelable: cancelable});
        eventObj.keyCode = keyCode;
        eventObj.which = keyCode;
        eventObj.charCode = charCode;
        eventObj.keyLocation = keyLocation;
        eventObj.ctrlKey = ctrlKey;
        eventObj.altKey = altKey;
        eventObj.shiftKey = shiftKey;
        eventObj.metaKey = metaKey;
      } else {
        eventObj = Event.eventType('KeyboardEvent', type, {canBubble: canBubble, cancelable: cancelable});
        Object.defineProperty(eventObj, 'keyCode', {
          get: function() {
            return this.keyCodeVal;
          }
        });
        Object.defineProperty(eventObj, 'which', {
          get: function() {
            return this.keyCodeVal;
          }
        });
        Object.defineProperty(eventObj, 'charCode', {
          get: function() {
            return this.charCodeVal;
          }
        });
        let keyIdentifier = KeyEvent._convertToHexString(charCode, keyCode);
        dart.dsend(eventObj, _initKeyboardEvent, type, canBubble, cancelable, view, keyIdentifier, keyLocation, ctrlKey, altKey, shiftKey, metaKey);
        eventObj.keyCodeVal = keyCode;
        eventObj.charCodeVal = charCode;
      }
      dart.dcall(/* Unimplemented unknown name */setDispatchProperty, eventObj, KeyEvent._keyboardEventDispatchRecord);
      let keyEvent = new KeyEvent.wrap(dart.as(eventObj, KeyboardEvent));
      if (keyEvent[_currentTarget] == null) {
        keyEvent[_currentTarget] = currentTarget == null ? exports.window : currentTarget;
      }
      return keyEvent;
    }
    static get canUseDispatchEvent() {
      return typeof document.body.dispatchEvent == "function" && document.body.dispatchEvent.length > 0;
    }
    get currentTarget() {
      return this[_currentTarget];
    }
    static _convertToHexString(charCode, keyCode) {
      if (charCode != -1) {
        let hex = charCode[dartx.toRadixString](16);
        let sb = new core.StringBuffer('U+');
        for (let i = 0; dart.notNull(i) < 4 - dart.notNull(hex[dartx.length]); i = dart.notNull(i) + 1)
          sb.write('0');
        sb.write(hex);
        return dart.toString(sb);
      } else {
        return KeyCode._convertKeyCodeToKeyName(keyCode);
      }
    }
    get clipboardData() {
      return this[_parent].clipboardData;
    }
    get ctrlKey() {
      return this[_parent].ctrlKey;
    }
    get detail() {
      return this[_parent].detail;
    }
    get keyLocation() {
      return this[_parent].keyLocation;
    }
    get layer() {
      return this[_parent].layer;
    }
    get metaKey() {
      return this[_parent].metaKey;
    }
    get page() {
      return this[_parent].page;
    }
    get shiftKey() {
      return this[_parent].shiftKey;
    }
    get view() {
      return dart.as(this[_parent].view, Window);
    }
    [_initUIEvent](type, canBubble, cancelable, view, detail) {
      dart.throw(new core.UnsupportedError("Cannot initialize a UI Event from a KeyEvent."));
    }
    get [_shadowKeyIdentifier]() {
      return this[_parent].keyIdentifier;
    }
    get [_charCode]() {
      return this.charCode;
    }
    get [_keyCode]() {
      return this.keyCode;
    }
    get [_keyIdentifier]() {
      dart.throw(new core.UnsupportedError("keyIdentifier is unsupported."));
    }
    [_initKeyboardEvent](type, canBubble, cancelable, view, keyIdentifier, keyLocation, ctrlKey, altKey, shiftKey, metaKey) {
      dart.throw(new core.UnsupportedError("Cannot initialize a KeyboardEvent from a KeyEvent."));
    }
    get [_layerX]() {
      return dart.throw(new core.UnsupportedError('Not applicable to KeyEvent'));
    }
    get [_layerY]() {
      return dart.throw(new core.UnsupportedError('Not applicable to KeyEvent'));
    }
    get [_pageX]() {
      return dart.throw(new core.UnsupportedError('Not applicable to KeyEvent'));
    }
    get [_pageY]() {
      return dart.throw(new core.UnsupportedError('Not applicable to KeyEvent'));
    }
    getModifierState(keyArgument) {
      return dart.throw(new core.UnimplementedError());
    }
    get location() {
      return dart.throw(new core.UnimplementedError());
    }
    get repeat() {
      return dart.throw(new core.UnimplementedError());
    }
    get [_get_view]() {
      return dart.throw(new core.UnimplementedError());
    }
  }
  KeyEvent[dart.implements] = () => [KeyboardEvent];
  dart.defineNamedConstructor(KeyEvent, 'wrap');
  dart.setSignature(KeyEvent, {
    constructors: () => ({
      wrap: [KeyEvent, [KeyboardEvent]],
      new: [KeyEvent, [core.String], {view: Window, canBubble: core.bool, cancelable: core.bool, keyCode: core.int, charCode: core.int, keyLocation: core.int, ctrlKey: core.bool, altKey: core.bool, shiftKey: core.bool, metaKey: core.bool, currentTarget: EventTarget}]
    }),
    methods: () => ({
      [_initUIEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.int]],
      [_initKeyboardEvent]: [dart.void, [core.String, core.bool, core.bool, Window, core.String, core.int, core.bool, core.bool, core.bool, core.bool]],
      getModifierState: [core.bool, [core.String]]
    }),
    statics: () => ({
      _makeRecord: [dart.dynamic, []],
      _convertToHexString: [core.String, [core.int, core.int]]
    }),
    names: ['_makeRecord', '_convertToHexString']
  });
  KeyEvent[dart.metadata] = () => [dart.const(new _metadata.Experimental())];
  class _CustomKeyEventStreamImpl extends _CustomEventStreamImpl$(KeyEvent) {
    _CustomKeyEventStreamImpl(type) {
      super._CustomEventStreamImpl(type);
    }
    add(event) {
      if (event.type == this[_type]) {
        event.currentTarget.dispatchEvent(event[_parent]);
        this[_streamController].add(event);
      }
    }
  }
  _CustomKeyEventStreamImpl[dart.implements] = () => [CustomStream$(KeyEvent)];
  dart.setSignature(_CustomKeyEventStreamImpl, {
    constructors: () => ({_CustomKeyEventStreamImpl: [_CustomKeyEventStreamImpl, [core.String]]}),
    methods: () => ({add: [dart.void, [KeyEvent]]})
  });
  const _subscriptions = Symbol('_subscriptions');
  const _controller = Symbol('_controller');
  const _StreamPool$ = dart.generic(function(T) {
    class _StreamPool extends core.Object {
      broadcast() {
        this[_subscriptions] = core.Map$(async.Stream$(T), async.StreamSubscription$(T)).new();
        this[_controller] = null;
        this[_controller] = async.StreamController$(T).broadcast({sync: true, onCancel: dart.bind(this, 'close')});
      }
      get stream() {
        return this[_controller].stream;
      }
      add(stream) {
        dart.as(stream, async.Stream$(T));
        if (dart.notNull(this[_subscriptions].containsKey(stream)))
          return;
        this[_subscriptions].set(stream, stream.listen(dart.bind(this[_controller], 'add'), {onError: dart.bind(this[_controller], 'addError'), onDone: dart.fn((() => this.remove(stream)).bind(this), dart.void, [])}));
      }
      remove(stream) {
        dart.as(stream, async.Stream$(T));
        let subscription = this[_subscriptions].remove(stream);
        if (subscription != null)
          subscription.cancel();
      }
      close() {
        for (let subscription of this[_subscriptions].values) {
          subscription.cancel();
        }
        this[_subscriptions].clear();
        this[_controller].close();
      }
    }
    dart.defineNamedConstructor(_StreamPool, 'broadcast');
    dart.setSignature(_StreamPool, {
      constructors: () => ({broadcast: [_StreamPool$(T), []]}),
      methods: () => ({
        add: [dart.void, [async.Stream$(T)]],
        remove: [dart.void, [async.Stream$(T)]],
        close: [dart.void, []]
      })
    });
    return _StreamPool;
  });
  let _StreamPool = _StreamPool$();
  class _Html5NodeValidator extends core.Object {
    _Html5NodeValidator(opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      this.uriPolicy = uriPolicy != null ? uriPolicy : UriPolicy.new();
      if (dart.notNull(_Html5NodeValidator._attributeValidators.isEmpty)) {
        for (let attr of _Html5NodeValidator._standardAttributes) {
          _Html5NodeValidator._attributeValidators.set(attr, _Html5NodeValidator._standardAttributeValidator);
        }
        for (let attr of _Html5NodeValidator._uriAttributes) {
          _Html5NodeValidator._attributeValidators.set(attr, _Html5NodeValidator._uriAttributeValidator);
        }
      }
    }
    allowsElement(element) {
      return _Html5NodeValidator._allowedElements.contains(Element._safeTagName(element));
    }
    allowsAttribute(element, attributeName, value) {
      let tagName = Element._safeTagName(element);
      let validator = _Html5NodeValidator._attributeValidators.get(`${tagName}::${attributeName}`);
      if (validator == null) {
        validator = _Html5NodeValidator._attributeValidators.get(`*::${attributeName}`);
      }
      if (validator == null) {
        return false;
      }
      return dart.as(dart.dcall(validator, element, attributeName, value, this), core.bool);
    }
    static _standardAttributeValidator(element, attributeName, value, context) {
      return true;
    }
    static _uriAttributeValidator(element, attributeName, value, context) {
      return context.uriPolicy.allowsUri(value);
    }
  }
  _Html5NodeValidator[dart.implements] = () => [NodeValidator];
  dart.setSignature(_Html5NodeValidator, {
    constructors: () => ({_Html5NodeValidator: [_Html5NodeValidator, [], {uriPolicy: UriPolicy}]}),
    methods: () => ({
      allowsElement: [core.bool, [Element]],
      allowsAttribute: [core.bool, [Element, core.String, core.String]]
    }),
    statics: () => ({
      _standardAttributeValidator: [core.bool, [Element, core.String, core.String, _Html5NodeValidator]],
      _uriAttributeValidator: [core.bool, [Element, core.String, core.String, _Html5NodeValidator]]
    }),
    names: ['_standardAttributeValidator', '_uriAttributeValidator']
  });
  _Html5NodeValidator._standardAttributes = dart.const(dart.list(['*::class', '*::dir', '*::draggable', '*::hidden', '*::id', '*::inert', '*::itemprop', '*::itemref', '*::itemscope', '*::lang', '*::spellcheck', '*::title', '*::translate', 'A::accesskey', 'A::coords', 'A::hreflang', 'A::name', 'A::shape', 'A::tabindex', 'A::target', 'A::type', 'AREA::accesskey', 'AREA::alt', 'AREA::coords', 'AREA::nohref', 'AREA::shape', 'AREA::tabindex', 'AREA::target', 'AUDIO::controls', 'AUDIO::loop', 'AUDIO::mediagroup', 'AUDIO::muted', 'AUDIO::preload', 'BDO::dir', 'BODY::alink', 'BODY::bgcolor', 'BODY::link', 'BODY::text', 'BODY::vlink', 'BR::clear', 'BUTTON::accesskey', 'BUTTON::disabled', 'BUTTON::name', 'BUTTON::tabindex', 'BUTTON::type', 'BUTTON::value', 'CANVAS::height', 'CANVAS::width', 'CAPTION::align', 'COL::align', 'COL::char', 'COL::charoff', 'COL::span', 'COL::valign', 'COL::width', 'COLGROUP::align', 'COLGROUP::char', 'COLGROUP::charoff', 'COLGROUP::span', 'COLGROUP::valign', 'COLGROUP::width', 'COMMAND::checked', 'COMMAND::command', 'COMMAND::disabled', 'COMMAND::label', 'COMMAND::radiogroup', 'COMMAND::type', 'DATA::value', 'DEL::datetime', 'DETAILS::open', 'DIR::compact', 'DIV::align', 'DL::compact', 'FIELDSET::disabled', 'FONT::color', 'FONT::face', 'FONT::size', 'FORM::accept', 'FORM::autocomplete', 'FORM::enctype', 'FORM::method', 'FORM::name', 'FORM::novalidate', 'FORM::target', 'FRAME::name', 'H1::align', 'H2::align', 'H3::align', 'H4::align', 'H5::align', 'H6::align', 'HR::align', 'HR::noshade', 'HR::size', 'HR::width', 'HTML::version', 'IFRAME::align', 'IFRAME::frameborder', 'IFRAME::height', 'IFRAME::marginheight', 'IFRAME::marginwidth', 'IFRAME::width', 'IMG::align', 'IMG::alt', 'IMG::border', 'IMG::height', 'IMG::hspace', 'IMG::ismap', 'IMG::name', 'IMG::usemap', 'IMG::vspace', 'IMG::width', 'INPUT::accept', 'INPUT::accesskey', 'INPUT::align', 'INPUT::alt', 'INPUT::autocomplete', 'INPUT::checked', 'INPUT::disabled', 'INPUT::inputmode', 'INPUT::ismap', 'INPUT::list', 'INPUT::max', 'INPUT::maxlength', 'INPUT::min', 'INPUT::multiple', 'INPUT::name', 'INPUT::placeholder', 'INPUT::readonly', 'INPUT::required', 'INPUT::size', 'INPUT::step', 'INPUT::tabindex', 'INPUT::type', 'INPUT::usemap', 'INPUT::value', 'INS::datetime', 'KEYGEN::disabled', 'KEYGEN::keytype', 'KEYGEN::name', 'LABEL::accesskey', 'LABEL::for', 'LEGEND::accesskey', 'LEGEND::align', 'LI::type', 'LI::value', 'LINK::sizes', 'MAP::name', 'MENU::compact', 'MENU::label', 'MENU::type', 'METER::high', 'METER::low', 'METER::max', 'METER::min', 'METER::value', 'OBJECT::typemustmatch', 'OL::compact', 'OL::reversed', 'OL::start', 'OL::type', 'OPTGROUP::disabled', 'OPTGROUP::label', 'OPTION::disabled', 'OPTION::label', 'OPTION::selected', 'OPTION::value', 'OUTPUT::for', 'OUTPUT::name', 'P::align', 'PRE::width', 'PROGRESS::max', 'PROGRESS::min', 'PROGRESS::value', 'SELECT::autocomplete', 'SELECT::disabled', 'SELECT::multiple', 'SELECT::name', 'SELECT::required', 'SELECT::size', 'SELECT::tabindex', 'SOURCE::type', 'TABLE::align', 'TABLE::bgcolor', 'TABLE::border', 'TABLE::cellpadding', 'TABLE::cellspacing', 'TABLE::frame', 'TABLE::rules', 'TABLE::summary', 'TABLE::width', 'TBODY::align', 'TBODY::char', 'TBODY::charoff', 'TBODY::valign', 'TD::abbr', 'TD::align', 'TD::axis', 'TD::bgcolor', 'TD::char', 'TD::charoff', 'TD::colspan', 'TD::headers', 'TD::height', 'TD::nowrap', 'TD::rowspan', 'TD::scope', 'TD::valign', 'TD::width', 'TEXTAREA::accesskey', 'TEXTAREA::autocomplete', 'TEXTAREA::cols', 'TEXTAREA::disabled', 'TEXTAREA::inputmode', 'TEXTAREA::name', 'TEXTAREA::placeholder', 'TEXTAREA::readonly', 'TEXTAREA::required', 'TEXTAREA::rows', 'TEXTAREA::tabindex', 'TEXTAREA::wrap', 'TFOOT::align', 'TFOOT::char', 'TFOOT::charoff', 'TFOOT::valign', 'TH::abbr', 'TH::align', 'TH::axis', 'TH::bgcolor', 'TH::char', 'TH::charoff', 'TH::colspan', 'TH::headers', 'TH::height', 'TH::nowrap', 'TH::rowspan', 'TH::scope', 'TH::valign', 'TH::width', 'THEAD::align', 'THEAD::char', 'THEAD::charoff', 'THEAD::valign', 'TR::align', 'TR::bgcolor', 'TR::char', 'TR::charoff', 'TR::valign', 'TRACK::default', 'TRACK::kind', 'TRACK::label', 'TRACK::srclang', 'UL::compact', 'UL::type', 'VIDEO::controls', 'VIDEO::height', 'VIDEO::loop', 'VIDEO::mediagroup', 'VIDEO::muted', 'VIDEO::preload', 'VIDEO::width'], core.String));
  _Html5NodeValidator._uriAttributes = dart.const(dart.list(['A::href', 'AREA::href', 'BLOCKQUOTE::cite', 'BODY::background', 'COMMAND::icon', 'DEL::cite', 'FORM::action', 'IMG::src', 'INPUT::src', 'INS::cite', 'Q::cite', 'VIDEO::poster'], core.String));
  dart.defineLazyProperties(_Html5NodeValidator, {
    get _allowedElements() {
      return core.Set$(core.String).from(['A', 'ABBR', 'ACRONYM', 'ADDRESS', 'AREA', 'ARTICLE', 'ASIDE', 'AUDIO', 'B', 'BDI', 'BDO', 'BIG', 'BLOCKQUOTE', 'BR', 'BUTTON', 'CANVAS', 'CAPTION', 'CENTER', 'CITE', 'CODE', 'COL', 'COLGROUP', 'COMMAND', 'DATA', 'DATALIST', 'DD', 'DEL', 'DETAILS', 'DFN', 'DIR', 'DIV', 'DL', 'DT', 'EM', 'FIELDSET', 'FIGCAPTION', 'FIGURE', 'FONT', 'FOOTER', 'FORM', 'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 'HEADER', 'HGROUP', 'HR', 'I', 'IFRAME', 'IMG', 'INPUT', 'INS', 'KBD', 'LABEL', 'LEGEND', 'LI', 'MAP', 'MARK', 'MENU', 'METER', 'NAV', 'NOBR', 'OL', 'OPTGROUP', 'OPTION', 'OUTPUT', 'P', 'PRE', 'PROGRESS', 'Q', 'S', 'SAMP', 'SECTION', 'SELECT', 'SMALL', 'SOURCE', 'SPAN', 'STRIKE', 'STRONG', 'SUB', 'SUMMARY', 'SUP', 'TABLE', 'TBODY', 'TD', 'TEXTAREA', 'TFOOT', 'TH', 'THEAD', 'TIME', 'TR', 'TRACK', 'TT', 'U', 'UL', 'VAR', 'VIDEO', 'WBR']);
    },
    get _attributeValidators() {
      return dart.map();
    }
  });
  class KeyCode extends core.Object {
    static isCharacterKey(keyCode) {
      if (dart.notNull(keyCode) >= dart.notNull(KeyCode.ZERO) && dart.notNull(keyCode) <= dart.notNull(KeyCode.NINE) || dart.notNull(keyCode) >= dart.notNull(KeyCode.NUM_ZERO) && dart.notNull(keyCode) <= dart.notNull(KeyCode.NUM_MULTIPLY) || dart.notNull(keyCode) >= dart.notNull(KeyCode.A) && dart.notNull(keyCode) <= dart.notNull(KeyCode.Z)) {
        return true;
      }
      if (dart.notNull(html_common.Device.isWebKit) && keyCode == 0) {
        return true;
      }
      return keyCode == KeyCode.SPACE || keyCode == KeyCode.QUESTION_MARK || keyCode == KeyCode.NUM_PLUS || keyCode == KeyCode.NUM_MINUS || keyCode == KeyCode.NUM_PERIOD || keyCode == KeyCode.NUM_DIVISION || keyCode == KeyCode.SEMICOLON || keyCode == KeyCode.FF_SEMICOLON || keyCode == KeyCode.DASH || keyCode == KeyCode.EQUALS || keyCode == KeyCode.FF_EQUALS || keyCode == KeyCode.COMMA || keyCode == KeyCode.PERIOD || keyCode == KeyCode.SLASH || keyCode == KeyCode.APOSTROPHE || keyCode == KeyCode.SINGLE_QUOTE || keyCode == KeyCode.OPEN_SQUARE_BRACKET || keyCode == KeyCode.BACKSLASH || keyCode == KeyCode.CLOSE_SQUARE_BRACKET;
    }
    static _convertKeyCodeToKeyName(keyCode) {
      switch (keyCode) {
        case KeyCode.ALT:
        {
          return _KeyName.ALT;
        }
        case KeyCode.BACKSPACE:
        {
          return _KeyName.BACKSPACE;
        }
        case KeyCode.CAPS_LOCK:
        {
          return _KeyName.CAPS_LOCK;
        }
        case KeyCode.CTRL:
        {
          return _KeyName.CONTROL;
        }
        case KeyCode.DELETE:
        {
          return _KeyName.DEL;
        }
        case KeyCode.DOWN:
        {
          return _KeyName.DOWN;
        }
        case KeyCode.END:
        {
          return _KeyName.END;
        }
        case KeyCode.ENTER:
        {
          return _KeyName.ENTER;
        }
        case KeyCode.ESC:
        {
          return _KeyName.ESC;
        }
        case KeyCode.F1:
        {
          return _KeyName.F1;
        }
        case KeyCode.F2:
        {
          return _KeyName.F2;
        }
        case KeyCode.F3:
        {
          return _KeyName.F3;
        }
        case KeyCode.F4:
        {
          return _KeyName.F4;
        }
        case KeyCode.F5:
        {
          return _KeyName.F5;
        }
        case KeyCode.F6:
        {
          return _KeyName.F6;
        }
        case KeyCode.F7:
        {
          return _KeyName.F7;
        }
        case KeyCode.F8:
        {
          return _KeyName.F8;
        }
        case KeyCode.F9:
        {
          return _KeyName.F9;
        }
        case KeyCode.F10:
        {
          return _KeyName.F10;
        }
        case KeyCode.F11:
        {
          return _KeyName.F11;
        }
        case KeyCode.F12:
        {
          return _KeyName.F12;
        }
        case KeyCode.HOME:
        {
          return _KeyName.HOME;
        }
        case KeyCode.INSERT:
        {
          return _KeyName.INSERT;
        }
        case KeyCode.LEFT:
        {
          return _KeyName.LEFT;
        }
        case KeyCode.META:
        {
          return _KeyName.META;
        }
        case KeyCode.NUMLOCK:
        {
          return _KeyName.NUM_LOCK;
        }
        case KeyCode.PAGE_DOWN:
        {
          return _KeyName.PAGE_DOWN;
        }
        case KeyCode.PAGE_UP:
        {
          return _KeyName.PAGE_UP;
        }
        case KeyCode.PAUSE:
        {
          return _KeyName.PAUSE;
        }
        case KeyCode.PRINT_SCREEN:
        {
          return _KeyName.PRINT_SCREEN;
        }
        case KeyCode.RIGHT:
        {
          return _KeyName.RIGHT;
        }
        case KeyCode.SCROLL_LOCK:
        {
          return _KeyName.SCROLL;
        }
        case KeyCode.SHIFT:
        {
          return _KeyName.SHIFT;
        }
        case KeyCode.SPACE:
        {
          return _KeyName.SPACEBAR;
        }
        case KeyCode.TAB:
        {
          return _KeyName.TAB;
        }
        case KeyCode.UP:
        {
          return _KeyName.UP;
        }
        case KeyCode.WIN_IME:
        case KeyCode.WIN_KEY:
        case KeyCode.WIN_KEY_LEFT:
        case KeyCode.WIN_KEY_RIGHT:
        {
          return _KeyName.WIN;
        }
        default:
        {
          return _KeyName.UNIDENTIFIED;
        }
      }
      return _KeyName.UNIDENTIFIED;
    }
  }
  dart.setSignature(KeyCode, {
    statics: () => ({
      isCharacterKey: [core.bool, [core.int]],
      _convertKeyCodeToKeyName: [core.String, [core.int]]
    }),
    names: ['isCharacterKey', '_convertKeyCodeToKeyName']
  });
  KeyCode.WIN_KEY_FF_LINUX = 0;
  KeyCode.MAC_ENTER = 3;
  KeyCode.BACKSPACE = 8;
  KeyCode.TAB = 9;
  KeyCode.NUM_CENTER = 12;
  KeyCode.ENTER = 13;
  KeyCode.SHIFT = 16;
  KeyCode.CTRL = 17;
  KeyCode.ALT = 18;
  KeyCode.PAUSE = 19;
  KeyCode.CAPS_LOCK = 20;
  KeyCode.ESC = 27;
  KeyCode.SPACE = 32;
  KeyCode.PAGE_UP = 33;
  KeyCode.PAGE_DOWN = 34;
  KeyCode.END = 35;
  KeyCode.HOME = 36;
  KeyCode.LEFT = 37;
  KeyCode.UP = 38;
  KeyCode.RIGHT = 39;
  KeyCode.DOWN = 40;
  KeyCode.NUM_NORTH_EAST = 33;
  KeyCode.NUM_SOUTH_EAST = 34;
  KeyCode.NUM_SOUTH_WEST = 35;
  KeyCode.NUM_NORTH_WEST = 36;
  KeyCode.NUM_WEST = 37;
  KeyCode.NUM_NORTH = 38;
  KeyCode.NUM_EAST = 39;
  KeyCode.NUM_SOUTH = 40;
  KeyCode.PRINT_SCREEN = 44;
  KeyCode.INSERT = 45;
  KeyCode.NUM_INSERT = 45;
  KeyCode.DELETE = 46;
  KeyCode.NUM_DELETE = 46;
  KeyCode.ZERO = 48;
  KeyCode.ONE = 49;
  KeyCode.TWO = 50;
  KeyCode.THREE = 51;
  KeyCode.FOUR = 52;
  KeyCode.FIVE = 53;
  KeyCode.SIX = 54;
  KeyCode.SEVEN = 55;
  KeyCode.EIGHT = 56;
  KeyCode.NINE = 57;
  KeyCode.FF_SEMICOLON = 59;
  KeyCode.FF_EQUALS = 61;
  KeyCode.QUESTION_MARK = 63;
  KeyCode.A = 65;
  KeyCode.B = 66;
  KeyCode.C = 67;
  KeyCode.D = 68;
  KeyCode.E = 69;
  KeyCode.F = 70;
  KeyCode.G = 71;
  KeyCode.H = 72;
  KeyCode.I = 73;
  KeyCode.J = 74;
  KeyCode.K = 75;
  KeyCode.L = 76;
  KeyCode.M = 77;
  KeyCode.N = 78;
  KeyCode.O = 79;
  KeyCode.P = 80;
  KeyCode.Q = 81;
  KeyCode.R = 82;
  KeyCode.S = 83;
  KeyCode.T = 84;
  KeyCode.U = 85;
  KeyCode.V = 86;
  KeyCode.W = 87;
  KeyCode.X = 88;
  KeyCode.Y = 89;
  KeyCode.Z = 90;
  KeyCode.META = 91;
  KeyCode.WIN_KEY_LEFT = 91;
  KeyCode.WIN_KEY_RIGHT = 92;
  KeyCode.CONTEXT_MENU = 93;
  KeyCode.NUM_ZERO = 96;
  KeyCode.NUM_ONE = 97;
  KeyCode.NUM_TWO = 98;
  KeyCode.NUM_THREE = 99;
  KeyCode.NUM_FOUR = 100;
  KeyCode.NUM_FIVE = 101;
  KeyCode.NUM_SIX = 102;
  KeyCode.NUM_SEVEN = 103;
  KeyCode.NUM_EIGHT = 104;
  KeyCode.NUM_NINE = 105;
  KeyCode.NUM_MULTIPLY = 106;
  KeyCode.NUM_PLUS = 107;
  KeyCode.NUM_MINUS = 109;
  KeyCode.NUM_PERIOD = 110;
  KeyCode.NUM_DIVISION = 111;
  KeyCode.F1 = 112;
  KeyCode.F2 = 113;
  KeyCode.F3 = 114;
  KeyCode.F4 = 115;
  KeyCode.F5 = 116;
  KeyCode.F6 = 117;
  KeyCode.F7 = 118;
  KeyCode.F8 = 119;
  KeyCode.F9 = 120;
  KeyCode.F10 = 121;
  KeyCode.F11 = 122;
  KeyCode.F12 = 123;
  KeyCode.NUMLOCK = 144;
  KeyCode.SCROLL_LOCK = 145;
  KeyCode.FIRST_MEDIA_KEY = 166;
  KeyCode.LAST_MEDIA_KEY = 183;
  KeyCode.SEMICOLON = 186;
  KeyCode.DASH = 189;
  KeyCode.EQUALS = 187;
  KeyCode.COMMA = 188;
  KeyCode.PERIOD = 190;
  KeyCode.SLASH = 191;
  KeyCode.APOSTROPHE = 192;
  KeyCode.TILDE = 192;
  KeyCode.SINGLE_QUOTE = 222;
  KeyCode.OPEN_SQUARE_BRACKET = 219;
  KeyCode.BACKSLASH = 220;
  KeyCode.CLOSE_SQUARE_BRACKET = 221;
  KeyCode.WIN_KEY = 224;
  KeyCode.MAC_FF_META = 224;
  KeyCode.WIN_IME = 229;
  KeyCode.UNKNOWN = -1;
  class KeyLocation extends core.Object {}
  KeyLocation.STANDARD = 0;
  KeyLocation.LEFT = 1;
  KeyLocation.RIGHT = 2;
  KeyLocation.NUMPAD = 3;
  KeyLocation.MOBILE = 4;
  KeyLocation.JOYSTICK = 5;
  class _KeyName extends core.Object {}
  _KeyName.ACCEPT = "Accept";
  _KeyName.ADD = "Add";
  _KeyName.AGAIN = "Again";
  _KeyName.ALL_CANDIDATES = "AllCandidates";
  _KeyName.ALPHANUMERIC = "Alphanumeric";
  _KeyName.ALT = "Alt";
  _KeyName.ALT_GRAPH = "AltGraph";
  _KeyName.APPS = "Apps";
  _KeyName.ATTN = "Attn";
  _KeyName.BROWSER_BACK = "BrowserBack";
  _KeyName.BROWSER_FAVORTIES = "BrowserFavorites";
  _KeyName.BROWSER_FORWARD = "BrowserForward";
  _KeyName.BROWSER_NAME = "BrowserHome";
  _KeyName.BROWSER_REFRESH = "BrowserRefresh";
  _KeyName.BROWSER_SEARCH = "BrowserSearch";
  _KeyName.BROWSER_STOP = "BrowserStop";
  _KeyName.CAMERA = "Camera";
  _KeyName.CAPS_LOCK = "CapsLock";
  _KeyName.CLEAR = "Clear";
  _KeyName.CODE_INPUT = "CodeInput";
  _KeyName.COMPOSE = "Compose";
  _KeyName.CONTROL = "Control";
  _KeyName.CRSEL = "Crsel";
  _KeyName.CONVERT = "Convert";
  _KeyName.COPY = "Copy";
  _KeyName.CUT = "Cut";
  _KeyName.DECIMAL = "Decimal";
  _KeyName.DIVIDE = "Divide";
  _KeyName.DOWN = "Down";
  _KeyName.DOWN_LEFT = "DownLeft";
  _KeyName.DOWN_RIGHT = "DownRight";
  _KeyName.EJECT = "Eject";
  _KeyName.END = "End";
  _KeyName.ENTER = "Enter";
  _KeyName.ERASE_EOF = "EraseEof";
  _KeyName.EXECUTE = "Execute";
  _KeyName.EXSEL = "Exsel";
  _KeyName.FN = "Fn";
  _KeyName.F1 = "F1";
  _KeyName.F2 = "F2";
  _KeyName.F3 = "F3";
  _KeyName.F4 = "F4";
  _KeyName.F5 = "F5";
  _KeyName.F6 = "F6";
  _KeyName.F7 = "F7";
  _KeyName.F8 = "F8";
  _KeyName.F9 = "F9";
  _KeyName.F10 = "F10";
  _KeyName.F11 = "F11";
  _KeyName.F12 = "F12";
  _KeyName.F13 = "F13";
  _KeyName.F14 = "F14";
  _KeyName.F15 = "F15";
  _KeyName.F16 = "F16";
  _KeyName.F17 = "F17";
  _KeyName.F18 = "F18";
  _KeyName.F19 = "F19";
  _KeyName.F20 = "F20";
  _KeyName.F21 = "F21";
  _KeyName.F22 = "F22";
  _KeyName.F23 = "F23";
  _KeyName.F24 = "F24";
  _KeyName.FINAL_MODE = "FinalMode";
  _KeyName.FIND = "Find";
  _KeyName.FULL_WIDTH = "FullWidth";
  _KeyName.HALF_WIDTH = "HalfWidth";
  _KeyName.HANGUL_MODE = "HangulMode";
  _KeyName.HANJA_MODE = "HanjaMode";
  _KeyName.HELP = "Help";
  _KeyName.HIRAGANA = "Hiragana";
  _KeyName.HOME = "Home";
  _KeyName.INSERT = "Insert";
  _KeyName.JAPANESE_HIRAGANA = "JapaneseHiragana";
  _KeyName.JAPANESE_KATAKANA = "JapaneseKatakana";
  _KeyName.JAPANESE_ROMAJI = "JapaneseRomaji";
  _KeyName.JUNJA_MODE = "JunjaMode";
  _KeyName.KANA_MODE = "KanaMode";
  _KeyName.KANJI_MODE = "KanjiMode";
  _KeyName.KATAKANA = "Katakana";
  _KeyName.LAUNCH_APPLICATION_1 = "LaunchApplication1";
  _KeyName.LAUNCH_APPLICATION_2 = "LaunchApplication2";
  _KeyName.LAUNCH_MAIL = "LaunchMail";
  _KeyName.LEFT = "Left";
  _KeyName.MENU = "Menu";
  _KeyName.META = "Meta";
  _KeyName.MEDIA_NEXT_TRACK = "MediaNextTrack";
  _KeyName.MEDIA_PAUSE_PLAY = "MediaPlayPause";
  _KeyName.MEDIA_PREVIOUS_TRACK = "MediaPreviousTrack";
  _KeyName.MEDIA_STOP = "MediaStop";
  _KeyName.MODE_CHANGE = "ModeChange";
  _KeyName.NEXT_CANDIDATE = "NextCandidate";
  _KeyName.NON_CONVERT = "Nonconvert";
  _KeyName.NUM_LOCK = "NumLock";
  _KeyName.PAGE_DOWN = "PageDown";
  _KeyName.PAGE_UP = "PageUp";
  _KeyName.PASTE = "Paste";
  _KeyName.PAUSE = "Pause";
  _KeyName.PLAY = "Play";
  _KeyName.POWER = "Power";
  _KeyName.PREVIOUS_CANDIDATE = "PreviousCandidate";
  _KeyName.PRINT_SCREEN = "PrintScreen";
  _KeyName.PROCESS = "Process";
  _KeyName.PROPS = "Props";
  _KeyName.RIGHT = "Right";
  _KeyName.ROMAN_CHARACTERS = "RomanCharacters";
  _KeyName.SCROLL = "Scroll";
  _KeyName.SELECT = "Select";
  _KeyName.SELECT_MEDIA = "SelectMedia";
  _KeyName.SEPARATOR = "Separator";
  _KeyName.SHIFT = "Shift";
  _KeyName.SOFT_1 = "Soft1";
  _KeyName.SOFT_2 = "Soft2";
  _KeyName.SOFT_3 = "Soft3";
  _KeyName.SOFT_4 = "Soft4";
  _KeyName.STOP = "Stop";
  _KeyName.SUBTRACT = "Subtract";
  _KeyName.SYMBOL_LOCK = "SymbolLock";
  _KeyName.UP = "Up";
  _KeyName.UP_LEFT = "UpLeft";
  _KeyName.UP_RIGHT = "UpRight";
  _KeyName.UNDO = "Undo";
  _KeyName.VOLUME_DOWN = "VolumeDown";
  _KeyName.VOLUMN_MUTE = "VolumeMute";
  _KeyName.VOLUMN_UP = "VolumeUp";
  _KeyName.WIN = "Win";
  _KeyName.ZOOM = "Zoom";
  _KeyName.BACKSPACE = "Backspace";
  _KeyName.TAB = "Tab";
  _KeyName.CANCEL = "Cancel";
  _KeyName.ESC = "Esc";
  _KeyName.SPACEBAR = "Spacebar";
  _KeyName.DEL = "Del";
  _KeyName.DEAD_GRAVE = "DeadGrave";
  _KeyName.DEAD_EACUTE = "DeadEacute";
  _KeyName.DEAD_CIRCUMFLEX = "DeadCircumflex";
  _KeyName.DEAD_TILDE = "DeadTilde";
  _KeyName.DEAD_MACRON = "DeadMacron";
  _KeyName.DEAD_BREVE = "DeadBreve";
  _KeyName.DEAD_ABOVE_DOT = "DeadAboveDot";
  _KeyName.DEAD_UMLAUT = "DeadUmlaut";
  _KeyName.DEAD_ABOVE_RING = "DeadAboveRing";
  _KeyName.DEAD_DOUBLEACUTE = "DeadDoubleacute";
  _KeyName.DEAD_CARON = "DeadCaron";
  _KeyName.DEAD_CEDILLA = "DeadCedilla";
  _KeyName.DEAD_OGONEK = "DeadOgonek";
  _KeyName.DEAD_IOTA = "DeadIota";
  _KeyName.DEAD_VOICED_SOUND = "DeadVoicedSound";
  _KeyName.DEC_SEMIVOICED_SOUND = "DeadSemivoicedSound";
  _KeyName.UNIDENTIFIED = "Unidentified";
  const _stream = Symbol('_stream');
  const _keyDownList = Symbol('_keyDownList');
  const _capsLockOn = Symbol('_capsLockOn');
  const _determineKeyCodeForKeypress = Symbol('_determineKeyCodeForKeypress');
  const _findCharCodeKeyDown = Symbol('_findCharCodeKeyDown');
  const _firesKeyPressEvent = Symbol('_firesKeyPressEvent');
  const _normalizeKeyCodes = Symbol('_normalizeKeyCodes');
  class _KeyboardEventHandler extends EventStreamProvider$(KeyEvent) {
    forTarget(e, opts) {
      let useCapture = opts && 'useCapture' in opts ? opts.useCapture : false;
      let handler = new _KeyboardEventHandler.initializeAllEventListeners(this[_type], e);
      return handler[_stream];
    }
    _KeyboardEventHandler(type) {
      this[_keyDownList] = dart.list([], KeyboardEvent);
      this[_type] = type;
      this[_stream] = new _CustomKeyEventStreamImpl('event');
      this[_target] = null;
      super.EventStreamProvider(_KeyboardEventHandler._EVENT_TYPE);
    }
    initializeAllEventListeners(type, target) {
      this[_keyDownList] = dart.list([], KeyboardEvent);
      this[_type] = type;
      this[_target] = target;
      this[_stream] = null;
      super.EventStreamProvider(_KeyboardEventHandler._EVENT_TYPE);
      Element.keyDownEvent.forTarget(this[_target], {useCapture: true}).listen(dart.bind(this, 'processKeyDown'));
      Element.keyPressEvent.forTarget(this[_target], {useCapture: true}).listen(dart.bind(this, 'processKeyPress'));
      Element.keyUpEvent.forTarget(this[_target], {useCapture: true}).listen(dart.bind(this, 'processKeyUp'));
      this[_stream] = new _CustomKeyEventStreamImpl(this[_type]);
    }
    get [_capsLockOn]() {
      return this[_keyDownList][dartx.any](dart.fn(element => dart.equals(dart.dload(element, 'keyCode'), KeyCode.CAPS_LOCK), core.bool, [dart.dynamic]));
    }
    [_determineKeyCodeForKeypress](event) {
      for (let prevEvent of this[_keyDownList]) {
        if (dart.equals(dart.dload(prevEvent, _shadowCharCode), event.charCode)) {
          return prevEvent.keyCode;
        }
        if ((dart.notNull(event.shiftKey) || dart.notNull(this[_capsLockOn])) && dart.notNull(event.charCode) >= dart.notNull("A"[dartx.codeUnits][dartx.get](0)) && dart.notNull(event.charCode) <= dart.notNull("Z"[dartx.codeUnits][dartx.get](0)) && dart.equals(dart.notNull(event.charCode) + dart.notNull(_KeyboardEventHandler._ROMAN_ALPHABET_OFFSET), dart.dload(prevEvent, _shadowCharCode))) {
          return prevEvent.keyCode;
        }
      }
      return KeyCode.UNKNOWN;
    }
    [_findCharCodeKeyDown](event) {
      if (event.keyLocation == 3) {
        switch (event.keyCode) {
          case KeyCode.NUM_ZERO:
          {
            return KeyCode.ZERO;
          }
          case KeyCode.NUM_ONE:
          {
            return KeyCode.ONE;
          }
          case KeyCode.NUM_TWO:
          {
            return KeyCode.TWO;
          }
          case KeyCode.NUM_THREE:
          {
            return KeyCode.THREE;
          }
          case KeyCode.NUM_FOUR:
          {
            return KeyCode.FOUR;
          }
          case KeyCode.NUM_FIVE:
          {
            return KeyCode.FIVE;
          }
          case KeyCode.NUM_SIX:
          {
            return KeyCode.SIX;
          }
          case KeyCode.NUM_SEVEN:
          {
            return KeyCode.SEVEN;
          }
          case KeyCode.NUM_EIGHT:
          {
            return KeyCode.EIGHT;
          }
          case KeyCode.NUM_NINE:
          {
            return KeyCode.NINE;
          }
          case KeyCode.NUM_MULTIPLY:
          {
            return 42;
          }
          case KeyCode.NUM_PLUS:
          {
            return 43;
          }
          case KeyCode.NUM_MINUS:
          {
            return 45;
          }
          case KeyCode.NUM_PERIOD:
          {
            return 46;
          }
          case KeyCode.NUM_DIVISION:
          {
            return 47;
          }
        }
      } else if (dart.notNull(event.keyCode) >= 65 && dart.notNull(event.keyCode) <= 90) {
        return dart.notNull(event.keyCode) + dart.notNull(_KeyboardEventHandler._ROMAN_ALPHABET_OFFSET);
      }
      switch (event.keyCode) {
        case KeyCode.SEMICOLON:
        {
          return KeyCode.FF_SEMICOLON;
        }
        case KeyCode.EQUALS:
        {
          return KeyCode.FF_EQUALS;
        }
        case KeyCode.COMMA:
        {
          return 44;
        }
        case KeyCode.DASH:
        {
          return 45;
        }
        case KeyCode.PERIOD:
        {
          return 46;
        }
        case KeyCode.SLASH:
        {
          return 47;
        }
        case KeyCode.APOSTROPHE:
        {
          return 96;
        }
        case KeyCode.OPEN_SQUARE_BRACKET:
        {
          return 91;
        }
        case KeyCode.BACKSLASH:
        {
          return 92;
        }
        case KeyCode.CLOSE_SQUARE_BRACKET:
        {
          return 93;
        }
        case KeyCode.SINGLE_QUOTE:
        {
          return 39;
        }
      }
      return event.keyCode;
    }
    [_firesKeyPressEvent](event) {
      if (!dart.notNull(html_common.Device.isIE) && !dart.notNull(html_common.Device.isWebKit)) {
        return true;
      }
      if (dart.notNull(html_common.Device.userAgent[dartx.contains]('Mac')) && dart.notNull(event.altKey)) {
        return KeyCode.isCharacterKey(event.keyCode);
      }
      if (dart.notNull(event.altKey) && !dart.notNull(event.ctrlKey)) {
        return false;
      }
      if (!dart.notNull(event.shiftKey) && (this[_keyDownList][dartx.last].keyCode == KeyCode.CTRL || this[_keyDownList][dartx.last].keyCode == KeyCode.ALT || dart.notNull(html_common.Device.userAgent[dartx.contains]('Mac')) && this[_keyDownList][dartx.last].keyCode == KeyCode.META)) {
        return false;
      }
      if (dart.notNull(html_common.Device.isWebKit) && dart.notNull(event.ctrlKey) && dart.notNull(event.shiftKey) && (event.keyCode == KeyCode.BACKSLASH || event.keyCode == KeyCode.OPEN_SQUARE_BRACKET || event.keyCode == KeyCode.CLOSE_SQUARE_BRACKET || event.keyCode == KeyCode.TILDE || event.keyCode == KeyCode.SEMICOLON || event.keyCode == KeyCode.DASH || event.keyCode == KeyCode.EQUALS || event.keyCode == KeyCode.COMMA || event.keyCode == KeyCode.PERIOD || event.keyCode == KeyCode.SLASH || event.keyCode == KeyCode.APOSTROPHE || event.keyCode == KeyCode.SINGLE_QUOTE)) {
        return false;
      }
      switch (event.keyCode) {
        case KeyCode.ENTER:
        {
          return !dart.notNull(html_common.Device.isIE);
        }
        case KeyCode.ESC:
        {
          return !dart.notNull(html_common.Device.isWebKit);
        }
      }
      return KeyCode.isCharacterKey(event.keyCode);
    }
    [_normalizeKeyCodes](event) {
      if (dart.notNull(html_common.Device.isFirefox)) {
        switch (event.keyCode) {
          case KeyCode.FF_EQUALS:
          {
            return KeyCode.EQUALS;
          }
          case KeyCode.FF_SEMICOLON:
          {
            return KeyCode.SEMICOLON;
          }
          case KeyCode.MAC_FF_META:
          {
            return KeyCode.META;
          }
          case KeyCode.WIN_KEY_FF_LINUX:
          {
            return KeyCode.WIN_KEY;
          }
        }
      }
      return event.keyCode;
    }
    processKeyDown(e) {
      if (dart.notNull(this[_keyDownList][dartx.length]) > 0 && (this[_keyDownList][dartx.last].keyCode == KeyCode.CTRL && !dart.notNull(e.ctrlKey) || this[_keyDownList][dartx.last].keyCode == KeyCode.ALT && !dart.notNull(e.altKey) || dart.notNull(html_common.Device.userAgent[dartx.contains]('Mac')) && this[_keyDownList][dartx.last].keyCode == KeyCode.META && !dart.notNull(e.metaKey))) {
        this[_keyDownList][dartx.clear]();
      }
      let event = new KeyEvent.wrap(e);
      event[_shadowKeyCode] = this[_normalizeKeyCodes](event);
      event[_shadowCharCode] = this[_findCharCodeKeyDown](event);
      if (dart.notNull(this[_keyDownList][dartx.length]) > 0 && event.keyCode != this[_keyDownList][dartx.last].keyCode && !dart.notNull(this[_firesKeyPressEvent](event))) {
        this.processKeyPress(e);
      }
      this[_keyDownList][dartx.add](event);
      this[_stream].add(event);
    }
    processKeyPress(event) {
      let e = new KeyEvent.wrap(event);
      if (dart.notNull(html_common.Device.isIE)) {
        if (e.keyCode == KeyCode.ENTER || e.keyCode == KeyCode.ESC) {
          e[_shadowCharCode] = 0;
        } else {
          e[_shadowCharCode] = e.keyCode;
        }
      } else if (dart.notNull(html_common.Device.isOpera)) {
        e[_shadowCharCode] = dart.notNull(KeyCode.isCharacterKey(e.keyCode)) ? e.keyCode : 0;
      }
      e[_shadowKeyCode] = this[_determineKeyCodeForKeypress](e);
      if (e[_shadowKeyIdentifier] != null && dart.notNull(_KeyboardEventHandler._keyIdentifier.containsKey(e[_shadowKeyIdentifier]))) {
        e[_shadowKeyCode] = _KeyboardEventHandler._keyIdentifier.get(e[_shadowKeyIdentifier]);
      }
      e[_shadowAltKey] = this[_keyDownList][dartx.any](dart.fn(element => dart.as(dart.dload(element, 'altKey'), core.bool), core.bool, [dart.dynamic]));
      this[_stream].add(e);
    }
    processKeyUp(event) {
      let e = new KeyEvent.wrap(event);
      let toRemove = null;
      for (let key of this[_keyDownList]) {
        if (key.keyCode == e.keyCode) {
          toRemove = key;
        }
      }
      if (toRemove != null) {
        this[_keyDownList][dartx.removeWhere](dart.fn(element => dart.equals(element, toRemove), core.bool, [dart.dynamic]));
      } else if (dart.notNull(this[_keyDownList][dartx.length]) > 0) {
        this[_keyDownList][dartx.removeLast]();
      }
      this[_stream].add(e);
    }
  }
  dart.defineNamedConstructor(_KeyboardEventHandler, 'initializeAllEventListeners');
  dart.setSignature(_KeyboardEventHandler, {
    constructors: () => ({
      _KeyboardEventHandler: [_KeyboardEventHandler, [core.String]],
      initializeAllEventListeners: [_KeyboardEventHandler, [core.String, EventTarget]]
    }),
    methods: () => ({
      forTarget: [CustomStream$(KeyEvent), [EventTarget], {useCapture: core.bool}],
      [_determineKeyCodeForKeypress]: [core.int, [KeyboardEvent]],
      [_findCharCodeKeyDown]: [core.int, [KeyboardEvent]],
      [_firesKeyPressEvent]: [core.bool, [KeyEvent]],
      [_normalizeKeyCodes]: [core.int, [KeyboardEvent]],
      processKeyDown: [dart.void, [KeyboardEvent]],
      processKeyPress: [dart.void, [KeyboardEvent]],
      processKeyUp: [dart.void, [KeyboardEvent]]
    })
  });
  _KeyboardEventHandler._EVENT_TYPE = 'KeyEvent';
  _KeyboardEventHandler._keyIdentifier = dart.const(dart.map({Up: KeyCode.UP, Down: KeyCode.DOWN, Left: KeyCode.LEFT, Right: KeyCode.RIGHT, Enter: KeyCode.ENTER, F1: KeyCode.F1, F2: KeyCode.F2, F3: KeyCode.F3, F4: KeyCode.F4, F5: KeyCode.F5, F6: KeyCode.F6, F7: KeyCode.F7, F8: KeyCode.F8, F9: KeyCode.F9, F10: KeyCode.F10, F11: KeyCode.F11, F12: KeyCode.F12, 'U+007F': KeyCode.DELETE, Home: KeyCode.HOME, End: KeyCode.END, PageUp: KeyCode.PAGE_UP, PageDown: KeyCode.PAGE_DOWN, Insert: KeyCode.INSERT}));
  dart.defineLazyProperties(_KeyboardEventHandler, {
    get _ROMAN_ALPHABET_OFFSET() {
      return dart.notNull("a"[dartx.codeUnits][dartx.get](0)) - dart.notNull("A"[dartx.codeUnits][dartx.get](0));
    }
  });
  class KeyboardEventStream extends core.Object {
    static onKeyPress(target) {
      return new _KeyboardEventHandler('keypress').forTarget(target);
    }
    static onKeyUp(target) {
      return new _KeyboardEventHandler('keyup').forTarget(target);
    }
    static onKeyDown(target) {
      return new _KeyboardEventHandler('keydown').forTarget(target);
    }
  }
  dart.setSignature(KeyboardEventStream, {
    statics: () => ({
      onKeyPress: [CustomStream$(KeyEvent), [EventTarget]],
      onKeyUp: [CustomStream$(KeyEvent), [EventTarget]],
      onKeyDown: [CustomStream$(KeyEvent), [EventTarget]]
    }),
    names: ['onKeyPress', 'onKeyUp', 'onKeyDown']
  });
  const _validators = Symbol('_validators');
  class NodeValidatorBuilder extends core.Object {
    NodeValidatorBuilder() {
      this[_validators] = dart.list([], NodeValidator);
    }
    common() {
      this[_validators] = dart.list([], NodeValidator);
      this.allowHtml5();
      this.allowTemplating();
    }
    allowNavigation(uriPolicy) {
      if (uriPolicy === void 0)
        uriPolicy = null;
      if (uriPolicy == null) {
        uriPolicy = UriPolicy.new();
      }
      this.add(_SimpleNodeValidator.allowNavigation(uriPolicy));
    }
    allowImages(uriPolicy) {
      if (uriPolicy === void 0)
        uriPolicy = null;
      if (uriPolicy == null) {
        uriPolicy = UriPolicy.new();
      }
      this.add(_SimpleNodeValidator.allowImages(uriPolicy));
    }
    allowTextElements() {
      this.add(_SimpleNodeValidator.allowTextElements());
    }
    allowInlineStyles(opts) {
      let tagName = opts && 'tagName' in opts ? opts.tagName : null;
      if (tagName == null) {
        tagName = '*';
      } else {
        tagName = tagName[dartx.toUpperCase]();
      }
      this.add(new _SimpleNodeValidator(null, {allowedAttributes: dart.list([`${tagName}::style`], core.String)}));
    }
    allowHtml5(opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      this.add(new _Html5NodeValidator({uriPolicy: uriPolicy}));
    }
    allowSvg() {
      this.add(new _SvgNodeValidator());
    }
    allowCustomElement(tagName, opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      let attributes = opts && 'attributes' in opts ? opts.attributes : null;
      let uriAttributes = opts && 'uriAttributes' in opts ? opts.uriAttributes : null;
      let tagNameUpper = tagName[dartx.toUpperCase]();
      let attrs = null;
      if (attributes != null) {
        attrs = attributes[dartx.map](dart.fn(name => `${tagNameUpper}::${dart.dsend(name, 'toLowerCase')}`, core.String, [dart.dynamic]));
      }
      let uriAttrs = null;
      if (uriAttributes != null) {
        uriAttrs = uriAttributes[dartx.map](dart.fn(name => `${tagNameUpper}::${dart.dsend(name, 'toLowerCase')}`, core.String, [dart.dynamic]));
      }
      if (uriPolicy == null) {
        uriPolicy = UriPolicy.new();
      }
      this.add(new _CustomElementNodeValidator(uriPolicy, dart.list([tagNameUpper], core.String), dart.as(attrs, core.Iterable$(core.String)), dart.as(uriAttrs, core.Iterable$(core.String)), false, true));
    }
    allowTagExtension(tagName, baseName, opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      let attributes = opts && 'attributes' in opts ? opts.attributes : null;
      let uriAttributes = opts && 'uriAttributes' in opts ? opts.uriAttributes : null;
      let baseNameUpper = baseName[dartx.toUpperCase]();
      let tagNameUpper = tagName[dartx.toUpperCase]();
      let attrs = null;
      if (attributes != null) {
        attrs = attributes[dartx.map](dart.fn(name => `${baseNameUpper}::${dart.dsend(name, 'toLowerCase')}`, core.String, [dart.dynamic]));
      }
      let uriAttrs = null;
      if (uriAttributes != null) {
        uriAttrs = uriAttributes[dartx.map](dart.fn(name => `${baseNameUpper}::${dart.dsend(name, 'toLowerCase')}`, core.String, [dart.dynamic]));
      }
      if (uriPolicy == null) {
        uriPolicy = UriPolicy.new();
      }
      this.add(new _CustomElementNodeValidator(uriPolicy, dart.list([tagNameUpper, baseNameUpper], core.String), dart.as(attrs, core.Iterable$(core.String)), dart.as(uriAttrs, core.Iterable$(core.String)), true, false));
    }
    allowElement(tagName, opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      let attributes = opts && 'attributes' in opts ? opts.attributes : null;
      let uriAttributes = opts && 'uriAttributes' in opts ? opts.uriAttributes : null;
      this.allowCustomElement(tagName, {uriPolicy: uriPolicy, attributes: attributes, uriAttributes: uriAttributes});
    }
    allowTemplating() {
      this.add(new _TemplatingNodeValidator());
    }
    add(validator) {
      this[_validators][dartx.add](validator);
    }
    allowsElement(element) {
      return this[_validators][dartx.any](dart.fn(v => dart.as(dart.dsend(v, 'allowsElement', element), core.bool), core.bool, [dart.dynamic]));
    }
    allowsAttribute(element, attributeName, value) {
      return this[_validators][dartx.any](dart.fn(v => dart.as(dart.dsend(v, 'allowsAttribute', element, attributeName, value), core.bool), core.bool, [dart.dynamic]));
    }
  }
  NodeValidatorBuilder[dart.implements] = () => [NodeValidator];
  dart.defineNamedConstructor(NodeValidatorBuilder, 'common');
  dart.setSignature(NodeValidatorBuilder, {
    constructors: () => ({
      NodeValidatorBuilder: [NodeValidatorBuilder, []],
      common: [NodeValidatorBuilder, []]
    }),
    methods: () => ({
      allowNavigation: [dart.void, [], [UriPolicy]],
      allowImages: [dart.void, [], [UriPolicy]],
      allowTextElements: [dart.void, []],
      allowInlineStyles: [dart.void, [], {tagName: core.String}],
      allowHtml5: [dart.void, [], {uriPolicy: UriPolicy}],
      allowSvg: [dart.void, []],
      allowCustomElement: [dart.void, [core.String], {uriPolicy: UriPolicy, attributes: core.Iterable$(core.String), uriAttributes: core.Iterable$(core.String)}],
      allowTagExtension: [dart.void, [core.String, core.String], {uriPolicy: UriPolicy, attributes: core.Iterable$(core.String), uriAttributes: core.Iterable$(core.String)}],
      allowElement: [dart.void, [core.String], {uriPolicy: UriPolicy, attributes: core.Iterable$(core.String), uriAttributes: core.Iterable$(core.String)}],
      allowTemplating: [dart.void, []],
      add: [dart.void, [NodeValidator]],
      allowsElement: [core.bool, [Element]],
      allowsAttribute: [core.bool, [Element, core.String, core.String]]
    })
  });
  class _SimpleNodeValidator extends core.Object {
    static allowNavigation(uriPolicy) {
      return new _SimpleNodeValidator(uriPolicy, {allowedElements: dart.const(dart.list(['A', 'FORM'], core.String)), allowedAttributes: dart.const(dart.list(['A::accesskey', 'A::coords', 'A::hreflang', 'A::name', 'A::shape', 'A::tabindex', 'A::target', 'A::type', 'FORM::accept', 'FORM::autocomplete', 'FORM::enctype', 'FORM::method', 'FORM::name', 'FORM::novalidate', 'FORM::target'], core.String)), allowedUriAttributes: dart.const(dart.list(['A::href', 'FORM::action'], core.String))});
    }
    static allowImages(uriPolicy) {
      return new _SimpleNodeValidator(uriPolicy, {allowedElements: dart.const(dart.list(['IMG'], core.String)), allowedAttributes: dart.const(dart.list(['IMG::align', 'IMG::alt', 'IMG::border', 'IMG::height', 'IMG::hspace', 'IMG::ismap', 'IMG::name', 'IMG::usemap', 'IMG::vspace', 'IMG::width'], core.String)), allowedUriAttributes: dart.const(dart.list(['IMG::src'], core.String))});
    }
    static allowTextElements() {
      return new _SimpleNodeValidator(null, {allowedElements: dart.const(dart.list(['B', 'BLOCKQUOTE', 'BR', 'EM', 'H1', 'H2', 'H3', 'H4', 'H5', 'H6', 'HR', 'I', 'LI', 'OL', 'P', 'SPAN', 'UL'], core.String))});
    }
    _SimpleNodeValidator(uriPolicy, opts) {
      let allowedElements = opts && 'allowedElements' in opts ? opts.allowedElements : null;
      let allowedAttributes = opts && 'allowedAttributes' in opts ? opts.allowedAttributes : null;
      let allowedUriAttributes = opts && 'allowedUriAttributes' in opts ? opts.allowedUriAttributes : null;
      this.allowedElements = core.Set$(core.String).new();
      this.allowedAttributes = core.Set$(core.String).new();
      this.allowedUriAttributes = core.Set$(core.String).new();
      this.uriPolicy = uriPolicy;
      this.allowedElements.addAll(dart.as((allowedElements != null ? allowedElements : dart.const([])), core.Iterable$(core.String)));
      allowedAttributes = dart.as((allowedAttributes != null ? allowedAttributes : dart.const([])), core.Iterable$(core.String));
      allowedUriAttributes = dart.as((allowedUriAttributes != null ? allowedUriAttributes : dart.const([])), core.Iterable$(core.String));
      let legalAttributes = allowedAttributes[dartx.where](dart.fn(x => !dart.notNull(_Html5NodeValidator._uriAttributes[dartx.contains](x)), core.bool, [dart.dynamic]));
      let extraUriAttributes = allowedAttributes[dartx.where](dart.fn(x => _Html5NodeValidator._uriAttributes[dartx.contains](x), core.bool, [dart.dynamic]));
      this.allowedAttributes.addAll(legalAttributes);
      this.allowedUriAttributes.addAll(allowedUriAttributes);
      this.allowedUriAttributes.addAll(extraUriAttributes);
    }
    allowsElement(element) {
      return this.allowedElements.contains(Element._safeTagName(element));
    }
    allowsAttribute(element, attributeName, value) {
      let tagName = Element._safeTagName(element);
      if (dart.notNull(this.allowedUriAttributes.contains(`${tagName}::${attributeName}`))) {
        return this.uriPolicy.allowsUri(value);
      } else if (dart.notNull(this.allowedUriAttributes.contains(`*::${attributeName}`))) {
        return this.uriPolicy.allowsUri(value);
      } else if (dart.notNull(this.allowedAttributes.contains(`${tagName}::${attributeName}`))) {
        return true;
      } else if (dart.notNull(this.allowedAttributes.contains(`*::${attributeName}`))) {
        return true;
      } else if (dart.notNull(this.allowedAttributes.contains(`${tagName}::*`))) {
        return true;
      } else if (dart.notNull(this.allowedAttributes.contains('*::*'))) {
        return true;
      }
      return false;
    }
  }
  _SimpleNodeValidator[dart.implements] = () => [NodeValidator];
  dart.setSignature(_SimpleNodeValidator, {
    constructors: () => ({
      allowNavigation: [_SimpleNodeValidator, [UriPolicy]],
      allowImages: [_SimpleNodeValidator, [UriPolicy]],
      allowTextElements: [_SimpleNodeValidator, []],
      _SimpleNodeValidator: [_SimpleNodeValidator, [UriPolicy], {allowedElements: core.Iterable$(core.String), allowedAttributes: core.Iterable$(core.String), allowedUriAttributes: core.Iterable$(core.String)}]
    }),
    methods: () => ({
      allowsElement: [core.bool, [Element]],
      allowsAttribute: [core.bool, [Element, core.String, core.String]]
    })
  });
  class _CustomElementNodeValidator extends _SimpleNodeValidator {
    _CustomElementNodeValidator(uriPolicy, allowedElements, allowedAttributes, allowedUriAttributes, allowTypeExtension, allowCustomTag) {
      this.allowTypeExtension = allowTypeExtension == true;
      this.allowCustomTag = allowCustomTag == true;
      super._SimpleNodeValidator(uriPolicy, {allowedElements: allowedElements, allowedAttributes: allowedAttributes, allowedUriAttributes: allowedUriAttributes});
    }
    allowsElement(element) {
      if (dart.notNull(this.allowTypeExtension)) {
        let isAttr = element.attributes.get('is');
        if (isAttr != null) {
          return dart.notNull(this.allowedElements.contains(isAttr[dartx.toUpperCase]())) && dart.notNull(this.allowedElements.contains(Element._safeTagName(element)));
        }
      }
      return dart.notNull(this.allowCustomTag) && dart.notNull(this.allowedElements.contains(Element._safeTagName(element)));
    }
    allowsAttribute(element, attributeName, value) {
      if (dart.notNull(this.allowsElement(element))) {
        if (dart.notNull(this.allowTypeExtension) && attributeName == 'is' && dart.notNull(this.allowedElements.contains(value[dartx.toUpperCase]()))) {
          return true;
        }
        return super.allowsAttribute(element, attributeName, value);
      }
      return false;
    }
  }
  dart.setSignature(_CustomElementNodeValidator, {
    constructors: () => ({_CustomElementNodeValidator: [_CustomElementNodeValidator, [UriPolicy, core.Iterable$(core.String), core.Iterable$(core.String), core.Iterable$(core.String), core.bool, core.bool]]})
  });
  const _templateAttrs = Symbol('_templateAttrs');
  class _TemplatingNodeValidator extends _SimpleNodeValidator {
    _TemplatingNodeValidator() {
      this[_templateAttrs] = core.Set$(core.String).from(_TemplatingNodeValidator._TEMPLATE_ATTRS);
      super._SimpleNodeValidator(null, {allowedElements: dart.list(['TEMPLATE'], core.String), allowedAttributes: _TemplatingNodeValidator._TEMPLATE_ATTRS[dartx.map](dart.fn(attr => `TEMPLATE::${attr}`, core.String, [dart.dynamic]))});
    }
    allowsAttribute(element, attributeName, value) {
      if (dart.notNull(super.allowsAttribute(element, attributeName, value))) {
        return true;
      }
      if (attributeName == 'template' && value == "") {
        return true;
      }
      if (element.attributes.get('template') == "") {
        return this[_templateAttrs].contains(attributeName);
      }
      return false;
    }
  }
  dart.setSignature(_TemplatingNodeValidator, {
    constructors: () => ({_TemplatingNodeValidator: [_TemplatingNodeValidator, []]})
  });
  _TemplatingNodeValidator._TEMPLATE_ATTRS = dart.const(dart.list(['bind', 'if', 'ref', 'repeat', 'syntax'], core.String));
  class _SvgNodeValidator extends core.Object {
    allowsElement(element) {
      if (dart.is(element, dart.dynamic)) {
        return false;
      }
      if (dart.is(element, dart.dynamic) && Element._safeTagName(element) == 'foreignObject') {
        return false;
      }
      if (dart.is(element, dart.dynamic)) {
        return true;
      }
      return false;
    }
    allowsAttribute(element, attributeName, value) {
      if (attributeName == 'is' || dart.notNull(attributeName[dartx.startsWith]('on'))) {
        return false;
      }
      return this.allowsElement(element);
    }
  }
  _SvgNodeValidator[dart.implements] = () => [NodeValidator];
  dart.setSignature(_SvgNodeValidator, {
    methods: () => ({
      allowsElement: [core.bool, [Element]],
      allowsAttribute: [core.bool, [Element, core.String, core.String]]
    })
  });
  class ReadyState extends core.Object {}
  ReadyState.LOADING = "loading";
  ReadyState.INTERACTIVE = "interactive";
  ReadyState.COMPLETE = "complete";
  const _list = Symbol('_list');
  const _WrappedList$ = dart.generic(function(E) {
    class _WrappedList extends collection.ListBase$(E) {
      _WrappedList(list) {
        this[_list] = list;
      }
      get iterator() {
        return new (_WrappedIterator$(E))(this[_list][dartx.iterator]);
      }
      get length() {
        return this[_list][dartx.length];
      }
      add(element) {
        dart.as(element, E);
        this[_list][dartx.add](element);
      }
      remove(element) {
        return this[_list][dartx.remove](element);
      }
      clear() {
        this[_list][dartx.clear]();
      }
      get(index) {
        return dart.as(this[_list][dartx.get](index), E);
      }
      set(index, value) {
        dart.as(value, E);
        this[_list][dartx.set](index, value);
        return value;
      }
      set length(newLength) {
        this[_list][dartx.length] = newLength;
      }
      sort(compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, dart.functionType(core.int, [E, E]));
        this[_list][dartx.sort](compare);
      }
      indexOf(element, start) {
        if (start === void 0)
          start = 0;
        return this[_list][dartx.indexOf](element, start);
      }
      lastIndexOf(element, start) {
        if (start === void 0)
          start = null;
        return this[_list][dartx.lastIndexOf](element, start);
      }
      insert(index, element) {
        dart.as(element, E);
        return this[_list][dartx.insert](index, element);
      }
      removeAt(index) {
        return dart.as(this[_list][dartx.removeAt](index), E);
      }
      setRange(start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        this[_list][dartx.setRange](start, end, iterable, skipCount);
      }
      removeRange(start, end) {
        this[_list][dartx.removeRange](start, end);
      }
      replaceRange(start, end, iterable) {
        dart.as(iterable, core.Iterable$(E));
        this[_list][dartx.replaceRange](start, end, iterable);
      }
      fillRange(start, end, fillValue) {
        if (fillValue === void 0)
          fillValue = null;
        dart.as(fillValue, E);
        this[_list][dartx.fillRange](start, end, fillValue);
      }
      get rawList() {
        return dart.as(this[_list], core.List$(Node));
      }
    }
    _WrappedList[dart.implements] = () => [html_common.NodeListWrapper];
    dart.setSignature(_WrappedList, {
      constructors: () => ({_WrappedList: [_WrappedList$(E), [core.List]]}),
      methods: () => ({
        add: [dart.void, [E]],
        get: [E, [core.int]],
        set: [dart.void, [core.int, E]],
        sort: [dart.void, [], [dart.functionType(core.int, [E, E])]],
        insert: [dart.void, [core.int, E]],
        removeAt: [E, [core.int]],
        setRange: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        replaceRange: [dart.void, [core.int, core.int, core.Iterable$(E)]],
        fillRange: [dart.void, [core.int, core.int], [E]]
      })
    });
    dart.defineExtensionMembers(_WrappedList, [
      'add',
      'remove',
      'clear',
      'get',
      'set',
      'sort',
      'indexOf',
      'lastIndexOf',
      'insert',
      'removeAt',
      'setRange',
      'removeRange',
      'replaceRange',
      'fillRange',
      'iterator',
      'length',
      'length'
    ]);
    return _WrappedList;
  });
  let _WrappedList = _WrappedList$();
  const _iterator = Symbol('_iterator');
  const _WrappedIterator$ = dart.generic(function(E) {
    class _WrappedIterator extends core.Object {
      _WrappedIterator(iterator) {
        this[_iterator] = iterator;
      }
      moveNext() {
        return this[_iterator].moveNext();
      }
      get current() {
        return dart.as(this[_iterator].current, E);
      }
    }
    _WrappedIterator[dart.implements] = () => [core.Iterator$(E)];
    dart.setSignature(_WrappedIterator, {
      constructors: () => ({_WrappedIterator: [_WrappedIterator$(E), [core.Iterator]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _WrappedIterator;
  });
  let _WrappedIterator = _WrappedIterator$();
  class _HttpRequestUtils extends core.Object {
    static get(url, onComplete, withCredentials) {
      let request = HttpRequest.new();
      request.open('GET', url, {async: true});
      request.withCredentials = withCredentials;
      request.onReadyStateChange.listen(dart.fn(e => {
        if (request.readyState == HttpRequest.DONE) {
          onComplete(request);
        }
      }));
      request.send();
      return request;
    }
  }
  dart.setSignature(_HttpRequestUtils, {
    statics: () => ({get: [HttpRequest, [core.String, dart.functionType(dart.dynamic, [HttpRequest]), core.bool]]}),
    names: ['get']
  });
  const _array = Symbol('_array');
  const _position = Symbol('_position');
  const _current = Symbol('_current');
  const FixedSizeListIterator$ = dart.generic(function(T) {
    class FixedSizeListIterator extends core.Object {
      FixedSizeListIterator(array) {
        this[_array] = array;
        this[_position] = -1;
        this[_length] = array[dartx.length];
        this[_current] = null;
      }
      moveNext() {
        let nextPosition = dart.notNull(this[_position]) + 1;
        if (dart.notNull(nextPosition) < dart.notNull(this[_length])) {
          this[_current] = this[_array][dartx.get](nextPosition);
          this[_position] = nextPosition;
          return true;
        }
        this[_current] = null;
        this[_position] = this[_length];
        return false;
      }
      get current() {
        return this[_current];
      }
    }
    FixedSizeListIterator[dart.implements] = () => [core.Iterator$(T)];
    dart.setSignature(FixedSizeListIterator, {
      constructors: () => ({FixedSizeListIterator: [FixedSizeListIterator$(T), [core.List$(T)]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return FixedSizeListIterator;
  });
  let FixedSizeListIterator = FixedSizeListIterator$();
  const _VariableSizeListIterator$ = dart.generic(function(T) {
    class _VariableSizeListIterator extends core.Object {
      _VariableSizeListIterator(array) {
        this[_array] = array;
        this[_position] = -1;
        this[_current] = null;
      }
      moveNext() {
        let nextPosition = dart.notNull(this[_position]) + 1;
        if (dart.notNull(nextPosition) < dart.notNull(this[_array][dartx.length])) {
          this[_current] = this[_array][dartx.get](nextPosition);
          this[_position] = nextPosition;
          return true;
        }
        this[_current] = null;
        this[_position] = this[_array][dartx.length];
        return false;
      }
      get current() {
        return this[_current];
      }
    }
    _VariableSizeListIterator[dart.implements] = () => [core.Iterator$(T)];
    dart.setSignature(_VariableSizeListIterator, {
      constructors: () => ({_VariableSizeListIterator: [_VariableSizeListIterator$(T), [core.List$(T)]]}),
      methods: () => ({moveNext: [core.bool, []]})
    });
    return _VariableSizeListIterator;
  });
  let _VariableSizeListIterator = _VariableSizeListIterator$();
  function _convertNativeToDart_Window(win) {
    if (win == null)
      return null;
    return _DOMWindowCrossFrame._createSafe(win);
  }
  dart.fn(_convertNativeToDart_Window, WindowBase, [dart.dynamic]);
  function _convertNativeToDart_EventTarget(e) {
    if (e == null) {
      return null;
    }
    if ("postMessage" in e) {
      let window = _DOMWindowCrossFrame._createSafe(e);
      if (dart.is(window, EventTarget)) {
        return window;
      }
      return null;
    } else
      return dart.as(e, EventTarget);
  }
  dart.fn(_convertNativeToDart_EventTarget, EventTarget, [dart.dynamic]);
  const _window = Symbol('_window');
  function _convertDartToNative_EventTarget(e) {
    if (dart.is(e, _DOMWindowCrossFrame)) {
      return dart.as(dart.dload(e, _window), EventTarget);
    } else {
      return dart.as(e, EventTarget);
    }
  }
  dart.fn(_convertDartToNative_EventTarget, EventTarget, [dart.dynamic]);
  function _convertNativeToDart_XHR_Response(o) {
    if (dart.is(o, Document)) {
      return o;
    }
    return html_common.convertNativeToDart_SerializedScriptValue(o);
  }
  dart.fn(_convertNativeToDart_XHR_Response);
  function _callConstructor(constructor, interceptor) {
    return dart.fn(receiver => {
      dart.dcall(/* Unimplemented unknown name */setNativeSubclassDispatchRecord, receiver, interceptor);
      receiver.constructor = receiver.__proto__.constructor;
      return constructor(receiver);
    });
  }
  dart.fn(_callConstructor);
  function _callAttached(receiver) {
    return dart.dsend(receiver, 'attached');
  }
  dart.fn(_callAttached);
  function _callDetached(receiver) {
    return dart.dsend(receiver, 'detached');
  }
  dart.fn(_callDetached);
  function _callAttributeChanged(receiver, name, oldValue, newValue) {
    return dart.dsend(receiver, 'attributeChanged', name, oldValue, newValue);
  }
  dart.fn(_callAttributeChanged);
  function _makeCallbackMethod(callback) {
    return (function(invokeCallback) {
      return function() {
        return invokeCallback(this);
      };
    })(dart.dcall(/* Unimplemented unknown name */convertDartClosureToJS, callback, 1));
  }
  dart.fn(_makeCallbackMethod);
  function _makeCallbackMethod3(callback) {
    return (function(invokeCallback) {
      return function(arg1, arg2, arg3) {
        return invokeCallback(this, arg1, arg2, arg3);
      };
    })(dart.dcall(/* Unimplemented unknown name */convertDartClosureToJS, callback, 4));
  }
  dart.fn(_makeCallbackMethod3);
  function _registerCustomElement(context, document, tag, type, extendsTagName) {
    let interceptorClass = dart.dcall(/* Unimplemented unknown name */findInterceptorConstructorForType, type);
    if (interceptorClass == null) {
      dart.throw(new core.ArgumentError(type));
    }
    let interceptor = interceptorClass.prototype;
    let constructor = dart.dcall(/* Unimplemented unknown name */findConstructorForNativeSubclassType, type, 'created');
    if (constructor == null) {
      dart.throw(new core.ArgumentError(`${type} has no constructor called 'created'`));
    }
    dart.dcall(/* Unimplemented unknown name */getNativeInterceptor, Element.tag('article'));
    let baseClassName = dart.as(dart.dcall(/* Unimplemented unknown name */findDispatchTagForInterceptorClass, interceptorClass), core.String);
    if (baseClassName == null) {
      dart.throw(new core.ArgumentError(type));
    }
    if (extendsTagName == null) {
      if (baseClassName != 'HTMLElement') {
        dart.throw(new core.UnsupportedError('Class must provide extendsTag if base ' + 'native class is not HtmlElement'));
      }
    } else {
      if (!(document.createElement(extendsTagName) instanceof window[baseClassName])) {
        dart.throw(new core.UnsupportedError('extendsTag does not match base native class'));
      }
    }
    let baseConstructor = context[baseClassName];
    let properties = {};
    properties.createdCallback = {value: _makeCallbackMethod(_callConstructor(constructor, interceptor))};
    properties.attachedCallback = {value: _makeCallbackMethod(_callAttached)};
    properties.detachedCallback = {value: _makeCallbackMethod(_callDetached)};
    properties.attributeChangedCallback = {value: _makeCallbackMethod3(_callAttributeChanged)};
    let baseProto = baseConstructor.prototype;
    let proto = Object.create(baseProto, properties);
    dart.dcall(/* Unimplemented unknown name */setNativeSubclassDispatchRecord, proto, interceptor);
    let options = {prototype: proto};
    if (extendsTagName != null) {
      options.extends = extendsTagName;
    }
    document.registerElement(tag, options);
  }
  dart.fn(_registerCustomElement, dart.void, [dart.dynamic, dart.dynamic, core.String, core.Type, core.String]);
  function _initializeCustomElement(e) {
  }
  dart.fn(_initializeCustomElement, dart.void, [Element]);
  const _interceptor = Symbol('_interceptor');
  const _constructor = Symbol('_constructor');
  const _nativeType = Symbol('_nativeType');
  class _JSElementUpgrader extends core.Object {
    _JSElementUpgrader(document, type, extendsTag) {
      this[_interceptor] = null;
      this[_constructor] = null;
      this[_nativeType] = null;
      let interceptorClass = dart.dcall(/* Unimplemented unknown name */findInterceptorConstructorForType, type);
      if (interceptorClass == null) {
        dart.throw(new core.ArgumentError(type));
      }
      this[_constructor] = dart.dcall(/* Unimplemented unknown name */findConstructorForNativeSubclassType, type, 'created');
      if (this[_constructor] == null) {
        dart.throw(new core.ArgumentError(`${type} has no constructor called 'created'`));
      }
      dart.dcall(/* Unimplemented unknown name */getNativeInterceptor, Element.tag('article'));
      let baseClassName = dart.dcall(/* Unimplemented unknown name */findDispatchTagForInterceptorClass, interceptorClass);
      if (baseClassName == null) {
        dart.throw(new core.ArgumentError(type));
      }
      if (extendsTag == null) {
        if (!dart.equals(baseClassName, 'HTMLElement')) {
          dart.throw(new core.UnsupportedError('Class must provide extendsTag if base ' + 'native class is not HtmlElement'));
        }
        this[_nativeType] = HtmlElement;
      } else {
        let element = document.createElement(extendsTag);
        if (!(element instanceof window[baseClassName])) {
          dart.throw(new core.UnsupportedError('extendsTag does not match base native class'));
        }
        this[_nativeType] = dart.runtimeType(element);
      }
      this[_interceptor] = interceptorClass.prototype;
    }
    upgrade(element) {
      if (!dart.equals(dart.runtimeType(element), this[_nativeType])) {
        dart.throw(new core.ArgumentError(`element is not subclass of ${this[_nativeType]}`));
      }
      dart.dcall(/* Unimplemented unknown name */setNativeSubclassDispatchRecord, element, this[_interceptor]);
      this[_constructor](element);
      return element;
    }
  }
  _JSElementUpgrader[dart.implements] = () => [ElementUpgrader];
  dart.setSignature(_JSElementUpgrader, {
    constructors: () => ({_JSElementUpgrader: [_JSElementUpgrader, [Document, core.Type, core.String]]}),
    methods: () => ({upgrade: [Element, [Element]]})
  });
  class _DOMWindowCrossFrame extends core.Object {
    get history() {
      return _HistoryCrossFrame._createSafe(this[_window].history);
    }
    get location() {
      return _LocationCrossFrame._createSafe(this[_window].location);
    }
    get closed() {
      return this[_window].closed;
    }
    get opener() {
      return _DOMWindowCrossFrame._createSafe(this[_window].opener);
    }
    get parent() {
      return _DOMWindowCrossFrame._createSafe(this[_window].parent);
    }
    get top() {
      return _DOMWindowCrossFrame._createSafe(this[_window].top);
    }
    close() {
      return this[_window].close();
    }
    postMessage(message, targetOrigin, messagePorts) {
      if (messagePorts === void 0)
        messagePorts = null;
      if (messagePorts == null) {
        this[_window].postMessage(html_common.convertDartToNative_SerializedScriptValue(message), targetOrigin);
      } else {
        this[_window].postMessage(html_common.convertDartToNative_SerializedScriptValue(message), targetOrigin, messagePorts);
      }
    }
    _DOMWindowCrossFrame(window) {
      this[_window] = window;
    }
    static _createSafe(w) {
      if (dart.notNull(core.identical(w, exports.window))) {
        return dart.as(w, WindowBase);
      } else {
        return new _DOMWindowCrossFrame(w);
      }
    }
    get on() {
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
    [_addEventListener](type, listener, useCapture) {
      if (type === void 0)
        type = null;
      if (listener === void 0)
        listener = null;
      if (useCapture === void 0)
        useCapture = null;
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
    addEventListener(type, listener, useCapture) {
      if (useCapture === void 0)
        useCapture = null;
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
    dispatchEvent(event) {
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
    [_removeEventListener](type, listener, useCapture) {
      if (type === void 0)
        type = null;
      if (listener === void 0)
        listener = null;
      if (useCapture === void 0)
        useCapture = null;
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
    removeEventListener(type, listener, useCapture) {
      if (useCapture === void 0)
        useCapture = null;
      return dart.throw(new core.UnsupportedError('You can only attach EventListeners to your own window.'));
    }
  }
  _DOMWindowCrossFrame[dart.implements] = () => [WindowBase];
  dart.setSignature(_DOMWindowCrossFrame, {
    constructors: () => ({_DOMWindowCrossFrame: [_DOMWindowCrossFrame, [dart.dynamic]]}),
    methods: () => ({
      close: [dart.void, []],
      postMessage: [dart.void, [dart.dynamic, core.String], [core.List]],
      [_addEventListener]: [dart.void, [], [core.String, EventListener, core.bool]],
      addEventListener: [dart.void, [core.String, EventListener], [core.bool]],
      dispatchEvent: [core.bool, [Event]],
      [_removeEventListener]: [dart.void, [], [core.String, EventListener, core.bool]],
      removeEventListener: [dart.void, [core.String, EventListener], [core.bool]]
    }),
    statics: () => ({_createSafe: [WindowBase, [dart.dynamic]]}),
    names: ['_createSafe']
  });
  class _LocationCrossFrame extends core.Object {
    set href(val) {
      return _LocationCrossFrame._setHref(this[_location], val);
    }
    static _setHref(location, val) {
      location.href = val;
    }
    _LocationCrossFrame(location) {
      this[_location] = location;
    }
    static _createSafe(location) {
      if (dart.notNull(core.identical(location, exports.window.location))) {
        return dart.as(location, LocationBase);
      } else {
        return new _LocationCrossFrame(location);
      }
    }
  }
  _LocationCrossFrame[dart.implements] = () => [LocationBase];
  dart.setSignature(_LocationCrossFrame, {
    constructors: () => ({_LocationCrossFrame: [_LocationCrossFrame, [dart.dynamic]]}),
    statics: () => ({
      _setHref: [dart.void, [dart.dynamic, dart.dynamic]],
      _createSafe: [LocationBase, [dart.dynamic]]
    }),
    names: ['_setHref', '_createSafe']
  });
  const _history = Symbol('_history');
  class _HistoryCrossFrame extends core.Object {
    back() {
      return this[_history].back();
    }
    forward() {
      return this[_history].forward();
    }
    go(distance) {
      return this[_history].go(distance);
    }
    _HistoryCrossFrame(history) {
      this[_history] = history;
    }
    static _createSafe(h) {
      if (dart.notNull(core.identical(h, exports.window.history))) {
        return dart.as(h, HistoryBase);
      } else {
        return new _HistoryCrossFrame(h);
      }
    }
  }
  _HistoryCrossFrame[dart.implements] = () => [HistoryBase];
  dart.setSignature(_HistoryCrossFrame, {
    constructors: () => ({_HistoryCrossFrame: [_HistoryCrossFrame, [dart.dynamic]]}),
    methods: () => ({
      back: [dart.void, []],
      forward: [dart.void, []],
      go: [dart.void, [core.int]]
    }),
    statics: () => ({_createSafe: [HistoryBase, [dart.dynamic]]}),
    names: ['_createSafe']
  });
  dart.defineLazyProperties(KeyEvent, {
    get _keyboardEventDispatchRecord() {
      return KeyEvent._makeRecord();
    },
    get keyDownEvent() {
      return new _KeyboardEventHandler('keydown');
    },
    set keyDownEvent(_) {},
    get keyUpEvent() {
      return new _KeyboardEventHandler('keyup');
    },
    set keyUpEvent(_) {},
    get keyPressEvent() {
      return new _KeyboardEventHandler('keypress');
    },
    set keyPressEvent(_) {}
  });
  class Platform extends core.Object {}
  Platform.supportsSimd = false;
  dart.defineLazyProperties(Platform, {
    get supportsTypedData() {
      return !!window.ArrayBuffer;
    }
  });
  function _wrapZone(callback) {
    if (dart.equals(async.Zone.current, async.Zone.ROOT))
      return callback;
    if (callback == null)
      return null;
    return async.Zone.current.bindUnaryCallback(callback, {runGuarded: true});
  }
  dart.fn(_wrapZone, dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic])]);
  function _wrapBinaryZone(callback) {
    if (dart.equals(async.Zone.current, async.Zone.ROOT))
      return callback;
    if (callback == null)
      return null;
    return async.Zone.current.bindBinaryCallback(callback, {runGuarded: true});
  }
  dart.fn(_wrapBinaryZone, dart.dynamic, [dart.functionType(dart.dynamic, [dart.dynamic, dart.dynamic])]);
  function query(relativeSelectors) {
    return exports.document.query(relativeSelectors);
  }
  dart.fn(query, Element, [core.String]);
  function queryAll(relativeSelectors) {
    return exports.document.queryAll(relativeSelectors);
  }
  dart.fn(queryAll, ElementList$(Element), [core.String]);
  function querySelector(selectors) {
    return exports.document.querySelector(selectors);
  }
  dart.fn(querySelector, Element, [core.String]);
  function querySelectorAll(selectors) {
    return exports.document.querySelectorAll(selectors);
  }
  dart.fn(querySelectorAll, ElementList$(Element), [core.String]);
  class ElementUpgrader extends core.Object {}
  class NodeValidator extends core.Object {
    static new(opts) {
      let uriPolicy = opts && 'uriPolicy' in opts ? opts.uriPolicy : null;
      return new _Html5NodeValidator({uriPolicy: uriPolicy});
    }
    static throws(base) {
      return new _ThrowsNodeValidator(base);
    }
  }
  dart.setSignature(NodeValidator, {
    constructors: () => ({
      new: [NodeValidator, [], {uriPolicy: UriPolicy}],
      throws: [NodeValidator, [NodeValidator]]
    })
  });
  class NodeTreeSanitizer extends core.Object {
    static new(validator) {
      return new _ValidatingTreeSanitizer(validator);
    }
  }
  dart.setSignature(NodeTreeSanitizer, {
    constructors: () => ({new: [NodeTreeSanitizer, [NodeValidator]]})
  });
  class _TrustedHtmlTreeSanitizer extends core.Object {
    _TrustedHtmlTreeSanitizer() {
    }
    sanitizeTree(node) {}
  }
  _TrustedHtmlTreeSanitizer[dart.implements] = () => [NodeTreeSanitizer];
  dart.setSignature(_TrustedHtmlTreeSanitizer, {
    constructors: () => ({_TrustedHtmlTreeSanitizer: [_TrustedHtmlTreeSanitizer, []]}),
    methods: () => ({sanitizeTree: [dart.void, [Node]]})
  });
  NodeTreeSanitizer.trusted = dart.const(new _TrustedHtmlTreeSanitizer());
  class UriPolicy extends core.Object {
    static new() {
      return new _SameOriginUriPolicy();
    }
  }
  dart.setSignature(UriPolicy, {
    constructors: () => ({new: [UriPolicy, []]})
  });
  const _hiddenAnchor = Symbol('_hiddenAnchor');
  const _loc = Symbol('_loc');
  class _SameOriginUriPolicy extends core.Object {
    _SameOriginUriPolicy() {
      this[_hiddenAnchor] = AnchorElement.new();
      this[_loc] = exports.window.location;
    }
    allowsUri(uri) {
      this[_hiddenAnchor].href = uri;
      return this[_hiddenAnchor].hostname == this[_loc].hostname && this[_hiddenAnchor].port == this[_loc].port && this[_hiddenAnchor].protocol == this[_loc].protocol || this[_hiddenAnchor].hostname == '' && this[_hiddenAnchor].port == '' && (this[_hiddenAnchor].protocol == ':' || this[_hiddenAnchor].protocol == '');
    }
  }
  _SameOriginUriPolicy[dart.implements] = () => [UriPolicy];
  dart.setSignature(_SameOriginUriPolicy, {
    methods: () => ({allowsUri: [core.bool, [core.String]]})
  });
  class _ThrowsNodeValidator extends core.Object {
    _ThrowsNodeValidator(validator) {
      this.validator = validator;
    }
    allowsElement(element) {
      if (!dart.notNull(this.validator.allowsElement(element))) {
        dart.throw(new core.ArgumentError(Element._safeTagName(element)));
      }
      return true;
    }
    allowsAttribute(element, attributeName, value) {
      if (!dart.notNull(this.validator.allowsAttribute(element, attributeName, value))) {
        dart.throw(new core.ArgumentError(`${Element._safeTagName(element)}[${attributeName}="${value}"]`));
      }
    }
  }
  _ThrowsNodeValidator[dart.implements] = () => [NodeValidator];
  dart.setSignature(_ThrowsNodeValidator, {
    constructors: () => ({_ThrowsNodeValidator: [_ThrowsNodeValidator, [NodeValidator]]}),
    methods: () => ({
      allowsElement: [core.bool, [Element]],
      allowsAttribute: [core.bool, [Element, core.String, core.String]]
    })
  });
  const _removeNode = Symbol('_removeNode');
  const _sanitizeElement = Symbol('_sanitizeElement');
  const _sanitizeUntrustedElement = Symbol('_sanitizeUntrustedElement');
  class _ValidatingTreeSanitizer extends core.Object {
    _ValidatingTreeSanitizer(validator) {
      this.validator = validator;
    }
    sanitizeTree(node) {
      const walk = (function(node, parent) {
        this.sanitizeNode(node, parent);
        let child = node.lastChild;
        while (child != null) {
          let nextChild = child.previousNode;
          walk(child, node);
          child = nextChild;
        }
      }).bind(this);
      dart.fn(walk, dart.void, [Node, Node]);
      walk(node, null);
    }
    [_removeNode](node, parent) {
      if (parent == null) {
        node.remove();
      } else {
        parent[_removeChild](node);
      }
    }
    [_sanitizeUntrustedElement](element, parent) {
      let corrupted = true;
      let attrs = null;
      let isAttr = null;
      try {
        attrs = dart.dload(element, 'attributes');
        isAttr = dart.dindex(attrs, 'is');
        let corruptedTest1 = Element._hasCorruptedAttributes(dart.as(element, Element));
        corrupted = dart.notNull(corruptedTest1) ? true : Element._hasCorruptedAttributesAdditionalCheck(dart.as(element, Element));
      } catch (e) {
      }

      let elementText = 'element unprintable';
      try {
        elementText = dart.toString(element);
      } catch (e) {
      }

      try {
        let elementTagName = Element._safeTagName(element);
        this[_sanitizeElement](dart.as(element, Element), parent, corrupted, elementText, elementTagName, dart.as(attrs, core.Map), dart.as(isAttr, core.String));
      } catch (e$) {
        if (dart.is(e$, core.ArgumentError)) {
          throw e$;
        } else {
          let e = e$;
          this[_removeNode](dart.as(element, Node), parent);
          exports.window.console.warn(`Removing corrupted element ${elementText}`);
        }
      }

    }
    [_sanitizeElement](element, parent, corrupted, text, tag, attrs, isAttr) {
      if (false != corrupted) {
        this[_removeNode](element, parent);
        exports.window.console.warn(`Removing element due to corrupted attributes on <${text}>`);
        return;
      }
      if (!dart.notNull(this.validator.allowsElement(element))) {
        this[_removeNode](element, parent);
        exports.window.console.warn(`Removing disallowed element <${tag}> from ${parent}`);
        return;
      }
      if (isAttr != null) {
        if (!dart.notNull(this.validator.allowsAttribute(element, 'is', isAttr))) {
          this[_removeNode](element, parent);
          exports.window.console.warn('Removing disallowed type extension ' + `<${tag} is="${isAttr}">`);
          return;
        }
      }
      let keys = attrs.keys[dartx.toList]();
      for (let i = dart.notNull(attrs.length) - 1; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
        let name = keys[dartx.get](i);
        if (!dart.notNull(this.validator.allowsAttribute(element, dart.as(dart.dsend(name, 'toLowerCase'), core.String), dart.as(attrs.get(name), core.String)))) {
          exports.window.console.warn('Removing disallowed attribute ' + `<${tag} ${name}="${attrs.get(name)}">`);
          attrs.remove(name);
        }
      }
      if (dart.is(element, TemplateElement)) {
        let template = element;
        this.sanitizeTree(template.content);
      }
    }
    sanitizeNode(node, parent) {
      switch (node.nodeType) {
        case Node.ELEMENT_NODE:
        {
          this[_sanitizeUntrustedElement](node, parent);
          break;
        }
        case Node.COMMENT_NODE:
        case Node.DOCUMENT_FRAGMENT_NODE:
        case Node.TEXT_NODE:
        case Node.CDATA_SECTION_NODE:
        {
          break;
        }
        default:
        {
          this[_removeNode](node, parent);
        }
      }
    }
  }
  _ValidatingTreeSanitizer[dart.implements] = () => [NodeTreeSanitizer];
  dart.setSignature(_ValidatingTreeSanitizer, {
    constructors: () => ({_ValidatingTreeSanitizer: [_ValidatingTreeSanitizer, [NodeValidator]]}),
    methods: () => ({
      sanitizeTree: [dart.void, [Node]],
      [_removeNode]: [dart.void, [Node, Node]],
      [_sanitizeUntrustedElement]: [dart.void, [dart.dynamic, Node]],
      [_sanitizeElement]: [dart.void, [Element, Node, core.bool, core.String, core.String, core.Map, core.String]],
      sanitizeNode: [dart.void, [Node, Node]]
    })
  });
  dart.copyProperties(exports, {
    get window() {
      return dart.as(wrap_jso(window), Window);
    },
    get document() {
      return dart.as(wrap_jso(document), HtmlDocument);
    }
  });
  const _F1 = dart.typedef('_F1', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  dart.defineLazyProperties(exports, {
    get _wrapper() {
      return Symbol("dart_wrapper");
    }
  });
  function unwrap_jso(wrapped) {
    if (dart.is(wrapped, DartHtmlDomObject)) {
      return dart.dload(wrapped, 'raw');
    }
    if (dart.is(wrapped, _F1)) {
      if (wrapped.hasOwnProperty(exports._wrapper)) {
        return wrapped[exports._wrapper];
      }
      let f = dart.fn(e => dart.dcall(wrapped, wrap_jso(e)));
      wrapped[exports._wrapper] = f;
      return f;
    }
    return wrapped;
  }
  dart.fn(unwrap_jso);
  function wrap_jso(jso) {
    if (jso == null || typeof jso == 'boolean' || typeof jso == 'number' || typeof jso == 'string') {
      return jso;
    }
    if (jso.hasOwnProperty(exports._wrapper)) {
      return jso[exports._wrapper];
    }
    let f = null;
    let constructor = jso.constructor;
    do {
      let name = constructor.name;
      f = getHtmlCreateFunction(name);
      constructor = constructor.__proto__;
    } while (f == null);
    let wrapped = dart.dcall(f);
    dart.dput(wrapped, 'raw', jso);
    jso[exports._wrapper] = wrapped;
    return wrapped;
  }
  dart.fn(wrap_jso);
  function createCustomUpgrader(customElementClass, $this) {
    return $this;
  }
  dart.fn(createCustomUpgrader, dart.dynamic, [core.Type, dart.dynamic]);
  dart.defineLazyProperties(exports, {
    get htmlBlinkFunctionMap() {
      return dart.map({Attr: dart.fn(() => _Attr.internalCreate_Attr, dart.functionType(_Attr, []), []), CSSStyleDeclaration: dart.fn(() => CssStyleDeclaration.internalCreateCssStyleDeclaration, dart.functionType(CssStyleDeclaration, []), []), DOMImplementation: dart.fn(() => DomImplementation.internalCreateDomImplementation, dart.functionType(DomImplementation, []), []), DOMTokenList: dart.fn(() => DomTokenList.internalCreateDomTokenList, dart.functionType(DomTokenList, []), []), DocumentFragment: dart.fn(() => DocumentFragment.internalCreateDocumentFragment, dart.functionType(DocumentFragment, []), []), HTMLAnchorElement: dart.fn(() => AnchorElement.internalCreateAnchorElement, dart.functionType(AnchorElement, []), []), HTMLBodyElement: dart.fn(() => BodyElement.internalCreateBodyElement, dart.functionType(BodyElement, []), []), HTMLButtonElement: dart.fn(() => ButtonElement.internalCreateButtonElement, dart.functionType(ButtonElement, []), []), HTMLDivElement: dart.fn(() => DivElement.internalCreateDivElement, dart.functionType(DivElement, []), []), HTMLDocument: dart.fn(() => HtmlDocument.internalCreateHtmlDocument, dart.functionType(HtmlDocument, []), []), HTMLElement: dart.fn(() => HtmlElement.internalCreateHtmlElement, dart.functionType(HtmlElement, []), []), HTMLHeadElement: dart.fn(() => HeadElement.internalCreateHeadElement, dart.functionType(HeadElement, []), []), HTMLHeadingElement: dart.fn(() => HeadingElement.internalCreateHeadingElement, dart.functionType(HeadingElement, []), []), HTMLInputElement: dart.fn(() => InputElement.internalCreateInputElement, dart.functionType(InputElement, []), []), HTMLLIElement: dart.fn(() => LIElement.internalCreateLIElement, dart.functionType(LIElement, []), []), HTMLLabelElement: dart.fn(() => LabelElement.internalCreateLabelElement, dart.functionType(LabelElement, []), []), HTMLParagraphElement: dart.fn(() => ParagraphElement.internalCreateParagraphElement, dart.functionType(ParagraphElement, []), []), HTMLScriptElement: dart.fn(() => ScriptElement.internalCreateScriptElement, dart.functionType(ScriptElement, []), []), HTMLSpanElement: dart.fn(() => SpanElement.internalCreateSpanElement, dart.functionType(SpanElement, []), []), HTMLStyleElement: dart.fn(() => StyleElement.internalCreateStyleElement, dart.functionType(StyleElement, []), []), HTMLTemplateElement: dart.fn(() => TemplateElement.internalCreateTemplateElement, dart.functionType(TemplateElement, []), []), HTMLUListElement: dart.fn(() => UListElement.internalCreateUListElement, dart.functionType(UListElement, []), []), KeyboardEvent: dart.fn(() => KeyboardEvent.internalCreateKeyboardEvent, dart.functionType(KeyboardEvent, []), []), MouseEvent: dart.fn(() => MouseEvent.internalCreateMouseEvent, dart.functionType(MouseEvent, []), []), NamedNodeMap: dart.fn(() => _NamedNodeMap.internalCreate_NamedNodeMap, dart.functionType(_NamedNodeMap, []), []), Navigator: dart.fn(() => Navigator.internalCreateNavigator, dart.functionType(Navigator, []), []), NodeList: dart.fn(() => NodeList.internalCreateNodeList, dart.functionType(NodeList, []), []), Range: dart.fn(() => Range.internalCreateRange, dart.functionType(Range, []), []), Text: dart.fn(() => Text.internalCreateText, dart.functionType(Text, []), []), Window: dart.fn(() => Window.internalCreateWindow, dart.functionType(Window, []), []), XMLHttpRequest: dart.fn(() => HttpRequest.internalCreateHttpRequest, dart.functionType(HttpRequest, []), []), XMLHttpRequestProgressEvent: dart.fn(() => _XMLHttpRequestProgressEvent.internalCreate_XMLHttpRequestProgressEvent, dart.functionType(_XMLHttpRequestProgressEvent, []), [])});
    },
    get typeNames() {
      return core.Set.new();
    },
    set typeNames(_) {},
    get typeKeys() {
      return [];
    },
    set typeKeys(_) {}
  });
  function getHtmlCreateFunction(key) {
    let result = null;
    if (!dart.notNull(exports.typeNames.contains(key))) {
      exports.typeNames.add(key);
      exports.typeKeys[dartx.add](key);
    }
    result = _getHtmlFunction(key);
    if (result != null) {
      return result;
    }
    return null;
  }
  dart.fn(getHtmlCreateFunction, dart.dynamic, [core.String]);
  function _getHtmlFunction(key) {
    if (dart.notNull(exports.htmlBlinkFunctionMap.containsKey(key))) {
      return dart.as(dart.dcall(exports.htmlBlinkFunctionMap.get(key)), core.Function);
    }
    return null;
  }
  dart.fn(_getHtmlFunction, core.Function, [core.String]);
  const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(core.bool, [Element]));
  const __CastType2 = dart.typedef('__CastType2', () => dart.functionType(dart.dynamic, [dart.dynamic]));
  // Exports:
  exports.DartHtmlDomObject = DartHtmlDomObject;
  exports.EventTarget = EventTarget;
  exports.Node = Node;
  exports.Element = Element;
  exports.HtmlElement = HtmlElement;
  exports.AnchorElement = AnchorElement;
  exports.BodyElement = BodyElement;
  exports.EventStreamProvider$ = EventStreamProvider$;
  exports.EventStreamProvider = EventStreamProvider;
  exports.Event = Event;
  exports.MessageEvent = MessageEvent;
  exports.PopStateEvent = PopStateEvent;
  exports.StorageEvent = StorageEvent;
  exports.ButtonElement = ButtonElement;
  exports.ChildNode = ChildNode;
  exports.Console = Console;
  exports.CssStyleDeclarationBase = CssStyleDeclarationBase;
  exports.CssStyleDeclaration = CssStyleDeclaration;
  exports.DivElement = DivElement;
  exports.Document = Document;
  exports.SecurityPolicyViolationEvent = SecurityPolicyViolationEvent;
  exports.DocumentFragment = DocumentFragment;
  exports.DomImplementation = DomImplementation;
  exports.DomTokenList = DomTokenList;
  exports.ElementList$ = ElementList$;
  exports.ElementList = ElementList;
  exports.UIEvent = UIEvent;
  exports.MouseEvent = MouseEvent;
  exports.KeyboardEvent = KeyboardEvent;
  exports.TouchEvent = TouchEvent;
  exports.Events = Events;
  exports.ElementEvents = ElementEvents;
  exports.HRElement = HRElement;
  exports.HashChangeEvent = HashChangeEvent;
  exports.HeadElement = HeadElement;
  exports.Headers = Headers;
  exports.HeadersForEachCallback = HeadersForEachCallback;
  exports.HeadingElement = HeadingElement;
  exports.History = History;
  exports.ImmutableListMixin$ = ImmutableListMixin$;
  exports.ImmutableListMixin = ImmutableListMixin;
  exports.HtmlCollection = HtmlCollection;
  exports.HtmlDocument = HtmlDocument;
  exports.HtmlFormControlsCollection = HtmlFormControlsCollection;
  exports.HtmlHtmlElement = HtmlHtmlElement;
  exports.HtmlOptionsCollection = HtmlOptionsCollection;
  exports.HttpRequestEventTarget = HttpRequestEventTarget;
  exports.HttpRequest = HttpRequest;
  exports.ProgressEvent = ProgressEvent;
  exports.HttpRequestUpload = HttpRequestUpload;
  exports.IFrameElement = IFrameElement;
  exports.ImageBitmap = ImageBitmap;
  exports.ImageData = ImageData;
  exports.ImageElement = ImageElement;
  exports.InjectedScriptHost = InjectedScriptHost;
  exports.InputElement = InputElement;
  exports.InputElementBase = InputElementBase;
  exports.HiddenInputElement = HiddenInputElement;
  exports.TextInputElementBase = TextInputElementBase;
  exports.SearchInputElement = SearchInputElement;
  exports.TextInputElement = TextInputElement;
  exports.UrlInputElement = UrlInputElement;
  exports.TelephoneInputElement = TelephoneInputElement;
  exports.EmailInputElement = EmailInputElement;
  exports.PasswordInputElement = PasswordInputElement;
  exports.RangeInputElementBase = RangeInputElementBase;
  exports.DateInputElement = DateInputElement;
  exports.MonthInputElement = MonthInputElement;
  exports.WeekInputElement = WeekInputElement;
  exports.TimeInputElement = TimeInputElement;
  exports.LocalDateTimeInputElement = LocalDateTimeInputElement;
  exports.NumberInputElement = NumberInputElement;
  exports.RangeInputElement = RangeInputElement;
  exports.CheckboxInputElement = CheckboxInputElement;
  exports.RadioButtonInputElement = RadioButtonInputElement;
  exports.FileUploadInputElement = FileUploadInputElement;
  exports.SubmitButtonInputElement = SubmitButtonInputElement;
  exports.ImageButtonInputElement = ImageButtonInputElement;
  exports.ResetButtonInputElement = ResetButtonInputElement;
  exports.ButtonInputElement = ButtonInputElement;
  exports.InputMethodContext = InputMethodContext;
  exports.LIElement = LIElement;
  exports.LabelElement = LabelElement;
  exports.Location = Location;
  exports.MenuItemElement = MenuItemElement;
  exports.MessageChannel = MessageChannel;
  exports.MessagePort = MessagePort;
  exports.MetaElement = MetaElement;
  exports.Metadata = Metadata;
  exports.MetadataCallback = MetadataCallback;
  exports.MeterElement = MeterElement;
  exports.MidiAccess = MidiAccess;
  exports.MidiConnectionEvent = MidiConnectionEvent;
  exports.MidiPort = MidiPort;
  exports.MidiInput = MidiInput;
  exports.MidiMessageEvent = MidiMessageEvent;
  exports.MidiInputMap = MidiInputMap;
  exports.MidiOutput = MidiOutput;
  exports.MidiOutputMap = MidiOutputMap;
  exports.MimeType = MimeType;
  exports.MimeTypeArray = MimeTypeArray;
  exports.ModElement = ModElement;
  exports.MutationCallback = MutationCallback;
  exports.MutationObserver = MutationObserver;
  exports.MutationRecord = MutationRecord;
  exports.Navigator = Navigator;
  exports.NavigatorCpu = NavigatorCpu;
  exports.NavigatorID = NavigatorID;
  exports.NavigatorLanguage = NavigatorLanguage;
  exports.NavigatorOnLine = NavigatorOnLine;
  exports.NavigatorUserMediaError = NavigatorUserMediaError;
  exports.NetworkInformation = NetworkInformation;
  exports.NodeFilter = NodeFilter;
  exports.NodeIterator = NodeIterator;
  exports.NodeList = NodeList;
  exports.Notification = Notification;
  exports.OListElement = OListElement;
  exports.ObjectElement = ObjectElement;
  exports.OptGroupElement = OptGroupElement;
  exports.OptionElement = OptionElement;
  exports.OutputElement = OutputElement;
  exports.OverflowEvent = OverflowEvent;
  exports.PageTransitionEvent = PageTransitionEvent;
  exports.ParagraphElement = ParagraphElement;
  exports.ParamElement = ParamElement;
  exports.ParentNode = ParentNode;
  exports.Path2D = Path2D;
  exports.Performance = Performance;
  exports.PerformanceEntry = PerformanceEntry;
  exports.PerformanceMark = PerformanceMark;
  exports.PerformanceMeasure = PerformanceMeasure;
  exports.PerformanceNavigation = PerformanceNavigation;
  exports.PerformanceResourceTiming = PerformanceResourceTiming;
  exports.PerformanceTiming = PerformanceTiming;
  exports.PictureElement = PictureElement;
  exports.Plugin = Plugin;
  exports.PluginArray = PluginArray;
  exports.PluginPlaceholderElement = PluginPlaceholderElement;
  exports.PositionError = PositionError;
  exports.PreElement = PreElement;
  exports.Presentation = Presentation;
  exports.ProgressElement = ProgressElement;
  exports.PushEvent = PushEvent;
  exports.PushManager = PushManager;
  exports.PushRegistration = PushRegistration;
  exports.QuoteElement = QuoteElement;
  exports.RtcStatsCallback = RtcStatsCallback;
  exports.Range = Range;
  exports.ReadableStream = ReadableStream;
  exports.RelatedEvent = RelatedEvent;
  exports.RequestAnimationFrameCallback = RequestAnimationFrameCallback;
  exports.ResourceProgressEvent = ResourceProgressEvent;
  exports.RtcDataChannel = RtcDataChannel;
  exports.RtcDataChannelEvent = RtcDataChannelEvent;
  exports.RtcDtmfSender = RtcDtmfSender;
  exports.RtcDtmfToneChangeEvent = RtcDtmfToneChangeEvent;
  exports.RtcIceCandidate = RtcIceCandidate;
  exports.RtcIceCandidateEvent = RtcIceCandidateEvent;
  exports.RtcPeerConnection = RtcPeerConnection;
  exports.RtcSessionDescription = RtcSessionDescription;
  exports.RtcStatsReport = RtcStatsReport;
  exports.RtcStatsResponse = RtcStatsResponse;
  exports.Screen = Screen;
  exports.ScreenOrientation = ScreenOrientation;
  exports.ScriptElement = ScriptElement;
  exports.SelectElement = SelectElement;
  exports.Selection = Selection;
  exports.ServiceWorkerClient = ServiceWorkerClient;
  exports.ServiceWorkerClients = ServiceWorkerClients;
  exports.ServiceWorkerContainer = ServiceWorkerContainer;
  exports.WorkerGlobalScope = WorkerGlobalScope;
  exports.ServiceWorkerGlobalScope = ServiceWorkerGlobalScope;
  exports.ServiceWorkerRegistration = ServiceWorkerRegistration;
  exports.ShadowElement = ShadowElement;
  exports.ShadowRoot = ShadowRoot;
  exports.SharedWorker = SharedWorker;
  exports.SharedWorkerGlobalScope = SharedWorkerGlobalScope;
  exports.SourceBuffer = SourceBuffer;
  exports.SourceBufferList = SourceBufferList;
  exports.SourceElement = SourceElement;
  exports.SourceInfo = SourceInfo;
  exports.SpanElement = SpanElement;
  exports.SpeechGrammar = SpeechGrammar;
  exports.SpeechGrammarList = SpeechGrammarList;
  exports.SpeechRecognition = SpeechRecognition;
  exports.SpeechRecognitionError = SpeechRecognitionError;
  exports.SpeechRecognitionEvent = SpeechRecognitionEvent;
  exports.SpeechRecognitionAlternative = SpeechRecognitionAlternative;
  exports.SpeechRecognitionResult = SpeechRecognitionResult;
  exports.SpeechSynthesis = SpeechSynthesis;
  exports.SpeechSynthesisEvent = SpeechSynthesisEvent;
  exports.SpeechSynthesisUtterance = SpeechSynthesisUtterance;
  exports.SpeechSynthesisVoice = SpeechSynthesisVoice;
  exports.Storage = Storage;
  exports.StorageErrorCallback = StorageErrorCallback;
  exports.StorageInfo = StorageInfo;
  exports.StorageQuota = StorageQuota;
  exports.StorageQuotaCallback = StorageQuotaCallback;
  exports.StorageUsageCallback = StorageUsageCallback;
  exports.StyleElement = StyleElement;
  exports.StyleMedia = StyleMedia;
  exports.StyleSheet = StyleSheet;
  exports.TableCaptionElement = TableCaptionElement;
  exports.TableCellElement = TableCellElement;
  exports.TableColElement = TableColElement;
  exports.TableElement = TableElement;
  exports.TableRowElement = TableRowElement;
  exports.TableSectionElement = TableSectionElement;
  exports.TemplateElement = TemplateElement;
  exports.CharacterData = CharacterData;
  exports.Text = Text;
  exports.TextAreaElement = TextAreaElement;
  exports.TextEvent = TextEvent;
  exports.TextMetrics = TextMetrics;
  exports.TextTrack = TextTrack;
  exports.TextTrackCue = TextTrackCue;
  exports.TextTrackCueList = TextTrackCueList;
  exports.TextTrackList = TextTrackList;
  exports.TrackEvent = TrackEvent;
  exports.TimeRanges = TimeRanges;
  exports.TimeoutHandler = TimeoutHandler;
  exports.Timing = Timing;
  exports.TitleElement = TitleElement;
  exports.Touch = Touch;
  exports.TouchList = TouchList;
  exports.TrackElement = TrackElement;
  exports.TransitionEvent = TransitionEvent;
  exports.TreeWalker = TreeWalker;
  exports.UListElement = UListElement;
  exports.UnknownElement = UnknownElement;
  exports.Url = Url;
  exports.UrlUtils = UrlUtils;
  exports.UrlUtilsReadOnly = UrlUtilsReadOnly;
  exports.ValidityState = ValidityState;
  exports.VoidCallback = VoidCallback;
  exports.Window = Window;
  exports.WindowBase64 = WindowBase64;
  exports.WindowEventHandlers = WindowEventHandlers;
  exports.Worker = Worker;
  exports.WorkerPerformance = WorkerPerformance;
  exports.XPathEvaluator = XPathEvaluator;
  exports.XPathExpression = XPathExpression;
  exports.XPathNSResolver = XPathNSResolver;
  exports.XPathResult = XPathResult;
  exports.XmlDocument = XmlDocument;
  exports.XmlSerializer = XmlSerializer;
  exports.XsltProcessor = XsltProcessor;
  exports.CanvasImageSource = CanvasImageSource;
  exports.WindowBase = WindowBase;
  exports.LocationBase = LocationBase;
  exports.HistoryBase = HistoryBase;
  exports.CssClassSet = CssClassSet;
  exports.CssRect = CssRect;
  exports.Dimension = Dimension;
  exports.EventListener = EventListener;
  exports.ElementStream$ = ElementStream$;
  exports.ElementStream = ElementStream;
  exports.CustomStream$ = CustomStream$;
  exports.CustomStream = CustomStream;
  exports.KeyEvent = KeyEvent;
  exports.KeyCode = KeyCode;
  exports.KeyLocation = KeyLocation;
  exports.KeyboardEventStream = KeyboardEventStream;
  exports.NodeValidatorBuilder = NodeValidatorBuilder;
  exports.ReadyState = ReadyState;
  exports.FixedSizeListIterator$ = FixedSizeListIterator$;
  exports.FixedSizeListIterator = FixedSizeListIterator;
  exports.Platform = Platform;
  exports.query = query;
  exports.queryAll = queryAll;
  exports.querySelector = querySelector;
  exports.querySelectorAll = querySelectorAll;
  exports.ElementUpgrader = ElementUpgrader;
  exports.NodeValidator = NodeValidator;
  exports.NodeTreeSanitizer = NodeTreeSanitizer;
  exports.UriPolicy = UriPolicy;
  exports.unwrap_jso = unwrap_jso;
  exports.wrap_jso = wrap_jso;
  exports.createCustomUpgrader = createCustomUpgrader;
  exports.getHtmlCreateFunction = getHtmlCreateFunction;
});
dart_library.library('dart/html_common', null, /* Imports */[
  "dart/_runtime",
  'dart/_metadata',
  'dart/core',
  'dart/html',
  'dart/typed_data',
  'dart/_js_helper',
  'dart/_native_typed_data',
  'dart/async',
  'dart/collection',
  'dart/_internal'
], /* Lazy imports */[
], function(exports, dart, _metadata, core, html, typed_data, _js_helper, _native_typed_data, async, collection, _internal) {
  'use strict';
  let dartx = dart.dartx;
  dart.export_(exports, _metadata);
  const _validateToken = Symbol('_validateToken');
  class CssClassSetImpl extends core.Object {
    [_validateToken](value) {
      if (dart.notNull(CssClassSetImpl._validTokenRE.hasMatch(value)))
        return value;
      dart.throw(new core.ArgumentError.value(value, 'value', 'Not a valid class token'));
    }
    toString() {
      return this.readClasses().join(' ');
    }
    toggle(value, shouldAdd) {
      if (shouldAdd === void 0)
        shouldAdd = null;
      this[_validateToken](value);
      let s = this.readClasses();
      let result = false;
      if (shouldAdd == null)
        shouldAdd = !dart.notNull(s.contains(value));
      if (dart.notNull(shouldAdd)) {
        s.add(value);
        result = true;
      } else {
        s.remove(value);
      }
      this.writeClasses(s);
      return result;
    }
    get frozen() {
      return false;
    }
    get iterator() {
      return this.readClasses().iterator;
    }
    [Symbol.iterator]() {
      return new dart.JsIterator(this.iterator);
    }
    forEach(f) {
      this.readClasses().forEach(f);
    }
    join(separator) {
      if (separator === void 0)
        separator = "";
      return this.readClasses().join(separator);
    }
    map(f) {
      return this.readClasses().map(f);
    }
    where(f) {
      return this.readClasses().where(f);
    }
    expand(f) {
      return this.readClasses().expand(f);
    }
    every(f) {
      return this.readClasses().every(f);
    }
    any(f) {
      return this.readClasses().any(f);
    }
    get isEmpty() {
      return this.readClasses().isEmpty;
    }
    get isNotEmpty() {
      return this.readClasses().isNotEmpty;
    }
    get length() {
      return this.readClasses().length;
    }
    reduce(combine) {
      return this.readClasses().reduce(combine);
    }
    fold(initialValue, combine) {
      return this.readClasses().fold(initialValue, combine);
    }
    contains(value) {
      if (!(typeof value == 'string'))
        return false;
      this[_validateToken](dart.as(value, core.String));
      return this.readClasses().contains(value);
    }
    lookup(value) {
      return dart.as(dart.notNull(this.contains(value)) ? value : null, core.String);
    }
    add(value) {
      this[_validateToken](value);
      return dart.as(this.modify(dart.fn(s => dart.dsend(s, 'add', value))), core.bool);
    }
    remove(value) {
      this[_validateToken](dart.as(value, core.String));
      if (!(typeof value == 'string'))
        return false;
      let s = this.readClasses();
      let result = s.remove(value);
      this.writeClasses(s);
      return result;
    }
    addAll(iterable) {
      this.modify(dart.fn((s => dart.dsend(s, 'addAll', iterable[dartx.map](dart.bind(this, _validateToken)))).bind(this)));
    }
    removeAll(iterable) {
      this.modify(dart.fn((s => dart.dsend(s, 'removeAll', iterable[dartx.map](dart.as(dart.bind(this, _validateToken), __CastType0)))).bind(this)));
    }
    toggleAll(iterable, shouldAdd) {
      if (shouldAdd === void 0)
        shouldAdd = null;
      iterable[dartx.forEach](dart.fn((e => this.toggle(dart.as(e, core.String), shouldAdd)).bind(this), core.bool, [dart.dynamic]));
    }
    retainAll(iterable) {
      this.modify(dart.fn(s => dart.dsend(s, 'retainAll', iterable)));
    }
    removeWhere(test) {
      this.modify(dart.fn(s => dart.dsend(s, 'removeWhere', test)));
    }
    retainWhere(test) {
      this.modify(dart.fn(s => dart.dsend(s, 'retainWhere', test)));
    }
    containsAll(collection) {
      return this.readClasses().containsAll(collection);
    }
    intersection(other) {
      return this.readClasses().intersection(other);
    }
    union(other) {
      return this.readClasses().union(other);
    }
    difference(other) {
      return this.readClasses().difference(other);
    }
    get first() {
      return this.readClasses().first;
    }
    get last() {
      return this.readClasses().last;
    }
    get single() {
      return this.readClasses().single;
    }
    toList(opts) {
      let growable = opts && 'growable' in opts ? opts.growable : true;
      return this.readClasses().toList({growable: growable});
    }
    toSet() {
      return this.readClasses().toSet();
    }
    take(n) {
      return this.readClasses().take(n);
    }
    takeWhile(test) {
      return this.readClasses().takeWhile(test);
    }
    skip(n) {
      return this.readClasses().skip(n);
    }
    skipWhile(test) {
      return this.readClasses().skipWhile(test);
    }
    firstWhere(test, opts) {
      let orElse = opts && 'orElse' in opts ? opts.orElse : null;
      return this.readClasses().firstWhere(test, {orElse: orElse});
    }
    lastWhere(test, opts) {
      let orElse = opts && 'orElse' in opts ? opts.orElse : null;
      return this.readClasses().lastWhere(test, {orElse: orElse});
    }
    singleWhere(test) {
      return this.readClasses().singleWhere(test);
    }
    elementAt(index) {
      return this.readClasses().elementAt(index);
    }
    clear() {
      this.modify(dart.fn(s => dart.dsend(s, 'clear')));
    }
    modify(f) {
      let s = this.readClasses();
      let ret = f(s);
      this.writeClasses(s);
      return ret;
    }
  }
  CssClassSetImpl[dart.implements] = () => [html.CssClassSet];
  dart.setSignature(CssClassSetImpl, {
    methods: () => ({
      [_validateToken]: [core.String, [core.String]],
      toggle: [core.bool, [core.String], [core.bool]],
      forEach: [dart.void, [dart.functionType(dart.void, [core.String])]],
      join: [core.String, [], [core.String]],
      map: [core.Iterable, [dart.functionType(dart.dynamic, [core.String])]],
      where: [core.Iterable$(core.String), [dart.functionType(core.bool, [core.String])]],
      expand: [core.Iterable, [dart.functionType(core.Iterable, [core.String])]],
      every: [core.bool, [dart.functionType(core.bool, [core.String])]],
      any: [core.bool, [dart.functionType(core.bool, [core.String])]],
      reduce: [core.String, [dart.functionType(core.String, [core.String, core.String])]],
      fold: [dart.dynamic, [dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic, core.String])]],
      contains: [core.bool, [core.Object]],
      lookup: [core.String, [core.Object]],
      add: [core.bool, [core.String]],
      remove: [core.bool, [core.Object]],
      addAll: [dart.void, [core.Iterable$(core.String)]],
      removeAll: [dart.void, [core.Iterable$(core.Object)]],
      toggleAll: [dart.void, [core.Iterable$(core.String)], [core.bool]],
      retainAll: [dart.void, [core.Iterable$(core.Object)]],
      removeWhere: [dart.void, [dart.functionType(core.bool, [core.String])]],
      retainWhere: [dart.void, [dart.functionType(core.bool, [core.String])]],
      containsAll: [core.bool, [core.Iterable$(core.Object)]],
      intersection: [core.Set$(core.String), [core.Set$(core.Object)]],
      union: [core.Set$(core.String), [core.Set$(core.String)]],
      difference: [core.Set$(core.String), [core.Set$(core.String)]],
      toList: [core.List$(core.String), [], {growable: core.bool}],
      toSet: [core.Set$(core.String), []],
      take: [core.Iterable$(core.String), [core.int]],
      takeWhile: [core.Iterable$(core.String), [dart.functionType(core.bool, [core.String])]],
      skip: [core.Iterable$(core.String), [core.int]],
      skipWhile: [core.Iterable$(core.String), [dart.functionType(core.bool, [core.String])]],
      firstWhere: [core.String, [dart.functionType(core.bool, [core.String])], {orElse: dart.functionType(core.String, [])}],
      lastWhere: [core.String, [dart.functionType(core.bool, [core.String])], {orElse: dart.functionType(core.String, [])}],
      singleWhere: [core.String, [dart.functionType(core.bool, [core.String])]],
      elementAt: [core.String, [core.int]],
      clear: [dart.void, []],
      modify: [dart.dynamic, [dart.functionType(dart.dynamic, [core.Set$(core.String)])]]
    })
  });
  dart.defineExtensionMembers(CssClassSetImpl, [
    'forEach',
    'join',
    'map',
    'where',
    'expand',
    'every',
    'any',
    'reduce',
    'fold',
    'contains',
    'toList',
    'toSet',
    'take',
    'takeWhile',
    'skip',
    'skipWhile',
    'firstWhere',
    'lastWhere',
    'singleWhere',
    'elementAt',
    'iterator',
    'isEmpty',
    'isNotEmpty',
    'length',
    'first',
    'last',
    'single'
  ]);
  dart.defineLazyProperties(CssClassSetImpl, {
    get _validTokenRE() {
      return core.RegExp.new('^\\S+$');
    }
  });
  const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(dart.dynamic, [core.Object]));
  function convertDartToNative_SerializedScriptValue(value) {
    return convertDartToNative_PrepareForStructuredClone(value);
  }
  dart.fn(convertDartToNative_SerializedScriptValue);
  function convertNativeToDart_SerializedScriptValue(object) {
    return convertNativeToDart_AcceptStructuredClone(object, {mustCopy: true});
  }
  dart.fn(convertNativeToDart_SerializedScriptValue);
  class _StructuredClone extends core.Object {
    _StructuredClone() {
      this.values = [];
      this.copies = [];
    }
    findSlot(value) {
      let length = this.values[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        if (dart.notNull(core.identical(this.values[dartx.get](i), value)))
          return i;
      }
      this.values[dartx.add](value);
      this.copies[dartx.add](null);
      return length;
    }
    readSlot(i) {
      return this.copies[dartx.get](i);
    }
    writeSlot(i, x) {
      this.copies[dartx.set](i, x);
    }
    cleanupSlots() {}
    walk(e) {
      if (e == null)
        return e;
      if (typeof e == 'boolean')
        return e;
      if (typeof e == 'number')
        return e;
      if (typeof e == 'string')
        return e;
      if (dart.is(e, core.DateTime)) {
        return convertDartToNative_DateTime(dart.as(e, core.DateTime));
      }
      if (dart.is(e, core.RegExp)) {
        dart.throw(new core.UnimplementedError('structured clone of RegExp'));
      }
      if (dart.is(e, dart.dynamic))
        return e;
      if (dart.is(e, dart.dynamic))
        return e;
      if (dart.is(e, dart.dynamic))
        return e;
      if (dart.is(e, html.ImageData))
        return e;
      if (dart.notNull(this.cloneNotRequired(e)))
        return e;
      if (dart.is(e, core.Map)) {
        let slot = this.findSlot(e);
        let copy = this.readSlot(slot);
        if (copy != null)
          return copy;
        copy = this.newJsMap();
        this.writeSlot(slot, copy);
        dart.dsend(e, 'forEach', dart.fn(((key, value) => {
          this.putIntoMap(copy, key, this.walk(value));
        }).bind(this)));
        return copy;
      }
      if (dart.is(e, core.List)) {
        let slot = this.findSlot(e);
        let copy = this.readSlot(slot);
        if (copy != null)
          return copy;
        copy = this.copyList(dart.as(e, core.List), slot);
        return copy;
      }
      dart.throw(new core.UnimplementedError('structured clone of other type'));
    }
    copyList(e, slot) {
      let i = 0;
      let length = e[dartx.length];
      let copy = this.newJsList(length);
      this.writeSlot(slot, copy);
      for (; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        dart.dsetindex(copy, i, this.walk(e[dartx.get](i)));
      }
      return copy;
    }
    convertDartToNative_PrepareForStructuredClone(value) {
      let copy = this.walk(value);
      this.cleanupSlots();
      return copy;
    }
  }
  dart.setSignature(_StructuredClone, {
    methods: () => ({
      findSlot: [core.int, [dart.dynamic]],
      readSlot: [dart.dynamic, [core.int]],
      writeSlot: [dart.dynamic, [core.int, dart.dynamic]],
      cleanupSlots: [dart.dynamic, []],
      walk: [dart.dynamic, [dart.dynamic]],
      copyList: [dart.dynamic, [core.List, core.int]],
      convertDartToNative_PrepareForStructuredClone: [dart.dynamic, [dart.dynamic]]
    })
  });
  class _AcceptStructuredClone extends core.Object {
    _AcceptStructuredClone() {
      this.values = [];
      this.copies = [];
      this.mustCopy = false;
    }
    findSlot(value) {
      let length = this.values[dartx.length];
      for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
        if (dart.notNull(this.identicalInJs(this.values[dartx.get](i), value)))
          return i;
      }
      this.values[dartx.add](value);
      this.copies[dartx.add](null);
      return length;
    }
    readSlot(i) {
      return this.copies[dartx.get](i);
    }
    writeSlot(i, x) {
      this.copies[dartx.set](i, x);
    }
    walk(e) {
      if (e == null)
        return e;
      if (typeof e == 'boolean')
        return e;
      if (typeof e == 'number')
        return e;
      if (typeof e == 'string')
        return e;
      if (dart.notNull(isJavaScriptDate(e))) {
        return convertNativeToDart_DateTime(e);
      }
      if (dart.notNull(isJavaScriptRegExp(e))) {
        dart.throw(new core.UnimplementedError('structured clone of RegExp'));
      }
      if (dart.notNull(isJavaScriptPromise(e))) {
        return convertNativePromiseToDartFuture(e);
      }
      if (dart.notNull(isJavaScriptSimpleObject(e))) {
        let slot = this.findSlot(e);
        let copy = this.readSlot(slot);
        if (copy != null)
          return copy;
        copy = dart.map();
        this.writeSlot(slot, copy);
        this.forEachJsField(e, dart.fn(((key, value) => dart.dsetindex(copy, key, this.walk(value))).bind(this)));
        return copy;
      }
      if (dart.notNull(isJavaScriptArray(e))) {
        let slot = this.findSlot(e);
        let copy = this.readSlot(slot);
        if (copy != null)
          return copy;
        let length = dart.as(dart.dload(e, 'length'), core.int);
        copy = dart.notNull(this.mustCopy) ? this.newDartList(length) : e;
        this.writeSlot(slot, copy);
        for (let i = 0; dart.notNull(i) < dart.notNull(length); i = dart.notNull(i) + 1) {
          dart.dsetindex(copy, i, this.walk(dart.dindex(e, i)));
        }
        return copy;
      }
      return e;
    }
    convertNativeToDart_AcceptStructuredClone(object, opts) {
      let mustCopy = opts && 'mustCopy' in opts ? opts.mustCopy : false;
      this.mustCopy = dart.as(mustCopy, core.bool);
      let copy = this.walk(object);
      return copy;
    }
  }
  dart.setSignature(_AcceptStructuredClone, {
    methods: () => ({
      findSlot: [core.int, [dart.dynamic]],
      readSlot: [dart.dynamic, [core.int]],
      writeSlot: [dart.dynamic, [core.int, dart.dynamic]],
      walk: [dart.dynamic, [dart.dynamic]],
      convertNativeToDart_AcceptStructuredClone: [dart.dynamic, [dart.dynamic], {mustCopy: dart.dynamic}]
    })
  });
  class _TypedContextAttributes extends core.Object {
    _TypedContextAttributes(alpha, antialias, depth, failIfMajorPerformanceCaveat, premultipliedAlpha, preserveDrawingBuffer, stencil) {
      this.alpha = alpha;
      this.antialias = antialias;
      this.depth = depth;
      this.failIfMajorPerformanceCaveat = failIfMajorPerformanceCaveat;
      this.premultipliedAlpha = premultipliedAlpha;
      this.preserveDrawingBuffer = preserveDrawingBuffer;
      this.stencil = stencil;
    }
  }
  dart.setSignature(_TypedContextAttributes, {
    constructors: () => ({_TypedContextAttributes: [_TypedContextAttributes, [core.bool, core.bool, core.bool, core.bool, core.bool, core.bool, core.bool]]})
  });
  function convertNativeToDart_ContextAttributes(nativeContextAttributes) {
    if (dart.is(nativeContextAttributes, dart.dynamic)) {
      return nativeContextAttributes;
    }
    return new _TypedContextAttributes(dart.as(nativeContextAttributes.alpha, core.bool), dart.as(nativeContextAttributes.antialias, core.bool), dart.as(nativeContextAttributes.depth, core.bool), dart.as(nativeContextAttributes.failIfMajorPerformanceCaveat, core.bool), dart.as(nativeContextAttributes.premultipliedAlpha, core.bool), dart.as(nativeContextAttributes.preserveDrawingBuffer, core.bool), dart.as(nativeContextAttributes.stencil, core.bool));
  }
  dart.fn(convertNativeToDart_ContextAttributes);
  class _TypedImageData extends core.Object {
    _TypedImageData(data, height, width) {
      this.data = data;
      this.height = height;
      this.width = width;
    }
  }
  _TypedImageData[dart.implements] = () => [html.ImageData];
  dart.setSignature(_TypedImageData, {
    constructors: () => ({_TypedImageData: [_TypedImageData, [typed_data.Uint8ClampedList, core.int, core.int]]})
  });
  function convertNativeToDart_ImageData(nativeImageData) {
    0;
    if (dart.is(nativeImageData, html.ImageData)) {
      let data = dart.dload(nativeImageData, 'data');
      if (data.constructor === Array) {
        if (typeof CanvasPixelArray !== "undefined") {
          data.constructor = CanvasPixelArray;
          data.BYTES_PER_ELEMENT = 1;
        }
      }
      return dart.as(nativeImageData, html.ImageData);
    }
    return new _TypedImageData(dart.as(nativeImageData.data, typed_data.Uint8ClampedList), dart.as(nativeImageData.height, core.int), dart.as(nativeImageData.width, core.int));
  }
  dart.fn(convertNativeToDart_ImageData, html.ImageData, [dart.dynamic]);
  function convertDartToNative_ImageData(imageData) {
    if (dart.is(imageData, _TypedImageData)) {
      return {data: imageData.data, height: imageData.height, width: imageData.width};
    }
    return imageData;
  }
  dart.fn(convertDartToNative_ImageData, dart.dynamic, [html.ImageData]);
  const _serializedScriptValue = 'num|String|bool|' + 'JSExtendableArray|=Object|' + 'Blob|File|NativeByteBuffer|NativeTypedData';
  const annotation_Creates_SerializedScriptValue = dart.const(new _js_helper.Creates(_serializedScriptValue));
  const annotation_Returns_SerializedScriptValue = dart.const(new _js_helper.Returns(_serializedScriptValue));
  function convertNativeToDart_Dictionary(object) {
    if (object == null)
      return null;
    let dict = dart.map();
    let keys = Object.getOwnPropertyNames(object);
    for (let key of dart.as(keys, core.Iterable)) {
      dict.set(key, object[key]);
    }
    return dict;
  }
  dart.fn(convertNativeToDart_Dictionary, core.Map, [dart.dynamic]);
  function convertDartToNative_Dictionary(dict, postCreate) {
    if (postCreate === void 0)
      postCreate = null;
    if (dict == null)
      return null;
    let object = {};
    if (postCreate != null) {
      dart.dcall(postCreate, object);
    }
    dict.forEach(dart.fn((key, value) => {
      object[key] = value;
    }, dart.dynamic, [core.String, dart.dynamic]));
    return object;
  }
  dart.fn(convertDartToNative_Dictionary, dart.dynamic, [core.Map], [dart.functionType(dart.void, [dart.dynamic])]);
  function convertDartToNative_StringArray(input) {
    return input;
  }
  dart.fn(convertDartToNative_StringArray, core.List, [core.List$(core.String)]);
  function convertNativeToDart_DateTime(date) {
    let millisSinceEpoch = date.getTime();
    return new core.DateTime.fromMillisecondsSinceEpoch(millisSinceEpoch, {isUtc: true});
  }
  dart.fn(convertNativeToDart_DateTime, core.DateTime, [dart.dynamic]);
  function convertDartToNative_DateTime(date) {
    return new Date(date.millisecondsSinceEpoch);
  }
  dart.fn(convertDartToNative_DateTime, dart.dynamic, [core.DateTime]);
  function convertDartToNative_PrepareForStructuredClone(value) {
    return new _StructuredCloneDart2Js().convertDartToNative_PrepareForStructuredClone(value);
  }
  dart.fn(convertDartToNative_PrepareForStructuredClone);
  function convertNativeToDart_AcceptStructuredClone(object, opts) {
    let mustCopy = opts && 'mustCopy' in opts ? opts.mustCopy : false;
    return new _AcceptStructuredCloneDart2Js().convertNativeToDart_AcceptStructuredClone(object, {mustCopy: mustCopy});
  }
  dart.fn(convertNativeToDart_AcceptStructuredClone, dart.dynamic, [dart.dynamic], {mustCopy: dart.dynamic});
  class _StructuredCloneDart2Js extends _StructuredClone {
    _StructuredCloneDart2Js() {
      super._StructuredClone();
    }
    newJsMap() {
      return {};
    }
    putIntoMap(map, key, value) {
      return map[key] = value;
    }
    newJsList(length) {
      return new Array(length);
    }
    cloneNotRequired(e) {
      return dart.is(e, _native_typed_data.NativeByteBuffer) || dart.is(e, _native_typed_data.NativeTypedData);
    }
  }
  dart.setSignature(_StructuredCloneDart2Js, {
    methods: () => ({
      newJsMap: [dart.dynamic, []],
      putIntoMap: [dart.void, [dart.dynamic, dart.dynamic, dart.dynamic]],
      newJsList: [dart.dynamic, [dart.dynamic]],
      cloneNotRequired: [core.bool, [dart.dynamic]]
    })
  });
  class _AcceptStructuredCloneDart2Js extends _AcceptStructuredClone {
    _AcceptStructuredCloneDart2Js() {
      super._AcceptStructuredClone();
    }
    newJsList(length) {
      return new Array(length);
    }
    newDartList(length) {
      return this.newJsList(length);
    }
    identicalInJs(a, b) {
      return core.identical(a, b);
    }
    forEachJsField(object, action) {
      for (let key of dart.as(Object.keys(object), core.Iterable)) {
        dart.dcall(action, key, object[key]);
      }
    }
  }
  dart.setSignature(_AcceptStructuredCloneDart2Js, {
    methods: () => ({
      newJsList: [dart.dynamic, [dart.dynamic]],
      newDartList: [dart.dynamic, [dart.dynamic]],
      identicalInJs: [core.bool, [dart.dynamic, dart.dynamic]],
      forEachJsField: [dart.void, [dart.dynamic, dart.dynamic]]
    })
  });
  function isJavaScriptDate(value) {
    return value instanceof Date;
  }
  dart.fn(isJavaScriptDate, core.bool, [dart.dynamic]);
  function isJavaScriptRegExp(value) {
    return value instanceof RegExp;
  }
  dart.fn(isJavaScriptRegExp, core.bool, [dart.dynamic]);
  function isJavaScriptArray(value) {
    return value instanceof Array;
  }
  dart.fn(isJavaScriptArray, core.bool, [dart.dynamic]);
  function isJavaScriptSimpleObject(value) {
    let proto = Object.getPrototypeOf(value);
    return proto === Object.prototype || proto === null;
  }
  dart.fn(isJavaScriptSimpleObject, core.bool, [dart.dynamic]);
  function isImmutableJavaScriptArray(value) {
    return !!value.immutable$list;
  }
  dart.fn(isImmutableJavaScriptArray, core.bool, [dart.dynamic]);
  function isJavaScriptPromise(value) {
    return typeof Promise != "undefined" && value instanceof Promise;
  }
  dart.fn(isJavaScriptPromise, core.bool, [dart.dynamic]);
  function convertNativePromiseToDartFuture(promise) {
    let completer = async.Completer.new();
    let then = dart.dcall(/* Unimplemented unknown name */convertDartClosureToJS, dart.fn(result => completer.complete(result), dart.void, [dart.dynamic]), 1);
    let error = dart.dcall(/* Unimplemented unknown name */convertDartClosureToJS, dart.fn(result => completer.completeError(result), dart.void, [dart.dynamic]), 1);
    let newPromise = promise.then(then).catch(error);
    return completer.future;
  }
  dart.fn(convertNativePromiseToDartFuture, async.Future, [dart.dynamic]);
  class Device extends core.Object {
    static get userAgent() {
      return html.window.navigator.userAgent;
    }
    static get isOpera() {
      if (Device._isOpera == null) {
        Device._isOpera = Device.userAgent[dartx.contains]("Opera", 0);
      }
      return Device._isOpera;
    }
    static get isIE() {
      if (Device._isIE == null) {
        Device._isIE = !dart.notNull(Device.isOpera) && dart.notNull(Device.userAgent[dartx.contains]("Trident/", 0));
      }
      return Device._isIE;
    }
    static get isFirefox() {
      if (Device._isFirefox == null) {
        Device._isFirefox = Device.userAgent[dartx.contains]("Firefox", 0);
      }
      return Device._isFirefox;
    }
    static get isWebKit() {
      if (Device._isWebKit == null) {
        Device._isWebKit = !dart.notNull(Device.isOpera) && dart.notNull(Device.userAgent[dartx.contains]("WebKit", 0));
      }
      return Device._isWebKit;
    }
    static get cssPrefix() {
      let prefix = Device._cachedCssPrefix;
      if (prefix != null)
        return prefix;
      if (dart.notNull(Device.isFirefox)) {
        prefix = '-moz-';
      } else if (dart.notNull(Device.isIE)) {
        prefix = '-ms-';
      } else if (dart.notNull(Device.isOpera)) {
        prefix = '-o-';
      } else {
        prefix = '-webkit-';
      }
      return Device._cachedCssPrefix = prefix;
    }
    static get propertyPrefix() {
      let prefix = Device._cachedPropertyPrefix;
      if (prefix != null)
        return prefix;
      if (dart.notNull(Device.isFirefox)) {
        prefix = 'moz';
      } else if (dart.notNull(Device.isIE)) {
        prefix = 'ms';
      } else if (dart.notNull(Device.isOpera)) {
        prefix = 'o';
      } else {
        prefix = 'webkit';
      }
      return Device._cachedPropertyPrefix = prefix;
    }
    static isEventTypeSupported(eventType) {
      try {
        let e = html.Event.eventType(eventType, '');
        return dart.is(e, html.Event);
      } catch (_) {
      }

      return false;
    }
  }
  dart.setSignature(Device, {
    statics: () => ({isEventTypeSupported: [core.bool, [core.String]]}),
    names: ['isEventTypeSupported']
  });
  Device._isOpera = null;
  Device._isIE = null;
  Device._isFirefox = null;
  Device._isWebKit = null;
  Device._cachedCssPrefix = null;
  Device._cachedPropertyPrefix = null;
  const _childNodes = Symbol('_childNodes');
  const _node = Symbol('_node');
  const _iterable = Symbol('_iterable');
  const _filtered = Symbol('_filtered');
  class FilteredElementList extends collection.ListBase$(html.Element) {
    FilteredElementList(node) {
      this[_childNodes] = node.nodes;
      this[_node] = node;
    }
    get [_iterable]() {
      return new _internal.WhereIterable(this[_childNodes], dart.fn(n => dart.is(n, html.Element), core.bool, [dart.dynamic]));
    }
    get [_filtered]() {
      return core.List$(html.Element).from(this[_iterable], {growable: false});
    }
    forEach(f) {
      this[_filtered][dartx.forEach](f);
    }
    set(index, value) {
      this.get(index).replaceWith(value);
      return value;
    }
    set length(newLength) {
      let len = this.length;
      if (dart.notNull(newLength) >= dart.notNull(len)) {
        return;
      } else if (dart.notNull(newLength) < 0) {
        dart.throw(new core.ArgumentError("Invalid list length"));
      }
      this.removeRange(newLength, len);
    }
    add(value) {
      this[_childNodes][dartx.add](value);
    }
    addAll(iterable) {
      for (let element of iterable) {
        this.add(element);
      }
    }
    contains(needle) {
      if (!dart.is(needle, html.Element))
        return false;
      let element = dart.as(needle, html.Element);
      return dart.equals(element.parentNode, this[_node]);
    }
    get reversed() {
      return this[_filtered][dartx.reversed];
    }
    sort(compare) {
      if (compare === void 0)
        compare = null;
      dart.throw(new core.UnsupportedError('Cannot sort filtered list'));
    }
    setRange(start, end, iterable, skipCount) {
      if (skipCount === void 0)
        skipCount = 0;
      dart.throw(new core.UnsupportedError('Cannot setRange on filtered list'));
    }
    fillRange(start, end, fillValue) {
      if (fillValue === void 0)
        fillValue = null;
      dart.throw(new core.UnsupportedError('Cannot fillRange on filtered list'));
    }
    replaceRange(start, end, iterable) {
      dart.throw(new core.UnsupportedError('Cannot replaceRange on filtered list'));
    }
    removeRange(start, end) {
      core.List.from(this[_iterable][dartx.skip](start)[dartx.take](dart.notNull(end) - dart.notNull(start)))[dartx.forEach](dart.fn(el => dart.dsend(el, 'remove')));
    }
    clear() {
      this[_childNodes][dartx.clear]();
    }
    removeLast() {
      let result = this[_iterable][dartx.last];
      if (result != null) {
        result.remove();
      }
      return result;
    }
    insert(index, value) {
      if (index == this.length) {
        this.add(value);
      } else {
        let element = this[_iterable][dartx.elementAt](index);
        element.parentNode.insertBefore(value, element);
      }
    }
    insertAll(index, iterable) {
      if (index == this.length) {
        this.addAll(iterable);
      } else {
        let element = this[_iterable][dartx.elementAt](index);
        element.parentNode.insertAllBefore(iterable, element);
      }
    }
    removeAt(index) {
      let result = this.get(index);
      result.remove();
      return result;
    }
    remove(element) {
      if (!dart.is(element, html.Element))
        return false;
      if (dart.notNull(this.contains(element))) {
        dart.as(element, html.Element).remove();
        return true;
      } else {
        return false;
      }
    }
    get length() {
      return this[_iterable][dartx.length];
    }
    get(index) {
      return this[_iterable][dartx.elementAt](index);
    }
    get iterator() {
      return this[_filtered][dartx.iterator];
    }
    get rawList() {
      return this[_node].childNodes;
    }
  }
  FilteredElementList[dart.implements] = () => [NodeListWrapper];
  dart.setSignature(FilteredElementList, {
    constructors: () => ({FilteredElementList: [FilteredElementList, [html.Node]]}),
    methods: () => ({
      forEach: [dart.void, [dart.functionType(dart.void, [html.Element])]],
      set: [dart.void, [core.int, html.Element]],
      add: [dart.void, [html.Element]],
      addAll: [dart.void, [core.Iterable$(html.Element)]],
      sort: [dart.void, [], [dart.functionType(core.int, [html.Element, html.Element])]],
      setRange: [dart.void, [core.int, core.int, core.Iterable$(html.Element)], [core.int]],
      fillRange: [dart.void, [core.int, core.int], [html.Element]],
      replaceRange: [dart.void, [core.int, core.int, core.Iterable$(html.Element)]],
      removeLast: [html.Element, []],
      insert: [dart.void, [core.int, html.Element]],
      insertAll: [dart.void, [core.int, core.Iterable$(html.Element)]],
      removeAt: [html.Element, [core.int]],
      get: [html.Element, [core.int]]
    })
  });
  dart.defineExtensionMembers(FilteredElementList, [
    'forEach',
    'set',
    'add',
    'addAll',
    'contains',
    'sort',
    'setRange',
    'fillRange',
    'replaceRange',
    'removeRange',
    'clear',
    'removeLast',
    'insert',
    'insertAll',
    'removeAt',
    'remove',
    'get',
    'length',
    'reversed',
    'length',
    'iterator'
  ]);
  class Lists extends core.Object {
    static indexOf(a, element, startIndex, endIndex) {
      if (dart.notNull(startIndex) >= dart.notNull(a[dartx.length])) {
        return -1;
      }
      if (dart.notNull(startIndex) < 0) {
        startIndex = 0;
      }
      for (let i = startIndex; dart.notNull(i) < dart.notNull(endIndex); i = dart.notNull(i) + 1) {
        if (dart.equals(a[dartx.get](i), element)) {
          return i;
        }
      }
      return -1;
    }
    static lastIndexOf(a, element, startIndex) {
      if (dart.notNull(startIndex) < 0) {
        return -1;
      }
      if (dart.notNull(startIndex) >= dart.notNull(a[dartx.length])) {
        startIndex = dart.notNull(a[dartx.length]) - 1;
      }
      for (let i = startIndex; dart.notNull(i) >= 0; i = dart.notNull(i) - 1) {
        if (dart.equals(a[dartx.get](i), element)) {
          return i;
        }
      }
      return -1;
    }
    static getRange(a, start, end, accumulator) {
      if (dart.notNull(start) < 0)
        dart.throw(new core.RangeError.value(start));
      if (dart.notNull(end) < dart.notNull(start))
        dart.throw(new core.RangeError.value(end));
      if (dart.notNull(end) > dart.notNull(a[dartx.length]))
        dart.throw(new core.RangeError.value(end));
      for (let i = start; dart.notNull(i) < dart.notNull(end); i = dart.notNull(i) + 1) {
        accumulator[dartx.add](a[dartx.get](i));
      }
      return accumulator;
    }
  }
  dart.setSignature(Lists, {
    statics: () => ({
      indexOf: [core.int, [core.List, core.Object, core.int, core.int]],
      lastIndexOf: [core.int, [core.List, core.Object, core.int]],
      getRange: [core.List, [core.List, core.int, core.int, core.List]]
    }),
    names: ['indexOf', 'lastIndexOf', 'getRange']
  });
  class NodeListWrapper extends core.Object {}
  // Exports:
  exports.CssClassSetImpl = CssClassSetImpl;
  exports.convertDartToNative_SerializedScriptValue = convertDartToNative_SerializedScriptValue;
  exports.convertNativeToDart_SerializedScriptValue = convertNativeToDart_SerializedScriptValue;
  exports.convertNativeToDart_ContextAttributes = convertNativeToDart_ContextAttributes;
  exports.convertNativeToDart_ImageData = convertNativeToDart_ImageData;
  exports.convertDartToNative_ImageData = convertDartToNative_ImageData;
  exports.annotation_Creates_SerializedScriptValue = annotation_Creates_SerializedScriptValue;
  exports.annotation_Returns_SerializedScriptValue = annotation_Returns_SerializedScriptValue;
  exports.convertNativeToDart_Dictionary = convertNativeToDart_Dictionary;
  exports.convertDartToNative_Dictionary = convertDartToNative_Dictionary;
  exports.convertDartToNative_StringArray = convertDartToNative_StringArray;
  exports.convertNativeToDart_DateTime = convertNativeToDart_DateTime;
  exports.convertDartToNative_DateTime = convertDartToNative_DateTime;
  exports.convertDartToNative_PrepareForStructuredClone = convertDartToNative_PrepareForStructuredClone;
  exports.convertNativeToDart_AcceptStructuredClone = convertNativeToDart_AcceptStructuredClone;
  exports.isJavaScriptDate = isJavaScriptDate;
  exports.isJavaScriptRegExp = isJavaScriptRegExp;
  exports.isJavaScriptArray = isJavaScriptArray;
  exports.isJavaScriptSimpleObject = isJavaScriptSimpleObject;
  exports.isImmutableJavaScriptArray = isImmutableJavaScriptArray;
  exports.isJavaScriptPromise = isJavaScriptPromise;
  exports.convertNativePromiseToDartFuture = convertNativePromiseToDartFuture;
  exports.Device = Device;
  exports.FilteredElementList = FilteredElementList;
  exports.Lists = Lists;
  exports.NodeListWrapper = NodeListWrapper;
});
dart_library.library('dart/isolate', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/async'
], /* Lazy imports */[
  'dart/_isolate_helper'
], function(exports, dart, core, async, _isolate_helper) {
  'use strict';
  let dartx = dart.dartx;
  class Capability extends core.Object {
    static new() {
      return new _isolate_helper.CapabilityImpl();
    }
  }
  dart.setSignature(Capability, {
    constructors: () => ({new: [Capability, []]})
  });
  class IsolateSpawnException extends core.Object {
    IsolateSpawnException(message) {
      this.message = message;
    }
    toString() {
      return `IsolateSpawnException: ${this.message}`;
    }
  }
  IsolateSpawnException[dart.implements] = () => [core.Exception];
  dart.setSignature(IsolateSpawnException, {
    constructors: () => ({IsolateSpawnException: [IsolateSpawnException, [core.String]]})
  });
  const _pause = Symbol('_pause');
  class Isolate extends core.Object {
    Isolate(controlPort, opts) {
      let pauseCapability = opts && 'pauseCapability' in opts ? opts.pauseCapability : null;
      let terminateCapability = opts && 'terminateCapability' in opts ? opts.terminateCapability : null;
      this.controlPort = controlPort;
      this.pauseCapability = pauseCapability;
      this.terminateCapability = terminateCapability;
    }
    static get current() {
      return Isolate._currentIsolateCache;
    }
    static spawn(entryPoint, message, opts) {
      let paused = opts && 'paused' in opts ? opts.paused : false;
      try {
        return _isolate_helper.IsolateNatives.spawnFunction(entryPoint, message, paused).then(dart.fn(msg => new Isolate(dart.as(dart.dindex(msg, 1), SendPort), {pauseCapability: dart.as(dart.dindex(msg, 2), Capability), terminateCapability: dart.as(dart.dindex(msg, 3), Capability)}), Isolate, [dart.dynamic]));
      } catch (e) {
        let st = dart.stackTrace(e);
        return async.Future$(Isolate).error(e, st);
      }

    }
    static spawnUri(uri, args, message, opts) {
      let paused = opts && 'paused' in opts ? opts.paused : false;
      let packageRoot = opts && 'packageRoot' in opts ? opts.packageRoot : null;
      if (packageRoot != null)
        dart.throw(new core.UnimplementedError("packageRoot"));
      try {
        if (dart.is(args, core.List)) {
          for (let i = 0; dart.notNull(i) < dart.notNull(args[dartx.length]); i = dart.notNull(i) + 1) {
            if (!(typeof args[dartx.get](i) == 'string')) {
              dart.throw(new core.ArgumentError(`Args must be a list of Strings ${args}`));
            }
          }
        } else if (args != null) {
          dart.throw(new core.ArgumentError(`Args must be a list of Strings ${args}`));
        }
        return _isolate_helper.IsolateNatives.spawnUri(uri, args, message, paused).then(dart.fn(msg => new Isolate(dart.as(dart.dindex(msg, 1), SendPort), {pauseCapability: dart.as(dart.dindex(msg, 2), Capability), terminateCapability: dart.as(dart.dindex(msg, 3), Capability)}), Isolate, [dart.dynamic]));
      } catch (e) {
        let st = dart.stackTrace(e);
        return async.Future$(Isolate).error(e, st);
      }

    }
    pause(resumeCapability) {
      if (resumeCapability === void 0)
        resumeCapability = null;
      if (resumeCapability == null)
        resumeCapability = Capability.new();
      this[_pause](resumeCapability);
      return resumeCapability;
    }
    [_pause](resumeCapability) {
      let message = core.List.new(3);
      message[dartx.set](0, "pause");
      message[dartx.set](1, this.pauseCapability);
      message[dartx.set](2, resumeCapability);
      this.controlPort.send(message);
    }
    resume(resumeCapability) {
      let message = core.List.new(2);
      message[dartx.set](0, "resume");
      message[dartx.set](1, resumeCapability);
      this.controlPort.send(message);
    }
    addOnExitListener(responsePort) {
      let message = core.List.new(2);
      message[dartx.set](0, "add-ondone");
      message[dartx.set](1, responsePort);
      this.controlPort.send(message);
    }
    removeOnExitListener(responsePort) {
      let message = core.List.new(2);
      message[dartx.set](0, "remove-ondone");
      message[dartx.set](1, responsePort);
      this.controlPort.send(message);
    }
    setErrorsFatal(errorsAreFatal) {
      let message = core.List.new(3);
      message[dartx.set](0, "set-errors-fatal");
      message[dartx.set](1, this.terminateCapability);
      message[dartx.set](2, errorsAreFatal);
      this.controlPort.send(message);
    }
    kill(priority) {
      if (priority === void 0)
        priority = Isolate.BEFORE_NEXT_EVENT;
      this.controlPort.send(["kill", this.terminateCapability, priority]);
    }
    ping(responsePort, pingType) {
      if (pingType === void 0)
        pingType = Isolate.IMMEDIATE;
      let message = core.List.new(3);
      message[dartx.set](0, "ping");
      message[dartx.set](1, responsePort);
      message[dartx.set](2, pingType);
      this.controlPort.send(message);
    }
    addErrorListener(port) {
      let message = core.List.new(2);
      message[dartx.set](0, "getErrors");
      message[dartx.set](1, port);
      this.controlPort.send(message);
    }
    removeErrorListener(port) {
      let message = core.List.new(2);
      message[dartx.set](0, "stopErrors");
      message[dartx.set](1, port);
      this.controlPort.send(message);
    }
    get errors() {
      let controller = null;
      let port = null;
      function handleError(message) {
        let errorDescription = dart.as(dart.dindex(message, 0), core.String);
        let stackDescription = dart.as(dart.dindex(message, 1), core.String);
        let error = new RemoteError(errorDescription, stackDescription);
        controller.addError(error, error.stackTrace);
      }
      dart.fn(handleError, dart.void, [dart.dynamic]);
      controller = async.StreamController.broadcast({sync: true, onListen: dart.fn((() => {
          port = RawReceivePort.new(handleError);
          this.addErrorListener(port.sendPort);
        }).bind(this)), onCancel: dart.fn((() => {
          this.removeErrorListener(port.sendPort);
          port.close();
          port = null;
        }).bind(this))});
      return controller.stream;
    }
  }
  dart.setSignature(Isolate, {
    constructors: () => ({Isolate: [Isolate, [SendPort], {pauseCapability: Capability, terminateCapability: Capability}]}),
    methods: () => ({
      pause: [Capability, [], [Capability]],
      [_pause]: [dart.void, [Capability]],
      resume: [dart.void, [Capability]],
      addOnExitListener: [dart.void, [SendPort]],
      removeOnExitListener: [dart.void, [SendPort]],
      setErrorsFatal: [dart.void, [core.bool]],
      kill: [dart.void, [], [core.int]],
      ping: [dart.void, [SendPort], [core.int]],
      addErrorListener: [dart.void, [SendPort]],
      removeErrorListener: [dart.void, [SendPort]]
    }),
    statics: () => ({
      spawn: [async.Future$(Isolate), [dart.functionType(dart.void, [dart.dynamic]), dart.dynamic], {paused: core.bool}],
      spawnUri: [async.Future$(Isolate), [core.Uri, core.List$(core.String), dart.dynamic], {paused: core.bool, packageRoot: core.Uri}]
    }),
    names: ['spawn', 'spawnUri']
  });
  Isolate.IMMEDIATE = 0;
  Isolate.BEFORE_NEXT_EVENT = 1;
  Isolate.AS_EVENT = 2;
  dart.defineLazyProperties(Isolate, {
    get _currentIsolateCache() {
      return _isolate_helper.IsolateNatives.currentIsolate;
    }
  });
  class SendPort extends core.Object {}
  SendPort[dart.implements] = () => [Capability];
  class ReceivePort extends core.Object {
    static new() {
      return new _isolate_helper.ReceivePortImpl();
    }
    static fromRawReceivePort(rawPort) {
      return new _isolate_helper.ReceivePortImpl.fromRawReceivePort(rawPort);
    }
  }
  ReceivePort[dart.implements] = () => [async.Stream];
  dart.setSignature(ReceivePort, {
    constructors: () => ({
      new: [ReceivePort, []],
      fromRawReceivePort: [ReceivePort, [RawReceivePort]]
    })
  });
  class RawReceivePort extends core.Object {
    static new(handler) {
      if (handler === void 0)
        handler = null;
      return new _isolate_helper.RawReceivePortImpl(handler);
    }
  }
  dart.setSignature(RawReceivePort, {
    constructors: () => ({new: [RawReceivePort, [], [dart.functionType(dart.void, [dart.dynamic])]]})
  });
  class _IsolateUnhandledException extends core.Object {
    _IsolateUnhandledException(message, source, stackTrace) {
      this.message = message;
      this.source = source;
      this.stackTrace = stackTrace;
    }
    toString() {
      return 'IsolateUnhandledException: exception while handling message: ' + `${this.message} \n  ` + `${dart.toString(this.source)[dartx.replaceAll]("\n", "\n  ")}\n` + 'original stack trace:\n  ' + `${dart.toString(this.stackTrace)[dartx.replaceAll]("\n", "\n  ")}`;
    }
  }
  _IsolateUnhandledException[dart.implements] = () => [core.Exception];
  dart.setSignature(_IsolateUnhandledException, {
    constructors: () => ({_IsolateUnhandledException: [_IsolateUnhandledException, [dart.dynamic, dart.dynamic, core.StackTrace]]})
  });
  const _description = Symbol('_description');
  class RemoteError extends core.Object {
    RemoteError(description, stackDescription) {
      this[_description] = description;
      this.stackTrace = new _RemoteStackTrace(stackDescription);
    }
    toString() {
      return this[_description];
    }
  }
  RemoteError[dart.implements] = () => [core.Error];
  dart.setSignature(RemoteError, {
    constructors: () => ({RemoteError: [RemoteError, [core.String, core.String]]})
  });
  const _trace = Symbol('_trace');
  class _RemoteStackTrace extends core.Object {
    _RemoteStackTrace(trace) {
      this[_trace] = trace;
    }
    toString() {
      return this[_trace];
    }
  }
  _RemoteStackTrace[dart.implements] = () => [core.StackTrace];
  dart.setSignature(_RemoteStackTrace, {
    constructors: () => ({_RemoteStackTrace: [_RemoteStackTrace, [core.String]]})
  });
  // Exports:
  exports.Capability = Capability;
  exports.IsolateSpawnException = IsolateSpawnException;
  exports.Isolate = Isolate;
  exports.SendPort = SendPort;
  exports.ReceivePort = ReceivePort;
  exports.RawReceivePort = RawReceivePort;
  exports.RemoteError = RemoteError;
});
dart_library.library('dart/js', null, /* Imports */[
  "dart/_runtime",
  'dart/core',
  'dart/collection',
  'dart/_js_helper'
], /* Lazy imports */[
], function(exports, dart, core, collection, _js_helper) {
  'use strict';
  let dartx = dart.dartx;
  dart.defineLazyProperties(exports, {
    get _global() {
      return dart.global;
    },
    get context() {
      return _wrapToDart(exports._global);
    }
  });
  const _jsObject = Symbol('_jsObject');
  class JsObject extends core.Object {
    _fromJs(jsObject) {
      this[_jsObject] = jsObject;
      dart.assert(this[_jsObject] != null);
    }
    static new(constructor, arguments$) {
      if (arguments$ === void 0)
        arguments$ = null;
      let ctor = constructor[_jsObject];
      if (arguments$ == null) {
        return _wrapToDart(new ctor());
      }
      return _wrapToDart(new ctor(...arguments$));
    }
    static fromBrowserObject(object) {
      if (typeof object == 'number' || typeof object == 'string' || typeof object == 'boolean' || object == null) {
        dart.throw(new core.ArgumentError("object cannot be a num, string, bool, or null"));
      }
      return _wrapToDart(_convertToJS(object));
    }
    static jsify(object) {
      if (!dart.is(object, core.Map) && !dart.is(object, core.Iterable)) {
        dart.throw(new core.ArgumentError("object must be a Map or Iterable"));
      }
      return _wrapToDart(JsObject._convertDataTree(object));
    }
    static _convertDataTree(data) {
      let _convertedObjects = collection.HashMap.identity();
      function _convert(o) {
        if (dart.notNull(_convertedObjects.containsKey(o))) {
          return _convertedObjects.get(o);
        }
        if (dart.is(o, core.Map)) {
          let convertedMap = {};
          _convertedObjects.set(o, convertedMap);
          for (let key of dart.as(dart.dload(o, 'keys'), core.Iterable)) {
            convertedMap[key] = _convert(dart.dindex(o, key));
          }
          return convertedMap;
        } else if (dart.is(o, core.Iterable)) {
          let convertedList = [];
          _convertedObjects.set(o, convertedList);
          convertedList[dartx.addAll](dart.as(dart.dsend(o, 'map', _convert), core.Iterable));
          return convertedList;
        } else {
          return _convertToJS(o);
        }
      }
      dart.fn(_convert);
      return _convert(data);
    }
    get(property) {
      if (!(typeof property == 'string') && !(typeof property == 'number')) {
        dart.throw(new core.ArgumentError("property is not a String or num"));
      }
      return _convertToDart(this[_jsObject][property]);
    }
    set(property, value) {
      if (!(typeof property == 'string') && !(typeof property == 'number')) {
        dart.throw(new core.ArgumentError("property is not a String or num"));
      }
      this[_jsObject][property] = _convertToJS(value);
      return value;
    }
    get hashCode() {
      return 0;
    }
    ['=='](other) {
      return dart.is(other, JsObject) && this[_jsObject] === dart.dload(other, _jsObject);
    }
    hasProperty(property) {
      if (!(typeof property == 'string') && !(typeof property == 'number')) {
        dart.throw(new core.ArgumentError("property is not a String or num"));
      }
      return property in this[_jsObject];
    }
    deleteProperty(property) {
      if (!(typeof property == 'string') && !(typeof property == 'number')) {
        dart.throw(new core.ArgumentError("property is not a String or num"));
      }
      delete this[_jsObject][property];
    }
    instanceof(type) {
      return this[_jsObject] instanceof _convertToJS(type);
    }
    toString() {
      try {
        return String(this[_jsObject]);
      } catch (e) {
        return super.toString();
      }

    }
    callMethod(method, args) {
      if (args === void 0)
        args = null;
      if (!(typeof method == 'string') && !(typeof method == 'number')) {
        dart.throw(new core.ArgumentError("method is not a String or num"));
      }
      if (args != null)
        args = core.List.from(args[dartx.map](_convertToJS));
      let fn = this[_jsObject][method];
      if (!(fn instanceof Function)) {
        dart.throw(new core.NoSuchMethodError(this[_jsObject], core.Symbol.new(dart.as(method, core.String)), args, dart.map()));
      }
      return _convertToDart(fn.apply(this[_jsObject], args));
    }
  }
  dart.defineNamedConstructor(JsObject, '_fromJs');
  dart.setSignature(JsObject, {
    constructors: () => ({
      _fromJs: [JsObject, [dart.dynamic]],
      new: [JsObject, [JsFunction], [core.List]],
      fromBrowserObject: [JsObject, [dart.dynamic]],
      jsify: [JsObject, [dart.dynamic]]
    }),
    methods: () => ({
      get: [dart.dynamic, [core.Object]],
      set: [dart.dynamic, [core.Object, dart.dynamic]],
      hasProperty: [core.bool, [dart.dynamic]],
      deleteProperty: [dart.void, [dart.dynamic]],
      instanceof: [core.bool, [JsFunction]],
      callMethod: [dart.dynamic, [dart.dynamic], [core.List]]
    }),
    statics: () => ({_convertDataTree: [dart.dynamic, [dart.dynamic]]}),
    names: ['_convertDataTree']
  });
  class JsFunction extends JsObject {
    static withThis(f) {
      return new JsFunction._fromJs(function() {
        let args = [_convertToDart(this)];
        for (let arg of arguments) {
          args.push(_convertToDart(arg));
        }
        return _convertToJS(f(...args));
      });
    }
    _fromJs(jsObject) {
      super._fromJs(jsObject);
    }
    apply(args, opts) {
      let thisArg = opts && 'thisArg' in opts ? opts.thisArg : null;
      return _convertToDart(this[_jsObject].apply(_convertToJS(thisArg), args == null ? null : core.List.from(args[dartx.map](_convertToJS))));
    }
  }
  dart.defineNamedConstructor(JsFunction, '_fromJs');
  dart.setSignature(JsFunction, {
    constructors: () => ({
      withThis: [JsFunction, [core.Function]],
      _fromJs: [JsFunction, [dart.dynamic]]
    }),
    methods: () => ({apply: [dart.dynamic, [core.List], {thisArg: dart.dynamic}]})
  });
  const _checkIndex = Symbol('_checkIndex');
  const _checkInsertIndex = Symbol('_checkInsertIndex');
  const JsArray$ = dart.generic(function(E) {
    class JsArray extends dart.mixin(JsObject, collection.ListMixin$(E)) {
      JsArray() {
        super._fromJs([]);
      }
      from(other) {
        super._fromJs((() => {
          let _ = [];
          _[dartx.addAll](other[dartx.map](dart.as(_convertToJS, __CastType0$(E))));
          return _;
        })());
      }
      _fromJs(jsObject) {
        super._fromJs(jsObject);
      }
      [_checkIndex](index) {
        if (typeof index == 'number' && (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(this.length))) {
          dart.throw(new core.RangeError.range(index, 0, this.length));
        }
      }
      [_checkInsertIndex](index) {
        if (typeof index == 'number' && (dart.notNull(index) < 0 || dart.notNull(index) >= dart.notNull(this.length) + 1)) {
          dart.throw(new core.RangeError.range(index, 0, this.length));
        }
      }
      static _checkRange(start, end, length) {
        if (dart.notNull(start) < 0 || dart.notNull(start) > dart.notNull(length)) {
          dart.throw(new core.RangeError.range(start, 0, length));
        }
        if (dart.notNull(end) < dart.notNull(start) || dart.notNull(end) > dart.notNull(length)) {
          dart.throw(new core.RangeError.range(end, start, length));
        }
      }
      get(index) {
        if (typeof index == 'number' && index == index[dartx.toInt]()) {
          this[_checkIndex](dart.asInt(index));
        }
        return dart.as(super.get(index), E);
      }
      set(index, value) {
        dart.as(value, E);
        if (typeof index == 'number' && index == index[dartx.toInt]()) {
          this[_checkIndex](dart.asInt(index));
        }
        super.set(index, value);
        return value;
      }
      get length() {
        let len = this[_jsObject].length;
        if (typeof len === "number" && len >>> 0 === len) {
          return len;
        }
        dart.throw(new core.StateError('Bad JsArray length'));
      }
      set length(length) {
        super.set('length', length);
      }
      add(value) {
        dart.as(value, E);
        this.callMethod('push', [value]);
      }
      addAll(iterable) {
        dart.as(iterable, core.Iterable$(E));
        let list = iterable instanceof Array ? iterable : core.List.from(iterable);
        this.callMethod('push', dart.as(list, core.List));
      }
      insert(index, element) {
        dart.as(element, E);
        this[_checkInsertIndex](index);
        this.callMethod('splice', [index, 0, element]);
      }
      removeAt(index) {
        this[_checkIndex](index);
        return dart.as(dart.dindex(this.callMethod('splice', [index, 1]), 0), E);
      }
      removeLast() {
        if (this.length == 0)
          dart.throw(new core.RangeError(-1));
        return dart.as(this.callMethod('pop'), E);
      }
      removeRange(start, end) {
        JsArray$()._checkRange(start, end, this.length);
        this.callMethod('splice', [start, dart.notNull(end) - dart.notNull(start)]);
      }
      setRange(start, end, iterable, skipCount) {
        dart.as(iterable, core.Iterable$(E));
        if (skipCount === void 0)
          skipCount = 0;
        JsArray$()._checkRange(start, end, this.length);
        let length = dart.notNull(end) - dart.notNull(start);
        if (length == 0)
          return;
        if (dart.notNull(skipCount) < 0)
          dart.throw(new core.ArgumentError(skipCount));
        let args = [start, length];
        args[dartx.addAll](iterable[dartx.skip](skipCount)[dartx.take](length));
        this.callMethod('splice', args);
      }
      sort(compare) {
        if (compare === void 0)
          compare = null;
        dart.as(compare, dart.functionType(core.int, [E, E]));
        this.callMethod('sort', compare == null ? [] : [compare]);
      }
    }
    dart.defineNamedConstructor(JsArray, 'from');
    dart.defineNamedConstructor(JsArray, '_fromJs');
    dart.setSignature(JsArray, {
      constructors: () => ({
        JsArray: [JsArray$(E), []],
        from: [JsArray$(E), [core.Iterable$(E)]],
        _fromJs: [JsArray$(E), [dart.dynamic]]
      }),
      methods: () => ({
        [_checkIndex]: [dart.dynamic, [core.int]],
        [_checkInsertIndex]: [dart.dynamic, [core.int]],
        get: [E, [core.Object]],
        set: [dart.void, [core.Object, E]],
        add: [dart.void, [E]],
        addAll: [dart.void, [core.Iterable$(E)]],
        insert: [dart.void, [core.int, E]],
        removeAt: [E, [core.int]],
        removeLast: [E, []],
        setRange: [dart.void, [core.int, core.int, core.Iterable$(E)], [core.int]],
        sort: [dart.void, [], [dart.functionType(core.int, [E, E])]]
      }),
      statics: () => ({_checkRange: [dart.dynamic, [core.int, core.int, core.int]]}),
      names: ['_checkRange']
    });
    dart.defineExtensionMembers(JsArray, [
      'get',
      'set',
      'add',
      'addAll',
      'insert',
      'removeAt',
      'removeLast',
      'removeRange',
      'setRange',
      'sort',
      'length',
      'length'
    ]);
    return JsArray;
  });
  let JsArray = JsArray$();
  function _isBrowserType(o) {
    return o instanceof Blob || o instanceof Event || window.KeyRange && o instanceof KeyRange || o instanceof ImageData || o instanceof Node || window.TypedData && o instanceof TypedData || o instanceof Window;
  }
  dart.fn(_isBrowserType, core.bool, [dart.dynamic]);
  const _dartObj = Symbol('_dartObj');
  class _DartObject extends core.Object {
    _DartObject(dartObj) {
      this[_dartObj] = dartObj;
    }
  }
  dart.setSignature(_DartObject, {
    constructors: () => ({_DartObject: [_DartObject, [dart.dynamic]]})
  });
  function _convertToJS(o) {
    if (o == null || typeof o == 'string' || typeof o == 'number' || typeof o == 'boolean' || dart.notNull(_isBrowserType(o))) {
      return o;
    } else if (dart.is(o, core.DateTime)) {
      return _js_helper.Primitives.lazyAsJsDate(o);
    } else if (dart.is(o, JsObject)) {
      return dart.dload(o, _jsObject);
    } else if (dart.is(o, core.Function)) {
      return _putIfAbsent(exports._jsProxies, o, _wrapDartFunction);
    } else {
      return _putIfAbsent(exports._jsProxies, o, dart.fn(o => new _DartObject(o), _DartObject, [dart.dynamic]));
    }
  }
  dart.fn(_convertToJS);
  function _wrapDartFunction(f) {
    let wrapper = function() {
      let args = Array.prototype.map.call(arguments, _convertToDart);
      return _convertToJS(f(...args));
    };
    dart.dsetindex(exports._dartProxies, wrapper, f);
    return wrapper;
  }
  dart.fn(_wrapDartFunction);
  function _convertToDart(o, isBrowserType) {
    if (isBrowserType === void 0)
      isBrowserType = null;
    if (isBrowserType == null)
      isBrowserType = _isBrowserType;
    if (o == null || typeof o == "string" || typeof o == "number" || typeof o == "boolean" || dart.notNull(dart.dcall(isBrowserType, o))) {
      return o;
    } else if (o instanceof Date) {
      let ms = o.getTime();
      return new core.DateTime.fromMillisecondsSinceEpoch(dart.asInt(ms));
    } else if (dart.is(o, _DartObject) && dart.jsobject != dart.realRuntimeType(o)) {
      return dart.dload(o, _dartObj);
    } else {
      return _putIfAbsent(exports._dartProxies, o, _wrapToDart);
    }
  }
  dart.fn(_convertToDart, core.Object, [dart.dynamic], [dart.functionType(core.bool, [dart.dynamic])]);
  function _wrapToDart(o) {
    if (typeof o == "function") {
      return new JsFunction._fromJs(o);
    }
    if (o instanceof Array) {
      return new JsArray._fromJs(o);
    }
    return new JsObject._fromJs(o);
  }
  dart.fn(_wrapToDart, JsObject, [dart.dynamic]);
  dart.defineLazyProperties(exports, {
    get _dartProxies() {
      return new WeakMap();
    },
    get _jsProxies() {
      return new WeakMap();
    }
  });
  function _putIfAbsent(weakMap, o, getValue) {
    let value = weakMap.get(o);
    if (value == null) {
      value = dart.dcall(getValue, o);
      weakMap.set(o, value);
    }
    return value;
  }
  dart.fn(_putIfAbsent, core.Object, [dart.dynamic, dart.dynamic, dart.functionType(dart.dynamic, [dart.dynamic])]);
  function allowInterop(f) {
    return f;
  }
  dart.fn(allowInterop, core.Function, [core.Function]);
  dart.defineLazyProperties(exports, {
    get _interopCaptureThisExpando() {
      return new (core.Expando$(core.Function))();
    },
    set _interopCaptureThisExpando(_) {}
  });
  function allowInteropCaptureThis(f) {
    let ret = exports._interopCaptureThisExpando.get(f);
    if (ret == null) {
      ret = dart.as(function() {
        let args = [this];
        for (let arg of arguments) {
          args.push(arg);
        }
        return f(...args);
      }, core.Function);
      exports._interopCaptureThisExpando.set(f, ret);
    }
    return ret;
  }
  dart.fn(allowInteropCaptureThis, core.Function, [core.Function]);
  const __CastType0$ = dart.generic(function(E) {
    const __CastType0 = dart.typedef('__CastType0', () => dart.functionType(dart.dynamic, [E]));
    return __CastType0;
  });
  let __CastType0 = __CastType0$();
  // Exports:
  exports.JsObject = JsObject;
  exports.JsFunction = JsFunction;
  exports.JsArray$ = JsArray$;
  exports.JsArray = JsArray;
  exports.allowInterop = allowInterop;
  exports.allowInteropCaptureThis = allowInteropCaptureThis;
});
dart_library.library('dart/math', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
  'dart/_js_helper'
], function(exports, dart, core, _js_helper) {
  'use strict';
  let dartx = dart.dartx;
  class _JenkinsSmiHash extends core.Object {
    static combine(hash, value) {
      hash = 536870911 & dart.notNull(hash) + dart.notNull(value);
      hash = 536870911 & dart.notNull(hash) + ((524287 & dart.notNull(hash)) << 10);
      return dart.notNull(hash) ^ dart.notNull(hash) >> 6;
    }
    static finish(hash) {
      hash = 536870911 & dart.notNull(hash) + ((67108863 & dart.notNull(hash)) << 3);
      hash = dart.notNull(hash) ^ dart.notNull(hash) >> 11;
      return 536870911 & dart.notNull(hash) + ((16383 & dart.notNull(hash)) << 15);
    }
    static hash2(a, b) {
      return _JenkinsSmiHash.finish(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(0, dart.as(a, core.int)), dart.as(b, core.int)));
    }
    static hash4(a, b, c, d) {
      return _JenkinsSmiHash.finish(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(_JenkinsSmiHash.combine(0, dart.as(a, core.int)), dart.as(b, core.int)), dart.as(c, core.int)), dart.as(d, core.int)));
    }
  }
  dart.setSignature(_JenkinsSmiHash, {
    statics: () => ({
      combine: [core.int, [core.int, core.int]],
      finish: [core.int, [core.int]],
      hash2: [core.int, [dart.dynamic, dart.dynamic]],
      hash4: [core.int, [dart.dynamic, dart.dynamic, dart.dynamic, dart.dynamic]]
    }),
    names: ['combine', 'finish', 'hash2', 'hash4']
  });
  const Point$ = dart.generic(function(T) {
    class Point extends core.Object {
      Point(x, y) {
        this.x = x;
        this.y = y;
      }
      toString() {
        return `Point(${this.x}, ${this.y})`;
      }
      ['=='](other) {
        if (!dart.is(other, Point$()))
          return false;
        return dart.equals(this.x, dart.dload(other, 'x')) && dart.equals(this.y, dart.dload(other, 'y'));
      }
      get hashCode() {
        return _JenkinsSmiHash.hash2(dart.hashCode(this.x), dart.hashCode(this.y));
      }
      ['+'](other) {
        dart.as(other, Point$(T));
        return new (Point$(T))(dart.notNull(this.x) + dart.notNull(other.x), dart.notNull(this.y) + dart.notNull(other.y));
      }
      ['-'](other) {
        dart.as(other, Point$(T));
        return new (Point$(T))(dart.notNull(this.x) - dart.notNull(other.x), dart.notNull(this.y) - dart.notNull(other.y));
      }
      ['*'](factor) {
        return new (Point$(T))(dart.notNull(this.x) * dart.notNull(factor), dart.notNull(this.y) * dart.notNull(factor));
      }
      get magnitude() {
        return sqrt(dart.notNull(this.x) * dart.notNull(this.x) + dart.notNull(this.y) * dart.notNull(this.y));
      }
      distanceTo(other) {
        dart.as(other, Point$(T));
        let dx = dart.notNull(this.x) - dart.notNull(other.x);
        let dy = dart.notNull(this.y) - dart.notNull(other.y);
        return sqrt(dart.notNull(dx) * dart.notNull(dx) + dart.notNull(dy) * dart.notNull(dy));
      }
      squaredDistanceTo(other) {
        dart.as(other, Point$(T));
        let dx = dart.notNull(this.x) - dart.notNull(other.x);
        let dy = dart.notNull(this.y) - dart.notNull(other.y);
        return dart.notNull(dx) * dart.notNull(dx) + dart.notNull(dy) * dart.notNull(dy);
      }
    }
    dart.setSignature(Point, {
      constructors: () => ({Point: [Point$(T), [T, T]]}),
      methods: () => ({
        '+': [Point$(T), [Point$(T)]],
        '-': [Point$(T), [Point$(T)]],
        '*': [Point$(T), [core.num]],
        distanceTo: [core.double, [Point$(T)]],
        squaredDistanceTo: [T, [Point$(T)]]
      })
    });
    return Point;
  });
  let Point = Point$();
  class Random extends core.Object {
    static new(seed) {
      if (seed === void 0)
        seed = null;
      return seed == null ? dart.const(new _JSRandom()) : new _Random(seed);
    }
  }
  dart.setSignature(Random, {
    constructors: () => ({new: [Random, [], [core.int]]})
  });
  const _RectangleBase$ = dart.generic(function(T) {
    class _RectangleBase extends core.Object {
      _RectangleBase() {
      }
      get right() {
        return dart.notNull(this.left) + dart.notNull(this.width);
      }
      get bottom() {
        return dart.notNull(this.top) + dart.notNull(this.height);
      }
      toString() {
        return `Rectangle (${this.left}, ${this.top}) ${this.width} x ${this.height}`;
      }
      ['=='](other) {
        if (!dart.is(other, Rectangle))
          return false;
        return dart.equals(this.left, dart.dload(other, 'left')) && dart.equals(this.top, dart.dload(other, 'top')) && dart.equals(this.right, dart.dload(other, 'right')) && dart.equals(this.bottom, dart.dload(other, 'bottom'));
      }
      get hashCode() {
        return _JenkinsSmiHash.hash4(dart.hashCode(this.left), dart.hashCode(this.top), dart.hashCode(this.right), dart.hashCode(this.bottom));
      }
      intersection(other) {
        dart.as(other, Rectangle$(T));
        let x0 = max(this.left, other.left);
        let x1 = min(dart.notNull(this.left) + dart.notNull(this.width), dart.notNull(other.left) + dart.notNull(other.width));
        if (dart.notNull(x0) <= dart.notNull(x1)) {
          let y0 = max(this.top, other.top);
          let y1 = min(dart.notNull(this.top) + dart.notNull(this.height), dart.notNull(other.top) + dart.notNull(other.height));
          if (dart.notNull(y0) <= dart.notNull(y1)) {
            return new (Rectangle$(T))(x0, y0, dart.notNull(x1) - dart.notNull(x0), dart.notNull(y1) - dart.notNull(y0));
          }
        }
        return null;
      }
      intersects(other) {
        return dart.notNull(this.left) <= dart.notNull(other.left) + dart.notNull(other.width) && dart.notNull(other.left) <= dart.notNull(this.left) + dart.notNull(this.width) && dart.notNull(this.top) <= dart.notNull(other.top) + dart.notNull(other.height) && dart.notNull(other.top) <= dart.notNull(this.top) + dart.notNull(this.height);
      }
      boundingBox(other) {
        dart.as(other, Rectangle$(T));
        let right = max(dart.notNull(this.left) + dart.notNull(this.width), dart.notNull(other.left) + dart.notNull(other.width));
        let bottom = max(dart.notNull(this.top) + dart.notNull(this.height), dart.notNull(other.top) + dart.notNull(other.height));
        let left = min(this.left, other.left);
        let top = min(this.top, other.top);
        return new (Rectangle$(T))(left, top, dart.notNull(right) - dart.notNull(left), dart.notNull(bottom) - dart.notNull(top));
      }
      containsRectangle(another) {
        return dart.notNull(this.left) <= dart.notNull(another.left) && dart.notNull(this.left) + dart.notNull(this.width) >= dart.notNull(another.left) + dart.notNull(another.width) && dart.notNull(this.top) <= dart.notNull(another.top) && dart.notNull(this.top) + dart.notNull(this.height) >= dart.notNull(another.top) + dart.notNull(another.height);
      }
      containsPoint(another) {
        return dart.notNull(another.x) >= dart.notNull(this.left) && dart.notNull(another.x) <= dart.notNull(this.left) + dart.notNull(this.width) && dart.notNull(another.y) >= dart.notNull(this.top) && dart.notNull(another.y) <= dart.notNull(this.top) + dart.notNull(this.height);
      }
      get topLeft() {
        return new (Point$(T))(this.left, this.top);
      }
      get topRight() {
        return new (Point$(T))(dart.notNull(this.left) + dart.notNull(this.width), this.top);
      }
      get bottomRight() {
        return new (Point$(T))(dart.notNull(this.left) + dart.notNull(this.width), dart.notNull(this.top) + dart.notNull(this.height));
      }
      get bottomLeft() {
        return new (Point$(T))(this.left, dart.notNull(this.top) + dart.notNull(this.height));
      }
    }
    dart.setSignature(_RectangleBase, {
      constructors: () => ({_RectangleBase: [_RectangleBase$(T), []]}),
      methods: () => ({
        intersection: [Rectangle$(T), [Rectangle$(T)]],
        intersects: [core.bool, [Rectangle$(core.num)]],
        boundingBox: [Rectangle$(T), [Rectangle$(T)]],
        containsRectangle: [core.bool, [Rectangle$(core.num)]],
        containsPoint: [core.bool, [Point$(core.num)]]
      })
    });
    return _RectangleBase;
  });
  let _RectangleBase = _RectangleBase$();
  const Rectangle$ = dart.generic(function(T) {
    class Rectangle extends _RectangleBase$(T) {
      Rectangle(left, top, width, height) {
        this.left = left;
        this.top = top;
        this.width = dart.notNull(width) < 0 ? -dart.notNull(width) * 0 : width;
        this.height = dart.notNull(height) < 0 ? -dart.notNull(height) * 0 : height;
        super._RectangleBase();
      }
      static fromPoints(a, b) {
        let left = min(a.x, b.x);
        let width = dart.notNull(max(a.x, b.x)) - dart.notNull(left);
        let top = min(a.y, b.y);
        let height = dart.notNull(max(a.y, b.y)) - dart.notNull(top);
        return new (Rectangle$(T))(left, top, width, height);
      }
    }
    dart.setSignature(Rectangle, {
      constructors: () => ({
        Rectangle: [Rectangle$(T), [T, T, T, T]],
        fromPoints: [Rectangle$(T), [Point$(T), Point$(T)]]
      })
    });
    return Rectangle;
  });
  let Rectangle = Rectangle$();
  const _width = Symbol('_width');
  const _height = Symbol('_height');
  const MutableRectangle$ = dart.generic(function(T) {
    class MutableRectangle extends _RectangleBase$(T) {
      MutableRectangle(left, top, width, height) {
        this.left = left;
        this.top = top;
        this[_width] = dart.notNull(width) < 0 ? _clampToZero(width) : width;
        this[_height] = dart.notNull(height) < 0 ? _clampToZero(height) : height;
        super._RectangleBase();
      }
      static fromPoints(a, b) {
        let left = min(a.x, b.x);
        let width = dart.notNull(max(a.x, b.x)) - dart.notNull(left);
        let top = min(a.y, b.y);
        let height = dart.notNull(max(a.y, b.y)) - dart.notNull(top);
        return new (MutableRectangle$(T))(left, top, width, height);
      }
      get width() {
        return this[_width];
      }
      set width(width) {
        dart.as(width, T);
        if (dart.notNull(width) < 0)
          width = _clampToZero(width);
        this[_width] = width;
      }
      get height() {
        return this[_height];
      }
      set height(height) {
        dart.as(height, T);
        if (dart.notNull(height) < 0)
          height = _clampToZero(height);
        this[_height] = height;
      }
    }
    MutableRectangle[dart.implements] = () => [Rectangle$(T)];
    dart.setSignature(MutableRectangle, {
      constructors: () => ({
        MutableRectangle: [MutableRectangle$(T), [T, T, T, T]],
        fromPoints: [MutableRectangle$(T), [Point$(T), Point$(T)]]
      })
    });
    return MutableRectangle;
  });
  let MutableRectangle = MutableRectangle$();
  function _clampToZero(value) {
    dart.assert(dart.notNull(value) < 0);
    return -dart.notNull(value) * 0;
  }
  dart.fn(_clampToZero, core.num, [core.num]);
  const E = 2.718281828459045;
  const LN10 = 2.302585092994046;
  const LN2 = 0.6931471805599453;
  const LOG2E = 1.4426950408889634;
  const LOG10E = 0.4342944819032518;
  const PI = 3.141592653589793;
  const SQRT1_2 = 0.7071067811865476;
  const SQRT2 = 1.4142135623730951;
  function min(a, b) {
    if (!(typeof a == 'number'))
      dart.throw(new core.ArgumentError(a));
    if (!(typeof b == 'number'))
      dart.throw(new core.ArgumentError(b));
    if (dart.notNull(a) > dart.notNull(b))
      return b;
    if (dart.notNull(a) < dart.notNull(b))
      return a;
    if (typeof b == 'number') {
      if (typeof a == 'number') {
        if (a == 0.0) {
          return (dart.notNull(a) + dart.notNull(b)) * dart.notNull(a) * dart.notNull(b);
        }
      }
      if (a == 0 && dart.notNull(b[dartx.isNegative]) || dart.notNull(b[dartx.isNaN]))
        return b;
      return a;
    }
    return a;
  }
  dart.fn(min, core.num, [core.num, core.num]);
  function max(a, b) {
    if (!(typeof a == 'number'))
      dart.throw(new core.ArgumentError(a));
    if (!(typeof b == 'number'))
      dart.throw(new core.ArgumentError(b));
    if (dart.notNull(a) > dart.notNull(b))
      return a;
    if (dart.notNull(a) < dart.notNull(b))
      return b;
    if (typeof b == 'number') {
      if (typeof a == 'number') {
        if (a == 0.0) {
          return dart.notNull(a) + dart.notNull(b);
        }
      }
      if (dart.notNull(b[dartx.isNaN]))
        return b;
      return a;
    }
    if (b == 0 && dart.notNull(a[dartx.isNegative]))
      return b;
    return a;
  }
  dart.fn(max, core.num, [core.num, core.num]);
  function atan2(a, b) {
    return Math.atan2(_js_helper.checkNum(a), _js_helper.checkNum(b));
  }
  dart.fn(atan2, core.double, [core.num, core.num]);
  function pow(x, exponent) {
    _js_helper.checkNum(x);
    _js_helper.checkNum(exponent);
    return Math.pow(x, exponent);
  }
  dart.fn(pow, core.num, [core.num, core.num]);
  function sin(x) {
    return Math.sin(_js_helper.checkNum(x));
  }
  dart.fn(sin, core.double, [core.num]);
  function cos(x) {
    return Math.cos(_js_helper.checkNum(x));
  }
  dart.fn(cos, core.double, [core.num]);
  function tan(x) {
    return Math.tan(_js_helper.checkNum(x));
  }
  dart.fn(tan, core.double, [core.num]);
  function acos(x) {
    return Math.acos(_js_helper.checkNum(x));
  }
  dart.fn(acos, core.double, [core.num]);
  function asin(x) {
    return Math.asin(_js_helper.checkNum(x));
  }
  dart.fn(asin, core.double, [core.num]);
  function atan(x) {
    return Math.atan(_js_helper.checkNum(x));
  }
  dart.fn(atan, core.double, [core.num]);
  function sqrt(x) {
    return Math.sqrt(_js_helper.checkNum(x));
  }
  dart.fn(sqrt, core.double, [core.num]);
  function exp(x) {
    return Math.exp(_js_helper.checkNum(x));
  }
  dart.fn(exp, core.double, [core.num]);
  function log(x) {
    return Math.log(_js_helper.checkNum(x));
  }
  dart.fn(log, core.double, [core.num]);
  const _POW2_32 = 4294967296;
  class _JSRandom extends core.Object {
    _JSRandom() {
    }
    nextInt(max) {
      if (dart.notNull(max) <= 0 || dart.notNull(max) > dart.notNull(_POW2_32)) {
        dart.throw(new core.RangeError(`max must be in range 0 < max ≤ 2^32, was ${max}`));
      }
      return Math.random() * max >>> 0;
    }
    nextDouble() {
      return Math.random();
    }
    nextBool() {
      return Math.random() < 0.5;
    }
  }
  _JSRandom[dart.implements] = () => [Random];
  dart.setSignature(_JSRandom, {
    constructors: () => ({_JSRandom: [_JSRandom, []]}),
    methods: () => ({
      nextInt: [core.int, [core.int]],
      nextDouble: [core.double, []],
      nextBool: [core.bool, []]
    })
  });
  const _lo = Symbol('_lo');
  const _hi = Symbol('_hi');
  const _nextState = Symbol('_nextState');
  class _Random extends core.Object {
    _Random(seed) {
      this[_lo] = 0;
      this[_hi] = 0;
      let empty_seed = 0;
      if (dart.notNull(seed) < 0) {
        empty_seed = -1;
      }
      do {
        let low = dart.notNull(seed) & dart.notNull(_Random._MASK32);
        seed = ((dart.notNull(seed) - dart.notNull(low)) / dart.notNull(_POW2_32))[dartx.truncate]();
        let high = dart.notNull(seed) & dart.notNull(_Random._MASK32);
        seed = ((dart.notNull(seed) - dart.notNull(high)) / dart.notNull(_POW2_32))[dartx.truncate]();
        let tmplow = dart.notNull(low) << 21;
        let tmphigh = dart.notNull(high) << 21 | dart.notNull(low) >> 11;
        tmplow = (~dart.notNull(low) & dart.notNull(_Random._MASK32)) + dart.notNull(tmplow);
        low = dart.notNull(tmplow) & dart.notNull(_Random._MASK32);
        high = ~dart.notNull(high) + dart.notNull(tmphigh) + ((dart.notNull(tmplow) - dart.notNull(low)) / 4294967296)[dartx.truncate]() & dart.notNull(_Random._MASK32);
        tmphigh = dart.notNull(high) >> 24;
        tmplow = dart.notNull(low) >> 24 | dart.notNull(high) << 8;
        low = dart.notNull(low) ^ dart.notNull(tmplow);
        high = dart.notNull(high) ^ dart.notNull(tmphigh);
        tmplow = dart.notNull(low) * 265;
        low = dart.notNull(tmplow) & dart.notNull(_Random._MASK32);
        high = dart.notNull(high) * 265 + ((dart.notNull(tmplow) - dart.notNull(low)) / 4294967296)[dartx.truncate]() & dart.notNull(_Random._MASK32);
        tmphigh = dart.notNull(high) >> 14;
        tmplow = dart.notNull(low) >> 14 | dart.notNull(high) << 18;
        low = dart.notNull(low) ^ dart.notNull(tmplow);
        high = dart.notNull(high) ^ dart.notNull(tmphigh);
        tmplow = dart.notNull(low) * 21;
        low = dart.notNull(tmplow) & dart.notNull(_Random._MASK32);
        high = dart.notNull(high) * 21 + ((dart.notNull(tmplow) - dart.notNull(low)) / 4294967296)[dartx.truncate]() & dart.notNull(_Random._MASK32);
        tmphigh = dart.notNull(high) >> 28;
        tmplow = dart.notNull(low) >> 28 | dart.notNull(high) << 4;
        low = dart.notNull(low) ^ dart.notNull(tmplow);
        high = dart.notNull(high) ^ dart.notNull(tmphigh);
        tmplow = dart.notNull(low) << 31;
        tmphigh = dart.notNull(high) << 31 | dart.notNull(low) >> 1;
        tmplow = dart.notNull(tmplow) + dart.notNull(low);
        low = dart.notNull(tmplow) & dart.notNull(_Random._MASK32);
        high = dart.notNull(high) + dart.notNull(tmphigh) + ((dart.notNull(tmplow) - dart.notNull(low)) / 4294967296)[dartx.truncate]() & dart.notNull(_Random._MASK32);
        tmplow = dart.notNull(this[_lo]) * 1037;
        this[_lo] = dart.notNull(tmplow) & dart.notNull(_Random._MASK32);
        this[_hi] = dart.notNull(this[_hi]) * 1037 + ((dart.notNull(tmplow) - dart.notNull(this[_lo])) / 4294967296)[dartx.truncate]() & dart.notNull(_Random._MASK32);
        this[_lo] = dart.notNull(this[_lo]) ^ dart.notNull(low);
        this[_hi] = dart.notNull(this[_hi]) ^ dart.notNull(high);
      } while (seed != empty_seed);
      if (this[_hi] == 0 && this[_lo] == 0) {
        this[_lo] = 23063;
      }
      this[_nextState]();
      this[_nextState]();
      this[_nextState]();
      this[_nextState]();
    }
    [_nextState]() {
      let tmpHi = 4294901760 * dart.notNull(this[_lo]);
      let tmpHiLo = dart.notNull(tmpHi) & dart.notNull(_Random._MASK32);
      let tmpHiHi = dart.notNull(tmpHi) - dart.notNull(tmpHiLo);
      let tmpLo = 55905 * dart.notNull(this[_lo]);
      let tmpLoLo = dart.notNull(tmpLo) & dart.notNull(_Random._MASK32);
      let tmpLoHi = dart.notNull(tmpLo) - dart.notNull(tmpLoLo);
      let newLo = dart.notNull(tmpLoLo) + dart.notNull(tmpHiLo) + dart.notNull(this[_hi]);
      this[_lo] = dart.notNull(newLo) & dart.notNull(_Random._MASK32);
      let newLoHi = dart.notNull(newLo) - dart.notNull(this[_lo]);
      this[_hi] = ((dart.notNull(tmpLoHi) + dart.notNull(tmpHiHi) + dart.notNull(newLoHi)) / dart.notNull(_POW2_32))[dartx.truncate]() & dart.notNull(_Random._MASK32);
      dart.assert(dart.notNull(this[_lo]) < dart.notNull(_POW2_32));
      dart.assert(dart.notNull(this[_hi]) < dart.notNull(_POW2_32));
    }
    nextInt(max) {
      if (dart.notNull(max) <= 0 || dart.notNull(max) > dart.notNull(_POW2_32)) {
        dart.throw(new core.RangeError(`max must be in range 0 < max ≤ 2^32, was ${max}`));
      }
      if ((dart.notNull(max) & dart.notNull(max) - 1) == 0) {
        this[_nextState]();
        return dart.notNull(this[_lo]) & dart.notNull(max) - 1;
      }
      let rnd32 = null;
      let result = null;
      do {
        this[_nextState]();
        rnd32 = this[_lo];
        result = dart.asInt(rnd32[dartx.remainder](max));
      } while (dart.notNull(rnd32) - dart.notNull(result) + dart.notNull(max) >= dart.notNull(_POW2_32));
      return result;
    }
    nextDouble() {
      this[_nextState]();
      let bits26 = dart.notNull(this[_lo]) & (1 << 26) - 1;
      this[_nextState]();
      let bits27 = dart.notNull(this[_lo]) & (1 << 27) - 1;
      return (dart.notNull(bits26) * dart.notNull(_Random._POW2_27_D) + dart.notNull(bits27)) / dart.notNull(_Random._POW2_53_D);
    }
    nextBool() {
      this[_nextState]();
      return (dart.notNull(this[_lo]) & 1) == 0;
    }
  }
  _Random[dart.implements] = () => [Random];
  dart.setSignature(_Random, {
    constructors: () => ({_Random: [_Random, [core.int]]}),
    methods: () => ({
      [_nextState]: [dart.void, []],
      nextInt: [core.int, [core.int]],
      nextDouble: [core.double, []],
      nextBool: [core.bool, []]
    })
  });
  _Random._POW2_53_D = 1.0 * 9007199254740992;
  _Random._POW2_27_D = 1.0 * (1 << 27);
  _Random._MASK32 = 4294967295;
  // Exports:
  exports.Point$ = Point$;
  exports.Point = Point;
  exports.Random = Random;
  exports.Rectangle$ = Rectangle$;
  exports.Rectangle = Rectangle;
  exports.MutableRectangle$ = MutableRectangle$;
  exports.MutableRectangle = MutableRectangle;
  exports.E = E;
  exports.LN10 = LN10;
  exports.LN2 = LN2;
  exports.LOG2E = LOG2E;
  exports.LOG10E = LOG10E;
  exports.PI = PI;
  exports.SQRT1_2 = SQRT1_2;
  exports.SQRT2 = SQRT2;
  exports.min = min;
  exports.max = max;
  exports.atan2 = atan2;
  exports.pow = pow;
  exports.sin = sin;
  exports.cos = cos;
  exports.tan = tan;
  exports.acos = acos;
  exports.asin = asin;
  exports.atan = atan;
  exports.sqrt = sqrt;
  exports.exp = exp;
  exports.log = log;
});
dart_library.library('dart/mirrors', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
  'dart/_js_mirrors'
], function(exports, dart, core, _js_mirrors) {
  'use strict';
  let dartx = dart.dartx;
  class MirrorSystem extends core.Object {
    findLibrary(libraryName) {
      return this.libraries.values[dartx.singleWhere](dart.fn(library => dart.equals(dart.dload(library, 'simpleName'), libraryName), core.bool, [dart.dynamic]));
    }
    static getName(symbol) {
      return _js_mirrors.getName(symbol);
    }
    static getSymbol(name, library) {
      if (library === void 0)
        library = null;
      return _js_mirrors.getSymbol(name, library);
    }
  }
  dart.setSignature(MirrorSystem, {
    methods: () => ({findLibrary: [LibraryMirror, [core.Symbol]]}),
    statics: () => ({
      getName: [core.String, [core.Symbol]],
      getSymbol: [core.Symbol, [core.String], [LibraryMirror]]
    }),
    names: ['getName', 'getSymbol']
  });
  function currentMirrorSystem() {
    return dart.as(_js_mirrors.currentJsMirrorSystem, MirrorSystem);
  }
  dart.fn(currentMirrorSystem, MirrorSystem, []);
  function reflect(reflectee) {
    return _js_mirrors.reflect(reflectee);
  }
  dart.fn(reflect, () => dart.definiteFunctionType(InstanceMirror, [core.Object]));
  function reflectClass(key) {
    if (!dart.is(key, core.Type) || dart.equals(key, dart.dynamic)) {
      dart.throw(new core.ArgumentError(`${key} does not denote a class`));
    }
    let tm = reflectType(key);
    if (!dart.is(tm, ClassMirror)) {
      dart.throw(new core.ArgumentError(`${key} does not denote a class`));
    }
    return dart.as(dart.as(tm, ClassMirror).originalDeclaration, ClassMirror);
  }
  dart.fn(reflectClass, () => dart.definiteFunctionType(ClassMirror, [core.Type]));
  function reflectType(key) {
    if (dart.equals(key, dart.dynamic)) {
      return currentMirrorSystem().dynamicType;
    }
    return _js_mirrors.reflectType(key);
  }
  dart.fn(reflectType, () => dart.definiteFunctionType(TypeMirror, [core.Type]));
  class Mirror extends core.Object {}
  class IsolateMirror extends core.Object {}
  IsolateMirror[dart.implements] = () => [Mirror];
  class DeclarationMirror extends core.Object {}
  DeclarationMirror[dart.implements] = () => [Mirror];
  class ObjectMirror extends core.Object {}
  ObjectMirror[dart.implements] = () => [Mirror];
  class InstanceMirror extends core.Object {}
  InstanceMirror[dart.implements] = () => [ObjectMirror];
  class ClosureMirror extends core.Object {}
  ClosureMirror[dart.implements] = () => [InstanceMirror];
  class LibraryMirror extends core.Object {}
  LibraryMirror[dart.implements] = () => [DeclarationMirror, ObjectMirror];
  class LibraryDependencyMirror extends core.Object {}
  LibraryDependencyMirror[dart.implements] = () => [Mirror];
  class CombinatorMirror extends core.Object {}
  CombinatorMirror[dart.implements] = () => [Mirror];
  class TypeMirror extends core.Object {}
  TypeMirror[dart.implements] = () => [DeclarationMirror];
  class ClassMirror extends core.Object {}
  ClassMirror[dart.implements] = () => [TypeMirror, ObjectMirror];
  class FunctionTypeMirror extends core.Object {}
  FunctionTypeMirror[dart.implements] = () => [ClassMirror];
  class TypeVariableMirror extends TypeMirror {}
  class TypedefMirror extends core.Object {}
  TypedefMirror[dart.implements] = () => [TypeMirror];
  class MethodMirror extends core.Object {}
  MethodMirror[dart.implements] = () => [DeclarationMirror];
  class VariableMirror extends core.Object {}
  VariableMirror[dart.implements] = () => [DeclarationMirror];
  class ParameterMirror extends core.Object {}
  ParameterMirror[dart.implements] = () => [VariableMirror];
  class SourceLocation extends core.Object {}
  class Comment extends core.Object {
    Comment(text, trimmedText, isDocComment) {
      this.text = text;
      this.trimmedText = trimmedText;
      this.isDocComment = isDocComment;
    }
  }
  dart.setSignature(Comment, {
    constructors: () => ({Comment: [Comment, [core.String, core.String, core.bool]]})
  });
  class MirrorsUsed extends core.Object {
    MirrorsUsed(opts) {
      let symbols = opts && 'symbols' in opts ? opts.symbols : null;
      let targets = opts && 'targets' in opts ? opts.targets : null;
      let metaTargets = opts && 'metaTargets' in opts ? opts.metaTargets : null;
      let override = opts && 'override' in opts ? opts.override : null;
      this.symbols = symbols;
      this.targets = targets;
      this.metaTargets = metaTargets;
      this.override = override;
    }
  }
  dart.setSignature(MirrorsUsed, {
    constructors: () => ({MirrorsUsed: [MirrorsUsed, [], {symbols: dart.dynamic, targets: dart.dynamic, metaTargets: dart.dynamic, override: dart.dynamic}]})
  });
  // Exports:
  exports.MirrorSystem = MirrorSystem;
  exports.currentMirrorSystem = currentMirrorSystem;
  exports.reflect = reflect;
  exports.reflectClass = reflectClass;
  exports.reflectType = reflectType;
  exports.Mirror = Mirror;
  exports.IsolateMirror = IsolateMirror;
  exports.DeclarationMirror = DeclarationMirror;
  exports.ObjectMirror = ObjectMirror;
  exports.InstanceMirror = InstanceMirror;
  exports.ClosureMirror = ClosureMirror;
  exports.LibraryMirror = LibraryMirror;
  exports.LibraryDependencyMirror = LibraryDependencyMirror;
  exports.CombinatorMirror = CombinatorMirror;
  exports.TypeMirror = TypeMirror;
  exports.ClassMirror = ClassMirror;
  exports.FunctionTypeMirror = FunctionTypeMirror;
  exports.TypeVariableMirror = TypeVariableMirror;
  exports.TypedefMirror = TypedefMirror;
  exports.MethodMirror = MethodMirror;
  exports.VariableMirror = VariableMirror;
  exports.ParameterMirror = ParameterMirror;
  exports.SourceLocation = SourceLocation;
  exports.Comment = Comment;
  exports.MirrorsUsed = MirrorsUsed;
});
dart_library.library('dart/typed_data', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
  'dart/_native_typed_data'
], function(exports, dart, core, _native_typed_data) {
  'use strict';
  let dartx = dart.dartx;
  class ByteBuffer extends core.Object {}
  class TypedData extends core.Object {}
  const _littleEndian = Symbol('_littleEndian');
  class Endianness extends core.Object {
    _(littleEndian) {
      this[_littleEndian] = littleEndian;
    }
  }
  dart.defineNamedConstructor(Endianness, '_');
  dart.setSignature(Endianness, {
    constructors: () => ({_: [Endianness, [core.bool]]})
  });
  Endianness.BIG_ENDIAN = dart.const(new Endianness._(false));
  Endianness.LITTLE_ENDIAN = dart.const(new Endianness._(true));
  dart.defineLazyProperties(Endianness, {
    get HOST_ENDIAN() {
      return ByteData.view(Uint16List.fromList(dart.list([1], core.int)).buffer).getInt8(0) == 1 ? Endianness.LITTLE_ENDIAN : Endianness.BIG_ENDIAN;
    }
  });
  class ByteData extends core.Object {
    static new(length) {
      return _native_typed_data.NativeByteData.new(length);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asByteData(offsetInBytes, length);
    }
  }
  ByteData[dart.implements] = () => [TypedData];
  dart.setSignature(ByteData, {
    constructors: () => ({
      new: [ByteData, [core.int]],
      view: [ByteData, [ByteBuffer], [core.int, core.int]]
    })
  });
  class Int8List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeInt8List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeInt8List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asInt8List(offsetInBytes, length);
    }
  }
  Int8List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Int8List, {
    constructors: () => ({
      new: [Int8List, [core.int]],
      fromList: [Int8List, [core.List$(core.int)]],
      view: [Int8List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Int8List.BYTES_PER_ELEMENT = 1;
  class Uint8List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeUint8List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeUint8List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asUint8List(offsetInBytes, length);
    }
  }
  Uint8List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Uint8List, {
    constructors: () => ({
      new: [Uint8List, [core.int]],
      fromList: [Uint8List, [core.List$(core.int)]],
      view: [Uint8List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Uint8List.BYTES_PER_ELEMENT = 1;
  class Uint8ClampedList extends core.Object {
    static new(length) {
      return _native_typed_data.NativeUint8ClampedList.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeUint8ClampedList.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asUint8ClampedList(offsetInBytes, length);
    }
  }
  Uint8ClampedList[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Uint8ClampedList, {
    constructors: () => ({
      new: [Uint8ClampedList, [core.int]],
      fromList: [Uint8ClampedList, [core.List$(core.int)]],
      view: [Uint8ClampedList, [ByteBuffer], [core.int, core.int]]
    })
  });
  Uint8ClampedList.BYTES_PER_ELEMENT = 1;
  class Int16List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeInt16List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeInt16List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asInt16List(offsetInBytes, length);
    }
  }
  Int16List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Int16List, {
    constructors: () => ({
      new: [Int16List, [core.int]],
      fromList: [Int16List, [core.List$(core.int)]],
      view: [Int16List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Int16List.BYTES_PER_ELEMENT = 2;
  class Uint16List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeUint16List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeUint16List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asUint16List(offsetInBytes, length);
    }
  }
  Uint16List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Uint16List, {
    constructors: () => ({
      new: [Uint16List, [core.int]],
      fromList: [Uint16List, [core.List$(core.int)]],
      view: [Uint16List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Uint16List.BYTES_PER_ELEMENT = 2;
  class Int32List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeInt32List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeInt32List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asInt32List(offsetInBytes, length);
    }
  }
  Int32List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Int32List, {
    constructors: () => ({
      new: [Int32List, [core.int]],
      fromList: [Int32List, [core.List$(core.int)]],
      view: [Int32List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Int32List.BYTES_PER_ELEMENT = 4;
  class Uint32List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeUint32List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeUint32List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asUint32List(offsetInBytes, length);
    }
  }
  Uint32List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Uint32List, {
    constructors: () => ({
      new: [Uint32List, [core.int]],
      fromList: [Uint32List, [core.List$(core.int)]],
      view: [Uint32List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Uint32List.BYTES_PER_ELEMENT = 4;
  class Int64List extends core.Object {
    static new(length) {
      dart.throw(new core.UnsupportedError("Int64List not supported by dart2js."));
    }
    static fromList(elements) {
      dart.throw(new core.UnsupportedError("Int64List not supported by dart2js."));
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asInt64List(offsetInBytes, length);
    }
  }
  Int64List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Int64List, {
    constructors: () => ({
      new: [Int64List, [core.int]],
      fromList: [Int64List, [core.List$(core.int)]],
      view: [Int64List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Int64List.BYTES_PER_ELEMENT = 8;
  class Uint64List extends core.Object {
    static new(length) {
      dart.throw(new core.UnsupportedError("Uint64List not supported by dart2js."));
    }
    static fromList(elements) {
      dart.throw(new core.UnsupportedError("Uint64List not supported by dart2js."));
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asUint64List(offsetInBytes, length);
    }
  }
  Uint64List[dart.implements] = () => [core.List$(core.int), TypedData];
  dart.setSignature(Uint64List, {
    constructors: () => ({
      new: [Uint64List, [core.int]],
      fromList: [Uint64List, [core.List$(core.int)]],
      view: [Uint64List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Uint64List.BYTES_PER_ELEMENT = 8;
  class Float32List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeFloat32List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeFloat32List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asFloat32List(offsetInBytes, length);
    }
  }
  Float32List[dart.implements] = () => [core.List$(core.double), TypedData];
  dart.setSignature(Float32List, {
    constructors: () => ({
      new: [Float32List, [core.int]],
      fromList: [Float32List, [core.List$(core.double)]],
      view: [Float32List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Float32List.BYTES_PER_ELEMENT = 4;
  class Float64List extends core.Object {
    static new(length) {
      return _native_typed_data.NativeFloat64List.new(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeFloat64List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asFloat64List(offsetInBytes, length);
    }
  }
  Float64List[dart.implements] = () => [core.List$(core.double), TypedData];
  dart.setSignature(Float64List, {
    constructors: () => ({
      new: [Float64List, [core.int]],
      fromList: [Float64List, [core.List$(core.double)]],
      view: [Float64List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Float64List.BYTES_PER_ELEMENT = 8;
  class Float32x4List extends core.Object {
    static new(length) {
      return new _native_typed_data.NativeFloat32x4List(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeFloat32x4List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asFloat32x4List(offsetInBytes, length);
    }
  }
  Float32x4List[dart.implements] = () => [core.List$(Float32x4), TypedData];
  dart.setSignature(Float32x4List, {
    constructors: () => ({
      new: [Float32x4List, [core.int]],
      fromList: [Float32x4List, [core.List$(Float32x4)]],
      view: [Float32x4List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Float32x4List.BYTES_PER_ELEMENT = 16;
  class Int32x4List extends core.Object {
    static new(length) {
      return new _native_typed_data.NativeInt32x4List(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeInt32x4List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asInt32x4List(offsetInBytes, length);
    }
  }
  Int32x4List[dart.implements] = () => [core.List$(Int32x4), TypedData];
  dart.setSignature(Int32x4List, {
    constructors: () => ({
      new: [Int32x4List, [core.int]],
      fromList: [Int32x4List, [core.List$(Int32x4)]],
      view: [Int32x4List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Int32x4List.BYTES_PER_ELEMENT = 16;
  class Float64x2List extends core.Object {
    static new(length) {
      return new _native_typed_data.NativeFloat64x2List(length);
    }
    static fromList(elements) {
      return _native_typed_data.NativeFloat64x2List.fromList(elements);
    }
    static view(buffer, offsetInBytes, length) {
      if (offsetInBytes === void 0)
        offsetInBytes = 0;
      if (length === void 0)
        length = null;
      return buffer.asFloat64x2List(offsetInBytes, length);
    }
  }
  Float64x2List[dart.implements] = () => [core.List$(Float64x2), TypedData];
  dart.setSignature(Float64x2List, {
    constructors: () => ({
      new: [Float64x2List, [core.int]],
      fromList: [Float64x2List, [core.List$(Float64x2)]],
      view: [Float64x2List, [ByteBuffer], [core.int, core.int]]
    })
  });
  Float64x2List.BYTES_PER_ELEMENT = 16;
  class Float32x4 extends core.Object {
    static new(x, y, z, w) {
      return new _native_typed_data.NativeFloat32x4(x, y, z, w);
    }
    static splat(v) {
      return new _native_typed_data.NativeFloat32x4.splat(v);
    }
    static zero() {
      return new _native_typed_data.NativeFloat32x4.zero();
    }
    static fromInt32x4Bits(x) {
      return _native_typed_data.NativeFloat32x4.fromInt32x4Bits(x);
    }
    static fromFloat64x2(v) {
      return new _native_typed_data.NativeFloat32x4.fromFloat64x2(v);
    }
  }
  dart.setSignature(Float32x4, {
    constructors: () => ({
      new: [Float32x4, [core.double, core.double, core.double, core.double]],
      splat: [Float32x4, [core.double]],
      zero: [Float32x4, []],
      fromInt32x4Bits: [Float32x4, [Int32x4]],
      fromFloat64x2: [Float32x4, [Float64x2]]
    })
  });
  Float32x4.XXXX = 0;
  Float32x4.XXXY = 64;
  Float32x4.XXXZ = 128;
  Float32x4.XXXW = 192;
  Float32x4.XXYX = 16;
  Float32x4.XXYY = 80;
  Float32x4.XXYZ = 144;
  Float32x4.XXYW = 208;
  Float32x4.XXZX = 32;
  Float32x4.XXZY = 96;
  Float32x4.XXZZ = 160;
  Float32x4.XXZW = 224;
  Float32x4.XXWX = 48;
  Float32x4.XXWY = 112;
  Float32x4.XXWZ = 176;
  Float32x4.XXWW = 240;
  Float32x4.XYXX = 4;
  Float32x4.XYXY = 68;
  Float32x4.XYXZ = 132;
  Float32x4.XYXW = 196;
  Float32x4.XYYX = 20;
  Float32x4.XYYY = 84;
  Float32x4.XYYZ = 148;
  Float32x4.XYYW = 212;
  Float32x4.XYZX = 36;
  Float32x4.XYZY = 100;
  Float32x4.XYZZ = 164;
  Float32x4.XYZW = 228;
  Float32x4.XYWX = 52;
  Float32x4.XYWY = 116;
  Float32x4.XYWZ = 180;
  Float32x4.XYWW = 244;
  Float32x4.XZXX = 8;
  Float32x4.XZXY = 72;
  Float32x4.XZXZ = 136;
  Float32x4.XZXW = 200;
  Float32x4.XZYX = 24;
  Float32x4.XZYY = 88;
  Float32x4.XZYZ = 152;
  Float32x4.XZYW = 216;
  Float32x4.XZZX = 40;
  Float32x4.XZZY = 104;
  Float32x4.XZZZ = 168;
  Float32x4.XZZW = 232;
  Float32x4.XZWX = 56;
  Float32x4.XZWY = 120;
  Float32x4.XZWZ = 184;
  Float32x4.XZWW = 248;
  Float32x4.XWXX = 12;
  Float32x4.XWXY = 76;
  Float32x4.XWXZ = 140;
  Float32x4.XWXW = 204;
  Float32x4.XWYX = 28;
  Float32x4.XWYY = 92;
  Float32x4.XWYZ = 156;
  Float32x4.XWYW = 220;
  Float32x4.XWZX = 44;
  Float32x4.XWZY = 108;
  Float32x4.XWZZ = 172;
  Float32x4.XWZW = 236;
  Float32x4.XWWX = 60;
  Float32x4.XWWY = 124;
  Float32x4.XWWZ = 188;
  Float32x4.XWWW = 252;
  Float32x4.YXXX = 1;
  Float32x4.YXXY = 65;
  Float32x4.YXXZ = 129;
  Float32x4.YXXW = 193;
  Float32x4.YXYX = 17;
  Float32x4.YXYY = 81;
  Float32x4.YXYZ = 145;
  Float32x4.YXYW = 209;
  Float32x4.YXZX = 33;
  Float32x4.YXZY = 97;
  Float32x4.YXZZ = 161;
  Float32x4.YXZW = 225;
  Float32x4.YXWX = 49;
  Float32x4.YXWY = 113;
  Float32x4.YXWZ = 177;
  Float32x4.YXWW = 241;
  Float32x4.YYXX = 5;
  Float32x4.YYXY = 69;
  Float32x4.YYXZ = 133;
  Float32x4.YYXW = 197;
  Float32x4.YYYX = 21;
  Float32x4.YYYY = 85;
  Float32x4.YYYZ = 149;
  Float32x4.YYYW = 213;
  Float32x4.YYZX = 37;
  Float32x4.YYZY = 101;
  Float32x4.YYZZ = 165;
  Float32x4.YYZW = 229;
  Float32x4.YYWX = 53;
  Float32x4.YYWY = 117;
  Float32x4.YYWZ = 181;
  Float32x4.YYWW = 245;
  Float32x4.YZXX = 9;
  Float32x4.YZXY = 73;
  Float32x4.YZXZ = 137;
  Float32x4.YZXW = 201;
  Float32x4.YZYX = 25;
  Float32x4.YZYY = 89;
  Float32x4.YZYZ = 153;
  Float32x4.YZYW = 217;
  Float32x4.YZZX = 41;
  Float32x4.YZZY = 105;
  Float32x4.YZZZ = 169;
  Float32x4.YZZW = 233;
  Float32x4.YZWX = 57;
  Float32x4.YZWY = 121;
  Float32x4.YZWZ = 185;
  Float32x4.YZWW = 249;
  Float32x4.YWXX = 13;
  Float32x4.YWXY = 77;
  Float32x4.YWXZ = 141;
  Float32x4.YWXW = 205;
  Float32x4.YWYX = 29;
  Float32x4.YWYY = 93;
  Float32x4.YWYZ = 157;
  Float32x4.YWYW = 221;
  Float32x4.YWZX = 45;
  Float32x4.YWZY = 109;
  Float32x4.YWZZ = 173;
  Float32x4.YWZW = 237;
  Float32x4.YWWX = 61;
  Float32x4.YWWY = 125;
  Float32x4.YWWZ = 189;
  Float32x4.YWWW = 253;
  Float32x4.ZXXX = 2;
  Float32x4.ZXXY = 66;
  Float32x4.ZXXZ = 130;
  Float32x4.ZXXW = 194;
  Float32x4.ZXYX = 18;
  Float32x4.ZXYY = 82;
  Float32x4.ZXYZ = 146;
  Float32x4.ZXYW = 210;
  Float32x4.ZXZX = 34;
  Float32x4.ZXZY = 98;
  Float32x4.ZXZZ = 162;
  Float32x4.ZXZW = 226;
  Float32x4.ZXWX = 50;
  Float32x4.ZXWY = 114;
  Float32x4.ZXWZ = 178;
  Float32x4.ZXWW = 242;
  Float32x4.ZYXX = 6;
  Float32x4.ZYXY = 70;
  Float32x4.ZYXZ = 134;
  Float32x4.ZYXW = 198;
  Float32x4.ZYYX = 22;
  Float32x4.ZYYY = 86;
  Float32x4.ZYYZ = 150;
  Float32x4.ZYYW = 214;
  Float32x4.ZYZX = 38;
  Float32x4.ZYZY = 102;
  Float32x4.ZYZZ = 166;
  Float32x4.ZYZW = 230;
  Float32x4.ZYWX = 54;
  Float32x4.ZYWY = 118;
  Float32x4.ZYWZ = 182;
  Float32x4.ZYWW = 246;
  Float32x4.ZZXX = 10;
  Float32x4.ZZXY = 74;
  Float32x4.ZZXZ = 138;
  Float32x4.ZZXW = 202;
  Float32x4.ZZYX = 26;
  Float32x4.ZZYY = 90;
  Float32x4.ZZYZ = 154;
  Float32x4.ZZYW = 218;
  Float32x4.ZZZX = 42;
  Float32x4.ZZZY = 106;
  Float32x4.ZZZZ = 170;
  Float32x4.ZZZW = 234;
  Float32x4.ZZWX = 58;
  Float32x4.ZZWY = 122;
  Float32x4.ZZWZ = 186;
  Float32x4.ZZWW = 250;
  Float32x4.ZWXX = 14;
  Float32x4.ZWXY = 78;
  Float32x4.ZWXZ = 142;
  Float32x4.ZWXW = 206;
  Float32x4.ZWYX = 30;
  Float32x4.ZWYY = 94;
  Float32x4.ZWYZ = 158;
  Float32x4.ZWYW = 222;
  Float32x4.ZWZX = 46;
  Float32x4.ZWZY = 110;
  Float32x4.ZWZZ = 174;
  Float32x4.ZWZW = 238;
  Float32x4.ZWWX = 62;
  Float32x4.ZWWY = 126;
  Float32x4.ZWWZ = 190;
  Float32x4.ZWWW = 254;
  Float32x4.WXXX = 3;
  Float32x4.WXXY = 67;
  Float32x4.WXXZ = 131;
  Float32x4.WXXW = 195;
  Float32x4.WXYX = 19;
  Float32x4.WXYY = 83;
  Float32x4.WXYZ = 147;
  Float32x4.WXYW = 211;
  Float32x4.WXZX = 35;
  Float32x4.WXZY = 99;
  Float32x4.WXZZ = 163;
  Float32x4.WXZW = 227;
  Float32x4.WXWX = 51;
  Float32x4.WXWY = 115;
  Float32x4.WXWZ = 179;
  Float32x4.WXWW = 243;
  Float32x4.WYXX = 7;
  Float32x4.WYXY = 71;
  Float32x4.WYXZ = 135;
  Float32x4.WYXW = 199;
  Float32x4.WYYX = 23;
  Float32x4.WYYY = 87;
  Float32x4.WYYZ = 151;
  Float32x4.WYYW = 215;
  Float32x4.WYZX = 39;
  Float32x4.WYZY = 103;
  Float32x4.WYZZ = 167;
  Float32x4.WYZW = 231;
  Float32x4.WYWX = 55;
  Float32x4.WYWY = 119;
  Float32x4.WYWZ = 183;
  Float32x4.WYWW = 247;
  Float32x4.WZXX = 11;
  Float32x4.WZXY = 75;
  Float32x4.WZXZ = 139;
  Float32x4.WZXW = 203;
  Float32x4.WZYX = 27;
  Float32x4.WZYY = 91;
  Float32x4.WZYZ = 155;
  Float32x4.WZYW = 219;
  Float32x4.WZZX = 43;
  Float32x4.WZZY = 107;
  Float32x4.WZZZ = 171;
  Float32x4.WZZW = 235;
  Float32x4.WZWX = 59;
  Float32x4.WZWY = 123;
  Float32x4.WZWZ = 187;
  Float32x4.WZWW = 251;
  Float32x4.WWXX = 15;
  Float32x4.WWXY = 79;
  Float32x4.WWXZ = 143;
  Float32x4.WWXW = 207;
  Float32x4.WWYX = 31;
  Float32x4.WWYY = 95;
  Float32x4.WWYZ = 159;
  Float32x4.WWYW = 223;
  Float32x4.WWZX = 47;
  Float32x4.WWZY = 111;
  Float32x4.WWZZ = 175;
  Float32x4.WWZW = 239;
  Float32x4.WWWX = 63;
  Float32x4.WWWY = 127;
  Float32x4.WWWZ = 191;
  Float32x4.WWWW = 255;
  class Int32x4 extends core.Object {
    static new(x, y, z, w) {
      return new _native_typed_data.NativeInt32x4(x, y, z, w);
    }
    static bool(x, y, z, w) {
      return new _native_typed_data.NativeInt32x4.bool(x, y, z, w);
    }
    static fromFloat32x4Bits(x) {
      return _native_typed_data.NativeInt32x4.fromFloat32x4Bits(x);
    }
  }
  dart.setSignature(Int32x4, {
    constructors: () => ({
      new: [Int32x4, [core.int, core.int, core.int, core.int]],
      bool: [Int32x4, [core.bool, core.bool, core.bool, core.bool]],
      fromFloat32x4Bits: [Int32x4, [Float32x4]]
    })
  });
  Int32x4.XXXX = 0;
  Int32x4.XXXY = 64;
  Int32x4.XXXZ = 128;
  Int32x4.XXXW = 192;
  Int32x4.XXYX = 16;
  Int32x4.XXYY = 80;
  Int32x4.XXYZ = 144;
  Int32x4.XXYW = 208;
  Int32x4.XXZX = 32;
  Int32x4.XXZY = 96;
  Int32x4.XXZZ = 160;
  Int32x4.XXZW = 224;
  Int32x4.XXWX = 48;
  Int32x4.XXWY = 112;
  Int32x4.XXWZ = 176;
  Int32x4.XXWW = 240;
  Int32x4.XYXX = 4;
  Int32x4.XYXY = 68;
  Int32x4.XYXZ = 132;
  Int32x4.XYXW = 196;
  Int32x4.XYYX = 20;
  Int32x4.XYYY = 84;
  Int32x4.XYYZ = 148;
  Int32x4.XYYW = 212;
  Int32x4.XYZX = 36;
  Int32x4.XYZY = 100;
  Int32x4.XYZZ = 164;
  Int32x4.XYZW = 228;
  Int32x4.XYWX = 52;
  Int32x4.XYWY = 116;
  Int32x4.XYWZ = 180;
  Int32x4.XYWW = 244;
  Int32x4.XZXX = 8;
  Int32x4.XZXY = 72;
  Int32x4.XZXZ = 136;
  Int32x4.XZXW = 200;
  Int32x4.XZYX = 24;
  Int32x4.XZYY = 88;
  Int32x4.XZYZ = 152;
  Int32x4.XZYW = 216;
  Int32x4.XZZX = 40;
  Int32x4.XZZY = 104;
  Int32x4.XZZZ = 168;
  Int32x4.XZZW = 232;
  Int32x4.XZWX = 56;
  Int32x4.XZWY = 120;
  Int32x4.XZWZ = 184;
  Int32x4.XZWW = 248;
  Int32x4.XWXX = 12;
  Int32x4.XWXY = 76;
  Int32x4.XWXZ = 140;
  Int32x4.XWXW = 204;
  Int32x4.XWYX = 28;
  Int32x4.XWYY = 92;
  Int32x4.XWYZ = 156;
  Int32x4.XWYW = 220;
  Int32x4.XWZX = 44;
  Int32x4.XWZY = 108;
  Int32x4.XWZZ = 172;
  Int32x4.XWZW = 236;
  Int32x4.XWWX = 60;
  Int32x4.XWWY = 124;
  Int32x4.XWWZ = 188;
  Int32x4.XWWW = 252;
  Int32x4.YXXX = 1;
  Int32x4.YXXY = 65;
  Int32x4.YXXZ = 129;
  Int32x4.YXXW = 193;
  Int32x4.YXYX = 17;
  Int32x4.YXYY = 81;
  Int32x4.YXYZ = 145;
  Int32x4.YXYW = 209;
  Int32x4.YXZX = 33;
  Int32x4.YXZY = 97;
  Int32x4.YXZZ = 161;
  Int32x4.YXZW = 225;
  Int32x4.YXWX = 49;
  Int32x4.YXWY = 113;
  Int32x4.YXWZ = 177;
  Int32x4.YXWW = 241;
  Int32x4.YYXX = 5;
  Int32x4.YYXY = 69;
  Int32x4.YYXZ = 133;
  Int32x4.YYXW = 197;
  Int32x4.YYYX = 21;
  Int32x4.YYYY = 85;
  Int32x4.YYYZ = 149;
  Int32x4.YYYW = 213;
  Int32x4.YYZX = 37;
  Int32x4.YYZY = 101;
  Int32x4.YYZZ = 165;
  Int32x4.YYZW = 229;
  Int32x4.YYWX = 53;
  Int32x4.YYWY = 117;
  Int32x4.YYWZ = 181;
  Int32x4.YYWW = 245;
  Int32x4.YZXX = 9;
  Int32x4.YZXY = 73;
  Int32x4.YZXZ = 137;
  Int32x4.YZXW = 201;
  Int32x4.YZYX = 25;
  Int32x4.YZYY = 89;
  Int32x4.YZYZ = 153;
  Int32x4.YZYW = 217;
  Int32x4.YZZX = 41;
  Int32x4.YZZY = 105;
  Int32x4.YZZZ = 169;
  Int32x4.YZZW = 233;
  Int32x4.YZWX = 57;
  Int32x4.YZWY = 121;
  Int32x4.YZWZ = 185;
  Int32x4.YZWW = 249;
  Int32x4.YWXX = 13;
  Int32x4.YWXY = 77;
  Int32x4.YWXZ = 141;
  Int32x4.YWXW = 205;
  Int32x4.YWYX = 29;
  Int32x4.YWYY = 93;
  Int32x4.YWYZ = 157;
  Int32x4.YWYW = 221;
  Int32x4.YWZX = 45;
  Int32x4.YWZY = 109;
  Int32x4.YWZZ = 173;
  Int32x4.YWZW = 237;
  Int32x4.YWWX = 61;
  Int32x4.YWWY = 125;
  Int32x4.YWWZ = 189;
  Int32x4.YWWW = 253;
  Int32x4.ZXXX = 2;
  Int32x4.ZXXY = 66;
  Int32x4.ZXXZ = 130;
  Int32x4.ZXXW = 194;
  Int32x4.ZXYX = 18;
  Int32x4.ZXYY = 82;
  Int32x4.ZXYZ = 146;
  Int32x4.ZXYW = 210;
  Int32x4.ZXZX = 34;
  Int32x4.ZXZY = 98;
  Int32x4.ZXZZ = 162;
  Int32x4.ZXZW = 226;
  Int32x4.ZXWX = 50;
  Int32x4.ZXWY = 114;
  Int32x4.ZXWZ = 178;
  Int32x4.ZXWW = 242;
  Int32x4.ZYXX = 6;
  Int32x4.ZYXY = 70;
  Int32x4.ZYXZ = 134;
  Int32x4.ZYXW = 198;
  Int32x4.ZYYX = 22;
  Int32x4.ZYYY = 86;
  Int32x4.ZYYZ = 150;
  Int32x4.ZYYW = 214;
  Int32x4.ZYZX = 38;
  Int32x4.ZYZY = 102;
  Int32x4.ZYZZ = 166;
  Int32x4.ZYZW = 230;
  Int32x4.ZYWX = 54;
  Int32x4.ZYWY = 118;
  Int32x4.ZYWZ = 182;
  Int32x4.ZYWW = 246;
  Int32x4.ZZXX = 10;
  Int32x4.ZZXY = 74;
  Int32x4.ZZXZ = 138;
  Int32x4.ZZXW = 202;
  Int32x4.ZZYX = 26;
  Int32x4.ZZYY = 90;
  Int32x4.ZZYZ = 154;
  Int32x4.ZZYW = 218;
  Int32x4.ZZZX = 42;
  Int32x4.ZZZY = 106;
  Int32x4.ZZZZ = 170;
  Int32x4.ZZZW = 234;
  Int32x4.ZZWX = 58;
  Int32x4.ZZWY = 122;
  Int32x4.ZZWZ = 186;
  Int32x4.ZZWW = 250;
  Int32x4.ZWXX = 14;
  Int32x4.ZWXY = 78;
  Int32x4.ZWXZ = 142;
  Int32x4.ZWXW = 206;
  Int32x4.ZWYX = 30;
  Int32x4.ZWYY = 94;
  Int32x4.ZWYZ = 158;
  Int32x4.ZWYW = 222;
  Int32x4.ZWZX = 46;
  Int32x4.ZWZY = 110;
  Int32x4.ZWZZ = 174;
  Int32x4.ZWZW = 238;
  Int32x4.ZWWX = 62;
  Int32x4.ZWWY = 126;
  Int32x4.ZWWZ = 190;
  Int32x4.ZWWW = 254;
  Int32x4.WXXX = 3;
  Int32x4.WXXY = 67;
  Int32x4.WXXZ = 131;
  Int32x4.WXXW = 195;
  Int32x4.WXYX = 19;
  Int32x4.WXYY = 83;
  Int32x4.WXYZ = 147;
  Int32x4.WXYW = 211;
  Int32x4.WXZX = 35;
  Int32x4.WXZY = 99;
  Int32x4.WXZZ = 163;
  Int32x4.WXZW = 227;
  Int32x4.WXWX = 51;
  Int32x4.WXWY = 115;
  Int32x4.WXWZ = 179;
  Int32x4.WXWW = 243;
  Int32x4.WYXX = 7;
  Int32x4.WYXY = 71;
  Int32x4.WYXZ = 135;
  Int32x4.WYXW = 199;
  Int32x4.WYYX = 23;
  Int32x4.WYYY = 87;
  Int32x4.WYYZ = 151;
  Int32x4.WYYW = 215;
  Int32x4.WYZX = 39;
  Int32x4.WYZY = 103;
  Int32x4.WYZZ = 167;
  Int32x4.WYZW = 231;
  Int32x4.WYWX = 55;
  Int32x4.WYWY = 119;
  Int32x4.WYWZ = 183;
  Int32x4.WYWW = 247;
  Int32x4.WZXX = 11;
  Int32x4.WZXY = 75;
  Int32x4.WZXZ = 139;
  Int32x4.WZXW = 203;
  Int32x4.WZYX = 27;
  Int32x4.WZYY = 91;
  Int32x4.WZYZ = 155;
  Int32x4.WZYW = 219;
  Int32x4.WZZX = 43;
  Int32x4.WZZY = 107;
  Int32x4.WZZZ = 171;
  Int32x4.WZZW = 235;
  Int32x4.WZWX = 59;
  Int32x4.WZWY = 123;
  Int32x4.WZWZ = 187;
  Int32x4.WZWW = 251;
  Int32x4.WWXX = 15;
  Int32x4.WWXY = 79;
  Int32x4.WWXZ = 143;
  Int32x4.WWXW = 207;
  Int32x4.WWYX = 31;
  Int32x4.WWYY = 95;
  Int32x4.WWYZ = 159;
  Int32x4.WWYW = 223;
  Int32x4.WWZX = 47;
  Int32x4.WWZY = 111;
  Int32x4.WWZZ = 175;
  Int32x4.WWZW = 239;
  Int32x4.WWWX = 63;
  Int32x4.WWWY = 127;
  Int32x4.WWWZ = 191;
  Int32x4.WWWW = 255;
  class Float64x2 extends core.Object {
    static new(x, y) {
      return new _native_typed_data.NativeFloat64x2(x, y);
    }
    static splat(v) {
      return new _native_typed_data.NativeFloat64x2.splat(v);
    }
    static zero() {
      return new _native_typed_data.NativeFloat64x2.zero();
    }
    static fromFloat32x4(v) {
      return new _native_typed_data.NativeFloat64x2.fromFloat32x4(v);
    }
  }
  dart.setSignature(Float64x2, {
    constructors: () => ({
      new: [Float64x2, [core.double, core.double]],
      splat: [Float64x2, [core.double]],
      zero: [Float64x2, []],
      fromFloat32x4: [Float64x2, [Float32x4]]
    })
  });
  // Exports:
  exports.ByteBuffer = ByteBuffer;
  exports.TypedData = TypedData;
  exports.Endianness = Endianness;
  exports.ByteData = ByteData;
  exports.Int8List = Int8List;
  exports.Uint8List = Uint8List;
  exports.Uint8ClampedList = Uint8ClampedList;
  exports.Int16List = Int16List;
  exports.Uint16List = Uint16List;
  exports.Int32List = Int32List;
  exports.Uint32List = Uint32List;
  exports.Int64List = Int64List;
  exports.Uint64List = Uint64List;
  exports.Float32List = Float32List;
  exports.Float64List = Float64List;
  exports.Float32x4List = Float32x4List;
  exports.Int32x4List = Int32x4List;
  exports.Float64x2List = Float64x2List;
  exports.Float32x4 = Float32x4;
  exports.Int32x4 = Int32x4;
  exports.Float64x2 = Float64x2;
});
dart_library.library('foo', null, /* Imports */[
  "dart/_runtime",
  'dart/core'
], /* Lazy imports */[
], function(exports, dart, core) {
  'use strict';
  let dartx = dart.dartx;
  function main() {
    core.print('hello');
  }
  dart.fn(main);
  // Exports:
  exports.main = main;
});
//# sourceMappingURL=foo.js.map
dart_library.import('dart/html');
dart_library.start('foo');
